/* $Id: fa_config.c,v 1.32 2001/08/11 14:35:44 jm Exp $
 * Foreign Agent - configuration reading routines
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <net/if.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <assert.h>
#include <arpa/inet.h>

#include "authorized.h"
#include "fileio.h"
#include "auth.h"
#include "ha.h"
#include "fa.h"
#include "util.h"
#ifdef WITH_DP
#include "proxy_debug.h"
#include "proxy_util.h"
#include "papi_interface.h"
#endif
#include "util/list.h"

struct load_fa_data {
	struct fa_config *cfg;
	int process_authorized_networks;
	int process_fa_spi_list;
	int process_interfaces;
	int iface_ups, iface_downs;
};

#ifdef WITH_DP
static int mmfa_config(struct load_fa_data *fa) ;
static int mmfaInit_interfaces(struct load_fa_data *fa, struct interface_entry ifent);
#endif

/* process fa authorized networks list */

static int process_load_fa_authorized_networks(struct load_fa_data *fa,
					       char *key, char *data)
{
        struct fa_config *cfg;
	struct in_addr network;
	struct in_addr netmask;

        if (strcmp(key, "AUTHORIZEDNETWORKS_END") == 0) {
		if (!fa->process_authorized_networks) {
			fprintf(stderr,
				"No AUTHORIZEDNETWORKS_BEGIN before _END\n");
			return -1;
		}
                fa->process_authorized_networks = FALSE;
                return 0;
        }

        cfg = fa->cfg;
	if (load_net_address(key, &network, &netmask) != TRUE) {
		fprintf(stderr, 
		       "process_load_fa_authorized_networks: invalid "
		       "authorized IP network address\n");
		return -1;
	}

	if (authorized_add(cfg->authorized_networks, network,
			   netmask) != TRUE) {
		fprintf(stderr,
			"process_load_fa_authorized_networks: "
			"unable to add address to authorized networks.\n");
		return -1;
	}
        return 0;
}


static int process_load_fa_spi_list(struct load_fa_data *fa, char *key,
				    char *data)
{
	struct fa_config *cfg;
	struct fa_spi_entry *spi;
	char *pos;
	int res;

	if (strcmp(key, "FA_SECURITY_END") == 0) {
		assert(fa->process_fa_spi_list == TRUE);
		fa->process_fa_spi_list = FALSE;
		return 0;
	}

	cfg = fa->cfg;
	spi = malloc(sizeof(struct fa_spi_entry));
	if (spi == NULL) {
		fprintf(stderr,
			"process_load_fa_spi_list: not enough memory for "
			"struct spi_entry\n");
		return -1;
	}

	list_init_node(&spi->node);

	if (key[0] == '0' && key[1] == 'x')
		res = sscanf(key, "%x", &spi->spi);
	else
		res = sscanf(key, "%d", &spi->spi);

	if (res != 1) {
		fprintf(stderr,
			"process_load_fa_spi_list: invalid SPI number\n");
		free(spi);
		return -1;
	}

	pos = data;

	while (*pos == ' ' || *pos == '\t') pos++;
	if (load_ip_address(pos, &spi->addr) != TRUE) {
		fprintf(stderr,
			"process_load_fa_authorized_list: invalid "
			"IP address\n");
		free(spi);
		return -1;
	}

	while (*pos != ' ' && *pos != '\t' && *pos != '\0') pos++;
	while (*pos == ' ' || *pos == '\t') pos++;
	if (*pos == '\0' || sscanf(pos, "%d", &spi->agent_type) != 1) {
		fprintf(stderr,
			"process_load_fa_spi_list: invalid agent type %i\n",
			spi->agent_type);
		free(spi);
		return -1;
	}

	while (*pos != ' ' && *pos != '\t' && *pos != '\0') pos++;
	while (*pos == ' ' || *pos == '\t') pos++;
	if (*pos == '\0' || sscanf(pos, "%d", &spi->alg) != 1) {
		fprintf(stderr,
			"process_load_fa_spi_list: invalid algorithm "
			"number %i\n", spi->alg);
		free(spi);
		return -1;
	}

	if (!auth_supported_auth_alg(spi->alg)) {
		fprintf(stderr, "process_load_fa_spi_list: unsupported "
			"algorithm %i\n", spi->alg);
		free(spi);
		return -1;
	}

	while (*pos != ' ' && *pos != '\t' && *pos != '\0') pos++;
	while (*pos == ' ' || *pos == '\t') pos++;
	if (*pos == '\0' || load_hex_table(pos, spi->shared_secret,
					   MAXSHAREDSECRETLEN,
					   &spi->shared_secret_len) == FALSE) {
		fprintf(stderr,
			"process_load_fa_spi_list: invalid shared secret\n");
		free(spi);
		return -1;
	}

	assert(spi->shared_secret_len >= 0);
	assert(spi->shared_secret_len <= MAXSHAREDSECRETLEN);
	list_add_tail(&cfg->fa_spi_list, &spi->node);
	
	return 0;
}


static int process_interfaces(struct load_fa_data *fa, char *key,
			      char *data)
{
	struct fa_config *cfg;
	struct interface_entry *iface;
	char *pos;

	if (strcmp(key, "INTERFACES_END") == 0) {
		assert(fa->process_interfaces == TRUE);
		fa->process_interfaces = FALSE;
		return 0;
	}

	cfg = fa->cfg;
	iface = malloc(sizeof(struct interface_entry));
	if (iface == NULL) {
		fprintf(stderr,
			"process_interfaces: not enough memory for "
			"struct interface_entry\n");
		return -1;
	}
	memset(iface, 0, sizeof(iface));
	list_init_node(&iface->node);

	dynamics_strlcpy(iface->dev, key, IFNAMSIZ);

	pos = data;
	while (*pos == ' ' || *pos == '\t') pos++;
	if (*pos == '\0' || sscanf(pos, "%d", &iface->type) != 1 ||
	    (iface->type < 1 || iface->type > 3)) {
		fprintf(stderr,
			"process_interfaces: invalid type (interface=%s)\n",
			iface->dev);
		free(iface);
		return -1;
	}
	if (iface->type == INTERFACE_TYPE_BOTH ||
	    iface->type == INTERFACE_TYPE_UP)
		fa->iface_ups++;
	if (iface->type == INTERFACE_TYPE_BOTH ||
	    iface->type == INTERFACE_TYPE_DOWN)
		fa->iface_downs++;

	while (*pos != ' ' && *pos != '\t' && *pos != '\0') pos++;
	while (*pos == ' ' || *pos == '\t') pos++;
	if (*pos == '\0' || sscanf(pos, "%d", &iface->agentadv) != 1) {
		fprintf(stderr, "process_interfaces: invalid agentadv "
			"(interface=%s)\n", iface->dev);
		free(iface);
		return -1;
	}

	while (*pos != ' ' && *pos != '\t' && *pos != '\0') pos++;
	while (*pos == ' ' || *pos == '\t') pos++;
	if (*pos == '\0' || sscanf(pos, "%d", &iface->interval) != 1) {
		fprintf(stderr, "process_interfaces: invalid interval "
			"(interface=%s)\n", iface->dev);
		free(iface);
		return -1;
	}

	if (iface->interval < 1) {
		fprintf(stderr, "process_interfaces: advertisement interval "
			"too short (must be at least one second)\n");
		free(iface);
		return -1;
	}

	while (*pos != ' ' && *pos != '\t' && *pos != '\0') pos++;
	while (*pos == ' ' || *pos == '\t') pos++;
	iface->force_addr.s_addr = 0;
	if (*pos != '\0' && load_ip_address(pos, &iface->force_addr) != TRUE) {
		fprintf(stderr,	"process_interfaces: invalid IP address "
			"(interface=%s)\n", iface->dev);
		free(iface);
		return -1;
	}

	iface->icmp_sock = -1;
	iface->udp_sock = -1;
	list_add_tail(&cfg->interfaces, &iface->node);

	return 0;
}

/* Process loading of the fa_data
 * Return values: -2: consistency error, -1: error, 0: ok, 1: end */

static int process_load_fa(void *voidptr, char *key, char *data)
{
	struct load_fa_data *fa;
	struct fa_config *cfg;

	fa = voidptr;
	cfg = fa->cfg;

	if (fa->process_authorized_networks == TRUE) {
		return process_load_fa_authorized_networks(fa, key, data);
	}
	if (fa->process_fa_spi_list == TRUE) {
		return process_load_fa_spi_list(fa, key, data);
	}
	if (fa->process_interfaces == TRUE) {
		return process_interfaces(fa, key, data);
	}
	if (strcmp(key, "NetworkAccessIdentifier") == 0) {
		if (load_nai(data, cfg->fa_nai, &cfg->fa_nai_len) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "HighestFAIPAddress") == 0) {
		if (load_ip_address(data, &cfg->highest_fa_addr) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "HighestFAPubKeyHash") == 0) {
		if (load_hex_table(data, cfg->hfa_pubkey_hash, MAX_HASH_LEN,
				   &cfg->hfa_pubkey_hash_len) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "UpperFAIPAddress") == 0) {
		if (load_ip_address(data, &cfg->upper_fa_addr) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "HighestFA") == 0) {
		if (load_bool(data, &cfg->highest_FA) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "UpperFAUDPPort") == 0) {
		if (load_int(data, &cfg->upper_fa_port) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "RoutingTableStart") == 0) {
		if (load_int(data, &cfg->routing_table_start) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "RoutingTableEnd") == 0) {
		if (load_int(data, &cfg->routing_table_end) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "TunnelDevice") == 0) {
		if (load_char_table(data, cfg->tunnel_device,
				    IFNAMSIZ) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "UDPPort") == 0) {
		if (load_int(data, &cfg->udp_port) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "HAUDPPort") == 0) {
		if (load_int(data, &cfg->ha_udp_port) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "FAKeyFile") == 0) {
		if (load_char_table(data, cfg->key_file,
				    MAXFILENAMELEN) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "EnableChallengeResponse") == 0) {
		if (load_bool(data, &cfg->enable_challenge_response) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "ChallengeWindow") == 0) {
		if (load_int(data, &cfg->challenge_window) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "ChallengeLength") == 0) {
		if (load_int(data, &cfg->challenge_length) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "RequireChallenge") == 0) {
		if (load_bool(data, &cfg->require_challenge) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "ChallengeInRegReply") == 0) {
		if (load_bool(data, &cfg->challenge_in_reg_reply) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "RequireMNFASecAssoc") == 0) {
		if (load_bool(data, &cfg->require_mnfa_sec_assoc) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "MaxBindings") == 0) {
		if (load_int(data, &cfg->max_bindings) == TRUE) return 0;
		return -1;
	}
	if (strcmp(key, "MaxPending") == 0) {
		if (load_int(data, &cfg->max_pending) == TRUE) return 0;
		return -1;
	}
	if (strcmp(key, "DeletePendingAfter") == 0) {
		if (load_int(data, &cfg->delete_pending_after) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "EnableFADecapsulation") == 0) {
		if (load_bool(data, &cfg->enable_fa_decapsulation) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "EnableTriangleTunneling") == 0) {
		if (load_bool(data, &cfg->enable_triangle_tunneling) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "ForceReverseTunneling") == 0) {
		if (load_bool(data, &cfg->force_reverse_tunneling) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "RegistrationRequired") == 0) {
		if (load_bool(data, &cfg->reg_required) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "EnableReverseTunneling") == 0) {
		if (load_bool(data, &cfg->enable_reverse_tunneling) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "AllowMobileNodes") == 0) {
		if (load_bool(data, &cfg->allow_mobile_nodes) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "FADefaultTunnelLifetime") == 0) {
		if (load_int(data, &cfg->fa_default_tunnel_lifetime) == TRUE) {
			return 0;
		}
		return -1;
	}
	if (strcmp(key, "PacketSocketMode") == 0) {
		if (load_int(data, &cfg->packet_socket_mode) == TRUE)
			return 0;
		return -1;
	}
	if (strcmp(key, "SyslogFacility") == 0) {
		if (load_syslog_facility(data,
					 &cfg->syslog_facility) == TRUE) {
			return 0;
		}
		return -1;
	}
        if (strcmp(key, "FAAPIReadSocketPath") == 0) {
                if (load_char_table(data, cfg->fa_api_read_socket_path,
                                    MAXFILENAMELEN) == TRUE) {
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "FAAPIReadSocketGroup") == 0) {
                if (load_char_table(data, cfg->fa_api_read_socket_group,
                                    MAXGROUPNAMELEN) == TRUE) {
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "FAAPIReadSocketOwner") == 0) {
                if (load_char_table(data, cfg->fa_api_read_socket_owner,
                                    MAXOWNERNAMELEN) == TRUE) {
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "FAAPIReadSocketPermissions") == 0) {
                if (load_int(data,
                             &cfg->fa_api_read_socket_permissions) == TRUE) {
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "FAAPIAdminSocketPath") == 0) {
                if (load_char_table(data, cfg->fa_api_admin_socket_path,
                                    MAXFILENAMELEN) == TRUE) {
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "FAAPIAdminSocketGroup") == 0) {
                if (load_char_table(data, cfg->fa_api_admin_socket_group,
                                    MAXGROUPNAMELEN) == TRUE) {
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "FAAPIAdminSocketOwner") == 0) {
                if (load_char_table(data, cfg->fa_api_admin_socket_owner,
                                    MAXOWNERNAMELEN) == TRUE) {
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "FAAPIAdminSocketPermissions") == 0) {
                if (load_int(data,
                             &cfg->fa_api_admin_socket_permissions) == TRUE) {
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "SocketPriority") == 0) {
                if (load_int(data, &cfg->socket_priority) == TRUE) {
                        return 0;
                }
                return -1;
        }
	if (strcmp(key, "SendSFADebugExt") == 0) {
		if (load_bool(data, &cfg->send_sfa_debug_ext) == TRUE)
			return 0;
		return -1;
	}
        if (strcmp(key, "RegistrationTTLCheck") == 0) {
                if (load_int(data, &cfg->reg_ttl_check) == TRUE) {
			if (cfg->reg_ttl_check != TTL_CHECK_NONE &&
			    cfg->reg_ttl_check != TTL_CHECK_REVERSE &&
			    cfg->reg_ttl_check != TTL_CHECK_ALL) {
				fprintf(stderr,
					"Invalid RegistrationTTLCheck: %i\n",
					cfg->reg_ttl_check);
				return -1;
			}
                        return 0;
                }
                return -1;
        }
        if (strcmp(key, "FARegistrationLifetime") == 0) {
                if (load_int(data, &cfg->fa_reg_lifetime) == TRUE)
                        return 0;
                return -1;
        }
#ifdef INCLUDE_IPAY
	if (strcmp(key, "IpayAAAFAddress") == 0) {
		if (load_ip_address(data, &cfg->aaaf_addr) == TRUE) {
			return 0;
		}
		return -1;
	}
        if (strcmp(key, "IpayAAAFPort") == 0) {
                if (load_int(data, &cfg->aaaf_port) == TRUE)
                        return 0;
                return -1;
        }
        if (strcmp(key, "IpayPort") == 0) {
                if (load_int(data, &cfg->ipay_port) == TRUE)
                        return 0;
                return -1;
        }
        if (strcmp(key, "IpayTimePrice") == 0) {
                if (load_int(data, &cfg->timePrice) == TRUE)
                        return 0;
                return -1;
        }
        if (strcmp(key, "IpayBytePrice") == 0) {
                if (load_int(data, &cfg->bytePrice) == TRUE)
                        return 0;
                return -1;
        }
#endif
	if (strcmp(key, "AUTHORIZEDNETWORKS_BEGIN") == 0) {
		if (fa->process_authorized_networks ||
		    fa->process_fa_spi_list || fa->process_interfaces) {
			fprintf(stderr, "List processing error while handling "
				"AUTHORIZEDNETWORKS_BEGIN\n");
			return -1;
		}
		fa->process_authorized_networks = TRUE;
		return 0;
	}
	if (strcmp(key, "FA_SECURITY_BEGIN") == 0) {
		if (fa->process_authorized_networks ||
		    fa->process_fa_spi_list || fa->process_interfaces) {
			fprintf(stderr, "List processing error while handling "
				"FA_SECURITY_BEGIN\n");
			return -1;
		}
		fa->process_fa_spi_list = TRUE;
		return 0;
	}
	if (strcmp(key, "INTERFACES_BEGIN") == 0) {
		if (fa->process_authorized_networks ||
		    fa->process_fa_spi_list || fa->process_interfaces) {
			fprintf(stderr, "List processing error while handling "
				"INTERFACES_BEGIN\n");
			return -1;
		}
		fa->process_interfaces = TRUE;
		return 0;
	}
        if (strcmp(key, "END") == 0) return 1;
	return -1;
}


/* load FA configuration */
int load_fa(struct fa_config *cfg, char *program_name, char *config_file)
{
        FILE *file;
        struct load_fa_data fa;
	int ret;
	u_int32_t switchip = cfg->highest_fa_addr.s_addr;

        fa.cfg = cfg;

	fa.process_authorized_networks = FALSE;
	fa.process_fa_spi_list = FALSE;
	fa.process_interfaces = FALSE;
	fa.iface_ups = 0;
	fa.iface_downs = 0;
        memset(cfg, 0, sizeof(struct fa_config));
	cfg->highest_fa_addr.s_addr = switchip;
	cfg->challenge_window = 2;
	cfg->challenge_length = 4;
        cfg->max_bindings = FA_DEFAULT_MAX_BINDINGS;
        cfg->fa_default_tunnel_lifetime = FA_DEFAULT_TUNNEL_LIFETIME;
        cfg->syslog_facility = FA_DEFAULT_SYSLOG_FACILITY;
	cfg->routing_table_start = FA_DEFAULT_ROUTING_TABLE_START;
	cfg->routing_table_end = FA_DEFAULT_ROUTING_TABLE_END;
	cfg->udp_port = cfg->upper_fa_port = FA_DEFAULT_UDP_PORT;
	cfg->ha_udp_port = cfg->upper_fa_port = FA_DEFAULT_UDP_PORT;
	cfg->highest_FA = TRUE;
	cfg->allow_mobile_nodes = TRUE;
	cfg->reg_required = TRUE;
	cfg->socket_priority = -1;
	cfg->reg_ttl_check = TTL_CHECK_REVERSE;
	cfg->fa_reg_lifetime = 3600;
#ifdef APP_SAVE_CONFIG
	//        cfg->indCurCmd = 0;
	cfg->cmdbuf = (char *)mmGet_memory(__FUNCTION__,MAX_CMDBUF_SIZE);
	if(!cfg->cmdbuf) {
	  LOG2(LOG_CRIT, "Error allocating cmd buffer");
	  return FALSE;
	}
	list_init(&cfg->facmdbuf);
#else
	cfg->cmdlist = NULL;
#endif
	cfg->authorized_networks = authorized_init();
	if (cfg->authorized_networks == NULL) {
		fprintf(stderr,
			"load_fa: Error while initializing list of authorized "
			"networks\n");
		return FALSE;
	}
	list_init(&cfg->fa_spi_list);
	list_init(&cfg->interfaces);
	file = fopen(config_file, "r");
        if (file == NULL) {
		fprintf(stderr,
			"%s: Could not open configuration file '%s'.\n",
			program_name, config_file);
		return FALSE;
	}

	ret = TRUE;

        if (load_data(&fa, file, process_load_fa) == FALSE) {
                fprintf(stderr,
                        "%s: Error while interpreting file '%s'!\n",
                        program_name, config_file);
                ret = FALSE;
        }
        fclose(file);
#ifdef WITH_DP
	mmfa_config(&fa);
#endif

	if (cfg->tunnel_device[0] == '\0') {
		fprintf(stderr, "load_fa: TunnelDevice not set\n");
		return FALSE;
	}

	if (fa.iface_ups != 1) {
		fprintf(stderr, "load_fa: %i interfaces to upper direction "
			"defined - exactly one needed (INTERFACES list)\n",
			fa.iface_ups);
		return FALSE;
	}

	if (fa.iface_downs < 1) {
		fprintf(stderr, "load_fa: no interfaces to lower direction "
			"defined - at least one needed (INTERFACES list)\n");
		return FALSE;
	}

	if (cfg->challenge_length > MAX_CHALLENGE_LEN) {
		fprintf(stderr, "load_fa: Invalid ChallengeLength %i "
			"(range: 0-%i)\n", cfg->challenge_length,
			MAX_CHALLENGE_LEN);
		return FALSE;
	}

	if (cfg->enable_challenge_response && cfg->challenge_window < 1) {
		fprintf(stderr, "load_fa: Invalid ChallengeWindow %i (must be "
			"at least 1)\n", cfg->challenge_window);
	}

	if (ret == FALSE)
		authorized_destroy(cfg->authorized_networks);



        return ret;
}


void cleanup_fa_config(struct fa_config *config)
{
	struct fa_spi_entry *fa_spi;
	struct interface_entry *iface;

	if (config == NULL) return;

	authorized_destroy(config->authorized_networks);

	fa_spi = (struct fa_spi_entry *)
		list_remove_first(&config->fa_spi_list);
	while (fa_spi != NULL) {
		free(fa_spi);
		fa_spi = (struct fa_spi_entry *)
			list_remove_first(&config->fa_spi_list);
	}

	iface = (struct interface_entry *)
		list_remove_first(&config->interfaces);
	while (iface != NULL) {
		close_socket(iface->icmp_sock);
		close_socket(iface->udp_sock);
		free(iface);
		iface = (struct interface_entry *)
			list_remove_first(&config->interfaces);
	}
}


#ifdef WITH_DP
static int
mmfa_config(struct load_fa_data *fa) 
{
  struct interface_entry ife;
  struct fa_config *cfg = fa->cfg;

  /* initialize the eth0 and lo interfaces 
   * For FA lo addr is 127.0.0.4
   */

  bzero((char *)&ife, sizeof(struct interface_entry));
  strncpy(ife.dev, "lo", IFNAMSIZ);
  ife.type = INTERFACE_TYPE_DOWN;
  ife.agentadv = -1;
  ife.interval = 1;
  ife.force_addr.s_addr = inet_addr(FA_LOOPBACK_ADDR);
  if(mmfaInit_interfaces(fa, ife)) {
    LOG2(LOG_CRIT, "%s: Error initializing interfaces...Exiting", __FUNCTION__);
    return -1;
  }
 
  //if(proxy_get_mswitchinfo(&cfg->highest_fa_addr, NULL)) {
  // LOG2(LOG_CRIT, "%s: Error getting switch ip...Exiting", __FUNCTION__);
  // return -1;
  // }

  DEBUG(DEBUG_INFO, "Highest FA Addr = %x", cfg->highest_fa_addr.s_addr);
  bzero((char *)&ife, sizeof(struct interface_entry));
  // strncpy(ife.dev, MM_VLAN1_INTERFACE, IFNAMSIZ);
  ife.type = INTERFACE_TYPE_UP;
  ife.agentadv = -1;
  ife.interval = 1;
  ife.force_addr = cfg->highest_fa_addr;
  if(mmfaInit_interfaces(fa, ife)) {
    LOG2(LOG_CRIT, "%s: Error initializing interfaces...Exiting", __FUNCTION__);
    return -1;
  }
 
  cfg->upper_fa_addr = cfg->highest_fa_addr;
  
  return 0;
  
}

static int mmfaInit_interfaces(struct load_fa_data *fa, struct interface_entry ifent)
{
  struct interface_entry *iface;
  struct fa_config *cfg;

  cfg = fa->cfg;
  iface = malloc(sizeof(struct interface_entry));
  if (iface == NULL) {
    fprintf(stderr,
	    "%s: not enough memory for "
	    "struct interface_entry\n", __FUNCTION__);
    return -1;
  }
  memset(iface, 0, sizeof(iface));
  list_init_node(&iface->node);
  
  memcpy((char *)iface, (char *)&ifent, sizeof(struct interface_entry));
  
  if (iface->type == INTERFACE_TYPE_BOTH ||
      iface->type == INTERFACE_TYPE_UP)
    fa->iface_ups++;
  if (iface->type == INTERFACE_TYPE_BOTH ||
      iface->type == INTERFACE_TYPE_DOWN)
    fa->iface_downs++;
  
  iface->icmp_sock = -1;
  iface->udp_sock = -1;
  list_add_tail(&cfg->interfaces, &iface->node);
  
  return 0;
}


int mmfa_spilist_add(int spival, u_int32_t ip, unsigned char *secret)
{
  struct fa_spi_entry *spi;
  extern struct fa_config *config; 


  if(!list_is_empty(&config->fa_spi_list)) {
    struct fa_spi_entry *s = (struct fa_spi_entry *)list_get_first(&config->fa_spi_list);
    while(s) {
      if(s->spi == spival && s->addr.s_addr == ip) {
	DEBUG(DEBUG_CONFIG, "SA between this HA and FA already configured. Updating secret");
	s->shared_secret_len = strlen(secret);
	strcpy(s->shared_secret, secret);
	return WARNING;
      }
      s = (struct fa_spi_entry *)list_get_next(&s->node);
    }
  }

  spi = malloc(sizeof(struct fa_spi_entry));
  if (spi == NULL) {
    fprintf(stderr,
	    "process_load_fa_spi_list: not enough memory for "
	    "struct spi_entry\n");
    return -1;
  }
  
  list_init_node(&spi->node);
  
  spi->spi = spival;
  spi->addr.s_addr = ip;
  spi->agent_type = SPI_AGENT_HA;
  spi->alg = 4;
  spi->shared_secret_len = strlen(secret);
  strcpy(spi->shared_secret, secret);

  assert(spi->shared_secret_len >= 0);
  assert(spi->shared_secret_len <= MAXSHAREDSECRETLEN);
  
  DEBUG(DEBUG_CONFIG, "Adding FA Security Association");
  DEBUG(DEBUG_CONFIG, "SPI: %d HA: %s Algo: %d Secret: %s",
	spi->spi, inet_ntoa(spi->addr), spi->alg, spi->shared_secret);
  list_add_tail(&config->fa_spi_list, &spi->node);

  DEBUG(DEBUG_FLAG, "Added securiry association to list");
  return 0;
}


int mmfa_spilist_del(int spival)
{
  struct fa_spi_entry *spi = NULL;
  extern struct fa_config *config; 
    
  for(spi = (struct fa_spi_entry *)list_get_first(&config->fa_spi_list); 
      spi && spival != spi->spi; spi = (struct fa_spi_entry *)list_get_next(&spi->node));
  
  if(spi) {
    DEBUG(DEBUG_CONFIG, "Removing FA-HA Security Association with spi %d", spi->spi);
    list_remove(&spi->node);
    free(spi);
    return 0;
  }
  
  return -1;
}

#endif
