/* $Id: fa_reply.c,v 1.56 2001/09/16 17:21:59 jm Exp $
 * Foreign Agent - registration reply handling
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <net/if.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <string.h>
#include <assert.h>

#include "ha.h"
#include "fa.h"
#include "debug.h"
#include "agent_utils.h"
#include "md5_mac.h"
#include "util.h"

#include "dispatcher/dispatcher.h"
#include "dispatcher/timer.h"
#include "papi/MHApi.h"


extern struct interface_entry *up_interface;
extern struct fa_config *config;
extern struct hashtable *tunnels_hash;
extern struct bindingtable *bindings_hash;
extern struct binding_counters bcounters;

static int forward_reply(struct bindingentry *binding,
			 struct msg_extensions *ext, int error_reply,
			 struct unconfirmed_request *unc);

/* returns:
 *   0 = auth. ext. ok
 *   1 = ff_auth missing/invalid
 *   2 = fh_auth missing/invalid
 */
static int
check_rep_auth_ext(struct packet_from_info *info, struct msg_extensions *ext)
{
	struct fa_spi_entry *fa_spi;

	fa_spi = get_fa_spi(0, info->src.sin_addr, SPI_AGENT_FA);
	if (fa_spi != NULL) {
		if (ext->ff_auth == NULL ||
		    !auth_check_vendor(AUTH_ALG_MD5, fa_spi->shared_secret,
				       fa_spi->shared_secret_len,
				       (unsigned char *) ext->rep,
				       ext->ff_auth))
			return 1;
		else
			return 0;
	}

	fa_spi = get_fa_spi(0, info->src.sin_addr, SPI_AGENT_HA);
	if (fa_spi != NULL) {
		if (ext->fh_auth == NULL ||
		    !auth_check(fa_spi->alg, fa_spi->shared_secret,
				fa_spi->shared_secret_len,
				(unsigned char *) ext->rep,
				ext->fh_auth))
			return 2;
		else
			return 0;
	}

	return 0;
}


/* Returns:
 * 0: valid reply
 * 1: valid reply, but no Dynamics extensions
 * -1: invalid reply
 */
static int
validate_reply(struct msg_extensions* ext, struct bindingentry *binding,
	       struct packet_from_info *info, struct unconfirmed_request *unc)
{
	int error_code = 0, i;
	char *error_text = NULL;
	struct sockaddr_in addr;
	struct tunnel_data *t_data;

	t_data = (struct tunnel_data *) binding->data;

	if ((i = check_rep_auth_ext(info, ext)) != 0) {
		if (i == 2) {
			error_text = "invalid hf_auth";
			error_code = REGREP_HA_FAILED_AUTH_FA;
		} else {
			/* actually this is FA-FA auth. failure, but there is
			 * not currently any error code for it */
			error_text = "invalid ff_auth";
			error_code = REGREP_REASON_UNSPEC_FA;
		}
	} else if (ext->double_auth_ext > 0) {
		error_text = "duplicate authentication extension";
		if (ext->double_auth_ext & DOUBLE_FH_AUTH)
			error_code = REGREP_HA_FAILED_AUTH_FA;
		else
			error_code = REGREP_REASON_UNSPEC_FA;
#if 0
	/* This check causes some interoperability problems at least with
	 * PSU Mobile IP implementation. Since the src address is not really
	 * protected, this check does not affect security and it is thus
	 * removed at least for now. */
	} else if (ext->rep->code != REGREP_UNKNOWN_HA_HA &&
		   config->highest_FA && info->src.sin_addr.s_addr !=
		   (unc ? unc->ha_addr.s_addr : binding->ha_addr.s_addr)) {
		error_text = "reply's source address does not match with the "
			"HA address";
		error_code = REGREP_ADMIN_PROHIBITED_FA;
#endif
	} else if (!config->highest_FA && info->src.sin_addr.s_addr !=
		   config->upper_fa_addr.s_addr) {
		error_text = "reply's source address does not match with "
			"the upper FA address";
		error_code = REGREP_ADMIN_PROHIBITED_FA;
	} else if (ext->rep->code != REGREP_UNKNOWN_HA_HA &&
		   ext->rep->ha_addr.s_addr !=
		   (unc ? unc->ha_addr.s_addr : binding->ha_addr.s_addr)) {
		error_text = "HA address changed from request";
		error_code = REGREP_ADMIN_PROHIBITED_FA;
	} else if (ext->rep->ha_addr.s_addr == 0) {
		error_text = "missing home agent";
		error_code = REGREP_MISSING_HOME_AGENT_FA;
	} else if (ext->rep->home_addr.s_addr == 0) {
		error_text = "missing home address";
		error_code = REGREP_MISSING_HOMEADDR_FA;
	} else if (unc && unc->mn_nai_included && ext->mn_nai == NULL) {
		error_text = "missing MN NAI";
		error_code = REGREP_MISSING_NAI_FA;
	} else if (unc && unc->challenge &&
		   (ext->challenge == NULL ||
		    !equal_challenge(unc->challenge, ext->challenge))) {
		error_text = "missing challenge";
		error_code = REGREP_MISSING_CHALLENGE_FA;
	} else if (ext->unknown_cvse) {
		error_text = "unknown CVSE";
		error_code = REGREP_UNSUPP_VENDOR_ID_HA_FA;
	} else if (config->require_mnfa_sec_assoc &&
		   ext->mn_fa_key_material_aaa == NULL &&
		   get_fa_spi(0, ext->rep->home_addr, SPI_AGENT_MN) == NULL) {
		error_text = "missing MN-FA sec. assoc.";
		error_code = REGREP_MISSING_MN_FA_FA;
	}

	if (error_code > 0) {
		char mn[30], ha[30];
		struct interface_entry *iface;

		dynamics_strlcpy(mn, inet_ntoa(ext->rep->home_addr), 30);
		dynamics_strlcpy(ha, inet_ntoa(ext->rep->ha_addr), 30);

		if (error_text != NULL) {
			LOG2(LOG_WARNING, "SRC=%s, MN=%s, HA=%s - "
			     "reply denied: %s\n",
			     inet_ntoa(info->src.sin_addr), mn,
			     ha, error_text);
			report_discarded_msg(ext->start, ext->len, &info->src,
					     error_text);
		}

		addr.sin_family = AF_INET;
		addr.sin_addr = unc ? unc->info.src.sin_addr :
			binding->lower_addr;
		addr.sin_port = unc ? unc->info.src.sin_port :
			binding->lower_port;
		iface = unc ? unc->info.iface : t_data->info.iface;

		if (iface == NULL || addr.sin_addr.s_addr == 0 ||
		    addr.sin_port == 0) {
			DEBUG(DEBUG_FLAG,
			      "validate_reply: unknown lower end (%s:%u:%p)\n",
			      inet_ntoa(addr.sin_addr), ntohs(addr.sin_port),
			      iface);
			return -1;
		}

		send_failure_reply(error_code, ext,
				   unc ? &unc->info : &t_data->info, NULL, 0);
		return -1;
	}

	/* check the existance of the Dynamics extensions */
	if (!ext->sk_auth && IS_REGREP_ACCEPTED(ext->rep->code)) {
		DEBUG(DEBUG_FLAG,
		      "validate_reply: missing sk_auth\n");
		return 1;
	}

	if (ext->rep->lifetime != 0 && !ext->fa_pubkeyrep &&
		!ext->fa_keyrep) {
		DEBUG(DEBUG_FLAG,
		      "validate_reply: missing fa_keyrep/fa_pubkeyrep\n");
		return 1;
	}

	return 0;
}


#ifdef USE_TEARDOWN
/* Forward tear down message in the special case when the binding id field
 * has changed in the lower part of the tunnel; i.e. the id must be changed
 * by this FA and new sk_auth MAC must be calculated.
 * This function removes a possible mh_auth as it cannot be recalculated, so
 * this should be called only for tear down messages. */
static int
handle_teardown_changed_id(struct bindingentry *binding,
			   struct msg_extensions *ext,
			   struct unconfirmed_request *unc)
{
	unsigned char msg[MAXMSG];
	int len, ret;
	struct reg_rep *reply;
	struct registration_ext_dynamics *dyn_ext;
	struct tunnel_data *t_data;
	struct in_addr addr;
	int port;

	DEBUG(DEBUG_FLAG, "handle_teardown_changed_id\n");
	t_data = (struct tunnel_data *) binding->data;

/*	if (unc != NULL) {
		addr = unc->lower_addr;
		port = unc->lower_port;
	} else { */
		addr = binding->lower_addr;
		port = binding->lower_port;
//	}

	memcpy(msg, ext->rep, sizeof(struct reg_rep));
	reply = (struct reg_rep *) msg;
	reply->id[0] = htonl(binding->id[0]);
	reply->id[1] = htonl(binding->id[1]);
	len = sizeof(struct reg_rep);

	if (ext->mh_auth != NULL)
		DEBUG(DEBUG_FLAG, "\tdropping mh_auth\n");

	/* add Dynamics options extension */
	dyn_ext = (struct registration_ext_dynamics *) &msg[len];
	dyn_ext->type = VENDOR_EXT_TYPE2;
	dyn_ext->length = sizeof(struct registration_ext_dynamics) - 2;
	dyn_ext->reserved = 0;
	dyn_ext->vendor_id = htonl(VENDOR_ID_DYNAMICS);
	dyn_ext->sub_type = htons(VENDOR_EXT_DYNAMICS_OPTIONS);
	dyn_ext->version = VENDOR_EXT_VERSION;
	dyn_ext->opts = REG_EXT_OWN_TEAR_DOWN;
	dyn_ext->seq = 0;
	len += sizeof(struct registration_ext_dynamics);

	len += auth_add_vendor(AUTH_ALG_MD5, binding->key, binding->keylen,
			       msg, (struct vendor_msg_auth *) (msg + len),
			       VENDOR_EXT_DYNAMICS_SK_AUTH,
			       htonl(binding->spi));

	len += add_fa_auth_ext(addr, reply->home_addr,
			       (unsigned char *) reply, msg + len);

	LOG2(LOG_WARNING, "handle_teardown_changed_id - "
	     "forwarding to %s:%i\n", inet_ntoa(addr), ntohs(port));
	assert(t_data->info.iface != NULL);
	ret = own_sendto(t_data->info.iface->udp_sock, addr, port, msg, len);
	dynamics_check_sendto(ret, len, "handle_teardown_changed_id");

	return 0;
}
#endif


enum { ID_NOT_FOUND, ID_CURRENT,
#ifdef USE_TEARDOWN
       ID_CHANGED,
#endif
       ID_NEW };

static struct unconfirmed_request*
find_matching_request(struct bindingentry *binding, __u32 id, int *type)
{
	struct unconfirmed_request *unc;
	struct tunnel_data *t_data;

	t_data = (struct tunnel_data *) binding->data;

	if (t_data->confirmed) {
		if (id == binding->id[1]) {
			DEBUG(DEBUG_FLAG, "Reply to current confirmed data\n");
			*type = ID_CURRENT;
			return NULL;
		}
#ifdef USE_TEARDOWN
		if (id == t_data->upper_id[1]) {
			DEBUG(DEBUG_FLAG, "Reply to confirmed data - ID "
			      "changed\n");
			*type = ID_CHANGED;
			return NULL; /* use current data */
		}
#endif
	}

	/* check the unconfirmed request list for id match */
	unc = t_data->unc_req;
	while (unc != NULL) {
		if (unc->id[1] == id) {
			DEBUG(DEBUG_FLAG, "Reply to unconfirmed data\n");
			*type = ID_NEW;
			return unc;
		}
		unc = unc->next;
	}

	*type = ID_NOT_FOUND;
	return NULL;
}


/* Takes care of the disconnect reply from HA or Switching FA */
static int
handle_reply_disconnect(struct bindingentry *binding,
			struct unconfirmed_request *unc,
			struct msg_extensions *ext, int idtype)
{
	struct tunnel_data *t_data;
	int is_lowest_fa;

	t_data = binding->data;

	if (idtype == ID_NEW && unc != NULL) {
		/* lower end changed */
		is_lowest_fa = unc->is_lowest_fa;
	} else
		is_lowest_fa = t_data->is_lowest_fa;

	/* Reply is forwarded to lower mobility agent in any case if it is
	 * a FA; if it is a MN, the tear down message is not forwarded */
#ifdef USE_TEARDOWN
	if (!is_lowest_fa && idtype == ID_CHANGED &&
	    binding->id[1] != ntohl(ext->rep->id[1]) &&
	    ext->ext_dyn != NULL &&
	    (ext->ext_dyn->opts & REG_EXT_OWN_TEAR_DOWN) != 0) {
		handle_teardown_changed_id(binding, ext, unc);
	} else
#endif
	if (!t_data->is_lowest_fa || ext->ext_dyn == NULL ||
	    (ext->ext_dyn->opts & REG_EXT_OWN_TEAR_DOWN) == 0) {
		forward_reply(binding, ext, 1, unc);
	} else {
		DEBUG(DEBUG_FLAG, "handle_reply_disconnect: tear down reply "
		      "not forwarded to MN\n");
	}

	/* Delay binding removal in tear down processing */
	if (t_data->confirmed == 1) {
		DEBUG(DEBUG_FLAG, "Tear down/disconnect: Delay binding "
		      "removal\n");

		/* eliminate old session key so that all the new requests will
		 * be forwarded upwards */
		memset(&binding->key, 0, MAX_SK_LEN);
		binding->keylen = 0;

#ifdef REMOVE_BINDING_OPTIM
		binding->timeout = 0;
		binding->exp_time = time(NULL);
		if (update_binding(bindings_hash, binding, &bcounters) != 0)
			LOG2(LOG_WARNING, "binding removal: "
			     "update_binding failed\n");
#else
		/* Leave the disconnect binding up for
		 * DELAY_BINDING_TIME seconds so that possible queued
		 * data can still be forwarded.  The tear down message
		 * does not change the binding timeout in order to
		 * avoid DoS attacks with tear down messages. The
		 * session key is still reset, so the returning MN
		 * does not cause problems. */
		if (ext->ext_dyn == NULL ||
		    (ext->ext_dyn->opts & REG_EXT_OWN_TEAR_DOWN) == 0) {
			if (binding->timeout > 2) {
				DEBUG(DEBUG_FLAG, "Deregistration reply - "
				      "setting tunnel lifetime to 2 secs\n");
				binding->timeout = 2;
			}
			binding->exp_time = time(NULL) + binding->timeout;
			if (update_binding(bindings_hash, binding,
					   &bcounters) != 0)
				LOG2(LOG_WARNING, "delay binding removal: "
				     "update_binding failed\n");
		} else {
#ifdef SHORT_BINDING_DELAY_TIME /* ONLY FOR TESTING */
			if (binding->timeout != DELAY_BINDING_TIME) {
				DEBUG(DEBUG_FLAG, "Deregistration reply - "
				      "setting tunnel lifetime to 2 secs\n");
				binding->timeout = DELAY_BINDING_TIME;
			}
			binding->exp_time = time(NULL) + binding->timeout;
			if (update_binding(bindings_hash, binding,
					   &bcounters) != 0)
				LOG2(LOG_WARNING, "delay binding removal: "
				     "update_binding failed\n");
#endif /* SHORT_BINDING_DELAY_TIME */
			DEBUG(DEBUG_FLAG, "Teardown - only resetting the "
			      "session key\n");
		}

		return 0;
#endif
	}

	/* NOTE: With lazy tunnel deletion the tunnel will be up long 
	   enough (make sure that tunnel.c:PURGE_TUNNEL_TIME >= 2) */
	/* remove the tunnels associated to this binding */
	remove_binding_tunnels(binding, tunnels_hash, &bcounters);
	return 0;
}


static int
forward_reply(struct bindingentry *binding, struct msg_extensions *ext,
	      int error_reply, struct unconfirmed_request *unc)
{
	char msg[MAXMSG];
	int len, ret;
	struct tunnel_data *t_data;
	struct reg_rep *reply;
	struct msg_key *key;
	struct fa_spi_entry *fa_spi;
	struct packet_from_info *dst;

	assert(binding != NULL && binding->data != NULL);

	if (ext->rep == NULL) return 1;

	t_data = binding->data;
	if (!t_data->confirmed)
		error_reply = TRUE;

	if (error_reply && !IS_REGREP_ACCEPTED(ext->rep->code)) {
		DEBUG(DEBUG_FLAG, "Forwarding error reply - code=%i - %s\n",
		      ext->rep->code, reply_code_str(ext->rep->code));
	}
	else {
		DEBUG(DEBUG_FLAG, "Forwarding reply\n");
	}

	if (ext->mh_auth != NULL) {
		/* copy the MAC protected area directly from the received
		 * message. This area should include at least the reply
		 * header, mn_keyrep and mh_auth. */
		len = (__u8 *) ext->mh_auth - (__u8 *) ext->rep +
			GET_AUTH_EXT_LEN(ext->mh_auth);
		if (len < sizeof(struct reg_rep) +
		    GET_AUTH_EXT_LEN(ext->mh_auth) || len > MAXMSG) {
			LOG2(LOG_WARNING, "forward_reply: invalid length %i\n",
			     len);
			return 1;
		}
		memcpy(msg, ext->rep, len);
		DEBUG(DEBUG_FLAG, " * copying up to and including mh_auth "
		      "(len ==> %i)\n", len);
	} else {
		/* no mh_auth - copy only the reply header */
		memcpy(msg, ext->rep, sizeof(struct reg_rep));
		len = sizeof(struct reg_rep);
		DEBUG(DEBUG_FLAG, " * copying reply header (len=%i)\n", len);
	}

	reply = (struct reg_rep *) msg;

	if (!error_reply && binding->fa_spi != 0 &&
	    ntohs(ext->rep->lifetime) > 0) {
		fa_spi = get_fa_spi(binding->fa_spi,
				    (unc ? unc->info.src.sin_addr :
				     binding->lower_addr), 0);
		if (fa_spi == NULL) {
			LOG2(LOG_WARNING, "unknown FA SPI %i, IP=%s\n",
			     binding->fa_spi,
			     inet_ntoa(binding->lower_addr));
			return 1;
		}
		key = (struct msg_key *) (msg + len);
		if (((char *) MSG_KEY_DATA(key)) + MAX_SK_LEN > msg + MAXMSG) {
			LOG2(LOG_WARNING, "not enough space for SK to FA\n");
			return 1;
		}
		len += auth_encrypt(fa_spi->alg, fa_spi->shared_secret,
				    fa_spi->shared_secret_len, binding->key,
				    key, reply, VENDOR_EXT_DYNAMICS_FA_KEYREP,
				    htonl(fa_spi->spi));
	} else if (!error_reply && t_data->fa_pubkey &&
		   ntohs(ext->rep->lifetime) > 0) {
		/* encrypt the SK with lower FA pubkey */
		if (t_data->last_sent_fa_pubkeyrep == NULL) {
			t_data->last_sent_fa_pubkeyrep =
				dynamics_do_rsa_encrypt(binding->key,
							binding->keylen,
							t_data->fa_pubkey);
			if (t_data->last_sent_fa_pubkeyrep == NULL) {
				LOG2(LOG_WARNING, "RSA encrypt failed\n");
				return 1;
			}
		}

		if (len + GET_KEY_EXT_LEN(t_data->last_sent_fa_pubkeyrep) >
		    MAXMSG) {
			LOG2(LOG_WARNING, "Message buffer overflow\n");
			return 1;
		}
		memcpy(msg + len, t_data->last_sent_fa_pubkeyrep,
		       GET_KEY_EXT_LEN(t_data->last_sent_fa_pubkeyrep));
		DEBUG(DEBUG_FLAG,
		      " * adding fa_pubkeyrep extension: pos=%i, len=%i\n",
		      len, GET_KEY_EXT_LEN(t_data->last_sent_fa_pubkeyrep));
		len += GET_KEY_EXT_LEN(t_data->last_sent_fa_pubkeyrep);
	} else if (!error_reply) {
		DEBUG(DEBUG_FLAG,
		      " * no pubkey known - FA encrypt session key not "
		      "added\n");
	}

	if (ext->ext_dyn != NULL) {
		int elen = ext->ext_dyn->length + 2;
		if (len + elen > MAXMSG) {
			LOG2(LOG_WARNING, "Message buffer overflow\n");
			return 1;
		}
		memcpy(msg + len, ext->ext_dyn, elen);
		DEBUG(DEBUG_FLAG, " * copying ext_dyn: pos=%i, len=%i\n",
		      len, elen);
		len += elen;
	}

	if (ext->sfa_debug != NULL) {
		int elen = GET_SFA_DEBUG_EXT_LEN(ext->sfa_debug);
		if (len + elen > MAXMSG) {
			LOG2(LOG_WARNING, "Message buffer overflow\n");
			return 1;
		}
		memcpy(msg + len, ext->sfa_debug, elen);
		DEBUG(DEBUG_FLAG, " * copying sfa_debug: pos=%i, len=%i\n",
		      len, elen);
		len += elen;
	}

	if (len + sizeof(struct msg_auth) + MD5_MAC_LEN > MAXMSG) {
		LOG2(LOG_WARNING, "Message buffer overflow\n");
		return 1;
	}

	if (!error_reply && ext->sk_auth != NULL) {
		DEBUG(DEBUG_FLAG,
		      " * adding sk_auth extension: pos=%i, len=%i\n",
		      len, MD5_MAC_LEN + SPI_LEN);
		len += auth_add_vendor(AUTH_ALG_MD5, binding->key,
				       binding->keylen,
				       (unsigned char *) reply,
				       (struct vendor_msg_auth *) (msg + len),
				       VENDOR_EXT_DYNAMICS_SK_AUTH,
				       ext->sk_auth->spi);
	}

	len += add_fa_auth_ext((unc ? unc->info.src.sin_addr :
				binding->lower_addr),
			       reply->home_addr, (unsigned char *) reply,
			       (unsigned char *) msg + len);

	assert(len <= MAXMSG);

	dst = NULL;
	if (unc && unc->is_lowest_fa)
		dst = &unc->info;
	else if (!unc && t_data->is_lowest_fa)
		dst = &t_data->info;

	if (config->packet_socket_mode != PACKET_SOCKET_MODE_ONLY_RECEIVE &&
	    dst != NULL) {
		DEBUG(DEBUG_FLAG, " * sending %i bytes using packet socket\n",
		      len);
		ret = packet_socket_send(dst, msg, len);
	} else if (unc) {
		DEBUG(DEBUG_FLAG, " * sending %i bytes to %s:%i (unc)\n",
		      len, inet_ntoa(unc->info.src.sin_addr),
		      ntohs(unc->info.src.sin_port));
		assert(unc->info.iface != NULL);
		ret = own_sendto(unc->info.iface->udp_sock,
				 unc->info.src.sin_addr,
				 unc->info.src.sin_port, msg, len);
	} else {
		DEBUG(DEBUG_FLAG, " * sending %i bytes to %s:%i\n",
		      len, inet_ntoa(binding->lower_addr),
		      ntohs(binding->lower_port));
		assert(t_data->info.iface != NULL);
#if 1
		ret = own_sendto(t_data->info.iface->udp_sock,
#else
     	        ret = own_sendto(up_interface->udp_sock,
#endif
				 binding->lower_addr,
				 MMGR_MIP_PORT, msg, len);
	}
	return dynamics_check_sendto(ret, len, "Forward reply");
}


/* try to confirm an unconfirmed binding */
static int
handle_unconfirmed_reply(struct bindingentry *binding,
			 struct msg_extensions *ext, unsigned char *sk,
			 int sk_len, struct unconfirmed_request *unc)
{
	struct tunnel_data *t_data = (struct tunnel_data *) binding->data;

	/* check that the session key and sk_auth extension match */
	if ((ext->fa_pubkeyrep != NULL || ext->fa_keyrep) && ext->sk_auth) {
		if (!auth_check_vendor(AUTH_ALG_MD5, sk, sk_len,
				       (unsigned char *) ext->rep,
				       ext->sk_auth)) {
			LOG2(LOG_WARNING, "Invalid sk_auth in reply\n");
			return 1;
		}
	} else if (t_data->dynamics_extensions) {
		/* this should not be reached as dynamics_extensions implies
		 * both fa_pubkeyrep and sk_auth extensions */
		LOG2(LOG_WARNING, "Extensions fa_pubkeyrep, fa_keyrep and/or "
		     "sk_auth missing\n");
		return 1;
	}

	/* okay, match found - fill the binding with confirmed data */
	unconfirmed_to_binding(binding, unc);

	t_data->confirmed = TRUE;
	DEBUG(DEBUG_FLAG, "\tconfirming the binding\n");
	bcounters.bindingcount++;
	bcounters.pendingcount--;
	free_unconfirmed_data(t_data->unc_req);
	t_data->unc_req = NULL;

	return 0;
}


/**
 * update_reply_binding_info:
 * @binding:
 * @unc: 
 * 
 * Update binding info based on request data
 */
static void
update_reply_binding_info(struct bindingentry *binding, 
			  const struct unconfirmed_request *unc)
{
	struct tunnel_data *t_data = (struct tunnel_data *) binding->data;

	/* update down interface pointer */
	memcpy(&t_data->info, &unc->info, sizeof(unc->info));
	t_data->old_if_index = unc->info.iface->if_index;

	/* update lower end type (MN/FA) */
	t_data->is_lowest_fa = unc->is_lowest_fa;

	/* update MN/FA decapsulation mode */
	t_data->fa_decapsulation = unc->fa_decapsulation;

	/* update encapsulating delivery mode */
	t_data->encaps_delivery = 
		!t_data->is_lowest_fa || unc->encaps_delivery;

	/* update down_type and down_key */
	t_data->down_type = unc->tunnel_type;
	t_data->down_key = unc->tunnel_key;

	/* update binding lower address and port */
	binding->lower_addr = unc->info.src.sin_addr;
	binding->lower_port = unc->info.src.sin_port;

	binding->fa_spi = 0;

	setup_device_forcing(t_data);

	/* update tunnel mode */
	t_data->tunnel_mode = unc->tunnel_mode;

#ifdef USE_TEARDOWN
	/* FIX-JM: update t_data->upper_id ?? */
#endif
}



/* handle a reply to an unconfirmed requests of a confirmed binding and switch
 * the lower side of the tunnel if needed
 * Returns: 0 = success, 1 = failure */
static int
handle_confirmed_reply(struct bindingentry *binding,
		       struct msg_extensions *ext,
		       struct unconfirmed_request *unc)
{
	struct tunnel_data *t_data = (struct tunnel_data *) binding->data;

	assert(unc != NULL);

	/* replace the binding data with the previously unconfirmed data */
	DEBUG(DEBUG_FLAG,
	      "Confirmed binding data - old lower_addr: %s, ",
	      inet_ntoa(binding->lower_addr));
	DEBUG(DEBUG_FLAG, "new lower_addr: %s\n",
	      inet_ntoa(unc->info.src.sin_addr));

	if (unc->info.src.sin_addr.s_addr != binding->lower_addr.s_addr ||
	    unc->info.src.sin_port != binding->lower_port ||
	    unc->tunnel_mode != t_data->tunnel_mode ||
	    unc->is_lowest_fa != t_data->is_lowest_fa ||
	    unc->fa_decapsulation != t_data->fa_decapsulation ||
	    unc->encaps_delivery != t_data->encaps_delivery ||
	    unc->info.iface != t_data->info.iface) {
		struct saved_binding_info old;
		/* the iface argument is NULL because the reply may not come
		 * from the same interface as the request */
		save_binding_info(&old, binding);
		update_reply_binding_info(binding, unc);
		if (handle_lower_switch(binding, &old, ext,
					&unc->info.src.sin_addr,
					FALSE) != 0)
			return 1;
	}

	unconfirmed_to_binding(binding, unc);
	free_unconfirmed_data(t_data->unc_req);
	t_data->unc_req = NULL;

	return 0;
}


static int
get_fa_session_key_sec(unsigned char *sk, int *sk_len, struct reg_rep *reply,
		       struct msg_key *fa_keyrep, struct in_addr addr)
{
	struct fa_spi_entry *fa_spi;
	unsigned char *tmp_sk;
#ifndef NODEBUGMODE
	int i;
#endif

	fa_spi = get_fa_spi(ntohl(fa_keyrep->spi), addr, 0);
	if (fa_spi == NULL) {
		LOG2(LOG_WARNING, "Could not handle fa_keyrep with "
		     "SPI %i\n", ntohl(fa_keyrep->spi));
		return 1;
	}

	tmp_sk = auth_decrypt(fa_spi->alg, fa_spi->shared_secret,
			      fa_spi->shared_secret_len,
			      fa_keyrep, reply, sk_len);
	if (*sk_len < 0) {
		LOG2(LOG_WARNING, "Could not decrypt fa_keyrep with "
		     "SPI %i\n", ntohl(fa_keyrep->spi));
		return 1;
	}
	if (*sk_len > MAX_SK_LEN) {
		LOG2(LOG_WARNING, "Too long session key\n");
		free(tmp_sk);
		return 1;
	}
	memcpy(sk, tmp_sk, *sk_len);
	free(tmp_sk);

#ifndef NODEBUGMODE
	DEBUG(DEBUG_FLAG, "Session key (MD5): ");
	for (i = 0; i < *sk_len; i++)
		DEBUG(DEBUG_FLAG, "%02x", sk[i]);
	DEBUG(DEBUG_FLAG, "\n");
#endif

	return 0;
}


static int
get_fa_session_key_pub(unsigned char *sk, int *sk_len,
		       struct tunnel_data *t_data,
		       struct msg_key *fa_pubkeyrep)
{
	if (t_data->last_sent_fa_pubkeyrep != NULL) {
		free(t_data->last_sent_fa_pubkeyrep);
		t_data->last_sent_fa_pubkeyrep = NULL;
	}

	if (t_data->last_recv_fa_pubkeyrep != NULL)
		free(t_data->last_recv_fa_pubkeyrep);
	t_data->last_recv_fa_pubkeyrep =
		(struct msg_key *) malloc(GET_KEY_EXT_LEN(fa_pubkeyrep));
	if (t_data->last_recv_fa_pubkeyrep == NULL)
		return 1;
	memcpy(t_data->last_recv_fa_pubkeyrep, fa_pubkeyrep,
	       GET_KEY_EXT_LEN(fa_pubkeyrep));

	*sk_len = MD5_SK_LEN;
	if (do_rsa_decrypt(fa_pubkeyrep, sk, *sk_len) != 0)
		return 1;

	return 0;
}

/** 
 * get_fa_session_key:
 * @sk: 
 * @sk_len:
 * @ext:
 * @addr:
 * @binding:
 *
 * Extract the session key from @ext (distributed either by symmetric 
 * secret encryption, or public key encryption). If the @ext does not
 * contain a key use an already decrypted key in @binding.
 *
 * The key is placed in @sk, and its length in @sk_len.
 *
 * Returns: 0 if successful
 */
static int
get_fa_session_key(unsigned char *sk, int *sk_len, struct msg_extensions *ext,
		   struct in_addr addr, struct bindingentry *binding)
{
	struct tunnel_data *t_data = (struct tunnel_data *) binding->data;

	if (ext->fa_keyrep) {
		/* shared secret based session key distribution */
		return get_fa_session_key_sec(sk, sk_len, ext->rep,
					      ext->fa_keyrep, addr);
	}

	/* public key based session key distribution */
	if (ext->fa_pubkeyrep != NULL && ntohs(ext->rep->lifetime) > 0 &&
	    !equal_pubkey(ext->fa_pubkeyrep, t_data->last_recv_fa_pubkeyrep)) {
		/* new key */
		return get_fa_session_key_pub(sk, sk_len, t_data,
					       ext->fa_pubkeyrep);
	}

	/* use already decrypted key */
	memcpy(sk, binding->key, binding->keylen);
	*sk_len = binding->keylen;

	return 0;
}


struct binding_search_data {
	struct in_addr mn_addr;
	__u32 id1;
};
static struct binding_search_data bsearch_data;

/* returns 1 if bsearch_data matches with the given binding or 0 if it does
 * not match; to be used with binding_fetch_func() */
static int
binding_search_check(struct bindingentry *binding)
{
	struct tunnel_data *t_data;
	struct unconfirmed_request *unc;

	if (binding->mn_addr.s_addr != bsearch_data.mn_addr.s_addr)
		return 0;

	if (binding->data == NULL)
		return 0;
	t_data = (struct tunnel_data *) binding->data;

	if (t_data->confirmed && bsearch_data.id1 == binding->id[1])
		return 1;

	unc = t_data->unc_req;
	while (unc != NULL) {
		if (unc->id[1] == bsearch_data.id1)
			return 1;
		unc = unc->next;
	}

	return 0;
}


/**
 * handle_reply:
 * @ext:
 * @info: source of reply message
 * @iface: interface where reply arrived
 *
 * Process a registration reply message.
 *
 * Returns: 0 if OK, else 1
 */
int
handle_reply(struct msg_extensions *ext, struct packet_from_info *info)
{
	struct bindingentry *binding;
	struct tunnel_data *t_data;
	unsigned char sk[MAX_SK_LEN];
	int prev_confirmed, sk_len, i, idtype;
	struct unconfirmed_request *unc;
	char tmp[30];
	struct bindingkey bkey;

	DEBUG(DEBUG_FLAG, "Handling reply\n");

	memcpy(&bkey.mn_addr, &(ext->rep->home_addr), sizeof(bkey.mn_addr));
	memcpy(&bkey.ha_addr, &(ext->rep->ha_addr), sizeof(bkey.ha_addr));
	if (ext->priv_ha != NULL)
		bkey.priv_ha = ntohl(ext->priv_ha->priv_ha);
	else
		bkey.priv_ha = 0;
	if (ext->rep->code == REGREP_UNKNOWN_HA_HA) {
		bsearch_data.mn_addr.s_addr = ext->rep->home_addr.s_addr;
		bsearch_data.id1 = ntohl(ext->rep->id[1]);
		binding = binding_fetch_func(bindings_hash,
					     binding_search_check);
	} else
		binding = binding_fetch(bindings_hash, &bkey);

	if (binding == NULL) {
		dynamics_strlcpy(tmp, inet_ntoa(info->src.sin_addr),
				 sizeof(tmp));
		LOG2(LOG_WARNING, "Received reply from %s:%i for MN %s, but "
		     "no binding found\n", tmp, ntohs(info->src.sin_port),
		     inet_ntoa(ext->rep->home_addr));
		return 1;
	}
	t_data = binding->data;
	t_data->pending_request = FALSE;

	unc = find_matching_request(binding, ntohl(ext->rep->id[1]), &idtype);
	if (idtype == ID_NOT_FOUND || (t_data->info.iface == NULL &&
				       unc == NULL)) {
		LOG2(LOG_WARNING, "Could not found a matching request for the "
		     "reply - dropping it\n");
		return 1;
	}
	assert(t_data->confirmed || unc != NULL);

	i = validate_reply(ext, binding, info, unc);
	if (i < 0) {
		LOG2(LOG_WARNING, "Reply validation failed - dropping it\n");
		return 1;
	}

	if (i == 1) {
		DEBUG(DEBUG_FLAG, "Reply did not have Dynamics extensions - "
		      "assuming that HA/upper FA does not support them\n");
		t_data->dynamics_extensions = 0;
	} else {
		t_data->dynamics_extensions = 1;
	}

	if (get_fa_session_key(sk, &sk_len, ext, info->src.sin_addr,
			       binding) != 0)
		return 1;

	if (!IS_REGREP_ACCEPTED(ext->rep->code)) {
		/* Note: RFC 2002 - chapter 3.7.2:
		 * "If the Registration Reply indicates that the Request (for
		 * registration or deregistration) was denied by the home
		 * agent, the existing visitor list entry for the mobile node
		 * MUST NOT be modified as a result of receiving the
		 * Registration Reply." */
		return forward_reply(binding, ext, 1, unc);
	}

	if (ntohs(ext->rep->lifetime) == 0) {
		/* Tear down or disconnect */
		return handle_reply_disconnect(binding, unc, ext, idtype);
	}

	prev_confirmed = t_data->confirmed;
	if (!t_data->confirmed) {
		if (handle_unconfirmed_reply(binding, ext, sk, sk_len, unc) !=
		    0) {
			LOG2(LOG_WARNING, "Reply to an unconfirmed binding "
			     "could not be confirmed - dropping reply\n");
			return 1;
		}
	} else {
		if (idtype == ID_NEW &&
		    handle_confirmed_reply(binding, ext, unc) != 0) {
			LOG2(LOG_WARNING, "Reply to a confirmed binding - "
			     "confirming request data failed\n");
			return 1;
		}
	}

	assert(t_data->confirmed);

	/* update the session key */
	memcpy(binding->key, sk, sk_len);
	binding->keylen = sk_len;

	time(&binding->mod_time);

	if (ntohs(ext->rep->lifetime) > t_data->req_lifetime) {
		/* RFC 2002, sec. 3.4:
		 * The home agent MUST NOT increase the Lifetime selected by
		 * the mobile node in the Registration Request, since doing so
		 * could increase it beyond the maximum Registration Lifetime
		 * allowed by the foreign agent. If the Lifetime received in
		 * the Registration Reply is greater than that in the
		 * Registration Request, the Lifetime in the Request MUST be
		 * used. */
		LOG2(LOG_WARNING, "Lifetime too long in reply (%i) - using the"
		     " lifetime of the request (%i)\n",
		     ntohs(ext->rep->lifetime), t_data->req_lifetime);
		binding->timeout = t_data->req_lifetime;
	} else
		binding->timeout = ntohs(ext->rep->lifetime);

	if (binding->timeout > config->fa_default_tunnel_lifetime) {
		DEBUG(DEBUG_FLAG, "Lifetime too long - using max lifetime\n");
		binding->timeout = config->fa_default_tunnel_lifetime;
	}

	/* increase timeout with one second so that the one second timer
	 * resolution doesn't cause too early expirations */
	binding->timeout++;
	binding->exp_time = time(NULL) + binding->timeout;
	if (update_binding(bindings_hash, binding, &bcounters) != 0) {
		LOG2(LOG_WARNING, "handle_reply: update_binding failed\n");
		return 1;
	}

	/* In LFA prepare the tunnels first but in IFA send the reply first
	 * and do the tunnel stuff afterwards. This speeds up the location
	 * update time when using IFAs */

	if (!t_data->is_lowest_fa) {
		DEBUG(DEBUG_FLAG, "handle_reply: not lowest FA: forwarding "
		      "reply first and then create_tunnels\n");
		i = forward_reply(binding, ext, 0, NULL);
	}

	/* create the tunnels if the binding was confirmed */
	if (!prev_confirmed && t_data->confirmed &&
	    create_tunnels(binding, tunnels_hash, config->highest_FA) != 0) {
		LOG2(LOG_WARNING, "handle_reply: create_tunnels failed\n");
		return 1;
	}

	if (t_data->is_lowest_fa) {
		DEBUG(DEBUG_FLAG, "handle_reply: lowest FA: forwarding reply "
		      "after create_tunnels ok\n");
		i = forward_reply(binding, ext, 0, NULL);
	}

	return i;	
}
