/* $Id: dyn_ip.c,v 1.79 2001/10/04 16:46:37 jm Exp $
 * Tunnel and route interface to kernel
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

/* This interface code is heavily based on Alexey Kuznetsov's iproute2 */


#define DEBUG_FLAG 'K'

#include "config.h"

#include "owntypes.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <netinet/ip.h>
#include <errno.h>
#include <unistd.h>
#ifdef DYN_TARGET_LINUX
#include <linux/if_tunnel.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <net/if_arp.h>
#endif
#include <time.h>
#include <fcntl.h>
#include <sys/uio.h>

#include "debug.h"
#include "dyn_ip.h"
#include "util.h"

#ifdef DYN_TARGET_WINDOWS
#include "windows_extra.h"
#endif

#ifndef MSG_TRUNC
#define MSG_TRUNC 0x20
#endif

#ifndef IPPROTO_GRE
#define IPPROTO_GRE 47
#endif


#ifdef DYN_TARGET_LINUX

struct rtnl_handle
{
	int fd;
	struct sockaddr_nl local;
	__u32 seq;
	__u32 dump;
};

static struct idxmap *idxmap = NULL;


/* generic tunnel ioctl caller for dyn_ip_tunnel_add(),
 * dyn_ip_tunnel_add_gre(), and dyn_ip_tunnel_del() */
static int tunnel_ioctl(int cmd, const char *dev, __u32 remote, __u32 local,
			int proto, __u32 key)
{
	struct ifreq ifr;
	struct ip_tunnel_parm p;
	int s;

	assert(dev != NULL && strlen(dev) < IFNAMSIZ);

	memset(&p, 0, sizeof(p));
	p.iph.version = 4;
	p.iph.ihl = 5;
	p.iph.frag_off = htons(IP_DF); /* Flag: Don't Fragment */
	p.iph.ttl = 64; /* fixed TTL for the outer IP header to make the
			 * tunnel transparent for traceroute like applications
			 */
	dynamics_strlcpy(p.name, dev, IFNAMSIZ);

	if (cmd == SIOCADDTUNNEL) {
		if (key != 0 && proto == IPPROTO_GRE) {
			p.i_key = p.o_key = htonl(key);
			p.i_flags |= htons(0x2000); /* GRE_KEY */
			p.o_flags |= htons(0x2000); /* GRE_KEY */
		}
		p.iph.protocol = proto;
		p.iph.daddr = remote;
		p.iph.saddr = local;
		p.iph.tos = 0x0004;
		if (proto == IPPROTO_IPIP)
			dynamics_strlcpy(ifr.ifr_name, "tunl0", IFNAMSIZ);
		else if (proto == IPPROTO_GRE)
			dynamics_strlcpy(ifr.ifr_name, "gre0", IFNAMSIZ);
	} else if (cmd == SIOCDELTUNNEL) {
		dynamics_strlcpy(ifr.ifr_name, dev, IFNAMSIZ);
	} else {
		DEBUG(DEBUG_FLAG, "tunnel_ioctl: Unknown cmd %i\n", cmd);
		return -1;
	}


	ifr.ifr_ifru.ifru_data = (void*) &p;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		DEBUG(DEBUG_FLAG, "tunnel_ioctl: socket: %s\n",
		      strerror(errno));
		return -1;
	}

	if (ioctl(s, cmd, &ifr) != 0) {
		DEBUG(DEBUG_FLAG, "tunnel_ioctl: ioctl: %s\n",
		      strerror(errno));
		close(s);
		return -1;
	}
	close(s);

	return 0;
}


/**
 * dyn_ip_tunnel_add:
 * @dev: the device name of the tunnel to add
 * @remote: IP address of remote end of tunnel
 * @local: IP address of local end of tunnel
 * 
 * Add an IP Encapsulation within IP (RFC 2003) tunnel named
 * @dev between @local and @remote.
 *
 * Returns: 0 if successful, else -1
 */
int dyn_ip_tunnel_add(const char *dev, struct in_addr remote, 
		      struct in_addr local)
{
	return tunnel_ioctl(SIOCADDTUNNEL, dev, remote.s_addr, local.s_addr,
			    IPPROTO_IPIP, 0);
}


/**
 * dyn_ip_tunnel_add_gre:
 * @dev: the device name of the tunnel to add
 * @remote: IP address of remote end of tunnel
 * @local: IP address of local end of tunnel
 * @key: Key field for the GRE header (used to index the tunnels between same
 * end-points)
 *
 * Add a Generic Routing Encapsulation (RFC 1701) tunnel named @dev
 * between @local and @remote.
 *
 * Returns: 0 if successful, else -1
 */
int dyn_ip_tunnel_add_gre(const char *dev, struct in_addr remote,
			  struct in_addr local, int key)
{
	return tunnel_ioctl(SIOCADDTUNNEL, dev, remote.s_addr, local.s_addr,
			    IPPROTO_GRE, key);
}


/**
 * dyn_ip_tunnel_del:
 * @dev: the name of the device to delete
 *
 * Delete tunnel named @dev.
 *
 * Returns: 0 if successful, else -1
 */
int dyn_ip_tunnel_del(const char *dev)
{
	return tunnel_ioctl(SIOCDELTUNNEL, dev, 0, 0, IPPROTO_IPIP, 0);
}


/**
 * dyn_ip_link_set_dev:
 * @dev: interface to be modified
 * @flag: flags to be set
 * @set: 1=set given flags, 0=unset given flags
 *
 * Modifies given interface @dev by setting or unsetting interface flags @flag.
 *
 * Returns: 0 on success or -1 on failure
 */
static int dyn_ip_link_set_dev(const char *dev, int flag, int set)
{
	struct ifreq ifr;
	int s, changed;

	assert(dev != NULL);
	assert(strlen(dev) < IFNAMSIZ);
	dynamics_strlcpy(ifr.ifr_name, dev, IFNAMSIZ);

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_link_set_dev: socket: %s\n",
		      strerror(errno));
		return -1;
	}

	if (ioctl(s, SIOCGIFFLAGS, &ifr) != 0) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_link_set_dev: ioctl SIOCGIFFLAGS: %s\n",
		      strerror(errno));
		close(s);
		return -1;
	}

	changed = 0;
	if (set) {
		if ((ifr.ifr_flags & flag) == 0) {
			ifr.ifr_flags |= flag;
			changed = 1;
		}
	} else {
		if ((ifr.ifr_flags & flag) != 0) {
			ifr.ifr_flags &= ~flag;
			changed = 1;
		}
	}

	if (changed) {
		if (ioctl(s, SIOCSIFFLAGS, &ifr) != 0) {
			DEBUG(DEBUG_FLAG,
			      "dyn_ip_link_set_dev_up: ioctl SIOCSIFFLAGS: "
			      "%s\n", strerror(errno));
			close(s);
			return -1;
		}
	}

	close(s);

	return 0;
}


/**
 * dyn_ip_link_set_dev_up:
 * @dev: network device name
 *
 * Set the link on device @dev up.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_link_set_dev_up(const char *dev)
{
	return dyn_ip_link_set_dev(dev, IFF_UP, 1);
}


/**
 * dyn_ip_link_set_dev_down:
 * @dev: network device name
 * 
 * Set the link on device @dev down.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_link_set_dev_down(const char *dev)
{
	return dyn_ip_link_set_dev(dev, IFF_UP, 0);
}


/**
 * rtnl_open:
 * @rth: 
 * @subscriptions: 
 * 
 * 
 * 
 * Returns: 
 */
static int rtnl_open(struct rtnl_handle *rth, unsigned subscriptions)
{
	unsigned int len;

	assert(rth != NULL);
	rth->fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (rth->fd < 0) {
		DEBUG(DEBUG_FLAG,
		      "rtnl_open: cannot open netlink socket: %s\n",
		      strerror(errno));
		return -1;
	}

	memset(&rth->local, 0, sizeof(rth->local));
	rth->local.nl_family = AF_NETLINK;
	rth->local.nl_groups = subscriptions;

	if (bind(rth->fd, (struct sockaddr *) &rth->local,
		 sizeof(rth->local)) < 0) {
		DEBUG(DEBUG_FLAG,
		      "rtnl_open: cannot bind netlink socket: %s\n",
		      strerror(errno));
		close(rth->fd);
		return -1;
	}

	len = sizeof(rth->local);
	if (getsockname(rth->fd, (struct sockaddr *) &rth->local, &len) < 0) {
		DEBUG(DEBUG_FLAG,
		      "rtnl_open: cannot getsockname: %s\n", strerror(errno));
		close(rth->fd);
		return -1;
	}
	if (len != sizeof(rth->local)) {
		DEBUG(DEBUG_FLAG,
		      "rtnl_open: wrong address length %i != %i\n",
		      len, sizeof(rth->local));
		close(rth->fd);
		return -1;
	}

	if (rth->local.nl_family != AF_NETLINK) {
		DEBUG(DEBUG_FLAG,
		      "rtnl_open: wrong address family %i != %i\n",
		      rth->local.nl_family, AF_NETLINK);
		close(rth->fd);
		return -1;
	}

	rth->seq = time(NULL);
	return 0;
}


/**
 * rtnl_talk:
 * @rth: 
 * @n: 
 * @answer: 
 * 
 * 
 * 
 * Returns: 
 */
static int rtnl_talk(struct rtnl_handle *rth, struct nlmsghdr *n,
		     struct nlmsghdr *answer)
{
	int status;
	struct nlmsghdr *h;
	struct sockaddr_nl nladdr;
	struct iovec iov;
	char   buf[8192];
	struct msghdr msg = {
		(void *) &nladdr, sizeof(nladdr),
		&iov, 1,
		NULL, 0,
		0
	};

	iov.iov_base = (void*) n;
	iov.iov_len = n->nlmsg_len;

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	nladdr.nl_pid = 0;
	nladdr.nl_groups = 0;

	n->nlmsg_seq = ++rth->seq;

	if (answer == NULL)
		n->nlmsg_flags |= NLM_F_ACK;

	status = sendmsg(rth->fd, &msg, 0);

	if (status < 0) {
		DEBUG(DEBUG_FLAG, "rtnl_talk: sendmsg: %s\n", strerror(errno));
		return -1;
	}

	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);

	do {
		status = recvmsg(rth->fd, &msg, 0);

		if (status < 0) {
			if (errno == EINTR)
				continue;
			DEBUG(DEBUG_FLAG, "rtnl_talk: OVERRUN: %s\n",
					 strerror(errno));
			continue;
		}
		if (status == 0) {
			DEBUG(DEBUG_FLAG, "rtnl_talk: EOF on netlink\n");
			return -1;
		}
		if (msg.msg_namelen != sizeof(nladdr)) {
			DEBUG(DEBUG_FLAG,
			      "rtnl_talk: sender address length == %i != %i\n",
			      msg.msg_namelen, sizeof(nladdr));
			return -1;
		}
		for (h = (struct nlmsghdr*) buf; status >= sizeof(*h); ) {
			int len = h->nlmsg_len;
			pid_t pid = h->nlmsg_pid;
			int l = len - sizeof(*h);
			unsigned seq = h->nlmsg_seq;

			if (l < 0 || len > status) {
				if (msg.msg_flags & MSG_TRUNC) {
					DEBUG(DEBUG_FLAG,
					      "rtnl_talk: "
					      "truncated message\n");
					return -1;
				}
				DEBUG(DEBUG_FLAG,
				      "rtnl_talk: malformed message: len=%d\n",
				      len);
				return -1;
			}

			if (h->nlmsg_pid != pid || h->nlmsg_seq != seq) {
				/* junk */
				continue;
			}

			if (h->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *err =
					(struct nlmsgerr *) NLMSG_DATA(h);
				if (l < sizeof(struct nlmsgerr)) {
					DEBUG(DEBUG_FLAG,
					      "rtnl_talk: ERROR truncated\n");
					return -1;
				}

				errno = -err->error;
				if (errno == 0) {
					if (answer)
						memcpy(answer, h,
						       h->nlmsg_len);
					return 0;
				}
				DEBUG(DEBUG_FLAG,
				      "rtnl_talk: RTNETLINK error: "
				      "%s\n", strerror(errno));

				return -1;
			}

			if (answer) {
				memcpy(answer, h, h->nlmsg_len);
				return 0;
			}

			DEBUG(DEBUG_FLAG, "rtnl_talk: unexpected reply\n");

			status -= NLMSG_ALIGN(len);
			h = (struct nlmsghdr *)
 				(((char *) h) + NLMSG_ALIGN(len));
		}
		if (msg.msg_flags & MSG_TRUNC) {
			DEBUG(DEBUG_FLAG, "rtnl_talk: message truncated\n");
			continue;
		}
		if (status) {
			DEBUG(DEBUG_FLAG, "rtnl_talk: remnant of size %i\n",
			      status);
			return -1;
		}
	} while (1);
}


/**
 * addattr32:
 * @n: 
 * @maxlen: 
 * @type: 
 * @data: 
 * 
 * 
 * 
 * Returns: 
 */
static int addattr32(struct nlmsghdr *n, int maxlen, int type, __u32 data)
{
	int len;
	struct rtattr *rta;

	len = RTA_LENGTH(4);
	if (NLMSG_ALIGN(n->nlmsg_len) + len > maxlen)
		return -1;
	rta = (struct rtattr *)(((char *) n) + NLMSG_ALIGN(n->nlmsg_len));
	rta->rta_type = type;
	rta->rta_len = len;
	memcpy(RTA_DATA(rta), &data, 4);
	n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + len;
	return 0;
}


/**
 * addattr_l:
 * @n: 
 * @maxlen: 
 * @type: 
 * @data: 
 * @alen: 
 * 
 * 
 * 
 * Returns: 
 */
static int addattr_l(struct nlmsghdr *n, int maxlen, int type,
		     const void *data, int alen)
{
	int len;
	struct rtattr *rta;

	len = RTA_LENGTH(alen);
	if (NLMSG_ALIGN(n->nlmsg_len) + len > maxlen)
		return -1;
	rta = (struct rtattr *)(((char *) n) + NLMSG_ALIGN(n->nlmsg_len));
	rta->rta_type = type;
	rta->rta_len = len;
	memcpy(RTA_DATA(rta), data, alen);
	n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + len;
	return 0;
}


static int rtnl_wilddump_request(struct rtnl_handle *rth, int family, int type)
{
	struct {
		struct nlmsghdr nlh;
		struct rtgenmsg g;
	} req;
	struct sockaddr_nl nladdr;

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;

	req.nlh.nlmsg_len = sizeof(req);
	req.nlh.nlmsg_type = type;
	req.nlh.nlmsg_flags = NLM_F_ROOT | NLM_F_MATCH | NLM_F_REQUEST;
	req.nlh.nlmsg_pid = 0;
	req.nlh.nlmsg_seq = rth->dump = ++rth->seq;
	req.g.rtgen_family = family;

	return sendto(rth->fd, (void *) &req, sizeof(req), 0,
		      (struct sockaddr *) &nladdr, sizeof(nladdr));
}


static int rtnl_dump_filter(struct rtnl_handle *rth,
			    int (*filter)(struct sockaddr_nl *,
					  struct nlmsghdr *n, void *),
			    void *arg1)
{
	char buf[8192];
	struct sockaddr_nl nladdr;
	struct iovec iov = { buf, sizeof(buf) };

	do {
		int status;
		struct nlmsghdr *h;

		struct msghdr msg = {
			(void *) &nladdr, sizeof(nladdr),
			&iov,	1,
			NULL,	0,
			0
		};

		status = recvmsg(rth->fd, &msg, 0);

		if (status < 0) {
			if (errno == EINTR)
				continue;
			DEBUG(DEBUG_FLAG,
			      "rtnl_dump_filter: recvmsg OVERRUN: %s\n",
			      strerror(errno));
			continue;
		}
		if (status == 0) {
			DEBUG(DEBUG_FLAG,
			      "rtnl_dump_filter: EOF on netlink\n");
			return -1;
		}
		if (msg.msg_namelen != sizeof(nladdr)) {
			DEBUG(DEBUG_FLAG,
			      "rtnl_dump_filter: sender address length "
			      "== %i\n", msg.msg_namelen);
			return -1;
		}

		h = (struct nlmsghdr *) buf;
		while (NLMSG_OK(h, status)) {
			int err;

			if (h->nlmsg_pid != rth->local.nl_pid ||
			    h->nlmsg_seq != rth->dump) {
				DEBUG(DEBUG_FLAG, "rtnl_dump_filter: junk?\n");
				h = NLMSG_NEXT(h, status);
				continue;
			}

			if (h->nlmsg_type == NLMSG_DONE)
				return 0;
			if (h->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *err =
					(struct nlmsgerr *) NLMSG_DATA(h);
				if (h->nlmsg_len <
				    NLMSG_LENGTH(sizeof(struct nlmsgerr))) {
					DEBUG(DEBUG_FLAG,
					      "rtnl_dump_filter: "
					      "ERROR truncated\n");
				} else {
					errno = -err->error;
					DEBUG(DEBUG_FLAG,
					      "rtnl_dump_filter: "
					      "RTNETLINK error: %s\n",
					      strerror(errno));
				}
				return -1;
			}
			err = filter(&nladdr, h, arg1);
			if (err < 0)
				return err;

			h = NLMSG_NEXT(h, status);
		}
		if (msg.msg_flags & MSG_TRUNC) {
			DEBUG(DEBUG_FLAG,
			      "rtnl_dump_filter: message truncated\n");
			continue;
		}
		if (status) {
			DEBUG(DEBUG_FLAG,
			      "rtnl_dump_filter: remnant of size %d\n",
			      status);
			return -1;
		}
	} while (1);
}


static int parse_rtattr(struct rtattr *tb[], int max, struct rtattr *rta,
			int len)
{
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= max)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta,len);
	}
	if (len)
		DEBUG(DEBUG_FLAG,  "parse_rtattr: Deficit %d, rta_len=%d\n",
		      len, rta->rta_len);
	return 0;
}


static int ll_remember_index(struct sockaddr_nl *who, struct nlmsghdr *n,
			     void *arg)
{
	struct idxmap **map = (struct idxmap **) arg;
	struct ifinfomsg *ifi = NLMSG_DATA(n);
	struct idxmap *im;
	struct rtattr *tb[IFLA_MAX + 1];

	if (map == NULL)
		map = &idxmap;

	if (n->nlmsg_type != RTM_NEWLINK)
		return 0;

	if (n->nlmsg_len < NLMSG_LENGTH(sizeof(ifi)))
		return -1;

	memset(tb, 0, sizeof(tb));
	parse_rtattr(tb, IFLA_MAX, IFLA_RTA(ifi), IFLA_PAYLOAD(n));
	if (tb[IFLA_IFNAME] == NULL)
		return 0;

	im = malloc(sizeof(*im));
	if (im == NULL)
		return -1;

	im->next = *map;
	im->index = ifi->ifi_index;
	dynamics_strlcpy(im->name, RTA_DATA(tb[IFLA_IFNAME]),
			 sizeof(im->name));
	*map = im;
	return 0;
}


static int ll_name_to_index(const char *name)
{
	struct idxmap *im;

	if (name == NULL)
		return 0;
	for (im = idxmap; im != NULL; im = im->next)
		if (strcmp(im->name, name) == 0)
			return im->index;
	return 0;
}


static char *ll_index_to_name(int idx)
{
	struct idxmap *im;
	static char nbuf[64];

	if (idx == 0)
		return "*";
	for (im = idxmap; im != NULL; im = im->next)
		if (im->index == idx)
			return im->name;
	snprintf(nbuf, sizeof(nbuf), "if%d", idx);
	return nbuf;
}


static int ll_init_map(struct rtnl_handle *rth)
{
	if (rtnl_wilddump_request(rth, AF_UNSPEC, RTM_GETLINK) < 0) {
		DEBUG(DEBUG_FLAG, "ll_init_map: cannot send dump request\n");
		return -1;
	}

	assert(idxmap == NULL);
	if (rtnl_dump_filter(rth, ll_remember_index, &idxmap) < 0) {
		DEBUG(DEBUG_FLAG, "ll_init_map: dump terminated\n");
		return -1;
	}
	return 0;
}


static void ll_free_map(void)
{
	struct idxmap *idx, *prev;

	idx = idxmap;
	while (idx != NULL) {
		prev = idx;
		idx = idx->next;
		free(prev);
	}
	idxmap = NULL;
}


/* Helper function for dyn_ip_addr_* functions */
static int dyn_ip_addr_ops(const char *dev, struct in_addr addr, int type)
{
	struct rtnl_handle rth;
	struct {
		struct nlmsghdr n;
		struct ifaddrmsg ifa;
		char buf[256];
	} req;

	memset(&req, 0, sizeof(req));

	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.n.nlmsg_type = type;
	req.ifa.ifa_family = AF_INET;
	req.ifa.ifa_prefixlen = 32; /* bits */

	addattr_l(&req.n, sizeof(req), IFA_LOCAL, (void *) &addr,
		  sizeof(struct in_addr));

	if (rtnl_open(&rth, 0) != 0)
		return -1;

	if (ll_init_map(&rth) != 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_addr_ops: ll_init_map() failed\n");
		ll_free_map();
		close(rth.fd);
		return -1;
	}

 	if ((req.ifa.ifa_index = ll_name_to_index(dev)) == 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_addr_ops: interface(%s) not found\n",
		      dev);
		ll_free_map();
		close(rth.fd);
		return -1;
	}
	ll_free_map();

	if (rtnl_talk(&rth, &req.n, NULL) < 0) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_addr_ops: unknown interface %s\n", dev);
		close(rth.fd);
		return -1;
	}

	close(rth.fd);
	return 0;
}


/**
 * dyn_ip_addr_add:
 * @dev: network device name
 * @addr: IP address
 * 
 * Add IP address @addr to device @dev.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_addr_add(const char *dev, struct in_addr addr)
{
	return dyn_ip_addr_ops(dev, addr, RTM_NEWADDR);
}


/**
 * dyn_ip_addr_del:
 * @dev: network device name
 * @addr: IP address
 * 
 * Remove IP address @addr from device @dev.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_addr_del(const char *dev, struct in_addr addr)
{
	return dyn_ip_addr_ops(dev, addr, RTM_DELADDR);
}


/**
 * rtnl_modify_route:
 * @cmd: 
 * @flags: 
 * @type: 
 * @dev: 
 * @table: 
 * @dst: 
 * @bits: 
 * @via: 
 * @prefsrc:
 * 
 * Handle route modification (add/del).
 * If table == -1 then table is not set (use RT_TABLE_MAIN).
 * If dst != NULL then dst contains destinations address for the route.
 * If dst == NULL then dst is ignored and default route is added.
 * If via != NULL then via contains the gateway address.
 * 
 * Returns: 0 on success or -1 on failure
 */
static int rtnl_modify_route(int cmd, int flags, int type, const char *dev,
			     int table, const struct in_addr *dst, int bits,
			     const struct in_addr *via,
			     const struct in_addr *prefsrc)
{
	struct rtnl_handle rth;
	struct {
		struct nlmsghdr n;
		struct rtmsg r;
		char buf[1024];
	} req;
	int idx;

	memset(&req, 0, sizeof(req));

	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = flags;
	req.n.nlmsg_type = cmd;
	req.r.rtm_table = RT_TABLE_MAIN;
	if (cmd == RTM_DELROUTE) {
		req.r.rtm_protocol = RTPROT_UNSPEC;
		req.r.rtm_scope = RT_SCOPE_NOWHERE;
	} else {
		req.r.rtm_protocol = RTPROT_BOOT;
		req.r.rtm_scope = RT_SCOPE_LINK;
	}
	req.r.rtm_type = type;
	req.r.rtm_family = AF_INET;

	if (dst != NULL) {
		addattr_l(&req.n, sizeof(req), RTA_DST, (void *) dst, 4);
		req.r.rtm_dst_len = bits;
	}

	if (table != -1) {
		if (table < 0 || table > 255) {
			DEBUG(DEBUG_FLAG,
			      "rtnl_modify_route: invalid table id %i\n",
			      table);
			return -1;
		}
		req.r.rtm_table = table;
	}

	if (via != NULL) {
		addattr_l(&req.n, sizeof(req), RTA_GATEWAY, (void *) via, 4);
		req.r.rtm_scope = RT_SCOPE_UNIVERSE;
	}

	if (prefsrc != NULL) {
		addattr_l(&req.n, sizeof(req), RTA_PREFSRC, (void *) prefsrc,
			  4);
	}

	if (rtnl_open(&rth, 0) != 0) {
		DEBUG(DEBUG_FLAG, "rtnl_modify_route: rtnl_open failed\n");
		return -1;
	}

	if (dev != NULL) {
		if (ll_init_map(&rth) != 0) {
			ll_free_map();
			close(rth.fd);
			DEBUG(DEBUG_FLAG,
			      "rtnl_modify_route: ll_init_map failed\n");
			return -1;
		}

		if ((idx = ll_name_to_index((char*)dev)) == 0) {
			DEBUG(DEBUG_FLAG,
			      "rtnl_modify_route: unknown interface %s\n",
			      dev);
			ll_free_map();
			close(rth.fd);
			return -1;
		}
		addattr32(&req.n, sizeof(req), RTA_OIF, idx);
		ll_free_map();
	}

	if (rtnl_talk(&rth, &req.n, NULL) < 0) {
		close(rth.fd);
		DEBUG(DEBUG_FLAG, "rtnl_modify_route: rtnl_talk failed\n");
		return -1;
	}

	close(rth.fd);
	return 0;
}


/**
 * dyn_ip_route_add_table:
 * @dev: network device name
 * @table: routing table number
 * @dst: destination IP address
 * 
 * Add a route in table @table through to the host address @dst
 * through network device @dev.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_route_add_table(const char *dev, int table, 
			   const struct in_addr *dst)
{
	return rtnl_modify_route(RTM_NEWROUTE, NLM_F_REQUEST | NLM_F_CREATE,
				 RTN_UNICAST, dev, table, dst, 32, NULL, NULL);
}


/**
 * dyn_ip_route_del_table:
 * @dev: network device name
 * @table: routing table number
 * @dst: destination IP address
 * 
 * Delete the route to the address @dst through device @dev
 * from table @table.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_route_del_table(const char *dev, int table, 
			   const struct in_addr *dst)
{
	return rtnl_modify_route(RTM_DELROUTE, NLM_F_REQUEST,
				 RTN_UNSPEC, dev, table, dst, 32, NULL, NULL);
}


/**
 * dyn_ip_route_add_net:
 * @dev: network device name
 * @dst: destination network address
 * @plen: prefix length of the network address
 * 
 * Add a route to network @dst/@plen through @dev.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_route_add_net(const char *dev, const struct in_addr *dst, int plen)
{
	return rtnl_modify_route(RTM_NEWROUTE, NLM_F_REQUEST | NLM_F_CREATE,
				 RTN_UNICAST, dev, -1, dst, plen, NULL, NULL);
}


/**
 * dyn_ip_route_replace_net:
 * @dev: network device name
 * @dst: destination network address
 * @plen: prefix length of the network address
 * 
 * Replace route to network @dst/@plen to go through @dev.
 * 
 * Returns:  0 on success or -1 on failure
 */
int dyn_ip_route_replace_net(const char *dev, const struct in_addr *dst,
			     int plen)
{
	return rtnl_modify_route(RTM_NEWROUTE, NLM_F_REQUEST | NLM_F_CREATE |
				 NLM_F_REPLACE, RTN_UNICAST, dev, -1, dst,
				 plen, NULL, NULL);
}


/**
 * dyn_ip_route_del_net:
 * @dev: network device name
 * @dst: destination network address
 * @plen: prefix length of the network address
 *
 * Delete a route to network @dst/@plen through @dev.
 *
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_route_del_net(const char *dev, const struct in_addr *dst, int plen)
{
	return rtnl_modify_route(RTM_DELROUTE, NLM_F_REQUEST,
				 RTN_UNSPEC, dev, -1, dst, plen, NULL, NULL);
}


/**
 * dyn_ip_route_add_blackhole:
 * @table: routing table number
 *
 * Add a blackhole default routing entry to the given routing @table.
 *
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_route_add_blackhole(int table)
{
	return rtnl_modify_route(RTM_NEWROUTE, NLM_F_REQUEST | NLM_F_CREATE,
				 RTN_BLACKHOLE, NULL, table, NULL, 32, NULL,
				 NULL);
}


/**
 * dyn_ip_route_del_blackhole:
 * @table: routing table number
 *
 * Delete the blackhole default routing entry from the given routing @table.
 *
 * Returns: 0 on sucess or -1 on failure
 */
int dyn_ip_route_del_blackhole(int table)
{
	return rtnl_modify_route(RTM_DELROUTE, NLM_F_REQUEST, RTN_BLACKHOLE,
				 NULL, table, NULL, 32, NULL, NULL);
}


/* generic routing rule help function for dyn_ip_rule_add_table() and
 * dyn_ip_rule_del_table() */
static int change_iprule(int cmd, int type, struct in_addr *src,
			 struct in_addr *dst, int table,
			 const char *device, int prio)
{
	struct rtnl_handle rth;
	struct {
		struct nlmsghdr n;
		struct rtmsg r;
		char buf[1024];
	} req;

	memset(&req, 0, sizeof(req));

	req.n.nlmsg_type = cmd;
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.r.rtm_family = AF_INET;
	req.r.rtm_protocol = RTPROT_BOOT;
	req.r.rtm_scope = RT_SCOPE_UNIVERSE;
	req.r.rtm_table = 0;
	req.r.rtm_type = type;

	if (cmd == RTM_NEWRULE) {
		req.n.nlmsg_flags |= NLM_F_CREATE | NLM_F_EXCL;
	}

	if (src != NULL) {
		req.r.rtm_src_len = 32; /* bits */
		addattr_l(&req.n, sizeof(req), RTA_SRC, (void *) src,
			  sizeof(struct in_addr));
	}

	if (dst != NULL) {
		req.r.rtm_dst_len = 32; /* bits */
		addattr_l(&req.n, sizeof(req), RTA_DST, (void *) dst,
			  sizeof(struct in_addr));
	}

	if (table < 0 || table > 255) {
		DEBUG(DEBUG_FLAG,
		      "change_iprule: invalid table id %i\n", table);
		return -1;
	}
	req.r.rtm_table = table;

	if (prio >= 0) {
		addattr32(&req.n, sizeof(req), RTA_PRIORITY, prio);
	}

	if (device != NULL) {
		/* Use also the incoming device to distinct packages */
		addattr_l(&req.n, sizeof(req), RTA_IIF, device,
			  strlen(device) + 1);
	}

	if (rtnl_open(&rth, 0) != 0)
		return -1;

	if (rtnl_talk(&rth, &req.n, NULL) < 0) {
		DEBUG(DEBUG_FLAG, "change_iprule: rtnl_talk failed\n");
		close(rth.fd);
		return -1;
	}

	close(rth.fd);
	return 0;
}



/* generic routing rule help function for dyn_ip_rule_add_table() and
 * dyn_ip_rule_del_table() */
static int get_iprule(int cmd, int type, struct in_addr *src,
			 struct in_addr *dst, const char *device, int prio)
{
	struct rtnl_handle rth;
	struct {
		struct nlmsghdr n;
		struct rtmsg r;
		char buf[1024];
	} req;

	memset(&req, 0, sizeof(req));

	req.n.nlmsg_type = cmd;
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.r.rtm_family = AF_INET;
	req.r.rtm_protocol = RTPROT_BOOT;
	req.r.rtm_scope = RT_SCOPE_UNIVERSE;
	req.r.rtm_table = 0;
	req.r.rtm_type = type;

	if (dst != NULL) {
		req.r.rtm_dst_len = 32; /* bits */
		addattr_l(&req.n, sizeof(req), RTA_DST, (void *) dst,
			  sizeof(struct in_addr));
	}

	req.r.rtm_table = 0;

	if (prio >= 0) {
		addattr32(&req.n, sizeof(req), RTA_PRIORITY, prio);
	}

	if (device != NULL) {
		/* Use also the incoming device to distinct packages */
		addattr_l(&req.n, sizeof(req), RTA_IIF, device,
			  strlen(device) + 1);
	}

	if (rtnl_open(&rth, 0) != 0)
		return -1;

	if (rtnl_talk(&rth, &req.n, NULL) < 0) {
		DEBUG(DEBUG_FLAG, "change_iprule: rtnl_talk failed\n");
		close(rth.fd);
		return -1;
	}

	close(rth.fd);

	return req.r.rtm_table;
}


/**
 * dyn_ip_rule_add_table:
 * @src: source IP address (from), or %NULL if not set
 * @dst: destination IP address (to), or %NULL if not set
 * @table: routing table number
 * @device: incoming network device name (iif), or %NULL if not set
 *
 * Add a routing rule for packets from @src to @dst using iif @device to
 * @table.
 *
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_rule_add_table(struct in_addr *src, struct in_addr *dst, int table,
			  const char *device)
{
	DEBUG(DEBUG_FLAG, "\tip rule add from %s to %s table %d iif %s\n",
	      src != NULL ? inet_ntoa(*src) : "N/A",
	      dst != NULL ? inet_ntoa(*dst) : "N/A",
	      table, device != NULL ? device : "N/A");
	return change_iprule(RTM_NEWRULE, RTN_UNICAST, src, dst, table, device,
			     -1);
}


/**
 * dyn_ip_rule_del_table:
 * @src: source IP address (from), or %NULL if not set
 * @dst: destination IP address (to), or %NULL if not set
 * @table: routing table number
 * @device: incoming network device name (iif), or %NULL if not set
 *
 * Delete the routing rule for packets from @src to @dst using iif @device to
 * @table.
 *
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_rule_del_table(struct in_addr *src, struct in_addr *dst, int table,
			  const char *device)
{
	DEBUG(DEBUG_FLAG, "\tip rule del from %s to %s table %d iif %s\n",
	      src != NULL ? inet_ntoa(*src) : "N/A",
	      dst != NULL ? inet_ntoa(*dst) : "N/A",
	      table, device != NULL ? device : "N/A");
	return change_iprule(RTM_DELRULE, RTN_UNICAST, src, dst, table, device,
			     -1);
}


/**
 * dyn_ip_rule_del_table:
 * @src: source IP address (from), or %NULL if not set
 * @dst: destination IP address (to), or %NULL if not set
 * @table: routing table number
 * @device: incoming network device name (iif), or %NULL if not set
 *
 * Delete the routing rule for packets from @src to @dst using iif @device to
 * @table.
 *
 * Returns: 0 on success or -1 on failure
 */
int ip_rule_get_table(struct in_addr *src, struct in_addr *dst,
			  const char *device)
{
	DEBUG(DEBUG_FLAG, "\tip rule del from %s to %s iif %s\n",
	      src != NULL ? inet_ntoa(*src) : "N/A",
	      dst != NULL ? inet_ntoa(*dst) : "N/A",
	      device != NULL ? device : "N/A");
	return get_iprule(RTM_GETRULE, RTN_UNICAST, src, dst, device,
			     -1);
}


/**
 * dyn_ip_route_get:
 * @dst: destination address
 * @dev: memory buffer for the interface name
 * @max_len: maximum length of the buffer (buffer will have to have space 
 * for the name including null termination)
 *
 * Find out the correct interface from the destination address
 *
 * Returns: 0 on success, -1 if failed, -2 if the buffer for the interface 
 * name is too small
 */
int dyn_ip_route_get(struct in_addr dst, char *dev, int max_len)
{
	struct rtnl_handle rth;
	struct {
		struct nlmsghdr n;
		struct rtmsg r;
		char buf[1024];
	} req;
	struct rtattr * tb[RTA_MAX+1];
	int len;
	char *tmp;

	memset(&req, 0, sizeof(req));
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.n.nlmsg_type = RTM_GETROUTE;
	req.r.rtm_table = 0;
	req.r.rtm_protocol = 0;
	req.r.rtm_scope = 0;
	req.r.rtm_type = 0;
	req.r.rtm_src_len = 0;
	req.r.rtm_tos = 0;

	req.r.rtm_family = AF_INET;
	req.r.rtm_dst_len = 32; /* bits */
	addattr_l(&req.n, sizeof(req), RTA_DST, (void*) &dst,
		  sizeof(struct in_addr));

	if (rtnl_open(&rth, 0) != 0)
		return -1;

	if (ll_init_map(&rth) != 0) {
		ll_free_map();
		close(rth.fd);
		return -1;
	}

	if (rtnl_talk(&rth, &req.n, &req.n) < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_get(%s): rtnl_talk failed\n",
		      inet_ntoa(dst));
		ll_free_map();
		close(rth.fd);
		return -1;
	}
	close(rth.fd);

	if (req.n.nlmsg_type != RTM_NEWROUTE) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_get: not a route\n");
		ll_free_map();
		return -1;
	}

	len = req.n.nlmsg_len - NLMSG_LENGTH(sizeof(req.r));
	if (len < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_get: invalid length\n");
		ll_free_map();
		return -1;
	}

	memset(tb, 0, sizeof(tb));
	parse_rtattr(tb, RTA_MAX, RTM_RTA(&req.r), len);

	if (tb[RTA_OIF] == 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_get: no RTA_OIF?\n");
		ll_free_map();
		return -1;
	}

	tmp = ll_index_to_name(*(int *) RTA_DATA(tb[RTA_OIF]));
	if (tmp == NULL) {
		ll_free_map();
		DEBUG(DEBUG_FLAG, "ll_index_to_name failed\n");
		return -1;
	}
	if (strlen(tmp) >= max_len) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_route_get: device name does not fit "
		      "to the given buffer\n");
		ll_free_map();
		return -2;
	}
	dynamics_strlcpy(dev, tmp, max_len);
	ll_free_map();

	return 0;
}


/**
 * dyn_ip_route_to_host:
 * @dst: destination IP address
 * 
 * Get current routing entry for a given host @dst and add a host route with
 * that information.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_route_to_host(struct in_addr dst)
{
	struct rtnl_handle rth;
	struct {
		struct nlmsghdr n;
		struct rtmsg r;
		char buf[1024];
	} req;
	struct rtattr *tb[RTA_MAX+1];
	int len;

	memset(&req, 0, sizeof(req));
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.n.nlmsg_type = RTM_GETROUTE;
	req.r.rtm_table = 0;
	req.r.rtm_protocol = 0;
	req.r.rtm_scope = 0;
	req.r.rtm_type = 0;
	req.r.rtm_src_len = 0;
	req.r.rtm_tos = 0;

	req.r.rtm_family = AF_INET;
	req.r.rtm_dst_len = 32; /* bits */
	addattr_l(&req.n, sizeof(req), RTA_DST, (void*) &dst,
		  sizeof(struct in_addr));

	if (rtnl_open(&rth, 0) != 0)
		return -1;

	if (rtnl_talk(&rth, &req.n, &req.n) < 0) {
		close(rth.fd);
		DEBUG(DEBUG_FLAG, "dyn_ip_route_to_host: rtnl_talk (get) "
		      "failed\n");
		return -1;
	}

	req.n.nlmsg_type = RTM_NEWROUTE;
	req.n.nlmsg_flags = NLM_F_REQUEST | NLM_F_CREATE | NLM_F_REPLACE;

	/* make sure this is a host route and remove cache information */
	req.r.rtm_dst_len = 32;
	req.r.rtm_flags &= ~RTM_F_CLONED;
	req.r.rtm_protocol = RTPROT_BOOT;
	req.r.rtm_type = RTN_UNICAST;

	memset(tb, 0, sizeof(tb));
	len = req.n.nlmsg_len - NLMSG_LENGTH(sizeof(req.r));
	if (len < 0) {
		close(rth.fd);
		DEBUG(DEBUG_FLAG, "dyn_ip_route_to_host: invalid len\n");
		return -1;
	}
	parse_rtattr(tb, RTA_MAX, RTM_RTA(&req.r), len);
	if (tb[RTA_METRICS]) {
		int i;
		unsigned mxlock = 0;
		struct rtattr *mxrta[RTAX_MAX+1];

		memset(mxrta, 0, sizeof(mxrta));

		parse_rtattr(mxrta, RTAX_MAX, RTA_DATA(tb[RTA_METRICS]),
			    RTA_PAYLOAD(tb[RTA_METRICS]));
		if (mxrta[RTAX_LOCK])
			mxlock = *(unsigned*)RTA_DATA(mxrta[RTAX_LOCK]);

		for (i=2; i<=RTAX_MAX; i++) {
			if (mxrta[i] != NULL)
				*(unsigned*)RTA_DATA(mxrta[i]) = 0;
		}

	}

	if (rtnl_talk(&rth, &req.n, NULL) < 0) {
		close(rth.fd);
		DEBUG(DEBUG_FLAG, "dyn_ip_route_to_host: rtnl_talk (set) "
		      "failed\n");
		return -1;
	}

	close(rth.fd);
	return 0;
}


/**
 * dyn_ip_get_ifaddr:
 * @dev: network interface name
 * @addr: address of the interface
 * 
 * Fill in @addr with the IP address of @dev.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_get_ifaddr(const char *dev, struct in_addr *addr)
{
	struct ifreq ifr;
	int s;
	struct sockaddr_in *saddr;

	memset(&ifr, 0, sizeof(ifr));
	assert(dev != NULL && strlen(dev) < IFNAMSIZ);
	dynamics_strlcpy(ifr.ifr_name, dev, IFNAMSIZ);

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifaddr: socket: %s\n",
		      strerror(errno));
		return -1;
	}

	if (ioctl(s, SIOCGIFADDR, &ifr) != 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifaddr(%s): ioctl: %s\n",
		      dev, strerror(errno));
		close(s);
		return -1;
	}
	close(s);
	saddr = (struct sockaddr_in *) &ifr.ifr_addr;
	if (saddr->sin_family != AF_INET) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_get_ifaddr(%s): invalid address family %i\n",
		      dev, saddr->sin_family);
		return -1;
	}
	memcpy(addr, &saddr->sin_addr, sizeof(*addr));
	return 0;
}


/**
 * dyn_ip_get_bcaddr:
 * @dev: network interface name
 * @addr: broadcast address of the interface
 * 
 * Fill in @addr with the broadcast IP address of @dev.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_get_bcaddr(const char *dev, struct in_addr *addr)
{
	struct ifreq ifr;
	int s;
	struct sockaddr_in *saddr;

	memset(&ifr, 0, sizeof(ifr));
	assert(dev != NULL && strlen(dev) < IFNAMSIZ);
	dynamics_strlcpy(ifr.ifr_name, dev, IFNAMSIZ);

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_bcaddr: socket: %s\n",
		      strerror(errno));
		return -1;
	}

	if (ioctl(s, SIOCGIFBRDADDR, &ifr) != 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_bcaddr(%s): ioctl: %s\n",
		      dev, strerror(errno));
		close(s);
		return -1;
	}
	close(s);
	saddr = (struct sockaddr_in *) &ifr.ifr_addr;
	if (saddr->sin_family != AF_INET) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_get_bcaddr(%s): invalid address family %i\n",
		      dev, saddr->sin_family);
		return -1;
	}
	memcpy(addr, &saddr->sin_addr, sizeof(*addr));
	return 0;
}


/**
 * dyn_ip_get_local_addr:
 * @remote: destination IP address
 * @local: buffer for the local IP address
 * 
 * Get the @local address of the interface to which the packet to @remote
 * address are sent.
 * 
 * Returns: 0 if successful, else -1
 */
int dyn_ip_get_local_addr(const struct in_addr *remote, struct in_addr *local)
{
	char dev[IFNAMSIZ];
	if (dyn_ip_route_get(*remote, dev, IFNAMSIZ) != 0)
		return -1;
	if (dyn_ip_get_ifaddr(dev, local) != 0)
		return -1;

	DEBUG(DEBUG_FLAG, "dyn_ip_get_local_addr: %s =>", inet_ntoa(*remote));
	DEBUG(DEBUG_FLAG, " %s (%s)\n", inet_ntoa(*local), dev);

	return 0;
}


/**
 * dyn_ip_route_replace:
 * @dst_addr: destination address
 * @dev: network device name
 * 
 * Replace the route to @dst_addr to go to @dev.
 * 
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_route_replace(struct in_addr dst_addr, const char *dev, const char *mnmac)
{
	return rtnl_modify_route(RTM_NEWROUTE, NLM_F_REQUEST | NLM_F_CREATE |
				 NLM_F_REPLACE, RTN_UNICAST,
				 dev, -1, &dst_addr, 32, NULL, NULL);
}


/**
 * dyn_ip_route_replace_table:
 * @dst_addr: destination address
 * @dev: network device name
 * @table: routing table number
 * 
 * Replace the route to @dst_addr in @table to go to @dev.
 * 
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_route_replace_table(struct in_addr dst_addr, const char *dev,
			       int table)
{
	return rtnl_modify_route(RTM_NEWROUTE, NLM_F_REQUEST | NLM_F_CREATE |
				 NLM_F_REPLACE, RTN_UNICAST,
				 dev, table, &dst_addr, 32, NULL, NULL);
}


/**
 * dyn_ip_route_add:
 * @dst_addr: destination address
 * @dev: network device name
 * 
 * Add a routing entry for @dst_addr to @dev.
 * 
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_route_add(struct in_addr dst_addr, const char *dev)
{
	return rtnl_modify_route(RTM_NEWROUTE, NLM_F_REQUEST | NLM_F_CREATE,
				 RTN_UNICAST, dev, -1, &dst_addr, 32, NULL,
				 NULL);
}


/**
 * dyn_ip_route_del:
 * @dst_addr: destination address
 * @dev: network device name
 *
 * Delete the route of @dst_addr to @dev.
 *
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_route_del(struct in_addr dst_addr, const char *dev)
{
	return rtnl_modify_route(RTM_DELROUTE, NLM_F_REQUEST, RTN_UNSPEC,
				 dev, -1, &dst_addr, 32, NULL, NULL);
}


#define MAX_ROUTE_ADDR_LEN 8
static int default_route_saved = 0;
static struct rtmsg saved_default_route;
static char saved_default_route_dev[IFNAMSIZ];
static int saved_default_route_dev_index;
static char saved_default_route_via[MAX_ROUTE_ADDR_LEN];
static int saved_default_route_via_len;
static char saved_default_route_prefsrc[MAX_ROUTE_ADDR_LEN];
static int saved_default_route_prefsrc_len;

/**
 * dyn_ip_get_saved_ifindex:
 * 
 * Return the interface index of the saved default route.
 * 
 * Returns: interface index or -1 on failure
 */
int dyn_ip_get_saved_ifindex(void)
{
	if (default_route_saved)
		return saved_default_route_dev_index;
	else
		return -1;
}


static int get_default_route(struct sockaddr_nl *who, struct nlmsghdr *n,
			     void *arg)
{
	struct in_addr *forced_gateway = (struct in_addr *) arg;
	struct rtattr *tb[RTA_MAX + 1];
	struct rtmsg *r = NLMSG_DATA(n);
	int len = n->nlmsg_len;

	if (n->nlmsg_type != RTM_NEWROUTE || r == NULL) {
		DEBUG(DEBUG_FLAG,
		      "get_default_route: not a route: %08x %08x %08x\n",
		      n->nlmsg_len, n->nlmsg_type, n->nlmsg_flags);
		return 0;
	}

	len -= NLMSG_LENGTH(sizeof(*r));
	if (len < 0) {
		DEBUG(DEBUG_FLAG, "get_default_route: wrong len %d\n", len);
		return -1;
	}

	memset(tb, 0, sizeof(tb));
	parse_rtattr(tb, RTA_MAX, RTM_RTA(r), len);

	if (tb[RTA_DST] || r->rtm_dst_len != 0 ||
	    r->rtm_table != RT_TABLE_MAIN) {
		/* not the default route */
		return 0;
	}

	if (forced_gateway != NULL) {
		saved_default_route_via_len = 4;
		memcpy(saved_default_route_via, forced_gateway, 4);
	} else if (tb[RTA_GATEWAY]) {
		if (tb[RTA_GATEWAY]->rta_len > MAX_ROUTE_ADDR_LEN) {
			saved_default_route_via_len = 0;
			DEBUG(DEBUG_FLAG,
			      "get_default_route: too long (%i) via address\n",
			      tb[RTA_GATEWAY]->rta_len);
		} else {
			saved_default_route_via_len = tb[RTA_GATEWAY]->rta_len;
			memcpy(saved_default_route_via,
			       RTA_DATA(tb[RTA_GATEWAY]),
			       saved_default_route_via_len);
		}
	} else {
		saved_default_route_via_len = 0;
	}

	if (tb[RTA_OIF]) {
		saved_default_route_dev_index = *(int *) RTA_DATA(tb[RTA_OIF]);
		memcpy(saved_default_route_dev,
		       ll_index_to_name(saved_default_route_dev_index),
		       IFNAMSIZ);
		DEBUG(DEBUG_FLAG, "get_default_route: idx=%i(%s)\n",
		      saved_default_route_dev_index, saved_default_route_dev);
	} else {
		DEBUG(DEBUG_FLAG,
		      "get_default_route: no device for default route\n");
		saved_default_route_dev[0] = '\0';
	}

	if (tb[RTA_PREFSRC]) {
		if (tb[RTA_PREFSRC]->rta_len > MAX_ROUTE_ADDR_LEN) {
			saved_default_route_prefsrc_len = 0;
			DEBUG(DEBUG_FLAG,
			      "get_default_route: too long (%i) prefsrc "
			      "address\n",
			      tb[RTA_PREFSRC]->rta_len);
		} else {
			saved_default_route_prefsrc_len =
				tb[RTA_PREFSRC]->rta_len;
			memcpy(saved_default_route_prefsrc,
			       RTA_DATA(tb[RTA_PREFSRC]),
			       saved_default_route_prefsrc_len);
		}
	}

	memcpy(&saved_default_route, r, sizeof(struct rtmsg));

	default_route_saved = 1;
	return 0;
}


/**
 * dyn_ip_route_save_default:
 * @forced_gateway: forced gateway address or %NULL if none
 *
 * Save the current default route. If @forced_gateway != %NULL, set the via
 * entry of the saved data (i.e., not of the routing table) to point to
 * @forced_gateway.
 *
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_route_save_default(struct in_addr *forced_gateway)
{
	struct rtnl_handle rth;
	int ret = 0;

	default_route_saved = 0;

	if (rtnl_open(&rth, 0) != 0)
		return -1;

        if (ll_init_map(&rth) != 0) {
		ret = -1;
        }

	if (ret == 0 &&
	    rtnl_wilddump_request(&rth, AF_INET, RTM_GETROUTE) < 0) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_route_get_default: cannot send dump request\n");
		ret = -1;
	}

	if (ret == 0 &&
	    rtnl_dump_filter(&rth, get_default_route, forced_gateway) < 0) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_route_get_default: dump terminated\n");
		ret = -1;
	}

	if (!default_route_saved && forced_gateway != NULL) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_get_default: "
		      "Using forced gateway\n");
		saved_default_route_via_len = 4;
		memcpy(saved_default_route_via, forced_gateway, 4);	

		if (idxmap != NULL) {
			memcpy(saved_default_route_dev, idxmap->name,
			       IFNAMSIZ);
			saved_default_route_dev_index = idxmap->index;
		} else {
			memset(saved_default_route_dev, 0, IFNAMSIZ);
			saved_default_route_dev_index = 0;
		}

		memset(&saved_default_route, 0, sizeof(saved_default_route));
		saved_default_route.rtm_table = RT_TABLE_MAIN;
		saved_default_route.rtm_protocol = RTPROT_BOOT;
		saved_default_route.rtm_type = RTN_UNICAST;
		saved_default_route.rtm_family = AF_INET;
		saved_default_route.rtm_scope = RT_SCOPE_UNIVERSE;

		default_route_saved = 1;
	}

        ll_free_map();
	close(rth.fd);
	return ret;
}


/**
 * dyn_ip_route_restore_default:
 * @force_dev: force given device for the new default route
 *
 * Restore the default route saved with dyn_ip_route_save_default().
 *
 * Returns: 0 on success or -1 on failure.
 */
int dyn_ip_route_restore_default(const char *force_dev)
{
	struct rtnl_handle rth;
	struct {
		struct nlmsghdr n;
		struct rtmsg r;
		char buf[1024];
	} req;
	int idx;
	const char *dev;

	if (!default_route_saved) {
		/* FIX: remove default route? */
		return 0;
	}

	if (force_dev != NULL)
		dev = force_dev;
	else
		dev = saved_default_route_dev;

	memset(&req, 0, sizeof(req));
	memcpy(&req.r, &saved_default_route, sizeof(struct rtmsg));

	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST | NLM_F_CREATE | NLM_F_REPLACE;
	req.n.nlmsg_type = RTM_NEWROUTE;

	if (saved_default_route_via_len > 0) {
		DEBUG(DEBUG_FLAG, 
		      "dyn_ip_route_restore_default: via %s\n",
		      inet_ntoa(*(struct in_addr *)saved_default_route_via));
		addattr_l(&req.n, sizeof(req), RTA_GATEWAY,
			  saved_default_route_via,
			  saved_default_route_via_len);
	}

	if (saved_default_route_prefsrc_len > 0) {
		DEBUG(DEBUG_FLAG, 
		      "dyn_ip_route_restore_default: prefsrc %s\n",
		      inet_ntoa(*(struct in_addr *)
				saved_default_route_prefsrc));
		addattr_l(&req.n, sizeof(req), RTA_PREFSRC,
			  saved_default_route_prefsrc,
			  saved_default_route_prefsrc_len);
	}

	if (rtnl_open(&rth, 0) != 0)
		return -1;

	if (ll_init_map(&rth) != 0) {
		ll_free_map();
		close(rth.fd);
		return -1;
	}

	idx = ll_name_to_index(dev);
	ll_free_map();
	DEBUG(DEBUG_FLAG,
	      "dyn_ip_route_restore_default: idx=%i(%s) (was %i)\n",
	      idx, dev, saved_default_route_dev_index);
	if (idx == 0) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_route_restore_default: unknown interface %s\n",
		      dev);
		ll_free_map();
		close(rth.fd);
		return -1;
	}
	addattr32(&req.n, sizeof(req), RTA_OIF, idx);

	if (rtnl_talk(&rth, &req.n, NULL) < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_restore_default: rtnl_talk "
		      "failed\n");
		close(rth.fd);
		return -1;
	}

	close(rth.fd);
	return 0;
}


/**
 * dyn_ip_route_replace_default:
 * @dev: new default route device
 * @via: new default route gateway address or %NULL if no gateway.
 * @prefsrc: new prefered source address or %NULL if not set.
 *
 * Replace the default route of the main routing table to @dev gateway address
 * @via.
 *
 * Returns: 0 on success or -1 on failure.
 */
int dyn_ip_route_replace_default(const char *dev, const struct in_addr *via,
				 const struct in_addr *prefsrc)
{
	return rtnl_modify_route(RTM_NEWROUTE,
				 NLM_F_REQUEST | NLM_F_CREATE | NLM_F_REPLACE,
				 RTN_UNICAST, dev, RT_TABLE_MAIN, NULL, 32,
				 via, prefsrc);
}


static int dummy_iterator(struct sockaddr_nl *who, struct nlmsghdr *n,
			     void *arg)
{
	int len = n->nlmsg_len;
	struct rtmsg *r = NLMSG_DATA(n);

	len -= NLMSG_LENGTH(sizeof(*r));
	if (len < 0) {
		DEBUG(DEBUG_FLAG, "dummy_iterator: wrong len %d\n", len);
		return -1;
	}

	return 0;
}


/**
 * dyn_ip_check_adv_routing:
 *
 * Check if kernel supports routing rules (advanced policy routing).
 *
 * Returns:
 *   -1: no support found,
 *    0: supported
 */
int dyn_ip_check_adv_routing(void)
{
	struct rtnl_handle rth;

	if (rtnl_open(&rth, 0) != 0) {
		fprintf(stderr, "Cannot open rtnetlink\n");
		return -1;
	}

	if (rtnl_wilddump_request(&rth, AF_INET, RTM_GETRULE) < 0) {
		fprintf(stderr, "Cannot send dump request\n");
		close(rth.fd);
		return -1;
	}

	if (rtnl_dump_filter(&rth, dummy_iterator, NULL) < 0) {
		DEBUG(DEBUG_FLAG,
		      "rtnl_dump terminated\n");
		close(rth.fd);
		return -1;
	}

	close(rth.fd);
	return 0;
}


/**
 * dyn_ip_check_ipip:
 *
 * Check if kernel supports IP-within-IP (ipip.o module).
 * 
 * Returns: 
 *   -1: no support found,
 *    0: supported
 */
int dyn_ip_check_ipip(void)
{
	struct ifreq ifr;
	struct ip_tunnel_parm p;
	int s;

	memset(&p, 0, sizeof(p));
	p.iph.version = 4;
	p.iph.ihl = 5;
	p.iph.frag_off = htons(IP_DF);
	dynamics_strlcpy(ifr.ifr_name, "tunl0", sizeof(ifr.ifr_name));
	ifr.ifr_ifru.ifru_data = (void*) &p;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		return -1;
	}

	if (ioctl(s, SIOCGETTUNNEL, &ifr) != 0) {
		perror("ioctl");
		close(s);
		return -1;
	}
	close(s);
	return 0;
}


/**
 * dyn_ip_check_gre:
 *
 * Check if kernel supports GRE (ip_gre.o module).
 *
 * Returns:
 *   -1: no support found,
 *    0: supported
 */
int dyn_ip_check_gre(void)
{
	struct ifreq ifr;
	struct ip_tunnel_parm p;
	int s;

	memset(&p, 0, sizeof(p));
	p.iph.version = 4;
	p.iph.ihl = 5;
	p.iph.frag_off = htons(IP_DF);
	dynamics_strlcpy(ifr.ifr_name, "gre0", sizeof(ifr.ifr_name));
	ifr.ifr_ifru.ifru_data = (void*) &p;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		return -1;
	}

	if (ioctl(s, SIOCGETTUNNEL, &ifr) != 0) {
		perror("ioctl");
		close(s);
		return -1;
	}
	close(s);
	return 0;
}


/**
 * dyn_ip_check_netlink:
 *
 * Check if kernel supports netlink socket.
 *
 * Returns:
 *   -1: no support found,
 *    0: supported
 */
int dyn_ip_check_netlink(void)
{
	int ret;
	struct rtnl_handle rth;

	/* test for CONFIG_NETLINK */
	if (rtnl_open(&rth, 0) != 0)
		return -1;

	/* test for CONFIG_RTNETLINK */
	ret = 0;
	if (ll_init_map(&rth) != 0)
		ret = 1;

	ll_free_map();
	close(rth.fd);

	return ret;
}


/**
 * dyn_ip_get_interface_map:
 *
 * Return the interface map from the kernel. The memory allocated for the map
 * must be released by the caller using dyn_ip_free_interface_map().
 *
 * Returns: interface map or %NULL on failure
 */
struct idxmap* dyn_ip_get_interface_map(void)
{
	struct rtnl_handle rth;
	struct idxmap *ret;

	if (rtnl_open(&rth, 0) != 0)
		return NULL;

	if (ll_init_map(&rth) != 0) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_get_interface_map: ll_init_map() failed\n");
		ll_free_map();
		close(rth.fd);
		return NULL;
	}

	close(rth.fd);

	ret = idxmap;
	idxmap = NULL; /* caller releases memory */
	return ret;
}


/**
 * dyn_ip_get_ifindex:
 * @dev: selected device name
 *
 * Get the interface index of the selected device @dev.
 *
 * Returns: interface index or -1 on failure
 */
int dyn_ip_get_ifindex(const char *dev)
{
	int s;
	struct ifreq ifr;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifindex - socket: %s\n",
		      strerror(errno));
		return -1;
	}

	memset(&ifr, 0, sizeof(ifr));
	memcpy(ifr.ifr_name, dev, IFNAMSIZ);
	if (ioctl(s, SIOCGIFINDEX, &ifr) != 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifindex(%s): ioctl: %s\n",
		      dev, strerror(errno));
		close(s);
		return -1;
	}

	close(s);
	return ifr.ifr_ifindex;
}


/**
 * dyn_ip_get_ifname:
 * @ifindex: selected interface
 * @dev: buffer for the interface name
 *
 * Get the device name @dev of the selected interface @ifindex.
 *
 * Returns: 0 on success or -1 on failure
 */
int dyn_ip_get_ifname(int ifindex, char *dev)
{
	struct rtnl_handle rth;
	struct idxmap *tmp;

	if (rtnl_open(&rth, 0) != 0)
		return -1;

	if (ll_init_map(&rth) != 0) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_get_ifname: ll_init_map() failed\n");
		ll_free_map();
		close(rth.fd);
		return -1;
	}

	close(rth.fd);

	tmp = idxmap;
	while (tmp != NULL) {
		if (tmp->index == ifindex) {
			memcpy(dev, tmp->name, IFNAMSIZ);
			break;
		}
		tmp = tmp->next;
	}

	ll_free_map();

	if (tmp == NULL) return -1;
	return 0;
}


/**
 * dyn_ip_route_cache_flush:
 *
 * Flush the routing cache.
 *
 * Returns: 0 on success or -1 on error
 */
int dyn_ip_route_cache_flush(void)
{
	int l, fd;
	char *buf = "0";

	fd = open("/proc/sys/net/ipv4/route/flush", O_WRONLY);
	if (fd < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_flush - cannot open "
		      "/proc/sys/net/ipv4/route/flush\n");
		return -1;
	}

	l = write(fd, buf, strlen(buf));
	if (l < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_flush - %s\n",
		      strerror(errno));
		close(fd);
		return -1;
	}
	close(fd);

	if (l < strlen(buf)) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_flush - only %i/%i wrote\n",
		      l, strlen(buf));
		return -1;
	}

	return 0;
}


/**
 * dyn_ip_get_ifflags:
 * @dev: interface
 *
 * Get interface flags for @dev.
 *
 * Returns: interface flags or -1 on failure
 */
int dyn_ip_get_ifflags(const char *dev)
{
	struct ifreq ifr;
	int s;

	assert(dev != NULL);
	assert(strlen(dev) < IFNAMSIZ);
	dynamics_strlcpy(ifr.ifr_name, dev, IFNAMSIZ);

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifflags: socket: %s\n",
		      strerror(errno));
		return -1;
	}

	if (ioctl(s, SIOCGIFFLAGS, &ifr) != 0) {
		DEBUG(DEBUG_FLAG,
		      "dyn_ip_get_ifflags: ioctl SIOCGIFFLAGS: %s\n",
		      strerror(errno));
		close(s);
		return -1;
	}
	close(s);
	return ifr.ifr_flags;
}



struct existance_data {
	const struct in_addr *addr;
	int prefix_len;
	int found;
};

static int check_existance(struct sockaddr_nl *who, struct nlmsghdr *n,
			   void *arg)
{
	struct rtattr *tb[RTA_MAX + 1];
	struct rtmsg *r = NLMSG_DATA(n);
	int len = n->nlmsg_len;
	struct existance_data *edata = (struct existance_data *) arg;
	struct in_addr addr;

	if (n->nlmsg_type != RTM_NEWROUTE || r == NULL) {
		DEBUG(DEBUG_FLAG,
		      "check_existance: not a route: %08x %08x %08x\n",
		      n->nlmsg_len, n->nlmsg_type, n->nlmsg_flags);
		return 0;
	}

	len -= NLMSG_LENGTH(sizeof(*r));
	if (len < 0) {
		DEBUG(DEBUG_FLAG, "check_existance: wrong len %d\n", len);
		return -1;
	}

	memset(tb, 0, sizeof(tb));
	parse_rtattr(tb, RTA_MAX, RTM_RTA(r), len);

	if (tb[RTA_DST] == NULL || r->rtm_dst_len != edata->prefix_len ||
	    r->rtm_table != RT_TABLE_MAIN) return 0;

	memcpy(&addr, RTA_DATA(tb[RTA_DST]), 4);
	if (addr.s_addr != edata->addr->s_addr) return 0;

	/* entry matches */
	edata->found = 1;
	return -2;
}

/**
 * dyn_ip_route_exists:
 * @addr: route destination address (host/network)
 * @prefix_len: prefix length for the network route
 *
 * Check whether a given route (@addr/@prefix_len) exists in the main routing
 * table.
 *
 * Returns: -1 on error, 0 if not found, 1 if found
 */
int dyn_ip_route_exists(const struct in_addr *addr, int prefix_len)
{
	struct rtnl_handle rth;
	struct existance_data edata;

	if (rtnl_open(&rth, 0) != 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_exists - rtnl_open failed\n");
		return -1;
	}

	if (rtnl_wilddump_request(&rth, AF_UNSPEC, RTM_GETROUTE) < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_exists - rtnl_wilddump_request"
		      "failed\n");
		close(rth.fd);
		return -1;
	}

	edata.addr = addr;
	edata.prefix_len = prefix_len;
	edata.found = 0;
	if (rtnl_dump_filter(&rth, check_existance, &edata) == -1) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_exists - rtnl_dump_filter "
		      "failed\n");
	}
	close(rth.fd);

	return edata.found;
}


/**
 * dyn_ip_route_delete_net_routes:
 * @addr: route destination address (host/network)
 * @prefix_len: prefix length for the network route
 *
 * Delete all routing entries to given network from main routing table.
 *
 * Returns: number of deleted routing entries or -1 on error.
 */
int dyn_ip_route_delete_net_routes(const struct in_addr *addr, int prefix_len)
{
	int count = 0;
	while (dyn_ip_route_exists(addr, prefix_len) > 0) {
		count++;
		if (dyn_ip_route_del_net(NULL, addr, prefix_len) != 0) {
			DEBUG(DEBUG_FLAG, "dyn_ip_route_delete_net_routes: "
			      "delete failed for (%s/%i)\n",
			      inet_ntoa(*addr), prefix_len);
			return -1;
		}
	}
	return count;
}


/**
 * dyn_ip_monitor_open:
 *
 * Open netlink monitor socket
 *
 * Returns: file descriptor for the netlink socket, or -1 on error
 */
int dyn_ip_monitor_open(void)
{
	struct rtnl_handle rth;

	if (rtnl_open(&rth, RTMGRP_LINK | RTMGRP_IPV4_IFADDR |
		RTMGRP_IPV4_ROUTE) != 0)
		return -1;
	return rth.fd;
}


static int handle_route_add(const struct current_route_info *cur,
			    struct nlmsghdr *h)
{
	struct in_addr tmp;
	struct rtattr *tb[RTA_MAX + 1];
	struct rtmsg *r;
	int len = h->nlmsg_len - NLMSG_LENGTH(sizeof(*r));
	int ret;
	char dev[IFNAMSIZ];

	if (cur == NULL || !cur->known) return -1;
	if (len < 0) {
		DEBUG(DEBUG_FLAG, "handle_route_add: len < 0\n");
		return -1;
	}

	r = NLMSG_DATA(h);
	if (r->rtm_family != AF_INET) {
		DEBUG(DEBUG_FLAG, "handle_route_add: unknown family %i\n",
		      r->rtm_family);
		return 0;
	}

	ret = 0;
	memset(tb, 0, sizeof(tb));
	parse_rtattr(tb, RTA_MAX, RTM_RTA(r), len);
	if (tb[RTA_DST] == NULL) {
		int ok = 1;
		/* handle default route */
		if (!tb[RTA_OIF])
			return 0;

		if (*(int*)RTA_DATA(tb[RTA_OIF]) != cur->ifindex)
			ok = 0;

		if (tb[RTA_GATEWAY] == NULL && cur->via.s_addr != 0)
			ok = 0;

		if (tb[RTA_GATEWAY] && cur->via.s_addr !=
		    *(int*)RTA_DATA(tb[RTA_GATEWAY]))
			ok = 0;

		if (tb[RTA_PREFSRC] == NULL && cur->prefsrc.s_addr != 0)
			ok = 0;

		if (tb[RTA_PREFSRC] && cur->prefsrc.s_addr !=
		    *(int*)RTA_DATA(tb[RTA_PREFSRC]))
			ok = 0;

		if (ok)
			return 0;

		DEBUG(DEBUG_FLAG,
		      "Enforcing default route to dev %s (index=%i)",
		      cur->ifname, cur->ifindex);
		if (cur->via.s_addr != 0)
			DEBUG(DEBUG_FLAG, " via %s", inet_ntoa(cur->via));
		if (cur->prefsrc.s_addr != 0)
			DEBUG(DEBUG_FLAG, " prefsrc %s",
			      inet_ntoa(cur->prefsrc));
		DEBUG(DEBUG_FLAG, "\n");
		ret = dyn_ip_route_replace_default(cur->ifname,
						   cur->via.s_addr == 0 ?
						   NULL : &cur->via,
						   cur->prefsrc.s_addr == 0 ?
						   NULL : &cur->prefsrc);
	} else {
		int ifindex;
		/* handle net/host route */
		memcpy(&tmp, RTA_DATA(tb[RTA_DST]), 4);
		if (tmp.s_addr != cur->net.s_addr ||
		    r->rtm_dst_len != cur->prefix_len)
			return 0;
		if (tb[RTA_OIF] == NULL)
			return 0;
		ifindex = *(int*)RTA_DATA(tb[RTA_OIF]);
		if (ifindex == cur->ifindex_net)
			return 0;
		if (cur->ifindex_net == -1) {
			DEBUG(DEBUG_FLAG,
			      "Enforcing no net route situation\n");
			if (dyn_ip_get_ifname(ifindex, dev) < 0)
				return -1;
			ret = dyn_ip_route_del_net(dev, &cur->net,
						   cur->prefix_len);
		} else {
			DEBUG(DEBUG_FLAG,
			      "Enforcing net route to dev %s (index=%i)\n",
			      cur->ifname_net, cur->ifindex_net);
			ret = dyn_ip_route_replace_net(cur->ifname_net,
						       &cur->net,
						       cur->prefix_len);
		}

	}
	if (ret < 0) return ret;

	return 3;
}

/**
 * dyn_ip_monitor_get:
 * @fd: file descriptor for the netlink monitor socket
 * @cur: information about the current default home network route
 *
 * Get a netlink message from the given file descriptor.
 * If @cur != NULL, then the added net routes with metric 0 will be changed
 * to use metric 1 (workaround for multiple PC Card situations).
 *
 * Returns:
 *  -1 on error,
 *   0 if not a link message,
 *   1 if a device changed (added or removed),
 *   2 if an address changed (added or remove),
 *   3 if net route metric changed
 */
int dyn_ip_monitor_get(int fd, const struct current_route_info *cur)
{
	int status;
	struct nlmsghdr *h;
	struct sockaddr_nl nladdr;
	struct iovec iov;
	char   buf[8192];
	struct msghdr msg = {
		(void *) &nladdr, sizeof(nladdr),
		&iov, 1,
		NULL, 0,
		0
	};

	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);
	status = recvmsg(fd, &msg, 0);
	if (status < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_monitor_get - recvmsg: %s\n",
		      strerror(errno));
		return -1;
	}
	if (status == 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_monitor_get - overrun\n");
		return -1;
	}

	for (h = (struct nlmsghdr*) buf; status >= sizeof(*h); ) {
		int len = h->nlmsg_len;
		int l = len - sizeof(*h);

		if (l < 0 || len > status) {
			if (msg.msg_flags & MSG_TRUNC) {
				DEBUG(DEBUG_FLAG, "dyn_ip_monitor_get: "
				      "truncated message\n");
				return -1;
			}
			DEBUG(DEBUG_FLAG, "dyn_ip_monitor_get: "
			      "malformed message: len=%d\n", len);
			return -1;
		}
		if (h->nlmsg_type == RTM_NEWLINK ||
		    h->nlmsg_type == RTM_DELLINK)
			return 1;
		if (h->nlmsg_type == RTM_NEWADDR ||
		    h->nlmsg_type == RTM_DELADDR)
			return 2;

		if (h->nlmsg_type == RTM_NEWROUTE && cur != NULL && cur->known)
			return handle_route_add(cur, h);

		status -= NLMSG_ALIGN(len);
		h = (struct nlmsghdr *) (((char *) h) + NLMSG_ALIGN(len));
	}

	return 0;
}


struct addr_iface_struct {
	int found;
	int ifindex;
	const struct in_addr *addr;
	int count;
};

static int check_iface_address(struct sockaddr_nl *who, struct nlmsghdr *n,
			       void *arg)
{
	struct addr_iface_struct *data = (struct addr_iface_struct *) arg;
	struct ifaddrmsg *ifa = NLMSG_DATA(n);
	struct rtattr *tb[IFA_MAX + 1];

	if (n->nlmsg_type != RTM_NEWADDR) {
		DEBUG(DEBUG_FLAG,
		      "check_iface_address: not RTM_NEWADDR (0x%x)\n",
		      n->nlmsg_type);
		return 0;
	}

	if (n->nlmsg_len < NLMSG_LENGTH(sizeof(ifa)))
		return -1;

        memset(tb, 0, sizeof(tb));
        parse_rtattr(tb, IFA_MAX, IFA_RTA(ifa),
		     n->nlmsg_len - NLMSG_LENGTH(sizeof(*ifa)));
	if (tb[IFA_LOCAL] == NULL)
		tb[IFA_LOCAL] = tb[IFA_ADDRESS];

	if (tb[IFA_LOCAL]) {
		int alen = RTA_PAYLOAD(tb[IFA_LOCAL]);
		if (alen != 4) {
			DEBUG(DEBUG_FLAG, "check_iface_address alen=%i != 4\n",
			      alen);
		}
		if (ifa->ifa_index == data->ifindex) {
			data->count++;
			if (memcmp(RTA_DATA(tb[IFA_LOCAL]), data->addr, alen)
			    == 0) {
				data->found = 1;
				return -2;
			}
		}
	}

	return 0;
}

/**
 * dyn_ip_iface_address:
 * @ifindex: interface index
 * @addr: address to be checked
 *
 * Check if address @addr is assigned in the interface @ifindex.
 *
 * Returns:
 * -1 on failure,
 *  0 address was not assigned,
 *  1 the given address is assigned to the interface,
 *  2 no addresses in the given interface
 */
int dyn_ip_iface_address(int ifindex, const struct in_addr *addr)
{
	struct addr_iface_struct data;
	struct rtnl_handle rth;

	if (rtnl_open(&rth, 0) != 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_iface_address: rtnl_open failed\n");
		return -1;
	}
	
	if (rtnl_wilddump_request(&rth, AF_INET, RTM_GETADDR) < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_iface_address: cannot send dump "
		      "request\n");
		close(rth.fd);
		return -1;
	}

	data.found = 0;
	data.ifindex = ifindex;
	data.addr = addr;
	data.count = 0;

	if (rtnl_dump_filter(&rth, check_iface_address, &data) == -1) {
		close(rth.fd);
		DEBUG(DEBUG_FLAG, "dyn_ip_iface_address: dump terminated\n");
		return -1;
	}

	close(rth.fd);
	if (data.count == 0 && !data.found)
		return 2;
	return data.found;
}


/**
 * dyn_ip_get_hwaddr:
 * @dev: selected interface
 * @hwaddr: data buffer for the MAC address
 * @max_len: size of @hwaddr buffer in bytes
 *
 * Get MAC address of the given interface.
 *
 * Returns: 0 on success, -1 on failure, or -2 on too small max_len.
 */
int dyn_ip_get_hwaddr(const char *dev, char *hwaddr, int max_len)
{
	int s, i;
	struct ifreq ifr;

	memcpy(ifr.ifr_name, dev, strlen(dev) + 1);
	s = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	if (s < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_hwaddr: socket failed: %s\n",
		      strerror(errno));
		return -1;
	}

	if (ioctl(s, SIOCGIFHWADDR, &ifr) < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_hwaddr: ioctl SIOGIFHWADDR "
		      "failed: %s\n", strerror(errno));
		close(s);
		return -1;
	}
	close(s);

	if (ifr.ifr_hwaddr.sa_family == ARPHRD_ETHER) {
		if (max_len < IFNAMSIZ * 3 + 1)
			return -2;
		dynamics_strlcpy(hwaddr,
			ether_hwtoa((unsigned char *) ifr.ifr_hwaddr.sa_data),
			max_len);
	} else {
		if (max_len < sizeof(ifr.ifr_hwaddr.sa_data) * 3 + 1)
			return -2;
		for (i = 0; i < sizeof(ifr.ifr_hwaddr.sa_data); i++) {
			sprintf(hwaddr + i * 3, "%02X",
				ifr.ifr_hwaddr.sa_data[i]);
			if (i + 1 == sizeof(ifr.ifr_hwaddr.sa_data))
				hwaddr[i * 3 + 2] = '\0';
			else
				hwaddr[i * 3 + 2] = ':';
		}
	}

	return 0;
}

#endif /* DYN_TARGET_LINUX */


#ifdef DYN_TARGET_WINDOWS

int dyn_ip_addr_add(const char *dev, struct in_addr addr) { return -1; }
int dyn_ip_addr_del(const char *dev, struct in_addr addr) { return -1; }
int dyn_ip_check_adv_routing(void) { return 0; }
int dyn_ip_check_ipip(void) { return 0; }
int dyn_ip_check_gre(void) { return 0; }
int dyn_ip_check_netlink(void) { return 0; }
int dyn_ip_get_hwaddr(const char *dev, char *hwaddr, int max_len)
{ return -1; }

int dyn_ip_get_ifaddr(const char *dev, struct in_addr *addr)
{
	char buf[8192];
	PMIB_IPADDRTABLE info = (PMIB_IPADDRTABLE) buf;
	DWORD res;
	ULONG blen;
	int i;
	int index = dyn_ip_get_ifindex(dev);

	if (index < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifaddr: dyn_ip_get_ifindex ==> "
		      "%i\n", index);
		return -1;
	}

	blen = sizeof(buf);
	res = GetIpAddrTable(info, &blen, 0);
	if (res != NO_ERROR) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifaddr - GetIpAddrTable"
		      " returned %li (blen=%li)\n", res, blen);
		return -1;
	}

	for (i = 0; i < info->dwNumEntries; i++) {
		if (info->table[i].dwIndex == index) {
			addr->s_addr = info->table[i].dwAddr;
			return 0;
		}
	}

	DEBUG(DEBUG_FLAG, "dyn_ip_get_ifaddr - could not found address for dev"
	      " '%s' (index=%i)\n", dev, index);
	return -1;
}

int dyn_ip_get_bcaddr(const char *dev, struct in_addr *addr) { return -1; }

int dyn_ip_get_ifflags(const char *dev) { return IFF_UP; }

int dyn_ip_get_ifindex(const char *dev)
{
	struct idxmap *map, *pos;
	int ret = -1;

	map = dyn_ip_get_interface_map();
	if (map == NULL) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifindex: could not get "
		      "interface map\n");
		return -1;
	}

	pos = map;
	while (pos != NULL) {
		/* FIX: Name is actually WCHAR (16bit characters), but at least
		 * some cases are 8bit strings.. */
		if (strncmp(dev, (char *) pos->name, IFNAMSIZ) == 0) {
			ret = pos->index;
			break;
		}
		pos = pos->next;
	}

	if (ret < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_ifindex: interface not found\n");
		if (map != NULL) {
			DEBUG(DEBUG_FLAG, "   using the index of the first "
			      "interface..\n");
			ret = map->index;
		}
	}

	dyn_ip_free_interface_map(map);

	return ret;
}

struct idxmap* dyn_ip_get_interface_map_alt(void)
{
	char buf[8192];
	PMIB_IFTABLE info = (PMIB_IFTABLE) buf;
	ULONG blen;
	DWORD res;
	struct idxmap *map = NULL, *nmap;
	int i;

	blen = sizeof(buf);
	res = GetIfTable(info, &blen, 0);
	if (res != NO_ERROR) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_interface_map_alt - "
		      "GetIfTable returned %li (blen=%li)\n", res, blen);
		if (res == ERROR_NOT_SUPPORTED)
			DEBUG(DEBUG_FLAG, "\tGetIfTable() not supported "
			      "in this OS\n");
		return NULL;
	}

	DEBUG(DEBUG_FLAG, "dwNumEntries=%li\n", info->dwNumEntries);
	for (i = 0; i < info->dwNumEntries; i++) {
		nmap = (struct idxmap *) malloc(sizeof(struct idxmap));
		if (nmap == NULL) {
			DEBUG(DEBUG_FLAG, "dyn_ip_get_interface_map_alt - "
			      "malloc failed\n");
			return map;
		}

		DEBUG(DEBUG_FLAG, "Index=%li Type=%li Descr='%s' Name='%s'\n",
		      info->table[i].dwIndex,
		      info->table[i].dwType,
		      info->table[i].bDescr,
		      info->table[i].wszName);

		if (info->table[i].dwType == MIB_IF_TYPE_LOOPBACK) {
			DEBUG(DEBUG_FLAG, "\tloopback iface - skipping\n");
			free(nmap);
			continue;
		}

		nmap->index = info->table[i].dwIndex;
		if (info->table[i].wszName[0] == 0) {
			/* WinNT does not seem to give interface names, so
			 * create a temporary name for the interface */
			snprintf(nmap->name, sizeof(nmap->name), "if#%i",
				 info->table[i].dwIndex);
		} else {
			/* FIX: Name is actually WCHAR (16bit characters), but
			 * at least some cases are 8bit strings.. */
			dynamics_strlcpy(nmap->name,
					 (char *) info->table[i].wszName,
					 IFNAMSIZ);
		}
		nmap->name[IFNAMSIZ-1] = '\0';

		nmap->next = map;
		map = nmap;
	}

	return map;
}

struct idxmap* dyn_ip_get_interface_map(void)
{
	char buf[8192];
	PIP_INTERFACE_INFO info = (PIP_INTERFACE_INFO) buf;
	struct idxmap *map = NULL, *nmap;
	DWORD res;
	ULONG blen;
	int i;

	blen = sizeof(buf);
	/* GetInterfaceInfo() works with Win98 and Win2k (not with WinNT), try
	 * GetIfEntry() if GetInterfaceInfo() failes (this should work also "
	 * with NT) */
	res = GetInterfaceInfo(info, &blen);
	if (res != NO_ERROR) {
		DEBUG(DEBUG_FLAG, "dyn_ip_get_interface_map - GetInterfaceInfo"
		      " returned %li (blen=%li)\n", res, blen);
		if (res == ERROR_NOT_SUPPORTED) {
			DEBUG(DEBUG_FLAG, "\tGetInterfaceInfo() not supported "
			      "in this OS\n");
			return dyn_ip_get_interface_map_alt();
		}
		return NULL;
	}

	for (i = 0; i < info->NumAdapters; i++) {
		nmap = (struct idxmap *) malloc(sizeof(struct idxmap));
		if (nmap == NULL) {
			DEBUG(DEBUG_FLAG, "dyn_ip_get_interface_map - malloc "
			      "failed\n");
			return map;
		}

		nmap->index = info->Adapter[i].Index;
		/* FIX: Name is actually WCHAR (16bit characters), but at least
		 * Win98 uses 8bit strings (NT and W2k should use unicode) */
		dynamics_strlcpy(nmap->name, (char *) info->Adapter[i].Name,
				 IFNAMSIZ);
		nmap->name[IFNAMSIZ-1] = '\0';

		nmap->next = map;
		map = nmap;
	}

	return map;
}

int dyn_ip_get_saved_ifindex(void) { return -1; }
int dyn_ip_iface_address(int ifindex, const struct in_addr *addr) { return 1; }
int dyn_ip_link_set_dev_up(const char *dev) { return -1; }
int dyn_ip_monitor_open(void) { return -1; }
int dyn_ip_monitor_get(int fd, const struct current_route_info *cur)
{ return -1; }

int dyn_ip_route_add(struct in_addr dst_addr, const char *dev)
{
	MIB_IPFORWARDROW route;
	DWORD res;
	int ifindex;
	struct in_addr own_addr;

	ifindex = dyn_ip_get_ifindex(dev);
	if (ifindex < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_add - dyn_ip_get_ifindex() "
		      "failed\n");
		return -1;
	}
	if (dyn_ip_get_ifaddr(dev, &own_addr) < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_add - dyn_ip_get_ifaddr() "
		      "failed\n");
		return -1;
	}

	memset(&route, 0, sizeof(route));
	route.dwForwardDest = dst_addr.s_addr;
	route.dwForwardMask = -1;
	route.dwForwardNextHop = own_addr.s_addr;
	route.dwForwardIfIndex = ifindex;
	route.dwForwardType = 3; /* The next hop is the final destination
				  * (local route). */
	route.dwForwardProto = 2; /* PROTO_IP_NETMGMT */
	route.dwForwardMetric1 = 1;
	res = CreateIpForwardEntry(&route);
	switch (res) {
	case NO_ERROR:
		return 0;
	case ERROR_INVALID_PARAMETER:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "ERROR_INVALID_PARAMETER\n");
		break;
	case ERROR_NOT_SUPPORTED:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "ERROR_NOT_SUPPORTED\n");
		break;
	default:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "Unknown error %li\n", res);
		break;
	}

	return -1;
}

int dyn_ip_route_replace(struct in_addr dst_addr, const char *dev)
{
	return dyn_ip_route_add(dst_addr, dev);
}

int dyn_ip_route_del(struct in_addr dst_addr, const char *dev)
{
	MIB_IPFORWARDROW route;
	DWORD res;
	int ifindex;
	struct in_addr own_addr;

	ifindex = dyn_ip_get_ifindex(dev);
	if (ifindex < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_del - dyn_ip_get_ifindex() "
		      "failed\n");
		return -1;
	}
	if (dyn_ip_get_ifaddr(dev, &own_addr) < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_del - dyn_ip_get_ifaddr() "
		      "failed\n");
		return -1;
	}

	memset(&route, 0, sizeof(route));
	route.dwForwardDest = dst_addr.s_addr;
	route.dwForwardMask = -1;
	route.dwForwardNextHop = own_addr.s_addr;
	route.dwForwardIfIndex = ifindex;
	route.dwForwardType = 3; /* The next hop is the final destination
				  * (local route). */
	route.dwForwardProto = 2; /* PROTO_IP_NETMGMT */
	route.dwForwardMetric1 = 1;
	res = DeleteIpForwardEntry(&route);
	if (res == NO_ERROR)
		return 0;

	DEBUG(DEBUG_FLAG, "DeleteIpForwardEntry: Unknown error %li\n", res);
	return -1;
}


int dyn_ip_route_add_net(const char *dev, const struct in_addr *dst, int plen)
{
	MIB_IPFORWARDROW route;
	DWORD res;
	int ifindex;
	struct in_addr own_addr;

	ifindex = dyn_ip_get_ifindex(dev);
	if (ifindex < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_add_net - "
		      "dyn_ip_get_ifindex() failed\n");
		return -1;
	}
	if (dyn_ip_get_ifaddr(dev, &own_addr) < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_add_net - dyn_ip_get_ifaddr() "
		      "failed\n");
		return -1;
	}

	memset(&route, 0, sizeof(route));
	route.dwForwardDest = dst->s_addr;
	route.dwForwardMask = dyn_ip_prefix_len_to_mask(plen);
	route.dwForwardNextHop = own_addr.s_addr;
	route.dwForwardIfIndex = ifindex;
	route.dwForwardType = 3; /* The next hop is the final destination
				  * (local route). */
	route.dwForwardProto = 2; /* PROTO_IP_NETMGMT */
	route.dwForwardMetric1 = 1;
	res = CreateIpForwardEntry(&route);
	switch (res) {
	case NO_ERROR:
		return 0;
	case ERROR_INVALID_PARAMETER:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "ERROR_INVALID_PARAMETER\n");
		break;
	case ERROR_NOT_SUPPORTED:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "ERROR_NOT_SUPPORTED\n");
		break;
	default:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "Unknown error %li\n", res);
		break;
	}

	return -1;
}


int dyn_ip_route_set_net(const char *dev, const struct in_addr *dst, int plen,
			 const struct in_addr *via, int metric)
{
	MIB_IPFORWARDROW route;
	DWORD res;
	int ifindex;
	struct in_addr own_addr;

	ifindex = dyn_ip_get_ifindex(dev);
	if (ifindex < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_add_net - "
		      "dyn_ip_get_ifindex() failed\n");
		return -1;
	}
	memset(&route, 0, sizeof(route));
	route.dwForwardDest = dst->s_addr;
	route.dwForwardMask = dyn_ip_prefix_len_to_mask(plen);
	route.dwForwardIfIndex = ifindex;
	if (via != NULL) {
		route.dwForwardNextHop = via->s_addr;
		route.dwForwardType = 4; /* The next hop is not the final
					  * destination (remote route). */
	} else {
		if (dyn_ip_get_ifaddr(dev, &own_addr) < 0) {
			DEBUG(DEBUG_FLAG, "dyn_ip_route_set_net - "
			      "dyn_ip_get_ifaddr() failed\n");
			return -1;
		}
		route.dwForwardNextHop = own_addr.s_addr;
		route.dwForwardType = 3; /* The next hop is the final
					  * destination (local route). */
	}
	route.dwForwardProto = 2; /* PROTO_IP_NETMGMT */
	route.dwForwardMetric1 = metric;
	res = CreateIpForwardEntry(&route);
	switch (res) {
	case NO_ERROR:
		return 0;
	case ERROR_INVALID_PARAMETER:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "ERROR_INVALID_PARAMETER\n");
		break;
	case ERROR_NOT_SUPPORTED:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "ERROR_NOT_SUPPORTED\n");
		break;
	default:
		DEBUG(DEBUG_FLAG, "CreateIpForwardEntry: "
		      "Unknown error %li\n", res);
		break;
	}

	return -1;
}


int dyn_ip_route_del_net(const char *dev, const struct in_addr *dst, int plen)
{ return -1; }

static const char *dword_addr_str(DWORD addr)
{
	struct in_addr tmp;
	tmp.s_addr = addr;
	return inet_ntoa(tmp);
}


int dyn_ip_route_exists(const struct in_addr *addr, int prefix_len)
{
	char buf[8192];
	PMIB_IPFORWARDTABLE table = (PMIB_IPFORWARDTABLE) buf;
	DWORD res;
	ULONG blen;
	int i;

	blen = sizeof(buf);
	res = GetIpForwardTable(table, &blen, 0);
	if (res != NO_ERROR) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_exists - GetIpForwardTable "
		      "returned %li (blen=%li)\n", res, blen);
		return -1;
	}

	for (i = 0; i < table->dwNumEntries; i++) {
		PMIB_IPFORWARDROW row = &table->table[i];
		if (addr->s_addr == row->dwForwardDest &&
		    prefix_len ==
		    dyn_ip_mask_to_prefix_len(row->dwForwardMask))
			return 1;
	}

	return 0;
}


static void dyn_ip_route_dump(PMIB_IPFORWARDROW row)
{
	DEBUG(DEBUG_FLAG, "MIP_IPFORWARDROW dump:\n"
	      "   dwForwardDest=%s\n",
	      dword_addr_str(row->dwForwardDest));
	DEBUG(DEBUG_FLAG, "   dwForwardMask=%s\n"
	      "   dwForwardPolicy=%lu\n",
	      dword_addr_str(row->dwForwardMask),
	      row->dwForwardPolicy);
	DEBUG(DEBUG_FLAG, "   dwForwardNextHop=%s\n"
	      "   dwForwardIfIndex=%lu\n"
	      "   dwForwardType=%lu\n"
	      "   dwForwardProto=%lu\n"
	      "   dwForwardAge=%lu\n"
	      "   dwForwardNextHopAS=%lu\n"
	      "   dwForwardMetric[1..5]=%lu %lu %lu %lu %lu\n",
	      dword_addr_str(row->dwForwardNextHop),
	      row->dwForwardIfIndex,
	      row->dwForwardType,
	      row->dwForwardProto,
	      row->dwForwardAge,
	      row->dwForwardNextHopAS,
	      row->dwForwardMetric1,
	      row->dwForwardMetric2,
	      row->dwForwardMetric3,
	      row->dwForwardMetric4,
	      row->dwForwardMetric5);
}

int dyn_ip_route_delete_net_routes(const struct in_addr *addr, int prefix_len)
{
	char buf[8192];
	PMIB_IPFORWARDTABLE table = (PMIB_IPFORWARDTABLE) buf;
	DWORD res;
	ULONG blen;
	int i, count = 0, err = 0;

	blen = sizeof(buf);
	res = GetIpForwardTable(table, &blen, 0);
	if (res != NO_ERROR) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_delete_net_routes - "
		      "GetIpForwardTable returned %li (blen=%li)\n",
		      res, blen);
		return -1;
	}

	for (i = 0; i < table->dwNumEntries; i++) {
		PMIB_IPFORWARDROW row = &table->table[i];
		if (addr->s_addr == row->dwForwardDest &&
		    prefix_len ==
		    dyn_ip_mask_to_prefix_len(row->dwForwardMask)) {
			res = DeleteIpForwardEntry(row);
			if (res != NO_ERROR) {
				dyn_ip_route_dump(row);
				DEBUG(DEBUG_FLAG,
				      "dyn_ip_route_delete_net_routes - "
				      "delete failed for route to %s/",
				      dword_addr_str(row->dwForwardDest));
				DEBUG(DEBUG_FLAG, "%s (res=%li)\n",
				      dword_addr_str(row->dwForwardMask),
				      res);
				err = 1;
			} else
				count++;
		}
	}

	if (err && count > 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_delete_net_routes - "
		      "error occurred, but at least one entry was removed\n");
		return -2;
	} else if (err)
		return -1;
	return count;
}


int dyn_ip_route_get(struct in_addr dst, char *dev, int max_len)
{ return -1; }

int dyn_ip_route_replace_default(const char *dev, const struct in_addr *via,
				 const struct in_addr *prefsrc)
{
	MIB_IPFORWARDROW route;
	DWORD res;
	int ifindex;
	struct in_addr own_addr, addr;

	ifindex = dyn_ip_get_ifindex(dev);
	if (ifindex < 0) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_replace_default - "
		      "dyn_ip_get_ifindex() failed\n");
		return -1;
	}
	if (via == NULL) {
		if (dyn_ip_get_ifaddr(dev, &own_addr) < 0) {
			DEBUG(DEBUG_FLAG, "dyn_ip_route_replace_default - "
			      "dyn_ip_get_ifaddr() failed\n");
			return -1;
		}
		via = &own_addr;
	}

	addr.s_addr = 0;
	dyn_ip_route_delete_net_routes(&addr, 0);

	memset(&route, 0, sizeof(route));
	route.dwForwardDest = 0;
	route.dwForwardMask = 0;
	route.dwForwardNextHop = via->s_addr;
	route.dwForwardIfIndex = ifindex;
	route.dwForwardType = 4; /* The next hop is not the final destination
				  * (remote route). */
	route.dwForwardProto = 2; /* PROTO_IP_NETMGMT */
	route.dwForwardMetric1 = 1;
	res = SetIpForwardEntry(&route);
	switch (res) {
	case NO_ERROR:
		return 0;
	case ERROR_INVALID_PARAMETER:
		DEBUG(DEBUG_FLAG, "SetIpForwardEntry: "
		      "ERROR_INVALID_PARAMETER\n");
		break;
	case ERROR_NOT_SUPPORTED:
		DEBUG(DEBUG_FLAG, "SetIpForwardEntry: "
		      "ERROR_NOT_SUPPORTED\n");
		break;
	default:
		DEBUG(DEBUG_FLAG, "SetIpForwardEntry: "
		      "Unknown error %li\n", res);
		break;
	}

	return -1;
}

int dyn_ip_route_replace_net(const char *dev, const struct in_addr *dst,
			     int plen)
{ return -1; }


static int default_route_saved = 0;
static MIB_IPFORWARDROW default_route_row;


int dyn_ip_route_restore_default(const char *force_dev)
{
	struct in_addr addr;
	DWORD res;

	if (!default_route_saved) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_restore_default - default "
		      "route not saved\n");
		return 0;
	}

	/* Note! force_dev is not currently support on Windows platforms */

	if (force_dev != NULL) {
		int ifindex = dyn_ip_get_ifindex(force_dev);
		if (ifindex < 0) {
			DEBUG(DEBUG_FLAG, "dyn_ip_route_restore_default - "
			      "could not force interface '%s'\n",
			      force_dev);
			/* continue anyway */
		} else {
			DEBUG(DEBUG_FLAG, "dyn_ip_route_restore_default - "
			      "save ifindex=%li, forcing %i\n",
			      default_route_row.dwForwardIfIndex,
			      ifindex);
			default_route_row.dwForwardIfIndex = ifindex;
		}
	}

	addr.s_addr = 0;
	dyn_ip_route_delete_net_routes(&addr, 0);

	res = SetIpForwardEntry(&default_route_row);
	switch (res) {
	case NO_ERROR:
		return 0;
	case ERROR_INVALID_PARAMETER:
		DEBUG(DEBUG_FLAG, "SetIpForwardEntry: "
		      "ERROR_INVALID_PARAMETER\n");
		break;
	case ERROR_NOT_SUPPORTED:
		DEBUG(DEBUG_FLAG, "SetIpForwardEntry: "
		      "ERROR_NOT_SUPPORTED\n");
		break;
	default:
		DEBUG(DEBUG_FLAG, "SetIpForwardEntry: "
		      "Unknown error %li\n", res);
		break;
	}

	return -1;
}


int dyn_ip_route_save_default(struct in_addr *forced_gateway)
{
	char buf[8192];
	PMIB_IPFORWARDTABLE table = (PMIB_IPFORWARDTABLE) buf;
	DWORD res;
	ULONG blen;
	int i;

	blen = sizeof(buf);
	res = GetIpForwardTable(table, &blen, 1);
	if (res != NO_ERROR) {
		DEBUG(DEBUG_FLAG, "dyn_ip_route_save_default - "
		      "GetIpForwardTable returned %li (blen=%li)\n",
		      res, blen);
		return -1;
	}

	for (i = 0; i < table->dwNumEntries; i++) {
		PMIB_IPFORWARDROW row = &table->table[i];
		if (row->dwForwardDest == 0 && row->dwForwardMask == 0) {
			default_route_saved = 1;
			memcpy(&default_route_row, row,
			       sizeof(default_route_row));
			DEBUG(DEBUG_FLAG, "dyn_ip_route_save_default - "
			      "default route saved\n");
			DEBUG(DEBUG_FLAG, "   NextHop=%s, ifIndex=%li\n",
			      dword_addr_str(row->dwForwardNextHop),
			      row->dwForwardIfIndex);
			if (forced_gateway != NULL) {
				DEBUG(DEBUG_FLAG, "   setting forced gateway "
				      "%s\n", inet_ntoa(*forced_gateway));
				default_route_row.dwForwardNextHop =
					forced_gateway->s_addr;
			}
			return 0;
		}
	}

	DEBUG(DEBUG_FLAG, "dyn_ip_route_save_default - could not find a "
	      "default route\n");
	return -1;
}


int dyn_ip_route_to_host(struct in_addr dst) { return -1; }
int dyn_ip_tunnel_add(const char *dev, struct in_addr remote, 
	struct in_addr local) { return -1; }
int dyn_ip_tunnel_del(const char *dev) { return -1; }


int dyn_ip_set_ttl(int ttl)
{
	DWORD res;

	res = SetIpTTL(ttl);

	if (res != NO_ERROR) {
		DEBUG(DEBUG_FLAG, "dyn_ip_set_ttl: SetIpTTL(%i) failed: "
		      "res=%li\n", ttl, res);
		return -1;
	}

	return 0;
}

#endif /* DYN_TARGET_WINDOWS */


/**
 * dyn_ip_free_interface_map:
 *
 * Free the given interface map (allocated with dyn_ip_get_interface_map()).
 */
void dyn_ip_free_interface_map(struct idxmap *map)
{
	struct idxmap *tmp, *prev;

	tmp = map;
	while (tmp != NULL) {
		prev = tmp;
		tmp = tmp->next;
		free(prev);
	}
}


/**
 * dyn_ip_mask_to_prefix_len:
 * @mask: network mask as a 32-bit field
 *
 * Converts address mask from 32-bit field format to prefix length.
 *
 * Returns: prefix length (number of prefix 1 bits) in @mask.
 */
int dyn_ip_mask_to_prefix_len(unsigned int mask)
{
	int prefix = 0;
	unsigned int u = ntohl(mask);
	while (prefix < 32 && (u & (1 << 31))) {
		prefix++;
		u <<= 1;
	}

	return prefix;
}


/**
 * dyn_ip_prefix_len_to_mask:
 * @plen: prefix length of the network mask
 *
 * Converts address mask from prefix length to 32-bit field.
 *
 * Returns: 32-bit network mask corresponding to @plen
 */
unsigned int dyn_ip_prefix_len_to_mask(int plen)
{
	unsigned int mask = 0;
	int i;

	if (plen < 0 || plen > 32) {
		DEBUG(DEBUG_FLAG, "dyn_ip_mask_to_prefix_len: invalid prefix "
		      "length %i\n", plen);
		return 0;
	}

	for (i = 0; i < plen; i++) {
		mask >>= 1;
		mask |= (1 << 31);
	}

	return htonl(mask);
}
