#include <asm/types.h>
#include <stdarg.h>
#include <string.h>
#include <msgHandler/msgHandler.h>
#include <papi/MHApi.h>
#include <util/list.h>
#include <util/hash.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <amapi.h>
#include <amapi_common.h>
#include <amapi_response.h>
#include "l2tp_cli.h"
#include <vpn/cli_handler.h>
#include <vpn/shared_cli.h>
#include <assert.h>
#include "l2tp.h"

static const char* shared_keywords[] = {
  "hello",
  "authentication",
  "LISTEN",
  "ANY"
};

const sHandlerKeywords SHAREDKeywords = {
  shared_keywords,
  sizeof(shared_keywords) / sizeof(*shared_keywords)
};

cli_chmode shared_modes[] = {
  { L2TPHandler, "L2TP", "PROMPT=(config-vpdn-l2tp)#  ", &shared_cli_vpdn },
  { NullHandler, NULL, NULL, NULL }
};

#ifdef ARUBA_USE_CLI
void shared_cli_show_vpdn_tunnel(char *buf, int *buflen, char *tag)
{
  struct tunnel *t;
  char           tmp[MAX_CLI_LINE];
  char          *tmp2;
  int            tagid = 0;

  //TODO: LOG !!
  if (!buf || !buflen)
    return;
  
  if (tag)
    tagid = atoi(tag);

  if (!tunnels.count)
    ADD_DATA(tmp, MAX_CLI_LINE, buf, "%% No active L2TP Tunnel");
  else
  {
    ADD_DATA(tmp, MAX_CLI_LINE, buf, 
	     "L2TP Tunnel Information (Total tunnels = %d)", tunnels.count);
    
    t = tunnels.head;
    while (t)
    {
      if (tag && tagid != t->ourtid)
      {
	t = t->next;
	continue;
      }

      ADD_DATA(tmp, MAX_CLI_LINE, buf,
	       "Tunnel ID = %d is up, remote id is %d", t->ourtid, t->tid);
      switch (t->state)
      {
      case SCCRQ:
	tmp2 = "SCCRQ: Connection Request";
	break;
      case SCCRP:
	tmp2 = "SCCRP: Connection Response";
	break;
      case SCCCN:
	tmp2 = "SCCCN: Connection Connected";
	break;
      case StopCCN:
	tmp2 = "StopCCN Stop Connection";
	break;
      default:
	tmp2 = "UNKNOWN";
	break;
      }
      ADD_DATA(tmp, MAX_CLI_LINE, buf, "  Tunnel state is %s", tmp2);
      ADD_DATA(tmp, MAX_CLI_LINE, buf, " Remote Internet Address %s, port %d",
	       IPADDY (t->peer.sin_addr), ntohs (t->peer.sin_port) );
#if 0
      // localaddr isn't set !!!
      ADD_DATA(tmp, MAX_CLI_LINE, buf, " Local Internet Address %s, port 1701",
	       IPADDY (t->lns->localaddr));
      ADD_DATA(tmp, MAX_CLI_LINE, buf, " Control Ns %d, Nr %d", 
	       t->control_seq_num, t->control_rec_seq_num);

      ADD_DATA(tmp, MAX_CLI_LINE, buf, " Local RWS %d, Remote RWS %d", 
	       t->ourrws, t->rws);
#endif

      t = t->next;
    }
  }

  return;
}

#if 0
/* for mem debugs */
extern void dmalloc_shutdown();
#endif

void shared_cli_clear_vpdn_tunnel(char *buf, int *buflen, char *tag)
{
  struct tunnel *t;
  struct tunnel *tmp;
  struct call   *sc;
  struct call   *tmpsc;  
  char           cleared = 0;
  int            tagid   = 0;

#if 0
/* for mem debugs */
  dmalloc_shutdown();
  exit(0);
#endif
     
  //TODO: LOG !!
  if (!buf || !buflen)
    return;
  
  if (tag)
    tagid = atoi(tag);

  t = tunnels.head;
  while (t)
  {
    if (tag && tagid != t->ourtid)
    {
      t = t->next;
      continue;
    }
    
    // cleared at least one tunnel !
    cleared =1;
    // Keep the next tunnel before calling destroy since it will do a free!!
    tmp = t->next;
    
    // close all the calls 
    sc = t->call_head;
    while (sc) {
      tmpsc = sc->next;
      call_close(sc);
      sc = tmpsc;
    }

    t = tmp;
  }

  if (!cleared)
  {
    if (tag)
      CLI_RETURN_ERROR(clibuf, SHARED_CLI_PAPI_PREALLOC_PACKET_SIZE,
		       "Tunnel %d does not exist", tagid);
    else
      CLI_RETURN_ERROR(clibuf, SHARED_CLI_PAPI_PREALLOC_PACKET_SIZE,
		       "No tunnel exist");
  }

}

#define IPTOSTR(a) (inet_ntoa(*(struct in_addr *)(&(a))))

int
tunndown(unsigned int innerip)
{
    struct tunnel *t, *nextt;
    struct call   *sc, *nextsc;
    int ret = 0;
    
    t = tunnels.head;
    while (t) {
        int found = 0;
        
        // Keep the next tunnel before calling destroy since it will do a free!!
        nextt = t->next;
        
        // close all the calls 
        sc = t->call_head;
        
        // search and match innerip, if found one, must close all calls...
        nextsc = sc;
        while(nextsc) {
            if (nextsc->addr == innerip) {
                found = 1;
                break;
            }
            nextsc = nextsc->next;
        }
        if (found) {
            arubaLog(LOG_INFO,
                     "Config changed.  Tearing down of L2TP innerIP=%s",
                     IPTOSTR(innerip));
            while (sc) {
                nextsc = sc->next;
                call_close(sc);
                sc = nextsc;
                ret = 1;
            }
        }
        t = nextt;
    }
    return ret;
}

#endif

static unsigned int g_magic[4] = {0x12345678, 0xabcdabcd, 0x87654321,
                                  0x43218765};
static unsigned int g_magicfree[4] = {0x10203040, 0xdcbadcba, 0x50607080,
                                      0x80706050};

typedef struct  {
    unsigned char str[8];
    unsigned int  line;
    unsigned int  magic[4];
} myhdr_t;


void *
l2tpmalloc(const char *str, int ln, size_t sz)
{
    int len = 0;
    unsigned char *ret = malloc(sz+sizeof(myhdr_t));
    myhdr_t *hdr;
    
    if (!ret) {
        arubaLog(LOG_CRIT, "L2TP fatal SW error.  l2tpmalloc failed at %s:%d", str, ln);
        return NULL;
    }
    hdr = (myhdr_t *)ret;
    
    memset(hdr, 0, sizeof(myhdr_t));
    
    if (str) {
        len = strlen(str);
        if (len >= sizeof(hdr->str)) {
            memcpy(hdr->str, &str[len-sizeof(hdr->str)], sizeof(hdr->str));
            hdr->str[sizeof(hdr->str)-1] = '\0';
        } else
            sprintf(hdr->str, "%s", str);
    }
    hdr->line = ln;
    memcpy(hdr->magic, g_magic, sizeof(hdr->magic));
    return (ret + sizeof(myhdr_t));
    
}


static char *g_nofree = NULL;

void
l2tpfree(const char *str, int ln, void *ptr)
{
    unsigned char *ret = ptr;
    myhdr_t *hdr;
    
    if (!ret) {
        arubaLog(LOG_CRIT, "L2TP fatal SW error, L2TPFREE NULL ptr from %s:%d", str, ln);
        // perfectly fine thing to do
        return; 
    }

    if (ptr == g_nofree) {
        arubaLog(LOG_CRIT, "L2TP fatal SW error, attempting to free no_free area from %s:%d",
                 str, ln);
        return;
    }
    hdr = (myhdr_t *)(ret - sizeof(myhdr_t));
    if (memcmp(hdr->magic, g_magic, sizeof(hdr->magic)) != 0) {
        char str2[sizeof(hdr->str)];
        memcpy(str2, hdr->str, sizeof(str2));
        str2[(sizeof(str2)-1)] = '\0';

        arubaLog(LOG_CRIT, "L2TP fatal SW error, L2TPFREE with bad ptr at %s:%d malloc'd from %s:%d, magic=0x%08x, 0x%08x, 0x%08x, 0x%08x",
                 str, ln, str2, hdr->line, hdr->magic[0], hdr->magic[1], hdr->magic[2], hdr->magic[3]);
    } else {
        // wipe out magic
        memcpy(hdr->magic, g_magicfree, sizeof(hdr->magic));
        free(hdr);
    }
}

char *
l2tpstrdup(const char *f, int ln, const char *s)
{
    int len;
    char *ret;
    if (!s)
        return NULL;
    len = strlen(s);
    ret = (char *)l2tpmalloc(f, ln, len+1);
    if (!ret)
        return NULL;
    memcpy(ret, s, len+1);
    return ret;
}


int
checksanity(const char *f, int ln, void *buf)
{
    myhdr_t *hdr;
    
    if (!buf) {
        arubaLog(LOG_ERR, "checksanity failed, buf is NULL");
        return -1;
    }
    hdr = (myhdr_t *)( ((unsigned char *)buf) - sizeof(myhdr_t));
    if (memcmp(hdr->magic, g_magic, sizeof(hdr->magic)) != 0) {
        char str2[sizeof(hdr->str)];
        memcpy(str2, hdr->str, sizeof(str2));
        str2[(sizeof(str2)-1)] = '\0';
        
        arubaLog(LOG_ERR, "checksanity bad ptr at %s:%d malloc'd from %s:%d, magic=0x%08x, 0x%08x, 0x%08x, 0x%08x",
                 f, ln, str2, hdr->line, hdr->magic[0], hdr->magic[1], hdr->magic[2], hdr->magic[3]);
        return -1;
    }
    return 0;
}


void
l2tpnofree(void *buf)
{
    if (!g_nofree) {
        g_nofree = buf;
    } else {
        arubaLog(LOG_ERR, "l2tpnofree called more than once");
        assert(0);
    }
}

int
isl2tpnofree(void *buf)
{
    return (buf == g_nofree);
}
