/*
 * Layer Two Tunnelling Protocol Daemon
 * Copyright (C) 1998 Adtran, Inc.
 *
 * Mark Spencer
 *
 * This software is distributed under the terms
 * of the GPL, which you should have received
 * along with this source.
 *
 * Network routines for UDP handling
 */
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include "vpn/cmn_ipsec.h"
#include "l2tp.h"
#include "util/sibyte_msg.h"

#ifdef ARUBA_USE_CLI
#include "l2tp_cli.h"
#include "vpn/shared_cli.h"
#define SERVER_SIDE
#include "vpn/getpool.h"
#endif

#ifdef ARUBA_USE_ARUBA_LOG
#include <syslog.h>
#include <syslog/syslogutil.h>
#include <syslog/arubalog.h>
#endif /* ARUBA_USE_ARUBA_LOG */

#include <include/authapi.h>

char hostname[256];
// This is set through messages to listen to the MX IP only
// XP has a bug with IPSEC if we don't do that !!
unsigned int listen_addy = INADDR_ANY;  /* Address to listen on */
unsigned int mswitch_addy = INADDR_ANY;  /* Address to listen on */

// for debugging
int g_tunnfreed = 0, g_callfreed;

struct sockaddr_in server, from;        /* Server and transmitter structs */
int server_socket = 0;              /* Server socket */
#ifdef USE_KERNEL
int kernel_support;             /* Kernel Support there or not? */
#endif

static int papi_fd;
static int pipefd;
static int dp_fd = -1;

static void safe_fdset(int fd, fd_set *readfds, int line);

/*
 * Debugging info
 */
//TODO ARUBA TURNED ALL THE DEBUG ON !!!
int debug_tunnel = 0;
int debug_network = 1;          /* Debug networking? */
int packet_dump = 1;            /* Dump packets? */
int debug_avp = 1;              /* Debug AVP negotiations? */
int debug_state = 1;            /* Debug state machine? */

#if 0
/* for mem debug */
extern void dmalloc_debug(unsigned int);
#endif

int init_network (void)
{
    long arg;
    int length = sizeof (server);
    gethostname (hostname, sizeof (hostname));
    server.sin_family = AF_INET;

#if 0
    /* for mem debug */
    dmalloc_debug(0x4f4e503);
#endif
    
    // init is also used for reinit !!
    if (server_socket)
      close(server_socket);

    //server.sin_addr.s_addr = htonl (listen_addy);
    server.sin_addr.s_addr = htonl (INADDR_ANY);
    server.sin_port = htons (gconfig.port);
    if ((server_socket = socket (PF_INET, SOCK_DGRAM, 0)) < 0)
    {
        log (LOG_CRIT, "%s: Unable to allocate socket. Terminating.\n",
             __FUNCTION__);
        return -EINVAL;
    };
    /* L2TP/IPSec: Set up SA for listening port here?  NTB 20011015
     */
    if (bind (server_socket, (struct sockaddr *) &server, sizeof (server)))
    {
        close (server_socket);
        log (LOG_CRIT, "%s: Unable to bind socket. Terminating.\n",
             __FUNCTION__);
        return -EINVAL;
    };
    if (getsockname (server_socket, (struct sockaddr *) &server, &length))
    {
        log (LOG_CRIT, "%s: Unable to read socket name.Terminating.\n",
             __FUNCTION__);
        return -EINVAL;
    }
#ifdef USE_KERNEL
    if (gconfig.forceuserspace)
    {
        log (LOG_LOG, "Not looking for kernel support.\n");
        kernel_support = 0;
    }
    else
    {
        if (ioctl (server_socket, SIOCSETL2TP, NULL) < 0)
        {
            log (LOG_LOG, "L2TP kernel support not detected.\n");
            kernel_support = 0;
        }
        else
        {
            log (LOG_LOG, "Using l2tp kernel support.\n");
            kernel_support = -1;
        }
    }
#else
    log (LOG_LOG, "This binary does not support kernel L2TP.\n");
#endif
    arg = fcntl (server_socket, F_GETFL);
    arg |= O_NONBLOCK;
    fcntl (server_socket, F_SETFL, arg);
    gconfig.port = ntohs (server.sin_port);
    return 0;
}

inline void extract (void *buf, int *tunnel, int *call)
{
    /*
     * Extract the tunnel and call #'s, and fix the order of the 
     * version
     */

    struct payload_hdr *p = (struct payload_hdr *) buf;
    if (PLBIT (p->ver))
    {
        *tunnel = p->tid;
        *call = p->cid;
    }
    else
    {
        *tunnel = p->length;
        *call = p->tid;
    }
}

inline void fix_hdr (void *buf)
{
    /*
     * Fix the byte order of the header
     */

    struct payload_hdr *p = (struct payload_hdr *) buf;
    _u16 ver = ntohs (p->ver);
    if (CTBIT (p->ver))
    {
        /*
         * Control headers are always
         * exactly 12 bytes big.
         */
        swaps (buf, 12);
    }
    else
    {
        int len = 6;
        if (PSBIT (ver))
            len += 4;
        if (PLBIT (ver))
            len += 2;
        if (PFBIT (ver))
            len += 4;
        swaps (buf, len);
    }
}

void dethrottle (void *call)
{
/*	struct call *c = (struct call *)call; */
/*	if (c->throttle) {
#ifdef DEBUG_FLOW
		log(LOG_DEBUG, "%s: dethrottling call %d, and setting R-bit\n",__FUNCTION__,c->ourcid); 
#endif 		c->rbit = RBIT;
		c->throttle = 0;
	} else {
		log(LOG_DEBUG, "%s:  call %d already dethrottled?\n",__FUNCTION__,c->ourcid); 	
	} */
}

/*
 * control_xmit: We no longer do hellos in the control path.
 */
void control_xmit (void *b)
{
    struct buffer *buf = (struct buffer *) b;
    struct tunnel *t;
    struct timeval tv;
    int ns;
    if (!buf)
    {
        log (LOG_WARN, "%s: called on NULL buffer!\n", __FUNCTION__);
        return;
    }

    //buf->retries++;
    t = buf->tunnel;
    ns = ntohs (((struct control_hdr *) (buf->start))->Ns);
    if (t)
    {
        if (ns < t->cLr)
        {
#ifdef DEBUG_CONTROL_XMIT
            log (LOG_DEBUG, "%s: Tossing packet %d\n", __FUNCTION__, ns);
#endif
            /* We don't send anything here */
            toss (buf);

            return;
        }
        /* send frame out */
        udp_xmit (buf);
        toss(buf);
    }
    else  {
        /* should never reach */
        abort();
    }
}

void udp_xmit (struct buffer *buf)
{
    int ret;
    /*
     * Just send it all out.
     */
    ret = sendto (server_socket, buf->start, buf->len, 0,
                  (struct sockaddr *) &buf->peer, sizeof (buf->peer));

    if (ret < 0) {
        arubaLog(LOG_ERR, 
                 "Failed to send %d bytes to peer %s, return value=%d, errno=%d...%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x",
                 buf->len,
                 inet_ntoa(buf->peer.sin_addr),
                 ret,
                 errno,
                 ((unsigned char *)(buf->start))[0],
                 ((unsigned char *)(buf->start))[1],
                 ((unsigned char *)(buf->start))[2],
                 ((unsigned char *)(buf->start))[3],
                 ((unsigned char *)(buf->start))[4],
                 ((unsigned char *)(buf->start))[5],
                 ((unsigned char *)(buf->start))[6],
                 ((unsigned char *)(buf->start))[7],
                 ((unsigned char *)(buf->start))[8],
                 ((unsigned char *)(buf->start))[9],
                 ((unsigned char *)(buf->start))[10],
                 ((unsigned char *)(buf->start))[11],
                 ((unsigned char *)(buf->start))[12],
                 ((unsigned char *)(buf->start))[13],
                 ((unsigned char *)(buf->start))[14],
                 ((unsigned char *)(buf->start))[15],
                 ((unsigned char *)(buf->start))[16],
                 ((unsigned char *)(buf->start))[17],
                 ((unsigned char *)(buf->start))[18],
                 ((unsigned char *)(buf->start))[19],
                 ((unsigned char *)(buf->start))[20],
                 ((unsigned char *)(buf->start))[21]);
        /* not good, restart server socket???  Or should we just abort? */
        sleep(1);
        init_network();
    } else {
        arubaDebugLog(DEBUG_VPDN_PACKETS, LOG_DEBUG,
                      "Sent %d bytes to peer %s...%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x", ret,
                      inet_ntoa(buf->peer.sin_addr),
                      ((unsigned char *)(buf->start))[0],
                      ((unsigned char *)(buf->start))[1],
                      ((unsigned char *)(buf->start))[2],
                      ((unsigned char *)(buf->start))[3],
                      ((unsigned char *)(buf->start))[4],
                      ((unsigned char *)(buf->start))[5],
                      ((unsigned char *)(buf->start))[6],
                      ((unsigned char *)(buf->start))[7],
                      ((unsigned char *)(buf->start))[8],
                      ((unsigned char *)(buf->start))[9],
                      ((unsigned char *)(buf->start))[10],
                      ((unsigned char *)(buf->start))[11],
                      ((unsigned char *)(buf->start))[12],
                      ((unsigned char *)(buf->start))[13],
                      ((unsigned char *)(buf->start))[14],
                      ((unsigned char *)(buf->start))[15],
                      ((unsigned char *)(buf->start))[16],
                      ((unsigned char *)(buf->start))[17],
                      ((unsigned char *)(buf->start))[18],
                      ((unsigned char *)(buf->start))[19],
                      ((unsigned char *)(buf->start))[20],
                      ((unsigned char *)(buf->start))[21]);
    }
}

#define MAX_PIPE_FD 64
static int g_numpipes = 0;
static int g_pipefd[MAX_PIPE_FD];

int
accept_pipe(int r)
{
    int s;
    errno = 0;
    s = accept(r, NULL, NULL);
    if (s < 0) {
        syslog(LOG_ERR, "accept_pipe error %s", strerror(errno));
        return -1;
    }
    if (g_numpipes >= MAX_PIPE_FD) {
        safe_close(&s);
        return -1;
    }
    g_pipefd[g_numpipes++] = s;

    arubaLog(LOG_DEBUG, "Accepted pipe %d from %d", s, r);
    
    return r;
}

int
close_pipe(int r, int index)
{
    int i;
    // should be index...
    for (i=index; i<g_numpipes; i++) {
        if (g_pipefd[i] == r) {
            arubaLog(LOG_DEBUG, "closed socket %d, index=%d, num=%d",
                     r, index, g_numpipes);
            safe_close(&r);
            g_pipefd[i] = g_pipefd[g_numpipes-1];
            g_numpipes--;
            return 0;
        }
    }
    arubaLog(LOG_ERR, "close_pipe, no match s=%d, index=%d", r, index);
    safe_close(&r);
    return -1;
}

static int
hellodown(char *buf, int buflen)
{
    int i;
    l2tp_down_t d;
    struct tunnel *t;
    
    memcpy(&d, &buf[sizeof(sibyte_msg_t)], sizeof(d));

    t = get_tunnel(d.tid, d.destip, 1701);
    if (t && t->self && t->self->needclose == 0 && t->self->closing == 0) {
        /* make it close immediately */
        t->self->needclose = 0;
        t->self->closing = -1;
    }
    if (d.truetimeout) {
        arubaLog(LOG_INFO, "L2TP Tunnel from %s timed out due to missed L2TP hellos",
                 inet_ntoa(*((struct in_addr *)(&d.destip))));
    }
    return 0;
}

/*
 * isfdvalid - verifies whether the fd is still open.
 */
static int
isfdvalid(int fd)
{
    fd_set readfds;             /* Descriptors to watch for reading */
    struct timeval tv;          /* Timeout for select */

    FD_ZERO(&readfds);
    
    tv.tv_sec = 0;
    tv.tv_usec = 0;

    FD_SET(fd, &readfds);
    
    if (select (fd + 1, &readfds, NULL, NULL, &tv) == -1 &&
        errno == EBADF) {
        return 0;
    } else
        return 1;
}

void network_thread ()
{
    /*
     * We loop forever waiting on either data from the ppp drivers or from
     * our network socket.  Control handling is no longer done here.
     */
    int i, fromlen;             /* Length of the address */
    int tunnel, call;           /* Tunnel and call */
    int recvsize;               /* Length of data received */
    int ret;
    struct buffer *buf;         /* Payload buffer */
    struct call *c, *sc;        /* Call to send this off to */
    struct tunnel *st;          /* Tunnel */
    fd_set readfds;             /* Descriptors to watch for reading */
    int max;                    /* Highest fd */
    struct timeval tv, now;     /* Timeout for select */
    struct timeval *ptv = NULL;
    
    /* This one buffer can be recycled for everything except control packets */
    buf = new_buf (MAX_RECV_SIZE);

    // make sure buf absolutely never gets freed or overwritten.
    l2tpnofree(buf);

#ifdef ARUBA_USE_CLI
    papi_fd = shared_cli_init(L2TPHandler, &l2tp_shared_ipc_handler, &tunndown,
                              hellodown, &dp_fd);
    if (papi_fd <= 0)
      return;

    if (client_auth_init())
      return; //TODO: LOG !!

    arubaLogInit(LOG_L2TP, "l2tp", LOG_INFO, vpdnDebugInfo);

    /* We only have one LNS, no LAC */
    if (!deflns)
      deflns = new_lns ();

    pipefd = initNamedPipe(L2TP_PIPENAME);
    if (pipefd < 0) {
        log(LOG_ERR, "L2TP named pipe at socket %d", pipefd);
        return;
    } else {
        // make it async.  Sometime accept to named pipe craps out in linux
        // for some reason.  You'd think they'd fix it by now.
        int on = 1;
        int ret = ioctl(pipefd, FIONBIO, &on);
        if (ret < 0) {
            log(LOG_ERR, "Unable to make named pipe server sock async");
        }
    }
#endif

    for (;;)
    {
        /*
           * First, let's send out any outgoing packets that are waiting on us.
           * xmit_udp should only
           * contain control packets in the unthreaded version!
         */
        // ugly debug for multiple frees
    restart:
        max = 0;
        FD_ZERO (&readfds);

        g_tunnfreed = 0;
        st = tunnels.head;
        while (st)
        {
            if (st->self->needclose ^ st->self->closing)
            {
                if (debug_tunnel)
                    log (LOG_DEBUG, "%S: closing down tunnel %d\n",
                         __FUNCTION__, st->ourtid);
                call_close (st->self);
#if 0
                /* Reset the while loop
                   and check for NULL */
                st = tunnels.head;
                if (!st)
                    break;
                continue;
#else
                goto restart;
#endif
            }
            sc = st->call_head;
            while (sc)
            {
                struct call *lastsc = sc;
                
                if (sc->needclose ^ sc->closing)
                {
                    call_close (sc);

                    if (g_tunnfreed) {
                        arubaLog(LOG_ERR,
                                 "Multiple tunnel free after call_close\n");
                        goto restart;
                    }
                    sc = st->call_head;
                    if (!sc)
                        break;
                    if(sc == lastsc)
                        sc = sc->next;
                    continue;
                }
                if (sc->fd > -1)
                {
/*					if (!sc->throttle && !sc->needclose && !sc->closing) { */
                    if (!sc->needclose && !sc->closing)
                    {
                        if (sc->fd > max)
                            max = sc->fd;
                        safe_fdset (sc->fd, &readfds, __LINE__);
                    }
                }
                sc = sc->next;
            }
            st = st->next;
        }
        
        safe_fdset (server_socket, &readfds, __LINE__);
        if (server_socket > max)
            max = server_socket;
        safe_fdset (control_fd, &readfds, __LINE__);
        if (control_fd > max)
            max = control_fd;
            
#ifdef ARUBA_USE_CLI
        safe_fdset (pipefd, &readfds, __LINE__);
        if (pipefd > max)
            max = pipefd;

        for (i=0; i<g_numpipes; i++) {
            safe_fdset(g_pipefd[i], &readfds, __LINE__);
            if (g_pipefd[i] > max)
                max = g_pipefd[i];
        }
        safe_fdset(papi_fd, &readfds, __LINE__);
        if (papi_fd > max)
            max = papi_fd;

        if (dp_fd > 0) {
            safe_fdset(dp_fd, &readfds, __LINE__);
            if (dp_fd > max)
                max = dp_fd;
        }
#endif
        tv.tv_sec = 1;
        tv.tv_usec = 0;
#if 0
        schedule_unlock ();
        if (getselectdiff(&tv)) {
            ptv = &tv;
        }
#endif
        ptv = &tv;
        ret = select (max + 1, &readfds, NULL, NULL, ptv);
        //schedule_lock ();

        if (ret < 0) {
            if (errno == EINTR) {
                errno = 0;
                continue;
            } else if (errno == EBADF) {
                //alarm_handler(0);
                if (!isfdvalid(server_socket)) {
                    /* dead meat, should never happen though */
                    arubaLog(LOG_CRIT, "Fatal SW error in L2TPD.  Server socket was closed.");
                    abort();
                }
                if (!isfdvalid(pipefd)) {
                    arubaLog(LOG_CRIT, "Fatal SW error in L2TPD.  Pipe socket was closed.");
                    abort();
                }
                /* scan through all the sockets and mark invalid ones closed.
                   alarm_handler should eventually kick in to free the
                   structures.  */
                st = tunnels.head;
                while (st) {
                    sc = st->call_head;
                    while (sc)  {
                        if (sc->fd > -1 && !isfdvalid(sc->fd)) {
                            sc->fd = -1;
                        }
                        sc = sc->next;
                    }
                    st = st->next;
                }
                continue;
            } else {
                //alarm_handler(0);
                arubaLog(LOG_CRIT, "Resource unavailable error code=%d",
                         errno);
                /* allows system to recover if it's RESOURCE_UNAVAIL */
                sleep(1);
                continue;
            }
        }

        if (ret == 0) {
            // only timeout, nothing else...
            //alarm_handler(0);
            continue;
        } else {
            // do timeouts first...
            //alarm_handler(0);
        }
        
#ifdef ARUBA_USE_CLI
        if (FD_ISSET(pipefd, &readfds)) {
            accept_pipe(pipefd);
        }
        if (FD_ISSET (papi_fd, &readfds))
        {
            shared_papi_rcv ();
        }
        if (dp_fd > 0 && FD_ISSET (dp_fd, &readfds))
        {
            shared_cli_dpfd(dp_fd);
        }
        for (i=0; i<g_numpipes; i++) {
            if (FD_ISSET(g_pipefd[i], &readfds)) {
                unsigned int oldip=0, newip=0;
                // return value is irrelevant here...
                getAddrFromSock(g_pipefd[i], &oldip, &newip);
                if (oldip && newip && (oldip != newip)) {
                    // search calls and change address so addr gets freed
                    // on tunnel down...
                    struct tunnel *st = tunnels.head;
                    while(st) {
                        sc = st->call_head;
                        while(sc) {
                            if (sc->addr == oldip) {
                                sc->addr = newip;
                                arubaLog(LOG_DEBUG,
                                         "Found call, replacing addr 0x%08x with 0x%08x", oldip, newip);
                                goto done;
                            }
                            sc = sc->next;
                        }
                        st = st->next;
                    }
                }
            done:
                if (close_pipe(g_pipefd[i], i) == 0)
                    i--;
            }
        }
#endif	
        if (FD_ISSET (control_fd, &readfds))
        {
            do_control ();
        }
        if (FD_ISSET (server_socket, &readfds))
        {
            unsigned int inSPI=0, outSPI=0, dstIP=0; /* quiet dumb compiler */
            unsigned int newns = 0, newnr = 0;
            
            /*
             * Okay, now we're ready for reading and processing new data.
             */
            recycle_buf (buf);
            /* Reserve space for expanding payload packet headers */
            buf->start += PAYLOAD_BUF;
            buf->len -= PAYLOAD_BUF;
            fromlen = sizeof (from);
            recvsize =
                recvfrom (server_socket, buf->start, buf->len, 0,
                          (struct sockaddr *) &from, &fromlen);
            /* check for l2tp_ipsecattr */
            if (recvsize > sizeof(l2tp_ipsecattr_t)) {
                l2tp_ipsecattr_t *avp =
                    (l2tp_ipsecattr_t *)(&((uint8_t *)(buf->start))[recvsize-sizeof(*avp)]);
                if (avp->udplen != recvsize ||
                    avp->magic != ARUBA_L2TP_AVP_MAGIC) {
                    arubaLog(LOG_DEBUG, "Unencrypted L2TP packet received from %s. len=%d", inet_ntoa(from.sin_addr), avp->udplen);
                    continue;
                } else {
                    arubaDebugLog(DEBUG_VPDN_PACKETS, LOG_DEBUG, "Received encrypted L2TP %d bytes from %s, SPI=0x%08x, 0x%08x,...%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x",
                                  recvsize, inet_ntoa(from.sin_addr),
                                  avp->inSPI, avp->outSPI,
                                  ((unsigned char *)(buf->start))[0],
                                  ((unsigned char *)(buf->start))[1],
                                  ((unsigned char *)(buf->start))[2],
                                  ((unsigned char *)(buf->start))[3],
                                  ((unsigned char *)(buf->start))[4],
                                  ((unsigned char *)(buf->start))[5],
                                  ((unsigned char *)(buf->start))[6],
                                  ((unsigned char *)(buf->start))[7],
                                  ((unsigned char *)(buf->start))[8],
                                  ((unsigned char *)(buf->start))[9],
                                  ((unsigned char *)(buf->start))[10],
                                  ((unsigned char *)(buf->start))[11],
                                  ((unsigned char *)(buf->start))[12],
                                  ((unsigned char *)(buf->start))[13],
                                  ((unsigned char *)(buf->start))[14],
                                  ((unsigned char *)(buf->start))[15],
                                  ((unsigned char *)(buf->start))[16],
                                  ((unsigned char *)(buf->start))[17],
                                  ((unsigned char *)(buf->start))[18],
                                  ((unsigned char *)(buf->start))[19],
                                  ((unsigned char *)(buf->start))[20],
                                  ((unsigned char *)(buf->start))[21]);

                    inSPI = avp->inSPI;
                    outSPI = avp->outSPI;
                    dstIP = avp->dstIP;
                    newns = avp->l2tpNs;
                    newnr = avp->l2tpNr;
                }
                recvsize -= sizeof(l2tp_ipsecattr_t);
            } else {
                /* invalid frame.  DOS attack or Windows bug? */
                continue;
            }
            if (recvsize < MIN_PAYLOAD_HDR_LEN)
            {
                if (recvsize < 0)
                {
                    if (errno != EAGAIN)
                        log (LOG_WARN,
                             "%s: recvfrom returned error %d (%s)\n",
                             __FUNCTION__, errno, strerror (errno));
                }
                else
                {
                    log (LOG_WARN, "%s: received too small a packet\n",
                         __FUNCTION__);
                }
            }
            else
            {
                buf->len = recvsize;
                fix_hdr (buf->start);
                extract (buf->start, &tunnel, &call);
                if (debug_network)
                {
                    log (LOG_DEBUG, "%s: recv packet from %s, size = %d, tunnel = %d, call = %d\n", __FUNCTION__, inet_ntoa (from.sin_addr), recvsize, tunnel, call);
                }
                if (packet_dump)
                {
                    do_packet_dump (buf);
                }
                g_tunnfreed = 0;
                g_callfreed = 0;
                
                if (!
                    (c =
                     get_call (tunnel, call, from.sin_addr.s_addr,
                               from.sin_port)))
                {
                    if ((c =
                         get_call_from_tunnel (tunnel, from.sin_addr.s_addr,
                                               from.sin_port)))
                    {
                        c->container->inSPI = inSPI;
                        c->container->outSPI = outSPI;
                        c->container->srcIP = dstIP;
                        /*
                         * It is theoretically possible that we could be sent
                         * a control message (say a StopCCN) on a call that we
                         * have already closed or some such nonsense.  To prevent
                         * this from closing the tunnel, if we get a call on a valid
                         * tunnel, but not with a valid CID, we'll just send a ZLB
                         * to ack receiving the packet.
                         */
                        if (debug_tunnel)
                            log (LOG_DEBUG,
                                 "%s: no such call %d on tunnel %d.  Sending special ZLB\n",
                                 __FUNCTION__);
                        handle_special (buf, c, call);
                    }
                    else
                        log (LOG_DEBUG,
                             "%s: unable to find call or tunnel to handle packet.  call = %d, tunnel = %d Dumping.\n",
                             __FUNCTION__, call, tunnel);

                }
                else
                {
                    /* update newns */
                    if (newns && newns >= c->container->control_seq_num) {
                        c->container->control_seq_num = newns;
                    }
                    /* datapath doesn't do any Nr */
                    if (newnr && newnr >= c->container->control_rec_seq_num) {
                        c->container->control_rec_seq_num = newnr;
                    }
                    
                    c->container->inSPI = inSPI;
                    c->container->outSPI = outSPI;
                    c->container->srcIP = dstIP;
                    buf->peer = from;

                    /* Handle the packet */
#if 0
                    c->container->chal_us.vector = NULL;
#endif
                    if (handle_packet (buf, c->container, c))
                    {
                        if (debug_tunnel)
                            log (LOG_DEBUG, "%s: bad packet\n", __FUNCTION__);
                    };
                    if (g_tunnfreed || g_callfreed) {
                        goto restart;
                    }
                    if (c->cnu)
                    {
                        /* Send Zero Byte Packet */
                        control_zlb (buf, c->container, c);
                        c->cnu = 0;
                    }
                    if (g_tunnfreed || g_callfreed) {
                        goto restart;
                    }
                }
            }
        };

        // ugly debugging for multiple frees...
    tunnrestart:
        st = tunnels.head;
        g_tunnfreed = 0;
        g_callfreed = 0;
        gettimeofday(&now, NULL);
        while (st)
        {
            if (st->isdestroyed) {
                // how did this happen?
                arubaLog(LOG_ERR,
                         "Tunnel already freed prior to payload send\n");
                break;
            }

#define GARBAGECOLL_TIME 1200
            
            if (st->indp == 0 &&
                (st->created.tv_sec + GARBAGECOLL_TIME) < now.tv_sec) {
                st->self->closing   = -1;
                st->self->needclose = 0;
            }

            sc = st->self;

            /* call closing, need to get out */
            if (sc->closing || sc->needclose) {
                sc = sc->next;
                st = st->next;
                continue;
            }
            if (sc->isdestroyed) {
            } else
                call_timers(st->self);
            
            sc = st->call_head;
            while (sc)
            {
                // ugly debugging for multiple frees...
                if (g_tunnfreed) {
                    arubaLog(LOG_ERR,
                             "Multiple tunnel free before payload send\n");
                    goto tunnrestart;
                }

                if (g_callfreed) {
                    arubaLog(LOG_ERR,
                             "Multiple call free before payload send\n");
                    goto tunnrestart;
                }

                if (sc->isdestroyed) {
                    // how did this happen?
                    arubaLog(LOG_ERR,
                             "Call already freed prior to payload send\n");
                    break;
                }

                /* call closing, need to get out */
                if (sc->closing || sc->needclose) {
                    sc = sc->next;
                    continue;
                }
                
                call_timers(sc);
            
                if ((sc->fd >= 0) && FD_ISSET (sc->fd, &readfds))
                {
                    /* Got some payload to send */
                    int result;
                    recycle_payload (buf, sc->container->peer);
#ifdef DEBUG_FLOW_MORE
                    log (LOG_DEBUG, "%s: rws = %d, pSs = %d, pLr = %d\n",
                         __FUNCTION__, sc->rws, sc->pSs, sc->pLr);
#endif
                    while ((result =
                            read_packet (buf, sc->fd, SYNC_FRAMING)) > 0)
                    {
                        if (!shared_cli_conf.enabled)
			     continue;

                        add_payload_hdr (sc->container, sc, buf);

                        if (packet_dump)
                        {                            do_packet_dump (buf);
                        }


                        sc->prx = sc->data_rec_seq_num;
#if 0
                        if (sc->zlb_xmit)
                        {
                            deschedule (sc->zlb_xmit);
                            sc->zlb_xmit = NULL;
                        }
#endif
                        sc->zlb_xmit.tv_sec = 0;
                        
                        sc->tx_bytes += buf->len;
                        sc->tx_pkts++;

                        if (add_aruba_avp(buf, sc->container) == 0)
                            udp_xmit (buf);
                        
                        recycle_payload (buf, sc->container->peer);
                    }
                    if (result != 0)
                    {
                        log (LOG_WARN,
                             "%s: tossing read packet, error = %s (%d).  Closing call.\n",
                             __FUNCTION__, strerror (-result), -result);
                        strcpy (sc->errormsg, strerror (-result));
                        sc->needclose = -1;
                    }
                }
                sc = sc->next;
            }
            st = st->next;
        }
    }

}


void
safe_close(int *fd)
{
    unsigned int retaddr = (unsigned int)(__builtin_return_address(0));
    
    if (*fd == papi_fd) {
        arubaLog(LOG_DEBUG, "Internal DEBUG papi_fd closed by 0x%x", retaddr);
    } else if (*fd == pipefd) {
        arubaLog(LOG_DEBUG, "Internal DEBUG pipe_fd closed by 0x%x", retaddr);
    } else if (*fd == server_socket) {
        arubaLog(LOG_DEBUG, "Internal DEBUG server_socket closed by 0x%x",
                 retaddr);
    } else if (*fd == control_fd) {
        arubaLog(LOG_DEBUG, "Internal DEBUG control_socket closed by 0x%x",
                 retaddr);
    } else if (*fd == client_auth_getfd()) {
        arubaLog(LOG_DEBUG, "Internal DEBUG client auth fd closed by 0x%x",
                 retaddr);
    } else if (*fd == get_ipc_fd()) {
        arubaLog(LOG_DEBUG, "Internal DEBUG invalid close socket by 0x%x",
                 retaddr);
    } else if (*fd == get_control_fd()) {
        arubaLog(LOG_DEBUG, "Internal DEBUG invalid close socket by 0x%x",
                 retaddr);
    } else if (*fd <= 2) {
        arubaLog(LOG_DEBUG, "Internal DEBUG invalid close socket by 0x%x",
                 retaddr);
    } else if (*fd == dp_fd) {
        arubaLog(LOG_DEBUG, "Internal DEBUG dp_fd closed by 0x%x", retaddr);
    } else {
        //arubaLog(LOG_ERR, "Socket %d closed by 0x%x", *fd, retaddr);
        close(*fd);
    }
    *fd = -1;
    return;
}

static void
safe_fdset(int fd, fd_set *toset, int line)
{
    unsigned int retaddr = (unsigned int)(__builtin_return_address(0));
    
    if (fd >= 0) {
#if 0
        /* enable this if we really need to be paranoid.  Does not seem to
           be necessary */
        fd_set readfds;
        struct timeval tv;
        
        FD_ZERO(&readfds);
    
        tv.tv_sec = 0;
        tv.tv_usec = 1;

        FD_SET(fd, &readfds);
    
        if (select (fd + 1, &readfds, NULL, NULL, &tv) == -1 &&
            errno == EBADF) {
            arubaLog(LOG_ERR,
                     "Fatal error, attempting to fdset an invalid FD %d from 0x%x:%d",
                     fd, retaddr, line);
            return;
        } else
#endif
        {
            FD_SET(fd, toset);
        }
    } else {
        arubaLog(LOG_ERR,
                 "Fatal error, attempting to fdset an invalid FD %d from 0x%x:%d",
                 fd, retaddr, line);
    }
}
