/*
 * Layer Two Tunnelling Protocol Daemon
 * Copyright (C) 1998 Adtran, Inc.
 *
 * Mark Spencer
 *
 * This software is distributed under the terms
 * of the GPL, which you should have received
 * along with this source.
 *
 * Scheduler code for time based functionality
 *
 */

#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include "l2tp.h"
#include "scheduler.h"

struct schedule_entry *events = NULL;
static struct timeval zero;
//static sigset_t alarm;

static int g_insched = 0;

void init_scheduler (void)
{
#if 0
    struct sigaction act;
    act.sa_handler = alarm_handler;
#ifdef LINUX
    act.sa_restorer = NULL;
#endif
    act.sa_flags = 0;
    sigemptyset (&act.sa_mask);
    sigaddset (&act.sa_mask, SIGALRM);
    sigaction (SIGALRM, &act, NULL);
    sigemptyset (&alarm);
    sigaddset (&alarm, SIGALRM);
#endif /* 0 */
    events = NULL;
    zero.tv_usec = 0;
    zero.tv_sec = 0;
}

void alarm_handler (int signal)
{
    /* Check queue for events which should be
       executed right now.  Execute them, then
       see how long we should set the next timer
     */
    struct schedule_entry *p = events;
    struct timeval now;
    //struct timeval then;
    //struct itimerval itv;
    static int cnt = 0;

    g_insched = 1;
    
    cnt++;
    if (cnt != 1)
    {
        /* Whoa, we got called from within ourselves! */
        // TODO: remove
        arubaLog(LOG_ERR, "alarm_handler calling itself...FATAL %d", cnt);
        //log (LOG_DEBUG, "%s : Whoa... cnt = %d\n", __FUNCTION__, cnt);
        g_insched = 0;
        return;
    }
    
    while (events)
    {
        gettimeofday (&now, NULL);
        p = events;
        if (TVLESSEQ (p->tv, now))
        {
            // TODO: remove
            //arubaLog(LOG_INFO, "alarm_handler has event");
            
            events = events->next;
            /* This needs to be executed, as it has expired.
               It is expected that p->func will free p->data
               if it is necessary */
            (*p->func) (p->data);
            L2TPFREE (__FUNCTION__, p);
        }
        else
            break;
    }
#if 0
    /* When we get here, either there are no more events
       in the queue, or the remaining events need to happen
       in the future, so we should schedule another alarm */
    if (events)
    {
        then.tv_sec = events->tv.tv_sec - now.tv_sec;
        then.tv_usec = events->tv.tv_usec - now.tv_usec;
        if (then.tv_usec < 0)
        {
            then.tv_sec -= 1;
            then.tv_usec += 1000000;
        }
        if ((then.tv_sec <= 0) && (then.tv_usec <= 0))
        {
            log (LOG_WARN, "%s: Whoa...  Scheduling for <=0 time???\n",
                 __FUNCTION__);
        }
        else
        {
            itv.it_interval = zero;
            itv.it_value = then;
            setitimer (ITIMER_REAL, &itv, NULL);
        }
    }
#endif
    cnt--;
    g_insched = 0;
}

void schedule_lock ()
{
    //while (sigprocmask (SIG_BLOCK, &alarm, NULL));
};

void schedule_unlock ()
{
#if 0
    /* See if we missed any events */
/*	alarm_handler(0); */
    while (sigprocmask (SIG_UNBLOCK, &alarm, NULL));
    raise (SIGALRM);
#endif
};

struct schedule_entry *schedule (struct timeval tv, void (*func) (void *),
                                 void *data)
{
    /* Schedule func to be run at relative time tv with data
       as arguments.  If it has already expired, run it 
       immediately.  The queue should be in order of
       increasing time */
    struct schedule_entry *p = events, *q = NULL;
    int need_timer = 0;
    struct timeval diff;

    //struct itimerval itv;
    diff = tv;
    gettimeofday (&tv, NULL);
    tv.tv_sec += diff.tv_sec;
    tv.tv_usec += diff.tv_usec;
    if (tv.tv_usec > 1000000)
    {
        tv.tv_sec++;
        tv.tv_usec -= 1000000;
    }
    while (p)
    {
        if (TVLESS (tv, p->tv))
            break;
        q = p;
        p = p->next;
    };
    if (q)
    {
        q->next =
            (struct schedule_entry *) L2TPMALLOC (__FUNCTION__, sizeof (struct schedule_entry));
        q = q->next;
    }
    else
    {
        q = (struct schedule_entry *) L2TPMALLOC (__FUNCTION__, sizeof (struct schedule_entry));
        events = q;
        need_timer = -1;
    }
    q->tv = tv;
    q->func = func;
    q->data = data;
    q->next = p;
#if 0
    if (need_timer)
    {
        itv.it_interval = zero;
        itv.it_value = diff;
        setitimer (ITIMER_REAL, &itv, NULL);

    }
#endif
    return q;

}

inline struct schedule_entry *aschedule (struct timeval tv,
                                         void (*func) (void *), void *data)
{
    /* Schedule func to be run at absolute time tv in the future with data
       as arguments */
    struct timeval now;
    gettimeofday (&now, NULL);
    tv.tv_usec -= now.tv_usec;
    if (tv.tv_usec < 0)
    {
        tv.tv_usec += 1000000;
        tv.tv_sec--;
    }
    tv.tv_sec -= now.tv_sec;
    return schedule (tv, func, data);
}

void deschedule (struct schedule_entry *s)
{
    struct schedule_entry *p = events, *q = NULL;


    if (g_insched) {
        abort();
        return;
    }
    
    if (!s)
        return;
    while (p)
    {
        if (p == s)
        {
            if (q)
            {
                q->next = p->next;
            }
            else
            {
                events = events->next;
            }
            L2TPFREE (__FUNCTION__, p);
            break;
        }
        q = p;
        p = p->next;
    }
}


/* returns 1 if there are events...*/
int
getselectdiff(struct timeval *v)
{
    struct timeval now;
    
    if (events) {
        gettimeofday(&now, NULL);
        if (TVLESSEQ(events->tv, now)) {
            // immediate timeout...
            v->tv_sec = 0;
            v->tv_usec = 0;
            return 1;
        } else {
            v->tv_sec = (events->tv.tv_sec - now.tv_sec);
            v->tv_usec = (events->tv.tv_usec - now.tv_usec);
            if (v->tv_usec < 0) {
                v->tv_sec-=1;
                v->tv_usec = (1000000 + v->tv_usec);
            }
            return 1;
        }
    }
    return 0;
}
