/*
 * arch/ppc/platforms/malbec_setup.c
 * 
 * Board setup routines for the Malbec Platform.
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/*
 * This file adds support for the Malbec Platform.
 * The malbec itself has a Broadcom 1250 network processor, an Intel 82559
 * ethernet controller, and a TI 1520 cardbus bridge.
 *
 * The interrupts are:
 *		Int	Source		Chip		Interface to 8241
 *		  0	Sibyte 		BCM1125		PCI
 *		  1	Crypto		Cavium or Hifn	PCI
 *		  2	Compact Flash	IDE		Local bus
 *		  3	FPGA		Aruba FPGA 	Local bus
 *		  4	Temp monitor	LM75		Local bus
 *
 */
#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/reboot.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/ide.h>
#include <linux/irq.h>
#include <linux/seq_file.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/time.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/prom.h>
#include <asm/smp.h>
#include <asm/keyboard.h>
#include <asm/vga.h>
#include <asm/open_pic.h>
#include <asm/todc.h>
#include <asm/bootinfo.h>
#include <asm/mpc10x.h>
#include <asm/pci-bridge.h>

#include "malbec.h"
#include "malbec_serial.h"

#include "../../../cmn/util/mal_cpld.h"

extern u_int openpic_irq(void);
extern void openpic_eoi(void);

extern int pckbd_setkeycode(unsigned int scancode, unsigned int keycode);
extern int pckbd_getkeycode(unsigned int scancode);
extern int pckbd_translate(unsigned char scancode, unsigned char *keycode,
		                           char raw_mode);
extern char pckbd_unexpected_up(unsigned char keycode);
extern void pckbd_leds(unsigned char leds);
extern void pckbd_init_hw(void);
extern unsigned char pckbd_sysrq_xlate[128];

static void	malbec_halt(void);


/*
 * *** IMPORTANT ***
 *
 * The first 16 entries of 'malbec_openpic_initsenses[]' are there and
 * initialized to 0 on purpose.  DO NOT REMOVE THEM as the 'offset' parameter
 * of 'openpic_init()' does not work for the malbec because the 8259
 * interrupt is NOT routed to the EPIC's IRQ 0 AND the EPIC's IRQ 0's offset is
 * the same as a normal openpic's IRQ 16 offset.
 */
static u_char malbec_openpic_initsenses[] __initdata = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0 - 15: reserved */
	1, /* 16: BCM1125 sibyte	= IRQ0 on the 8241 */
	1, /* 17: Crypto Processor	= IRQ1 */
	1, /* 18: Compact Flash 	= IRQ2 */
	1, /* 19: FPGA			= IRQ3 */
	1, /* 20: LM75 (not used)	= IRQ4 */
	0, /* 21: reserved */
	0, /* 22: reserved */
	0, /* 23: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, /* 24 - 31: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 32 - 47: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 48 - 63: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 64 - 79: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 80 - 95: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 96 - 111: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 112 - 127: reserved */
	0, /* 128: reserved */
	1, /* 129: I2C */
	0, /* 130: DMA CH0 */
	0, /* 131: DMA CH1 */
	0, /* 132: EPIC Timer 1 */
	0, /* 133: EPIC Timer 2 */
	0, /* 134: EPIC Timer 3 */
	0, /* 135: EPIC Timer 4 */
	0, /* 136: MU */
	1, /* 137: DUART CH1 */
	1, /* 138: DUART CH2 */
};

static void __init
malbec_setup_arch(void)
{
	extern int panic_timeout;

	/* reboot on panic in 1 second */
	panic_timeout = 1;

	loops_per_jiffy = 100000000 / HZ;

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	else
#endif
#ifdef	CONFIG_ROOT_NFS
		ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
		ROOT_DEV = to_kdev_t(0x0301);	/* /dev/hda1 IDE disk */
#endif

	/* Lookup PCI host bridges */
	malbec_find_bridges();

#ifdef CONFIG_DUMMY_CONSOLE
	conswitchp = &dummy_con;
#endif


	/* The Malbec rom doesn't enable any caches.  Do that now.
	 * The 7450 portion will also set up the L3s once I get enough
	 * information do do so.  If the processor running doesn't have
	 * and L2, the _set_L2CR is a no-op.
	 */
	if (cur_cpu_spec[0]->cpu_features & CPU_FTR_SPEC7450) {
		/* Just enable L2, the bits are different from others.
		*/
		_set_L2CR(L2CR_L2E);
	}
	else {
		/* The magic number for Malbec/74xx PrPMCs.
		*/
		_set_L2CR(0xbd014000);
	}
}

static void __init
malbec_init2(void)
{
	/* Do Malbec board specific initialization.  */

	/* write to DUART Configuration Register (DCR) to enable second UART */
	*(volatile unsigned char *)(MALBEC_EUMB_BASE + 0x4511) = 0x1;

	return;
}

/*
 * Interrupt setup and service.  Interrrupts on the Malbec come
 * from the four PCI slots plus the 8259 in the Winbond Super I/O (SIO).
 * These interrupts are sent to one of four IRQs on the EPIC.
 * The SIO shares its interrupt with either slot 2 or slot 3 (INTA#).
 * Slot numbering is confusing.  Sometimes in the documentation they
 * use 0,1,2,3 and others 1,2,3,4.  We will use slots 1,2,3,4 and
 * map this to IRQ 16, 17, 18, 19.
 * For Malbec X3, this has been better designed.  The 8259 is
 * cascaded from EPIC IRQ0, IRQ1-4 map to PCI slots 1-4, IDE is on
 * EPIC 7 and 8.
 */
static void __init
malbec_init_IRQ(void)
{
	OpenPIC_InitSenses = malbec_openpic_initsenses;
	OpenPIC_NumInitSenses = sizeof(malbec_openpic_initsenses);

	openpic_init(1, 0, NULL, -1);
}

static int
malbec_get_irq(struct pt_regs *regs)
{
        int	irq;

	irq = openpic_irq();

	if (irq == OPENPIC_VEC_SPURIOUS) {
		irq = -1;
	}

	return irq;
}

static u32
malbec_irq_cannonicalize(u32 irq)
{
	if (irq == 2)
	{
		return 9;
	}
	else
	{
		return irq;
	}
}

static ulong __init
malbec_find_end_of_memory(void)
{
	ulong	size;

	size = mpc10x_get_mem_size(MPC10X_MEM_MAP_B);

	return size;
}

static void __init
malbec_map_io(void)
{
	io_block_mapping(0xf8000000, 0xf8000000, 0x08000000, _PAGE_IO);
	io_block_mapping(0x70000000, 0x70000000, 0x01000000, _PAGE_IO);
}

static void
malbec_restart(char *cmd)
{
	__cli();

	MAL_CPLD_REG_WRITE(MAL_CPLD_BASE_ADDR, MAL_CPLD_RESET_ENABLE,
			  MAL_CPLD_RESET_ENABLE_VALUE);
	mb();
	MAL_CPLD_REG_WRITE(MAL_CPLD_BASE_ADDR, MAL_CPLD_RESET,
			  MAL_CPLD_RESET_MASTER);
	mb();

	for(;;);	/* Spin until reset happens */
}

static void
malbec_power_off(void)
{
	__cli();
	for(;;);  /* No way to shut power off with software */
	/* NOTREACHED */
}

static void
malbec_halt(void)
{
	// set Status led RED
	MAL_CPLD_REG_WRITE(MAL_CPLD_BASE_ADDR, MAL_CPLD_FPGA_INTERFACE,
                       MAL_LED_STAT_COLOR_SET(MAL_LED_COLOR_RED));

	malbec_power_off();
	/* NOTREACHED */
}

static int
malbec_show_cpuinfo(struct seq_file *m)
{
	uint pvid;

	pvid = mfspr(PVR);

	seq_printf(m, "vendor\t\t: Aruba Wireless Networks Inc.\n");
	seq_printf(m, "machine\t\t: Malbec\n");
	seq_printf(m, "processor\t: PVID: 0x%x, vendor: %s\n",
			pvid, (pvid & (1<<15) ? "IBM" : "Motorola"));

	return 0;
}
#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_IDE_MODULE)
/*
 * IDE support.
 */
static int		malbec_ide_ports_known = 0;
static ide_ioreg_t	malbec_ide_regbase[MAX_HWIFS];
static ide_ioreg_t	malbec_ide_ctl_regbase[MAX_HWIFS];
static ide_ioreg_t	malbec_idedma_regbase;

static void
malbec_ide_probe(void)
{

	malbec_ide_regbase[0]=MALBEC_COMPACT_FLASH_BASE + (0x1F0<<1);
	malbec_ide_ctl_regbase[0]=MALBEC_COMPACT_FLASH_BASE + (0x3F6<<1);
	malbec_idedma_regbase=0;
        malbec_ide_ports_known = 1;
	return;
}

static int
malbec_ide_default_irq(ide_ioreg_t base)
{
        if (malbec_ide_ports_known == 0)
	        malbec_ide_probe();

	/* Compact flash controller on malbec */
	return 18;
}

static ide_ioreg_t
malbec_ide_default_io_base(int index)
{
        if (malbec_ide_ports_known == 0)
	        malbec_ide_probe();

	return malbec_ide_regbase[index];
}

static int
malbec_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
	return check_region(from, extent);
}

static void
malbec_ide_request_region(ide_ioreg_t from,
			unsigned int extent,
			const char *name)
{
        request_region(from, extent, name);
	return;
}

static void
malbec_ide_release_region(ide_ioreg_t from,
			unsigned int extent)
{
        release_region(from, extent);
	return;
}

static void /* sameer hack __init */
malbec_ide_init_hwif_ports(hw_regs_t *hw, ide_ioreg_t data_port,
			      ide_ioreg_t ctrl_port, int *irq)
{
	ide_ioreg_t reg = data_port;
	uint	alt_status_base;
	int	i;

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg; 
		reg += 2; /* Malbec registers are 2 bytes apart */
	}

	if (data_port == malbec_ide_regbase[0]) {
		alt_status_base = malbec_ide_ctl_regbase[0];
		hw->irq = 18;
		
	} else {
		alt_status_base = 0;
		hw->irq = 0;
	}

	if (ctrl_port) {
		hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	} else {
		hw->io_ports[IDE_CONTROL_OFFSET] = alt_status_base;
	}

	if (irq != NULL) {
		*irq = hw->irq;
	}

	return;
}
#endif

/*
 * Set BAT 3 to map 0xf8000000 to end of physical memory space 1-to-1.
 */
static __inline__ void
malbec_set_bat(void)
{
	unsigned long   bat3u, bat3l;
	static int	mapping_set = 0;

	if (!mapping_set) {

		__asm__ __volatile__(
		" lis %0,0xf800\n \
		  ori %1,%0,0x002a\n \
		  ori %0,%0,0x0ffe\n \
		  mtspr 0x21e,%0\n \
		  mtspr 0x21f,%1\n \
		  isync\n \
		  sync "
		  : "=r" (bat3u), "=r" (bat3l));

		mapping_set = 1;
	}

	return;
}

#ifdef	CONFIG_SERIAL_TEXT_DEBUG
#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <asm/serial.h>

static struct serial_state rs_table[RS_TABLE_SIZE] = {
	SERIAL_PORT_DFNS	/* Defined in <asm/serial.h> */
};

static void
malbec_progress(char *s, unsigned short hex)
{
	volatile char c;
	volatile unsigned long com_port;
	u16 shift;

	com_port = rs_table[0].port;
	shift = rs_table[0].iomem_reg_shift;

	while ((c = *s++) != 0) {
		while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
		                ;
	        *(volatile unsigned char *)com_port = c;
		
		if (c == '\n') {
			while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
					;
	        	*(volatile unsigned char *)com_port = '\r';
		}
	}
}
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */

__init void malbec_setup_pci_ptrs(void);

TODC_ALLOC();

static void __init
malbec_calibrate_decr(void)
{
#if 1
        unsigned long freq;
	int divisor = 4;

	/* freq = 133333333; */
	freq = 100000000;

	tb_ticks_per_jiffy = freq / (HZ * divisor);
	tb_to_us = mulhwu_scale_factor(freq/divisor, 1000000);
#else

	todc_calibrate_decr();
#endif
}

ulong
mal_time_init(void)
{
    return 0;
}

ulong
mal_get_rtc_time(void)
{
    return 0;
}

int
mal_set_rtc_time(unsigned long nowtime)
{
    return 0;
}

void __init
platform_init(unsigned long r3, unsigned long r4, unsigned long r5,
	      unsigned long r6, unsigned long r7)
{
	parse_bootinfo(find_bootinfo());

	/* Map in board regs, etc. */
	malbec_set_bat();

	isa_io_base = 0;
	isa_mem_base = 0;
	pci_dram_offset = MPC10X_MAPB_DRAM_OFFSET;
	ISA_DMA_THRESHOLD = 0x00ffffff;
	DMA_MODE_READ = 0x44;
	DMA_MODE_WRITE = 0x48;

	ppc_md.setup_arch = malbec_setup_arch;
	ppc_md.show_cpuinfo = malbec_show_cpuinfo;
	ppc_md.irq_cannonicalize = malbec_irq_cannonicalize;
	ppc_md.init_IRQ = malbec_init_IRQ;
	ppc_md.get_irq = malbec_get_irq;
	ppc_md.init = malbec_init2;

	ppc_md.restart = malbec_restart;
	ppc_md.power_off = malbec_power_off;
	ppc_md.halt = malbec_halt;

	ppc_md.find_end_of_memory = malbec_find_end_of_memory;
	ppc_md.setup_io_mappings = malbec_map_io;

#if 0
	/* SAMEER TODO */
	TODC_INIT(TODC_TYPE_MK48T35, NULL, NULL, 0xFF000000, NULL);

	ppc_md.time_init = todc_time_init;
	ppc_md.set_rtc_time = todc_set_rtc_time;
	ppc_md.get_rtc_time = todc_get_rtc_time;
#else
	ppc_md.time_init = mal_time_init;
	ppc_md.set_rtc_time = mal_set_rtc_time;
	ppc_md.get_rtc_time = mal_get_rtc_time;
#endif

	ppc_md.calibrate_decr = malbec_calibrate_decr;

	ppc_md.nvram_read_val = todc_direct_read_val;
	ppc_md.nvram_write_val = todc_direct_write_val;

	ppc_md.heartbeat = NULL;
	ppc_md.heartbeat_reset = 0;
	ppc_md.heartbeat_count = 0;


#ifdef	CONFIG_SERIAL_TEXT_DEBUG
	ppc_md.progress = malbec_progress;
#else	/* !CONFIG_SERIAL_TEXT_DEBUG */
	ppc_md.progress = NULL;
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */

#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_IDE_MODULE)
        ppc_ide_md.default_irq =malbec_ide_default_irq;
        ppc_ide_md.default_io_base = malbec_ide_default_io_base;
        ppc_ide_md.ide_check_region = malbec_ide_check_region;
        ppc_ide_md.ide_request_region = malbec_ide_request_region;
        ppc_ide_md.ide_release_region = malbec_ide_release_region;
        ppc_ide_md.ide_init_hwif = malbec_ide_init_hwif_ports;
#endif

	return;
}
