/*
 * arch/ppc/platforms/mswitch_setup.c
 * 
 * Board setup routines for the Mswitch Platform.
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/*
 * This file adds support for the Mswitch Platform.
 * The mswitch itself has a Broadcom 1250 network processor, an Intel 82559
 * ethernet controller, and a TI 1520 cardbus bridge.
 *
 * The interrupts are:
 *		EPIC	Source		Chip
 *		  0	Sibyte 		BCM1250
 *		  1	PCI To PCMCIA 	TI1520 
 *		  2	Ethernet Ctrl	Intel82559
 *		  3	Aruba FPGA	SCCI
 *		  4	Daughter Card	
 *
 */
#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/reboot.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/ide.h>
#include <linux/irq.h>
#include <linux/seq_file.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/time.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/prom.h>
#include <asm/smp.h>
#include <asm/keyboard.h>
#include <asm/vga.h>
#include <asm/open_pic.h>
#include <asm/todc.h>
#include <asm/bootinfo.h>
#include <asm/mpc10x.h>
#include <asm/pci-bridge.h>

#include "mswitch.h"
#include "mswitch_serial.h"

#include "../../../cmn/util/sc_cpld.h"
#include "../../../cmn/util/scci.h"
#include "../../../cmn/util/lcci.h"

extern u_int openpic_irq(void);
extern void openpic_eoi(void);

extern int pckbd_setkeycode(unsigned int scancode, unsigned int keycode);
extern int pckbd_getkeycode(unsigned int scancode);
extern int pckbd_translate(unsigned char scancode, unsigned char *keycode,
		                           char raw_mode);
extern char pckbd_unexpected_up(unsigned char keycode);
extern void pckbd_leds(unsigned char leds);
extern void pckbd_init_hw(void);
extern unsigned char pckbd_sysrq_xlate[128];

static void	mswitch_halt(void);

static void
setup_nmi(void)
{
unsigned int hid0;

    asm volatile("mfspr %0,1008" : "=r" (hid0) :);
    hid0 = (hid0 | HID0_EMCP);
    asm volatile("mtspr 1008,%0" : : "r" (hid0));
}

/*
 * *** IMPORTANT ***
 *
 * The first 16 entries of 'mswitch_openpic_initsenses[]' are there and
 * initialized to 0 on purpose.  DO NOT REMOVE THEM as the 'offset' parameter
 * of 'openpic_init()' does not work for the mswitch because the 8259
 * interrupt is NOT routed to the EPIC's IRQ 0 AND the EPIC's IRQ 0's offset is
 * the same as a normal openpic's IRQ 16 offset.
 */
static u_char mswitch_openpic_initsenses[] __initdata = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0 - 15: reserved */
	1, /* 16: BCM1250 sibyte 			= IRQ0 on the 8245*/
	1, /* 17: i82559 Intel Ethernet controller 	= IRQ1 */
	1, /* 18: TI PCI1520 PCI-PCMCIA controller 	= IRQ2 */
	1, /* 19: ARUBA FPGA SCCI  			= IRQ3 */
	1, /* 20: ARUBA Daughter Card  			= IRQ4 */
	0, /* 21: reserved */
	0, /* 22: reserved */
	0, /* 23: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, /* 24 - 31: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 32 - 47: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 48 - 63: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 64 - 79: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 80 - 95: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 96 - 111: reserved */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 112 - 127: reserved */
	0, /* 128: reserved */
	1, /* 129: I2C */
	0, /* 130: DMA CH0 */
	0, /* 131: DMA CH1 */
	0, /* 132: EPIC Timer 1 */
	0, /* 133: EPIC Timer 2 */
	0, /* 134: EPIC Timer 3 */
	0, /* 135: EPIC Timer 4 */
	0, /* 136: MU */
	1, /* 137: DUART CH1 */
	1, /* 138: DUART CH2 */
};

static void __init
mswitch_setup_arch(void)
{
	extern int panic_timeout;

	/* reboot on panic in 1 second */
	panic_timeout = 1;

	loops_per_jiffy = 100000000 / HZ;

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	else
#endif
#ifdef	CONFIG_ROOT_NFS
		ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
		ROOT_DEV = to_kdev_t(0x0301);	/* /dev/hda1 IDE disk */
#endif

	/* Lookup PCI host bridges */
	mswitch_find_bridges();

#ifdef CONFIG_DUMMY_CONSOLE
	conswitchp = &dummy_con;
#endif


	/* The Mswitch rom doesn't enable any caches.  Do that now.
	 * The 7450 portion will also set up the L3s once I get enough
	 * information do do so.  If the processor running doesn't have
	 * and L2, the _set_L2CR is a no-op.
	 */
	if (cur_cpu_spec[0]->cpu_features & CPU_FTR_SPEC7450) {
		/* Just enable L2, the bits are different from others.
		*/
		_set_L2CR(L2CR_L2E);
	}
	else {
		/* The magic number for Mswitch/74xx PrPMCs.
		*/
		_set_L2CR(0xbd014000);
	}
}

static void __init
mswitch_init2(void)
{
	/* Do Mswitch board specific initialization.  */

	/* write to DUART Configuration Register (DCR) to enable second UART */
	*(volatile unsigned char *)(MSWITCH_EUMB_BASE + 0x4511) = 0x1;

	return;
}

/*
 * Interrupt setup and service.  Interrrupts on the Mswitch come
 * from the four PCI slots plus the 8259 in the Winbond Super I/O (SIO).
 * These interrupts are sent to one of four IRQs on the EPIC.
 * The SIO shares its interrupt with either slot 2 or slot 3 (INTA#).
 * Slot numbering is confusing.  Sometimes in the documentation they
 * use 0,1,2,3 and others 1,2,3,4.  We will use slots 1,2,3,4 and
 * map this to IRQ 16, 17, 18, 19.
 * For Mswitch X3, this has been better designed.  The 8259 is
 * cascaded from EPIC IRQ0, IRQ1-4 map to PCI slots 1-4, IDE is on
 * EPIC 7 and 8.
 */
static void __init
mswitch_init_IRQ(void)
{
	OpenPIC_InitSenses = mswitch_openpic_initsenses;
	OpenPIC_NumInitSenses = sizeof(mswitch_openpic_initsenses);

	openpic_init(1, 0, NULL, -1);
}

static int
mswitch_get_irq(struct pt_regs *regs)
{
        int	irq;

	irq = openpic_irq();

	if (irq == OPENPIC_VEC_SPURIOUS) {
		irq = -1;
	}

	return irq;
}

static u32
mswitch_irq_cannonicalize(u32 irq)
{
	if (irq == 2)
	{
		return 9;
	}
	else
	{
		return irq;
	}
}

static ulong __init
mswitch_find_end_of_memory(void)
{
	ulong	size;

	size = mpc10x_get_mem_size(MPC10X_MEM_MAP_B);

	return size;
}

static void __init
mswitch_map_io(void)
{
	io_block_mapping(0xfc000000, 0xfc000000, 0x04000000, _PAGE_IO);
	io_block_mapping(SC_CPLD_BASE_ADDR, SC_CPLD_BASE_ADDR,
			 0x02000000, _PAGE_IO);
}

static void
mswitch_restart(char *cmd)
{
	__cli();

	SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_RESET_ENABLE,
			  SC_CPLD_RESET_ENABLE_VALUE);
	mb();
	SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_LOCAL_RESET,
			  SC_CPLD_LOCAL_RESET_MASTER);
	mb();

	for(;;);	/* Spin until reset happens */
}

static void
mswitch_power_off(void)
{
	__cli();
	for(;;);  /* No way to shut power off with software */
	/* NOTREACHED */
}

static void
mswitch_halt(void)
{
	int lc,j;
	char data = 5;   // from platform/watchdog/watchdog.c
	unsigned long junk;
        unsigned long mod_status;
	__cli();

	/* Make the LEDs more visible to show users that we are in an unusual state. */

        /* Set SC status LED blinking RED */
        SCCI_REG_WRITE(SCCI_BASE_ADDR, SCCI_LED_CONTROL,
                       (SCCI_LED_STAT_COLOR_SET(SCCI_LED_COLOR_RED) |
                        SCCI_LED_STAT_STATE_SET(SCCI_LED_STATE_BLINK) |
                        SCCI_LED_ACTIVE_COLOR_SET(SCCI_LED_COLOR_RED) |
                        SCCI_LED_ACTIVE_STATE_SET(SCCI_LED_STATE_BLINK) |
                        SCCI_LED_POWER_COLOR_SET(SCCI_LED_COLOR_GREEN)));

        SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_MODULE_STATUS, mod_status);
	for(lc = 0; lc < SCCI_LCS; lc++) {
                if (!(mod_status & SCCI_MODULE_LC_PRES(lc))) {
                        continue;
                }
                if (mod_status & SCCI_MODULE_PSC_PRES) {
                        if (SCCI_MODULE_ID(mod_status) == SCCI_MODULE_ID_LEFT) {
                                /* left SC slot */
                                if (lc == 1) {
                                        continue; /* not owned by this SC */
                                }
                        } else {
                                /* right SC slot */
                                if (lc == 0) {
                                        continue; /* not owned by this SC */
                                }
                        }
                }
		/* set status LED blinking RED */
		LCCI_REG_WRITE(SCCI_BASE_ADDR, lc, LCCI_LED_CONTROL,
                               ((LCCI_LED_COLOR_RED << LCCI_LED_STAT_COLOR_SHIFT) |
                                (LCCI_LED_STATE_BLINK << LCCI_LED_STAT_STATE_SHIFT) |
                                (LCCI_LED_COLOR_GREEN << LCCI_LED_POWER_COLOR_SHIFT) |
                                (LCCI_LED_STATE_SOLID << LCCI_LED_POWER_STATE_SHIFT)),
                               junk);
                /* Set LC AP LEDs solid RED and POE LEDs solid YELLOW */
		for(j = 0; j < LCCI_POES; j++) {
			LCCI_REG_WRITE(SCCI_BASE_ADDR, lc, LCCI_POE_LED_0 + j, 0x33333333, junk);
                }
		for(j = 0; j < LCCI_APS; j++) {
			LCCI_REG_WRITE(SCCI_BASE_ADDR, lc, LCCI_AP_LED_0 + j, 0x11111111, junk);
		}
	}

	/* service the watchdog so we don't reboot */
	while(1) {
		SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_WATCHDOG, data);
		mb();
	}
	/* NOTREACHED */
}

static int
mswitch_show_cpuinfo(struct seq_file *m)
{
	uint pvid;

	pvid = mfspr(PVR);

	seq_printf(m, "vendor\t\t: Aruba Wireless Networks Inc.\n");
	seq_printf(m, "machine\t\t: Mswitch\n");
	seq_printf(m, "processor\t: PVID: 0x%x, vendor: %s\n",
			pvid, (pvid & (1<<15) ? "IBM" : "Motorola"));

	return 0;
}
#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_IDE_MODULE)
/*
 * IDE support.
 */
static int		mswitch_ide_ports_known = 0;
static ide_ioreg_t	mswitch_ide_regbase[MAX_HWIFS];
static ide_ioreg_t	mswitch_ide_ctl_regbase[MAX_HWIFS];
static ide_ioreg_t	mswitch_idedma_regbase;

static void
mswitch_ide_probe(void)
{

        struct pci_dev *pdev = pci_find_device(PCI_VENDOR_ID_WINBOND,
					       PCI_DEVICE_ID_WINBOND_82C105,
					       NULL);

        if(pdev) {
                mswitch_ide_regbase[0]=pdev->resource[0].start;
                mswitch_ide_regbase[1]=pdev->resource[2].start;
                mswitch_ide_ctl_regbase[0]=pdev->resource[1].start;
                mswitch_ide_ctl_regbase[1]=pdev->resource[3].start;
                mswitch_idedma_regbase=pdev->resource[4].start;
        }

        mswitch_ide_ports_known = 1;
	return;
}

static int
mswitch_ide_default_irq(ide_ioreg_t base)
{
#if 0
        if (mswitch_ide_ports_known == 0)
	        mswitch_ide_probe();

	if (base == mswitch_ide_regbase[0])
		return SANDPOINT_IDE_INT0;
	else if (base == mswitch_ide_regbase[1])
		return SANDPOINT_IDE_INT1;
	else
		return 0;
#else
	/* pcmcia controller on mswitch */
	return 18;
#endif
}

static ide_ioreg_t
mswitch_ide_default_io_base(int index)
{
        if (mswitch_ide_ports_known == 0)
	        mswitch_ide_probe();

	return mswitch_ide_regbase[index];
}

static int
mswitch_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
	return check_region(from, extent);
}

static void
mswitch_ide_request_region(ide_ioreg_t from,
			unsigned int extent,
			const char *name)
{
        request_region(from, extent, name);
	return;
}

static void
mswitch_ide_release_region(ide_ioreg_t from,
			unsigned int extent)
{
        release_region(from, extent);
	return;
}

static void /* sameer hack __init */
mswitch_ide_init_hwif_ports(hw_regs_t *hw, ide_ioreg_t data_port,
			      ide_ioreg_t ctrl_port, int *irq)
{
	ide_ioreg_t reg = data_port;
	uint	alt_status_base;
	int	i;

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg++;
	}

	if (data_port == mswitch_ide_regbase[0]) {
		alt_status_base = mswitch_ide_ctl_regbase[0] + 2;
		hw->irq = 14;
	}
	else if (data_port == mswitch_ide_regbase[1]) {
		alt_status_base = mswitch_ide_ctl_regbase[1] + 2;
		hw->irq = 15;
	}
	else {
		alt_status_base = 0;
		hw->irq = 0;
	}

	if (ctrl_port) {
		hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	} else {
		hw->io_ports[IDE_CONTROL_OFFSET] = alt_status_base;
	}

	if (irq != NULL) {
		*irq = hw->irq;
	}

	return;
}
#endif

/*
 * Set BAT 3 to map 0xf8000000 to end of physical memory space 1-to-1.
 */
static __inline__ void
mswitch_set_bat(void)
{
	unsigned long   bat3u, bat3l;
	static int	mapping_set = 0;

	if (!mapping_set) {

		__asm__ __volatile__(
		" lis %0,0xf800\n \
		  ori %1,%0,0x002a\n \
		  ori %0,%0,0x0ffe\n \
		  mtspr 0x21e,%0\n \
		  mtspr 0x21f,%1\n \
		  isync\n \
		  sync "
		  : "=r" (bat3u), "=r" (bat3l));

		mapping_set = 1;
	}

	return;
}

#ifdef	CONFIG_SERIAL_TEXT_DEBUG
#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <asm/serial.h>

static struct serial_state rs_table[RS_TABLE_SIZE] = {
	SERIAL_PORT_DFNS	/* Defined in <asm/serial.h> */
};

static void
mswitch_progress(char *s, unsigned short hex)
{
	volatile char c;
	volatile unsigned long com_port;
	u16 shift;

	com_port = rs_table[0].port;
	shift = rs_table[0].iomem_reg_shift;

	while ((c = *s++) != 0) {
		while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
		                ;
	        *(volatile unsigned char *)com_port = c;
		
		if (c == '\n') {
			while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
					;
	        	*(volatile unsigned char *)com_port = '\r';
		}
	}
}
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */

__init void mswitch_setup_pci_ptrs(void);

TODC_ALLOC();

static void __init
mswitch_calibrate_decr(void)
{
#if 1
        unsigned long freq;
	int divisor = 4;

	freq = 133333333;
	/* freq = 100000000; */

	tb_ticks_per_jiffy = freq / (HZ * divisor);
	tb_to_us = mulhwu_scale_factor(freq/divisor, 1000000);
#else

	todc_calibrate_decr();
#endif
}


void __init
platform_init(unsigned long r3, unsigned long r4, unsigned long r5,
	      unsigned long r6, unsigned long r7)
{
	parse_bootinfo(find_bootinfo());

	/* Map in board regs, etc. */
	mswitch_set_bat();

	isa_io_base = MPC10X_MAPB_ISA_IO_BASE;
	isa_mem_base = MPC10X_MAPB_ISA_MEM_BASE;
	pci_dram_offset = MPC10X_MAPB_DRAM_OFFSET;
	ISA_DMA_THRESHOLD = 0x00ffffff;
	DMA_MODE_READ = 0x44;
	DMA_MODE_WRITE = 0x48;

	ppc_md.setup_arch = mswitch_setup_arch;
	ppc_md.show_cpuinfo = mswitch_show_cpuinfo;
	ppc_md.irq_cannonicalize = mswitch_irq_cannonicalize;
	ppc_md.init_IRQ = mswitch_init_IRQ;
	ppc_md.get_irq = mswitch_get_irq;
	ppc_md.init = mswitch_init2;

	ppc_md.restart = mswitch_restart;
	ppc_md.power_off = mswitch_power_off;
	ppc_md.halt = mswitch_halt;

	ppc_md.find_end_of_memory = mswitch_find_end_of_memory;
	ppc_md.setup_io_mappings = mswitch_map_io;

	TODC_INIT(TODC_TYPE_MK48T35, NULL, NULL, 0xFF000000, NULL);

	ppc_md.time_init = todc_time_init;
	ppc_md.set_rtc_time = todc_set_rtc_time;
	ppc_md.get_rtc_time = todc_get_rtc_time;

	ppc_md.calibrate_decr = mswitch_calibrate_decr;

	ppc_md.nvram_read_val = todc_direct_read_val;
	ppc_md.nvram_write_val = todc_direct_write_val;

	ppc_md.heartbeat = NULL;
	ppc_md.heartbeat_reset = 0;
	ppc_md.heartbeat_count = 0;

#ifdef	CONFIG_SERIAL_TEXT_DEBUG
	ppc_md.progress = mswitch_progress;
#else	/* !CONFIG_SERIAL_TEXT_DEBUG */
	ppc_md.progress = NULL;
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */

#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_IDE_MODULE)
        ppc_ide_md.default_irq =mswitch_ide_default_irq;
        ppc_ide_md.default_io_base = mswitch_ide_default_io_base;
        ppc_ide_md.ide_check_region = mswitch_ide_check_region;
        ppc_ide_md.ide_request_region = mswitch_ide_request_region;
        ppc_ide_md.ide_release_region = mswitch_ide_release_region;
        ppc_ide_md.ide_init_hwif = mswitch_ide_init_hwif_ports;
#endif


        setup_nmi();

	return;
}
