/*
 * arch/ppc/platforms/sap_setup.c
 * 
 * Board setup routines for the SAP Platform.
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/*
 * This file adds support for the SAP Platform.
 * The SAP itself has a Broadcom 1250 network processor, an Intel 82559
 * ethernet controller, and a TI 1520 cardbus bridge.
 *
 * The interrupts are:
 *		EPIC	Source
 *		  0	
 *		  1	PCI #1
 *		  2	PCI #2
 *		  3	PCI #3
 *		  4	PCI #4
 *
 */
#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/reboot.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/ide.h>
#include <linux/irq.h>
#include <linux/seq_file.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/time.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/prom.h>
#include <asm/smp.h>
#include <asm/keyboard.h>
#include <asm/vga.h>
#include <asm/open_pic.h>
#ifndef CONFIG_ASAP
#include <asm/i8259.h>
#endif
#include <asm/todc.h>
#include <asm/bootinfo.h>
#include <asm/mpc10x.h>
#include <asm/pci-bridge.h>

#include "sap.h"

extern u_int openpic_irq(void);
extern void openpic_eoi(void);

extern int pckbd_setkeycode(unsigned int scancode, unsigned int keycode);
extern int pckbd_getkeycode(unsigned int scancode);
extern int pckbd_translate(unsigned char scancode, unsigned char *keycode,
		                           char raw_mode);
extern char pckbd_unexpected_up(unsigned char keycode);
extern void pckbd_leds(unsigned char leds);
extern void pckbd_init_hw(void);
extern unsigned char pckbd_sysrq_xlate[128];

static void	sap_halt(void);
int asap_model = ASAP_AP52;

/*
 * *** IMPORTANT ***
 *
 * The first 16 entries of 'switch_openpic_initsenses[]' are there and
 * initialized to 0 on purpose.  DO NOT REMOVE THEM as the 'offset' parameter
 * of 'openpic_init()' does not work for the ASAP.
 */
static u_char sap_openpic_initsenses[] __initdata = {
    0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, /* 0 - 15: reserved */
    1, 1, 1, 1, /* 16 - 19 used for wlan interfaces + ethernet(18) */
    1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 20 used by airespace AP1200 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 
    1, /* 137: DUART CH1 */
    1, /* 138: DUART CH2 */
};

static void __init
sap_setup_arch(void)
{
    loops_per_jiffy = 100000000 / HZ;

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	else
#endif
#ifdef	CONFIG_ROOT_NFS
		ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
		ROOT_DEV = to_kdev_t(0x0301);	/* /dev/hda1 IDE disk */
#endif

	/* Lookup PCI host bridges */
	sap_find_bridges();

#ifndef CONFIG_ASAP
#ifdef CONFIG_DUMMY_CONSOLE
	conswitchp = &dummy_con;
#endif
#endif
}

static void __init
sap_init2(void)
{
	
    /* Do SAP board specific initialization.  */

    /* write to DUART Configuration Register (DCR) to enable second UART */
    *(volatile unsigned char *)(SAP_EUMB_BASE + 0x4511) = 0x1;
    *(volatile unsigned char *)(SAP_EUMB_BASE + 0x4611) = 0x1;

    return;
}

/*
 * Interrupt setup and service.  Interrrupts on the SAP come
 * from the four PCI slots (17-20).
 * These interrupts are sent to one of four IRQs on the EPIC.
 * Slot numbering is confusing.  Sometimes in the documentation they
 * use 0,1,2,3 and others 1,2,3,4.  We will use slots 1,2,3,4 and
 * map this to IRQ 16, 17, 18, 19.
 */
static void __init
sap_init_IRQ(void)
{
    /*
	 * 3 things cause us to jump through some hoops:
	 *   1) the EPIC on the 8240 & 107 are not full-blown openpic pic's
	 *   2) the 8259 is NOT cascaded on the openpic IRQ 0
	 *   3) the 8259 shares its interrupt line with some PCI interrupts.
	 *
	 * What we'll do is set up the 8259 to be level sensitive, active low
	 * just like a PCI device.  Then, when an interrupt on the IRQ that is
	 * shared with the 8259 comes in, we'll take a peek at the 8259 to see
	 * it its generating an interrupt.  If it is, we'll handle the 8259
	 * interrupt.  Otherwise, we'll handle it just like a normal PCI
	 * interrupt.  This does give the 8259 interrupts a higher priority
	 * than the EPIC ones--hopefully, not a problem.
	 */
	OpenPIC_InitSenses = sap_openpic_initsenses;
	OpenPIC_NumInitSenses = sizeof(sap_openpic_initsenses);

    openpic_init(1, 0, NULL, -1);
   
}

static int
sap_get_irq(struct pt_regs *regs)
{
    int	irq;
    irq = openpic_irq();
    
    if (irq == OPENPIC_VEC_SPURIOUS) {
       //printk("SUPRIOURS INTER. irq=%d\n", irq);
       irq = -1;
   }
   
   return irq;
}

static u32
sap_irq_cannonicalize(u32 irq)
{
    if (irq == 2)
	{
        return 9;
    }
	else
	{
        return irq;
	}
}

static ulong __init
sap_find_end_of_memory(void)
{
    ulong	size;

    size = mpc10x_get_mem_size(MPC10X_MEM_MAP_B);

    return size;
}

static void __init
sap_map_io(void)
{
#ifdef ARUBA_LINUX_BIN
  // needed to build the initial image to boot from VxWorks boot loader
  io_block_mapping(SAP_CPLD_LED_BASE, SAP_CPLD_LED_BASE,
                   0x20000, _PAGE_IO);
  io_block_mapping(SAP_CPLD_WDG_BASE, SAP_CPLD_WDG_BASE,
		   0x20000, _PAGE_IO);
#else
  io_block_mapping(SAP_CPLD_LED_BASE, SAP_CPLD_LED_BASE,
                   0x20000, _PAGE_IO);
#endif
  /* increase window to cover FLASH and RCS1 (AP1200 wdog),
   * not just PPC registers */
  io_block_mapping(SAP_EUMB_BASE, SAP_EUMB_BASE,
                   0x04000000, _PAGE_IO); 
}

void(*fResetVector)(void);

static void
sap_restart(char *cmd)
{

  volatile __u8 *cpld = (volatile __u8 *)SAP_CPLD_WDG_BASE;

  printk("sap_restart: Rebooting...");

  *cpld = 0xff;
  *cpld = 0xbf;
  *cpld = 0xff;
  *cpld = 0xbf;
	__cli();
  for(;;);	/* Spin until reset happens */
}

static void
sap_power_off(void)
{
    printk("Powering SAP off...<stopping interrupts>");
	__cli();
    printk("Done.\n");
	for(;;);  /* No way to shut power off with software */
	/* NOTREACHED */                   }

static void
sap_halt(void)
{   
    printk("Halting SAP...");
    sap_power_off();
    printk("Done\n");
	/* NOTREACHED */
}

static int
sap_show_cpuinfo(struct seq_file *m)
{
#ifdef ARUBA_DEBUG
	uint pvid;
	pvid = mfspr(PVR);
	printk("\nvendor\t\t: Aruba Wireless Networks Inc.\n");
	printk("machine\t\t: ASAP\n");
	printk("processor\t: PVID: 0x%x\n", pvid);
#endif
	return 0;
}

/*
 * Set BAT 3 to map 0xf8000000 to end of physical memory space 1-to-1.
 */
static __inline__ void
sap_set_bat(void)
{
	unsigned long   bat3u, bat3l;
	static int	mapping_set = 0;

	if (!mapping_set) {

		__asm__ __volatile__(
		" lis %0,0xf800\n \
		  ori %1,%0,0x002a\n \
		  ori %0,%0,0x0ffe\n \
		  mtspr 0x21e,%0\n \
		  mtspr 0x21f,%1\n \
		  isync\n \
		  sync "
		  : "=r" (bat3u), "=r" (bat3l));

		mapping_set = 1;
	}

	return;
}

#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <asm/serial.h>
static struct serial_state rs_table[RS_TABLE_SIZE] = {
	SERIAL_PORT_DFNS	/* Defined in <asm/serial.h> */
};


#ifdef	CONFIG_SERIAL_TEXT_DEBUG
static void
sap_progress(char *s, unsigned short hex)
{
	volatile char c;
	volatile unsigned long com_port;
	u16 shift;

	com_port = rs_table[0].port;
	shift = rs_table[0].iomem_reg_shift;

	while ((c = *s++) != 0) {
		while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
		                ;
	        *(volatile unsigned char *)com_port = c;
		
		if (c == '\n') {
			while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
					;
	        	*(volatile unsigned char *)com_port = '\r';
		}
	}
}
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */

__init void mswitch_setup_pci_ptrs(void);

static void
sap_console_write(struct console *co, const char *ptr,
            unsigned nb)
{ 
    int i;
	volatile char c;
	volatile unsigned long com_port;
	u16 shift;

	com_port = rs_table[0].port;
	shift = rs_table[0].iomem_reg_shift;

	for (i = 0; i < nb; ++i) {
	  c = ptr[i];
	  
	  while ((*((volatile unsigned char *)com_port +
		    (UART_LSR << shift)) & UART_LSR_THRE) == 0)
	    ;
	  *(volatile unsigned char *)com_port = c;
	  
	  if (c == '\n') {
	    while ((*((volatile unsigned char *)com_port +
		      (UART_LSR << shift)) & UART_LSR_THRE) == 0)
	      ;
	    *(volatile unsigned char *)com_port = '\r';
	    
	  }
	  
	}
	return;

}
          
static struct console sap_sercons = {
	name:		"SapCon",
	write:		sap_console_write,
        flags:		CON_PRINTBUFFER,
	index:		-1,
};

#define PPC8241_PVID 0x80811014
#define PPC8245_PVID 0x80811015

TODC_ALLOC();
static void __init 
sap_calibrate_decr(void)
{
  unsigned long freq;
  int divisor = 4;
  uint pvid = mfspr(PVR);
  int is_8245 = (pvid == PPC8245_PVID);
  

  if (is_8245){
    freq = 133333333;
  } else {
    //freq = powerpmc250_get_bus_speed();
    switch (asap_model) {
    case ASAP_AP1200:
      freq = 75000000;
      break;
    case ASAP_AP52:
    default:
      freq = 100000000; // SCA; updated from 96 Mhz to 100 Mhz per Thomas
      break;
    }
  }
  
  tb_ticks_per_jiffy = freq / (HZ * divisor);
  tb_to_us = mulhwu_scale_factor(freq/divisor, 1000000);
}



void __init
platform_init(unsigned long r3, unsigned long r4, unsigned long r5,
	      unsigned long r6, unsigned long r7)
{
  parse_bootinfo(find_bootinfo());
  char signature[32];
  
  /* Map in board regs, etc. */
  sap_set_bat();

  memcpy(signature, AP1200_SIG_ADDR, sizeof(signature)-1);
  signature[31] = 0;
  if (strstr(signature, AP1200_SIG)) {
      extern char *linux_model;
      asap_model = ASAP_AP1200;
      linux_model = "1200";
  }
#if 0
  isa_io_base = MPC10X_MAPB_ISA_IO_BASE;
  isa_mem_base = MPC10X_MAPB_ISA_MEM_BASE;
#endif
  pci_dram_offset = MPC10X_MAPB_DRAM_OFFSET;
  ISA_DMA_THRESHOLD = 0x00ffffff;
  DMA_MODE_READ = 0x44;
  DMA_MODE_WRITE = 0x48;
  
  ppc_md.setup_arch = sap_setup_arch;
  ppc_md.show_cpuinfo = sap_show_cpuinfo;
  ppc_md.irq_cannonicalize = sap_irq_cannonicalize;
  ppc_md.init_IRQ = sap_init_IRQ;
  ppc_md.get_irq = sap_get_irq;
  ppc_md.init = sap_init2;
  
  ppc_md.restart = sap_restart;
  ppc_md.power_off = sap_power_off;
  ppc_md.halt = sap_halt;
  
  ppc_md.find_end_of_memory = sap_find_end_of_memory;
  ppc_md.setup_io_mappings = sap_map_io;
  
  ppc_md.time_init = NULL;
  ppc_md.set_rtc_time = NULL;
  ppc_md.get_rtc_time = NULL;
  ppc_md.calibrate_decr = sap_calibrate_decr;
  
  ppc_md.nvram_read_val = NULL;
  ppc_md.nvram_write_val = NULL;
  
  ppc_md.heartbeat = NULL;
  ppc_md.heartbeat_reset = 0;
  ppc_md.heartbeat_count = 0;
  
#ifdef	CONFIG_SERIAL_TEXT_DEBUG
  ppc_md.progress = sap_progress;
#else	/* !CONFIG_SERIAL_TEXT_DEBUG */
  ppc_md.progress = NULL;
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */
  
  //register_console(&sap_sercons); 
  
  // unregistering this console will disable
  // output of initial bootup msgs..keep out for now!
  //unregister_console(&sap_sercons);
  
  return;
}



