#define	Z_WAKE

/*
 *  linux/drivers/char/malbec.c
 *
 * This file contains the driver for the malbec switch
 * that contain serial ports on ethernet ports.
 *
 * Based upon cyclades.c
 *
 * Much of the design and some of the code came from serial.c
 * which was copyright (C) 1991, 1992  Linus Torvalds.  It was
 * extensively rewritten by Theodore Ts'o, 8/16/92 -- 9/14/92,
 * and then fixed as suggested by Michael K. Johnson 12/12/92.
 */

/* If you need to install more boards than NR_CARDS, change the constant
   in the definition below. No other change is necessary to support up to
   eight boards. Beyond that you'll have to extend mal_isa_addresses. */

#define NR_CARDS        1 /* not to be confused with line cards */
                          /* there is only one FPGA, and hence 1 card */

/*
   If the total number of ports is larger than NR_PORTS, change this
   constant in the definition below. No other change is necessary to
   support more boards/ports. */
#define NR_PORTS        1

#undef	MAL_DEBUG_OPEN
#undef	MAL_DEBUG_THROTTLE
#undef	MAL_DEBUG_OTHER
#undef	MAL_DEBUG_IO
#undef	MAL_DEBUG_IOCTL
#undef	MAL_DEBUG_COUNT
#undef	MAL_DEBUG_DTR
#undef	MAL_DEBUG_WAIT_UNTIL_SENT
#undef	MAL_DEBUG_INTERRUPTS
#undef	MAL_DEBUG_NVRAM
#undef	MAL_ENABLE_MONITORING

/*
 * Include section 
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/malbec.h>
#include <linux/compile.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#define	MAL_LOCK(info,flags) \
    do { \
	spin_lock_irqsave(&mal_card[info->card].card_lock, flags); \
    } while (0)
		
#define	MAL_UNLOCK(info,flags) \
    do { \
	spin_unlock_irqrestore(&mal_card[info->card].card_lock, flags); \
    } while (0)

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/version.h>

#include <linux/stat.h>
#include <linux/proc_fs.h>

#include "../../../cmn/util/mal_cpld.h"
#include "../../../cmn/util/util_arp.h"
#include "../../../cmn/util/mxc_ioctl.h"

#ifndef MIN
#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#endif

#ifndef SERIAL_XMIT_SIZE
#define	SERIAL_XMIT_SIZE	(MIN(PAGE_SIZE, 4096))
#endif
#define WAKEUP_CHARS		256
#define STD_COM_FLAGS (0)

static DECLARE_TASK_QUEUE(tq_malbec);

static struct tty_driver mal_driver;
static int serial_refcount;

/* This is the per-card data structure containing address, irq, number of
   channels, etc. This driver supports a maximum of NR_CARDS cards.
*/
static struct malbec_card mal_card[NR_CARDS];

/* This is the per-channel data structure containing pointers, flags
 and variables for the port. This driver supports a maximum of NR_PORTS.
*/
static struct malbec_port mal_port[NR_PORTS];
static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

static unsigned short	mal_nboard;
static void mal_start(struct tty_struct *);

static int malbec_get_proc_info(char *, char **, off_t , int , int *, void *);

/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver
 * (also known as the "bottom half").  This can be called any
 * number of times for any channel without harm.
 */
static inline void
mal_sched_event(struct malbec_port *info, int event)
{
    info->event |= 1 << event; /* remember what kind of event and who */
    queue_task(&info->tqueue, &tq_malbec); /* it belongs to */
    mark_bh(CYCLADES_BH);                   /* then trigger event */
} /* mal_sched_event */


/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * mx#/_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using mal_sched_event(), and they get done here.
 *
 * This is done through one level of indirection--the task queue.
 * When a hardware interrupt service routine wants service by the
 * driver's bottom half, it enqueues the appropriate tq_struct (one
 * per port) to the tq_malbec work queue and sets a request flag
 * via mark_bh for processing that queue.  When the time is right,
 * do_malbec_bh is called (because of the mark_bh) and it requests
 * that the work queue be processed.
 *
 * Although this may seem unwieldy, it gives the system a way to
 * pass an argument (in this case the pointer to the malbec_port
 * structure) to the bottom half of the driver.  Previous kernels
 * had to poll every port to see if that port needed servicing.
 */
static void
do_malbec_bh(void)
{
    run_task_queue(&tq_malbec);
} /* do_malbec_bh */

static void
do_softint(void *private_)
{
  struct malbec_port *info = (struct malbec_port *) private_;
  struct tty_struct    *tty;

    tty = info->tty;
    if (!tty)
        return;

    if (test_and_clear_bit(MAL_EVENT_HANGUP, &info->event)) {
        tty_hangup(info->tty);
        wake_up_interruptible(&info->open_wait);
        info->flags &= ~(ASYNC_NORMAL_ACTIVE|
                             ASYNC_CALLOUT_ACTIVE);
    }
    if (test_and_clear_bit(MAL_EVENT_OPEN_WAKEUP, &info->event)) {
        wake_up_interruptible(&info->open_wait);
    }
    if (test_and_clear_bit(MAL_EVENT_WRITE_WAKEUP, &info->event)) {
        if((tty->flags & (1<< TTY_DO_WRITE_WAKEUP))
        && tty->ldisc.write_wakeup){
            (tty->ldisc.write_wakeup)(tty);
        }
        wake_up_interruptible(&tty->write_wait);
    }
#ifdef Z_WAKE
    if (test_and_clear_bit(MAL_EVENT_SHUTDOWN_WAKEUP, &info->event)) {
        wake_up_interruptible(&info->shutdown_wait);
    }
#endif
} /* do_softint */


/* The real interrupt service routine is called
   whenever the card wants its hand held--chars
   received, out buffer empty, modem change, etc.
 */
static void
mal_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    struct malbec_card *cinfo;

    if ((cinfo = (struct malbec_card *)dev_id) == 0) {
#ifdef MAL_DEBUG_INTERRUPTS
	printk("mal_interrupt: spurious interrupt %d\n\r", irq);
#endif
        return; /* spurious interrupt */
    }
} /* mal_interrupt */

/*
 * This routine shuts down a serial port; interrupts are disabled,
 * and DTR is dropped if the hangup on close termio flag is on.
 */
static void
shutdown(struct malbec_port * info)
{
    info->flags &= ~ASYNC_INITIALIZED;

#ifdef MAL_DEBUG_OPEN
    printk(" mxc: shutdown done\n");
#endif
    return;
} /* shutdown */


/*
 * ------------------------------------------------------------
 * mal_open() and friends
 * ------------------------------------------------------------
 */

/*
 * This routine is called whenever a serial port is opened.  It
 * performs the serial-specific initialization for the tty structure.
 */
static int
mal_open(struct tty_struct *tty, struct file * filp)
{
    struct malbec_port  *info;
    int line;

    MOD_INC_USE_COUNT;
    line = MINOR(tty->device) - tty->driver.minor_start;
    if ((line < 0) || (NR_PORTS <= line)){
        MOD_DEC_USE_COUNT;
        return -ENODEV;
    }
    info = &mal_port[line];
    if (info->line < 0) {
        MOD_DEC_USE_COUNT;
        return -ENODEV;
    }

    if ((MAJOR(tty->device) == MALBEC_MAJOR) &&
	!strcmp(tty->driver.name, mal_driver.name)) {
#ifdef MAL_DEBUG_OPEN
        printk("mxc: mal_open mxc %d %s\n", MALBEC_MAJOR, tty->driver.name);
#endif
	tty->driver_data = info;
	return 0;
    }

    tty->driver_data = info;
    info->tty = tty;

    if ((info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS)) {
        *tty->termios = info->normal_termios;
    }

    info->session = current->session;
    info->pgrp = current->pgrp;

#ifdef MAL_DEBUG_OPEN
    printk(" mxc: mal_open done\n");/**/
#endif

    return 0;
} /* mal_open */


/*
 * mal_wait_until_sent() --- wait until the transmitter is empty
 */
static void 
mal_wait_until_sent(struct tty_struct *tty, int timeout)
{
    struct malbec_port * info = (struct malbec_port *)tty->driver_data;
    unsigned long orig_jiffies, char_time;
	
    if (info->xmit_fifo_size == 0)
	return; /* Just in case.... */


    orig_jiffies = jiffies;
    /*
     * Set the check interval to be 1/5 of the estimated time to
     * send a single character, and make it at least 1.  The check
     * interval should also be less than the timeout.
     * 
     * Note: we have to use pretty tight timings here to satisfy
     * the NIST-PCTS.
     */
    char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
    char_time = char_time / 5;
    if (char_time == 0)
	char_time = 1;
    if (timeout < 0)
	timeout = 0;
    if (timeout)
	char_time = MIN(char_time, timeout);
    /*
     * If the transmitter hasn't cleared in twice the approximate
     * amount of time to send the entire FIFO, it probably won't
     * ever clear.  This assumes the UART isn't doing flow
     * control, which is currently the case.  Hence, if it ever
     * takes longer than info->timeout, this is probably due to a
     * UART bug of some kind.  So, we clamp the timeout parameter at
     * 2*info->timeout.
     */
    if (!timeout || timeout > 2*info->timeout)
	timeout = 2*info->timeout;
#ifdef MAL_DEBUG_WAIT_UNTIL_SENT
    printk("In mal_wait_until_sent(%d) check=%lu...", timeout, char_time);
    printk("jiff=%lu...", jiffies);
#endif

    /* Run one more char cycle */
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(char_time * 5);
    current->state = TASK_RUNNING;
#ifdef MAL_DEBUG_WAIT_UNTIL_SENT
    printk("Clean (jiff=%lu)...done\n", jiffies);
#endif
}

/*
 * This routine is called when a particular tty device is closed.
 */
static void
mal_close(struct tty_struct *tty, struct file *filp)
{
    struct malbec_port *info = (struct malbec_port *)tty->driver_data;

#ifdef MAL_DEBUG_OTHER
    printk("mxc: mal_close mxc%d\n", info->line);
#endif

    if ((MAJOR(tty->device) == MALBEC_MAJOR) &&
	!strcmp(tty->driver.name, mal_driver.name)) {
	MOD_DEC_USE_COUNT;
        return;
    }

    /* If the TTY is being hung up, nothing to do */
    if (tty_hung_up_p(filp)) {
	MOD_DEC_USE_COUNT;
        return;
    }
        
#ifdef MAL_DEBUG_OPEN
    printk("mxc: mal_close mxc%d, count = %d\n", info->line, info->count);
#endif
    if ((tty->count == 1) && (info->count != 1)) {
        /*
         * Uh, oh.  tty->count is 1, which means that the tty
         * structure will be freed.  Info->count should always
         * be one in these conditions.  If it's greater than
         * one, we've got real problems, since it means the
         * serial port won't be shutdown.
         */
        printk("mxc: mal_close: bad serial port count; tty->count is 1, "
           "info->count is %d\n", info->count);
        info->count = 1;
    }
#ifdef MAL_DEBUG_COUNT
    printk("mxc: mal_close at (%d): decrementing count to %d\n",
        current->pid, info->count - 1);
#endif
    if (--info->count < 0) {
#ifdef MAL_DEBUG_COUNT
        printk("mxc: mxc_close setting count to 0\n");
#endif
        info->count = 0;
    }
    if (info->count) {
	MOD_DEC_USE_COUNT;
        return;
    }
    info->flags |= ASYNC_CLOSING;
    /*
     * Save the termios structure, since this port may have
     * separate termios for callout and dialin.
     */
    if (info->flags & ASYNC_NORMAL_ACTIVE)
        info->normal_termios = *tty->termios;

    /*
    * Now we wait for the transmit buffer to clear; and we notify
    * the line discipline to only process XON/XOFF characters.
    */
    tty->closing = 1;
    if (info->closing_wait != MAL_CLOSING_WAIT_NONE) {
	tty_wait_until_sent(tty, info->closing_wait);
    }
    {
	if (info->flags & ASYNC_INITIALIZED) {
	    /* Waiting for on-board buffers to be empty before closing 
	       the port */
	    mal_wait_until_sent(tty, info->timeout);
	}
    }

    shutdown(info);
    if (tty->driver.flush_buffer)
        tty->driver.flush_buffer(tty);
    if (tty->ldisc.flush_buffer)
        tty->ldisc.flush_buffer(tty);

    tty->closing = 0;
    info->event = 0;
    info->tty = 0;
    if (info->blocked_open) {
        if (info->close_delay) {
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(info->close_delay);
        }
        wake_up_interruptible(&info->open_wait);
    }
    info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
                     ASYNC_CLOSING);
    wake_up_interruptible(&info->close_wait);

#ifdef MAL_DEBUG_OTHER
    printk(" mxc: mal_close done\n");
#endif

    MOD_DEC_USE_COUNT;
    return;
} /* mal_close */


/* This routine gets called when tty_write has put something into
 * the write_queue.  The characters may come from user space or
 * kernel space.
 *
 * This routine will return the number of characters actually
 * accepted for writing.
 *
 * If the port is not already transmitting stuff, start it off by
 * enabling interrupts.  The interrupt service routine will then
 * ensure that the characters are sent.
 * If the port is already active, there is no need to kick it.
 *
 */
static int
mal_write(struct tty_struct * tty, int from_user,
	 const unsigned char *buf, int count)
{
    return 0;
} /* mal_write */

/*
 * This routine is called by the kernel to write a single
 * character to the tty device.  If the kernel uses this routine,
 * it must call the flush_chars() routine (if defined) when it is
 * done stuffing characters into the driver.  If there is no room
 * in the queue, the character is ignored.
 */
static void
mal_put_char(struct tty_struct *tty, unsigned char ch)
{
    struct malbec_port *info = (struct malbec_port *)tty->driver_data;

#ifdef MAL_DEBUG_IO
    printk("mxc: mal_put_char mxc%d\n", info->line);
#endif

    if (!tty || !info->xmit_buf)
        return;

    if (info->xmit_cnt >= SERIAL_XMIT_SIZE - 1) {
	return;
    }

    info->xmit_buf[info->xmit_head++] = ch;
    info->xmit_head &= SERIAL_XMIT_SIZE - 1;
    info->xmit_cnt++;
    info->idle_stats.xmit_bytes++;
    info->idle_stats.xmit_idle = jiffies;
} /* mal_put_char */


static void
mal_flush_chars(struct tty_struct *tty)
{
} /* mal_flush_chars */

static int
mal_write_room(struct tty_struct *tty)
{
    return 0;
} /* mal_write_room */


static int
mal_chars_in_buffer(struct tty_struct *tty)
{
    return 0;
} /* mal_chars_in_buffer */


/*
 * ------------------------------------------------------------
 * mal_ioctl() and friends
 * ------------------------------------------------------------
 */

static void
mal_break(struct tty_struct *tty, int break_state)
{
} /* mal_break */

static int
get_mon_info(struct malbec_port * info, struct malbec_monitor * mon)
{

    if(copy_to_user(mon, &info->mon, sizeof(struct malbec_monitor)))
    	return -EFAULT;
    info->mon.int_count  = 0;
    info->mon.char_count = 0;
    info->mon.char_max   = 0;
    info->mon.char_last  = 0;
    return 0;
}/* get_mon_info */

static int
mal_cpld_read(struct malbec_port * info,
	     mal_cpld_ioctl_t *u)
{
    mal_cpld_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    MAL_CPLD_REG_READ(mal_card[info->card].ctl_addr, k.reg, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
mal_cpld_write(struct malbec_port * info,
	      mal_cpld_ioctl_t *u)
{
    mal_cpld_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    MAL_CPLD_REG_WRITE(mal_card[info->card].ctl_addr, k.reg, k.data);
    return 0;
}

static int
mal_arp_add(struct malbec_port * info, arp_entry_t *u)
{   
    return (arp_entry_add(u));
}

static int
mal_arp_del(struct malbec_port * info, arp_entry_t *u)
{   
    return (arp_entry_del(u));
}

static int
mal_arp_clear_all(struct malbec_port * info, arp_entry_t *u)
{   
    return (arp_clear_all());
}

static int
mal_sw_version(struct malbec_port * info,
	       mxc_string_t *u)
{
    mxc_string_t k;
    if (strlen(ARUBA_VERSION)+1 >= sizeof(k.string))
        return -EFAULT;
    sprintf(k.string, "%s", ARUBA_VERSION);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mal_lc_status(struct malbec_port * info,
	      mxc_lc_t *u)
{
    mxc_lc_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.lc > MXC_MAX_LCS)
	return -ENODEV;
    k.data = (MXC_LC_PRES |
	      MXC_LC_SC_PRES(0) |
	      MXC_LC_SC_ACTIVE(0));
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mal_lc_switch_rev(struct malbec_port * info, mxc_lc_t *u)
{
    mxc_lc_t k;

    if (copy_from_user(&k, u,sizeof(k)))
        return -EFAULT;

    if (k.lc > MXC_MAX_LCS)
        return -ENODEV;

    k.data = 0;
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
mal_fan_status(struct malbec_port * info,
	       mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    k.data = 0;
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mal_ps_status(struct malbec_port * info,
	      mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    int i;
    k.data = 0;
    for (i = 0; i < MXC_MAX_PSS; i++) {
        k.data |= MXC_PS_PRES(i) | MXC_PS_OK(i);
    }
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mal_temperature(struct malbec_port * info,
		mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    k.data = 0; /* XXX i2c */
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mal_sc_slot(struct malbec_port * info,
	    mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    k.data = 0;
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mal_sc_id(struct malbec_port * info,
	    mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    k.data = 0;
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mal_lc_id(struct malbec_port * info,
	    mxc_ioctl_t *u)
{
    mxc_lc_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.lc > MXC_MAX_LCS)
	return -ENODEV;
    k.data = 0x100;
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mal_wd_reset(struct malbec_port * info)
{
    /* XXX i2c */
    return 0;
}

#ifdef MAL_DEBUG_NVRAM
static void 
mal_print_sector(volatile char *buffer)
{
	int i;
	for(i=0;i<MXC_FLASH_SECTOR_SIZE;i++) {
		if((i%16)==0) printk("\n0x%08x: ", (unsigned int)&buffer[i]);
		printk("%02x ", (unsigned char)buffer[i]);
	}
	printk("\n");
}
#endif

static int
mal_nvram_read(struct malbec_port * info,
	       mxc_nvram_t *u)
{
    mxc_nvram_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.offset + k.length > NVRAM_SIZE)
	return -EFAULT;
    if (copy_to_user(k.data, (char *)(NVRAM_BASE_ADDRESS + k.offset), k.length))
        return -EFAULT;
    return 0;
}

void
mal_flash_write(char *flash_addr, char *mem_addr, int len)
{
    int i;
    char buffer[MXC_FLASH_SECTOR_SIZE];
    volatile char *flash_sect = (char *)((long)flash_addr & ~(MXC_FLASH_SECTOR_SIZE-1));
    volatile char *flash_base = (char *)MXC_FLASH_BASE_ADDR;
    int copy_len;

    do {
	copy_len = MIN(len, MXC_FLASH_SECTOR_SIZE - (flash_addr - flash_sect));
	if (copy_len < MXC_FLASH_SECTOR_SIZE) {
	    /* read entire flash sector if we are only updating some of it */
	    for(i=0;i<MXC_FLASH_SECTOR_SIZE;i++) buffer[i]=flash_sect[i];
	}

#ifdef MAL_DEBUG_NVRAM
	 printk("READ VALUES:\n"); mal_print_sector(flash_sect);
#endif
        /* update buffer with write data */
	memcpy(&buffer[flash_addr - flash_sect], mem_addr, copy_len);

	/* tell flash we want to program one sector */
	flash_base[0x5555] = 0xaa;
	flash_base[0x2aaa] = 0x55;
	flash_base[0x5555] = 0xa0;
        /* program the sector */
	for(i=0;i<MXC_FLASH_SECTOR_SIZE;i++) {
		flash_sect[i]=buffer[i];
	}
	/* wait for the Write Cycle time before moving to next sector */
	udelay(20000);

#ifdef MAL_DEBUG_NVRAM
	printk("AFTER WRITE: VALUES ARE:\n"); mal_print_sector(flash_sect);
#endif

        /* advance ptrs for additional sectors if any */
	len -= copy_len;
	flash_addr += copy_len;
	flash_sect += MXC_FLASH_SECTOR_SIZE;
	mem_addr += copy_len;
    } while (len);
}

static int
mal_nvram_write(struct malbec_port * info,
		mxc_nvram_t *u)
{
    mxc_nvram_t k;
    char buf[256];
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.offset + k.length > NVRAM_SIZE)
	return -EFAULT;
    if (k.length > sizeof(buf))
	return -EFAULT;
    if (copy_from_user(buf, k.data, k.length))
	return -EFAULT;
    mal_flash_write((char *)NVRAM_BASE_ADDRESS + k.offset,
		    buf,
		    k.length);
    return 0;
}

static int
mal_panic_read(struct malbec_port * info,
	      mxc_panic_t *u)
{
    mxc_panic_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.block >= NVRAM_PANICS)
	return -EFAULT;
    if (copy_to_user(k.record,
	   (char *)NVRAM_BASE_ADDRESS + NVRAM_PANIC_OFFSET(k.block),
	   NVRAM_PANIC_SIZE))
        return -EFAULT;
    return 0;
}

static int
mal_panic_clear(struct malbec_port * info,
	       mxc_panic_t *u)
{
    mxc_panic_t k;
    char buf[NVRAM_PANIC_SIZE];
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.block >= NVRAM_PANICS)
	return -EFAULT;
    memset(buf, 0, NVRAM_PANIC_SIZE);
    mal_flash_write((char *)NVRAM_BASE_ADDRESS + NVRAM_PANIC_OFFSET(k.block),
		    buf,
		    NVRAM_PANIC_SIZE);
    return 0;
}

static int
mal_reboot_str_read(struct malbec_port * info,
		   mxc_string_t *u)
{
    mxc_string_t k;
    memcpy(k.string,
	   (char *)NVRAM_BASE_ADDRESS + NVRAM_REBOOT_STRING_OFFSET,
	   NVRAM_REBOOT_STRING_SIZE);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
mal_reboot_str_write(struct malbec_port * info,
		    mxc_string_t *u)
{
    mxc_string_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    mal_flash_write((char *)NVRAM_BASE_ADDRESS + NVRAM_REBOOT_STRING_OFFSET,
		    k.string,
		    NVRAM_REBOOT_STRING_SIZE);
    return 0;
}

/*
 * This routine allows the tty driver to implement device-
 * specific ioctl's.  If the ioctl number passed in cmd is
 * not recognized by the driver, it should return ENOIOCTLCMD.
 */
static int
mal_ioctl(struct tty_struct *tty, struct file * file,
	 unsigned int cmd, unsigned long arg)
{
    struct malbec_port * info = (struct malbec_port *)tty->driver_data;
    int ret_val = 0;

#ifdef MAL_DEBUG_IOCTL
    printk("mxc: mal_ioctl mxc%d, cmd = %x arg = %lx\n",
	   info->line, cmd, arg); /* */
#endif

    switch (cmd) {
    case MXGETMON:
        ret_val = get_mon_info(info, (struct malbec_monitor *)arg);
	break;
    case MXGETCARDINFO:
        if (copy_to_user((void *)arg, (void *)&mal_card[info->card], 
			 sizeof (struct malbec_card))) {
	    ret_val = -EFAULT;
	    break;
	}
	ret_val = 0;
	break;
    case MXSETWAIT:
        info->closing_wait = (unsigned short)arg * HZ/100;
	ret_val = 0;
	break;
    case MXGETWAIT:
        ret_val = info->closing_wait / (HZ/100);
	break;
    case MXPHYREAD:
        ret_val = 0;
        break;
    case MXPHYWRITE:
        ret_val = 0;
        break;
    case MXCPLDREAD:
        ret_val = mal_cpld_read(info, (mal_cpld_ioctl_t *) arg);
        break;
    case MXCPLDWRITE:
        ret_val = mal_cpld_write(info, (mal_cpld_ioctl_t *) arg);
        break;
    case MXFPGAPROG:
        ret_val = 0;
        break;

	/* arp ioctls */
    case MXARPENTRYADD:
        ret_val = mal_arp_add(info, (arp_entry_t *) arg);
        break;
    case MXARPENTRYDEL:
        ret_val = mal_arp_del(info, (arp_entry_t *) arg);
        break;
    case MXARPCLEARALL:
        ret_val = mal_arp_clear_all(info, (arp_entry_t *) arg);
        break;

	/* AP led ioctls */
    case MXAPLEDREAD:
        ret_val = 0;
	break;
    case MXAPLEDWRITE:
        ret_val = 0;
	break;
 
	/* status ioctls */
    case MXSWVERSION:
        ret_val = mal_sw_version(info, (mxc_string_t *) arg);
	break;
    case MXSCID:
        ret_val = mal_sc_id(info, (mxc_ioctl_t *) arg);
	break;
    case MXFANSTATUS:
        ret_val = mal_fan_status(info, (mxc_ioctl_t *) arg);
	break;
    case MXPSSTATUS:
        ret_val = mal_ps_status(info, (mxc_ioctl_t *) arg);
	break;
    case MXTEMPERATURE:
        ret_val = mal_temperature(info, (mxc_ioctl_t *) arg);
	break;
    case MXSCSLOT:
        ret_val = mal_sc_slot(info, (mxc_ioctl_t *) arg);
	break;
    case MXLCID:
        ret_val = mal_lc_id(info, (mxc_ioctl_t *) arg);
	break;
    case MXLCSWITCHREV:
        ret_val = mal_lc_switch_rev(info, (mxc_lc_t *) arg);
	break;
    case MXLCSTATUS:
        ret_val = mal_lc_status(info, (mxc_lc_t *) arg);
	break;

	/* reboot ioctls */
    case MXWDRESET:
        ret_val = mal_wd_reset(info);
        break;
    case MXPANIC:
        panic("test panic");
        break;
    case MXNULLPTR:
        *(char *)0x0 = 1;
	printk("NULLPTR ptr val = 0x%x\n", *(char *)0x0);
        break;
    case MXHANG:
        while(1);
        break;

	/* "NVRAM" access ioctls */
    case MXNVRAMREAD:
        ret_val = mal_nvram_read(info, (mxc_nvram_t *) arg);
        break;
    case MXNVRAMWRITE:
        ret_val = mal_nvram_write(info, (mxc_nvram_t *) arg);
        break;
    case MXPANICREAD:
        ret_val = mal_panic_read(info, (mxc_panic_t *) arg);
        break;
    case MXPANICCLEAR:
        ret_val = mal_panic_clear(info, (mxc_panic_t *) arg);
        break;
    case MXREBOOTSTRREAD:
        ret_val = mal_reboot_str_read(info, (mxc_string_t *) arg);
        break;
    case MXREBOOTSTRWRITE:
        ret_val = mal_reboot_str_write(info, (mxc_string_t *) arg);
        break;

	/* scratch ioctls */
    case MXSCRATCHREAD:
        ret_val = 0;
        break;
    case MXSCRATCHWRITE:
        ret_val = 0;
	break;

	/* serial ioctls */
    case TIOCGSERIAL:
        ret_val = 0;
	break;
    case TIOCSSERIAL:
        ret_val = 0;
	break;
    default:
        ret_val = -ENOIOCTLCMD;
    }

#ifdef MAL_DEBUG_IOCTL
    printk(" mxc: mal_ioctl done %x\n", ret_val);
#endif

    return ret_val;
} /* mal_ioctl */


/*
 * This routine allows the tty driver to be notified when
 * device's termios settings have changed.  Note that a
 * well-designed tty driver should be prepared to accept the case
 * where old == NULL, and try to do something rational.
 */
static void
mal_set_termios(struct tty_struct *tty, struct termios * old_termios)
{
    return;
} /* mal_set_termios */

/* This routine is called by the upper-layer tty layer to signal
   that incoming characters should be throttled because the input
   buffers are close to full.
 */
static void
mal_throttle(struct tty_struct * tty)
{
} /* mal_throttle */


/*
 * This routine notifies the tty driver that it should signal
 * that characters can now be sent to the tty without fear of
 * overrunning the input buffers of the line disciplines.
 */
static void
mal_unthrottle(struct tty_struct * tty)
{
} /* mal_unthrottle */


/* mal_start and mal_stop provide software output flow control as a
   function of XON/XOFF, software CTS, and other such stuff.
*/
static void
mal_stop(struct tty_struct *tty)
{
} /* mal_stop */


static void
mal_start(struct tty_struct *tty)
{
} /* mal_start */


static void
mal_flush_buffer(struct tty_struct *tty)
{
} /* mal_flush_buffer */


/*
 * mal_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void
mal_hangup(struct tty_struct *tty)
{
    struct malbec_port * info = (struct malbec_port *)tty->driver_data;
        
#ifdef MAL_DEBUG_OTHER
    printk("mxc: mal_hangup mxc%d\n", info->line); /* */
#endif

    mal_flush_buffer(tty);
    shutdown(info);
    info->event = 0;
    info->count = 0;
#ifdef MAL_DEBUG_COUNT
    printk("mxc: mal_hangup (%d): setting count to 0\n", current->pid);
#endif
    info->tty = 0;
    info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
    wake_up_interruptible(&info->open_wait);
} /* mal_hangup */


/*
 * This routine prints out the appropriate serial driver version number
 * and identifies which options were configured into this driver.
 */
static inline void
show_version(void)
{
    printk("malbec control driver, built %s %s\n",
	   __DATE__, __TIME__);
} /* show_version */

static int 
malbec_get_proc_info(char *buf, char **start, off_t offset, int length,
		       int *eof, void *data)
{
    int len=0;
    off_t begin=0;
    off_t pos=0;
    int size;

    size = sprintf(buf, "\n");

    pos += size;
    len += size;
    *eof = 1;
    *start = buf + (offset - begin);	/* Start of wanted data */
    len -= (offset - begin);		/* Start slop */
    if (len > length)
	len = length;			/* Ending slop */
    if (len < 0)
	len = 0;
    return len;
}

/* The serial driver boot-time initialization code!
    Hardware I/O ports are mapped to character special devices on a
    first found, first allocated manner.  That is, this code searches
    for Mxclom cards in the system.  As each is found, it is probed
    to discover how many chips (and thus how many ports) are present.
    These ports are mapped to the tty ports 32 and upward in monotonic
    fashion.  If an 8-port card is replaced with a 16-port card, the
    port mapping on a following card will shift.

    This approach is different from what is used in the other serial
    device driver because the Mxclom is more properly a multiplexer,
    not just an aggregation of serial ports on one card.

    If there are more cards with more ports than have been
    statically allocated above, a warning is printed and the
    extra ports are ignored.
 */

int __init
mal_init(void)
{
    struct malbec_port  *info;
    struct malbec_card *cinfo;
    int board,port,i;
    int nports;

    init_bh(CYCLADES_BH, do_malbec_bh);

#ifdef MAL_DEBUG_OPEN
    show_version();
#endif

    /* Initialize the tty_driver structure */
    
    memset(&mal_driver, 0, sizeof(struct tty_driver));
    mal_driver.magic = TTY_DRIVER_MAGIC;
    mal_driver.driver_name = "malbec";
    mal_driver.name = "mxc";
    mal_driver.major = MALBEC_MAJOR;
    mal_driver.minor_start = 0;
    mal_driver.num = NR_PORTS;
    mal_driver.type = TTY_DRIVER_TYPE_SERIAL;
    mal_driver.subtype = SERIAL_TYPE_NORMAL;
    mal_driver.init_termios = tty_std_termios;
    mal_driver.init_termios.c_cflag =
            B9600 | CS8 | CREAD | HUPCL | CLOCAL;
    mal_driver.flags = TTY_DRIVER_REAL_RAW;
    mal_driver.refcount = &serial_refcount;
    mal_driver.table = serial_table;
    mal_driver.termios = serial_termios;
    mal_driver.termios_locked = serial_termios_locked;

    mal_driver.open = mal_open;
    mal_driver.close = mal_close;
    mal_driver.write = mal_write;
    mal_driver.put_char = mal_put_char;
    mal_driver.flush_chars = mal_flush_chars;
    mal_driver.write_room = mal_write_room;
    mal_driver.chars_in_buffer = mal_chars_in_buffer;
    mal_driver.flush_buffer = mal_flush_buffer;
    mal_driver.ioctl = mal_ioctl;
    mal_driver.throttle = mal_throttle;
    mal_driver.unthrottle = mal_unthrottle;
    mal_driver.set_termios = mal_set_termios;
    mal_driver.stop = mal_stop;
    mal_driver.start = mal_start;
    mal_driver.hangup = mal_hangup;
    mal_driver.break_ctl = mal_break;
    mal_driver.wait_until_sent = mal_wait_until_sent;
    mal_driver.read_proc = malbec_get_proc_info;

    if (tty_register_driver(&mal_driver))
            panic("Couldn't register Malbec serial driver\n");

    for (i = 0; i < NR_CARDS; i++) {
            /* base_addr=0 indicates board not found */
            mal_card[i].base_addr = 0;
    }

    /* the code below is responsible to find the boards. Each different
       type of board has its own detection routine. If a board is found,
       the next mal_card structure available is set by the detection
       routine. These functions are responsible for checking the
       availability of mal_card and mal_port data structures */

    /* assume it always exists */
    mal_card[0].ctl_phys = MAL_CPLD_BASE_ADDR;
    mal_card[0].ctl_addr = (ulong)ioremap(mal_card[0].ctl_phys, 0x100);
    mal_card[0].base_phys = 0;
    mal_card[0].base_addr = 0;
    mal_card[0].irq = 19; /* malbec_setup.c */
    mal_card[0].num_chips = 1;
    mal_card[0].first_line = 0;
    if(request_irq(mal_card[0].irq, mal_interrupt, SA_SHIRQ, "Malbec", &mal_card[0])) {
        printk("mxc: could not allocate IRQ#%d\n", mal_card[0].irq);
	return 1;
    }

#ifdef ARUBA_DEBUG
    printk("mxc: Malbec Control Driver, IRQ %d\n", mal_card[0].irq);
#endif

    mal_nboard = 1;

    /* invalidate remaining mal_port structures */
    for (i = 0; i < NR_PORTS ; i++) {
        mal_port[i].line = -1;
        mal_port[i].magic = -1;
    }

    nports = NR_PORTS;
    /* initialize per-port data structures for each valid board found */
    for (board = 0; board < mal_nboard; board++) {
        cinfo = &mal_card[board];
	cinfo->nports = nports;
	spin_lock_init(&cinfo->card_lock);
	for (port = cinfo->first_line ;
	     port < cinfo->first_line + nports;
	     port++) {
	    info = &mal_port[port];
	    info->magic = MALBEC_MAGIC;
	    info->type = PORT_PHILIPS;
	    info->card = board;
	    info->line = port;
	    info->flags = STD_COM_FLAGS;
	    info->tty = 0;
	    info->xmit_fifo_size = MAL_FIFO_SIZE;
	    info->custom_divisor = 0;
	    info->close_delay = 5*HZ/10;
	    info->closing_wait = CLOSING_WAIT_DELAY;
	    info->icount.cts = info->icount.dsr = 
	      info->icount.rng = info->icount.dcd = 0;
	    info->icount.rx = info->icount.tx = 0;
	    info->icount.frame = info->icount.parity = 0;
	    info->icount.overrun = info->icount.brk = 0;
	    info->event = 0;
	    info->count = 0;
	    info->blocked_open = 0;
	    info->tqueue.routine = do_softint;
	    info->tqueue.data = info;
	    info->normal_termios =
	      mal_driver.init_termios;
	    init_waitqueue_head(&info->open_wait);
	    init_waitqueue_head(&info->close_wait);
	    init_waitqueue_head(&info->shutdown_wait);
	    /* info->session */
	    /* info->pgrp */
	    /* info->timeout */
	    /* Bentson's vars */
	    info->jiffies[0] = 0;
	    info->jiffies[1] = 0;
	    info->jiffies[2] = 0;
	}
    }

    return 0;
    
} /* mal_init */

#ifdef MODULE
void
mal_cleanup_module(void)
{
    int i;
    int e;
    unsigned long flags;

    save_flags(flags); cli();
    remove_bh(MALBEC_BH);

    if ((e = tty_unregister_driver(&mal_driver)))
            printk("mxc: failed to unregister Malbec driver(%d)\n",
		   e);

    restore_flags(flags);

    for (i = 0; i < NR_CARDS; i++) {
        if (mal_card[i].base_addr != 0) {
	    iounmap((void *)mal_card[i].base_addr);
	    if (mal_card[i].ctl_addr != 0)
		iounmap((void *)mal_card[i].ctl_addr);
	    if (mal_card[i].irq)
		free_irq(mal_card[i].irq, &mal_card[i]);
        }
    }
} /* mal_cleanup_module */

/* Module entry-points */
module_init(mal_init);
module_exit(mal_cleanup_module);

#else /* MODULE */
/* called by linux/init/main.c to parse command line options */
void
mal_setup(char *str, int *ints)
{
} /* mal_setup */
#endif /* MODULE */

MODULE_LICENSE("GPL");
