/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

/*
 * Functions to perform wlan buffer memory management.
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/ap/apmem.c#1 $"

#include <linux/netdevice.h>

#include "wlantype.h"
#include "ui.h"
#include "wlanproto.h"
#include "wlanext.h"
#include "apdefs.h"
#include "apmem.h"
#include "halApi.h"
#include "ar5hwcEnd.h"

A_STATUS init_buff_pool            (void);
A_UINT16 get_position              (void *pVirtAddr);                     
A_UINT16 size_of_mem_pool_in_buffs (void);
A_STATUS return_buff_to_mem_pool   (void *pVirtBuffAddr);
A_STATUS grab_buff_from_mem_pool   (A_UINT16 bufSize, 
                                    A_UINT32 *pPhysBuffAddr,
                                    void     **pVirtBuffAddr);


                     
#ifdef DEBUG
int minfreecnt = 0;
int maxfreecnt = 0;
#endif

/**************************************************************************
* getDescFromQueue - get a descriptor from the given queue
*
* RETURNS: ATHEROS_DESC *
*/
ATHEROS_DESC *
getDescFromQueue(QUEUE_DETAILS *pQueue)
{
    ATHEROS_DESC *pDesc;

    pDesc = pQueue->pDescQueueHead;
    if (pDesc) {
        pQueue->pDescQueueHead = pDesc->pNextVirtPtr;
        if (pQueue->pDescQueueHead == NULL) {
            pQueue->pDescQueueTail = NULL;
        }

        pDesc->pNextVirtPtr = NULL;
        pDesc->nextPhysPtr = 0;
    }

    return pDesc;
}

/**************************************************************************
* addDescToQueueTail - add the descriptor to the tail of queue struct
*
* Adds the given descriptor the tail of the queue structure
*
* RETURNS: N/A
*/
void
addDescToQueueTail(ATHEROS_DESC *pDesc,     /* head of new chain */
                   ATHEROS_DESC *pTail,     /* tail of new chain */
                   QUEUE_DETAILS *pQueue)
{
    /* reject attempts to requeue the same descriptor */
    ASSERT(pQueue->pDescQueueTail != pDesc);
    if (pQueue->pDescQueueTail == pDesc) {
        return;
    }

    /* check to see if anything is in the queue */
    if (pQueue->pDescQueueTail == NULL) {
        ASSERT(pQueue->pDescQueueHead == NULL);
        pQueue->pDescQueueHead = pDesc;
    } else {
        pQueue->pDescQueueTail->pNextVirtPtr = pDesc;
        pQueue->pDescQueueTail->nextPhysPtr = pDesc->thisPhysPtr;
    }
    pQueue->pDescQueueTail = pTail;

    return;
}

/**************************************************************************
 * memAllocateDescriptor - allocate a descriptor from the descriptor pool
 *
 * This routine will get a single ATHEROS descriptor from the descriptor pool
 *
 * RETURNS: N/A
 */
ATHEROS_DESC *
memAllocateDescriptor(WLAN_DEV_INFO *pdevInfo)
{
    ATHEROS_DESC  *pDesc;
    A_UINT32      physAddr;

    pDesc = getDescFromQueue(&pdevInfo->emptyDescQueue);
    if (pDesc) {
        /* clear state of desc */
        physAddr = pDesc->thisPhysPtr;
        A_MEM_ZERO(pDesc, sizeof(ATHEROS_DESC));
        pDesc->thisPhysPtr = physAddr;
        pDesc->freeQueue   = &pdevInfo->emptyDescQueue;

#ifdef DEBUG
        pdevInfo->freeDescCount--;
        minfreecnt = min(pdevInfo->freeDescCount, minfreecnt);
#endif
    }

    return pDesc;
}

/**************************************************************************
* memFreeDescriptor - free a descriptor
*
* This routine will free a descriptor that was already created by
* memAllocateDescriptor()
*
* RETURNS: N/A
*/
void
memFreeDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    QUEUE_DETAILS * pQueue = pDesc->freeQueue;

    ASSERT(pQueue == &pdevInfo->emptyDescQueue);

    pDesc->nextPhysPtr = 0;
    pDesc->pNextVirtPtr = NULL;

    addDescToQueueTail(pDesc, pDesc, pQueue);

#ifdef DEBUG
    pdevInfo->freeDescCount++;
    maxfreecnt = max(pdevInfo->freeDescCount, maxfreecnt);
#endif

    return;
}

A_STATUS
hwAllocatePhysBuffer(WLAN_DEV_INFO *pdevInfo,
                     A_UINT16 buffSize,         /* size to create */
                     A_UINT32 *pPhysAddress,    /* return physical address */
                     void **ppVirtAddress,      /* return virtual address */
                     void **ppOSDescPtr)        /* return OS desc address */
{
    // creates an skb and stores skb->data in ppVirtAddress
    // if we want to pass this skb to gre, then we need to create enough headroom
	// for the gre header
	
	#define GRE_HEADROOM (24 + 16)  //ATHEROS_HW_CTRL_AND_STATUS=16
	int ALIGN_OFFSET = ((GRE_HEADROOM + 31) & ~31);

    int skb_buffSize = ((buffSize +31) & ~31);
	struct sk_buff *skb = alloc_skb(skb_buffSize + ALIGN_OFFSET, GFP_ATOMIC);
		if (skb == NULL){
			return A_NO_MEMORY;
		}
		
		// reserve enough room in the front for the gre header
		skb_reserve(skb, ALIGN_OFFSET);
		
        skb->data = (char*)(((A_UINT32)skb->data + 31) & ~31);
        *ppVirtAddress = skb->data;
		*ppOSDescPtr = skb;
		*pPhysAddress = A_DESC_V2P(*ppVirtAddress);
		
		return A_OK;
	
}

#if 0

/**************************************************************************
* hwAllocatePhysBuffer - allocate a physically contiguous buffer
*
* This routine will create a physically contiguous buffer and get
* the virtual and physical address of it.
*
* RETURNS: A_OK if got buffer, A_ERROR otherwise
*/
A_STATUS
hwAllocatePhysBuffer(WLAN_DEV_INFO *pdevInfo,
                     A_UINT16 buffSize,         /* size to create */
                     A_UINT32 *pPhysAddress,    /* return physical address */
                     void **ppVirtAddress,      /* return virtual address */
                     void **ppOSDescPtr)        /* return OS desc address */
{
    /* kmalloc should give us the contiguous mem */
    /* this will also get called from an ISR, processRxInterrupt, so 
       use GFP_ATOMIC */
    if ((*ppVirtAddress = kmalloc(buffSize, GFP_ATOMIC)) == NULL) {
        aPrintk("FAILED TO ALLOC MEM for %dB in createBuffandDesc\n", buffSize);
        return A_NO_MEMORY;
    }

    *pPhysAddress = A_DESC_V2P(*ppVirtAddress);

    if ((NULL==pPhysAddress) || (NULL==*ppVirtAddress) || (NULL==ppVirtAddress)) {
        printk("ONE OF THE BUFFER ADDRESS IS NULLLLLLLLLLLLLLLLLLLLLLLLLLLLLL\n");
    }

    *ppOSDescPtr  = NULL;

    return A_OK;
}

#endif // USE_SKB



void
hwFreePhysBuffer(WLAN_DEV_INFO *pdevInfo, void *pVirtAddress, void *pOSDescPtr)
{
	
	// if the skb is passed to gre, then we should not free it here since it
	// will be freed in the eth0 driver. Freeing it here will cause kernel panic
	// maybe we change this to pass a flag to indicate whether we need kfree_skb

	struct sk_buff *skb = (struct sk_buff *)pOSDescPtr;

	// need to call this since the fre could happen in an interrupt
	// context too.
	dev_kfree_skb_any(skb);
	pVirtAddress = pOSDescPtr = NULL;
}

#if 0

/**************************************************************************
* hwFreePhysBuffer - free the physically contiguous buffer
*
* This routine will free the physically contiguous buffer previously
* allocated by hwAllocatePhysBuffer()
*
* RETURNS: N/A
*/
void
hwFreePhysBuffer(WLAN_DEV_INFO *pdevInfo, void *pVirtAddress, void *pOSDescPtr)
{
        
    kfree(pVirtAddress);
	  pVirtAddress = NULL;
    pOSDescPtr = NULL;

    return;
}

#endif  // USE_SKB

/**************************************************************************
* createHeaderBuffandDesc - create a physically contiguous buffer
*                           and descriptor for 802.11 header
*
* Create a physically contiguous data buffer and create a descriptor
* to go with it.  An ATHEROS_DESC gets returned, pointer to data buffer
* can be assertained from the descriptor
*
* RETURNS: ATHEROS_DESC if successful, NULL otherwise
*/
ATHEROS_DESC *
createHeaderBuffandDesc(WLAN_DEV_INFO *pdevInfo, A_UINT16 buffSize)
{
    ATHEROS_DESC *pDesc;

    /* Get the physically contiguous buffer and desc from
     * the pre-linked txHeaderQueue
     */
    pDesc = getDescFromQueue(&pdevInfo->txHeaderQueue);
    if (pDesc) {
        /* Fill in the descriptor fields */
        pDesc->pDestSibEntry = NULL;
        pDesc->hwControl.tx.bufferLength = buffSize;
    }

    return pDesc;
}

/**************************************************************************
* memCopyToPSBuffers - since power save frames are going to be a while
*   its important to release their buffers back to the sending device
*   and copy the data over to buffers reserved for power save; the wlan
*   driver can reserve buffers for power save based on the amount of
*   memory available on the system
*
* RETURNS: A_OK if successful, A_ERROR otherwise
*/

ATHEROS_DESC *
memCopyToPSBuffers(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *inChain)
{
    ATHEROS_DESC *outChain = NULL;
    ATHEROS_DESC *inDesc, *outDesc = NULL;
    HAL_DESC_TX_CONTROL *outTxControl;
    char *ptr;

    for (inDesc = inChain; inDesc; inDesc = inDesc->pNextVirtPtr) {
        if (!outChain) {
            outChain = getDescFromQueue(&pdevInfo->psBufferQueue);
            if (!outChain) {
                freeBuffandDescChain(pdevInfo, inChain);
                return NULL;
            }
            outDesc = outChain;
        } else {
            outDesc->pNextVirtPtr = getDescFromQueue(&pdevInfo->psBufferQueue);
            if (!outDesc->pNextVirtPtr) {
                freeBuffandDescChain(pdevInfo, outChain);
                freeBuffandDescChain(pdevInfo, inChain);
                return NULL;
            }
            outDesc->nextPhysPtr = outDesc->pNextVirtPtr->thisPhysPtr;
            outDesc = outDesc->pNextVirtPtr;
        }

        ASSERT(!inDesc->psFlag);

        outDesc->pSrcSibEntry = inDesc->pSrcSibEntry;
        outDesc->pDestSibEntry = inDesc->pDestSibEntry;
        outDesc->pTxLastDesc = outDesc;
        outDesc->pTxFirstDesc = outDesc;

        outTxControl = &outDesc->hwControl.tx;
	/* Capture the state of inDesc tx control */
	A_BCOPY(&inDesc->hwControl.tx, outTxControl, sizeof(*outTxControl));

        outTxControl->more = 0;
        halSetTxDescPktType(pdevInfo, outDesc, HAL_DESC_PKT_TYPE_NORMAL);
        outTxControl->frameLength = inDesc->hwControl.tx.frameLength;

        outTxControl->bufferLength = 0;
        for (ptr = outDesc->pBufferVirtPtr.ptr; inDesc; inDesc = inDesc->pNextVirtPtr) {
            A_BCOPY(inDesc->pBufferVirtPtr.ptr, ptr, inDesc->hwControl.tx.bufferLength);
            outTxControl->bufferLength += inDesc->hwControl.tx.bufferLength;
            ptr += inDesc->hwControl.tx.bufferLength;
            if (!inDesc->hwControl.tx.more) {
                break;
            }
        }
        ASSERT(inDesc);
    }

    freeBuffandDescChain(pdevInfo, inChain);

    return outChain;
}

/**************************************************************************
* createBuffandDesc - create a physically contiguous buffer and descriptor
*
* Create a physically contiguous data buffer and create a descriptor
* to go with it.  An ATHEROS_DESC gets returned, pointer to data buffer
* can be assertained from the descriptor
*
* RETURNS:A_OK if successful, A_ERROR otherwise
*/
A_STATUS
createBuffandDesc(WLAN_DEV_INFO *pdevInfo, A_UINT16 buffSize,
                  ATHEROS_DESC **ppDesc)
{
    ATHEROS_DESC *pDesc;
    A_STATUS status = A_ERROR;

    /* allocate the descriptor */
    pDesc = memAllocateDescriptor(pdevInfo);
    *ppDesc = pDesc;
    if (!pDesc) {
        printk("createBuffandDesc: NO MEMORY IN CREATEBUFANDDESC..\n");
        return A_NO_MEMORY;
    }

    /* Get the physically contiguous buffer */
    status = hwAllocatePhysBuffer(pdevInfo, buffSize, &pDesc->bufferPhysPtr,
                &pDesc->pBufferVirtPtr.ptr, &pDesc->pOSDescPtr);

    if (status != A_OK) {
        printk("hwAllocatePhysBuffer: failed to allocated skb space!!!\n");
        memFreeDescriptor(pdevInfo, pDesc);
        return status;
    }
    pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
    pDesc->pDestSibEntry      = NULL;
    pDesc->hwControl.tx.frameLength  = buffSize;
    pDesc->hwControl.tx.bufferLength = buffSize;

    return A_OK;
}

/**************************************************************************
* freeBuffandDesc - free buffer and descriptor
*
* Free the buffer and descriptor, created with createBuffandDesc()
*
* RETURNS: N/A
*/
void
freeBuffandDesc(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    QUEUE_DETAILS * pQueue = pDesc->freeQueue;

    if ((pQueue == &pdevInfo->txHeaderQueue)
        || (pQueue == &pdevInfo->psBufferQueue)) {

					
				pDesc->pNextVirtPtr = NULL;
        pDesc->nextPhysPtr  = 0;

        pDesc->swretryCount = pDesc->fragSeqNumValid = 0;
        pDesc->psFlag       = pDesc->staleFlag       = 0;
        pDesc->pTxFirstDesc = pDesc->pTxLastDesc     = NULL;

        /* zero out the control and status fields */
        pDesc->hwControl.word[0] = 0;
        pDesc->hwControl.word[1] = 0;
        pDesc->hwStatus.word[0]  = 0;
        pDesc->hwStatus.word[1]  = 0;

        A_SEM_LOCK(pQueue->qSem, WAIT_FOREVER);
        addDescToQueueTail(pDesc, pDesc, pQueue);
        A_SEM_UNLOCK(pQueue->qSem);

        return;
    }

    hwFreePhysBuffer(pdevInfo, pDesc->pOrigBufferVirtPtr, pDesc->pOSDescPtr);
    memFreeDescriptor(pdevInfo, pDesc);

    return;
}

/**************************************************************************
* freeBuffandDescChain - free buffer and descriptor chain
*
* Free the buffer and descriptor chain, created with createBuffandDesc()
*
* RETURNS: N/A
*/
void
freeBuffandDescChain(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    ATHEROS_DESC *pNextDesc;

    while (pDesc != NULL) {
        /* clean the current buffer and descriptor */
        pNextDesc = pDesc->pNextVirtPtr;
        freeBuffandDesc(pdevInfo, pDesc);
        pDesc = pNextDesc;
    }
    return;
}

/**************************************************************************
* descInfoDup - duplicate os dependent fields in descriptor chain
*
* read os dependent fields in the original descriptor and copy it over to
* to be duplicated descriptor
*
* RETURNS: N/A
*/
void
descInfoDup(ATHEROS_DESC *pOrigDesc, ATHEROS_DESC *pDupDesc)
{
    /* duplicate the os dependent info fields */
    pDupDesc->pOSHandle     = pOrigDesc->pOSHandle;
    pDupDesc->memFlag       = pOrigDesc->memFlag;
    pDupDesc->pOSDescPtr    = pOrigDesc->pOSDescPtr;

    pDupDesc->pDestSibEntry = pOrigDesc->pDestSibEntry;
    pDupDesc->pSrcSibEntry  = pOrigDesc->pSrcSibEntry;

    return;
}

/**************************************************************************
* bufferRefUpdate - update the reference info for the buffer
*
* for vxWorks this is increase the the reference counter for the buffer
*
* for the NDIS this could be set flag
*
* RETURNS: N/A
*/
void
bufferRefUpdate(ATHEROS_DESC *pDesc)
{
    printk("bufferRefUpdate() if BLANK ! \n");
#if FADI
    ((CL_BLK_ID)pDesc->pOSDescPtr)->clRefCnt++;
#endif /* FADI */
    return;
}

/**************************************************************************
* getPktFromQueueStart - get a packet from the start of queue
*
* Removes the packet's descriptor chain from the head of the queue.
* Sets the new head of the queue to the next packet's descriptor chain
*
* In case of the fragmented packet, the whole packet has to be send out
*
* RETURNS: Pointer to the descriptor chain of the packet if successful
*      NULL if failed
*/
ATHEROS_DESC *
getPktFromQueueStart(WLAN_DEV_INFO *pdevInfo, QUEUE_DETAILS *pQueue)
{
    ATHEROS_DESC      *pDesc;           /* descriptor chain to get */
    WLAN_FRAME_HEADER *pWlanHdr;
    ATHEROS_DESC      *pVarDesc;


    /* don't do anything if the queue is empty */
    if (pQueue->pDescQueueHead == NULL) {
        A_SEM_UNLOCK(pQueue->qSem);
        return NULL;
    }

    /* get descriptor from head of queue */
    pDesc = pQueue->pDescQueueHead;

    /* walk through the descriptor chain for the packet */
    pVarDesc = pDesc;

    pWlanHdr = pVarDesc->pBufferVirtPtr.header;

    /* walk through all the fragments */
    while (pWlanHdr->frameControl.moreFrag != 0) {
        /* walk through the current fragment */
        while (pVarDesc->hwControl.tx.more) {
            pVarDesc = pVarDesc->pNextVirtPtr;
        }

        /* next fragment */
        pVarDesc = pVarDesc->pNextVirtPtr;
        pWlanHdr = pVarDesc->pBufferVirtPtr.header;

        pQueue->qFrameCount--;
    }

    /* walk through the last fragment */
    while (pVarDesc->hwControl.tx.more) {
        pVarDesc = pVarDesc->pNextVirtPtr;
    }

    pQueue->qFrameCount--;

    /* set the new queue head */
    pQueue->pDescQueueHead = pVarDesc->pNextVirtPtr;

    if (pQueue->pDescQueueHead == NULL) {
        pQueue->pDescQueueTail = NULL;
    }

    pVarDesc->pNextVirtPtr = NULL;
    pVarDesc->nextPhysPtr = 0;


    return (pDesc);
}

/**************************************************************************
* ageQueue - aging out the buffered tx traffic from the queue
*
* RETURNS: N/A
*/
void
ageQueue(WLAN_DEV_INFO  *pdevInfo, QUEUE_DETAILS *pQueue,
         A_UINT32 agingInterval,        /* aging time period in MS */
         A_UINT32 now)                  /* current time in MS */
{
    ATHEROS_DESC *pHeadDesc, *pVarDesc = NULL;

    if (agingInterval == 0) {
        return;
    }

    /* aging the buffered traffic */
    pHeadDesc = pQueue->pDescQueueHead;

    while ((pHeadDesc != NULL) &&
           ((now - pHeadDesc->timeStamp) > agingInterval))
    {
        pVarDesc = pHeadDesc;
        while (pVarDesc->hwControl.tx.more == 1) {
            pVarDesc = pVarDesc->pNextVirtPtr;
        }

        /* next frame */
        pHeadDesc = pVarDesc->pNextVirtPtr;
        pQueue->qFrameCount--;
    }

    /* aged out the old frames if any */
    if (pVarDesc != NULL) {
        pVarDesc->pNextVirtPtr = NULL;
        pVarDesc->nextPhysPtr = 0;

        freeBuffandDescChain(pdevInfo, pQueue->pDescQueueHead);

        /* set the new head of the queue */
        if ((pQueue->pDescQueueHead = pHeadDesc) == NULL) {
            pQueue->pDescQueueTail = NULL;
        }
    }

    return;
}

#ifdef DEBUG
int
qShow(void)
{
    WLAN_DEV_INFO *pdevInfo;
    SIB_ENTRY     *pSib;
    QUEUE_DETAILS *pQueue;
    ATHEROS_DESC  *pDesc;
    int           cnt, unit = 0, idx;
    int           mode;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];

    uiPrintf("Free Count   | min:  %10d max:  %10d\n", minfreecnt, maxfreecnt);

    pQueue = &pdevInfo->emptyDescQueue;
    uiPrintf("Free Queue   | head: %10p tail: %10p count: %5d\n",
             pQueue->pDescQueueHead,
             pQueue->pDescQueueTail,
             pdevInfo->freeDescCount);

    pQueue = &pdevInfo->txHeaderQueue;
    pDesc  = pQueue->pDescQueueHead;
    for (cnt = 0; pDesc; pDesc = pDesc->pNextVirtPtr) {
        cnt++;
    }
    uiPrintf("Tx Hdr Queue | head: %10p tail: %10p count: %5d\n",
             pQueue->pDescQueueHead, pQueue->pDescQueueTail, cnt);

    pQueue = &pdevInfo->txQueue;
    pDesc  = pQueue->pDescQueueHead;
    mode   = pdevInfo->staConfig.swretryEnabled;
    for (cnt = 0; pDesc; pDesc = pDesc->pNextVirtPtr) {
        if (mode && pDesc->pTxFirstDesc->pDestSibEntry != NULL &&
            pDesc->pTxFirstDesc->hwControl.tx.clearDestMask == 1) {
            uiPrintf("\tSIB %10p desc %10p has ClearDestMask set\n",
                     pDesc->pTxFirstDesc->pDestSibEntry, pDesc);
        }
        cnt++;
    }

    uiPrintf("Tx Queue     | head: %10p tail: %10p count: %5d\n",
             pQueue->pDescQueueHead, pQueue->pDescQueueTail, cnt);

    pQueue = &pdevInfo->altQueue;
    pDesc  = pQueue->pDescQueueHead;
    for (cnt = 0; pDesc; pDesc = pDesc->pNextVirtPtr) {
        cnt++;
    }

    uiPrintf("CAB/Alt Queue| head: %10p tail: %10p count: %5d\n",
             pQueue->pDescQueueHead, pQueue->pDescQueueTail, cnt);

    pQueue = &pdevInfo->rxQueue;
    pDesc  = pQueue->pDescQueueHead;

    uiPrintf("Rx Queue     | head: %10p tail: %10p count: %5d\n",
             pQueue->pDescQueueHead, pQueue->pDescQueueTail, pdevInfo->rxQueueDepth);

    A_SIB_TAB_LOCK(pSibTable);
    if (pSibTable->currentEntryNum) {
        pSib = &pSibTable->sta[MAX_STA_NUM - 1];
        uiPrintf("\n");
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--, pSib--) {
            if (pSib->used & (SIB_VALID | SIB_USED)) {
                pQueue = &pSib->txPSQueue;
                pDesc  = pQueue->pDescQueueHead;
                for (cnt = 0; pDesc; pDesc = pDesc->pNextVirtPtr) {
                    cnt++;
                }
                uiPrintf("PS Q AID=%04d| head: %10p tail: %10p count: %5d %s\n",
                         (pSib->assocId & 0x3fff), pQueue->pDescQueueHead,
                         pQueue->pDescQueueTail, cnt,
                         (pSib->psMode ? "Asleep" : "Awake"));
                uiPrintf("             | sib:  %10p frames: %8d pend: %6d\n",
                         pSib, pQueue->qFrameCount, pSib->numPSTxPending);
                uiPrintf("             | needClrDest: %s   qFrames: %7d filtPend:  %d\n",
                         (pSib->swRetry.needClearDest ? "Y" : "N"),
                         pSib->swRetry.qFrames,
                         pSib->swRetry.numFilteredPending);
            }
        }
    }
    A_SIB_TAB_UNLOCK(pSibTable);
    return 0;
}

int
txDescLen(void)
{
    WLAN_DEV_INFO *pdevInfo;
    int           cnt, unit = 0;
    QUEUE_DETAILS *pQueue;
    ATHEROS_DESC  *pDesc;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];

    pQueue = &pdevInfo->txQueue;
    pDesc  = pQueue->pDescQueueHead;
    for (cnt = 0; pDesc; pDesc = pDesc->pNextVirtPtr) {
        cnt++;
    }

    return cnt;
}

int
dumpResourceUsage(int unit)
{
    WLAN_DEV_INFO *pdevInfo;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];

    uiPrintf("Free Count:\n  min: %d\n  max: %d\n  current: %d\n",
             minfreecnt, maxfreecnt, pdevInfo->freeDescCount);
    uiPrintf("Num Tx Frames Dropped: %d\n", pdevInfo->txFramesDropped);
    uiPrintf("Num Rx Frames Dropped: %d\n", pdevInfo->rxFramesDropped);
    uiPrintf("Num Power Save Frames Dropped: %d\n", pdevInfo->psFramesDropped);

    return 0;
}

const char* typeToName[] = {
    "ASSOC_REQ",    "ASSOC_RESP",   "REASSOC_REQ",  "REASSOC_RESP",
    "PROBE_REQ",    "PROBE_RESP",   "MGT:???",      "MGT:???",
    "BEACON",       "ATIM",         "DISASSOC",     "AUTH",
    "DEAUTH",       "MGT:???",      "MGT:???",      "MGT:???",

    "CTRL:???",     "CTRL:???",     "CTRL:???",     "CTRL:???",
    "CTRL:???",     "CTRL:???",     "CTRL:???",     "CTRL:???",
    "CTRL:???",     "CTRL:???",     "PSPOLL",       "RTS",
    "CTS",          "ACK",          "CFEND",        "CFENDACK",

    "DATA",         "DATA_CFACK",   "DATA_CFPOLL",  "DATA_CFACK_CFPOLL",
    "NODATA",       "NODATA_CFACK", "NODATA_CFPOLL","NODATA_CFACK_CFPOLL",
    "DATA:???",     "DATA:???",     "DATA:???",     "DATA:???",
    "DATA:???",     "DATA:???",     "DATA:???",     "DATA:???",

    "???", "???", "???", "???",
    "???", "???", "???", "???",
    "???", "???", "???", "???",
    "???", "???", "???", "???"
};

int
tdDump(ATHEROS_DESC *pDesc)
{
    FRAME_CONTROL_FIELD_LITTLE *fc;

    A_DESC_CACHE_INVAL(pDesc);

    uiPrintf("0x%08x:", (A_UINT32)pDesc);
    if (!pDesc->hwStatus.tx.done) {
        uiPrintf(" .");
    } else {
        if (pDesc->hwStatus.tx.pktTransmitOK) {
            uiPrintf(" ok");
        } else if (pDesc->hwStatus.tx.excessiveRetries) {
            uiPrintf("  x");
        } else if (pDesc->hwStatus.tx.filtered) {
            uiPrintf("  f");
        } else {
            uiPrintf("  ?");
        }
    }
    if (pDesc->pDestSibEntry) {
        uiPrintf(" sta%d", pDesc->pDestSibEntry->assocId & 0x3fff);
    }
    if (pDesc->pTxFirstDesc == pDesc) {
        fc = &pDesc->pBufferVirtPtr.headerLittle->frameControl;
        uiPrintf(" %s", typeToName[(fc->fType << 4) + fc->fSubtype]);
        uiPrintf(" len:%d", pDesc->hwControl.tx.bufferLength);
        uiPrintf("/%d", pDesc->hwControl.tx.frameLength);
        if (fc->wep) {
            uiPrintf(" wep");
        }
        if (fc->retry) {
            uiPrintf(" retry");
        }
        if (fc->moreFrag) {
            uiPrintf(" moreFrag");
        }
    } else {
        uiPrintf(" ---- len:%d", pDesc->hwControl.tx.bufferLength);
    }
    uiPrintf(" key%d", halGetTxDescEncryptKeyIndex(pwlanDevInfo, pDesc));
    if (pDesc->hwControl.tx.clearDestMask) {
        uiPrintf(" clearDestMask");
    }
    if (pDesc->staleFlag) {
        uiPrintf(" stale");
    }
    if (pDesc->psFlag) {
        uiPrintf(" ps");
    }

    uiPrintf(" %s\n", pDesc->hwControl.tx.more ? "->" : "|");

    return 0;
}

int
tqDump(ATHEROS_DESC *pDesc, const char* msg)
{
    if (!pDesc) {
        pDesc = pwlanDevInfo->txQueue.pDescQueueHead;
        msg = "txQueue";
    }
    if (!msg) {
        msg = "txQueue";
    }
    uiPrintf("-- %s --------\n", msg);
    while (pDesc) {
        tdDump(pDesc);
        pDesc = pDesc->pNextVirtPtr;
    }
    uiPrintf("---------\n");

    return 0;
}

#endif /* DEBUG */


#ifdef DONT_NEED_THIS
// partha

#include "wlanproto.h"
#include "wlantype.h"

#define MAX_MEM_BUF_DEPTH  100

// WE ARE MAKING THE ASSUMPTION THAT THESE BUFFERS ARE USED AND RETURNED
// IN ORDER (used to be queued in skbs then freed by tasklets in same order)
struct _ar_mem_pool {
    void *       pArPoolVirtBuf[MAX_MEM_BUF_DEPTH];
    A_UINT32     ArPoolPhysAddr[MAX_MEM_BUF_DEPTH];
    A_UINT16     ar_buff_pool_tail;
    A_UINT16     ar_buff_pool_head;
};

struct _ar_mem_pool ar_mem_pool;


A_STATUS init_buff_pool (void)
{
    A_STATUS status = A_ERROR;
    int i;
    ar_mem_pool.ar_buff_pool_tail = ar_mem_pool.ar_buff_pool_head = 0;

    for (i=0; i < MAX_MEM_BUF_DEPTH; i++) {
        /* kmalloc should give us the contiguous mem */
        ar_mem_pool.pArPoolVirtBuf[i] = (void*)kmalloc(MAX_WLAN_FRAME_SIZE, GFP_DMA);
        if (NULL == ar_mem_pool.pArPoolVirtBuf[i]) {
            printk("init_buff_pool FAILED to alloc for poolbuf#%d\n", i);
            break;
        }
        else
            ar_mem_pool.ArPoolPhysAddr[i] = A_DESC_V2P(ar_mem_pool.pArPoolVirtBuf[i]);
    }

    if (i == MAX_MEM_BUF_DEPTH) {
        status = A_OK;
    }

    printk("INIT_CRE: Sizeof pool left=%d buffs\n", size_of_mem_pool_in_buffs());
    return status;
}

// the only time both would be sitting togther is:
//#1 - initially both at 0
//#2 - if tail sneeks up on head and catches up with it (OK)
A_STATUS grab_buff_from_mem_pool (A_UINT16 bufSize, A_UINT32 *pPhysBuffAddr,
                                  void **pVirtBuffAddr)
{
    
    if (bufSize > MAX_WLAN_FRAME_SIZE) {
        printk("BAD_CRE: bufSize %d exceeds cluster size = %d\n", bufSize, MAX_WLAN_FRAME_SIZE);
        return A_ERROR;
    }

    // reset the head to the start
    if ((ar_mem_pool.ar_buff_pool_head + 1) == MAX_MEM_BUF_DEPTH) {
        if (ar_mem_pool.ar_buff_pool_tail != 0) {
            // wrap around safely
            ar_mem_pool.ar_buff_pool_head = 0;
            // return buffer sitting at the 0 position
            *pVirtBuffAddr = ar_mem_pool.pArPoolVirtBuf[ar_mem_pool.ar_buff_pool_head];
            *pPhysBuffAddr = ar_mem_pool.ArPoolPhysAddr[ar_mem_pool.ar_buff_pool_head];
            printk("GOOD_CRE: reseting head to the start..tail=%d\n", 
                    ar_mem_pool.ar_buff_pool_tail);
        }
        else {
            printk("BAD_CRE: list is full w tail still at start\n");
            // list is full w tail still at start!
            return A_ERROR;
        }
    }
    // is head catching up with tail from behind !?
    else if ((ar_mem_pool.ar_buff_pool_head+1) == ar_mem_pool.ar_buff_pool_tail) {
        printk("BAD_CRE: head caught up with tail from behind! illegal!\n");
        // head caught up midway
        return A_ERROR;
    }
    else {
        // return buffer sitting at that head position
        *pVirtBuffAddr = ar_mem_pool.pArPoolVirtBuf[ar_mem_pool.ar_buff_pool_head];
        *pPhysBuffAddr = ar_mem_pool.ArPoolPhysAddr[ar_mem_pool.ar_buff_pool_head];
        // advance head
        ar_mem_pool.ar_buff_pool_head++;

        printk("GOOD_CRE: head = %d, tail= %d[pVirtAddr=%p][pPhysBuff=0x%lx]\n", 
                ar_mem_pool.ar_buff_pool_head,ar_mem_pool.ar_buff_pool_tail, 
                *pVirtBuffAddr, *pPhysBuffAddr);

    }

    printk("GOOD_CRE: Sizeof pool left=%d buffs\n", size_of_mem_pool_in_buffs());
    return A_OK;
}

A_STATUS return_buff_to_mem_pool (void *pVirtBuffAddr)
{
    printk("DEL_DEL: &pVirtBuffAddr = %p\n", &pVirtBuffAddr);
    printk("DEL_DEL: pVirtBuffAddr = %p\n", pVirtBuffAddr);
    
    // we're assuming in order use1,use2,use3,free1,use4,free2,free3,free4...etc
    ASSERT(pVirtBuffAddr == 
           ar_mem_pool.pArPoolVirtBuf[ar_mem_pool.ar_buff_pool_tail]);
    // ok, so this is the buffer where tail is pointing
    if (pVirtBuffAddr != ar_mem_pool.pArPoolVirtBuf[ar_mem_pool.ar_buff_pool_tail]) {
        printk("BAD_DEL: freeing buff%p [%d pos][head=%d][tail=%d]\n", 
                pVirtBuffAddr, get_position(pVirtBuffAddr),
                ar_mem_pool.ar_buff_pool_head, ar_mem_pool.ar_buff_pool_tail);
        return A_ERROR;
    }

    if (ar_mem_pool.ar_buff_pool_tail == ar_mem_pool.ar_buff_pool_head) {
        printk("BAD_DEL: tail going past head..fail [pos=%d]\n", 
                ar_mem_pool.ar_buff_pool_head);
        // don't go past head
        return A_ERROR;
    }

    if ((ar_mem_pool.ar_buff_pool_tail + 1) == MAX_MEM_BUF_DEPTH) {
        // wrap around safely even if catches up with head
        ar_mem_pool.ar_buff_pool_tail = 0;
        printk("GOOD_DEL: reseting tail[ok to wrap]...head = %p\n", 
                ar_mem_pool.ar_buff_pool_head);
    }
    else {
        // advance tail even if catches up with head
        ar_mem_pool.ar_buff_pool_tail++;
        printk("GOOD_DEL: Sizeof pool left=%d buffs\n", size_of_mem_pool_in_buffs());
    }

    return A_OK;
}


A_UINT16 size_of_mem_pool_in_buffs (void)
{
    if (ar_mem_pool.ar_buff_pool_head > ar_mem_pool.ar_buff_pool_tail) {
        return (A_UINT16)(MAX_MEM_BUF_DEPTH - (ar_mem_pool.ar_buff_pool_head+1));
    }
    else {
        return (A_UINT16)(ar_mem_pool.ar_buff_pool_tail - ar_mem_pool.ar_buff_pool_head);
    }
}

A_UINT16 get_position (void *pVirtAddr)
{
    int i=0;
    while ((i < MAX_MEM_BUF_DEPTH) && 
           (pVirtAddr != ar_mem_pool.pArPoolVirtBuf[i])) {
        i++;
    }
    
    if (i == MAX_MEM_BUF_DEPTH) i = 0xffff;

    return i;
}

#endif // DONT_NEED_THIS
