/*
 *  Copyright  2000-2001 Atheros Communications, Inc.,  All Rights Reserved.
 */

/* apvlan.h - functions for vlan support in the AP */

/* VLANs theory of operation:
 *  - vlan support in this driver is for privacy vlans i.e. vlans
 *    distinguished by their privacy policy
 *  - the default vlan is a tagless vlan; there can be only one
 *    default vlan; the ap host stack belongs only to the default
 *    tagless vlan for now
 *  - addPrivacyVlan is the configuration call which associates
 *    a vlan tag to a given crypto; a crypto implies - wep, aes,
 *    ..., or no crypto at all
 *  - on authentication/association, a station is allowed to
 *    authenticate/associate only if a valid vlan is present for
 *    the type of crypto it wants to do i.e. its privacy policy
 *    (see below)
 *  - if the AP's privacy policy is ON i.e. the AP is configured
 *    with privacyInvoked true, then unless an explicitly configured
 *    vlan for "no crypto" stations is present, all stations who
 *    do not want to use encryption are assigned VLAN_TAG_INVALID
 *    thereby preventing association/authentication with them
 *  - if the AP's privacy policy is OFF i.e. privacyInvoked is
 *    false, then unless there are explicitly configured vlan(s)
 *    for stations wanting to do crypto, all stations who want to
 *    do crypto are assigned VLAN_TAG_INVALID
 *  - on a station attempting authentication/association implying
 *    the use of a particular crypto for encryption (or no crypto),
 *    the sme code is expected to request getPrivacyVlanTag for
 *    the given keytype and refuse authentication/association if
 *    the tag supplied is VLAN_TAG_INVALID
 */

#ifndef _AP_VLAN_H_
#define _AP_VLAN_H_

typedef A_UINT16 VLAN_TAG;
#define VLAN_TAG_DEFAULT    0x0001
#define VLAN_TAG_INVALID    0x0fff

typedef enum VlanPrivacy {
    VLAN_PRIVACY_AES    = 0,
    VLAN_PRIVACY_WEP    = 1,
    VLAN_PRIVACY_NONE   = 2,

    VLAN_ETH_NATIVE     = 3,
    VLAN_LOCAL_STACK    = 4,

    VLAN_PRIVACY_NELEMENTS
} VLAN_PRIVACY;

void
initPrivacyVlan(WLAN_DEV_INFO* pdevInfo);

extern A_STATUS
setPrivacyVlan(WLAN_DEV_INFO* pdevInfo, VLAN_PRIVACY priv, VLAN_TAG tag);
extern VLAN_TAG
getPrivacyVlan(WLAN_DEV_INFO* pdevInfo, VLAN_PRIVACY priv);
extern A_UINT16
getMulticastKeyIdForVlan(WLAN_DEV_INFO* pdevInfo, VLAN_TAG tag);

extern void
getLanHdrInfoFromVlanHdr(LAN_FRAME_HEADER* pLanHdr, A_UINT16* pTypeOrLen, A_UINT16* pLanHdrLen);
extern VLAN_TAG
lookupVlanTag(WLAN_DEV_INFO* pdevInfo, LAN_FRAME_HEADER* pLanHdr);

extern A_UINT16
getEthVlanHdrLenForVlanTag(VLAN_TAG tag);
extern void
fillEthVlanHdr(LAN_FRAME_HEADER* pLanHdr, VLAN_TAG tag);

#ifdef VXWORKS // M_BLKS and Z_BULKS are vxworks internal bufs
extern M_BLK*
vlanEgressToDSM(M_BLK* pMblk, VLAN_TAG tag);
extern M_BLK*
vlanEgressToProto(M_BLK* pMblk, VLAN_TAG tag);
#endif

extern VLAN_TAG
apCfgVlanTagGet(VLAN_PRIVACY vp);

#ifdef VXWORKS
extern M_BLK*
vlanToProto(M_BLK* p_Mblk);
extern M_BLK*
vlanFromProto(M_BLK* p_Mblk);
#endif

extern VLAN_TAG
getVlanFromCsn(SIB_ENTRY *pSib);


#endif /* _AP_VLAN_H_ */

