/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * Routines used to initialize and generated beacons for the AR5211/AR5311.
 *
 * $Id: //depot/sw/releases/2.0.1/src/hal/ar5211/ar5211Beacon.c#1 $
 */

#ifdef BUILD_AR5211

#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/interrupt.h>   
   
/* Standard HAL Headers */
#include "wlantype.h"
#include "wlandrv.h"
#include "wlanchannel.h"
#include "halApi.h"
#include "hal.h"

/* Headers for HW private items */
#include "ar5hwcEnd.h"

#include "ar5211Reg.h"
#include "ar5211Beacon.h"
#include "ar5211.h"

#include "wlanbeacon.h"     /* BEACON_INFO */

extern struct pci_dev * gPciDev;

struct beacon_tasklet_param BeaconDevReset;
void resetChipForBeacons(A_UINT32 param);
void DisableInterruptsGlobaly   (WLAN_DEV_INFO *pdevInfo);
DECLARE_TASKLET(reset_beacon_tasklet, resetChipForBeacons, (A_UINT32)&BeaconDevReset);

#define MAX_BEACON_RETRIES 10

/**************************************************************
 * ar5211WaitForBeaconDone
 *
 * Waits for the beacon queue and CAB queue to finish.
 *
 * Returns: A_OK    if everything stopped successfully
 *          A_ERROR if something went wrong
 */
A_STATUS
ar5211WaitForBeaconDone(WLAN_DEV_INFO *pDev)
{
    A_UINT32    cabQStatus;
    BEACON_INFO *pInfo;

    ASSERT(pDev && pDev->pBeaconInfo);

    cabQStatus = MAC_Q0_STS + (pDev->cabQueueNum * sizeof(A_UINT32));
    pInfo      = pDev->pBeaconInfo;
									    
    /* Make sure that the beacon went out */
    if (readPlatformReg(pDev, MAC_Q_TXE) & (1 << pDev->beaconQueueNum)) {
        pInfo->missedBeaconCount++;
        if (pInfo->missedBeaconCount < MAX_BEACON_RETRIES) {
            A_UINT32 beaconPfc =
                readPlatformReg(pDev, MAC_Q0_STS + (pDev->beaconQueueNum * sizeof(A_UINT32))) &&
                MAC_Q_STS_PEND_FR_CNT_M;
#ifdef DEBUG
            //aPrintk("ar5211WaitForBeaconDone: TXE still high on beacon QCU. PFC = %ld (%d in a row).\n",
            //          beaconPfc, pInfo->missedBeaconCount);
#endif
        } else {
            /* If we feel we are stuck let beacon watchdog reboot us */
            if (pInfo->missedBeaconCount == MAX_BEACON_RETRIES) {
#ifdef DEBUG
                isrPrintf("ar5211WaitForBeaconDone: TXE on beacon QCU stuck.\n");
#endif
                struct net_device *pNetDev = NULL;
                if (NULL != gPciDev) pNetDev = pci_get_drvdata(gPciDev);
                if (NULL != pNetDev) {
                    
                    DisableInterruptsGlobaly (pDev);
                    printk("DISABLED INTS GLOBALY..in prep. for beacon reset (%p)\n", pDev);
                    // reset the h/w for this but not from an ISR context
                    BeaconDevReset.pdevInfo = pDev;
                    BeaconDevReset.dev      = pNetDev;
                    tasklet_schedule(&reset_beacon_tasklet);
                    // resetChipForBeacons(pDev, pNetDev);
                }
                else printk("ar5211WaitForBeaconDone: CANNOT RESET DEVICE...\n");

            }
        }
        return A_ERROR;
    } else {
        if (pInfo->missedBeaconCount) {
#ifdef DEBUG
            //aPrintk("ar5211WaitForBeaconDone: Resuming beacon transmission!\n");
#endif
        }
        pInfo->missedBeaconCount = 0;
    }

    /*
     * See if TXE is still asserted on the CAB queue.  It shouldn't be as its
     * ReadyTime should have expired prior to this SWBA.
     */
    if (readPlatformReg(pDev, MAC_Q_TXE) & (1 << pDev->cabQueueNum)) {
#ifdef DEBUG
        A_UINT32 pfc  = readPlatformReg(pDev, cabQStatus) & MAC_Q_STS_PEND_FR_CNT_M;
        A_UINT32 rtsd = readPlatformReg(pDev, MAC_Q_RDYTIMESHDN) & (1 << pDev->cabQueueNum);

        isrPrintf("TXE still high on CAB QCU, PFC = %ld, RTSD = %d\n", pfc, rtsd ? 1 : 0);
#endif

        /*
         * If we don't return A_OK, we will not queue up a new beacon.  Despite
         * appearances to the contrary given the documentation, TXE on the CAB
         * QCU will never deassert and we will end up never transmitting another
         * beacon and spewing lots of annoying messages to the console.
         */
    }
#ifdef DEBUG
    /* Clear the ReadyTime shutdown status bits */
    writePlatformReg(pDev, MAC_Q_RDYTIMESHDN, 1 << pDev->cabQueueNum);
#endif

    return A_OK;
}

/**************************************************************
 * ar5211UpdateBeaconDesc
 *
 * Updates the HW-specific fields of the beacon frame's descriptor.
 *
 * Assumes: frameLen does not include the FCS
 */
void
ar5211UpdateBeaconDesc(ATHEROS_DESC *pDesc, A_UINT32 frameLen)
{
    pDesc->hwControl.tx.frameLength  = frameLen + FCS_FIELD_SIZE;
    pDesc->hwControl.tx.bufferLength = frameLen;
    pDesc->hwStatus.word[0] = 0;
    pDesc->hwStatus.word[1] = 0;
}

/**************************************************************
 * ar5211QueueBeaconAndCab
 *
 * Queues the beacon and CAB (crap after beacon) to their respective
 * HW queues.
 *
 * Assumes: pBeacon is not NULL; pCab may be NULL
 */
void
ar5211QueueBeaconAndCab(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pBeacon, ATHEROS_DESC *pCab)
{
    A_UINT32 beaconQcuTxdp, cabQcuTxdp;

    ASSERT(pDev && pBeacon);

    beaconQcuTxdp = MAC_Q0_TXDP + (pDev->beaconQueueNum * sizeof(A_UINT32));
    cabQcuTxdp    = MAC_Q0_TXDP + (pDev->cabQueueNum * sizeof(A_UINT32));

    writePlatformReg(pDev, beaconQcuTxdp, pBeacon->thisPhysPtr);

    A_DATA_CACHE_FLUSH(pBeacon->pBufferVirtPtr,
                       pBeacon->hwControl.tx.bufferLength);
    A_DESC_CACHE_FLUSH(pBeacon);
    A_PIPEFLUSH();

    if (pCab != NULL) {
        writePlatformReg(pDev, cabQcuTxdp, pCab->thisPhysPtr);
        writePlatformReg(pDev, MAC_Q_TXE, 1 << pDev->cabQueueNum);
    }

    /*
     * Enable the beacon queue here to avoid the remotely possible race
     * condition where the beacon queue is ready, fires, and triggers
     * the CAB queue, but the CAB queue is not yet ready.
     */
    writePlatformReg(pDev, MAC_Q_TXE, 1 << pDev->beaconQueueNum);
}

/**************************************************************
 * ar5211SetupBeaconDesc
 *
 * Fills in the control fields of the beacon descriptor
 *
 * Assumes: frameLen does not include the FCS
 */
void
ar5211SetupBeaconDesc(ATHEROS_DESC *pDesc, A_BOOL isAp)
{
    AR5211_TX_CONTROL *pTxControl= (AR5211_TX_CONTROL *) &pDesc->hwControl.tx;

    if (isAp) {
        pDesc->pNextVirtPtr = NULL;
        pDesc->nextPhysPtr  = 0;
        pTxControl->Veol    = 0;

    } else {
        /* If we're a client, link the beacon descriptor back onto itself
         * with the Veol bit set
         */
        pDesc->pNextVirtPtr = pDesc;
        pDesc->nextPhysPtr  = pDesc->thisPhysPtr;
        pTxControl->Veol    = 1;
    }

    pTxControl->PktType = HAL_DESC_PKT_TYPE_BEACON;
    pTxControl->noAck   = 1;
}

#define DMA_BEACON_RESPONSE_TIME    2   /* 2  TUs */
#define SW_BEACON_RESPONSE_TIME     10  /* 10 TUs */
#define MAX_FRAME_TIME_ON_AIR       3   /* 3  TUs for max len frame @ 6Mbps */
#define MAX_PENDING_PCU_FRAMES      2   /* 2  frames sitting in the PCU ready to send */
#define ADDITIONAL_SWBA_BACKOFF     (MAX_FRAME_TIME_ON_AIR * MAX_PENDING_PCU_FRAMES)

/**************************************************************
 * ar5211BeaconInit
 *
 * Initializes all of the hardware registers used to send beacons.
 *
 */
void
ar5211BeaconInit(WLAN_DEV_INFO *pDev, A_UINT32 tsf, A_BOOL isAp)
{
    A_UINT32        value = 0;
    A_UINT32        beaconQOffset  = pDev->beaconQueueNum * sizeof(A_UINT32);
    A_UINT32        cabQOffset     = pDev->cabQueueNum * sizeof(A_UINT32);
    A_UINT32        beaconInterval = pDev->bssDescr->beaconInterval;
    A_UINT32        nextTbtt       = ( (tsf/beaconInterval) + 1) * beaconInterval;
    ATHEROS_DESC    *pDesc;
    A_UINT32        cwMin;

    /* Configure QCU for beacons */
    writePlatformReg(pDev, MAC_Q0_MISC + beaconQOffset,
        readPlatformReg(pDev, MAC_Q0_MISC + beaconQOffset) |
            MAC_Q_MISC_FSP_DBA_GATED | MAC_Q_MISC_BEACON_USE |
            MAC_Q_MISC_CBR_INCR_DIS1);

    /* Configure DCU for beacons */
    value = (MAC_D_MISC_ARB_LOCKOUT_CNTRL_GLOBAL << MAC_D_MISC_ARB_LOCKOUT_CNTRL_S) |
            MAC_D_MISC_BEACON_USE | MAC_D_MISC_POST_FR_BKOFF_DIS;
    if (pDev->macVersion < MAC_SREV_VERSION_OAHU) {
        value |= MAC_5311_D_MISC_SEQ_NUM_CONTROL;
    }
    writePlatformReg(pDev, MAC_D0_MISC + beaconQOffset, value);

    /* For adhoc beacons, set beacon DCU CW_MIN and CW_MAX to
     * twice the "usual" CW_MIN value */
    if (!isAp) {
        if ((pDev->staConfig.phwChannel->channelFlags & CHANNEL_B) == CHANNEL_B) {
            cwMin = INIT_CWMIN_11B;
        } else {
            cwMin = INIT_CWMIN;
        }
        writePlatformReg(pDev, MAC_D0_LCL_IFS + beaconQOffset,
                         ((2*cwMin+1) & MAC_D_LCL_IFS_CWMIN_M) |
                         (((2*cwMin+1) << MAC_D_LCL_IFS_CWMAX_S) & MAC_D_LCL_IFS_CWMAX_M) |
                         (( pDev->staConfig.aifs << MAC_D_LCL_IFS_AIFS_S) & MAC_D_LCL_IFS_AIFS_M) );

    }

    /* Configure QCU for CAB (Crap After Beacon) frames */
    writePlatformReg(pDev, MAC_Q0_MISC + cabQOffset,
        readPlatformReg(pDev, MAC_Q0_MISC + cabQOffset) |
            MAC_Q_MISC_FSP_DBA_GATED | MAC_Q_MISC_CBR_INCR_DIS1 |
            MAC_Q_MISC_CBR_INCR_DIS0 | MAC_Q_MISC_RDYTIME_EXP_POLICY);
    /*
     * We subtract ADDITIONAL_SWBA_BACKOFF from the ReadyTime to account for the
     * frames that might be sitting in the PCU that still have to go out on the
     * air at the slowest data rate.  We want TXE to go low before we hit SWBA
     * the next time around.
     */
    value = (beaconInterval - (SW_BEACON_RESPONSE_TIME - DMA_BEACON_RESPONSE_TIME) -
            ADDITIONAL_SWBA_BACKOFF) * 1024;
    writePlatformReg(pDev, MAC_Q0_RDYTIMECFG + cabQOffset, value | MAC_Q_RDYTIMECFG_EN);
    
    /* Configure DCU for CAB */
    value = (MAC_D_MISC_ARB_LOCKOUT_CNTRL_GLOBAL << MAC_D_MISC_ARB_LOCKOUT_CNTRL_S);
    if (pDev->macVersion < MAC_SREV_VERSION_OAHU) {
        value |= MAC_5311_D_MISC_SEQ_NUM_CONTROL;
    }
    writePlatformReg(pDev, MAC_D0_MISC + cabQOffset, value);
    
    /* Enable TXOK and TXERR ISRs for CAB */
    writePlatformReg(pDev, MAC_IMR_S0,
        readPlatformReg(pDev, MAC_IMR_S0) | (1 << pDev->cabQueueNum));  /* TXOK */
    writePlatformReg(pDev, MAC_IMR_S1,
        readPlatformReg(pDev, MAC_IMR_S1) | (1 << pDev->cabQueueNum));  /* TXERR */
    
    /* Set the next beacon time register */
    writePlatformReg(pDev, MAC_TIMER0, nextTbtt);
    
    /* Set DMA Beacon Alert (DBA) timer (in 1/8 TU = 128 us) */
    value = (nextTbtt - DMA_BEACON_RESPONSE_TIME) * 8;
    writePlatformReg(pDev, MAC_TIMER1, value);
    
    /* Set Software Beacon Alert (SWBA) timer (in 1/8 TU = 128 us) */
    value = (nextTbtt - SW_BEACON_RESPONSE_TIME) * 8;
    writePlatformReg(pDev, MAC_TIMER2, value);
    
    // Update active Q mask
    pDev->activeQueueMask |= (1 << pDev->beaconQueueNum) | (1 << pDev->cabQueueNum);

    if (!isAp) {
        /* The SWBA interrupt is not used for beacons in ad hoc mode
         * as we don't yet support ATIMs. So since the beacon never changes,
         * the beacon descriptor is set up once with the descriptor linked
         * to itself and the Veol bit set.
         */

        /* Set the ATIM window */
        value = nextTbtt + pDev->staConfig.atimWindow;
        writePlatformReg(pDev, MAC_TIMER3, value);
        
        /* Set beacon QCU TXDP */
        pDesc = pDev->pBeaconInfo->pBeaconDesc;
        writePlatformReg(pDev, MAC_Q0_TXDP + beaconQOffset, pDesc->thisPhysPtr);
        
        /* Enable the beacon queue */
        writePlatformReg(pDev, MAC_Q_TXE, 1 << pDev->beaconQueueNum);
        
    } else {
        /* Enable SWBA interrupt. */
        writePlatformReg(pDev, MAC_IER, 0);
        pDev->MaskReg |= MAC_IMR_SWBA;
        writePlatformReg(pDev, MAC_IMR, pDev->MaskReg);
        writePlatformReg(pDev, MAC_IER, MAC_IER_ENABLE);
    }

    /* Set the Beacon Control register */
    value = beaconInterval | MAC_BEACON_EN;
    if (tsf == 0) {
        value |= MAC_BEACON_RESET_TSF;
    }
    writePlatformReg(pDev, MAC_BEACON, value);
}

#endif /* BUILD_AR5211 */
