/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 *  Chips-specific device miscellaneous functions including hardware queries,
 *  EEPROM routines, gpio funcs, beacon creation, ...
 */

#ifdef BUILD_AR5211

#ident "$Id: //depot/sw/releases/2.0.1/src/hal/ar5211/ar5211Misc.c#4 $"

/* Standard HAL Headers */
#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"
#include "halUtil.h"
#include "wlanchannel.h"

/* Headers for HW private items */
#include "ar5hwcEnd.h"

#include "ar5211Reg.h"
#include "ar5211Misc.h"
#include "ar5211Reset.h"
#include "ar5211Power.h"
#include "ar5211.h"

#if defined(AR5311)
#include "ar5311reg.h"
#endif

#define A_MAX(x, y)         (((x) > (y)) ? (x) : (y))

static A_STATUS
ar5211EepromWrite(WLAN_DEV_INFO *pDev, A_UINT32 offset, A_UINT16 data);

/**************************************************************
 * ar5211SetRegulatoryDomain
 *
 * Attempt to change the cards operating regulatory domain to the given value
 * Returns: A_EINVAL for an unsupported regulatory domain.
 *          A_HARDWARE for an unwritable EEPROM or bad EEPROM version
 */
A_STATUS
ar5211SetRegulatoryDomain(WLAN_DEV_INFO *pDev, A_UINT16 regDomain)
{
    A_STATUS status = A_OK;
    EEP_MAP *pEep;

    ASSERT(pDev->pHalInfo && pDev->pHalInfo->pEepData);

    pEep = pDev->pHalInfo->pEepData;
    if (pEep->regDmnEnum != regDomain) {
        if (!(pEep->protect & EEPROM_PROTECT_WP_128_191))
        {
            if(wlanIsRegCcValid(pDev)) {
                status = ar5211EepromWrite(pDev, REGULATORY_DOMAIN_OFFSET, regDomain);
                if (status == A_OK) {
                    pEep->regDmnEnum = regDomain;
                } else {
                    uiPrintf("Failed to write EEPROM\n");
                    status = A_HARDWARE;
                }
            } else {
                uiPrintf("EEPROM does not contain matching information for this country setting\n");
                status = A_EINVAL;
            }
        } else {
            uiPrintf("Protect bits set or incorrect EEPROM version\n");
            status = A_HARDWARE;
        }
    } else {
        status = A_EINVAL;
    }
    if (status != A_OK) {
        uiPrintf("Cannot change current EEPROM Country/Reg 0x%02X to given value 0x%02X\n",
                 pEep->regDmnEnum, regDomain);
    }

    return status;
}

/**************************************************************
 * ar5211GetRegDmn
 *
 * Pull the current regulatory domain out of hardware.
 */
A_UINT16
ar5211GetRegDmn(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev->pHalInfo && pDev->pHalInfo->pEepData);
    return pDev->pHalInfo->pEepData->regDmnEnum;
}

/**************************************************************
 * ar5211GetRfKill
 *
 * Accessor to get rfkill from private EEPROM structure
 */
A_BOOL
ar5211GetRfKill(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev->pHalInfo && pDev->pHalInfo->pEepData);
    return (A_BOOL)pDev->pHalInfo->pEepData->rfKill;
}

/**************************************************************
 * ar5211GetTurboDisable
 *
 * Accessor to get turbo disable setting from private EEPROM structure
 */
A_BOOL
ar5211GetTurboDisable(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev->pHalInfo && pDev->pHalInfo->pEepData);
    return (A_BOOL)pDev->pHalInfo->pEepData->pEepHeader->turboDisable;
}

/**************************************************************
 * ar5211GetMacAddr
 *
 * Attempt to get the MAC address from the wireless EEPROM
 * Returns: A_HARDWARE for an unreadable EEPROM
 */
A_STATUS
ar5211GetMacAddr(WLAN_DEV_INFO *pDev, WLAN_MACADDR *mac)
{
#if defined(PCI_INTERFACE)
    A_UINT32 total = 0;
    A_UINT16 half;

    if (ar5211EepromRead(pDev, EEPROM_MAC_MSW_OFFSET, &half) != A_OK) {
        goto ee_error;
    }
    total += half;
    mac->octets[1] = half & 0xff;
    mac->octets[0] = half >> 8;

    if (ar5211EepromRead(pDev, EEPROM_MAC_MID_OFFSET, &half) != A_OK) {
        goto ee_error;
    }
    total += half;
    mac->octets[3] = half & 0xff;
    mac->octets[2] = half >> 8;
    if (ar5211EepromRead(pDev, EEPROM_MAC_LSW_OFFSET, &half) != A_OK) {
        goto ee_error;
    }
    total += half;
    mac->octets[5] = half & 0xff;
    mac->octets[4] = half >> 8;

    if ((total == 0) || total == (3*0xffff)) {
        uiPrintf("ar5211GetMacAddr: EEPROM MAC is all 0's or all F's\n");
        return A_HARDWARE;
    }

    printk("Atheros ar5211 MacAddr : %x:%x:%x:%x:%x:%x\n", 
           mac->octets[0], mac->octets[1], mac->octets[2], 
           mac->octets[3], mac->octets[4], mac->octets[5]);


    return A_OK;

ee_error:
    uiPrintf("ar5211GetMacAddr: EEPROM read failed\n");
    return A_HARDWARE;
#elif defined(AR5311)
    /* The ar5311 may store the WLAN mac addr in the system board config */
    A_BCOPY(sysBoardData.wlanMac, mac->octets, 6);
    return A_OK;
#endif
}

/**************************************************************
 * ar5211EepromRead
 *
 * Read 16 bits of data from offset into *data
 */
A_STATUS
ar5211EepromRead(WLAN_DEV_INFO *pDev, A_UINT32 offset, A_UINT16 *data)
{
#if defined(PCI_INTERFACE)
    A_UINT32 status;
    int to = 10000;     // 10ms timeout

    writePlatformReg(pDev, MAC_EEPROM_ADDR, offset);
    writePlatformReg(pDev, MAC_EEPROM_CMD, MAC_EEPROM_CMD_READ);

    while (to > 0) {
        udelay(1);
        status = readPlatformReg(pDev, MAC_EEPROM_STS);
        if (status & MAC_EEPROM_STS_READ_COMPLETE) {
            status = readPlatformReg(pDev, MAC_EEPROM_DATA);
            *data = (A_UINT16)(status & 0xffff);
            return A_OK;
        }

        if (status & MAC_EEPROM_STS_READ_ERROR) {
            uiPrintf("ar5211EepromRead: eeprom read error at offset %ld.\n", 
                     (unsigned long)offset);
            return A_HARDWARE;
        }
        to--;
    }
    uiPrintf("ar5211EepromRead: eeprom read timeout at offset %ld.\n", 
             (unsigned long)offset);
    return A_HARDWARE;
#elif defined(AR5311)
    /* radio configuration data is stored in the system flash */
    *data = sysFlashConfigRead(FLC_RADIOCFG, (offset * 2) + 1) |
            (sysFlashConfigRead(FLC_RADIOCFG, offset * 2) << 8);
    return A_OK;
#endif /* PLATFORM */
}

/**************************************************************
 * ar5211EepromWrite
 *
 * Read 16 bits of data from offset into *data
 */
A_STATUS
ar5211EepromWrite(WLAN_DEV_INFO *pDev, A_UINT32 offset, A_UINT16 data)
{
#if defined(PCI_INTERFACE)
    A_UINT32 status;
    int to = 15000;     // 15ms timeout

    // send write data
    writePlatformReg(pDev, MAC_EEPROM_ADDR, offset);
    writePlatformReg(pDev, MAC_EEPROM_DATA, data);
    writePlatformReg(pDev, MAC_EEPROM_CMD, MAC_EEPROM_CMD_WRITE);

    while (to > 0) {
        udelay(1);
        status = readPlatformReg(pDev, MAC_EEPROM_STS);
        if (status & MAC_EEPROM_STS_WRITE_COMPLETE) {
            return A_OK;
        }

        if (status & MAC_EEPROM_STS_WRITE_ERROR)    {
            uiPrintf("ar5211EepromWrite: eeprom write error at offset %ld.\n", 
                     (unsigned long)offset);
            return A_HARDWARE;;
        }
        to--;
    }

    uiPrintf("ar5211EepromWrite: eeprom write timeout at offset %ld.\n", 
             (unsigned long)offset);
    return A_HARDWARE;
#elif defined(AR5311)
    char str[2];

    /* Radio configuration data is stored in system flash (in reverse endian) */
    str[0] = (data >> 8) & 0xff;
    str[1] = data & 0xff;
    sysFlashConfigWrite(FLC_RADIOCFG, offset<<1, str, 2);

    return A_OK;
#endif /* PLATFORM */
}

/**************************************************************************
 * ar5211EnableRfKill
 *
 * Called if RfKill is supported (according to EEPROM).  Set the interrupt and
 * GPIO values so the ISR and can disable RF on a switch signal
 */
void
ar5211EnableRfKill(WLAN_DEV_INFO *pDev)
{
    /* TODO - can this really be above the hal on the GPIO interface for
     * TODO - the client only?
     */
#if !defined(BUILD_AP)  /* AP uses GPIOs for LEDs and buttons */
    // if radio disable switch connection to GPIO bit 0 is enabled
    // program GPIO interrupt.
    // If rfkill bit on eeprom is 1, setupeeprommap routine has already
    // verified that it is a later version of eeprom, it has a place for
    // rfkill bit and it is set to 1, indicating that GPIO bit 0 hardware
    // connection is present.

    if ( (pDev->gpio0Bit = ar5211GpioGet(pDev, 0)) == 0 ) {
        // switch already closed, force no rf
        ar5211Gpio0SetIntr(pDev, 1);
    } else {
        ar5211Gpio0SetIntr(pDev, 0);
    }
#endif /* !BUILD_AP */
}

/**************************************************************
 * ar5211GpioCfgOutput
 *
 * Configure GPIO Output lines
 */
void
ar5211GpioCfgOutput(WLAN_DEV_INFO *pDev, A_UINT32 gpio)
{
#if defined(PCI_INTERFACE)
    A_UINT32 reg;

    ASSERT(gpio < MAC_NUM_GPIO);

    reg =  readPlatformReg(pDev, MAC_GPIOCR);
    reg &= ~(MAC_GPIOCR_0_CR_A << (gpio * MAC_GPIOCR_CR_SHIFT));
    reg |= MAC_GPIOCR_0_CR_1 << (gpio * MAC_GPIOCR_CR_SHIFT);

    writePlatformReg(pDev, MAC_GPIOCR, reg);
#endif /* PCI_INTERFACE */
}

/**************************************************************
 * ar5211GpioCfgInput
 *
 * Configure GPIO Input lines
 */
void
ar5211GpioCfgInput(WLAN_DEV_INFO *pDev, A_UINT32 gpio)
{
#if defined(PCI_INTERFACE)
    A_UINT32 reg;

    ASSERT(gpio < MAC_NUM_GPIO);

    reg =  readPlatformReg(pDev, MAC_GPIOCR);
    reg &= ~(MAC_GPIOCR_0_CR_A << (gpio * MAC_GPIOCR_CR_SHIFT));
    reg |= MAC_GPIOCR_0_CR_N << (gpio * MAC_GPIOCR_CR_SHIFT);

    writePlatformReg(pDev, MAC_GPIOCR, reg);
#endif /* PCI_INTERFACE */
}

/**************************************************************
 * ar5211GpioSet
 *
 * Once configured for I/O - set output lines
 */
void
ar5211GpioSet(WLAN_DEV_INFO *pDev, A_UINT32 gpio, A_UINT32 val)
{
#if defined(PCI_INTERFACE)
    A_UINT32 reg;

    ASSERT(gpio < MAC_NUM_GPIO);

    reg =  readPlatformReg(pDev, MAC_GPIODO);
    reg &= ~(1 << gpio);
    reg |= (val&1) << gpio;

    writePlatformReg(pDev, MAC_GPIODO, reg);
#endif /* PCI_INTERFACE */
}

/**************************************************************
 * ar5211GpioGet
 *
 * Once configured for I/O - get input lines
 */
A_UINT32
ar5211GpioGet(WLAN_DEV_INFO *pDev, A_UINT32 gpio)
{
#if defined(PCI_INTERFACE)
    A_UINT32 reg;

    ASSERT(gpio < MAC_NUM_GPIO);

    reg =  readPlatformReg(pDev, MAC_GPIODI);
    reg = ((reg & MAC_GPIOD_MASK) >> gpio) & 0x1;

    return reg;
#else /* !PCI_INTERFACE */
    return 0;
#endif /* PCI_INTERFACE */
}

/**************************************************************
 * ar5211Gpio0SetIntr
 *
 * Set the GPIO 0 Interrupt
 */
void
ar5211Gpio0SetIntr(WLAN_DEV_INFO *pDev, A_UINT32 ilevel)
{
#if defined(PCI_INTERFACE)
    A_UINT32    reg;

    reg = readPlatformReg(pDev, MAC_GPIOCR);
    // clear the bits that we will modify
    reg &= ~(MAC_GPIOCR_INT_SEL0 | MAC_GPIOCR_INT_EN | MAC_GPIOCR_0_CR_A | MAC_GPIOCR_INT_SELH);
    reg |= MAC_GPIOCR_INT_SEL0 | MAC_GPIOCR_INT_EN | MAC_GPIOCR_0_CR_N;
    if (ilevel) {
        reg |= MAC_GPIOCR_INT_SELH;
    }
    // don't need to change anything for low level interrupt.
    writePlatformReg(pDev, MAC_GPIOCR, reg);
    // change the interrupt mask
    writePlatformReg(pDev, MAC_IMR, readPlatformReg(pDev, MAC_IMR) | MAC_IMR_GPIO);
    pDev->MaskReg = readPlatformReg(pDev, MAC_IMR);
#endif /* PCI_INTERFACE */
}

/******************************************************************
 * ar5211SetLedState
 *
 * Change the LED blinking pattern to correspond to the connectivity
 */
void
ar5211SetLedState(WLAN_DEV_INFO *pDev, A_BOOL bConnected)
{
#if defined(PCI_INTERFACE)
    A_UINT32 reg = readPlatformReg(pDev, MAC_PCICFG) & ~MAC_PCICFG_ASSOC_STATUS_M;

    if (bConnected) {
        /* normal blink when connected... */
        reg |= MAC_PCICFG_ASSOC_STATUS_ASSOCIATED << MAC_PCICFG_ASSOC_STATUS_S;
    } else {
        /* alternate blink when not. */
        reg |= MAC_PCICFG_ASSOC_STATUS_PENDING << MAC_PCICFG_ASSOC_STATUS_S;
    }
    writePlatformReg(pDev, MAC_PCICFG, reg);
#elif defined(AR5311)
    A_UINT32 reg = sysRegRead(AR5311_PCICFG) & ~LED_MODE_M;

    if (bConnected) {
        reg |= ASSOC_STATUS_ASSOCIATED;
    } else {
        reg |= ASSOC_STATUS_PENDING;
    }
    sysRegWrite(AR5311_PCICFG, reg);
#endif /* PLATFORM */
}

/**************************************************************
 * ar5211AntennaGet
 *
 * Return 1 or 2 for the corresponding antenna that is in use
 */
int
ar5211AntennaGet(WLAN_DEV_INFO *pDev)
{
    return 0;                   // TODO ar5211AntennaGet
}

/**************************************************************************
 * ar5211WriteAssocid - Change association related fields programmed into the hardware.
 *
 * Writing a valid BSSID to the hardware effectively enables the hardware
 * to synchronize its TSF to the correct beacons and receive frames coming
 * from that BSSID. It is called by the SME JOIN operation.
 */
void
ar5211WriteAssocid(WLAN_DEV_INFO *pDev, WLAN_MACADDR *bssid, A_UINT16 assocId, A_UINT16 timOffset)
{
    A_UINT32    ulAddressLow = 0;   /* default values, if bssid is not passed in */
    A_UINT32    ulAddressHigh = 0;

#ifdef BIG_ENDIAN
    if (bssid != NULL) {
        ulAddressLow  = cpu2le32(*((A_UINT32 *) &(bssid->octets[0])));
        ulAddressHigh = 0x0000ffff & (cpu2le16((*((A_UINT16 *)(&bssid->octets[4])))));
    }
    writePlatformReg(pDev, MAC_BSS_ID0, ulAddressLow);
    writePlatformReg(pDev, MAC_BSS_ID1, (ulAddressHigh | ((assocId & 0x3fff) << 16)));
#else
    if (bssid != NULL) {
        ulAddressLow  = *((A_UINT32 *) &(bssid->octets[0]));
        ulAddressHigh = 0x0000ffff & (*((A_UINT16 *)&(bssid->octets[4])));
    }
    writePlatformReg(pDev, MAC_BSS_ID0, ulAddressLow);
    writePlatformReg(pDev, MAC_BSS_ID1, (ulAddressHigh | ((assocId & 0x3fff) << 16)));
#endif /* BIG_ENDIAN */

    if (assocId != 0) {
        /* Now set the TIM_OFFSET. We are starting a new association
         * Enable PS-Polls
         */
        writePlatformReg(pDev, MAC_BEACON,
                (readPlatformReg(pDev, MAC_BEACON) & ~MAC_BEACON_TIM_MASK) |
                ((((A_UINT32)(timOffset ? (timOffset + 4) : 0) << MAC_BEACON_TIM_S)) & MAC_BEACON_TIM_MASK));

        ar5211EnablePSPoll(pDev, bssid, assocId);

    } else {
        /* cancel association, if any. Disable PS-Polls*/
        ar5211DisablePSPoll(pDev);
    }
}


/**************************************************************
 * ar5211SetStaBeaconTimers
 *
 *  sets all the beacon related bits on the h/w for stations
 *  i.e. initializes the corresponding h/w timers;
 *  also tells the h/w whether to anticipate PCF beacons
 *
 *  tsf - current tsf in TU or msec.
 *  dtim_count and cfp_count from the current beacon - their current
 *  values aren't necessarily maintained in the device struct
 */
void
ar5211SetStaBeaconTimers(WLAN_DEV_INFO *pDev, A_UINT32 tsf, A_UINT32 dtim_count, A_UINT32 cfp_count)
{
    BSSDESCR        *bp = pDev->bssDescr;
    CF_PARAM_SET    *cfp = &bp->cfParamSet;
    A_UINT32        cfp_period, time_to_cfp;
    A_UINT32        nextTbtt = ( (tsf/bp->beaconInterval) + 1) * bp->beaconInterval;
    A_UINT32        bmissTime = pDev->staConfig.noBeaconTimeout;
    A_UINT32        bmissCount;

    uiPrintf("set_sta_pcf_mode: DEBUG: setting beacon timers\n");

    /* set the TIMER0(1.024ms) register to anticipated time of the next beacon */
    writePlatformReg(pDev, MAC_TIMER0, nextTbtt);

    /* if the AP will do PCF */
    if (VALID_CF_ELEMENT(cfp)) {
        /* tell the h/w that the associated AP is PCF capable */
        writePlatformReg(pDev, MAC_STA_ID1,
                         readPlatformReg(pDev, MAC_STA_ID1) | MAC_STA_ID1_PCF);

        /* set CFP_PERIOD(1.024ms) register */
        cfp_period = cfp->cfpPeriod * bp->DTIMPeriod * bp->beaconInterval;
        writePlatformReg(pDev, MAC_CFP_PERIOD, cfp_period);

        /* set CFP_DUR(1.024ms) register to max cfp duration */
        writePlatformReg(pDev, MAC_CFP_DUR, cfp->cfpMaxDuration);

        /* set TIMER2(128us) to anticipated time of next CFP */
        time_to_cfp = (cfp_count * bp->DTIMPeriod + dtim_count) * bp->beaconInterval;
        if (time_to_cfp == 0) {
            time_to_cfp = cfp_period;
        }
        writePlatformReg(pDev, MAC_TIMER2, (tsf + time_to_cfp) << 3);
    } else {
        /* tell the h/w that the associated AP is not PCF capable */
        writePlatformReg(pDev, MAC_STA_ID1,
                         readPlatformReg(pDev, MAC_STA_ID1) & ~MAC_STA_ID1_PCF);
    }

    /* Start the beacon timers by setting the BEACON register to the beacon
     * interval; also write the tim offset which we should know by now.  The
     * code, in ar5211WriteAssocid, also sets the tim offset once the AID is
     * known which can be left as such for now.
     */
    writePlatformReg(
        pDev,
        MAC_BEACON,
        (readPlatformReg(pDev, MAC_BEACON) &
            (~(MAC_BEACON_PERIOD_MASK | MAC_BEACON_TIM_MASK))) |
            (((A_UINT32)bp->beaconInterval) & MAC_BEACON_PERIOD_MASK) |
            ((((A_UINT32)(bp->timOffset ? (bp->timOffset + 4) : 0)) << MAC_BEACON_TIM_S) & MAC_BEACON_TIM_MASK));

    /*
     * Calculate the number of consecutive beacons to miss before HW issues
     * a BMISS interrupt.  The beacon intenrval is in TUs.  bmissTime is in
     * milliseconds. We take the maximum of the calculated number of beacons
     * or 10 (in case of long beacon intervals).
     */
    ASSERT(bmissTime > 0);
    bmissTime  = bmissTime * 1000 / 1024;
    bmissCount = bmissTime / pDev->bssDescr->beaconInterval;
    bmissCount = (bmissCount & 1) ? bmissCount + 1 : bmissCount;
    bmissCount = A_MAX(bmissCount, 10) / 2;
    ASSERT(bmissCount <= (MAC_RSSI_THR_BM_THR_M >> MAC_RSSI_THR_BM_THR_S));

    /* Clear any pending BMISS interrupt so far */
    pDev->globISRReg &= ~MAC_IMR_BMISS;

    writePlatformReg(pDev, MAC_RSSI_THR, 
                     (readPlatformReg(pDev, MAC_RSSI_THR) & ~MAC_RSSI_THR_BM_THR_M) |
                     (bmissCount << MAC_RSSI_THR_BM_THR_S));

    /* Now enable the interrupt */
    halDisableInterrupts(pDev, HAL_INT_GLOBAL);
    pDev->MaskReg |= MAC_IMR_BMISS;
    writePlatformReg(pDev, MAC_IMR, pDev->MaskReg);
    halEnableInterrupts(pDev, HAL_INT_GLOBAL);
    ASSERT(pDev->MaskReg == readPlatformReg(pDev, MAC_IMR));


}

/******************************************************************
 * ar5211SetBmiss
 *
 * This routine sets missing beacon counter and interrupt bit
 * It's called during SME JOIN
 * We don't really need to clear it, it will be cleared by reset.
 */
void
ar5211SetBmiss(WLAN_DEV_INFO *pDev, A_UINT16 count)
{
    A_UINT32    temp;

    count &= 0xFF; // Keep only lowest 8 bits since it's a 8 bit counter

    if (count == 0) {
        // count of 0 with BMISS interrupt enabled will hang the system
        // with too many interrupts
        return;
    }

    // first disable the interrupt
    halDisableInterrupts(pDev, HAL_INT_GLOBAL);
    pDev->MaskReg &= ~MAC_IMR_BMISS;
    writePlatformReg(pDev, MAC_IMR, pDev->MaskReg);
    // clear any pending BMISS interrupt so far
    pDev->globISRReg &= ~MAC_IMR_BMISS;

    temp =  readPlatformReg(pDev, MAC_RSSI_THR) & ~(0x0000FF00);
    temp |= count << 8;
    writePlatformReg(pDev, MAC_RSSI_THR, temp);

    // now enable the interrupt
    pDev->MaskReg |= MAC_IMR_BMISS;
    writePlatformReg(pDev, MAC_IMR, pDev->MaskReg);
    halEnableInterrupts(pDev, HAL_INT_GLOBAL);
    ASSERT(pDev->MaskReg == readPlatformReg(pDev, MAC_IMR));
}

/******************************************************************
 * ar5211GetTsf
 *
 * Get the current hardware tsf for stamlme
 */
void
ar5211GetTsf(WLAN_DEV_INFO *pDev, WLAN_TIMESTAMP *tsf)
{
    tsf->high = readPlatformReg(pDev, MAC_TSF_U32);
    tsf->low = readPlatformReg(pDev, MAC_TSF_L32);
}

/******************************************************************
 * ar5211ResetTsf
 *
 * Reset the current hardware tsf for stamlme
 */
void
ar5211ResetTsf(WLAN_DEV_INFO *pDev)
{
    writePlatformReg(pDev, MAC_BEACON,
                    (readPlatformReg(pDev, MAC_BEACON) | MAC_BEACON_RESET_TSF));
}

/******************************************************************
 * ar5211SetAdhocMode
 *
 * Set adhoc mode for stamlme
 */
void
ar5211SetAdhocMode(WLAN_DEV_INFO *pDev)
{
    A_UINT32 id1 = readPlatformReg(pDev, MAC_STA_ID1);

    writePlatformReg(pDev, MAC_CFG,
                     readPlatformReg(pDev, MAC_CFG) | MAC_CFG_AP_ADHOC_INDICATION);

    /* Update antenna mode */
    id1 &= ~(MAC_STA_ID1_USE_DEFANT | MAC_STA_ID1_DEFANT_UPDATE);
    writePlatformReg(pDev, MAC_STA_ID1, id1 |
                     MAC_STA_ID1_AD_HOC |
                     MAC_STA_ID1_RTS_USE_DEF);
}

/**************************************************************
 * ar5211GetChannelEdges
 *
 * Store the channel edges for the requested operational mode
 */
A_BOOL
ar5211GetChannelEdges(WLAN_DEV_INFO *pDev, A_UINT16 channelFlags, A_UINT16 *pLowChannel, A_UINT16 *pHighChannel)
{
    A_BOOL bMode = FALSE;

    bMode = (A_BOOL)pDev->pHalInfo->pEepData->pEepHeader->Bmode;
    if (CHANNEL_5GHZ & channelFlags) {
        *pLowChannel = 5120;
        *pHighChannel = 6100;

        return TRUE;
    }
    else if ((CHANNEL_2GHZ & channelFlags) && (bMode)) {
        *pLowChannel = 2412;
        *pHighChannel = 2732;
        return TRUE;
    }
    else {
        return FALSE;
    }
}

/**************************************************************
 * ar5211GetMaxTurboRate
 *
 * Return the maximum rate for running turbo
 */
A_UINT32
ar5211GetMaxTurboRate(void)
{
    return 54;
}

/**************************************************************
 * ar5211GetRandomSeed
 *
 * Grab a semi-random value from hardware registers - may not
 * change often
 */

A_UINT32
ar5211GetRandomSeed(WLAN_DEV_INFO *pDev)
{
    A_UINT32 seed, nf;

    nf = (readPlatformReg(pDev, PHY_BASE+(25<<2)) >> 19) & 0x1ff;
    if (nf & 0x100) {
        nf = 0 - ((nf ^ 0x1ff) + 1);
    }

    seed = readPlatformReg(pDev, MAC_TSF_U32) ^ readPlatformReg(pDev, MAC_TSF_L32) ^
        nf;

    return seed;
}

/**************************************************************
 * ar5211DetectCardPresent
 *
 * Detect if our card is present
 */

A_BOOL
ar5211DetectCardPresent(WLAN_DEV_INFO *pDev)
{
#if defined(PCI_INTERFACE)
    A_UINT32 value;
    A_UINT16  macVersion, macRev;

    // Read the SREV register and compare to what we read at initialization
    value     = readPlatformReg(pDev, MAC_SREV) & MAC_SREV_ID_M;
    macVersion= (A_UINT16) (value >> MAC_SREV_ID_S);
    macRev    = (A_UINT16) (value & MAC_SREV_REVISION_M);

    if ( (pDev->macVersion == macVersion) && (pDev->macRev == macRev) ) {
        return TRUE;
    } else {
        return FALSE;
    }
#else
    return TRUE;                        /* the spirit mac is not going anywhere */
#endif
}

/**************************************************************
 * ar5211UpdateMibCounters
 *
 * Update MIB Counters
 */

void
ar5211UpdateMibCounters(WLAN_DEV_INFO *pDev, A_BOOL bUpdateAll)
{
    pDev->localSta->stats.AckRcvFailures += readPlatformReg(pDev, MAC_ACK_FAIL);
    pDev->localSta->stats.RtsFailCnt     += readPlatformReg(pDev, MAC_RTS_FAIL);
    pDev->localSta->stats.FcsFailCnt     += readPlatformReg(pDev, MAC_FCS_FAIL);
    pDev->localSta->stats.RtsSuccessCnt  += readPlatformReg(pDev, MAC_RTS_OK);

    if (bUpdateAll) {
        // Read MIB variables that change less frequently.
        pDev->localSta->stats.BeaconCount += readPlatformReg(pDev, MAC_BEACON_CNT);
    }
}

/**************************************************************
 * ar5211IsHwAesSupported
 *
 * Detect if HW AES supported
 */
A_BOOL
ar5211IsHwAesSupported(WLAN_DEV_INFO *pDev)
{
    return TRUE;
}

/**************************************************************
 * ar5211IsSwAesSupported
 *
 * Detect if SW AES supported, i.e. if HW AES engine is stoppable
 */
A_BOOL
ar5211IsSwAesSupported(WLAN_DEV_INFO *pDev)
{
    return TRUE;
}

/**************************************************************
 * ar5211EnableHwEncryption
 *
 * Enable HW encryption engine
 */
void
ar5211EnableHwEncryption(WLAN_DEV_INFO *pDev)
{
    return;   
}

/**************************************************************
 * ar5211DisableHwEncryption
 *
 * Disable HW encryption engine
 */
void
ar5211DisableHwEncryption(WLAN_DEV_INFO *pDev)
{
    return;  
}

/**************************************************************
 * ar5211EnableRadarDetection
 *
 * Enable radar detection in the hardware - may later be changed
 * to modify radar "sensitivity"
 */
void
ar5211EnableRadarDetection(WLAN_DEV_INFO *pDev)
{
    /* Enable pulse detection */
    writePlatformReg(pDev, PHY_RADAR_0, readPlatformReg(pDev, PHY_RADAR_0) |
        PHY_RADAR_0_EN);

    /* Enable RXPHY interrupt. */
    writePlatformReg(pDev, MAC_IER, 0);
    pDev->MaskReg |= MAC_IMR_RXPHY;
    writePlatformReg(pDev, MAC_IMR, pDev->MaskReg);
    writePlatformReg(pDev, MAC_IER, MAC_IER_ENABLE);
}

#define DUMP_REG_SET(x, y, z)  { int i; for (i = (y); i <= (z); i+=4) { uiPrintf("=== 0x%04X: 0x%08lX\n", i, (unsigned long)readPlatformReg((x), i)); }}

/**************************************************************
 * ar5211DisableHwEncryption
 *
 * Disable HW encryption engine
 */
void
ar5211DumpRegisters(WLAN_DEV_INFO *pDev)
{
    uiPrintf("MAC Registers\n");
    DUMP_REG_SET(pDev, 0x0008, 0x00b4)
    uiPrintf("\nQCU Registers\n");
    DUMP_REG_SET(pDev, 0x0800, 0x0a40)
    uiPrintf("\nDCU Registers\n");
    DUMP_REG_SET(pDev, 0x1000, 0x10F0)
    DUMP_REG_SET(pDev, 0x1230, 0x1230)
    uiPrintf("\nEEP Registers\n");
    DUMP_REG_SET(pDev, 0x6000, 0x6010)
    uiPrintf("\nPCU Registers\n");
    DUMP_REG_SET(pDev, 0x8000, 0x8058)
    uiPrintf("\nBB Registers\n");
    DUMP_REG_SET(pDev, 0x9800, 0x9878)
    DUMP_REG_SET(pDev, 0x9900, 0x995C)
    DUMP_REG_SET(pDev, 0x9C00, 0x9C1C)
}

/******************************************************************
 * ar5211GetCurRssi
 *
 * Get the rssi of frame curently being received.
 */
A_UINT32
ar5211GetCurRssi(WLAN_DEV_INFO *pDev)
{
    return (readPlatformReg(pDev, PHY_CURRENT_RSSI) & 0xff);
}

A_UINT32
ar5211GetDefAntenna(WLAN_DEV_INFO *pDev)
{
    return (readPlatformReg(pDev, MAC_DEF_ANTENNA) & 0x7);
}

void
ar5211SetDefAntenna(WLAN_DEV_INFO *pDev, A_UINT32 antenna)
{
    writePlatformReg(pDev, MAC_DEF_ANTENNA, (antenna & 0x7));
}

#endif // #ifdef BUILD_AR5211

