/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 *  Chips-specific power management routines.
 */

#ifdef BUILD_AR5211

#ident "$Id: //depot/sw/releases/2.0.1/src/hal/ar5211/ar5211Power.c#1 $"

/* Standard HAL Headers */
#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"
#include "halUtil.h"

/* Headers for HW private items */
#include "ar5hwcEnd.h"

#include "ar5211Reg.h"
#include "ar5211Power.h"
#include "ar5211.h"

/**************************************************************
 * ar5211SetPowerModeAuto
 *
 * Notify Power Mgt is disabled in self-generated frames.
 * If requested, set Power Mode of chip to auto/normal.
 * Duration in units of 128us (1/8 TU)
 */
void
ar5211SetPowerModeAuto(WLAN_DEV_INFO *pDev, A_BOOL setChip, A_UINT16 sleepDuration)
{
    writePlatformReg(pDev, MAC_STA_ID1,
                    readPlatformReg(pDev, MAC_STA_ID1) | MAC_STA_ID1_PWR_SAV);

#if defined(PCI_INTERFACE)
    if (setChip)
    {
        writePlatformReg(pDev, MAC_SCR, MAC_SCR_SLE_NORMAL | sleepDuration);
    }

#endif
}

/**************************************************************
 * ar5211SetPowerModeAwake
 *
 * Notify Power Mgt is enabled in self-generated frames.
 * If requested, force chip awake.
 * 
 * Returns A_OK if chip is awake or successfully forced awake.
 *
 * WARNING WARNING WARNING
 * There is a problem with the chip where sometimes it will not wake up.
 */
A_STATUS
ar5211SetPowerModeAwake(WLAN_DEV_INFO *pDev, A_BOOL setChip)
{
#if defined(PCI_INTERFACE)
    A_STATUS status = A_HARDWARE;
    A_UINT32 i;

    if (setChip) {
        writePlatformReg(pDev, MAC_SCR, MAC_SCR_SLE_FWAKE);
        udelay(1000);  //used to be 10, Give chip the chance to awake

        for ( i = 0; i < POWER_UP_TIME / 200; i++) {
            if ( (readPlatformReg(pDev, MAC_PCICFG) & MAC_PCICFG_SPWR_DN) == 0 ) {
                status = A_OK;
                break;
            }
            udelay(200);
            writePlatformReg(pDev, MAC_SCR, MAC_SCR_SLE_FWAKE);
        }
        if (status != A_OK) {
            uiPrintf("ar5211SetPowerModeAwake: Failed to leave sleep\n");
        }
    }
    else {
        status = A_OK;
    }

    if (status == A_OK) {
        writePlatformReg(pDev, MAC_STA_ID1,
            readPlatformReg(pDev, MAC_STA_ID1) & ~MAC_STA_ID1_PWR_SAV);
    }

    return status;
#else
    return A_OK;
#endif
}

/**************************************************************
 * ar5211SetPowerModeSleep
 *
 * Notify Power Mgt is disabled in self-generated frames.
 * If requested, force chip to sleep.
 */
void
ar5211SetPowerModeSleep(WLAN_DEV_INFO *pDev, A_BOOL setChip)
{
    writePlatformReg(pDev, MAC_STA_ID1,
                    readPlatformReg(pDev, MAC_STA_ID1) | MAC_STA_ID1_PWR_SAV);

#if defined(PCI_INTERFACE)
    if (setChip)
    {
        writePlatformReg(pDev, MAC_SCR, MAC_SCR_SLE_FSLEEP);
    }
#endif
}


/**************************************************************
 * ar5211QueryPSPollSupport
 *
 * Query hardware support for PS-Polls.
 * Returns TRUE if the hardware automatically generates PS-Poll frames
 */
A_BOOL
ar5211QueryPSPollSupport(WLAN_DEV_INFO *pDev)
{
    /* ar5211 does not automatically generate PS-Polls */
    return FALSE;
}


/**************************************************************
 * ar5211InitPSPoll
 *
 * Initialize for PS-Polls
 */
A_STATUS
ar5211InitPSPoll(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev->pPSPollDesc);

    // Set PS-Poll QCU TXDP
    writePlatformReg(pDev, MAC_Q0_TXDP + (pDev->psPollQueueNum * sizeof(A_UINT32)),
        pDev->pPSPollDesc->thisPhysPtr);

    // Configure psPoll QCU to be TIM-gated.
    writePlatformReg(pDev, MAC_Q0_MISC + (pDev->psPollQueueNum * sizeof(A_UINT32)),
         readPlatformReg(pDev, MAC_Q0_MISC + (pDev->psPollQueueNum * sizeof(A_UINT32))) |
         MAC_Q_MISC_FSP_TIM_GATED |
         MAC_Q_MISC_CBR_INCR_DIS1);
    return A_OK;
}

/**************************************************************
 * ar5211EnablePSPoll
 *
 * Enable PS-Polls on the STA
 */
A_STATUS
ar5211EnablePSPoll(WLAN_DEV_INFO *pDev, WLAN_MACADDR *pBssid, A_UINT16 assocId)
{
    A_STATUS status=A_OK;
    WLAN_FRAME_PSPOLL *pPSPollBuffer;   // pointer to the PS-Poll frame buffer

    ASSERT(pDev->pPSPollDesc);
    ASSERT(pBssid);

    pPSPollBuffer = pDev->pPSPollDesc->pBufferVirtPtr.psPoll;

    // Disable PS-Poll Tx Queue before updating the PS-Poll descriptor frame buffer
    status = ar5211DisablePSPoll(pDev);

    if (status == A_OK) {

        // Update PS-Poll AID
        pPSPollBuffer->durationId.value = assocId;

        // Update PS-Poll BSSID
        A_MACADDR_COPY(pBssid, &pPSPollBuffer->bssId);

        // Update active Q mask
        pDev->activeQueueMask |= 1 << pDev->psPollQueueNum;

        // Enable PS-Poll Tx Queue.
        writePlatformReg(pDev, MAC_Q_TXE, 1 << pDev->psPollQueueNum);
    }

    return status;
}

/**************************************************************
 * ar5211DisablePSPoll
 *
 * Disable PS-Polls on the STA
 */
A_STATUS
ar5211DisablePSPoll(WLAN_DEV_INFO *pDev)
{
    A_STATUS status = A_OK;
    A_UINT32 regVal;
    A_UINT32     i;

    // Return OK if Q is inactive
    if (!(pDev->activeQueueMask & (1 << pDev->psPollQueueNum))) {
        return A_OK;
    }

    // Disable PS-Poll Tx queue
    writePlatformReg(pDev, MAC_Q_TXD, (1 << pDev->psPollQueueNum));

    for( i = 0; i < 10000; i++) {
        regVal = readPlatformReg(pDev, MAC_Q_TXE) & (1 << pDev->psPollQueueNum);
        if (regVal==0) {
            break;
        }
        udelay (1000); // used to be 10
    }

    // Wait for all pending frames to be completed
    for (i = 0; i < 10000; i++) {
        regVal = readPlatformReg(pDev, MAC_Q0_STS + (pDev->psPollQueueNum * sizeof(A_UINT32)));
        if ((regVal & MAC_Q_STS_PEND_FR_CNT_M) == 0) {
            break;
        }
        udelay (1000); // used to be 10
    }

    // Reset Transmit queue disable, otherwise,
    // we won't be able to restart queue
    writePlatformReg(pDev, MAC_Q_TXD, 0);

    // Final check to make sure queue has stopped.
    if (regVal & MAC_Q_STS_PEND_FR_CNT_M) {
        // Still have pending tx frames
        uiPrintf("ar5211DisablePSPoll: Hardware error: Pending frame count non-zero\n");
        status = A_HARDWARE;
        ASSERT(0);
    }

    // Update active Q mask
    pDev->activeQueueMask &= ~(1 << pDev->psPollQueueNum);

    return status;
}

#endif // #ifdef BUILD_AR5211

