/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 *  Transmit HAL functions
 */

#ifdef BUILD_AR5211

#ident "$Id: //depot/sw/releases/2.0.1/src/hal/ar5211/ar5211Transmit.c#1 $"

/* Standard HAL Headers */
#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"
#include "display.h"

/* Headers for HW private items */
#include "ar5hwcEnd.h"

#include "ar5211Reg.h"
#include "ar5211Transmit.h"
#include "ar5211.h"

/**************************************************************
 * ar5211GetTxDP
 *
 * Get the TXDP for the "main" data queue.  Needs to be extended
 * for multiple Q functionality
 */
A_UINT32
ar5211GetTxDP(WLAN_DEV_INFO *pDev)
{
    A_UINT32 txdpReg;

    txdpReg = MAC_Q0_TXDP + (pDev->dataQueueNum * sizeof(A_UINT32));
    return readPlatformReg(pDev, txdpReg);
}

/**************************************************************
 * ar5211SetTxDP
 *
 * Set the TxDP for the "main" data queue.  Needs to be extended
 * for multiple data Q functionality
 */
void
ar5211SetTxDP(WLAN_DEV_INFO *pDev, A_UINT32 txdp)
{
    A_UINT32 txdpReg;

    /*
     * Make sure that TXE is deasserted before setting the TXDP.  If TXE
     * is still asserted, setting TXDP will have no effect.
     */
    ASSERT(!(readPlatformReg(pDev, MAC_Q_TXE) &
             (1 << pDev->dataQueueNum)));

    txdpReg = MAC_Q0_TXDP + (pDev->dataQueueNum * sizeof(A_UINT32));
    writePlatformReg(pDev, txdpReg, txdp);
}

/**************************************************************
 * ar5211UpdateTxTrigLevel
 *
 * Update Tx FIFO trigger level.
 *
 * Set bIncTrigLevel to TRUE to increase the trigger level.
 * Set bIncTrigLevel to FALSE to decrease the trigger level.
 *
 * Returns TRUE if the trigger level was updated
 */
A_BOOL
ar5211UpdateTxTrigLevel(WLAN_DEV_INFO *pDev, A_BOOL bIncTrigLevel)
{
    A_UINT32 regValue, curTrigLevel;

    ASSERT(pDev);

    /* Disable chip interrupts. This is because halUpdateTrigLevel
     * is called from both ISR and non-ISR contexts.
     */
    halDisableInterrupts(pDev, HAL_INT_GLOBAL);

    regValue     = readPlatformReg(pDev, MAC_TXCFG);
    curTrigLevel = (regValue & MAC_FTRIG_M) >> MAC_FTRIG_S;

    if (bIncTrigLevel) {
        /* increase the trigger level */
        curTrigLevel = curTrigLevel + ((MAX_TX_FIFO_THRESHOLD - curTrigLevel) / 2);
    } else {

        /* decrease the trigger level if not already at the minimum */
        if (curTrigLevel > MIN_TX_FIFO_THRESHOLD) {
            /* decrease the trigger level */
            curTrigLevel--;
        } else {
            /* no update to the trigger level */

            /* re-enable chip interrupts */
            halEnableInterrupts(pDev, HAL_INT_GLOBAL);

            return FALSE;
        }
    }

    /* Update the trigger level */
    writePlatformReg(pDev, MAC_TXCFG, (regValue & (~MAC_FTRIG_M)) |
                     ((curTrigLevel << MAC_FTRIG_S) & MAC_FTRIG_M));

    /* re-enable chip interrupts */
    halEnableInterrupts(pDev, HAL_INT_GLOBAL);

    return TRUE;
}

/**************************************************************
 * ar5211StartDmaTransmit
 *
 * Set Transmit Enable bits for the specified queues.
 */
void
ar5211StartDmaTransmit(WLAN_DEV_INFO *pDev, A_UINT16 queueTypeMask)
{
    A_UINT16 queueMask=0;

    if (queueTypeMask & HAL_TX_QUEUE_DATA) {
        queueMask |= (1 << pDev->dataQueueNum);
    }

    if (queueTypeMask & HAL_TX_QUEUE_CAB) {
        queueMask |= (1 << pDev->cabQueueNum);
    }

    if (queueTypeMask & HAL_TX_QUEUE_BEACON) {
        queueMask |= (1 << pDev->beaconQueueNum);
    }

    if (queueTypeMask & HAL_TX_QUEUE_PS_POLL) {
        queueMask |= (1 << pDev->psPollQueueNum);
    }

    // Only allow active Q to be enabled
    queueMask &= pDev->activeQueueMask;

    if (queueMask) {
        // Check to be sure we're not enabling a queue that has its TXD bit set.
        ASSERT( !(readPlatformReg(pDev, MAC_Q_TXD) & queueMask) );

        writePlatformReg(pDev, MAC_Q_TXE, queueMask);
    }
}

/**************************************************************
 * ar5211IsTxQueueStopped
 *
 * Returns:
 *   TRUE   if all queues in the type mask are stopped
 *   FALSE  otherwise
 */
A_BOOL
ar5211IsTxQueueStopped(WLAN_DEV_INFO *pDev, A_UINT16 queueTypeMask)
{
    A_UINT16    queueMask = 0;
    A_UINT32    pfcReg    = MAC_Q0_STS;

    if (queueTypeMask & HAL_TX_QUEUE_DATA) {
        queueMask |= (1 << pDev->dataQueueNum);
    }
    if (queueTypeMask & HAL_TX_QUEUE_CAB) {
        queueMask |= (1 << pDev->cabQueueNum);
    }
    if (queueTypeMask & HAL_TX_QUEUE_BEACON) {
        queueMask |= (1 << pDev->beaconQueueNum);
    }
    if (queueTypeMask & HAL_TX_QUEUE_PS_POLL) {
        queueMask |= (1 << pDev->psPollQueueNum);
    }

    /* Only check active queues */
    queueMask &= pDev->activeQueueMask;

    /* See if any queues are still enabled */
    if (readPlatformReg(pDev, MAC_Q_TXE) & queueMask) {
        return FALSE;
    }

    /*
     * Check the number of pending frames in the queues only after
     * making sure that TXE is deasserted as the pending frame count (PFC)
     * can momentarily go to zero while TXE remains asserted.  In other
     * words a PFC of zero is not sufficient to say that the queue has
     * stopped.
     */
    while (queueMask) {
        if (queueMask & 1) {
            if (readPlatformReg(pDev, pfcReg) & MAC_Q_STS_PEND_FR_CNT_M) {
                return FALSE;
            }
        }

        pfcReg += sizeof(A_UINT32);
        queueMask >>= 1;
    }

    return TRUE;
}

/**************************************************************
 * ar5211ClearTransmitFilter
 *
 * Clear transmit filter
 *
 */
void
ar5211ClearTransmitFilter(WLAN_DEV_INFO *pDev)
{
    // Clear DCU 0 transmit filter bits (all 128)
    writePlatformReg(pDev, TXBLK_FROM_MMR(48), MAC_DCU_0);
}

/**************************************************************
 * ar5211GetTransmitFilterIndex
 *
 * Get transmit filter by index
 *
 * Returns TRUE if transmit filter bit is set
 */
A_BOOL
ar5211GetTransmitFilterIndex(WLAN_DEV_INFO *pDev, A_UINT32 index)
{
    A_UINT32 reg, mask, value;

    ASSERT(index < 128);

    // DCU 0 only.
    reg  = CALC_TXBLK_ADDR(0, index);
    mask = CALC_TXBLK_VALUE(index);

    value = readPlatformReg(pDev, reg);

    return ((value & mask) != 0);
}

/**************************************************************
 * ar5211GetShortRetryLimit
 *
 * Get short frame retry limit for the main data queue.
 * Needs to be extended for multiple Q functionality
 *
 */
A_UINT32
ar5211GetShortRetryLimit(WLAN_DEV_INFO *pDev)
{
    A_UINT32 value;

    value = readPlatformReg(pDev, MAC_D0_RETRY_LIMIT);
    return (value & MAC_D_RETRY_LIMIT_FR_SH_M);
}

/**************************************************************
 * ar5211GetLongRetryLimit
 *
 * Get long frame retry limit for the main data queue.
 * Needs to be extended for multiple Q functionality
 *
 */
A_UINT32
ar5211GetLongRetryLimit(WLAN_DEV_INFO *pDev)
{
    A_UINT32 value;

    value = readPlatformReg(pDev, MAC_D0_RETRY_LIMIT);
    return (value & MAC_D_RETRY_LIMIT_FR_LG_M) >> MAC_D_RETRY_LIMIT_FR_LG_S;
}


/* Descriptor Acess Functions */

void
ar5211SetTxDescHdrLen(ATHEROS_DESC *pDesc, A_UINT16 value)
{
    // Does nothing on Maui
}

void
ar5211SetTxDescLongPkt(ATHEROS_DESC *pDesc, A_UINT16 value)
{
    // Does nothing on Maui
}

void
ar5211SetTxDescAntModeXmit(ATHEROS_DESC *pDesc, A_UINT16 value)
{
    AR5211_TX_CONTROL *pTxControl = (AR5211_TX_CONTROL *)&pDesc->hwControl.tx;

    /* Use auto antenna select mode */
    pTxControl->antModeXmit = 0;
}

void
ar5211SetTxDescPktType(ATHEROS_DESC *pDesc, HAL_DESC_PKT_TYPE value)
{
    AR5211_TX_CONTROL *pTxControl = (AR5211_TX_CONTROL *)&pDesc->hwControl.tx;

    pTxControl->PktType = value;
}

A_UINT16
ar5211GetTxDescEncryptKeyIndex(ATHEROS_DESC *pDesc)
{
    AR5211_TX_CONTROL *pTxControl = (AR5211_TX_CONTROL *)&pDesc->hwControl.tx;

    return (A_UINT16)pTxControl->encryptKeyIndex;
}

void
ar5211SetTxDescEncryptKeyIndex(ATHEROS_DESC *pDesc, A_UINT16 value)
{
    AR5211_TX_CONTROL *pTxControl = (AR5211_TX_CONTROL *)&pDesc->hwControl.tx;

    pTxControl->encryptKeyIndex = value;
    pTxControl->encryptKeyValid = 1;
}

void
ar5211SetTxDescRtsDuration(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 rateIndex, A_UINT16 controlRate)
{
    // Does nothing on Maui
}

void
ar5211SetTxDescNoAck(ATHEROS_DESC *pDesc, A_UINT16 value)
{
    AR5211_TX_CONTROL *pTxControl = (AR5211_TX_CONTROL *)&pDesc->hwControl.tx;

    pTxControl->noAck = value;
}

void
ar5211SetTxDescVeol(ATHEROS_DESC *pDesc, A_UINT16 value)
{
    AR5211_TX_CONTROL *pTxControl = (AR5211_TX_CONTROL *)&pDesc->hwControl.tx;

    pTxControl->Veol = value;
}

void
ar5211DebugPrintTxDesc(ATHEROS_DESC *pDesc)
{
    AR5211_TX_CONTROL *pTxControl = (AR5211_TX_CONTROL *)&pDesc->hwControl.tx;

    uiPrintf(FRAME_LEN      "%03x       ",   pTxControl->frameLength);
    uiPrintf(XMIT_RATE      "%01x\n",        pTxControl->transmitRate);
    uiPrintf(RTS_CTS_EN     "%01x         ", pTxControl->RTSCTSEnable);
    uiPrintf(CLR_DEST_MSK   "%01x\n",        pTxControl->clearDestMask);
    uiPrintf(ANT_MODE_TX    "%01x         ", pTxControl->antModeXmit);
    uiPrintf(ATIM           "%01x         ", pTxControl->PktType);
    uiPrintf(INT_REQ        "%01x\n",        pTxControl->interruptReq);
    uiPrintf(ENCRYPT_KEY_VLD"%01x         ", pTxControl->encryptKeyValid);
    uiPrintf(BUFFER_LEN     "%03x       ",   pTxControl->bufferLength);
    uiPrintf(MORE           "%01x\n",        pTxControl->more);
    uiPrintf(ENCRYPT_KEY_IDX"%02x\n",        pTxControl->encryptKeyIndex);
}

#endif // #ifdef BUILD_AR5211
