/* ar5311Sio.c - Atheros AR5311 (ns16550 core) DUART driver */

/* Copyright 1984-1997 Wind River Systems, Inc. */

#include "copyright_wrs.h"

/*
 *  Copyright  2001-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/ar5311Sio.c#1 $"

/*
 * DESCRIPTION
 * This is the driver for the AR5311 integrated DUART.   It is mostly
 * compatible with the WRS ns16550 driver, but requires word read/writes
 * to the register space to move each byte.
 *
 * The baud clock can be derived from either an external crystal or
 * the internal system clock.
 *
 * A AR5311SIO_CHAN structure is used to describe the serial channel. This
 * data structure is defined in ar5311Sio.h.
 *
 * Only asynchronous serial operation is supported by this driver.
 * The default serial settings are 8 data bits, 1 stop bit, no parity, 9600
 * baud, and software flow control.
 *
 * This driver handles setting of hardware options such as parity(odd, even) and
 * number of data bits(5, 6, 7, 8). Hardware flow control is provided with the
 * handshakes RTS/CTS. The function HUPCL(hang up on last close) is available.
 * When hardware flow control is enabled, the signals RTS and DTR are set TRUE
 * and remain set until a HUPCL is performed.
 */

#include "vxWorks.h"
#include "intLib.h"
#include "errnoLib.h"
#include "errno.h"
#include "sioLib.h"
#include "ar5311Sio.h"

#ifndef SIO_HUP
#   define SIO_OPEN     0x100A
#   define SIO_HUP      0x100B
#endif

/* min/max baud rate */
#define AR5311SIO_MIN_RATE 300          /* limited by 80Mhz internal clock */
#define AR5311SIO_MAX_RATE 115200

/* Register access macros.  Uses pointer math for 4 byte register stride. */
#define REGPTR(pChan, reg)      (pChan->regs + (reg))
#define REG(pChan, reg)         (*REGPTR(pChan, reg))

/* DMA engine register pair access */
#define dmaRegRead(reg, n)              sysRegRead(((reg) + (n*4)))
#define dmaRegWrite(reg, n, val)        sysRegWrite(((reg) + (n*4)), val)

LOCAL STATUS    ar5311SioDummyCallback(void);
LOCAL void      ar5311SioInitChannel(AR5311SIO_CHAN *);
LOCAL STATUS    ar5311SioBaudSet(AR5311SIO_CHAN *, UINT);
LOCAL STATUS    ar5311SioModeSet(AR5311SIO_CHAN *, UINT);
LOCAL STATUS    ar5311SioOptsSet(AR5311SIO_CHAN *, UINT);
LOCAL STATUS    ar5311SioOpen(AR5311SIO_CHAN * pChan);
LOCAL STATUS    ar5311SioHup(AR5311SIO_CHAN * pChan);

LOCAL int ar5311SioIoctl(AR5311SIO_CHAN *, int, void *);
LOCAL int ar5311SioTxStartup(AR5311SIO_CHAN *);
LOCAL int ar5311SioDmaTxStartup(AR5311SIO_CHAN *);
LOCAL int ar5311SioCallbackInstall(SIO_CHAN *, int, STATUS (*)(), void *);
LOCAL int ar5311SioPollInput(AR5311SIO_CHAN *, char *);
LOCAL int ar5311SioPollOutput(AR5311SIO_CHAN *, char);

static SIO_DRV_FUNCS ar5311SioDrvFuncs = {
    (int (*)(SIO_CHAN *, int, void *))ar5311SioIoctl,
    (int (*)(SIO_CHAN *))ar5311SioTxStartup,
    ar5311SioCallbackInstall,
    (int (*)(SIO_CHAN *, char *))ar5311SioPollInput,
    (int (*)(SIO_CHAN *, char))ar5311SioPollOutput
};

UINT32 ar5311SioStrayInt;               /* keep track of stray interrupts */

/*
 * ar5311SioDummyCallback - dummy callback routine.
 */
LOCAL STATUS
ar5311SioDummyCallback(void)
{
    return ERROR;
}

/*
 * ar5311SioDevInit - intialize an ar5311Sio channel
 *
 * This routine initializes some SIO_CHAN function pointers and then resets
 * the chip in a quiescent state.  Before this routine is called, the BSP
 * must already have initialized all the device addresses, etc. in the
 * AR5311SIO_CHAN structure.
 *
 * RETURNS: N/A
 */
void
ar5311SioDevInit(AR5311SIO_CHAN *pChan)
{
    int oldlevel = intLock();

    /* initialize the driver function pointers in the SIO_CHAN's */
    pChan->sio.pDrvFuncs = &ar5311SioDrvFuncs;

    /* set the non BSP-specific constants */
    pChan->getTxChar = ar5311SioDummyCallback;
    pChan->putRcvChar = ar5311SioDummyCallback;
    pChan->channelMode = 0;    /* undefined */
    pChan->options = (CLOCAL | CREAD | CS8);
    pChan->mcr = MCR_OUT2;

    pChan->overrun = pChan->parity = pChan->framing = pChan->breaks = 0;

    ar5311SioInitChannel(pChan);        /* reset the chip */

    intUnlock(oldlevel);
}

/*
 * ar5311SioInitChannel - initialize UART
 *
 * Initialize the number of data bits, parity and set the selected
 * baud rate.
 * Set the modem control signals if the option is selected.
 *
 * RETURNS: N/A
 */
LOCAL void
ar5311SioInitChannel(AR5311SIO_CHAN *pChan)
{
    ar5311SioBaudSet(pChan, pChan->baudRate);   /* set the baud rate */
    ar5311SioOptsSet(pChan, pChan->options);    /* set the options */
    return;
}

/*
 * ar5311SioOptsSet - set the serial options
 *
 * Set the channel operating mode to that specified.  All sioLib options
 * are supported: CLOCAL, HUPCL, CREAD, CSIZE, PARENB, and PARODD.
 * When the HUPCL option is enabled, a connection is closed on the last
 * close() call and opened on each open() call.
 *
 * Note, this routine disables the transmitter.  The calling routine
 * may have to re-enable it.
 *
 * RETURNS:
 * Returns OK to indicate success, otherwise ERROR is returned
 */
LOCAL STATUS
ar5311SioOptsSet(AR5311SIO_CHAN *pChan, UINT options)
{
    int oldlevel;
    UINT8 fcr;

    pChan->lcr = 0;
    pChan->mcr &= (~(MCR_RTS | MCR_DTR)); /* clear RTS and DTR bits */

    if (pChan == NULL || options & 0xffffff00) {
        return ERROR;
    }
    if ((pChan->dmaTx == TRUE) && !(options & CLOCAL)) {
        return ERROR;
    }

    switch (options & CSIZE) {
    case CS5:
        pChan->lcr = CHAR_LEN_5;
        break;
    case CS6:
        pChan->lcr = CHAR_LEN_6;
        break;
    case CS7:
        pChan->lcr = CHAR_LEN_7;
        break;
    default:
    case CS8:
        pChan->lcr = CHAR_LEN_8;
        break;
    }

    if (options & STOPB) {
        pChan->lcr |= LCR_STB;
    } else {
        pChan->lcr |= ONE_STOP;
    }

    switch (options & (PARENB | PARODD)) {
    case PARENB|PARODD:
        pChan->lcr |= LCR_PEN;
        break;
    case PARENB:
        pChan->lcr |= (LCR_PEN | LCR_EPS);
        break;
    default:
    case 0:
        pChan->lcr |= PARITY_NONE;
        break;
    }

    REG(pChan, IER) = 0;

    if (!(options & CLOCAL)) {
        /* !clocal enables hardware flow control(DTR/DSR) */
        pChan->mcr |= MCR_DTR | MCR_RTS;
        pChan->ier &= ~IER_ETHREI;      /* disable tx fifo intr */
        pChan->ier |= IER_EMSI;         /* enable modem status intr */
    } else {
        pChan->ier &= ~IER_EMSI;        /* disable modem status intr */
    }

    oldlevel = intLock();

    REG(pChan, LCR) = pChan->lcr;
    REG(pChan, MCR) = pChan->mcr;

    /* now reset the channel mode registers */
    fcr = FCR_RXTRIG_1 | FCR_RXCLR | FCR_TXCLR;
#if FIFO_DEPTH > 1
    fcr |= FCR_EN;
#endif
    if ((pChan->dmaTx == TRUE) && (pChan->dmaRx == TRUE)) {
        fcr |= FCR_DMA;
    }
    if (pChan->dmaTx == TRUE) {
        pChan->ier &= ~IER_ETHREI;      /* disable tx fifo intr */
    }
    if (pChan->dmaRx == TRUE) {
        pChan->ier &= ~IER_ERDAI;       /* rx data avail */
    } else if (options & CREAD) {
        pChan->ier |= IER_ERDAI;        /* rx data avail */
    }

    if (pChan->channelMode == SIO_MODE_INT) {
        REG(pChan, IER) = pChan->ier;
    }

    intUnlock(oldlevel);

    pChan->options = options;

    return OK;
}

/*
 * ar5311SioHup - hang up the modem control lines
 *
 * Resets the RTS and DTR signals and clears both the receiver and
 * transmitter sections.
 *
 * RETURNS: OK
 */
LOCAL STATUS
ar5311SioHup(AR5311SIO_CHAN *pChan)
{
    int oldlevel = intLock();

    pChan->mcr &= ~(MCR_RTS | MCR_DTR);
    REG(pChan, MCR) = pChan->mcr;
#if FIFO_DEPTH > 1
    REG(pChan, FCR) = FCR_RXCLR | FCR_TXCLR | FCR_EN;
#else
    REG(pChan, FCR) = FCR_RXCLR | FCR_TXCLR;
#endif

    intUnlock(oldlevel);
    return OK;
}

/*
 * ar5311SioOpen - Set the modem control lines
 *
 * Set the modem control lines(RTS, DTR) TRUE if not already set.
 * It also clears the receiver, transmitter and enables the fifo.
 *
 * RETURNS: OK
 */
LOCAL STATUS
ar5311SioOpen(AR5311SIO_CHAN *pChan)
{
    int oldlevel;       /* current interrupt level mask */
    char mask;

    mask = REG(pChan, MCR) & (MCR_RTS | MCR_DTR);

    if (mask != (MCR_RTS | MCR_DTR)) {
        /* RTS and DTR not set yet */
        oldlevel = intLock();

        /* set RTS and DTR TRUE */
        pChan->mcr |= MCR_DTR | MCR_RTS;
        REG(pChan, MCR) = pChan->mcr;

        /* clear Tx and receive and enable FIFO */
#if FIFO_DEPTH > 1
        REG(pChan, FCR) = FCR_RXCLR | FCR_TXCLR | FCR_EN;
#else
        REG(pChan, FCR) = FCR_RXCLR | FCR_TXCLR;
#endif

        intUnlock (oldlevel);
    }

    return OK;
}

/*
 * ar5311SioBaudSet - change baud rate for channel
 *
 * This routine sets the baud rate for the UART. The interrupts are disabled
 * during chip access.
 *
 * RETURNS: OK
 */
LOCAL STATUS
ar5311SioBaudSet(AR5311SIO_CHAN *pChan, UINT baud)
{
    int oldlevel = intLock();

    /* Enable access to the divisor latches by setting DLAB in LCR. */
    REG(pChan, LCR) = LCR_DLAB | pChan->lcr;

    /* Set divisor latches. */
    REG(pChan, DLL) = pChan->xtal/(16 * baud);
    REG(pChan, DLM) = (pChan->xtal/(16 * baud)) >> 8;

    /* Restore line control register */
    REG(pChan, LCR) = pChan->lcr;

    pChan->baudRate = baud;

    /* DMA timeout is in ~4 character periods on fixed HW */
    if ((pChan->dmaRx == TRUE) &&
        ((sysRegRead(AR5311_REV) & REV_CHIP) >= REV_AR5311_SPIN)) {
        sysRegWrite(AR5311_TIMEOUT,
                    (sysBoardData.sysFreq / (pChan->baudRate * 10)) * 4);
    }

    intUnlock(oldlevel);

    return OK;
}

/*
 * ar5311SioModeSet - change channel mode setting
 *
 * This driver supports both polled and interrupt modes and is capable of
 * switching between modes dynamically.
 *
 * If interrupt mode is desired this routine enables the channels receiver and
 * transmitter interrupts. If the modem control option is TRUE, the Tx interrupt
 * is disabled if the CTS signal is FALSE. It is enabled otherwise.
 *
 * If polled mode is desired the device interrupts are disabled.
 *
 * RETURNS:
 * Returns a status of OK if the mode was set else ERROR.
 */
LOCAL STATUS
ar5311SioModeSet(AR5311SIO_CHAN *pChan, UINT newMode)
{
    int oldlevel;       /* current interrupt level mask */
    char mask;

    if ((newMode != SIO_MODE_POLL) && (newMode != SIO_MODE_INT)) {
        return ERROR;
    }

    oldlevel = intLock();

    if (newMode == SIO_MODE_INT) {
        /* Enable appropriate interrupts */
        if (pChan->options & CLOCAL) {
            if ((pChan->dmaTx == TRUE) && (pChan->dmaRx == TRUE)) {
                REG(pChan, IER) = pChan->ier;
            } else if (pChan->dmaTx == TRUE) {
                REG(pChan, IER) = pChan->ier | IER_ERDAI;
            } else if (pChan->dmaRx == TRUE) {
                REG(pChan, IER) = pChan->ier | IER_ETHREI;
            }else {
                REG(pChan, IER) = pChan->ier | IER_ERDAI | IER_ETHREI;
            }
        } else {
            mask = REG(pChan, MSR) & MSR_CTS;

            /* if the CTS is asserted enable Tx interrupt */
            if (mask & MSR_CTS) {
                pChan->ier |= IER_ETHREI;  /* enable Tx int */
            } else {
                pChan->ier &= ~IER_ETHREI; /* disable Tx int */
            }

            REG(pChan, IER) = pChan->ier;
        }
    } else {
        /* disable all ar5311Sio interrupts */
        REG(pChan, IER) = 0;
    }

    pChan->channelMode = newMode;

    intUnlock(oldlevel);

    return OK;
}

/*
 * ar5311SioIoctl - special device control
 *
 * Includes commands to get/set baud rate, mode(INT,POLL), hardware options(
 * parity, number of data bits), and modem control(RTS/CTS and DTR/DSR).
 * The ioctl command SIO_HUP is sent by ttyDrv when the last close() function
 * call is made. Likewise SIO_OPEN is sent when the first open() function call
 * is made.
 *
 * RETURNS: OK on success, EIO on device error, ENOSYS on unsupported
 *          request.
 */

LOCAL int
ar5311SioIoctl(AR5311SIO_CHAN *pChan, int request, void *arg)
{
    int iarg = (int)arg;
    STATUS status = OK;

    switch (request) {
    case SIO_BAUD_SET:
        if (iarg < AR5311SIO_MIN_RATE || iarg > AR5311SIO_MAX_RATE) {
            status = EIO;               /* baud rate out of range */
        } else {
            status = ar5311SioBaudSet(pChan, iarg);
        }
        break;

    case SIO_BAUD_GET:
        *(int *)arg = pChan->baudRate;
        break;

    case SIO_MODE_SET:
        status = (ar5311SioModeSet (pChan, iarg) == OK) ? OK : EIO;
        break;

    case SIO_MODE_GET:
        *(int *)arg = pChan->channelMode;
        break;

    case SIO_AVAIL_MODES_GET:
        *(int *)arg = SIO_MODE_INT | SIO_MODE_POLL;
        break;

    case SIO_HW_OPTS_SET:
        status = (ar5311SioOptsSet (pChan, iarg) == OK) ? OK : EIO;
        break;

    case SIO_HW_OPTS_GET:
        *(int *)arg = pChan->options;
        break;

    case SIO_HUP:
        /* check if hupcl option is enabled */
        if (pChan->options & HUPCL) {
            status = ar5311SioHup(pChan);
        }
        break;

    case SIO_OPEN:
        /* check if hupcl option is enabled */
        if (pChan->options & HUPCL) {
            status = ar5311SioOpen(pChan);
        }
        break;

    default:
        status = ENOSYS;
    }

    return status;
}

/*
 * ar5311SioInt - interrupt level processing
 *
 * This routine handles four sources of interrupts from the UART. They are
 * prioritized in the following order by the Interrupt Identification Register:
 * Receiver Line Status, Received Data Ready, Transmit Holding Register Empty
 * and Modem Status.
 *
 * When a modem status interrupt occurs, the transmit interrupt is enabled if
 * the CTS signal is TRUE.
 *
 * RETURNS: N/A
 */
void
ar5311SioInt(AR5311SIO_CHAN *pChan)
{
    volatile char intStatus;
    char msr, lsr, byte;
    int to;

    /* read the Interrrupt Status Register (Int. Ident.) */
    intStatus = REG(pChan, IIR) & 0x0f;

    /*
     * This UART chip always produces level active interrupts, and the IIR
     * only indicates the highest priority interrupt.
     * In the case that receive and transmit interrupts happened at
     * the same time, we must clear both interrupt pending to prevent
     * edge-triggered interrupt(output from interrupt controller) from
     * locking up. One way doing it is to disable all the interrupts at
     * the beginning of the ISR and enable at the end.
     */
    REG(pChan, IER) = 0;    /* disable interrupt */
    sysWbFlush();

    switch (intStatus) {
    case IIR_RLS:
        /* Overrun, parity error and break interrupt */
        intStatus = REG(pChan, LSR); /* read LSR to reset interrupt */
        if (intStatus & LSR_OE) {
            pChan->overrun++;
        }
        if (intStatus & LSR_PE) {
            pChan->parity++;
        }
        if (intStatus & LSR_FE) {
            pChan->framing++;
        }
        if (intStatus & LSR_BI) {
            pChan->breaks++;
        }
        break;

    case IIR_RDA:               /* received data available */
    case IIR_TIMEOUT:
        /*
         * Receiver FIFO interrupt. In some case, IIR_RDA will
         * not be indicated in IIR register when there is more
         * than one character in the rx FIFO.
         *
         * Read character from Receive Holding Reg and forward it.
         */
        if (pChan->dropOutput) pChan->dropOutput = 0;   /* wake-up */
        for (to=0; to < FIFO_DEPTH; to++) {
            lsr = REG(pChan, LSR);
            if ((lsr & LSR_DR) == 0) {
                break;
            }
            byte = REG(pChan, RBR);
            (*pChan->putRcvChar)(pChan->putRcvArg, byte);
        }

        /*
         * Check for pending output.  There are issues with losing
         * THRE interrupt if it happens close to a RDA interrupt.
         */
        if ((lsr & LSR_THRE) == 0) {
            break;
        }
        /* else FALLSTHRU */

    case IIR_THRE:  /* transmitter holding register ready */
        /* Fill the tx fifo as much as we can. */
        if ((*pChan->getTxChar)(pChan->getTxArg, &byte) != ERROR) {
            to = FIFO_DEPTH - 1;
            REG(pChan, THR) = byte;   /* output character */
            while ((REG(pChan, LSR) & LSR_THRE) &&
                   ((*pChan->getTxChar)(pChan->getTxArg, &byte) != ERROR)) {
                REG(pChan, THR) = byte;   /* output character */
                if (--to <= 0) {
                    break;
                }
            }
        } else {
            /* Disable tx interrupt as no more pending output */
            pChan->ier &= ~IER_ETHREI;
        }
        break;

    case IIR_MSTAT: /* modem status register */
        msr = REG(pChan, MSR);

        /* if CTS is asserted by modem, enable tx interrupt */
        if ((msr & MSR_CTS) && (msr & MSR_DCTS)) {
            pChan->ier |= IER_ETHREI;
        } else {
            pChan->ier &= ~IER_ETHREI;
        }
        break;

    default:
        ar5311SioStrayInt++;
        break;
    }

    REG(pChan, IER) = pChan->ier;       /* enable interrupts accordingly */
    sysWbFlush();

    return;
}

/*
 * ar5311SioTxStartup - transmitter startup routine
 *
 * Call interrupt level character output routine and enable interrupt if it is
 * in interrupt mode with no hardware flow control.
 * If the option for hardware flow control is enabled and CTS is set TRUE,
 * then the Tx interrupt is enabled.
 */
LOCAL int
ar5311SioTxStartup(AR5311SIO_CHAN *pChan)
{
    char mask, byte;

    /* Eat any output if in "quiet" mode. */
    if (pChan->dropOutput) {
        while ((*pChan->getTxChar)(pChan->getTxArg, &byte) != ERROR)
            ;
    }

    if (pChan->channelMode == SIO_MODE_INT) {
        if (pChan->dmaTx == TRUE) {
            return ar5311SioDmaTxStartup(pChan);
        }
        if (pChan->options & CLOCAL) {
            /* No modem control */
            pChan->ier |= IER_ETHREI;
            REG(pChan, IER) = pChan->ier;
        } else {
            mask = REG(pChan, MSR);

            /* if the CTS is asserted enable Tx interrupt */
            if (mask & MSR_CTS) {
                pChan->ier |= IER_ETHREI;
            } else {
                pChan->ier &= ~IER_ETHREI;
            }

            REG(pChan, IER) = pChan->ier;
        }

        /*
         * Prime the tx pump.  Should not have to do this, but
         * many UARTs seem to require it.
         */
        if (pChan->ier & IER_ETHREI) {
            if ((*pChan->getTxChar)(pChan->getTxArg, &byte) != ERROR) {
                REG(pChan, THR) = byte;
            }
        }
    }

    return OK;
}

/*
 * ar5311SioPollOutput - output a character in polled mode.
 *
 * RETURNS: OK if a character arrived, EIO on device error, EAGAIN
 * if the output buffer if full.
 */
LOCAL int
ar5311SioPollOutput(AR5311SIO_CHAN *pChan, char outChar)
{
    char pollStatus;
    char msr;

    pollStatus = REG(pChan, LSR);
    msr = REG(pChan, MSR);

    /* Is the transmitter ready to accept a character? */
    if ((pollStatus & LSR_THRE) == 0x00) {
        return EAGAIN;
    }

    if (!(pChan->options & CLOCAL)) {    /* modem flow control ? */
        if (msr & MSR_CTS) {
            REG(pChan, THR) = outChar;
        } else {
            return EAGAIN;
        }
    } else {
        REG(pChan, THR) = outChar;       /* transmit character */
    }

    return OK;
}

/*
 * ar5311SioPollInput - poll the device for input.
 *
 * RETURNS: OK if a character arrived, EIO on device error, EAGAIN
 * if the input buffer if empty.
 */
LOCAL int
ar5311SioPollInput(AR5311SIO_CHAN *pChan, char *pChar)
{
    char pollStatus;

    pollStatus = REG(pChan, LSR);

    if ((pollStatus & LSR_DR) == 0x00) {
        return EAGAIN;
    }

    /* got a character */
    *pChar = REG(pChan, RBR);

    return OK;
}

/*
 * ar5311SioCallbackInstall - install ISR callbacks to get/put chars.
 *
 * This routine installs the callback functions for the driver
 *
 * RETURNS: OK on success or ENOSYS on unsupported callback type.
 */
LOCAL int
ar5311SioCallbackInstall(SIO_CHAN *pSioChan, int callbackType,
                         STATUS (*callback)(), void *callbackArg)
{
    AR5311SIO_CHAN *pChan = (AR5311SIO_CHAN *)pSioChan;

    switch (callbackType) {
    case SIO_CALLBACK_GET_TX_CHAR:
        pChan->getTxChar = callback;
        pChan->getTxArg = callbackArg;
        return OK;

    case SIO_CALLBACK_PUT_RCV_CHAR:
        pChan->putRcvChar = callback;
        pChan->putRcvArg  = callbackArg;
        return OK;

    default:
        return ENOSYS;
    }
}

/*
 * ar5311SioDmaInit - initialize DMA engine.
 *
 * RESET_DMA should happen in sysHwInit.
 */
void
ar5311SioDmaInit(AR5311SIO_CHAN *pChan)
{
    (void)sysRegRead(AR5311_DMAISR);    /* Clear + disable interrupts */
    sysRegWrite(AR5311_DMAIMR, 0);

    pChan->dmaImr = 0;
    pChan->dmaTxBusy[0] = pChan->dmaTxBusy[1] = FALSE;

    /* Allocate DMA buffers in cached space */
    if (pChan->dmaTx) {
        pChan->dmaTxBuf[0] = memalign(UARTDMA_ALIGN, UARTDMA_TXSIZE);
        pChan->dmaTxBuf[1] = memalign(UARTDMA_ALIGN, UARTDMA_TXSIZE);
        if ((pChan->dmaTxBuf[0] == NULL) ||
            (pChan->dmaTxBuf[1] == NULL)) {
            pChan->dmaTx = FALSE;           /* disable DMA */
            if (pChan->dmaTxBuf[0] == NULL) free(pChan->dmaTxBuf[0]);
            if (pChan->dmaTxBuf[1] == NULL) free(pChan->dmaTxBuf[1]);
            return;
        }
        /* Initialize DMA engine registers */
        sysRegWrite(AR5311_TXADDR0, (UINT32)pChan->dmaTxBuf[0]);
        sysRegWrite(AR5311_TXADDR1, (UINT32)pChan->dmaTxBuf[1]);

        sysRegWrite(AR5311_TXSTAT0, 0);
        sysRegWrite(AR5311_TXSTAT1, 0);
        sysRegWrite(AR5311_TXCTRL0, 0);
        sysRegWrite(AR5311_TXCTRL1, 0);
        if (sysRegRead(AR5311_TXCTRL0) & TXCTRL_READY) {
            pChan->dmaNextTx = 0;
        } else {
            pChan->dmaNextTx = 1;
        }

        /* Enable TX interrupts */
        pChan->dmaImr |= DMAIMR_RXOK | DMAIMR_RXERR | DMAIMR_RXTIM;
        pChan->dmaImr |= DMAIMR_TXOK | DMAIMR_TXERR;
        sysRegWrite(AR5311_DMAIMR, pChan->dmaImr);
    }

    if (pChan->dmaRx) {
        pChan->dmaRxBuf[0] = memalign(UARTDMA_ALIGN, UARTDMA_RXSIZE);
        pChan->dmaRxBuf[1] = memalign(UARTDMA_ALIGN, UARTDMA_RXSIZE);
        if ((pChan->dmaRxBuf[0] == NULL) ||
            (pChan->dmaRxBuf[1] == NULL)) {
            pChan->dmaRx = FALSE;           /* disable rx DMA */
            if (pChan->dmaRxBuf[0] == NULL) free(pChan->dmaRxBuf[0]);
            if (pChan->dmaRxBuf[1] == NULL) free(pChan->dmaRxBuf[1]);
            return;
        }
        if ((sysRegRead(AR5311_REV) & REV_CHIP) >= REV_AR5311_SPIN) {
            /* DMA timeout is in ~4 character periods */
            sysRegWrite(AR5311_TIMEOUT,
                        (sysBoardData.sysFreq / (pChan->baudRate*10)) * 4);
        } else {
            /* DMA timeout is 1/2 second for spirit 1.0 */
            sysRegWrite(AR5311_TIMEOUT, sysBoardData.sysFreq >> 1);
        }

        sysRegWrite(AR5311_RXADDR0, (UINT32)pChan->dmaRxBuf[0]);
        sysRegWrite(AR5311_RXADDR1, (UINT32)pChan->dmaRxBuf[1]);

        sysRegWrite(AR5311_RXSTAT0, 0);
        sysRegWrite(AR5311_RXSTAT1, 0);
        sysRegWrite(AR5311_RXCTRL0, RXCTRL_OWN | UARTDMA_RXSIZE);
        sysRegWrite(AR5311_RXCTRL1, RXCTRL_OWN | UARTDMA_RXSIZE);

	/* use buffer zero on start */
	pChan->dmaNextRx = 0;

        /* Enable RX interrupts */
        pChan->dmaImr |= DMAIMR_RXOK | DMAIMR_RXERR | DMAIMR_RXTIM;
        sysRegWrite(AR5311_DMAIMR, pChan->dmaImr);
    }

    return;
}

/*
 * ar5311SioDmaTxStartup - transmitter startup routine for DMA output.
 */
LOCAL int
ar5311SioDmaTxStartup(AR5311SIO_CHAN *pChan)
{
    int s, n, i;
    UINT8 *p;

    s = intLock();

    /* Interrupt handler will wake us */
    i = pChan->dmaNextTx;
    if (pChan->dmaTxBusy[i] == TRUE) {
        intUnlock(s);
        return OK;
    }

    /* Fill this buffer - one is enough when priming the pump */
    p = pChan->dmaTxBuf[i];
    n = 0;
    while ((n < UARTDMA_TXSIZE) &&
           ((*pChan->getTxChar)(pChan->getTxArg, p) != ERROR)) {
        p++;
        n++;
    }

    if (n == 0) {
        intUnlock(s);
        return EIO;
    }

    /* Let HW have after the write buffer is flushed. */
    sysWbFlush();
    pChan->dmaTxBusy[i] = TRUE;
    dmaRegWrite(AR5311_TXSTAT0, i, 0);
    dmaRegWrite(AR5311_TXCTRL0, i, TXCTRL_OWN | n);
    pChan->dmaNextTx ^= 1;

    intUnlock(s);

    return OK;
}


/*
 * ar5311SioDmaReset(AR5311SIO_CHAN *pChan)
 */
static void
ar5311SioDmaReset(AR5311SIO_CHAN *pChan)
{
    UINT32 reset = sysRegRead(AR5311_RESET);

    /* reset DMA and UART HW */
    sysRegWrite(AR5311_RESET, reset | RESET_DMA | RESET_UART0);
    sysWbFlush();
    sysUDelay(10);
    sysRegWrite(AR5311_RESET, reset);

    /* try to reset the software state as it knows about the HW */
    sysRegWrite(AR5311_RXSTAT0, 0);
    sysRegWrite(AR5311_RXSTAT1, 0);
    sysRegWrite(AR5311_RXCTRL0, RXCTRL_OWN | UARTDMA_RXSIZE);
    sysRegWrite(AR5311_RXCTRL1, RXCTRL_OWN | UARTDMA_RXSIZE);

    /* use buffer zero on start */
    pChan->dmaNextRx = 0;

    sysRegWrite(AR5311_TXSTAT0, 0);
    sysRegWrite(AR5311_TXSTAT1, 0);
    sysRegWrite(AR5311_TXCTRL0, 0);
    sysRegWrite(AR5311_TXCTRL1, 0);
    if (sysRegRead(AR5311_TXCTRL0) & TXCTRL_READY) {
        pChan->dmaNextTx = 0;
    } else {
        pChan->dmaNextTx = 1;
    }

    return;
}

/*
 * ar5311SioDmaInt - interrupt level processing for DMA engine.
 *
 * This probably blocks interrupts a long time.  For higher
 * speed interfaces it probably needs to move to a task context.
 */
void
ar5311SioDmaInt(AR5311SIO_CHAN *pChan)
{
    UINT32 isr = sysRegRead(AR5311_DMAISR) & pChan->dmaImr;
    UINT32 stat, ctrl;
    int n, i, x;
    INT8 *p;

    /* Service RX first to help prevent overruns */
    if (isr & (DMAISR_RXOK | DMAISR_RXERR | DMAISR_RXTIM)) {
        for (x=0; x < 2; x++) {
            i = pChan->dmaNextRx;
            ctrl = dmaRegRead(AR5311_RXCTRL0, i);
            if ((ctrl & RXCTRL_OWN) != 0) {
                break;
            }
            stat = dmaRegRead(AR5311_RXSTAT0, i);
            if (stat & RXSTAT_AHBERR) {
                ar5311SioDmaReset(pChan);
            }
            else if (stat & RXSTAT_DONE) {
                n = (stat & RXSTAT_COUNT) >> RXSTAT_COUNT_S;
                if (n > 0) {
                    p = pChan->dmaRxBuf[i];
                    cacheDataInvalidate(p, UARTDMA_RXSIZE);
                    while (n > 0) {
                        (*pChan->putRcvChar)(pChan->putRcvArg, *p);
                        p++;
                        n--;
                    }
                }
            }
            dmaRegWrite(AR5311_RXSTAT0, i, 0);
            dmaRegWrite(AR5311_RXCTRL0, i, RXCTRL_OWN | UARTDMA_RXSIZE);

            /* alternate between the two buffers */
            pChan->dmaNextRx = 1 - pChan->dmaNextRx;
        }
    }

    if (isr & (DMAISR_TXOK | DMAISR_TXERR)) {
        stat = sysRegRead(AR5311_TXSTAT0);
        if (stat & TXSTAT_AHBERR) {
            ar5311SioDmaReset(pChan);
        }
        if (stat & (TXSTAT_DONE | TXSTAT_OK)) {
            sysRegWrite(AR5311_RXSTAT0, 0);
            pChan->dmaTxBusy[0] = FALSE;
        }
        stat = sysRegRead(AR5311_TXSTAT1);
        if (stat & TXSTAT_AHBERR) {
            ar5311SioDmaReset(pChan);
        }
        if (stat & (TXSTAT_DONE | TXSTAT_OK)) {
            sysRegWrite(AR5311_RXSTAT1, 0);
            pChan->dmaTxBusy[1] = FALSE;
        }

        /* Fill more output, each buffer if we have enough */
        for (x = 0 ; x < 2; x++) {
            i = pChan->dmaNextTx;
            if (pChan->dmaTxBusy[i] == FALSE) {
                p = pChan->dmaTxBuf[i];
                n = 0;
                while ((n < UARTDMA_TXSIZE) &&
                       ((*pChan->getTxChar)(pChan->getTxArg, p) != ERROR)) {
                    p++;
                    n++;
                }
                if (n > 0) {
                    sysWbFlush();
                    pChan->dmaTxBusy[i] = TRUE;
                    dmaRegWrite(AR5311_TXCTRL0, i, TXCTRL_OWN | n);
                    pChan->dmaNextTx ^= 1;
                }
            }
        }
    }

    return;
}
