/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

/*
 *  ar5hwcEnd.c - Atheros VxWorks END-style network interface driver for
 *                802.11a Access Point devices
 */

/*

SUPPORTED DRIVER FEATURES:

The driver supports the following features:
Transmission rates of 6, 12, 18, 36, 48 and 54 Mbps
Distribution service (DS)
Beacon generation
Stable configuration storage
RTS/CTS
Fragmentation
WEP with 48, 104 or 128 bit shared keys
Station Mgmt Entity (SME)
Multicast traffic following DTIMs
Dynamic rate selection
Turbo mode
EEPROM support
SNMP
Rate backoff, with histogram
Software retries with rate backoff
Unique wep key

UNSUPPORTED FEATURES
IAPP
Frame sizes > 1500 bytes

DRIVER OPERATION

The entry point to the driver from the MUX layer is the ar5hwcEndLoad() routine
which initializes the device and the driver to an operational state.  The
ar5hwcEndLoad() routine gets parameters via the <initString>.

The driver makes use of two types of buffers for each device, DMA buffers and
non-DMA buffers. DMA buffers are used to allocate receive and transmit
descriptors, and buffers (clusters); while non-DMA buffers are used to
allocate other management data viz., m-blocks, cluster blocks, and device
objects.

The receive and transmit descriptors are arranged in two circular
singly-linked rings. The transmit and receive buffers (clusters),
cluster blocks, and m-blocks are managed by a vxWorks netpool. The descriptor
rings are shared between the driver and the device. The device accesses these
descriptors via DMA.

During the initialization of the transmit and receive rings, an
mBlk-clBlk-cluster construct is allocated and associated with each descriptor.
The driver single-threads transmit operations. All transmit buffers are copied
into the drivers transmit queue, while all receive buffers are loaned with no
copy to the upper protocol layers.

The driver makes use of appropriate macros to adjust for endian-ness of the
host CPU, PCI bus, Ethernet, and the Atheros device. It also performs appropriate
translations of addresses for device use. These builtin features make the
driver extremely portable to run on various CPUs with minimum changes.

DEVICE PARAMETERS

The <initString>, parameter to ar5hwcEndLoad() - see comment at top of ar5hwcParse

The driver allocates a DMA buffer of the required size if the buffer address
is specified either as NONE (-1) or NULL (0) in the <initString>. Otherwise
the address is assumed to point to a memory region which is cache-safe, and of
the required size. Refer to the manual pages of ar5hwcEndLoad() for more
details.

SEE ALSO: muxLib, endLib, netBufLib

REFERENCES
"Atheros AR5000 MAC Hardware/Software Interface Specification
*/

#ident "ACI $Id: //depot/sw/linux-2.4.17_mvl21/drivers/net/atheros/ar5hwcEnd.c#4 $"

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include "stdlib.h"
#include "stdio.h"
#include "config.h"

#include "cacheLib.h"
#include "intLib.h"
#include "end.h"                /* Common END structures. */
#include "endLib.h"
#include "lstLib.h"             /* Needed to maintain protocol list. */
#include "wdLib.h"
#include "iv.h"
#include "semLib.h"
#include "tickLib.h"
#include "usrLib.h"
#include "rebootLib.h"                                     
#include "ctype.h"

#undef  ETHER_MAP_IP_MULTICAST

#include "etherMultiLib.h"      /* multicast stuff. */
#include "assert.h"

#include "apdefs.h"
#include "apcfg.h"
#include "ar5hwcEnd.h"
#include "halApi.h"
#include "wlanchannel.h"
#include "wlandrv.h"
#include "vxdrv.h"
#include "wlanext.h"
#include "wlansmeext.h"
#include "radardefs.h"
#include "radarext.h"

#include "diags/display.h"

#if defined(PCI_INTERFACE)
#include "drv/pci/pciIntLib.h"
#endif

#include "wlanapDs.h"
#include "wlanCapDe.h"
#include "wlanbeacon.h"
#include "bufconv.h"
#include "apmem.h"
#include "ui.h"
#include "trace.h"
#include "histograms.h"

#ifdef INCLUDE_SNMPD
#include "snmpdefs.h"
#include "snmpIoLib.h"
#endif

/* need this for Crete; also need it in case the
 * rx overrun on small AES frames bug persists in
 * the hw; the workaround for the latter is to
 * reset the chip on rx overrun, and we'd rather
 * have buggy overruns only and not real overruns
 * - real overruns are informative only and thus
 * resetting the chip on real overruns would be
 * very expensive
 */
#define RXEOL_WAR   1

/* The followng macros are used for debugging the driver, all the debug messages
 * are sent out to the terminal through Log task.
 */
#ifdef  DEBUG
#define DRV_DEBUG_OFF           0x00000000
#define DRV_DEBUG_RX            0x00000001
#define DRV_DEBUG_TX            0x00000002
#define DRV_DEBUG_INT           0x00000004
#define DRV_DEBUG_POLL          (DRV_DEBUG_POLL_RX | DRV_DEBUG_POLL_TX)
#define DRV_DEBUG_POLL_RX       0x00000008
#define DRV_DEBUG_POLL_TX       0x00000010
#define DRV_DEBUG_LOAD          0x00000020
#define DRV_DEBUG_IOCTL         0x00000040
#define DRV_DEBUG_POLL_MODE     0x00000080
#define DRV_DEBUG_POLL_REDIR    0x00010000
#define DRV_DEBUG_LOG_NVRAM     0x00020000
#define DRV_DEBUG_LOG_ALL       0x00080000
#define DRV_DEBUG_BEACON        0x00001000
#define DRV_DEBUG_TEMP          0x00002000
#define DRV_DEBUG_ALT           0x00004000

int arDebug = 0;

#define DRV_LOG(FLG,args)               \
    if (arDebug & FLG)                  \
        logMsg args
#else /* DEBUG */

#define DRV_LOG(DBG_SW, args...)
const int arDebug = 0;

#endif /* DEBUG */

static const int RateIdxTbl[] =
/* 0    1    2    3    4    5    6    7    8    9    A    B    C    D    E    F */
{ 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x6, 0x4, 0x2, 0x0, 0x7, 0x5, 0x3, 0x1 };

/* Set number of rx descriptors to process before rescheduling */
#define AR_RX_RESCHED   AR_RX_QUEUE_DEPTH      /* define as 0 to remove this fcn */

/* Set the default value for reset period to be 15 seconds */
#define RT_DEFAULT_VAL  15

/* Set the offset calibration waiting time */
#define CAL_WAIT        4000
#define POLL_WAIT       4000

/* Local variables used in this file */

LOCAL MSG_Q_ID  calMsgQId = NULL;
LOCAL BOOL      wlanStop  = FALSE;

#ifdef DEBUG
int bcVerbose    = 0;
int rxVerbose    = 0;
int txVerbose    = 0;
int swRetryDebug = 0;
void       dumpFrame(ATHEROS_DESC *pDesc, char *pMsg);
static int swRetryQueueVerify(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib);
static void swrDump(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib);
#else
#define swRetryQueueVerify(d, s)
const int swRetryDebug = 0;
#endif /* DEBUG */


#ifdef INCLUDE_TRACE
static void eventDump_wlanrx(struct event_s *eventBuf);
static void eventDump_wlantx(struct event_s *eventBuf);
#endif

/* Beacon Watchdog stuff */
LOCAL WDOG_ID  beaconWatchDogId = NULL;
LOCAL A_UINT32 txBeaconNum      = 0;
LOCAL A_UINT16 beaconNum        = 0;
LOCAL A_UINT16 calBeaconCount   = 0;

/*
 * End specific functions: The following function are local, but can be
 * exposed to other modules through NET_FUNCS.
 */
LOCAL STATUS   ar5hwcEndStart    (END_OBJ *);
LOCAL STATUS   ar5hwcEndStop     (END_OBJ *);
LOCAL STATUS   ar5hwcEndUnload   (END_OBJ *);
LOCAL int      ar5hwcEndIoctl    (END_OBJ *, int cmd, caddr_t data);
LOCAL STATUS   ar5hwcEndMCastAdd (END_OBJ *, char* pAddress);
LOCAL STATUS   ar5hwcEndMCastDel (END_OBJ *, char* pAddress);
LOCAL STATUS   ar5hwcEndMCastGet (END_OBJ *, MULTI_TABLE* pTable);
LOCAL STATUS   ar5hwcEndPollSend (END_OBJ *, M_BLK_ID pBuf);
LOCAL STATUS   ar5hwcEndPollRecv (END_OBJ *, M_BLK_ID pBuf);
LOCAL STATUS   ar5hwcEndPollStart(END_OBJ *);
LOCAL STATUS   ar5hwcEndPollStop (END_OBJ *);

/* The following functions are local */
LOCAL void     ar5hwcEndInt(WLAN_DEV_INFO *pdevInfo);
LOCAL STATUS   ar5hwcParse(WLAN_DEV_INFO *pdevInfo, char *initString);
LOCAL STATUS   ar5hwcMemInit (WLAN_DEV_INFO *pdevInfo);
LOCAL STATUS   ar5hwcQueueCreate(WLAN_DEV_INFO *pdevInfo, QUEUE_DETAILS *pQueue, QUEUE_DETAILS *freeQueue,
                                 int count, int bufSize);
LOCAL void     ar5hwcQueueDelete(QUEUE_DETAILS *pQueue);
LOCAL STATUS   hashIndex(char *pMacAddress);
LOCAL void     ar5RxFilterConfig(WLAN_DEV_INFO *);
LOCAL void     queueTxDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, ATHEROS_DESC *pTail);
LOCAL void     processTxInterrupt(WLAN_DEV_INFO *);
LOCAL void     processRxInterrupt(WLAN_DEV_INFO *);
LOCAL void     processAltQueue(WLAN_DEV_INFO *);
LOCAL void     resetOnError(WLAN_DEV_INFO *);
LOCAL void     resetRxQueue(WLAN_DEV_INFO *);
LOCAL void     resetwlan(WLAN_DEV_INFO *pdevInfo);
LOCAL void     athCalibrationTask(void);
LOCAL void     beaconWatchDogISR(WLAN_DEV_INFO *pdevInfo);

LOCAL INLINE void txFramePendIncrement(WLAN_DEV_INFO *, SIB_ENTRY *, ATHEROS_DESC *);
LOCAL INLINE void txFramePendDecrement(WLAN_DEV_INFO *, SIB_ENTRY *, ATHEROS_DESC *);
LOCAL INLINE int  processTxSibInfo(WLAN_DEV_INFO *, ATHEROS_DESC *);

extern char *mgtTypeStr[];
extern char *ctrlTypeStr[];
extern char *dataTypeStr[];

extern struct ar5ConfigItems *par5Config;
#define TSF_ARRAY_SIZE  512
static WLAN_TIMESTAMP   tsfStamp[TSF_ARRAY_SIZE];
static A_UINT32         tsfIndex = 0;

/*
 * Declare our function table.  This is static across all driver
 * instances.
 */
LOCAL NET_FUNCS ar5hwcFuncTable = {
    (FUNCPTR)ar5hwcEndStart,        /* Function to start the device. */
    (FUNCPTR)ar5hwcEndStop,         /* Function to stop the device. */
    (FUNCPTR)ar5hwcEndUnload,       /* Unloading function for the driver.*/
    (FUNCPTR)ar5hwcEndIoctl,        /* Ioctl function for the driver. */
    (FUNCPTR)wlanDsFromMux,         /* Send function for the driver. */
    (FUNCPTR)ar5hwcEndMCastAdd,     /* Multicast add function.*/
    (FUNCPTR)ar5hwcEndMCastDel,     /* Multicast delete function.  */
    (FUNCPTR)ar5hwcEndMCastGet,     /* Multicast retrieve function . */
    (FUNCPTR)ar5hwcEndPollSend,     /* Polling send function */
    (FUNCPTR)ar5hwcEndPollRecv,     /* Polling receive function */
    endEtherAddressForm,            /* put address info into a NET_BUFFER */
    endEtherPacketDataGet,          /* get pointer to data in NET_BUFFER */
    endEtherPacketAddrGet,          /* Get packet addresses. */
    NULL                            /* Bind */
};

#ifdef PERF_DEBUG
HISTOGRAM_DEFINE_BINS(arTx, 1, AR_TX_QUEUE_DEPTH + AR_TX_QUEUE_FRAG_DELTA, 2);
HISTOGRAM_DEFINE_BINS(arRx, 1, AR_RX_QUEUE_DEPTH, 2);
HISTOGRAM_DEFINE_BINS(arTxDelta, 1, 16, 0);
LOCAL A_UINT32 onFullQueue  = 0;
LOCAL A_UINT32 onEmptyQueue = 0;

int wlanDump(void)
{
    extern void dumpResourceUsage(int);

    HISTOGRAM_DUMP("wlan tx processing", arTx, 0);
    HISTOGRAM_DUMP("wlan rx processing", arRx, 0);
    HISTOGRAM_DUMP("wlan good tx deltas (unit=1TU)", arTxDelta, 0);
    uiPrintf("Tx on empty = %u, on full = %u\n", onEmptyQueue, onFullQueue);
    onFullQueue  = 0;
    onEmptyQueue = 0;
    dumpResourceUsage(0);
    return 0;
}

#endif /* PERF_DEBUG */

#ifdef AR5311_ES
/* Initial work around for bug #3403, 2nd word of
 * status corrupted with RX data.  If 1st word is
 * written, the HW is really done with the frame.
 *
 * Mark this frame done, let swretry do it's thing with
 * the corrupt sequence # to be dropped by the STA and
 * eventually finish the frame.  This is only a start.
 *
 * Incomming descriptor must already be synced with the cache.
 *
 * XXX This routine should be removed and old chips used 'as is'.
 */
LOCAL INLINE void
ar5311TxDoneWAR(ATHEROS_DESC *pTxDesc)
{
    if (!pTxDesc->hwStatus.tx.done &&
        !pTxDesc->hwStatus.tx.pktTransmitOK &&
        (pTxDesc->hwStatus.tx.excessiveRetries || pTxDesc->hwStatus.tx.filtered))
    {
#ifdef DEBUG    /* this is annoying but want to keep an eye on it. */
        uiPrintf("WAR 3403: %p %x %x\n", pTxDesc,
                 pTxDesc->hwStatus.word[0], pTxDesc->hwStatus.word[1]);
#endif
        pTxDesc->hwStatus.tx.done = 1;
    }
}
#else /* AR5311_ES */
#define ar5311TxDoneWAR(desc)
#endif

/*
 *  Control our state variables for frames pending
 */
LOCAL INLINE void
txFramePendIncrement(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib, ATHEROS_DESC *pDesc)
{
    pdevInfo->numTxFramesPending++;
    if (pSib != NULL) {
        if (pDesc->psFlag) {
            pSib->numPSTxPending++;
        }
        pSib->numTxPending++;
    }
}

LOCAL INLINE void
txFramePendDecrement(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib, ATHEROS_DESC *pDesc)
{
    ASSERT(pdevInfo->numTxFramesPending > 0);
    pdevInfo->numTxFramesPending--;
    if (pSib != NULL) {
        ASSERT(pSib->numTxPending > 0);
        pSib->numTxPending--;
        if (pDesc->psFlag) {
            ASSERT(pSib->numPSTxPending > 0);
            pSib->numPSTxPending--;
            /* update the power save queue state and TIM */
            if ((pSib->numPSTxPending == 0) &&
                (pSib->txPSQueue.qFrameCount == 0) &&
                (pSib->staState & STATE_ASSOC)) {
                wlanTimUpdate(pdevInfo, pSib->assocId, FALSE);
            }
            /*
             * this should be changed in the future when swRetry
             * implements the special process for ps frame XXX
             */
            pDesc->psFlag = 0;
        }
    }
}

/*
 *  This will return the final descriptor of a frame.  Asserts the
 *  chain is properly formed and verifies its input is non-NULL.
 */
LOCAL INLINE ATHEROS_DESC *
lastDescGet(ATHEROS_DESC *pDesc)
{
    ATHEROS_DESC *p;

    for (p = pDesc; p && p->hwControl.tx.more != 0; p = p->pNextVirtPtr)
        /*LOOP*/;
    ASSERT(p != NULL); // badly formed chain
    A_DESC_CACHE_INVAL(p);

    ar5311TxDoneWAR(p);         /* bug #3403 */

    return p;
}

AR5_PARAM arParam;

#include "swretry.inl"

/*****************************************************************************
* ar5hwcEndLoad - initialize the driver and device
*
* Function Description - This routine initializes the driver and the device
* to the operational state.
*
* Input Parameters: initString, this string is defined in configNet.h
*   For description of <initString> see ar5hwcParse
*
* RETURNS: An END object pointer or NULL on error.
*/
END_OBJ *
ar5hwcEndLoad(char *initString, void *p)
{
    WLAN_DEV_INFO   *pdevInfo = NULL;
    A_UINT16        cfgCountry;
    A_UINT16        opModes;

    if (initString == NULL) {
        return NULL;
    }
    if (initString[0] == '\0') {
        bcopy(AR_DEV_NAME, initString, strlen(AR_DEV_NAME));
        return 0;
    }

#if defined(SYSUNMAPPAGEZERO)
    /* Want to do this ASAP, this is the first hook we have. */
    sysUnmapPageZero();
#endif

    /* Always log malloc errors.  For DEBUG check free() calls. */
#ifdef DEBUG
    memOptionsSet(MEM_ALLOC_ERROR_LOG_FLAG | MEM_BLOCK_ERROR_LOG_FLAG);
#else
    memOptionsSet(MEM_ALLOC_ERROR_LOG_FLAG);
#endif

#ifdef SWCOHERENCY
    /* Double check that ATHEROS_DESC is nicely sized */
    if (sizeof(ATHEROS_DESC) != 128) {
        uiPrintf("ATHEROS_DESC is fishy!\n");
        goto errorExit;
    }
#if CACHE_LINE_SIZE > 32
#error Need to tune ATHEROS_DESC hw/sw alignment to large line size.
#endif
#endif

    /* allocate the device structure; calloc intializes to 0's */
    pdevInfo = (WLAN_DEV_INFO *)calloc(sizeof(WLAN_DEV_INFO) + sizeof(END_OBJ), 1);
    if (pdevInfo == NULL) {
        uiPrintf("Could not alloc memory for device structures\n");
        goto errorExit;
    }
    pwlanDevInfo = pdevInfo;
    pdevInfo->pOSHandle = (END_OBJ *)(pdevInfo + 1);
    DRV_LOG (DRV_DEBUG_LOAD,("pdevInfo = %p, pEnd = %x\n", pdevInfo, pdevInfo->pOSHandle));

    /* Get parameters from the BSP and fill in the device structure */
    pdevInfo->devno         = 0;
    pdevInfo->baseAddress   = arParam.pciMemBase;
    pdevInfo->ivec          = arParam.ivec;

#ifdef DEBUG
    uiPrintf("Base address = %lx, irq %d\n", arParam.pciMemBase, arParam.ivec);
#endif

    /*
     * The pciInfo struct needs to be filled in prior to calling halAttach().
     * halAttach() needs to be called prior to calling apCfgInit() as it may
     * use HAL functions.
     */
    pdevInfo->pciInfo.VendorID          = arParam.pciVendorID;
    pdevInfo->pciInfo.DeviceID          = arParam.pciDeviceID;
    pdevInfo->pciInfo.SubVendorID       = arParam.pciSubVendorID;
    pdevInfo->pciInfo.SubVendorDeviceID = arParam.pciSubDeviceID;
    pdevInfo->pciInfo.ChipRevision      = arParam.pciRevisionID;

    if (halAttach(pdevInfo) != A_OK) {
        uiPrintf("Error initializing wireless software state.\n");
        goto errorExit;
    }

    /* halAttached initialized macPermAddr.  Update other users. */
    A_MACADDR_COPY(&pdevInfo->staConfig.macPermAddr,
                   &pdevInfo->staConfig.macAddr);
    A_MACADDR_COPY(&pdevInfo->staConfig.macPermAddr,
                   &pdevInfo->staConfig.bssId);

    /* Load configuration.  Depends on halAttach(). */
    if (apCfgInit() != OK) {
        goto errorExit;
    }

    eventDisplayFuncRegister(SUBSYS_WLANTX, eventDump_wlantx);
    eventDisplayFuncRegister(SUBSYS_WLANRX, eventDump_wlanrx);

    /* move some config values into the staConfig struct */
    pdevInfo->staConfig.defaultFragThreshold    = apCfgFragThresholdGet();
    pdevInfo->staConfig.defaultDot11RTSThreshold= apCfgRtsThresholdGet();
    pdevInfo->staConfig.rateCtrlEnable          = apCfgRateCtrlGet();
    pdevInfo->staConfig.swretryEnabled          = apCfgSwRetryModeGet();
    pdevInfo->staConfig.limitTxRetries          = apCfgLimitTxRetriesGet();
    pdevInfo->staConfig.swRetryMaxRetries       = DEFAULT_SWRETRY_MAXRETRIES;
    pdevInfo->staConfig.psQueueMaxDepth         = DEFAULT_PS_QDEPTH;
    pdevInfo->staConfig.calibrationTime         = apCfgCalibrationTimeGet();
    pdevInfo->staConfig.tpScale                 = apCfgPowerGet();
    pdevInfo->staConfig.QosEnabled              = apCfgQosGet();
    pdevInfo->staConfig.cwmin                   = apCfgCwminGet();
    pdevInfo->staConfig.aifs                    = apCfgAifsGet();

    cfgCountry = apCfgCountryCodeGet();
    if (wlanIsCountryCodeValid(pdevInfo, cfgCountry)) {
        pdevInfo->staConfig.countryCode = cfgCountry;
    } else {
        /* An incorrect Country Code is Set */
        pdevInfo->staConfig.countryCode = wlanGetDefaultCountry(pdevInfo);
        apCfgCountryCodeSet(pdevInfo->staConfig.countryCode);
    }
#if 0 /* Remove 802.11d Beacon support until STA beacon parsing is cleaned */
    if ((pdevInfo->staConfig.countryCode != CTRY_DEFAULT) &&
        (pdevInfo->staConfig.countryCode != CTRY_DEBUG)) {
        pdevInfo->staConfig.multiDomainCapEnabled = TRUE;
    }
#endif

    /* Build channel list based on config Info */
    opModes = apCfgFreqSpecGet();
    if(((opModes >> 3) & 0x1) + ((opModes >> 2) & 0x1) +
        ((opModes >> 1) & 0x1) + (opModes & 0x1) != 1)
    {
        uiPrintf("More than one wireless modes are specified - setting to 11A operation\n");
        opModes = MODE_SELECT_11A;
        apCfgFreqSpecSet(opModes);
    }

    wlanInitChannelList(pdevInfo, &(pdevInfo->staConfig.pClist), pdevInfo->staConfig.countryCode,
        opModes, apCfgOutdoorChanEnabledGet());

    /* Setup the channel */
    pdevInfo->staConfig.pChannel = wlanGetChannelPtr(pdevInfo->staConfig.pClist,
        apCfgRadioChannelGet(), opModes);

    /* Fix channel config */
    if (pdevInfo->staConfig.pChannel == NULL) {
        pdevInfo->staConfig.pChannel = &(pdevInfo->staConfig.pClist->chanArray[0]);
        apCfgRadioChannelSetNow(pdevInfo->staConfig.pChannel->channel);
        uiPrintf("Setting channel frequency to be a member of the configured regulatory domain and operational modes\n");
    }
    pdevInfo->staConfig.phwChannel = pdevInfo->staConfig.pChannel;

    calBeaconCount = (pdevInfo->staConfig.calibrationTime * 1000) /
                     apCfgBeaconIntervalGet();

    /* parse the init string, filling in the device structure */
    if (ar5hwcParse(pdevInfo, initString) == ERROR) {
        DRV_LOG(DRV_DEBUG_LOAD,("Error in parsing string...\n"));
        goto errorExit;
    }

    pdevInfo->NicResetInProgress    = FALSE;
    pdevInfo->hwResetCount          = 0;
    pdevInfo->txUrn                 = 0;
    pdevInfo->txUrnBurst            = 0;
    pdevInfo->txUrnBurstMs          = 0;
    pdevInfo->globISRReg            = 0;

    if (halReset(pdevInfo, pdevInfo->staConfig.pChannel, FALSE) != A_OK) {
        uiPrintf("Error initializing wireless hardware.\n");
        goto errorExit;
    }

    halSetLedState(pdevInfo, TRUE);

    /* initialize the END and MIB2 parts of the structure */
    /* What's the speed of the wlan? */
    if ((END_OBJ_INIT(pdevInfo->pOSHandle, (DEV_OBJ *)pdevInfo, AR_DEV_NAME,
                      pdevInfo->devno, &ar5hwcFuncTable,
                      AR_DRV_NAME) == ERROR) ||
        (END_MIB_INIT(pdevInfo->pOSHandle, M2_ifType_ethernet_csmacd,
                      pdevInfo->staConfig.macAddr.octets, 6, ETHERMTU,
                      END_SPEED) == ERROR))
    {
        uiPrintf("Error: could not set MIB state.\n");
        goto errorExit;
    }

    /* Perform memory allocation/distribution */
    if (ar5hwcMemInit(pdevInfo) == ERROR) {
        goto errorExit;
    }

    /* set the flags to indicate readiness */
    END_OBJ_READY(pdevInfo->pOSHandle,
                  IFF_NOTRAILERS /*| IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST */);

    DRV_LOG(DRV_DEBUG_LOAD,("Done loading ar5\n"));

    return pdevInfo->pOSHandle;

errorExit:
    if (pdevInfo != NULL) {
        free(pdevInfo);
        pwlanDevInfo = NULL;
    }
    DRV_LOG(DRV_DEBUG_LOAD,("errorExit from ar5hwcEndLoad...\n"));
    return NULL;
}

/*****************************************************************************
* ar5hwcParse - parse the init string
*
* Function Description: Parse the input string.  Fill in values in the driver
* control structure.
*
* The initialization string format is:
* "unit:pDmaBuf:dmaBufSize:txDescCount:rxDescCount:extraClCount:extraTxDescCount"
*
* .bS
* unit              Device unit number, a small integer
* pDmaBuf           Shared memory address
* dmaBufSize        Size of memory reserved
* txDescCount       Transmit descriptor count - depth of the tx queue; this
*                     generally determines the latency through the system,
*                     trading it off against the burst tolerance of the system
* rxDescCount       Receive descriptor count - depth of the rx queue; this is
*                     generally a function of the scheduling granularity of the
*                     system
* extraClCount      Extra data buffers for loan to devices outside like
*                     ethernet, local stack etc
* extraTxDescCount  Extra transmit descriptors for power save etc; these
*                     are suggested based on the memory availability
* .bE
*
* NULL (0) in these fields are used to specify corresponding default values
*
* If <pDmaBuf> is specified as NULL (0) or NONE (-1), the driver allocates
* the required DMA buffer area using malloc() (do not want cacheDmaMalloc()
* as may be uncached unless this is requested).
*
* Otherwise, <pDmaBuf> should contain the desired
* address, whose size is specified by <dmaBufSize> that is safe to use
* with the BSPs CACHE_DMA_*() routines.
*
* RETURNS: OK or ERROR for invalid arguments.
*/
LOCAL STATUS
ar5hwcParse(WLAN_DEV_INFO *pdevInfo, char *initString)
{
    A_UINT16 txQueueDepth;
    A_UINT32 tmp32;
    char     *pHolder = NULL;
    char     *tok;

    /* Parse the initString */

    /* Unit number. */
    if ((tok = strtok_r(initString, ":", &pHolder)) == NULL) {
        return ERROR;
    }
    pdevInfo->devno = strtoul(tok, NULL, 0);
    DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: got devno\n", pdevInfo->devno));

    /* Device DMA buffer address. */
    if ((tok = strtok_r(NULL, ":", &pHolder)) == NULL) {
        return ERROR;
    }
    pdevInfo->pDmaBuf = (char *)strtoul(tok, NULL, 0);
    if (pdevInfo->pDmaBuf == (char *)NONE) {
        pdevInfo->pDmaBuf = NULL;
    }
    DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: pDmaBuf = 0x%08x\n", pdevInfo->devno, (A_UINT32)pdevInfo->pDmaBuf));

    /* DMA buffer size. */
    if ((tok = strtok_r(NULL, ":", &pHolder)) == NULL) {
        return ERROR;
    }
    pdevInfo->dmaBufSize = strtoul(tok, NULL, 0);
    DRV_LOG (DRV_DEBUG_LOAD,("wlan%d: dmaBufSize = %d\n", pdevInfo->devno, pdevInfo->dmaBufSize));

    /* Transmit descriptor count. */
    if ((tok = strtok_r(NULL, ":", &pHolder)) == NULL) {
        return ERROR;
    }
    tmp32 = strtoul(tok, NULL, 0);
    if ((tmp32 == 0) || (tmp32 > USHRT_MAX)) {
        txQueueDepth = AR_TX_QUEUE_DEPTH;
        /*
         * For fragmentation, in order to maintain the similar burst
         * tolerance to non-fragmented we need a deeper tx queue.  It
         * is acceptable for the system to have higher latency that a
         * deeper tx queue may bring.
         */
        if (pdevInfo->staConfig.defaultFragThreshold < 1500) {
            txQueueDepth += AR_TX_QUEUE_FRAG_DELTA;
        }
    } else {
        txQueueDepth = tmp32;
    }

    DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: txQueueDepth = %d\n", pdevInfo->devno, txQueueDepth));

    /* Receive descriptor count. */
    if ((tok = strtok_r(NULL, ":", &pHolder)) == NULL) {
        return ERROR;
    }
    tmp32 = strtoul(tok, NULL, 0);
    pdevInfo->rxQueueDepth = ((tmp32 == 0) || (tmp32 > USHRT_MAX)) ? AR_RX_QUEUE_DEPTH : tmp32;
    DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: rxDescCount = %d\n", pdevInfo->devno, pdevInfo->rxQueueDepth));

    /* Extra cluster count - buffers for loaning outside. */
    if ((tok = strtok_r(NULL, ":", &pHolder)) == NULL) {
        return ERROR;
    }
    tmp32 = strtoul(tok, NULL, 0);
    pdevInfo->maxLoanedCount = ((tmp32 == 0) || (tmp32 > USHRT_MAX)) ? AR_MAX_LOANED_COUNT : tmp32;
    DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: maxLoanedCount = %d\n", pdevInfo->devno, pdevInfo->maxLoanedCount));

    /* Extra transmit descriptor count - power save queue depth really. */
    if ((tok = strtok_r(NULL, ":", &pHolder)) == NULL) {
        return ERROR;
    }
    tmp32 = strtoul(tok, NULL, 0);
    pdevInfo->psQueueDepth = ((tmp32 == 0) || (tmp32 > USHRT_MAX)) ? AR_PS_QUEUE_DEPTH : tmp32;
    DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: psQueueDepth = %d\n", pdevInfo->devno, pdevInfo->psQueueDepth));

    /* we need one header buffer per transmitted frame/frag */
    pdevInfo->txHeaderDescCount = txQueueDepth;

    /* the free descriptors are used for data portion of each
     * tx frame, to be put on the receive queue, and each sme
     * frame transmitted;  sme frames transmitted always get
     * charged to the trasmit queue directly;  the rx queue is
     * always maintained at its constant depth;  thus the max
     * number of free descriptors available on the actual empty
     * descriptor queue at any time corresponds to the desired
     * limit on the transmit queue
     */
    pdevInfo->freeDescCount = txQueueDepth;

    /* for now */
    pdevInfo->numDefragLanes = DEFRAG_ENTRIES_DEFAULT;

    return OK;
}

#ifdef DEBUG
NET_POOL_ID pNetPool;
#endif

/******************************************************************************
* ar5hwcMemInit - initialize queues and netpool for the device
*
* Function Description: This function allocate the memroy pool required by this
* driver. Also, it creates the software tx queue and software rx queue
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS
ar5hwcMemInit(WLAN_DEV_INFO *pdevInfo)
{
    M_CL_CONFIG     mclBlkConfig = {0, 0, NULL, 0};
    CL_DESC         clConfig = {0, 0, NULL, 0};
    ATHEROS_DESC    *pDesc;
    A_UINT16        numDescriptors;
    char            *pMemBuf;
    A_UINT32        totalSize;
    A_UINT32        error;

    /* Allocate memory for the net pool */
    if ((pdevInfo->pOSHandle->pNetPool = malloc(sizeof(NET_POOL))) == NULL) {
        DRV_LOG (DRV_DEBUG_LOAD,("wlan%d: Could not allocate net pool memory\n", pdevInfo->devno));
        goto arMemInitFail;
    }
#ifdef  DEBUG
    pNetPool = pdevInfo->pOSHandle->pNetPool;
#endif

    /* Allocate memory for mBlks and clBlks */
    mclBlkConfig.mBlkNum  = pdevInfo->maxLoanedCount;
    mclBlkConfig.clBlkNum = pdevInfo->rxQueueDepth + pdevInfo->maxLoanedCount;

    /*
     * Because every mBlk and cluster is prepended by a 4-byte header -
     * which is why the size calculations for clusters and mBlk structures
     * contain an extra sizeof(long)
     */
    mclBlkConfig.memSize = (mclBlkConfig.mBlkNum * (M_BLK_SZ + sizeof(long))) +
                           (mclBlkConfig.clBlkNum * (CL_BLK_SZ + sizeof(long)));
    mclBlkConfig.memArea = memalign(sizeof(long), mclBlkConfig.memSize);
    pdevInfo->pMclMem    = mclBlkConfig.memArea;

    if (mclBlkConfig.memArea == NULL) {
        DRV_LOG (DRV_DEBUG_LOAD,("wlan%d: Could not allocate mclBlk memory\n", pdevInfo->devno));
        goto arMemInitFail;
    }

    /* memory for clusters: vxworks prepends a reserved long to each
     * cluster; we want real clusters to begin cache line aligned; so
     * make sure that the size of the cluster we use is 4 bytes less
     * than a multiple of cache line size
     */
    clConfig.clSize  = A_ROUNDUP(AR_BUF_SIZE + sizeof(long), CACHE_LINE_SIZE) - sizeof(long);
    clConfig.clNum   = mclBlkConfig.clBlkNum;
    clConfig.memSize = clConfig.clNum * (clConfig.clSize + sizeof(long));


    /* calculate total number of descriptors needed; this is
     * the number of free descriptors; plus each header buffer
     * for the scattered wlan header needs a descriptor; plus
     * descriptors for the receive ring; once they leave the
     * rx ring they are freed or charged to defrag/SME/transmit
     */
    numDescriptors = pdevInfo->freeDescCount + pdevInfo->txHeaderDescCount +
                     pdevInfo->psQueueDepth + pdevInfo->rxQueueDepth;

    /* allocate memory for clusters, free and rx queues; begin
     * by accounting for cluster memory while allowing a cache
     * line in the beginning so that clusters can begin at 4
     * bytes less than a cache line boundary
     */
    totalSize = CACHE_LINE_SIZE + clConfig.memSize;

    /* descriptor size needs to be multiple of a cache line; account
     * for descriptor memory while allowing a cache line round up from
     * cluster memory
     */
    ASSERT((AR_DESC_SIZE & (CACHE_LINE_SIZE - 1)) == 0);
    totalSize += CACHE_LINE_SIZE + (AR_DESC_SIZE * numDescriptors);

    /* memory for header buffers */
    ASSERT((AR_HEADER_SIZE & (CACHE_LINE_SIZE - 1)) == 0);
    totalSize += pdevInfo->txHeaderDescCount * AR_HEADER_SIZE;

    /* memory for power save buffers */
    totalSize += pdevInfo->psQueueDepth * A_ROUNDUP(AR_BUF_SIZE, CACHE_LINE_SIZE);

    if ((pdevInfo->pDmaBuf != NULL) && (pdevInfo->pDmaBuf != (char *)NONE)) {
        if (pdevInfo->dmaBufSize < totalSize) {
            DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: not enough memory provided, need %ul got %d\n",
                    pdevInfo->devno, totalSize, pdevInfo->dmaBufSize));
            goto arMemInitFail;
        }
    } else {
        /* Allocate the whole memory for cluster, rx and empty queue */
#ifdef VX_CACHE_DMA_MALLOC
        pdevInfo->pDmaBuf = cacheDmaMalloc(totalSize);
#else
        pdevInfo->pDmaBuf = malloc(totalSize);
#endif
        pdevInfo->dmaBufSize = totalSize;
        if (pdevInfo->pDmaBuf == NULL) {
            goto arMemInitFail;
        }
        pdevInfo->flags |= AR_BUF_LOCAL;
        DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: DevMem allocated, pDmaBuf=%p\n", pdevInfo->devno, pdevInfo->pDmaBuf));
    }
    A_MEM_ZERO(pdevInfo->pDmaBuf, totalSize);

    /* setup memory for data clusters - align to 4 bytes ahead of a cacheline
     * boundary so that the data buffer falls on a cacheline boundary - vxworks
     * prepends a 4 byte overhead to each cluster
     */
    clConfig.memArea = (char *)A_ROUNDUP((A_UINT32)pdevInfo->pDmaBuf + sizeof(long), CACHE_LINE_SIZE) - sizeof(long);
    pMemBuf = (char *)clConfig.memArea + clConfig.memSize;

    /* set up memory for descriptor queues - aligned to cacheline boundaries */
    pdevInfo->emptyDescQueue.pDescQueueHead = (ATHEROS_DESC *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)pdevInfo->emptyDescQueue.pDescQueueHead + (AR_DESC_SIZE * pdevInfo->freeDescCount);

    pdevInfo->txHeaderQueue.pDescQueueHead = (ATHEROS_DESC *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)pdevInfo->txHeaderQueue.pDescQueueHead + (AR_DESC_SIZE * pdevInfo->txHeaderDescCount);

    pdevInfo->psBufferQueue.pDescQueueHead = (ATHEROS_DESC *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)pdevInfo->psBufferQueue.pDescQueueHead + (AR_DESC_SIZE * pdevInfo->psQueueDepth);

    pdevInfo->rxQueue.pDescQueueHead = (ATHEROS_DESC *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)pdevInfo->rxQueue.pDescQueueHead + (AR_DESC_SIZE * pdevInfo->rxQueueDepth);


    /* Initialize the memory pool. */
    if ((error = netPoolInit(pdevInfo->pOSHandle->pNetPool, &mclBlkConfig, &clConfig, 1, NULL)) != OK) {
        DRV_LOG(DRV_DEBUG_LOAD, ("wlan%d: Failed to intialize memory pool error = %x\n", pdevInfo->devno, error));
        goto arMemInitFail;
    }
    DRV_LOG(DRV_DEBUG_LOAD, ("wlan%d: finished initializing memory pool\n", pdevInfo->devno));

    /* Create a tx descriptor queue */
    if (ar5hwcQueueCreate(pdevInfo, &pdevInfo->emptyDescQueue, &pdevInfo->emptyDescQueue,
                          pdevInfo->freeDescCount, 0) != OK)
    {
        DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: Failed to create tx desc queue, %p\n", pdevInfo->devno,
                pdevInfo->emptyDescQueue.pDescQueueHead));
        goto arMemInitFail;
    }
    DRV_LOG(DRV_DEBUG_LOAD, ("wlan%d: finished allocating tx desc queue\n", pdevInfo->devno));

    /* Create a header queue */
    if (ar5hwcQueueCreate(pdevInfo, &pdevInfo->txHeaderQueue, &pdevInfo->txHeaderQueue,
                          pdevInfo->txHeaderDescCount, 0) != OK)
    {
        DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: Failed to create tx header queue, %p\n", pdevInfo->devno,
                pdevInfo->txHeaderQueue.pDescQueueHead));
        goto arMemInitFail;
    }
    pMemBuf = (char *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    for (pDesc = pdevInfo->txHeaderQueue.pDescQueueHead; pDesc; pDesc = pDesc->pNextVirtPtr) {
        pDesc->pBufferVirtPtr.ptr = pMemBuf;
        pDesc->bufferPhysPtr      = (A_UINT32)pMemBuf;
        pDesc->pOSDescPtr         = NULL;
        pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
        pMemBuf += AR_HEADER_SIZE;
    }
    DRV_LOG(DRV_DEBUG_LOAD, ("wlan%d: finished allocating tx header queue\n", pdevInfo->devno));

    /* Create a header queue */
    if (ar5hwcQueueCreate(pdevInfo, &pdevInfo->psBufferQueue, &pdevInfo->psBufferQueue,
                          pdevInfo->psQueueDepth, 0) != OK)
    {
        DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: Failed to create power save buffers queue, %p\n", pdevInfo->devno,
                pdevInfo->psBufferQueue.pDescQueueHead));
        goto arMemInitFail;
    }
    pMemBuf = (char *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    for (pDesc = pdevInfo->psBufferQueue.pDescQueueHead; pDesc; pDesc = pDesc->pNextVirtPtr) {
        pDesc->pBufferVirtPtr.ptr = pMemBuf;
        pDesc->bufferPhysPtr      = (A_UINT32)pMemBuf;
        pDesc->pOSDescPtr         = NULL;
        pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
        pMemBuf += AR_BUF_SIZE;
    }
    DRV_LOG(DRV_DEBUG_LOAD, ("wlan%d: finished allocating power save buffer queue\n", pdevInfo->devno));

    /* Create a Rx queue */
    if (ar5hwcQueueCreate(pdevInfo, &pdevInfo->rxQueue, &pdevInfo->emptyDescQueue,
                          pdevInfo->rxQueueDepth, AR_BUF_SIZE) != OK)
    {
        DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: Failed to create rx queue, %p\n", pdevInfo->devno,
                pdevInfo->rxQueue.pDescQueueHead));
        goto arMemInitFail;
    }
    DRV_LOG(DRV_DEBUG_LOAD, ("wlan%d: finished allocating rx queue\n", pdevInfo->devno));

    ASSERT(pMemBuf <= (pdevInfo->pDmaBuf + pdevInfo->dmaBufSize));

    /* Initialize the rxQueue, txQueue and altQueue */

    resetRxQueue(pdevInfo);

    pdevInfo->txQueue.pDescQueueHead = NULL;
    pdevInfo->txQueue.pDescQueueTail = NULL;
    A_SEM_INIT(pdevInfo->txQueue.qSem, SEM_Q_PRIORITY, CREATE_UNLOCKED);

    pdevInfo->altQueue.pDescQueueHead = NULL;
    pdevInfo->altQueue.pDescQueueTail = NULL;
    A_SEM_INIT(pdevInfo->altQueue.qSem, SEM_Q_PRIORITY, CREATE_UNLOCKED);

    if (pdevInfo->txQueue.qSem == NULL || pdevInfo->altQueue.qSem == NULL) {
        goto arMemInitFail;
    }

    pdevInfo->lastTxFrameDesc = NULL;

    DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: Memory setup complete\n", pdevInfo->devno));
    return OK;

arMemInitFail:
    /* Clean all the queues, tx, rx, alt, tx and header queues*/
    ar5hwcQueueDelete(&pdevInfo->altQueue);
    ar5hwcQueueDelete(&pdevInfo->txQueue);
    ar5hwcQueueDelete(&pdevInfo->rxQueue);
    ar5hwcQueueDelete(&pdevInfo->psBufferQueue);
    ar5hwcQueueDelete(&pdevInfo->txHeaderQueue);
    ar5hwcQueueDelete(&pdevInfo->emptyDescQueue);

    /* free memory allocated for clusters and Free/Rx Queues */
    if ((pdevInfo->flags & AR_BUF_LOCAL) && (pdevInfo->pDmaBuf != NULL)) {
#ifdef VX_CACHE_DMA_MALLOC
        cacheDmaFree(pdevInfo->pDmaBuf);
#else
        free(pdevInfo->pDmaBuf);
#endif
    }

    /* free memory allocated for mBlk-clBlks */
    if (mclBlkConfig.memArea != NULL) {
        free(mclBlkConfig.memArea);
    }

    /* free memory allocated for net pool */
    if (pdevInfo->pOSHandle->pNetPool != NULL) {
        free(pdevInfo->pOSHandle->pNetPool);
    }

    DRV_LOG(DRV_DEBUG_LOAD,("wlan%d: Memory Init failed\n", pdevInfo->devno));

    return ERROR;
}

/*****************************************************************************
* ar5hwcEndStart - start the device
*
* This function calls BSP functions to connect interrupts and start the
* device running in interrupt mode.
*
* RETURNS: OK or ERROR
*/
LOCAL STATUS
ar5hwcEndStart(END_OBJ *pEnd)
{
    WLAN_DEV_INFO *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;

    /* Install interrupt vector */
#if defined(PCI_INTERFACE)
    STATUS result;

    /* Install interrupt vector */
    result = pciIntConnect(INUM_TO_IVEC(pdevInfo->ivec), ar5hwcEndInt, (int)pdevInfo);
    if (result == ERROR) {
        return ERROR;
    }
#elif defined(AR5311)
    intConnect(INUM_TO_IVEC(pdevInfo->ivec), ar5hwcEndInt, (int)pdevInfo);
#else
#error interrupt connect.
#endif

    /* mark the interface -- up */
    END_FLAGS_SET(pdevInfo->pOSHandle, IFF_UP | IFF_RUNNING);

    /* Create the task to do periodic calibration */
    if ((wlanStop == FALSE) && (pdevInfo->staConfig.calibrationTime != 0)) {
        /* Create a message queue for offset calibration */
        calMsgQId = msgQCreate(10, 20, MSG_Q_FIFO);
        taskSpawn("tWlanCal", CALIBRATION_TASK_PRIORITY, 0, 2048,
                  (FUNCPTR)athCalibrationTask,
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    }

    return OK;
}

/*****************************************************************************
* ar5hwcEndStop - stop the device
*
* This function calls BSP functions to disconnect interrupts, disable the
* interrupts, disable the tx and rx
*
* RETURNS: OK or ERROR
*/
LOCAL STATUS
ar5hwcEndStop(END_OBJ *pEnd)
{
    WLAN_DEV_INFO *pdevInfo;

    if (pEnd->pFuncTable == &ar5hwcFuncTable) {
        pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;
    } else {
        /* muxDevStopAll pases pDevice instead pEnd, which is a bug! */
        pdevInfo = (WLAN_DEV_INFO *)pEnd;
    }

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    if (!(END_FLAGS_GET(pdevInfo->pOSHandle) & (IFF_UP | IFF_RUNNING))) {
        logMsg("ar5hwcEndStop -- already stopped!\n");
        return ERROR;
    }

    if (beaconWdgCancel() != OK) {
        logMsg("ar5hwcEndStop: Error shutting down beacon watchdog timer!\n");
        return ERROR;
    }

    /* mark the interface -- down */
    END_FLAGS_CLR(pdevInfo->pOSHandle, IFF_UP | IFF_RUNNING);

    /* system is shutting down.... Turn out the lights. */
    sysLed(0);

    /*
     * Steps in stopping:
     *  - Disable, and disconnect the interrupt.
     *  - Put the chip into the reset state -- will disable all rx/tx.
     *  - Ignore any errors!  Need to make sure we reset the chip.
     */
    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    intDisable(arParam.ilevel);

#if defined(PCI_INTERFACE)
    (void)pciIntDisconnect(INUM_TO_IVEC(arParam.ivec), ar5hwcEndInt);
#endif

    halDisable(pdevInfo);
    pdevInfo->localSta->staState = STATE_AP_DOWN;
    wlanStop = TRUE;

    return OK;
}


/******************************************************************************
* ar5hwcQueueCreate - create a link list queue of descriptors
*
* This routine allocates a descriptor buffer array aligned  on a word
* boundary, initializes and links the array.
*
* RETURNS: OK or ERROR
*/
LOCAL STATUS
ar5hwcQueueCreate(WLAN_DEV_INFO *pdevInfo, QUEUE_DETAILS *pQueue, QUEUE_DETAILS *freeQueue,
                  int count, int bufSize)
{
    ATHEROS_DESC *pDesc;
    int          i;

    /* Initialize queue data */
    pQueue->pDescQueueTail = pQueue->pDescQueueHead + (count - 1);

    A_SEM_INIT (pQueue->qSem, SEM_Q_PRIORITY, CREATE_UNLOCKED);
    if (pQueue->qSem == NULL) {
        logMsg("Error:  could not create a semaphore for queue %p\n", pQueue);
        return ERROR;
    }

    /* Initialize each descriptor */
    for (i = 0, pDesc = pQueue->pDescQueueHead; i < count; ++i, pDesc = pDesc->pNextVirtPtr) {
        pDesc->nextPhysPtr  = A_DESC_V2P((UINT32)pDesc + sizeof(ATHEROS_DESC));
        pDesc->pNextVirtPtr = pDesc + 1;
        pDesc->thisPhysPtr  = A_DESC_V2P(pDesc);
        pDesc->freeQueue    = freeQueue;

        if (bufSize) {
            if (hwAllocatePhysBuffer(pdevInfo, bufSize, &pDesc->bufferPhysPtr,
                &pDesc->pBufferVirtPtr.ptr, &pDesc->pOSDescPtr) != A_OK)
            {
                logMsg("Error: could not allocate buffer for descriptor\n");
                return ERROR;
            }
            ASSERT(((A_UINT32)pDesc->pBufferVirtPtr.ptr & (CACHE_LINE_SIZE - 1)) == 0);
            pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
        }
    }
    pQueue->pDescQueueTail->nextPhysPtr  = 0;
    pQueue->pDescQueueTail->pNextVirtPtr = NULL;

    return OK;
}

/******************************************************************************
* ar5hwcQueueDelete - Delete a queue
*
* This routine clean all the queue field in QUEUE_DETAIL, the memeory clean is
* not done here. This function and ar5hwcQueueCreate are only called when we load
* the driver.
*
* RETURNS: NONE
*/
LOCAL void
ar5hwcQueueDelete(QUEUE_DETAILS *pQueue)
{
    pQueue->pDescQueueHead = NULL;
    pQueue->pDescQueueTail = NULL;
    pQueue->physQueueHead  = 0;

    if (pQueue->qSem != NULL) {
        A_SEM_DELETE(pQueue->qSem);
        pQueue->qSem = NULL;
    }
}

/*
 * If this is a WEP frame we need to:
 *  - generate the random IV (incrementing is good enough)
 *  - Add the ICV to the last descriptor's length.
 *  - perform sw encryption if needed
 *  - Adjust the header length to accomodate the PCU's need to include
 *    need to include the 32 bit phys pointer to the descriptor.
 *  - Set the key cache index.
 *  - Set the descriptor key valid flag.
 */
INLINE A_STATUS
setupEncrypt(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    WLAN_FRAME_3_DATA_ENCRYPT   *hdr;
    WLAN_PRIV_RECORD            *pKey;
    SIB_ENTRY                   *pSib;
    A_UINT32                    hwidx;
    A_BOOL                      faulted;
    A_UINT32                    iv;
    A_UINT32                    ivmask;
    A_UINT32                    icvsize;

    hdr  = pDesc->pBufferVirtPtr.header3Iv;
    pSib = pDesc->pDestSibEntry;

    if (hdr->macHeader.frameControl.wep) {
        /* multicast frames do not have a SIB - use shared key. */
        if (pSib == NULL) {
            /*
             * after vlan support encryptKeyIndex is carried as part of
             * the descriptor for multicast frames
             */
            hwidx = halGetTxDescEncryptKeyIndex(pdevInfo, pDesc);
            pKey  = &pdevInfo->keyTable[hwidx];

            /* the power saved multicast frame has been set up */
            if (pDesc->psFlag) {
                return A_OK;
            }
        } else {
            hwidx = wlanKeyCacheFault(pdevInfo, pSib, &faulted);
            if (hwidx == KEYINDEX_INVALID) {
                return A_NO_RESOURCE;
            }
            if (faulted) {
                swretryDestMaskFault(pDesc->pDestSibEntry, pDesc);
            }
            pKey = pSib->pPriv;
        }

        ASSERT(pKey);

        /* Increment (randomize) the per key IV */
        iv = pKey->keyIV + 1;

        if (pKey->keyType == PRIV_KEY_TYPE_AES) {
            ivmask = 0x0fffffff;                /* 28 bit IV */
            icvsize = AES_ICV_FIELD_SIZE;
        } else {
            ivmask = 0x00ffffff;
            icvsize = WEP_ICV_FIELD_SIZE;

            /*
             * Filter keys that are weak according to "Weaknesses in the
             * Key Scheduling Algorithm of RC4" by Scott Fluhrer, Itsik
             * Martin and Adi Shimir.  We do not filter other "weak"
             * keys, so this is only a patch.
             *
             * Form: any,0xff,0x3-0x8,0xf,0x13 (40, 104, 128b keys)
             */
            if ( ((iv & 0xffff) >= 0xff03) &&
                 ((iv & 0xffff) <= (0xff03 + (pKey->keyLength >> 3))) )
            {
                iv = (iv & 0xffff00) | (4 + (pKey->keyLength >> 3));
            }
        }

        pKey->keyIV = iv;
        hdr->iv     = (iv & ivmask) | ((pKey->keyFlags & PRIV_FRAMEKEY) << 30);

        /* sw encryption support for Crete Chip */
        if (pdevInfo->localSta->swEncryptionEnabled &&
            (halIsSwAesSupported(pdevInfo) ||
            (!halIsSwAesSupported(pdevInfo) &&
            (pKey->keyType != PRIV_KEY_TYPE_WEP))))
        {
            if (swEnCipher(pdevInfo, pKey, pDesc) != A_OK) {
                return A_ERROR;
            } else {
                /* add icv to the last buffer and frame length */
                pDesc->pTxLastDesc->hwControl.tx.bufferLength += icvsize;

                /*
                 * reset the hdr since the old buffer is freed
                 * for some AES case
                 */
                hdr = pDesc->pBufferVirtPtr.header3Iv;

                /* support AES_OCB_REVA mode */
                if (!halIsSwAesSupported(pdevInfo) &&
                    (pKey->keyType == PRIV_KEY_TYPE_AES))
                {
                        hdr->iv = A_swab32(hdr->iv);
                        hdr->macHeader.frameControl.wep = 0;
                }
            }
        }

        /*
         * Set Desc Header Length for AR5210 WEP'ed packets.
         * this field is dc for nonWEP, so needs to check the wep bit again
         * since AES_OCB_REVA frame has wep bit cleared
         */
        if (hdr->macHeader.frameControl.wep) {
            halSetTxDescHdrLen(pdevInfo, pDesc,
                               macHeaderLenGet(&(hdr->macHeader.frameControl)) + 4);
        }
        pDesc->hwControl.tx.frameLength += icvsize;
        halSetTxDescEncryptKeyIndex(pdevInfo, pDesc, hwidx);
    } else if (pSib != NULL) {
        /*
         * Always allocate a key cache entry for the non-wep case.
         * This is needed both by software retry, and for antenna
         * diversity with the ar5211 in AP mode.
         */
        hwidx = wlanKeyCacheFault(pdevInfo, pDesc->pDestSibEntry, &faulted);
        if (hwidx == KEYINDEX_INVALID) {
            return A_NO_RESOURCE;
        }
        if (faulted) {
            swretryDestMaskFault(pDesc->pDestSibEntry, pDesc);
        }
        halSetTxDescEncryptKeyIndex(pdevInfo, pDesc, hwidx);
    }

    return A_OK;
}

/*
 * If this is a WEP frame we need to:
 *- get the random IV from the frame
 *- get the key from sib entry
 *- sw decrypt the frame if needed
 *- update decrypt error counter
 */
INLINE A_STATUS
setupDecrypt(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    WLAN_FRAME_3_DATA_ENCRYPT   *hdr;
    WLAN_PRIV_RECORD            *pKey;
    SIB_ENTRY                   *pSib;
    A_UINT32                    index;
    A_UINT32                    ivmask;
    A_UINT32                    icvsize = 0;
    A_UINT32                    status;

    hdr  = pDesc->pBufferVirtPtr.header3Iv;
    pSib = pDesc->pSrcSibEntry;

    /* support AES_OCB_REVA mode */
    if (!hdr->macHeader.frameControl.wep &&
        hdr->macHeader.frameControl.fType == FRAME_DATA)
    {
        /*
         * AES' works on older h/w by not setting the WEP bit;
         * must fake it for s/w sake, for unicast AES' negotiated frames
         */
        if (pSib && pSib->swEncryptionEnabled) {
            /* but only for frames that really have some data */
            if (pDesc->hwStatus.rx.dataLength > MIN_AES_OCB_DATA_LENGTH) {
                hdr->macHeader.frameControl.wep = TRUE;
                hdr->iv = le2cpu32(hdr->iv);
            }
        }
    }

    if (hdr->macHeader.frameControl.wep) {
        /* key index from IV */
        index = hdr->iv;
        /* check the key ID for shared key */
        /* ID bits are the high-order ones */
        index >>= 30;

        if (index != 0) {
            pKey = &pdevInfo->keyTable[index];

            if ((pKey == NULL || pKey->keyLength == 0) ||
                (pKey->keyFlags & PRIV_UNIQUEKEY)) {
                pdevInfo->localSta->stats.RcvWEPExcludedCount++;
                if (pSib) {
                    pSib->stats.RcvWEPExcludedCount++;
                }

                return A_ERROR;
            }
        } else {
            if (!pSib || (pSib->keyIndex == KEYINDEX_INVALID)) {
                return A_ERROR;
            } else {
                pKey = pSib->pPriv;
                if (pKey == NULL || pKey->keyLength == 0) {
                    pdevInfo->localSta->stats.RcvWEPExcludedCount++;
                    pSib->stats.RcvWEPExcludedCount++;

                    return A_ERROR;
                }
            }
        }

        ASSERT(pKey);

        switch (pKey->keyType) {
        case PRIV_KEY_TYPE_WEP:
            ivmask  = 0x00ffffff;           /* 24 bit IV */
            icvsize = WEP_ICV_FIELD_SIZE;   /* WEP */
            break;

        case PRIV_KEY_TYPE_AES:
            ivmask  = 0x0fffffff;           /* 28 bit IV */
            icvsize = AES_ICV_FIELD_SIZE;   /* AES */
            break;

        default:
#ifdef DEBUG
            uiPrintf("unsupported key type!\n");
#endif
            ivmask  = 0x00ffffff;           /* 24 bit IV */
            icvsize = WEP_ICV_FIELD_SIZE;   /* WEP */
            return A_ERROR;
        }

        pKey->keyIV = (hdr->iv) & ivmask;

        if (pdevInfo->localSta->swEncryptionEnabled ||
            (pSib && pSib->swEncryptionEnabled))
        {
            status = swDeCipher(pdevInfo, pKey, pDesc);
            if (status != A_OK) {
                pdevInfo->localSta->stats.RcvDecipherErrors++;
                if (pSib) {
                    pSib->stats.RcvDecipherErrors++;
                }
                return status;
            }
        }
    }

    /* remove ICV field */
    pDesc->hwStatus.rx.dataLength -= icvsize;

    return A_OK;
}

/*******************************************************************************
* ar5hwcDrvSend - the driver send routine
*
* This routine takes a Atheros Descriptor and sends out to the DMA engine.
*/
STATUS ar5hwcDrvSend(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pHead, A_BOOL callerHoldsLock)
{
    ATHEROS_DESC    *pTxDesc;
    ATHEROS_DESC    *pTrailingDesc;
    ATHEROS_DESC    *newDesc;
    SIB_ENTRY       *pSib;
    A_STATUS         status;

    if ((pHead == NULL) || (pdevInfo->NicResetInProgress == TRUE)) {
#ifdef DEBUG
        uiPrintf("ar5hwcDrvSend: Error\n");
#endif
        return ERROR;
    }

    pTrailingDesc = NULL;
    pTxDesc       = pHead;

    /* lock out the tx completion code */
    if (!callerHoldsLock) {
        A_SEM_LOCK(pdevInfo->txQueue.qSem, WAIT_FOREVER);
    }

    while (pTxDesc != NULL) {
        /*
         *  This routine might do the following with the desc:
         *  1) enqueue it and set pTxDesc to the next frame
         *  2) set some state (clearDestMask) then pass it thru
         *
         *  Because of 1), we must call this before setupEncrypt().
         */
        newDesc = swretrySetup(pdevInfo, pTxDesc, pTrailingDesc);

        if (pTxDesc != newDesc) {
            /*
             *  This frame ended up on the swretry queue! newDesc
             *  now pts to the subsequent frame, if any; if the
             *  frame was the first frame, update the head pointer.
             */
            if (pTxDesc == pHead) {
                pHead = newDesc;
            }

            pTxDesc = newDesc;

            /*
             *  We don't update the qTail because we'll
             *  get it below.  If pTxDesc == NULL, then we
             *  want to keep its prev value anyway.
             */
            continue;
        }

        /* we get here iff swretrySetup() didn't eat the frame */
        pSib = pTxDesc->pTxFirstDesc->pDestSibEntry;

        txFramePendIncrement(pdevInfo, pSib, pTxDesc);

        /* Handle key cache, update descriptor + IV for WEP frames */
        status = setupEncrypt(pdevInfo, pTxDesc);
        if (status != A_OK) {
            ATHEROS_DESC *pLast, *pNext;

            /* oversubscribed key cache or swencryption failure */
            LOG_DROP_FRAME(pdevInfo, tx);
            if (pSib) {
                /* restore needClearDest flag */
                pSib->swRetry.needClearDest = pTxDesc->hwControl.tx.clearDestMask;
            }
            txFramePendDecrement(pdevInfo, pSib, pTxDesc);
#ifdef DEBUG    /* For now we want to know this is happening */
            uiPrintf("setupEncrypt failed: frame dropped for SIB %p\n", pSib);
#endif
            /* Free this frame, and move to the next leaving pTrailingDesc */
            pLast = lastDescGet(pTxDesc);
            pNext = pLast->pNextVirtPtr;
            if (pTrailingDesc != NULL) {
                pTrailingDesc->pNextVirtPtr = pNext;
                pTrailingDesc->nextPhysPtr  = pNext->thisPhysPtr;
            }
            pLast->pNextVirtPtr = NULL;
            pLast->nextPhysPtr  = 0;
            if (pTxDesc == pHead) {
                pHead = pNext;
            }
            freeBuffandDescChain(pdevInfo, pTxDesc);
            pTxDesc = pNext;                    /* next frame */
            continue;
        }

#ifdef DEBUG
        if (txVerbose >= 10) {
            dumpFrame(pTxDesc, "tx");
        }
#endif

        /* swap */
        FRAME_CONVERT_CPUTOLE(pTxDesc->pBufferVirtPtr.header,
                              pTxDesc->hwControl.tx.bufferLength);

        /* Proclaim our freshness */
        pTxDesc->staleFlag = 0;
        eventPost(SUBSYS_WLANTX, EV_TX_Q, "TX:", pTxDesc,
                  pTxDesc->hwControl.tx.more);
        while (pTxDesc->hwControl.tx.more != 0) {
            pTxDesc = pTxDesc->pNextVirtPtr;
            pTxDesc->staleFlag = 0;
            eventPost(SUBSYS_WLANTX, EV_TX_Q, "TX:",
                      pTxDesc, pTxDesc->hwControl.tx.more);
        }

        pTrailingDesc = pTxDesc;                /* last frame */
        pTxDesc = pTxDesc->pNextVirtPtr;        /* next frame */
    }

    if (pHead) {
        ASSERT(pTrailingDesc != NULL);

        /* Add to normal tx queue -  kick-start the hw if first-time use */
        if (pdevInfo->txQueue.pDescQueueTail == NULL) {
            queueTxDescriptor(pdevInfo, pHead, pTrailingDesc);
            halSetTxDP(pdevInfo, pHead->thisPhysPtr);
        } else {
            queueTxDescriptor(pdevInfo, pHead, pTrailingDesc);
        }

        /*
         *  Note this write is not strictly necessary if the sw is
         *  ahead of the hw, and we could defer it until we get
         *  a TXEOL interrupt.  But if we issue it here, we may
         *  in fact avoid that interrupt.
         */
        halStartDmaTransmit(pdevInfo, HAL_TX_QUEUE_DATA);
    }

    if (!callerHoldsLock) {
        A_SEM_UNLOCK(pdevInfo->txQueue.qSem);
    }

    return A_OK;
}

/*******************************************************************************
* ar5hwcEndInt - interrupt service routine
*
* This routine is called at interrupt level in response to an interrupt from
* the controller.  We do the minimal work at interrupt level.   Most
* of the processing is done via netJobAdd().
*/
LOCAL void
ar5hwcEndInt(WLAN_DEV_INFO *pdevInfo)
{
    HAL_INT_TYPE    isrValue, unmaskedIsrValue, disableMask = 0;
    UINT32          msgBuffer[3];
    UINT32          now, delta;

#if defined(PCI_INTERFACE)
    if (!halIsInterruptPending(pdevInfo)) {
        DRV_LOG(DRV_DEBUG_INT,("no interrupt pending.\n"));
        return;
    }
#endif

    /* Find out which interrupts are pending, and deassert the interrupt line */
    isrValue = halGetInterrupts(pdevInfo, &unmaskedIsrValue);

    if (pdevInfo->NicResetInProgress == TRUE) {
        DRV_LOG(DRV_DEBUG_TEMP,("resetting wlan device\n"));
        return;
    }

    ASSERT(isrValue != HAL_INT_NOCARD);

    DRV_LOG(DRV_DEBUG_INT, ("isrValue = 0x%08x, unmaskedIsrValue = 0x%08x\n",
                            isrValue, unmaskedIsrValue));

    if ((isrValue & HAL_INT_RXPHY) && pdevInfo->staConfig.radarDetectEnable) {
        /* radar signal ? */
        if (tsfIndex < TSF_ARRAY_SIZE) {
            halGetTsf(pdevInfo, &tsfStamp[tsfIndex++]);
        }
        if (radarDur) {
            wakeupRadarTxThrottleHandler();
        }
        /*
         * don't record it here. Better not to manipulate events array
         * from interrupt handler, it will cause mutual exclusion problems.
         */
    }

    /*
     * On the ar5211 RXORN is not fatal.  It is acceptable to see
     * this message occasionally, maybe once an hour.  If it is
     * occuring more often than that the system is not giving us
     * enough write bandwidth and there is a system level problem.
     *
     * When this happens, it can repeatedly interrupt until the condition
     * clears.  To handle this cleanly with one message, and w/o spending
     * undue time in the ISR we disable the interrupt and re-enable it in
     * the receive processing job.
     */
    if (isrValue & HAL_INT_RXORN) {
        logMsg("RXORN: RXDP 0x%x (%d)\n", halGetRxDP(pdevInfo),
               ++pdevInfo->rxOrnCnt);
        disableMask |= HAL_INT_RXORN;
    }

    /* For all fatal interrupts, we reset the HW */
    if (isrValue & HAL_INT_FATAL) {
        logMsg("Fatal HW Interrupt: Resetting...\n");
        resetwlan(pdevInfo);
        return;
    }

    /* EOL error handling -- disable receive at the PCU */
    if (isrValue & HAL_INT_RXEOL) {
#ifdef RXEOL_WAR
        apPanic("where from rxEOL?\n");
#endif
        /* rxEOL is non fatal - just keep stats; also prevent it from
         * repeatedly interrupting the processor by diabling RXEOL here
         * and re-enabling in receive processing job
         */
        ++pdevInfo->rxEOLcount;
        disableMask |= HAL_INT_RXEOL;
        DRV_LOG (DRV_DEBUG_INT, ("wlan%d: got rxEOL\n", pdevInfo->devno));
    }

    if (isrValue & HAL_INT_TXURN) {
        pdevInfo->txUrn++;
        now   = A_MS_TICKGET();
        delta = DELTA(now, pdevInfo->txUrnBurstMs); /* timer wraps in 828 days */

        /*  If we get too many in a short time then we reset. */
        if (delta < TXURN_PERIOD_ALLOW) {
            pdevInfo->txUrnBurst++;
            if (pdevInfo->txUrnBurst > TXURN_COUNT_ALLOW) {
                logMsg("%d TXURNs received in a burst. Resetting... TXDP = 0x%x\n",
                       pdevInfo->txUrnBurst, halGetTxDP(pdevInfo));
                pdevInfo->txUrnBurstMs = now;
                pdevInfo->txUrnBurst   = 0;
                resetwlan(pdevInfo);
                return;
            }
        } else {
            pdevInfo->txUrnBurstMs = now;
            pdevInfo->txUrnBurst   = 1;
        }

        /*
         *  Allow a few TXURNs per TXURN_COUNT_ALLOW.  This can
         *  happen sometimes on the 824x platform, and it is ok.
         *  If your platform is seeing this message, the IO system's
         *  performance needs to be debugged.
         */
        if (pdevInfo->txUrnBurst >= 2) {
            logMsg("TXURN #%d (%d) at %d\n", pdevInfo->txUrnBurst,
                   pdevInfo->txUrn, now);
        }
    }

    if (isrValue & HAL_INT_BNR) {
#ifdef DEBUG
        logMsg("BNR (Beacon Not Ready) Interrupt: Informational...\n");
#endif
    }

    if (isrValue & HAL_INT_SWBA) {
        wlanApBeaconGen(pdevInfo);
        beaconNum++;

        if ((pdevInfo->staConfig.calibrationTime != 0) &&
            (beaconNum == calBeaconCount))
        {
            beaconNum    = 0;
            msgBuffer[0] = (UINT32)pdevInfo;
            if (msgQSend(calMsgQId, (char*)&msgBuffer[0], sizeof(msgBuffer), NO_WAIT,
                         MSG_PRI_NORMAL) == ERROR)
            {
                logMsg("Error sending message to perform HW calibration!\n");
            }
        }
    }

    if (isrValue & HAL_INT_RX) {
        DRV_LOG(DRV_DEBUG_INT,("adding processRxInterrupt() job...\n"));
        if (netJobAdd((FUNCPTR)processRxInterrupt, (int)pdevInfo, 0, 0, 0, 0) == OK) {
            disableMask      |= HAL_INT_RX;
            unmaskedIsrValue &= ~HAL_INT_RX;
        } else {
            logMsg("Unable to add processRxInterrupt() job to tNetTask!\n");
        }
    }

    if (isrValue & HAL_INT_TX) {
        DRV_LOG(DRV_DEBUG_INT,("adding processTxInterrupt() job...\n"));
        if (netJobAdd((FUNCPTR)processTxInterrupt, (int)pdevInfo, 0, 0, 0, 0) == OK) {
            disableMask |= HAL_INT_TX;
            unmaskedIsrValue &= ~HAL_INT_TX;
        } else {
            logMsg("Unable to add processTxInterrupt() job to tNetTask!\n");
        }
    }

    if (isrValue & HAL_INT_MIB) {
        /*
         * The on-chip MIB-related counters have overflown.  We read the
         * registers here and add their values to our running totals.
         */
        halUpdateMibCounters(pdevInfo, TRUE);
    }

    if (disableMask) {
        halDisableInterrupts(pdevInfo, disableMask);
    }

    /*
     * To avoid missed interrupts and, therefore, stranded frames, we keep
     * track of TX and RX interrupts that are set in the ISR, but didn't
     * actually assert an interrupt because they are masked.
     */
    pdevInfo->globISRReg |= unmaskedIsrValue & (HAL_INT_RX | HAL_INT_TX);
}

/*******************************************************************************
 * chipReset
 *
 * Reset chip and some selected sw state
 */
void
chipReset(WLAN_DEV_INFO *pdevInfo, A_BOOL unicast, CHAN_VALUES *pChannel)
{
    ATHEROS_DESC    *pDesc;

#ifdef  INCLUDE_SNMPD
    snmpIoTrapSend(LINK_DOWN, 0);
#endif  /* INCLUDE_SNMPD */

    A_SEM_LOCK(pdevInfo->txQueue.qSem, WAIT_FOREVER);

    pdevInfo->globISRReg = 0;

    /* Disable interrupts and stop receive. */
    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    halStopPcuReceive(pdevInfo);

    if (halReset(pdevInfo, pChannel, FALSE) != A_OK) {
        A_SEM_UNLOCK(pdevInfo->txQueue.qSem);
        /* reboot the whole system */
        uiPrintf("chipReset: Error resetting WLAN HW!  Rebooting...\n");
        apReboot();
        return;
    }

    /* do the needful for rates, etc. after switching baseband */
    if ((pChannel->channelFlags & CHANNEL_CCK) ^
        (pdevInfo->staConfig.phwChannel->channelFlags & CHANNEL_CCK))
    {
#ifdef RATECTRL_C
        drvBBChange(pdevInfo, pChannel);
#endif
    }

    /* Let our little light shine... */
    sysLed(1);

    /* Clean up the ps, rx, tx, alt queue */
    A_SEM_LOCK(pdevInfo->localSta->txPSQueue.qSem, WAIT_FOREVER);
    pDesc = (pdevInfo->localSta->txPSQueue).pDescQueueHead;
    if (pDesc != NULL) {
        freeBuffandDescChain(pdevInfo, pDesc);
    }

    pdevInfo->numTxFramesPending = 0;

    pdevInfo->localSta->txPSQueue.pDescQueueHead = NULL;
    pdevInfo->localSta->txPSQueue.pDescQueueTail = NULL;
    A_SEM_UNLOCK(pdevInfo->localSta->txPSQueue.qSem);

    sibTableStateReset(pdevInfo);

    pDesc = (pdevInfo->txQueue).pDescQueueHead;
    if (pDesc != NULL) {
        freeBuffandDescChain(pdevInfo, pDesc);
    }

    pdevInfo->txQueue.pDescQueueHead = NULL;
    pdevInfo->txQueue.pDescQueueTail = NULL;

    pdevInfo->lastTxFrameDesc = NULL;
    A_SEM_UNLOCK(pdevInfo->txQueue.qSem);

    A_SEM_LOCK(pdevInfo->altQueue.qSem, WAIT_FOREVER);
    pDesc = pdevInfo->altQueue.pDescQueueHead;
    if (pDesc != NULL) {
        freeBuffandDescChain(pdevInfo, pDesc);
    }

    pdevInfo->altQueue.pDescQueueHead = NULL;
    pdevInfo->altQueue.pDescQueueTail = NULL;
    A_SEM_UNLOCK(pdevInfo->altQueue.qSem);

    A_PIPEFLUSH();

    /* Reset the receive queue */
    resetRxQueue(pdevInfo);

    /* enable sw encryption if needed */
    if (pdevInfo->staConfig.privacyInvoked) {
        setupSwEncryption(pdevInfo);
    }

    halEnableReceive(pdevInfo);
    halStartPcuReceive(pdevInfo);

    if (unicast) {
        halRxFilter(pdevInfo, HAL_RX_UCAST, HAL_RX_FILTER_SET);
    }

    halEnableInterrupts(pdevInfo, HAL_INT_GLOBAL);

    if (pdevInfo->localSta->staState != STATE_QUIET) {
        pdevInfo->localSta->staState = STATE_AP_UP;
    }

#ifdef  INCLUDE_SNMPD
    snmpIoTrapSend(WARM_START, 0);
    snmpIoTrapSend(LINK_UP, 0);
#endif  /* INCLUDE_SNMPD */

    if (pdevInfo->staConfig.radarDetectEnable) {
        hwRadarInit(pdevInfo);
    }

    /* Set the running flag */
    END_FLAGS_SET(pdevInfo->pOSHandle, IFF_UP | IFF_RUNNING);
}

/*******************************************************************************
* resetOnError - Reset chip and some selected sw state
*
* This routine is called by EndInt to handle the fatal error detected.
* We have to reset the chip.  Also, all the descriptor queues have to be
* reinitialized.  The WEP key cache survives the hardware reset.
* XXX Need to clear some stateful items from the SIB.
*
* RETURNS: OK or ERROR
*/
LOCAL void
resetOnError(WLAN_DEV_INFO *pdevInfo)
{
    /* Keep track of number of resets */
    pdevInfo->localSta->stats.hwResetCount++;
    pdevInfo->hwResetCount++;

#ifdef DEBUG
    logMsg("(%08ld):  Number of resets = %d\n", tickGet(), pdevInfo->hwResetCount);
#endif

    beaconWdgCancel();
    chipReset(pdevInfo, TRUE, pdevInfo->staConfig.pChannel);
    wlanBeaconFree(pdevInfo);
    wlanBeaconInit(pdevInfo, 0);
    beaconWdgStart(pdevInfo);

#ifdef DEBUG
    logMsg("Finished resetting the wlan chip\n");
#endif

    /* Clear the reset flag */
    pdevInfo->NicResetInProgress = FALSE;
}

/*
 *  This routine is run from ISR context.  It queues a reset job
 *  onto the tNetTask and sets the NicResetInProgress flag.  Code that
 *  depends on the state of the sw queues should check this flag
 *  and exit accordingly.
 */
void
resetwlan(WLAN_DEV_INFO *pdevInfo)
{
    /* If things look very bleak, then reboot */
    if (pdevInfo->localSta->stats.hwResetCount > RESET_COUNT_ALLOW) {
        logMsg("resetwlan: Tried to reset too many times (%d). Rebooting...\n",
               pdevInfo->localSta->stats.hwResetCount);
        apReboot();
    }

    if (pdevInfo->NicResetInProgress == TRUE) {
        return;
    }

    enterDebugger();

    /* Queue up a reset job. */
    if (netJobAdd((FUNCPTR)resetOnError, (int)pdevInfo, 0,0,0,0) == OK) {
        pdevInfo->NicResetInProgress = TRUE;
        pdevInfo->localSta->staState = STATE_AP_DOWN;

        halRxFilter(pdevInfo, 0, HAL_RX_FILTER_INIT);
        halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);

        /* Clear the running flag */
        END_FLAGS_CLR (pdevInfo->pOSHandle, IFF_UP | IFF_RUNNING);
    } else {
        logMsg("resetwlan: netJobAdd failed!  Rebooting...\n");

        /* reboot the system */
        apReboot();
    }
}

/*******************************************************************************
* resetRxQueue - reset the receive queue
*
* This routine is called at bootup of the device and chipReset
*
* RETURNS: void
*/
LOCAL void
resetRxQueue(WLAN_DEV_INFO *pdevInfo)
{
    ATHEROS_DESC *pRxDesc;

    A_SEM_LOCK(pdevInfo->rxQueue.qSem, WAIT_FOREVER);

    pdevInfo->partialRxDesc = FALSE;
    pdevInfo->rxEOLcount    = 0;

    pRxDesc = pdevInfo->rxQueue.pDescQueueHead;
    while (pRxDesc) {
        pRxDesc->hwControl.word[0] = 0;
        pRxDesc->hwControl.word[1] = 0;
        pRxDesc->hwStatus.word[0]  = 0;
        pRxDesc->hwStatus.word[1]  = 0;
        pRxDesc->hwControl.rx.bufferLength = AR_BUF_SIZE;
        A_DATA_CACHE_INVAL(pRxDesc->pBufferVirtPtr, AR_BUF_SIZE);
        A_DESC_CACHE_FLUSH(pRxDesc);

#ifdef RXEOL_WAR
        if (pRxDesc == pdevInfo->rxQueue.pDescQueueTail) {
            pRxDesc->nextPhysPtr  = pRxDesc->thisPhysPtr;
            pRxDesc->pNextVirtPtr = pRxDesc;
            A_DESC_CACHE_FLUSH(pRxDesc);
            break;
        }
#endif
        pRxDesc = pRxDesc->pNextVirtPtr;
    }

    A_PIPEFLUSH();

    halSetRxDP(pdevInfo, A_DESC_V2P(pdevInfo->rxQueue.pDescQueueHead));

    A_SEM_UNLOCK(pdevInfo->rxQueue.qSem);
}

/*******************************************************************************
* queueRxDescriptor - queue descriptor to receive queue tail
*
* This routine is only called from processRxInterrupt!
*
* RETURNS: void
*/
LOCAL void
queueRxDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pRxDesc)
{
    QUEUE_DETAILS *pQueue = &pdevInfo->rxQueue;

    pRxDesc->hwControl.word[0] = 0;
    pRxDesc->hwControl.word[1] = 0;
    pRxDesc->hwStatus.word[0]  = 0;
    pRxDesc->hwStatus.word[1]  = 0;

#ifdef RXEOL_WAR
    pRxDesc->nextPhysPtr  = pRxDesc->thisPhysPtr;
    pRxDesc->pNextVirtPtr = pRxDesc;
#else
    pRxDesc->nextPhysPtr  = 0;
    pRxDesc->pNextVirtPtr = NULL;
#endif

    pRxDesc->hwControl.rx.bufferLength = AR_BUF_SIZE;
    A_DATA_CACHE_INVAL(pRxDesc->pBufferVirtPtr, AR_BUF_SIZE);
    A_DESC_CACHE_FLUSH(pRxDesc);

    ASSERT(pQueue->pDescQueueTail != NULL);
    pQueue->pDescQueueTail->pNextVirtPtr = pRxDesc;
    pQueue->pDescQueueTail->nextPhysPtr  = pRxDesc->thisPhysPtr;
    A_DESC_CACHE_FLUSH(pQueue->pDescQueueTail);

    pQueue->pDescQueueTail = pRxDesc;

    A_PIPEFLUSH();
}

/*******************************************************************************
* processRxInterrupt - interrupt service routine
*
* This routine is queued as a job by the ISR to handle the received interrupt.
* It sends the good frames to DS service routine. For the error ones, it frees
* the descriptors and buffers.
*
* RETURNS: OK or ERROR
*/
LOCAL void
processRxInterrupt(WLAN_DEV_INFO *pdevInfo)
{
    WLAN_FRAME_HEADER   *pWlanHdr;
    WLAN_MACADDR        *pSrcStaMacAddr;
    HAL_DESC_RX_STATUS  *pRxStatus;
    ATHEROS_DESC        *pRxDesc, *pTailDesc;
    SIB_ENTRY           *pSib;
    A_BOOL              doneEarly = FALSE;
    A_UINT32            foundPhyErr = 0;
    A_STATUS            status;
    A_UINT32            antRx;
    WLAN_STATS          *pLocalStats = &pdevInfo->localSta->stats;
#if AR_RX_RESCHED > 0
    int                 rxCountdown = AR_RX_RESCHED;
#endif

    HISTOGRAM_START_COUNTER(arRx);

    A_SEM_LOCK(pdevInfo->rxQueue.qSem, WAIT_FOREVER);

    /* update global MIB items XXX these should move to a background task */
    halUpdateMibCounters(pdevInfo, FALSE);

    while (pdevInfo->localSta->staState & STATE_AP_UP) {
        ASSERT(pdevInfo->rxQueue.pDescQueueHead);
        ASSERT(pdevInfo->rxQueue.pDescQueueHead != pdevInfo->rxQueue.pDescQueueTail);

        pRxDesc = pdevInfo->rxQueue.pDescQueueHead;
        A_DESC_CACHE_INVAL(pRxDesc);
        pRxStatus = &pRxDesc->hwStatus.rx;

        if (!pRxStatus->done) {
            break;
        }

#ifdef RXEOL_WAR
        /* given the use of a self-linked tail be very sure that the hw is
         * done with this descriptor; the hw may have done this descriptor
         * once and picked it up again.. make sure the hw has moved on
         */
        A_DESC_CACHE_INVAL(pRxDesc->pNextVirtPtr);
        if ( (!pRxDesc->pNextVirtPtr->hwStatus.rx.done) &&
             (halGetRxDP(pdevInfo) == (A_UINT32)pRxDesc) )
        {
            break;
        }
#endif

#if AR_RX_RESCHED > 0
        /*
        *  RX loop to relinquish and reschedule after processing
        *  AR_RX_RESCHED descriptors.
        */
        if (0 == rxCountdown--) {
            doneEarly = TRUE;
            break;
        }
#endif

        pRxDesc->decryptError = FALSE;          /* assume frame is ok */

        /*
         * Drop multiple-descriptor frames.  This is not WECA-compliant if our
         * buffer size (sans wlan header size) is larger than 1514 bytes.
         */
        if (pRxStatus->more || pdevInfo->partialRxDesc) {
            DRV_LOG(DRV_DEBUG_RX, ("rx frame too long!\n"));

            /* capture if the next one has to be discarded */
            pdevInfo->partialRxDesc = pRxStatus->more ? TRUE : FALSE;

            /* discard this one */
            pdevInfo->rxQueue.pDescQueueHead = pRxDesc->pNextVirtPtr;
            queueRxDescriptor(pdevInfo, pRxDesc);

            pLocalStats->RxInvalidFrames++;
            continue;
        }
#ifdef DEBUG_PKT_LOG
	{
	    /*
	     * Log receive packet info
	     */
	    extern void logEventRxPkt(A_UINT32 length,
				       A_UINT32 rate,
				       A_UINT32 rssi,
				       A_UINT32 frameType,
				       A_UINT32 seqNum,
				       A_UINT32 status,
				       A_UINT32 antRx,
				       A_UINT16 timeStamp);
	    logEventRxPkt(pRxStatus->dataLength,
		  pRxStatus->rxRate,
		  pRxStatus->rxSigStrength,
		  *(A_UINT16*)&pRxDesc->pBufferVirtPtr.header3->frameControl,
		  *(A_UINT16*)&pRxDesc->pBufferVirtPtr.header3->sequenceControl,
		  pRxStatus->phyError << 4
		| pRxStatus->decryptCRCError << 3
		| pRxStatus->fifoOverrun << 2
		| pRxStatus->CRCError << 1
		| pRxStatus->pktReceivedOK << 0,
		  pRxStatus->hwSpecific0,
		  (A_UINT16)((pRxStatus->hwSpecific2 >> 7) & 0x7fff));
	}
#endif /* DEBUG_PKT_LOG */

        if (!pRxStatus->pktReceivedOK) {
            /* do we *have* to maintain two counters? */
            END_ERR_ADD(pdevInfo->pOSHandle, MIB2_IN_ERRS, +1);
            pLocalStats->ReceiveErrors++;

            /*
             * which of all these counters do we ever use? these
             * and other stats counters should be macroized
             */
            if (pRxStatus->fifoOverrun) {
                pLocalStats->RcvDmaOverrunErrors++;
            } else if (pRxStatus->phyError) {
                pLocalStats->RcvPhyErrors++;
                phyErrorCounts[(pRxStatus->phyError & 0x7)]++;
                if (pdevInfo->staConfig.radarDetectEnable && (tsfIndex > 0)) {
                    int i;
                    A_SEM_LOCK(radarDataSemId, WAIT_FOREVER);
                    for (i = 0; i < tsfIndex; i++) {
                        recordRadarEvent(tsfStamp[i].low, tsfStamp[i].high);
                    }
                    tsfIndex = 0;
                    foundPhyErr++;
                    A_SEM_UNLOCK(radarDataSemId);
                }
            } else if (pRxStatus->CRCError) {
                pLocalStats->RcvCrcErrors++;
            } else if (pRxStatus->decryptCRCError) {
                A_BOOL faulted = FALSE;

                pWlanHdr = pRxDesc->pBufferVirtPtr.header;
                /* looking at address2 only - don't pay the byte swap cost */
                pSrcStaMacAddr = &pWlanHdr->address2;
                pSib = sibEntryFind(pSrcStaMacAddr);
                if (pSib && (pRxStatus->keyIndexValid == 0)) {
                    pLocalStats->RcvKeyCacheMisses++;
                    pSib->stats.RcvKeyCacheMisses++;
                    (void)wlanKeyCacheFault(pdevInfo, pSib, &faulted);
                }
                if (faulted == FALSE) {
                    pLocalStats->RcvDecryptCrcErrors++;
                    if (pSib) {
                        pSib->stats.RcvDecryptCrcErrors++;
                    }
                    /*
                     * Decrypt errors still get sent to the mac layer
                     * to deal with appropriately.
                     */
                    pRxDesc->decryptError = TRUE;       /* frame is corrupt */
                }
            }
            eventPost(SUBSYS_WLANRX, EV_RX_ERR, "RX: error", pRxDesc,
                      pRxDesc->hwStatus.word[0], pRxDesc->hwStatus.word[1]);

            if (pRxDesc->decryptError == FALSE) {
                /* Take desc off the recv queue, and requeue */
                pdevInfo->rxQueue.pDescQueueHead = pRxDesc->pNextVirtPtr;
                queueRxDescriptor(pdevInfo, pRxDesc);

                continue;
            }
        }

        /*
         * no point reaping frames that are going to be discarded down
         * the forwarding path due to lack of resources... therefore,
         * throttle the receive processing based on if our own tx or
         * the ethernet's tx is going to drop the frame! since our own
         * tx is throttled by availability of descriptors on the empty
         * queue (and/or avail of tx header buffers really.. but) - an
         * attempt to get a descriptor below will achieve the desired
         * effect;  ideally availability of resources on ethernet side
         * should get throttled by loaned buffers returning to us and
         * (non)-availability of a buffer below - but currently we're
         * stumped by lack of proactive buffer release on tx completion
         * by the ethernet driver; we can live with that for now - as
         * for the common case we're the slower interface
         */
        if (createBuffandDesc(pdevInfo, AR_BUF_SIZE, &pTailDesc) != OK) {
            doneEarly = TRUE;
            break;
        }

        /* take this off, and queue the new one */
        pdevInfo->rxQueue.pDescQueueHead = pRxDesc->pNextVirtPtr;
        pRxDesc->nextPhysPtr             = 0;
        pRxDesc->pNextVirtPtr            = NULL;
        queueRxDescriptor(pdevInfo, pTailDesc);

        HISTOGRAM_COUNT(arRx);

        pWlanHdr = pRxDesc->pBufferVirtPtr.header;
        FRAME_CONVERT_LETOCPU(pWlanHdr, pRxStatus->dataLength);
        pSrcStaMacAddr        = &pWlanHdr->address2;
        pSib                  = sibEntryFind(pSrcStaMacAddr);
        pRxDesc->pSrcSibEntry = pSib;

        /* save RX WEP key in an easy place key for wlanFrameReceive() */
        pRxDesc->rxKeyIndex = halGetRxDescKeyIndex(pdevInfo, pRxDesc);

        /* subtract off ICV and do sw encryption if needed */
        status = setupDecrypt(pdevInfo, pRxDesc);
        if (status != A_OK) {
            pLocalStats->ReceiveErrors++;
            if (pSib) {
                pSib->stats.ReceiveErrors++;
            }

            if (status == A_DECRYPT_ERROR) {
                /* Frame is corrupt but pass up to the MAC to handle */
                pRxDesc->decryptError = TRUE;
            } else {
                freeBuffandDesc(pdevInfo, pRxDesc);   /* drop frame */
                continue;
            }
        }

        /* subtract off FCS */
        pRxDesc->hwStatus.rx.dataLength -= FCS_FIELD_SIZE;

        if (pRxDesc->decryptError == FALSE) {
            /* do we use *all* these stats; macroize them someday */
            END_ERR_ADD(pdevInfo->pOSHandle, MIB2_IN_UCAST, +1);
            pLocalStats->GoodReceives++;
            /* rxRate is in PLCP format. */
            pLocalStats->RcvRate =
                pdevInfo->staConfig.pRateTable->plcpRateToIndex[pRxStatus->rxRate];
            /* RSSI can be negative. Treat 8 bit value as signed. */
            pLocalStats->RcvSignalStrength = (A_INT32)((A_INT8)pRxStatus->rxSigStrength);
            antRx = halGetRxDescRxAntenna(pdevInfo, pRxDesc);
            pLocalStats->RxAntenna = antRx;
            pLocalStats->AntCnt[pLocalStats->RxAntenna&1]++;
            eventPost(SUBSYS_WLANRX, EV_RX_OK, "RX: ok", pRxDesc,
                      pRxDesc->hwStatus.word[0], pRxDesc->hwStatus.word[1]);

            if (pSib != NULL) {
                TRACEBUF_LOG_START(Rx, pSib);
                TRACEBUF_LOG_DONE(Rx, pSib, _desc == NULL,
                    _powersave = pRxDesc->pBufferVirtPtr.header->frameControl.pwrMgt);

                pSib->stats.RcvSignalStrength = (A_INT32)((A_INT8)pRxStatus->rxSigStrength);
                pSib->stats.RcvRate           = pdevInfo->staConfig.pRateTable->plcpRateToIndex[pRxStatus->rxRate];
                pSib->stats.RxAntenna         = antRx;
                pSib->stats.GoodReceives++;
            }

#ifdef DEBUG
            if (rxVerbose >= 10) {
                dumpFrame(pRxDesc, "rx");
            }
#endif
        }

        /* good + decrypt error frames are sent to the mac/DS */
        wlanFrameReceive(pdevInfo, pRxDesc);
    }

    HISTOGRAM_BIN(arRx, 0);

    A_SEM_UNLOCK(pdevInfo->rxQueue.qSem);

    /*
     * Check to see if there was another RX interrupt recorded in the ISR while
     * handling a non-masked interrupt source.  This will avoid leaving
     * descriptors "stranded" on the queue waiting for another received frame
     * to kickstart their completion processing.  We queue another job if we
     * aborted processing descriptors early, or the ISR read another HAL_INT_RX
     * while RX ints were masked.
     */
    intDisable(arParam.ilevel);
    if ((pdevInfo->globISRReg & HAL_INT_RX) || doneEarly) {
        if (netJobAdd((FUNCPTR)processRxInterrupt, (int)pdevInfo, 0, 0, 0, 0) != OK) {
            logMsg("processRxInterrupt: netJobAdd() call failed!\n");
        }
        pdevInfo->globISRReg &= ~HAL_INT_RX;
    } else {
        /* Re-enable RXORN|RXEOL, as well, in case we disabled it in the ISR */
        halEnableInterrupts(pdevInfo, HAL_INT_RX | HAL_INT_RXORN | HAL_INT_RXEOL);
    }
    intEnable(arParam.ilevel);

    if (foundPhyErr) {
        wakeupRadarEventHandler();      /* Offload the processing of radar events */
    }
}

/*
 * processTxSibInfo:
 *
 * This routine is called by processInt to handle the transmitted interrupt.
 * It updates the per sib counters, handles rate adaptation feedback, and
 * forwards MGT frams the the SME code via mlmeMgtTxCompletion().
 */
LOCAL INLINE int
processTxSibInfo(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pTxDesc)
{
    ATHEROS_DESC            *pFirst     = pTxDesc->pTxFirstDesc;
    SIB_ENTRY               *pSib       = pFirst->pDestSibEntry;
    HAL_DESC_TX_STATUS      *pTxStatus  = &pTxDesc->hwStatus.tx;
    HAL_DESC_TX_CONTROL     *pTxControl = &pFirst->hwControl.tx;
    int                     txRate      = pTxControl->transmitRate;


    /* Update per-SIB stats */

    if (pSib == NULL) {
        return 1;
    }

    if (pTxStatus->pktTransmitOK) {
        TRACEBUF_LOG_DONE(Tx, pSib, _desc == pFirst, _completionCode = 1);

        pSib->stats.GoodTransmits++;
        pSib->stats.CurrentAckSignalStrenth =
            pdevInfo->localSta->stats.CurrentAckSignalStrenth;
        /*
         * Update stats if there were any retries for the good frame.
         *
         * If the frame saw excessive retries, do not change the antenna.
         * This matches Oahu, and is ok for Crete unless the HW retry count
         * is set to less than three.
         *
         * Toggle antenna setting in SIB.  The hardware switches the
         * antennas every two transmit failures.  This is for the 2nd retry
         * (3rd transmit),  3th retry (4th transmit), etc.
         *
         * For excessive retries keep the antenna the same.
         */
        if (pTxStatus->shortRetryCount || pTxStatus->longRetryCount) {
            int retries;

            pSib->stats.TotalRetries++;
            if (pTxStatus->shortRetryCount) {
                pSib->stats.shortFrameRetryBins[pTxStatus->shortRetryCount]++;
                retries = pTxStatus->shortRetryCount & 3;
            } else { /* if (pTxStatus->longRetryCount) */
                pSib->stats.RetryBins[pTxStatus->longRetryCount]++;
                retries = pTxStatus->longRetryCount & 3;
            }
            if (retries >= 2) {
                pSib->antTx ^= 1;               /* flip */
                pSib->stats.AntSwCnt++;
            }
        }
        /* update the tx antenna with what with actual antenna used */
        pSib->stats.AntCnt[pSib->antTx]++;
    } else {
        pSib->stats.TransmitErrors++;

        if (pTxStatus->excessiveRetries) {
            pSib->stats.TxExcessiveRetries++;
        } else if (pTxStatus->filtered) {
            pSib->stats.TxFiltered++;
        }

        if (pTxStatus->fifoUnderrun) {
            pSib->stats.TxDmaUnderrun++;
        }
    }

    if ( txRate && !pTxStatus->filtered ) {
	    int antTx = pTxDesc->hwControl.tx.hwSpecific0;
	    A_UINT8 flipReq = 0;
        RateCtrlTxRetryCntUpdate(
             pdevInfo->staConfig.pRateTable->plcpRateToIndex[txRate],
		     pTxStatus->excessiveRetries,
		     pTxStatus->shortRetryCount | pTxStatus->longRetryCount,
		     pTxControl->frameLength, pSib,
		     pTxStatus->ackSigStrength,
		     antTx,
		     &flipReq
		 );
    }

    /* Look for management packets
     * only call mlmeMgtTxCompletion() after all retries are done
     */
    if ((pFirst->pBufferVirtPtr.headerLittle->frameControl.fType == FRAME_MGT) &&
        (!pdevInfo->staConfig.swretryEnabled ||
        (pTxStatus->pktTransmitOK ||
        (pFirst->swretryCount == pdevInfo->staConfig.swRetryMaxRetries)))){
        ASSERT(pFirst == pTxDesc); // mgmt frames are single desc

        /* Put mgmt frames back in host order for processing */
        FRAME_CONVERT_LETOCPU(pFirst->pBufferVirtPtr.header,
                              pFirst->hwControl.tx.bufferLength);
        mlmeMgtTxCompletion(pdevInfo, pFirst);
        // perform pending descr processing. It isn't done by mlmeMgtTxCompletion anymore.
        smePendDescProcess(pdevInfo, pSib);

        /* Put mgmt frames back in network order for sw retry processing */
        FRAME_CONVERT_CPUTOLE(pFirst->pBufferVirtPtr.header,
                              pFirst->hwControl.tx.bufferLength);
    }

    return 1;
}


/*******************************************************************************
* queueTxDescriptor - queue descriptor to transmit queue tail
*
* RETURNS: void
*/
LOCAL void
queueTxDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, ATHEROS_DESC *pTail)
{
    QUEUE_DETAILS *pQueue = &pdevInfo->txQueue;

    ASSERT(pTail->pNextVirtPtr == NULL);

    if (pQueue->pDescQueueTail == NULL) {
        ASSERT(pQueue->pDescQueueHead == NULL);
        pQueue->pDescQueueHead = pDesc;
    } else {
        pQueue->pDescQueueTail->nextPhysPtr  = pDesc->thisPhysPtr;
        pQueue->pDescQueueTail->pNextVirtPtr = pDesc;
#ifdef PERF_DEBUG
        A_DESC_CACHE_INVAL(pQueue->pDescQueueTail);
        if (pQueue->pDescQueueTail->hwStatus.tx.done) {
            ++onEmptyQueue;
        } else {
            ++onFullQueue;
        }
#endif
        A_DESC_CACHE_FLUSH(pQueue->pDescQueueTail);
    }
    pQueue->pDescQueueTail = pTail;

#ifdef SWCOHERNECYWB
    /* skip loop for writethru cache */
    while (pDesc) {
        A_DATA_CACHE_FLUSH(pDesc->pBufferVirtPtr, pDesc->hwControl.tx.bufferLength);
        A_DESC_CACHE_FLUSH(pDesc);
        pDesc = pDesc->pNextVirtPtr;
    }
#endif
    A_PIPEFLUSH();

    return;
}

/*******************************************************************************
* processTxQueue/processTxInterrupt - interrupt service routines
*
* This routine is called by processInt to handle the transmitted interrupt.  It
* releases the tx buffers to the buffer pool.  It also releases the descriptors
* to the tx queue.
*/
void
processTxQueue(WLAN_DEV_INFO *pdevInfo)
{
    HAL_DESC_TX_STATUS  *pTxStatus;
    ATHEROS_DESC        *pTxDesc;
    ATHEROS_DESC        *pPreLastDesc = NULL;
    WLAN_STATS          *pLocalStats  = &pdevInfo->localSta->stats;

    HISTOGRAM_START_COUNTER(arTx);

    /* Lock the tx queue */
    DRV_LOG(DRV_DEBUG_INT,("acquiring the tx queue lock\n"));
    A_SEM_LOCK(pdevInfo->txQueue.qSem, WAIT_FOREVER);

    pTxDesc = pdevInfo->txQueue.pDescQueueHead;

    /*
     * We iterate over all the descriptors.  However, only the first descriptor
     * of a frame contains the control information and only the last descriptor
     * of a frame contains the status information information.  Furthermore,
     * when the hardware runs out of tx descs, hw will expect to use the final
     * desc as a springboard to get to a (newly-chained) desc when it is finally
     * restarted; so software must not free this descriptor.  When we fully
     * process a descriptor, it is marked stale so we can see our own
     * breadcrumbs and skip over that desc the next time thru the loop.
     */
    while (pTxDesc) {
#ifdef DEBUG
        if (txVerbose == 1) {
            A_DESC_CACHE_INVAL(pTxDesc);
            uiPrintf("tx isr -- Tx desc: %p, more: %d, done: %d\n",
                     pTxDesc, pTxDesc->hwControl.tx.more,
                     pTxDesc->hwStatus.tx.done);
        }
#endif
        eventPost(SUBSYS_WLANTX, EV_TX_TXDITER, "TXintr:",
                  pTxDesc, pTxDesc->hwControl.tx.more,
                  pTxDesc->staleFlag);

        /* only process complete frames we've not touched before */
        if (pTxDesc->hwControl.tx.more || pTxDesc->staleFlag) {
            /* Go to the next descriptor in the same frame */
            pPreLastDesc = pTxDesc;
            pTxDesc      = pTxDesc->pNextVirtPtr;
            continue;
        }

        /* ensure we have the status correctly */
        A_DESC_CACHE_INVAL(pTxDesc);
        pTxStatus = &pTxDesc->hwStatus.tx;

        ar5311TxDoneWAR(pTxDesc);               /* bug #3403 */
        if (!pTxStatus->done) {
            /* whole frame has not finished transmit */
            if (lastDescGet(pdevInfo->txQueue.pDescQueueTail)->hwStatus.tx.done) {
                A_DESC_CACHE_INVAL(pTxDesc);
                if (!pTxStatus->done) {
                    apPanic("Transmit queue in inconsistent state.");
                }
            }
            break;
        }

        HISTOGRAM_COUNT(arTx);

#ifdef PERF_DEBUG
{
        static A_UINT16 lastSend = 0;

        if (pTxStatus->pktTransmitOK && (!pTxStatus->shortRetryCount)) {
            A_UINT16 thisDiff = (A_UINT16)pTxStatus->sendTimestamp - lastSend;
            HISTOGRAM_BIN_VALUE(arTxDelta, 0, thisDiff);
        }
        lastSend = pTxStatus->sendTimestamp;
}
#endif
#ifdef DEBUG_PKT_LOG
	{
	    WLAN_DATA_MAC_HEADER3 *pWlanHdr = pTxDesc->pBufferVirtPtr.header3;
	    /*
	     * Log transmit complete event
	     */
	    extern void logEventTxPkt(A_UINT32 length, A_UINT32 rate,
				      A_UINT32 ackRssi, A_UINT32 retries,
				      A_UINT32 status,
				      A_UINT32 frameCtrl,
				      A_UINT32 seqCtrl,
				      A_UINT32 antTx,
				      A_UINT16 timeStamp);

	    int antTx = pTxDesc->hwControl.tx.hwSpecific0;
	    if ( antTx == 0 ) {
		/*
		 * Get the last tx antenna from the key cache
		 */
		int keyIdx = pTxDesc->hwControl.tx.hwSpecific3 & 0x7f;
		antTx = (readPlatformReg(pdevInfo, 0x8814 + 32 * keyIdx) >> 3)
				& 0x1;
		antTx = antTx | keyIdx << 1;
	    }

	    logEventTxPkt(pTxDesc->hwControl.tx.frameLength,
			pTxDesc->hwControl.tx.transmitRate,
			pTxStatus->ackSigStrength,
			pTxStatus->shortRetryCount | pTxStatus->longRetryCount,
			pTxStatus->filtered << 3
		      | pTxStatus->fifoUnderrun << 2
		      | pTxStatus->excessiveRetries << 1
		      | pTxStatus->pktTransmitOK << 0,
			*(A_UINT16*)&pWlanHdr->frameControl,
			*(A_UINT16*)&pWlanHdr->sequenceControl,
			antTx,
			(A_UINT16)pTxStatus->sendTimestamp
	    );
	}
#endif

        /* finished transmit of the frame? */
        if (pTxStatus->pktTransmitOK) {
            DRV_LOG(DRV_DEBUG_INT,("pkt Transmit OK\n"));

            /* Update the successful send statistics */
            END_ERR_ADD(pdevInfo->pOSHandle, MIB2_OUT_UCAST, +1);
            pLocalStats->GoodTransmits++;
            pLocalStats->CurrentAckSignalStrenth = (A_INT32)pTxStatus->ackSigStrength;

            if (pTxStatus->shortRetryCount) {
                pLocalStats->TotalRetries++;
                pLocalStats->shortFrameRetryBins[pTxStatus->shortRetryCount]++;
            } else if (pTxStatus->longRetryCount) {
                pLocalStats->TotalRetries++;
                pLocalStats->RetryBins[pTxStatus->longRetryCount]++;
            }
        } else {
            /*
             *  Before we collect stats, make sure we'll
             *  be able to fully process the error.
             */
            if (swretryLookAhead(pdevInfo, pTxDesc) != 0) {
                break;
            }

            /* Update the error send statistics */
            END_ERR_ADD(pdevInfo->pOSHandle, MIB2_OUT_ERRS, +1);
            pLocalStats->TransmitErrors++;

            if (pTxStatus->excessiveRetries) {
                DRV_LOG(DRV_DEBUG_INT, ("Transmit excessive retries\n"));
                pLocalStats->TxExcessiveRetries++;
            } else if (pTxStatus->filtered) {
                DRV_LOG(DRV_DEBUG_INT,("Transmit filtered\n"));
                pLocalStats->TxFiltered++;
            } else if (pTxStatus->fifoUnderrun) {
                DRV_LOG(DRV_DEBUG_INT,("Transmit fifo underrun\n"));
                pLocalStats->TxDmaUnderrun++;
            } else {
                apPanic("Unknown transmit error.");
            }
        }

        /*
         * Collect per-SIB Tx stats and/or Tx
         * completion for management frames
         */
        ASSERT(pTxDesc->pTxFirstDesc != NULL);
        processTxSibInfo(pdevInfo, pTxDesc);

        if (pdevInfo->staConfig.swretryEnabled &&
            swretry(pdevInfo, pPreLastDesc, pTxDesc) != 0)
        {
            /*
             *  swretry() has queued up more stuff. Reset the
             *  current TxDesc ptr to head of queue and loop again.
             *  We've also lost a useful setting for pPreLastDesc,
             *  so reset it and tell kick hw just in case it had
             *  gotten to the end of queue already (quite likely).
             */
            pTxDesc = pdevInfo->txQueue.pDescQueueHead;
            pPreLastDesc = NULL;
            halStartDmaTransmit(pdevInfo, HAL_TX_QUEUE_DATA);
            continue;
        }

        txFramePendDecrement(pdevInfo, pTxDesc->pTxFirstDesc->pDestSibEntry,
                             pTxDesc->pTxFirstDesc);

        if (pPreLastDesc != NULL) {
            /* terminate the to-be-released descriptor chain */
            pPreLastDesc->nextPhysPtr  = 0;
            pPreLastDesc->pNextVirtPtr = NULL;
            freeBuffandDescChain(pdevInfo, pdevInfo->txQueue.pDescQueueHead);
        }

        /* Advance the trailing descriptor */
        pPreLastDesc = pTxDesc;

        /* Head is now this desc, in case the next is NULL */
        pdevInfo->txQueue.pDescQueueHead = pTxDesc;
        pdevInfo->txQueue.pDescQueueHead->pTxFirstDesc = pTxDesc;
        pTxDesc->staleFlag = 1;

        /* finally advance pTxDesc to next descriptor */
        pTxDesc = pTxDesc->pNextVirtPtr;

    } /* while (pTxDesc != NULL) */

    DRV_LOG(DRV_DEBUG_INT,("release the tx queue lock\n"));

    HISTOGRAM_BIN(arTx, 0);

    A_SEM_UNLOCK(pdevInfo->txQueue.qSem);
}

LOCAL void
processTxInterrupt(WLAN_DEV_INFO *pdevInfo)
{
    /* XXX - why check to see if AP is up here? */
    if (pdevInfo->localSta->staState & STATE_AP_UP) {
        processAltQueue(pdevInfo);
    }

    processTxQueue(pdevInfo);

    /* Check to see if there was another TX interrupt recorded in the ISR while
     * handling a non-masked interrupt source.  This will avoid leaving
     * descriptors "stranded" on the queue waiting for another transmitted
     * frame to kickstart their completion processing.  If we aborted processing
     * descriptors early (to be fair to other jobs running in the task) or if
     * the ISR recorded a TX interrupt while TX interrupts were masked, we queue
     * up another job to run later.
     */
    intDisable(arParam.ilevel);
    if (pdevInfo->globISRReg & HAL_INT_TX) {
        if (netJobAdd((FUNCPTR)processTxInterrupt, (int)pdevInfo, 0, 0, 0, 0) != OK) {
            logMsg("processTxInterrupt: netJobAdd() call failed!\n");
        }
        pdevInfo->globISRReg &= ~HAL_INT_TX;
    } else {
        halEnableInterrupts(pdevInfo, HAL_INT_TX);
    }
    intEnable(arParam.ilevel);
}

/*******************************************************************************
* processAltQueue - altQ process routine
*
* This routine is called by wlanBeaconGen to handle the transmitted frames in
* AltQ.  It releases the buffer and descriptors back to the right queues.
*/
void
processAltQueue(WLAN_DEV_INFO *pdevInfo)
{
    HAL_DESC_TX_STATUS  *pAltStatus;
    ATHEROS_DESC        *pAltDesc;
    ATHEROS_DESC        *pAltHeadDesc;

    pAltDesc = pdevInfo->altQueue.pDescQueueHead;

    DRV_LOG(DRV_DEBUG_INT, ("starting to process alt queue\n"));

    while (pAltDesc != NULL) {
        DRV_LOG(DRV_DEBUG_INT,("  walking alt queue\n"));
        pAltHeadDesc = pAltDesc;
        pAltDesc     = lastDescGet(pAltDesc);
        pAltStatus   = &pAltDesc->hwStatus.tx;

        if (pAltStatus->done == 0) {
            break;
        }

        /* update the alt queue */
        pdevInfo->altQueue.pDescQueueHead = pAltDesc->pNextVirtPtr;

        DRV_LOG(DRV_DEBUG_INT,("  alt queue frame done\n"));

        if (pAltStatus->pktTransmitOK) {
            DRV_LOG(DRV_DEBUG_INT,("    frame status = ok\n"));

            /* Update the correct sent statistics */
            END_ERR_ADD(pdevInfo->pOSHandle, MIB2_OUT_UCAST, +1);
            pdevInfo->localSta->stats.GoodTransmits++;
        } else {
            DRV_LOG(DRV_DEBUG_INT,("    frame status = not ok!\n"));
            /* Update the incorrect sent statistics */
            END_ERR_ADD(pdevInfo->pOSHandle, MIB2_OUT_ERRS, +1);

            if (pAltStatus->fifoUnderrun) {
                DRV_LOG(DRV_DEBUG_INT,("    frame status = fifo underrun!\n"));
                pdevInfo->localSta->stats.TxDmaUnderrun++;
            }
        }

        /* terminate the current frame */
        pAltDesc->nextPhysPtr  = 0;
        pAltDesc->pNextVirtPtr = NULL;

        /* free the done frames */
        freeBuffandDescChain(pdevInfo, pAltHeadDesc);

        /* next frame */
        pAltDesc = pdevInfo->altQueue.pDescQueueHead;

    } /* while loop to walk through the alt queue */

    /* disable int to handle race condition with swba ISR */
    intDisable(arParam.ilevel);
    if (pdevInfo->altQueue.pDescQueueHead == NULL) {
        pdevInfo->altQueue.pDescQueueTail = NULL;
    }
    intEnable(arParam.ilevel);

    DRV_LOG(DRV_DEBUG_INT,("done processing alt queue\n"));
}

/*******************************************************************************
* ar5hwcEndIoctl - the driver I/O control routine
*
* Process an ioctl request.
*
* RETURNS: A command specific response, usually OK or ERROR.
*/
LOCAL int
ar5hwcEndIoctl(END_OBJ *pEnd, int cmd, caddr_t data)
{
    WLAN_DEV_INFO   *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;
    int             error = 0;
    long            value;

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    DRV_LOG(DRV_DEBUG_IOCTL,("ioctl CMD=0x%x\n", cmd));

    switch (cmd) {

    case EIOCSADDR:
        /* XXX this really should dribble into pdevInfo's copies? */
        if (data == NULL) {
            return EINVAL;
        }
        bcopy(data, END_HADDR(pEnd), END_HADDR_LEN(pEnd));
        break;

    case EIOCGADDR:
        if (data == NULL) {
            return EINVAL;
        }
        bcopy(END_HADDR(pEnd), data, END_HADDR_LEN(pEnd));
        break;

    case EIOCSFLAGS:
        value = (long)data;
        if (value < 0) {
            value = -(--value);
            END_FLAGS_CLR(pEnd, value);
        } else {
            END_FLAGS_SET(pEnd, value);
        }
        ar5RxFilterConfig(pdevInfo);
        break;

    case EIOCGFLAGS:
        *(int *)data = END_FLAGS_GET(pEnd);
        break;

    case EIOCPOLLSTART: /* Begin polled operation */
        ar5hwcEndPollStart(pEnd);
        break;

    case EIOCPOLLSTOP:  /* End polled operation */
        ar5hwcEndPollStop(pEnd);
        break;

    case EIOCGMIB2:     /* return MIB information */
        if (data == NULL) {
            return EINVAL;
        }
        bcopy((char *)&pEnd->mib2Tbl, data, sizeof(pEnd->mib2Tbl));
        break;

    case EIOCMULTIADD:
        error = ar5hwcEndMCastAdd(pEnd, data);
        break;

    case EIOCMULTIDEL:
        error = ar5hwcEndMCastDel(pEnd, data);
        break;

    case EIOCMULTIGET:
        error = ar5hwcEndMCastGet(pEnd, (MULTI_TABLE *)data);
        break;

    case EIOCGNAME:
        strcpy((char *)data, AR_DEV_NAME);
        break;

    case EIOCBSSSTART:
        if (pdevInfo->localSta->staState == STATE_AP_UP) {
            if (!pdevInfo->pBeaconInfo->beaconEnabled) {
                resetwlan(pdevInfo);
            }
        } else {
            /*
             *  This should be done for all BSPs.
             */
            intEnable(arParam.ilevel);

            /* Enable Rx */
            halStartPcuReceive(pdevInfo);
            halEnableReceive(pdevInfo);

            /* Allow DMA of RX now that we're ready to rock'n'roll */
            halRxFilter(pdevInfo, HAL_RX_UCAST, HAL_RX_FILTER_INIT);

            /* Enable global interrupts */
            halEnableInterrupts(pdevInfo, HAL_INT_GLOBAL);

            sysLed(1);      /* transition to solid LED */
        }
        break;

    case EIOCSKEY:
    {
    struct ioctlKeyDesc_s *iocKeyDesc = (struct ioctlKeyDesc_s *)data;
    SIB_ENTRY *pSib;

    /* install the newly derived key in AP hw */

    ASSERT(iocKeyDesc->usage == KEYDESC_USAGE_UNICAST);

    pSib = sibEntryFind(iocKeyDesc->macAddr);
    if (wlanKeyTableAddDynamic(pdevInfo, pSib,
                   PRIV_KEY_TYPE_WEP, // We'll reset below
                   iocKeyDesc->keyLength,
                   iocKeyDesc->secret) == FALSE) {
        uiPrintf("Could not add key to hw.\n");
        /* XXX key update failed...send NULL keymsg to STA? */
        return EINVAL;
    }

    if (wlanKeyUpdate(pdevInfo, pSib, CONTEXT_DOT1X) != A_OK)
    {
        return EINVAL;
    }
    bzero((char *)iocKeyDesc, sizeof(*iocKeyDesc));
    /* Ship the length back to user so EAPOL-msg can use correct key */
    iocKeyDesc->keyLength = pSib->pPriv->keyLength;
    ASSERT(iocKeyDesc->keyLength != 0);
    }

    break;

    default:
        error = EINVAL;
    }

    return error;
}

/*******************************************************************************
* ar5hwcPollRecv - routine to receive a frame in polled mode.
*
* This routine is called by a user to try and get a frame from the
* device.
*
* RETURNS: OK upon success.  EAGAIN is returned when no frame is available.
*/
LOCAL STATUS
ar5hwcEndPollRecv(END_OBJ *pEnd, M_BLK_ID pMblk)
{
    WLAN_DEV_INFO *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    /* XXX this is unused right now */

    return OK;
}

/*******************************************************************************
* ar5hwcPollSend - routine to send a frame in polled mode.
*
* This routine is called by a user to try and send a frame on the
* device.
*
* RETURNS: OK upon success.  EAGAIN if device is busy.
*/
LOCAL STATUS
ar5hwcEndPollSend(END_OBJ *pEnd, M_BLK_ID pMblk)
{
    WLAN_DEV_INFO *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    if ((pdevInfo->flags & AR_POLLING) == 0) {
        return ERROR;
    }

    /* XXX this is ununsed right now */

    return OK;
}

/*******************************************************************************
* ar5hwcPollStart - start polled mode operations
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS
ar5hwcEndPollStart(END_OBJ *pEnd)
{
    WLAN_DEV_INFO *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;
    UINT oldLevel;

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    /* XXX this is ununsed right now */

    /* Turn off device interrupts */
    oldLevel = intLock();
    pdevInfo->flags |= AR_POLLING;
    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    intUnlock(oldLevel);

    return OK;
}

/*******************************************************************************
* ar5hwcPollStop - stop polled mode operations
*
* This function terminates polled mode operation.  The device returns to
* interrupt mode.
*
* The device interrupts are enabled, the current mode flag is switched
* to indicate interrupt mode and the device is then reconfigured for
* interrupt operation.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS
ar5hwcEndPollStop(END_OBJ *pEnd)
{
    WLAN_DEV_INFO *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;
    UINT oldLevel;

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    /* XXX this is unused right now */

    /* Turn on device interrupts */
    oldLevel = intLock();
    pdevInfo->flags &= ~AR_POLLING;
    halEnableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    intUnlock(oldLevel);

    return OK;
}

/*****************************************************************************
* ar5hwcMCastAdd - add a multicast address for the device
*
* This
routine adds a multicast address to whatever the driver
* is already listening for.  It then resets the address filter.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS
ar5hwcEndMCastAdd(END_OBJ *pEnd, char *pAddress)
{
    WLAN_DEV_INFO *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;
    int  error;
    char hashIdx;

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    /* XXX this is ununsed right now */

    if ((error = etherMultiAdd(&pEnd->multiList, pAddress)) == ENETRESET) {
        hashIdx = hashIndex(pAddress);
        halMulticastFilterIndex(pdevInfo, hashIdx, TRUE);
    }

    return OK;
}

/*****************************************************************************
* ar5hwcMCastDel - delete a multicast address for the device
*
* This routine removes a multicast address from whatever the driver
* is listening for.  It then resets the address filter.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS
ar5hwcEndMCastDel(END_OBJ *pEnd, char *pAddress)
{
    WLAN_DEV_INFO *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;
    char hashIdx;
    int  error;

    /* XXX this is unused right now */

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    if ((error = etherMultiDel(&pEnd->multiList, pAddress)) == ENETRESET) {
        hashIdx = hashIndex(pAddress);
        halMulticastFilterIndex(pdevInfo, hashIdx, FALSE);
    }

    return OK;
}

/*****************************************************************************
* ar5hwcMCastGet - get the multicast address list for the device
*
* This routine gets the multicast list of whatever the driver
* is already listening for.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS
ar5hwcEndMCastGet(END_OBJ *pEnd, MULTI_TABLE *pTable)
{
    return (etherMultiGet(&pEnd->multiList, pTable));
}

/******************************************************************************
* ar5RxFilterConfig - configure rx frame filtering
*
* Enables appropritate receive filtering on the device.
*
* RETURNS: N/A
*/
LOCAL void
ar5RxFilterConfig(WLAN_DEV_INFO *pdevInfo)
{
    A_UINT32        rfcrFlags = 0;
    A_UINT32        devFlags = END_FLAGS_GET(pdevInfo->pOSHandle);
    ETHER_MULTI     *pMcastAddr;
    char            hashIdx;

    DRV_LOG(DRV_DEBUG_LOAD,("RxFilterConfig devFlags=0x%x\n", devFlags));

    /* Setup Broadcast */
    if (devFlags & IFF_BROADCAST) {
        rfcrFlags |= HAL_RX_BCAST;            /* All broadcast */
    }

    /* Setup Promiscuous, all multicast, Or multicast filter */
    if (devFlags & IFF_PROMISC) {
        rfcrFlags |= HAL_RX_PROM;
    } else if (devFlags & IFF_ALLMULTI) {
        rfcrFlags |= HAL_RX_MCAST;
        halSetMulticastFilter(pdevInfo, 0xFFFFFFFF, 0xFFFFFFFF);
    } else if (devFlags & IFF_MULTICAST) {
        rfcrFlags |= HAL_RX_MCAST;

        /* Setup to receive programmed multicast frames */
        for ( pMcastAddr = END_MULTI_LST_FIRST(pdevInfo->pOSHandle);
              pMcastAddr;
              pMcastAddr=END_MULTI_LST_NEXT(pMcastAddr) )
        {
            hashIdx = hashIndex(pMcastAddr->addr);
            halMulticastFilterIndex(pdevInfo, hashIdx, TRUE);
        }
    }

    halRxFilter(pdevInfo, rfcrFlags, HAL_RX_FILTER_INIT);

    DRV_LOG(DRV_DEBUG_LOAD,("RxFilter = 0x%x\n",rfcrFlags));
    return;
}

/******************************************************************************
* hashIndex - Caculate the hash index for multicast address according to ar5210
*             spec.
*
* Input: Mac address
*
* RETURNS: hash index
*/
LOCAL STATUS
hashIndex(char *pMacAddress)
{
    char    macAddr[8];
    char    hashIdx;

    macAddr[0] = pMacAddress[0] & 0x3F;
    macAddr[1] = ((pMacAddress[1] & 0x0F)<<2) | ((pMacAddress[0] & 0xC0) >> 6);
    macAddr[2] = ((pMacAddress[2] & 0x03)<<4) | ((pMacAddress[1] & 0xF0) >> 4);
    macAddr[3] = (pMacAddress[2] & 0xFC) >> 2;
    macAddr[4] = pMacAddress[3]& 0x3F;
    macAddr[5] = ((pMacAddress[4] & 0x0F)<<2) | ((pMacAddress[3] & 0xC0) >> 6);
    macAddr[6] = ((pMacAddress[5] & 0x03)<<4) | ((pMacAddress[4] & 0xF0) >> 4);
    macAddr[7] = (pMacAddress[5] & 0xFC) >> 2;

    hashIdx = macAddr[0]^macAddr[1]^macAddr[2]^macAddr[3]^macAddr[4]^macAddr[5]
        ^macAddr[6]^macAddr[7];

    return hashIdx;
}

/******************************************************************************
* ar5hwcEndUnload - unload a driver from the system
*
* This function first brings down the device, and then frees any
* stuff that was allocated by the driver in the load function.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS
ar5hwcEndUnload(END_OBJ *pEnd)
{
    WLAN_DEV_INFO *pdevInfo = (WLAN_DEV_INFO *)pEnd->devObject.pDevice;

    if (pdevInfo->NicResetInProgress == TRUE) {
        return ERROR;
    }

    /* Bring down the device */
    END_OBJECT_UNLOAD(pEnd);

    if (!pdevInfo) {
        return OK;
    }

    /* free memory allocated for clusters and Tx/Rx Queues */
    if ((pdevInfo->flags & AR_BUF_LOCAL) && (pdevInfo->pDmaBuf != NULL)) {
#ifdef VX_CACHE_DMA_MALLOC
        cacheDmaFree(pdevInfo->pDmaBuf);
#else
        free(pdevInfo->pDmaBuf);
#endif
    }

    /* free memory allocated for mBlks and cluster blocks */
    if (pdevInfo->pMclMem) {
        free(pdevInfo->pMclMem);
    }

    /* free memory allocated for net pool */
    if (pdevInfo->pOSHandle->pNetPool != NULL) {
        free(pdevInfo->pOSHandle->pNetPool);
    }

    /* free memory allocated for device structure */
    free(pdevInfo);

    return OK;
}

/******************************************************************************
* athCalibrationTask
*
* This task is used to calibrate the HW
*/
LOCAL void
athCalibrationTask(void)
{
    UINT32          msgBuffer[3] = {0, 0, 0};
    WLAN_DEV_INFO   *pdevInfo;
    UINT32          status;

    while (1) {
        if (msgQReceive(calMsgQId, (char *)&msgBuffer[0], sizeof(msgBuffer),
                        WAIT_FOREVER) == ERROR)
        {
            logMsg("Could not do offset/nf cal on time\n");
            continue;
        }

        pdevInfo = (WLAN_DEV_INFO *)msgBuffer[0];

        if (pdevInfo->NicResetInProgress == TRUE) {
            continue;
        }

        status = halPerCalibration(pdevInfo, pdevInfo->staConfig.pChannel);
        if (status != A_OK) {
            /*
             * Noise cal failed.  Possible CW radio interference.
             * Reboot the AP, hopefully the interference will have gone away or
             * AP will detect during reboot and move over to another channel.
             */
            uiPrintf("WARNING! WARNING! Radio interference is detected in the selected channel\n");
            uiPrintf("WARNING! WARNING! Access point operation is being terminated\n");
            apReboot();
        }
    }
}

#define AP_BM_ALERT_TH  4
/*******************************************************************************
* beaconWdgStart - start beacon watchdog timer
*
* This routine is called when the bss starts
*
* RETURNS:  OK if successful
*           ERROR if unsuccessful
*/
STATUS
beaconWdgStart(WLAN_DEV_INFO *pdevInfo)
{
    if (beaconWatchDogId == NULL) {
        if ((beaconWatchDogId = wdCreate()) == NULL) {
            apPanic("beacon watchdog start ERROR: watchdog creation failed\n");
            return ERROR;
        }
    }

    pdevInfo->pBeaconInfo->wdgDelayTicks = AP_BM_ALERT_TH * sysClkRateGet() *
        pdevInfo->bssDescr->beaconInterval / 1000;

    /* set timer goes off in beaconInterval */
    if (wdStart(beaconWatchDogId, pdevInfo->pBeaconInfo->wdgDelayTicks,
                (FUNCPTR)beaconWatchDogISR, (int)pdevInfo) == ERROR)
    {
        /* clean up */
        wdDelete(beaconWatchDogId);
        apPanic("beacon watchdog start ERROR: watchdog start failed\n");
        return ERROR;
    }

    return OK;
}

/*******************************************************************************
* beaconWdgCancel - cancel beacon watchdog timer
*
* This routine is called when the chip resets
*
* RETURNS:  OK if successful
*           ERROR if unsuccessful
*/
STATUS
beaconWdgCancel(void)
{
    /* Cancel the beacon watchdog timer */
    if (beaconWatchDogId != NULL) {
        if (wdCancel(beaconWatchDogId) != OK) {
            return ERROR;
        }
    }
    return OK;
}

/*******************************************************************************
* beaconWatchDogISR - isr routine for beacon watchdog timer
*
* This routine is called when the beacon watch dog timer expires
*
* RETURNS:  OK if successful
*           ERROR if unsuccessful
*/
void
beaconWatchDogISR(WLAN_DEV_INFO *pdevInfo)
{
    /* check if the beacon transmitted counter advanced */
#ifndef DEBUG
    if (txBeaconNum == pdevInfo->pBeaconInfo->txBeaconCount) {
        apPanic("beacon transmission problem at beacon #%d!\n", txBeaconNum);
    } else
#else
    if (txBeaconNum == pdevInfo->pBeaconInfo->txBeaconCount) {
        logMsg("beacon transmission problem at beacon #%d!\n", txBeaconNum);
    } else
#endif
    {
        txBeaconNum = pdevInfo->pBeaconInfo->txBeaconCount;

        /* Restart beacon watchdog timer */
        if (wdStart(beaconWatchDogId, pdevInfo->pBeaconInfo->wdgDelayTicks,
                    (FUNCPTR)beaconWatchDogISR, (int)pdevInfo) == ERROR)
        {
            wdDelete(beaconWatchDogId);
#ifdef DEBUG
            logMsg("AP beacon watch dog timer can not be activated.\n");
#endif
        }
    }
}

#ifdef DEBUG
void
dumpFrame(ATHEROS_DESC* pDesc, char* pMsg)
{
    if (rxVerbose || txVerbose) {
        A_UINT16 *fc;

        if (strcmp(pMsg, "rx") != 0)
            if (!bcVerbose && pDesc->pBufferVirtPtr.header->address1.octets[0] & 1)
                return;

        if (pDesc->hwControl.word[0] && pDesc->hwControl.word[1])
            uiPrintf ("%s desc: %p, hwctrl: %08x, %08x, hwstat: %08x, %08x\n",
                      pMsg, pDesc, pDesc->hwControl.word[0], pDesc->hwControl.word[1],
                  pDesc->hwStatus.word[0], pDesc->hwStatus.word[1]);
        else
            uiPrintf ("%s desc: %p, hwctrl: %08x, %08x\n",
                      pMsg, pDesc, pDesc->hwControl.word[0], pDesc->hwControl.word[1]);
        fc = (A_UINT16 *)&pDesc->pBufferVirtPtr.header->frameControl;
        uiPrintf ("fc: %x, order: %d, wep: %d, more: %d, pwr: %d, retry: %d, frag: %d, to: %d, from: %d\n",
                  *fc,
                  pDesc->pBufferVirtPtr.header->frameControl.order,
                  pDesc->pBufferVirtPtr.header->frameControl.wep,
                  pDesc->pBufferVirtPtr.header->frameControl.moreData,
                  pDesc->pBufferVirtPtr.header->frameControl.pwrMgt,
                  pDesc->pBufferVirtPtr.header->frameControl.retry,
                  pDesc->pBufferVirtPtr.header->frameControl.moreFrag,
                  pDesc->pBufferVirtPtr.header->frameControl.ToDS,
                  pDesc->pBufferVirtPtr.header->frameControl.FromDS);
        uiPrintf ("adr1 ");
        printMacAddress (pDesc->pBufferVirtPtr.header->address1);
        uiPrintf (" adr2 ");
        printMacAddress (pDesc->pBufferVirtPtr.header->address2);
        uiPrintf(", type: ");

        switch (pDesc->pBufferVirtPtr.header->frameControl.fType) {
            case FRAME_MGT:
                uiPrintf("mgt -- %s\n",
                         mgtTypeStr[pDesc->pBufferVirtPtr.header->frameControl.fSubtype]);
                break;
            case FRAME_CTRL:
                uiPrintf("ctrl -- %s\n",
                         ctrlTypeStr[pDesc->pBufferVirtPtr.header->frameControl.fSubtype]);
                break;
            case FRAME_DATA:
                uiPrintf("data -- %s\n",
                         dataTypeStr[pDesc->pBufferVirtPtr.header->frameControl.fSubtype]);
                break;
        }
        if (rxVerbose == 99 || txVerbose == 99)
            taskSuspend (0);
    }
}

void
swrDump(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib)
{
    extern int tqDump(ATHEROS_DESC *pDesc, const char* msg);

    uiPrintf("pSib = 0x%x; AID = %d; numFilteredPending = %d; txPending = %d\n",
            (A_UINT32)pSib, pSib->assocId  & 0x3fff,
            pSib->swRetry.numFilteredPending, pSib->numTxPending);
    tqDump(pSib->swRetry.retryQueue.pDescQueueHead, "swretryQueue");
    tqDump(pdevInfo->txQueue.pDescQueueHead, "txQueue");

    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    eventDump(SUBSYS_WLANTX, 0);

    ASSERT(0);
}

int
swRetryQueueVerify(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib)
{
    ATHEROS_DESC    *p;
    int             cnt = 0;
    int             filtered = 0, xsr = 0, cdm = 0;
    int             sibCnt = 0;
    int             notDone = 0;
    int             keyIdx = 0;

    if (swRetryDebug == 0 || pSib == NULL) {
        return 0;
    }

    for (p = pdevInfo->txQueue.pDescQueueHead; p; p = p->pNextVirtPtr) {
        ATHEROS_DESC*   first = p;
        ATHEROS_DESC*   last = lastDescGet(p);

        if (p->staleFlag) {
            continue;
        }

        cnt++;
        if (first->pDestSibEntry != pSib) {
            continue;
        }

        ASSERT(first->pTxFirstDesc == p);

        sibCnt++;

        keyIdx = halGetTxDescEncryptKeyIndex(pdevInfo, first);

        if (first->hwControl.tx.clearDestMask == 1) {
            cdm++;
        }

        if (last->hwStatus.tx.done == 1) {
            if (last->hwStatus.tx.excessiveRetries == 1) {
                xsr++;
            } else if (last->hwStatus.tx.filtered == 1) {
                filtered++;
            }
        } else {
            notDone++;
        }
        p = last;
    }

    if (sibCnt != pSib->numTxPending) {
        swrDump(pdevInfo, pSib);
    }

    //  if ((filtered+notDone) != pSib->swRetry.numFilteredPending)
    //          enterDebugger();

    if (halGetTransmitFilterIndex(pdevInfo, keyIdx) &&
           cdm == 0 && pSib->swRetry.needClearDest == 0)
    {
        swrDump(pdevInfo, pSib);
    }

    if (xsr > 1) {
        swrDump(pdevInfo, pSib);
    }

    if (cdm > 1) {
        swrDump(pdevInfo, pSib);
    }

    // add check for needCDM=0 if cdm=1
    return 0;
}

int bssStart ()
{
    return muxIoctl(pwlanCookie[0], EIOCBSSSTART, 0);
}


#endif /* DEBUG */

#ifdef INCLUDE_TRACE
static void
eventDump_wlantx(struct event_s *event)
{
    struct event_s *ev = event;
    int            j;

    switch (ev->eventID) {
    case EV_TX_RETRYQFLUSH:
    case EV_TX_MUSTEAT:
        uiPrintf("%s pSib %p pDesc %p numFiltP %d\n",
                 ev->msg, ev->arg[0].ptr, ev->arg[1].ptr, ev->arg[2].word);
        break;

    case EV_TX_NODUPREQ:
    case EV_TX_DUPREQ:
    case EV_TX_PKTOK:
    case EV_TX_XSR:
    case EV_TX_FILTER:
    case EV_TX_FILTNEEDCDM:
    case EV_TX_XSRDROP:
    case EV_TX_NORSRC:
    case EV_TX_QDOK:
    case EV_TX_CDM:
    case EV_TX_SQUASH:
        uiPrintf("%s pSib %p pDesc %p numFiltP %d\n",
                 ev->msg, ev->arg[0].ptr, ev->arg[1].ptr, ev->arg[2].word);
        break;

    case EV_TX_TXDITER:
        uiPrintf("%s pDesc %p more %d stale %d\n", ev->msg,
                 ev->arg[0].ptr, ev->arg[1].word, ev->arg[2].word);
        break;
    case EV_TX_Q:
        uiPrintf("%s pDesc %p more %d\n", ev->msg,
                 ev->arg[0].ptr, ev->arg[1].word);
        break;

    case EV_TX_NOFRAME:
        uiPrintf("%s\n", ev->msg);
        break;

    default:
        uiPrintf("Ev:(%d) %s ", ev->eventID, ev->msg);
        for (j = 0; j < N_EVENT_ARGS; j++) {
            uiPrintf("%d(%p) ", j, ev->arg[j].ptr);
        }
        uiPrintf("\n");
        break;
    }

    return;
}

void
eventDump_wlanrx(struct event_s *event)
{
    struct event_s *ev = event;

    switch (ev->eventID) {
    case EV_RX_EOLREC:
        uiPrintf("%s desc count %d\n", ev->msg, ev->arg[0].word);
        break;
    case EV_RX_OK:
    case EV_RX_ERR:
        uiPrintf("%s pDesc %p hwStatus %08x:%08x numRxDesc %d\n", ev->msg,
                 ev->arg[0].ptr, ev->arg[1].word, ev->arg[2].word, ev->arg[3].word);
        break;
    case EV_RX_SIBFIND:
        uiPrintf("%s Sib %p\n", ev->msg, ev->arg[0].ptr);
        break;
    case EV_RX_OVF:
        uiPrintf("%s pDesc %p active desc count %d\n", ev->msg,
                 ev->arg[0].ptr, ev->arg[1].word);
        break;
    }
}

void
ltfShow(void)
{
    eventDump(SUBSYS_WLANTX, 100);
}

#endif /* INCLUDE_TRACE */
