/*
 * Soft AP driver - based on the Host AP driver
 * Partha Narasimhan, Aruba Networks
 * 6/25/02
 *
 * Host AP (software wireless LAN access point) driver for ATHEROS 5210/5211 chipset
 * Copyright (c) 2001-2002, SSH Communications Security Corp
 * Jouni Malinen <jkm@ssh.com> and/or <jkmaline@cc.hut.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 *
 * FIX:
 * - shutting down pcmcia-cs while wlan# interface is up does not work
 *   (it does not remove the driver and eventually the kernel will panic when
 *   card is removed); when interface is first configured down, pcmcia-cs can
 *   be shut down; update: this seems to work at least with kernel-based
 *   pcmcia and Linux 2.4.13..
 * - SIOCSIWRATE can set TX rate for adhoc and managed modes, but this is not
 *   currently used in host-based TX rate control; it could be used also for
 *   this (i.e., allow AP to be configured to limit/force TX rate)
 * - there is currently no way of associating TX packets to correct wds device
 *   when TX Exc/OK event occurs, so all tx_packets and some
 *   tx_errors/tx_dropped are added to the main netdevice; using sw_support
 *   field in txdesc might be used to fix this (using Alloc event to increment
 *   tx_packets would need some further info in txfid table)
 *
 * Buffer Access Path (BAP) usage:
 *   BAP0 is used for sending data from the host computer to the card and for
 *   RID read/writes; it is protected with local->baplock (spin_lock_irqsave)
 *   and it must not be used from hardware interrupt context.
 *   BAP1 is used for receiving data from the card to the host computer; it is
 *   used only in hardware interrupt handler and kernel is assumed to not call
 *   the same handler simultaneously for the same interrupt even on SMP
 *   systems (this removes need for spin_lock protecting BAP1 access.
 *   Note! There seems to be some problems with BAP0 access being interrupted
 *   with RX/info frame (BAP1). To remove these problems, baplock is used with
 *   spin_lock_irqsave (it could be _bh otherwise) and RX/info handling locks
 *   baplock during BAP1 setup to make sure that BAP0 use has been completed.
 */


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wireless.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/interrupt.h>


#include <asm/io.h>
#include <asm/byteorder.h>

#ifdef ATHEROS_MONITOR
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
/* net/sock.h (at least in 2.2.17) does not like min()/max() macros from
 * pcmcia-cs's kernel compat header */
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif
#endif /* kernel < 2.4.0 */

#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#endif /* ATHEROS_MONITOR */



#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
/* 2.2 compatibility */
#ifndef spin_lock_bh
#define spin_lock_bh(lock) spin_lock_irq(lock)
#define spin_unlock_bh(lock) spin_unlock_irq(lock)
#endif
#ifndef __constant_cpu_to_le16
#define __constant_cpu_to_le16 __cpu_to_le16
#endif

#define ATHEROS_NETDEV_EXTRA IFNAMSIZ
#define atheros_set_dev_name(dev, pos) (dev)->name = (char *) (pos)
/* end 2.2 compatibility */
#else /* kernel < 2.4.0 */
/* no extra space needed for 2.4.x net_device */
#define ATHEROS_NETDEV_EXTRA 0
#define atheros_set_dev_name(dev, pos) do { }  while (0)
#endif /* kernel < 2.4.0 */
                                            
#include "linuxdrv.h"

#include "atheros_wlan.h"
#include "atheros_sap.h"
#include "wlantype.h"
#include "wlandrv.h"                              
#include "wlanchannel.h"
#include "wlanext.h"
#include "wlanfragde.h"
#include "halApi.h"
#include "apdefs.h"
#ifdef BUILD_AR5210
#include "ar5210.h"
#include "ar5210Reg.h"
#else
#include "ar5211.h"
#include "ar5211Reg.h"
#endif
#include "ui.h"
#include "display.h"
#include "wlanbeacon.h"
#include "hal.h"
#include "stacserv.h"
#include "wlansmeext.h"
#include "apmem.h"

#include "ar5hwcEnd.h"
#include "wifi/wifi.h"
#include "wifi/sap_config.h"

#define RRA_MSG_INIT
#include "wifi/rra_msg.h"
#undef  RRA_MSG_INIT

/* ATHEROS PROTOTYPES */

/* reset */
void resetOnError(WLAN_DEV_INFO *pdevInfo);
void resetwlan(WLAN_DEV_INFO *pdevInfo, struct net_device *dev);
void chipReset(WLAN_DEV_INFO *pdevInfo, A_BOOL unicast, CHAN_VALUES *pChannel);
void resetChipForBeacons(WLAN_DEV_INFO *pdevInfo, struct net_device *dev);
/* initialization */
A_STATUS default_driver_mem_desc (WLAN_DEV_INFO * pdevInfo);
A_STATUS init_driver_mem_pool(WLAN_DEV_INFO *pdevInfo);
A_STATUS init_driver_queue_create(WLAN_DEV_INFO *pdevInfo, QUEUE_DETAILS *pQueue, 
                                QUEUE_DETAILS *freeQueue,int count, int bufSize);
void init_driver_queue_delete(QUEUE_DETAILS *pQueue);
void init_driver_queue_reset(WLAN_DEV_INFO *pdevInfo);
/* receive */
void queueRxDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pRxDesc);
void processRxInterrupt(WLAN_DEV_INFO *);
inline A_STATUS setupDecrypt(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);
void atheros_handle_mgt_frame (WLAN_DEV_INFO *pdevInfo, struct ieee80211_hdr* phdr, A_UINT16 len);

/* interrupts */
A_BOOL HandleAtherosRxIntAndStats (WLAN_DEV_INFO *pdevInfo);
void   HandleAtherosTxIntAndStats (WLAN_DEV_INFO *pdevInfo);
void   DisableInterruptsGlobaly   (WLAN_DEV_INFO *pdevInfo);
void   EnableInterruptsGlobaly    (WLAN_DEV_INFO *pdevInfo);

/* transmit */
A_STATUS wlanFrameTransmit(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_BOOL bypass);
A_STATUS ar5MacSend(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_BOOL bypass);
A_STATUS txService(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pTxDesc, A_BOOL bypass);
A_STATUS atheros_send(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pHead, 
                      A_BOOL callerHoldsLock, A_BOOL bypass);
inline ATHEROS_DESC *swretrySetup(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, ATHEROS_DESC *pTrailer);
void queueTxDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, ATHEROS_DESC *pTail);
void processAltQueue(WLAN_DEV_INFO *);
void processTxQueue(WLAN_DEV_INFO *pdevInfo);
inline void processTxInterrupt(WLAN_DEV_INFO *pdevInfo);
inline void txFramePendDecrement(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib, ATHEROS_DESC *pDesc);
inline int processTxSibInfo(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pTxDesc);
A_STATUS ar521x_transmit  (void *ignore, struct sk_buff *skb, struct net_device *dev);


ATHEROS_DESC *nullDataFrameCreate(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSibEntry);
inline int swretryLookAhead(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pTxDesc);
void RateCtrlBBChange(WLAN_DEV_INFO *pdevInfo, A_UINT16 channelFlags);
#ifdef RATECTRL_C
void drvBBChange(WLAN_DEV_INFO *pdevInfo, CHAN_VALUES *pChval);
#endif
void AR5210_setup_power_level (int power_level);
void AR5211_setup_power_level (int power_level);

/* AIR MONITOR */
monitor_type g_monitor_type = ATHEROS_MONITOR_OFF; // by default, AM is off

/* CLI and MGT interface */
A_STATUS atheros_set_tx_power        (int power_level);
A_STATUS atheros_set_dtim_interval   (int dtim_interval);
A_STATUS atheros_set_beacon_interval (int beacon_interval);
A_STATUS atheros_set_short_pream     (A_BOOL short_pream);
A_STATUS atheros_set_tx_rate         (int tx_rate);
A_STATUS atheros_set_enabled             (A_BOOL enable);


int process_am_frame_a(struct ar5211_rx_frame *, char *);
A_UINT16 atheros_get_channel (void);
A_STATUS atheros_set_channel (A_UINT16 channel);
A_UINT16 atheros_get_ammode  (void);
int atheros_set_tim(struct net_device* dev, u16 aid, int set);
void     atheros_set_ammode  (monitor_type mode);
A_STATUS atheros_set_essid   (char* essid);
char*    atheros_get_essid   (void);

/* channel/frequency routines */
A_UINT16 g_channel_operational = CHAN_11A_DEFAULT;


#ifdef DEBUG
int gNumOfFrames = 0;
int bcVerbose    = 0;
int rxVerbose    = 10; //fadi, turn this off eventually
int txVerbose    = 10;
int swRetryDebug = 0;
void       dumpFrame(ATHEROS_DESC *pDesc, char *pMsg);
int swRetryQueueVerify(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib);
void swrDump(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib);
#else
#define swRetryQueueVerify(d, s)
const int swRetryDebug = 0;
#endif /* DEBUG */

#include <linux/types.h>
#include <linux/ipc.h>
#include <linux/sem.h>

struct sembuf sem_lock = {0, -1, SEM_UNDO};     
struct sembuf sem_unlock = {0, 1, SEM_UNDO};    
int    global_sem_key = 1234; // any unique key to create semaphores

#undef  WAIT_FOREVER
#define WAIT_FOREVER (-1)

A_SEM_TYPE    radarTaskSemId;
A_SEM_TYPE    radarTxThrottleSemId;
A_SEM_TYPE    radarDataSemId;

#define ATHEROS_HW_CTRL_AND_STATUS       (sizeof(HAL_DESC_RX_CONTROL) + sizeof(HAL_DESC_RX_STATUS))
#define ATHEROS_GRE_HEADROOM             24
#define ATHEROS_MONITOR                  1
#undef  ETHER_MAP_IP_MULTICAST
/* need this for Crete; also need it in case the
 * rx overrun on small AES frames bug persists in
 * the hw; the workaround for the latter is to
 * reset the chip on rx overrun, and we'd rather
 * have buggy overruns only and not real overruns
 * - real overruns are informative only and thus
 * resetting the chip on real overruns would be
 * very expensive
 */
#define RXEOL_WAR  1
//#undef RXEOL_WAR

/* Set number of rx descriptors to process before rescheduling */
#define AR_RX_RESCHED   AR_RX_QUEUE_DEPTH      /* define as 0 to remove this fcn */

// track different types of phy errors
A_UINT32 phyErrorCounts[8] = 
{ 0, 0, 0, 0, 0, 0, 0, 0, }; 

/* ATHEROS DEFINES */

static const int contention_window[] = {
           0,   15,   31,   63,  127,  255,  511, 1023, 1023, 1023,
        1023, 1023, 1023, 1023, 1023, 1023, 1023, 1023, 1023, 1023,
        1023
};


#define RC_BKGND_MODE           1       /* Background operating mode */
#define RC_FOREGND_MODE         2       /* Foreground operating mode */

#define LOWER_RELRATE           0
#define CURRENT_RELRATE         1
#define HIGHER_RELRATE          2

#define LO_RATE_MAX_CNT         10
#define HI_RATE_MAX_CNT         10

#define RRTHLD1PCT              99      /* rate reduction threshold -- 1% */
#define RRTHLD10PCT             90      /* rate reduction threshold -- 10% */
#define RETHLD100               100     /* rate elevation threshold -- 100% */
#define RETHLD110               110     /* rate elevation threshold -- 110% */


#define MAX_FRAME_LOSS          1000    /* Mulitpler */
#define FRAME_LOSS_THRESHOLD    50

#define MAX_RTS_CTS_AGING       20

#define SHORT_FRAME_LEN         500
#define LONG_FRAME_LEN          1530
#define MIN_FRAME_LEN           100
#define FRAME_BUCKET_LEN        100


#define MAX_THROTTLE_HISTORY    200



struct RateCtrl_s {
    int      rate;              /* Data rate */
    A_UINT16 bkoffRidx;         /* Software retry backoff data rate index */
    A_UINT16 valid;             /* Valid data rate */
    int      mb;                /* Valid broadcast/multicast rate */
    A_UINT32 dn_threshold;
    A_UINT32 up_threshold;
};


extern int csDebug;
extern int csDebug2;
extern int csrDebug;
extern int rcDebug;
extern int csVerbose;
extern int autoRate;
extern int csInitialized;
extern int sfTxRateCntmax;
extern int lfTxRateCntmax;
extern int iterations;
extern int curRate;
extern int curTxRate;
extern int RateCtrlTaskDelay; 
extern char *dataRateTbl[];
extern char *dataRate2xTbl[];
extern int RateAlgSelect;
extern char **pDataRate;
extern struct TxRetryHisto_s TxRetryHistogram;
extern struct TxRetryHistoBB_s TxRetryHistoCCK;
extern struct TxRetryHistoBB_s TxRetryHistoOFDM;
extern struct TxRetryHistoBB_s *pTxRetryHistoBB;
extern struct RateCtrl_s RateCtrlTblOFDM[];
extern struct RateCtrl_s *RateCtrlTbl;
#ifndef MIN
#define MIN(x,y)        ((x) <= (y) ? (x) : (y))
#endif 
extern A_UINT8 RssiAckValidMin11a[];
extern A_UINT8 RssiAckDeltaMin11a[]; 
extern A_UINT8 RssiAckValidMin11aTurbo[];
extern A_UINT8 RssiAckDeltaMin11aTurbo[];
extern A_UINT8 RssiAckDeltaMinCCK[];
extern A_UINT8 *RssiAckValidMin;
extern A_UINT8 *RssiAckDeltaMin;




extern char *mgtTypeStr[];
extern char *ctrlTypeStr[];
extern char *dataTypeStr[];

static int atheros_close(struct net_device *dev);

extern WLAN_DEV_INFO *pwlanDevInfo;
extern struct pci_dev * gPciDev;
 

/* For more info, look at 9-5, table 9-1 in AR5001 Programmer's manual
   DMA BUF SIZE = txDescCount*theirSize +    // 256 x (64 | 1596)
                  rxDescCount*theirSize +    // 128 x 1596
                  extraClCount*Size     +    // 144 x 1596
                  extraTxDescCount*Size =    // 256 x 1596
                                        a max of 1.2MB
  Before you sweat bullets, this DmaBufSize is not allocated 
  but used to check against totalSize later and that is the
  final size that is finally allocated...to a max of 1.2MB
*/              

#define MAX_DMA_AREA (128*1024)

#define DELTA(a,b)	(((A_UINT32)(a)) - ((A_UINT32)(b)))

int atheros_rx_kernel(WLAN_DEV_INFO *pWlanDev, struct net_device *dev, 
                      ATHEROS_DESC *pRxDesc, int am_monitor_flag);
void atheros_sap_switch_uplink(unsigned long data);
DECLARE_TASKLET(atheros_sap_tasklet, atheros_sap_switch_uplink, 0);

struct sk_buff_head atheros_sap_readq; 

#define ATHEROS_MAX_INTERRUPT_EVENTS    40
#ifndef CLI_MIN_DATA_RATE
#define CLI_MIN_DATA_RATE				6
#endif
#ifndef CLI_MAX_DATA_RATE
#define CLI_MAX_DATA_RATE				54
#endif


extern A_BOOL   init_buff_pool (void);

/* #define final_version */
#define MAX_PARM_DEVICES 8
#define PARM_MIN_MAX "1-" __MODULE_STRING(MAX_PARM_DEVICES)
#define DEF_INTS -1, -1, -1, -1, -1, -1, -1
#define GET_INT_PARM(var,idx) var[var[idx] < 0 ? 0 : idx]

#define TSF_ARRAY_SIZE  512
static WLAN_TIMESTAMP   tsfStamp[TSF_ARRAY_SIZE];
static A_UINT32         tsfIndex = 0;

static int mtu = 1600;
MODULE_PARM(mtu, "i");
MODULE_PARM_DESC(mtu, "Maximum transfer unit");

static int channel[MAX_PARM_DEVICES] = { 1, DEF_INTS };
MODULE_PARM(channel, PARM_MIN_MAX "i");
MODULE_PARM_DESC(channel, "Initial channel");

static char *essid[MAX_PARM_DEVICES] = { "asap" };
MODULE_PARM(essid, PARM_MIN_MAX "s");
MODULE_PARM_DESC(essid, "Host AP's ESSID 802.11a");

static int beacon_int[MAX_PARM_DEVICES] = { 100, DEF_INTS };
MODULE_PARM(beacon_int, PARM_MIN_MAX "i");
MODULE_PARM_DESC(beacon_int, "Beacon interval (1 = 1024 usec)");

static int dtim_period[MAX_PARM_DEVICES] = { 1, DEF_INTS };
MODULE_PARM(dtim_period, PARM_MIN_MAX "i");
MODULE_PARM_DESC(dtim_period, "DTIM period");

static int disable_on_close = 0;
MODULE_PARM(disable_on_close, "i");
MODULE_PARM_DESC(disable_on_close, "Disable MAC port on netdevice close");

#if defined(ATHEROS_PCI) && defined(ATHEROS_BUS_MASTER)
static int bus_master_threshold_rx[MAX_PARM_DEVICES] = { 100, DEF_INTS };
MODULE_PARM(bus_master_threshold_rx, "i");
MODULE_PARM_DESC(bus_master_threshold_rx, "Packet length threshold for using "
		 "PCI bus master on RX");

static int bus_master_threshold_tx[MAX_PARM_DEVICES] = { 100, DEF_INTS };
MODULE_PARM(bus_master_threshold_tx, "i");
MODULE_PARM_DESC(bus_master_threshold_tx, "Packet length threshold for using "
		 "PCI bus master on TX");
#endif /* ATHEROS_PCI and ATHEROS_BUS_MASTER */



/* ca. 1 usec */
#define ATHEROS_CMD_BUSY_TIMEOUT 1000
#define ATHEROS_BAP_BUSY_TIMEOUT 5000

/* ca. 10 usec */
#define ATHEROS_INIT_TIMEOUT 50000
#define ATHEROS_CMD_COMPL_TIMEOUT 20000
#define ATHEROS_ALLOC_COMPL_TIMEOUT 1000


#define TX_TIMEOUT (2 * HZ)
#define ATHEROS_MAX_FRAME_SIZE 2304
#define ATHEROS_MIN_MTU 256
/* FIX: */
#define ATHEROS_MAX_MTU (ATHEROS_MAX_FRAME_SIZE)


/* Beacon Watchdog stuff */
//LOCAL WDOG_ID  beaconWatchDogId = NULL;
LOCAL A_UINT32 txBeaconNum      = 0;
LOCAL A_UINT16 beaconNum        = 0;
LOCAL A_UINT16 calBeaconCount   = 0;

int atheros_recv_frame(atheros_local_info_t *, struct sk_buff *);

/* atheros_ioctl.c */
struct iw_statistics *atheros_get_wireless_stats(struct net_device *);
int atheros_ioctl(struct net_device *, struct ifreq *, int );


#ifdef DEBUG
void
dumpFrame(ATHEROS_DESC* pDesc, char* pMsg)
{
    if (rxVerbose || txVerbose) {
        A_UINT16 *fc;

        if (strcmp(pMsg, "rx") != 0)
            if (!bcVerbose && pDesc->pBufferVirtPtr.header->address1.octets[0] & 1)
                return;

        if (pDesc->hwControl.word[0] && pDesc->hwControl.word[1])
            aPrintk ("%s desc: %p, hwctrl: %08x, %08x, hwstat: %08x, %08x\n",
                      pMsg, pDesc, pDesc->hwControl.word[0], pDesc->hwControl.word[1],
                  pDesc->hwStatus.word[0], pDesc->hwStatus.word[1]);
        else
            aPrintk ("%s desc: %p, hwctrl: %08x, %08x\n",
                      pMsg, pDesc, pDesc->hwControl.word[0], pDesc->hwControl.word[1]);
        fc = (A_UINT16 *)&pDesc->pBufferVirtPtr.header->frameControl;
        aPrintk ("fc: %x, order: %d, wep: %d, more: %d, pwr: %d, retry: %d, frag: %d, to: %d, from: %d\n",
                  *fc,
                  pDesc->pBufferVirtPtr.header->frameControl.order,
                  pDesc->pBufferVirtPtr.header->frameControl.wep,
                  pDesc->pBufferVirtPtr.header->frameControl.moreData,
                  pDesc->pBufferVirtPtr.header->frameControl.pwrMgt,
                  pDesc->pBufferVirtPtr.header->frameControl.retry,
                  pDesc->pBufferVirtPtr.header->frameControl.moreFrag,
                  pDesc->pBufferVirtPtr.header->frameControl.ToDS,
                  pDesc->pBufferVirtPtr.header->frameControl.FromDS);
        aPrintk ("adr1 ");
        printMacAddress (pDesc->pBufferVirtPtr.header->address1);
        aPrintk (" adr2 ");
        printMacAddress (pDesc->pBufferVirtPtr.header->address2);
        aPrintk(", type: ");

        switch (pDesc->pBufferVirtPtr.header->frameControl.fType) {
            case FRAME_MGT:
                aPrintk("mgt -- %s\n",
                         mgtTypeStr[pDesc->pBufferVirtPtr.header->frameControl.fSubtype]);
                break;
            case FRAME_CTRL:
                aPrintk("ctrl -- %s\n",
                         ctrlTypeStr[pDesc->pBufferVirtPtr.header->frameControl.fSubtype]);
                break;
            case FRAME_DATA:
                aPrintk("data -- %s\n",
                         dataTypeStr[pDesc->pBufferVirtPtr.header->frameControl.fSubtype]);
                break;
        }
        if (rxVerbose == 99 || txVerbose == 99) {
            aPrintk("BAD rxVerbose!!! PANICCCCCCC?!\n");
        }
        
    }
}

#endif // DEBUG

  



static int atheros_hw_init(struct net_device *dev, int initial)
{
  atheros_local_info_t *local = (atheros_local_info_t *) dev->priv;
  
  halDisableInterrupts (local->pWlanDev, HAL_INT_GLOBAL);
  aPrintk("\tDisabled interrupts...\n");

  dev->addr_len = 6;
  
  return 0;
}


static int atheros_hw_enable(struct net_device *dev)
{
  atheros_local_info_t *local = (atheros_local_info_t *) dev->priv;
  
  local->hw_ready = 1;
  local->hw_reset_tries = 0;

  halEnableInterrupts ((WLAN_DEV_INFO *)local->pWlanDev, HAL_INT_COMMON);
  aPrintk("atheros interrupts enabled\n");

  return 0;
}


int atheros_hw_config(struct net_device *dev, int initial)
{
  // #1- disable interrupts + more init
  if (atheros_hw_init(dev, initial))
    return 1;

  // #2 set power to up or poll or auto
  // #3 init beacons and their descriptors
  // #4 enable receive
  // LAST: enable interrupts
  
  A_STATUS status;
  atheros_local_info_t *local 		= (atheros_local_info_t *)(dev->priv);
  WLAN_DEV_INFO * pDevInfo  = (WLAN_DEV_INFO *)(local->pWlanDev);
  
  ASSERT (local);

  if (!initial) {
      local->dev_enabled = 1;
  }
  
  //#2 Force powermode to awake (maybe set it to PSPoll later!)
  if (A_OK != (status = halSetPowerModeAwake (pDevInfo, TRUE)) ) {
      aPrintk("Atheros_Hw_Config: Failed to Force PowerMode to AWAKE!\n");
      
      return 0;
  }
  aPrintk("\tSetPowerMode to AWAKE..\n");

  //#3 Initialize and start beacons
  if (A_OK != (status = wlanBeaconInit(pDevInfo, 0)) ) {
	  aPrintk("Atheros_Hw_Config: Failed to Initialize Beacon!\n");
      return 0;
  }
  
  //#4 Enabling receiving (DMA receive?? don't toggle for now!) 
  halEnableRxZlf (pDevInfo);
  aPrintk("\thalEnableRxZlf....\n");
  halEnableReceive (pDevInfo);
  aPrintk("\thalEnableReceive....\n");
  
  /* Allow DMA of RX now that we're ready to rock'n'roll */
  halRxFilter(pDevInfo, HAL_RX_UCAST, HAL_RX_FILTER_INIT);
  aPrintk("\thalRxFilter...\n");

  pDevInfo->localSta->staState = STATE_AP_UP;
  
  local->hw_ready = 1;
  local->hw_reset_tries = 0;

#if _FADI // this was causing us to break??
  /* AP software is up, change the state. */
  if ((NULL != pSibEntry) && (NULL != pDevInfo))       
      pSibEntry->staState = STATE_AP_UP;
  else {
      aPrintk("[pDevInfo=0x%lx][pSibEntry=0x%lx][StaState=0x%lx]\n", 
             (unsigned long)pDevInfo, (unsigned long)pSibEntry,
             (unsigned long)pSibEntry->staState);
  }
#endif
  // MAYBE!??? Scan channels if configured and enable radarDetect...
  //if (apCfgAutoChanSelectGet() == TRUE) {... (look into apAppInit.c)

  return 0;
}


void atheros_hw_shutdown(struct net_device *dev, int no_disable)
{
    return;
}


void atheros_hw_reset(struct net_device *dev)
{
    return;
}


struct net_device_stats *atheros_get_stats(struct net_device *dev)
{
  aPrintk("net_device_stats\n");
  atheros_local_info_t *local = (atheros_local_info_t *) dev->priv;
  return &local->stats;
}


/* stop all netif queues in use */
void atheros_netif_stop_queues(struct net_device *dev)
{
  atheros_local_info_t *local = (atheros_local_info_t *) dev->priv;
  
  if (local->dev)
    netif_stop_queue(local->dev);
}

static int atheros_open(struct net_device *dev)
{
  atheros_local_info_t *local = (atheros_local_info_t *) dev->priv;
  
  // SHOULD MIMC ar5hwcEndStart()

  if (local->func->dev_open && local->func->dev_open(local)) {
  	aPrintk("ERROR ATHEROS_OPEN....dev_open unassigned or failed!\n");
    return 1;
  }
  
  //  MOD_INC_USE_COUNT;
  
  if (!local->dev_enabled && atheros_hw_enable(dev)) {
    aPrintk("%s: could not enable MAC port !!!!! \n",dev->name);
    atheros_close(dev);
    return 1;
  }
  local->dev_enabled = 1;
  
  netif_device_attach(dev);
  netif_start_queue(dev);
  
  aPrintk("Netif attached and netif queue started!\n");

  return 0;
}


static int atheros_close(struct net_device *dev)
{
  // SHOULD MIMC ar5hwcEndStop()

  atheros_local_info_t *local = (atheros_local_info_t *) dev->priv;
  
  aPrintk("%s: atheros_close\n", dev->name);
  
  if (local->func->dev_close && local->func->dev_close(local)) {
    aPrintk("Going into dev_close !!!!!!!!!!1 \n");
    return 0;
  }
  
  if (disable_on_close) {
    local->dev_enabled = 0;
    atheros_hw_shutdown(dev, 0);
	aPrintk("Shutting down atheros HW **********\n");
  }
  
  if (netif_running(dev)) {
    netif_stop_queue(dev);
    netif_device_detach(dev);
	aPrintk("STOPPING NETIF queues and detaching...\n");
  }
  
  //  MOD_DEC_USE_COUNT;
  
  return 0;
}



static int atheros_change_mtu(struct net_device *dev, int new_mtu)
{
  if (new_mtu < ATHEROS_MIN_MTU || new_mtu > ATHEROS_MAX_MTU)
    return -EINVAL;
  
  dev->mtu = new_mtu;
  return 0;
}







enum {
  ATHEROS_RX_MONITOR, ATHEROS_RX_MGMT, ATHEROS_RX_NON_ASSOC,
  ATHEROS_RX_NULLFUNC_ACK, ATHEROS_RX_RAW
};

int atheros_80211_header_parse(struct sk_buff *skb, unsigned char *haddr)
{
  memcpy(haddr, skb->mac.raw + 10, ETH_ALEN); /* addr2 */
  return ETH_ALEN;
}



int atheros_rx_monitor_mode(struct net_device *);




static int atheros_tx_null(struct sk_buff *skb, struct net_device *dev){

    dev_kfree_skb(skb);
    return 0;
}

static void atheros_setup_dev(struct net_device *dev, atheros_local_info_t *local,
			     int main_dev)
{
  ether_setup(dev); //in drivers/net/net_init.c
  
  /* kernel callbacks */
  dev->get_stats = atheros_get_stats;
#ifdef WIRELESS_EXT
#ifdef FADI_TEMP // specific to H/W receives
  dev->get_wireless_stats = main_dev ? atheros_get_wireless_stats : NULL;
#else
  dev->get_wireless_stats = NULL;
#endif
#endif
  dev->open = atheros_open;
  dev->stop = atheros_close;
  dev->hard_start_xmit = atheros_tx_null;
  //  dev->hard_start_xmit = NULL;
#ifdef HAVE_MULTICAST
  /* FIX: to be implemented as soon as Prism2 supports GroupAddresses
   * and correct documentation is available */
  
  /* dev->set_multicast_list = atheros_set_multicast_list; */
#endif
#ifdef HAVE_PRIVATE_IOCTL
#ifdef FADI_TEMP
  dev->do_ioctl = main_dev ? atheros_ioctl : NULL;
#else
  dev->do_ioctl = NULL;
#endif
#endif
#ifdef HAVE_CHANGE_MTU
  dev->change_mtu = atheros_change_mtu;
#endif
#ifdef HAVE_TX_TIMEOUT
  dev->tx_timeout = NULL;
  dev->watchdog_timeo = TX_TIMEOUT;
#endif
  
  dev->mtu = mtu;
  
  netif_stop_queue(dev);
}




/* Called only from hardware IRQ */
void atheros_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  struct net_device    *dev     = (struct net_device *) dev_id;
  atheros_local_info_t *local   = (atheros_local_info_t *) dev->priv;
  WLAN_DEV_INFO * pdevInfo      = (WLAN_DEV_INFO *)local->pWlanDev;
  HAL_INT_TYPE    isrValue, unmaskedIsrValue = 0, disableMask = 0;
  A_UINT32        now, delta;

#if defined(PCI_INTERFACE)
    if (!halIsInterruptPending(pdevInfo)) {
        aPrintk("ISR: no interrupt pending.\n");
        return;
    }
#endif

    /* Find out which interrupts are pending, and deassert the interrupt line */
    isrValue = halGetInterrupts(pdevInfo, &unmaskedIsrValue);

    if (pdevInfo->NicResetInProgress == TRUE) {
        aPrintk("ISR: Inside of a NicReset ALREADY\n");
        return;
    }

    ASSERT(isrValue != HAL_INT_NOCARD);

    if ((isrValue & HAL_INT_RXPHY) && pdevInfo->staConfig.radarDetectEnable) {
        /* radar signal ? */
        if (tsfIndex < TSF_ARRAY_SIZE) {
            halGetTsf(pdevInfo, &tsfStamp[tsfIndex++]);
	    }

        /*
         * don't record it here. Better not to manipulate events array
         * from interrupt handler, it will cause mutual exclusion problems.
         */
    }
	/*
     * On the ar5211 RXORN is not fatal.  It is acceptable to see
     * this message occasionally, maybe once an hour.  If it is
     * occuring more often than that the system is not giving us
     * enough write bandwidth and there is a system level problem.
     *
     * When this happens, it can repeatedly interrupt until the condition
     * clears.  To handle this cleanly with one message, and w/o spending
     * undue time in the ISR we disable the interrupt and re-enable it in
     * the receive processing job.
     */
    if (isrValue & HAL_INT_RXORN) {
        printk("RXORN: RxDP 0x%lx (%ld)\n", halGetRxDP(pdevInfo),
               ++pdevInfo->rxOrnCnt);       
        disableMask |= HAL_INT_RXORN;
    }

    /* For all fatal interrupts, we reset the HW */
    if (isrValue & HAL_INT_FATAL) {
        logMsg("Fatal HW Interrupt: Resetting\n");
        resetwlan(pdevInfo, dev);
        return;
    }

    /* EOL error handling -- disable receive at the PCU */
    if (isrValue & HAL_INT_RXEOL) {
#ifdef RXEOL_WAR
        aPrintk("ISR: SUPPOSED TO PANIC: apPanic where from rxEOL?\n");
#endif
        /* rxEOL is non fatal - just keep stats; also prevent it from
         * repeatedly interrupting the processor by diabling RXEOL here
         * and re-enabling in receive processing job
         */
        ++pdevInfo->rxEOLcount;
        disableMask |= HAL_INT_RXEOL;
        printk("ISR: wlan%ld: got rxEOL\n", pdevInfo->devno);
    }

    if (isrValue & HAL_INT_TXURN) {
        pdevInfo->txUrn++;
        now   = A_MS_TICKGET();
        if (now > pdevInfo->txUrnBurstMs) 
            delta = DELTA(now, pdevInfo->txUrnBurstMs); /* timer wraps in 828 days */
        else
            delta = DELTA(pdevInfo->txUrnBurstMs, now); /* timer wraps in 828 days */

        /*  If we get too many in a short time then we reset. */
        if (delta < TXURN_PERIOD_ALLOW) {
            pdevInfo->txUrnBurst++;
            if (pdevInfo->txUrnBurst > TXURN_COUNT_ALLOW) {
                logMsg("ISR: %ld TXURNs received in a burst. Resetting... TXDP = 0x%lx\n",
                       pdevInfo->txUrnBurst, halGetTxDP(pdevInfo));
                printk("ISR: %ld TXURNs received in a burst. Resetting... TXDP = 0x%lx\n",
                       pdevInfo->txUrnBurst, halGetTxDP(pdevInfo));
                pdevInfo->txUrnBurstMs = now;
                pdevInfo->txUrnBurst   = 0;
                resetwlan(pdevInfo, dev);
                return;
            }
        } else {
            pdevInfo->txUrnBurstMs = now;
            pdevInfo->txUrnBurst   = 1;
        }

        /*
         *  Allow a few TXURNs per TXURN_COUNT_ALLOW.  This can
         *  happen sometimes on the 824x platform, and it is ok.
         *  If your platform is seeing this message, the IO system's
         *  performance needs to be debugged.
         */
        if (pdevInfo->txUrnBurst >= 2) {
            printk("ISR: TXURN #%ld (%ld) at %ld\n", pdevInfo->txUrnBurst,
                   pdevInfo->txUrn, now);
        }
    }

    if (isrValue & HAL_INT_BNR) {
#ifdef DEBUG
        logMsg("BNR (Beacon Not Ready) Interrupt: Informational...\n");
#endif
    }

    if (isrValue & HAL_INT_SWBA) {
        wlanApBeaconGen(pdevInfo);
        beaconNum++;
    }

    if (isrValue & HAL_INT_RX) {
        aPrintk("adding processRxInterrupt() job...\n");
        processRxInterrupt(pdevInfo);
        // we want to disable at start of processRxInterrupt and re-enable at end
#if FADI
        disableMask      |= HAL_INT_RX;
        unmaskedIsrValue &= ~HAL_INT_RX;
#endif
    }

    if (isrValue & HAL_INT_TX) {
        aPrintk("adding processTxInterrupt() job...\n");
        processTxInterrupt(pdevInfo);

        // we want to disable at start of processRxInterrupt and re-enable at end
#if FADI

        disableMask      |= HAL_INT_TX;
        unmaskedIsrValue &= ~HAL_INT_TX;
#endif    
    }

    if (isrValue & HAL_INT_MIB) {
        /*
         * The on-chip MIB-related counters have overflown.  We read the
         * registers here and add their values to our running totals.
         */
        halUpdateMibCounters(pdevInfo, TRUE);
    }

    if (disableMask) {
        halDisableInterrupts(pdevInfo, disableMask);
    }

    /*
     * To avoid missed interrupts and, therefore, stranded frames, we keep
     * track of TX and RX interrupts that are set in the ISR, but didn't
     * actually assert an interrupt because they are masked.
     */
    pdevInfo->globISRReg |= unmaskedIsrValue & (HAL_INT_RX | HAL_INT_TX);
}



extern struct pci_device_id atheros_pci_id_table[] __devinitdata;

atheros_local_info_t *atheros_init_local_data(struct atheros_helper_functions *funcs,
				     int card_idx, unsigned long mem_base)
{
  atheros_local_info_t *local;
  int 		 len, i, index=0;
  A_UINT32   basicRateCount = 0;
  A_UINT32   rate;
  //  char mySsid[ELE_SSID_SIZE] = "asap"; //DEFAULT_SSID


  WLAN_DEV_INFO * pWlanaDev;
  SIB_ENTRY 	* pWlanaSibEntry;
  BSSDESCR 		* pWlanaBssDescr;
  SCAN_INFO 	* pWlanaScanInfo;
  struct cservInfoStruct * pWlanaCservInfo;
  
  WLAN_DEV_INFO * pdevInfo = NULL;
  A_UINT16        opModes;
  
  if (funcs == NULL)
    return NULL;
  
  /**/int localSize = A_ROUNDUP(sizeof(atheros_local_info_t), CACHE_LINE_SIZE);
  local = kmalloc(localSize, GFP_KERNEL);
  if (local == NULL)
    return NULL;
  memset(local, 0, localSize);

  /**************************************************/
  /* AR5210/5211 specific initialization, see ar5hwEnd.c */
  /**************************************************/

  /* allocate the device structure(s), all memory usage here is about 16KB */
  pWlanaDev        = (WLAN_DEV_INFO *)A_DRIVER_MALLOC((sizeof(WLAN_DEV_INFO) + 100), 0, 0);
  pWlanaSibEntry   = (SIB_ENTRY *)A_DRIVER_MALLOC(sizeof(SIB_ENTRY),0,0);
  pWlanaBssDescr   = (BSSDESCR *)A_DRIVER_MALLOC(sizeof(BSSDESCR),0,0);
  pWlanaScanInfo   = (SCAN_INFO *)A_DRIVER_MALLOC(sizeof(SCAN_INFO),0,0);
  pWlanaCservInfo  = (struct cservInfoStruct *)A_DRIVER_MALLOC(sizeof(struct cservInfoStruct),0,0);
  

  if ( (NULL == pWlanaDev) 		 || (NULL == pWlanaBssDescr)   || 
	   (NULL == pWlanaSibEntry)  || (NULL == pWlanaScanInfo)   || 
	   (NULL == pWlanaCservInfo) 								) {
      aPrintk("NOTE: Could not alloc memory for WLAN_DEV_INFO device structures\n");
      goto fail;
  }
  else {
	  // zero all these structs out
	  A_MEM_ZERO(pWlanaDev,(sizeof(WLAN_DEV_INFO) + 100));
      A_MEM_ZERO(pWlanaSibEntry,sizeof(SIB_ENTRY));
      A_MEM_ZERO(pWlanaBssDescr,sizeof(BSSDESCR));
      A_MEM_ZERO(pWlanaScanInfo,sizeof(SCAN_INFO));
      A_MEM_ZERO(pWlanaCservInfo,sizeof(struct cservInfoStruct));
  }

  pdevInfo			    = pWlanaDev;
  pdevInfo->localSta    = pWlanaSibEntry;
  pdevInfo->bssDescr    = pWlanaBssDescr;
  pdevInfo->pscanInfo   = pWlanaScanInfo;
  pdevInfo->pCsInfo     = pWlanaCservInfo;
  
  //pdevInfo->pHwFunc, pdevInfo->pHalInfo in halAttach will be assigned to const struct
  //pdevInfo->pdkInfo     = (struct dkDevInfo *)A_DRIVER_MALLOC(sizeof(struct dkDevInfo),0,0);
  //pdevInfo->pSibTable   = (struct sibTable *)A_DRIVER_MALLOC(sizeof(struct sibTable),0,0);
  //pdevInfo->pBeaconInfo, allocated later in wlanIeaconInit() called from atheros_hw_config()
    
  pdevInfo->pOSHandle = NULL;
  // the following memory bufs/descriptors should not be used since we're using
  // the intersil's model of memory/skb buffer usage
  pdevInfo->pdkInfo			= NULL;
  /*
  pdevInfo->pDescMemory 	= NULL;       
  pdevInfo->pLargebufMemory = NULL;   
  pdevInfo->pSmallbufMemory = NULL;   
  pdevInfo->pCryptBuf 		= NULL;         
  */
  pdevInfo->pQueueCurrent 	= NULL;     
  pdevInfo->lastTxFrameDesc = NULL;
  pdevInfo->holdingTxDesc 	= NULL;
  pdevInfo->holdingAltDesc 	= NULL;
  pdevInfo->TxTailDesc 		= NULL;
  pdevInfo->altTailDesc 	= NULL;
  pdevInfo->pPSPollDesc 	= NULL;       
  pdevInfo->powerChangeTxDesc = NULL; 

  // save this devInfo to the global wlan dev pointer
  pwlanDevInfo = pdevInfo;

  // we are not using any of these globals now!! - fadi
  //globDrvInfo.pDevInfoArray[pdevInfo->devno] = pdevInfo;
  //globDrvInfo.devCount = 1;

#ifdef DEBUG
  aPrintk("pdevInfo = %p, pOsHandle = %x\n", pdevInfo, (unsigned int)pdevInfo->pOSHandle);
#endif

  /* Get parameters from the BSP (hardcoded in this case) and fill in the device structure */
  pdevInfo->devno         = 0;
  pdevInfo->baseAddress   = mem_base;
//fadi #ifdef BUILD_AP
  pdevInfo->ivec          = 20; //using IRQ=20 for the atheros chip slot
//fadi #endif

#ifdef DEBUG
  aPrintk("Base address = %lx\n", mem_base);
#endif

  /***********************/
  /*  apInit in apops.c	 */
  /***********************/

  /*
   * The pciInfo struct needs to be filled in prior to calling halAttach().
   * halAttach() needs to be called prior to calling apCfgInit() as it may
   * use HAL functions.
   */
  pdevInfo->pciInfo.VendorID          = 0x168c;

#ifdef BUILD_AR5210
  pdevInfo->pciInfo.DeviceID          = DEV_AR5210_AP;
#else //AR5211
  pdevInfo->pciInfo.DeviceID          = DEV_AR5212; 
#endif

  pdevInfo->pciInfo.SubVendorID       = 0;
  pdevInfo->pciInfo.SubVendorDeviceID = 0;
  pdevInfo->pciInfo.ChipRevision      = 0x0;

  // halAttach will call hwAttach which initializes the hardware
  if (halAttach(pdevInfo) != A_OK) {
        aPrintk("Error initializing wireless software state.\n");
        goto fail;
    }

  /******************/
  /*   STA CONFIG   */
  /******************/

  /* halAttached initialized macPermAddr.  Update other users. */
  A_MACADDR_COPY(&pdevInfo->staConfig.macPermAddr,
                 &pdevInfo->staConfig.macAddr);
  A_MACADDR_COPY(&pdevInfo->staConfig.macPermAddr,
                 &pdevInfo->staConfig.bssId);

  // will need better default values! For now, use default values apCfgDefaultSet()
  // as used by Atheros manufacturer (in cli.c)
  pdevInfo->staConfig.defaultFragThreshold    = DEFAULT_FRAG_THRESHOLD;
  pdevInfo->staConfig.defaultDot11RTSThreshold= DEFAULT_RTS_THRESHOLD;
  pdevInfo->staConfig.rateCtrlEnable          = TRUE;
  pdevInfo->staConfig.swretryEnabled          = FALSE; //TRUE;
  pdevInfo->staConfig.swRetryMaxRetries       = 0; //DEFAULT_SWRETRY_MAXRETRIES;
  pdevInfo->staConfig.limitTxRetries          = DEFAULT_LIMIT_TXRETRIES;
  pdevInfo->staConfig.psQueueMaxDepth         = DEFAULT_PS_QDEPTH;
  pdevInfo->staConfig.calibrationTime         = DEFAULT_CALIBRATION_TIME;
  pdevInfo->staConfig.tpScale                 = DEFAULT_TXPOWER;
  pdevInfo->staConfig.QosEnabled              = FALSE;
  pdevInfo->staConfig.cwmin                   = DEFAULT_CWMIN;
  pdevInfo->staConfig.aifs                    = DEFAULT_AIFS;
  pdevInfo->staConfig.countryCode             = CTRY_UNITED_STATES; //CTRY_DEBUG; 
  pdevInfo->staConfig.NetBand				  = MODE_SELECT_11A;
  pdevInfo->staConfig.privacyOptionImplemented = 1;
  pdevInfo->staConfig.maxTransmitMsduLifetime = DEFAULT_MAX_TRANSMIT_MSDU_LIFETIME;
  pdevInfo->staConfig.maxReceiveLifeTime 	  = DEFAULT_MAX_RECEIVE_LIFETIME;
  pdevInfo->staConfig.beaconInterval		  = DEFAULT_BEACON_PERIOD;
  pdevInfo->staConfig.multiDomainCapEnabled   = FALSE;

  // station environment initialization, again defaults (from locatsta.c)
  // assuming INFRASTRUCTURE_BSS
  pdevInfo->staConfig.capInfo.cfPollable 	 = 0;  // CF poll ability of STA
  pdevInfo->staConfig.capInfo.cfPollReq 	 = 0;  // CF poll request
  pdevInfo->staConfig.capInfo.privacy 		 = 0;  // no wep required
  pdevInfo->staConfig.capInfo.shortPreamble = 0;  // short preamble allowed (a 11B mode!)
  pdevInfo->staConfig.capInfo.ess           = 1;  // Set for ESS type network
  pdevInfo->staConfig.capInfo.ibss          = 0;  // Set for IBSS type network

  // initialize rate set to a default value
  // CCK element has only 4 rates.
  pdevInfo->staConfig.cckRateSet.elementID = ELE_SUPPORTED_RATES;
  pdevInfo->staConfig.cckRateSet.length = (A_UINT8) globalRateTable11b.rateCount;
  for (index = 0; index < pdevInfo->staConfig.cckRateSet.length; index++ )
	pdevInfo->staConfig.cckRateSet.rates[index] = globalRateTable11b.info[index].dot11Rate;
    
  pdevInfo->staConfig.ofdmRateSet.elementID = ELE_SUPPORTED_RATES;
  pdevInfo->staConfig.ofdmRateSet.length = (A_UINT8) globalRateTable11a.rateCount;
  for (index = 0; index < globalRateTable11a.rateCount; index++)
    pdevInfo->staConfig.ofdmRateSet.rates[index] = globalRateTable11a.info[index].dot11Rate;
    
  pdevInfo->staConfig.pRateTable = &globalRateTable11a;
        
  // these security values are based on apCfgSecurityInit() routines setup in ap/apcfg.c
  pdevInfo->staConfig.bssType 				  = INFRASTRUCTURE_BSS;
  pdevInfo->staConfig.privacyInvoked 		  = FALSE;
  pdevInfo->staConfig.authAllowOpen 		  = FALSE;
  pdevInfo->staConfig.authAllowShared 		  = FALSE;
  pdevInfo->staConfig.enableACL 		   	  = FALSE;
  pdevInfo->staConfig.excludeUnlisted 		  = FALSE;
  pdevInfo->staConfig.dot1xMode 		      = DEFAULT_DOT1X_MODE; //key in flash!?
  pdevInfo->staConfig.encryptionAlg			  = DEFAULT_CIPHER_SELECTION; // which is wep
  pdevInfo->staConfig.aesVlanKey 			  = KEYINDEX_INVALID;
  pdevInfo->staConfig.defaultKey			  = KEYINDEX_INVALID;
  pdevInfo->staConfig.ssid.elementID 		  = ELE_SSID;
  pdevInfo->staConfig.ssid.length			  = DEFAULT_SSID_LEN; // asap=4 + \0
  //pdevInfo->staConfig.ssid.ssid	 		  = DEFAULT_SSID = "asap"
#if FADI
  A_DRIVER_BCOPY (mySsid, pdevInfo->staConfig.ssid.ssid, 
				  (pdevInfo->staConfig.ssid.length-1));
  pdevInfo->staConfig.ssid.ssid[pdevInfo->staConfig.ssid.length]='\0';
#else /* FADI */
  strncpy(pdevInfo->staConfig.ssid.ssid, DEFAULT_SSID, DEFAULT_SSID_LEN); //"asap" = 4
  pdevInfo->staConfig.ssid.ssid[pdevInfo->staConfig.ssid.length]='\0';
#endif /* FADI */

  mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ofdmRateSet, 
				  (INFO_ELEMENT *)&pdevInfo->staConfig.operationalRateSet);
  
  /**********************/
  /*   BSSDESC CONFIG   */
  /**********************/

  /* Build a BSS Description set
   * Normally this step is necessary only for an access point. The station
   * builds it from the process of scanning. (also see localsta.c for info)
   */
  
  /* clear ucse & mcse set */
  A_MEM_ZERO(&pdevInfo->bssDescr->bssUcseSet, sizeof(UCSE_SET));
  A_MEM_ZERO(&pdevInfo->bssDescr->bssMcseSet, sizeof(MCSE_SET));

//fadi  pdevInfo->bssDescr->bssUcseSet.elementID   = ELE_UCSE_SET; // unicast cipher suite
//fadi  pdevInfo->bssDescr->bssMcseSet.elementID   = ELE_MCSE_SET; // multicast
//fadi  pdevInfo->bssDescr->bssUcseSet.length      = ELE_UCSE_SIZE;
//fadi  pdevInfo->bssDescr->bssMcseSet.length      = ELE_MCSE_SIZE;
//fadi  pdevInfo->bssDescr->bssUcseSet.uCiphers[0] = CSE_NULL; 
//fadi  pdevInfo->bssDescr->bssUcseSet.uCiphers[1] = CSE_NULL; //CSE_AES_OCB_REVA;
//fadi  pdevInfo->bssDescr->bssUcseSet.uCiphers[2] = CSE_NULL; //CSE_BASIC_WEP;
//fadi  pdevInfo->bssDescr->bssMcseSet.mCipher[0]  = CSE_NULL; //CSE_BASIC_WEP;

  /* Basic fields, addresses and sssid */
  mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ssid,
                  (INFO_ELEMENT *)&pdevInfo->bssDescr->ssid);
  //pdevInfo->bssDescr->capabilityInfo = pdevInfo->staConfig.capInfo;

  /* Assume that ap has same rate set as station */
  mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ofdmRateSet, 
				  (INFO_ELEMENT *)&pdevInfo->bssDescr->bssRateSet);
  
  /* set basic rate set */
  pdevInfo->bssDescr->bssBasicRateSet.elementID = ELE_SUPPORTED_RATES;
  for (index = 0; index < pdevInfo->staConfig.pRateTable->rateCount; index++) {
      if (pdevInfo->staConfig.pRateTable->info[index].dot11Rate & 0x80) {
          
		  pdevInfo->bssDescr->bssBasicRateSet.rates[basicRateCount++] =
              pdevInfo->staConfig.pRateTable->info[index].dot11Rate;
      }
  }
  pdevInfo->bssDescr->bssBasicRateSet.length = basicRateCount;

  A_MACADDR_COPY(&pdevInfo->staConfig.bssId, &pdevInfo->bssDescr->bssid);
  pdevInfo->bssDescr->bsstype				  = INFRASTRUCTURE_BSS; // NOT INDEPENDENT_BSS
  pdevInfo->bssDescr->ssid.elementID		  = ELE_SSID;
  pdevInfo->bssDescr->ssid.length 			  = DEFAULT_SSID_LEN; // "asap" = 4 + '\0'
  //pdevInfo->bssDescr->ssid.ssid				  = DEFAULT_SSID;
  strncpy(pdevInfo->bssDescr->ssid.ssid, DEFAULT_SSID, DEFAULT_SSID_LEN); //"asap" = 4
  pdevInfo->bssDescr->ssid.ssid[pdevInfo->bssDescr->ssid.length]='\0';
  pdevInfo->bssDescr->beaconInterval 		  = DEFAULT_BEACON_PERIOD;
  pdevInfo->bssDescr->DTIMPeriod    		  = DEFAULT_DTIM_PERIOD;
  pdevInfo->bssDescr->timOffset      		  = 0;
  pdevInfo->bssDescr->ibssParamSet.elementID  = ELE_IBSS_PARAM_SET;
  pdevInfo->bssDescr->ibssParamSet.length     = ELE_IBSS_PARAM_SIZE;
  pdevInfo->bssDescr->ibssParamSet.atimWindow = 0;

  /* set phy parameter set */
  pdevInfo->bssDescr->phyParamSet.aSlotTime 		  = 0;
  pdevInfo->bssDescr->phyParamSet.aSIFSTime 		  = 9;
  pdevInfo->bssDescr->phyParamSet.aCCATime 			  = 0;
  pdevInfo->bssDescr->phyParamSet.aRxTxTurnaroundTime = 0;
  pdevInfo->bssDescr->phyParamSet.aTxPLCPDelay 		  = 0;
  pdevInfo->bssDescr->phyParamSet.aRxPLCPDelay 		  = 0;
  pdevInfo->bssDescr->phyParamSet.aRxTxSwitchTime 	  = 0;
  pdevInfo->bssDescr->phyParamSet.aTxRampOnTime 	  = 0;
  pdevInfo->bssDescr->phyParamSet.aTxRampOffTime 	  = 0;
  pdevInfo->bssDescr->phyParamSet.aTxRFDelay 		  = 0;
  pdevInfo->bssDescr->phyParamSet.aRxRFDelay 		  = 0;
  pdevInfo->bssDescr->phyParamSet.aAirPropagationTime = 0;
  pdevInfo->bssDescr->phyParamSet.aMACProcessingDelay = 0;
  pdevInfo->bssDescr->phyParamSet.aPreambleLength 	  = 0;
  pdevInfo->bssDescr->phyParamSet.aPLCPHeaderLength   = 0;
  pdevInfo->bssDescr->phyParamSet.aMPDUDurationFactor = 0;
  pdevInfo->bssDescr->phyParamSet.aMPDUMaxLength 	  = 0;
  pdevInfo->bssDescr->phyParamSet.aCWmin 			  = 0;
  pdevInfo->bssDescr->phyParamSet.aCWmax 			  = 0;

  /* Set capability info:
   * - cfPollable and cfPollReq fields are currently set to 0
   *   because our AP is not a point coordinator.
   * - privacy is set later if needed.
   */
  pdevInfo->bssDescr->capabilityInfo.cfPollable 	= 0;
  pdevInfo->bssDescr->capabilityInfo.cfPollReq 		= 0;
  pdevInfo->bssDescr->capabilityInfo.ess 			= 1;
  pdevInfo->bssDescr->capabilityInfo.ibss 			= 0;
  pdevInfo->bssDescr->capabilityInfo.privacy 		= 0;
  pdevInfo->bssDescr->capabilityInfo.shortPreamble  = 0;  // short preamble allowed (a 11B mode!)

  /***********************/
  /*   LocalSTA CONFIG   */
  /***********************/

  /* init the local STA sib */
#if FADI /* this code is not needed since we're using one SIB entry
            and do not plan on using statab.c code..so for our case, 
			simple malloc and init of one SIB_ENTRY at the start should
			do for now */
  pdevInfo->localSta = sibEntryAlloc(NULL, NULL, &pdevInfo->bssDescr->bssid);
  if (pdevInfo->localSta == NULL) {
      aPrintk("Initialization ERROR: sibEntryAlloc failed\n");
      goto fail;
  }
#endif /* FADI */

  pdevInfo->localSta->serviceType 					= WLAN_AP_SERVICE; // this sw runs on AP
  A_MACADDR_COPY(&pdevInfo->bssDescr->bssid, &pdevInfo->localSta->bssId);
  A_MACADDR_COPY(&pdevInfo->bssDescr->bssid, &pdevInfo->localSta->macAddr);
  pdevInfo->localSta->devNo							= 0;
  pdevInfo->localSta->staState 						= STATE_QUIET;
  pdevInfo->localSta->transState 					= STATE_AP_UP;
  pdevInfo->localSta->privacyInvoked 		  		= pdevInfo->staConfig.privacyInvoked;
  pdevInfo->localSta->antTx 						= 0;
  pdevInfo->localSta->capInfo	 					= pdevInfo->bssDescr->capabilityInfo;

  mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->bssRateSet, 
				  (INFO_ELEMENT *)&pdevInfo->localSta->operationalRateSet);
  mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->bssBasicRateSet, 
				  (INFO_ELEMENT *)&pdevInfo->localSta->basicRateSet);
  pdevInfo->localSta->defaultRateIndex 				= 0;
  pdevInfo->localSta->basicRateIndex   				= 0;

  mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->bssUcseSet,
                  (INFO_ELEMENT *)&pdevInfo->localSta->ucseSet);

  mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->bssMcseSet,
                  (INFO_ELEMENT *)&pdevInfo->localSta->mcseSet);

  // DEFAULT_DATA_RATE = "best" isntead of 56, experiemental
  //rateVar = strtoul(DEFAULT_DATA_RATE, &pEnd, 10); 
  //if (rateVar) rate = (A_UINT32)(rateVar * 1000);
  //else   	   rate = (A_UINT32)(CLI_MIN_DATA_RATE * 1000);
  // override previous formula as copied from apCfgDataRateValGet() in apcfg.c
  rate = (A_UINT32)(CLI_MAX_DATA_RATE * 1000); // try 6000 Kbps

  for (index = 0; index < pdevInfo->staConfig.pRateTable->rateCount; index++) {
    if (pdevInfo->staConfig.pRateTable->info[index].rateKbps == rate) {
        pdevInfo->localSta->currentOperationalRateIndex = index;
        break;
    }
  }

  pdevInfo->staConfig.defaultOfdmRateIndex = index;

  /* The input rate is not supported by our software */
  if (index >= pdevInfo->staConfig.pRateTable->rateCount) {
#ifdef DEBUG
    aPrintk("apInit ERROR: %ldMbps not supported by software\n",rate);
#endif
    goto fail;
  }

  /* Set usrFragThreshold from base config */
  pdevInfo->staConfig.userFragThreshold = 
	  pdevInfo->staConfig.defaultFragThreshold;

  /* SME Initialization not likely needed here, again since we're using SKBs
   * Initialize SME data structures.
   * wlanMlmeStartRequest() copies the data from our local BSSDESCR into
   * a newly-allocated one which hangs off of WLAN_DEV_INFO
   */

  pdevInfo->keyTable = NULL;
  //index = 0 ;
  //while (index < (CFG_MAX_SHARED_KEY+1)) // we have this many keys
  //    pdevInfo->staConfig.keys[index++] = NULL;
  
  // we're setting the wireless mode to 11A
  opModes =     MODE_SELECT_11A;

  if (A_OK != wlanInitChannelList(pdevInfo, &(pdevInfo->staConfig.pClist), 
					  pdevInfo->staConfig.countryCode, opModes, TRUE)) {
	  aPrintk("FAILED TO INITIALIZE CHANNEL LIST under this country code in 11A mode\n");
	  goto fail;
  }
  
  /* Setup the channels */
  pdevInfo->staConfig.pChannel = 
	  wlanGetChannelPtr(pdevInfo->staConfig.pClist, CHAN_11A_DEFAULT, opModes);


  /* Fix channel config */
  if (pdevInfo->staConfig.pChannel == NULL) {
      pdevInfo->staConfig.pChannel = &(pdevInfo->staConfig.pClist->chanArray[0]);
      aPrintk("Setting channel frequency to be a member of the configured regulatory domain and operational modes\n");
  }
  pdevInfo->staConfig.phwChannel = pdevInfo->staConfig.pChannel;
  pdevInfo->bssDescr->pChannel   = pdevInfo->staConfig.pChannel;
  pdevInfo->localSta->Channel 	 = pdevInfo->bssDescr->pChannel->channel;

  // skipping the pdevInfo->staConfig.xxxxxx channel initialization

  calBeaconCount = (pdevInfo->staConfig.calibrationTime * 1000) /
                     DEFAULT_BEACON_PERIOD;

  // ar5hwcParse() which initializes memory pool/params to those
  // passed by driver initString..if null, give defaults - our case
  if (A_OK != default_driver_mem_desc (pdevInfo)) {
      aPrintk ("*** ERROR DEFAULTING DRIVER MEM DESC SETTINGS !!\n");
      goto fail;
  }


  pdevInfo->NicResetInProgress    = FALSE;
  pdevInfo->hwResetCount          = 0;
  pdevInfo->txUrn                 = 0;
  pdevInfo->txUrnBurst            = 0;
  pdevInfo->txUrnBurstMs          = 0;
  pdevInfo->globISRReg            = 0;
  
  // halReset will call hwReset (i.e. ar5211/5210Reset() ) resetting the chip
  if (halReset(pdevInfo, pdevInfo->staConfig.pChannel, FALSE) != A_OK) {
      aPrintk("Error initializing wireless hardware.\n");
      goto fail;
  }

  /* Perform memory allocation/distribution */
  if (A_OK != init_driver_mem_pool(pdevInfo) ) {
      aPrintk ("*** ERROR INITIALIZAING DRIVER MEM POOL !!\n");
      goto fail;
  }

  // save our WLAN_DEV_INFO struct since most other drivers
  // use atheros_local_info struct to pass params
  local->pWlanDev = (WLAN_DEV_INFO*)pdevInfo;

  
  /****************************************************/
  /* local structs/variables  + memory initialization */
  /****************************************************/

    int localBusBufSize = A_ROUNDUP((WLAN_HDR_SIZE + MAX_WLAN_FRAME_SIZE), CACHE_LINE_SIZE);
    
    localBusBufSize = A_ROUNDUP(MAX_WLAN_BODY_SIZE, CACHE_LINE_SIZE);

    int localDevSize = A_ROUNDUP(((sizeof(struct net_device)) + ATHEROS_NETDEV_EXTRA),CACHE_LINE_SIZE);
    local->dev = kmalloc(localDevSize, GFP_KERNEL);
    if (local->dev == NULL)
        goto fail;
    memset(local->dev, 0, localDevSize);
    
    atheros_set_dev_name(local->dev, local->dev + 1);
  
    local->dev->priv = local;
    local->func = funcs;
    local->func->asap.tx_asap = ar521x_transmit;
    local->func->asap.rx_asap = NULL;
    local->func->asap.rx_monitor_asap = NULL;
    local->func->asap.set_channel = atheros_set_channel;
    local->func->asap.set_txpower = atheros_set_tx_power;
    local->func->asap.set_mode = atheros_set_ammode;
    local->func->asap.set_tim = atheros_set_tim;
    local->func->set_essid = atheros_set_essid;


    local->iw_mode = IW_MODE_INFRA;
    
    memcpy(local->dev->dev_addr, &pdevInfo->staConfig.macPermAddr, 6);
    strcpy(local->essid, "asap");
  
    spin_lock_init(&local->txfidlock);
    spin_lock_init(&local->cmdlock);
    spin_lock_init(&local->baplock);
    spin_lock_init(&local->wdslock);
  
    if (card_idx < 0 || card_idx >= MAX_PARM_DEVICES)
        card_idx = 0;
    local->card_idx = card_idx;
  
    i = essid[card_idx] == NULL ? 0 : card_idx;
    len = strlen(essid[i]);
    memcpy(local->essid, essid[i], len > MAX_SSID_LEN ? MAX_SSID_LEN : len);
    local->essid[MAX_SSID_LEN] = '\0';
    local->channel        = GET_INT_PARM(channel, card_idx);
    local->beacon_int     = GET_INT_PARM(beacon_int, card_idx);
    local->dtim_period    = GET_INT_PARM(dtim_period, card_idx);
    local->manual_retry_count = -1;
#if defined(ATHEROS_PCI) && defined(ATHEROS_BUS_MASTER)
    //	local->bus_master_threshold_rx = GET_INT_PARM(bus_master_threshold_rx,card_idx);
    //	local->bus_master_threshold_tx = GET_INT_PARM(bus_master_threshold_tx,card_idx);
#endif /* ATHEROS_PCI and ATHEROS_BUS_MASTER */

    skb_queue_head_init(&local->bridge_list);
  
  /* Initialize task queue structure for bridged packet handling */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
    local->bridge_queue.next = NULL;
#else
    INIT_LIST_HEAD(&local->bridge_queue.list);
#endif
    local->bridge_queue.sync = 0;
    local->bridge_queue.routine = NULL;
    local->bridge_queue.data = local;
  
    atheros_setup_dev(local->dev, local, 1);
  
    //local->dev->type = ARPHRD_IEEE80211;
    local->dev->hard_header_parse = NULL;
  
    return local;
  
fail:
    if (local->dev != NULL)
        kfree(local->dev);
    
    kfree(local);

    return NULL;
}


int atheros_init_dev(atheros_local_info_t *local)
{
  struct net_device *dev = local->dev;
  
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
  {
    int i = 0;
    do {
      sprintf(dev->name, "wlanA", i++);
    } while (dev_get(dev->name));
  }
#else
  memcpy(dev->name, "wlanA", 6);
#endif
  
  dev->type = ARPHRD_ETHER;
  dev->addr_len = 6;
  dev->hard_header_len = 14;
  dev->mtu = 4500; //mtu; //1600;
  
  dev->tx_queue_len = 1000; //100;
  
  if (register_netdev(dev)) {
    aPrintk(KERN_WARNING "register_netdev() failed!\n");
    return 1;
  }
  aPrintk(KERN_INFO "Registered netdevice %s\n", 
	 dev->name);
  

  skb_queue_head_init(&atheros_sap_readq);
  
  return 0;
}


void atheros_free_local_data(atheros_local_info_t *local)
{
  if (local == NULL)
    return;
  
  
  if (local->dev && local->dev->name && local->dev->name[0]) {
    unregister_netdev(local->dev);
    aPrintk(KERN_INFO "Netdevice %s unregistered\n",
	   local->dev->name);
    kfree(local->dev);
  }

  
  
  
  if (local->pda)
    kfree(local->pda);
  
  kfree(local);
  
}


void resetwlan(WLAN_DEV_INFO *pdevInfo, struct net_device *dev)
{
    if (NULL == dev) {
        aPrintk("ISR-RESET: device not yet attached\n");
        return;
    }

    aPrintk("ISR-RESET: inside of resetwlan...\n");
    

    /* If things look very bleak, then reboot */
    if (pdevInfo->localSta->stats.hwResetCount > RESET_COUNT_ALLOW) {
        printk("--------------------------------------------------------------\n");
        printk("ISR-RESET: resetwlan: reset too many times (%ld). Rebooting...\n",
               pdevInfo->localSta->stats.hwResetCount);
        printk("--------------------------------------------------------------\n");
        // we'll need a reboot func call here
    }

    if (pdevInfo->NicResetInProgress == TRUE) {
		aPrintk("ISR-RESET: !!! NIC already in reset process\n");
        return;
    }

	/* Queue up a reset job. */
    pdevInfo->NicResetInProgress = TRUE;
    pdevInfo->localSta->staState = STATE_AP_DOWN;
    aPrintk("ISR-RESET: setting staState -> down\n");
    halRxFilter(pdevInfo, 0, HAL_RX_FILTER_INIT);
    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);

    /* Clear the running flag */
    netif_stop_queue(dev);	    

    resetOnError (pdevInfo);

    aPrintk("ISR-RESET: RE-enabling GLOBAL INTS\n");

}

void resetOnError(WLAN_DEV_INFO *pdevInfo)
{
	aPrintk("ISR-RESET: Entering resetOnError #####\n");

    /* Keep track of number of resets */
    pdevInfo->localSta->stats.hwResetCount++;
    pdevInfo->hwResetCount++;

#ifdef DEBUG
    logMsg("ISR-RESET: Number of resets = %ld\n", pdevInfo->hwResetCount);
#endif

    // FN - beaconWdgCancel(); will need to cancel beacon watchdog
    if ((NULL == pdevInfo) || (NULL == pdevInfo->staConfig.pChannel)) {
        aPrintk("DEVICE NOT READY FOR RESETonERROR\n");
        return;
    }
    chipReset(pdevInfo, TRUE, pdevInfo->staConfig.pChannel);
    aPrintk("ISR-RESET: ### finished chipreset...\n");
    wlanBeaconFree(pdevInfo);
    aPrintk("ISR-RESET: ### finished wlanBeaconFree...\n");
    if (A_OK != wlanBeaconInit(pdevInfo, 0)) printk("FAILED TO RESET CHIP!\n");
    aPrintk("ISR-RESET: ### finished wlanBeaconInit...\n");
    // FN - beaconWdgStart(pdevInfo); will need to start beacon watchdog

#ifdef DEBUG
    logMsg("ISR-RESET: Finished resetting the wlan chip\n");
#endif

    /* Clear the reset flag */
    pdevInfo->NicResetInProgress = FALSE;

	aPrintk("ISR-RESET: ##### done resetOnError\n");
}

/*******************************************************************************
 * chipReset
 *
 * Reset chip and some selected sw state
 */
void 
chipReset(WLAN_DEV_INFO *pdevInfo, A_BOOL unicast, CHAN_VALUES *pChannel)
{
    ATHEROS_DESC    *pDesc;

	 aPrintk("ISR-RESET: Entering chipReset ++++++++\n");

     A_SEM_LOCK(pdevInfo->txQueue.qSem, WAIT_FOREVER);

    pdevInfo->globISRReg = 0;

    /* Disable interrupts and stop receive. */
    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    halStopPcuReceive(pdevInfo);

    if (halReset(pdevInfo, pChannel, FALSE) != A_OK) {
        A_SEM_UNLOCK(pdevInfo->txQueue.qSem);
        /* reboot the whole system */
        aPrintk("ISR-RESET: chipReset: Error resetting WLAN HW!  Rebooting...\n");
        aPrintk("ISR-RESET: need some server action here...\n");
        return;
    }

    /* do the needful for rates, etc. after switching baseband */
    if ((pChannel->channelFlags & CHANNEL_CCK) ^
        (pdevInfo->staConfig.phwChannel->channelFlags & CHANNEL_CCK))
    {
#ifdef RATECTRL_C
        drvBBChange(pdevInfo, pChannel);
        aPrintk("ISR-RESET: done drvBBChange...\n");
#endif
    }

    /* Clean up the ps, rx, tx, alt queue */
    A_SEM_LOCK(pdevInfo->localSta->txPSQueue.qSem, WAIT_FOREVER);
    pDesc = (pdevInfo->localSta->txPSQueue).pDescQueueHead;
    if (pDesc != NULL) {
        freeBuffandDescChain(pdevInfo, pDesc);
    }

    pdevInfo->numTxFramesPending = 0;

    pdevInfo->localSta->txPSQueue.pDescQueueHead = NULL;
    pdevInfo->localSta->txPSQueue.pDescQueueTail = NULL;
    A_SEM_UNLOCK(pdevInfo->localSta->txPSQueue.qSem);

    pDesc = (pdevInfo->txQueue).pDescQueueHead;
    if (pDesc != NULL) {
        freeBuffandDescChain(pdevInfo, pDesc);
    }

    pdevInfo->txQueue.pDescQueueHead = NULL;
    pdevInfo->txQueue.pDescQueueTail = NULL;

    pdevInfo->lastTxFrameDesc = NULL;
    A_SEM_UNLOCK(pdevInfo->txQueue.qSem);

    A_SEM_LOCK(pdevInfo->altQueue.qSem, WAIT_FOREVER);
    pDesc = pdevInfo->altQueue.pDescQueueHead;
    if (pDesc != NULL) {
        freeBuffandDescChain(pdevInfo, pDesc);
    }

    pdevInfo->altQueue.pDescQueueHead = NULL;
    pdevInfo->altQueue.pDescQueueTail = NULL;
    A_SEM_UNLOCK(pdevInfo->altQueue.qSem);

    A_PIPEFLUSH();

    /* Reset the receive queue */
    init_driver_queue_reset(pdevInfo);

    halEnableReceive(pdevInfo);
    halStartPcuReceive(pdevInfo);

    if (unicast) {
        halRxFilter(pdevInfo, HAL_RX_UCAST, HAL_RX_FILTER_SET);
    }

    halEnableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    pdevInfo->localSta->staState = STATE_AP_UP;
    
    /* Set the running flag */
    struct net_device *pNetDev = NULL;
    if (NULL != gPciDev) pNetDev = pci_get_drvdata(gPciDev);
    if (NULL != pNetDev) netif_start_queue(pNetDev); //IFF_UP | IFF_RUNNING
    aPrintk("ISR-RESET: setting staState -> UP\n");

    aPrintk("ISR-RESET: +++ Done with chipReset\n");
}

/*****************************************************************************
* default_driver_mem_desc - replaces atheros' parse the init string routine
*
* Fill in values in the driver control structure.
*
* The initialization string format is:
* "unit:pDmaBuf:dmaBufSize:txDescCount:rxDescCount:extraClCount:extraTxDescCount"
*
* .bS
* unit              Device unit number, a small integer
* pDmaBuf           Shared memory address
* dmaBufSize        Size of memory reserved
* txDescCount       Transmit descriptor count - depth of the tx queue; this
*                     generally determines the latency through the system,
*                     trading it off against the burst tolerance of the system
* rxDescCount       Receive descriptor count - depth of the rx queue; this is
*                     generally a function of the scheduling granularity of the
*                     system
* extraClCount      Extra data buffers for loan to devices outside like
*                     ethernet, local stack etc
* extraTxDescCount  Extra transmit descriptors for power save etc; these
*                     are suggested based on the memory availability
* .bE
*
* NULL (0) in these fields are used to specify corresponding default values
*
* If <pDmaBuf> is specified as NULL (0) or NONE (-1), the driver allocates
* the required DMA buffer area using malloc() (do not want cacheDmaMalloc()
* as may be uncached unless this is requested).
*
* Otherwise, <pDmaBuf> should contain the desired
* address, whose size is specified by <dmaBufSize> that is safe to use
* with the BSPs CACHE_DMA_*() routines.
*
* RETURNS: OK or ERROR for invalid arguments.
*/
A_STATUS default_driver_mem_desc (WLAN_DEV_INFO * pdevInfo)
{
    A_UINT16 txQueueDepth;
    
    /* Parse the initString or in our case, go with defaults */

    /* Unit number. */
    pdevInfo->devno = 0; // 1st wlan-11a device
    aPrintk("wlan%ld: got devno\n", pdevInfo->devno);

    /* Device DMA buffer address. */
    pdevInfo->pDmaBuf = NULL;
    aPrintk("wlan%ld: pDmaBuf = %p\n", pdevInfo->devno, pdevInfo->pDmaBuf);

    /* DMA buffer size. */
    pdevInfo->dmaBufSize = MAX_DMA_AREA; 
    aPrintk("wlan%ld: dmaBufSize = %d\n", pdevInfo->devno, pdevInfo->dmaBufSize);

    /* Transmit descriptor count. */
    txQueueDepth = AR_TX_QUEUE_DEPTH;
    /*
     * For fragmentation, in order to maintain the similar burst
     * tolerance to non-fragmented we need a deeper tx queue.  It
     * is acceptable for the system to have higher latency that a
     * deeper tx queue may bring.
     */
    if (pdevInfo->staConfig.defaultFragThreshold < 1500) {
        aPrintk ("Added more txQueueDepth to support small defaultFragThreshold\n");
        txQueueDepth += AR_TX_QUEUE_FRAG_DELTA; 
    }
    aPrintk("wlan%ld: txQueueDepth = %d\n", pdevInfo->devno, txQueueDepth);

    /* Receive descriptor count. */
    pdevInfo->rxQueueDepth = AR_RX_QUEUE_DEPTH;
    aPrintk("wlan%ld: rxDescCount = %d\n", pdevInfo->devno, pdevInfo->rxQueueDepth);

    /* Extra cluster count - buffers for loaning outside. */
    pdevInfo->maxLoanedCount = AR_MAX_LOANED_COUNT;
    aPrintk("wlan%ld: maxLoanedCount = %d\n", pdevInfo->devno, pdevInfo->maxLoanedCount);

    /* Extra transmit descriptor count - power save queue depth really. */
    pdevInfo->psQueueDepth = AR_PS_QUEUE_DEPTH;
    aPrintk("wlan%ld: psQueueDepth = %d\n", pdevInfo->devno, pdevInfo->psQueueDepth);

    /* we need one header buffer per transmitted frame/frag */
    pdevInfo->txHeaderDescCount = txQueueDepth;

    /* the free descriptors are used for data portion of each
     * tx frame, to be put on the receive queue, and each sme
     * frame transmitted;  sme frames transmitted always get
     * charged to the trasmit queue directly;  the rx queue is
     * always maintained at its constant depth;  thus the max
     * number of free descriptors available on the actual empty
     * descriptor queue at any time corresponds to the desired
     * limit on the transmit queue
     */
    pdevInfo->freeDescCount = txQueueDepth;

    /* for now */
    pdevInfo->numDefragLanes = DEFRAG_ENTRIES_DEFAULT;
                     
    return A_OK;
}

/******************************************************************************
* init_driver_mem_pool - initialize queues and netpool for the device
*
* Function Description: This function allocate the memory pool required by this
* driver. Also, it creates the software tx queue and software rx queue
*
* RETURNS: A_OK or A_ERROR.
*/
A_STATUS
init_driver_mem_pool(WLAN_DEV_INFO *pdevInfo)
{
    ATHEROS_DESC    *pDesc;
    A_UINT16        numDescriptors;
    char            *pMemBuf;
    A_UINT32        totalSize;
    A_UINT32        memSize = 0, clSize = 0, clMemSize  = 0;
    char            *memArea = NULL, *clMemArea = NULL;
    

#define BLK_SZ  32 // in vx, M_BLK_SZ & CL_BLK_SIZE = 32

    memSize = (pdevInfo->maxLoanedCount * 32) + 
              ((pdevInfo->rxQueueDepth + pdevInfo->maxLoanedCount) * 32);

#define DRV_ALIGN_ADDRESS(alignment, addr)  \
        MEM_ALIGN_ADDRESS(addr, alignment)
#define MEM_ALIGN_ADDRESS(addr, alignment)  \
        ((((A_UINT32)(addr)) + (((A_UINT32)(alignment))-1)) & ~(((A_UINT32)(alignment)) - 1))

    // pls note that since we're memory aligning this ptr, freeing it
    // will have to be aware of this alignment!!!

    A_UINT32 alignedMemSize = A_ROUNDUP(memSize, CACHE_LINE_SIZE);
    memArea = (char*)kmalloc(alignedMemSize, GFP_DMA); // used to be GFP_KERNEL
    printk("ALLOCATED MEMSIZE =  %p[size=%ld]\n", memArea, alignedMemSize);
    if (NULL != memArea) 
        memArea = (char*)DRV_ALIGN_ADDRESS(sizeof(long), (memArea)); 
    printk("AFTER MEMALIGN: new memArea is at %p\n", memArea);

    /**/memset(memArea, 0, alignedMemSize);

    pdevInfo->pMclMem = memArea;
    aPrintk("mblk (equivalent) memory location=0x%lx[size] = %ld\n", (A_UINT32)memArea, memSize);
    
    if (NULL == memArea) {
        logMsg("wlan%ld: Could not allocate mclBlk memory\n", pdevInfo->devno);
        goto arMemInitFail;
    }

    clSize    = A_ROUNDUP(AR_BUF_SIZE, CACHE_LINE_SIZE);
    clMemSize = 0;


    /* calculate total number of descriptors needed; this is
     * the number of free descriptors; plus each header buffer
     * for the scattered wlan header needs a descriptor; plus
     * descriptors for the receive ring; once they leave the
     * rx ring they are freed or charged to defrag/SME/transmit
     */
    numDescriptors = pdevInfo->freeDescCount + pdevInfo->txHeaderDescCount +
                     pdevInfo->psQueueDepth + pdevInfo->rxQueueDepth;

    totalSize = CACHE_LINE_SIZE + clMemSize;

    /* descriptor size needs to be multiple of a cache line; account
     * for descriptor memory while allowing a cache line round up from
     * cluster memory
     */
    ASSERT((AR_DESC_SIZE & (CACHE_LINE_SIZE - 1)) == 0);
    totalSize += (CACHE_LINE_SIZE + (AR_DESC_SIZE * numDescriptors));
    
    /* memory for header buffers */
    ASSERT((AR_HEADER_SIZE & (CACHE_LINE_SIZE - 1)) == 0);
    totalSize += (pdevInfo->txHeaderDescCount * AR_HEADER_SIZE);

    /* memory for power save buffers */
    totalSize += (pdevInfo->psQueueDepth * (A_ROUNDUP(AR_BUF_SIZE, CACHE_LINE_SIZE)));

    if (pdevInfo->pDmaBuf != NULL) {
        if (pdevInfo->dmaBufSize < totalSize) {
            logMsg("wlan%ld: not enough memory provided, need %ld got %d\n",
                    pdevInfo->devno, totalSize, pdevInfo->dmaBufSize);
            goto arMemInitFail;
        }
        aPrintk("WE SHOULD NOT BE IN HERE AS DMABUF = NULL\n");
    } else {
        /* Allocate the whole memory for cluster, rx and empty queue */
        /**/totalSize = A_ROUNDUP(totalSize, CACHE_LINE_SIZE);
        /* +CLS */
        pdevInfo->pDmaBuf = kmalloc(totalSize, GFP_DMA); // eligible for DMA

        if (NULL == pdevInfo->pDmaBuf) {
            goto arMemInitFail;
        }
        else 
            printk("-> JUST ALLOCATED %ldB for our DMA bufs(%p)\n", totalSize, pdevInfo->pDmaBuf);

        pdevInfo->dmaBufSize = totalSize;
        pdevInfo->flags     |= AR_BUF_LOCAL;
        logMsg("wlan%ld: DevMem allocated, pDmaBuf=%p\n", pdevInfo->devno, pdevInfo->pDmaBuf);
    }
    A_MEM_ZERO(pdevInfo->pDmaBuf, totalSize);

    /* setup memory for data clusters - align to 4 bytes ahead of a cacheline
     * boundary so that the data buffer falls on a cacheline boundary - vxworks
     * prepends a 4 byte overhead to each cluster (overhead only in vxWorks)
     */
    clMemArea = (char *)A_ROUNDUP((A_UINT32)pdevInfo->pDmaBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)clMemArea;
    aPrintk("This is the area where WIRELESS PKTS SHOULD BE RECEIVED &&&&&&\n");
    

    /* set up memory for descriptor queues - aligned to cacheline boundaries */
    pdevInfo->emptyDescQueue.pDescQueueHead = (ATHEROS_DESC *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)pdevInfo->emptyDescQueue.pDescQueueHead + (AR_DESC_SIZE * pdevInfo->freeDescCount);
    
    printk("-emptyDescQueue is @ %p [size = 0x%x][# = %d]\n", pdevInfo->emptyDescQueue.pDescQueueHead,
           (AR_DESC_SIZE * pdevInfo->freeDescCount), pdevInfo->freeDescCount);

    pdevInfo->txHeaderQueue.pDescQueueHead = (ATHEROS_DESC *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)pdevInfo->txHeaderQueue.pDescQueueHead + (AR_DESC_SIZE * pdevInfo->txHeaderDescCount);

    printk("-txHdrQueue is @ %p [size = 0x%x][# = %d]\n", pdevInfo->txHeaderQueue.pDescQueueHead,
           (AR_DESC_SIZE * pdevInfo->txHeaderDescCount), pdevInfo->txHeaderDescCount);


    pdevInfo->psBufferQueue.pDescQueueHead = (ATHEROS_DESC *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)pdevInfo->psBufferQueue.pDescQueueHead + (AR_DESC_SIZE * pdevInfo->psQueueDepth);

    printk("-psBufQueue is @ %p [size = 0x%x][# = %d]\n", pdevInfo->psBufferQueue.pDescQueueHead,
           (AR_DESC_SIZE * pdevInfo->psQueueDepth), pdevInfo->psQueueDepth);


    pdevInfo->rxQueue.pDescQueueHead = (ATHEROS_DESC *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    pMemBuf = (char *)pdevInfo->rxQueue.pDescQueueHead + (AR_DESC_SIZE * pdevInfo->rxQueueDepth);

    printk("-rxQueue is @ %p [size = 0x%x][# = %d]\n", pdevInfo->rxQueue.pDescQueueHead,
           (AR_DESC_SIZE * pdevInfo->rxQueueDepth), pdevInfo->rxQueueDepth);


    aPrintk ("wlan%ld: finished initializing memory pool\n", pdevInfo->devno);

    /* Create a tx descriptor queue */
    if (init_driver_queue_create(pdevInfo, &pdevInfo->emptyDescQueue, &pdevInfo->emptyDescQueue,
                          pdevInfo->freeDescCount, 0) != A_OK)
    {
        aPrintk("wlan%ld: Failed to create tx desc queue, %p\n", pdevInfo->devno,
                pdevInfo->emptyDescQueue.pDescQueueHead);
        goto arMemInitFail;
    }
    aPrintk ("wlan%ld: finished allocating tx desc queue\n", pdevInfo->devno);

    /* Create a header queue */
    if (init_driver_queue_create(pdevInfo, &pdevInfo->txHeaderQueue, &pdevInfo->txHeaderQueue,
                          pdevInfo->txHeaderDescCount, 0) != A_OK)
    {
        aPrintk("wlan%ld: Failed to create tx header queue, %p\n", pdevInfo->devno,
                pdevInfo->txHeaderQueue.pDescQueueHead);
        goto arMemInitFail;
    }
    pMemBuf = (char *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    for (pDesc = pdevInfo->txHeaderQueue.pDescQueueHead; pDesc; pDesc = pDesc->pNextVirtPtr) {
        pDesc->pBufferVirtPtr.ptr = pMemBuf;
        pDesc->bufferPhysPtr      = (A_UINT32)pMemBuf;
        pDesc->pOSDescPtr         = NULL;
        pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
        pMemBuf += AR_HEADER_SIZE;
    }
    aPrintk ("wlan%ld: finished allocating tx header queue\n", pdevInfo->devno);

    /* Create a header queue */
    if (init_driver_queue_create(pdevInfo, &pdevInfo->psBufferQueue, &pdevInfo->psBufferQueue,
                          pdevInfo->psQueueDepth, 0) != A_OK)
    {
        aPrintk("wlan%ld: Failed to create power save buffers queue, %p\n", pdevInfo->devno,
                pdevInfo->psBufferQueue.pDescQueueHead);
        goto arMemInitFail;
    }
#if FADI
    pMemBuf = (char *)A_ROUNDUP((A_UINT32)pMemBuf, CACHE_LINE_SIZE);
    for (pDesc = pdevInfo->psBufferQueue.pDescQueueHead; pDesc; pDesc = pDesc->pNextVirtPtr) {
        pDesc->pBufferVirtPtr.ptr = pMemBuf;
        pDesc->bufferPhysPtr      = (A_UINT32)pMemBuf;
        pDesc->pOSDescPtr         = NULL;
        pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
        pMemBuf += AR_BUF_SIZE;
    }
    aPrintk ("wlan%ld: finished allocating power save buffer queue\n", pdevInfo->devno);
#else
    for (pDesc = pdevInfo->psBufferQueue.pDescQueueHead; pDesc; pDesc = pDesc->pNextVirtPtr) {
            pDesc->pBufferVirtPtr.ptr = NULL;
            pDesc->bufferPhysPtr      = 0;
            pDesc->pOSDescPtr         = NULL;
            pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
    }

#endif //FADI

    /* Create a Rx queue */
    if (init_driver_queue_create(pdevInfo, &pdevInfo->rxQueue, &pdevInfo->emptyDescQueue,
                          pdevInfo->rxQueueDepth, AR_BUF_SIZE) != A_OK)
    {
        aPrintk("wlan%ld: Failed to create rx queue, %p\n", pdevInfo->devno,
                pdevInfo->rxQueue.pDescQueueHead);
        goto arMemInitFail;
    }
    aPrintk ("wlan%ld: finished allocating rx queue\n", pdevInfo->devno);

    ASSERT(pMemBuf <= (pdevInfo->pDmaBuf + pdevInfo->dmaBufSize));
    
    /* Initialize the rxQueue, txQueue and altQueue */

    init_driver_queue_reset(pdevInfo);

    pdevInfo->txQueue.pDescQueueHead = NULL;
    pdevInfo->txQueue.pDescQueueTail = NULL;
    A_SEM_INIT(pdevInfo->txQueue.qSem, global_sem_key++, CREATE_UNLOCKED); //SEM_Q_PRIORITY
    
    pdevInfo->altQueue.pDescQueueHead = NULL;
    pdevInfo->altQueue.pDescQueueTail = NULL;
    A_SEM_INIT(pdevInfo->altQueue.qSem, global_sem_key++, CREATE_UNLOCKED); //SEM_Q_PRIORITY
    
    pdevInfo->lastTxFrameDesc = NULL;

    aPrintk("wlan%ld: Memory setup complete\n", pdevInfo->devno);
    return A_OK;

arMemInitFail:
    
    aPrintk("OOOOOOOooops, entering arMemInitFail...\n");
    /* Clean all the queues, tx, rx, alt, tx and header queues*/
    init_driver_queue_delete(&pdevInfo->altQueue);
    init_driver_queue_delete(&pdevInfo->txQueue);
    init_driver_queue_delete(&pdevInfo->rxQueue);
    init_driver_queue_delete(&pdevInfo->psBufferQueue);
    init_driver_queue_delete(&pdevInfo->txHeaderQueue);
    init_driver_queue_delete(&pdevInfo->emptyDescQueue);

    /* free memory allocated for clusters and Free/Rx Queues */
    if ((pdevInfo->flags & AR_BUF_LOCAL) && (pdevInfo->pDmaBuf != NULL)) {
#ifdef VX_CACHE_DMA_MALLOC
        cacheDmaFree(pdevInfo->pDmaBuf);
#else
        kfree(pdevInfo->pDmaBuf);
#endif
    }

    /* free memory allocated for mBlk-clBlks */
    if (memArea != NULL) {
        kfree(memArea);
    }

    logMsg("wlan%ld: Memory Init failed\n", pdevInfo->devno);

    return A_ERROR;
}

/******************************************************************************
* init_driver_queue_create - create a link list queue of descriptors
*
* This routine allocates a descriptor buffer array aligned  on a word
* boundary, initializes and links the array.
*
* RETURNS: OK or ERROR
*/
A_STATUS
init_driver_queue_create(WLAN_DEV_INFO *pdevInfo, QUEUE_DETAILS *pQueue, QUEUE_DETAILS *freeQueue,
                  int count, int bufSize)
{
    ATHEROS_DESC *pDesc;
    int          i;

    /* Initialize queue data */
    pQueue->pDescQueueTail = pQueue->pDescQueueHead + (count - 1);

    A_SEM_INIT (pQueue->qSem, global_sem_key++, CREATE_UNLOCKED); // SEM_Q_PRIORITY
    /* Initialize each descriptor */
    for (i = 0, pDesc = pQueue->pDescQueueHead; i < count; ++i, pDesc = pDesc->pNextVirtPtr) {
        pDesc->nextPhysPtr  = A_DESC_V2P((A_UINT32)pDesc + sizeof(ATHEROS_DESC));
        pDesc->pNextVirtPtr = pDesc + 1;
        pDesc->thisPhysPtr  = A_DESC_V2P(pDesc);
        pDesc->freeQueue    = freeQueue;

        if (bufSize) {
            if (hwAllocatePhysBuffer(pdevInfo, bufSize, &pDesc->bufferPhysPtr,
                &pDesc->pBufferVirtPtr.ptr, &pDesc->pOSDescPtr) != A_OK)
            {
                logMsg("Error: could not allocate buffer for descriptor\n");
                return A_ERROR;
            }
            ASSERT(((A_UINT32)pDesc->pBufferVirtPtr.ptr & (CACHE_LINE_SIZE - 1)) == 0);
            pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
        }
    }
    pQueue->pDescQueueTail->nextPhysPtr  = 0;
    pQueue->pDescQueueTail->pNextVirtPtr = NULL;

    return A_OK;
}

/******************************************************************************
* init_driver_queue_delete - Delete a queue
*
* This routine clean all the queue field in QUEUE_DETAIL, the memeory clean is
* not done here. This function and ar5hwcQueueCreate are only called when we load
* the driver.
*
* RETURNS: NONE
*/
void
init_driver_queue_delete(QUEUE_DETAILS *pQueue)
{
    pQueue->pDescQueueHead = NULL;
    pQueue->pDescQueueTail = NULL;
    pQueue->physQueueHead  = 0;

    if (pQueue->qSem != (int)NULL) {
        A_SEM_DELETE(pQueue->qSem);                    
        pQueue->qSem = (int)NULL;
    }
}

/*******************************************************************************
* init_driver_queue_reset - reset the receive queue
*
* This routine is called at bootup of the device and chipReset
*
* RETURNS: void
*/
void
init_driver_queue_reset(WLAN_DEV_INFO *pdevInfo)
{
    ATHEROS_DESC *pRxDesc;

    A_SEM_LOCK(pdevInfo->rxQueue.qSem, WAIT_FOREVER);

    pdevInfo->partialRxDesc = FALSE;
    pdevInfo->rxEOLcount    = 0;

    pRxDesc = pdevInfo->rxQueue.pDescQueueHead;
    while (pRxDesc) {
        pRxDesc->hwControl.word[0] = 0;
        pRxDesc->hwControl.word[1] = 0;
        pRxDesc->hwStatus.word[0]  = 0;
        pRxDesc->hwStatus.word[1]  = 0;
        pRxDesc->hwControl.rx.bufferLength = AR_BUF_SIZE;
        A_DATA_CACHE_INVAL(pRxDesc->pBufferVirtPtr, AR_BUF_SIZE);
        A_DESC_CACHE_FLUSH(pRxDesc);

#ifdef RXEOL_WAR
        if (pRxDesc == pdevInfo->rxQueue.pDescQueueTail) {
            pRxDesc->nextPhysPtr  = pRxDesc->thisPhysPtr;
            pRxDesc->pNextVirtPtr = pRxDesc;
            A_DESC_CACHE_FLUSH(pRxDesc);
            break;
        }
#endif
        pRxDesc = pRxDesc->pNextVirtPtr;
    }

    A_PIPEFLUSH();

    halSetRxDP(pdevInfo, A_DESC_V2P(pdevInfo->rxQueue.pDescQueueHead));

    A_SEM_UNLOCK(pdevInfo->rxQueue.qSem);
}

/*
 * If this is a WEP frame we need to:
 *- get the random IV from the frame
 *- get the key from sib entry
 *- sw decrypt the frame if needed
 *- update decrypt error counter
 */
inline A_STATUS
setupDecrypt(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    WLAN_FRAME_3_DATA_ENCRYPT   *hdr;
    WLAN_PRIV_RECORD            *pKey = NULL;
    A_UINT32                    index;
    A_UINT32                    ivmask;
    A_UINT32                    icvsize = 0;
    

    hdr  = pDesc->pBufferVirtPtr.header3Iv;

    /* support AES_OCB_REVA mode */
    if (!hdr->macHeader.frameControl.wep &&
        hdr->macHeader.frameControl.fType == FRAME_DATA)
    {
        
        aPrintk("SHOULD NOT BE IN HERE!!!WEP DISABLED\n");        
        /*
         * AES' works on older h/w by not setting the WEP bit;
         * must fake it for s/w sake, for unicast AES' negotiated frames
         */
        hdr->macHeader.frameControl.wep = FALSE;

    }

    if (TRUE == hdr->macHeader.frameControl.wep) {
        aPrintk("RX: WE HAVE WEP ENABLED!!!!!!!!11\n");
        /* key index from IV */
        index = hdr->iv;
        /* check the key ID for shared key */
        /* ID bits are the high-order ones */
        index >>= 30;

        if (index != 0) {
            pKey = &pdevInfo->keyTable[index];

            if ((pKey == NULL || pKey->keyLength == 0) ||
                (pKey->keyFlags & PRIV_UNIQUEKEY)) {
                pdevInfo->localSta->stats.RcvWEPExcludedCount++;
                
                return A_ERROR;
            }
        } else {
            if (pKey == NULL || pKey->keyLength == 0) {
                pdevInfo->localSta->stats.RcvWEPExcludedCount++;
                return A_ERROR;
            }
        }

        ASSERT(pKey);

        switch (pKey->keyType) {
        case PRIV_KEY_TYPE_WEP:
            ivmask  = 0x00ffffff;           /* 24 bit IV */
            icvsize = WEP_ICV_FIELD_SIZE;   /* WEP */
            break;

        case PRIV_KEY_TYPE_AES:
            ivmask  = 0x0fffffff;           /* 28 bit IV */
            icvsize = AES_ICV_FIELD_SIZE;   /* AES */
            break;

        default:
#ifdef DEBUG
            aPrintk("unsupported key type!\n");
#endif
            ivmask  = 0x00ffffff;           /* 24 bit IV */
            icvsize = WEP_ICV_FIELD_SIZE;   /* WEP */
            return A_ERROR;
        }

        pKey->keyIV = (hdr->iv) & ivmask;

        if (pdevInfo->localSta->swEncryptionEnabled)
        {
            pdevInfo->localSta->stats.RcvDecipherErrors++;
            printk("Encryption should not be enabled!\n");
            return A_ERROR;
            
        }

        /* FN - remove ICV field, only in WEP case I believe */
        pDesc->hwStatus.rx.dataLength -= icvsize;
    }

    /* remove ICV field */
    //FN - pDesc->hwStatus.rx.dataLength -= icvsize;

    return A_OK;
}

/*******************************************************************************
* processRxInterrupt - interrupt service routine
*
* In vxWorks, this routine was queued as a job by the ISR to handle the received 
* interrupt (using NetJobAdd so it runs in a task context and not Interrupt 
* context)...BE AWARE OF THIS UNDER LINUX, right now, call from ISR temporarily
* but we might need to fork it later!!
*
* It sends the good frames to DS service routine. For the error ones, it frees
* the descriptors and buffers.
*
* Invoked through a tasklet_schedule call from ISR..i.e. runs in normal context
*
* RETURNS: OK or ERROR
*/
void
processRxInterrupt(WLAN_DEV_INFO *pdevInfo)
{
    A_BOOL doneEarly = FALSE;
       
    // lock out interrupts till the end of this routine in ISR
    // while we copy descriptors/hw receive contents into skbs
    halDisableInterrupts (pdevInfo, HAL_INT_TX | HAL_INT_RX);
    
#ifdef PARTHA
    A_SEM_LOCK(pdevInfo->rxQueue.qSem, WAIT_FOREVER);

    /* updates all necessary stats then calls rx_kernel which task-lets
       as many times as we detect RX INTs, up to a ATHEROS_MAX_INTERRUPT_EVENT */
    doneEarly = HandleAtherosRxIntAndStats (pdevInfo);

    A_SEM_UNLOCK(pdevInfo->rxQueue.qSem);
#endif

    /*
     * Check to see if there was another RX interrupt recorded in the ISR while
     * handling a non-masked interrupt source.  This will avoid leaving
     * descriptors "stranded" on the queue waiting for another received frame
     * to kickstart their completion processing.  We queue another job if we
     * aborted processing descriptors early, or the ISR read another HAL_INT_RX
     * while RX ints were masked.
     */
    
    struct net_device *pNetDev = NULL;
    if (NULL != gPciDev) pNetDev = pci_get_drvdata(gPciDev);

    if (NULL != pNetDev) {

        // drain any RxInts uninterrupted..could have a performance hit!!??
        DisableInterruptsGlobaly (pdevInfo);
        
        if ((pdevInfo->globISRReg & HAL_INT_RX) || doneEarly) {
            doneEarly = HandleAtherosRxIntAndStats(pdevInfo); 
            pdevInfo->globISRReg &= ~HAL_INT_RX;
        }
        // if last HandleAtherosRxIntAndStats() call went well,
        else   {
            halEnableInterrupts(pdevInfo, HAL_INT_RX | HAL_INT_RXORN | HAL_INT_RXEOL);
        }

        EnableInterruptsGlobaly (pdevInfo);
    }

    // unlock interrupts till the end of this routine in ISR
    halEnableInterrupts (pdevInfo, HAL_INT_TX | HAL_INT_RX | HAL_INT_RXEOL);
    
}

/*****************************************************************
 *  HandleAtherosRxIntAndStats - handles a max number of received
 *  RX interrupts, update some minimal stats, checks for errors
 * (CRC, decryption...etc) and then puts the data in an assigned
 * buffer using after the CREATEANDBUFDESC call (which needs to
 * be freed after). Then, it calls atheros-rx-kernel which 
 * moves data to an skb and schedules a task to handles the
 * data received.
 *
 */
A_BOOL HandleAtherosRxIntAndStats (WLAN_DEV_INFO *pdevInfo) {

    WLAN_FRAME_HEADER   *pWlanHdr;
    WLAN_MACADDR        *pSrcStaMacAddr;
    HAL_DESC_RX_STATUS  *pRxStatus;
    ATHEROS_DESC        *pRxDesc;
    ATHEROS_DESC        *pTailDesc;
    A_BOOL              doneEarly = FALSE;
    A_UINT32            foundPhyErr = 0;
    
    WLAN_STATS          *pLocalStats = &pdevInfo->localSta->stats;
#if AR_RX_RESCHED > 0
    int                 rxCountdown = AR_RX_RESCHED;
#endif

    aPrintk("\nISR-RX: ENTERING HandleAtherosRxIntAndStats ------\n");
    aPrintk("ISR-RX: localSta->staState = %s\n", 
             (pdevInfo->localSta->staState & STATE_AP_UP)?"AP_UP":"AP_DOWN");

    while (pdevInfo->localSta->staState & STATE_AP_UP) {
      
      ASSERT(pdevInfo->rxQueue.pDescQueueHead);
      ASSERT(pdevInfo->rxQueue.pDescQueueHead != pdevInfo->rxQueue.pDescQueueTail);

      pRxDesc = pdevInfo->rxQueue.pDescQueueHead;
      A_DESC_CACHE_INVAL(pRxDesc);
      pRxStatus = &pRxDesc->hwStatus.rx;
      
      if (!pRxStatus->done) {
          aPrintk("ISR-RX: NO MORE! RX HEAD QUEUE HARDWARE STATUS != DONE!!\n");
          break;
      }

      
#ifdef RXEOL_WAR
      /* given the use of a self-linked tail be very sure that the hw is
       * done with this descriptor; the hw may have done this descriptor
       * once and picked it up again.. make sure the hw has moved on
       */
      A_DESC_CACHE_INVAL(pRxDesc->pNextVirtPtr);
      if ( (!pRxDesc->pNextVirtPtr->hwStatus.rx.done) &&
           (halGetRxDP(pdevInfo) == (A_UINT32)pRxDesc) )
      {
          aPrintk("ISR-RX: RX HEAD QUEUE VIRT PTR/HW STATUS is not DONE\n");
          break;
      }
#endif

#if AR_RX_RESCHED > 0
      /*
      *  RX loop to relinquish and reschedule after processing
      *  AR_RX_RESCHED descriptors.
      */
      if (0 == rxCountdown--) {
          aPrintk("ISR-RX: (((((((((DONE EARLY))))))))))\n");
          doneEarly = TRUE;
          break;
      }
#endif                         

      pRxDesc->decryptError = FALSE;          /* assume frame is ok */

      /*
       * Drop multiple-descriptor frames.  This is not WECA-compliant if our
       * buffer size (sans wlan header size) is larger than 1514 bytes.
       */
      if (pRxStatus->more || pdevInfo->partialRxDesc) {
          aPrintk("ISR-RX: rx frame toooooo long!\n");
      
          /* capture if the next one has to be discarded */
          pdevInfo->partialRxDesc = pRxStatus->more ? TRUE : FALSE;

          /* discard this one */
          pdevInfo->rxQueue.pDescQueueHead = pRxDesc->pNextVirtPtr;
          queueRxDescriptor(pdevInfo, pRxDesc);

          pLocalStats->RxInvalidFrames++;
          aPrintk("ISR-RX: RxInvalidFrames = %ld\n", pLocalStats->RxInvalidFrames);
          continue;
      }

      if (!pRxStatus->pktReceivedOK) {
          // tell the interface (ifconfig) to increment it's error counters
          //aPrintk("ISR-RX: RECORD AN ERROR IN IF_STATS\n");
          pLocalStats->ReceiveErrors++;

          /*
           * which of all these counters do we ever use? these
           * and other stats counters should be macroized
           */
          if (pRxStatus->fifoOverrun) {
              pLocalStats->RcvDmaOverrunErrors++;
              aPrintk("ISR-RX: FIFO OVER RUNNNNNs [so far = %ld]\n", pLocalStats->RcvDmaOverrunErrors);
          } else if (pRxStatus->phyError) {
              pLocalStats->RcvPhyErrors++;
              aPrintk("ISR-RX: PHY ERRORs [so far = %ld]\n", pLocalStats->RcvPhyErrors);

              phyErrorCounts[(pRxStatus->phyError & 0x7)]++;
              if (pdevInfo->staConfig.radarDetectEnable && (tsfIndex > 0)) {
                  aPrintk("ISR-RX: WHY IS RADARDETECTENABLE ON!??[i=%d][tsfIndex=%ld]\n", i, tsfIndex);
                  tsfIndex = 0;
                  foundPhyErr++;
              }
          } else if (pRxStatus->CRCError) {
              pLocalStats->RcvCrcErrors++;
              aPrintk("ISR-RX: RCV CRC ERRORs [so far = %ld]\n", pLocalStats->RcvCrcErrors);
          } else if (pRxStatus->decryptCRCError) {
              A_BOOL faulted = FALSE;

              pWlanHdr = pRxDesc->pBufferVirtPtr.header;
              /* looking at address2 only - don't pay the byte swap cost */
              pSrcStaMacAddr = &pWlanHdr->address2;
              if (pRxStatus->keyIndexValid == 0) {
                  pLocalStats->RcvKeyCacheMisses++;
                  aPrintk("ISR-RX: RcvKEY CACHE MISSES !!!![so far=%ld]\n", pLocalStats->RcvKeyCacheMisses);
              }

              if (faulted == FALSE) {
                  aPrintk("ISR-RX: Rcv DECRYPT CRC ERRORS !!!![so far=%ld]\n", pLocalStats->RcvDecryptCrcErrors);
                  /*
                   * Decrypt errors still get sent to the mac layer
                   * to deal with appropriately.
                   */
                  pRxDesc->decryptError = TRUE;       /* frame is corrupt */
              }
          }

          if (pRxDesc->decryptError == FALSE) {
              /* Take desc off the recv queue, and requeue */
              pdevInfo->rxQueue.pDescQueueHead = pRxDesc->pNextVirtPtr;
              queueRxDescriptor(pdevInfo, pRxDesc);

              continue;
          }
      }



      /*
       * no point reaping frames that are going to be discarded down
       * the forwarding path due to lack of resources... therefore,
       * throttle the receive processing based on if our own tx or
       * the ethernet's tx is going to drop the frame! since our own
       * tx is throttled by availability of descriptors on the empty
       * queue (and/or avail of tx header buffers really.. but) - an
       * attempt to get a descriptor below will achieve the desired
       * effect;  ideally availability of resources on ethernet side
       * should get throttled by loaned buffers returning to us and
       * (non)-availability of a buffer below - but currently we're
       * stumped by lack of proactive buffer release on tx completion
       * by the ethernet driver; we can live with that for now - as
       * for the common case we're the slower interface
       */
      if (createBuffandDesc(pdevInfo, AR_BUF_SIZE, &pTailDesc) != A_OK) {
          printk("ISR-RX: xxxxx Out of buff desc!!![createBuffandDesc FAILED]\n");
          doneEarly = TRUE;
          break;
      }

      struct net_device *pNetDev = NULL;
      if (NULL != gPciDev)
          pNetDev = pci_get_drvdata(gPciDev);

      if (pNetDev != NULL){
          atheros_local_info_t *local = (atheros_local_info_t *)pNetDev->priv;
          sap_config *config = (sap_config *)local->config;
          if ((config != NULL) && (config->am_ack == 0)) {
              // disable ACK processing in monitor mode
              struct ieee80211_hdr *hdr 
                  = (struct ieee80211_hdr *)pRxDesc->pBufferVirtPtr.ptr;
              u16 fc = cpu_to_le16(hdr->frame_control);
              if ((WLAN_FC_GET_TYPE(fc) == WLAN_FC_TYPE_CTRL) &&
                  (WLAN_FC_GET_STYPE(fc) == WLAN_FC_STYPE_ACK)){
                  /* take this off, and queue the new one */
                  pdevInfo->rxQueue.pDescQueueHead = pRxDesc->pNextVirtPtr;
                  pRxDesc->nextPhysPtr             = 0;
                  pRxDesc->pNextVirtPtr            = NULL;
                  queueRxDescriptor(pdevInfo, pTailDesc);
                  
                  freeBuffandDesc(pdevInfo, pRxDesc); 
                  continue;
              }
          }
      }
      

      /* take this off, and queue the new one */
      pdevInfo->rxQueue.pDescQueueHead = pRxDesc->pNextVirtPtr;
      pRxDesc->nextPhysPtr             = 0;
      pRxDesc->pNextVirtPtr            = NULL;
      queueRxDescriptor(pdevInfo, pTailDesc);

      pWlanHdr = pRxDesc->pBufferVirtPtr.header;
      /* subtract off FCS */
      pRxDesc->hwStatus.rx.dataLength -= FCS_FIELD_SIZE;
      
      atheros_rx_kernel (pdevInfo, pNetDev, pRxDesc, 0); 

  } // end of receive for loop servicing all RX INTs

  return doneEarly;
}

/*****************************************************************
 *  HandleAtherosTxIntAndStats - 
 *
 *
 */
void HandleAtherosTxIntAndStats (WLAN_DEV_INFO *pdevInfo) {
    
    aPrintk("\nISR-TX: ENTERING HandleAtherosTxIntAndStats ------\n");

    /* XXX - why check to see if AP is up here? */
    if (pdevInfo->localSta->staState & STATE_AP_UP) {
        processAltQueue(pdevInfo);
    }

    processTxQueue(pdevInfo);
}

/*****************************************************************
 * atheros_rx_kernel - moves by queueing received frame to an SKB 
 * and then calls a task-let to schedule a task/process to wake up 
 * sometime later and dequeue all queued up frames and handle 
 * them.
 * 
 * STILL RUNNING IN ISR context ...should change soon to some task-let
 *
 * Returns - Always 0 for now 
 */
int debugPrint = 0;
int atheros_rx_kernel(WLAN_DEV_INFO *pdevInfo, struct net_device *dev, 
		              ATHEROS_DESC  *pRxDesc,  int am_monitor_flag)
{
    // two skbs need to be created, one for the packet to the forwarding 
    // engine and another for the header/frame to AM
    struct sk_buff *skb;
    int    len;
    WLAN_FRAME_HEADER    * pWlanHdr;
    int    discard = 0;
  
    aPrintk("\nISR-RX: ENTERING atheros-rx-kernel ++++ \n");

    ASSERT(pdevInfo);
    ASSERT(pRxDesc);

    if (NULL == dev) {
        aPrintk("ISR_RX: device not attached yet...dropping this frame\n");
        goto error;
    }

    /*****************************************/
    /*   First, some desc maintanance work   */
    /*  (wlanFrameReceive)                   */
    /*****************************************/
   
    pWlanHdr = pRxDesc->pBufferVirtPtr.header;
    if (NULL == pWlanHdr) {
        printk("ISR_RX: WlanHdr is empty!! ???????\n");
        goto error;
    }
  
    /*****************************************/        
    /*  Then, skb buffering and GRE forward  */
    /*****************************************/
    // len is size of data in recv'd buffer pRxDesc 
    len = (pRxDesc->hwStatus.rx.dataLength);

    if (len > MAX_80211_FRAME_BODY_SIZE) {
        printk("ISR-RX WARNING: %s: atheros_rx_80211: len(%d) > "
               "MAX(%d)\n", dev->name, len, MAX_80211_FRAME_BODY_SIZE);
        debugPrint = 1;
        goto error;
    }

    // ensure that the skb has the received data (preferrably, no copies)
    if (len > 0 && len < MAX_80211_FRAME_BODY_SIZE) {    //2312B
        
        A_UINT8 *pos;

        if (NULL == pRxDesc->pOSDescPtr) {
            printk("SKB DATA NOT FOUND IN pOsDescPtr..\n");
            goto error;
        }

        // keep a copy of net_device
        skb = (struct sk_buff*)pRxDesc->pOSDescPtr;
        skb->dev  = dev;
        // data is auto received in skb->data
        pos = skb->data - (ATHEROS_GRE_HEADROOM + ATHEROS_HW_CTRL_AND_STATUS);
        memcpy(pos, (void*)(&(pRxDesc->hwControl.rx)), sizeof(HAL_DESC_RX_CONTROL));
        pos += sizeof(HAL_DESC_RX_CONTROL);
        memcpy(pos, (void*)(&(pRxDesc->hwStatus.rx)), sizeof(HAL_DESC_RX_STATUS));
        skb->len = len;

        skb_queue_tail  (&atheros_sap_readq, skb);
        tasklet_schedule(&atheros_sap_tasklet);

        if (debugPrint) printk("ISR-RX: QUEUEING&SCHEDULING A TASKLET....\n");
    }  // if len is valid
  
    /**************************************/
    /* Finally, any extra needed cleanup  */
    /**************************************/
    if (discard) {
        pdevInfo->localSta->stats.RxDiscardFrames++;
        aPrintk("ISR-RX: FREE UP THE BUF DUE TO discard=%d\n", discard);
    }

    // we need to do the same job as freeBuffandDesc() except for 
    // freeing up the skb(s).
    QUEUE_DETAILS * pQueue = pRxDesc->freeQueue;
        
    if ((pQueue == &pdevInfo->txHeaderQueue)
        || (pQueue == &pdevInfo->psBufferQueue)) {

        printk("DOING THE FREEING EXTRA PART!!\n");

        pRxDesc->pNextVirtPtr = NULL;
        pRxDesc->nextPhysPtr  = 0;

        pRxDesc->swretryCount = pRxDesc->fragSeqNumValid = 0;
        pRxDesc->psFlag       = pRxDesc->staleFlag       = 0;
        pRxDesc->pTxFirstDesc = pRxDesc->pTxLastDesc     = NULL;

        /* zero out the control and status fields */
        pRxDesc->hwControl.word[0] = 0;
        pRxDesc->hwControl.word[1] = 0;
        pRxDesc->hwStatus.word[0]  = 0;
        pRxDesc->hwStatus.word[1]  = 0;

        A_SEM_LOCK(pQueue->qSem, WAIT_FOREVER);
        addDescToQueueTail(pRxDesc, pRxDesc, pQueue);
        A_SEM_UNLOCK(pQueue->qSem);

        return 0;
    }

    // don't free up the skb since it's forwarded to the ethernet layer for use
    memFreeDescriptor(pdevInfo, pRxDesc);
    return 0;

error:
    /* free desc used for this Rx: Contents are copied into SKB before here! */
    /* we need to free up descriptors allocated in the HandleAtherosRx..()   */
    freeBuffandDesc(pdevInfo, pRxDesc); 
    return 0;
}

/*************************************************************************
*
* atheros_sap_switch_uplink - The tasklet entry point to handle recvd
*                             mgt packets
* 
* RETURNS:
*/
void atheros_sap_switch_uplink(unsigned long data) {

    // called when a .11 frame is received from the radio interface
    struct sk_buff       *skb = NULL;
    struct ieee80211_hdr *hdr = NULL;
    A_UINT16              fc, type, stype;
    struct net_device    *dev   = NULL;        
    atheros_local_info_t *local = NULL;

    aPrintk("\nRX: ENTERING atheros_sap_switch_uplink\n");

    if (NULL != gPciDev) {
        dev = pci_get_drvdata(gPciDev);
        if (NULL !=  dev)
            local = (atheros_local_info_t *) dev->priv;
	else {
	    printk("dev = NULL...free skb\n");
	    goto cleanup;
	}
    }
    else {
	printk("DEVICE IS NULL...free skb\n");
	goto cleanup;
    }


    // if a reset process is awake and doing it's thing while
    // we woke up, don't dequeue
    if (pwlanDevInfo->NicResetInProgress == TRUE) {
        printk("RX: NICRESETINPROGRESS....NO DEQUEUING [try later]\n");
        goto cleanup;
    }

    // while there are queued up skbs
    while ((skb = skb_dequeue(&atheros_sap_readq))) {

        // filter unnecessary traffic
        hdr = (struct ieee80211_hdr *)skb->data;
        fc = __cpu_to_le16(hdr->frame_control);
        type  = WLAN_FC_GET_TYPE(fc);
        stype = WLAN_FC_GET_STYPE(fc);
        
        /* now dispatch the frame */
        if ( (type == WLAN_FC_TYPE_MGMT) && (stype == WLAN_FC_STYPE_PROBE_REQ) ) {

           // from CPU to bus
           FRAME_CONVERT_CPUTOLE (hdr, skb->len);
           aPrintk ("RX: GOOD, A PROBE REQUEST..RESPOND\n");
           // call probe request handling routine directly
           mlmeProbeRequestReceive(pwlanDevInfo,(WLAN_FRAME_PROBE_REQ*)hdr, skb->len);
           // rx_asap and rx_monitor_asap should free skb

       }

        // forward onto MONITOR
        if (local) {
            sap_config *config = (sap_config *)local->config;
            WLAN_DEV_INFO *pdevInfo = local->pWlanDev;
            pdevInfo->localSta->stats.GoodReceives++;
            if (config) {
                if (config->mode == WIFI_MODE_MONITOR) {
                    if ((local->func->asap.rx_monitor_asap) && 
                        (g_monitor_type == ATHEROS_MONITOR_ON)) // if we're in AM mode
                        local->func->asap.rx_monitor_asap(skb, skb->dev);
                    else
                        dev_kfree_skb(skb);
                }
                else {
                    if (local->func->asap.rx_asap)
                        local->func->asap.rx_asap(skb, skb->dev);
                    else
                        dev_kfree_skb(skb);
                }
            }
            else
                dev_kfree_skb(skb);
        }
        else 
            dev_kfree_skb(skb);

    } // while dequeuing

    aPrintk("RX: Done atheros_sap_switch_uplink\n\n");

    return;

cleanup:
    printk("CLEANING UP THE UPLINK TASKLET\n");
    if (skb)
      dev_kfree_skb(skb);
    return;

}

/*************************************************************************
*
* atheros_handle_mgt_frame - post processing the validated mgt 
*                            receiving frame.
*
* Called when a frame has been received at hwc layer and needs
* to be handled by MAC layer. Based on wlanFrameReceive() but only 
* handling the mgt frames
*
* RETURNS:
*/
void atheros_handle_mgt_frame (WLAN_DEV_INFO *pdevInfo, 
                               struct ieee80211_hdr* hdr, 
                               A_UINT16 data_len)
{
    A_UINT16 fc, type, stype;
    
    fc = __cpu_to_le16(hdr->frame_control);
    type = WLAN_FC_GET_TYPE(fc);
    stype = WLAN_FC_GET_STYPE(fc);
    
    /* now dispatch the frame */
    if ( (type == WLAN_FC_TYPE_MGMT) && (stype == WLAN_FC_STYPE_PROBE_REQ) ) {
            // from CPU to bus
           FRAME_CONVERT_CPUTOLE (hdr, data_len);
           aPrintk ("RX: GOOD, A PROBE REQUEST..RESPOND\n");
           // call probe request handling routine directly
           mlmeProbeRequestReceive(pdevInfo,(WLAN_FRAME_PROBE_REQ*)hdr, data_len);
    }

    pdevInfo->localSta->stats.RxMgmtFrames++;
            
    // all packets (including probe_req) should be forwarded onto the next stage

#if HANDLE_AUTH
            // AUTHENTICATE REQUESTS: as an experiement
            else if  (stype == WLAN_FC_STYPE_AUTH) {

                aPrintk ("RX: %%%%%%GOOD, AN AUTH. REQUEST..RESPOND\n");
                // call probe request handling routine directly
                WLAN_MGT_MAC_HEADER *pAuthReqFrame = (WLAN_MGT_MAC_HEADER *)hdr;
                int i=0;
                for (i=0; i<6;i++) {
                    pAuthReqFrame->destAddr.octets[i] ^= pAuthReqFrame->srcAddr.octets[i];
                    pAuthReqFrame->srcAddr.octets[i]  ^= pAuthReqFrame->destAddr.octets[i];
                    pAuthReqFrame->destAddr.octets[i] ^= pAuthReqFrame->srcAddr.octets[i];
                }
                
                A_UINT16* pBuf = (A_UINT16 *)(hdr + 1);
                pBuf[1] = 2;
                pBuf[2] = 0;
                
                ATHEROS_DESC* pDesc;
                if ((createBuffandDesc(pdevInfo, data_len, (void **) &pDesc)) != 0) {
                     printk("TX: Failed to allocate buffer for frame and descriptor\n");
                     return;
                }
                WLAN_FRAME_AUTH* pFrame = pDesc->pBufferVirtPtr.auth;
                memcpy(pFrame, pAuthReqFrame, data_len);
                
                // call the atheros_sap.c FrameTransmit version
                wlanFrameTransmit(pdevInfo, pDesc, FALSE);
                // no one will explicitly cleanup this created pDesc, so we do
                //freeBuffandDesc(pdevInfo, pDesc);
            }
            else if (stype == WLAN_FC_STYPE_ASSOC_REQ){
                printk("+++++++++++++++++++++++++++++++++++++++\n");
                printk("Received an association request! yeah!!\n");
                printk("+++++++++++++++++++++++++++++++++++++++\n");
            }
            else if ( (stype == WLAN_FC_STYPE_PROBE_RESP) ||
                      (stype == WLAN_FC_STYPE_BEACON)  ) 
                printk("RX: WARNING WARNING *** fSubtype = %d[PASS TO FRWDR]\n", stype);
#endif // HANDLE_AUTH
        
    return;
}


/*************************************************************************
 *
 * wlanFrameTransmit - send the tx frame to driver
 *
 * called when a frame has been completely assembled, except for fragmentation,
 * and is ready for final xmit prep. The bypass option is to avoid overriding
 * the forwarder set bits of the HAL_DESC_TX_CONTROL structure in buffer sent
 * down to us.
 *
 * RETURNS: error status
 */
A_STATUS
wlanFrameTransmit(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_BOOL bypass)
{
    WLAN_FRAME_HEADER   *pWlanHdr;
    
    aPrintk("\nTX: Inside wlanFrameTransmit...\n");

    /* before calling txServices, apply the basic tx rules to the descriptor */
    pWlanHdr = pDesc->pBufferVirtPtr.header;

    switch (pWlanHdr->frameControl.fType) {

    case FRAME_DATA:
        pdevInfo->localSta->stats.TxDataFrames++;
        break;
    case FRAME_CTRL:
        pdevInfo->localSta->stats.TxCtrlFrames++;
        break;
    case FRAME_MGT:
        pdevInfo->localSta->stats.TxMgmtFrames++;
        break;
    default:
        pdevInfo->localSta->stats.TxInvalidFrames++;
#ifdef DEBUG
        printk("TXTXTX: wlanFrameTransmit: invalid frame type %d\n",
                 pWlanHdr->frameControl.fType);
#endif
        freeBuffandDescChain(pdevInfo, pDesc);
        return A_ERROR;
    }

    int frameLen      = (A_UINT16)pDesc->hwControl.tx.frameLength;
    int fragThreshold = (A_UINT16)pdevInfo->staConfig.userFragThreshold;
    /* Skip fragmentation if it's too short */
    if ((frameLen <= fragThreshold)) 
        aPrintk("TX: FRAG IS NOT NEEDED [len=%d][thresh=%d]\n", frameLen, fragThreshold);
    else
        aPrintk("TX: HEY< FRAG IS NEEDED INDEEEEEED[len=%d][thresh=%d]\n", frameLen, fragThreshold);

    /*
     * handle the power management here
     * and assuming the STA is awake for management response frames
     */


    /* prepare desc chain and send it to the hw */
    if (ar5MacSend(pdevInfo, pDesc, bypass) == A_ERROR) {
        freeBuffandDescChain(pdevInfo, pDesc);
#ifdef DEBUG
        aPrintk("TX: wlanFrameTransmit: ar5MacSend failed\n");
#endif
        return A_ERROR;
    }

    return A_OK;
}

/*************************************************************************
*
* ar5MacSend - mac layer transmission routine.
*
* called before a frame is passed to the DMA engine.
*
* RETURNS: A_OK
*          A_ERROR
*/
A_STATUS
ar5MacSend(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_BOOL bypass)
{
    /* perform tx service */
    if (txService(pdevInfo, pDesc, bypass) != A_OK) {
#ifdef DEBUG
        printk("TX: ar5MacSend: txService failed\n");
#endif
        /* drop the frames */
        freeBuffandDescChain(pdevInfo, pDesc);
        return A_ERROR;
    }

    if (atheros_send(pdevInfo, pDesc, FALSE, bypass) != A_OK) {
#ifdef DEBUG
        printk("TX: ar5MAcSend: atheros_send failed\n");
#endif
        /* drop the frames */
        freeBuffandDescChain(pdevInfo, pDesc);
        return A_ERROR;
    }

    return A_OK;
}


/*******************************************************************************
*
* txService - the tx service routine
*
* This routine takes ATHEROS descriptors. It will fill the frame header with tx
* rate, the still empty descriptor fields. Also, it will fill the MAC header
* duration field. Then it will send the descriptor to the tx queue and put the
* first descriptor pointer to tx queue register.
*
* RETURNS: OK or ERROR.
*/

A_STATUS
txService(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pTxDesc, A_BOOL bypass)
{
    HAL_DESC_TX_CONTROL *pHwTxControl = NULL;
    A_UINT32            rateIndex = pdevInfo->staConfig.defaultOfdmRateIndex;
    /* this always pts to the head desc of a MPDU (ie frame or frag) */
    ATHEROS_DESC        *pCurrentPktTxDesc = pTxDesc;

    ATHEROS_DESC        *pCurrentTxDesc;
    ATHEROS_DESC        *txd;
    ATHEROS_DESC        *pFragDesc; // to keep track of multi-desc frags
    
    A_UINT16            fragNum = 0;
    int                 controlRate, toGroup;

    WLAN_DATA_MAC_HEADER3 *hdr;
    
    aPrintk("TX: Entering txService...\n");

    /* iterate over all MPDUs in this chain */
    while (pCurrentPktTxDesc != NULL) {
            pCurrentTxDesc = pCurrentPktTxDesc;
            pFragDesc      = pCurrentPktTxDesc;

            aPrintk("TX: inside of txService's all MPDUs processing loop!!\n");

            /*
            *  due to the loop at the bottom of this routine,
            *  this loop only visits the initial desc of each
            *  MPDU
            */
            while (pCurrentTxDesc != NULL) {
                aPrintk("TX: currentTxDesc != NULL\n");

                pHwTxControl = &pCurrentTxDesc->hwControl.tx;
                hdr          = pCurrentTxDesc->pBufferVirtPtr.header3;

                if (hdr == NULL) {
                    aPrintk("TX: ERROR hdr == NULL\n");
                    return A_ERROR;
                }

                toGroup = isGrp(&hdr->address1);

                /*
                * The spec and WECA definition is currently to use 6, 12 & 24 as
                * the base modes. I think all multicast, management & control frames
                * have to go at those rates. That may change after future debate, some
                * think that 6 should be the base rate to maximize the range of control
                * packets.                                         - AndyD
                */
                if ( (hdr->frameControl.fType == FRAME_MGT) &&
                        (pdevInfo->staConfig.rateCtrlEnable==TRUE) ) {
                    /* If we are not at a fixed rate, send all mgt frames at lowest rate.
                    * It gives more reliable transmission and the stations that are at
                    * the boundary have a better chance of completing the connection process.
                    */
                    rateIndex = LOWEST_RATE_INDEX;
                    aPrintk("TX: WE ARE A FRAME_MGT\n");
                } 
                
                if (FALSE == bypass) { // otherwise, this was set by forwarder
                    /* Select transmit rate */
                    pHwTxControl->transmitRate =
                        pdevInfo->staConfig.pRateTable->info[rateIndex].plcpRate;
                } else {
                    

                    // the forwarder layer should pass it's desired rate index onto us!
                    rateIndex = pHwTxControl->transmitRate;
                    
                    if ((rateIndex < 0) || (rateIndex >= RATE_TABLE_SIZE)) {
                        printk("txService: Invalid rateIndex[%ld] passed down\n", rateIndex);
                        rateIndex = pdevInfo->staConfig.defaultOfdmRateIndex;
                        printk("Transmitting at default rateIndex [%ld]\n", rateIndex);
                    }
                    
                    pHwTxControl->transmitRate =
                        pdevInfo->staConfig.pRateTable->info[rateIndex].plcpRate;
                }

                aPrintk("TX: TransmitRate = %d\n", pHwTxControl->transmitRate);
                
                /* Select control rate */
                controlRate = pdevInfo->staConfig.pRateTable->info[rateIndex].controlRate;
                aPrintk("TX: controlRate = %d\n", controlRate);
                
                /* Increase the buffer length with the FCS_FIELD_SIZE */
                pHwTxControl->frameLength += FCS_FIELD_SIZE;
                aPrintk("TX: HwTxControl->frameLen = %d[inc FCS_SIZE=%d]\n", 
                        pHwTxControl->frameLength, FCS_FIELD_SIZE);

                /* If not unicast, set duration to be zero */
                if (toGroup && (fragNum == 0)) {
                    
                    aPrintk("TX: WE ARE A GROUP ADDRESS and no frags yet!\n");
                    halSetTxDescNoAck(pdevInfo, pCurrentTxDesc, 1);
                    
                    pCurrentTxDesc->pBufferVirtPtr.header->durationNav.duration = 0;
                    pCurrentTxDesc->pBufferVirtPtr.header->durationNav.clearBit = 0;
                    
                    // otherwise, this was set by forwarder
                    if (FALSE == bypass) { 
                        pHwTxControl->transmitRate =
                            pdevInfo->staConfig.pRateTable->info[rateIndex].plcpRate;
                    }

                    pdevInfo->localSta->stats.TxMulticastFrames++;
                    break;
                }

                if (pdevInfo->staConfig.pChannel->channelFlags & CHANNEL_CCK) {
                    printk("TX: setting channel to LONG PREAMBLE_ENABLED\n");
                    // in init, we set capInfo.shortPreamble to OFF
                    // force long preamble by clearing short preamble bit
                    // otherwise, this was set by forwarder
                    if (FALSE == bypass)
                        pHwTxControl->transmitRate &= ~CCK_SHORT_PREAMBLE_ENABLE;
                }

                /* Set RTSCTS and LongPkt */
                if ( (pHwTxControl->frameLength >
                      pdevInfo->staConfig.defaultDot11RTSThreshold) && 
                     !toGroup && (fragNum == 0) )
                {
                    aPrintk("TX: Enabling RTSCTS\n");
                    if (FALSE == bypass) pHwTxControl->RTSCTSEnable = 1;
                    halSetTxDescLongPkt(pdevInfo, pCurrentTxDesc, 1);
                } else {
                    aPrintk("TX: Disabling RTSCTS\n");
                    if (FALSE == bypass) pHwTxControl->RTSCTSEnable = 0;
                    halSetTxDescLongPkt(pdevInfo, pCurrentTxDesc, 0);
                }
  
                // safe to go with long preamble. It doesn't matter for m/bcast
                descDurationSet(pdevInfo, pCurrentTxDesc, rateIndex, bypass, FALSE);

                /* Check if there exist more fragments */
                if (pCurrentTxDesc->pBufferVirtPtr.header->frameControl.moreFrag == 1) {

                    aPrintk("TXTXTX: I sure hope we're not here...coz we skipped FRAG\n");
                    /* Look for the next fragment header */
                    while (pHwTxControl->more != 0) {
                        /* init FirstDesc for each desc in this fragment */
                        pFragDesc->pTxFirstDesc = pCurrentTxDesc;
                        pFragDesc               = pFragDesc->pNextVirtPtr;

                        ASSERT(pFragDesc != NULL);
                        if (pFragDesc == NULL) {
                            return A_ERROR;
                        }
                        pHwTxControl = &pFragDesc->hwControl.tx;
                    }

                    /* init FirstDesc for each desc in this fragment */
                    pFragDesc->pTxFirstDesc = pCurrentTxDesc;
                    
                    /* set the last desc field for the first desc only */
                    pCurrentTxDesc->pTxLastDesc = pFragDesc;

                    /*
                    *  the next desc holds the next frag header;
                    *  moreFrag is 1, so there must be another
                    */
                    pFragDesc = pFragDesc->pNextVirtPtr;
                    if (pFragDesc == NULL) {
                        aPrintk("TX: ERROR pFragDesc should hold next frag header!!\n");
                        ASSERT(pFragDesc != NULL);
                        return A_ERROR;
                    }
                    
                    pCurrentTxDesc = pFragDesc;

                    fragNum++;
                    aPrintk("TX: fragNum = %d\n", fragNum);
                    
                } else {
                    aPrintk("TX: GOOD, no frag bits set!!!!\n");
                    /* No more fragments or the frame is not fragmented */
                    break;
                }

            } /* end while loop for descriptor chain for the current frame */

            /* set the first desc field */

            aPrintk("TX: ARE WE SKIPPING the for loop?[%s]\n", (txd->hwControl.tx.more == 0)?"true":"false");

            for (txd = pCurrentTxDesc; txd->hwControl.tx.more != 0;
                 txd = txd->pNextVirtPtr) {
                txd->pTxFirstDesc = pCurrentTxDesc;

                aPrintk("TX: Setting the first desc field\n");
            }
            txd->pTxFirstDesc = pCurrentTxDesc;

            /* set the last desc field for the first desc only */
            pCurrentTxDesc->pTxLastDesc = txd;

            /* advance to the first desc in the next frame */
            pCurrentTxDesc    = txd->pNextVirtPtr;
            pCurrentPktTxDesc = pCurrentTxDesc;
            
    } /* end while loop for frame chain */

    aPrintk("TX: Done txService...\n");

    return A_OK;
}


/*******************************************************************************
* ar5hwcDrvSend - the driver send routine
*
* This routine takes a Atheros Descriptor and sends out to the DMA engine.
*/
A_STATUS atheros_send(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pHead, 
                      A_BOOL callerHoldsLock, A_BOOL bypass)
{
    ATHEROS_DESC    *pTxDesc;
    ATHEROS_DESC    *pTrailingDesc;
    
    if ((pHead == NULL) || (pdevInfo->NicResetInProgress == TRUE)) {
#ifdef DEBUG
        uiPrintf("ar5hwcDrvSend: Error\n");
#endif
        return A_ERROR;
    }

    pTrailingDesc = NULL;
    pTxDesc       = pHead;

    while (pTxDesc != NULL) {
	
            aPrintk("TX: pTxDesc: %p[something to transmit]\n", pTxDesc);
	
            /*
            *  This routine might do the following with the desc:
            *  1) enqueue it and set pTxDesc to the next frame
            *  2) set some state (clearDestMask) then pass it thru
            *
            *  Because of 1), we must call this before setupEncrypt().
            */
            //fadi newDesc = swretrySetup(pdevInfo, pTxDesc, pTrailingDesc);
            //fadi txFramePendIncrement(pdevInfo, pSib, pTxDesc);
            //fadi status = setupEncrypt(pdevInfo, pTxDesc);
            pdevInfo->numTxFramesPending++;
            if (FALSE == bypass) pTxDesc->hwControl.tx.clearDestMask = 1; /* sync */
            
            if ((pTxDesc->pBufferVirtPtr.header3Iv)->macHeader.frameControl.wep) 
                aPrintk("TX: WHATT??? WE HAVE WEP in this direction!!???\n");
            else
                aPrintk("TX: NO WEP in this direction!!\n");
            
#ifdef DEBUG
            if (txVerbose >= 10) {
                //fadi            dumpFrame(pTxDesc, "tx");
            }
#endif
	
            /* swap */
            // cpu to pci bus...not needed in the case of FORWADER PASSED TRAFFIC
            if (FALSE == bypass) {
                FRAME_CONVERT_CPUTOLE(pTxDesc->pBufferVirtPtr.header,
				pTxDesc->hwControl.tx.bufferLength);
            }
	
            /* Proclaim our freshness */
            pTxDesc->staleFlag = 0;
            while (pTxDesc->hwControl.tx.more != 0) {
                aPrintk("TX: PROCLAIMING OUR FRESHNESS!!!!! huhhhhh???\n");
                pTxDesc = pTxDesc->pNextVirtPtr;
                pTxDesc->staleFlag = 0;
            }
	
            pTrailingDesc = pTxDesc;                /* last frame */
            pTxDesc = pTxDesc->pNextVirtPtr;        /* next frame */
        
    }  // while (pTxDesc != NULL)

    if (pHead) {
        ASSERT(pTrailingDesc != NULL);

        /* Add to normal tx queue -  kick-start the hw if first-time use */
        if (pdevInfo->txQueue.pDescQueueTail == NULL) {
            queueTxDescriptor(pdevInfo, pHead, pTrailingDesc);
            halSetTxDP(pdevInfo, pHead->thisPhysPtr);
        } else {
            queueTxDescriptor(pdevInfo, pHead, pTrailingDesc);
        }

        /*
        *  Note this write is not strictly necessary if the sw is
        *  ahead of the hw, and we could defer it until we get
        *  a TXEOL interrupt.  But if we issue it here, we may
        *  in fact avoid that interrupt.
        */
        halStartDmaTransmit(pdevInfo, HAL_TX_QUEUE_DATA);
    }

    if (!callerHoldsLock) {
        A_SEM_UNLOCK(pdevInfo->txQueue.qSem);
    }

    return A_OK;
}

/*
 *  swretrySetup() -- Decide if we want to set the clearDestMask or not.
 *      pdevInfo - device info struct
 *      pSib - SIB entry
 *      pDesc - 1st desc of current frame
 *      pTrailer - final desc of previous frame, used for linkage
 *
 *      Returns:  tx descriptor, possibly different from argument
 */
inline ATHEROS_DESC *
swretrySetup(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, ATHEROS_DESC *pTrailer)
{
    WLAN_FRAME_HEADER *pHdr;
    int     needClearDest = 1; // like in swretryStateReset() 
    
    aPrintk("TX: in swretrySetup...\n");

    if (pDesc == NULL) {
        printk("TX: ERROR, pDesc passed in is NULL\n");
        return NULL;
    }

    pHdr = pDesc->pBufferVirtPtr.header;

    if (pdevInfo->staConfig.swretryEnabled) {
        
        printk("TX: SWRETRY IS ENABLED!!\n");

        /* 
         * STA drains queue in DPC; AP drains queue in context
         * of filtered frame count decrementing to 0
         */
        if (needClearDest) {
            pDesc->hwControl.tx.clearDestMask = 1;
            needClearDest = 0;
        }
        if (pDesc->hwControl.tx.clearDestMask == 1) 
            needClearDest = 0;
        else 
            pDesc->hwControl.tx.clearDestMask = 0;
        

        // h/w hasn't given it a seq num yet
        pDesc->fragSeqNumValid = 0;

    } // if (pdevInfo->staConfig.swretryEnabled)
    else {

        aPrintk("TX: SWRETRY IS DISABLED!!\n");
        /*
         *  Non-head fragments should have a clearDestMask always off,
         *  and we should never swretry a frag if the preceding frag
         *  got dropped.
         */
        pHdr = pDesc->pBufferVirtPtr.header;
        if (pHdr->sequenceControl.fragNo == 0) {
            pDesc->hwControl.tx.clearDestMask = 1;
        } else {
            pDesc->hwControl.tx.clearDestMask = 0;
        }
    }

    pHdr->frameControl.retry = 0;

    return pDesc;
}

/*******************************************************************************
* queueRxDescriptor - queue descriptor to receive queue tail
*
* This routine is only called from processRxInterrupt!
*
* RETURNS: void
*/
void
queueRxDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pRxDesc)
{
    QUEUE_DETAILS *pQueue = &pdevInfo->rxQueue;

    pRxDesc->hwControl.word[0] = 0;
    pRxDesc->hwControl.word[1] = 0;
    pRxDesc->hwStatus.word[0]  = 0;
    pRxDesc->hwStatus.word[1]  = 0;

#ifdef RXEOL_WAR
    pRxDesc->nextPhysPtr  = pRxDesc->thisPhysPtr;
    pRxDesc->pNextVirtPtr = pRxDesc;
#else
    pRxDesc->nextPhysPtr  = 0;
    pRxDesc->pNextVirtPtr = NULL;
#endif

    pRxDesc->hwControl.rx.bufferLength = AR_BUF_SIZE;
    A_DATA_CACHE_INVAL(pRxDesc->pBufferVirtPtr, AR_BUF_SIZE);
    A_DESC_CACHE_FLUSH(pRxDesc);

    ASSERT(pQueue->pDescQueueTail != NULL);
    pQueue->pDescQueueTail->pNextVirtPtr = pRxDesc;
    pQueue->pDescQueueTail->nextPhysPtr  = pRxDesc->thisPhysPtr;
    A_DESC_CACHE_FLUSH(pQueue->pDescQueueTail);

    pQueue->pDescQueueTail = pRxDesc;

    A_PIPEFLUSH();
}

/*******************************************************************************
* queueTxDescriptor - queue descriptor to transmit queue tail
*
* RETURNS: void
*/
void
queueTxDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, ATHEROS_DESC *pTail)
{
    QUEUE_DETAILS *pQueue = &pdevInfo->txQueue;

    ASSERT(pTail->pNextVirtPtr == NULL);

    if (pQueue->pDescQueueTail == NULL) {
        ASSERT(pQueue->pDescQueueHead == NULL);
        pQueue->pDescQueueHead = pDesc;
    } else {
        pQueue->pDescQueueTail->nextPhysPtr  = pDesc->thisPhysPtr;
        pQueue->pDescQueueTail->pNextVirtPtr = pDesc;
#ifdef PERF_DEBUG
        A_DESC_CACHE_INVAL(pQueue->pDescQueueTail);
        if (pQueue->pDescQueueTail->hwStatus.tx.done) {
            ++onEmptyQueue;
        } else {
            ++onFullQueue;
        }
#endif
        A_DESC_CACHE_FLUSH(pQueue->pDescQueueTail);
    }
    pQueue->pDescQueueTail = pTail;

#ifdef SWCOHERNECYWB
    /* skip loop for writethru cache */
    while (pDesc) {
        A_DATA_CACHE_FLUSH(pDesc->pBufferVirtPtr, pDesc->hwControl.tx.bufferLength);
        A_DESC_CACHE_FLUSH(pDesc);
        pDesc = pDesc->pNextVirtPtr;
    }
#endif
    A_PIPEFLUSH();

    return;
}

inline void
processTxInterrupt(WLAN_DEV_INFO *pdevInfo)
{
    aPrintk("ISR-TX: Entering processTxInterrupt...\n");
    // lock out interrupts till the end of this routine in ISR
    // while we copy descriptors/hw receive contents into skbs
    halDisableInterrupts (pdevInfo, HAL_INT_TX | HAL_INT_RX);
    aPrintk("ISR-TX: DIS-abling TX | RX ints\n");


    HandleAtherosTxIntAndStats (pdevInfo);

    /* Check to see if there was another TX interrupt recorded in the ISR while
     * handling a non-masked interrupt source.  This will avoid leaving
     * descriptors "stranded" on the queue waiting for another transmitted
     * frame to kickstart their completion processing.  If we aborted processing
     * descriptors early (to be fair to other jobs running in the task) or if
     * the ISR recorded a TX interrupt while TX interrupts were masked, we queue
     * up another job to run later.
     */
    struct net_device *pNetDev = NULL;
    if (NULL != gPciDev) pNetDev = pci_get_drvdata(gPciDev);

    if (NULL != pNetDev) {
        //fadi halDisableInterrupts(pdevInfo, HAL_INT_TX);
        DisableInterruptsGlobaly (pdevInfo);
        if (pdevInfo->globISRReg & HAL_INT_TX) {
            HandleAtherosTxIntAndStats(pdevInfo); 
            pdevInfo->globISRReg &= ~HAL_INT_TX;
        
        }
        else 
            halEnableInterrupts(pdevInfo, HAL_INT_TX | HAL_INT_RX);

        EnableInterruptsGlobaly (pdevInfo);
    }  // if dev exists

        // lock out interrupts till the end of this routine in ISR
    // while we copy descriptors/hw receive contents into skbs
    halEnableInterrupts (pdevInfo, HAL_INT_TX | HAL_INT_RX);
    aPrintk("ISR-TX: Enabling TX\n");
}

#define lastDescGet(x)  ((x)->pTxLastDesc)

/*******************************************************************************
* processAltQueue - altQ process routine
*
* This routine is called by wlanBeaconGen to handle the transmitted frames in
* AltQ.  It releases the buffer and descriptors back to the right queues.
*/
void
processAltQueue(WLAN_DEV_INFO *pdevInfo)
{
    HAL_DESC_TX_STATUS  *pAltStatus;
    ATHEROS_DESC        *pAltDesc;
    ATHEROS_DESC        *pAltHeadDesc;

    pAltDesc = pdevInfo->altQueue.pDescQueueHead;

    aPrintk ("\nTX: Inside processAltQueue...\n");

    while (pAltDesc != NULL) {
        aPrintk("TX: walking alt queue\n");
        pAltHeadDesc = pAltDesc;
        pAltDesc     = lastDescGet(pAltDesc);
        pAltStatus   = &pAltDesc->hwStatus.tx;

        if (pAltStatus->done == 0) {
            break;
        }

        /* update the alt queue */
        pdevInfo->altQueue.pDescQueueHead = pAltDesc->pNextVirtPtr;

        aPrintk("TX: alt queue frame done\n");

        if (pAltStatus->pktTransmitOK) {
            aPrintk("TX: frame status = ok\n");

            /* Update the correct sent statistics */
            //fadi END_ERR_ADD(pdevInfo->pOSHandle, MIB2_OUT_UCAST, +1);
            pdevInfo->localSta->stats.GoodTransmits++;
        } else {
            aPrintk("TX: frame status = not ok!\n");
            /* Update the incorrect sent statistics */
            //fadi END_ERR_ADD(pdevInfo->pOSHandle, MIB2_OUT_ERRS, +1);

            if (pAltStatus->fifoUnderrun) {
                aPrintk("TX: frame status = fifo underrun!\n");
                pdevInfo->localSta->stats.TxDmaUnderrun++;
            }
        }

        /* terminate the current frame */
        pAltDesc->nextPhysPtr  = 0;
        pAltDesc->pNextVirtPtr = NULL;

        /* free the done frames */
        freeBuffandDescChain(pdevInfo, pAltHeadDesc);

        /* next frame */
        pAltDesc = pdevInfo->altQueue.pDescQueueHead;

    } /* while loop to walk through the alt queue */

    /* disable int to handle race condition with swba ISR */
    DisableInterruptsGlobaly (pdevInfo);
    if (pdevInfo->altQueue.pDescQueueHead == NULL) {
        pdevInfo->altQueue.pDescQueueTail = NULL;
    }
    EnableInterruptsGlobaly (pdevInfo);

    aPrintk("TX: done processing alt queue\n");
}

/*******************************************************************************
* processTxQueue/processTxInterrupt - interrupt service routines
*
* This routine is called by processInt to handle the transmitted interrupt.  It
* releases the tx buffers to the buffer pool.  It also releases the descriptors
* to the tx queue.
*/
void
processTxQueue(WLAN_DEV_INFO *pdevInfo)
{
    HAL_DESC_TX_STATUS  *pTxStatus;
    ATHEROS_DESC        *pTxDesc;
    ATHEROS_DESC        *pPreLastDesc = NULL;
    WLAN_STATS          *pLocalStats  = &pdevInfo->localSta->stats;

    /* Lock the tx queue */
    aPrintk("\nTX: Inside of processTxQueue\n");
    A_SEM_LOCK(pdevInfo->txQueue.qSem, WAIT_FOREVER);

    pTxDesc = pdevInfo->txQueue.pDescQueueHead;

    /*
     * We iterate over all the descriptors.  However, only the first descriptor
     * of a frame contains the control information and only the last descriptor
     * of a frame contains the status information information.  Furthermore,
     * when the hardware runs out of tx descs, hw will expect to use the final
     * desc as a springboard to get to a (newly-chained) desc when it is finally
     * restarted; so software must not free this descriptor.  When we fully
     * process a descriptor, it is marked stale so we can see our own
     * breadcrumbs and skip over that desc the next time thru the loop.
     */
    while (pTxDesc) {

        aPrintk("TX: pTxDesc =%p[also something to send]\n", pTxDesc);

        /* only process complete frames we've not touched before */
        if (pTxDesc->hwControl.tx.more || pTxDesc->staleFlag) {
            /* Go to the next descriptor in the same frame */
            pPreLastDesc = pTxDesc;
            pTxDesc      = pTxDesc->pNextVirtPtr;
            continue;
        }

        /* ensure we have the status correctly */
        A_DESC_CACHE_INVAL(pTxDesc);
        pTxStatus = &pTxDesc->hwStatus.tx;

        if (!pTxStatus->done) {
            /* whole frame has not finished transmit */
            if (lastDescGet(pdevInfo->txQueue.pDescQueueTail)->hwStatus.tx.done) {
                A_DESC_CACHE_INVAL(pTxDesc);
                if (!pTxStatus->done) {
                    aPrintk("TX: APPANIC: Transmit queue in inconsistent state.");
                }
            }
            break;
        }

#ifdef DEBUG_PKT_LOG
#else //DEBUG_PKT_LOG
	    //int antTx = pTxDesc->hwControl.tx.hwSpecific0;
#endif //DEBUG_PKT_LOG


        /* finished transmit of the frame? */
        if (pTxStatus->pktTransmitOK) {
            aPrintk("TX: pkt Transmit OK\n");

            /* Update the successful send statistics */
            //fadi END_ERR_ADD(pdevInfo->pOSHandle, MIB2_OUT_UCAST, +1);
            pLocalStats->GoodTransmits++;
            pLocalStats->CurrentAckSignalStrenth = (A_INT32)pTxStatus->ackSigStrength;

            if (pTxStatus->shortRetryCount) {
                pLocalStats->TotalRetries++;
                pLocalStats->shortFrameRetryBins[pTxStatus->shortRetryCount]++;
            } else if (pTxStatus->longRetryCount) {
                pLocalStats->TotalRetries++;
                pLocalStats->RetryBins[pTxStatus->longRetryCount]++;
            }
        } else {
            aPrintk("TX: !! pktTransmit NOT OK!\n");
            /*
             *  Before we collect stats, make sure we'll
             *  be able to fully process the error.
             */
//fadi            if (swretryLookAhead(pdevInfo, pTxDesc) != 0) {
//fadi                break;
//fadi            }

            /* Update the error send statistics */
            //fadi END_ERR_ADD(pdevInfo->pOSHandle, MIB2_OUT_ERRS, +1);
            pLocalStats->TransmitErrors++;

            if (pTxStatus->excessiveRetries) {
                aPrintk ("TX: Transmit excessive retries\n");
                pLocalStats->TxExcessiveRetries++;
            } else if (pTxStatus->filtered) {
                aPrintk("TX: Transmit filtered\n");
                pLocalStats->TxFiltered++;
            } else if (pTxStatus->fifoUnderrun) {
                aPrintk("TX: Transmit fifo underrun\n");
                pLocalStats->TxDmaUnderrun++;
            } else {
                aPrintk("TX: APPANIC: Unknown transmit error.");
            }
        }

        /*
         * Collect per-SIB Tx stats and/or Tx
         * completion for management frames
         */
        ASSERT(pTxDesc->pTxFirstDesc != NULL);
        processTxSibInfo(pdevInfo, pTxDesc);

        if (pdevInfo->staConfig.swretryEnabled) {
            //fadi swretry(pdevInfo, pPreLastDesc, pTxDesc) != 0)
            /*
             *  swretry() has queued up more stuff. Reset the
             *  current TxDesc ptr to head of queue and loop again.
             *  We've also lost a useful setting for pPreLastDesc,
             *  so reset it and tell kick hw just in case it had
             *  gotten to the end of queue already (quite likely).
             */
            pTxDesc = pdevInfo->txQueue.pDescQueueHead;
            pPreLastDesc = NULL;
            halStartDmaTransmit(pdevInfo, HAL_TX_QUEUE_DATA);
            
            continue;
        }

        //sib = pTxDesc->pTxFirstDesc->pDestSibEntry = NULL for now
        txFramePendDecrement(pdevInfo, NULL, pTxDesc->pTxFirstDesc);

        if (pPreLastDesc != NULL) {
            /* terminate the to-be-released descriptor chain */
            pPreLastDesc->nextPhysPtr  = 0;
            pPreLastDesc->pNextVirtPtr = NULL;
            freeBuffandDescChain(pdevInfo, pdevInfo->txQueue.pDescQueueHead);
        }

        /* Advance the trailing descriptor */
        pPreLastDesc = pTxDesc;

        /* Head is now this desc, in case the next is NULL */
        pdevInfo->txQueue.pDescQueueHead = pTxDesc;
        pdevInfo->txQueue.pDescQueueHead->pTxFirstDesc = pTxDesc;
        pTxDesc->staleFlag = 1;

        /* finally advance pTxDesc to next descriptor */
        pTxDesc = pTxDesc->pNextVirtPtr;

    } /* while (pTxDesc != NULL) */

    aPrintk("TX: release the tx queue lock\n");

    A_SEM_UNLOCK(pdevInfo->txQueue.qSem);
}

inline void
txFramePendDecrement(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib, ATHEROS_DESC *pDesc)
{
    ASSERT(pdevInfo->numTxFramesPending > 0);
    pdevInfo->numTxFramesPending--;
#if FADI
    if (pSib != NULL) {
        ASSERT(pSib->numTxPending > 0);
        pSib->numTxPending--;
        if (pDesc->psFlag) {
            ASSERT(pSib->numPSTxPending > 0);
            pSib->numPSTxPending--;
            /* update the power save queue state and TIM */
            if ((pSib->numPSTxPending == 0) &&
                (pSib->txPSQueue.qFrameCount == 0) &&
                (pSib->staState & STATE_ASSOC)) {
                wlanTimUpdate(pdevInfo, pSib->assocId, FALSE);
            }
            /*
             * this should be changed in the future when swRetry
             * implements the special process for ps frame XXX
             */
            pDesc->psFlag = 0;
        }
    }
#else  // FADI
    if (pDesc->psFlag) 
    /*
     * this should be changed in the future when swRetry
     * implements the special process for ps frame XXX
     */
        pDesc->psFlag = 0;
#endif // FADI
}


/*
 * processTxSibInfo:
 *
 * This routine is called by processInt to handle the transmitted interrupt.
 * It updates the per sib counters, handles rate adaptation feedback, and
 * forwards MGT frams the the SME code via mlmeMgtTxCompletion().
 */
inline int
processTxSibInfo(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pTxDesc)
{
    return 1;
}

void   DisableInterruptsGlobaly   (WLAN_DEV_INFO *pDev) {
    /*
     * We have to disable interrupts here before reading & modifying
     * the mask so that the ISR does not modify the mask out from under us.
     */
    if (readPlatformReg(pDev, MAC_IER) == MAC_IER_ENABLE) {
        writePlatformReg(pDev, MAC_IER, MAC_IER_DISABLE);
        aPrintk("DISABLING GLOBAL INTERRUPTSSSSSSSSSSSSSSS\n");
    }
}

void   EnableInterruptsGlobaly    (WLAN_DEV_INFO *pDev) {

    /* Re-enable interrupts if they were disabled before. */
    if (readPlatformReg(pDev, MAC_IER) == MAC_IER_DISABLE) {
        writePlatformReg(pDev, MAC_IER, MAC_IER_ENABLE);
        aPrintk("ENABLING GLOBAL INTERRUPTSSSSSSSSSSSSSSS\n");
    }
}

#ifdef RATECTRL_C
void
drvBBChange(WLAN_DEV_INFO *pdevInfo, CHAN_VALUES *pChval)
{
    aPrintk("ISR-RESET: ofdm mode, resetting rate table\n");
    // ofdm mode. can be 5 GHz or 2.4 GHz 'g' mode
    pdevInfo->staConfig.pRateTable = &globalRateTable11a;
    mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ofdmRateSet,
                    (INFO_ELEMENT *)&pdevInfo->staConfig.operationalRateSet);
    mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ofdmRateSet,
                    (INFO_ELEMENT *)&pdevInfo->localSta->operationalRateSet);
    pdevInfo->localSta->currentOperationalRateIndex = pdevInfo->staConfig.defaultOfdmRateIndex; // user configured value for 11a
    pdevInfo->localSta->defaultRateIndex = pdevInfo->staConfig.defaultOfdmRateIndex; // user configured value for 11a
    
    // Let rate control switch its tables too.
    RateCtrlBBChange(pdevInfo, pChval->channelFlags);
}

void RateCtrlBBChange(WLAN_DEV_INFO *pdevInfo, A_UINT16 channelFlags)
{

    RateCtrlTbl = &RateCtrlTblOFDM[0];
    pTxRetryHistoBB = &TxRetryHistoOFDM;
    if (channelFlags & CHANNEL_TURBO) {
        printk("I SURE HOPE WE'RE NOT IN TURBO MODE YET!!!\n");
        pDataRate = &dataRate2xTbl[0];
        if ( RateAlgSelect ) {
            RssiAckValidMin = RssiAckValidMin11aTurbo;
            RssiAckDeltaMin = RssiAckDeltaMin11aTurbo;
        }
    } else {
        pDataRate = &dataRateTbl[0];
        if ( RateAlgSelect ) {
            RssiAckValidMin = RssiAckValidMin11a;
            RssiAckDeltaMin = RssiAckDeltaMin11a;
        }
    }
    
    // recompute the local SIB parameters
    //fadi RateCtrlSibInit (pdevInfo->localSta);
}
#endif 

/*
 *  This routine will check that the following frames are marked done:
 *  If no fragmentation, the immediately subsequent frame
 *  If the current frame is a non-terminal fragment, the balance of
 *  the fragment train.
 *  All frames including the current frame are still in network order.
 */
inline int
swretryLookAhead(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pTxDesc)
{
    ATHEROS_DESC *pDesc;

    aPrintk("TX: Entering swretryLookAhead\n");
    
    /* an errored fragment train needs to be fully baked
     * for us to handle it
     */
    while (pTxDesc->pTxFirstDesc->pBufferVirtPtr.headerLittle->frameControl.moreFrag) {
        pTxDesc = lastDescGet(pTxDesc->pNextVirtPtr);
        
        aPrintk("TX: More frags detected\n");
        if (!pTxDesc->hwStatus.tx.done) {
            aPrintk("TX: tx.done == false\n");
            return 1;
        }
    }

    /* if we're not the last descriptor in the transmit queue
     * swretry can be attempted if the next descriptor is done
     */
    if (pTxDesc->pNextVirtPtr) {
        pTxDesc = lastDescGet(pTxDesc->pNextVirtPtr);
        aPrintk("TX: we're not last Desc\n");
        if (pTxDesc->hwStatus.tx.done) {
            aPrintk("TX: tx.done == TRUE..done\n");
            return 0;
        }
        return 1;
    }

    /* this frame is the last frame on the tx queue and has errored
     * out; our current code base is much happier with a holding
     * descriptor in place.. which means we can't take this frame
     * off till the next one arrives! now this situation could turn
     * into nuisance of varying severities mostly if this is the
     * only destination sta active for now - if this is the frame
     * needing a sw retry after having undergone excessive retries,
     * we might be stuck till TCP times out and sends us another
     * frame to this destination! a more severe conditon will develop
     * if this frame got filtered and will be the one clearing the
     * numFilteredPending to this destination.. new frames to this
     * destination are ending up on the swretry queue meanwhile..
     * thus nothing arrives to dislodge us from this situation unless
     * a bc/mc frame or frame to another destination happens by!
     * so the following hack proactively broadcasts a null Data
     * frame through the normal tx queue.. its not the high perf
     * case (there are no frames following this one!!) anyway..
     * so don't criticise this approach but live with it for now :-)
     */
    if (pTxDesc->pDestSibEntry != NULL) {
        aPrintk("TX: OH OH...SIBENTRY != NULL\n");
        pTxDesc->pDestSibEntry->stats.swRetryHungCount++;
    }
    pdevInfo->localSta->stats.swRetryHungCount++;
    pDesc = nullDataFrameCreate(pdevInfo, NULL);

    if (pDesc == NULL) {
#ifdef DEBUG
        aPrintk("swretryLookAhead: error creating Null Data frame!\n");
#endif
        return 1;
    }
    pDesc->hwControl.tx.frameLength += FCS_FIELD_SIZE;
    if (pdevInfo->staConfig.pChannel->channelFlags & CHANNEL_CCK) {
        aPrintk("TX: setting up RATE_11_INDEX\n");
        pDesc->hwControl.tx.transmitRate = globalRateTable11b.info[RATE_11_INDEX].plcpRate;
    } else {
        aPrintk("TX: setting up RATE_54_INDEX\n");
        pDesc->hwControl.tx.transmitRate = globalRateTable11a.info[RATE_54_INDEX].plcpRate;
    }
    halSetTxDescAntModeXmit(pdevInfo, pDesc, pdevInfo->localSta->antTx);

    /*
     * TODO: This is duplicating the code in txService() & ar5hwcDrvSend()
     * need to be clean up when the tx path re-design takes place.
     */
    pdevInfo->numTxFramesPending++;
    pDesc->pTxFirstDesc = pDesc;
    pDesc->pTxLastDesc  = pDesc;
    pDesc->staleFlag    = 0;
    halSetTxDescNoAck(pdevInfo, pDesc, 1);
    swretrySetup(pdevInfo, pDesc, NULL);
    FRAME_CONVERT_CPUTOLE(pDesc->pBufferVirtPtr.ptr,
                          pDesc->hwControl.tx.bufferLength);
    queueTxDescriptor(pdevInfo, pDesc, pDesc);
    aPrintk("TX: QUEUEING NEW TX DESC for transmission\n");


    /* We have to restart the HW queue, as it has reached its end. */
    halStartDmaTransmit(pdevInfo, HAL_TX_QUEUE_DATA);
    aPrintk("TX: Done StartDmaTransmit\n");

    return 1;
}

/*************************************************************************
*
* nullDataFrameCreate - build up a null data frame
*
* RETURNS: NULL if there is an error
*          pointer to the Null Data frame descriptor, otherwise
*/
ATHEROS_DESC *
nullDataFrameCreate(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSibEntry)
{
    A_UINT16                frameLen;
    WLAN_DATA_MAC_HEADER3   *pFrame;
    ATHEROS_DESC            *pDesc;
    WLAN_MACADDR            *pAddr;

    aPrintk("TX: Entering nullDataFrameCreate\n");

    // Compute frame length.
    frameLen = sizeof(WLAN_DATA_MAC_HEADER3);

    if (createBuffandDesc(pdevInfo, frameLen, &pDesc) != A_OK) {
        /* Caller handles any messaging */
        return NULL;
    }

    aPrintk("TX: createdbuffanddesc for frameLen=%d\n", frameLen);

    pFrame = pDesc->pBufferVirtPtr.header3;

    /* intialize the header, zero them out first */
    A_MEM_ZERO(pFrame, frameLen);

    pFrame->frameControl.protoVer = PROTOCOL_VER_0;
    pFrame->frameControl.fType    = FRAME_DATA;
    pFrame->frameControl.fSubtype = SUBT_NODATA;
    pFrame->frameControl.ToDS     = 0;
    pFrame->frameControl.FromDS   = 1;
    pFrame->frameControl.wep      = 0;

    /* initialize the 802.11 frame header address fields */
    if (pSibEntry) {
        pAddr = &pSibEntry->macAddr;
    } else {
        pAddr = &beaconBroadcastAddr;
    }

    A_MACADDR_COPY(pAddr, &pFrame->address1);
    A_MACADDR_COPY(&pdevInfo->localSta->macAddr, &pFrame->address2);
    A_MACADDR_COPY(&pdevInfo->localSta->macAddr, &pFrame->address3);

    aPrintk("TX: addr1: %x-%x-%x-%x-%x-%x\n", pFrame->address1.octets[0],
            pFrame->address1.octets[1],pFrame->address1.octets[2],
            pFrame->address1.octets[3],pFrame->address1.octets[4],
            pFrame->address1.octets[5]);
    aPrintk("TX: addr2: %x-%x-%x-%x-%x-%x\n", pFrame->address2.octets[0],
            pFrame->address2.octets[1],pFrame->address2.octets[2],
            pFrame->address2.octets[3],pFrame->address2.octets[4],
            pFrame->address2.octets[5]);
    aPrintk("TX: addr3: %x-%x-%x-%x-%x-%x\n", pFrame->address3.octets[0],
                pFrame->address3.octets[1],pFrame->address3.octets[2],
                pFrame->address3.octets[3],pFrame->address3.octets[4],
                pFrame->address3.octets[5]);


    pDesc->hwControl.tx.frameLength  = frameLen;

    return pDesc;
}


/*************************************************************************
*
* ar521x_transmit - transmit raw data out the wireless (pkts from Forwarder)
*
* RETURNS: A_OK is successful
*          A_ERROR otherwise
*
* This routine will be called by the forwarder to hand me raw data (including
* 802.11 frames) to be sent down to DMA (then to wireless). This raw data will
* start with the HAL_DESC_TX_CONTROL info and then the HAL_DESC_TX_STATUS
* then followed by the data portion. So we will have to bypass the re-setting
* of the important bits in the TX_CONTROL/STATUS bits later on by the
* wlanFrameTransmit() routine.
*
* ignore: this argument is deemed necessary for intersil portion of this
* common interface.
*
* derived from AUTH_answer piece in atheros_handle_mgt_frame()
*/
A_STATUS ar521x_transmit  (void *ignore, 
                           struct sk_buff *skb,
                           struct net_device *dev)
{
    atheros_local_info_t *local = (atheros_local_info_t *) dev->priv;
    WLAN_DEV_INFO *pdevInfo = local->pWlanDev;
    // skb starts points to data only (802.11 contents w/o hwcontrol/status)
    unsigned char* hdr = skb->data;
    A_UINT16 data_len  = skb->len;
    ATHEROS_DESC* pDesc;

    pDesc = memAllocateDescriptor (pdevInfo);

    if (!pDesc) {
        printk("TX: FAILED TO ALLOC DESC\n");
        dev_kfree_skb(skb);
        return A_ERROR;
    }
    pDesc->pBufferVirtPtr.ptr = skb->data;
    pDesc->bufferPhysPtr = A_DESC_V2P (skb->data);
    pDesc->pOSDescPtr = skb;
    pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;

    // need to handle fragmented IP pkts
    if (skb_shinfo(skb)->frag_list) {
        struct sk_buff *skb2;
        int tlen;
        // the data exists on multiple skbs
        tlen = skb_headlen(skb);
        

        skb2 = skb_shinfo(skb)->frag_list;
        while (skb2 && (tlen < skb->len))   {
            if(skb->tail < skb->end) {
                memcpy(skb->tail, skb2->data, skb2->len);
                skb->tail += (unsigned int)skb2->len;
                skb->len  += (unsigned int)skb2->len;
            }
            tlen += skb2->len;           
            skb2 = skb2->next;
        }
      
        pDesc->pBufferVirtPtr.ptr = skb->data;
        pDesc->bufferPhysPtr = A_DESC_V2P (skb->data);
        pDesc->pOSDescPtr = skb;
        pDesc->pOrigBufferVirtPtr = pDesc->pBufferVirtPtr.ptr;
    }


    // in skb, before data starts, the immediate 16B should contain
    // the HAL_DESC_TX_CONTROL/HAL_DESC_TX_STATUS
    memcpy( (void*)(&(pDesc->hwControl.tx)), 
            (hdr-(sizeof(HAL_DESC_TX_CONTROL)+sizeof(HAL_DESC_TX_STATUS))), 
            sizeof(HAL_DESC_TX_CONTROL) );

    // set these fields in the pDesc/pFrame portion
    //pDesc->hwControl.tx.frameLength, set in forwarder, fcs added in txService
    pDesc->hwControl.tx.frameLength  = data_len;
    pDesc->hwControl.tx.bufferLength = data_len;

  
    pDesc->pDestSibEntry = NULL;

    // call the atheros_sap.c FrameTransmit version
    if (wlanFrameTransmit(pdevInfo, pDesc, TRUE) == A_OK)
    {
        // FN - no need to freeBuff freeBuffandDesc(pdevInfo, pDesc);
    }
    
    return A_OK;
}

A_STATUS ar5210SetChannel(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval);
A_STATUS ar5211SetChannel(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval);
/*************************************************************************
*
* atheros_set_channel - Sets the channel number on the hardware
*
* RETURNS: A_OK is successful
*          A_ERROR otherwise
*
* We expect a channel number, such as 1,2,3... corresponding to
* equivalent 5210, 5218, ....etc
*
*/
A_STATUS atheros_set_channel (A_UINT16 channel)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; // global saved copy
    A_STATUS status = A_ERROR;

    // we're setting the wireless mode to 11A
    A_UINT16 opModes = MODE_SELECT_11A;
 
    /*staConfig.pClist is already initialized in the init_local routine */

    /* Setup the channels */
    pdevInfo->staConfig.pChannel = 
        wlanGetChannelPtr(pdevInfo->staConfig.pClist, channel, opModes);
        
    /* Fix channel config */
    if (pdevInfo->staConfig.pChannel == NULL) {
        pdevInfo->staConfig.pChannel = &(pdevInfo->staConfig.pClist->chanArray[0]);
        printk("WARNING: Setting channel frequency to be a member of the configured regulatory domain and operational modes\n");
    }
    pdevInfo->staConfig.phwChannel = pdevInfo->staConfig.pChannel;
    pdevInfo->bssDescr->pChannel   = pdevInfo->staConfig.pChannel;
    pdevInfo->localSta->Channel    = pdevInfo->bssDescr->pChannel->channel;

#ifdef BUILD_AR5210
    if (A_OK != (status = ar5210SetChannel(pdevInfo, pdevInfo->staConfig.pChannel))) {
#else //AR5211
    if (A_OK != (status = ar5211SetChannel(pdevInfo, pdevInfo->staConfig.pChannel))) {
#endif // AR5210
        printk("atheros_set_channel: Failed to set channel #%d\n", channel);
    }
    else 
        g_channel_operational = channel;

    return status;
}


/*************************************************************************
*
* atheros_get_channel - Gets the channel number
*
* RETURNS: int, channel number
*
*
*/

A_UINT16 atheros_get_channel (void)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; // global saved copy
    A_UINT16 chan_num = 0;

    // we're setting the wireless mode to 11A
    A_UINT16 opModes = MODE_SELECT_11A;
    
    /*staConfig.pClist is already initialized in the init_local routine */

    pdevInfo->staConfig.pChannel = 
        wlanGetChannelPtr(pdevInfo->staConfig.pClist, g_channel_operational, opModes);
        //wlanGetChannelPtr(pdevInfo->staConfig.pClist, CHAN_11A_DEFAULT, opModes);
    
    if (pdevInfo->staConfig.pChannel) {
        chan_num = MHZ2IEEE_CH(g_channel_operational);
    }
    
    return chan_num;
}

/*************************************************************************
*
* atheros_get_ammode - get operational air monitor mode
*
* RETURNS: int, channel number
*
*
*/

A_UINT16 atheros_get_ammode  (void)
{
    return g_monitor_type;
}

/*************************************************************************
*
* atheros_set_ammode - set operational air monitor mode
*
* RETURNS: N/A
*
*
*/
void atheros_set_ammode  (monitor_type mode)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; // global saved copy

    g_monitor_type = mode;
    if (mode == ATHEROS_MONITOR_ON) 
        halRxFilter(pdevInfo, HAL_RX_PROM, HAL_RX_FILTER_SET);
    else if (mode == ATHEROS_MONITOR_OFF) 
        halRxFilter(pdevInfo, HAL_RX_PROM, HAL_RX_FILTER_CLEAR);
    else
        printk("atheros_set_ammode: INVALID MONITOR_TYPE MODE!!\n");
}


/*************************************************************************
*
* atheros_set_essid - set operational ESSID going out on beacons
*
* RETURNS: N/A
*
*
*/
A_STATUS atheros_set_essid  (char* essid)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; 
    struct net_device *dev = NULL;
    atheros_local_info_t *local;

    if (NULL != gPciDev) 
        dev = pci_get_drvdata(gPciDev);
    else 
        return A_ERROR;
    
    local = (atheros_local_info_t *) dev->priv;
                                
    if (strlen(essid) >= MAX_SSID_LEN) {
        printk("INVALID Essid (length=%d)\n", strlen(essid));
        return A_ERROR;
    }

    strcpy(local->essid, essid);

    pdevInfo->staConfig.ssid.elementID 		  = ELE_SSID;
    pdevInfo->staConfig.ssid.length			  = strlen(essid); 
    strcpy(pdevInfo->staConfig.ssid.ssid, essid); 
    pdevInfo->staConfig.ssid.ssid[pdevInfo->staConfig.ssid.length]='\0';


    // Basic fields, addresses and sssid..BSSID is what makes it out on BEACONS
    pdevInfo->bssDescr->ssid.elementID		  = ELE_SSID;
    pdevInfo->bssDescr->ssid.length 		  = strlen(essid); 
    strcpy(pdevInfo->bssDescr->ssid.ssid, essid); 
    pdevInfo->bssDescr->ssid.ssid[pdevInfo->bssDescr->ssid.length]='\0';

    DisableInterruptsGlobaly(pdevInfo);

    wlanBeaconFree(pdevInfo);
    wlanBeaconInit(pdevInfo, 0);

    EnableInterruptsGlobaly(pdevInfo);
    
    return A_OK;
}

/*************************************************************************
*
* atheros_get_essid - get operational ESSID going out on beacons
*
* RETURNS: Essid
*
*
*/
char* atheros_get_essid  (void)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; // global saved copy
    return (pdevInfo->bssDescr->ssid.ssid);
}



void resetChipForBeacons(WLAN_DEV_INFO *pdevInfo, struct net_device *dev)
{
    if (NULL == dev) {
        aPrintk("ISR-RESET: device not yet attached\n");
        return;
    }

    aPrintk("ISR-RESET: inside of resetwlan...\n");
    

    /* If things look very bleak, then reboot */
    if (pdevInfo->localSta->stats.hwResetCount > RESET_COUNT_ALLOW) {
        printk("--------------------------------------------------------------\n");
        printk("ISR-RESET: resetwlan: reset too many times (%ld). Rebooting...\n",
               pdevInfo->localSta->stats.hwResetCount);
        printk("--------------------------------------------------------------\n");
        // we'll need a reboot func call here
    }

    if (pdevInfo->NicResetInProgress == TRUE) {
		aPrintk("ISR-RESET: !!! NIC already in reset process\n");
        return;
    }

	/* Queue up a reset job. */
    pdevInfo->NicResetInProgress = TRUE;
    pdevInfo->localSta->staState = STATE_AP_DOWN;
    aPrintk("ISR-RESET: setting staState -> down\n");
    halRxFilter(pdevInfo, 0, HAL_RX_FILTER_INIT);
    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);

    /* Clear the running flag */
    netif_stop_queue(dev);	    

    aPrintk("ISR-RESET: Entering resetOnError #####\n");

    /* Keep track of number of resets */
    pdevInfo->localSta->stats.hwResetCount++;
    pdevInfo->hwResetCount++;

#ifdef DEBUG
    logMsg("ISR-RESET: Number of resets = %ld\n", pdevInfo->hwResetCount);
#endif

    if ((NULL == pdevInfo) || (NULL == pdevInfo->staConfig.pChannel)) {
        aPrintk("DEVICE NOT READY FOR RESETonERROR\n");
        return;
    }
    
    aPrintk("ISR-RESET: Entering chipReset ++++++++\n");

    A_SEM_LOCK(pdevInfo->txQueue.qSem, WAIT_FOREVER);

    pdevInfo->globISRReg = 0;

    /* Disable interrupts and stop receive. */
    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    halStopPcuReceive(pdevInfo);

    if (halReset(pdevInfo, pdevInfo->staConfig.pChannel, FALSE) != A_OK) {
        A_SEM_UNLOCK(pdevInfo->txQueue.qSem);
        /* reboot the whole system */
        printk("ISR-RESET: chipReset: Error resetting WLAN HW!  Rebooting...\n");
        printk("ISR-RESET: need some server action here...\n");
        return;
    }

    /* do the needful for rates, etc. after switching baseband */
    if ((pdevInfo->staConfig.pChannel->channelFlags & CHANNEL_CCK) ^
        (pdevInfo->staConfig.phwChannel->channelFlags & CHANNEL_CCK))
    {
#ifdef RATECTRL_C
        drvBBChange(pdevInfo, pdevInfo->staConfig.pChannel);
        aPrintk("ISR-RESET: done drvBBChange...\n");
#endif
    }

    halEnableReceive(pdevInfo);
    halStartPcuReceive(pdevInfo);

    halRxFilter(pdevInfo, HAL_RX_UCAST, HAL_RX_FILTER_SET);
    
    halEnableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    pdevInfo->localSta->staState = STATE_AP_UP;

    /* Set the running flag */
    struct net_device *pNetDev = NULL;
    if (NULL != gPciDev) pNetDev = pci_get_drvdata(gPciDev);
    if (NULL != pNetDev) netif_start_queue(pNetDev); //IFF_UP | IFF_RUNNING
    aPrintk("ISR-RESET: setting staState -> UP\n");

    wlanBeaconFree(pdevInfo);
    printk("ISR-RESET: ### finished wlanBeaconFree...\n");
    if (A_OK != wlanBeaconInit(pdevInfo, 0)) printk("FAILED TO RESET CHIP!\n");
    printk("ISR-RESET: ### finished wlanBeaconInit...\n");
    //fadi beaconWdgStart(pdevInfo); will need to start beacon watchdog

#ifdef DEBUG
    logMsg("ISR-RESET: Finished resetting the wlan chip\n");
#endif

    /* Clear the reset flag */
    pdevInfo->NicResetInProgress = FALSE;

    aPrintk("ISR-RESET: RE-enabling GLOBAL INTS\n");
}


/**********************************************************
 *
 *
 */
int atheros_set_tim(struct net_device *dev, u16 aid, int set){
  // set == 1 if TIM is to be set, 0 otherwise

  atheros_local_info_t *local = (atheros_local_info_t *)dev->priv;

  wlanTimUpdate(local->pWlanDev, aid, set);

  return 0;
}


/**********************************************************
 *
 *
 */
extern __s16 power_levels[3][5];

A_STATUS atheros_set_tx_power (int power_level)
{
    // we're expecting a power level between 350 and 2750 from 
    // power_levelsp[][] in rra_msg.h

    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; 

    // for atomicity's sake!??
    DisableInterruptsGlobaly(pdevInfo);

#ifdef CONFIG_ATHEROS_5210_PCI
    AR5210_setup_power_level (power_level);
#else
    AR5211_setup_power_level (power_level);
#endif

    EnableInterruptsGlobaly(pdevInfo);
    return A_OK;
}       


/**********************************************************
 *
 *
 */
A_STATUS atheros_set_dtim_interval (int dtim_interval)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; 
    struct net_device *dev = NULL;
    atheros_local_info_t *local;
    A_STATUS status = A_ERROR;

    if (NULL != gPciDev) {
        dev = pci_get_drvdata(gPciDev);
        if (dev) local = (atheros_local_info_t *)dev->priv;
        if (NULL == local) return A_ERROR;
    }
    else 
        return A_ERROR;

    // for atomicity's sake!??
    DisableInterruptsGlobaly(pdevInfo);

    pdevInfo->bssDescr->DTIMPeriod  = dtim_interval;
    local->dtim_period              = dtim_interval;

    wlanBeaconFree(pdevInfo);
    status = wlanBeaconInit(pdevInfo, 0);

    EnableInterruptsGlobaly(pdevInfo);
    return status;
}

/**********************************************************
 *
 *
 */
A_STATUS atheros_set_beacon_interval (int beacon_interval)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; 
    struct net_device *dev = NULL;
    atheros_local_info_t *local;
    A_STATUS status = A_ERROR;

    if (NULL != gPciDev) {
        dev = pci_get_drvdata(gPciDev);
        if (dev) local = (atheros_local_info_t *)dev->priv;
        if (NULL == local) return A_ERROR;
    }
    else 
        return A_ERROR;

    // for atomicity's sake!??
    DisableInterruptsGlobaly(pdevInfo);

    pdevInfo->staConfig.beaconInterval = beacon_interval;
    pdevInfo->bssDescr->beaconInterval = beacon_interval;
    local->beacon_int                  = beacon_interval;

    wlanBeaconFree(pdevInfo);
    status = wlanBeaconInit(pdevInfo, 0);

    EnableInterruptsGlobaly(pdevInfo);
    return status;
}


/**********************************************************
 *
 *
 */
A_STATUS atheros_set_short_pream (A_BOOL short_pream)
{
    // as far as the atheros 5210/11 chipsets, in .a mode, they
    // only work in long-preamble mode, so just ignore short
    // mode...
    return A_OK;
}

/**********************************************************
 *
 * P.S. conversion of rate indices needed, look at sap_config.h
 * macros vs. wlandrv.h ones.
 */
A_STATUS atheros_set_tx_rate (int tx_rate)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; 
    
    // for atomicity's sake!??
    DisableInterruptsGlobaly(pdevInfo);

    switch (tx_rate) {
    case WIFI_RATE_6MBPS:
        tx_rate = RATE_6_INDEX;
        break;
    case WIFI_RATE_9MBPS:
        tx_rate = RATE_9_INDEX;
        break;
    case WIFI_RATE_12MBPS:
        tx_rate = RATE_12_INDEX;
        break;
    case WIFI_RATE_18MBPS:
        tx_rate = RATE_18_INDEX;
        break;
    case WIFI_RATE_24MBPS:
        tx_rate = RATE_24_INDEX;
        break;
    case WIFI_RATE_36MBPS:
        tx_rate = RATE_36_INDEX;
        break;
    case WIFI_RATE_48MBPS:
        tx_rate = RATE_48_INDEX;
        break;
    case WIFI_RATE_54MBPS:
        tx_rate = RATE_54_INDEX;
        break;
    default:
        printk("UNSUPPORTED 802.11b rate value. Try again\n");
        return A_ERROR;
        break;
    };

    // these new rates should be picked up by the txService routine
    pdevInfo->staConfig.defaultOfdmRateIndex        = (A_UINT8)tx_rate;
    pdevInfo->localSta->currentOperationalRateIndex = (A_UINT8)tx_rate; 
    pdevInfo->localSta->defaultRateIndex            = (A_UINT8)tx_rate; 
    pdevInfo->localSta->basicRateIndex              = (A_UINT8)tx_rate;

#ifdef RATECTRL_C
    drvBBChange (pdevInfo, pdevInfo->staConfig.pChannel);
#endif
    EnableInterruptsGlobaly(pdevInfo);

    return A_OK;
}

/**********************************************************
 *
 * Will enable the device is enabled = TRUE, otherwise, shut
 * it off by stopping the receive and disabling the PHY so no
 * transmitions out on the air (mgmt or other)
 *
 */
A_STATUS atheros_set_enable (A_BOOL enable)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; 
    
    // for atomicity's sake!??
    DisableInterruptsGlobaly(pdevInfo);

    if (TRUE == enable) {
        /* Activate the PHY */
        writePlatformReg(pdevInfo, PHY_ACTIVE, PHY_ACTIVE_EN);
        // Wait a bit 
        udelay(1000);

        pdevInfo->localSta->staState = STATE_AP_UP;
        halEnableReceive (pdevInfo);
        halStartPcuReceive (pdevInfo);
    }
    else {
        
        // FN - look at apStopWlan() for more stopping action!!

        /* Deactivate the PHY */
        writePlatformReg(pdevInfo, PHY_ACTIVE, PHY_ACTIVE_DIS);
        // Wait a bit 
        udelay(1000);

        halStopPcuReceive (pdevInfo);
        pdevInfo->localSta->staState = STATE_AP_DOWN;
    }

    EnableInterruptsGlobaly(pdevInfo);
    return A_OK;
}

#ifdef CONFIG_ATHEROS_5210_PCI

/* Default Power Settings for channels outside of EEPROM range */
static const A_UINT8 ar5k0007_pwrSettings[17] = {
//        gain delta                       pc dac
// 54  48  36  24  18  12   9   54  48  36  24  18  12   9   6  ob  db
    9,  9,  0,  0,  0,  0,  0,   2,  2,  6,  6,  6,  6,  6,  6,  2,  2
};

A_UCHAR getGainF(TPC_MAP *pRD, A_UCHAR pcdac, A_UCHAR *dBm);
A_UCHAR getPcdac(TPC_MAP *pRD, A_UCHAR dBm);
A_UINT32 reverseBits(A_UINT32 val, A_UINT32 bitCount);

void AR5210_setup_power_level (int power_level)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; 
    A_UCHAR         gainFRD, gainF36, gainF48, gainF54;
    A_UCHAR         dBmRD, dBm36, dBm48, dBm54, dontcare;
    A_UINT32        rd, group;
    TPC_MAP         *pRD;
    struct eepMap   *pEepData;
    A_UINT16        i, num_pwr_regs;
    A_UCHAR         cp[17];
    A_UINT32        pwr_regs[17] = {
        0x00000000, 0x00000000, 0x00000000,
        0x00000000, 0x00000000, 0xf0000000,
        0xcc000000, 0x00000000, 0x00000000,
        0x00000000, 0x0a000000, 0x000000e2,
        0x0a000020, 0x01000002, 0x01000018,
        0x40000000, 0x00000418
    };
    A_UINT16        channel = pdevInfo->staConfig.pChannel->channel;

    A_DRIVER_BCOPY(ar5k0007_pwrSettings, cp, 17);

    // following is based on setupPowerSettings()
    ASSERT(pdevInfo->pHalInfo && pdevInfo->pHalInfo->pEepData);

    pEepData = pdevInfo->pHalInfo->pEepData;

    // Set OB/DB Values regardless of channel
    cp[15] = (A_UCHAR)((pEepData->biasCurrents >> 4) & 0x7);
    cp[16] = (A_UCHAR)(pEepData->biasCurrents & 0x7);
                                                                                       
    if ((channel < 5170) || (channel > 5320)) {
        return;
    }

    if (((pEepData->version >> 12) & 0xF) == 1) {
        /* Match regulatory domain */
        for (rd = 0; rd < REGULATORY_DOMAINS; rd++) {
            if ((pEepData->currentRD) == pEepData->regDomain[rd]) {
                break;
            }
        }
        if (rd == REGULATORY_DOMAINS) {
            uiPrintf("setupPowerSettings: No calibrated regulatory domain matches the set value 0x%0lx\n", rd);
            return;
        }
        
        for (i=0; i < 5; i++) {
            if (group == power_levels[1][i]) { // row1 contains the .a values
                group = i;
                break;
            }
        }
        
        pRD   = &pEepData->tpc[group];

        /* Set PC DAC Values */
        cp[14] = pRD->regdmn[rd];
        cp[9]  = A_MIN(pRD->regdmn[rd], pRD->rate36);
        cp[8]  = A_MIN(pRD->regdmn[rd], pRD->rate48);
        cp[7]  = A_MIN(pRD->regdmn[rd], pRD->rate54);

        /* Find Corresponding gainF values for RD, 36, 48, 54 */
        gainFRD = getGainF(pRD, pRD->regdmn[rd], &dBmRD);
        gainF36 = getGainF(pRD, cp[9], &dBm36);
        gainF48 = getGainF(pRD, cp[8], &dBm48);
        gainF54 = getGainF(pRD, cp[7], &dBm54);

        /* Power Scale if requested */
        if (pdevInfo->staConfig.tpScale != TP_SCALE_MAX) {
            switch (pdevInfo->staConfig.tpScale) {
            case TP_SCALE_50:
                dBmRD -= 3;
                break;
            case TP_SCALE_25:
                dBmRD -= 6;
                break;
            case TP_SCALE_12:
                dBmRD -= 9;
                break;
            case TP_SCALE_MIN:
                dBmRD = 3;
                break;
            default:
                uiPrintf("Illegal Transmit Power Scale Value: %d\n",
                    pdevInfo->staConfig.tpScale);
            }
            cp[14]  = getPcdac(pRD, dBmRD);
            gainFRD = getGainF(pRD, cp[14], &dontcare);
            dBm36   = A_MIN(dBm36, dBmRD);
            cp[9]   = getPcdac(pRD, dBm36);
            gainF36 = getGainF(pRD, cp[9], &dontcare);
            dBm48   = A_MIN(dBm48, dBmRD);
            cp[8]   = getPcdac(pRD, dBm48);
            gainF48 = getGainF(pRD, cp[8], &dontcare);
            dBm54   = A_MIN(dBm54, dBmRD);
            cp[7]   = getPcdac(pRD, dBm54);
            gainF54 = getGainF(pRD, cp[7], &dontcare);
        }
        /* Record current dBm at rate 6 */
        pdevInfo->tx6PowerInDbm = dBmRD;

        cp[13] = cp[12] = cp[11] = cp[10] = cp[14];

        /* Set GainF Values */
        cp[0] = (A_UCHAR)(gainFRD - gainF54);
        cp[1] = (A_UCHAR)(gainFRD - gainF48);
        cp[2] = (A_UCHAR)(gainFRD - gainF36);
        cp[3] = cp[4] = cp[5] = cp[6] = 0;  // 9, 12, 18, 24 have no gain_delta from 6
    }

    
    // following is based on ar5210InitTransmitPower()
    if ( (cp[15] < 1) || (cp[15] > 5) ) {
        uiPrintf("ar5210InitTransmitPower: OB should be between 1 and 5 inclusively, not %d", cp[15]);
        printk("ar5210InitTransmitPower: OB should be between 1 and 5 inclusively, not %d", cp[15]);
        return;
    }
    if ( (cp[16] < 1) || (cp[16] > 5) ) {
        uiPrintf("ar5210InitTransmitPower: DB should be between 1 and 5 inclusively, not %d", cp[16]);
        printk("ar5210InitTransmitPower: DB should be between 1 and 5 inclusively, not %d", cp[16]);
        return;
    }

    /* reverse bits of the transmit power array */
    for (i = 0; i < 7; i++) {
        cp[i] = (A_UINT8)reverseBits(cp[i], 5);
    }
    for (i = 7; i < 15; i++) {
        cp[i] = (A_UINT8)reverseBits(cp[i], 6);
    }

    /* merge transmit power values into the register - quite gross */
    pwr_regs[0] |= ((cp[1] << 5) & 0xE0) | (cp[0] & 0x1F);
    pwr_regs[1] |= ((cp[3] << 7) & 0x80) | ((cp[2] << 2) & 0x7C) | ((cp[1] >> 3) & 0x03);
    pwr_regs[2] |= ((cp[4] << 4) & 0xF0) | ((cp[3] >> 1) & 0x0F);
    pwr_regs[3] |= ((cp[6] << 6) & 0xC0) | ((cp[5] << 1) & 0x3E) | ((cp[4] >> 4) & 0x01);
    pwr_regs[4] |= ((cp[7] << 3) & 0xF8) | ((cp[6] >> 2) & 0x07);
    pwr_regs[5] |= ((cp[9] << 7) & 0x80) | ((cp[8] << 1) & 0x7E) | ((cp[7] >> 5) & 0x01);
    pwr_regs[6] |= ((cp[10] << 5) & 0xE0) | ((cp[9] >> 1) & 0x1F);
    pwr_regs[7] |= ((cp[11] << 3) & 0xF8) | ((cp[10] >> 3) & 0x07);
    pwr_regs[8] |= ((cp[12] << 1) & 0x7E) | ((cp[11] >> 5) & 0x01);
    pwr_regs[9] |= ((cp[13] << 5) & 0xE0);
    pwr_regs[10] |= ((cp[14] << 3) & 0xF8) | ((cp[13] >> 3) & 0x07);
    pwr_regs[11] |= ((cp[14] >> 5) & 0x01);

    /* Set OB */
    pwr_regs[8] |=  (reverseBits(cp[15], 3) << 7) & 0x80;
    pwr_regs[9] |=  (reverseBits(cp[15], 3) >> 1) & 0x03;

    /* Set DB */
    pwr_regs[9] |=  (reverseBits(cp[16], 3) << 2) & 0x1C;

    /* Write the registers */
    num_pwr_regs = sizeof(pwr_regs) / sizeof(A_UINT32);
    for (i = 0; i < num_pwr_regs; i++) {
        if (i != num_pwr_regs - 1) {
            writePlatformReg(pdevInfo, 0x0000989c, pwr_regs[i]);
        } else {
            /* last write is a flush */
            writePlatformReg(pdevInfo, 0x000098d4, pwr_regs[i]);
        }
    }

    return;
}

#else // CONFIG_ATHEROS_5210_PCI

A_UINT16
ar5211GetScaledPower(A_UINT16 channel, A_UINT16 pcdacValue, PCDACS_EEPROM *pSrcStruct);
A_UINT16
ar5211GetInterpolatedValue(A_UINT16 target, A_UINT16 srcLeft, A_UINT16 srcRight,
                           A_UINT16 targetLeft, A_UINT16 targetRight, A_BOOL scaleUp);

void AR5211_setup_power_level (int power_level)
{
    WLAN_DEV_INFO* pdevInfo = pwlanDevInfo; 
    static FULL_PCDAC_STRUCT pcdacStruct;
    static A_UINT16          pcdacTable[PWR_TABLE_SIZE];
    PCDACS_EEPROM *pSrcStruct;
    A_UINT16     i, j;
    A_UINT16     *pPcdacValues;
    A_INT16      *pScaledUpDbm;
    A_INT16      minScaledPwr;
    A_INT16      maxScaledPwr;
    A_UINT16     pcdacMin = 0;
    A_UINT16     pcdacMax = 63;
    A_UINT16     pcdacTableIndex;
    A_UINT16     scaledPcdac;
    A_UINT32     addr;
    A_UINT32     temp32;
    
    A_UINT16        numChannels = 0;
    TRGT_POWER_INFO *pPowerInfo = NULL;
    PCDACS_EEPROM   eepromPcdacs;
    struct eepMap   *pData;
    
    CHAN_VALUES  *pChval = pdevInfo->staConfig.pChannel;
    A_UINT16     channel = pdevInfo->staConfig.pChannel->channel;

    printk("Entering AR5211_setup_power_level..[for chan=%d][power_level=%d]\n", channel, power_level);

    // status = ar5211SetTransmitPower (pdevInfo, pdevInfo->staConfig.pChannel);
    ASSERT(pdevInfo->pHalInfo && pdevInfo->pHalInfo->pEepData);

    pData = pdevInfo->pHalInfo->pEepData;

    /* setup the pcdac struct to point to the correct info, based on mode */
    switch (pChval->channelFlags &
        (CHANNEL_OFDM | CHANNEL_CCK | CHANNEL_5GHZ | CHANNEL_2GHZ))
    {
    case (CHANNEL_OFDM | CHANNEL_5GHZ):
        printk("\t5Ghz...channel\n");
        eepromPcdacs.numChannels     = pData->pPcdacInfo->numChannels11a;
        eepromPcdacs.pChannelList    = pData->pPcdacInfo->Channels11a;
        eepromPcdacs.pDataPerChannel = pData->pPcdacInfo->DataPerChannel11a;
        numChannels = pData->pTrgtPowerInfo->numTargetPwr_11a;
        pPowerInfo  = pData->pTrgtPowerInfo->trgtPwr_11a;

        printk("\tnumChannels\%d[eepromNumChannels=%d]\n", numChannels, eepromPcdacs.numChannels);

        break;

    case (CHANNEL_OFDM | CHANNEL_2GHZ):
        printk("\tOFDM/2Ghz .a channel\n");
        eepromPcdacs.numChannels     = pData->pPcdacInfo->numChannels2_4;
        eepromPcdacs.pChannelList    = pData->pPcdacInfo->Channels2_4;
        eepromPcdacs.pDataPerChannel = pData->pPcdacInfo->DataPerChannel11g;
        numChannels = pData->pTrgtPowerInfo->numTargetPwr_11g;
        pPowerInfo  = pData->pTrgtPowerInfo->trgtPwr_11g;
        break;

    case (CHANNEL_CCK | CHANNEL_2GHZ):
        printk("\twhat..a 11.b channel??\n");
        eepromPcdacs.numChannels     = pData->pPcdacInfo->numChannels2_4;
        eepromPcdacs.pChannelList    = pData->pPcdacInfo->Channels2_4;
        eepromPcdacs.pDataPerChannel = pData->pPcdacInfo->DataPerChannel11b;
        numChannels = pData->pTrgtPowerInfo->numTargetPwr_11b;
        pPowerInfo  = pData->pTrgtPowerInfo->trgtPwr_11b;
        break;
    default:
        ASSERT(0);
        break;
    }

    pSrcStruct = &eepromPcdacs;

    // the following is based on ar5211SetPowerTable()
    A_MEM_ZERO(&pcdacStruct, sizeof(FULL_PCDAC_STRUCT));
    A_MEM_ZERO(pcdacTable, sizeof(A_UINT16) * PWR_TABLE_SIZE);
    pPcdacValues = pcdacStruct.PcdacValues;
    pScaledUpDbm = pcdacStruct.PwrValues;

    /* Initialize the pcdacs to dBM structs pcdacs to be 1 to 63 */
    for (i = PCDAC_START, j = 0; i <= PCDAC_STOP; i+= PCDAC_STEP, j++) {
        pPcdacValues[j] = i;
    }

    pcdacStruct.numPcdacValues = j;
    pcdacStruct.pcdacMin = PCDAC_START;
    pcdacStruct.pcdacMax = PCDAC_STOP;

    /* Fill out the power values for this channel */
    for (j = 0; j < pcdacStruct.numPcdacValues; j++ ) {
        pScaledUpDbm[j] = ar5211GetScaledPower(channel, pPcdacValues[j], pSrcStruct);
    }

    /* Now scale the pcdac values to fit in the 64 entry power table */
    minScaledPwr = pScaledUpDbm[0];
    maxScaledPwr = pScaledUpDbm[pcdacStruct.numPcdacValues - 1];

    printk("\tminScaledPwr=%d, MaxScaledPwr=%d\n", minScaledPwr, maxScaledPwr);

    /* find minimum and make monotonic */
    for (j = 0; j < pcdacStruct.numPcdacValues; j++) {
        if (minScaledPwr >= pScaledUpDbm[j]) {
            minScaledPwr = pScaledUpDbm[j];
            pcdacMin = j;
        }
        /*
         * Make the full_hsh monotonically increasing otherwise interpolation algorithm will get fooled
         * gotta start working from the top, hence i = 63 - j.
         */
        i = (A_UINT16)(pcdacStruct.numPcdacValues - 1 - j);
        if (i == 0) {
            break;
        }
        if (pScaledUpDbm[i-1] > pScaledUpDbm[i]) {
            //it could be a glitch, so make the power for this pcdac,
            //the same as the power from the next highest pcdac
            pScaledUpDbm[i - 1] = pScaledUpDbm[i];
        }
    }

    for (j = 0; j < pcdacStruct.numPcdacValues; j++) {
        if (maxScaledPwr < pScaledUpDbm[j]) {
            maxScaledPwr = pScaledUpDbm[j];
            pcdacMax = j;
        }
    }

    /* Find the first power level with a pcdac */
//    pwr = (A_UINT16)(PWR_STEP * ((minScaledPwr - PWR_MIN + PWR_STEP / 2) / PWR_STEP)  + PWR_MIN);

    /* Write all the first pcdac entries based off the pcdacMin */
    pcdacTableIndex = 0;
    for (i = 0; i < (2 * (power_level - PWR_MIN) / EEP_SCALE + 1); i++) {
        pcdacTable[pcdacTableIndex++] = pcdacMin;
    }

    i = 0;

    printk("\tpower_level=%d,  pScaledUpDbm[pcdacStruct.numPcdacValues - 1]=%d\n",
           power_level, pScaledUpDbm[pcdacStruct.numPcdacValues - 1]);

    if (power_level < pScaledUpDbm[pcdacStruct.numPcdacValues - 1]) {
//        pwr += PWR_STEP;
        /* stop if dbM > max_power_possible */
        while ((power_level < pScaledUpDbm[pcdacStruct.numPcdacValues - 1]) &&
               ((power_level - pScaledUpDbm[i])*(power_level - pScaledUpDbm[i+1]) > 0))
        {
            i++;
        }

        printk("\tmax_power_possible[i=%d]\n", i);

        /* scale by 2 and add 1 to enable round up or down as needed */
        scaledPcdac = (A_UINT16)(ar5211GetInterpolatedValue(power_level, pScaledUpDbm[i],
            pScaledUpDbm[i+1], (A_UINT16)(pPcdacValues[i]*2),
            (A_UINT16)(pPcdacValues[i+1]*2), 0 ) + 1);

        pcdacTable[pcdacTableIndex] = scaledPcdac / 2;
        if (pcdacTable[pcdacTableIndex] > pcdacMax) {
            pcdacTable[pcdacTableIndex] = pcdacMax;
        }
        pcdacTableIndex++;
    }

    /* Write all the last pcdac entries based off the last valid pcdac */
    
    printk("\tpcdacTableIndex = %d\n", pcdacTableIndex);

    while (pcdacTableIndex < PWR_TABLE_SIZE) {
        pcdacTable[pcdacTableIndex] = pcdacTable[pcdacTableIndex - 1];
        pcdacTableIndex++;
    }

    /* Finally, write the power values into the baseband power table */
    addr = PHY_BASE + (608 << 2);
    for (i = 0; i < 32; i++) {
        temp32 = 0xffff & ((pcdacTable[2 * i + 1] << 8) | 0xff);
        temp32 = (temp32 << 16) | (0xffff & ((pcdacTable[2 * i] << 8) | 0xff));
        writePlatformReg(pdevInfo, addr, temp32);
        printk("\tWriting [addr=%lx][val=%x]\n", addr, temp32);
        addr += 4;
    }
    
    return;
}

#endif // CONFIG_ATHEROS_5210_PCI
