/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

#ident  "ACI $Id: //depot/sw/releases/2.0.1/src/ap/bufconv.c#1 $"

/*
 * This module implements the conversion between mBlk and Atheros descriptors
 * functions.
 *
 * mBlk are specially designed for VxWorks.  It is defined as a structure of
 * clusters.
 *
 * typedef struct mBlk {
 *        M_BLK_HDR   mBlkHdr;        // header
 *        M_PKT_HDR   mBlkPktHdr;     // pkthdr
 *        CL_BLK     *pClBLK;         // pointer to cluster blk
 * } M_BLK;
 *
 * // header at beginning of each mBlk
 *
 * typedef struct mHdr {
 *        struct mBlk *mNext;      // next buffer in chain
 *        struct mBlk *mNextPkt;   // next chain in queue/record
 *        char        *mData;      // location of data
 *        int          mLen;       // amount of data in this mBlk
 *        UCHAR        mType;      // type of data in this mBlk
 *        UCHAR        mFlags;     // flags; see below
 *        USHORT       reserved;
 * } M_BLK_HDR;
 *
 * record/packet header in first mBlk of chain; valid if M_PKTHDR set
 *
 * typedef struct  pktHdr {
 *        struct ifnet *rcvif;  // rcv interface
 *        int           len;    // total packet length
 * } M_PKT_HDR;
 *
 * description of external storage mapped into mBlk, valid if M_EXT set
 *
 * typedef struct clBlk {
 *        CL_BLK_LIST clNode;         // union of next clBlk, buffer ptr
 *        UINT        clSize;         // cluster size
 *        int         clRefCnt;       // reference count of the cluster
 *        FUNCPTR     pClFreeRtn;     // pointer to cluster free routine
 *        int         clFreeArg1;     // free routine arg 1
 *        int         clFreeArg2;     // free routine arg 2
 *        int         clFreeArg3;     // free routine arg 3
 *        struct netPool *pNetPool;   // pointer to the netPool
 *  } CL_BLK;
 *
 * The Atheros descriptors, ATHEROS_DESC, is a combination of the MAC
 * hardware DMA descriptor and software only fields.
 *
 * The header and frame body are put into separate descriptors.  Chances are the
 * header will always be a contingous physical buffer and the frame body may
 * occupy serveral smaller physical buffers.

 * All clusters are allocated as received buffers for the incoming data.
 *
 * The MBlk and Atheros structures are the different way of the link for
 * the input data buffers.
 *
 * To convert Atheros Descriptors to mBlk structures: XXX
 *
 * To convert mBlk structures to Atheros Descriptors: XXX
 *
 */

/* includes */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wireless.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>

#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/io.h>

#include "wlantype.h"
#include "wlanproto.h"
#include "ui.h"
#include "apdefs.h"
#include "apmem.h"
#include "halApi.h"

/*******************************************************************************
*
* vxDescToSkb - converts Atheros Descriptors to linux skb structures
*
* The input data from MAC has to be put into clusters directly. The clusters
* are reserved for the receive function during initialization phase by calling
* netClusterGet(). And the receive descriptors are pre-allocated and built up
* using clusters.
* After the incoming frame is received and put into the descriptor, i.e. cluster
* Then the clusters are linked using Mblk structures chain and passed to the
* protocol or DSM layer.
*
* the input descriptor is kept for DS purpose. The calling part should free the
* input descriptor when it is no longer needed or error condition
*
*
* RETURNS:  NULL if unsuccesful
*           pointer of mBlk structures if successful
*/
struct sk_buff *
vxDescToSkb(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_BOOL dup)
{
#if FADI
    ATHEROS_DESC    *pCurrDesc;
    M_BLK           *pMblk = NULL;
    M_BLK           *pPrevMblk = NULL;
    M_BLK           *pVarMblk, *pAllocMblk;
    ATHEROS_DESC    *pHeadDesc = NULL;

    pHeadDesc = pDesc;

    /* any field can we use to check the integrity of descriptor ??? */

    while (pDesc != NULL) {
        /* reserve a Mblk */
        if ( (pAllocMblk = netMblkGet(pNetPool, M_DONTWAIT, MT_DATA)) == NULL ) {
            /* ran out of mblks, a receive resource => loaned resources aren't
             * being returned in time; slow ethernet? slow stack?
             */
            LOG_DROP_FRAME(pdevInfo, rx);
            if (pMblk != NULL) {
                netMblkClChainFree(pMblk);
            }
            freeBuffandDescChain(pdevInfo, dup ? pHeadDesc : pDesc);
            return NULL;
        }

        /* join Mblk */
        if ((pVarMblk = netMblkClJoin(pAllocMblk, (CL_BLK *)pDesc->pOSDescPtr)) == NULL ) {
            /* generally shouldn't happen */
            uiPrintf("\nvxDescToMblk: error in join mBlk\n");
            netMblkFree(pNetPool, pAllocMblk);
            if (pMblk != NULL) {
                netMblkClChainFree(pMblk);
            }
            freeBuffandDescChain(pdevInfo, dup ? pHeadDesc : pDesc);
            return NULL;
        }

        /* assign the first M_BLK */
        if (pMblk == NULL) {
            pMblk = pVarMblk;
            pMblk->mBlkHdr.mFlags |= M_PKTHDR;
            pMblk->mBlkPktHdr.len = 0;
        }

        pVarMblk->mBlkHdr.mNext = NULL;
        pVarMblk->mBlkHdr.mNextPkt = NULL;
        pVarMblk->mBlkHdr.mData = (A_UCHAR *)pDesc->bufferPhysPtr;
        pVarMblk->mBlkHdr.mLen = pDesc->hwStatus.rx.dataLength;

        /* calculate the total length of the frame */
        pMblk->mBlkPktHdr.len += pDesc->hwStatus.rx.dataLength;

        /* get the previous mblk in the chain */
        if (pPrevMblk != NULL) {
            pPrevMblk->mBlkHdr.mNext = pVarMblk;
        }

        /* next descriptor */
        pCurrDesc = pDesc;
        pDesc = pDesc->pNextVirtPtr;
        pPrevMblk = pVarMblk;

        /* free the current descriptor */
        if (dup) {
            ((CL_BLK *)(pCurrDesc->pOSDescPtr))->clRefCnt += 1;
        } else {
            memFreeDescriptor(pdevInfo, pCurrDesc);
        }
    }

    return pMblk;
#else // FADI
    
    ATHEROS_DESC    *pCurrDesc;
    ATHEROS_DESC    *pHeadDesc = NULL;

    pHeadDesc = pDesc;
    /* any field can we use to check the integrity of descriptor ??? */

    while (pDesc != NULL) {
        /* next descriptor */
        pCurrDesc = pDesc;
        pDesc = pDesc->pNextVirtPtr;
        
        /* free the current descriptor */
        if (!dup) {
            memFreeDescriptor(pdevInfo, pCurrDesc);
        }
    }

    // we want to return SKB here
    return 0;

#endif // FADI

}

/*******************************************************************************
*
* vxMblkToDesc - converts VxWorks mBlk structures to Atheros Descriptors
*
* This routine allocates a descriptor to put the 802.3 header in it and then
* walk through all the strucures to link the data.  this routine could be part
* of the driver sending routine
*
* The calling routine is responsible for the releasing of input mBlk due to DS
*
* Ref: Tornado BSP Developer's kit for VxWorks pp 366
*
* RETURNS:  NULL if unsuccesful
*           pointer of Atheros Descriptors if successful
*/
ATHEROS_DESC*
vxSkbkToDesc(WLAN_DEV_INFO *pdevInfo, struct sk_buff *skb)
{
    ATHEROS_DESC    	*pDesc = NULL;
    ATHEROS_DESC    	*pVarDesc;
    ATHEROS_DESC    	*pLastDesc = NULL;
    HAL_DESC_TX_CONTROL *pHwControl;
    M_BLK           	*pCurrMblk;

    /* verify the first MBLK contains the header */
    if ((pMblk->mBlkHdr.mFlags & M_PKTHDR) != M_PKTHDR) {
        uiPrintf("vxMblkToDesc: M_PKTHDR not set on first mblk\n");
        netMblkClChainFree(pMblk);
        return NULL;
    }

    /* generate a descriptor for each of the Mblk */
    while (pMblk) {
        /* sometimes the vxWorks protocol stack passes Mblk with buffer length = 0 */
        if (pMblk->mBlkHdr.mLen == 0) {
            pCurrMblk = pMblk;
            pMblk = pMblk->mBlkHdr.mNext;
            netMblkClFree(pCurrMblk);
            continue;
        }

        /* allocate a descriptor for Mblk, i.e. frame data */
        pVarDesc = memAllocateDescriptor(pdevInfo);
        if (!pVarDesc) {
            LOG_DROP_FRAME(pdevInfo, tx);
            /* clean up before leaving */
            if (pDesc) {
                /* we should clean all descriptors and buffers */
                freeBuffandDescChain(pdevInfo, pDesc);
            }
            netMblkClChainFree(pMblk);
            return NULL;
        }

        /* Initialize the fields of descriptor */
        pVarDesc->bufferPhysPtr      = (A_UINT32)A_DATA_V2P(pMblk->mBlkHdr.mData);
        pVarDesc->pBufferVirtPtr.ptr = pMblk->mBlkHdr.mData;
        pVarDesc->pOSDescPtr         = pMblk->pClBlk;

        /* fill in the frame size */
        pHwControl = &pVarDesc->hwControl.tx;
        pVarDesc->hwControl.tx.bufferLength = pMblk->mBlkHdr.mLen;
        halSetTxDescPktType(pdevInfo, pVarDesc, HAL_DESC_PKT_TYPE_NORMAL);

        /*
         * first descriptor with header, but protocol stack also passes us the
         * second Mblk with M_PKTHDR set ???
         */
        if (pDesc == NULL) {
            pDesc = pVarDesc;
            pDesc->hwControl.tx.frameLength = pMblk->mBlkPktHdr.len;
        }

        /* if there was a descriptor before this, so need to link */
        if (pLastDesc != NULL) {
            pLastDesc->nextPhysPtr = pVarDesc->thisPhysPtr;
            pLastDesc->pNextVirtPtr = pVarDesc;

            /* set more descriptor bit */
            pLastDesc->hwControl.tx.more = 1;
        }

        /* update variables for next time round loop */
        pLastDesc = pVarDesc;
        pCurrMblk = pMblk;
        pMblk = pMblk->mBlkHdr.mNext;

        /* grab the reference to the current cluster and free the mblk */
        pCurrMblk->pClBlk->clRefCnt += 1;
        netMblkClFree(pCurrMblk);
    }

    return pDesc;
}
