/*
 * Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

#ident  "ACI $Id: //depot/sw/releases/2.0.1/src/common/fragde.c#1 $"

/*
DESCRIPTION

This module implements the fragmentation/de-fragmentation functions.

Please refer to the Access Point Software Desgin Doc for details

Fragmentation
=============

An output frame is presented as a descriptor chain. For each descriptor
the physical buffer length could be different. The fragmentation will
have to first examine the length of frame, if it is not longer than
the dot11FragmentionThreshold for the STA then no fragmentation is
needed. Otherwise we walk through the descriptors until the partial
length of the frame reaches dot11FragmentionThreshold, we need to
allocate a new descriptor chain for the next dot11FragmentionThreshold
length and so on until the last descriptor chain is shorter than
the dot11FragmentionThreshold. We need to add a 802.11 header for each
new allocated descriptor chain which is the duplicate of the original
header and with each header including the original one but except the
last one has More Frag bit set.

The result would a descriptor chain which chains all fragments together

Defragmentation
===============

An incoming frame could be just a fragment of a complete frame by
checking the More Frag bit in its 802.11 header in the first descriptor
of the descriptor chain. A partial frame descriptor chain ( only limited
number < 3 is allowed ) is reserved for the complete frame. First we need
to examine the frame length and More Frag bit of the incoming descriptor
chain.

if ((frame length < dot11FragmentionThreshold) & More Frag bit clear)
    no defragmentation needed
    this is the only fragment of the frame
    pass up, Done

if ((frame length < dot11FragmentionThreshold) & More Frag bit set)
    defragmentation needed
    this is the last fragment of the frame

if ((frame length = dot11FragmentionThreshold) & More Frag bit set)
    defragmentation needed
    this is one of the fragment of the frame

for the latter two situation

we need allocate a 802.11 header which is the duplicate of the header
of the incoming fragment with More Frag bit cleared and link the rest
descriptors of all fragments together and get ride of the 802.11 headers.
Also need to update the frame length.
*/

/* includes */

#ifdef    VXWORKS

#include "vxWorks.h"
#include "semLib.h"
#include "stdio.h"
#include "tickLib.h"
#include "assert.h"

#else

//#include <stdio.h>
//#include <stdlib.h>

#endif

#include "wlantype.h"
#include "wlanfragde.h"
#include "wlanproto.h"
#include "wlandrv.h"
#include "wlanframe.h"
#include "ui.h"

//#ifdef VXWORKS
#include "apmem.h"
//#else
#include "wlanext.h"
//#include "mem.h"
//#include "hwext.h"
//#endif

#ifdef DEBUG
int fragVerbose = 0;
int membufDebug;
int fragErrs = 0;
#endif

ATHEROS_DESC *tailDescFind(ATHEROS_DESC *pDesc);
A_UINT32 wlanDeFragArFree(WLAN_DEV_INFO *);
A_UINT32 wlanDeFragArSearch(WLAN_DEV_INFO *, WLAN_MACADDR *, A_UINT16 seqNum, A_UINT16 fragNum);
void wlanDeFragArAge(WLAN_DEV_INFO *);

#ifdef VXWORKS
#define CREATE_HEADER_BUFF_AND_DESC(_pdevInfo, _hdrlen, _ppDesc)    \
    ( ((*(_ppDesc)) = createHeaderBuffandDesc(_pdevInfo, _hdrlen)) ? A_OK : A_ERROR )
#define MEM_ALLOCATE_DESCRIPTOR(_pdevInfo, _ppDesc)                 \
    ( ((*(_ppDesc)) = memAllocateDescriptor(_pdevInfo)) ? A_OK : A_ERROR )
#else // linux
#define CREATE_HEADER_BUFF_AND_DESC(_pdevInfo, _hdrlen, _ppDesc)    \
    createBuffandDesc(_pdevInfo, _hdrlen, (void **)(_ppDesc))
#define MEM_ALLOCATE_DESCRIPTOR(_pdevInfo, _ppDesc)                 \
( ((*(_ppDesc)) = memAllocateDescriptor(_pdevInfo)) ? A_OK : A_ERROR )
    //memAllocateDescriptor(_pdevInfo, (void **)(_ppDesc))
#endif

/*******************************************************************************
 *
 *  wlanFrag - fragmentation function
 *
 * AP: No fragmentation for broadcast/multicast frames including beacons
 *     No fragmentation for control frames cause they are short enough.
 * STA: Infrastructure mode all frames are unicast frames since they are all
 *      sent to AP.
 *      Ad-hoc mode the broadcast/multicast including beacons are not
 *      fragmentable.
 *
 * The function is called during TX service
 *
 * RETURNS: ERROR if unsuccessful
 *           OK if successful
 */

ATHEROS_DESC *
wlanFrag(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc) /* fragment pDesc */
{
    WLAN_DATA_MAC_HEADER3 *pWlanHdr, *pHdrHdr;
    ATHEROS_DESC *pFragDesc;            /* output descriptor */
    ATHEROS_DESC *pVarDesc;             /* variable desc */
    ATHEROS_DESC *pHdrDesc;
    ATHEROS_DESC *pDataDesc;
    A_UINT16     pduSize, frameLen, hdrLen, fragLen;    /* length variables */
    A_UINT16     pos, fragNum = 0, numFrags;            /* frag vasiables */
    A_UINT16     fragThreshold;

    /* decide if we need to frag the frame */

    frameLen      = (A_UINT16)pDesc->hwControl.tx.frameLength;
    fragThreshold = (A_UINT16)pdevInfo->staConfig.userFragThreshold;

    /* Skip fragmentation if it's too short */
    if ((frameLen <= fragThreshold)) {
        return pDesc;
    }

    /* Skip or for AP/ad hoc multicast */
    if ((pdevInfo->localSta->serviceType == WLAN_AP_SERVICE) &&
        isGrp(&pDesc->pBufferVirtPtr.header3->address1)) {
        return pDesc;
    }

    /* fragmentation begins ... */

    /* calculate the number of fragments, the pdu size */
    hdrLen = headerLengthGet(pDesc);
    pduSize = fragThreshold - hdrLen;
    numFrags = (frameLen - hdrLen)/pduSize +
        (((frameLen - hdrLen)%pduSize == 0) ? 0 : 1);

    /* assign a master header buffer */
    pWlanHdr = pDesc->pBufferVirtPtr.header3;

    /* assign the temp and return descriptor */
    pFragDesc = pVarDesc = pDesc;

#ifdef DEBUG
    if (membufDebug)
        uiPrintf("\nfragmentation starts...\n");
#endif

    /* for the first descriptor with header */
    pWlanHdr->frameControl.moreFrag = 1;
    pVarDesc->hwControl.tx.frameLength = pduSize + hdrLen;
    fragLen = (A_UINT16)pVarDesc->hwControl.tx.bufferLength - hdrLen;

    while (fragNum < numFrags) {

        while (fragLen >= pduSize) {
            /* calculate new buffer length */
            fragLen -= pduSize;
            pos = (A_UINT16)pVarDesc->hwControl.tx.bufferLength - fragLen;
            pVarDesc->hwControl.tx.bufferLength = pos;

            /* complete a fragment if buffer length is equal to fragThreshold */
            /* no more descripor for this fragment */
            pVarDesc->hwControl.tx.more = 0;

            if ((fragLen == 0) && (pVarDesc->pNextVirtPtr == NULL)) {
                break;
            }

#ifdef DEBUG
            if (membufDebug)
                uiPrintf("fragLen: %d, pduSize: %d\n", fragLen, pduSize);
#endif
            /* start a new fragment by creating a new header */
            if (CREATE_HEADER_BUFF_AND_DESC(pdevInfo, hdrLen, &pHdrDesc) != A_OK) {
                /* clean up before leave */
                freeBuffandDescChain(pdevInfo, pFragDesc);
#ifdef VXWORKS
                LOG_DROP_FRAME(pdevInfo, tx);
#endif
#ifdef DEBUG
                if (fragVerbose)
                    uiPrintf("\nFragment header buffer allocation failed.\n");
                ++fragErrs;
#endif
                return NULL;
            }

            A_BCOPY(pWlanHdr, pHdrDesc->pBufferVirtPtr.ptr, hdrLen);
            pHdrHdr = pHdrDesc->pBufferVirtPtr.header3;

            /* set the header and descriptor */
            pHdrHdr->sequenceControl.fragNo = ++fragNum;
            if (fragNum != (A_UINT16)(numFrags - 1)) {
                pHdrHdr->frameControl.moreFrag = 1;
                pHdrDesc->hwControl.tx.frameLength = pduSize + hdrLen;
            } else {
                pHdrHdr->frameControl.moreFrag = 0;
                pHdrDesc->hwControl.tx.frameLength = frameLen - pduSize*fragNum;
            }

            pHdrDesc->hwControl.tx.more = 1;
            pHdrDesc->pDestSibEntry = pFragDesc->pDestSibEntry;
            pHdrDesc->pSrcSibEntry = pFragDesc->pSrcSibEntry;

#ifdef DEBUG
            if (membufDebug)
                uiPrintf("bufRefUpdate -- fragLen: %d, pduSize: %d\n", fragLen, pduSize);
#endif

            if (fragLen == 0) {
#ifndef VXWORKS
                /* for NDIS, increasing the reference count for the packet */
                bufferRefUpdate(pVarDesc);
#endif
                /* link the descriptors */
                pDataDesc = pVarDesc->pNextVirtPtr;

                pVarDesc->pNextVirtPtr = pHdrDesc;
                pVarDesc->nextPhysPtr = pHdrDesc->thisPhysPtr;

                pHdrDesc->pNextVirtPtr = pDataDesc;
                pHdrDesc->nextPhysPtr = pDataDesc->thisPhysPtr;

                pVarDesc = pHdrDesc;
            } else {
                /* update the reference count for buffer(vxWorks), packet(NDIS) */
                bufferRefUpdate(pVarDesc);
                /* allocate the descriptor for remaining data portion */
                if (MEM_ALLOCATE_DESCRIPTOR(pdevInfo, &pDataDesc) != A_OK) {
#ifdef VXWORKS
                    LOG_DROP_FRAME(pdevInfo, tx);
#endif
                    /* clean up before leave */
                    freeBuffandDescChain(pdevInfo, pFragDesc);
                    freeBuffandDescChain(pdevInfo, pHdrDesc);
#ifdef DEBUG
                    if (fragVerbose)
                        uiPrintf("\ndescriptor allocation error!\n");
                    ++fragErrs;
#endif
                    return NULL;
                }

                /* update the OS dependent info in the descriptor */
                descInfoDup(pVarDesc, pDataDesc);

                /* set descriptor field */
                pDataDesc->pBufferVirtPtr.byte = pVarDesc->pBufferVirtPtr.byte + pos;
                pDataDesc->bufferPhysPtr = pVarDesc->bufferPhysPtr + pos;

                pDataDesc->hwControl.tx.more = 1;
                pDataDesc->hwControl.tx.bufferLength = fragLen;

                /* keep the original buffer characteristics here */
                pDataDesc->pOSDescPtr = pVarDesc->pOSDescPtr;
                pDataDesc->pOrigBufferVirtPtr = pVarDesc->pOrigBufferVirtPtr;

                /* link descriptors */
                pDataDesc->pNextVirtPtr = pVarDesc->pNextVirtPtr;
                pDataDesc->nextPhysPtr = pVarDesc->nextPhysPtr;

                pVarDesc->pNextVirtPtr = pHdrDesc;
                pVarDesc->nextPhysPtr = pHdrDesc->thisPhysPtr;

                pHdrDesc->pNextVirtPtr = pDataDesc;
                pHdrDesc->nextPhysPtr = pDataDesc->thisPhysPtr;

                pVarDesc = pDataDesc;
            }
        }

        /* continue a fragment */
        if (pVarDesc->pNextVirtPtr == NULL) {
            pVarDesc->hwControl.tx.more = 0;
            break;
        } else {
            pVarDesc->hwControl.tx.more = 1;
            pVarDesc = pVarDesc->pNextVirtPtr;
            fragLen += (A_UINT16)pVarDesc->hwControl.tx.bufferLength;
        }

    }

    return pFragDesc;
}

/*******************************************************************************
*
*  wlanDeFrag - De-fragmentation function
*
* AP: de-fragmentation for broadcast/multicast frames from WM
*     de-fragmentation for data and management frames
* STA: no de-fragmentation for broadcast/multicast frames in any mode
*
* the function is called in the RX service
*
* RETURNS:    NULL if the fragment is consumed by the pool
*             pointer of assemblied complete frame structure.
*/

ATHEROS_DESC *
wlanDeFrag(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pFragDesc)
{
    PARTIAL_DESC *deFragDescArray = pdevInfo->defragLanes;
    ATHEROS_DESC *pLastDesc, *pReassembled = NULL;
    A_UINT32     bufIndex;      /* index for pre-allocated defrag structures */
    A_UINT32    moreFrag;

    // printf("\nwe are in de-frag\n");

    // fadi added...
    //FRAME_CONVERT_CPUTOLE(pFragDesc->pBufferVirtPtr.header,
	  //  		pFragDesc->hwStatus.rx.bufferLength);

    FRAME_CONVERT_LETOCPU(pFragDesc->pBufferVirtPtr.header,
				pFragDesc->hwStatus.rx.dataLength);

    


    /* wind up our aging clock */
    wlanDeFragArAge(pdevInfo);

    /* check the current fragment frame whether fragmented or not */
    if (((moreFrag = moreFragGet(pFragDesc)) == 0) &&
        (fragNumGet(pFragDesc) == 0)) {
        printk ("*** NOT FRAGMENTED[len=%d][frag=%d][seq=%d]\n", 
        lengthGet(pFragDesc),fragNumGet(pFragDesc), seqNumGet(pFragDesc));
        FRAME_CONVERT_CPUTOLE(pFragDesc->pBufferVirtPtr.header,
	    		pFragDesc->hwStatus.rx.dataLength);

        /* not fragmented, return  */
        // uiPrintf("\n non fraged frame\n");
        return pFragDesc;
    }

    printk("DeFrag info: \n");
    printk("\tlengthGet = %d\n", lengthGet(pFragDesc));
    printk("\tHdrLengthGet = %d\n", headerLengthGet(pFragDesc));
    printk("\tisFrag = %s\n", (isFrag(pFragDesc->pBufferVirtPtr.header))?"TRUE":"FALSE");
    printk("\tmoreFragGet = %d\n", moreFragGet(pFragDesc));
    printk("\tfragNumGet = %d\n", fragNumGet(pFragDesc));
    printk("\tseqNumGet = %d\n", seqNumGet(pFragDesc));



    /* initial MPDU of fragmentated MSDU. */
    if (fragNumGet(pFragDesc) == 0) {
        printk("initial mpdu of fragmented msdu\n");
        /* first descriptor and find free partial frame to begin Msdu rx */
        if ((bufIndex = wlanDeFragArFree(pdevInfo)) == A_ERROR) {
            printk("error11111\n");
            /* give us one more shot at it */
            wlanDeFragArAge(pdevInfo);

            /* check again. if not, reaches max limit partial frame numbers */
            if ((bufIndex = wlanDeFragArFree(pdevInfo)) == A_ERROR) {
                printk("error22222\n");
                freeBuffandDescChain(pdevInfo, pFragDesc);
                return NULL;
            }
        }

        deFragDescArray[bufIndex].inUse = TRUE;
        deFragDescArray[bufIndex].rta = &pFragDesc->pBufferVirtPtr.header3->address2;
        deFragDescArray[bufIndex].rsn  = seqNumGet(pFragDesc);
        deFragDescArray[bufIndex].rCur = fragNumGet(pFragDesc);
        deFragDescArray[bufIndex].timeStamp = A_MS_TICKGET();
        deFragDescArray[bufIndex].pDesc  = pFragDesc;
        deFragDescArray[bufIndex].hdrLen = headerLengthGet(pFragDesc);
        deFragDescArray[bufIndex].pduLen = lengthGet(pFragDesc) -
            deFragDescArray[bufIndex].hdrLen;

        pFragDesc->pBufferVirtPtr.header->frameControl.moreFrag = 0;
        return NULL;

    } else {    /* intermediate or final Mpdu of the fragmented Msdu */

        printk("intermediate or final mpdu of frag msdu\n");
        /* search for the right buffer */
        bufIndex = wlanDeFragArSearch(pdevInfo,
                                      &pFragDesc->pBufferVirtPtr.header3->address2,
                                      seqNumGet(pFragDesc),
                                      fragNumGet(pFragDesc));

        if (bufIndex == A_ERROR) {
            printk("error33333333333\n");
            /* free the input descriptor and buffer */
            // uiPrintf("\ninvalid fragment!\n");
            freeBuffandDescChain(pdevInfo, pFragDesc);
            return NULL;
        }

        /* make sure it fits into WLAN pdu space */
        deFragDescArray[bufIndex].pduLen += lengthGet(pFragDesc)
            - deFragDescArray[bufIndex].hdrLen;

        if (deFragDescArray[bufIndex].pduLen > MAX_WLAN_BODY_SIZE) {
            /* frame is too long, throw away all fragments */

            printk("frame is tooooo long\n");
            /* free the input descriptor and buffer */
            freeBuffandDescChain(pdevInfo, pFragDesc);

            /* free the defrag partial frame buffer back to pool*/
            freeBuffandDescChain(pdevInfo, deFragDescArray[bufIndex].pDesc);
            deFragDescArray[bufIndex].inUse = FALSE;

            return NULL;
        }

        /* frame fits into the WLAN pdu space */
        deFragDescArray[bufIndex].rCur = fragNumGet(pFragDesc);
        pLastDesc = tailDescFind(deFragDescArray[bufIndex].pDesc);
        pLastDesc->hwStatus.rx.more = 1;

        /* skip duplicate headers */
        if (deFragDescArray[bufIndex].hdrLen == pFragDesc->hwStatus.rx.dataLength) {
            printk("skipping dup headers\n");
            /* first descriptor only contains the header */
            pLastDesc->nextPhysPtr  = pFragDesc->nextPhysPtr;
            pLastDesc->pNextVirtPtr = pFragDesc->pNextVirtPtr;

            /* free the input descriptor and buffer */
            freeBuffandDesc(pdevInfo, pFragDesc);
        } else {

            printk("treating first desc..len=%d\n", pFragDesc->hwStatus.rx.dataLength);

            /* first descriptor contains the header + data */
            pLastDesc->nextPhysPtr  = pFragDesc->thisPhysPtr;
            pLastDesc->pNextVirtPtr = pFragDesc;

            /* adjust buffer pointer */
            pFragDesc->bufferPhysPtr += deFragDescArray[bufIndex].hdrLen;
            pFragDesc->pBufferVirtPtr.byte += deFragDescArray[bufIndex].hdrLen;

            pFragDesc->hwStatus.rx.dataLength -= deFragDescArray[bufIndex].hdrLen;
        }

        /* got em all */
        if (moreFrag == 0) {
            /* free the partial frame back to the pool */
            pReassembled = deFragDescArray[bufIndex].pDesc;
            deFragDescArray[bufIndex].inUse = FALSE;

            /* return the completed assembled frame */
            FRAME_CONVERT_CPUTOLE(pReassembled->pBufferVirtPtr.header,
	    		pReassembled->hwStatus.rx.dataLength);
            printk("ASSEMBLED: new_len = %d\n", pReassembled->hwStatus.rx.dataLength);
            return pReassembled;
        }

        return NULL;
    }
}

/*******************************************************************************
*
*  wlanArFree - free entities of De-fragmentation Mpdu partial arrays
*
* the function is called in the de-fragmentation process
*
* RETURNS:    A_ERROR if unsuccessful, no entries free
*            the index of a free entry if successful
*/
A_UINT32 wlanDeFragArFree(WLAN_DEV_INFO *pdevInfo)
{
    PARTIAL_DESC *deFragDescArray = pdevInfo->defragLanes;
    A_UINT32 arrayIndex;

    for (arrayIndex = 0; arrayIndex < pdevInfo->numDefragLanes; arrayIndex++) {
        if (deFragDescArray[arrayIndex].inUse == FALSE) {
            return arrayIndex;
        }
    }

    return A_ERROR;
}

/*******************************************************************************
*
*  wlanArSearch - search entities of De-fragmentation Mpdu partial arrays to
*         find the entity with matching sequence number, fragment number
*         destination address.
*
* the function is called in the de-fragmentation process
*
* RETURNS:    -1 if unsuccessful, no match found
*            the index of the first match entry if successful
*/
A_UINT32 wlanDeFragArSearch
(
    WLAN_DEV_INFO   *pdevInfo,
    WLAN_MACADDR    *addr,                    /* destination address */
    A_UINT16        seqNum,                   /* sequence number */
    A_UINT16        fragNum                   /* fragment number */
)
{
    PARTIAL_DESC *deFragDescArray = pdevInfo->defragLanes;
    A_UINT32 arrayIndex;

    for (arrayIndex = 0; arrayIndex < pdevInfo->numDefragLanes; arrayIndex++) {
        if ((deFragDescArray[arrayIndex].inUse == TRUE) &&
            (deFragDescArray[arrayIndex].rsn == (A_UINT16)seqNum) &&
            (deFragDescArray[arrayIndex].rCur == (fragNum - 1)) &&
            (A_MACADDR_COMP(deFragDescArray[arrayIndex].rta, addr) == 0)) {
            return arrayIndex;
        }
    }

    return A_ERROR;
}


/*******************************************************************************
*
*  lanDeFragArAge - free entities of De-fragmentation Mpdu partial arrays which
*                   age > maxReceiveLifeTime.
*/
void wlanDeFragArAge(WLAN_DEV_INFO *pdevInfo)
{
    PARTIAL_DESC *deFragDescArray = pdevInfo->defragLanes;
    A_UINT32 arrayIndex, now;

    now = A_MS_TICKGET();
    for (arrayIndex = 0; arrayIndex < pdevInfo->numDefragLanes; arrayIndex++) {
        if ((deFragDescArray[arrayIndex].inUse == TRUE) &&
            ((now - deFragDescArray[arrayIndex].timeStamp) >=
             pdevInfo->staConfig.maxReceiveLifeTime)) {

            deFragDescArray[arrayIndex].inUse = FALSE;
            freeBuffandDescChain(pdevInfo, deFragDescArray[arrayIndex].pDesc);
            deFragDescArray[arrayIndex].pDesc = NULL;
        }
    }
}

/*******************************************************************************
*
* tailDescFind - search for the tail of the descriptor chain.
*
* This function is called in the de-fragmentation process.
*
* RETURNS: the pointer of the tail descriptor
*/
ATHEROS_DESC *tailDescFind(ATHEROS_DESC *pDesc)
{
    /* traverse list of input descriptors to get tail descriptor */
    while (pDesc->pNextVirtPtr != NULL) {
        pDesc = pDesc->pNextVirtPtr;
    }

    return pDesc;
}

#if defined(VXWORKS) && defined(DEBUG)
void fragShow(void)
{
    PARTIAL_DESC *deFragDescArray = pwlanDevInfo->defragLanes;
    A_UINT32 arrayIndex;

    uiPrintf("FragErrs: %d\n", fragErrs);
    for (arrayIndex = 0; arrayIndex < pwlanDevInfo->numDefragLanes; arrayIndex++) {
        uiPrintf("deFragDescArray[%d] inUse:%d, seqNum:%d, fragNum:%d, hdrLen:%d, pduLen:%d\n",
                 arrayIndex, deFragDescArray[arrayIndex].inUse,
                 deFragDescArray[arrayIndex].rsn,
                 deFragDescArray[arrayIndex].rCur,
                 deFragDescArray[arrayIndex].hdrLen,
                 deFragDescArray[arrayIndex].pduLen);
    }
}
#endif
