/*
 * Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * frameutil.c - General support routines for 802.11 frames
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/common/frameutil.c#2 $"

#ifdef VXWORKS

#include "VxWorks.h"
#include "stdio.h"
#include "sysLib.h"

#else

//#include <malloc.h>
//#include <memory.h>
//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
#include "linuxdefs.h"

#endif

#include "wlantype.h"
#include "wlanproto.h"
#include "wlanext.h"
#include "wlansmeext.h"
#include "ui.h"
#include "display.h"
#include "halApi.h"

/* mgtElementGet - Get a pointer to an element in the frame.
 *
 * This routine walks through the frame looking for the
 * information element (802.11 section 7.3.2) in the frame
 * and returns a pointer to the desired element.
 *
 * PARAMETERS:
 *  pmacHdr     Pointer to the 802.11 frame
 *  franeSize   Total size of the frame, including FCS.
 *  elementId   Element ID to search for, e.g. ELE_TIM
 *
 * RETURNS: pointer to INFO_ELEMENT data structure.
 *
 * Returns NULL if it is not a management frame or matching
 * elementId is not found.
 */

INFO_ELEMENT *
mgtElementGet(WLAN_MGT_MAC_HEADER *pmacHdr, A_UINT16 frameSize, A_UINT8 elementId)
{
    A_UINT16        subtype;
    A_UINT8         *p1;
    A_UINT16        len, elen;

    if (pmacHdr->frameControl.fType != FRAME_MGT) {
        return NULL;
    }

    subtype = (A_UINT16) pmacHdr->frameControl.fSubtype;
    switch (subtype) {
    case SUBT_ASSOC_REQ:
        p1 = (A_UINT8 *) &(((WLAN_FRAME_ASSOC_REQ *)pmacHdr)->ssid);
        break;

    case SUBT_ASSOC_RESP:
    case SUBT_REASSOC_RESP:
        p1 = (A_UINT8 *) &(((WLAN_FRAME_ASSOC_RESP *)pmacHdr)->rateSet);
        break;

    case SUBT_REASSOC_REQ:
        p1 = (A_UINT8 *) &(((WLAN_FRAME_REASSOC_REQ *)pmacHdr)->ssid);
        break;

    case SUBT_AUTH:
        if (((WLAN_FRAME_AUTH *)pmacHdr)->transSeqNo == 2) {
            p1 = (A_UINT8 *) &(((WLAN_FRAME_AUTH_CHLG *)pmacHdr)->challenge);
        } else if (((WLAN_FRAME_AUTH *)pmacHdr)->transSeqNo == 3) {
            p1 = (A_UINT8 *) &(((WLAN_FRAME_AUTH_ENCRYPT *)pmacHdr)->challenge);
        } else {
            p1 = NULL;
        }
        break;

    case SUBT_PROBE_REQ:
        p1 = (A_UINT8 *) &(((WLAN_FRAME_PROBE_REQ *)pmacHdr)->ssid);
        break;

    case SUBT_PROBE_RESP:
        p1 = (A_UINT8 *) &(((WLAN_FRAME_PROBE_RESP *)pmacHdr)->buffer);
        break;

    case SUBT_BEACON:
        p1 = (A_UINT8 *) &(((WLAN_FRAME_BEACON *)pmacHdr)->buffer);
        break;

    default:
        p1 = NULL;
    }

    if (p1 == NULL) {
        return NULL;
    }

    /* Now, p1 is the beginning of the variable size fields */
    /* len is the offset into the frame */
    len = (int) (p1 - (A_UINT8 *)pmacHdr);

    /* A valid element should have at least 2 bytes - elementID,
     * length. value byte may not be there if length is 0 (null ssid)
     */
    while (len <= frameSize - 2) {
        elen = ((INFO_ELEMENT *)p1)->length + 2;
        if (((INFO_ELEMENT *)p1)->elementID == elementId && elen > 2) {
            return( (INFO_ELEMENT *) p1 );
        } else {
            len += elen;
            p1 += elen;
        }
    }

    return NULL;
}

/* dk header file compatability */
#ifndef WEP_IV_FIELD_SIZE
#define WEP_IV_FIELD_SIZE WEP_IV_FIELD
#endif

/*
 * macHeaderLenGet - Get MAC header length of a packet
 *
 * MAC header length depends upon the frame type, subtype and in the case
 * of data packets, ToDS and FromDS bits too.
 * PARAMETER
 * Pointer to frame control field of the packet
 * RETURNS
 * MAC header length appropriate for the frame type and subtype.
 */
A_UINT16
macHeaderLenGet( FRAME_CONTROL_FIELD *frameControl)
{
    A_UINT16 len;

    switch (frameControl->fType) {
    case FRAME_MGT:
        len = MAC_HDR_MGT_SIZE;
        break;

    case FRAME_CTRL:
        if (frameControl->fSubtype == SUBT_CTS ||
            frameControl->fSubtype == SUBT_ACK) {
            len = MAC_HDR_CTRL1_SIZE;
        } else {
            len = MAC_HDR_CTRL2_SIZE;
        }
        break;
    case FRAME_DATA:
        if (frameControl->ToDS && frameControl->FromDS) {
            len = MAC_HDR_DATA4_SIZE;
        } else {
            len = MAC_HDR_DATA3_SIZE;
        }
        break;
    default:
        /* Illegal frame type */
        return 0;
    }

    if (frameControl->wep) {
        len += WEP_IV_FIELD_SIZE;
    }

    return len;
}


/* frameEndianChange - Endian conversion of kinds of 802.11 frames.
 *
 * Byte oriented fields do not need to be swapped, beccause they are
 * already in byte order. e.e. MAC addresses, INFO_ELEMENT, RATE_SET, etc.
 *
 * direction - 1 == tx, 0 == rx
 */
void
frameEndianChange(WLAN_FRAME_HEADER *pFrame, A_UINT16 len, int direction)
{
    FRAME_CONTROL_FIELD fCntl;
    A_UINT16            tmp16;
    A_UINT32            i;
    A_UINT32            tmp32;

    fCntl = pFrame->frameControl;
    if (!direction) {
        tmp16 = A_swab16(*((A_UINT16 *)&pFrame->frameControl));
        fCntl = *((FRAME_CONTROL_FIELD *)&tmp16);
    }

    // don't byteswap pFrame->frameControl just yet. We need it for mgtElementGet() calls

    tmp16 = A_swab16(*((A_UINT16 *)&pFrame->durationNav));
    pFrame->durationNav = *((DURATION_NAV *)&tmp16);
    tmp16 = A_swab16(*((A_UINT16 *)&pFrame->sequenceControl));
    pFrame->sequenceControl = *((SEQ_CTRL *)&tmp16);

    if (fCntl.fType == FRAME_DATA) {
        if (fCntl.wep) {
            WLAN_FRAME_3_DATA_ENCRYPT *p3ef;        /* 3 addr encrypted frame */
            WLAN_FRAME_DATA_ENCRYPT *p4ef;          /* 4 addr encrypted frame */

            if (fCntl.ToDS && fCntl.FromDS) {
                p4ef = (WLAN_FRAME_DATA_ENCRYPT *)pFrame;
                p4ef->iv = A_swab32(p4ef->iv);
            } else {
                p3ef = (WLAN_FRAME_3_DATA_ENCRYPT *)pFrame;
                p3ef->iv = A_swab32(p3ef->iv);
            }
        }
        tmp16 = A_swab16(*((A_UINT16 *)&pFrame->frameControl));
        pFrame->frameControl = *((FRAME_CONTROL_FIELD *)&tmp16);
        return;
    }

    if (fCntl.fType == FRAME_MGT) {
        switch(fCntl.fSubtype) {
        case SUBT_ASSOC_REQ:
        case SUBT_REASSOC_REQ:
            {
            UCSE_SET    *pUcseSet;
            tmp16 = A_swab16(*((A_UINT16 *)&((WLAN_FRAME_ASSOC_REQ *)pFrame)->capInfo));
            ((WLAN_FRAME_ASSOC_REQ *)pFrame)->capInfo = *(CAP_INFO *)&tmp16;
            ((WLAN_FRAME_ASSOC_REQ *)pFrame)->listenInterval =
                A_swab16(((WLAN_FRAME_ASSOC_REQ *)pFrame)->listenInterval);
            pUcseSet = (UCSE_SET *)mgtElementGet((WLAN_MGT_MAC_HEADER *)pFrame,
                                                len, ELE_UCSE_SET);
            for (i=0; pUcseSet != NULL && i < (A_UINT32)(pUcseSet->length)/4; i++) {
                tmp32 = A_swab32(pUcseSet->uCiphers[i]);
                pUcseSet->uCiphers[i] = tmp32;
            }
            }
            break;

        case SUBT_ASSOC_RESP:
        case SUBT_REASSOC_RESP:
            {
            UCSE_SET    *pUcseSet;
            MCSE_SET    *pMcseSet;
            tmp16 = A_swab16(*((A_UINT16 *)&((WLAN_FRAME_ASSOC_REQ *)pFrame)->capInfo));
            ((WLAN_FRAME_ASSOC_REQ *)pFrame)->capInfo = *(CAP_INFO *)&tmp16;
            ((WLAN_FRAME_ASSOC_RESP *)pFrame)->statusCode =
                A_swab16(((WLAN_FRAME_ASSOC_RESP *)pFrame)->statusCode);
            ((WLAN_FRAME_ASSOC_RESP *)pFrame)->assocId =
                A_swab16(((WLAN_FRAME_ASSOC_RESP *)pFrame)->assocId);
            pUcseSet = (UCSE_SET *)mgtElementGet((WLAN_MGT_MAC_HEADER *)pFrame,
                                                len, ELE_UCSE_SET);
            for (i=0; pUcseSet != NULL && i < (A_UINT32)(pUcseSet->length)/4; i++) {
                tmp32 = A_swab32(pUcseSet->uCiphers[i]);
                pUcseSet->uCiphers[i] = tmp32;
            }
            pMcseSet = (MCSE_SET *)mgtElementGet((WLAN_MGT_MAC_HEADER *)pFrame,
                                                len, ELE_MCSE_SET);
            for (i=0; pMcseSet != NULL && i < (A_UINT32)(pMcseSet->length)/4; i++) {
                tmp32 = A_swab32(pMcseSet->mCipher[i]);
                pMcseSet->mCipher[i] = tmp32;
            }
            }
            break;

        case SUBT_PROBE_RESP:
        case SUBT_BEACON:
    {
            UCSE_SET    *pUcseSet;
            MCSE_SET    *pMcseSet;

            /* swab 32 bit fields independently, don't need to swap them */
            ((WLAN_FRAME_PROBE_RESP *)pFrame)->timestamp.high =
                A_swab32(((WLAN_FRAME_PROBE_RESP *)pFrame)->timestamp.high);
            ((WLAN_FRAME_PROBE_RESP *)pFrame)->timestamp.low =
                A_swab32(((WLAN_FRAME_PROBE_RESP *)pFrame)->timestamp.low);
            ((WLAN_FRAME_PROBE_RESP *)pFrame)->beaconInterval =
                A_swab16(((WLAN_FRAME_PROBE_RESP *)pFrame)->beaconInterval);
            tmp16 = A_swab16(*((A_UINT16 *)&((WLAN_FRAME_PROBE_RESP *)pFrame)->capInfo));
            ((WLAN_FRAME_PROBE_RESP *)pFrame)->capInfo = *(CAP_INFO *)&tmp16;

            pUcseSet = (UCSE_SET *)mgtElementGet((WLAN_MGT_MAC_HEADER *)pFrame, len, ELE_UCSE_SET);
            for (i=0; pUcseSet != NULL && i < (A_UINT32)(pUcseSet->length)/4; i++) {
                tmp32 = A_swab32(pUcseSet->uCiphers[i]);
                pUcseSet->uCiphers[i] = tmp32;
            }
            pMcseSet = (MCSE_SET *)mgtElementGet((WLAN_MGT_MAC_HEADER *)pFrame, len, ELE_MCSE_SET);
            for (i=0; pMcseSet != NULL && i < (A_UINT32)(pMcseSet->length)/4; i++) {
                tmp32 = A_swab32(pMcseSet->mCipher[i]);
                pMcseSet->mCipher[i] = tmp32;
            }
        }
            break;

        case SUBT_AUTH:
            if (fCntl.wep) {
                ((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->algNo =
                    A_swab16(((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->algNo);
                ((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->transSeqNo =
                    A_swab16(((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->transSeqNo);
                ((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->statusCode =
                    A_swab16(((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->statusCode);
                ((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->iv = 
                    A_swab32(((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->iv);
            } else {
                ((WLAN_FRAME_AUTH *)pFrame)->algNo =
                    A_swab16(((WLAN_FRAME_AUTH *)pFrame)->algNo);
                ((WLAN_FRAME_AUTH *)pFrame)->transSeqNo =
                    A_swab16(((WLAN_FRAME_AUTH *)pFrame)->transSeqNo);
                ((WLAN_FRAME_AUTH *)pFrame)->statusCode =
                    A_swab16(((WLAN_FRAME_AUTH *)pFrame)->statusCode);
            }
            break;

        case SUBT_DEAUTH:
        case SUBT_DISASSOC:
            /* reasonCode has same offset in both frame types */
            ((WLAN_FRAME_DEAUTH *)pFrame)->reasonCode =
                A_swab16(((WLAN_FRAME_DEAUTH *)pFrame)->reasonCode);
            break;
        }
        tmp16 = A_swab16(*((A_UINT16 *)&pFrame->frameControl));
        pFrame->frameControl = *((FRAME_CONTROL_FIELD *)&tmp16);
        return;
    }

    /* Control frames and rest of management subtypes don't have any
     * fields that need to be byte-swapped. Leave them alone.
     */
     if (fCntl.fType == FRAME_CTRL) {
        tmp16 = A_swab16(*((A_UINT16 *)&pFrame->frameControl));
        pFrame->frameControl = *((FRAME_CONTROL_FIELD *)&tmp16);
        return;
     }

}

// Following routines are general purpose utility routines.

/***********************************************************
 * asciiToMacaddr - Converts ascii string to ethernet mac address
 * DESCRIPTION:
 * This routine parses an ASCII string that looks like a standard
 * ethernet mac address like nnnnnnnnnnnn or nn:nn:nn:nn:nn:nn or
 * '.' or '-' instead of ':'. It is flexible about single digit
 * values too. It places resulting 6 bytes address in pmac.
 *
 * RETURNS: FALSE if conversion failed, TRUE if it succeeds
 */
A_BOOL
asciiToMacaddr(A_UCHAR      *pstr,  // Input ascii string, may not be null terminated
               A_UINT16     alen,   // length of input ascii string
               WLAN_MACADDR *pmac)  // pointer to resulting mac address
{
    int     macIndex   = 0;
    int     asciiIndex = 0;
    int     count      = 0;
    A_UINT8 tval       = 0;

    while (asciiIndex < alen && macIndex < 6) {
        if (pstr[asciiIndex] >= '0' && pstr[asciiIndex] <= '9') {
            tval = (tval << 4) | (pstr[asciiIndex] - '0');
            count++;
        } else if (pstr[asciiIndex] >= 'a' && pstr[asciiIndex] <= 'f') {
            tval = (tval << 4) | (pstr[asciiIndex] - 'a' + 10);
            count++;
        } else if (pstr[asciiIndex] >= 'A' && pstr[asciiIndex] <= 'F') {
            tval = (tval << 4) | (pstr[asciiIndex] - 'A' + 10);
            count++;
        } else if (pstr[asciiIndex] == ':' ||
                   pstr[asciiIndex] == '-' ||
                   pstr[asciiIndex] == '.' ||
                   asciiIndex == (alen-1)) {
            count = 2;
        } else {
            /* bad character */
            break;
        }
        if (count == 2) {
            pmac->octets[macIndex++] = tval;
            count = 0;
            tval = 0;
        }
        asciiIndex++;
    }

    if (macIndex != 6) {
        return FALSE;
    }

    return TRUE;
}

/***********************************************************
 * asciiToKey - Converts ascii string to WEP key
 * DESCRIPTION:
 * This routine parses an ASCII string of hex digits
 * and converts them into a byte array of corresponding values.
 * It also updates key length field.
 *
 * RETURNS: FALSE if conversion failed, TRUE if it succeeds
 *
 */

A_BOOL
asciiToKey(A_UCHAR          *pstr,              // Input ascii string, may not be null terminated
           A_UINT16         alen,               // length of input string
           WLAN_PRIV_RECORD *pKey,              // WEP key record pointer
           A_UINT16         keylen,             // length of desired hardware key in bytes
           A_UINT16         encryptedKeyLen)    // length of encrypted key stored in memory
{
    int     keyIndex   = 0;
    int     asciiIndex = 0;
    int     count      = 0;
    A_UINT8 tval       = 0;

    while (asciiIndex < alen && keyIndex < encryptedKeyLen) {
        if (pstr[asciiIndex] >= '0' && pstr[asciiIndex] <= '9') {
            tval = (tval << 4) | (pstr[asciiIndex] - '0');
            count++;
        } else if (pstr[asciiIndex] >= 'a' && pstr[asciiIndex] <= 'f') {
            tval = (tval << 4) | (pstr[asciiIndex] - 'a' + 10);
            count++;
        } else if (pstr[asciiIndex] >= 'A' && pstr[asciiIndex] <= 'F') {
            tval = (tval << 4) | (pstr[asciiIndex] - 'A' + 10);
            count++;
        } else {
            /* bad character */
            pKey->keyLength = 0;        // make it an invalid key
            return FALSE;
        }
        if (count == 2) {
            pKey->keyVal[keyIndex++] = tval;
            count = 0;
            tval = 0;
        }
        asciiIndex++;
    }

    if (keyIndex != encryptedKeyLen) {
        pKey->keyLength = 0;            // make it an invalid key
        return FALSE;
    }

    pKey->keyLength = keylen;
    return TRUE;
}

/*******************************************************************************
*
* lengthGet - calculate the length of the descriptor chain.
*
* This function is called in the de-fragmentation process
*
* RETURNS: length
*/
A_UINT16
lengthGet(ATHEROS_DESC *pDesc)
{
    A_UINT32 len = 0;

    /* traverse list of input descriptors to get length */
    while (pDesc != NULL) {
        len += pDesc->hwStatus.rx.dataLength;
        pDesc = pDesc->pNextVirtPtr;
    }

    return (A_UINT16)len;
}

/*******************************************************************************
*
* headerLengthGet - calculate the header length of the descriptor chain.
*
* the function is called in the de-fragmentation process
*
* assuming the IV, ICV and FCS are striped off already
*
* RETURNS: frame header length
*/
A_UINT16
headerLengthGet(ATHEROS_DESC *pDesc)
{
    WLAN_DATA_MAC_HEADER3 *pWlanHdr;
    A_UINT16               hdrLen = 0;

    pWlanHdr = pDesc->pBufferVirtPtr.header3;

    if ((pWlanHdr->frameControl.FromDS == 1) && (pWlanHdr->frameControl.ToDS == 1)) {
        hdrLen = WLAN_HDR4_SIZE;
    } else {
        hdrLen = WLAN_HDR_SIZE;
    }

    if (pWlanHdr->frameControl.wep == 1) {
        hdrLen += WEP_IV_FIELD_SIZE;
    }

    return hdrLen;
}

void
descDurationSet(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_UINT16 rateIdx,
                A_BOOL swap, A_BOOL shortPreamble)
{
    if (pDesc->hwControl.tx.RTSCTSEnable == 1) {
        halSetTxDescRtsDuration(pdevInfo,
                                pDesc,
                                rateIdx,
                                pdevInfo->staConfig.pRateTable->
                                info[rateIdx].controlRate);
    }

    /* Set the frame control duration field */
    if (pDesc->pBufferVirtPtr.header->frameControl.moreFrag == 0) {

        pDesc->pBufferVirtPtr.header->durationNav.duration =
            halComputeTxTime(pdevInfo->staConfig.pRateTable->ackBits, pdevInfo->staConfig.pRateTable,
                             pdevInfo->staConfig.pRateTable->
                             info[rateIdx].controlRate, shortPreamble) +
            pdevInfo->staConfig.pRateTable->sifs;

    } else {

        /* moreFrag = 1, so NextFrag cannot be NULL */
        ATHEROS_DESC *pNextFrag = pDesc->pNextVirtPtr;

        pDesc->pBufferVirtPtr.header->durationNav.duration =
            (halComputeTxTime(pdevInfo->staConfig.pRateTable->ackBits,
                              pdevInfo->staConfig.pRateTable,
                              pdevInfo->staConfig.pRateTable->
                              info[rateIdx].controlRate, shortPreamble) << 1) +
            halComputeTxTime(((pNextFrag->hwControl.tx.frameLength +
                               FCS_FIELD_SIZE) << 3) + 22,
                             pdevInfo->staConfig.pRateTable,
                             rateIdx, shortPreamble) +
            3*pdevInfo->staConfig.pRateTable->sifs;
    }

    pDesc->pBufferVirtPtr.header->durationNav.clearBit = 0;

    if (swap) {
        /* frame header is in network order, swap it */
        pDesc->pBufferVirtPtr.headerLittle->durationNav.data =
            cpu2le16(pDesc->pBufferVirtPtr.headerLittle->durationNav.data);
    }

    return;
}


