/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 *  Connects reset Reg Vectors, EEPROM Data, and device Functions to pDev
 */

#ident "$Id: //depot/sw/releases/2.0.1/src/hal/halAttach.c#1 $"

#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"

#ifdef CONFIG_ATHEROS_5210_PCI            
#include "ar5210Attach.h"
#else

#ifdef BUILD_AR5211
#include "ar5211Attach.h"
#include "ar5211Reg.h"
#endif

#endif 

typedef struct DeviceAttachData {
    A_UINT16 deviceID;
    A_STATUS (*hwAttach)(WLAN_DEV_INFO *, A_UINT16);
} DEVICE_ATTACH_DATA;

static DEVICE_ATTACH_DATA ar5kAttachData[] = {
//#ifdef BUILD_AR5210
#ifdef CONFIG_ATHEROS_5210_PCI
    {0x1107, ar5210Attach},  /* Default AR5210 DevID */
    {0x0007, ar5210Attach},
    {0x0207, ar5210Attach},
#else // BUILD_AR5211
    {DEV_AR5212, ar5211Attach},
    {DEV_AR5211, ar5211Attach},         /* PreProd AR5211 DevID */
    {0x1112, ar5211Attach},             /* Default AR5211 DevID */
    {DEV_AR5211_FPGA11B, ar5211Attach}, /* 11b emulation FPGA */
#endif
    {0x5AA5, NULL}
};

const static A_UINT32 NumDevids = (sizeof(ar5kAttachData) / sizeof(DEVICE_ATTACH_DATA)) - 1;

/**************************************************************
 * halAttach
 *
 * HalAttach is the first call into the HAL layer.  The HAL
 * Attach detects device chip revisions, initializes the hwLayer
 * function list, reads EEPROM information,
 * selects reset vectors, and performs a short self test.
 * Any failures will return an error that should cause a hardware
 * disable.
 */

A_STATUS
halAttach(WLAN_DEV_INFO *pDev)
{
    A_STATUS      status;
    unsigned int  i;
    A_UINT32      numDeviceIDs;

    ASSERT(pDev);

    /* Find a matching DeviceID attach routine */
    numDeviceIDs = (sizeof(ar5kAttachData)/sizeof(DEVICE_ATTACH_DATA));
    for (i = 0; i < numDeviceIDs; i++) {
        if (pDev->pciInfo.DeviceID == ar5kAttachData[i].deviceID) {
            break;
        }
    }

    if (i == numDeviceIDs) {
        printk("halAttach: Failed to find an attachable driver for devid 0x%04X\n", pDev->pciInfo.DeviceID);
        return A_DEVICE_NOT_FOUND;
    }

    pDev->pHalInfo = kmalloc(sizeof(HAL_INFO), GFP_KERNEL);
    if (pDev->pHalInfo == NULL) {
        printk("halAttach: Error allocating memory for info struct\n");
        return A_ERROR;
    }

    /* Call the device specific attach function */
    status = ar5kAttachData[i].hwAttach(pDev, pDev->pciInfo.DeviceID);

    /* If unsuccessful, free any allocated memory */
    if (status != A_OK) {
        A_DRIVER_FREE(pDev->pHalInfo, sizeof(HAL_INFO), 0);
        pDev->pHalInfo = NULL;
    }
    
    return status;
}

/**************************************************************
 * halDetach
 *
 * Remove HAL layer allocated structures
 */
A_STATUS
halDetach(WLAN_DEV_INFO *pDev)
{
    A_STATUS status;

    ASSERT(pDev);

    if (pDev->pHalInfo == NULL) {
        return A_ERROR;
    }

    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwDetach);
    status = pDev->pHwFunc->hwDetach(pDev);

    /* Free HAL info struct */
    A_DRIVER_FREE(pDev->pHalInfo, sizeof(HAL_INFO), 0);
    pDev->pHalInfo = NULL;

    return status;
}

/**************************************************************
 * halGetNumDevid
 *
 * Returns the number of devids in the table
 */
A_UINT32
halGetNumDevid(void)
{
    return NumDevids;
}

/**************************************************************
 * halGetDevids
 *
 * Return the devid at the given index
 */
A_UINT16
halGetDevids(A_UINT32 index)
{
    ASSERT(index < NumDevids);

    return ar5kAttachData[index].deviceID;
}
