/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * Hardware-abstracted functions used to deal with interrupts.  These functions
 * are primarily called in the driver code residing above the HAL layer.
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/hal/halInterrupts.c#1 $"

#include "wlantype.h"   /* A_UINTxx, etc. */
#include "wlandrv.h"    /* WLAN_DEV_INFO struct */
#include "halApi.h"     /* fn prototypes */
#include "hal.h"        /* HW_FUNCS struct */

void halPrintIntStr (HAL_INT_TYPE ints);

/**************************************************************
 * halIsInteruptPending
 *
 * Checks to see if an interrupt is pending on our NIC
 *
 * Returns: TRUE    if an interrupt is pending
 *          FALSE   if not
 */
A_BOOL
halIsInterruptPending(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwIsInterruptPending);

    return pDev->pHwFunc->hwIsInterruptPending(pDev);
}

/**************************************************************
 * halGetInterrupts
 *
 * Reads the Interrupt Status Register value from the NIC, thus deasserting
 * the interrupt line, and returns both the masked and unmasked mapped ISR
 * values.  The value returned is mapped to abstract the hw-specific bit
 * locations in the Interrupt Status Register.
 *
 * Returns: A hardware-abstracted bitmap of all non-masked-out
 *          interrupts pending, as well as an unmasked value
 */
HAL_INT_TYPE
halGetInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE *pUnmaskedValue)
{
    HAL_INT_TYPE ints;
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwGetInterrupts);

    ints =  pDev->pHwFunc->hwGetInterrupts(pDev, pUnmaskedValue);

    // print all except annoying strictly Beacon ints
    if (ints != HAL_INT_SWBA) {
        aPrintk("\nINTS PENDING: "); halPrintIntStr (ints);
    }
    if (*pUnmaskedValue != HAL_INT_SWBA) {
        aPrintk("INTS UNMASKED: "); halPrintIntStr (*pUnmaskedValue);
    }

    return ints;
}

/**************************************************************
 * halEnableInterrupts
 *
 * Atomically enables NIC interrupts.  Interrupts are passed in
 * via the enumerated bitmask in "ints".
 */
void
halEnableInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE ints)
{
    aPrintk("++++++++++++++++++++++ ENABLED INTERRUPTS:0x%8x..", ints); 
    halPrintIntStr (ints);

    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwEnableInterrupts);
    ASSERT((ints & HAL_INT_INVALID) == 0);

    pDev->pHwFunc->hwEnableInterrupts(pDev, ints);
}

/**************************************************************
 * halDisableInterrupts
 *
 * Atomically disables NIC interrupts.  Interrupts are passed in
 * via the enumerated bitmask in "ints".
 */
void
halDisableInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE ints)
{
    aPrintk("-------------------- DISABLED INTERRUPTS: 0x%8x..", ints);
    halPrintIntStr (ints);

    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwDisableInterrupts);
    ASSERT((ints & HAL_INT_INVALID) == 0);

    pDev->pHwFunc->hwDisableInterrupts(pDev, ints);
}

void halPrintIntStr (HAL_INT_TYPE ints) 
{
    if (ints == HAL_INT_COMMON) {
        aPrintk("[COMMON INTS]\n");
        return;
    }
    if (ints & HAL_INT_RX   )  aPrintk("[RX]");
    if (ints & HAL_INT_RXEOL)  aPrintk("[RXEOL]");
    if (ints & HAL_INT_RXORN)  aPrintk("[RXORN]");
    if (ints & HAL_INT_TX   )  aPrintk("[TX]");
    if (ints & HAL_INT_TXURN)  aPrintk("[TXURN]");
    if (ints & HAL_INT_MIB  )  aPrintk("[MIB]");
    if (ints & HAL_INT_RXPHY)  aPrintk("[RXPHY]");
    if (ints & HAL_INT_RXKCM)  aPrintk("[RXKCM]");
    if (ints & HAL_INT_SWBA )  aPrintk("[SWBA]");
    if (ints & HAL_INT_BMISS)  aPrintk("[BMISS]");
    if (ints & HAL_INT_BNR  )  aPrintk("[BNR]");
    if (ints & HAL_INT_GPIO )  aPrintk("[GPIO]");
    if (ints & HAL_INT_FATAL)  aPrintk("[FATAL]");
    if (ints & HAL_INT_GLOBAL) aPrintk("[GLOBAL]");
    aPrintk("\n");
}
