/*
 *  Copyright (c) 2001 Atheros Communications, Inc., All Rights Reserved
 */

#ident  "ACI $Id: //depot/sw/releases/2.0.1/src/common/ratectrl.c#5 $"

/*
 *  DESCRIPTION
 *  This file contains the data structures and routines for the connection
 *  services.

 1. Introduction

 The rate adaptation algorithm is an optional feature, when enabled, will always attempt to
 transmit unicast packets at the highest possible optimal data rate. If there are obstacles or
 interference, the rate adaptation algorithm will automatically steps down to an optimal rate that
 allows for reliable data transmission. The optimal data rate is selected periodically by the rate
 adaptation algorithm, based on past performance of the data transmission at different adjacent
 data rates.

 For every transmission of a unicast packet, the rate adaptation algorithm provides a
 pre-determined optimal rate to transmit the packet. The result of the transmission is recorded
 for evaluation of a new optimal rate later. Not all unicast packets are sent at the optimal rate.
 Some packets are sent at a different selected rate. The selected rate could be one rate higher or
 lower than the pre-determined optimal rate.

 The rate adaptation algorithm will use the transmission results to determine a new optimal rate
 periodically. The transmission results of the optimal rate, the optimal + 1 rate and the
 optimal - 1 rate are compared to see which rate yields the best outcome. The one with the best
 outcome will be selected as the new optimal rate.

 On the Access Point, the optimal rate is determined independently for each station.


 2. Optimal Data Rate
 The optimal data rate is the highest data rate chosen by the rate adaptation algorithm for the
 reliable transmission of unicast packets. Once chosen, the rate remains unchanged until the re
 adaptation algorithm chooses a new rate when operating condition varies.


 3. Data Rate Bracketing
 To help establish an optimal data rate, a small percentage of the data transmissions are sent at
 two different rates: one is above and one is below the optimal data rate. By bracketing the
 optimal data rate with two adjacent rates and monitoring the performance of these transmissions,
 the rate adaptation algorithm will be able to determine whether to raise or lower the optimal
 data rate.


 4. Sampling Window
 The optimal data rate is chosen at the end of the sampling window by the rate adaptation
 algorithm. During this window period, unicast packets are sent either at optimal rate or at one
 of the bracketed optimal rates and the results of the transmission are saved for evaluation at
 the end of the window period.


 5. RTS/CTS
 RTS/CTS is enabled when there is the number of retries is much higher than the successful
 transmits for the optimal & its neighboring rates. When enabled, the RTS/CTS length is set to
 500 bytes and the duration of the RTS/CTS usage is set to 10 seconds. After 10 seconds, the
 RTS/CTS will be disabled by restoring its original RTS/CTS length.


 6. Data Structures
 -  Rate Control Table. This table contains a list of data rates, thresholds and counters.
 -  Lookup Table. This table is generated during system initialization to speed up the time it
    takes to figure out the transmission time of a given length packet at a given rate.
 -  SIB Table. This table contains the status and various data structures for the station. A
    data structure called TxRateCtrl_s contains the rate control info for the station.
 -  TxRateCtrl_s. This data structure contains counters, the optimal data rate, the bracket data
    rates, and their corresponding transmission results and outcomes.
 -  Packet Length Classes. There are two classes of packet length: short and long. Packet length
    of less than 500 bytes is called short packet and packet length greater or equal to 500 bytes
    is called long packet. Separate optimal data rates are applied to both short and long packets
    for each station.

 7. Primitives

 a) TxRateFind
 This primitive returns a data rate when a calling function wants to transmit a unicast packet.
 It will returns the current operational data rate if rate adaptation is disabled. If enabled, the
 optimal data rate for the target station of a particular class of data length will be returned.
 If one of the bracket data rate usage counters rolls over, then one of the bracket data rate is
 returned to the caller.

 b) RateCtrlTxretryCntUpdate
 This primitive is called by the transmit interrupt service routine to update the TxRateCtrl_s
 data structure with the result of the data transmission. The caller supplies the primitive with
 the station ID, the data rate used in the transmission, the length of the data packet and the
 number of transmit retry. The primitive uses packet length, data rate to looks up one of the
 lookup table to determine the transmission time. The transmission time is then multiplied by the
 number of transmit retries and the result is added to a total transmit time for that data rate.
 Counter is incremented to keep track of number of long or short frame.


 c) Rate Control Scheduler (RateCtrlFunc)
 The rate control scheduler is called at the end of the sampling window to choose a new optimal
 data rate for each station. It is done by going through the TxRateCtrl_s data structure for short
 and long packet of the optimal data rate and the two bracket data rates to calculate the
 throughput for each rate. The throughput is calculated by dividing the total number of packets
 transmitted by the entire sum of all the transmission times during the sampling window.

 The three throughputs are then used to determine the new optimal data rate. If the current
 optimal data rate is the lowest data rate and the bracket high data rate throughput is higher
 than the optimal data rate throughput plus threshold, then raise the optimal data rate to the
 one above it else leave the rate unchanged. If the current optimal data rate is already highest
 data rate and the bracketed lower data rate throughput is higher than the optimal data rate
 throughput plus threshold, then lower the optimal data rate to the one below it else leave the
 rate unchanged.

 If bracketed higher data rate throughput is higher than both the optimal data rate throughput
 plus threshold and the bracketed lower data rate throughput, then raise the optimal data rate to
 the one above it. Otherwise, if the bracketed lower data rate throughput is higher than the
 optimal data rate throughput plus threshold, and then lower the optimal data rate to the one
 below it.

 */

/*
 *      Revision History
 *
 *      1.0 - Initial Release
 *      1.1 - Faster rate climb when packet flow is rapidly increasing.
 *            Stop lowering the rate to 6 for low packet flow.
 *      1.2 - Don't drop to 6Mb when traffic slows down.
 *      1.3 - Added dual operating mode to rate control task on the AP.
 *            Accelerate the optimal rate usage count to allow the faster
 *            use of the exploratory rate to determine the next optimal
 *            rate when the traffic slows down. Backoff the optimal
 *            rate on excessive retries. Reduced the number of txTime table
 *            to only two: one for short (<500 bytes) frame and long (>=500
 *            bytes) frame. Moved optimal rate usage count and exploratory
 *            rate flags to SIB_ENTRY.
 *
 *            Previously, the rate control task runs at a higher task
 *            priority every second to select the optimal rate. This is
 *            now called the foreground mode.
 *            To provide a more up-to-date optimal rate, a new background
 *            mode is added to allow the rate control task to select
 *            the optimal rate more frequently. This is done by running
 *            the rate control task at the lowest task priority and select
 *            the optimal task every 1/10 of a second. Initially the rate
 *            control task runs in the foreground mode before switching
 *            to the background mode where it will remain until the
 *            watchdog timer switches it back to the foreground mode to
 *            restart the cycle again.
 *      1.3.2 - Added corner case rate index conversion to swretryTxRateBackoff
 *            Modified RateCtrlTxRetryCntUpdate to test for the inUse bit
 *            before clearing the txRateCtrl table.
 *            RateCtrlFunc no longer sets the rate.
 *            RateCtrlAdjuster works on a local copy of the txRateCtrl.
 *      1.3.3 - Fixed storing of erranous data rate index in xxCurRateSet. Should
 *            have store 6Mbps instead of 9Mbps.
 *      2.0 - Used HAL API to test for turbo mode.
 *            Managed both minimum and maximum tx rates from SIB on a per station
 *            basis. Minimum and maximum tx rates are extracted from workingRateSet
 *            of the SIB. The workingRateSet is initialized during association of
 *            of the station.
 *            Initial 802.11a/b roaming support for station. Search for either
 *            802.11a or 802.11b if the rate has fallen below the threshold.
 *            Reset the multicast and broadcast data rate to 6Mbps if no station
 *            is found in the BSS or no activity from any station for over 60 seconds.
 *            Reconfigure the validity of turbo data rate whenever the mode is changed.
 */

#ifdef VXWORKS

#include "vxWorks.h"
#include "time.h"
#include "tickLib.h"
#include "wdLib.h"
#include "logLib.h"

#else

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>

#endif

#ifdef BUILD_AP
#include "apdefs.h"
#include "apcfg.h"
#endif
#include "wlantype.h"
#include "wlanproto.h"
#include "wlandrv.h"
#include "ui.h"
#include "wlanext.h"
#include "halApi.h"
#include "wlanchannel.h"
#ifndef BUILD_AP
#include "stacserv.h"
#endif

/*
 * Local declaration(s)
 */
LOCAL A_UINT8 findHiMbRate(A_UINT8 rateIdx);

#ifdef  DEBUG
#define RC_HISTORY
#endif  /* DEBUG */

#define RC_BKGND_MODE           1       /* Background operating mode */
#define RC_FOREGND_MODE         2       /* Foreground operating mode */

#define LOWER_RELRATE           0
#define CURRENT_RELRATE         1
#define HIGHER_RELRATE          2

#define LO_RATE_MAX_CNT         10
#define HI_RATE_MAX_CNT         10

#define RRTHLD1PCT              99      /* rate reduction threshold -- 1% */
#define RRTHLD10PCT             90      /* rate reduction threshold -- 10% */
#define RETHLD100               100     /* rate elevation threshold -- 100% */
#define RETHLD110               110     /* rate elevation threshold -- 110% */


#define MAX_FRAME_LOSS          1000    /* Mulitpler */
#define FRAME_LOSS_THRESHOLD    50

#define MAX_RTS_CTS_AGING       20

#define SHORT_FRAME_LEN         500
#define LONG_FRAME_LEN          1530
#define MIN_FRAME_LEN           100
#define FRAME_BUCKET_LEN        100


#define MAX_THROTTLE_HISTORY    200

#define CLOCK_TICKS_PER_SEC     60

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))

#define LOFLOW_FRAMES           10

#define LONGEST_FRAME_LEN       2400
#define MAX_TBL_RATE            8

#define LOWEST_TX_RATE          pTxRetryHistoBB->minTxRate
#define HIGHEST_TX_RATE         pTxRetryHistoBB->maxTxRate

/*
 *  The following definitions are used to calculate the Tx time of a packet
 */
#define DIFS            (SIFS+2*SLOT)   /* usecs */
#define SIFS            16              /* usecs */
#define SLOT            9               /* usecs */
#define PHYOD           (16+4)
#define DATAMACOD       (24+4+2)
#define ACKMACOD        (14+2+6/8)      /* bytes */
#define ACKTMOUT        MAX(DIFS, 27)


static const int contention_window[] = {
           0,   15,   31,   63,  127,  255,  511, 1023, 1023, 1023,
        1023, 1023, 1023, 1023, 1023, 1023, 1023, 1023, 1023, 1023,
        1023
};



struct RateCtrl_s {
    int      rate;              /* Data rate */
    A_UINT16 bkoffRidx;         /* Software retry backoff data rate index */
    A_UINT16 valid;             /* Valid data rate */
    int      mb;                /* Valid broadcast/multicast rate */
    A_UINT32 dn_threshold;
    A_UINT32 up_threshold;
};

struct txTime_s {
    struct {
        struct {
            /* 16+1 because 0 is used to store no retry count */
            A_UINT32 time[NO_RETRY+MAX_TX_RETRIES];
        } txflen[2];
    } rate[MAX_TBL_RATE];
};

struct  throttle_history_item_s {
    A_UINT8     curRate;                /* Prior selected optimal rate */
    A_UINT8     newRate;                /* New optimal rate selected by rate adaptation */
    A_UINT16    len;                    /* Frame length type: short or long */
    A_UINT32    thruput[MAX_REL_RATE];  /* bytes/sec */
    A_UINT32    frameLoss[MAX_REL_RATE];/* Ratio of retry and no-retry packets */
    A_UINT32    ticks;
    A_UINT32    mode;
    SIB_ENTRY   *pSib;
    struct      TxRetry_s sf;           /* Tx retry bin for shorter length frame */
    struct      TxRetry_s lf;           /* Tx retry bin for longer length frame */
};

struct  throttle_history_s {
    int         max_count;
    int         count;
    int         head;
    int         tail;
    struct throttle_history_item_s item[1];
};

struct TxRetryHisto_s {
    A_UINT8     rateCtrlEnable;
    A_UINT8     dontUpdate;
    A_UINT16    shortFrameLen;
    A_UINT8     rtsCtsEnable;
    A_UINT8     rtsCtsInUse;
    A_UINT32    rtsCtsCount;
    A_UINT32    defaultDot11RTSThreshold;
    A_UINT32    frameLossThreshold;
    A_UINT32    throttle_count;
    A_UINT32    lowPerCount;
    A_UINT32    shortFrameCount;
    A_UINT32    longFrameCount;
    A_UINT32    txShortFrameCount;
    A_UINT32    txLongFrameCount;
    A_UINT32    txShortMbfcnt;
    A_UINT32    txLongMbfcnt;
    A_UINT32    txFrameCount;
    A_UINT32    opmode;         /* Background or foreground mode */
    A_UINT32    work2do;        /* Work to do count */
    A_UINT32    swRetrySfRateCount;
    A_UINT32    swRetryLfRateCount;
#ifdef  VXWORKS
    WDOG_ID     wdgId;
    int         taskId;
#endif
#ifdef  RC_HISTORY
    struct      throttle_history_s *pHistory;
#endif  /* RC_HISTORY */
    int         turboMode;
};

/* Baseband specific fields of TxRetryHistogram */
struct TxRetryHistoBB_s {
    A_UINT16    bbType;
    A_UINT8     minTableTxRate;
    A_UINT8     maxTableTxRate;
    A_UINT8     minTxRate;
    A_UINT8     maxTxRate;
    A_UINT8     sMbTxRateIdx;
    A_UINT8     lMbTxRateIdx;
};

int csDebug = 0;
int csDebug2 = 0;
int csrDebug = 0;
int rcDebug = 0;
int csVerbose = 0;
int autoRate = 0;
int csInitialized = 0;
int sfTxRateCntmax = 5;
int lfTxRateCntmax = 5;
int iterations = 20;
int curRate = 0;
int curTxRate = 0;
int RateCtrlTaskDelay = CLOCK_TICKS_PER_SEC;    /* CLOCKS_PER_SEC; */
struct TxRetryHisto_s TxRetryHistogram;
struct TxRetryHistoBB_s TxRetryHistoCCK;
struct TxRetryHistoBB_s TxRetryHistoOFDM;
struct TxRetryHistoBB_s *pTxRetryHistoBB;

struct txTime_s *pTxTimeTbl = (struct txTime_s *)NULL;
char **pDataRate;

static char *dataRateTbl[] =
{ "6", "6", "12", "18", "24", "36", "48", "54", NULL };

static char *dataRate2xTbl[] =
{ "12", "12", "24", "36", "48", "72", "96", "108", NULL };

static char *dataRateCCKTbl[] =
{ "1", "2", "5.5", "11", NULL };

/*
 *  Rate Contrl Table
 *
 *  The table index is the transmit rate.
 *  Since it is hopeless to use 9 Mbps and we don't want to leave a hole
 *  in the table. The table entry for 9Mbps is swapped with 6Mbps so that
 *  it can used effortlessly by the rate control alogirthm.
 *  The valid bit for the 9 Mbps is disabled, so that the table entry is
 *  not selected. Other changes are made to get and save current transmit
 *  rate to and from the database. If the current transmit rate in the
 *  database is 0 (6 Mbps), then convert it to 1, so that it can index to
 *  the correct table entry. If the new transmit rate is 1, then convert
 *  it to 0 before storing it to the database.
 *
 */
struct RateCtrl_s RateCtrlTblOFDM[] =
{
    { 9, 1, 0, 0, RRTHLD10PCT, RETHLD100 },         /* 9 Mbps */
    { 6, 1, 1, 1, RRTHLD10PCT, RETHLD110 },         /* 6 Mbps */
    {12, 1, 1, 1, RRTHLD10PCT, RETHLD110 },         /* 12 Mbps */
    {18, 1, 1, 0, RRTHLD10PCT, RETHLD110 },         /* 18 Mbps */
    {24, 2, 1, 1, RRTHLD10PCT, RETHLD110 },         /* 24 Mbps */
    {36, 2, 1, 0, RRTHLD10PCT, RETHLD110 },         /* 36 Mbps */
    {48, 4, 1, 0, RRTHLD10PCT, RETHLD110 },         /* 48 Mbps */
    {54, 4, 1, 0, RRTHLD10PCT, RETHLD110 }          /* 54 Mbps */
};

struct RateCtrl_s RateCtrlTblCCK[] =
{
    { 1, 0, 1, 1, RRTHLD10PCT, RETHLD110 },         /* 1 Mbps */
    { 2, 0, 1, 1, RRTHLD10PCT, RETHLD110 },         /* 2 Mbps */
    { 5, 1, 1, 1, RRTHLD10PCT, RETHLD110 },         /* 5.5 Mbps */
    {11, 2, 1, 1, RRTHLD10PCT, RETHLD110 },         /* 11 Mbps */
    {11, 2, 0, 0, RRTHLD10PCT, RETHLD110 },         /* unused */
    {11, 2, 0, 0, RRTHLD10PCT, RETHLD110 },         /* unused */
    {11, 2, 0, 0, RRTHLD10PCT, RETHLD110 },         /* unused */
    {11, 2, 0, 0, RRTHLD10PCT, RETHLD110 }          /* unused */
};

struct RateCtrl_s *RateCtrlTbl = &RateCtrlTblOFDM[0];

/*
 * New rate adaptation code
 */
#define MIN(x,y)        ((x) <= (y) ? (x) : (y))

/*
 * 5GHz 11a OFDM minimum valid rssi values.
 */
static A_UINT8 RssiAckValidMin11a[] = {
     1,                 /* 6 Mbps */
     2,                 /* 9 Mbps */
     3,                 /* 12 Mbps */
     5,                 /* 18 Mbps */
     8,                 /* 24 Mbps */
    11,                 /* 36 Mbps */
    16,                 /* 48 Mbps */
    20                  /* 54 Mbps */
};

/*
 * Minimum differences in dB between rates in the 11a table
 */
A_UINT8 RssiAckDeltaMin11a[] = {
     1,                 /* 6 Mbps */
     1,                 /* 9 Mbps */
     2,                 /* 12 Mbps */
     2,                 /* 18 Mbps */
     3,                 /* 24 Mbps */
     3,                 /* 36 Mbps */
     3,                 /* 48 Mbps */
     3                  /* 54 Mbps */
};

/*
 * 5GHz 11a OFDM Turbo minimum valid rssi values.
 */
static A_UINT8 RssiAckValidMin11aTurbo[] = {
     1,                 /* 12 Mbps */
     5,                 /* 18 Mbps */
     7,                 /* 24 Mbps */
     8,                 /* 36 Mbps */
    11,                 /* 48 Mbps */
    14,                 /* 72 Mbps */
    19,                 /* 96 Mbps */
    23                  /* 108 Mbps */
};

/*
 * Minimum differences in dB between rates in the 11a table
 */
A_UINT8 RssiAckDeltaMin11aTurbo[] = {
     1,                 /* 12 Mbps */
     1,                 /* 18 Mbps */
     2,                 /* 24 Mbps */
     2,                 /* 36 Mbps */
     3,                 /* 48 Mbps */
     3,                 /* 72 Mbps */
     3,                 /* 96 Mbps */
     3                  /* 108 Mbps */
};

/*
 * 2.4GHz CCK minimum valid rssi values.
 */
static A_UINT8 RssiAckValidMinCCK[] = {
     0,                 /* 1 Mbps */
     1,                 /* 2 Mbps */
     2,                 /* 5.5 Mbps */
     4,                 /* 11 Mbps */
    200,                    /* invalid */
    200,                    /* invalid */
    200,                    /* invalid */
    200                 /* invalid */
};

/*
 * Minimum differences in dB between rates in the 11a table
 */
A_UINT8 RssiAckDeltaMinCCK[] = {
     1,                 /* 1 Mbps */
     1,                 /* 2 Mbps */
     2,                 /* 5.5 Mbps */
     100,                   /* 11 Mbps */
     3,                 /* invalid */
     3,                 /* invalid */
     3,                 /* invalid */
     3                  /* invalid */
};


/*
 * These pointers are dynamically switched to the correct tables
 * above based on the mode (11a, CCK, turbo).  We default them
 * to the 11a values.
 */
static A_UINT8 *RssiAckValidMin = RssiAckValidMin11a;
static A_UINT8 *RssiAckDeltaMin = RssiAckDeltaMin11a;

/*
 * Whether to use to new or existing rate adaptation algorithm.
 * 0 => original, 1 => new.
 */
int RateAlgSelect = 1;

static int txRateFindCB(SIB_ENTRY *pSib, int frameLen);
static void swretryTxRateBackoffCB(ATHEROS_DESC *pErrorDesc);
static void RateCtrlTxRetryCntUpdateCB
(
    int txRate, int Xretries, int retries, int frameLen, SIB_ENTRY *pSib,
    A_UINT8 rssiAck, A_UINT8 antTx, A_UINT8 *flipReq
);

LOCAL void
genTxTime(int fLen, int rate, A_UINT32 *pSum)
{
    int         i, retries;
    A_UINT32    cwtime, pkttime, acktime, time, acctime[NO_RETRY+MAX_TX_RETRIES+1];
    A_UINT32    sum;

    for (retries = 0; retries <= MAX_TX_RETRIES; retries++) {
        sum = 0;
        for (i = 1; i <= MAX_TX_RETRIES+1; i++) {
            cwtime  = ((contention_window[i] * SLOT)+1)/2;
            pkttime = ((DATAMACOD + fLen) * 8)/rate;
            pkttime = ((pkttime + 3)/4) * 4;
            pkttime = PHYOD + pkttime;
            acktime = (ACKMACOD * 8)/rate;
            acktime = ((acktime + 3)/4) * 4;
            acktime = PHYOD + acktime + SIFS;
            time = ((i == 1) ? DIFS : ACKTMOUT) + cwtime + pkttime + acktime;
            if (i == 1) {
                acctime[1] = time;
            } else {
                acctime[i] = time + acctime[i-1];
            }
            sum += (retries == (i - 1)) ? acctime[i] : 0;
#ifdef  DEBUG
            if (csDebug > 1) {
              uiPrintf("cw: %4d, pkt: %d, ack: %d, time: %d, acc: %d, sum: %d\n",
                        cwtime, pkttime, acktime, time, acctime[i], sum);
            }
#endif  /* DEBUG */
          }
#ifdef  DEBUG
        if (csDebug) {
            uiPrintf("retry %2d, rate: %d, sum: %d\n", retries, rate, sum);
        }
#endif  /* DEBUG */
        if (pSum) {
          *pSum++ = sum;
        }
    }
}

/*
 *  Construct a table for Tx time of different frame size, Tx rates and Tx retries.
 *  The table is then used by the rate control algorithm to compute the
 *  Tx thruput
 */
int txTimeTableInit(void)
{
    A_UINT32 *pSum;
    int      rIdx;

    pTxTimeTbl = (struct txTime_s *)A_DRIVER_MALLOC(sizeof(struct txTime_s), 0, 0);
    if (!pTxTimeTbl) {
        uiPrintf("can't alloc memory for TxTimeTbl!!!\n");
        return A_ERROR;
    }
    A_MEM_ZERO((char *)pTxTimeTbl, sizeof(struct txTime_s));

    for (rIdx = 0; rIdx < MAX_TBL_RATE; rIdx++) {
        if (RateCtrlTbl[rIdx].valid) {
            pSum = (A_UINT32 *)&pTxTimeTbl->rate[rIdx].txflen[0].time;
            genTxTime(SHORT_FRAME_LEN, RateCtrlTbl[rIdx].rate, pSum);
            pSum = (A_UINT32 *)&pTxTimeTbl->rate[rIdx].txflen[1].time;
            genTxTime(LONG_FRAME_LEN, RateCtrlTbl[rIdx].rate, pSum);
        }
    }
    return A_OK;
}

#if     0
void ratechange(int newrate)
{
    WLAN_DEV_INFO *pdevInfo;   /* device information structure */
    int     unit = 0;
    int     oldrate;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    oldrate = pdevInfo->localSta->currentOperationalRateIndex;
    pdevInfo->localSta->currentOperationalRateIndex = (A_UINT8) newrate;
    uiPrintf("old rate: %d, new rate: %d\n", oldrate, newrate);
}


/*
 *  This function returns the Tx rate index from the database
 *  If the current Tx rate index is 1 (9 Mbps), then convert it to 6 Mbps
 */
int txRateGet(void)
{
    WLAN_DEV_INFO * pdevInfo;   /* device information structure */
    int     unit = 0;
    int     rate;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    rate     = pdevInfo->localSta->currentOperationalRateIndex;
    if (rate == 0) {
        rate = LOWEST_TX_RATE;
    }
    return rate;
}
#endif

/*
 *  This function returns the current Tx rate index of a short frame
 *  If the current Tx rate index is 1 (9 Mbps), then convert it to 6 Mbps
 */
A_UINT16 sfCurRateGet(SIB_ENTRY *pSib)
{
    A_UINT16 rateIdx;

    rateIdx = pSib->txRateCtrl.sf.rateIdx;
    if (rateIdx < LOWEST_TX_RATE) {
        rateIdx = LOWEST_TX_RATE;
    }
    if (rateIdx > pSib->txRateCtrl.maxTxRateIdx) {
        rateIdx = pSib->txRateCtrl.maxTxRateIdx;
    }
    return rateIdx;
}


/*
 *  This function modifies the current Tx rate of a short frame
 *  If the current Tx rate is 1 (6 Mbps), then convert it to 0
 */
void sfCurRateSet(SIB_ENTRY *pSib, int newTxRate)
{
    WLAN_DEV_INFO *pdevInfo;   /* device information structure */
    int     unit = 0;

    if ((newTxRate < 0) || (newTxRate > pSib->txRateCtrl.maxTxRateIdx)) {
      return;
    }
    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    pSib->txRateCtrl.sf.rateIdx = (A_UINT8) newTxRate;
    if (pTxRetryHistoBB->bbType & CHANNEL_OFDM) {
        /* Swap 9 and 6 in OFDM table */
        if (newTxRate == LOWEST_TX_RATE) {
            pdevInfo->localSta->currentOperationalRateIndex = 0;
        } else {
            pdevInfo->localSta->currentOperationalRateIndex = (A_UINT8) newTxRate;
        }
    } else {
        /* Map 0 thru 3 into 0 to 3 for operationalrateindex */
        pdevInfo->localSta->currentOperationalRateIndex = (A_UINT8) newTxRate;
    }
    pSib->txRateCtrl.sf.rateCnt[newTxRate]++;
}


/*
 *  This function returns the current Tx rate table index of a long frame
 *  If the current Tx rate table index is 1 (9 Mbps), then convert it to 6 Mbps
 */
A_UINT16 lfCurRateGet(SIB_ENTRY *pSib)
{
    A_UINT16 rateIdx;

    rateIdx = pSib->txRateCtrl.lf.rateIdx;
    if (rateIdx < LOWEST_TX_RATE) {
        rateIdx = LOWEST_TX_RATE;
    }
    if (rateIdx > pSib->txRateCtrl.maxTxRateIdx) {
        rateIdx = pSib->txRateCtrl.maxTxRateIdx;
    }
    return rateIdx;
}


/*
 *  This function modifies the current Tx rate table index of a long frame
 *  If the current Tx rate table index is 1 (6 Mbps), then convert it to 0
 */
void lfCurRateSet(SIB_ENTRY *pSib, int newTxRate)
{
    WLAN_DEV_INFO *pdevInfo;    /* device information structure */
    int     unit = 0;

    if ((newTxRate < 0) || (newTxRate > pSib->txRateCtrl.maxTxRateIdx)) {
      return;
    }
    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    pSib->txRateCtrl.lf.rateIdx = (A_UINT8) newTxRate;
    if (pTxRetryHistoBB->bbType & CHANNEL_OFDM) {
        /* Swap 9 and 6 in OFDM table */
        if (newTxRate == LOWEST_TX_RATE) {
            pdevInfo->localSta->currentOperationalRateIndex = 0;
        } else {
            pdevInfo->localSta->currentOperationalRateIndex = (A_UINT8) newTxRate;
        }
    } else {
        /* Map 0 thru 3 into 0 to 3 for operationalrateindex */
        pdevInfo->localSta->currentOperationalRateIndex = (A_UINT8) newTxRate;
    }
     pSib->txRateCtrl.lf.rateCnt[newTxRate]++;
}

/*
 *  This function finds the next lower Tx rate
 */
int txLowerRateFind(int curRate)
{
    int rate;

    rate = curRate;
    if (curRate) {
        for (rate = curRate - 1; rate >= LOWEST_TX_RATE; rate--) {
            if (RateCtrlTbl[rate].valid) {
                break;
            }
          }
        if (rate <= 0) {
          rate = LOWEST_TX_RATE;
        }
    } else {
      rate = LOWEST_TX_RATE;
    }

    return rate;
}

/*
 *  This function finds the next higher Tx rate
 */
static int txHigherRateFind(SIB_ENTRY *pSib, int curRate)
{
    int rate;

    for (rate = curRate + 1; rate <= pSib->txRateCtrl.maxTxRateIdx; rate++) {
        if (RateCtrlTbl[rate].valid) {
            break;
        }
    }
    if (rate > pSib->txRateCtrl.maxTxRateIdx) {
        rate = curRate;
    }

    return rate;
}


/*
 *  This function determines and returns the new Tx rate index
 *  The rate control algorithm will occasionally (1 out of 10)
 *  send packets at higher and lower rates to gather statistics
 *  on thruput at those rates.
 */
int txRateFind(SIB_ENTRY *pSib, int frameLen)
{
    WLAN_DEV_INFO *pdevInfo;   /* device information structure */
    int     curRate, rate;
    struct  RateCtrl_s *pRc;
    struct  TxRetry_s *pX;
    struct bin_s *pRtbl;
    int     unit = 0;

    if ( RateAlgSelect ) {
        return txRateFindCB(pSib, frameLen);
    }

    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    if (!TxRetryHistogram.rateCtrlEnable) {
      return pdevInfo->localSta->currentOperationalRateIndex;
    }

    pSib->txRateCtrl.work2do++;
    TxRetryHistogram.txFrameCount++;
    TxRetryHistogram.work2do++;

    /*
     *  Get current rate for frame length
     */
    if (frameLen < TxRetryHistogram.shortFrameLen) {
        curRate = sfCurRateGet(pSib);
        pX = &pSib->txRateCtrl.sf;
        pRtbl = &pSib->txRateCtrl.sf.tbl[CURRENT_RELRATE];
        TxRetryHistogram.txShortFrameCount++;
    } else {
        curRate = lfCurRateGet(pSib);
        pX = &pSib->txRateCtrl.lf;
        pRtbl = &pSib->txRateCtrl.lf.tbl[CURRENT_RELRATE];
        TxRetryHistogram.txLongFrameCount++;
    }

    rate = curRate;
    pRc = (struct RateCtrl_s *)&RateCtrlTbl[curRate];

    pX->txCnt++;
    /*
     *  Use the exploratory rate after sending out a number
     *  of frames in optimal rate.
     */
    if (++pX->optRateCnt > pX->optRateMax) {
        pX->optRateCnt = 0;
        pX->xLoRate = 0;
        pX->xHiRate = 1;
    }

    if (pX->xHiRate) {
        pX->xLoRate = 1;        /* Use the lower rate in the next transmission */
        pX->xHiRate = 0;
        pX->optRateCnt = 0;
        /*
         *  If not the highest rate, then check if PER is acceptable
         */
        if (rate != pSib->txRateCtrl.maxTxRateIdx) {
            rate = txHigherRateFind(pSib, curRate);
        } else {
            pX->xLoRate = 0;
            rate = txLowerRateFind(curRate);
        }
    } else {
        if (pX->xLoRate) {
            pX->xLoRate = 0;
            pX->xHiRate = 0;
            pX->optRateCnt = 0;
            if (rate != LOWEST_TX_RATE) {
                rate = txLowerRateFind(curRate);
            } else {
                rate = txHigherRateFind(pSib, curRate);
            }
        }
    }

#ifdef  DEBUG
    if (csDebug == 8) {
        uiPrintf("txRateFind -- rate %s => %s, len: %d, cnt: %d, Lo: %d, Hi: %d\n",
                 pDataRate[curRate], pDataRate[rate], frameLen,
                  pX->optRateCnt, pX->xLoRate, pX->xHiRate);
    }
#endif  /* DEBUG */

    if (rate == pTxRetryHistoBB->minTableTxRate) {
        rate = 0;           /* Convert it back to real rate index */
    }

    return rate;
}




/*
 *  This function determines and returns the new Tx rate index
 *  for broadcast and multicast traffic
 */
int txMbRateFind(int frameLen)
{
    int     curRate;
    WLAN_DEV_INFO * pdevInfo;   /* device information structure */
    int     unit = 0;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    if (!TxRetryHistogram.rateCtrlEnable) {
        return findHiMbRate(pdevInfo->localSta->currentOperationalRateIndex);
    }

    TxRetryHistogram.txFrameCount++;
/***    TxRetryHistogram.work2do++;     ***/

    if (frameLen < TxRetryHistogram.shortFrameLen) {
        curRate = pTxRetryHistoBB->sMbTxRateIdx;
        TxRetryHistogram.txShortMbfcnt++;
    } else {
        curRate = pTxRetryHistoBB->lMbTxRateIdx;
        TxRetryHistogram.txLongMbfcnt++;
    }

#ifdef  DEBUG
    if (csDebug == 88) {
        uiPrintf("txMbRateFind -- rate %d\n", curRate);
    }
#endif  /* DEBUG */
    if (curRate == pTxRetryHistoBB->minTableTxRate) {
        curRate = 0;
    }
    return curRate;
}


/*
 *  This routine is called by the Tx interrupt service routine to update the
 *  Tx retry count table
 */
void RateCtrlTxRetryCntUpdate(int txRate, int Xretries, int retries,
                              int frameLen, SIB_ENTRY *pSib,
                  A_UINT8 rssiAck, A_UINT8 antTx,
                  A_UINT8 *flipReq)
{
    WLAN_DEV_INFO *pdevInfo;   /* device information structure */
    int       curRate;
    int       relRate;
    A_UINT32  txTime;
    struct bin_s *pRtbl;
    int       unit = 0;

    if ( RateAlgSelect ) {
        RateCtrlTxRetryCntUpdateCB(txRate, Xretries, retries,
                              frameLen, pSib, rssiAck, antTx, flipReq);
        return;
    }

    if (pTxRetryHistoBB->bbType & CHANNEL_OFDM) {
        if (txRate == 0) {  /* Translate actual rate to rate used by rate table */
            txRate = 1;
        }
    }

    pdevInfo = globDrvInfo.pDevInfoArray[unit];

    if (!TxRetryHistogram.rateCtrlEnable) {
        return;
    }

    if (retries < 0 || retries >= MAX_TX_RETRIES) {
        return;
    }

    if (!pSib) {
        return;
    }

    /*
     *    Because the rate control task only looks at three rates: lower, current
     *    and higher rates. We need to calculate the relative rate in order
     *    to update the retry histogram. The relative rate, RelRate, has three values:
     *        0 = lower,
     *        1 = current,
     *        2 = higher
     *    If the current rate changes recently, then the relative rate
     *    might be out of bound.
     */
    if (frameLen < TxRetryHistogram.shortFrameLen) {
        curRate = sfCurRateGet(pSib);
    } else {
        curRate = lfCurRateGet(pSib);
    }

    relRate = txRate - curRate + 1;

    /*
     *    Do bound checking
     */
    if (relRate < 0) {
        relRate = 0;
    }
    if (relRate >= MAX_REL_RATE) {
        relRate = 2;
    }

    if (frameLen < TxRetryHistogram.shortFrameLen) {
        pRtbl = &pSib->txRateCtrl.sf.tbl[relRate];
        txTime = pTxTimeTbl->rate[txRate].txflen[0].time[retries];
        if (retries) {
            TxRetryHistogram.shortFrameCount++;
        }
    } else {
        pRtbl = &pSib->txRateCtrl.lf.tbl[relRate];
        txTime = pTxTimeTbl->rate[txRate].txflen[1].time[retries];
        if (retries) {
            TxRetryHistogram.longFrameCount++;
        }
    }

    /*
     *  Compute the transmission time
     */
    pRtbl->total_time += txTime;
    pRtbl->retryBin[retries]++;

#ifdef  DEBUG
    if (csDebug == 8 && csDebug2) {
        uiPrintf("retry cnt [%d][%d]: %d, tx: %s, cur: %s, rel: %d, len: %d, time: %d, %d\n",
                 relRate, retries,  pRtbl->retryBin[retries],
                 pDataRate[txRate],
                 pDataRate[curRate], relRate, frameLen, txTime,
                 pRtbl->total_time);
    }
#endif  /* DEBUG */

    if (retries) {
        /*
         *    Slam on the brakes if this is an excessive retries
         */
        if (Xretries && txRate <= curRate) {
            int newRate;

            newRate = RateCtrlTbl[curRate].bkoffRidx;
#ifdef  DEBUG
            if (csrDebug) {
                uiPrintf("%d sw retries -- txrate %s, curRate: %s => %s, len: %d\n",
                         retries, pDataRate[txRate], pDataRate[curRate],
                         pDataRate[newRate], frameLen);
            }
#endif

            if (frameLen < TxRetryHistogram.shortFrameLen) {
                TxRetryHistogram.swRetrySfRateCount++;
                pSib->txRateCtrl.sf.optRateCnt = 0;
                pSib->txRateCtrl.sf.xLoRate = 0;
                pSib->txRateCtrl.sf.xHiRate = 0;
                /*
                 *    Don't update the histogram if it is used by the rate control task
                 */
                if (!pSib->txRateCtrl.inUse) {
                    A_MEM_ZERO((char *)&pSib->txRateCtrl.sf.tbl,
                               sizeof(pSib->txRateCtrl.sf.tbl));
                }
                sfCurRateSet(pSib, newRate);
            } else {
                TxRetryHistogram.swRetryLfRateCount++;
                pSib->txRateCtrl.lf.optRateCnt = 0;
                pSib->txRateCtrl.lf.xLoRate = 0;
                pSib->txRateCtrl.lf.xHiRate = 0;
                /*
                 *    Don't update the histogram if it is used by the rate control task
                 */
                if (!pSib->txRateCtrl.inUse) {
                    A_MEM_ZERO((char *)&pSib->txRateCtrl.lf.tbl,
                               sizeof(pSib->txRateCtrl.lf.tbl));
                }
                lfCurRateSet(pSib, newRate);
            }
        }
    }

#ifdef  DEBUG
    if (csDebug == 80) {
        uiPrintf("retry cnt [%d][%d]: %d, tx: %d, cur: %d, fLen: %d, time: %d\n",
                 relRate, retries, pRtbl->retryBin[retries],
                 txRate, curRate, frameLen, pRtbl->total_time);
    }
#endif  /* DEBUG */
}


/*
 *  This routine is called to initialize the rate control parameters
 *  in the SIB. It is called initially during system initialization
 *  or when a station is associated with the AP.
 */
void RateCtrlSibInit(SIB_ENTRY *pSib)
{
    WLAN_DEV_INFO * pdevInfo;   /* device information structure */
    int     unit = 0;
    struct  TxRateCtrlCB_s *pRc = &pSib->txRateCtrl.cb;
    int     rate;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    A_MEM_ZERO((char *)&pSib->txRateCtrl, sizeof(pSib->txRateCtrl));
    pSib->txRateCtrl.sf.rateIdx = pSib->txRateCtrl.lf.rateIdx =
        (HIGHEST_TX_RATE - LOWEST_TX_RATE)/2 + 1;
    pSib->txRateCtrl.minTxRateIdx  = LOWEST_TX_RATE;
    pSib->txRateCtrl.maxTxRateIdx  = HIGHEST_TX_RATE;
        RateCtrlSibTxRateInit(pSib);
    pSib->txRateCtrl.sf.optRateMax = LO_RATE_MAX_CNT;
    pSib->txRateCtrl.sf.optRateMax = HI_RATE_MAX_CNT;
    pSib->txRateCtrl.lf.optRateMax = LO_RATE_MAX_CNT;
    pSib->txRateCtrl.lf.optRateMax = HI_RATE_MAX_CNT;

    for ( rate = 0 ; rate < MAX_TBL_RATE ; rate++ ) {
        pRc->state[rate].rssiThres = RssiAckValidMin[rate] + 3;
        pRc->state[rate].cost = 0;
    }
    pRc->rssiLast     = 0;
    pRc->rssiLastPrev = 0;
    pRc->rssiLastLkup = 0;
    pRc->rateMax      = MIN(4, pSib->txRateCtrl.maxTxRateIdx);
    pRc->rssiTime     = 0;
    pRc->rssiDownTime = A_MS_TICKGET();
}

/*
 * This routine is called when the driver changes channel and it results in changing
 * the baseband from CCK to OFDM or vice versa. Certain tables need to be swapped.
 */

void RateCtrlBBChange(WLAN_DEV_INFO *pdevInfo, A_UINT16 channelFlags)
{

    if (channelFlags & CHANNEL_CCK) {
        RateCtrlTbl = &RateCtrlTblCCK[0];
        pTxRetryHistoBB = &TxRetryHistoCCK;
        pDataRate = &dataRateCCKTbl[0];
    if ( RateAlgSelect ) {
        RssiAckValidMin = RssiAckValidMinCCK;
        RssiAckDeltaMin = RssiAckDeltaMinCCK;
    }
    } else {
        RateCtrlTbl = &RateCtrlTblOFDM[0];
        pTxRetryHistoBB = &TxRetryHistoOFDM;
        if (channelFlags & CHANNEL_TURBO) {
            pDataRate = &dataRate2xTbl[0];
        if ( RateAlgSelect ) {
        RssiAckValidMin = RssiAckValidMin11aTurbo;
        RssiAckDeltaMin = RssiAckDeltaMin11aTurbo;
        }
        } else {
            pDataRate = &dataRateTbl[0];
        if ( RateAlgSelect ) {
        RssiAckValidMin = RssiAckValidMin11a;
        RssiAckDeltaMin = RssiAckDeltaMin11a;
        }
        }
    }
    // recompute the local SIB parameters
    RateCtrlSibInit (pdevInfo->localSta);
}

/*
 *  This routine is called to set up the minimum and maximum tx rates
 *  by extracting the info from workingRateSet of the SIB
 */
// TODO: Modify to use valid rates, short/long, etc.
void RateCtrlSibTxRateInit(SIB_ENTRY *pSib)
{
    int ii;
    int lo = 99999, hi = 0;
    int rate, rateIdx;

    if (!pSib->workingRateSet.length) {
        return;
    }

    for (ii = 0; ii < pSib->workingRateSet.length; ii++) {
        int ik;

        rateIdx = 0;
        rate = ((pSib->workingRateSet.rates[ii] & ~0x80) * 5)/10;
        for (ik = 0; ik < MAX_TBL_RATE; ik++) {
            if (RateCtrlTbl[ik].rate == rate) {
                rateIdx = ik;
                break;
            }
        }
        lo = min(lo, rateIdx);
        hi = max(hi, rateIdx);
    }
    pSib->txRateCtrl.minTxRateIdx  = (A_UINT8)lo;
    pSib->txRateCtrl.maxTxRateIdx  = (A_UINT8)hi;
}


void RateCtrlDataRateInit (WLAN_DEV_INFO *pdevInfo)
{
    A_UINT8     rate;
    A_UINT16    netBand;

#ifdef BUILD_AP
    netBand = (A_UINT16) apCfgFreqSpecGet();
#else
    netBand = pdevInfo->staConfig.NetBand;
#endif

    /* Perform OFDM initialization */
    TxRetryHistoOFDM.bbType = CHANNEL_OFDM;
    TxRetryHistoOFDM.minTableTxRate = 1;
    TxRetryHistoOFDM.maxTableTxRate = 7;
    for (rate = 1; rate <= TxRetryHistoOFDM.maxTableTxRate; rate++) {
        RateCtrlTblOFDM[rate].valid = 1;
    }

    /*
     * If turbo mode is enabled, then put a lower ceiling on max rate
     */
    pDataRate = &dataRateTbl[0];
    TxRetryHistogram.turboMode = (netBand & MODE_SELECT_TURBO) || 0;
    if (TxRetryHistogram.turboMode) {
        pDataRate = &dataRate2xTbl[0];
        for (rate = 0; rate <= TxRetryHistoOFDM.maxTableTxRate; rate++) {
            if (RateCtrlTblOFDM[rate].rate > (int)halGetMaxTurboRate(pdevInfo)) {
                RateCtrlTblOFDM[rate].valid = 0;
            }
        }
    }

    /* Compute OFDM rates */
    for (rate = 0; rate <= TxRetryHistoOFDM.maxTableTxRate; rate++) {
        if (RateCtrlTblOFDM[rate].valid) {
            TxRetryHistoOFDM.minTxRate = rate;
            TxRetryHistoOFDM.sMbTxRateIdx =
                TxRetryHistoOFDM.lMbTxRateIdx = rate;
            break;
        }
    }

    for (rate = TxRetryHistoOFDM.minTableTxRate;
         rate <= TxRetryHistoOFDM.maxTableTxRate; rate++) {
        if (RateCtrlTblOFDM[rate].valid) {
            TxRetryHistoOFDM.maxTxRate = rate;
        }
    }

    /* Compute CCK rates */
    TxRetryHistoCCK.bbType = CHANNEL_CCK;
    TxRetryHistoCCK.minTableTxRate = 0;
    TxRetryHistoCCK.maxTableTxRate = 3;
    for (rate = 0; rate <= TxRetryHistoOFDM.maxTableTxRate; rate++) {
        RateCtrlTblOFDM[rate].valid = 1;
    }

    for (rate = 0; rate <= TxRetryHistoCCK.maxTableTxRate; rate++) {
        if (RateCtrlTblCCK[rate].valid) {
            TxRetryHistoCCK.minTxRate = rate;
            TxRetryHistoCCK.sMbTxRateIdx =
                TxRetryHistoCCK.lMbTxRateIdx = rate;
            break;
        }
    }

    for (rate = TxRetryHistoCCK.minTableTxRate;
         rate <= TxRetryHistoCCK.maxTableTxRate; rate++) {
        if (RateCtrlTblCCK[rate].valid) {
            TxRetryHistoCCK.maxTxRate = rate;
        }
    }

    /*
     * Now set a default wireless mode so that table pointers are
     * initialized to a valid table
     */
    RateCtrlBBChange(pdevInfo, pdevInfo->staConfig.pChannel->channelFlags);
}

void RateCtrlInit(void)
{
    SIB_ENTRY   *pSib;
    WLAN_DEV_INFO * pdevInfo;   /* device information structure */
    int         unit = 0;

#ifdef  RC_HISTORY
    struct      throttle_history_s *pHistory;
#endif  /* RC_HISTORY */

#ifdef VXWORKS
    RateCtrlTaskDelay = sysClkRateGet();
#endif
    pdevInfo = globDrvInfo.pDevInfoArray[unit];
#ifdef  RC_HISTORY
    pHistory = TxRetryHistogram.pHistory;
    A_MEM_ZERO((char *)&TxRetryHistogram, sizeof(TxRetryHistogram));

    /*
     *  In case the NDIS driver gets unloaded and re-loaded again,
     *  don't allocate memory.
     */
    if (!csInitialized) {
        csInitialized = 1;
        pHistory = (struct throttle_history_s *)A_DRIVER_MALLOC
          (sizeof(struct throttle_history_s)+sizeof(struct throttle_history_item_s)*MAX_THROTTLE_HISTORY, 0, 0);
        if (!pHistory) {
            uiPrintf("RateCtrlInit error -- out of memory\n");
            return;
        }
    }

    TxRetryHistogram.pHistory = pHistory;
#endif  /* RC_HISTORY */
    TxRetryHistogram.dontUpdate = 1;

#ifdef  BUILD_AP
    /*
     *  By definition, AP is heard by all stations in its BSS
     *  and will never be a hidden node. Don't enable RTS/CTS
     *  unless is forced to do so.
     */
    TxRetryHistogram.rtsCtsEnable = 0;
#else
    TxRetryHistogram.rtsCtsEnable = 1;
#endif
    TxRetryHistogram.rtsCtsInUse = 0;
    TxRetryHistogram.rtsCtsCount = 0;
    TxRetryHistogram.rateCtrlEnable = 0;
    TxRetryHistogram.shortFrameLen = SHORT_FRAME_LEN;
    TxRetryHistogram.frameLossThreshold = FRAME_LOSS_THRESHOLD;
    TxRetryHistogram.defaultDot11RTSThreshold = 0;
#ifdef  RC_HISTORY
    if (TxRetryHistogram.pHistory) {
        TxRetryHistogram.pHistory->max_count = MAX_THROTTLE_HISTORY;
        TxRetryHistogram.pHistory->count = 0;
        TxRetryHistogram.pHistory->head = 0;
        TxRetryHistogram.pHistory->tail = 0;
    }
#endif  /* RC_HISTORY */

    TxRetryHistogram.opmode = RC_FOREGND_MODE;
#ifdef  VXWORKS
    TxRetryHistogram.wdgId = wdCreate();
    if (!TxRetryHistogram.wdgId) {
        uiPrintf("RateCtrlInit error -- can't create watchdog timer\n");
        return;
    }
    TxRetryHistogram.taskId = taskIdSelf();
#endif  /* VXWORKS */

    RateCtrlDataRateInit (pdevInfo);
    pSib = NULL;

    while ((pSib = sibNextStaFind(pSib)) != NULL) {
        RateCtrlSibInit(pSib);
    }

    if (txTimeTableInit() == A_ERROR) {
        uiPrintf("RateCtrlInit error -- can't create txTime table!\n");
        return;
    }

}


void RateCtrlRtsCtsEnable(int v)
{
    TxRetryHistogram.rtsCtsEnable = (A_UINT8) v;
}

int RateCtrlIsRtsCtsEnabled(void)
{
    return TxRetryHistogram.rtsCtsEnable;
}

/*
 *  The Rate Control Handler periodically examine the retry
 *  statistics and recompute the transmission rates.
 */
void RateCtrlFunc(int fLen, struct TxRetry_s *pBin, int *pRateSet,
                  SIB_ENTRY *pSib)
{
    int         retry_found;
    int         rate;
    int         curRate;
    int         newRate;
    int         relRate;
    int         retries;
    int         numPkts = 0;
    A_UINT32    total_pkts;
    A_UINT32    total_time;
    WLAN_DEV_INFO *pdevInfo;    /* device information structure */
    int         unit = 0;
    A_UINT32    hthput, othput, lthput;
    A_UINT32    hfloss, ofloss, lfloss;
#ifdef  RC_HISTORY
    struct throttle_history_item_s *pItem;
    int         idx;
    int         maxcount;
    int         update_history = 0;
#endif  /* RC_HISTORY */

    struct thruput_s {
      struct bin_s {
        A_UINT32    retry_pkts;
        A_UINT32    thruput;
      } bin[MAX_REL_RATE];
    } stats;
    struct bin_s *pRbin;

    curRate = *pRateSet;
    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    newRate = curRate;

    if ( RateAlgSelect ) {
        return;
    }

    /*
     *  Calculate the thruputs for all relative rates
     */

    retry_found = 0;
    pRbin = (struct bin_s *)stats.bin;
    A_MEM_ZERO((char *)&stats, sizeof(stats));
    for (relRate = 0; relRate < MAX_REL_RATE; relRate++, pRbin++) {
        A_UINT32  good_pkts;

        /*
         *    Calculate the actual rate from relative rate
         */
        if (curRate) {
            rate = curRate + relRate - 1;
        } else {
            rate = curRate;
        }
        total_time = pBin->tbl[relRate].total_time;

        retries = 1;
        good_pkts = total_pkts = pBin->tbl[relRate].retryBin[0];
#if     0
        if (RateCtrlTbl[curRate].up_threshold == RETHLD100) {
            /*
             * Consider no retry and 1 retry as good packets
             */
            good_pkts = total_pkts += pBin->tbl[relRate].retryBin[retries++];
        }
#endif
        for (; retries <= MAX_TX_RETRIES; retries++) {
            if (pBin->tbl[relRate].retryBin[retries]) {
                total_pkts += pBin->tbl[relRate].retryBin[retries];
            }
        }

        numPkts += total_pkts;
        if (total_pkts && total_time) {
            A_UINT32  retry_pkts = total_pkts - good_pkts;

            if (total_pkts && retry_pkts) {
                pBin->tbl[relRate].frameLoss = (retry_pkts*100)/total_pkts;
            }
            pRbin->thruput = (total_pkts * 1000000)/total_time;
            if (pRbin->thruput == 0) {        /* watch out for overflow */
                pRbin->thruput = (A_UINT32)-1;
            }
            retry_found++;
        }
    }

    if (retry_found == 0) {
        A_MEM_ZERO((char *)&pBin->tbl, sizeof(pBin->tbl));
        return;         /* Nothing to do */
    }

    hthput = stats.bin[HIGHER_RELRATE].thruput;
    othput = stats.bin[CURRENT_RELRATE].thruput;
    lthput = stats.bin[LOWER_RELRATE].thruput;
    lfloss = pBin->tbl[LOWER_RELRATE].frameLoss;
    ofloss = pBin->tbl[CURRENT_RELRATE].frameLoss;
    hfloss = pBin->tbl[HIGHER_RELRATE].frameLoss;

    /*
     *  Use RTS/CTS if frame loss did not improved after a rate change
     *  (An indication of the hidden node problem)
     */
    if (TxRetryHistogram.rtsCtsEnable) {
        if (!TxRetryHistogram.rtsCtsInUse) {
            int enableRts = 0;

            /*
             * Make sure that we've suffer subtantial frame loss for all relative rates
             */
            enableRts = ofloss > TxRetryHistogram.frameLossThreshold ? 1 : 0;
            if (lthput) {
              if (lfloss > TxRetryHistogram.frameLossThreshold) {
                enableRts++;
              }
            }
            if (hthput) {
                if (hfloss > TxRetryHistogram.frameLossThreshold ) {
                    enableRts++;
                }
            }
            if (fLen && enableRts >= 3) {
#if 0
                if (curRate == pSib->txRateCtrl.rtsCts.newRate &&
                    pSib->txRateCtrl.rtsCts.frameLoss[LOWER_RELRATE] > 10 &&
                    pSib->txRateCtrl.rtsCts.frameLoss[CURRENT_RELRATE] > 10 &&
                    pSib->txRateCtrl.rtsCts.frameLoss[HIGHER_RELRATE] > 10)
#endif
                  {
#ifdef  DEBUG
                    if (csDebug == 3) {
                      uiPrintf ("SIB %d enable RTS, rate: %d, thruput: %d, %d, %d, loss: %d%%, %d%%, %d%%\n",
                                pSib->assocId & 0x3fff, curRate, lthput, othput, hthput,
                                lfloss, ofloss, hfloss);
                    }
#endif  /* DEBUG */
                    TxRetryHistogram.defaultDot11RTSThreshold =
                      pdevInfo->staConfig.defaultDot11RTSThreshold;
                    pdevInfo->staConfig.defaultDot11RTSThreshold =
                      pSib->txRateCtrl.rtsCts.enable = 500;
                    pSib->txRateCtrl.rtsCts.aging = MAX_RTS_CTS_AGING;
                    pSib->txRateCtrl.rtsCts.count++;
                    TxRetryHistogram.rtsCtsInUse = 1;
                    TxRetryHistogram.rtsCtsCount++;
                  }
                pSib->txRateCtrl.rtsCts.frameLoss[LOWER_RELRATE] = lfloss;
                pSib->txRateCtrl.rtsCts.frameLoss[CURRENT_RELRATE] = ofloss;
                pSib->txRateCtrl.rtsCts.frameLoss[HIGHER_RELRATE] = hfloss;
                pSib->txRateCtrl.rtsCts.newRate = (A_UINT8) newRate;
              }
        } else {
            /*
             *  After enabling RTS/CTS for some time, then disable it
             */
            if (pSib->txRateCtrl.rtsCts.aging) {
                pSib->txRateCtrl.rtsCts.aging--;
            } else {
#ifdef  DEBUG
                if (csDebug == 3) {
                  uiPrintf("SIB %d disable RTS, thruput: %d, %d, %d\n",
                            pSib->assocId & 0x3fff, lfloss, ofloss, hfloss);
                }
#endif  /* DEBUG */
                pdevInfo->staConfig.defaultDot11RTSThreshold =
                  (A_UINT16) TxRetryHistogram.defaultDot11RTSThreshold;
                TxRetryHistogram.rtsCtsInUse = 0;
                pSib->txRateCtrl.rtsCts.frameLoss[LOWER_RELRATE] = 0;
                pSib->txRateCtrl.rtsCts.frameLoss[CURRENT_RELRATE] = 0;
                pSib->txRateCtrl.rtsCts.frameLoss[HIGHER_RELRATE] = 0;
                pSib->txRateCtrl.rtsCts.newRate = 0;
            }
        }
    }

    /*
     *  Make sure we've collected at least one sample for each rate before
     *  invoking the rate control algorithm.
     */
    if ( (retry_found != MAX_REL_RATE &&
          curRate > LOWEST_TX_RATE && curRate < pSib->txRateCtrl.maxTxRateIdx) ||
         ((curRate <= LOWEST_TX_RATE && (hthput == 0 || othput == 0)) ||
          (curRate >= pSib->txRateCtrl.maxTxRateIdx && (lthput == 0 || othput == 0))) )
      {
#if     0
          /*
           *    Bump up the count to speed up the use of exploratory rates
           *    if the traffic is too slow
           */
          if (TxRetryHistogram.opmode == RC_FOREGND_MODE) {
              if (numPkts < pBin->optRateMax) {
                  if ((pBin->optRateCnt + 2) < pBin->optRateMax)
                      pBin->optRateCnt += pBin->optRateMax - pBin->optRateCnt - 2;
              }
          }
#endif
          return;
      }

    /*
     *  See if we should modify the current rate
     *  If the higher rate thruput is greater than the thruput of the lower
     *  rate and greater or equal to the current rate, then raise the rate.
     *  Otherwise, if the lower rate thruput is greater than the current
     *  rate thruput, then it is time to lower the rate.
     */
    if (curRate == pSib->txRateCtrl.maxTxRateIdx) {
        if (lthput > othput) {
            if ((othput*100)/lthput < RateCtrlTbl[curRate].dn_threshold) {
                if (ofloss) {
                    if ((curRate - 1) >= LOWEST_TX_RATE) {
                        newRate = txLowerRateFind (curRate);
                    }
                }
            }
        }
    } else {
        if (curRate == LOWEST_TX_RATE) {
            if ((hthput > othput) &&
                ((hthput*100/othput) > RateCtrlTbl[curRate].up_threshold)) {
                newRate = txHigherRateFind(pSib, curRate);
                /*
                 *  Down at the bottom. Might as well push for a higher rate
                 */
            }
        } else {
            if ((hthput >= lthput) && (hthput > othput) &&
                (((hthput*100)/othput > RateCtrlTbl[curRate].up_threshold) ||
                 (hthput > othput && ofloss == 0 && hfloss == 0)))
              {
                if ((curRate + 1) <= pSib->txRateCtrl.maxTxRateIdx) {
                    if (hfloss < 10) {
                        newRate = txHigherRateFind(pSib, curRate);
                    }
                }
              } else {
                  if (lthput > othput) {
                      if ((othput*100)/lthput < RateCtrlTbl[curRate].dn_threshold) {
                          if (ofloss) {
                              if ((curRate - 1) >= LOWEST_TX_RATE) {
                                  newRate = txLowerRateFind(curRate);
                              }
                          }
                      }
                  }
              }
          }
      }

#ifndef BUILD_AP
    // more aggressive scheme. Whether we drop the rate or not, try to find options
    // check pChannelpointer for safety.
    if (pSib->rateAdaptationOff == 0 &&
        pdevInfo->staConfig.pChannel != NULL &&
        newRate <= curRate && newRate <=
            ((pdevInfo->staConfig.pChannel->channelFlags & CHANNEL_CCK) ?
            pdevInfo->pCsInfo->roamRateB : pdevInfo->pCsInfo->roamRateA)) {
        cservTxrateNotify(pdevInfo, newRate,
                          pSib->stats.CurrentAckSignalStrenth);
    }
#endif

    pBin->frameCnt = numPkts;

    if (newRate > curRate) {
        ++pBin->throttle_upcnt;
#ifdef  RC_HISTORY
        ++update_history;
#endif  /* RC_HISTORY */
    }
    if (newRate < curRate) {
        ++pBin->throttle_dncnt;
#ifdef  RC_HISTORY
        ++update_history;
#endif  /* RC_HISTORY */
    }

    ++pBin->throttle_count;

    /*
     *  Set new optimal rate
     */
    *pRateSet = newRate;

#ifdef  DEBUG
    if (csDebug == 8 || csDebug == 1) {
        int j;

#ifdef  VXWORKS
        uiPrintf("%s %s %ld %ld.%ld\n",
                 TxRetryHistogram.opmode == RC_BKGND_MODE ? "B" : "F",
                 fLen ? "L" : "S", tickGet(), tickGet()/CLOCK_TICKS_PER_SEC,
                 tickGet() % CLOCK_TICKS_PER_SEC);
#endif  /* VXWORKS */
        uiPrintf("packets: %d, thruputs: %d, %d, %d, rate: %s => %s\n",
                 numPkts, lthput, othput, hthput,
                 pDataRate[curRate], pDataRate[newRate]);
        uiPrintf("High/Opt: %d, UpThld: %d, Opt/Low: %d, DnThld: %d\n",
                 (hthput*100)/othput, RateCtrlTbl[curRate].up_threshold,
                 (othput*100)/lthput, RateCtrlTbl[curRate].dn_threshold);
        for (j = 0; j < 3; j++) {
            int r, pkts;
            pkts = 0;
            for (r = 0; r <= MAX_TX_RETRIES; r++) {
                if (pBin->tbl[j].retryBin[r]) {
                    pkts += pBin->tbl[j].retryBin[r];
                }
            }
            if (pkts) {
                uiPrintf("time: %5d, pkts: %4d, retries: %2d, fLoss: %3d\n",
                         pBin->tbl[j].total_time, pkts, pkts - pBin->tbl[j].retryBin[0],
                         pBin->tbl[j].frameLoss);
            }
        }
    }
#endif  /* DEBUG */

#ifdef  RC_HISTORY
    if ((!TxRetryHistogram.dontUpdate && update_history) || csVerbose) {
        idx = TxRetryHistogram.pHistory->tail;
        pItem = (struct throttle_history_item_s *)&TxRetryHistogram.pHistory->item[idx];

#ifdef  VXWORKS
        pItem->ticks   = tickGet();
#endif
        pItem->pSib    = pSib;
        pItem->len     = (A_UINT16) fLen;
        pItem->curRate = (A_UINT8) curRate;
        pItem->newRate = (A_UINT8) newRate;
        pItem->frameLoss[LOWER_RELRATE]    = lfloss;
        pItem->frameLoss[CURRENT_RELRATE]  = ofloss;
        pItem->frameLoss[HIGHER_RELRATE]   = hfloss;
        pItem->thruput[LOWER_RELRATE]   = lthput;
        pItem->thruput[CURRENT_RELRATE] = othput;
        pItem->thruput[HIGHER_RELRATE]  = hthput;
        pItem->mode = TxRetryHistogram.opmode;

        /*
         *  If tail runs into head, then, move the head pointer
         *  before moving the tail pointer.
         */
        maxcount = TxRetryHistogram.pHistory->max_count;
        if (TxRetryHistogram.pHistory->head == ((idx + 1) % maxcount)) {
            TxRetryHistogram.pHistory->head = (TxRetryHistogram.pHistory->head + 1) % maxcount;
        }
        TxRetryHistogram.pHistory->tail = (idx + 1) % maxcount;
        TxRetryHistogram.pHistory->count++;
#if 0
        if (csDebug == 1) {
            uiPrintf("head: %d, tail: %d\n",
                     TxRetryHistogram.pHistory->head,
                     TxRetryHistogram.pHistory->tail);
        }
#endif
    }
#endif  /* RC_HISTORY */

    A_MEM_ZERO((char *)&pBin->tbl, sizeof(pBin->tbl));
}


A_UINT8 validMbRate(A_UINT8 rateIdx)
{
    A_UINT8     i;

    for (i = rateIdx; i >= pTxRetryHistoBB->minTableTxRate; i--) {
        if (RateCtrlTbl[i].mb) {
            break;
        }
    }
    return i;
}


A_UINT8 findHiMbRate(A_UINT8 rateIdx)
{
    A_UINT8     i;

    for (i = rateIdx; i >= pTxRetryHistoBB->minTableTxRate; i--) {
        if (RateCtrlTbl[i].mb) {
            return i == pTxRetryHistoBB->minTableTxRate ? 0 : i;
        }
    }

    /*
     * Convert from table index to actual rate index
     */
    return 0;
}


/*
 *  Rate Control Adjuster
 *
 *  Rate Control Algorithm:
 *  Use the statistics of the number of retries per packet over a small subset
 *  of rates to determine the expected "net thruput" at each rate due to
 *  prevailing packet error rates.
 */

void RateCtrlAdjuster(void)
{
    int         sfRate;
    int         lfRate;
    A_UINT32    sfRetryCnt, lfRetryCnt;
    A_UINT8     smbRateIdx, lmbRateIdx;
    SIB_ENTRY   *pSib;
    A_UINT32    curtime = A_MS_TICKGET();
    struct TxRetry_s *pBin;
    struct TxRateCtrl_s tbl;

    if (!TxRetryHistogram.rateCtrlEnable) {
        return;
    }

    if (RateAlgSelect) {
        return;
    }
    if (!TxRetryHistogram.work2do) {
        /*
         *  Reset the broadcast/multicast rate if there is no station
         */
        if (sibNextStaFind((SIB_ENTRY *)NULL) == NULL) {
            pTxRetryHistoBB->sMbTxRateIdx =
                pTxRetryHistoBB->lMbTxRateIdx = validMbRate (pTxRetryHistoBB->minTableTxRate);
        }
        return;
    }

    TxRetryHistogram.work2do = 0;

#ifdef BUILD_AP
    RateCtrlRtsCtsEnable(apCfgRtsCtsEnable());
#endif  /* BUILD_AP */

    smbRateIdx = lmbRateIdx = pTxRetryHistoBB->maxTableTxRate;
    pSib = NULL;

    while ((pSib = sibNextStaFind(pSib)) != NULL) {
        if (!pSib->txRateCtrl.work2do) {
            continue;
        }

        sfRetryCnt = TxRetryHistogram.swRetrySfRateCount;
        lfRetryCnt = TxRetryHistogram.swRetryLfRateCount;

        pSib->txRateCtrl.work2do = 0;
        pSib->txRateCtrl.inUse   = 1;   /* Lock critical region */
        A_DRIVER_BCOPY((char *)&pSib->txRateCtrl, (char *)&tbl, sizeof(tbl));
        sfRate = sfCurRateGet(pSib);
        lfRate = lfCurRateGet(pSib);
        pSib->txRateCtrl.inUse = 0;     /* Unlock critical region */

        /*
         *    Calculate the thruputs for all relative rates
         */

        pBin = (struct TxRetry_s *)&tbl.sf;
        RateCtrlFunc(0, pBin, &sfRate, pSib);
        smbRateIdx = min(smbRateIdx, sfRate);

        pBin = (struct TxRetry_s *)&tbl.lf;
        RateCtrlFunc(1, pBin, &lfRate, pSib);
        lmbRateIdx = min (lmbRateIdx, lfRate);

        pSib->txRateCtrl.inUse = 1;       /* Lock critical region */
        A_DRIVER_BCOPY((char *)&tbl, (char *)&pSib->txRateCtrl, sizeof(pSib->txRateCtrl));
        if (sfRetryCnt == TxRetryHistogram.swRetrySfRateCount) {
            sfCurRateSet(pSib, sfRate);
        }
        if (lfRetryCnt == TxRetryHistogram.swRetryLfRateCount) {
            lfCurRateSet(pSib, lfRate);
        }

        /*
         *  Reset the broadcast/multicast rate if there is no activity
         *  from anyone for over 1 minute
         */
        if ((curtime - pSib->staLastActivityTime) >= (60 * 1 * 1000))
            smbRateIdx = lmbRateIdx = pTxRetryHistoBB->minTableTxRate;

        pSib->txRateCtrl.inUse = 0;       /* Unlock critical region */
    }

    /*
     *  Determine the broadcast & multicast tx rates for short &
     *  long packets
     */
    pTxRetryHistoBB->sMbTxRateIdx = validMbRate(smbRateIdx);
    pTxRetryHistoBB->lMbTxRateIdx = validMbRate(lmbRateIdx);
}

void RateCtrlSet(int state)
{
    TxRetryHistogram.rateCtrlEnable = state ? 1 : 0;
}


void RateCtrlCfg(int cfg)
{
    WLAN_DEV_INFO *pdevInfo;    /* device information structure */
    int     unit = 0;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    pdevInfo->staConfig.rateCtrlEnable = cfg ? 1 : 0;
}

/*
 *  Used to back off rate of current frame immediately...like
 *  when it suffered and excessive retry
 */
// TODO: Modify to use only valid rates
void
swretryTxRateBackoff(ATHEROS_DESC *pErrorDesc)
{
    int                 curRateIdx;
    A_UINT16            bkoffRateIdx;
    WLAN_DEV_INFO       *pdevInfo = globDrvInfo.pDevInfoArray[0];
    A_BOOL              shortPreamble;

    if ( RateAlgSelect ) {
        swretryTxRateBackoffCB(pErrorDesc);
        return;
    }

    curRateIdx = pdevInfo->staConfig.pRateTable->plcpRateToIndex[pErrorDesc->hwControl.tx.transmitRate];
    bkoffRateIdx = RateCtrlTbl[curRateIdx].bkoffRidx;
    shortPreamble = (pErrorDesc->hwControl.tx.transmitRate & CCK_SHORT_PREAMBLE_ENABLE) || 0;

    /*
     * If this is the lowest rate, then convert it into the actual tx rate
     * index because the lowest rate is the RateCtrlTable index and not
     * an actual tx rate. If we don't convert it, then we'll end up using
     * the useless 9Mbps instead of 6Mbps.
     */
    if (bkoffRateIdx == LOWEST_TX_RATE) {
        bkoffRateIdx = 0;
    }
    pErrorDesc->hwControl.tx.transmitRate =
        pdevInfo->staConfig.pRateTable->info[bkoffRateIdx].plcpRate;

    /* swretry has the frame still in network order */
    descDurationSet(pdevInfo, pErrorDesc, bkoffRateIdx, TRUE, shortPreamble);
    return;
}

// TODO: Modify to merge rate sets
void rateCtrlTblUpdate(SIB_ENTRY *pSib, INFO_ELEMENT *pOperationalRateSet)
{
}


/*
 *  Used to normalize frame to what the connection's default rate is.
 *  That rate will be dropped back when excessive retries happen.  This
 *  routine writes that (new) rate into frames.  Usually, these will
 *  be frames which were filtered.
 */
void
swretryDefaultTxRateSet(ATHEROS_DESC *pDesc)
{
    WLAN_DEV_INFO   *pdevInfo = globDrvInfo.pDevInfoArray[0];
    A_UINT16 rateIdx;
    static void swretryDefaultTxRateSetCB(ATHEROS_DESC *pDesc);

    if ( RateAlgSelect ) {
        swretryDefaultTxRateSetCB(pDesc);
        return;
    }
    if (pDesc->hwControl.tx.frameLength < TxRetryHistogram.shortFrameLen) {
        rateIdx = sfCurRateGet(pDesc->pDestSibEntry);
    } else {
        rateIdx = lfCurRateGet(pDesc->pDestSibEntry);
    }
    pDesc->hwControl.tx.transmitRate =
        pdevInfo->staConfig.pRateTable->info[rateIdx].plcpRate;

    /* swretry has the frame still in network order */
    descDurationSet(pdevInfo, pDesc, rateIdx, TRUE, (A_BOOL) pDesc->pDestSibEntry->capInfo.shortPreamble);
    return;
}

A_BOOL isRateCtrlEnabled(void)
{
    return TxRetryHistogram.rateCtrlEnable ? TRUE : FALSE;
}

#ifdef  VXWORKS

void RateCtrlTaskModeSet(int mode);


void rcMBrateGet(int *pShort, int *pLong)
{
    *pShort = RateCtrlTbl[pTxRetryHistoBB->sMbTxRateIdx].rate;
    *pLong  = RateCtrlTbl[pTxRetryHistoBB->lMbTxRateIdx].rate;
}

int RateCtrlTaskModeGet(void)
{
    return TxRetryHistogram.opmode;
}

FUNCPTR RateCtrlForeGndTimeout(int flag)
{
    if (rcDebug) {
        logMsg ("timeout -- mode: %d, tid: %d\n",
                RateCtrlTaskModeGet(),
                TxRetryHistogram.taskId, 0, 0, 0, 0);
    }

    RateCtrlTaskModeSet(RC_FOREGND_MODE);
    return OK;
}


void RateCtrlTaskModeSet(int mode)
{
    TxRetryHistogram.opmode = mode;

    taskPrioritySet(TxRetryHistogram.taskId, mode == RC_BKGND_MODE ? RC_TASK_BG_MODE_PRIORITY : RC_TASK_FG_MODE_PRIORITY);
    if (rcDebug) {
        logMsg("set mode %d -- start watchdog timer\n", mode, 0, 0, 0, 0, 0);
    }
    wdCancel(TxRetryHistogram.wdgId);
    wdStart(TxRetryHistogram.wdgId, RateCtrlTaskDelay, (FUNCPTR)RateCtrlForeGndTimeout, 0);
}


/*
 *  Rate Control Task
 *
 *  Rate Control Algorithm:
 *  Use the statistics of the number of retries per packet over a small subset
 *  of rates to determine the expected "net thruput" at each rate due to
 *  prevailing packet error rates.
 *  The Rate Control Task runs in either foreground or background mode.
 *  Every one second interval, the rate control task switches to foreground mode
 *  and runs at a higher task priority. Afterwards, the rate control task reverts
 *  to background mode and runs every a tenth of a second at the lowest task priority.
 *
 */

void RateCtrlTask(void)
{
    WLAN_DEV_INFO * pdevInfo;   /* device information structure */
    int         unit = 0;
    int         delay, mode;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    RateCtrlInit();
    RateCtrlTaskModeSet(RC_FOREGND_MODE);

    for (;;) {
        /*
         *  Save the current operating mode now
         */
        mode = RateCtrlTaskModeGet();

        if (pdevInfo->staConfig.rateCtrlEnable) {
            RateCtrlSet(1);
        } else {
              RateCtrlSet(0);
        }
        delay = RateCtrlTaskDelay;
        if (TxRetryHistogram.rateCtrlEnable) {
            RateCtrlAdjuster();
            /*
             *  Check the last but not the current mode
             *  because the watchdog timer can change
             *  the mode. We need to recognize the mode
             *  change in order to fire off the watchdog
             *  timer
             */
            if (mode == RC_FOREGND_MODE) {
                RateCtrlTaskModeSet(RC_BKGND_MODE);
            }
            delay = RateCtrlTaskDelay/10;
            delay = delay ? delay : 1;
        }
        taskDelay(delay);
    }
}

#else /* VXWORKS */

/*
 *  Rate Control Task (Station Side)
 *
 *  Rate Control Algorithm:
 *  Use the statistics of the number of retries per packet over a small
 *  subset of rates to determine the expected "net thruput" at each rate
 *  due to prevailing packet error rates.
 */

void RateCtrlTask(void)
{
    WLAN_DEV_INFO *pDev;
    int           unit = 0;

    pDev = globDrvInfo.pDevInfoArray[unit];

    if (pDev->staConfig.rateCtrlEnable) {
        RateCtrlSet(1);
    } else {
        RateCtrlSet(0);
    }

    if (TxRetryHistogram.rateCtrlEnable) {
        if (RateAlgSelect) {
            /*
             * In the new rate control algorithm, we notify connection services
             * of a rate drop differently from the old algorithm.
             */
            SIB_ENTRY *pApSib;
            int       rate, rssi;

            if (pDev->bssDescr == NULL ||
                (pApSib = sibEntryFind(&pDev->bssDescr->bssid)) == NULL)
            {
                return;
            }

            rate = txRateFindCB(pApSib, LONG_FRAME_LEN);
            rssi = pApSib->stats.CurrentAckSignalStrenth;

            /*
             * The following checks are performed to be consistent with
             * previous code and may not actually be necessary.
             */
            if (pDev->staConfig.pChannel != NULL &&
                rate <= (IS_CHAN_CCK(pDev->staConfig.pChannel->channelFlags) ?
                         pDev->pCsInfo->roamRateB : pDev->pCsInfo->roamRateA))
            {
                cservTxrateNotify(pDev, rate, rssi);
            }
        } else {
            RateCtrlAdjuster();
        }
    }
}
#endif /* VXWORKS */

#ifdef  DEBUG
/*
 *  Display the Tx time table
 */
void tableShow(int len)
{
    int fidx, i, rIdx;


    for (fidx = 0; fidx < 2; fidx++) {
        for (i = 0; i <= MAX_TX_RETRIES; i++) {
            if (i == 0) {
                uiPrintf("    ");
                for (rIdx = 0; rIdx < MAX_TBL_RATE; rIdx++) {
                    if (RateCtrlTbl[rIdx].valid) {
                        uiPrintf("%s ", pDataRate[rIdx]);
                    }
                }
                uiPrintf("\n");
            }
            uiPrintf("%2d) ", i);
            for (rIdx = 0; rIdx < MAX_TBL_RATE; rIdx++) {
                if (RateCtrlTbl[rIdx].valid) {
                    uiPrintf("%6d ", pTxTimeTbl->rate[rIdx].txflen[fidx].time[i]);
                }
            }
            uiPrintf("\n");
        }
    }
}

void rcShow(void)
{
    SIB_ENTRY   *pSib;
    WLAN_DEV_INFO * pdevInfo;   /* device information structure */
    int         unit = 0;
    int         rate;


    if (!TxRetryHistogram.rateCtrlEnable) {
        return;
    }

    pdevInfo = globDrvInfo.pDevInfoArray[unit];

    uiPrintf("Lowest Tx Rate[%d]: %s Mbps, Highest Tx Rate[%d]: %s Mbps\n",
             pTxRetryHistoBB->minTxRate,
             pDataRate[pTxRetryHistoBB->minTxRate],
             pTxRetryHistoBB->maxTxRate,
             pDataRate[pTxRetryHistoBB->maxTxRate]);
    uiPrintf("Broadcast Tx Rate %s Mbps (short), %s Mbps (long). %d short frames, %d long frames\n",
             pDataRate[pTxRetryHistoBB->sMbTxRateIdx],
             pDataRate[pTxRetryHistoBB->lMbTxRateIdx],
             TxRetryHistogram.txShortMbfcnt,
             TxRetryHistogram.txLongMbfcnt);
    uiPrintf("RTS/CTS: %s, %s, cnt: %d, frameLossThreshold: %d, dot11RTSThreshold: %d\n",
             TxRetryHistogram.rtsCtsEnable ? "enable" : "disable",
             TxRetryHistogram.rtsCtsInUse ? "In Use" : "Not In Use",
             TxRetryHistogram.rtsCtsCount,
             TxRetryHistogram.frameLossThreshold,
             pdevInfo->staConfig.defaultDot11RTSThreshold);

    for (rate = pTxRetryHistoBB->minTableTxRate; rate <= pTxRetryHistoBB->maxTableTxRate; rate++) {
          if (RateCtrlTbl[rate].valid) {
              uiPrintf("rate[%d}: %s, backoff rate: %s, up threshold: %d, dn threshold: %d\n",
                       rate,
                       pDataRate[rate],
                       pDataRate[RateCtrlTbl[rate].bkoffRidx],
                       RateCtrlTbl[rate].up_threshold,
                       RateCtrlTbl[rate].dn_threshold);
          }
    }

    pSib = NULL;

    while ((pSib = sibNextStaFind(pSib)) != NULL) {
        int   i;
        uiPrintf("SIB (%p) %3d, short frame rate %s Mbps, long frame rate %s Mbps, RTS/CTS count: %d\n",
                 pSib, pSib->assocId & 0x3fff,
                 pDataRate[pSib->txRateCtrl.sf.rateIdx],
                 pDataRate[pSib->txRateCtrl.lf.rateIdx],
                 pSib->txRateCtrl.rtsCts.count);
        uiPrintf("Lo: %s, Hi: %s\n",
                 pDataRate[pSib->txRateCtrl.minTxRateIdx],
                 pDataRate[pSib->txRateCtrl.maxTxRateIdx]);
        for (i = pTxRetryHistoBB->minTableTxRate; i <= pTxRetryHistoBB->maxTableTxRate; i++) {
            if (pSib->txRateCtrl.sf.rateCnt[i] || pSib->txRateCtrl.lf.rateCnt[i]) {
                uiPrintf("[%s]: %d, %d ",
                         pDataRate[i],
                         pSib->txRateCtrl.sf.rateCnt[i],
                         pSib->txRateCtrl.lf.rateCnt[i]);
            }
        }
        uiPrintf("\n");
    }
}


void frShow(int snr, int len)
{
    SIB_ENTRY   *pSib;


    if (!TxRetryHistogram.rateCtrlEnable) {
        return;
    }

    pSib = NULL;

    while ((pSib = sibNextStaFind(pSib)) != NULL) {
        pSib->txRateCtrl.inUse = 1;     /* Lock critical region */
        if (len == 0 || len == 1) {
            uiPrintf("SIB %3d, short frame rate %s -- throttle cnt: %8d, up: %8d, down: %8d\n",
                     pSib->assocId & 0x3fff,
                     pDataRate[pSib->txRateCtrl.sf.rateIdx],
                     pSib->txRateCtrl.sf.throttle_count,
                     pSib->txRateCtrl.sf.throttle_upcnt,
                     pSib->txRateCtrl.sf.throttle_dncnt);
        }
        if (len == 0 || len == 2) {
            uiPrintf("SIB %3d,  long frame rate %s -- throttle cnt: %8d, up: %8d, down: %8d\n",
                     pSib->assocId & 0x3fff,
                     pDataRate[pSib->txRateCtrl.lf.rateIdx],
                     pSib->txRateCtrl.lf.throttle_count,
                     pSib->txRateCtrl.lf.throttle_upcnt,
                     pSib->txRateCtrl.lf.throttle_dncnt);
        }

        pSib->txRateCtrl.inUse = 0;     /* Unlock critical region */
    }

    uiPrintf("Rate Backoff -- short: %d, long: %d\n",
             TxRetryHistogram.swRetrySfRateCount,
             TxRetryHistogram.swRetryLfRateCount);

    uiPrintf("short retry frame: %d, long retry frame: %d\nTx Frame -- total: %d, short: %d (%d%%), long: %d (%d%%)\n",
             TxRetryHistogram.shortFrameCount, TxRetryHistogram.longFrameCount,
             TxRetryHistogram.txFrameCount,
             TxRetryHistogram.txShortFrameCount,
             (TxRetryHistogram.txShortFrameCount * 100)/TxRetryHistogram.txFrameCount,
             TxRetryHistogram.txLongFrameCount,
             (TxRetryHistogram.txLongFrameCount * 100)/TxRetryHistogram.txFrameCount);
}

void histoShow(int aid, int len, int verbose)
{
    int         noUpdate;
#ifdef  RC_HISTORY
    int         idx;
    A_UINT32    tick0;
    struct throttle_history_item_s *pItem;
    A_UINT32    hThput, oThput, lThput;
    A_UINT32    hfloss, ofloss, lfloss;
#endif  /* RC_HISTORY */

    if (!TxRetryHistogram.rateCtrlEnable) {
        return;
    }

    noUpdate = TxRetryHistogram.dontUpdate;
    TxRetryHistogram.dontUpdate = 1;
    frShow(0, len);

#ifdef  RC_HISTORY
    if (TxRetryHistogram.pHistory->head || TxRetryHistogram.pHistory->tail) {
        uiPrintf("head: %d, tail: %d\n",
                 TxRetryHistogram.pHistory->head, TxRetryHistogram.pHistory->tail);
    }

    idx = TxRetryHistogram.pHistory->head;
    pItem = (struct throttle_history_item_s *)&TxRetryHistogram.pHistory->item[idx];
    tick0 = pItem->ticks;
    while (TxRetryHistogram.pHistory->tail != idx) {
        if (pItem->ticks < tick0) {
            tick0 = pItem->ticks;
        }
        idx = (idx + 1) % TxRetryHistogram.pHistory->max_count;
        pItem = (struct throttle_history_item_s *)&TxRetryHistogram.pHistory->item[idx];
    }

    idx = TxRetryHistogram.pHistory->head;
    pItem = (struct throttle_history_item_s *)&TxRetryHistogram.pHistory->item[idx];
    while (TxRetryHistogram.pHistory->tail != idx) {
        int   k, r, rmax;
        int   day, hr, min, sec;

        pItem = (struct throttle_history_item_s *)&TxRetryHistogram.pHistory->item[idx];
        if ((len == 0) || (len == 1 && pItem->len == 0) || (len == 2 && pItem->len)) {
            if ((aid == 0) || (aid && (aid == (pItem->pSib->assocId & 0x3fff)))) {
                day = hr = min = sec = 0;
                sec = (pItem->ticks - tick0)/CLOCK_TICKS_PER_SEC;
                day = (sec /(60*60)) / 24;
                hr  = (sec /(60*60)) >= 24 ? sec/(60*60) % 24 : sec/(60*60);
                min = (sec /60) >= 60 ? sec/60 % 60 : sec/60;
                sec %= 60;
                lThput = pItem->thruput[LOWER_RELRATE];
                oThput = pItem->thruput[CURRENT_RELRATE];
                hThput = pItem->thruput[HIGHER_RELRATE];
                lfloss = pItem->frameLoss[LOWER_RELRATE];
                ofloss = pItem->frameLoss[CURRENT_RELRATE];
                hfloss = pItem->frameLoss[HIGHER_RELRATE];

                uiPrintf("%3d) SIB %3d %d:%02d:%02d:%02d.%02d %s %s %s => %s  ",
                         idx,
                         pItem->pSib->assocId & 0x3fff,
                         day, hr, min, sec,
                         (pItem->ticks - tick0) % CLOCK_TICKS_PER_SEC,
                         pItem->mode == RC_BKGND_MODE ? "B" : "F",
                         pItem->len ? "L" : "S",
                         pDataRate[pItem->curRate],
                         pDataRate[pItem->newRate]);
                if (pItem->curRate == pItem->pSib->txRateCtrl.maxTxRateIdx) {
                    uiPrintf ("%5d, %5d  (%s%2d%%) %2d%%,%2d%%\n",
                              lThput, oThput,
                              lThput > oThput ? "-" : "",
                              100 - (lThput >= oThput ? (oThput*100)/lThput : (lThput*100)/oThput),
                              lfloss, ofloss);
                } else {
                    if (pItem->curRate == LOWEST_TX_RATE) {
                        uiPrintf("%5d, %5d  (%s%2d%%) %2d%%,%2d%%\n",
                                 oThput, hThput,
                                 hThput > oThput ? "-" : "",
                                 100 - (hThput >= oThput ? (oThput*100)/hThput : (hThput*100)/oThput),
                                 ofloss, hfloss);
                    } else {
                        uiPrintf("%5d, %5d, %5d  (%s%2d%%, %s%2d%%) %2d%%,%2d%%,%2d%%\n",
                                 lThput, oThput, hThput,
                                 lThput > oThput ? "-" : "",
                                 100 - (lThput >= oThput ? (oThput*100)/lThput : (lThput*100)/oThput),
                                 hThput > oThput ? "-" : "",
                                 100 - (hThput >= oThput ? (oThput*100)/hThput : (hThput*100)/oThput),
                                 lfloss, ofloss, hfloss);
                    }
                }
                rmax = MAX_REL_RATE;
                if (pItem->curRate == pItem->pSib->txRateCtrl.maxTxRateIdx) {
                    rmax = HIGHER_RELRATE;
                }
                r = LOWER_RELRATE;
                if (pItem->curRate == LOWEST_TX_RATE) {
                    r = CURRENT_RELRATE;
                }
                if (verbose) {
                    for (; r < rmax; r++) {
                        if (!pItem->len) {
                            uiPrintf("R%d, sf -- ", r);
                            for (k = 0; k <= MAX_TX_RETRIES; k++) {
                                if (pItem->sf.tbl[r].retryBin[k]){
                                    uiPrintf("[%2d]:%2d%s", k, pItem->sf.tbl[r].retryBin[k], (k+1) < MAX_TX_RETRIES ? " " : "\n");
                                }
                            }
                        } else {
                            uiPrintf("R%d, lf -- ", r);
                            for (k = 0; k <= MAX_TX_RETRIES; k++) {
                                if (pItem->lf.tbl[r].retryBin[k]) {
                                    uiPrintf("[%2d]:%2d%s", k, pItem->lf.tbl[r].retryBin[k], (k+1) < MAX_TX_RETRIES ? " " : "\n");
                                }
                            }
                        }
                        uiPrintf("\n");
                    }
                }
            }
        }
        idx = (idx + 1) % TxRetryHistogram.pHistory->max_count;
    }
#endif  /* RC_HISTORY */
    TxRetryHistogram.dontUpdate = (A_UINT8) noUpdate;
}


void histoClear(void)
{
#ifdef  RC_HISTORY
    struct  throttle_history_s *p;
#endif  /* RC_HISTORY */
    int enable, opmode;
    int len, noUpdate;

    noUpdate = TxRetryHistogram.dontUpdate;
    TxRetryHistogram.dontUpdate = 1;
    enable = TxRetryHistogram.rateCtrlEnable;
    opmode = TxRetryHistogram.opmode;
#ifdef  RC_HISTORY
    p = TxRetryHistogram.pHistory;
#endif
    len = TxRetryHistogram.shortFrameLen;
    A_MEM_ZERO((char *)&TxRetryHistogram, sizeof(TxRetryHistogram));
    TxRetryHistogram.shortFrameLen = (A_UINT16) len;
#ifdef  RC_HISTORY
    TxRetryHistogram.pHistory->max_count = MAX_THROTTLE_HISTORY;
    TxRetryHistogram.pHistory = p;
    TxRetryHistogram.pHistory->count = 0;
    TxRetryHistogram.pHistory->head = 0;
    TxRetryHistogram.pHistory->tail = 0;
#endif  /* RC_HISTORY */
    TxRetryHistogram.rateCtrlEnable = (A_UINT8) enable;
    TxRetryHistogram.dontUpdate = (A_UINT8) noUpdate;
    TxRetryHistogram.opmode = opmode;
}


void histolog(int state)
{
    TxRetryHistogram.dontUpdate = state ? 0 : 1;
    csVerbose = state ? 0 : 1;
}


void frameLossThreshold (int v)
{
    TxRetryHistogram.rtsCtsEnable = 1;
    TxRetryHistogram.frameLossThreshold = v;
}


void rateCfg(int idx, int valid)
{
    RateCtrlTbl[idx].valid = valid;
}


void threshold(int idx, int up, int dn)
{
    int         rate;

    if (idx) {
        if (up) {
            RateCtrlTbl[idx].up_threshold = up;
        }
        if (dn) {
            RateCtrlTbl[idx].dn_threshold = dn;
        }
    } else {
        for (rate = 0; rate <= pTxRetryHistoBB->maxTableTxRate; rate++) {
            if (up) {
                RateCtrlTbl[rate].up_threshold = up;
            }
            if (dn) {
                RateCtrlTbl[rate].dn_threshold = dn;
            }
        }
    }
}

#ifdef BUILD_AP
void jumpdotTask(void)
{
    int i;

    for (;;) {
        i++;
    }
}

void jmpdot(void)
{
    taskSpawn("jumpdot", 250, 0, 2000, (FUNCPTR)jumpdotTask,
              (int)0,
              0, 0, 0, 0, 0, 0, 0, 0, 0);
}
#endif /* BUILD_AP */
#endif /* DEBUG */

#ifndef BUILD_AP
void txRateFree(void)
{
    if (pTxTimeTbl) {
        A_DRIVER_FREE(pTxTimeTbl, sizeof(struct txTime_s), 0);
        pTxTimeTbl = 0;
    }

#ifdef  RC_HISTORY
    if (TxRetryHistogram.pHistory) {
        A_DRIVER_FREE(TxRetryHistogram.pHistory, sizeof(struct throttle_history_item_s), 0);
        TxRetryHistogram.pHistory = 0;
        csInitialized = 0;
#ifdef  BUILD_AP
        /*
         *  By definition, AP is heard by all stations in its BSS
         *  and will never be a hidden node. Don't enable RTS/CTS
         *  unless is forced to do so.
         */
        TxRetryHistogram.rtsCtsEnable = 0;
#else
        TxRetryHistogram.rtsCtsEnable = 1;
#endif
        TxRetryHistogram.rtsCtsInUse = 0;
        TxRetryHistogram.rtsCtsCount = 0;
        TxRetryHistogram.rateCtrlEnable = 0;
        TxRetryHistogram.shortFrameLen = SHORT_FRAME_LEN;
        TxRetryHistogram.frameLossThreshold = FRAME_LOSS_THRESHOLD;
        TxRetryHistogram.defaultDot11RTSThreshold = 0;
        TxRetryHistogram.opmode = RC_FOREGND_MODE;
    }
#endif  /* RC_HISTORY */
}
#endif  /* BUILD_AP */

/************************************************************************/

#ifdef DEBUG_PKT_LOG
/*
 * Logging code
 */

#define LOG_TYPE_RX     1
#define LOG_TYPE_TX     2
#define LOG_TYPE_RATE_GET   3
#define LOG_TYPE_RATE_UPDATE    4

typedef struct TxRateCtrlLog_s {
    A_UINT8 logType;
    struct TxRateCtrlCB_s rc;
    union {
    struct {
        A_UINT16 frameLen;
        A_UINT8 rate;
        A_UINT8 defAnt;
        A_UINT8 sibAnt;
    } g;
    struct {
        A_UINT8 rate;
        A_UINT8 xRetry;
        A_UINT8 retries;
        A_UINT16 frameLen;
        A_UINT8 rssiAck;
        A_UINT8 defAnt;
        A_UINT8 sibAnt;
    } t;
    } u;
} TxRateCtrlLog;

typedef struct {
    A_UINT32 timeStamp;
    union {
    struct {
        A_UINT8 logType;
        A_UINT8 rate;
        A_UINT8 rssi;
        A_UINT8 status;
        A_UINT16 length;
        A_UINT16 seqCtrl;
        A_UINT16 frameCtrl;
        A_UINT8 antRx;
        A_UINT8 dummy;
        A_UINT16 timeStamp;
    } rx;
    struct {
        A_UINT8 logType;
        A_UINT8 rate;
        A_UINT8 retries;
        A_UINT8 status;
        A_UINT16 length;
        A_UINT16 seqCtrl;
        A_UINT16 frameCtrl;
        A_UINT8 ackRssi;
        A_UINT8 antTx;
        A_UINT16 timeStamp;
    } tx;
    TxRateCtrlLog rate;
    } d;
} LogInfo;

#define LOG_MAX   10000

static LogInfo *LogData = NULL;
static int LogIdx = -1;

void logEventClear(void)
{
    if ( !LogData ) {
    LogData = (LogInfo*)A_DRIVER_MALLOC(sizeof(LogInfo) * LOG_MAX, 0, 0);
    if ( !LogData ) {
        LogIdx = -1;
        return;
    }
    }
    memset(LogData, 0, sizeof(LogInfo) * LOG_MAX);
    LogIdx = 0;
}

void logEventRxPkt(
       A_UINT32 length,
       A_UINT32 rate,
       A_UINT32 rssi,
       A_UINT32 frameCtrl,
       A_UINT32 seqCtrl,
       A_UINT32 status,
       A_UINT32 antRx,
       A_UINT16 timeStamp)
{
    LogInfo log;

    if ( LogIdx < 0 ) {
    logEventClear();
    if ( LogIdx < 0 ) return;
    }

    log.timeStamp    = A_MS_TICKGET();
    log.d.rx.logType = LOG_TYPE_RX;
    log.d.rx.rate    = (A_UINT8)rate;
    log.d.rx.rssi    = (A_UINT8)rssi;
    log.d.rx.frameCtrl = (A_UINT16)frameCtrl;
    log.d.rx.seqCtrl  = (A_UINT16)seqCtrl;
    log.d.rx.status  = (A_UINT8)status;
    log.d.rx.length  = (A_UINT16)length;
    log.d.rx.antRx   = (A_UINT8)antRx;
    log.d.rx.timeStamp = timeStamp;
    LogData[LogIdx++] = log;
    if ( LogIdx >= LOG_MAX ) LogIdx = 0;
}

void logEventTxPkt(A_UINT32 length, A_UINT32 rate,
    A_UINT32 ackRssi, A_UINT32 retries,
    A_UINT32 status,
    A_UINT32 frameCtrl,
    A_UINT32 seqCtrl,
    A_UINT32 antTx,
    A_UINT16 timeStamp)
{
    LogInfo log;

    if ( LogIdx < 0 ) {
        logEventClear();
        if ( LogIdx < 0 ) 
            return;
    }
    log.timeStamp  = A_MS_TICKGET();
    log.d.tx.logType = LOG_TYPE_TX;
    log.d.tx.rate    = (A_UINT8)rate;
    log.d.tx.length  = (A_UINT16)length;
    log.d.tx.retries = (A_UINT8)retries;
    log.d.tx.status  = (A_UINT8)status;
    log.d.tx.ackRssi = (A_UINT8)ackRssi;
    log.d.tx.antTx   = (A_UINT8)antTx;
    log.d.tx.frameCtrl = (A_UINT16)frameCtrl;
    log.d.tx.seqCtrl  = (A_UINT16)seqCtrl;
    log.d.tx.timeStamp = timeStamp;
    LogData[LogIdx++] = log;
    if ( LogIdx >= LOG_MAX ) 
        LogIdx = 0;
}

void logEventRateCtrl(TxRateCtrlLog *rateLog)
{
    LogInfo log;

    if ( LogIdx < 0 ) {
        logEventClear();
        if ( LogIdx < 0 ) 
            return;
    }
    log.timeStamp  = A_MS_TICKGET();
    log.d.rate     = *rateLog;
    LogData[LogIdx++] = log;
    if ( LogIdx >= LOG_MAX ) 
        LogIdx = 0;
}

void logEventGet(A_UINT32 *len, A_UINT32 *maxCnt, A_UINT32 *currIdx,
        A_UINT32 getIdx, void *data)
{
    if ( LogIdx < 0 ) {
        *len = sizeof(LogInfo);
        *maxCnt = 0;
        *currIdx = 0;
        return;
    }
    *len = sizeof(LogInfo);
    *maxCnt = LOG_MAX;
    *currIdx = LogIdx;
    if ( getIdx >= LOG_MAX ) 
        getIdx = LOG_MAX - 1;
    if ( LogIdx >= 0 ) {
        *(LogInfo*)data = LogData[getIdx];
    }
}

#endif /* #ifdef DEBUG_PKT_LOG */

/************************************************************************/

/*
 * New rate adaptation code
 */

/*
 *  This function determines and returns the new Tx rate index
 *  The rate control algorithm will occasionally (1 out of 10)
 *  send packets at higher and lower rates to gather statistics 
 *  on thruput at those rates.
 */
static int txRateFindCB(SIB_ENTRY *pSib, int frameLen)
{
    int rate;
    struct TxRateCtrlCB_s *pRc;
    A_UINT8 rssiLast, rssiReduce = 0;
#ifdef DEBUG_PKT_LOG
    TxRateCtrlLog log;
#endif

    pRc = &pSib->txRateCtrl.cb;
    rssiLast = MIN(pRc->rssiLast, pRc->rssiLastPrev);

    pSib->antTx = pRc->antTx;

    /*
     * Age (reduce) last ack rssi based on how old it is.
     * The bizarre numbers are so the delta is 160msec,
     * meaning we divide by 16.
     *   0msec   <= dt <= 41msec:   don't derate
     *   41msec  <= dt <= 185msec:  derate linearly by 10dB
     *   185msec <= dt:             derate by 10dB
     */
    {
        A_UINT32 dt = A_MS_TICKGET() - pRc->rssiTime;

        if ( dt >= 185 ) {
            rssiReduce = 10;
        } else if ( dt >= 25 ) {
            rssiReduce = (A_UINT8)((dt - 25) / 16);
        }
    }

    /*
     * Reduce rssi for long packets
     */
    if ( frameLen < 256 ) {
        /* no change */
    } else if ( frameLen < 1000 ) {
        rssiReduce += 1;    /* need 1 more dB for packets 256-1000 */
    } else {
        rssiReduce += 2;    /* need 2 more dB for packets >1000 */
    }

    /*
     * Now reduce rssiLast by rssiReduce
     */
    if ( rssiLast < rssiReduce ) {
        rssiLast = 0;
    } else {
        rssiLast -= rssiReduce;
    }

    /*
     * Now look up the rate in the rssi table and return it.
     * If no rates match then we return 0 (lowest rate)
     */
    for ( rate = 0 ; rate < MAX_TBL_RATE - 1 ; rate++ ) {
        if ( rssiLast < pRc->state[rate+1].rssiThres ) break;
        }
    
    pRc->rssiLastLkup = rssiLast;

    if ( rate > pRc->rateMax ) {
    A_UINT32 nowMsec = A_MS_TICKGET();
    rate = pRc->rateMax;
    if ( rate < MAX_TBL_RATE - 1
                && nowMsec - pRc->probeTime > 100
                && pRc->hwMaxRetryPktCnt >= 4 ) {
            rate++;
            pRc->probeRate = (A_UINT8)rate;
            pRc->probeTime = nowMsec;
            pRc->hwMaxRetryPktCnt = 0;
        }
    }

    /*
     * Make sure rate is not higher than the allowed maximum.
     * We should also enfore the min, but I suspect the min is
     * normally 1 rather than 0 because of the rate 9 vs 6 issue
     * in the old code.
     */
    if ( rate > pSib->txRateCtrl.maxTxRateIdx ) {
        rate = pSib->txRateCtrl.maxTxRateIdx;
    }

    if ( !TxRetryHistogram.rateCtrlEnable ) {
        int unit = 0;
        WLAN_DEV_INFO *pdevInfo = globDrvInfo.pDevInfoArray[unit];
        rate = pdevInfo->localSta->currentOperationalRateIndex;
    }


#ifdef DEBUG_PKT_LOG
    log.rc         = *pRc;
    log.logType    = LOG_TYPE_RATE_GET;
    log.u.g.frameLen = (A_UINT16)frameLen;
    log.u.g.rate   = (A_UINT8)rate;
    log.u.g.defAnt = (A_UINT8)0;      // defaultAntennaGet();
    log.u.g.sibAnt = (A_UINT8)pSib->antTx;
    logEventRateCtrl(&log);
#endif

    return rate;
}

/*
 * This routine is called by the Tx interrupt service routine to update the
 * Tx retry count table
 */
static void RateCtrlTxRetryCntUpdateCB
(
    int txRate, int Xretries, int retries, int frameLen, SIB_ENTRY *pSib,
    A_UINT8 rssiAck, A_UINT8 antTx, A_UINT8 *flipReq
)
{
    WLAN_DEV_INFO *pdevInfo;    /* device information structure */
    int unit = 0;
    struct TxRateCtrlCB_s *pRc;
    int rate;
    A_UINT32 nowMsec = A_MS_TICKGET();
    A_UINT8 costDelta;
#ifdef DEBUG_PKT_LOG
    TxRateCtrlLog log;
#endif
    A_BOOL  stateChange = FALSE;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];

    if ( retries < 0 || retries >= MAX_TX_RETRIES )
    return;

    if ( txRate < 0 || txRate >= MAX_TBL_RATE )
    return;

    if (!pSib)
    return;

    pRc = &pSib->txRateCtrl.cb;
    if ( flipReq ) 
        *flipReq = 0;

    /*
     * treat negative rssi as zero
     */
    if ( rssiAck >= 128 ) 
        rssiAck = 0;

    /*
     * Update the cost.
     */
    costDelta = (2 << retries) + Xretries * 32;
    if ( costDelta > 2 ) {
        if ( pRc->state[txRate].cost < 192 ) {
            pRc->state[txRate].cost += costDelta;
            stateChange = TRUE;
        }
    } else {
        if ( pRc->state[txRate].cost > 0 ) {
            pRc->state[txRate].cost--;
            stateChange = TRUE;
        } else {
            pRc->state[txRate].cost = 0;
        }
    }

    if ( Xretries ) {
        /*
         * Oops, it didn't work at all.  Set the required rssi
         * to the rssiAck we used to lookup the rate, plus 4dB.
         * The immediate effect is we won't try this rate again
         * until we get an rssi at least 4dB higher.
         */
        if ( txRate > 0 ) {
            int rssi = MAX(pRc->rssiLastLkup, pRc->rssiLast - 2);
            if ( pRc->state[txRate].rssiThres + 2 < rssi ) {
                pRc->state[txRate].rssiThres += 4;
            } else if ( pRc->state[txRate].rssiThres < rssi + 2 ) {
                pRc->state[txRate].rssiThres = rssi + 2;
            } else {
                pRc->state[txRate].rssiThres += 2;
            }
            pRc->hwMaxRetryRate = (A_UINT8)txRate;
            stateChange = TRUE;
        }

        /*
         * Also, if we are not doing a probe, we force a significant
         * backoff by reducing rssiLast.  This doesn't have a big
         * impact since we will bump the rate back up as soon as we
         * get any good ack.
         */
        if ( pRc->probeRate == 0 ) {
            if ( pRc->state[txRate].rssiThres >= 6 )
                pRc->rssiLast = pRc->state[txRate].rssiThres - 6;
            else
                pRc->rssiLast = 0;
        }
        pRc->probeRate = 0;
    } else {
#ifndef BUILD_AP
    if ( pRc->antFlipCnt < 0 ) {
        pRc->antFlipCnt++;
    } else if ( retries == 1 || retries == 2 ) {
        antTx = !antTx;
        pRc->antFlipCnt++;
        if ( (retries == 1 && rssiAck >= pRc->rssiLast + 2)
            || pRc->antFlipCnt >= 3 ) {
        pSib->antTx = pRc->antTx = antTx;
        // writePlatformReg(pdevInfo, 0x8058, pSib->antTx ? 2 : 1);
        halSetDefAntenna(pdevInfo, pSib->antTx ? 2 : 1);
        pRc->antFlipCnt = -20;
        }
    } else {
        pRc->antFlipCnt = 0;
    }
#endif
    if ( pRc->antTx != antTx && rssiAck >= pRc->rssiLast ) {
        pSib->antTx = pRc->antTx = antTx;
    }
    // if ( pRc->antTx == antTx && retries == 0 ) {
    pRc->rssiLastPrev = pRc->rssiLast;
    pRc->rssiLast = rssiAck;
    pRc->rssiTime = nowMsec;
    if ( retries > 0 ) {
        /*
         * Don't update anything.  We don't know if this was because
         * of collisions or poor signal.
         *
         * Later: if rssiAck is close to pRc->state[txRate].rssiThres
         * and we see lots of retries, then we could increase
         * pRc->state[txRate].rssiThres.
         */
        pRc->hwMaxRetryPktCnt = 0;
        pRc->probeRate = 0;
#if 1
        if ( retries >= 2 && rssiAck >= pRc->state[txRate].rssiThres
                  && rssiAck < pRc->state[txRate].rssiThres + 3 ) {
            pRc->state[txRate].rssiThres += 1;
            stateChange = TRUE;
        }
#endif
    } else {
        /*
         * It worked with no retries.  First ignore bogus (small)
         * rssiAck values.
         */
        if ( pRc->hwMaxRetryPktCnt < 255 )
        pRc->hwMaxRetryPktCnt++;
        if ( pRc->probeRate ) {
        pRc->rateMax = pRc->probeRate;
        if ( pRc->state[pRc->probeRate].cost > 30 )
            pRc->state[pRc->probeRate].cost = 30;
        pRc->probeRate = 0;
        }
        if ( rssiAck < RssiAckValidMin[txRate] ) {
#ifdef  DEBUG
        uiPrintf("**** Discarding rssi: %d < %d\n", rssiAck,
                        RssiAckValidMin[txRate]);
#endif
        } else {
        /*
         * Average the rssi
         */
        if ( txRate != pRc->rssiSumRate ) {
            pRc->rssiSumRate = (A_UINT8)txRate;
            pRc->rssiSum = pRc->rssiSumCnt = 0;
        }
        pRc->rssiSum += rssiAck;
        pRc->rssiSumCnt++;
        if ( pRc->rssiSumCnt >= 4 ) {
            A_UINT8 rssiAckAvg = (pRc->rssiSum + 2) / 4;
            pRc->rssiSum = pRc->rssiSumCnt = 0;
            /*
             * Now reduce the current rssi threshold.
             */
            if ( rssiAckAvg < pRc->state[txRate].rssiThres + 2 )
            pRc->state[txRate].rssiThres--;
            stateChange = TRUE;
        }
        }
    }
    }

    /*
     * If this rate looks bad (high cost) then stop using it for
     * a while (except if we are probing).
     */
    if ( pRc->state[txRate].cost > 100
            && !pRc->probeRate
            && txRate > 0
            && txRate <= pRc->rateMax ) {
    pRc->rateMax = txRate - 1;
    /*
     * Don't probe for a little while
     */
    pRc->probeTime = nowMsec;
    }

    if (stateChange) {
        /* 
         * Make sure the rates above this have higher rssi thresholds
         */
        for ( rate = txRate ; rate < MAX_TBL_RATE - 1 ; rate++ ) {
            if ( pRc->state[rate].rssiThres + RssiAckDeltaMin[rate]
                > pRc->state[rate+1].rssiThres ) {
                pRc->state[rate+1].rssiThres = pRc->state[rate].rssiThres
                        + RssiAckDeltaMin[rate];
            }
        }

        /* 
         * Make sure the rates below this have lower rssi thresholds
         */
        for ( rate = txRate - 1 ; rate >= 0 ; rate-- ) {
            if ( pRc->state[rate].rssiThres + RssiAckDeltaMin[rate]
                        > pRc->state[rate+1].rssiThres ) {
                if ( pRc->state[rate+1].rssiThres < RssiAckDeltaMin[rate] ) {
                    pRc->state[rate].rssiThres = 0;
                } else {
                    pRc->state[rate].rssiThres = pRc->state[rate+1].rssiThres
                                - RssiAckDeltaMin[rate];
                }
                if ( pRc->state[rate].rssiThres < RssiAckValidMin[txRate] ) {
                    pRc->state[rate].rssiThres = RssiAckValidMin[txRate];
                }
            }
            if ( pRc->state[rate].cost > pRc->state[rate+1].cost ) {
                pRc->state[rate].cost = pRc->state[rate+1].cost;
            }
        }
    }
    /*
     * Every 50msec or so we reduce the thresholds
     */
    if ( nowMsec - pRc->rssiDownTime >= 50 ) {
        for ( rate = 0 ; rate < MAX_TBL_RATE ; rate++ ) {
            if ( pRc->state[rate].rssiThres > RssiAckValidMin[rate] + 3 ) {
                pRc->state[rate].rssiThres -= 1;
            }
        }
        pRc->rssiDownTime = nowMsec;
    }

#if DEBUG_PKT_LOG
    log.rc         = *pRc;
    log.rc.misc[0] = (A_UINT8)pSib->numTxPending;
    log.logType    = LOG_TYPE_RATE_UPDATE;
    log.u.t.rate   = (A_UINT8)txRate;
    log.u.t.xRetry = (A_UINT8)Xretries;
    log.u.t.retries = (A_UINT8)retries;
    log.u.t.frameLen = (A_UINT16)frameLen;
    log.u.t.rssiAck = rssiAck;
    log.u.t.defAnt = (A_UINT8)0;        // defaultAntennaGet();
    log.u.t.sibAnt = (A_UINT8)pSib->antTx;
    logEventRateCtrl(&log);
#endif
}

static void swretryTxRateBackoffCB(ATHEROS_DESC *pErrorDesc)
{
    A_UINT16            bkoffRateIdx;
    WLAN_DEV_INFO       *pdevInfo = globDrvInfo.pDevInfoArray[0];
    A_BOOL              shortPreamble;

    bkoffRateIdx = (A_UINT16)txRateFindCB(pErrorDesc->pDestSibEntry, 
                pErrorDesc->hwControl.tx.frameLength);
    shortPreamble = (pErrorDesc->hwControl.tx.transmitRate & CCK_SHORT_PREAMBLE_ENABLE) || 0;

    pErrorDesc->hwControl.tx.transmitRate =
        pdevInfo->staConfig.pRateTable->info[bkoffRateIdx].plcpRate;

    /* swretry has the frame still in network order */
    descDurationSet(pdevInfo, pErrorDesc, bkoffRateIdx, TRUE, shortPreamble);
    return;
}

static void swretryDefaultTxRateSetCB(ATHEROS_DESC *pDesc)
{
    WLAN_DEV_INFO   *pdevInfo = globDrvInfo.pDevInfoArray[0];
    A_UINT16 rateIdx;

    rateIdx = (A_UINT16)txRateFind(pDesc->pDestSibEntry,
             pDesc->hwControl.tx.frameLength);
    pDesc->hwControl.tx.transmitRate =
        pdevInfo->staConfig.pRateTable->info[rateIdx].plcpRate;

    /* swretry has the frame still in network order */
    descDurationSet(pdevInfo, pDesc, rateIdx, TRUE, 
        (A_BOOL) pDesc->pDestSibEntry->capInfo.shortPreamble);
    return;
}
