/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

/*
DESCRIPTION
This file contains the data structures and routines to manage
the shared data items in STA table
*/

#ident  "ACI $Id: //depot/sw/releases/2.0.1/src/common/statab.c#1 $"

#ifdef VXWORKS

#include "vxWorks.h"
#include "stdio.h"
#include "endLib.h"
#include "time.h"
#include "tickLib.h"

#else

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#endif

#include "wlantype.h"
#include "wlanproto.h"
#include "wlansme.h"
#include "wlandrv.h"
#include "wlanext.h"
#include "display.h"
#include "ui.h"
#include "hwext.h"
#include "wlansmeext.h"
#include "halApi.h"

#ifdef VXWORKS
#include "net/if_dl.h"
#ifdef PRJ_BUILD
#include "prjParams.h"
#endif
#endif

#ifdef BUILD_AP
#include "apcfg.h"
#include "apvlan.h"
#include "dot1x.h"
#ifdef INCLUDE_RCW
#include "rc.h"
#include "ws_global.h"
#endif /* INCLUDE_RCW */
#endif /* BUILD_AP */

/*
 *  Forward references
 */
LOCAL A_STATUS  sibEntryAddLocked(SIB_ENTRY *);
LOCAL SIB_ENTRY *sibEntryFindLocked(WLAN_MACADDR *);
LOCAL void      sibEntryDelLocked(SIB_ENTRY *);
LOCAL SIB_ENTRY *sibVictimFindLocked(SIB_ENTRY *localSib, BSSDESCR *apDescr);

#ifdef  DEBUG
int agingDebug = 0;
#endif

/*
DESCRIPTION

This module implements the management functions for AP's Station
Information Database.

The SIB is related to the AP host key table and MAC key Cache.

At the first step only one semaphore is utilized to implement the mutual
exclusion.

Functions:

*   find the entry which has the expected macAddr
*   add an entry
*   delete an entry
*   get an used entry
*   get a free entry
*   modify an entry
*
*/

static char *dataRateTbl[] =
    { "6", "9", "12", "18", "24", "36", "48", "54" };

static char *dataRate2xTbl[] =
    { "12", "18", "24", "36", "48", "72", "96", "108" };

static char *dataRate11bTbl[] =
    { "1", "2", "5.5", "11" };

static char *rcRateTbl[] =
    { "6", "6", "9", "12", "18", "24", "36", "48", "54" };
static char *rcRate2xTbl[] =
    { "12", "12", "18", "24", "36", "48", "72", "96", "108"};
static char *rcRate11bTbl[] =
    { "1", "2", "5.5", "11" };


/***************************************************
*
* sibTableInit - Initialize the SIB table at system initialization stage
*
* This routine initilizes the SIB Table during the system initialization
* stage
*
* RETURNS: always A_OK.
*/

A_STATUS
sibTableInit(WLAN_DEV_INFO *pdevInfo)
{
    SIB_TABLE *sibtab = pSibTable;
    A_UINT16 i;

    /* Initialize the table semaphores */
    A_SEM_INIT(sibtab->sibTableSem, 0, CREATE_LOCKED);

    /*
     * As we go through the loop, make a linked list of
     * all the entries into pFreeList. Searching for a free
     * entry becomes easy with this linked list.
     */
    sibtab->pFreeList = NULL;

    for (i = 0; i < MAX_STA_NUM; i++) {
        A_MEM_ZERO(&sibtab->sta[i], sizeof(struct sibEntry));
        A_SEM_INIT(sibtab->sta[i].sibLock, 0, CREATE_LOCKED);

        sibtab->sta[i].used = SIB_FREE; /* unused entry */
        sibtab->sta[i].pNext = sibtab->pFreeList;
        sibtab->pFreeList = &sibtab->sta[i];
        sibtab->sta[i].staIndex = i;        /* Index for itself */
        sibtab->sta[i].pSibTab = sibtab;

#ifdef BUILD_AP
        A_SEM_INIT(sibtab->sta[i].txPSQueue.qSem, 0, CREATE_UNLOCKED);
#endif /* BUILD_AP */
        A_SEM_UNLOCK(sibtab->sta[i].sibLock);
    }

    /* assign back ptrs */
    sibtab->pdevInfo = pdevInfo;
    pdevInfo->pSibTable = sibtab;

    sibtab->currentEntryNum = 0;

    for (i = 0; i < MAX_STA_HASH_NUM; i++) {
        pSibHashTable->pFirst[i] =  NULL;
    }
    A_SEM_UNLOCK(sibtab->sibTableSem);
    return A_OK;
}

A_INT32 wlanMACHash(WLAN_MACADDR *pmacAddr)
{
    return((A_INT32)
        (pmacAddr->octets[3] ^
         pmacAddr->octets[4] ^
         pmacAddr->octets[5]) & (MAX_STA_HASH_NUM-1));
}

/*******************************************************************************
*
* sibEntryFind - find the entry which has matched macAddres
*
* This routine searches the entire SIB database to find the match entry which has
* the same MAC address
*
* RETURNS:  the pointer of the entry if successful
*           NULL if unsuccessful
*/
LOCAL INLINE SIB_ENTRY *
sibEntryFindLocked(WLAN_MACADDR *macAddr)
{
    SIB_ENTRY   *sibEntry;
    A_UINT32     hash;

    hash = wlanMACHash(macAddr);
    sibEntry = pSibHashTable->pFirst[hash];
    while (sibEntry) {
        /* parsing through all the entries in linked list */
        if ((sibEntry->used & SIB_VALID) &&
           (A_MACADDR_COMP(&sibEntry->macAddr, macAddr) == 0)) {
            sibEntry->staLastActivityTime = A_MS_TICKGET();
            return sibEntry;
        }
        sibEntry = sibEntry->pNext;
    }

    return NULL;
}

SIB_ENTRY *
sibEntryFind(WLAN_MACADDR *macAddr)
{
    SIB_ENTRY   *sibEntry;          /* a entry from SIB Table */

    A_SIB_TAB_LOCK(pSibTable);
    sibEntry = sibEntryFindLocked(macAddr);
    A_SIB_TAB_UNLOCK(pSibTable);
    return sibEntry;
}

/*******************************************************************************
*
* sibEntryGet - get the entry which is indexed
*
* This routine searches the entire SIB database to find the entry which has
* the expected index #
*
* RETURNS:  the pointer of the entry if successful
*           NULL if unsuccessful
*/

SIB_ENTRY *
sibEntryGet(A_UINT32 entryIndex)    /* the index of the entry */
{
    SIB_ENTRY   *sibEntry;          /* a entry from SIB Table */

    if( entryIndex >= MAX_STA_NUM ) {
        return NULL;
    }
    sibEntry = &pSibTable->sta[entryIndex];

    if (sibEntry->used & SIB_VALID) {
        return sibEntry;
    }
    return NULL;
}

/*******************************************************************************
*
* sibEntryDel - delete the #entryIndex entry which is used
*
* This routine delete the entry with index # of entryIndex
*
* RETURNS:  OK if successful
*           ERROR if unsuccessful
*/
A_STATUS
sibEntryDel(SIB_ENTRY *pSib)
{
    A_SIB_TAB_LOCK(pSibTable);
    A_SIB_ENTRY_LOCK(pSib);

    /* check the status of the entry */
    if (!(pSib->used & (SIB_USED|SIB_VALID)) ||
        (pSib->used & SIB_BUSY) || pSib->numTxPending != 0) {
        A_SIB_ENTRY_UNLOCK(pSib);
        A_SIB_TAB_UNLOCK(pSibTable);
        return A_ERROR;
    }

    sibEntryDelLocked(pSib);

    A_SIB_ENTRY_UNLOCK(pSib);
    A_SIB_TAB_UNLOCK(pSibTable);

    return A_OK;
}

/*
 *  Note, there are some members of the SIB entry that cannot
 *  be set to zero...the SIB entry lock in particular.
 */
LOCAL void
sibEntryDelLocked(SIB_ENTRY *pSib)          /* a entry from SIB Table */
{
    SIB_ENTRY   *pEntry;            /* temporary pointer */
    A_UINT32    hindex;             /* hash table index */

    if (pSib->assocId != 0) {
        /* Release the association ID */
        mlmeAssocIdRelease(pSib->assocId);
        pSib->assocId = 0;
    }
    pSib->used = SIB_FREE; // after this, sibEntryFind() won't find this entry

    /* change the status of the entry */
    pSib->staState = STATE_QUIET;

    if (pSib->pChlg) {
        /* This could be left over */
        A_DRIVER_FREE(pSib->pChlg, 0, 0);
        pSib->pChlg = NULL;
    }

    /* change the current number of the SIBTable entries */
    pSibTable->currentEntryNum -= 1;

    /* remove it from the linked list */
    hindex = wlanMACHash(&pSib->macAddr);
    pEntry = pSibHashTable->pFirst[hindex];
    if( pEntry == pSib ) {
        pSibHashTable->pFirst[hindex] = pSib->pNext;
    } else {
        while (pEntry->pNext != NULL && pEntry->pNext != pSib) {
            pEntry = pEntry->pNext;
        }
        /* We must find a match for pSib, if not, something
         * is seriously wrong. It will crash.
         */
        pEntry->pNext = pSib->pNext;
        /* Place it back at the beginning of the free list */
    }

    A_MACADDR_COPY(&nullMacAddr, &pSib->macAddr);

    /* free any keyTable and keyCache resources */
    wlanKeyFree(pSib->pSibTab->pdevInfo, pSib);

    pSib->staLifeStartTime = pSib->staLastActivityTime = 0;
    A_MEM_ZERO(&pSib->stats, sizeof(pSib->stats));
    pSib->psMode = 0;
    pSib->lastRxSeqNo = 0;
    pSib->lastRxFragNo = 0;
    pSib->oneBeforeLastRxSeqNo = 0;
    pSib->seqNum = 0;
    pSib->privacyInvoked = FALSE;
#ifdef BUILD_AP
    pSib->dot1xState = PORTSTATUS_UNAUTH;
#endif
    pSib->pNext = pSibTable->pFreeList;
    pSibTable->pFreeList = pSib;
    return;
}

/*******************************************************************************
*
* sibEntryAlloc - Allocate a SIB entry from sibTable.
*
* This routine searches the SIB table for a free entry, marks
* it used and returns it to the caller. It is not marked valid or
* added to the hash table.
*
* RETURNS:  the pointer of the entry if successful
*           NULL if unsuccessful
*/

SIB_ENTRY *
sibEntryAlloc(SIB_ENTRY *localSib, BSSDESCR *apDescr, WLAN_MACADDR *macAddr)
{
    SIB_ENTRY   *sibEntry, *victim;

    A_SIB_TAB_LOCK(pSibTable);

    sibEntry = sibEntryFindLocked(macAddr);
    if (sibEntry)
    {
        A_SIB_TAB_UNLOCK(pSibTable);
        return sibEntry;
    }

    /*
     *  If SIB table is full, go looking for a stale STA
     */
    if (pSibTable->currentEntryNum >= MAX_STA_NUM) {
        if ((victim = sibVictimFindLocked(localSib, apDescr)) != NULL) {
            A_SIB_TAB_UNLOCK(pSibTable);
            if (sibEntryDel(victim) == A_ERROR) {
                uiPrintf("sibEntryAlloc:  Could not remove inactive "
                         "SIB to make room for new SIB.\n");
                return NULL;
            }
            A_SIB_TAB_LOCK(pSibTable);
        } else {
            A_SIB_TAB_UNLOCK(pSibTable);
            uiPrintf("Access denied ");
            printMacAddress(*macAddr);
            uiPrintf(" out of resources (SIB)\n");
            return NULL;
        }
    }

    sibEntry = pSibTable->pFreeList;
    ASSERT(sibEntry != NULL);

    if (sibEntry->used != SIB_FREE) {
        A_SIB_TAB_UNLOCK(pSibTable);
        uiPrintf("sibEntryAlloc: ERROR: Program bug! Entry not marked free\n");
        return NULL;
    }
    /* Move free list pointer to the next entry */
    pSibTable->pFreeList = sibEntry->pNext;

    /* Initialize various fields to defaults */
    A_MACADDR_COPY(macAddr, &sibEntry->macAddr);
    sibEntry->pNext = NULL;
    sibEntry->used = SIB_USED;

    /* Initial timestamps */
    sibEntry->staLifeStartTime      = A_MS_TICKGET();

    sibEntry->pDevInfo = pSibTable->pdevInfo;
    sibEntry->keyIndex = KEYINDEX_INVALID;
    sibEntry->pPriv = NULL;
#ifdef BUILD_AP
    sibEntry->vlanTag = VLAN_TAG_DEFAULT;
    sibEntry->dot1xState = PORTSTATUS_UNAUTH;
    TRACEBUF_INIT(Tx, sibEntry);
    TRACEBUF_INIT(Rx, sibEntry);
#endif

    sibEntry->bridged = FALSE;

    sibEntry->pChlg = NULL;
    sibEntry->lastRxSeqNo = -1; /* impossible value */
    /*
     *  The previous user of this TXMASK entry may have left it
     *  set, so be sure our first frame clears it.
     */
    sibEntry->swRetry.needClearDest = 1;
    sibEntry->antTx = 0;
    A_MEM_ZERO((char *)&sibEntry->stats, sizeof(sibEntry->stats));

    A_SIB_TAB_UNLOCK(pSibTable);

    return sibEntry;
}

/**************************************************************************
*
* sibEntryAdd - add a entry to the SIBTable
*
* This routine check the specific entry and set it to the valid status
*
* The caller must pass an entry already marked as "used".
*
* RETURNS: A_OK or A_WIN_DRIVER_ERROR.
*/
A_STATUS
sibEntryAdd(SIB_ENTRY *sibEntry)
{
    A_STATUS status;

    A_SIB_TAB_LOCK(pSibTable);
    if (sibEntryFindLocked((WLAN_MACADDR *)&sibEntry->macAddr))
    {
        A_SIB_TAB_UNLOCK(pSibTable);
        return A_OK;
    }
    status = sibEntryAddLocked(sibEntry);
    A_SIB_TAB_UNLOCK(pSibTable);
    return status;
}

LOCAL A_STATUS
sibEntryAddLocked(SIB_ENTRY *sibEntry)
{
    A_UINT32    hindex;

    if (sibEntry->used != SIB_USED) {
        /* must be wrong if reach here */
        uiPrintf("sibEntryAdd: ERROR: sibEntry in wrong state!\n");
        return A_WIN_DRIVER_ERROR;
    }
    /* It is in the correct state. */

    /* Add it to the linked list
     * Add it to the beginning of the list. Since sibEntryAdd
     * will be soon followed by transmit/receive, some thread
     * will soon start searching for it. Better to keep it
     * at the beginning.
     */
    hindex = wlanMACHash(&sibEntry->macAddr);
    sibEntry->used = SIB_USED | SIB_VALID;
    A_SIB_ENTRY_LOCK(sibEntry);
    sibEntry->staLastActivityTime = A_MS_TICKGET();
    sibEntry->pNext = pSibHashTable->pFirst[hindex];
    A_SIB_ENTRY_UNLOCK(sibEntry);

    pSibHashTable->pFirst[hindex] = sibEntry;
    pSibTable->currentEntryNum++ ;

#ifdef DEBUG
    /* Only need a little information usually. */
    uiPrintf("Added STA: ");
    printMacAddress(sibEntry->macAddr);
    uiPrintf(" (%d)\n", A_MS_TICKGET());
#endif

    return A_OK;
}

/*******************************************************************************
*
* sibEntryAgeCheck - Check if a sib entry is stale, and should be removed.
*
* Assumes caller holds the per-SIB lock.
*
* RETURNS:  FALSE if this sib is still valid. TRUE if the sib can be
*           removed with sibEntryAge().
*/
A_BOOL
sibEntryAgeCheck(SIB_ENTRY *pSib, A_UINT32 aging_period)
{
    QUEUE_DETAILS *pQueue;
    A_UINT32       lasttime;
    A_BOOL         idle;

    if (aging_period == 0) {
        return FALSE;
    }

    lasttime = pSib->staLastActivityTime;
    pQueue   = &pSib->txPSQueue;
    idle     = TRUE;
    if (pQueue->pDescQueueHead ||
        pQueue->qFrameCount ||
        pSib->numPSTxPending ||
        pSib->numTxPending ||
        pSib->swRetry.numFilteredPending ||
        pSib->swRetry.qFrames)
    {
        idle = FALSE;
    }

    if ((A_MS_TICKGET() - lasttime) >= aging_period) {
#ifdef DEBUG
        if (agingDebug) {
            char buf[80];

            cliUptimeGet(buf);
            printf("%s -- ", buf);
            printMacAddress(pSib->macAddr);
            printf(" %s\n",
                   !idle ? "busy" : pSib->assocId & 0x3fff ? "disassociated" : "aged out");
        }
#endif
        return idle;
    }
    return FALSE;                       /* still has time left */
}

/*
 * sibEntryAge - handle an elderly SIB in 2 easy steps.
 *   1a) clean up any association state.  This does not send a frame.
 *   1b) send a deauth frame.  For power save stations this will likely
 *       be dropped, but currently is sent anyway.  The station will
 *       happily think it's associated.
 *   2)  remove the sib entry if disassociated.
 *
 * Should not hold the sib lock when calling.
 *
 * XXX Should handle the deauth frame better.  Either not send for
 *     PS stations to save air, or send it before tearing down
 *     the SIB to let it go on the PS queue and tear down the SIB
 *     after the tx is done.  In some ways it's better for the STA
 *     to not know the AP has aged it to avoid thrashing.
 */
void
sibEntryAge(SIB_ENTRY *pSib)
{
    if (pSib->assocId & 0x3fff) {
        /* This only cleans-up state, it does not send a frame */
        wlanMlmeDisassocRequest(pSib->pDevInfo, &pSib->macAddr,
                                REASON_INACTIVE);
    }
    if (pSib->staState & STATE_AUTH) {
#ifdef DEBUG
        /* Only need a little information usually. */
        uiPrintf("Aged STA: ");
        printMacAddress(pSib->macAddr);
        uiPrintf(" (%d)\n", A_MS_TICKGET());
#endif
        wlanMlmeDeauthRequest(pSib->pDevInfo, &pSib->macAddr,
                              REASON_AUTH_EXPIRED);
    } else {
        sibEntryDel(pSib);
    }
}

LOCAL INLINE void
swretryStateReset(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib)
{
    if (pSib->swRetry.retryQueue.pDescQueueHead != NULL)
        freeBuffandDescChain(pdevInfo, pSib->swRetry.retryQueue.pDescQueueHead);
    pSib->swRetry.retryQueue.pDescQueueHead =
        pSib->swRetry.retryQueue.pDescQueueTail = NULL;
    pSib->swRetry.qFrames = 0;

    pSib->swRetry.numFilteredPending = 0;
    pSib->swRetry.needClearDest = 1;

    return;
}

LOCAL INLINE void
psStateReset(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib)
{
    /* grab the psQueue lock */
    A_SEM_LOCK(pSib->txPSQueue.qSem, WAIT_FOREVER);

    if (pSib->txPSQueue.pDescQueueHead != NULL)
        freeBuffandDescChain(pdevInfo, pSib->txPSQueue.pDescQueueHead);
    pSib->txPSQueue.pDescQueueHead =
        pSib->txPSQueue.pDescQueueTail = NULL;
    pSib->txPSQueue.qFrameCount = 0;

    A_SEM_UNLOCK(pSib->txPSQueue.qSem);

    return;
}


void
sibTableStateReset(WLAN_DEV_INFO *pdevInfo)
{
    SIB_ENTRY *pSib;
    int i;

    A_SIB_TAB_LOCK(pSibTable);
    for (i = 0; i < MAX_STA_NUM; i++) {
        pSib = &pSibTable->sta[i];
        if ((pSib->used & SIB_VALID) == 0 || pSib == pdevInfo->localSta)
            continue;

        /* free software retry states */
        swretryStateReset(pdevInfo, pSib);
        pSib->numTxPending = 0;

        /* free PS states */
        psStateReset(pdevInfo, pSib);
        pSib->numPSTxPending = 0;
        pSib->psMode = 0;

    }
    A_SIB_TAB_UNLOCK(pSibTable);
    return;
}

char *sibStateStrGet(SIB_ENTRY *pSib)
{
    char *p = "none";

    if (pSib->staState & STATE_AUTH)
        p = "authenticated";
    if (pSib->staState & STATE_ASSOC)
        p = "associated";
    if (pSib->staState & STATE_PS)
        p = "powersave";
    if (pSib->staState & STATE_ADHOC)
        p = "adhoc";
    if (pSib->staState & STATE_JOINED)
        p = "joined";
    if (pSib->staState & STATE_AP_UP)
        p = "up";
    if (pSib->staState & STATE_AP_STARTING)
        p = "starting";
    if (pSib->staState & STATE_AP_DOWN)
        p = "down";
    if (pSib->staState & STATE_STA_LOCKED)
        p = "locked";

    return p;
}


void sibStateSet
    (
    A_UINT16    idx,
    A_UINT16    state
    )
{
    SIB_ENTRY   *pSib;  /* a entry from SIB Table */

    A_SIB_TAB_LOCK(pSibTable);
    pSib = &pSibTable->sta[idx];
    pSib->staState = state;
    A_SIB_TAB_UNLOCK(pSibTable);
}

void sibTstateSet
    (
    A_UINT16    idx,
    A_UINT16    tstate
    )
{
    SIB_ENTRY   *pSib;  /* a entry from SIB Table */

    A_SIB_TAB_LOCK(pSibTable);
    pSib = &pSibTable->sta[idx];
    pSib->transState = tstate;
    A_SIB_TAB_UNLOCK(pSibTable);
}

void staStatsClear(int staId)
{
    SIB_ENTRY   *pSib;
    int idx;

    A_SIB_TAB_LOCK(pSibTable);
    if (staId == -1) {
        pSib = &pSibTable->sta[MAX_STA_NUM-1];
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--, pSib--) {
            if (pSib->used & SIB_VALID)
                A_MEM_ZERO((char *)&pSib->stats, sizeof(pSib->stats));
        }
    } else {
        pSib = sibTblFind(staId);
        if (pSib) {
            A_MEM_ZERO((char *)&pSib->stats, sizeof(pSib->stats));
        }
    }
    A_SIB_TAB_UNLOCK(pSibTable);
}

SIB_ENTRY *sibAddr(int idx)
{
  return (SIB_ENTRY *)&pSibTable->sta[idx];
}


SIB_ENTRY *sibTblFind(A_UINT32 staId)
{
    int idx;
    int cnt;
    SIB_ENTRY   *pSib;  /* a entry from SIB Table */
    SIB_ENTRY   *v = (SIB_ENTRY *)NULL;

    A_SIB_TAB_LOCK(pSibTable);

    cnt = pSibTable->currentEntryNum;
    if (cnt) {
        pSib = &pSibTable->sta[MAX_STA_NUM-1];
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--, pSib--)
            if (pSib->used & SIB_VALID && staId != 0) {
                if (staId == (A_UINT32)(0x3fff & pSib->assocId)) {
                    v = pSib;
                    break;
                }
            }
    }

    A_SIB_TAB_UNLOCK(pSibTable);

    return v;
}


void sibStatsClear(int idx)
{
    SIB_ENTRY   *pSib;  /* a entry from SIB Table */

    A_SIB_TAB_LOCK(pSibTable);
    pSib = &pSibTable->sta[idx];
    A_MEM_ZERO((char *)&pSib->stats, sizeof(pSib->stats));
    A_SIB_TAB_UNLOCK(pSibTable);
}


void ssidStringGet(char *p)
{
  WLAN_DEV_INFO *   pdevInfo;   /* device information structure */
  int       unit = 0;

  pdevInfo = globDrvInfo.pDevInfoArray[unit];
  sprintf(p, "%s", pdevInfo->bssDescr->ssid.ssid);
}

void ssidStringPut(char *p)
{
  WLAN_DEV_INFO *   pdevInfo;   /* device information structure */
  int       unit = 0;

  pdevInfo = globDrvInfo.pDevInfoArray[unit];
  sprintf(pdevInfo->bssDescr->ssid.ssid, "%s", p);
}


void bssidStringGet(char *p)
{
  WLAN_DEV_INFO *   pdevInfo;   /* device information structure */
  int       unit = 0;
  SIB_ENTRY *pSib;

  pdevInfo = globDrvInfo.pDevInfoArray[unit];
  pSib = pdevInfo->localSta;
  sprintf(p, "%02X:%02X:%02X:%02X:%02X:%02X",
          pSib->macAddr.octets[0], pSib->macAddr.octets[1],
          pSib->macAddr.octets[2], pSib->macAddr.octets[3],
          pSib->macAddr.octets[4], pSib->macAddr.octets[5]);
}

/*
 *  We've run out of free slots in the SIB table.
 *  Return the LRU item so we can kick it out.
 *
 *  Do not victimize ourselves, or the AP BSSDESCR.
 *
 *  Assumption: A_SIB_TAB_LOCK(pSibTable) has been called.
 */
LOCAL SIB_ENTRY *
sibVictimFindLocked(SIB_ENTRY *localSib, BSSDESCR *apDescr)
{
    SIB_ENTRY *pSib, *victim = NULL;
    int idx;
    const WLAN_MACADDR dummyMacAddr = { {0x00, 0x03, 0x7f, 0x00, 0x00, 0x00}};

    if (localSib == NULL) {
        return NULL;
    }

    for (idx = 0; idx < MAX_STA_NUM; ++idx) {
        pSib = &pSibTable->sta[idx];

        if (!(pSib->used & SIB_VALID)) {
            continue;
        }

        /* Don't torch ourself */
        if (pSib == localSib) {
            continue;
        }

        /* Don't torch our partner AP if there is one */
        if (localSib->serviceType == WLAN_STA_SERVICE &&
            pSib->serviceType == WLAN_AP_SERVICE &&
            apDescr != NULL &&
            A_MACADDR_COMP(&pSib->bssId, &apDescr->bssid) == 0) {
            continue;
        }

	if (localSib->serviceType == WLAN_STA_SERVICE &&
	    A_MACADDR_COMP(&pSib->macAddr, &dummyMacAddr) == 0) {
	    continue;
	}

        /* if we already have a better victim */
        if (victim && (pSib->staLastActivityTime >= victim->staLastActivityTime)) {
            continue;
        }

        /* assert a limit (60s) to prevent thrashing and ensure
         * no ref counts are out to this one
         */
        if (!sibEntryAgeCheck(pSib, 60 * 1000)) {
            continue;
        }

        victim = pSib;
    }

    return victim;
}

SIB_ENTRY *
sibNextStaFind(SIB_ENTRY *pCurSib)
{
    SIB_ENTRY   *pSib;          /* a entry from SIB Table */
    SIB_ENTRY   *pNxtSib;       /* a entry from SIB Table */
    WLAN_DEV_INFO * pdevInfo;   /* device information structure */
    int     unit = 0;
    int     cnt, idx;

    pdevInfo = globDrvInfo.pDevInfoArray[unit];
    A_SIB_TAB_LOCK(pSibTable);

    pNxtSib = (SIB_ENTRY *)NULL;
    cnt = pSibTable->currentEntryNum;
    if (cnt > 1) {
        pSib = &pSibTable->sta[MAX_STA_NUM-1];
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--, pSib--) {
            if (pSib->used & SIB_VALID) {
                if (pSib != pdevInfo->localSta) {
                    if (pCurSib == NULL || (pSib < pCurSib)) {
                        pNxtSib = pSib;
                        break;
                    }
                }
                cnt--;
                if (cnt == 0)
                    break;
            }
        }
    }

    A_SIB_TAB_UNLOCK(pSibTable);

    return pNxtSib;
}

int macaddrcmp(SIB_ENTRY **pp1, SIB_ENTRY **pp2)
{
    SIB_ENTRY *p1;
    SIB_ENTRY *p2;

    p1 = pp1[0];
    p2 = pp2[0];
    return A_BCOMP(p1->macAddr.octets, p2->macAddr.octets, sizeof(WLAN_MACADDR));
}

#ifdef VXWORKS
#ifdef INCLUDE_RCW

SIB_ENTRY   *pStaListTbl[MAX_STA_NUM];
int     staListNum;

LOCAL int macAddr2Sib(char *addr, SIB_ENTRY *pSibCopy)
{
    int         cnt, idx;
    SIB_ENTRY   *pSib;          /* a entry from SIB Table */
    char        macaddr[6];
    WLAN_DEV_INFO *pdevInfo;


    strtoWlanMAC(addr, (WLAN_MACADDR *)&macaddr);

    pdevInfo = pSibTable->pdevInfo;

    A_MEM_ZERO(pSibCopy, sizeof(struct sibEntry));

    A_SIB_TAB_LOCK(pSibTable);
    cnt = pSibTable->currentEntryNum;
    if (cnt)
    {
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--)
        {
            pSib = &pSibTable->sta[idx];
            if (bcmp((char *)pSib->macAddr.octets, (char *)&macaddr, sizeof(macaddr)) == 0)
            {
                A_BCOPY((char *)pSib, (char *)pSibCopy, sizeof(SIB_ENTRY));
                if (pSib->used & SIB_VALID)
                {
                    A_SIB_TAB_UNLOCK(pSibTable);
                    return OK;
                }
                break;
            }
        }
    }
    A_SIB_TAB_UNLOCK(pSibTable);

    return ERROR;
}


int staListTblInit(void)
{
    int         idx, j;
    int         cnt;
    SIB_ENTRY   *pSib;          /* a entry from SIB Table */

    if (pSibTable == NULL) {
        return staListNum = 0;
    }

    j = 0;
    A_SIB_TAB_LOCK(pSibTable);
    cnt = pSibTable->currentEntryNum;
    if (cnt)
    {
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--)
        {
            pSib = &pSibTable->sta[idx];
            if (pSib->used & SIB_VALID)
            {
                pStaListTbl[j++] = pSib;
                cnt--;
                if (cnt == 0)
                    break;
            }
        }
    }
    A_SIB_TAB_UNLOCK(pSibTable);

    if (j > 1) {
        qsort(&pStaListTbl[1],
              j-1,
              sizeof(SIB_ENTRY *),
              (void *)&macaddrcmp);
    }

    return staListNum = j;
}


int staMacAddrGet(int index, char *p)
{
    char macaddr[20];
    SIB_ENTRY   *pSib;

    pSib = pStaListTbl[index];
    sprintf(macaddr, "%02X:%02X:%02X:%02X:%02X:%02X",
            pSib->macAddr.octets[0], pSib->macAddr.octets[1],
            pSib->macAddr.octets[2], pSib->macAddr.octets[3],
            pSib->macAddr.octets[4], pSib->macAddr.octets[5]);
    strcpy(p, macaddr);
    return index;
}

int staStatusGet(int index, char *p)
{
    SIB_ENTRY   *pSib;

    pSib = pStaListTbl[index];
    strcpy(p, sibStateStrGet(pSib));
    return index;
}


int staIdGet(int index, char *p)
{
    SIB_ENTRY   *pSib;

    pSib = pStaListTbl[index];
    if (pSib->serviceType == WLAN_STA_SERVICE)
        sprintf(p, "STA %d", pSib->assocId & 0x3fff);
    else
        strcpy(p, "AP");
    return index;
}


void staInfoGet(char *pMacAddr, rcw_SibTbl *pInfo)
{
    SIB_ENTRY     sibCopy;      /* Copy of SIB entry */
    SIB_ENTRY     *pSib;        /* a entry from SIB Table */
    WLAN_DEV_INFO *pdevInfo;    /* device information structure */
    int           i, idx;
    char          *pCipher, *pRate;

    pdevInfo = pSibTable->pdevInfo;
    pSib = (SIB_ENTRY *)&sibCopy;
    if (macAddr2Sib(pMacAddr, pSib) != OK)
        return;

    sprintf(pInfo->sib_Id, "%d", pSib->assocId & 0x3fff);
    sprintf(pInfo->sib_MacAddr, "%02X:%02X:%02X:%02X:%02X:%02X",
            pSib->macAddr.octets[0], pSib->macAddr.octets[1],
            pSib->macAddr.octets[2], pSib->macAddr.octets[3],
            pSib->macAddr.octets[4], pSib->macAddr.octets[5]);
    strcpy(pInfo->sib_State, sibStateStrGet(pSib));
    strcpy(pInfo->sib_PwrSave, pSib->psMode ? "on" : "off");
    strcpy(pInfo->sib_Wep, pSib->privacyInvoked ? "yes" : "no");

    strcpy(pInfo->sib_CipherType, "None");
    strcpy(pInfo->sib_UcastCipher, "None");
    strcpy(pInfo->sib_McastCipher, "None");

    /* Need to show this STAs vlan id independant from privacyInvoked */
    if (pSib->vlanTag == VLAN_TAG_INVALID) {
        strcpy(pInfo->sib_VlanId, "invalid");
    }
    else {
        sprintf(pInfo->sib_VlanId, "%d", pSib->vlanTag);
    }

    if (pSib->privacyInvoked) {
        if (pSib->serviceType == WLAN_STA_SERVICE) {

            if (VALID_UCSE_ELEMENT(&pSib->workingUcseSet)) {
                strcpy(pInfo->sib_UcastCipher, "UNKNOWN");
                if (pSib->workingUcseSet.uCiphers[0] == CSE_NULL) {
                    strcpy(pInfo->sib_UcastCipher, "NULL");
                } else {
                    if (pSib->workingUcseSet.uCiphers[0] == CSE_BASIC_WEP) {
                        strcpy(pInfo->sib_UcastCipher, "WEP");
                    } else {
                        if ((pSib->workingUcseSet.uCiphers[0] == CSE_AES_OCB) ||
                            (pSib->workingUcseSet.uCiphers[0] == CSE_AES_OCB_REVA)) {
                            strcpy(pInfo->sib_UcastCipher, "AES");
                        }
                    }
                }
            }

            if (VALID_MCSE_ELEMENT(&pSib->mcseSet)) {
                strcpy(pInfo->sib_McastCipher, "UNKNOWN");
                if (pSib->mcseSet.mCipher[0] == CSE_NULL) {
                    strcpy(pInfo->sib_McastCipher, "NULL");
                } else {
                    if (pSib->mcseSet.mCipher[0] == CSE_BASIC_WEP) {
                        strcpy(pInfo->sib_McastCipher, "WEP");
                    } else {
                        if ((pSib->mcseSet.mCipher[0] == CSE_AES_OCB) ||
                            (pSib->mcseSet.mCipher[0] == CSE_AES_OCB_REVA)) {
                            strcpy(pInfo->sib_McastCipher, "AES");
                        }
                    }
                }
            }
        }

        if (VALID_UCSE_ELEMENT(&pSib->ucseSet)) {
            pCipher = " Unknown";
            
            strcpy (pInfo->sib_CipherType, "");
            for (i = 0; i < pSib->ucseSet.length/4; i++) {
                if (pSib->ucseSet.uCiphers[i] == CSE_NULL) {
                    pCipher = " CLEAR";
                } else {
                    if (pSib->ucseSet.uCiphers[i] == CSE_BASIC_WEP) {
                        pCipher = " WEP";
                    } else {
                        if ((pSib->ucseSet.uCiphers[i] == CSE_AES_OCB) ||
                            (pSib->ucseSet.uCiphers[i] == CSE_AES_OCB_REVA)) {
                            if (i == 0) {
                                pCipher = " AES";
                            }
                        }
                    }
                }
                strcat (pInfo->sib_CipherType, pCipher);
                pCipher = "";
            }
        }
    }

    strcpy(pInfo->sib_AuthType, pSib->authenAlgo == 0 ?
           "open-system" : (pSib->authenAlgo == 1 ? "shared-key" :
                            (pSib->authenAlgo == 2 ? "802.1x" : "unknown")));
    pInfo->sib_AssocCnt = pSib->stats.Associations;
    pInfo->sib_DisassocCnt = pSib->stats.DisAssociations;
    pInfo->sib_AuthCnt = pSib->stats.Authentications;
    pInfo->sib_DeauthCnt = pSib->stats.DeAuthentications;
    pInfo->rx_Cnt = pSib->stats.GoodReceives;
    pInfo->rx_Data = pSib->stats.RxDataFrames;
    pInfo->rx_Mcast = pSib->stats.MulticastReceives;
    pInfo->rx_Mgmt = pSib->stats.RxMgmtFrames;
    pInfo->rx_Ctrl = pSib->stats.RxCtrlFrames;
    pInfo->rx_Errs = pSib->stats.ReceiveErrors;
    pInfo->rx_AckErr = pSib->stats.AckRcvFailures;
    pInfo->rx_CrcErr = pSib->stats.RcvCrcErrors + pSib->stats.RcvDecryptCrcErrors;
    pInfo->rx_DiscardErr = pSib->stats.RxDiscardFrames;
    pInfo->rx_DmaErr = pSib->stats.RcvDmaOverrunErrors;
    pInfo->rx_DupFrameErr = pSib->stats.MultipleRxDuplicates;
    pInfo->rx_FcsErr = pSib->stats.FcsFailCnt;
    pInfo->rx_PhyErr = pSib->stats.RcvPhyErrors;
    pInfo->rx_RtsErr = pSib->stats.RtsFailCnt;
    pInfo->rx_WepXErr = pSib->stats.RcvWEPExcludedCount;
    idx = pSib->stats.RcvRate;
    if (apCfgFreqSpecGet() == MODE_SELECT_11B) {
        pRate = dataRate11bTbl[idx & 3];
    } else if (apCfgFreqSpecGet() == MODE_SELECT_TURBO) {
        pRate = dataRate2xTbl[idx & 7];
    } else {
        pRate = dataRateTbl[idx & 7];
    }
    strcpy (pInfo->rx_Rate, pRate);
    pInfo->rx_Ssl = pSib->stats.RcvSignalStrength;

    pInfo->tx_Cnt = pSib->stats.GoodTransmits;
    pInfo->tx_Data = pSib->stats.TxDataFrames;
    pInfo->tx_Mcast = pSib->stats.TxMulticastFrames;
    pInfo->tx_Mgmt = pSib->stats.TxMgmtFrames;
    pInfo->tx_Ctrl = pSib->stats.TxCtrlFrames;
    pInfo->tx_Errs = pSib->stats.TransmitErrors;
    pInfo->tx_DiscardErr = pSib->stats.TxDiscardFrames;
    pInfo->tx_DmaErr = pSib->stats.TxDmaUnderrun;
    pInfo->tx_XretryErr = pSib->stats.TxExcessiveRetries;
    if (pdevInfo->staConfig.rateCtrlEnable) {
        idx = pSib->txRateCtrl.sf.rateIdx;
        if (apCfgFreqSpecGet() == MODE_SELECT_11B) {
            pRate = rcRate11bTbl[idx];
        } else if (apCfgFreqSpecGet() == MODE_SELECT_TURBO) {
            pRate = rcRate2xTbl[idx];
        } else {
            pRate = rcRateTbl[idx];
        }
        strcpy (pInfo->tx_sfRate, pRate);

        idx = pSib->txRateCtrl.lf.rateIdx;
        if (apCfgFreqSpecGet () == MODE_SELECT_11B) {
            pRate = rcRate11bTbl[idx];
        } else if (apCfgFreqSpecGet() == MODE_SELECT_TURBO) {
            pRate = rcRate2xTbl[idx];
        } else {
            pRate = rcRateTbl[idx];
        }

        strcpy (pInfo->tx_lfRate, pRate);
    } else {
        idx = pdevInfo->localSta->currentOperationalRateIndex;
        if (apCfgFreqSpecGet() == MODE_SELECT_11B) {
            pRate = dataRate11bTbl[idx];
        } else if (apCfgFreqSpecGet() == MODE_SELECT_TURBO) {
            pRate = dataRate2xTbl[idx];
        } else {
            pRate = dataRateTbl[idx];
        }
        strcpy (pInfo->tx_sfRate, pRate);
        strcpy (pInfo->tx_lfRate, pRate);
    }
    pInfo->tx_Ssl = pSib->stats.CurrentAckSignalStrenth;
}


SCB *dot1xSessionGet (SIB_ENTRY *);
int dot1xReAuthPeriodGet ();
extern char *dot1xStateStr[];

LOCAL SIB_ENTRY *findMacAddrSib(char *addr)
{
    int         cnt, idx;
    SIB_ENTRY   *pSib;          /* a entry from SIB Table */
    char        macaddr[6];

    strtoWlanMAC(addr, (WLAN_MACADDR *)&macaddr);

    A_SIB_TAB_LOCK(pSibTable);
    cnt = pSibTable->currentEntryNum;
    if (cnt)
    {
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--)
        {
            pSib = &pSibTable->sta[idx];
            if (bcmp((char *)pSib->macAddr.octets, (char *)&macaddr, sizeof(macaddr)) == 0)
            {
                if (pSib->used & SIB_VALID)
                {
                    A_SIB_TAB_UNLOCK(pSibTable);
                    return pSib;
                }
                break;
            }
        }
    }
    A_SIB_TAB_UNLOCK(pSibTable);
    return (SIB_ENTRY *)NULL;
}


void staDot1xInfoGet (char *pMacAddr, rcw_Dot1x *pInfo)
{
    SCB         *s;
    SIB_ENTRY   *pSib;

    bzero ((char *)pInfo, sizeof(rcw_Dot1x));
    pInfo->reAuthPeriod = dot1xReAuthPeriodGet();

    pSib = findMacAddrSib(pMacAddr);
    if (!pSib) return;

    s = dot1xSessionGet (pSib);
    if (s == NULL)
        return;

    strcpy (pInfo->identity, s->radiusInfo.identity);
    strcpy (pInfo->authState, dot1xStateStr[s->stateAuth.state]);
    strcpy (pInfo->asState, dot1xStateStr[s->stateAS.state]);
    pInfo->EntersConnecting =
        s->AuthCounter.authEntersConnecting;
    pInfo->EapLogoffsWhileConnecting =
        s->AuthCounter.authEapLogoffsWhileConnecting;
    pInfo->EntersAuthenticating =
        s->AuthCounter.authEntersAuthenticating;
    pInfo->SuccessesWhileAuthenticating =
        s->AuthCounter.authAuthSuccessesWhileAuthenticating;
    pInfo->TimeoutsWhileAuthenticating =
        s->AuthCounter.authAuthTimeoutsWhileAuthenticating;
    pInfo->FailWhileAuthenticating =
        s->AuthCounter.authAuthFailWhileAuthenticating;
    pInfo->FailWhileAuthenticating =
        s->AuthCounter.authAuthFailWhileAuthenticating;
    pInfo->ReauthsWhileAuthenticating =
        s->AuthCounter.authAuthReauthsWhileAuthenticating;
    pInfo->EapStartsWhileAuthenticating =
        s->AuthCounter.authAuthEapStartsWhileAuthenticating;
    pInfo->EapLogoffWhileAuthenticating =
        s->AuthCounter.authAuthEapLogoffWhileAuthenticating;
    pInfo->ReauthsWhileAuthenticated =
        s->AuthCounter.authAuthReauthsWhileAuthenticated;
    pInfo->EapStartsWhileAuthenticated =
        s->AuthCounter.authAuthEapStartsWhileAuthenticated;
    pInfo->EapLogoffWhileAuthenticated =
        s->AuthCounter.authAuthEapLogoffWhileAuthenticated;
    pInfo->Responses =
        s->AuthServerCounter.backendResponses;
    pInfo->AccessChallenges =
        s->AuthServerCounter.backendAccessChallenges;
    pInfo->OtherRequestsToSupplicant =
        s->AuthServerCounter.backendOtherRequestsToSupplicant;
    pInfo->NonNakResponsesFromSupplicant =
        s->AuthServerCounter.backendNonNakResponsesFromSupplicant;
    pInfo->AuthSuccesses =
        s->AuthServerCounter.backendAuthSuccesses;
    pInfo->AuthFails =
        s->AuthServerCounter.backendAuthFails;
}

#endif  /* INCLUDE_RCW */

/*
 *  Display AP association table
 */
void sibSmeShow(void)
{
    int         cnt, idx, j;
    SIB_ENTRY   *pSib;          /* a entry from SIB Table */
    SIB_ENTRY   *pTbl[MAX_STA_NUM];

    j = 0;
    A_SIB_TAB_LOCK(pSibTable);
    cnt = pSibTable->currentEntryNum;
    if (cnt)
    {
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--)
        {
            pSib = &pSibTable->sta[idx];
            if (pSib->used & SIB_VALID)
            {
                pTbl[j++] = pSib;
                cnt--;
                if (cnt == 0)
                    break;
            }
        }
    }
    A_SIB_TAB_UNLOCK(pSibTable);

    if (j > 1)
        qsort(&pTbl[1],
              j-1,
              sizeof(SIB_ENTRY *),
              (void *)&macaddrcmp);

    cliWrite("STA\tMAC Address\t\tState\n");
    for (idx = 0; idx < j; idx++)
    {
        pSib = pTbl[idx];

        if (pSib->serviceType == WLAN_AP_SERVICE)
            cliWrite("%4s","AP");
        else
            cliWrite("%4d", pSib->assocId & 0x3fff);

        cliWrite("\t%02X:%02X:%02X:%02X:%02X:%02X\t%s\n",
                 pSib->macAddr.octets[0], pSib->macAddr.octets[1],
                 pSib->macAddr.octets[2], pSib->macAddr.octets[3],
                 pSib->macAddr.octets[4], pSib->macAddr.octets[5],
                 sibStateStrGet(pSib));
    }
    cliWrite("\n");
}

void sibStationShow(int staId)
{
    int         idx;
    int         cnt;
    SIB_ENTRY   sibCopy;        /* Copy of SIB entry */
    SIB_ENTRY   *pSib;          /* a entry from SIB Table */
#if defined(DEBUG)
    int         i;
#endif

    if (pSibTable == NULL) {
        return;
    }

    cnt = pSibTable->currentEntryNum;
    if (cnt == 0) {
        return;
    }
    for (idx = MAX_STA_NUM - 1 ; idx >= 0; idx--) {
        pSib = &pSibTable->sta[idx];
        A_SIB_TAB_LOCK(pSibTable);
        A_BCOPY((char *)pSib, (char *)&sibCopy, sizeof(SIB_ENTRY));
        A_SIB_TAB_UNLOCK(pSibTable);
        pSib = (SIB_ENTRY *)&sibCopy;
        if (pSib->used & SIB_VALID) {
            if (staId != -1 && staId != 0) {
                if (staId != (pSib->assocId & 0x3fff)) {
                    continue;
                }
            }
            if (pSib->serviceType == WLAN_STA_SERVICE) {
                cliWrite("MacAddr: %02x:%02x:%02x:%02x:%02x:%02x, State: %s, AID: %d\n",
                         pSib->macAddr.octets[0], pSib->macAddr.octets[1],
                         pSib->macAddr.octets[2], pSib->macAddr.octets[3],
                         pSib->macAddr.octets[4], pSib->macAddr.octets[5],
                         sibStateStrGet(pSib),
                         pSib->assocId & 0x3fff);
                cliWrite("Authentication Type: %s\n", pSib->authenAlgo == 0 ?
                         "Open System" : (pSib->authenAlgo == 1 ? "Shared Key" : "Unknown"));
                if (pSib->privacyInvoked) {
                    cliWrite("Encryption: %s",
                             pSib->privacyInvoked ? "on" : "off");
                    if (pSib->pPriv) {
                        if (pSib->pPriv->keyFlags & PRIV_DYNAMIC) {
                            cliWrite(" dynamic");
                        } else if (pSib->pPriv->keyFlags & PRIV_UNIQUEKEY) {
                            cliWrite(" unique");
                        } else {
                            cliWrite(" shared[%d]",
                                     (pSib->pPriv->keyFlags & PRIV_FRAMEKEY) + 1);
                        }

                        if (pSib->pPriv->keyType == PRIV_KEY_TYPE_AES) {
                            cliWrite(" AES");
                        } else {
                            if (pSib->pPriv->keyType == PRIV_KEY_TYPE_WEP) {
                                cliWrite(" WEP");
                            }
                        }

#if defined(DEBUG)      /* do not show the key data normally */
                        cliWrite(" %db key: ", pSib->pPriv->keyLength);
                        for (i = 0; i < (pSib->pPriv->keyLength/8); i++) {
                            cliWrite("%02x", pSib->pPriv->keyVal[i]);
                        }
#endif
                    }
                    if (pSib->keyIndex == KEYINDEX_INVALID) {
                        cliWrite(" hwIndex none\n");
                    } else {
                        cliWrite(" hwIndex %d\n", pSib->keyIndex);
                    }
                    sibCipherShow(pSib);
                    cliWrite("\n");
                } else {
                    cliWrite("Encryption: OFF\n");
                }

#ifdef VXWORKS
                cliWrite("VLAN TAG: 0x%04x\n", pSib->vlanTag);
#endif

                cliWrite("Power Save Mode: %s\n", pSib->psMode ? "ON" : "OFF");
                if (pSib->stats.RcvDecryptCrcErrors)
                    cliWrite("Encryption decryption Errors: %d - key may be misconfigured.\n",
                             pSib->stats.RcvDecryptCrcErrors);
                cliWrite("Rx Data Rate: %s, RxSignalStrength: %d, AckSignalStrength: %d\n",
                         apCfgFreqSpecGet() == MODE_SELECT_11B ?
                         dataRate11bTbl[pSib->stats.RcvRate & 0x3] :
                         ((apCfgFreqSpecGet() == MODE_SELECT_TURBO) ?
                          dataRate2xTbl[pSib->stats.RcvRate & 0x7] : dataRateTbl[pSib->stats.RcvRate & 0x7]),
                         pSib->stats.RcvSignalStrength,
                         pSib->stats.CurrentAckSignalStrenth);

                cliWrite("              MSDU       Data    Mcast     Mgmt     Ctrl    Errors\n");
                cliWrite("Rx\t%10u %10d %8u %8u %8u %9u\n",
                         pSib->stats.GoodReceives,
                         pSib->stats.RxDataFrames,
                         pSib->stats.MulticastReceives,
                         pSib->stats.RxMgmtFrames,
                         pSib->stats.RxCtrlFrames,
                         pSib->stats.ReceiveErrors);
                cliWrite("Tx\t%10u %10u %8u %8u %8u %9u\n",
                         pSib->stats.GoodTransmits,
                         pSib->stats.TxDataFrames,
                         pSib->stats.TxMulticastFrames,
                         pSib->stats.TxMgmtFrames,
                         pSib->stats.TxCtrlFrames,
                         pSib->stats.TransmitErrors);
                cliWrite("\n");
            }
        }
    }
}

void sibCipherShow(SIB_ENTRY *pSib)
{
    int i;

    uiPrintf("Ciphers:");
    if (pSib->serviceType == WLAN_STA_SERVICE) {
        if (VALID_UCSE_ELEMENT(&pSib->workingUcseSet)) {
            if (pSib->workingUcseSet.uCiphers[0] == CSE_NULL) {
                uiPrintf(" NULL");
            } else {
                if (pSib->workingUcseSet.uCiphers[0] == CSE_BASIC_WEP) {
                    uiPrintf(" WEP");
                } else {
                    if ((pSib->workingUcseSet.uCiphers[0] == CSE_AES_OCB) ||
                        (pSib->workingUcseSet.uCiphers[0] == CSE_AES_OCB_REVA))
                    {
                        uiPrintf(" AES");
                    } else {
                        uiPrintf(" UNKNOWN");
                    }
                }
            }
        } else {
            uiPrintf(" NONE");
        }
        uiPrintf("(unicast),");

        if (VALID_MCSE_ELEMENT(&pSib->mcseSet)) {
            if (pSib->mcseSet.mCipher[0] == CSE_NULL) {
                uiPrintf(" NULL");
            } else {
                if (pSib->mcseSet.mCipher[0] == CSE_BASIC_WEP) {
                    uiPrintf(" WEP");
                } else {
                    if ((pSib->mcseSet.mCipher[0] == CSE_AES_OCB) ||
                        (pSib->mcseSet.mCipher[0] == CSE_AES_OCB_REVA))
                    {
                        uiPrintf(" AES");
                    } else {
                        uiPrintf(" UNKNOWN");
                    }
                }
            }
        } else {
            uiPrintf(" NONE");
        }
        uiPrintf("(multicast), ");

        if (VALID_UCSE_ELEMENT(&pSib->ucseSet)) {
            uiPrintf("supports");
            for (i=0; i < pSib->ucseSet.length/4; i++) {
                if (pSib->ucseSet.uCiphers[i] == CSE_NULL) {
                    uiPrintf(" NULL");
                } else {
                    if (pSib->ucseSet.uCiphers[i] == CSE_BASIC_WEP) {
                        uiPrintf(" WEP");
                    } else {
                        if ((pSib->ucseSet.uCiphers[i] == CSE_AES_OCB) ||
                            (pSib->ucseSet.uCiphers[i] == CSE_AES_OCB_REVA))
                        {
                            if (i == 0) {
                                uiPrintf(" AES");
                            }
                        } else {
                            uiPrintf(" UNKNOWN");
                        }
                    }
                }
            }
        } else {
            uiPrintf(" NONE");
        }

    } else {
        if (VALID_UCSE_ELEMENT(&pSib->ucseSet)) {
            for (i=0; i < pSib->ucseSet.length/4; i++) {
                if (pSib->ucseSet.uCiphers[i] == CSE_NULL) {
                    uiPrintf(" NULL");
                } else {
                    if (pSib->ucseSet.uCiphers[i] == CSE_BASIC_WEP) {
                        uiPrintf(" WEP");
                    } else {
                        if ((pSib->ucseSet.uCiphers[i] == CSE_AES_OCB) ||
                            (pSib->ucseSet.uCiphers[i] == CSE_AES_OCB_REVA))
                        {
                            if (i == 0) {
                                uiPrintf(" AES");
                            }
                        } else {
                            uiPrintf(" UNKNOWN");
                        }
                    }
                }
            }

        } else {
            uiPrintf(" NONE");
        }
    }
}

#if defined(DEBUG)

WLAN_MACADDR  *pStaMacAddr;


SIB_ENTRY *addsta(int cnt, int serialize, int verbose)
{
    WLAN_DEV_INFO *pdevInfo;
    SIB_ENTRY     *pSib;
    WLAN_MACADDR  macAddr;
    int ii = 0;
    int seed = 1;

    pdevInfo = pwlanDevInfo;
    srand(halGetRandomSeed(pdevInfo));

    if (cnt == 0) cnt = 1;

    while (cnt--) {
        macAddr.words[0] = serialize ? 0x0011 : rand() & 0xfeff;
        macAddr.words[1] = serialize ? 0x2233 : rand();
        macAddr.words[2] = serialize ? seed++ : rand();

        pSib = sibEntryAlloc(pdevInfo->localSta, pdevInfo->bssDescr, &macAddr);
        if (!pSib) {
            printf("Can't allocate SIB\n");
            return pSib;
        }
        pStaMacAddr     = &pSib->macAddr;
        pSib->assocId   = mlmeAssocIdNew() & 0x3fff;
        pSib->staState |= (STATE_AUTH | STATE_ASSOC);
        if (sibEntryAdd(pSib) == OK) {
            if (verbose)
            printf("sib[%d]: %p, %d added\n", ++ii, pSib, pSib->assocId & 0x3fff);
        } else {
            printf("Can't add SIB[%d]: %d added\n", ++ii, pSib->assocId & 0x3fff);
        }
    }
    return pSib;
}

void sibShow(int staId)
{
    int         i, idx;
    int         cnt;
    SIB_ENTRY   sibCopy;        /* Copy of SIB entry */
    SIB_ENTRY   *pSib;          /* a entry from SIB Table */
    WLAN_DEV_INFO *pdevInfo;    /* device information structure */

    pdevInfo = pSibTable->pdevInfo;

    cnt = pSibTable->currentEntryNum;
    if (cnt) {
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--) {
            pSib = &pSibTable->sta[idx];
            A_SIB_TAB_LOCK(pSibTable);
            A_BCOPY((char *)pSib, (char *)&sibCopy, sizeof(SIB_ENTRY));
            A_SIB_TAB_UNLOCK(pSibTable);
            pSib = (SIB_ENTRY *)&sibCopy;
            if (pSib->used & SIB_VALID) {
                if (staId != -1 && staId != 0) {
                    if (staId != (pSib->assocId & 0x3fff)) {
                        continue;
                    }
                }
            }
            else
                continue;

            if (pSib->serviceType == WLAN_STA_SERVICE) {
                uiPrintf("AID %d, ", pSib->assocId & 0x3fff);
            }
            printMacAddress(pSib->macAddr);
            uiPrintf("%4s, idx: %d\n",
                     (pSib->serviceType == WLAN_STA_SERVICE) ? "STA" : "AP",
                      idx);
            displayStates(pSib->staState, pSib->transState);
            uiPrintf("Current Ant: %d, Rx Ant: %d\n",
                      pSib->antTx + 1, pSib->stats.RxAntenna + 1);
            uiPrintf("Ant 1: %d, Ant 2: %d, Switch: %d, Delta: %d\n",
                      pSib->stats.AntCnt[0], pSib->stats.AntCnt[1],
                      pSib->stats.AntSwCnt,
                      pSib->stats.AntCnt[0] - pSib->stats.AntCnt[1]);
            if (pSib->serviceType != WLAN_STA_SERVICE) {
                uiPrintf("Channel %d\n", pSib->Channel);
                uiPrintf("Turbo: %s\n", (apCfgFreqSpecGet() == MODE_SELECT_TURBO) ? "ON" : "OFF");
                uiPrintf("Encryption: %s\n", pdevInfo->staConfig.privacyInvoked ? "ON" : "OFF");
                sibCipherShow(pSib);
                uiPrintf("\n");

            } else {
                int sfidx, lfidx;
                uiPrintf("Start: %d, Last active: %d\n",
                          pSib->staLifeStartTime/1000,
                          pSib->staLastActivityTime/1000);
                sfidx = pSib->txRateCtrl.sf.rateIdx;
                lfidx = pSib->txRateCtrl.lf.rateIdx;
                uiPrintf("sfRate: %s, lfRate: %s\n",
                         apCfgFreqSpecGet() == MODE_SELECT_11B ?
                         dataRate11bTbl[sfidx] :
                         ((apCfgFreqSpecGet() == MODE_SELECT_TURBO) ?
                          dataRate2xTbl[sfidx] : dataRateTbl[sfidx]),
                         apCfgFreqSpecGet() == MODE_SELECT_11B ?
                         dataRate11bTbl[lfidx] :
                         ((apCfgFreqSpecGet() == MODE_SELECT_TURBO) ?
                          dataRate2xTbl[lfidx] : dataRateTbl[lfidx]));
                uiPrintf("Authentication Type:\t %s\n", pSib->authenAlgo == 0 ?
                          "Open System" : (pSib->authenAlgo == 1 ? "Shared Key" : "Unknown"));
                uiPrintf("Encryption: %s slot ",
                          pSib->privacyInvoked ? "ON" : "OFF");
                if (pSib->keyIndex == KEYINDEX_INVALID)
                    uiPrintf("none");
                else
                    uiPrintf("%d", pSib->keyIndex);
                if (pSib->pPriv) {
                    if (pSib->pPriv->keyFlags & PRIV_UNIQUEKEY) {
                        uiPrintf(" unique");
                    } else {
                        uiPrintf(" shared[%d]",
                                  pSib->pPriv->keyFlags & PRIV_FRAMEKEY);
                    }

                    if (pSib->pPriv->keyType == PRIV_KEY_TYPE_AES) {
                        uiPrintf(" AES");
                    } else {
                        if (pSib->pPriv->keyType == PRIV_KEY_TYPE_WEP) {
                            uiPrintf(" WEP");
                        }
                    }

                    uiPrintf(" %db key: ", pSib->pPriv->keyLength);
                    for (i = 0; i < (pSib->pPriv->keyLength/8); i++) {
                        uiPrintf("%02x", pSib->pPriv->keyVal[i]);
                    }
                }
                uiPrintf("\n");

                sibCipherShow(pSib);
                uiPrintf("\n");

#ifdef VXWORKS
                uiPrintf("VLAN TAG: 0x%04x\n", pSib->vlanTag);
#endif
            }

            if (pSib->stats.Authentications || pSib->stats.DeAuthentications) {
                uiPrintf("Authentications: \t%5d  Deauthentications: %5d",
                          pSib->stats.Authentications,
                          pSib->stats.DeAuthentications);
                if (pSib->stats.DeAuthReasonCode) {
                    uiPrintf("  Reason Code: %d\n",
                              pSib->stats.DeAuthReasonCode);
                } else {
                    uiPrintf("\n");
                }
            }
            if (pSib->stats.Associations || pSib->stats.DisAssociations) {
                uiPrintf("Associations:    \t%5d  Disassociations:   %5d",
                         pSib->stats.Associations,
                         pSib->stats.DisAssociations);
                if (pSib->stats.DisAssocReasonCode) {
                    uiPrintf("  Reason Code: %d\n",
                              pSib->stats.DisAssocReasonCode);
                } else {
                    uiPrintf("\n");
                }
            }
            uiPrintf("Power Save Mode: %s\n", pSib->psMode ? "ON" : "OFF");
            if (pSib->stats.RcvSignalStrength)
                uiPrintf("RcvRate:\t\t %10s\nAckSignalStrength:\t %10d\t RcvSignalStrength:\t %10d\n",
                         apCfgFreqSpecGet() == MODE_SELECT_11B ?
                         dataRate11bTbl[pSib->stats.RcvRate & 0x3] :
                         ((apCfgFreqSpecGet() == MODE_SELECT_TURBO) ?
                          dataRate2xTbl[pSib->stats.RcvRate & 0x7] : dataRateTbl[pSib->stats.RcvRate & 0x7]),
                         pSib->stats.CurrentAckSignalStrenth,
                         pSib->stats.RcvSignalStrength);
            uiPrintf("Tx MSDUs:\t\t %10d\t Rx MSDUs:\t\t %10d\n",
                     pSib->stats.GoodTransmits,
                     pSib->stats.GoodReceives);
            if (pSib->stats.TxMulticastFrames || pSib->stats.MulticastReceives)
                uiPrintf("TxMulticasts:\t\t %10d\t MulticastReceives:\t %10d\n",
                         pSib->stats.TxMulticastFrames,
                         pSib->stats.MulticastReceives);
            if (pSib->stats.TxCtrlFrames ||  pSib->stats.RxCtrlFrames)
                uiPrintf("TxCtrlFrames:\t\t %10d\t RxCtrlFrames:\t\t %10d\n",
                         pSib->stats.TxCtrlFrames, pSib->stats.RxCtrlFrames);
            if (pSib->stats.TxDataFrames ||  pSib->stats.RxDataFrames)
                uiPrintf("TxDataFrames:\t\t %10d\t RxDataFrames:\t\t %10d\n",
                         pSib->stats.TxDataFrames, pSib->stats.RxDataFrames);
            if (pSib->stats.TxMgmtFrames || pSib->stats.RxMgmtFrames)
                uiPrintf("TxMgmtFrames:\t\t %10d\t RxMgmtFrames:\t\t %10d\n",
                         pSib->stats.TxMgmtFrames, pSib->stats.RxMgmtFrames);
            if (pSib->stats.TransmitErrors || pSib->stats.ReceiveErrors)
                uiPrintf("TxErrors:\t\t %10d\t RxErrors:\t\t %10d\n",
                         pSib->stats.TransmitErrors,
                         pSib->stats.ReceiveErrors);
            if (pSib->stats.RcvPhyErrors)
                uiPrintf("RxPhyErrors:\t\t %10d\n",
                         pSib->stats.RcvPhyErrors);
            if (pSib->stats.TxExcessiveRetries || pSib->stats.RcvCrcErrors)
                uiPrintf("TxExcessiveRetries:\t %10d\t RxCrcErrors:\t\t %10d\n",
                         pSib->stats.TxExcessiveRetries,
                         pSib->stats.RcvCrcErrors);
            if (pSib->stats.RcvDecryptCrcErrors || pSib->stats.RcvKeyCacheMisses)
                uiPrintf("\t\t\t\t\t RcvDecryptCrcErrors:\t %10d\n"
                         "\t\t\t\t\t RxKeyCacheMisses:\t %10d\n",
                          pSib->stats.RcvDecryptCrcErrors,
                          pSib->stats.RcvKeyCacheMisses);
            if (pSib->numTxPending || pSib->swRetry.numFilteredPending)
                uiPrintf("Pending transmit frames: %10d\t Pending Filtered frames: %9d\n",
                         pSib->numTxPending, pSib->swRetry.numFilteredPending);
            if (pSib->stats.swRetryTotalCnt || pSib->stats.swRetryMaxRetriesExceeded)
                uiPrintf("SW Retries:\t\t %10d\t Exceeded SW Retry limits: %8d\n",
                         pSib->stats.swRetryTotalCnt,
                         pSib->stats.swRetryMaxRetriesExceeded
                         );
            if (pSib->stats.TxFiltered || pSib->stats.MultipleRxDuplicates)
                uiPrintf("TxFiltered:\t\t %10d\t RxDuplicateFrames:\t %10d\n",
                         pSib->stats.TxFiltered,
                         pSib->stats.MultipleRxDuplicates);
            if (pSib->stats.TxDiscardFrames || pSib->stats.RxDiscardFrames)
                uiPrintf("TxDiscardFrames:\t %10d\t RxDiscardFrames:\t %10d\n",
                         pSib->stats.TxDiscardFrames,
                         pSib->stats.RxDiscardFrames);
            if (pSib->stats.RtsFailCnt || pSib->stats.FcsFailCnt)
                uiPrintf("RtsFailCnt:\t\t %10d\t FcsFailCnt:\t\t %10d\n",
                         pSib->stats.RtsFailCnt,
                         pSib->stats.FcsFailCnt);
            if (pSib->stats.TotalRetries || pSib->stats.AckRcvFailures)
                uiPrintf("TxTotalRetries:\t\t %10d\t AckRcvFailures:\t %10d\n",
                         pSib->stats.TotalRetries,
                         pSib->stats.AckRcvFailures);
            if (pSib->stats.RcvWEPExcludedCount || pSib->stats.swRetryHungCount)
                uiPrintf("SwRetryHungCount:\t %10d\t RcvWEPExcludedCount:\t %10d\n",
                         pSib->stats.swRetryHungCount,
                         pSib->stats.RcvWEPExcludedCount);

            for (i = 1; i < 16; i++) {
                if (pSib->stats.RetryBins[i] || pSib->stats.shortFrameRetryBins[i]) {
                    if (i == 1)
                        uiPrintf("%2d   TxShortFrameRetry:\t %10d\t %2d   TxLongFrameRetry:\t %10d\n",
                                 i, pSib->stats.shortFrameRetryBins[i],
                                 i, pSib->stats.RetryBins[i]);
                    else
                        uiPrintf("%2d TxShortFrameRetries:\t %10d\t %2d TxLongFrameRetries:\t %10d\n",
                                 i, pSib->stats.shortFrameRetryBins[i],
                                  i, pSib->stats.RetryBins[i]);
                }
            }
            /* power save related statistics */
            if (pSib->serviceType == WLAN_STA_SERVICE) {
                if (pSib->stats.psRequestFrames || pSib->stats.psSavedFrames)
                    uiPrintf("psRequestFrames:\t %10d\t psSavedFrames:\t\t %10d\n",
                             pSib->stats.psRequestFrames,
                             pSib->stats.psSavedFrames);
                if (pSib->stats.psDiscardFrames || pSib->stats.psAgedFrames)
                    uiPrintf("psDiscardFrames:\t %10d\t psAgedFrames:\t\t %10d\n",
                             pSib->stats.psDiscardFrames,
                             pSib->stats.psAgedFrames);
            } else {
                if (pSib->stats.psMulticastFrames)
                    uiPrintf("psMulticastFrames:\t %10d\n",
                             pSib->stats.psMulticastFrames);
            }
            uiPrintf("\n");

            cnt--;
            if (cnt == 0)
                break;
        }
    }
}

void
bssShow(int unit)
{
  WLAN_DEV_INFO *pdevInfo;      /* device information structure */
  int            i, len;

  pdevInfo = pSibTable->pdevInfo;

  uiPrintf("SSID: %s\n",
           pdevInfo->bssDescr->ssid.ssid);
  uiPrintf("BSS type: %d\n",
           pdevInfo->bssDescr->bsstype);
  uiPrintf("Channel: %d\n",
           pdevInfo->bssDescr->pChannel ?
           pdevInfo->bssDescr->pChannel->channel : 0);
  uiPrintf("Beacon Interval: %d\n",
           pdevInfo->bssDescr->beaconInterval);
  uiPrintf("DTIM Interval: %d\n",
           pdevInfo->bssDescr->DTIMPeriod);
  len = pdevInfo->bssDescr->bssRateSet.length;
  for (i = 0; i < len; i++) {
      uiPrintf("rate[%d]: %d\n", i, pdevInfo->bssDescr->bssRateSet.rates[i]);
  }
}

void sibAgingShow()
{
    SIB_ENTRY *pSib;
    int         idx;
    int         cnt;
    SIB_ENTRY   sibCopy;        /* Copy of SIB entry */
    WLAN_DEV_INFO *pdevInfo;    /* device information structure */
    QUEUE_DETAILS *pQueue;
    A_UINT32       lasttime;


    pdevInfo = pSibTable->pdevInfo;

    cnt = pSibTable->currentEntryNum;
    if (cnt) {
        for (idx = MAX_STA_NUM - 1; idx >= 0; idx--) {
            pSib = &pSibTable->sta[idx];
            A_SIB_TAB_LOCK(pSibTable);
            A_BCOPY((char *)pSib, (char *)&sibCopy, sizeof(SIB_ENTRY));
            A_SIB_TAB_UNLOCK(pSibTable);
            pSib = (SIB_ENTRY *)&sibCopy;
            if (!(pSib->used & SIB_VALID))
                continue;

            printMacAddress(pSib->macAddr);
            printf (" AID: %d\n", pSib->assocId & 0x3fff);
            lasttime = pSib->staLastActivityTime;
            pQueue   = &pSib->txPSQueue;
            printf ("QueueHead: %p, qFrameCount: %d, numPSTxPending: %d\nnumTxPending: %d, swRetry.numFilteredPending: %d, swRetry.qFrames: %d\n",
                    pQueue->pDescQueueHead,
                    pQueue->qFrameCount,
                    pSib->numPSTxPending,
                    pSib->numTxPending,
                    pSib->swRetry.numFilteredPending,
                    pSib->swRetry.qFrames);
            printf("time: %d\n", A_MS_TICKGET() - lasttime);

            cnt--;
            if (cnt == 0)
                break;
        }
    }
}


void ocrap (int cnt)
{
    SIB_ENTRY *pSib;

    pSib = addsta (1,1,0);
    if (cnt == 0)
        cnt = 256;
    while (cnt--) {
        wlanMlmeDeauthRequest(pwlanDevInfo, &pSib->macAddr, REASON_INACTIVE);
    }
}

void
crapFrameHdrFill(WLAN_DEV_INFO *pdevInfo, WLAN_MGT_MAC_HEADER *pFrame,
                WLAN_MACADDR  *destAddr, int subtype)
{
    /* Fill in Frame Control fields */
    pFrame->frameControl.fType    = FRAME_DATA;
    pFrame->frameControl.fSubtype = (A_UINT16) subtype;
    pFrame->frameControl.FromDS   = 1;
    pFrame->frameControl.ToDS     = 0;  /* and ToDS as 0                     */
    pFrame->frameControl.protoVer = PROTOCOL_VER_0;
    pFrame->frameControl.pwrMgt   = 0;  /* default value */
    pFrame->frameControl.wep      = 0;  /* Most management frames not envrypted */
    pFrame->frameControl.moreData = 0;  /* default 0 */
    pFrame->frameControl.moreFrag = 0;  /* default 0 */

    /* BSSID */
    /* If it's a probe request, calling program should replace the value
     * of bssId anyway.
     */
    A_MACADDR_COPY(&pdevInfo->localSta->bssId, &pFrame->bssId);

    /* Source address */
    A_MACADDR_COPY(&pdevInfo->localSta->macAddr, &pFrame->srcAddr);

    /* Destination address */
    A_MACADDR_COPY(destAddr, &pFrame->destAddr);

    /* duration, seq control fields are not initialized here */
    pFrame->sequenceControl.fragNo = 0;
}

int sendcrapdataframe(WLAN_DEV_INFO *pdevInfo,
                      WLAN_MACADDR  *pPeerSTAAddress)
{
    SIB_ENTRY         *pSib;
    A_UINT16          frameLen;
    ATHEROS_DESC      *pDesc;
    WLAN_FRAME_DEAUTH *pFrame;    /* Pointer to deauth req frame */

    // It's ok to not find the destination SIB.
    pSib = sibEntryFind(pPeerSTAAddress);

    /* Compute frame length. */
    frameLen = sizeof(1400);

    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.deAuth;
    crapFrameHdrFill(pdevInfo, &pFrame->macHeader, pPeerSTAAddress, 0);

    /* displayMgtFrame(&pFrame->macHeader, frameLen); */
    pDesc->pDestSibEntry = pSib;

    powerFrameTransmit(pdevInfo, pDesc); // return status doesn't matter

    return MLME_SUCCESS;
}

void ocrapdata (int cnt)
{
    SIB_ENTRY *pSib;

    pSib = addsta (1,1,0);
    if (cnt == 0)
        cnt = 256;
    pSib->staState |= (STATE_AUTH | STATE_ASSOC);

    while (cnt--) {
        sendcrapdataframe(pwlanDevInfo, &pSib->macAddr);
    }
}

void webset (SIB_ENTRY *pSib)
{
    pSib->privacyInvoked = 1;
    pSib->ucseSet.elementID = ELE_UCSE_SET;
    pSib->ucseSet.length = 4*2;
    pSib->ucseSet.uCiphers[0] = CSE_AES_OCB;
    pSib->ucseSet.uCiphers[1] = CSE_BASIC_WEP;    

    pSib->mcseSet.elementID = ELE_MCSE_SET;
    pSib->mcseSet.length = ELE_MCSE_SIZE;
    pSib->mcseSet.mCipher[0] = CSE_AES_OCB;
}

    


#endif /* DEBUG */
#endif /* VXWORKS */
