/*
 *  Copyright (C) 2001-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

/* sysTffs.c - ar5311 system-dependent TrueFFS library */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/sysTffs.c#1 $"

/* Copyright 1984-1997 Wind River Systems, Inc. */
#include "copyright_wrs.h"

/* FAT-FTL Lite Software Development Kit
 * Copyright (C) M-Systems Ltd. 1995-1997       */

/*
 * DESCRIPTION
 * This library provides board-specific hardware access routines for TrueFFS.
 * In effect, these routines comprise the socket component driver (or drivers)
 * for your flash device hardware.  At socket registration time, TrueFFS stores
 * pointers to the functions of this socket component driver in an 'FLSocket'
 * structure.  When TrueFFS needs to access the flash device, it uses these
 * functions.
 *
 * Because this file is, for the most part, a device driver that exports its
 * functionality by registering function pointers with TrueFFS, very few of the
 * functions defined here are externally callable.  For the record, these
 * external functions are flFitInSocketWindow() and flDelayLoop().  You should
 * never have any need to call these functions.
 *
 * However, one of the most import functions defined in this file is neither
 * referenced in an 'FLSocket' structure, nor is it externally callable.  This
 * function is sysTffsInit().  TrueFFS calls this function at initialization
 * time to register socket component drivers for all the flash devices attached
 * to your target.  It is this call to sysTffs() that results in assigning
 * drive numbers to the flash devices on your target hardware.  Drive numbers
 * are assigned by the order in which the socket component drivers are
 * registered.  The first to be registered is drive 0, the second is drive 1,
 * and so on up to 4.  As shipped, TrueFFS supports up to five flash drives.
 *
 * After registering socket component drivers for a flash device, you may
 * format the flash medium even though there is not yet a block device driver
 * associated with the flash (see the reference entry for the tffsDevCreate()
 * routine).  To format the flash medium for use with TrueFFS,
 * call tffsDevFormat() or, for some BSPs, sysTffsFormat().
 *
 * The sysTffsFormat() routine is an optional but BSP-specific externally
 * callable helper function.  Internally, it calls tffsDevFormat() with a
 * pointer to a 'FormatParams' structure initialized to values that leave a
 * space on the flash device for a boot image. This space is outside the
 * region managed by TrueFFS.  This special region is necessary for boot
 * images because the normal translation and wear-leveling services of TrueFFS
 * are incompatible with the needs of the boot program and the boot image it
 * relies upon.  To write a boot image (or any other data) into this area,
 * use tffsBootImagePut().
 *
 * Finally, this file also contains define statements for symbolic constants
 * that determine which MTDs, translation layer modules, and other utilities
 * are ultimately included in TrueFFS.  These defines are as follows:
 */

#include "vxWorks.h"
#include "stdio.h"
#include "config.h"

#ifdef INCLUDE_TFFS
#include "tffs/flsocket.h"
#include "ar5311reg.h"
#include "ioLib.h"
#include "dosFsLib.h"

/*
 * The ar5311 hardware supports 3 flash banks:
 *   AR5311_FLASH0      boot flash, required, aliased to 0x1fc00000.
 *   AR5311_FLASH1      not supported by software currently.
 *   AR5311_FLASH2      not supported by software currently.
 */
#define FLASH_BASE_LEN  0x10000000              /* 16MB window */

static FLBoolean        rfaCardDetected(FLSocket vol);
static void             rfaVccStub(FLSocket vol);
#ifdef SOCKET_12_VOLTS
static FLStatus         rfaVppStub(FLSocket vol);
#endif
static FLBoolean        rfaGetAndClearCardChangeIndicator(FLSocket vol);
static FLBoolean        rfaWriteProtected(FLSocket vol);
static void             rfaSetWindow(FLSocket vol);
static void             rfaSetMappingContext(FLSocket vol, unsigned page);
static FLStatus         rfaSocketInit(FLSocket vol);

#ifndef DOC
#include "tffs/tffsConfig.c"
#endif

/*
 * sysTffsInit - initialize TFFS flash socket interface.
 */
void
sysTffsInit(void)
{
    FLSocket vol = flSocketOf(noOfDrives);

    if (noOfDrives >= DRIVES) {
        return;
    }

    tffsSocket[noOfDrives] = "RFA";
    noOfDrives++;

    vol.window.baseAddress = PHYS_TO_K1(AR5311_FLASH0) >> 12;
    vol.serialNo = 0;

    /* fill in function pointers */
    vol.cardDetected = rfaCardDetected;
    vol.VccOn = rfaVccStub;
    vol.VccOff = rfaVccStub;
#ifdef SOCKET_12_VOLTS
    vol.VppOn = rfaVppStub;
    vol.VppOff = rfaVccStub;
#endif
    vol.initSocket = rfaSocketInit;
    vol.setWindow = rfaSetWindow;
    vol.setMappingContext = rfaSetMappingContext;
    vol.getAndClearCardChangeIndicator = rfaGetAndClearCardChangeIndicator;
    vol.writeProtected = rfaWriteProtected;

    return;
}

char *
sysTffsVers(void)
{
    return tffsVersion;         /* for -Wall -Werror */
}

/*
 * rfaCardDetected - detect if a card is present (inserted).
 */
static FLBoolean
rfaCardDetected(FLSocket vol)
{
    return TRUE;                /* our flash is always there */
}

/* rfaVccStub
 * rfaVccOn - turn on Vcc (3.3/5 Volts).  Vcc must be known good on exit.
 * rfaVccOff - turn off Vcc (3.3/5 Volts)
 */
static void
rfaVccStub(FLSocket vol)
{
}

#ifdef SOCKET_12_VOLTS
/*
 * rfaVppStub - rfaVppOn, rfaVppOff.
 */
static FLStatus
rfaVppStub(FLSocket vol)
{
    return flOK;
}
#endif

/*
 * rfaSocketInit - perform all necessary initializations of the socket
 */
static FLStatus
rfaSocketInit(FLSocket vol)             /* pointer identifying drive */
{
    vol.cardChanged = FALSE;

    /* enable memory window and map it at address 0 */
    rfaSetWindow(&vol);

    return flOK;
}

/*
 * rfaSetWindow - set current window attributes, Base address, size, etc
 *
 * This routine sets current window hardware attributes: Base address, size,
 * speed and bus width.  The requested settings are given in the 'vol.window'
 * structure.  If it is not possible to set the window size requested in
 * 'vol.window.size', the window size should be set to a larger value,
 * if possible. In any case, 'vol.window.size' should contain the
 * actual window size (in 4 KB units) on exit.
 *
 * RETURNS: N/A
 */
#define pgno(a)         ((a)>>12)
static void
rfaSetWindow(FLSocket vol)
{
    /* Physical base as a 4K page */
    vol.window.baseAddress = pgno(PHYS_TO_K1(AR5311_FLASH0));
    flSetWindowSize(&vol, pgno(FLASH_BASE_LEN));
}

/*
 * rfaSetMappingContext - sets the window mapping register to a card address
 * The window should be set to the value of 'vol.window.currentPage',
 * which is the card address divided by 4KB.  An address over 128MB,
 * (page over 32K) specifies an attribute-space address.
 *
 * The page to map is guaranteed to be on a full window-size boundary.
 */
static void
rfaSetMappingContext(FLSocket vol, unsigned page)
{
    /* no mapping required */
}

/*
 * rfaGetAndClearCardChangeIndicator - return the hardware card-change
 * indicator and clears it if set.
 */
static FLBoolean
rfaGetAndClearCardChangeIndicator(FLSocket vol)
{
    return FALSE;               /* we cannot remove our flash */
}

/*
 * rfaWriteProtected - return the write-protect state of the media.
 */
static FLBoolean
rfaWriteProtected(FLSocket vol)
{
    return FALSE;               /* we do not have write protection */
}

/*
 * flFitInSocketWindow - Checks whether the flash array fits in the socket
 * window.
 *
 *  chipSize            // size of single physical chip in bytes
 *  interleaving        // flash chip interleaving (1,2,4 etc)
 *  windowSize          // socket window size in bytes
 *
 * RETURNS: A chip size guaranteed to fit in the socket window.
 */
long int
flFitInSocketWindow(long int chipSize, int interleaving, long int windowSize)
{
    return chipSize;
}

/*
 * flDelayLoop - consume the specified time.
 */
void
flDelayLoop(int cycles)
{
    cycles >>= 8;                       /* divide by 256 */
    sysUDelay(cycles ? cycles : 1);
}

/*
 * sysTffsFormat - This routine formats the flash memory, taking care of
 * the boot area is not touched.
 */
STATUS
sysTffsFormat(int clear)
{
    static tffsDevFormatParams params = {
        {
            0,                  /* boot sector of ROM_SIZE bytes */
            99,                 /* percentUse */
            1,                  /* spare units for update */
            0x00010000,         /* in core buffer */
            NULL,               /* erase callback */
            {0,0,0,1},          /* DOS volume ID */
            "Atheros AP",       /* DOS volume label */
            2,                  /* DOS FAT copies (standard is 2) */
            0,                  /* PCMCIA CIS length */
            NULL                /* PCMCIA CIS header */
        },
        FTL_FORMAT_IF_NEEDED    /* DOS, FTL only if needed */
    };

    params.formatParams.bootImageLen = ROM_SIZE; /* runtime */

    /* This is a long operation, disable the watchdog */
    sysWatchdogDisable();

    /* Force complete low-level format? */
    if (clear) {
        params.formatFlags = FTL_FORMAT;
    }

    return tffsDevFormat(0, (int)&params);
}

/*
 * Mount ttfs filesystem if it exists.  Called early so we can
 * initialize the radio config from the file system.
 *
 * Returns 1 if we should use the flash FS, 0 otherwise.
 */
int
sysTffsMount(void)
{
    int status, fd;

    status = usrTffsConfig(0, 0, "/fl") == OK;
    if (status) {
        struct timespec ts;

        /* Set time later than 1/99 toa void start-up warnings */
        ts.tv_sec  = 32 * 365 * 24 * 60 * 60;            /* ~2002 */
        ts.tv_nsec = 0;
        clock_settime(CLOCK_REALTIME,&ts);

        /* Check filesystem consistancy at start-up -- unfortunately
         * it does print that the fs is clean and really isn't silent.
         */
        fd = open("/fl/.", O_RDONLY, 0);
        if (fd >= 0) {
            ioctl(fd, FIOCHKDSK, DOS_CHK_REPAIR|DOS_CHK_VERB_SILENT);
            close(fd);
        }
    }
    return status;
}
#endif  /* INCLUDE_TFFS */

