/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * wlanchannel.c - contains regulatory domain table and creates the channel lists
 */

#ident  "ACI $Id: //depot/sw/releases/2.0.1/src/common/wlanchannel.c#1 $"

/*
 *  This module contains the regulatory domain table and accessor functions
 *  for the information in the table.
 *
 *  The channel list creation is also contained in this module.
 */
#include <linux/sched.h>
                        
#include "display.h"
#include "wlantype.h"
#include "wlandrv.h"
#include "wlanext.h"
#include "stacserv.h"
#include "halApi.h"
#include "ui.h"
#include "wlanchannel.h"
#include "radarext.h"
#include "wlansmeext.h"
#include "wlanbeacon.h"

#include "ar5hwcEnd.h"                      
#include "apcfg.h"
                      
#ifdef VXWORKS
#include "ar5hwcEnd.h"
#include "vxWorks.h"
#include "sysLib.h"
#include "taskLib.h"
#include "apcfg.h"
#else
//fadi #include "ndisdrvext.h"
#endif

/* Remove to enforce EEPROM restriction of country codes */
#define ALLOW_DEBUG_COUNTRY

#define FCC_OUTDOOR_FIRST_FREQ 5725

/* Enumerated Regulatory Domain Information */
enum EnumRd {
    NO_ENUMRD     = 0x00,
    FCC2_FCCA     = 0x20, // Canada
    FCC2_WORLD    = 0x21, // Australia
    ETSI2_WORLD   = 0x35, // Germany
    ETSI3_WORLD   = 0x36, // Austria
    ETSI3_ETSIA   = 0x32, // France
    ETSI1_WORLD   = 0x30, // UK
    FCC1_WORLD    = 0x11, // Hong Kong
    FCC1_FCCA     = 0x10, // USA
    APL2_WORLD    = 0x45, // Korea
    APL1_WORLD    = 0x52, // Latin Amer
    APL1_FCCA     = 0x53, // Mexico
    APL1_APLA     = 0x54, // Chili
    APL1_APLB     = 0x44, // China
    MKK1_MKKA     = 0x40, // Japan
    APL4_WORLD    = 0x42, // Singapore
    APL3_FCCA     = 0x43, // Taiwan

    APL1          = 0x0150,
    APL2          = 0x0250,
    APL3          = 0x0350,
    APL4          = 0x0450,
    ETSI1         = 0x0130,
    ETSI2         = 0x0230,
    ETSI3         = 0x0330,
    FCC1          = 0x0110,
    FCC2          = 0x0120, 
    MKK1          = 0x0140,
    APLA          = 0x0A50,
    APLB          = 0x0B50,
    ETSIA         = 0x0A30,
    FCCA          = 0x0A10,
    MKKA          = 0x0A40,
    WORLD         = 0x0199,
    DEBUG_REG_DMN = 0x01ff
};

#define DEFAULT_REG_CODE_FOR_BLANK_EEPROMS FCC1_FCCA
#define COUNTRY_ERD_FLAG 0x8000

typedef struct RegDmnLegacyNames {
    A_UCHAR  regDmnName[8];
    A_UINT16 regDmn;
    A_UINT16 countryCode;
} REG_DMN_LEGACY_NAMES;

static const REG_DMN_LEGACY_NAMES allLegacyNames[] = {
    {"NONE", NO_ENUMRD,   CTRY_DEBUG          },
    {"FCC",  FCC1_FCCA,   CTRY_UNITED_STATES  },
    {"MKK",  MKK1_MKKA,   CTRY_JAPAN          },
    {"ETSI", ETSI1_WORLD, CTRY_UNITED_KINGDOM }
};

typedef struct CountryCodeToEnumRd {
    A_UINT16        countryCode;       
    A_UINT16        regDmnEnum;
    const A_UCHAR   isoName[3];
    const A_UCHAR   name[16];
} COUNTRY_CODE_TO_ENUM_RD;
  
/*
 * Country Code Table to Enumerated RD
 */
static const COUNTRY_CODE_TO_ENUM_RD allCountries[] = {
    {CTRY_DEBUG,       NO_ENUMRD,     "DB", "DEBUG"       },
    {CTRY_DEFAULT,     DEFAULT_REG_CODE_FOR_BLANK_EEPROMS, "NA", "NO_COUNTRY_SET"}, 
    {CTRY_ARGENTINA,   APL1_WORLD,    "AR", "ARGENTINA"   },
    {CTRY_AUSTRALIA,   FCC2_WORLD,    "AU", "AUSTRALIA"   },
    {CTRY_AUSTRIA,     ETSI3_WORLD,   "AT", "AUSTRIA"     },
    {CTRY_BELGIUM,     ETSI1_WORLD,   "BE", "BELGUIM"     },
    {CTRY_BRAZIL,      APL1_FCCA,     "BR", "BRAZIL"      },
    {CTRY_CANADA,      FCC2_FCCA,     "CA", "CANADA"      },
    {CTRY_CHILE,       APL1_APLA,     "CL", "CHILE"       },
    {CTRY_PRCHINA,     APL1_APLB,     "CN", "CHINA"       },
    {CTRY_COLOMBIA,    APL1_FCCA,     "CO", "COLOMBIA"    },
    {CTRY_DENMARK,     ETSI1_WORLD,   "DK", "DENMARK"     },
    {CTRY_ECUADOR,     APL1_WORLD,    "EC", "ECUADOR"     },
    {CTRY_FINLAND,     ETSI1_WORLD,   "FI", "FINLAND"     },
    {CTRY_FRANCE,      ETSI3_ETSIA,   "FR", "FRANCE"      },
    {CTRY_GERMANY,     ETSI2_WORLD,   "DE", "GERMANY"     },
    {CTRY_HONG_KONG,   FCC2_WORLD,    "HK", "HONG_KONG"   },
    {CTRY_IRELAND,     ETSI1_WORLD,   "IE", "IRELAND"     },
    {CTRY_ITALY,       ETSI2_WORLD,   "IT", "ITALY"       },
    {CTRY_JAPAN,       MKK1_MKKA,     "JP", "JAPAN"       },
    {CTRY_SOUTH_KOREA, APL2_WORLD,    "KR", "KOREA_REPUBLIC" },
    {CTRY_MACAU,       FCC1_WORLD,    "MO", "MACAU"       },
    {CTRY_MEXICO,      APL1_FCCA,     "MX", "MEXICO"      },
    {CTRY_NETHERLANDS, ETSI3_WORLD,   "NL", "NETHERLANDS" },
    {CTRY_NEW_ZEALAND, FCC2_WORLD,    "NZ", "NEW_ZEALAND" },
    {CTRY_NORWAY,      ETSI3_WORLD,   "NO", "NORWAY"      },
    {CTRY_PANAMA,      APL1_WORLD,    "PA", "PANAMA"      },
    {CTRY_PERU,        APL1_WORLD,    "PE", "PERU"        },
    {CTRY_PHILIPPINES, FCC1_WORLD,    "PH", "PHILIPPINES" },
    {CTRY_POLAND,      ETSI2_WORLD,   "PL", "POLAND"      },
    {CTRY_PUERTO_RICO, FCC1_FCCA,     "PR", "PUERTO_RICO" },
    {CTRY_SINGAPORE,   APL4_WORLD,    "SG", "SINGAPORE"   },
    {CTRY_SWEDEN,      ETSI1_WORLD,   "SE", "SWEDEN"      },
    {CTRY_SWITZERLAND, ETSI2_WORLD,   "CH", "SWITZERLAND" },
    {CTRY_TAIWAN,      APL3_FCCA,     "TW", "TAIWAN"      },
    {CTRY_UNITED_KINGDOM, ETSI1_WORLD, "GB", "UNITED_KINGDOM"},
    {CTRY_UNITED_STATES, FCC1_FCCA,   "US", "UNITED_STATES"},
    {CTRY_VENEZUELA,   APL1_WORLD,    "VE", "VENEZUELA"   }
};

typedef struct EnumRdToWirelessModeRd {
    A_UINT16  regDmnEnum;
    A_UINT16  regDmnWirelessMode5;
    A_UINT16  regDmnWirelessMode2;
} ENUM_RD_TO_WIRELESS_MODE_RD;

static const ENUM_RD_TO_WIRELESS_MODE_RD allEnumRds[] = {
    {NO_ENUMRD, DEBUG_REG_DMN, DEBUG_REG_DMN},
    {FCC2_FCCA,     FCC2,    FCCA    },
    {FCC2_WORLD,    FCC2,    WORLD   },
    {ETSI2_WORLD,   ETSI2,   WORLD   },
    {ETSI3_WORLD,   ETSI3,   WORLD   },
    {ETSI3_ETSIA,   ETSI3,   ETSIA   },
    {ETSI1_WORLD,   ETSI1,   WORLD   },
    {FCC1_WORLD,    FCC1,    WORLD   },
    {FCC1_FCCA,     FCC1,    FCCA    },
    {APL2_WORLD,    APL2,    WORLD   },
    {APL1_WORLD,    APL1,    WORLD   },
    {APL1_FCCA,     APL1,    FCCA    },
    {APL1_APLA,     APL1,    APLA    },
    {APL1_APLB,     APL1,    APLB    },
    {MKK1_MKKA,     MKK1,    MKKA    },
    {APL4_WORLD,    APL4,    WORLD   },
    {APL3_FCCA,     APL3,    FCCA    }
};

#define TURBO_CHANNEL_SEPARATION (40)
#define OFDM5_CHANNEL_SEPARATION (20)
#define CCK_CHANNEL_SEPARATION   (5)

#define MAX_CHANNEL_GROUPS 4

/* 5 GHz table settings */
typedef struct RegDmnWmChannelGroup {
    A_UINT16    lowChannel;   // Low Channel center in MHz
    A_UINT16    highChannel;  // High Channel center in MHz
    A_UINT8     powerDfs;     // Maximum power (dBm) for this channel range when using DFS
    A_UINT8     antennaMax;   // Maximum allowed antenna gain
} REG_DMN_WM_CHANNEL_GROUP;

typedef struct RegDmnWmFreqTable {
    A_UINT16                 regDmnWirelessMode;
    A_UINT16                 entries;
    A_UINT8                  conformanceTestLimit;
    REG_DMN_WM_CHANNEL_GROUP chanGroup[MAX_CHANNEL_GROUPS];
} REG_DMN_WM_FREQ_TABLE;

enum ConformanceTestLimits {
    FCC    = 0x10,
    MKK    = 0x40,
    ETSI   = 0x30,
    NO_CTL = 0xFF
};

/*
 * Regulatory Domain Tables
 *
 * Tables have entries ordered with AP/Adhoc planting preference in mind.
 *  -Indoor channels are ordered before outdoor
 *  -Higher power level indoor channels are first
 */

static const REG_DMN_WM_FREQ_TABLE regDmnWm5[] = {
    {DEBUG_REG_DMN, 2, FCC, {
    {5120, 5700,  5,  6},
    {5745, 5825,  5,  6},
    {   0,    0,  0,  0},
    {   0,    0,  0,  0}}},

    {APL1, 1, FCC, {
    {5745, 5825, 30, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {APL2, 1, FCC, {
    {5745, 5805, 23, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {APL3, 2, FCC, {
    {5260, 5320, 17, 6},
    {5745, 5825, 30, 6},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {APL4, 2, FCC, {
    {5180, 5240, 20, 0},
    {5745, 5825, 20, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    /*
     * The ETSI1 domain requires TPC
     * As the TPC specification are unfinished, 3dBi are
     * removed from each of the ETSI1 power selections
     */
    {ETSI1, 2, ETSI, {
    {5180, 5320, 20, 0},  // Should be 23 with TPC
    {5500, 5700, 27, 0},  // Should be 30 with TPC
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {ETSI2, 1, ETSI, {
    {5180, 5320, 18, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {ETSI3, 1, ETSI, {
    {5180, 5320, 20, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    /*
     * Values artificially decreased to meet FCC testing
     * procedures
     */
    {FCC1, 3, FCC, {
    {5260, 5320, 18, 6},   // Should allow 23 dBm
    {5180, 5240, 14, 6},   // Should allow 17 dBm
    {5745, 5825, 18, 6},   // Should allow 30 dBm
    {   0,    0,  0, 0}}},

    {FCC2, 2, FCC, {
    {5180, 5320, 23, 6},
    {5745, 5825, 30, 6},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {MKK1, 1, MKK, {
    {5170, 5230, 23, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}}
};

static const REG_DMN_WM_FREQ_TABLE regDmnWm5Turbo[] = {
    {DEBUG_REG_DMN, 2, FCC, {
    {5130, 5670,  5, 6},
    {5150, 5650,  5, 6},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {FCC1, 3, FCC, {
    {5250, 5290, 23, 6},
    {5210, 5210, 17, 6},
    {5760, 5800, 30, 6},
    {   0,    0,  0, 0}}}
};

static const REG_DMN_WM_FREQ_TABLE regDmnWm2[] = {
    {DEBUG_REG_DMN, 2, NO_CTL, {
    {2412, 2472,  5, 6},
    {2484, 2484,  5, 6},
    /* TODO: Add Debug channels 15-26 */
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {APLA, 1, NO_CTL, {
    {2412, 2472, 17, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {APLB, 1, NO_CTL, {
    {2412, 2472, 10, 6},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {ETSIA, 1, NO_CTL, {
    {2452, 2472, 20, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    /*
     * Transmit Power is artificially decreased to meet
     * regulatory domain testing for FCC
     */
    {FCCA, 1, NO_CTL, {
    {2412, 2472, 18, 6}, // Should be 27
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {MKKA, 2, NO_CTL, {
    {2412, 2472, 20, 0},
    {2484, 2484, 20, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},

    {WORLD, 1, NO_CTL, {
    {2412, 2472, 20, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0},
    {   0,    0,  0, 0}}},
};

/* Forward declarations */
static A_UINT16
wlanGetNumChannelsForDomain(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode, A_UINT16 modeSelect,
    A_UINT16 enableOutdoor);

static A_STATUS
wlanAllocScanList(WLAN_SCAN_LIST **pSList, A_UINT16 entries);

static A_STATUS
wlanAllocChannelList(WLAN_CHANNEL_LIST **pCList, A_UINT16 entries);

static CHAN_VALUES *
wlanFindChannel(WLAN_DEV_INFO *pDevInfo, BSSDESCR_SET *pBssSet, WLAN_CHANNEL_LIST *pCList,
    BSSDESCR_SET *pIntBssSet);

#define TOTAL_CLIST_MODES   4
typedef struct {
    A_UINT16 modeSelect;
    A_UINT16 channelFlags;
} cListModes;

const static cListModes modeTable[TOTAL_CLIST_MODES] = {
    {MODE_SELECT_TURBO, CHANNEL_T},
    {MODE_SELECT_11A,   CHANNEL_A},
    {MODE_SELECT_11B,   CHANNEL_B},
    {MODE_SELECT_PUREG, CHANNEL_PUREG}
};

static const REG_DMN_WM_FREQ_TABLE *
wlanGetCcTable(A_UINT16 channelFlags)
{
    if (channelFlags & CHANNEL_2GHZ) {
        return (regDmnWm2);
    }
    else if (channelFlags & CHANNEL_5GHZ) {
        if (channelFlags & CHANNEL_TURBO) {
            return(regDmnWm5Turbo);
        } else {
            return(regDmnWm5);
        }
    }
    // either 2 or 5 GHz flags must be set
    ASSERT(0);
    return NULL;
}

static int
wlanGetCcEntries(A_UINT16 channelFlags)
{
    if (channelFlags & CHANNEL_2GHZ) {
        return(sizeof(regDmnWm2)/sizeof(REG_DMN_WM_FREQ_TABLE));
    }
    else if (channelFlags & CHANNEL_5GHZ) {
        if (channelFlags & CHANNEL_TURBO) {
            return(sizeof(regDmnWm5Turbo)/sizeof(REG_DMN_WM_FREQ_TABLE));
        } else {
            return(sizeof(regDmnWm5)/sizeof(REG_DMN_WM_FREQ_TABLE));
        }
    }
    // either 2 or 5 GHz flags must be set
    ASSERT(0);
    return 0;
}

static A_UINT16
wlanGetChannelSpread(A_UINT16 channelFlags)
{
    if (channelFlags & CHANNEL_2GHZ) {
        // TODO: currently we return same value for CCK or OFDM in 2 GHz
        return (CCK_CHANNEL_SEPARATION);
    }
    else if (channelFlags & CHANNEL_5GHZ) {
        if (channelFlags & CHANNEL_TURBO) {
            return(TURBO_CHANNEL_SEPARATION);
        } else {
            return(OFDM5_CHANNEL_SEPARATION);
        }
    }
    // either 2 or 5 GHz flags must be set
    ASSERT(0);
    return 20; // default. returning 0 may throw it into a loop
}

/***********************************************************
 * wlanGetCountryIndex
 *
 * Match the given countryCode in the static country Table
 */
A_UINT16
wlanGetCountryIndex(A_UINT16 countryCode)
{
    A_UINT16 index;

    for (index = 0; index < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); index++) {
        if (countryCode == allCountries[index].countryCode) {
            return index;
        }
    }
    uiPrintf("wlanGetCountryIndex: Failed to find index in country code table %d\n", countryCode);
    ASSERT(0);
    return 0;
}

/***********************************************************
 * wlanGetRegDmnEnumIndex
 *
 * Get the Wireless Mode Regulatory Domain based on the country code
 * and wireless mode
 */
A_INT16
wlanGetRegDmnEnumIndex(A_UINT16 regDmnEnum)
{
    A_UINT16 index;

    for (index = 0; index < sizeof(allEnumRds) / sizeof(ENUM_RD_TO_WIRELESS_MODE_RD); index++) {
        if (regDmnEnum == allEnumRds[index].regDmnEnum) {
            return index;
        }
    }
    uiPrintf("wlanGetRegDmnEnumIndex: Failed to find index in regDmnEnum table %d\n", regDmnEnum);
    return -1;
}

/***********************************************************
 * wlanGetWirelessModeRegDmn
 *
 * Get the Wireless Mode Regulatory Domain based on the country code
 * and wireless mode
 */
A_UINT16
wlanGetWirelessModeRegDmn(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode, A_UINT16 channelFlags)
{
    A_UINT16 countryIndex, eepRdCc;
    A_INT16  regDmnEnumIndex;

    if (countryCode == CTRY_DEFAULT) {
        eepRdCc = halGetRegDmn(pDevInfo);
        /* No country set - USE EEPROM value */
        if (!eepRdCc) {
            regDmnEnumIndex = wlanGetRegDmnEnumIndex(DEFAULT_REG_CODE_FOR_BLANK_EEPROMS);
        } else if ((regDmnEnumIndex = wlanGetRegDmnEnumIndex(eepRdCc)) == -1) {
            uiPrintf("wlanGetWirelessModeRegDmn: Could not determine country using default\n");
            regDmnEnumIndex = wlanGetRegDmnEnumIndex(DEFAULT_REG_CODE_FOR_BLANK_EEPROMS);
        }
    } else {
        countryIndex = wlanGetCountryIndex(countryCode);
        regDmnEnumIndex = wlanGetRegDmnEnumIndex(allCountries[countryIndex].regDmnEnum);
    }
    if (channelFlags & CHANNEL_2GHZ) {
        return allEnumRds[regDmnEnumIndex].regDmnWirelessMode2;
    } else {
        return allEnumRds[regDmnEnumIndex].regDmnWirelessMode5;
    }
}

/*******************************************************************************
 * wlanInitChannelList
 *
 * This routine takes a regulatory domain (enumerated) and builds the channel list for base
 * and adds turbo mode channels if selected
 *
 * Returns: A_OK, A_EINVAL
 */
A_STATUS
wlanInitChannelList(WLAN_DEV_INFO *pDevInfo, WLAN_CHANNEL_LIST **pCList, A_UINT16 countryCode,
    A_UINT16 modeSelect, A_UINT16 enableOutdoor)
{
    const REG_DMN_WM_FREQ_TABLE *pCcTable;
    A_STATUS     status = A_OK;
    A_UINT16     lowChannel, highChannel, totalChannels = 0;
    A_UINT16     maxRdChannel = 7000;
    A_UINT16     addChannel, channelSpread;
    int          i, j, k, regDmnEnumEntries;
    A_UINT16     regDmnWirelessMode;
    A_UINT16     clistSize;

    regDmnEnumEntries = 0;             /* for gcc -O2 */

    clistSize = wlanGetNumChannelsForDomain(pDevInfo, countryCode, modeSelect, enableOutdoor);
    if (clistSize == 0) {
        return A_EINVAL;
    }

    status = wlanAllocChannelList(pCList, clistSize);
    if (status != A_OK || *pCList == NULL) {
        return status;
    }

    if ((!enableOutdoor) && (countryCode == CTRY_UNITED_STATES)) {
        maxRdChannel = FCC_OUTDOOR_FIRST_FREQ;
    }

    for (k = 0; k < TOTAL_CLIST_MODES; k++) {
        if (!(modeTable[k].modeSelect & modeSelect)) {
            // Corresponding bit in channelFlags is not set,
            // we don't want add that type of channels.
            continue;
        }
        /* Skip channel modes not supported by this hardware */
        if (halGetChannelEdges(pDevInfo, modeTable[k].channelFlags, &lowChannel, &highChannel) == FALSE) {
            continue;
        }
        pCcTable = wlanGetCcTable(modeTable[k].channelFlags);
        regDmnEnumEntries = wlanGetCcEntries(modeTable[k].channelFlags);
        channelSpread = wlanGetChannelSpread(modeTable[k].channelFlags);
        regDmnWirelessMode = wlanGetWirelessModeRegDmn(pDevInfo, countryCode, modeTable[k].channelFlags);

        /* Search through regulatory domain enumeration table for a match */
        for (i = 0; i < regDmnEnumEntries; i++) {
            if (pCcTable[i].regDmnWirelessMode == regDmnWirelessMode) {
                /* Build channel list */
                for (j = 0; j < pCcTable[i].entries; j++) {
                    /* Add entries to the channel list */
                    for(addChannel = pCcTable[i].chanGroup[j].lowChannel;
                        addChannel <= pCcTable[i].chanGroup[j].highChannel;
                        addChannel += channelSpread)
                    {
                        if ((addChannel >= lowChannel) && (addChannel <= highChannel) &&
                            (addChannel <= maxRdChannel))
                        {
                            (*pCList)->chanArray[totalChannels].channel = addChannel;
                            (*pCList)->chanArray[totalChannels].channelFlags = modeTable[k].channelFlags;
                            if (((regDmnWirelessMode & 0xFF) == ETSI) &&
                                (modeTable[k].channelFlags & CHANNEL_5GHZ))
                            {
                                (*pCList)->chanArray[totalChannels].channelFlags |= CHANNEL_PASSIVE;
								aPrintk("\t---channel=0x%x------flags=0x%x\n", 
									   (*pCList)->chanArray[totalChannels].channel,
									   (*pCList)->chanArray[totalChannels].channelFlags);
                            }
                            totalChannels++;
                            ASSERT(totalChannels <= clistSize);
                        }
                    }
                }
                /* List is built for this table - exit loop */
                break;
            }
        }
    } /* End for loop around table entries */

    (*pCList)->listSize = totalChannels;
    if (totalChannels == 0) {
#ifdef DEBUG
        uiPrintf("wlanInitChannelList: Matching Country Code: %s produced zero entries\n",
            wlanGetCountryCodeName(pDevInfo, countryCode));
#endif
        return A_EINVAL;
    }
    return status;
}

/***********************************************************
 * wlanGetNumChannelsForDomain
 *
 * Count the number of channels in this domain table adding in
 * all possible modes.
 */
A_UINT16
wlanGetNumChannelsForDomain(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode, A_UINT16 modeSelect,
    A_UINT16 enableOutdoor)
{
    const REG_DMN_WM_FREQ_TABLE *pCcTable;
    A_UINT16   lowChannel, highChannel, totalChannels = 0;
    A_UINT16   searchChannel, channelSpread;
    A_UINT16   maxRdChannel = 7000;
    int        i, j, k, regDmnEnumEntries;
    A_UINT16   regDmnWirelessMode;

    
    if ((!enableOutdoor) && (countryCode == CTRY_UNITED_STATES)) {
        maxRdChannel = FCC_OUTDOOR_FIRST_FREQ;
    }

    i = regDmnEnumEntries = 0;             /* for gcc -O2 */

    for (k = 0; k < TOTAL_CLIST_MODES; k++) {
        if (!(modeTable[k].modeSelect & modeSelect)) {
            // Corresponding bit in channelFlags is not set,
            // we don't want add that type of channels.
            continue;
        }
        /* Skip channel modes not supported by this hardware */
        if (halGetChannelEdges(pDevInfo, modeTable[k].channelFlags, &lowChannel, &highChannel) == FALSE) {
            continue;
        }
		
		//aPrintk("\t\tlowchannel=%d hichannel=%d for this mode(11a)\n", lowChannel, highChannel);

        pCcTable = wlanGetCcTable(modeTable[k].channelFlags);
        regDmnEnumEntries = wlanGetCcEntries(modeTable[k].channelFlags);
        channelSpread = wlanGetChannelSpread(modeTable[k].channelFlags);
        regDmnWirelessMode = wlanGetWirelessModeRegDmn(pDevInfo, countryCode, modeTable[k].channelFlags);
        /* Search through Regulatory Domain (Enum) table for a match */
        for (i = 0; i < regDmnEnumEntries; i++) {
            if (pCcTable[i].regDmnWirelessMode == regDmnWirelessMode) {
                for (j = 0; j < pCcTable[i].entries; j++) {
                    for(searchChannel = pCcTable[i].chanGroup[j].lowChannel;
                        searchChannel <= pCcTable[i].chanGroup[j].highChannel;
                        searchChannel += channelSpread)
                    {
                        if ((searchChannel >= lowChannel) && (searchChannel <= highChannel) &&
                            (searchChannel <= maxRdChannel))
                        {
                            totalChannels++;
                        }
                    }
                }
                /* List is built for this table - exit loop */
                break;
            }
        }
    }

	//aPrintk("\t\ttotal channels found=%d\n", totalChannels);

    if (totalChannels == 0) {
#ifdef DEBUG 
        uiPrintf("wlanGetNumChannelsForDomain: No matching channels were found\n");
#endif
        return A_EINVAL;
    }
    return totalChannels;
}

/***********************************************************
 * asciiToClist - Converts ascii string to channel list
 *
 * This routine parses an ASCII string of decimal digits and ';' and '-'
 * and converts them into a list of numbers.  Ranges are incremented by 4
 * (IEEE 5Mhz channels) to bump by 20Mhz.
 *
 * RETURNS: FALSE if conversion failed, TRUE if it succeeds
 */
#define CHANNEL_LIST_SIZE 128

A_STATUS
asciiToClist(WLAN_DEV_INFO *pDevInfo, A_UCHAR *pStr, A_UINT16 alen, WLAN_CHANNEL_LIST **pList)
{
    A_UINT16 lindex = 0;
    A_UINT16 asciiIndex, addChannel;
    A_UINT16 fnum = 0, cnum = 0;
    A_UINT16 lowChannel, highChannel;
    A_UINT16 channelSpread = OFDM5_CHANNEL_SEPARATION;
    A_UINT16 channelFlags = CHANNEL_A;
    CHAN_VALUES tempChannels[CHANNEL_LIST_SIZE];
    A_UINT16 i;
    A_STATUS status = A_OK;

    ASSERT(pList);
    if ((pStr == NULL) || (alen <= 0)) {
        return A_EINVAL;
    }

    for (asciiIndex = 0; (asciiIndex < alen) && (lindex < CHANNEL_LIST_SIZE); asciiIndex++) {
        switch (pStr[asciiIndex]) {
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            cnum = cnum * 10 + (pStr[asciiIndex] - '0');
            break;
        case '-':
            fnum = cnum;
            cnum = 0;
            break;
        case 'g':
        case 'G':
            channelFlags = CHANNEL_PUREG;
            channelSpread = CCK_CHANNEL_SEPARATION;
            break;
        case 't':
        case 'T':
            channelFlags = CHANNEL_T;
            channelSpread = TURBO_CHANNEL_SEPARATION;
            break;
        case 'b':
        case 'B':
            channelFlags = CHANNEL_B;
            channelSpread = CCK_CHANNEL_SEPARATION;
            break;
        case 'a':
        case 'A':
            channelFlags = CHANNEL_A;
            channelSpread = OFDM5_CHANNEL_SEPARATION;
            break;
        case ',':
            if (fnum == 0) {
                fnum = cnum;
            }
            /* Channel mode must be supported in hardware to be in the channel list */
            if (halGetChannelEdges(pDevInfo, channelFlags, &lowChannel, &highChannel)) {
                /* Convert channels up for scripted testing */
                if ((fnum > 0) && (fnum < 256)) {
                    fnum = IEEE_CH2MHZ(fnum);
                }
                if ((cnum > 0) && (cnum < 256)) {
                    cnum = IEEE_CH2MHZ(cnum);
                }
                if (fnum < lowChannel || cnum > highChannel)
                {
                    // one of the nuber doesn't fit into the valid range
                    fnum = cnum = 0;
                    break;
                }
                for (addChannel = fnum; (addChannel <= cnum) && (fnum != 0) && (lindex < CHANNEL_LIST_SIZE);
                    addChannel += channelSpread)
                {
                    if ((addChannel >= lowChannel) && (addChannel <= highChannel) &&
                        (lindex < CHANNEL_LIST_SIZE))
                    {
                        tempChannels[lindex].channel = addChannel;
                        tempChannels[lindex].channelFlags = channelFlags;
                        lindex++;
                    }
                }
            }
            fnum = cnum = 0;
            break;
        default:
            return A_ERROR;
        }
    }
    if (asciiIndex == alen) {
        if (fnum == 0) {
            fnum = cnum;
        }
        /* Channel mode must be supported in hardware to be in the channel list */
        if (halGetChannelEdges(pDevInfo, channelFlags, &lowChannel, &highChannel)) {
            /* Convert channels up for scripted testing */
            if ((fnum > 0) && (fnum < 256)) {
                fnum = IEEE_CH2MHZ(fnum);
            }
            if ((cnum > 0) && (cnum < 256)) {
                cnum = IEEE_CH2MHZ(cnum);
            }
            if (fnum >= lowChannel && cnum <= highChannel) {
                for (addChannel = fnum; (addChannel <= cnum) && (fnum != 0) && (lindex < CHANNEL_LIST_SIZE);
                    addChannel += channelSpread)
                {
                    if ((addChannel >= lowChannel) && (addChannel <= highChannel) &&
                        (lindex < CHANNEL_LIST_SIZE))
                    {
                        tempChannels[lindex].channel = addChannel;
                        tempChannels[lindex].channelFlags = channelFlags;
                        lindex++;
                    }
                }
            }
        }
    }

    /* If we got here with no channels set - don't use clist */
    if (lindex == 0) {
        return A_ERROR;
    }

    status = wlanAllocChannelList(pList, lindex);
    if (status != A_OK) {
        return status;
    }
    for (i = 0; i < lindex; i++) {
        (*pList)->chanArray[i].channel = tempChannels[i].channel;
        (*pList)->chanArray[i].channelFlags = tempChannels[i].channelFlags;
        uiPrintf("%2d. New channel freq %d, flags %x\n", i, (*pList)->chanArray[i].channel,
            (*pList)->chanArray[i].channelFlags);
    }
    (*pList)->listSize = lindex;

    return status;
}

/***********************************************************
 * wlanCopyChannelList
 *
 * Copy the first channel list into the second
 */
A_STATUS
wlanCopyChannelList(WLAN_CHANNEL_LIST *pCListFrom, WLAN_CHANNEL_LIST **pCListTo)
{
    A_STATUS status = A_OK;

    wlanFreeChannelList(*pCListTo);
    status = wlanAllocChannelList(pCListTo, pCListFrom->listSize);
    if (status != A_OK) {
        return status;
    }
    A_BCOPY(pCListFrom->chanArray, (*pCListTo)->chanArray,
        (pCListFrom->listSize * sizeof(struct channelValues)));
    (*pCListTo)->listSize = pCListFrom->listSize;
    return status;
}

/***********************************************************
 * wlanAllocChannelList
 *
 * Allocates the number of entries into the provided channel list pointer
 */
A_STATUS
wlanAllocChannelList(WLAN_CHANNEL_LIST **pCList, A_UINT16 entries)
{
    /* Allocate the Channel List */
    *pCList = (struct wlanChannelList *)A_DRIVER_MALLOC((sizeof(struct wlanChannelList) +
        (entries * sizeof(struct channelValues))), 0, 0);
    (*pCList)->chanArray = (struct channelValues *)(((A_UINT8 *)*pCList) + sizeof(struct wlanChannelList));
    if (!(*pCList)) {
        uiPrintf("wlanAllocChannelList: Could not allocate memory for channel list of size %d entries\n",
            entries);
        return A_NO_MEMORY;
    }
    return A_OK;
}

/***********************************************************
 * wlanFreeChannelList
 *
 * Frees all entries in this channel list
 */
A_STATUS
wlanFreeChannelList(WLAN_CHANNEL_LIST *pCList)
{
    A_UINT16 entries;
    if (pCList) {
        entries = pCList->listSize;
        A_DRIVER_FREE(pCList, sizeof(struct wlanChannelList) +
            (entries * sizeof(struct channelValues)), 0);
        pCList = NULL;
    }
    return A_OK;
}

/***********************************************************
 * wlanInitScanList
 *
 * Create a scanList which points into the selected portion of
 * the given channelList
 */
A_STATUS
wlanInitScanList(WLAN_CHANNEL_LIST *pCListFrom, WLAN_SCAN_LIST **pSListTo)
{
    A_STATUS status = A_OK;
    int i;

    ASSERT(*pSListTo == NULL);

    status = wlanAllocScanList(pSListTo, pCListFrom->listSize);
    if (status != A_OK) {
        return status;
    }
    for (i = 0; i < pCListFrom->listSize; i++) {
        (*pSListTo)->pChanArray[i] = &(pCListFrom->chanArray[i]);
    }
    (*pSListTo)->listSize = pCListFrom->listSize;
    return status;
}

/***********************************************************
 * wlanInitSingleScanList
 *
 * Create a scanList which points into the selected portion of
 * the given channelList
 */
A_STATUS
wlanInitSingleScanList(CHAN_VALUES *pChannelFrom, WLAN_SCAN_LIST **pSListTo)
{
    A_STATUS status = A_OK;

    ASSERT(*pSListTo == NULL);

    status = wlanAllocScanList(pSListTo, 1);
    if (status != A_OK) {
        return status;
    }
    (*pSListTo)->pChanArray[0] = pChannelFrom;
    (*pSListTo)->listSize = 1;
    return status;
}

/***********************************************************
 * wlanAllocScanList
 *
 * Allocates the number of ptr entries into the scan list
 */
A_STATUS
wlanAllocScanList(WLAN_SCAN_LIST **pSList, A_UINT16 entries)
{
    /* Allocate the Channel List */
    *pSList = (struct wlanScanList *)A_DRIVER_MALLOC(sizeof(struct wlanScanList) +
        (entries * sizeof(struct channelValues *)), 0, 0);
    (*pSList)->pChanArray = (struct channelValues **)(((A_UINT8 *)*pSList) + sizeof(struct wlanScanList));
    if (!(*pSList)) {
        uiPrintf("wlanAllocScanList: Could not allocate memory for scan list of size %d entries\n",
            entries);
        return A_NO_MEMORY;
    }
    return A_OK;
}

/***********************************************************
 * wlanFreeScanList
 *
 * Frees the scan list
 */
A_STATUS
wlanFreeScanList(WLAN_SCAN_LIST *pSList)
{
    A_UINT16 entries;
    if (pSList) {
        entries = pSList->listSize;
        A_DRIVER_FREE(pSList, sizeof(struct wlanScanList) +
            (entries * sizeof(struct channelValues *)), 0);
        pSList = NULL;
    }
    return A_OK;
}

/***********************************************************
 * wlanGetChannelPtr
 *
 * Find this channel's entry in the channel list
 * Returns NULL if no channel ptr is found in the list
 */
CHAN_VALUES *
wlanGetChannelPtr(WLAN_CHANNEL_LIST *pCList, A_UINT16 channel, A_UINT16 modeSelect)
{
    int i, j;                                                  ASSERT(pCList);

    /* Match all selected modes to find the channel flags */
    for (i = 0; i < TOTAL_CLIST_MODES; i++) {
        if (modeSelect & modeTable[i].modeSelect) {
            for (j = 0; j < pCList->listSize; j++) {
                /* If the same flags exist then we've hit the right channel */
				aPrintk("\t\t>>>>Trying to find a channel match...\n");
                if ((pCList->chanArray[j].channelFlags & CHANNEL_ALL) ==
                    modeTable[i].channelFlags) {
					aPrintk("\t\t>>>> Found an matching entry in list w channel=%d\n", modeTable[i].channelFlags);
                    if (pCList->chanArray[j].channel == channel) {
						aPrintk("\t\t>>>> FOUND A MATCH at channel=%d\n", channel);
                        return &(pCList->chanArray[j]);
                    }
				}
            }

        }
    }
    return NULL;
}

/***********************************************************
 * wlanIsValidChannel
 *
 * Search the regulatory table to see if this channel exists
 * for this device
 */
A_BOOL
wlanIsValidChannel(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode, A_UINT16 modeSelect, A_UINT16 channel)
{
    int mode, i, j, regDmnEnumEntries;
    const REG_DMN_WM_FREQ_TABLE *pCcTable;
    A_UINT16 channelSpread, searchChannel, lowChannel, highChannel;
    A_UINT16 regDmnWirelessMode;

    /* Match all selected modes to find the channel flags */
    for (mode = 0; mode < TOTAL_CLIST_MODES; mode++) {
        if (modeSelect & modeTable[mode].modeSelect) {

            pCcTable = wlanGetCcTable(modeTable[mode].channelFlags);
            regDmnEnumEntries = wlanGetCcEntries(modeTable[mode].channelFlags);
            channelSpread = wlanGetChannelSpread(modeTable[mode].channelFlags);
            halGetChannelEdges(pDevInfo, modeTable[mode].channelFlags, &lowChannel, &highChannel);
            regDmnWirelessMode = wlanGetWirelessModeRegDmn(pDevInfo, countryCode, modeTable[mode].channelFlags);

            /* Search through Regulatory Domain (Enum) table for a match */
            for (i = 0; i < regDmnEnumEntries; i++) {
                if (pCcTable[i].regDmnWirelessMode == regDmnWirelessMode) {
                    for (j = 0; j < pCcTable[i].entries; j++) {
                        /* Find this channel's entry */
                        for(searchChannel = pCcTable[i].chanGroup[j].lowChannel;
                            searchChannel <= pCcTable[i].chanGroup[j].highChannel;
                            searchChannel += channelSpread)
                        {
                            if ((searchChannel >= lowChannel) && (searchChannel <= highChannel) &&
                                (searchChannel == channel))
                            {
                                return TRUE;
                            }
                        }
                    }
                    break;
                }
            }
        }
    }
    return FALSE;
}

/***********************************************************
 * wlanGetAntennaReduction
 *
 * Find the max allowed antenna gain and apply any regulatory
 * domain specific changes.
 *
 * NOTE: a negative reduction is possible in RD's that only
 * measure radiated power (e.g., ETSI) which would increase
 * that actual conducted output power (though never beyond
 * the calibrated target power).
 */
A_INT8
wlanGetAntennaReduction(WLAN_DEV_INFO *pDevInfo, CHAN_VALUES *pChannel, A_INT8 twiceAntennaGain)
{
    A_INT8 antennaMax;
    int i, j, regDmnEnumEntries;
    const REG_DMN_WM_FREQ_TABLE *pCcTable;
    A_UINT16 regDmnWirelessMode;

    pCcTable = wlanGetCcTable(pChannel->channelFlags);
    regDmnEnumEntries = wlanGetCcEntries(pChannel->channelFlags);
    regDmnWirelessMode = wlanGetWirelessModeRegDmn(pDevInfo, pDevInfo->staConfig.countryCode, pChannel->channelFlags);

    /* Search through Regulatory Domain (Enum) table for a match */
    for (i = 0; i < regDmnEnumEntries; i++) {
        if (pCcTable[i].regDmnWirelessMode == regDmnWirelessMode) {
            for (j = 0; j < pCcTable[i].entries; j++) {
                /* Find this channel's entry */
                if ((pChannel->channel >= pCcTable[i].chanGroup[j].lowChannel) &&
                    (pChannel->channel <= pCcTable[i].chanGroup[j].highChannel))
                {
                    antennaMax = twiceAntennaGain - 
                        (pCcTable[i].chanGroup[j].antennaMax * 2);
                    if (antennaMax < 0) {
                        antennaMax = 0;
                    }
                    return antennaMax;
                }
            }
            break;
        }
    }
    /* Failed to find the correct index - may be a debug channel */
    return 0;
 }

/***********************************************************
 * wlanGetCtl
 *
 * Search the regulatory table and return the test group
 */
A_UINT8
wlanGetCtl(WLAN_DEV_INFO *pDevInfo, CHAN_VALUES *pChannel)
{
    int i, regDmnEnumEntries;
    const REG_DMN_WM_FREQ_TABLE *pCcTable;
    A_UINT16 regDmnWirelessMode;

    pCcTable = wlanGetCcTable(pChannel->channelFlags);
    regDmnEnumEntries = wlanGetCcEntries(pChannel->channelFlags);
    regDmnWirelessMode = wlanGetWirelessModeRegDmn(pDevInfo, pDevInfo->staConfig.countryCode, pChannel->channelFlags);

    /* Search through Regulatory Domain (Enum) table for a match */
    for (i = 0; i < regDmnEnumEntries; i++) {
        if (pCcTable[i].regDmnWirelessMode == regDmnWirelessMode) {
            return pCcTable[i].conformanceTestLimit;
        }
    }
    /* Failed to find the correct index - may be a debug channel */
    return NO_CTL;
}

/***********************************************************
 * wlanGetChannelPower
 *
 * Search the regulatory table and return the channel's power
 * (in dBm)
 */
A_UINT8
wlanGetChannelPower(WLAN_DEV_INFO *pDevInfo, CHAN_VALUES *pChannel)
{
    int i, j, regDmnEnumEntries;
    const REG_DMN_WM_FREQ_TABLE *pCcTable;
    A_UINT16 regDmnWirelessMode;

    pCcTable = wlanGetCcTable(pChannel->channelFlags);
    regDmnEnumEntries = wlanGetCcEntries(pChannel->channelFlags);
    regDmnWirelessMode = wlanGetWirelessModeRegDmn(pDevInfo, pDevInfo->staConfig.countryCode, pChannel->channelFlags);

    /* Search through Regulatory Domain (Enum) table for a match */
    for (i = 0; i < regDmnEnumEntries; i++) {
        if (pCcTable[i].regDmnWirelessMode == regDmnWirelessMode) {
            for (j = 0; j < pCcTable[i].entries; j++) {
                /* Find this channel's entry */
                if ((pChannel->channel >= pCcTable[i].chanGroup[j].lowChannel) &&
                    (pChannel->channel <= pCcTable[i].chanGroup[j].highChannel))
                {
                    return (pCcTable[i].chanGroup[j].powerDfs);
                }
            }
            break;
        }
    }
    /* Failed to find the correct index - may be a debug channel */
    return 5;
}


/***********************************************************
 * wlanPrintChannelList
 *
 * Print out the channel list based only on the given reg
 * domain and wireless mode(s)
 */
void
wlanPrintChannelList(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode,
    A_UINT16 wirelessMode, A_UINT16 enableOutdoor)
{
    int i, j, k, regDmnEnumEntries, newPrint = 0;
    const REG_DMN_WM_FREQ_TABLE *pCcTable;
    A_UINT16 channelSpread, searchChannel;
    A_UINT16 lowChannel = 0, highChannel = 0;
    A_UINT16 maxRdChannel = 7000;
    A_UINT16 regDmnWirelessMode;

    if ((!enableOutdoor) && (countryCode == CTRY_UNITED_STATES)) {
        maxRdChannel = FCC_OUTDOOR_FIRST_FREQ;
    }

    for (k = 0; k < TOTAL_CLIST_MODES; k++) {
        if (!(modeTable[k].modeSelect & wirelessMode)) {
            // Corresponding bit in channelFlags is not set,
            // we don't want add that type of channels.
            continue;
        }
        pCcTable = wlanGetCcTable(modeTable[k].channelFlags);
        regDmnEnumEntries = wlanGetCcEntries(modeTable[k].channelFlags);
        channelSpread = wlanGetChannelSpread(modeTable[k].channelFlags);
        regDmnWirelessMode = wlanGetWirelessModeRegDmn(pDevInfo, countryCode, modeTable[k].channelFlags);
        /* Skip channel modes not supported by this hardware */
        if (halGetChannelEdges(pDevInfo, modeTable[k].channelFlags, &lowChannel, &highChannel) == FALSE) {
            continue;
        }

        /* Search through Regulatory Domain (Enum) table for a match */
        for (i = 0; i < regDmnEnumEntries; i++) {
            if (pCcTable[i].regDmnWirelessMode == regDmnWirelessMode) {
                uiPrintf("\t");
                for (j = 0; j < pCcTable[i].entries; j++) {
                    /* Find this channel's entry */
                    for(searchChannel = pCcTable[i].chanGroup[j].lowChannel;
                        searchChannel <= pCcTable[i].chanGroup[j].highChannel;
                        searchChannel += channelSpread)
                    {
                        if ((searchChannel >= lowChannel) && (searchChannel <= highChannel) &&
                            (searchChannel <= maxRdChannel))
                        {
                            uiPrintf("%04d%s", searchChannel,
                                ((++newPrint) % 8) ? (" ") : ("\n\t"));
                        }
                    }
                }
                uiPrintf("\n");
            }
        }
    }
}

/***********************************************************
 * wlanGetCountryCodeTableSize
 *
 * Return the static size of the entire country code table
 */
A_UINT16
wlanGetCountryCodeTableSize(WLAN_DEV_INFO *pDevInfo)
{
    A_UINT16 eepRdCc, totalCountries = 0, i;

    eepRdCc = halGetRegDmn(pDevInfo);
    if (eepRdCc & COUNTRY_ERD_FLAG) {
        /* EEP setting is a country - config shall match */
        totalCountries = 1;
    } else if ((eepRdCc == DEBUG_REG_DMN) || (eepRdCc == NO_ENUMRD)) {
        /* Set to Debug or AllowAnyCountry - allow any setting - 2 to remove Debug */
        totalCountries =  (sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD)) - 1;
    } else {
        for (i = 1; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            if (allCountries[i].regDmnEnum == eepRdCc) {
                totalCountries++;
            }
        }
    }
    ASSERT(totalCountries);
    return totalCountries;
}

/***********************************************************
 * wlanGetIndexedCountryCodeName
 *
 * Return a ptr to the country name at the given index
 */
const A_UCHAR *
wlanGetIndexedCountryCodeName(WLAN_DEV_INFO *pDevInfo, int index)
{
    A_UINT16       eepRdCc, localIndex = 0, i;
    const A_UCHAR  *name = NULL;

    eepRdCc = halGetRegDmn(pDevInfo);
    if (eepRdCc & COUNTRY_ERD_FLAG) {
        /* EEP setting is a country - config shall match */
        ASSERT(index == 0);
        name = allCountries[wlanGetCountryIndex((A_UINT16)(eepRdCc & ~COUNTRY_ERD_FLAG))].name;
    } else if ((eepRdCc == DEBUG_REG_DMN) || (eepRdCc == NO_ENUMRD)) {
        /* Set to Debug or AllowAnyCountry - allow any setting - 2 to remove Debug */
        ASSERT(index < (sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD)) - 1);
        name = allCountries[index + 1].name;
    } else {
        for (i = 1; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            if (allCountries[i].regDmnEnum == eepRdCc) {
                if (index == localIndex) {
                    name = allCountries[i].name;
                    break;
                }
                localIndex++;
            }
        }
    }
    ASSERT(name);
    return name;
}

/***********************************************************
 * wlanGetIndexedCountryCodeIsoName
 *
 * Return the static size of the entire country code table
 */
const A_UCHAR *
wlanGetIndexedCountryCodeIsoName(WLAN_DEV_INFO *pDevInfo, int index)
{
    A_UINT16       eepRdCc, localIndex = 0, i;
    const A_UCHAR  *isoName = NULL;

    eepRdCc = halGetRegDmn(pDevInfo);
    if (eepRdCc & COUNTRY_ERD_FLAG) {
        /* EEP setting is a country - config shall match */
        ASSERT(index == 0);
        isoName = allCountries[wlanGetCountryIndex((A_UINT16)(eepRdCc & ~COUNTRY_ERD_FLAG))].isoName;
    } else if ((eepRdCc == DEBUG_REG_DMN) || (eepRdCc == NO_ENUMRD)) {
        /* Set to Debug or AllowAnyCountry - allow any setting - 2 to remove Debug */
        ASSERT(index < (sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD)) - 1);
        isoName = allCountries[index + 1].isoName;
    } else {
        for (i = 1; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            if (allCountries[i].regDmnEnum == eepRdCc) {
                if (index == localIndex) {
                    isoName = allCountries[i].isoName;
                    break;
                }
                localIndex++;
            }
        }
    }
    ASSERT(isoName);
    return isoName;
}

/***********************************************************
 * wlanGetCountryCodeIndex
 *
 * Find the table index matching the set country code
 */
A_UINT16
wlanGetCountryCodeIndex(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode)
{
    A_UINT16       eepRdCc, localIndex = 0, i;

    /* Set country display to NA if debug is present */
    if (countryCode == CTRY_DEBUG) {
        return 0;
    }

    eepRdCc = halGetRegDmn(pDevInfo);
    if (eepRdCc & COUNTRY_ERD_FLAG) {
        /* EEP setting is a country - config shall match */
        return 0;
    } else if ((eepRdCc == DEBUG_REG_DMN) || (eepRdCc == NO_ENUMRD)) {
        /* Set to Debug or AllowAnyCountry - allow any setting -  remove Debug */
        return wlanGetCountryIndex(countryCode) - 1;
    } else {
        for (i = 1; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            if (allCountries[i].regDmnEnum == eepRdCc) {
                if (countryCode == allCountries[i].countryCode) {
                    return localIndex;
                }
                localIndex++;
            }
        }
    }

    uiPrintf("wlanGetCountryCodeIndex: No matching country\n");
    return 0;
}

/***********************************************************
 * wlanGetCountryCodeName
 *
 * Return a pointer to the ISO name of the country
 */
const A_UCHAR *
wlanGetCountryCodeName(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode)
{
    int i;

    for (i = 0; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
        if (countryCode == allCountries[i].countryCode) {
            return allCountries[i].isoName;
        }
    }
    uiPrintf("wlanGetCountryCodeName: Could not find matching name for country code\n");
    return allCountries[0].isoName;
}

/***********************************************************
 * wlanGetCountryCodeByName
 *
 * Return the country code that matches the given string
 */
A_UINT16
wlanGetCountryCodeByName(A_UCHAR *countryCodeName)
{
    int i;

    for (i = 0; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
        if (strncmp(countryCodeName, allCountries[i].isoName, 2) == 0) {
            return allCountries[i].countryCode;
        }
    }
    uiPrintf("wlanGetCountryCodeByName: Could not find matching country code for country %s\n", countryCodeName);
    return CTRY_INVALID;
}

/***********************************************************
 * wlanIsCountryCodeValid
 *
 * Returns TRUE if cfg'ed country is allowed by the EEPROM setting
 */
A_BOOL
wlanIsCountryCodeValid(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode)
{
    A_UINT16 eepRdCc, i;

    /* Default setting requires no checks */
    if (countryCode == CTRY_DEFAULT) {
        return TRUE;
    }
#ifdef ALLOW_DEBUG_COUNTRY
    if (countryCode == CTRY_DEBUG) {
        return TRUE;
    }
#endif

    eepRdCc = halGetRegDmn(pDevInfo);
    if (eepRdCc & COUNTRY_ERD_FLAG) {
        /* EEP setting is a country - config shall match */
        return (countryCode == (eepRdCc & ~COUNTRY_ERD_FLAG)) ? TRUE : FALSE;
    } else if ((eepRdCc == DEBUG_REG_DMN) || (eepRdCc == NO_ENUMRD)) {
        /* Set to Debug or AllowAnyCountry mode - allow any setting */
        return TRUE;
    } else {
        for (i = 0; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            if ((countryCode == allCountries[i].countryCode) && 
                (allCountries[i].regDmnEnum == eepRdCc))
            {
                return TRUE;
            }
        }
    }
    return FALSE;
}

/***********************************************************
 * wlanPrintCountryCodeList
 *
 * Prints the list of available countries for this EEPROM setting
 */
void
wlanPrintCountryCodeList(WLAN_DEV_INFO *pDevInfo)
{
    int i;
    A_UINT16 eepRdCc, countryCode;

#ifdef ALLOW_DEBUG_COUNTRY
    uiPrintf("DB - DEBUG\n");
#endif

    eepRdCc = halGetRegDmn(pDevInfo);
    if ((eepRdCc == DEBUG_REG_DMN) || (eepRdCc == NO_ENUMRD)) {
        /* Print all countries */
        for (i = 0; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            uiPrintf("%s - %s\n", allCountries[i].isoName, allCountries[i].name);
        }
    } else if (eepRdCc & COUNTRY_ERD_FLAG) {
        /* Print the country that matches the hardware EEPROM setting */
        countryCode = eepRdCc & ~COUNTRY_ERD_FLAG;
        for (i = 0; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            if (countryCode == allCountries[i].countryCode) {
                uiPrintf("%s - %s\n", allCountries[i].isoName, allCountries[i].name);
                break;
            }
        }
        if (i == sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD)) {
            uiPrintf("Hardware setting does not match a known value\n");
        }
    } else {
        /* Print only countries that fall in the set reg domain */
        for (i = 0; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            if (eepRdCc == allCountries[i].regDmnEnum) {
                uiPrintf("%s - %s\n", allCountries[i].isoName, allCountries[i].name);
            }
        }
    }
}

/***********************************************************
 * wlanGetDefaultCountry
 *
 * Returns country from the EEPROM or CTRY_DEFAULT
 */
A_UINT16
wlanGetDefaultCountry(WLAN_DEV_INFO *pDevInfo)
{
    A_UINT16 eepRdCc, i;

    eepRdCc = halGetRegDmn(pDevInfo);
    if (eepRdCc & COUNTRY_ERD_FLAG) {
        eepRdCc = eepRdCc & ~COUNTRY_ERD_FLAG;
        for (i = 0; i < sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD); i++) {
            if (eepRdCc == allCountries[i].countryCode) {
                return eepRdCc;
            }
        }
    }
    return CTRY_DEFAULT;
}

/***********************************************************
 * wlanIsRegCcValid
 *
 * Checks the EEPROM to see if this is a valid country code
 */
A_BOOL
wlanIsRegCcValid(WLAN_DEV_INFO *pDevInfo)
{
    /* TODO: needs to return the CTL for the given software RegCcValid setting */
    return TRUE;
}

/***********************************************************
 * wlanGetCountryInfoList
 *
 * Search the eeprom and return the country string +
 * all available subband channel group info, such as first 
 * channel number in the group, number of channels in the
 * group, and the max transmit power.
 */
void
wlanGetCountryInfoList(WLAN_DEV_INFO *pDevInfo, COUNTRY_INFO_LIST *pCountryInfo)
{
#ifdef DEBUG
    int dummyCountryInfoList;
#endif 
    int i, j = 0, regDmnEnumEntries;
    const REG_DMN_WM_FREQ_TABLE *pCcTable;
    A_UINT16 channelSpread;
    A_UINT16 lowChannel = 0, highChannel = 0;
    A_UINT16 regDmnWirelessMode;
    A_UINT16 countryCode, channelFlags;

    countryCode = pDevInfo->staConfig.countryCode;
    ASSERT(countryCode != CTRY_DEFAULT);
    ASSERT(countryCode != CTRY_DEBUG);

    for (i = 0; i < (sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD)); i++) {
        if (countryCode == allCountries[i].countryCode) {
            break;
        }
    }
    ASSERT(i < (sizeof(allCountries) / sizeof(COUNTRY_CODE_TO_ENUM_RD)));

    pCountryInfo->countryString[0] = allCountries[i].isoName[0];
    pCountryInfo->countryString[1] = allCountries[i].isoName[1];
    /* TODO: outdoor/indoor: the question man has asked forever */
    pCountryInfo->countryString[2] = 'I';

    channelFlags = pDevInfo->staConfig.pChannel->channelFlags;

    pCcTable = wlanGetCcTable(channelFlags);
    regDmnEnumEntries = wlanGetCcEntries(channelFlags);
    channelSpread = wlanGetChannelSpread(channelFlags);
    halGetChannelEdges(pDevInfo, channelFlags, &lowChannel, &highChannel);
    regDmnWirelessMode = wlanGetWirelessModeRegDmn(pDevInfo, countryCode, channelFlags);

    /* Search through Regulatory Domain (Enum) table for a match */
    for (i = 0; i < regDmnEnumEntries; i++) {
        if (pCcTable[i].regDmnWirelessMode == regDmnWirelessMode) {
            for (j = 0; j < pCcTable[i].entries; j++) {
                if ((pCcTable[i].chanGroup[j].lowChannel >= lowChannel) && 
                    (pCcTable[i].chanGroup[j].highChannel <= highChannel))
                    /* TODO: Account for no outdoor channels? */
                {
                    pCountryInfo->subBand[j].firstChannel = 
                        wlanConvertGHztoCh(pCcTable[i].chanGroup[j].lowChannel, channelFlags);
                    pCountryInfo->subBand[j].numChannel = 
                        ((pCcTable[i].chanGroup[j].highChannel - pCcTable[i].chanGroup[j].lowChannel) /
                        channelSpread) + 1;
                    pCountryInfo->subBand[j].maxTxPwr = pCcTable[i].chanGroup[j].powerDfs;
                }
            }
            break;
        }
    }
    ASSERT(i < regDmnEnumEntries);
    pCountryInfo->len = sizeof(pCountryInfo->countryString) +
        (sizeof(A_UINT8) * 3 * j);

#if 0
    uiPrintf("--Country Code of Beacon--\n");
    uiPrintf("RAW WORDS\n");
    for (i = 0; i < sizeof(dummyCountryInfoList)/sizeof(A_UINT32); i++) {
        uiPrintf("\t0x%08X\n", ((A_UINT32 *)(&dummyCountryInfoList))[i]);
    } 
    uiPrintf("Fields\n");
    uiPrintf("\tLength: %d\n", pCountryInfo->len);
    uiPrintf("\tCountry: %c%c%c\n", pCountryInfo->countryString[0],
        pCountryInfo->countryString[1],
        pCountryInfo->countryString[2]);
    for (i = 0; i < j; i++) {
        uiPrintf("\tFirst Channel %3d\tNum Channels %3d\tChannel Power %3d\n",
        pCountryInfo->subBand[i].firstChannel,
        pCountryInfo->subBand[i].numChannel,
        pCountryInfo->subBand[i].maxTxPwr);
    }
#endif

    return;
}

/***********************************************************
 * wlanGetCountryCodeByRegDmnName
 *
 * Search the regulatory table by name and return the country code
 */
A_UINT16
wlanGetCountryCodeByRegDmnName(A_UCHAR *regName)
{
    int i;

    /* Search through Regulatory Domain (Enum) table for a match */
    for (i = 0; i < sizeof(allLegacyNames)/sizeof(REG_DMN_LEGACY_NAMES); i++) {
        if (strcmp(allLegacyNames[i].regDmnName, regName) == 0) {
            return allLegacyNames[i].countryCode;
        }
    }
    return CTRY_INVALID;
}

/***********************************************************
 * wlanGetRegDmnByName
 *
 * Search the regulatory table by name and return the domain
 */
A_UINT16 
wlanGetRegDmnByName(A_UCHAR *regName)
{
    int i;

    /* Search through Regulatory Domain (Enum) table for a match */
    for (i = 0; i < sizeof(allLegacyNames)/sizeof(REG_DMN_LEGACY_NAMES); i++) {
        if (strcmp(allLegacyNames[i].regDmnName, regName) == 0) {
            return allLegacyNames[i].regDmn;
        }
    }
    /* Previous call to get country code should have stopped this call */
    ASSERT(0);
    return FCC;
}

/***********************************************************
 * wlanPrintRegDmnList
 *
 * Fill the given string with all reg domain names space separated
 */
void
wlanPrintRegDmnList(void)
{
    A_UINT16 i;

    for (i = 0; i < sizeof(allLegacyNames)/sizeof(REG_DMN_LEGACY_NAMES); i++) {
        uiPrintf("%s ", allLegacyNames[i].regDmnName);
    }
    uiPrintf("\n");
}

/***********************************************************
 * wlanConvertGHztoCh
 *
 * Convert 2.4 & 5 GHz frequency to IEEE channel number
 *
 * 11B range uses 1 thru 14 with 14 being a special case
 *
 *
 */
A_UINT8
wlanConvertGHztoCh(A_UINT16 freq, A_UINT16 channelFlags)
{
    A_UINT32 channel;

    if (channelFlags & CHANNEL_2GHZ) {
        // 2 GHz band
        if ( freq == 2484 ) {
            channel = 14;
        } else if ( freq < 2484 ) {
            channel =  ( freq - 2407  ) / 5;
        } else {
            // additional 2 GHz channels (15-26)
            channel = (( freq - 2512 ) / 20) + 15;
        }
    } else {
        // 5 GHz band, channel 0 = 5 GHz, increments of 5 MHz
        channel = ( freq - 5000 ) / 5;
    }
    return (A_UINT8) channel;
}


/***********************************************************
 * wlanConvertChtoGHz
 *
 * Convert 2 & 5 GHz channel number to frequency
 *
 *
 */
A_UINT16 wlanConvertChtoGHz(A_UINT8 ch, A_UINT16 channelFlags)
{
    A_UINT32 freq;

    if (channelFlags & CHANNEL_2GHZ) {
        if ( ch == 14 ) {
            freq = 2484;
        } else if ( ch < 14 ) {
            freq = 2407 + ch * 5;
        } else {
            // additional 2 GHz channels (15-26)
            freq = 2512 + ((ch-15) * 20);
        }
    } else {
        // 5 GHz band
        freq = 5000 + ch * 5;
    }

    return (A_UINT16) freq;
}

/***********************************************************
 * wlanEnableDFS
 *
 * Check the EEPROM set regulatory domain and enable DFS if
 * required for operation.
 *
 * RETURNS: TRUE if DFS is enabled
 */
A_BOOL
wlanEnableDFS(WLAN_DEV_INFO *pDevInfo)
{
    A_UINT16 wirelessModeRegDmn;

    if (pDevInfo->staConfig.phwChannel->channelFlags & CHANNEL_2GHZ) {
        return FALSE;
    }

    wirelessModeRegDmn = wlanGetWirelessModeRegDmn(pDevInfo, 
        pDevInfo->staConfig.countryCode,
        pDevInfo->staConfig.phwChannel->channelFlags);
    if ((wirelessModeRegDmn & 0xFF) != ETSI) {
        return FALSE;
    }

#ifdef DEBUG
    uiPrintf("wlanEnableDFS: Enabling radar detection features\n");
#endif
#ifdef BUILD_AP
    etsiFeaturesEnable();
#endif
    return TRUE;
}

/***********************************************************
 * wlanIsActiveScanAllowed
 *
 * Return the scantype allowed for the given ERD
 */
A_BOOL
wlanIsActiveScanAllowed(WLAN_DEV_INFO *pDevInfo)
{
    A_UINT16 wirelessModeRegDmn;

    ASSERT(pDevInfo->staConfig.phwChannel);
    if (pDevInfo->staConfig.phwChannel->channelFlags & CHANNEL_2GHZ) {
        return TRUE;
    }
    wirelessModeRegDmn = wlanGetWirelessModeRegDmn(pDevInfo, 
        pDevInfo->staConfig.countryCode,
        pDevInfo->staConfig.phwChannel->channelFlags);
    if ((wirelessModeRegDmn & 0xFF) != ETSI) {
        return TRUE;
    }
    return FALSE;
}

/***********************************************************
 * wlanIsNfCheckRequired
 *
 * Return the scantype allowed for the given ERD
 */
A_BOOL
wlanIsNfCheckRequired(WLAN_DEV_INFO *pDevInfo)
{
    A_UINT16 wirelessModeRegDmn;

    ASSERT(pDevInfo->staConfig.phwChannel);
    wirelessModeRegDmn = wlanGetWirelessModeRegDmn(pDevInfo, 
        pDevInfo->staConfig.countryCode,
        pDevInfo->staConfig.phwChannel->channelFlags);
    if ((wirelessModeRegDmn & 0xFF) == MKK) {
        return TRUE;
    }
    return FALSE;
}

/***********************************************************
 * wlanGetWirelessMode
 *
 * Return the radio's frequency specifications
 */
A_UINT16
wlanGetWirelessMode(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode)
{
#if !defined(BUILD_AP) || defined(DEBUG)
    A_UINT16 lowChannel = 0, highChannel= 0;
#endif
    A_UINT16 turbo = 0, i;
    A_UINT16 wirelessModeRegDmn;

    /* Return minimal config if the device is not attached to get config items up */
    if (pDevInfo == NULL) {
        return MODE_SELECT_11A;
    }

    wirelessModeRegDmn = wlanGetWirelessModeRegDmn(pDevInfo, 
        countryCode, CHANNEL_T);
    for (i = 0; i < sizeof(regDmnWm5Turbo) / sizeof(REG_DMN_WM_FREQ_TABLE); i++) {
        if (wirelessModeRegDmn == regDmnWm5Turbo[i].regDmnWirelessMode) {
            turbo = 1;
            break;
        }
    }

#if !defined(BUILD_AP) || defined(DEBUG)
    /* Determine by the underlying hardware */
    if (halGetChannelEdges(pDevInfo, CHANNEL_2GHZ, &lowChannel, &highChannel) == TRUE) {
        return MODE_SELECT_11A | (turbo ? MODE_SELECT_TURBO : 0) | MODE_SELECT_11B | MODE_SELECT_PUREG;
    }
#endif

    return MODE_SELECT_11A | (turbo ? MODE_SELECT_TURBO : 0);
}

/*******************************************************************************
 * AP Scan and related functions
 */

/* Scan time in milliSeconds */
enum {
    AP_ACTIVE_SCAN_MSTIME = 300,
    AP_PASSIVE_SCAN_MSTIME = 4000
};

/*******************************************************************************
 * bssDescCmp
 *
 * Sort the BSS descriptor elements by their channel
 * - used by qsort in bssDescShow
 */
static int
bssDescCmp(BSSDESCR *p1, BSSDESCR *p2)
{
    if (p1->pChannel->channel > p2->pChannel->channel) {
        return +1;
    }
    else if (p1->pChannel->channel == p2->pChannel->channel) {
        return 0;
    }
    return -1;
}

/*******************************************************************************
 * bssDescShow
 *
 * Sort and Print the BSS Description set
 */
static void
bssDescShow(BSSDESCR_SET *pBssSet)
{
#ifndef VXWORKS
    aPrintk("WE SHOULDN'T BE CALLING THIS...\n");
#else
    BSSDESCR    *pbssDescr;    /* BSS Description set */
    int         index, ap = 0, adhoc = 0;
    char        buf[ELE_SSID_SIZE+1];
        
    /* Sort the BSS descriptor elements by their channel */
    qsort(&(pBssSet->bssDescrArray[0]),
          pBssSet->count,
          sizeof (BSSDESCR),
          (void *)&bssDescCmp);

    /* Print the BSS descriptors and their info */
    uiPrintf("BSS Type        Channel  RSSI  BSSID              SSID\n");
    for (index = 0; index < pBssSet->count; index++) {
        pbssDescr = &pBssSet->bssDescrArray[index];
        if (pbssDescr->bsstype == INFRASTRUCTURE_BSS) {
            ap++;
        } else {
            adhoc++;
        }
        uiPrintf("%14s   %d.%03d   %3d   ",
                 pbssDescr->bsstype == INFRASTRUCTURE_BSS ? "Infrastructure" : "Ad-Hoc",
                 pbssDescr->pChannel->channel / 1000, pbssDescr->pChannel->channel % 1000,
                 pbssDescr->rssi);
        printMacAddress(pbssDescr->bssid);
        A_BCOPY (pbssDescr->ssid.ssid, buf, pbssDescr->ssid.length);
        buf[pbssDescr->ssid.length] = 0;
        uiPrintf("  %s\n", buf);
    }
    uiPrintf("AP: %d, Ad-Hoc: %d. Total BSS: %d\n", ap,
             adhoc, pBssSet->count);
#endif /* VXWORKS */
}

#ifdef BUILD_AP
/*******************************************************************************
 * wlanScanMode
 *
 * Initialize the scan followed by scanning the scanInfo channels
 * using mlmeScanFunction
 */
static A_STATUS
wlanScanMode(WLAN_DEV_INFO *pDevInfo, SCANTYPE scanType)
{
    SCAN_INFO      *pScanInfo;
    SSID           *pSsid;
    int            resultCode;
    A_INT32        maxChanTicks, maxChanTime;
    WLAN_SCAN_LIST *pScanList = NULL;
    A_UCHAR        wirelessModeName[30];

    pSsid = (SSID *)"";

    /* maxChanTime in milliseconds */
    if (scanType == ACTIVE_SCAN) {
        maxChanTime = AP_ACTIVE_SCAN_MSTIME;
    } else  {
        if (pDevInfo->staConfig.radarDetectEnable) {
            maxChanTime = radarScanTime * 1000;
        } else {
            maxChanTime = AP_PASSIVE_SCAN_MSTIME;
        }
    }
    maxChanTicks = (maxChanTime * sysClkRateGet())/1000;

    /* TODO: commonize these strings for web and other printing */
    switch (pDevInfo->staConfig.pClist->chanArray[0].channelFlags & CHANNEL_ALL) {
    case CHANNEL_A:
        strcpy(wirelessModeName, "5 GHz 54Mbps (802.11a)");
        break;
    case CHANNEL_T:
        strcpy(wirelessModeName, "5 GHz 108Mbps (Turbo)");
        break;
    case CHANNEL_B:
        strcpy(wirelessModeName, "2.4GHz 11Mbps (802.11b)");
        break;
    case CHANNEL_PUREG:
        strcpy(wirelessModeName, "2.4GHz 54Mbps");
        break;
    }

    /* collect all parameters required for mlmeScanRequest */
    wlanInitScanList(pDevInfo->staConfig.pClist, &pScanList);
    uiPrintf("\n%s scanning %s channels. This will take %d seconds\n",
        scanType == PASSIVE_SCAN ? "Passive" : "Active", wirelessModeName,
        (pScanList->listSize * maxChanTime + 999)/1000);
    resultCode = wlanMlmeScanRequest(pDevInfo, pDevInfo->staConfig.bssType,
                                     &broadcastBssid,
                                     pSsid, scanType, 1,
                                     pScanList, maxChanTime,
                                     maxChanTime, TRUE);
    if (resultCode) {
        uiPrintf("Scan Request result %x\n", resultCode);
        return A_EINVAL;
    }

    pScanInfo = pDevInfo->pscanInfo;
    ASSERT(pScanInfo);

    /* Scan from current channel index through listSize */
    for (; pScanInfo->curChannelIndex < pScanList->listSize; pScanInfo->curChannelIndex++) {
        if (isApShutdown() == TRUE) {
            exit(0);
        }
        taskDelay(maxChanTicks);
        /* turn off any more frames coming in */
        halRxFilter(pDevInfo, 0xffffffff, HAL_RX_FILTER_CLEAR);
        /* give up CPU for a fraction of a second. allow received beacons to get processed. */
        taskDelay(5);
        if (pScanList->pChanArray[pScanInfo->curChannelIndex]->channelFlags & (CHANNEL_CW_INT | CHANNEL_RADAR_INT)) {
            uiPrintf(" Channel Interference (radar or carrier wave) detected on channel %d MHz\n", 
                pScanList->pChanArray[pScanInfo->curChannelIndex]->channel);
        }
        if (pDevInfo->staConfig.radarDetectEnable) {
            initRadarDetect(pDevInfo);
        }
        /* Scan curChannelIndex */
        mlmeScanFunction(pDevInfo);
    }
    wlanMlmeScanConfirm(pDevInfo, &pScanInfo->bssSet, MLME_SUCCESS);

    return A_OK;
}

/*******************************************************************************
 * apScan
 *
 *  Scan all available channels in search for AP
 *  If scanType is passive scan, then listen for beacon
 *  If scanType is active scan, then send probe request and listen for reply
 */
A_STATUS
apScan(WLAN_DEV_INFO *pDevInfo, APSCAN_OP scanOp, SCANTYPE scanType)
{
    SCAN_INFO         *pScanInfo;
    A_STATUS          status;
    CHAN_VALUES       *pNewChannel = NULL;
    A_UINT16          intMode = 0;
    WLAN_CHANNEL_LIST *pMainCList = NULL, *pIntCList = NULL;
    BSSDESCR_SET      MainBssSet;
    BSSDESCR_SET      *pIntBssSet = NULL;
    int               i;

    if (scanOp != AP_SCAN_SELECT_CHAN) {
        /* stop the eth drivers first */
        for (i=0; planCookie[i]; i++) {
            muxDevStop(planCookie[i]);
        }

        /* let traffic drain */
        taskDelay(sysClkRateGet());

        /* Cancel the Beacon watchdog and Beacons*/
        beaconWdgCancel();
        halDisableInterrupts(pDevInfo, HAL_INT_SWBA);
    }

    status = wlanScanMode(pDevInfo, scanType);
    if (status != A_OK) {
        uiPrintf("Um... scanMan problem\n");
        ASSERT(0);
    }

    /* Determine interference mode */
    switch(pDevInfo->staConfig.pClist->chanArray[0].channelFlags & CHANNEL_ALL) {
    case CHANNEL_A:
        intMode = (halGetTurboDisable(pDevInfo)) ? 0 : MODE_SELECT_TURBO;
        break;
    case CHANNEL_T:
        intMode = MODE_SELECT_11A;
        break;
    case CHANNEL_B:
#ifdef DEBUG
        uiPrintf("apScan: WARNING: 802.11B apScan currently uses incorrect channel criteria\n");
#endif
        intMode = MODE_SELECT_PUREG;
        break;
    case CHANNEL_PUREG:
#ifdef DEBUG
        uiPrintf("apScan: WARNING: 802.11B apScan currently uses incorrect channel criteria\n");
#endif
        intMode = MODE_SELECT_11A;
        break;
    default:
        ASSERT(0);
        break;
    }
    pMainCList = pDevInfo->staConfig.pClist;

    /* Save the main wireless mode scan list */
    A_BCOPY((char *)&pDevInfo->pscanInfo->bssSet, (char *)&MainBssSet, sizeof(BSSDESCR_SET));
    A_MEM_ZERO(&pDevInfo->pscanInfo->bssSet, sizeof(BSSDESCR_SET));

    if (intMode) {
        /* Scan any interfering wireless mode - if possible */
        /*
         * Some list "acrobatics" are required here as the scan tools expect
         * to use the staConfig channel list and the scanInfo BSS list.  Hence
         * we "pointer save" the channel list and have to copy the BSS list from
         * the first scan.
         */
        pMainCList = pDevInfo->staConfig.pClist;
        pDevInfo->staConfig.pClist = NULL;
        if (wlanInitChannelList(pDevInfo, &(pDevInfo->staConfig.pClist), pDevInfo->staConfig.countryCode,
            intMode, apCfgOutdoorChanEnabledGet()) != A_OK)
        {
#ifdef DEBUG
            uiPrintf("Auto Channel Scan: Interfering wireless mode cannot be scanned\n");
#endif
            pDevInfo->staConfig.pClist = pMainCList;
        } else {
            /* Scan interfering mode */
            status = wlanScanMode(pDevInfo, scanType);
            if (status != A_OK) {
                uiPrintf("Um... scanMan problem on interferer's mode\n");
                ASSERT(0);
            }
            pIntCList = pDevInfo->staConfig.pClist;
            pDevInfo->staConfig.pClist = pMainCList;
            pIntBssSet = &pDevInfo->pscanInfo->bssSet;
        }
    }

    if (scanOp != AP_SCAN_SELECT_CHAN) {
        /* Re-enable beacons */
        wlanBeaconFree(pDevInfo);
        wlanBeaconInit(pDevInfo, 0);
        beaconWdgStart(pDevInfo);
        /* restart the eth drivers */
        for (i = 0; planCookie[i]; i++) {
            muxDevStart(planCookie[i]);
        }
    }

    halRxFilter (pDevInfo, HAL_RX_UCAST, HAL_RX_FILTER_SET);

    /* pScanInfo is not malloc'ed until after the wlanScanMode is run */
    pScanInfo = pDevInfo->pscanInfo;
    ASSERT(pScanInfo);

    if (scanOp == AP_SCAN_ALL_BSS) {
        /*
         * Finish with the last channel on the list.
         * Clear the state. Don't let any more bss's creep in
         */
        if (MainBssSet.count > 0) {
            uiPrintf("=> BSS'es from the selected wireless mode <=\n");
            bssDescShow(&MainBssSet);
        }
        if (pIntBssSet && (pIntBssSet->count > 0)) {
            uiPrintf("=> BSS'es from the interfering wireless mode <=\n");
            bssDescShow(pIntBssSet);
        }
        if (pIntCList) {
            wlanFreeChannelList(pIntCList);
        }
        return A_OK;
    }

    pNewChannel = wlanFindChannel(pDevInfo, &MainBssSet, pMainCList, pIntBssSet);
    if (scanOp == AP_SCAN_CLEAN_CHAN) {
        if (pNewChannel) {
            uiPrintf("Auto Channel Scan would select channel %d as the best channel\n", 
                pNewChannel->channel);
        } else {
            uiPrintf("Auto Channel Scan did not detect a good channel\n");
        }
        if (pIntCList) {
            wlanFreeChannelList(pIntCList);
        }
        return A_OK;
    }
    if (pNewChannel && (pNewChannel != pDevInfo->staConfig.pChannel)) {
        uiPrintf("Auto Channel Scan selected channel %d\n", pNewChannel->channel);
        pDevInfo->staConfig.pChannel = pNewChannel;
        pDevInfo->bssDescr->pChannel = pNewChannel;
        apCfgRadioChannelSet(pNewChannel->channel);
    }
    if (pIntCList) {
        wlanFreeChannelList(pIntCList);
    }

    /* Retune to the configured channel */
    chipReset(pDevInfo, TRUE, pDevInfo->staConfig.pChannel);

    if (pDevInfo->staConfig.radarDetectEnable) {
        initRadarDetect(pDevInfo);
    }

    return A_OK;
}
#endif

/* 
 * The defines below increase or decrease the cost accounting that selects
 * a channel for initial placement
 */
enum ChanCostEnum {
    CHAN_COST_BASE_NEXT_TO_TURBO = 10,
    CHAN_COST_NEW_GEN            = -20,
    CHAN_COST_ADJACENT           = 18,
    CHAN_COST_MIRROR             = 50,
    CHAN_COST_UNUSABLE           = 512
} CHAN_COST_ENUM;

#define CHAN_COST_TP_MULTIPLIER    1.5

#define PRINTCOST(x, y, z)                                               \
    { int a;                                                             \
    for(a = 0; a < (z); a++)                                             \
       uiPrintf(" -- Channel %d Cost %d\n", (y)[a].channel, (x)[a].cost);  \
    }
#define A_MAX(x, y)     (((x) > (y)) ? (x) : (y))

typedef struct ChannelListCost {
    A_INT16 cost;
    A_INT16 maxCost;
    A_INT16 numBadGuys;
} CHANNEL_LIST_COST;

/*******************************************************************************
 * wlanAddToCost
 *
 * Adds the given cost into the Cost List and updates the max and number of
 * interferers.
 */
static void
wlanAddToCost(CHANNEL_LIST_COST *pCListCost, A_UINT16 index, A_INT16 cost)
{
    if (cost > 0 ) {
       pCListCost[index].numBadGuys++;
    }
    if (cost > pCListCost[index].maxCost) {
        /* New signal is bigger - set it as the main interfering cost */
        pCListCost[index].maxCost = cost;
    }
}

#define WLAN_IS_OVERLAP(channel1, spread1, channel2, spread2)                   \
    (                                                                           \
    ((((channel2) - ((spread2) / 2) + 1) > ((channel1) - ((spread1) / 2))) &&   \
     (((channel2) - ((spread2) / 2) + 1) < ((channel1) + ((spread1) / 2)))) ||  \
    ((((channel2) + ((spread2) / 2) - 1) > ((channel1) - ((spread1) / 2))) &&   \
     (((channel2) + ((spread2) / 2) - 1) < ((channel1) + ((spread1) / 2))))     \
    )

/*******************************************************************************
 * wlanFindChannel
 *
 * Given the bss/channel list, the interferer bss/channel list, the configured
 * channel - find an acceptable runtime channel.
 *
 * A channelCost list is created.  The given channel statistics
 * will increase (less desirable) and decrease (more desirable) 
 */
/* TODO: Need to provide channelSpread and cross-channel signal bleed for 2.4 channels */
CHAN_VALUES *
wlanFindChannel(WLAN_DEV_INFO *pDevInfo, BSSDESCR_SET *pBssSet, WLAN_CHANNEL_LIST *pCList,
    BSSDESCR_SET *pIntBssSet)
{
#ifdef FADI_TEMP
    CHANNEL_LIST_COST *pCListCost;
    CHAN_VALUES       *pNewChannel = NULL;
    A_UINT16          i, j;
    A_UINT16          channelSpread, intSpread, bssChannel, bssRssi;
    int               winningLowChannel, numLowestCost, lowestCost;

    /* Allocate cost-function list holder to match channel list size */
    pCListCost = (struct ChannelListCost *) 
        A_DRIVER_MALLOC(sizeof(struct ChannelListCost) * pCList->listSize, 0, 0);
    A_MEM_ZERO(pCListCost, sizeof(struct ChannelListCost) * pCList->listSize);

    /* 
     * Initialize Cost
     * Subtract cost for the relative power levels - give higher power channels preference 
     * Only power levels up to 23 dBm are counted
     */
    /* Add cost for carrier wave or radar detection present to remove the channel from the list */
    for (i = 0; i < pCList->listSize; i++) {
        if((pCList->chanArray[i].channel < 5120) || (pCList->chanArray[i].channel > 5430)) {
            /* Increase cost of using non 1st gen channels */
            pCListCost[i].cost = CHAN_COST_NEW_GEN;
        } else {
            pCListCost[i].cost = (A_UINT16)(wlanGetChannelPower(pDevInfo, &(pCList->chanArray[i])) *
                (-CHAN_COST_TP_MULTIPLIER));
        }
        if (pCList->chanArray[i].channelFlags & (CHANNEL_CW_INT | CHANNEL_RADAR_INT)) {
            pCListCost[i].cost = CHAN_COST_UNUSABLE;
        }
    }

    ASSERT(pBssSet);
    channelSpread = wlanGetChannelSpread(pCList->chanArray[0].channelFlags);
    /* Add cost for detected BSS'es */
    for (j = 0; j < pBssSet->count; j++) {
        bssChannel = pBssSet->bssDescrArray[j].pChannel->channel;
        bssRssi = (A_UINT16)pBssSet->bssDescrArray[j].rssi;
        for (i = 0; i < pCList->listSize; i++) {
            if (pCList->chanArray[i].channel == bssChannel) {
                /* Add this BSS'es cost to its channel */
                wlanAddToCost(pCListCost, i, bssRssi);
            }
            if ((bssChannel == (pCList->chanArray[i].channel + channelSpread)) ||
                (bssChannel == (pCList->chanArray[i].channel - channelSpread)))
            {
                /* Add this BSS'es cost to any adjacent channels */
                wlanAddToCost(pCListCost, i, A_MAX(bssRssi - CHAN_COST_ADJACENT, 0));
            }
            /* Add this BSS'es Cost to any possible mirrors */
            if (pCList->chanArray[i].channelFlags & CHANNEL_5GHZ) {
                if ((bssChannel == (pCList->chanArray[i].channel + 80)) ||
                   (bssChannel == (pCList->chanArray[i].channel - 80)) ||
                   (bssChannel == (pCList->chanArray[i].channel + 160)) ||
                   (bssChannel == (pCList->chanArray[i].channel - 160)))
                {
                    wlanAddToCost(pCListCost, i, A_MAX(bssRssi - CHAN_COST_MIRROR, 0));
                }
            }
        }
    }

    /* Add Cost of channels and BSS'es from an interfering wireless mode */
    if (pIntBssSet != NULL) {
        for (j = 0; j < pIntBssSet->count; j++) {
            intSpread = wlanGetChannelSpread(pIntBssSet->bssDescrArray[j].pChannel->channelFlags);
            bssChannel = pIntBssSet->bssDescrArray[j].pChannel->channel;
            bssRssi = (A_UINT16)pIntBssSet->bssDescrArray[j].rssi;
            for (i = 0; i < pCList->listSize; i++) {
                if (WLAN_IS_OVERLAP(bssChannel, intSpread, pCList->chanArray[i].channel, channelSpread)) {
                    /* Add this BSS'es cost to any overlapping channel */
                    wlanAddToCost(pCListCost, i, bssRssi);
                }
                if (WLAN_IS_OVERLAP(bssChannel, intSpread, pCList->chanArray[i].channel + channelSpread, channelSpread) ||
                    WLAN_IS_OVERLAP(bssChannel, intSpread, pCList->chanArray[i].channel - channelSpread, channelSpread))
                {
                    /* Add this BSS'es cost to any adjacent channels */
                    wlanAddToCost(pCListCost, i, A_MAX(bssRssi - CHAN_COST_ADJACENT, 0));
                }
                /* Add this BSS'es Cost to any possible mirrors */
                if (pCList->chanArray[i].channelFlags & CHANNEL_5GHZ) {
                    if (WLAN_IS_OVERLAP(bssChannel, intSpread, pCList->chanArray[i].channel + 80, channelSpread) ||
                        WLAN_IS_OVERLAP(bssChannel, intSpread, pCList->chanArray[i].channel - 80, channelSpread) ||
                        WLAN_IS_OVERLAP(bssChannel, intSpread, pCList->chanArray[i].channel + 160, channelSpread) ||
                        WLAN_IS_OVERLAP(bssChannel, intSpread, pCList->chanArray[i].channel - 160, channelSpread))
                    {
                        wlanAddToCost(pCListCost, i, A_MAX(bssRssi - CHAN_COST_MIRROR, 0));
                    }
                }
            }

        }
    }

    lowestCost = CHAN_COST_UNUSABLE - 1;
    numLowestCost = 0;
    for (i = 0; i < pCList->listSize; i++) {
        /* Add the worst Cost plus 3 per additional interferer to the cost value */
        if (pCListCost[i].numBadGuys) {
            pCListCost[i].cost += (pCListCost[i].maxCost + ((pCListCost[i].numBadGuys - 1) * 3)); 
        }

        /* Find lowest cost and count how many are at this cost */
        if (pCListCost[i].cost == lowestCost) {
            numLowestCost++;
        }
        else if (pCListCost[i].cost < lowestCost) {
            lowestCost = pCListCost[i].cost;
            numLowestCost = 1;
        }
    }

    if (numLowestCost == 0) {
        uiPrintf("wlanFindChannel: WARNING - All channels are unusable\n");
#ifdef DEBUG
        PRINTCOST(pCListCost, pCList->chanArray, pCList->listSize)
#endif
        A_DRIVER_FREE(pCListCost, 0, sizeof(struct ChannelListCost) * pCList->listSize);
        return NULL;
    }

    /*
     * Randomize the set of low cost channels
     * for the case of multiple AP's booting at once
     */
    winningLowChannel = jiffies % numLowestCost;

    /* Return the winner */
    for (i = 0; i < pCList->listSize; i++) {
        if (pCListCost[i].cost == lowestCost) {
            if (winningLowChannel == 0) {
                pNewChannel = &(pCList->chanArray[i]);
                break;
            }
            winningLowChannel--;
        }
    }
    ASSERT(i != pCList->listSize);

#ifdef DEBUG
    uiPrintf("apScan Cost Histogram: higher values indicate more interference\n");
    /* Display cost histogram */
    for (i = 0; i < pCList->listSize; i++) {
        /* Normalize */
        pCListCost[i].cost -= lowestCost;
        uiPrintf("%5d ", pCListCost[i].cost);
    }
    uiPrintf("\n");
    for (i = 0; i < pCList->listSize; i++) {
        uiPrintf("%5d ", pCList->chanArray[i].channel);
    }
    uiPrintf("\n");
#endif

    A_DRIVER_FREE(pCListCost, 0, sizeof(struct ChannelListCost) * pCList->listSize);
    return pNewChannel;
#else
    return 0;
#endif

}


