/*
 *  Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * wlanproto.h - Data structures for 802.11 protocol
 *
 * $Id: //depot/sw/releases/2.0.1/src/include/wlanproto.h#1 $
 */

#ifndef __INCwlanprotoh
#define __INCwlanprotoh

/*
 * This file has the declarations specific to the 802.11 protocol.
 * It has all the fields and frames (that we will use) and also
 * mnemonic defines for constants. Although the MLME/PLME is part
 * of the protocol too, we have kept it in a separate file for
 * convenience.
 */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include "wlantype.h"

#ifdef WIN32
#pragma pack (push, 1)
#endif

/* PHY related definitions */

/* Predefined PHY types from dot11PHYType field of 802.11 MIB specification */
#define PHY_FHSS                1       // Frequency hopping
#define PHY_DSSS                2       // Direct Sequence Spread Spectrum (original 1, 2 Mbps)
#define PHY_IR                  3       // Infra Red Baseband
#define PHY_OFDM                4       // OFDM is not in the original MIB spec, but we use this value
#define PHY_HRDSSS              5       // High Rate DSSS, new 11b with extensions for 5.5 and 11 Mbps
#define PHY_CCK             PHY_HRDSSS  // 11b is commonly called as CCK in our code.

/* MAC related definitions */
/* MAC Address prefix allocated for Atheros devices */
#define Atheros_MACADDR0        0x00
#define Atheros_MACADDR1        0x03
#define Atheros_MACADDR2        0x7f
#define Atheros_MACADDR3        0x00
#define Atheros_MACADDR4        0x00
#define Atheros_MACADDR5        0x00

/* List of Element IDs */
#define ELE_SSID                0       /* SSID */
#define ELE_SUPPORTED_RATES     1       /* Supported Rates */
#define ELE_FH_PARAM_SET        2       /* FH Parameter set */
#define ELE_DS_PARAM_SET        3       /* DS Parameter set */
#define ELE_CF_PARAM_SET        4       /* CF Parameter set */
#define ELE_TIM                 5       /* TIM vector */
#define ELE_IBSS_PARAM_SET      6       /* IBSS parameter set */
#define ELE_COUNTRY_INFO        7       /* Country Information */    
#define ELE_CHLG_TEXT           16      /* Challenge text */
                                /* 17-31 reserved for challenge txt extension */
#define ELE_UCSE_SET            33      /* unicast cipher suite set */
#define ELE_MCSE_SET            34      /* multicast cipher suite set */

/* Max sizes of variable size information elements */
#define ELE_INFO_SIZE           255     /* Max possible size for any element */
#define ELE_SSID_SIZE           32      /* SSID 0 - 32 octets */
#define ELE_RATE_SIZE           8       /* Supported rates 1 - 8 */
#define ELE_BITMAP_SIZE         251     /* TIM bitmap */
#define ELE_CHLG_TEXT_SIZE      253     /* Challenge text size */
#define ELE_CF_PARAM_SIZE       6       /* CF param element size */
#define ELE_IBSS_PARAM_SIZE     2       /* IBSS param element size */
#define ELE_UCSE_SIZE           32      /* UCSE element size */
#define ELE_MCSE_SIZE           4       /* MCSE element size */
#define ELE_COUNTRY_SIZE        3       /* Country string size */


/* Frame control field of the MAC header */

/* values for the protocol version field */
#define PROTOCOL_VER_0          0       /* Standard specifies version 0 */

/* Values for frame type */
#define FRAME_MGT               0       /* Managemnt frame */
#define FRAME_CTRL              1       /* Control frame */
#define FRAME_DATA              2       /* Data frame */
#define FRAME_RESERVED          3       /* 3 is reserved value */

/* Values for frame subtype */
#define SUBT_ASSOC_REQ          0       /* Association request */
#define SUBT_ASSOC_RESP         1       /* Association response */
#define SUBT_REASSOC_REQ        2       /* Reassociation request */
#define SUBT_REASSOC_RESP       3       /* Reassociation response */
#define SUBT_PROBE_REQ          4       /* Probe request */
#define SUBT_PROBE_RESP         5       /* Probe response */
#define SUBT_BEACON             8       /* Beacon frame */
#define SUBT_ATIM               9       /* ATIM message */
#define SUBT_DISASSOC           10      /* Disassociation */
#define SUBT_AUTH               11      /* Authentication */
#define SUBT_DEAUTH             12      /* Deauthentication */

#define SUBT_PSPOLL             10      /* Power save poll */
#define SUBT_RTS                11      /* Request to Send */
#define SUBT_CTS                12      /* Clear to Send */
#define SUBT_ACK                13      /* ACK frame */
#define SUBT_CFEND              14      /* Contention-free end */
#define SUBT_CFENDACK           15      /* CF-End + CF-ACK */

#define SUBT_DATA               0       /* Just data */
#define SUBT_DATA_CFACK         1       /* Data + CF-Ack */
#define SUBT_DATA_CFPOLL        2       /* Data + CF-Poll */
#define SUBT_DATA_CFACK_CFPOLL  3       /* Data + CF-Ack + CF-Poll */
#define SUBT_NODATA             4       /* Null function, no data */
#define SUBT_NODATA_CFACK       5       /* No data + CF-Ack */
#define SUBT_NODATA_CFPOLL      6       /* No data + CF-Poll */
#define SUBT_NODATA_CFACK_CFPOLL 7      /* No data + CF-Ack + */
                                        /* CF-Poll */

/*
 * Define the frame size related constants for ethernet and 802.11 frames
 */
#define MAX_ETHER_FRAME_SIZE    1514    /* max length for Ethernet/802.3 frame */
#define ETHER_HDR_SIZE          14      /* Ethernet/802.3  header length */
#define MAX_ETHER_BODY_SIZE     1500    /* max body length for Ethernet/802.3 */
#define ETHER_TYPE_SIZE         2       /* size of type or length for Ethernet/802.3 */
#define MAX_WLAN_FRAME_SIZE     2342    /* max length for 802.11 frame */
#define WLAN_HDR_SIZE           24      /* length for 802.11 frame header without address 4 */
#define WLAN_HDR4_SIZE          30      /* length for 802.11 frame header with adress 4 */
#define MAX_WLAN_BODY_SIZE      2312    /* max length for 802.11 frame body */

#define FCS_FIELD_SIZE          4       /* size of FCS */
#define WEP_IV_FIELD_SIZE       4       /* wep IV field size */
#define AES_IV_FIELD_SIZE       4       /* aes IV field size */
#define WEP_ICV_FIELD_SIZE      4       /* wep ICV field size */
#define AES_ICV_FIELD_SIZE      8       /* AES ICV field size */
#define WEP_FIELDS_SIZE (WEP_IV_FIELD_SIZE + WEP_ICV_FIELD_SIZE) /* total size of wep fields needed */

/* minimum size of AES_OCB data frame length excluding FCS */
#define MIN_AES_OCB_DATA_LENGTH \
    (MAC_HDR_DATA3_SIZE + AES_IV_FIELD_SIZE + AES_ICV_FIELD_SIZE)


/* EAPOL/802.1x frames are identified by their EAPOL typeOrLen field in SNAP header. */
#define EAPOL_TYPEORLEN         0x888e

/* MAC header sizes in octets for various types of frames */
#define MAC_HDR_MGT_SIZE        24      /* All management frames */
#define MAC_HDR_DATA3_SIZE      24      /* Data frames with 3 addr */
#define MAC_HDR_DATA4_SIZE      30      /* Data frames with 4 addr */
                                        /* when ToDS and FromDS are set */
#define MAC_HDR_CTRL1_SIZE      10      /* Control frames with 1 addr */
                                        /* CTS, ACK types */
#define MAC_HDR_CTRL2_SIZE      16      /* Control frames with 2 addr */
                                        /* all other types */

/* Power management modes */
#define PS_MODE_ACTIVE          0       /* Station is active */
#define PS_MODE_POWER_SAVE      1       /* Station going in power save mode */

/* Type of wireless lan service - access point or station */
#define WLAN_STA_SERVICE        0       /* We are a Station */
#define WLAN_AP_SERVICE         1       /* We are Access Point */

#define WLAN_OK                 A_OK

/* MAC Address definitions */
#define WLAN_MAC_ADDR_SIZE      6
typedef union wlanMACAddr {
    A_UINT8  octets[WLAN_MAC_ADDR_SIZE];
    A_UINT16 words[WLAN_MAC_ADDR_SIZE/2];
    struct {
        A_UINT32 word;
        A_UINT16 half;
    } __ATTRIB_PACK st;
} WLAN_MACADDR;


#ifdef BIG_ENDIAN

typedef struct wlanFrameControlField {
    A_UINT16    order:1,                /* True for StrictOrdered service class */
                wep:1,                  /* True for WEP encrypted body */
                moreData:1,             /* True if more DUs are buffered */
                pwrMgt:1,               /* True if STA going into PS mode */
                retry:1,                /* True if frame is being retransmitted */
                moreFrag:1,             /* True if more fragments still to come */
                FromDS:1,               /* FromDS bit, frame coming from DS */
                ToDS:1,                 /* ToDS bit, frame going to DS */
                fSubtype:4,             /* frame subtype */
                fType:2,                /* frame type - contro, mgt, data */
                protoVer:2;             /* protocol version */
} FRAME_CONTROL_FIELD;

/*
 * Little Endian (802.11 order) swap16()ed version of the structure.
 * Used to avoid some swaps on big endian systems.
 */
typedef struct wlanFrameControlFieldLittle {
    A_UINT16    fSubtype:4,             /* frame subtype */
                fType:2,                /* frame type - contro, mgt, data */
                protoVer:2,             /* protocol version */
                order:1,                /* True for StrictOrdered service class */
                wep:1,                  /* True for WEP encrypted body */
                moreData:1,             /* True if more DUs are buffered */
                pwrMgt:1,               /* True if STA going into PS mode */
                retry:1,                /* True if frame is being retransmitted */
                moreFrag:1,             /* True if more fragments still to come */
                FromDS:1,               /* FromDS bit, frame coming from DS */
                ToDS:1;                 /* ToDS bit, frame going to DS */
} FRAME_CONTROL_FIELD_LITTLE;


#else /* Little_Endian */

typedef struct wlanFrameControlField {
    A_UINT16    protoVer:2,             /* protocol version */
                fType:2,                /* frame type - contro, mgt, data */
                fSubtype:4,             /* frame subtype */
                ToDS:1,                 /* ToDS bit, frame going to DS */
                FromDS:1,               /* FromDS bit, frame coming from DS */
                moreFrag:1,             /* True if more fragments still to come */
                retry:1,                /* True if frame is being retransmitted */
                pwrMgt:1,               /* True if STA going into PS mode */
                moreData:1,             /* True if more DUs are buffered */
                wep:1,                  /* True for WEP encrypted body */
                order:1;                /* True for StrictOrdered service class */
} FRAME_CONTROL_FIELD, FRAME_CONTROL_FIELD_LITTLE;

#endif /* Endianness */

/*
 * Duration/ID field is used in two different ways.
 * When bit 15 is 0, bits 0:14 are used to update NAV.
 * When bit 15 is 1 and bit 14 is 1, bits 0:13 are Association ID.
 * Bit 15 and 14 are 1 during CFP
 * Hence We define 2 different data types for the 2 octet duration
 */

#ifdef BIG_ENDIAN

typedef struct wlanDurationNAV {
    A_UINT16    clearBit:1,             /* Bit 15 must be 0 */
                duration:15;            /* 0-32767 duration value */
} DURATION_NAV;

typedef struct wlanDurationID {
    A_UINT16    bit15:1,                /* Must be 1 in this frame */
                bit14:1,
                value:14;               /* Bits 0 to 13 hold the value */
} DURATION_ID;

#else /* Little_Endian */

typedef struct wlanDurationNAV {
    A_UINT16    duration:15,            /* 0-32767 duration value */
                clearBit:1;             /* Bit 15 must be 0 */
} DURATION_NAV;

typedef struct wlanDurationID {
    A_UINT16    value:14,               /* Bits 0 to 13 hold the value */
                bit14:1,
                bit15:1;                /* Must be 1 in this frame */
} DURATION_ID;

#endif /* Endianness */

typedef WLAN_MACADDR BSSID;             /* BSSID is 48 bit field, same format */
                                        /* as MAC address, but not called MAC */
                                        /* address. Hence a separate definition */

#ifdef BIG_ENDIAN

typedef struct wlanSeqCtrl {
    A_UINT16    seqNo:12,               /* Sequence number 0 - 4095 */
                fragNo:4;               /* Fragment number */
} SEQ_CTRL;

typedef struct wlanCapabilityInfo {
    A_UINT16    reserved:8,            /* Reserved bits */
                agility:1,              /* Channel agility in use */
                pbcc:1,                 /* 1 = PBCC modulation allowed */
                shortPreamble:1,        /* 1 = Short preamble allowed, 0 = not */
                privacy:1,              /* Set by AP if WEP required */
                cfPollReq:1,            /* CF poll request */
                cfPollable:1,           /* CF poll ability of STA */
                ibss:1,                 /* Set for IBSS type network */
                ess:1;                  /* Set for ESS type network */
} CAP_INFO;

#else /* Little_Endian */

typedef struct wlanSeqCtrl {
    A_UINT16     fragNo:4,              /* Fragment number */
                 seqNo:12;              /* Sequence number 0 - 4095 */
} SEQ_CTRL;

typedef struct wlanCapabilityInfo {
    A_UINT16    ess:1,                  /* Set for ESS type network */
                ibss:1,                 /* Set for IBSS type network */
                cfPollable:1,           /* CF poll ability of STA */
                cfPollReq:1,            /* CF poll request */
                privacy:1,              /* Set by AP if WEP required */
                shortPreamble:1,        /* 1 = Short preamble allowed, 0 = not */
                pbcc:1,                 /* 1 = PBCC modulation allowed */
                agility:1,              /* Channel agility in use */
                reserved:8;            /* Reserved bits */
} CAP_INFO;

#endif /* Endianness */

typedef struct wlanTimeStamp {
    A_UINT32    low;                    /* Timestamp low order 32 bits */
    A_UINT32    high;                   /* Timestamp high order 32 bits */
} __ATTRIB_PACK WLAN_TIMESTAMP;


typedef struct wlanInfoElement {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT8     value[ELE_INFO_SIZE];   /* Max 255 byte array of octets */
} __ATTRIB_PACK INFO_ELEMENT;

typedef struct wlanSSID {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT8     ssid[ELE_SSID_SIZE];    /* Variable size 0-32 array of octets */
} __ATTRIB_PACK SSID;

typedef struct wlanRateSet {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT8     rates[ELE_RATE_SIZE];   /* Rates 1-8 array of octets */
} __ATTRIB_PACK RATE_SET;

typedef struct wlanFHParamSet {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT16    dwellTime;              /* Dwell time in TU */
    A_UINT8     hopSet;                 /* Hop set */
    A_UINT8     hopPattern;             /* Hop Pattern */
    A_UINT8     hopIndex;               /* Hop index selects current index in pattern */
} __ATTRIB_PACK FH_PARAM_SET;

typedef struct wlanDSParamSet {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT8     currentChannel;         /* 1 byte channel */
} __ATTRIB_PACK DS_PARAM_SET;

typedef struct wlanCFParamSet {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT8     cfpCount;               /* how many DTIMs before next CFP start */
    A_UINT8     cfpPeriod;              /* No.of DTIMs between CFPs */
    A_UINT16    cfpMaxDuration;         /* Max duration of CFP in TU */
    A_UINT16    cfpDurRemain;           /* Remaining duration */
} __ATTRIB_PACK CF_PARAM_SET;

#define VALID_CF_ELEMENT(_cfp)                          \
        (  ((_cfp)->elementID == ELE_CF_PARAM_SET)      \
        && ((_cfp)->length == ELE_CF_PARAM_SIZE)        \
        && ((_cfp)->cfpPeriod > 0) )

typedef struct wlanTIMElement {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT8     dtimCount;              /* How many beacons before next DTIM */
    A_UINT8     dtimPeriod;             /* No. of beacons between DTIMs */
    A_UINT8     bitmapControl;
    A_UINT8     bitmap[ELE_BITMAP_SIZE];        /* 251 bytes Partial virtual Bitmap */
} __ATTRIB_PACK TIM_ELEMENT;

typedef struct wlanIBSSParamSet {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT16    atimWindow;             /* ATIM window length in TU */
} __ATTRIB_PACK IBSS_PARAM_SET;

typedef struct wlanChlgText {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT8     chlgText[ELE_CHLG_TEXT_SIZE];   /* Challenge text, 1-253 byte */
} __ATTRIB_PACK CHLG_TEXT;

typedef struct wlanUcseSet {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT32    uCiphers[ELE_UCSE_SIZE/4];/* ciphers 1-63 array of 4 octets */
} __ATTRIB_PACK UCSE_SET;

#define VALID_UCSE_ELEMENT(_ucsep)                      \
        (((_ucsep)->elementID == ELE_UCSE_SET)          \
         && ((_ucsep)->length >= 4)                     \
         && ((_ucsep)->uCiphers[0] >= 0))

typedef struct wlanMcseSet {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT32    mCipher[ELE_MCSE_SIZE/4];/* ciphers 4 octets */
} __ATTRIB_PACK MCSE_SET;

#define VALID_MCSE_ELEMENT(_mcsep)                      \
        (((_mcsep)->elementID == ELE_MCSE_SET)          \
         && ((_mcsep)->length == ELE_MCSE_SIZE)         \
         && ((_mcsep)->mCipher[0] >= 0))

typedef struct wlanChannelSubBand {
    A_UINT8     firstChannel;           /* First channel */
    A_UINT8     numChannel;             /* Number of channels */
    A_UINT8     maxTxPower;             /* Max Transmit Power Level */
} __ATTRIB_PACK CHANNEL_SUBBAND;

typedef struct wlanCountryInfoSet {
    A_UINT8     elementID;              /* 1 byte element ID */
    A_UINT8     length;                 /* Length field */
    A_UINT8     country[ELE_COUNTRY_SIZE];   /* Country string array of octets */
    CHANNEL_SUBBAND subband;            /* One or more subband channel info */
} __ATTRIB_PACK COUNTRY_INFO_SET;

/**********************************************************/

/*
 * This section defines all types of 802.11 frames
 */

/* PLCP Header
 * Frames delivered by MAC to the host will have the PLCP
 * header before the actual MAC header begins.
 */

#ifdef BIG_ENDIAN

typedef struct wlanPLCPHeader {
    A_UINT32    tail:6,                 /* 6 tail bits */
                parity:1,               /* parity bit */
                length:12,              /* 12 bits length field */
                reserved:1,             /* Reserved bit */
                rate:4,                 /* 4 bits rate */
                rssi:8;                 /* receiver signal strength */
    A_UINT16    service;                /* 16 service bits, ignored */
} __ATTRIB_PACK WLAN_PLCP_HEADER;

#else /* Little_Endian */

typedef struct wlanPLCPHeader {
    A_UINT32    rssi:8,                 /* receiver signal strength */
                rate:4,                 /* 4 bits rate */
                reserved:1,             /* Reserved bit */
                length:12,              /* 12 bits length field */
                parity:1,               /* parity bit */
                tail:6;                 /* 6 tail bits */
    A_UINT16    service;                /* 16 service bits, ignored */
} __ATTRIB_PACK WLAN_PLCP_HEADER;

#endif /* Endianness */

/* Control frames, section 7.2.1 of the spec.
 * The MAC Header part of control frames varies, hence we
 * haven't declared a separate structure for it.
 */

typedef struct wlanFrameRTS {
    FRAME_CONTROL_FIELD frameCtrl;      /* Frame control */
    DURATION_NAV        durationNav;    /* Duration 0-32767 */
    WLAN_MACADDR        recAddr;        /* Receiver address */
    WLAN_MACADDR        transAddr;      /* Transmitter addr */
} __ATTRIB_PACK WLAN_FRAME_RTS;

typedef struct wlanFrameCTS {
    FRAME_CONTROL_FIELD frameCtrl;      /* Frame control */
    DURATION_NAV        durationNav;    /* Duration 0-32767 */
    WLAN_MACADDR        recAddr;        /* Receiver address */
} __ATTRIB_PACK WLAN_FRAME_CTS;

typedef struct wlanFrameACK {
    FRAME_CONTROL_FIELD frameCtrl;      /* Frame control */
    DURATION_NAV        durationNav;    /* Duration 0-32767 */
    WLAN_MACADDR        recAddr;        /* Receiver address */
} __ATTRIB_PACK WLAN_FRAME_ACK;

typedef struct wlanFramePspoll {
    FRAME_CONTROL_FIELD frameCtrl;      /* Frame control */
    DURATION_ID         durationId;     /* bits 0:10 AID */
    BSSID               bssId;          /* BSS ID */
    WLAN_MACADDR        transAddr;      /* Transmitter addr */
} __ATTRIB_PACK WLAN_FRAME_PSPOLL;

typedef struct wlanFrameCFEnd {
    FRAME_CONTROL_FIELD frameCtrl;      /* Frame control */
    DURATION_ID         durationId;     /* set to 0 */
    WLAN_MACADDR        recAddr;        /* Receiver address */
    BSSID               bssId;          /* BSS ID */
} __ATTRIB_PACK WLAN_FRAME_CFEND;

typedef struct wlanFrameCFEndAck {
    FRAME_CONTROL_FIELD frameCtrl;      /* Frame control */
    DURATION_ID         durationId;     /* set to 0 */
    WLAN_MACADDR        recAddr;        /* Receiver address */
    BSSID               bssId;          /* BSS ID */
} __ATTRIB_PACK WLAN_FRAME_CFENDACK;

/*
 * Data frames, section 7.2.2
 */

/* Maximum size of payload, i.e. frame body in data packet */
#define MAX_80211_FRAME_BODY_SIZE       MAX_WLAN_BODY_SIZE

/*
 * Size of empty packet, i.e. data packet with null body.
 * When you want to allocate a complete data packet, add the
 * payload size to EMPTY_80211_DATA_FRAME_SIZE and then
 * typecast it back to WLAN_FRAME_DATA. However, the FCS
 * field's offset depends on the data body size.
 */

#define EMPTY_80211_DATA_FRAME_SIZE     34

/* MAC header for the data packets is defined separately. */

/* 3 Address frame */
typedef struct wlanDataMacHeader3 {
    FRAME_CONTROL_FIELD frameControl;
    DURATION_NAV        durationNav;
    WLAN_MACADDR        address1;
    WLAN_MACADDR        address2;
    WLAN_MACADDR        address3;
    SEQ_CTRL            sequenceControl;
} __ATTRIB_PACK WLAN_DATA_MAC_HEADER3;

/* 3 Address frame with encryption fields */
typedef struct wlanFrame3DataEncrypt {
    WLAN_DATA_MAC_HEADER3 macHeader;
    A_UINT32            iv;             /* IV for encryption */
    A_UINT8             FrameBody[MAX_80211_FRAME_BODY_SIZE];
    A_UINT32            icv;            /* ICV for encryption */
} __ATTRIB_PACK WLAN_FRAME_3_DATA_ENCRYPT;

/* 4 Address frame from DS to DS */
/* This data structure is also not only the largest size frame
 * header but a superset of all possible frame headers.
 * Hence it can be safely used as a generic frame header called
 * WLAN_FRAME_HEADER.
 */
typedef struct wlanDataMacHeader {
    FRAME_CONTROL_FIELD frameControl;
    DURATION_NAV        durationNav;
    WLAN_MACADDR        address1;
    WLAN_MACADDR        address2;
    WLAN_MACADDR        address3;
    SEQ_CTRL            sequenceControl;
    WLAN_MACADDR        address4;
} __ATTRIB_PACK WLAN_DATA_MAC_HEADER, WLAN_FRAME_HEADER;

/*
 * Frame header with fields reversed so they can be accessed directly
 * w/o a swap.  Not all fields are availiable.
 */
typedef struct wlanDataMacHeaderLittle {
    FRAME_CONTROL_FIELD_LITTLE  frameControl;
    union {
        A_UINT16                data;
        DURATION_NAV            duration;
    } durationNav;
    WLAN_MACADDR                address1;
    WLAN_MACADDR                address2;
    WLAN_MACADDR                address3;
    union {
        A_UINT16                data;                   /* user has to swap */
        SEQ_CTRL                scf;
    } sequenceControl;
} __ATTRIB_PACK WLAN_FRAME_HEADER_LITTLE;

/* Maximum size data frame */
typedef struct wlanFrameData {
    WLAN_DATA_MAC_HEADER        macHeader;
    A_UINT8                     FrameBody[MAX_80211_FRAME_BODY_SIZE];
} __ATTRIB_PACK WLAN_FRAME_DATA;

/* Maximum size data frame witn encryption fields */
typedef struct wlanFrameDataEncrypt {
    WLAN_DATA_MAC_HEADER        macHeader;
    A_UINT32                    iv;     /* IV for encryption */
    A_UINT8                     FrameBody[MAX_80211_FRAME_BODY_SIZE];
    A_UINT32                    icv;    /* ICV for encryption */
} __ATTRIB_PACK WLAN_FRAME_DATA_ENCRYPT;

/* The standard Ethernet frame is also defined for convenience */

typedef struct lanFrameHeader {
    WLAN_MACADDR        destAddr;
    WLAN_MACADDR        srcAddr;
    A_UINT16            lanTypeOrLen;
} __ATTRIB_PACK LAN_FRAME_HEADER;

/* Maximum size LAN data frame */
typedef struct lanFrameData {
    LAN_FRAME_HEADER    lanHeader;
    A_UINT8             frameData[MAX_ETHER_BODY_SIZE];
} __ATTRIB_PACK LAN_FRAME_DATA;

typedef struct {
    LAN_FRAME_HEADER    lanHeader;
    A_UINT8             dsap;
    A_UINT8             ssap;
    A_UINT8             control;
    A_UINT8             xid[3];
} __ATTRIB_PACK L2_UPDATE_FRAME;

#define L2UF_LEN (sizeof(L2_UPDATE_FRAME) - sizeof(LAN_FRAME_HEADER))

/*
 * Management frames, section 7.2.3
 */

/*
 * MAC header for the management packets is defined separately.
 * since it is common to all of them.
 */

typedef struct wlanMgtMacHeader {
    FRAME_CONTROL_FIELD frameControl;
    DURATION_NAV        durationNav;
    WLAN_MACADDR        destAddr;
    WLAN_MACADDR        srcAddr;
    WLAN_MACADDR        bssId;
    SEQ_CTRL            sequenceControl;
} __ATTRIB_PACK WLAN_MGT_MAC_HEADER;

/*
 * The beacon frame is variable length. First 5 fields are
 * always present, but the information elements, i.e. sets
 * of different parameters are present only when the beacon
 * is generated by that type of AP. These sets can be variable
 * size. We just declare 'buffer' field as a place holder. The
 * algorithms should look at the first byte, i.e. element ID,
 * then parse rest of the element and look for the next one.
 */

typedef struct wlanFrameBeacon {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    WLAN_TIMESTAMP      timestamp;      /* 64 bit timestamp */
    A_UINT16            beaconInterval; /* Beacon interval in TU */
    CAP_INFO            capInfo;        /* AP capability info set */
    A_UCHAR             buffer[2];      /* Generic buffer. Variable size */
                                        /* data follows */
} __ATTRIB_PACK WLAN_FRAME_BEACON;


#define WLAN_BEACON_HEADER_LENGTH       36      /* The length for macheader, timestamp,
                                                 * beaconinterval and capinfo.  It is
                                                 * always constant for beacon frames.
                                                 */
/* Cipher Suite Selectors, defined such that they go out on the air as little endian */
#define CSE_AES_OCB                     0x007f0300
#define CSE_AES_OCB_REVA                0x017f0300
#define CSE_AES                         0x027f0300
#define CSE_TKIP                        0x107f0300
#define CSE_BASIC_WEP                   0x207f0300
#define CSE_NULL                        0xff7f0300
#define CSE_DEFAULT                     0xff7f0300

/* Status code field values in the management frames 802.11 Spec section 7.3.1.9 */

#define WLAN_SUCCESS                    WLAN_OK /* Successful */
#define WLAN_UNKNOWN_FAILURE            1       /* Unspecified failure */
#define WLAN_UNSUPPORTED_CAP            10      /* Can't support all requested */
                                                /* capabilities */
#define WLAN_UNCONFIRMED_ASSOC          11      /* Can't confirm current association */
#define WLAN_OUT_OF_SCOPE               12      /* Association denied due to a reason */
                                                /* outside the scope of 802.11 standard */
#define WLAN_UNSUPPORTED_ALG            13      /* Unsupported algorithm */
#define WLAN_OUT_OF_SEQUENCE            14      /* Authentication frame out of sequence */
#define WLAN_CHLG_FAILURE               15      /* Challenge failure in authentication */
#define WLAN_AUTH_TIMEOUT               16      /* Auth timeout waiting for next frame */
#define WLAN_TOOMANY_ASSOC              17      /* AP not able to handle more assoc */
#define WLAN_UNSUPPORTED_RATE           18      /* Requesting station doesn't support */
                                                /* all the rates in BSSBasicRateSet */
#define WLAN_NO_SHORT_PREAMBLE          19      /* Requesting station doesn't support short preamble */
#define WLAN_NO_PBCC                    20      /* Requesting station doesn't support PBCC */
#define WLAN_NO_CHANNEL_AGILITY         21      /* Requesting station doesn't support channel agility */

/* Reason Code values for Disassoc and Deauth frames 802.11 Spec section 7.3.1.7 */

#define REASON_RESERVED                 0       /* Reserved */
#define REASON_UNSPECIFIED              1       /* Unspecified reason */
#define REASON_AUTH_EXPIRED             2       /* Past authentication no longer valid */
#define REASON_AUTH_LEAVING             3       /* Sending station leaving BSS */
#define REASON_INACTIVE                 4       /* Disassocd due to inactivity */
#define REASON_TOOMANY_ASSOC            5       /* AP not able to handle all assoc's */
#define REASON_CLASS2                   6       /* Class 2 frame received from non-auth */
#define REASON_CLASS3                   7       /* Class 3 frame received from non-assoc */
#define REASON_ASSOC_LEAVING            8       /* Sending station leaving BSS */
#define REASON_NOT_AUTH                 9       /* Sender not auth'ed before requesting */
                                                /* association */
                                                /* 10 - 65535 reserved codes */

#define ASSOCID_MIN                     1       /* Min value of Association ID */
#define ASSOCID_MAX                     2007    /* Max value of Association ID */

typedef struct wlanFrameDisassoc {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    A_UINT16            reasonCode;     /* Reason for action */
} __ATTRIB_PACK WLAN_FRAME_DISASSOC;

/* Associate request frame. SSID and supported rates elements
 * are variable size. Hence, you should compute correct offset
 * to actual rates field. This structure assumes max size ssid
 * and rateSet fields.
 */
typedef struct wlanFrameAssocReq {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    CAP_INFO            capInfo;        /* capability information */
    A_UINT16            listenInterval; /* interval value in TU */
    SSID                ssid;           /* SSID, variable size */
    RATE_SET            rateSet;        /* rates element */
    UCSE_SET            ucseSet;        /* ucse element */
} __ATTRIB_PACK WLAN_FRAME_ASSOC_REQ;

/* Associate response frame. Supported rates elements
 * is variable size.
 */
typedef struct wlanFrameAssocResp {
        WLAN_MGT_MAC_HEADER macHeader;  /* Standard mgt header */
        CAP_INFO        capInfo;        /* capability information */
        A_UINT16        statusCode;     /* status code */
        A_UINT16        assocId;        /* association ID */
        RATE_SET        rateSet;        /* rates element */
        UCSE_SET        ucseSet;        /* ucse element */
        MCSE_SET        mcseSet;        /* mcse element */
} __ATTRIB_PACK WLAN_FRAME_ASSOC_RESP;


/* Reassociate request frame. SSID and supported rates elements
 * are variable size. Hence, you should compute correct offset
 * to actual rates field. This structure assumes max size ssid
 * field.
 */
typedef struct wlanFrameReassocReq {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    CAP_INFO            capInfo;        /* capability information */
    A_UINT16            listenInterval; /* interval value in TU */
    WLAN_MACADDR        curAPAddr;      /* Current AP Address */
    SSID                ssid;           /* SSID, variable size */
    RATE_SET            rateSet;        /* rates element */
    UCSE_SET            ucseSet;        /* ucse element */
} __ATTRIB_PACK WLAN_FRAME_REASSOC_REQ;

/* Frames for associate and reassociate response are the same */

typedef WLAN_FRAME_ASSOC_RESP   WLAN_FRAME_REASSOC_RESP;

/* Probe request frame. SSID and supported rates elements
 * are variable size. Hence, you should compute correct offset
 * to actual rates field. This structure assumes max size ssid
 * field.
 */
typedef struct wlanFrameProbeReq {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    SSID                ssid;           /* SSID, variable size */
    RATE_SET            rateSet;        /* rates element */
} __ATTRIB_PACK WLAN_FRAME_PROBE_REQ;

/*
 * The probe response frame is variable length. First 5 fields are
 * always present, but the information elements, i.e. sets
 * of different parameters are present only when the frame
 * is generated by that type of AP. These sets can be variable
 * size. We just declare 'buffer' field as a place holder. The
 * algorithms should look at the first byte, i.e. element ID,
 * then parse rest of the element and look for the next one.
 */

typedef struct wlanFrameProbeResp {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    WLAN_TIMESTAMP      timestamp;      /* 64 bit timestamp */
    A_UINT16            beaconInterval; /* Beacon interval in TU */
    CAP_INFO            capInfo;        /* AP capability info set */
    A_UCHAR             buffer[2];      /* Generic buffer. Variable size */
                                        /* data follows */
} __ATTRIB_PACK WLAN_FRAME_PROBE_RESP;

/* Two authentication algorithms are currently defined */
/* Use them in the field algNo of authentication frames */

#define AUTH_OPEN_SYSTEM        (A_UINT16)0     /* Open systems, no encryption */
#define AUTH_SHARED_KEY         (A_UINT16)1     /* Shared key, use WEP */
#define AUTH_SHARED_KEY_CHLG_SIZE       128     /* Section 8.1.2.2, 128 byte challenge */
                                                /* text for shared key authentication */
/*
 * Authentication frames
 */

/*
 * Structure for #1 and #4 frames, no encryption fields.
 */

typedef struct wlanFrameAuth {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    A_UINT16            algNo;          /* Algorithm number */
    A_UINT16            transSeqNo;     /* Authentication transaction */
                                        /* sequence number */
    A_UINT16            statusCode;     /* Status (result) code */
} __ATTRIB_PACK WLAN_FRAME_AUTH;

/*
 * Structure for #2 frame, challenge text, but no encryption or IV
 */

typedef struct wlanFrameAuthChlg {
        WLAN_MGT_MAC_HEADER macHeader;  /* Standard mgt header */
        A_UINT16            algNo;      /* Algorithm number */
        A_UINT16            transSeqNo; /* Authentication transaction */
                                        /* sequence number */
        A_UINT16            statusCode; /* Status (result) code */
        CHLG_TEXT           challenge;  /* Challenge text for #3 frame */
} __ATTRIB_PACK WLAN_FRAME_AUTH_CHLG;

/*
 * Structure for #3 frame, with challenge and iv/icv fields.
 */

typedef struct wlanFrameAuthEncrypt {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    A_UINT32            iv;             /* IV for encryption */
    A_UINT16            algNo;          /* Algorithm number */
    A_UINT16            transSeqNo;     /* Authentication transaction */
                                        /* sequence number */
    A_UINT16            statusCode;     /* Status (result) code */
    CHLG_TEXT           challenge;      /* Challenge text for #3 frame */
} __ATTRIB_PACK WLAN_FRAME_AUTH_ENCRYPT;

/* Deauthentication frame */

typedef struct wlanFrameDeauth {
    WLAN_MGT_MAC_HEADER macHeader;      /* Standard mgt header */
    A_UINT16            reasonCode;     /* Reason for action */
} __ATTRIB_PACK WLAN_FRAME_DEAUTH;

/* Many frame types to avoid most casts */
typedef union wlanFrameUnion {
    void                        *ptr;
    A_UCHAR                     *byte;
    A_UINT32                    *word;
    WLAN_FRAME_HEADER           *header;
    WLAN_FRAME_HEADER_LITTLE    *headerLittle;
    WLAN_DATA_MAC_HEADER3       *header3;
    WLAN_FRAME_3_DATA_ENCRYPT   *header3Iv;
    WLAN_FRAME_DATA_ENCRYPT     *header4Iv;
    WLAN_FRAME_PSPOLL           *psPoll;
    WLAN_FRAME_CFEND            *cfEnd;
    WLAN_FRAME_CFENDACK         *cfEndAck;
    WLAN_FRAME_ASSOC_REQ        *assoc;
    WLAN_FRAME_ASSOC_RESP       *assocResp;
    WLAN_FRAME_DISASSOC         *disAssoc;
    WLAN_FRAME_REASSOC_REQ      *reAssoc;
    WLAN_FRAME_REASSOC_RESP     *reAssocResp;
    WLAN_FRAME_PROBE_REQ        *probe;
    WLAN_FRAME_PROBE_RESP       *probeResp;
    WLAN_FRAME_AUTH             *auth;
    WLAN_FRAME_AUTH_CHLG        *authChal;
    WLAN_FRAME_AUTH_ENCRYPT     *authChalIv;
    WLAN_FRAME_DEAUTH           *deAuth;
    WLAN_FRAME_BEACON           *beacon;
    WLAN_MGT_MAC_HEADER         *mgtHeader;
    LAN_FRAME_HEADER            *lanHeader;
} WLAN_FRAME_UNION;

#ifdef WIN32
#pragma pack (pop)
#endif

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __INCwlanprotoh */
