#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/in.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/if_arp.h>
#include <linux/mroute.h>
#include <linux/init.h>
#include <linux/inetdevice.h>

#include <net/sock.h>
#include <net/ip.h>
#include <net/icmp.h>
#include <net/protocol.h>
#include <net/ipip.h>
#include <net/arp.h>
#include <net/checksum.h>
#include <net/inet_ecn.h>
#include <linux/netfilter_ipv4.h>

#define ARUBA_80211_FRAME 0x8211

#define GRE_OVERHEAD 4

static struct net_device *wlan_dev = NULL;

int aruba_rcv(struct sk_buff *, struct net_device *,
	      struct packet_type *pt);



struct gre_tunnel_t {
  struct iphdr iph;
};

static struct gre_tunnel_t *gre_tun = NULL;

int gre_set_tunnel_params(struct iphdr *iph){

  if (gre_tun)
    return -EINVAL;

  gre_tun = kmalloc(sizeof(struct gre_tunnel_t), GFP_KERNEL);
  if (gre_tun == NULL)
    return -ENOMEM;

  memcpy(&gre_tun->iph, iph, sizeof(*iph));
  
  return 0;

}


static int gre_rcv(struct sk_buff *skb){

  struct iphdr *iph = skb->nh.iph;
  u8 *ghdr = skb->data;
  u16 ptype = *(u16 *)(ghdr+2);

  if ((gre_tun == NULL) || (wlan_dev == NULL))
    goto drop_pkt;

  // do we need this?
  if (ptype != htons(ARUBA_80211_FRAME)){
    printk("gre: rx pkt with type %04x\n", ptype);
    goto drop_pkt;
  }

  if ((iph->saddr != gre_tun->iph.daddr) && 
      (iph->daddr != gre_tun->iph.saddr))
    goto drop_pkt;

  skb->mac.raw = skb->nh.raw;
  skb->nh.raw = __pskb_pull(skb, 4);
  memset(&(IPCB(skb)->opt), 0, sizeof(struct ip_options));
  if (skb->ip_summed == CHECKSUM_HW)
    skb->csum = 
      csum_sub(skb->csum,
	       csum_partial(skb->mac.raw, skb->nh.raw-skb->mac.raw, 0));
  
  aruba_rcv(skb, NULL, NULL);

  return 0;

 drop_pkt:
  kfree_skb(skb);
  return 0;

}


static void gre_err(struct sk_buff *skb, u32 info){
  printk("gre_err: skb %p, info %x\n", skb, info);
  return;
}

/* Need this wrapper because NF_HOOK takes the function address */
static inline int do_ip_send(struct sk_buff *skb){
  return ip_send(skb);
}


int gre_tx(struct sk_buff *skb){

  struct rtable *rt;
  struct iphdr *iph;
  struct net_device *odev;
  u16 *gre_hdr;
  int retval;

  if (gre_tun == NULL)
    goto tx_drop;

  if (ip_route_output(&rt, gre_tun->iph.daddr, gre_tun->iph.saddr, 0, 0))
    goto tx_drop;


  odev = rt->u.dst.dev;
  skb->h.raw = skb->nh.raw;
  
  // assume headroom exists for the IP + GRE headers (24 bytes)

  skb->nh.raw = skb_push(skb, sizeof(struct iphdr)+4);
  memset(&(IPCB(skb)->opt), 0, sizeof(struct ip_options));
  skb->dst = &rt->u.dst;
  
  iph = skb->nh.iph;
  iph->version = 4;
  iph->ihl = sizeof(struct iphdr) >> 2;
  iph->frag_off = 0;
  iph->protocol = IPPROTO_GRE;
  iph->tos = 0;
  iph->daddr = rt->rt_dst;
  iph->saddr = rt->rt_src;

  iph->ttl = 64;

  gre_hdr = (u16 *)(iph+1);
  gre_hdr[0] = 0;
  gre_hdr[1] = skb->protocol;
  
  skb->ip_summed = CHECKSUM_NONE;
  iph->tot_len = htons(skb->len);
  //  ip_select_ident(iph, &rt->u.dst, NULL);
  ip_send_check(iph);

  retval = NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, rt->u.dst.dev, do_ip_send);
  if ((retval != NET_XMIT_SUCCESS) && (retval != NET_XMIT_CN)){
    printk("gre_tx: ip_send returns %d\n", retval);
  }


  return 0;


 tx_drop:
  kfree_skb(skb);
  return 0;

}

static struct inet_protocol gre_proto = {
  gre_rcv,
  gre_err,
  0,
  IPPROTO_GRE,
  0,
  NULL,
  "GRE"
};

int gre_init(struct net_device *dev){

  wlan_dev = dev;

  inet_add_protocol(&gre_proto);

  printk("Registered GRE handler\n");

  return 0;

}

void gre_cleanup(void){

  if (inet_del_protocol(&gre_proto) < 0){
    printk("gre_cleanup: unable to remove GRE protocol\n");
  }

  wlan_dev = NULL;

  if (gre_tun)
    kfree(gre_tun);
  gre_tun = NULL;
  

  return;

}
