/*
 * Soft AP driver - based on the Host AP driver
 * Partha Narasimhan, Aruba Networks
 * 6/25/02
 *
 * Host AP (software wireless LAN access point) driver for Intersil Prism2
 * Copyright (c) 2001-2002, SSH Communications Security Corp
 * Jouni Malinen <jkm@ssh.com> and/or <jkmaline@cc.hut.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 *
 * FIX:
 * - shutting down pcmcia-cs while wlan# interface is up does not work
 *   (it does not remove the driver and eventually the kernel will panic when
 *   card is removed); when interface is first configured down, pcmcia-cs can
 *   be shut down; update: this seems to work at least with kernel-based
 *   pcmcia and Linux 2.4.13..
 * - SIOCSIWRATE can set TX rate for adhoc and managed modes, but this is not
 *   currently used in host-based TX rate control; it could be used also for
 *   this (i.e., allow AP to be configured to limit/force TX rate)
 * - there is currently no way of associating TX packets to correct wds device
 *   when TX Exc/OK event occurs, so all tx_packets and some
 *   tx_errors/tx_dropped are added to the main netdevice; using sw_support
 *   field in txdesc might be used to fix this (using Alloc event to increment
 *   tx_packets would need some further info in txfid table)
 *
 * Buffer Access Path (BAP) usage:
 *   BAP0 is used for sending data from the host computer to the card and for
 *   RID read/writes; it is protected with local->baplock (spin_lock_irqsave)
 *   and it must not be used from hardware interrupt context.
 *   BAP1 is used for receiving data from the card to the host computer; it is
 *   used only in hardware interrupt handler and kernel is assumed to not call
 *   the same handler simultaneously for the same interrupt even on SMP
 *   systems (this removes need for spin_lock protecting BAP1 access.
 *   Note! There seems to be some problems with BAP0 access being interrupted
 *   with RX/info frame (BAP1). To remove these problems, baplock is used with
 *   spin_lock_irqsave (it could be _bh otherwise) and RX/info handling locks
 *   baplock during BAP1 setup to make sure that BAP0 use has been completed.
 */


#include <linux/config.h>

#if defined(PRISM2_PCCARD) && defined(__IN_PCMCIA_PACKAGE__) 
#include <pcmcia/k_compat.h>
#endif 

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wireless.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>

#include <linux/delay.h>
#include <asm/io.h>
#include <linux/sched.h>
#include <linux/interrupt.h>

#include "prism2_config.h"
#if defined(PRISM2_PCCARD) && defined(__IN_PCMCIA_PACKAGE__) 
#include <pcmcia/config.h>
#endif

#ifdef PRISM2_PCCARD
#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>
#endif


#include "prism2_wlan.h"
#include "wifi/sap_config.h"

#ifdef PRISM2_PCCARD
static dev_info_t dev_info = "softap_cs";
#endif


#ifdef PRISM2_MONITOR
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
/* net/sock.h (at least in 2.2.17) does not like min()/max() macros from
 * pcmcia-cs's kernel compat header */
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif
#endif /* kernel < 2.4.0 */

#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#endif /* PRISM2_MONITOR */



#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
/* 2.2 compatibility */
#ifndef spin_lock_bh
#define spin_lock_bh(lock) spin_lock_irq(lock)
#define spin_unlock_bh(lock) spin_unlock_irq(lock)
#endif
#ifndef __constant_cpu_to_le16
#define __constant_cpu_to_le16 __cpu_to_le16
#endif

#define PRISM2_NETDEV_EXTRA IFNAMSIZ
#define prism2_set_dev_name(dev, pos) (dev)->name = (char *) (pos)
/* end 2.2 compatibility */
#else /* kernel < 2.4.0 */
/* no extra space needed for 2.4.x net_device */
#define PRISM2_NETDEV_EXTRA 0
#define prism2_set_dev_name(dev, pos) do { } while (0)
#endif /* kernel < 2.4.0 */

#include "prism2_wlan.h"
#include "prism2_sap.h"

#include "wifi/wifi.h"


#ifdef ASAP_MOD

#include "wifi/asap_drv.h"
#include "wifi/sap_config.h"

#else // ASAP_MOD

int prism2_rx_kernel(struct net_device *, struct hfa384x_rx_frame *, int);

#include "forwarder.h"

DECLARE_TASKLET(sap_tasklet, sap_switch_uplink, 0);
extern struct sk_buff_head sap_readq; 

int gre_init(struct net_device *);
void gre_cleanup(void);


int sap_chr_init(struct net_device *);
void sap_chr_cleanup(void);

void softap_init_proc(struct net_device *);
void softap_takedown_proc(void);
#endif // ASAP_MOD

int prism2_tx_kernel(void *desc, struct sk_buff *skb, struct net_device *dev);

/* #define final_version */

#define MAX_PARM_DEVICES 8
#define PARM_MIN_MAX "1-" __MODULE_STRING(MAX_PARM_DEVICES)
#define DEF_INTS -1, -1, -1, -1, -1, -1, -1
#define GET_INT_PARM(var,idx) var[var[idx] < 0 ? 0 : idx]

static int mtu = 1600;
MODULE_PARM(mtu, "i");
MODULE_PARM_DESC(mtu, "Maximum transfer unit");

static int channel[MAX_PARM_DEVICES] = { 1, DEF_INTS };
MODULE_PARM(channel, PARM_MIN_MAX "i");
MODULE_PARM_DESC(channel, "Initial channel");

static char *essid[MAX_PARM_DEVICES] = { "aruba-ppc-sap" };
MODULE_PARM(essid, PARM_MIN_MAX "s");
MODULE_PARM_DESC(essid, "Host AP's ESSID");

static int iw_mode[MAX_PARM_DEVICES] = { IW_MODE_MASTER, DEF_INTS };
MODULE_PARM(iw_mode, PARM_MIN_MAX "i");
MODULE_PARM_DESC(iw_mode, "Initial operation mode");

static int beacon_int[MAX_PARM_DEVICES] = { 100, DEF_INTS };
MODULE_PARM(beacon_int, PARM_MIN_MAX "i");
MODULE_PARM_DESC(beacon_int, "Beacon interval (1 = 1024 usec)");

static int dtim_period[MAX_PARM_DEVICES] = { 1, DEF_INTS };
MODULE_PARM(dtim_period, PARM_MIN_MAX "i");
MODULE_PARM_DESC(dtim_period, "DTIM period");

static int delayed_enable = 0;
MODULE_PARM(delayed_enable, "i");
MODULE_PARM_DESC(delayed_enable, "Delay MAC port enable until netdevice open");

static int disable_on_close = 0;
MODULE_PARM(disable_on_close, "i");
MODULE_PARM_DESC(disable_on_close, "Disable MAC port on netdevice close");

#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
static int bus_master_threshold_rx[MAX_PARM_DEVICES] = { 100, DEF_INTS };
MODULE_PARM(bus_master_threshold_rx, "i");
MODULE_PARM_DESC(bus_master_threshold_rx, "Packet length threshold for using "
		 "PCI bus master on RX");

static int bus_master_threshold_tx[MAX_PARM_DEVICES] = { 100, DEF_INTS };
MODULE_PARM(bus_master_threshold_tx, "i");
MODULE_PARM_DESC(bus_master_threshold_tx, "Packet length threshold for using "
		 "PCI bus master on TX");
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */


/* See IEEE 802.1H for LLC/SNAP encapsulation/decapsulation */
/* Ethernet-II snap header (RFC1042 for most EtherTypes) */
static unsigned char rfc1042_header[] =
{ 0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00 };
/* Bridge-Tunnel header (for EtherTypes ETH_P_AARP and ETH_P_IPX) */
#if 0
static unsigned char bridge_tunnel_header[] =
{ 0xaa, 0xaa, 0x03, 0x00, 0x00, 0xf8 };
#endif
/* No encapsulation header if EtherType < 0x600 (=length) */


static const long freq_list[] = { 2412, 2417, 2422, 2427, 2432, 2437, 2442,
				  2447, 2452, 2457, 2462, 2467, 2472, 2484 };
#define FREQ_COUNT (sizeof(freq_list) / sizeof(freq_list[0]))


#ifdef final_version
#define EXTRA_EVENTS_WTERR 0
#else
/* check WTERR events (Wait Time-out) in development versions */
#define EXTRA_EVENTS_WTERR HFA384X_EV_WTERR
#endif

#ifdef PRISM2_USE_TX_INTERRUPT
#define EXTRA_EVENTS_TX HFA384X_EV_TX
#define EXTRA_TX_CTRL HFA384X_TX_CTRL_TX_OK
#else
#define EXTRA_EVENTS_TX 0
#define EXTRA_TX_CTRL 0
#endif

#ifdef PRISM2_USE_CMD_COMPL_INTERRUPT
#define EXTRA_EVENTS_CMD HFA384X_EV_CMD
#else
#define EXTRA_EVENTS_CMD 0
#endif

#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
#define EXTRA_EVENTS_BUS_MASTER (HFA384X_EV_PCI_M0 | HFA384X_EV_PCI_M1)
#else
#define EXTRA_EVENTS_BUS_MASTER 0
#endif

/* event mask, i.e., events that will result in an interrupt */
#define HFA384X_EVENT_MASK \
	(HFA384X_EV_TXEXC | HFA384X_EV_RX | HFA384X_EV_ALLOC | \
	HFA384X_EV_INFDROP | \
	HFA384X_EV_INFO | \
        EXTRA_EVENTS_WTERR | EXTRA_EVENTS_TX | \
	EXTRA_EVENTS_CMD | EXTRA_EVENTS_BUS_MASTER)


#define HFA384X_MONITOR_EVENT_MASK \
	(HFA384X_EV_TXEXC | HFA384X_EV_RX | HFA384X_EV_ALLOC | \
	EXTRA_EVENTS_CMD)

#define HFA384X_TX_CTRL_FLAGS \
	(HFA384X_TX_CTRL_802_11 | HFA384X_TX_CTRL_TX_EX | EXTRA_TX_CTRL)


/* ca. 1 usec */
#define HFA384X_CMD_BUSY_TIMEOUT 1000
#define HFA384X_BAP_BUSY_TIMEOUT 5000

/* ca. 10 usec */
#define HFA384X_INIT_TIMEOUT 50000
#define HFA384X_CMD_COMPL_TIMEOUT 20000
#define HFA384X_ALLOC_COMPL_TIMEOUT 1000


#define TX_TIMEOUT (2 * HZ)
#define PRISM2_MAX_FRAME_SIZE 2304
#define PRISM2_MIN_MTU 256
/* FIX: */
#define PRISM2_MAX_MTU (PRISM2_MAX_FRAME_SIZE - sizeof(rfc1042_header))

static void prism2_check_sta_fw_version(intersil_local_info_t *local);

//int prism2_recv_frame(intersil_local_info_t *, struct sk_buff *);

/* prism2_ioctl.c */
struct iw_statistics *prism2_get_wireless_stats(struct net_device *);
int prism2_ioctl(struct net_device *, struct ifreq *, int );

u16 hfa384x_tx_ctrl_flags = HFA384X_TX_CTRL_FLAGS;

#ifndef final_version
/* magic value written to SWSUPPORT0 reg. for detecting whether card is still
 * present */
#define HFA384X_MAGIC 0x8A32
#endif

#ifdef PRISM2_PCI
#define HFA384X_OUTB(v,a) writeb((v), dev->mem_start + (a))
#define HFA384X_INB(a) (u8) readb(dev->mem_start + (a))
#define HFA384X_OUTW(v,a) writew((v), dev->mem_start + (a))
#define HFA384X_INW(a) (u16) readw(dev->mem_start + (a))
#define HFA384X_OUTW_DATA(v,a) writew(cpu_to_le16(v), dev->mem_start + (a))
#define HFA384X_INW_DATA(a) (u16) le16_to_cpu(readw(dev->mem_start + (a)))
#endif /* PRISM2_PCI */


//static struct sock *sap_sock;

u16 hfa384x_read_reg(struct net_device *dev, u16 reg)
{
  return HFA384X_INW(reg);
}


int hfa384x_cmd(struct net_device *dev, u16 cmd, u16 param0, u16 *param1,
		u16 *resp0)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int tries, res;
  unsigned long flags;
  u16 reg;
  
#ifdef PRISM2_USE_CMD_COMPL_INTERRUPT
  /* Wait for a possible pending transmit command completion without
   * disabling IRQs */
  tries = 10000;
  for (;;) {
    while (local->last_txfid_idx != PRISM2_DUMMY_FID &&
	   tries > 0) {
      tries--;
      udelay(1);
    }
    
    if (tries == 0) {
      printk(KERN_DEBUG "hfa384x_cmd: timeout on waiting "
	     "pending TX\n");
      return -ETIMEDOUT;
    }
    
    spin_lock_irqsave(&local->cmdlock, flags);
    if (local->last_txfid_idx == PRISM2_DUMMY_FID)
      break;
    
    printk(KERN_DEBUG "hfa384x_cmd: new TX started - retrying\n");
    spin_unlock_irqrestore(&local->cmdlock, flags);
  }
#else /* PRISM2_USE_CMD_COMPL_INTERRUPT */
  spin_lock_irqsave(&local->cmdlock, flags);
#endif /* PRISM2_USE_CMD_COMPL_INTERRUPT */
  
  /* wait until busy bit is clear */
  tries = HFA384X_CMD_BUSY_TIMEOUT;
  while (HFA384X_INW(HFA384X_CMD_OFF) & HFA384X_CMD_BUSY && tries > 0) {
    tries--;
    udelay(1);
  }
  if (tries == 0) {
    reg = HFA384X_INW(HFA384X_CMD_OFF);
    spin_unlock_irqrestore(&local->cmdlock, flags);
    printk("%s: hfa384x_cmd - timeout - reg=0x%04x\n", dev->name,
	   reg);
    return -ETIMEDOUT;
  }
  
  /* write command */
  HFA384X_OUTW(param0, HFA384X_PARAM0_OFF);
  if (param1 != NULL)
    HFA384X_OUTW(*param1, HFA384X_PARAM1_OFF);
  HFA384X_OUTW(cmd, HFA384X_CMD_OFF);
  
  /* wait for command completion */
  tries = HFA384X_CMD_COMPL_TIMEOUT;
  while (!(HFA384X_INW(HFA384X_EVSTAT_OFF) & HFA384X_EV_CMD) &&
	 tries > 0) {
    tries--;
    udelay(10);
  }
  if (tries == 0) {
    reg = HFA384X_INW(HFA384X_EVSTAT_OFF);
    spin_unlock_irqrestore(&local->cmdlock, flags);
    printk("%s: hfa384x_cmd - timeout2 - reg=0x%04x\n", dev->name,
	   reg);
    return -ETIMEDOUT;
  }
  
  if (resp0 != NULL)
    *resp0 = HFA384X_INW(HFA384X_RESP0_OFF);
  res = (HFA384X_INW(HFA384X_STATUS_OFF) &
	 (BIT(14) | BIT(13) | BIT(12) | BIT(11) | BIT(10) | BIT(9) |
	  BIT(8))) >> 8;
#ifndef final_version
  if (res) {
    u16 resp0 = HFA384X_INW(HFA384X_RESP0_OFF);
    printk("%s: CMD=0x%04x => res=0x%02x, resp0=0x%04x\n",
	   dev->name, cmd, res, resp0);
  }
#endif
  
  HFA384X_OUTW(HFA384X_EV_CMD, HFA384X_EVACK_OFF);
  
  spin_unlock_irqrestore(&local->cmdlock, flags);
  return res;
}


static int hfa384x_cmd_no_wait(struct net_device *dev, u16 cmd, u16 param0)
{
  int tries;
  u16 reg;
  
  /* wait until busy bit is clear */
  tries = HFA384X_CMD_BUSY_TIMEOUT;
  while (HFA384X_INW(HFA384X_CMD_OFF) & HFA384X_CMD_BUSY && tries > 0) {
    tries--;
    udelay(1);
  }
  if (tries == 0) {
    reg = HFA384X_INW(HFA384X_CMD_OFF);
    printk("%s: hfa384x_cmd - timeout - reg=0x%04x\n", dev->name,
	   reg);
    return -ETIMEDOUT;
  }
  
  /* write command */
  HFA384X_OUTW(param0, HFA384X_PARAM0_OFF);
  HFA384X_OUTW(cmd, HFA384X_CMD_OFF);
  
  return 0;
}


static inline int hfa384x_wait_offset(struct net_device *dev, u16 o_off)
{
  int tries = HFA384X_BAP_BUSY_TIMEOUT;
  
  while ((HFA384X_INW(o_off) & HFA384X_OFFSET_BUSY) && tries > 0) {
    tries--;
    udelay(1);
  }
  return (HFA384X_INW(o_off) & HFA384X_OFFSET_BUSY);
}


/* Offset must be even */
int hfa384x_setup_bap(struct net_device *dev, u16 bap, u16 id, int offset)
{
  u16 o_off, s_off;
  int ret = 0;
  
  if (offset % 2 || bap > 1)
    return -EINVAL;
  
  if (bap == BAP1) {
    o_off = HFA384X_OFFSET1_OFF;
    s_off = HFA384X_SELECT1_OFF;
  } else {
    o_off = HFA384X_OFFSET0_OFF;
    s_off = HFA384X_SELECT0_OFF;
  }
  
  if (hfa384x_wait_offset(dev, o_off)) {
    printk(KERN_DEBUG "%s: hfa384x_setup_bap - timeout before\n",
	   dev->name);
    ret = -ETIMEDOUT;
    goto out;
  }
  
  HFA384X_OUTW(id, s_off);
  HFA384X_OUTW(offset, o_off);
  
  if (hfa384x_wait_offset(dev, o_off)) {
    printk(KERN_DEBUG "%s: hfa384x_setup_bap - timeout after\n",
	   dev->name);
    ret = -ETIMEDOUT;
    goto out;
  }
#ifndef final_version
  if (HFA384X_INW(o_off) & HFA384X_OFFSET_ERR) {
    printk(KERN_DEBUG "%s: hfa384x_setup_bap - offset error "
	   "(%d,%d,%d)\n",
	   dev->name, bap, id, offset);
    ret = -EINVAL;
  }
#endif
  
 out:
  return ret;
}


/* FIX: hfa384x_{from,to}_bap could be moved to prism2_{pccard,plx,pci}.o since
 * they differ on different hardware versions */

int hfa384x_from_bap(struct net_device *dev, u16 bap, void *buf, int len)
{
  u16 d_off;
  u16 *pos;
  
  d_off = (bap == 1) ? HFA384X_DATA1_OFF : HFA384X_DATA0_OFF;
  pos = (u16 *) buf;
  
#ifdef PRISM2_PCI
  for ( ; len > 1; len -= 2)
    *pos++ = HFA384X_INW_DATA(d_off);
#else /* PRISM2_PCI */
  if (len / 2)
    HFA384X_INSW(d_off, buf, len / 2);
  pos += len / 2;
#endif /* PRISM2_PCI */
  
  if (len & 1)
    *((char *) pos) = HFA384X_INB(d_off);
  
  return 0;
}


int hfa384x_to_bap(struct net_device *dev, u16 bap, void *buf, int len)
{
  u16 d_off;
  u16 *pos;
  
  d_off = (bap == 1) ? HFA384X_DATA1_OFF : HFA384X_DATA0_OFF;
  pos = (u16 *) buf;
  
#ifdef PRISM2_PCI
  for ( ; len > 1; len -= 2)
    HFA384X_OUTW_DATA(*pos++, d_off);
#else /* PRISM2_PCI */
  if (len / 2)
    HFA384X_OUTSW(d_off, buf, len / 2);
  pos += len / 2;
#endif /* PRISM2_PCI */
  
  if (len & 1)
    HFA384X_OUTB(*((char *) pos), d_off);
  
  return 0;
}


int hfa384x_get_rid(struct net_device *dev, u16 rid, void *buf, int len,
		    int exact_len)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int res, rlen = 0;
  struct hfa384x_rid_hdr rec;
  unsigned long flags;
  
  res = hfa384x_cmd(dev, HFA384X_CMDCODE_ACCESS, rid, NULL, NULL);
  if (res) {
    printk("%s: hfa384x_get_rid: CMDCODE_ACCESS failed (res=%d)\n",
	   dev->name, res);
    return res;
  }
  
  spin_lock_irqsave(&local->baplock, flags);
  
  res = hfa384x_setup_bap(dev, BAP0, rid, 0);
  if (!res)
    res = hfa384x_from_bap(dev, BAP0, &rec, sizeof(rec));
  if (res) {
    printk("%s: hfa384x_get_rid - from BAP0 failed\n",
	   dev->name);
    goto fail;
  }
  
  rlen = (__le16_to_cpu(rec.len) - 1) * 2;
  if (exact_len && rlen != len) {
    printk("RID len mismatch: rid=0x%04x, len=%d (expected %d)\n",
	   rid, rlen, len);
    res = -ENODATA;
    goto fail;
  }
  
  res = hfa384x_from_bap(dev, BAP0, buf, len);
  if (res) {
    printk("%s: hfa384x_get_rid - from BAP0(2) failed\n",
	   dev->name);
    goto fail;
  }
  
 fail:
  spin_unlock_irqrestore(&local->baplock, flags);
  
  if (res) {
    if (res == -ETIMEDOUT)
      prism2_hw_reset(dev);
    return res;
  }
  return rlen;
}


int hfa384x_set_rid(struct net_device *dev, u16 rid, void *buf, int len)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int res;
  struct hfa384x_rid_hdr rec;
  unsigned long flags;
  
  rec.rid = __cpu_to_le16(rid);
  /* RID len in words and +1 for rec.rid */
  rec.len = __cpu_to_le16(len / 2 + len % 2 + 1);
  
  spin_lock_irqsave(&local->baplock, flags);
  
  res = hfa384x_setup_bap(dev, BAP0, rid, 0);
  if (!res)
    res = hfa384x_to_bap(dev, BAP0, &rec, sizeof(rec));
  if (res) {
    printk("%s: hfa384x_set_rid - to BAP0 failed\n",
	   dev->name);
    goto fail;
  }
  
  res = hfa384x_to_bap(dev, BAP0, buf, len);
  if (res) {
    printk("%s: hfa384x_set_rid - to BAP0(2) failed\n", dev->name);
    goto fail;
  }
  
  res = hfa384x_cmd(dev, HFA384X_CMDCODE_ACCESS_WRITE, rid, NULL, NULL);
  if (res) {
    printk("%s: hfa384x_set_rid: CMDCODE_ACCESS_WRITE failed "
	   "(res=%d)\n", dev->name, res);
    goto fail;
  }
  
 fail:
  spin_unlock_irqrestore(&local->baplock, flags);
  
  if (res == -ETIMEDOUT)
    prism2_hw_reset(dev);
  
  return res;
}


void hfa384x_disable_interrupts(struct net_device *dev)
{
  /* disable interrupts and clear event status */
  HFA384X_OUTW(0, HFA384X_INTEN_OFF);
  HFA384X_OUTW(0xffff, HFA384X_EVACK_OFF);
}


static void hfa384x_enable_interrupts(struct net_device *dev)
{
  /* ack pending events and enable interrupts from selected events */
  HFA384X_OUTW(0xffff, HFA384X_EVACK_OFF);
  HFA384X_OUTW(HFA384X_EVENT_MASK, HFA384X_INTEN_OFF);
}

static void hfa384x_enable_monitor_mode_interrupts(struct net_device *dev){
  /* ack pending events and enable interrupts from selected events */
  HFA384X_OUTW(0xffff, HFA384X_EVACK_OFF);
  HFA384X_OUTW(HFA384X_MONITOR_EVENT_MASK, HFA384X_INTEN_OFF);
}
  


static u16 hfa384x_allocate_fid(struct net_device *dev, int len)
{
  int tries;
  u16 fid;
  
  if (hfa384x_cmd(dev, HFA384X_CMDCODE_ALLOC, len, NULL, NULL)) {
    printk("%s: cannot allocate fid, len=%d\n", dev->name, len);
    return 0xffff;
  }
  
  tries = HFA384X_ALLOC_COMPL_TIMEOUT;
  while (!(HFA384X_INW(HFA384X_EVSTAT_OFF) & HFA384X_EV_ALLOC) &&
	 tries > 0) {
    tries--;
    udelay(10);
  }
  if (tries == 0) {
    printk("%s: fid allocate, len=%d - timeout\n", dev->name, len);
    return 0xffff;
  }
  
  fid = HFA384X_INW(HFA384X_ALLOCFID_OFF);
  HFA384X_OUTW(HFA384X_EV_ALLOC, HFA384X_EVACK_OFF);
  
  return fid;
}


int prism2_reset_port(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int res;
  
  if (!local->dev_enabled)
    return 0;
  
  res = hfa384x_cmd(dev, HFA384X_CMDCODE_DISABLE, 0,
		    NULL, NULL);
  if (!res)
    res = hfa384x_cmd(dev, HFA384X_CMDCODE_ENABLE, 0,
		      NULL, NULL);
  
  return res;
}


int prism2_set_tim(struct net_device *dev, u16 aid, int set)
{
  u16 tmp;
  
  if (set) {
    PDEBUG(DEBUG_PS2, "Setting TIM bit for AID %i\n", aid);
    tmp = __cpu_to_le16(0x8000 | aid);
  } else {
    PDEBUG(DEBUG_PS2, "Clearing TIM bit for AID %i\n", aid);
    tmp = __cpu_to_le16(aid);
  }
  if (hfa384x_set_rid(dev, HFA384X_RID_CNFTIMCTRL, &tmp, 2)) {
    PDEBUG(DEBUG_PS, "TIM AID %d setting (set=%d) failed\n",
	   aid, set);
    return -1;
  }
  return 0;
}


static void prism2_get_version_info(struct net_device *dev, u16 rid,
				    const char *txt)
{
  struct hfa384x_comp_ident comp;
  
  if (hfa384x_get_rid(dev, rid, &comp, sizeof(comp), 1) < 0)
    printk("Could not get RID for component %s\n", txt);
  else
    printk("%s: %s: id=0x%02x v%d.%d.%d\n", dev->name, txt,
	   __le16_to_cpu(comp.id), __le16_to_cpu(comp.major),
	   __le16_to_cpu(comp.minor), __le16_to_cpu(comp.variant));
}


/* val is in host byte order */
int prism2_set_word(struct net_device *dev, int rid, u16 val)
{
  u16 tmp = cpu_to_le16(val);
  return hfa384x_set_rid(dev, rid, &tmp, 2);
}


int prism2_set_string(struct net_device *dev, int rid, const char *val)
{
  char buf[MAX_SSID_LEN + 2];
  int len;
  
  len = strlen(val);
  if (len > MAX_SSID_LEN)
    return -1;
  memset(buf, 0, sizeof(buf));
  buf[0] = len; /* little endian 16 bit word */
  memcpy(buf + 2, val, len);
  
  return hfa384x_set_rid(dev, rid, &buf, MAX_SSID_LEN + 2);
}


u16 prism2_get_porttype(intersil_local_info_t *local)
{
  if (local->iw_mode == IW_MODE_ADHOC && local->pseudo_adhoc)
    return HFA384X_PORTTYPE_PSEUDO_IBSS;
  if (local->iw_mode == IW_MODE_ADHOC)
    return HFA384X_PORTTYPE_IBSS;
  if (local->iw_mode == IW_MODE_INFRA)
    return HFA384X_PORTTYPE_BSS;
  return HFA384X_PORTTYPE_HOSTAP;
}


int prism2_set_encryption(intersil_local_info_t *local)
{
  u16 val;
  
  if (hfa384x_get_rid(local->dev, HFA384X_RID_CNFWEPFLAGS, &val, 2, 1) <
      0) {
    printk(KERN_DEBUG "Could not read current WEP flags.\n");
    goto fail;
  }
  le16_to_cpus(&val);

  val &= ~HFA384X_WEPFLAGS_EXCLUDEUNENCRYPTED;
#if 1  
  val &= ~HFA384X_WEPFLAGS_PRIVACYINVOKED;
#else
  val |= HFA384X_WEPFLAGS_PRIVACYINVOKED;
#endif
  val |= HFA384X_WEPFLAGS_HOSTENCRYPT;
  val |= HFA384X_WEPFLAGS_HOSTDECRYPT;
  
  printk(KERN_DEBUG "Wrote 0x%x into CNFWEPFLAGS\n", val);
  
  if (prism2_set_word(local->dev, HFA384X_RID_CNFWEPFLAGS, val)) {
    printk(KERN_DEBUG "Could not write new WEP flags (0x%x)\n",
	   val);
    goto fail;
  }
  
  return 0;
  
 fail:
  printk(KERN_DEBUG "%s: encryption setup failed\n", local->dev->name);
  return -1;
}


static int prism2_setup_rids(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  u16 tmp;
  int ret = 0;
  
  if (!local->fw_ap) {
    tmp = prism2_get_porttype(local);
    ret = prism2_set_word(dev, HFA384X_RID_CNFPORTTYPE, tmp);
    if (ret) {
      printk("%s: Port type setting to %d failed\n",
	     dev->name, tmp);
      goto fail;
    }
  }
  
  /* Setting SSID to empty string seems to kill the card in Host AP mode
   */
  if (local->iw_mode != IW_MODE_MASTER || local->essid[0] != '\0') {
    ret = prism2_set_string(dev, HFA384X_RID_CNFOWNSSID,
			    local->essid);
    if (ret) {
      printk("%s: AP own SSID setting failed\n", dev->name);
      goto fail;
    }
  }
  
  ret = prism2_set_word(dev, HFA384X_RID_CNFMAXDATALEN,
			PRISM2_DATA_MAXLEN);
  if (ret) {
    printk("%s: MAC data length setting to %d failed\n",
	   dev->name, PRISM2_DATA_MAXLEN);
    goto fail;
  }
  
  if (hfa384x_get_rid(dev, HFA384X_RID_CHANNELLIST, &tmp, 2, 1) < 0) {
    printk("%s: Channel list read failed\n", dev->name);
    ret = -EINVAL;
    goto fail;
  }
  local->channel_mask = __le16_to_cpu(tmp);
  
  if (local->channel < 1 || local->channel > FREQ_COUNT ||
      !(local->channel_mask & (1 << (local->channel - 1)))) {
    printk(KERN_WARNING "%s: Channel setting out of range "
	   "(%d)!\n", dev->name, local->channel);
    ret = -EBUSY;
    goto fail;
  }
  
  ret = prism2_set_word(dev, HFA384X_RID_CNFOWNCHANNEL, local->channel);
  if (ret) {
    printk("%s: Channel setting to %d failed\n",
	   dev->name, local->channel);
    goto fail;
  }
  
  ret = prism2_set_word(dev, HFA384X_RID_CNFBEACONINT,
			local->beacon_int);
  if (ret) {
    printk("%s: Beacon interval setting to %d failed\n",
	   dev->name, local->beacon_int);
    if (!local->is_symbol && !local->is_lucent)
      goto fail;
  }
  
  ret = prism2_set_word(dev, HFA384X_RID_CNFBASICRATES,
			HFA384X_RATES_1MBPS | HFA384X_RATES_2MBPS);
  if (ret) {
    printk("%s: Basic rates setting failed\n", dev->name);
    if (!local->is_symbol && !local->is_lucent)
      goto fail;
  }
  
  ret = prism2_set_word(dev, HFA384X_RID_CNFSUPPORTEDRATES,
			HFA384X_RATES_1MBPS | HFA384X_RATES_2MBPS |
			HFA384X_RATES_5MBPS | HFA384X_RATES_11MBPS);
  if (ret) {
    printk("%s: Supported rates setting failed\n", dev->name);
    if (!local->is_symbol && !local->is_lucent)
      goto fail;
  }
  
  ret = prism2_set_word(dev, HFA384X_RID_CNFOWNDTIMPERIOD,
			local->dtim_period);
  if (ret) {
    printk("%s: DTIM period setting to %d failed\n",
	   dev->name, local->dtim_period);
    if (!local->is_symbol && !local->is_lucent)
      goto fail;
  }
  
  if (!local->fw_ap) {
    ret = prism2_set_string(dev, HFA384X_RID_CNFDESIREDSSID,
			    local->essid);
    if (ret) {
      printk("%s: Desired SSID setting failed\n", dev->name);
      goto fail;
    }
  }
  
  /* Setup TXRateControl, defaults to allow use of 1, 2, 5.5, and
   * 11 Mbps in automatic TX rate fallback */
  if (local->tx_rate_control == 0) {
    local->tx_rate_control =
      HFA384X_RATES_1MBPS |
      HFA384X_RATES_2MBPS |
      HFA384X_RATES_5MBPS |
      HFA384X_RATES_11MBPS;
  }
  if (!local->fw_ap) {
    ret = prism2_set_word(dev, HFA384X_RID_TXRATECONTROL,
			  local->tx_rate_control);
    if (ret) {
      printk("%s: TXRateControl setting to %d failed\n",
	     dev->name, local->tx_rate_control);
      goto fail;
    }
    
    ret = prism2_set_word(dev, HFA384X_RID_CREATEIBSS, 1);
    if (ret) {
      printk("%s: Create IBSS setting to 1 failed\n",
	     dev->name);
    }
  }
  
  if (local->name_set)
    (void) prism2_set_string(dev, HFA384X_RID_CNFOWNNAME,
			     local->name);
  
  if (prism2_set_encryption(local)) {
    printk(KERN_INFO "%s: could not configure encryption\n",
	   dev->name);
  }
  
 fail:
  return ret;
}



static int prism2_hw_init(struct net_device *dev, int initial)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int ret, i, len;
  
  PDEBUG(DEBUG_FLOW, "prism2_hw_init()\n");
  
#ifdef PRISM2_USE_CMD_COMPL_INTERRUPT
  local->last_txfid_idx = PRISM2_DUMMY_FID;
#endif
  
  /* initialize HFA 384x */
  ret = hfa384x_cmd_no_wait(dev, HFA384X_CMDCODE_INIT, 0);
  if (ret) {
    printk("prism2: first command failed - is the card compatible?\n");
    goto failed;
  }
  i = HFA384X_INIT_TIMEOUT;
  while (!(HFA384X_INW(HFA384X_EVSTAT_OFF) & HFA384X_EV_CMD) && i > 0) {
    i--;
    udelay(10);
  }
  if (i == 0) {
    printk("prism2: card initialization timed out\n");
    goto failed;
  }
  printk(KERN_DEBUG "prism2_hw_config: initialized in %d iterations\n",
	 HFA384X_INIT_TIMEOUT - i);
  HFA384X_OUTW(HFA384X_EV_CMD, HFA384X_EVACK_OFF);
  
#ifdef PRISM2_DOWNLOAD_SUPPORT
  local->pda = prism2_read_pda(dev);
#endif /* PRISM2_DOWNLOAD_SUPPORT */
  
  hfa384x_disable_interrupts(dev);
  
#ifndef final_version
  HFA384X_OUTW(HFA384X_MAGIC, HFA384X_SWSUPPORT0_OFF);
  if (HFA384X_INW(HFA384X_SWSUPPORT0_OFF) != HFA384X_MAGIC) {
    printk("SWSUPPORT0 write/read failed: %04X != %04X\n",
	   HFA384X_INW(HFA384X_SWSUPPORT0_OFF), HFA384X_MAGIC);
    goto failed;
  }
#endif
  
  /* allocate TX FIDs */
  len = local->is_symbol ? 1600 : PRISM2_TXFID_LEN;
  for (i = 0; i < PRISM2_TXFID_COUNT; i++) {
    local->txfid[i] = hfa384x_allocate_fid(dev, len);
    if (local->txfid[i] == 0xffff)
      goto failed;
    local->intransmitfid[i] = PRISM2_TXFID_EMPTY;
  }
  
#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
  local->bus_m1_in_use = 0;
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */
  
  if (initial) {
    /* get card version information */
    prism2_get_version_info(dev, HFA384X_RID_NICID, "NIC");
    prism2_get_version_info(dev, HFA384X_RID_PRIID, "PRI");
    prism2_get_version_info(dev, HFA384X_RID_STAID, "STA");
    
    prism2_check_sta_fw_version(local);
    
    if (hfa384x_get_rid(dev, HFA384X_RID_CNFOWNMACADDR,
			&dev->dev_addr, 6, 1) < 0) {
      printk("%s: could not get own MAC address\n",
	     dev->name);
    }
    dev->addr_len = 6;
  } else if (local->fw_ap)
    prism2_check_sta_fw_version(local);
  
  prism2_setup_rids(dev);
  
  /* MAC is now configured, but port 0 is not yet enabled */
  return 0;
  
 failed:
  printk(KERN_WARNING "prism2: Initialization failed\n");
  return 1;
}


static int prism2_hw_enable(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  
  if (hfa384x_cmd(dev, HFA384X_CMDCODE_ENABLE, 0, NULL, NULL)) {
    printk("%s: MAC port 0 enabling failed\n", dev->name);
    return 1;
  }
  
  /* at least D-Link DWL-650 seems to require additional port reset
   * before it starts acting as an AP, so reset port automatically
   * here just in case */
  if (prism2_reset_port(dev)) {
    printk("%s: MAC port 0 reseting failed\n", dev->name);
    return 1;
  }
  
  local->hw_ready = 1;
  local->hw_reset_tries = 0;
  hfa384x_enable_interrupts(dev);
  printk("prism2 interrupts enabled\n");
  
  return 0;
}


int prism2_hw_config(struct net_device *dev, int initial)
{
  if (prism2_hw_init(dev, initial))
    return 1;
  
  if (!initial || !delayed_enable) {
    intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
    local->dev_enabled = 1;
    return prism2_hw_enable(dev);
  }
  
  return 0;
}


void prism2_hw_shutdown(struct net_device *dev, int no_disable)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  
  local->hw_ready = 0;
  if (local->func->card_present && !local->func->card_present(local)) {
    printk(KERN_DEBUG "%s: card already removed or not configured "
	   "during shutdown\n", dev->name);
    return;
  }
  
  hfa384x_disable_interrupts(dev);
  
  if (!no_disable &&
      hfa384x_cmd(dev, HFA384X_CMDCODE_DISABLE, 0, NULL, NULL))
    printk(KERN_WARNING "prism2: Shutdown failed\n");
}


void prism2_hw_reset(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  
#if 0
  static long last_reset = 0;
  
  /* do not reset card more than once per second to avoid ending up in a
   * busy loop reseting the card */
  if (last_reset + HZ >= jiffies)
    return;
  last_reset = jiffies;
#endif
  
  if (local->hw_resetting) {
    printk(KERN_WARNING "%s: already resetting card "
	   "ignoring reset request\n", dev->name);
    return;
  }
  
  local->hw_reset_tries++;
  if (local->hw_reset_tries > 10) {
    printk(KERN_WARNING "%s: too many reset tries, skipping\n",
	   dev->name);
    return;
  }
  
  printk(KERN_WARNING "%s: resetting card, jiffies %ld\n", 
	 dev->name, jiffies);
  local->hw_resetting = 1;
  hfa384x_disable_interrupts(dev);
  if (local->func->cor_sreset)
    local->func->cor_sreset(local);
  prism2_hw_shutdown(dev, 1);
  prism2_hw_config(dev, 0);

  // check if monitor mode was enabled.
  // if yes, restore monitor mode

  if (local->monitor_type == PRISM2_MONITOR_ON){
    // restore monitor mode

    int res;

    res = prism2_set_word(dev, HFA384X_RID_CNFPORTTYPE,
			  HFA384X_PORTTYPE_PSEUDO_IBSS);
    if (res) {
      printk(KERN_DEBUG "prism2_hw_reset : Port type setting for monitor mode "
	     "failed\n");
      //      return ret;
    }
    hfa384x_cmd(dev, HFA384X_CMDCODE_TEST |
		(HFA384X_TEST_MONITOR << 8),
		0, NULL, NULL);

    res = prism2_set_word(dev, HFA384X_RID_CNFOWNCHANNEL, local->channel);
    if (res) {
      printk("%s: Channel setting to %d failed\n",
	     dev->name, local->channel);
    }
    
  }

  local->hw_resetting = 0;
}


struct net_device_stats *prism2_get_stats(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  return &local->stats;
}


/* wake all netif queues in use */
static void prism2_netif_wake_queues(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  
  if (local->dev)
    netif_wake_queue(local->dev);
  
}


/* stop all netif queues in use */
void prism2_netif_stop_queues(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  
  if (local->dev)
    netif_stop_queue(local->dev);
}

static int prism2_close(struct net_device *dev);

static int prism2_open(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  
  PDEBUG(DEBUG_FLOW, "%s: prism2_open\n", dev->name);
  
  if (local->func->dev_open && local->func->dev_open(local))
    return 1;
  
  //  MOD_INC_USE_COUNT;
  
  if (!local->dev_enabled && prism2_hw_enable(dev)) {
    printk(KERN_WARNING "%s: could not enable MAC port\n",
	   dev->name);
    prism2_close(dev);
    return 1;
  }
  local->dev_enabled = 1;
  
  netif_device_attach(dev);
  netif_start_queue(dev);
  
  return 0;
}


static int prism2_close(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  
  PDEBUG(DEBUG_FLOW, "%s: prism2_close\n", dev->name);
  
  if (local->func->dev_close && local->func->dev_close(local))
    return 0;
  
  if (disable_on_close) {
    local->dev_enabled = 0;
    prism2_hw_shutdown(dev, 0);
  }
  
  if (netif_running(dev)) {
    netif_stop_queue(dev);
    netif_device_detach(dev);
  }
  
  //  MOD_DEC_USE_COUNT;
  
  return 0;
}



static int prism2_change_mtu(struct net_device *dev, int new_mtu)
{
  if (new_mtu < PRISM2_MIN_MTU || new_mtu > PRISM2_MAX_MTU)
    return -EINVAL;
  
  dev->mtu = new_mtu;
  return 0;
}


#ifdef HAVE_TX_TIMEOUT
static void prism2_tx_timeout(struct net_device *dev)
{
  printk(KERN_WARNING "%s Tx timed out! Resetting card\n",
	 dev->name);
  
  printk(KERN_DEBUG "%s: CMD=%04x EVSTAT=%04x "
	 "OFFSET0=%04x OFFSET1=%04x SWSUPPORT0=%04x\n",
	 dev->name,
	 HFA384X_INW(HFA384X_CMD_OFF),
	 HFA384X_INW(HFA384X_EVSTAT_OFF),
	 HFA384X_INW(HFA384X_OFFSET0_OFF),
	 HFA384X_INW(HFA384X_OFFSET1_OFF),
	 HFA384X_INW(HFA384X_SWSUPPORT0_OFF));
  
  prism2_hw_reset(dev);
  
  if (netif_queue_stopped(dev)) {
    int i;
    intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
    
    for (i = 0; i < PRISM2_TXFID_COUNT; i++)
      if (local->intransmitfid[i] == PRISM2_TXFID_EMPTY) {
	PDEBUG(DEBUG_EXTRA, "prism2_tx_timeout: "
	       "wake up queue\n");
	netif_wake_queue(dev);
	break;
      }
  }
}
#endif /* HAVE_TX_TIMEOUT */


int prism2_get_txfid_idx(intersil_local_info_t *local)
{
  int idx, end;
  
  spin_lock_bh(&local->txfidlock);
  end = idx = local->next_txfid;
  do {
    if (local->intransmitfid[idx] == PRISM2_TXFID_EMPTY) {
      local->intransmitfid[idx] = PRISM2_TXFID_RESERVED;
      spin_unlock_bh(&local->txfidlock);
      return idx;
    }
    idx++;
    if (idx >= PRISM2_TXFID_COUNT)
      idx = 0;
  } while (idx != end);
  spin_unlock_bh(&local->txfidlock);
  
  PDEBUG(DEBUG_EXTRA2, "prism2_get_txfid_idx: no room in txfid buf: "
	 "packet dropped\n");
  local->stats.tx_dropped++;
  
  return -1;
}


#ifdef PRISM2_USE_CMD_COMPL_INTERRUPT

/* Called only from software IRQ if PCI bus master is not used (with bus master
 * this can be called both from software and hardware IRQ) */
int prism2_transmit(struct net_device *dev, int idx)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int res;
  unsigned long flags;
  
  /* The driver tries to stop netif queue so that there would not be
   * more than one attempt to transmit frames going on; check that this
   * is really the case */
  if (local->last_txfid_idx != PRISM2_DUMMY_FID) {
    int tries = 1000;
    while (tries > 0 && local->last_txfid_idx != PRISM2_DUMMY_FID)
      {
	udelay(10);
	tries--;
      }
    
    if (local->last_txfid_idx != PRISM2_DUMMY_FID) {
      printk(KERN_DEBUG "%s: timeout on waiting pending "
	     "transmit command..\n", dev->name);
      return 1;
    }
    //    printk(KERN_DEBUG "%s: %d x 10 usec cleared pending TX\n",
    //	   dev->name, 1000 - tries);
  }
  
  /* stop the queue for the time that last_txfid_idx is in use */
  prism2_netif_stop_queues(dev);
  
  local->last_txfid_idx = idx;
  
  /* transmit packet */
  spin_lock_irqsave(&local->cmdlock, flags);
  res = hfa384x_cmd_no_wait(dev,
			    HFA384X_CMDCODE_TRANSMIT | HFA384X_CMD_TX_RECLAIM,
			    local->txfid[idx]);
  spin_unlock_irqrestore(&local->cmdlock, flags);
  
  if (res) {
    struct net_device_stats *stats;
    printk(KERN_DEBUG "%s: prism2_transmit: CMDCODE_TRANSMIT "
	   "failed (res=%d)\n", dev->name, res);
    stats = prism2_get_stats(dev);
    stats->tx_dropped++;
    prism2_netif_wake_queues(dev);
    return 1;
  }
  dev->trans_start = jiffies;
  
  /* Since we did not wait for command completion, the card continues
   * to process on the background and we will finish handling when
   * command completion event is handled (prism2_cmd_ev() function) */
  
  return 0;
}


/* Called only from hardware IRQ */
static void prism2_cmd_ev(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int idx = local->last_txfid_idx;
  u16 resp0;

#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
#ifdef PRISM2_ENABLE_BEFORE_TX_BUS_MASTER
	if (local->bus_m0_pending_len) {
		unsigned long addr;
		addr = virt_to_phys(local->bus_m0_buf);
		HFA384X_OUTW((addr & 0xffff0000) >> 16,
			     HFA384X_PCI_M0_ADDRH_OFF);
		HFA384X_OUTW(addr & 0x0000ffff, HFA384X_PCI_M0_ADDRL_OFF);
		HFA384X_OUTW(local->bus_m0_pending_len / 2,
			     HFA384X_PCI_M0_LEN_OFF);
		HFA384X_OUTW(HFA384X_PCI_CTL_TO_BAP, HFA384X_PCI_M0_CTL_OFF);

		local->bus_m0_pending_len = 0;
		return;
	}
#endif /* PRISM2_ENABLE_BEFORE_TX_BUS_MASTER */
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */

  
  if (local->last_txfid_idx == PRISM2_DUMMY_FID) {
    /* no pending transmit command; should not happen since we
     * wait for command completion (and ACK event) with every other
     * command */
    printk(KERN_DEBUG "%s: CMD EV - no pending transmit\n",
	   dev->name);
    return;
  }
  
  local->last_txfid_idx = PRISM2_DUMMY_FID;
  
  resp0 = HFA384X_INW(HFA384X_RESP0_OFF);
  
  if (netif_queue_stopped(dev)) {
    /* last_txfid_idx is free for next TX, so wake up queue that
     * was stopped in prism2_transmit() */
    prism2_netif_wake_queues(dev);
  }
  
  /* FIX: is some locking needed for txfid data? */
  
  /* With reclaim, Resp0 contains new txfid for transmit; the old txfid
   * will be automatically allocated for the next TX frame */
  local->intransmitfid[idx] = resp0;
  
  PDEBUG(DEBUG_FID, "%s: prism2_cmd_ev: txfid[%d]=0x%04x, resp0=0x%04x, "
	 "transmit_txfid=0x%04x\n", dev->name, idx, local->txfid[idx],
	 resp0, local->intransmitfid[local->next_txfid]);
  
  idx++;
  if (idx >= PRISM2_TXFID_COUNT)
    idx = 0;
  local->next_txfid = idx;
  
  /* check if all TX buffers are occupied */
  do {
    if (local->intransmitfid[idx] == PRISM2_TXFID_EMPTY) {
      return;
    }
    idx++;
    if (idx >= PRISM2_TXFID_COUNT)
      idx = 0;
  } while (idx != local->next_txfid);
  
  /* no empty TX buffers, stop queue */
  prism2_netif_stop_queues(dev);
}

#else /* PRISM2_USE_CMD_COMPL_INTERRUPT */

/* Called only from software IRQ */
int prism2_transmit(struct net_device *dev, int idx)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int res, end;
  unsigned long flags;
  u16 resp0;
  
  spin_lock_irqsave(&local->txfidlock, flags);
  
  /* transmit packet */
  res = hfa384x_cmd(dev,
		    HFA384X_CMDCODE_TRANSMIT | HFA384X_CMD_TX_RECLAIM,
		    local->txfid[idx], NULL, &resp0);
  if (res) {
    struct net_device_stats *stats;
    spin_unlock_irqrestore(&local->txfidlock, flags);
    printk("%s: prism2_transmit: CMDCODE_TRANSMIT failed "
	   "(res=%d, resp0=0x%04x)\n", dev->name, res, resp0);
    stats = prism2_get_stats(dev);
    stats->tx_dropped++;
    return 1;
  }
  dev->trans_start = jiffies;
  
  /* With reclaim, Resp0 contains new txfid for transmit; the old txfid
   * will be automatically allocated for the next TX frame */
  local->intransmitfid[idx] = resp0;
  
  PDEBUG(DEBUG_FID, "prism2_transmit: [%d] txfid=0x%04x, "
	 "transmit_txfid=0x%04x\n", idx, local->txfid[idx],
	 local->intransmitfid[local->next_txfid]);
  
  idx++;
  if (idx >= PRISM2_TXFID_COUNT)
    idx = 0;
  end = local->next_txfid = idx;
  
  /* check if all TX buffers are occupied */
  do {
    if (local->intransmitfid[idx] == PRISM2_TXFID_EMPTY) {
      spin_unlock_irqrestore(&local->txfidlock, flags);
      return 0;
    }
    idx++;
    if (idx >= PRISM2_TXFID_COUNT)
      idx = 0;
  } while (idx != end);
  spin_unlock_irqrestore(&local->txfidlock, flags);
  
  /* no empty TX buffers, stop queue */
  prism2_netif_stop_queues(dev);
  
  return 0;
}

#endif /* PRISM2_USE_CMD_COMPL_INTERRUPT */

void prism2_dump_tx_header(const char *name, const struct hfa384x_tx_frame *tx)
{
  u16 fc;
  
  printk(KERN_DEBUG "%s: TX status=0x%04x retry_count=%d tx_rate=%d "
	 "tx_control=0x%04x; jiffies=%ld\n",
	 name, __le16_to_cpu(tx->status), tx->retry_count, tx->tx_rate,
	 __le16_to_cpu(tx->tx_control), jiffies);
  
  fc = __le16_to_cpu(tx->frame_control);
  printk(KERN_DEBUG "   FC=0x%04x (type=%d:%d) dur=0x%04x seq=0x%04x "
	 "data_len=%d%s%s\n",
	 fc, WLAN_FC_GET_TYPE(fc), WLAN_FC_GET_STYPE(fc),
	 __le16_to_cpu(tx->duration_id), __le16_to_cpu(tx->seq_ctrl),
	 __le16_to_cpu(tx->data_len),
	 fc & WLAN_FC_TODS ? " [ToDS]" : "",
	 fc & WLAN_FC_FROMDS ? " [FromDS]" : "");
  
  printk(KERN_DEBUG "   A1=" MACSTR " A2=" MACSTR " A3=" MACSTR " A4="
	 MACSTR "\n",
	 MAC2STR(tx->addr1), MAC2STR(tx->addr2), MAC2STR(tx->addr3),
	 MAC2STR(tx->addr4));
  
  printk(KERN_DEBUG "   dst=" MACSTR " src=" MACSTR " len=%d\n",
	 MAC2STR(tx->dst_addr), MAC2STR(tx->src_addr),
	 __be16_to_cpu(tx->len));
}



enum {
  PRISM2_RX_MONITOR, PRISM2_RX_MGMT, PRISM2_RX_NON_ASSOC,
  PRISM2_RX_NULLFUNC_ACK, PRISM2_RX_RAW
};

int prism2_80211_header_parse(struct sk_buff *skb, unsigned char *haddr)
{
  memcpy(haddr, skb->mac.raw + 10, ETH_ALEN); /* addr2 */
  return ETH_ALEN;
}


#ifdef ASAP_MOD

static int prism2_rx(struct net_device *dev){

  intersil_local_info_t *local = (intersil_local_info_t *)dev->priv;
  sap_config *config = (sap_config *)local->config;

  if (!config)
    return 0;

  if (config->mode == WIFI_MODE_MONITOR) {
    if (local->func->asap.rx_monitor_asap)
      return local->func->asap.rx_monitor_asap(NULL, dev);
    else return 0;
  } else {
    if (local->func->asap.rx_asap)
      return local->func->asap.rx_asap(NULL, dev);
  }

  return 0;
  

}

#else
int prism2_rx_monitor_mode(struct net_device *);



/* Called only from hardware IRQ */
static int prism2_rx(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int res, ret = 0;
  u16 rxfid, status, macport;
  struct net_device_stats *stats;
  struct hfa384x_rx_frame *rxdesc = &local->rxdesc;
  int monitor_flag = 0;

  
  if (local->monitor_type == PRISM2_MONITOR_ON)
    return prism2_rx_monitor_mode(dev);

  stats = prism2_get_stats(dev);
  
  rxfid = HFA384X_INW(HFA384X_RXFID_OFF);
  
  /* Prism2 should support concurrent BAP0 (TX) and BAP1 (RX) access,
   * but it seems to bring some errors when TX path is interrupted by RX
   * event, so require baplock to be free before starting RX handling. */
  spin_lock(&local->baplock);
  res = hfa384x_setup_bap(dev, BAP1, rxfid, 0);
  if (!res)
    res = hfa384x_from_bap(dev, BAP1, rxdesc, sizeof(*rxdesc));
  spin_unlock(&local->baplock);
  if (res) {
    printk(KERN_DEBUG "copy from BAP1 failed %d\n", res);
    if (res == -ETIMEDOUT)
      prism2_hw_reset(dev);
    goto rx_dropped;
  }
  
  
  status = le16_to_cpu(rxdesc->status);
  macport = (status >> 8) & 0x07;
  
  if (macport != 0) {
    // should not happen, so drop the pkt
#if 0
    if (macport == 7) {
      u16 fc;
      monitor_flag = 2;
      fc = le16_to_cpu(rxdesc->frame_control);
      if (WLAN_FC_GET_TYPE(fc) == WLAN_FC_TYPE_DATA)
	monitor_flag |= 1;
      if (WLAN_FC_GET_TYPE(fc) == WLAN_FC_TYPE_CTRL)
	rxdesc->data_len = 0;
      goto read_frame;
    } else {
      printk(KERN_DEBUG "RX: Unknown MACPort %d\n", macport);
      goto rx_dropped;
    }
#else
    goto rx_dropped;
#endif
  }
    
  if (rxdesc->data_len & __constant_cpu_to_le16(0x8000)) {
    printk(KERN_DEBUG "%s: Received frame with invalid length "
	   "0x%04x\n", dev->name, le16_to_cpu(rxdesc->data_len));
    goto rx_dropped;
    
  }
  
  /* FCS errors should not come this far, but let's make sure that frames
   * with errors will be dropped even in Host AP mode */
  if (status & HFA384X_RX_STATUS_FCSERR) {
    printk(KERN_DEBUG "%s: prism2_rx: dropped FCSErr frame "
	   "(status=%02X)\n", dev->name, status);
    goto rx_dropped;
  }
  
#ifdef ASAP_MOD
  if (local->func->asap.rx_asap)
    local->func->asap.rx_asap(dev, rxdesc, monitor_flag);
#else
  prism2_rx_kernel(dev, rxdesc, monitor_flag);
#endif
  
 rx_exit:
  return ret;
  
 rx_dropped:
  stats->rx_dropped++;
  goto rx_exit;
}

#endif // ASAP_MOD

static int prism2_tx_null(struct sk_buff *skb, struct net_device *dev){
	dev_kfree_skb(skb);
	return 0;
}

static void prism2_setup_dev(struct net_device *dev, intersil_local_info_t *local,
			     int main_dev)
{
  ether_setup(dev);
  
  /* kernel callbacks */
  dev->get_stats = prism2_get_stats;
#ifdef WIRELESS_EXT
  dev->get_wireless_stats = main_dev ? prism2_get_wireless_stats : NULL;
#endif
  dev->open = prism2_open;
  dev->stop = prism2_close;
  dev->hard_start_xmit = prism2_tx_null;
  //  dev->hard_start_xmit = NULL;
#ifdef HAVE_MULTICAST
  /* FIX: to be implemented as soon as Prism2 supports GroupAddresses
   * and correct documentation is available */
  
  /* dev->set_multicast_list = prism2_set_multicast_list; */
#endif
#ifdef HAVE_PRIVATE_IOCTL
  dev->do_ioctl = main_dev ? prism2_ioctl : NULL;
#endif
#ifdef HAVE_CHANGE_MTU
  dev->change_mtu = prism2_change_mtu;
#endif
#ifdef HAVE_TX_TIMEOUT
  dev->tx_timeout = prism2_tx_timeout;
  dev->watchdog_timeo = TX_TIMEOUT;
#endif
  
  dev->mtu = mtu;
  
  netif_stop_queue(dev);
}


/* Called only from hardware IRQ */
static void prism2_alloc_ev(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int idx;
  u16 fid;
  
  fid = HFA384X_INW(HFA384X_ALLOCFID_OFF);
  
  PDEBUG(DEBUG_FID, "FID: interrupt: ALLOC - fid=0x%04x\n", fid);
  
  idx = local->next_alloc;
  
  do {
    if (local->txfid[idx] == fid) {
      PDEBUG(DEBUG_FID, "FID: found matching txfid[%d]\n",
	     idx);
      
#ifndef PRISM2_USE_TX_INTERRUPT
      /* Alloc events are used only for TX frames, so
       * increase TX packet counter here. This will make
       * separate TX event handling unnecessary. */
      local->stats.tx_packets++;
#endif /* PRISM2_USE_TX_INTERRUPT */
      
#ifndef final_version
      if (local->intransmitfid[idx] == PRISM2_TXFID_EMPTY)
	printk("Already released txfid found at idx "
	       "%d\n", idx);
      if (local->intransmitfid[idx] == PRISM2_TXFID_RESERVED)
	printk("Already reserved txfid found at idx "
	       "%d\n", idx);
#endif
      local->intransmitfid[idx] = PRISM2_TXFID_EMPTY;
      idx++;
      local->next_alloc = idx >= PRISM2_TXFID_COUNT ? 0 :
	idx;
      
#ifdef PRISM2_USE_CMD_COMPL_INTERRUPT
      if (local->last_txfid_idx == PRISM2_DUMMY_FID &&
	  netif_queue_stopped(dev))
	prism2_netif_wake_queues(dev);
#else /* PRISM2_USE_CMD_COMPL_INTERRUPT */
      if (netif_queue_stopped(dev))
	prism2_netif_wake_queues(dev);
#endif /* PRISM2_USE_CMD_COMPL_INTERRUPT */
      
      goto out;
    }
    
    idx++;
    if (idx >= PRISM2_TXFID_COUNT)
      idx = 0;
  } while (idx != local->next_alloc);
  
  printk(KERN_WARNING "%s: could not found matching txfid (0x%04x) for "
	 "alloc event\n", dev->name, HFA384X_INW(HFA384X_ALLOCFID_OFF));
  
 out:
}

static void prism2_txexc(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  u16 fid;
  int res;
  struct hfa384x_tx_frame *txdesc;
  struct sk_buff *skb;
  
  local->stats.tx_errors++;
  fid = HFA384X_INW(HFA384X_TXCOMPLFID_OFF);
  //  PDEBUG(DEBUG_EXTRA, "%s: TXEXC - fid=0x%04x - ", dev->name, fid);
  
  skb = dev_alloc_skb(sizeof(*txdesc)+16);
  if (!skb) {
    printk(KERN_WARNING "%s: prism2_rx_80211 cannot allocate "
	   "buffer\n", dev->name);
    return;
  }
  
  /* align IP on 16b boundary */
  skb_reserve(skb, 2);
  
  skb_put(skb, sizeof(*txdesc));
  txdesc = (struct hfa384x_tx_frame *)skb->data;

  res = hfa384x_setup_bap(dev, BAP1, fid, 0);
  if (!res)
    res = hfa384x_from_bap(dev, BAP1, txdesc, sizeof(*txdesc));
  if (res) {
    PDEBUG2(DEBUG_EXTRA, "could not get TXEXC txframe\n");
    if (res == -ETIMEDOUT)
      prism2_hw_reset(dev);
    return;
  }

  txdesc->len = txdesc->data_len = 0;

  skb->dev = dev;
  skb->mac.raw = skb->data;
  skb->pkt_type = PACKET_HOST;
  skb->protocol = 0;
  netif_rx(skb);
  
  return;


}
  

static const char* hfa384x_linkstatus_str(u16 linkstatus)
{
  switch (linkstatus) {
  case HFA384X_LINKSTATUS_CONNECTED:
    return "Connected";
  case HFA384X_LINKSTATUS_DISCONNECTED:
    return "Disconnected";
  case HFA384X_LINKSTATUS_AP_CHANGE:
    return "Access point change";
  case HFA384X_LINKSTATUS_AP_OUT_OF_RANGE:
    return "Access point out of range";
  case HFA384X_LINKSTATUS_AP_IN_RANGE:
    return "Access point in range";
  case HFA384X_LINKSTATUS_ASSOC_FAILED:
    return "Association failed";
  default:
    return "Unknown";
  }
}


/* Called only from hardware IRQ */
static void prism2_info(struct net_device *dev)
{
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  u16 fid, val, *pos, len;
  int res, i, left;
  struct hfa384x_info_frame info;
  unsigned char buf[1024];
  struct hfa384x_comm_tallies *tallies;
  struct hfa384x_scan_result *scanres;
  
  fid = HFA384X_INW(HFA384X_INFOFID_OFF);
  
  /* Prism2 should support concurrent BAP0 (TX) and BAP1 (info) access,
   * but it seems to bring some errors when TX path is interrupted by
   * info event, so require baplock to be free before starting info frame
   * handling. */
  spin_lock(&local->baplock);
  res = hfa384x_setup_bap(dev, BAP1, fid, 0);
  if (!res)
    res = hfa384x_from_bap(dev, BAP1, &info, sizeof(info));
  spin_unlock(&local->baplock);
  if (res) {
    printk(KERN_DEBUG "Could not get info frame (fid=0x%04x)\n",
	   fid);
    if (res == -ETIMEDOUT)
      prism2_hw_reset(dev);
    return;
  }
  
  le16_to_cpus(&info.len);
  le16_to_cpus(&info.type);
  left = (info.len - 1) * 2;
  
  if (info.len & 0x8000 || info.len == 0 || left > sizeof(buf)) {
    /* data register seems to give 0x8000 in some error cases even
     * though busy bit is not set in offset register;
     * in addition, length must be at least 1 due to type field */
    printk(KERN_DEBUG "%s: Received info frame with invalid "
	   "length 0x%04x\n", dev->name, info.len);
    return;
  }
  
  if (left > 0 && hfa384x_from_bap(dev, BAP1, &buf, left)) {
    printk(KERN_WARNING "%s: Info frame read failed (fid=0x%04x, "
	   "len=0x%04x, type=0x%04x\n",
	   dev->name, fid, info.len, info.type);
    return;
  }
  
  if (info.type != HFA384X_INFO_COMMTALLIES)
    PDEBUG(DEBUG_EXTRA, "%s: INFO - fid=0x%04x - len=%d "
	   "type=0x%04x\n", dev->name, fid, info.len, info.type);
  
  switch (info.type) {
  case HFA384X_INFO_COMMTALLIES:
    if (left < sizeof(struct hfa384x_comm_tallies)) {
      printk(KERN_DEBUG "%s: two short (len=%d) commtallies "
	     "info frame\n", dev->name, left);
      break;
    }
    
    tallies = (struct hfa384x_comm_tallies *) buf;
#define ADD_COMM_TALLIES(name) local->comm_tallies.name += tallies->name
    ADD_COMM_TALLIES(tx_unicast_frames);
    ADD_COMM_TALLIES(tx_multicast_frames);
    ADD_COMM_TALLIES(tx_fragments);
    ADD_COMM_TALLIES(tx_unicast_octets);
    ADD_COMM_TALLIES(tx_multicast_octets);
    ADD_COMM_TALLIES(tx_deferred_transmissions);
    ADD_COMM_TALLIES(tx_single_retry_frames);
    ADD_COMM_TALLIES(tx_multiple_retry_frames);
    ADD_COMM_TALLIES(tx_retry_limit_exceeded);
    ADD_COMM_TALLIES(tx_discards);
    ADD_COMM_TALLIES(rx_unicast_frames);
    ADD_COMM_TALLIES(rx_multicast_frames);
    ADD_COMM_TALLIES(rx_fragments);
    ADD_COMM_TALLIES(rx_unicast_octets);
    ADD_COMM_TALLIES(rx_multicast_octets);
    ADD_COMM_TALLIES(rx_fcs_errors);
    ADD_COMM_TALLIES(rx_discards_no_buffer);
    ADD_COMM_TALLIES(tx_discards_wrong_sa);
    ADD_COMM_TALLIES(rx_discards_wep_undecryptable);
    ADD_COMM_TALLIES(rx_message_in_msg_fragments);
    ADD_COMM_TALLIES(rx_message_in_bad_msg_fragments);
#undef ADD_COMM_TALLIES
    break;
    
  case HFA384X_INFO_LINKSTATUS:
    if (left != 2) {
      printk(KERN_DEBUG "%s: invalid linkstatus info frame "
	     "length %d\n", dev->name, left);
      break;
    }
    val = buf[0] | (buf[1] << 8);
    PDEBUG(DEBUG_EXTRA, "  LinkStatus=%d (%s)\n",
	   val, hfa384x_linkstatus_str(val));
    break;
    
  case HFA384X_INFO_SCANRESULTS:
    if (left < 4) {
      printk(KERN_DEBUG "%s: invalid scanresult info frame "
	     "length %d\n", dev->name, left);
      break;
    }
    PDEBUG(DEBUG_EXTRA, "  ScanResults:\n");
    pos = (u16 *) buf;
    PDEBUG(DEBUG_EXTRA, "    Reserved=0x%04x\n",
	   __le16_to_cpu(*pos));
    pos++;
    PDEBUG(DEBUG_EXTRA, "    ScanReason=%d\n",
	   __le16_to_cpu(*pos));
    pos++;
    left -= 4;
    
    scanres = (struct hfa384x_scan_result *) pos;
    val = 1;
    while (left >= sizeof(struct hfa384x_scan_result)) {
      /* FIX: it might not be that wise to print a lot
       * from here.. this is called from hw interrupt
       * handler.. */
      PDEBUG(DEBUG_EXTRA, "    ScanResult %d:\n", val);
      PDEBUG(DEBUG_EXTRA, "      CHID=%d\n",
	     __le16_to_cpu(scanres->chid));
      PDEBUG(DEBUG_EXTRA, "      ANL=%d\n",
	     __le16_to_cpu(scanres->anl));
      PDEBUG(DEBUG_EXTRA, "      SL=%d\n",
	     __le16_to_cpu(scanres->sl));
      PDEBUG(DEBUG_EXTRA, "      bssid=" MACSTR "\n",
	     MAC2STR(scanres->bssid));
      PDEBUG(DEBUG_EXTRA, "      BcnInt=%d\n",
	     __le16_to_cpu(scanres->beacon_interval));
      PDEBUG(DEBUG_EXTRA, "      Capability=0x%04x\n",
	     __le16_to_cpu(scanres->capability));
      PDEBUG(DEBUG_EXTRA, "      SSID='");
      len = __le16_to_cpu(*(u16 *) &scanres->ssid);
      if (len > 32)
	len = 32;
      for (i = 0; i < len; i++) {
	unsigned char c = scanres->ssid[2 + i];
	if (c >= 32 && c < 127)
	  PDEBUG2(DEBUG_EXTRA, "%c", c);
	else
	  PDEBUG2(DEBUG_EXTRA, "<%02x>", c);
      }
      PDEBUG2(DEBUG_EXTRA, "'\n");
      PDEBUG(DEBUG_EXTRA, "      SupRates=");
      for (i = 0; i < sizeof(scanres->sup_rates); i++)
	PDEBUG2(DEBUG_EXTRA, "%02x ",
		scanres->sup_rates[i]);
      PDEBUG2(DEBUG_EXTRA, "\n");
      PDEBUG(DEBUG_EXTRA, "      Rate=%d\n",
	     __le16_to_cpu(scanres->rate));
      val++;
      left -= sizeof(struct hfa384x_scan_result);
      scanres++;
    }
    
    if (left > 0) {
      unsigned char *c = (unsigned char *) scanres;
      PDEBUG(DEBUG_EXTRA, "ScanRes extra: left=%d "
	     "(expected %dx):", left,
	     sizeof(struct hfa384x_scan_result));
      for (i = 0; i < left; i++)
	PDEBUG2(DEBUG_EXTRA, " %02x", *c++);
      PDEBUG2(DEBUG_EXTRA, "\n");
    }
    
    break;
    
  default:
    PDEBUG(DEBUG_EXTRA, "Unknown info frame:");
    for (i = 0; i < (left < 100 ? left : 100); i++)
      PDEBUG2(DEBUG_EXTRA, " %02x", buf[i]);
    PDEBUG2(DEBUG_EXTRA, "\n");
  }
}


#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
/* Called only from hardware IRQ */
static void prism2_bus_master_ev(struct net_device *dev, int bap)
{
	if (bap == BAP1)
		prism2_rx2(dev, NULL);
	else {
		intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
		if (prism2_transmit(dev, local->bus_m0_tx_idx)) {
			printk(KERN_DEBUG "%s: prism2_transmit() failed "
			       "when called from bus master event\n",
			       dev->name);
			local->intransmitfid[local->bus_m0_tx_idx] =
				PRISM2_TXFID_EMPTY;
			prism2_hw_reset(dev);
		}

		local->bus_m0_in_use = 0;
	}
}
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */


/* Called only from hardware IRQ */
void prism2_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  struct net_device *dev = (struct net_device *) dev_id;
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int events = 0;
  u16 ev;
  
  if (local->func->card_present && !local->func->card_present(local)) {
    printk(KERN_DEBUG "%s: Interrupt, but dev not OK\n",
	   dev->name);
    return;
  }
  
  if (!local->hw_ready || local->hw_resetting || !local->dev_enabled) {
    ev = HFA384X_INW(HFA384X_EVSTAT_OFF);
    if (local->dev_enabled)
      printk(KERN_DEBUG "%s: prism2_interrupt: hw not ready;"
	     " skipping events 0x%04x\n", dev->name, ev);
    HFA384X_OUTW(ev, HFA384X_EVACK_OFF);
    return;
  }
  

  for (;;) {
#ifdef PRISM2_USE_CMD_COMPL_INTERRUPT
    /* SMP platforms may get command completion events on other
     * CPUs during hfa384x_cmd() execution, so use spin lock to
     * prevent problems */
    spin_lock(&local->cmdlock);
    ev = HFA384X_INW(HFA384X_EVSTAT_OFF);
    spin_unlock(&local->cmdlock);
#else /* PRISM2_USE_CMD_COMPL_INTERRUPT */
    ev = HFA384X_INW(HFA384X_EVSTAT_OFF);
#endif /* PRISM2_USE_CMD_COMPL_INTERRUPT */
    if ((ev & HFA384X_EVENT_MASK) == 0)
      break;
    if (ev == 0xffff) {
      HFA384X_OUTW(0xffff, HFA384X_EVACK_OFF);
      printk(KERN_DEBUG "%s: prism2_interrupt: ev=0xffff\n",
	     dev->name);
      prism2_hw_reset(dev);
      return;
    }

#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
    if (local->bus_m1_in_use &&
	(ev & (HFA384X_EV_RX | HFA384X_EV_INFO))) {
      /* delay RX and info handling until bus master read of
       * previously received frame is completed */
      ev &= ~(HFA384X_EV_RX | HFA384X_EV_INFO);
      if ((ev & HFA384X_EVENT_MASK) == 0)
	break;
    }

    if (ev & HFA384X_EV_PCI_M0) {
      prism2_bus_master_ev(dev, BAP0);
      HFA384X_OUTW(HFA384X_EV_PCI_M0, HFA384X_EVACK_OFF);
    }
    
    if (ev & HFA384X_EV_PCI_M1) {
      /* previous RX has been copied can be ACKed now */
      HFA384X_OUTW(HFA384X_EV_RX, HFA384X_EVACK_OFF);
      
      prism2_bus_master_ev(dev, BAP1);
      HFA384X_OUTW(HFA384X_EV_PCI_M1, HFA384X_EVACK_OFF);
    }
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */
    
    
#ifdef PRISM2_USE_CMD_COMPL_INTERRUPT
    if (ev & HFA384X_EV_CMD) {
      prism2_cmd_ev(dev);
      HFA384X_OUTW(HFA384X_EV_CMD, HFA384X_EVACK_OFF);
    }
#endif /* PRISM2_USE_CMD_COMPL_INTERRUPT */
    
    if (ev & HFA384X_EV_ALLOC) {
      prism2_alloc_ev(dev);
      HFA384X_OUTW(HFA384X_EV_ALLOC, HFA384X_EVACK_OFF);
    }
    
#ifdef PRISM2_USE_TX_INTERRUPT
    if (ev & HFA384X_EV_TX) {
      local->stats.tx_packets++;
      PDEBUG(DEBUG_FID, "interrupt: TX - fid=0x%04x\n",
	     HFA384X_INW(HFA384X_TXCOMPLFID_OFF));
      HFA384X_OUTW(HFA384X_EV_TX, HFA384X_EVACK_OFF);
    }
#endif /* PRISM2_USE_TX_INTERRUPT */
    
    if (ev & HFA384X_EV_TXEXC) {
      prism2_txexc(dev);
      HFA384X_OUTW(HFA384X_EV_TXEXC, HFA384X_EVACK_OFF);
    }    
    if (ev & HFA384X_EV_RX) {
      if (!prism2_rx(dev))
	HFA384X_OUTW(HFA384X_EV_RX, HFA384X_EVACK_OFF);
    }
    
#ifndef final_version
    if (ev & HFA384X_EV_WTERR) {
      PDEBUG(DEBUG_EXTRA, "%s: WTERR event\n", dev->name);
      HFA384X_OUTW(HFA384X_EV_WTERR, HFA384X_EVACK_OFF);
    }
#endif /* final_version */
    
    if (ev & HFA384X_EV_INFO) {
      prism2_info(dev);
      HFA384X_OUTW(HFA384X_EV_INFO, HFA384X_EVACK_OFF);
    }
    
    if (ev & HFA384X_EV_INFDROP) {
      PDEBUG(DEBUG_EXTRA, "%s: INFDROP event\n", dev->name);
      HFA384X_OUTW(HFA384X_EV_INFDROP, HFA384X_EVACK_OFF);
    }
    
    events++;
    if (events >= PRISM2_MAX_INTERRUPT_EVENTS) {
      PDEBUG(DEBUG_EXTRA, "prism2_interrupt: >%d events "
	     "(EvStat=0x%04x)\n",
	     PRISM2_MAX_INTERRUPT_EVENTS,
	     HFA384X_INW(HFA384X_EVSTAT_OFF));
      break;
    }
  }
}



int prism2_monitor_enable(void *data){

  int retval;
  intersil_local_info_t *local = (intersil_local_info_t *)data;
  struct net_device *dev = local->dev;
  
  retval = prism2_set_word(dev, HFA384X_RID_CNFPORTTYPE, 
			   HFA384X_PORTTYPE_PSEUDO_IBSS);

  if (retval){
    printk("Port type setting for monitor mode failed\n");
    return -1;
  }

  prism2_reset_port(dev);
  hfa384x_cmd(dev, HFA384X_CMDCODE_TEST |
	      (HFA384X_TEST_MONITOR << 8),
	      0, NULL, NULL);

  printk("Monitor mode enabled\n");

  return 0;

}

int prism2_monitor_disable(void *data){
  int retval;
  intersil_local_info_t *local = (intersil_local_info_t *)data;
  sap_config *config = (sap_config *)local->config;
  struct net_device *dev = local->dev;
  
  hfa384x_cmd(dev, HFA384X_CMDCODE_TEST |
	      (HFA384X_TEST_STOP << 8),
	      0, NULL, NULL);
  retval = prism2_set_word(dev, HFA384X_RID_CNFPORTTYPE,
			   config->mode);
  prism2_reset_port(dev);
  if (retval){
    printk("Error setting porttype to %d\n", config->mode);
  } else {
    printk("Monitor mode disabled\n");
  }
  return retval;
  

}



static void prism2_check_sta_fw_version(intersil_local_info_t *local)
{
  struct hfa384x_comp_ident comp;
  int id, variant, major, minor;
  
  if (hfa384x_get_rid(local->dev, HFA384X_RID_STAID,
		      &comp, sizeof(comp), 1) < 0)
    return;
  
  local->fw_ap = 0;
  id = le16_to_cpu(comp.id);
  if (id != HFA384X_COMP_ID_STA) {
    if (id == HFA384X_COMP_ID_FW_AP)
      local->fw_ap = 1;
    return;
  }
  
  major = __le16_to_cpu(comp.major);
  minor = __le16_to_cpu(comp.minor);
  variant = __le16_to_cpu(comp.variant);
  
  /* no firmware version specific configuration for the base driver yet,
   * so just call AP specific parts */
  //ap_check_sta_fw_version(local->ap, major, minor, variant);
}

/* Called only from software IRQ */
static void handle_bridged_queue(void *data)
{
  intersil_local_info_t *local = (intersil_local_info_t *) data;
  struct sk_buff *skb;
  
  while ((skb = skb_dequeue(&local->bridge_list)) != NULL)
    dev_queue_xmit(skb);
  
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
  //  MOD_DEC_USE_COUNT;
#endif
}


intersil_local_info_t *prism2_init_local_data(struct prism2_helper_functions *funcs,
				     int card_idx)
{
  intersil_local_info_t *local;
  int len, i;
  
  if (funcs == NULL)
    return NULL;
  
  local = kmalloc(sizeof(intersil_local_info_t), GFP_KERNEL);
  if (local == NULL)
    return NULL;
  
  memset(local, 0, sizeof(intersil_local_info_t));

#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
  local->bus_m0_buf = (u8 *) kmalloc(sizeof(struct hfa384x_tx_frame) +
				     PRISM2_DATA_MAXLEN, GFP_DMA);
  if (local->bus_m0_buf == NULL)
    goto fail;
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */
  
  local->bus_m1_buf = (u8 *) kmalloc(PRISM2_DATA_MAXLEN, GFP_DMA);
  if (local->bus_m1_buf == NULL)
    goto fail;
  
  local->dev = kmalloc(sizeof(struct net_device) + PRISM2_NETDEV_EXTRA,
		       GFP_KERNEL);
  if (local->dev == NULL)
    goto fail;
  memset(local->dev, 0, sizeof(struct net_device) + PRISM2_NETDEV_EXTRA);
  prism2_set_dev_name(local->dev, local->dev + 1);
  local->dev->priv = local;
  
  local->func = funcs;
  
  spin_lock_init(&local->txfidlock);
  spin_lock_init(&local->cmdlock);
  spin_lock_init(&local->baplock);
  spin_lock_init(&local->wdslock);
  
  if (card_idx < 0 || card_idx >= MAX_PARM_DEVICES)
    card_idx = 0;
  local->card_idx = card_idx;
  
  i = essid[card_idx] == NULL ? 0 : card_idx;
  len = strlen(essid[i]);
  memcpy(local->essid, essid[i],
	 len > MAX_SSID_LEN ? MAX_SSID_LEN : len);
  local->essid[MAX_SSID_LEN] = '\0';
#ifdef WIRELESS_EXT
  i = GET_INT_PARM(iw_mode, card_idx);
  if (i >= 1 && i <= 3) {
    local->iw_mode = i;
  } else {
    printk(KERN_WARNING "prism2: Unknown iw_mode %d; using "
	   "IW_MODE_MASTER\n", i);
    local->iw_mode = IW_MODE_MASTER;
  }
#endif
  local->channel = GET_INT_PARM(channel, card_idx);
  local->beacon_int = GET_INT_PARM(beacon_int, card_idx);
  local->dtim_period = GET_INT_PARM(dtim_period, card_idx);
  local->wds_max_connections = 16;
  local->tx_control = HFA384X_TX_CTRL_FLAGS;
  local->manual_retry_count = -1;
#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
	local->bus_master_threshold_rx = GET_INT_PARM(bus_master_threshold_rx,
						      card_idx);
	local->bus_master_threshold_tx = GET_INT_PARM(bus_master_threshold_tx,
						      card_idx);
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */

  skb_queue_head_init(&local->bridge_list);
  
  /* Initialize task queue structure for bridged packet handling */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
  local->bridge_queue.next = NULL;
#else
  INIT_LIST_HEAD(&local->bridge_queue.list);
#endif
  local->bridge_queue.sync = 0;
  local->bridge_queue.routine = handle_bridged_queue;
  local->bridge_queue.data = local;
  
  prism2_setup_dev(local->dev, local, 1);
  
  //  local->dev->type = ARPHRD_IEEE80211;
  //	local->dev->hard_header_parse = prism2_80211_header_parse;
  local->dev->hard_header_parse = NULL;
  
  local->saved_eth_header_parse = local->dev->hard_header_parse;

#ifdef ASAP_MOD

  local->func->reset_port = prism2_reset_port;
  local->func->hw_reset = prism2_hw_reset;
  local->func->from_bap = hfa384x_from_bap;
  local->func->cmd = hfa384x_cmd;
  local->func->set_rid = hfa384x_set_rid;
  local->func->asap.set_string = prism2_set_string;
  local->func->asap.set_word = prism2_set_word;
  local->func->asap.setup_bap = hfa384x_setup_bap;
  local->func->asap.monitor_enable = prism2_monitor_enable;
  local->func->asap.monitor_disable = prism2_monitor_disable;

  local->func->asap.rx_asap = NULL;
  local->func->asap.tx_asap = prism2_tx_kernel;
  local->func->asap.rx_monitor_asap = NULL;


#endif

  
  return local;
  
 fail:
#if 0
  if (local->ap != NULL)
    kfree(local->ap);
#endif
  if (local->dev != NULL)
    kfree(local->dev);
#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
  if (local->bus_m0_buf)
    kfree(local->bus_m0_buf);
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */
  if (local->bus_m1_buf)
    kfree(local->bus_m1_buf);
  kfree(local);
  return NULL;
}


int prism2_init_dev(intersil_local_info_t *local)
{
  struct net_device *dev = local->dev;
  
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
  {
    int i = 0;
    do {
      sprintf(dev->name, "wlan%d", i++);
    } while (dev_get(dev->name));
  }
#else
  memcpy(dev->name, "wlanB", 7);
#endif
  
  dev->type = ARPHRD_ETHER;
  dev->addr_len = 6;
  dev->hard_header_len = 14;
  dev->mtu = 1600;
  
  dev->tx_queue_len = 100;
  
  if (register_netdev(dev)) {
    printk(KERN_WARNING "prism2: register_netdev() failed!\n");
    return 1;
  }
  printk(KERN_INFO "Registered netdevice %s\n", 
	 dev->name);
  
#ifndef ASAP_MOD
  // initialize the forwarder

  sap_switch_init(dev);
  gre_init(dev);

  sap_chr_init(dev);

  softap_init_proc(dev);
#endif // ASAP_MOD
  
  return 0;
}


void prism2_free_local_data(intersil_local_info_t *local)
{
  int i;
  
  if (local == NULL)
    return;
  
  if (local->dev && local->dev->name && local->dev->name[0]) {
    unregister_netdev(local->dev);
    printk(KERN_INFO "Netdevice %s unregistered\n",
	   local->dev->name);
    kfree(local->dev);
  }

#ifndef ASAP_MOD
  sap_switch_cleanup();
  gre_cleanup();
  sap_chr_cleanup();

  softap_takedown_proc();
#endif // ASAP_MOD

#ifdef PRISM2_MONITOR
  if (local->nl_monitor != NULL) {
    if (local->nl_monitor->socket == NULL)
      printk("nl_monitor->socket == NULL\n");
    /* this seems to crash the kernel.. */
    /* sock_release(local->nl_monitor->socket); */
  }
#endif /* PRISM2_MONITOR */
  
  for (i = 0; i < PRISM2_FRAG_CACHE_LEN; i++) {
    if (local->frag_cache[i].skb != NULL)
      dev_kfree_skb(local->frag_cache[i].skb);
  }
  
  
#if defined(PRISM2_PCI) && defined(PRISM2_BUS_MASTER)
  if (local->bus_m0_buf)
    kfree(local->bus_m0_buf);
#endif /* PRISM2_PCI and PRISM2_BUS_MASTER */
  if (local->bus_m1_buf)
    kfree(local->bus_m1_buf);
  
  if (local->pda)
    kfree(local->pda);
  
  kfree(local);
  
}

#ifdef ASAP_MOD

int prism2_tx_kernel(void *desc, struct sk_buff *skb, struct net_device *dev){
  
  struct hfa384x_tx_frame *txdesc = (struct hfa384x_tx_frame *)desc;
  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int res, idx = -1, ret = 1;
  struct net_device_stats *stats = &local->stats;
  unsigned long flags;

  if ((local->func->card_present && !local->func->card_present(local)) ||
      !local->hw_ready) {
    printk(KERN_DEBUG "%s: prism2_tx: hw not ready - skipping\n",
	   dev->name);
    ret = 0;
    goto fail;
  }
  

  if (local->monitor_type == PRISM2_MONITOR_ON){
    
    // need to get it out of monitor mode before attempting to transmit
    
    hfa384x_cmd(dev, HFA384X_CMDCODE_TEST |
		(HFA384X_TEST_STOP << 8),
		0, NULL, NULL);
    ret = prism2_set_word(dev, HFA384X_RID_CNFPORTTYPE,
			  HFA384X_PORTTYPE_HOSTAP);
  }
  
  idx = prism2_get_txfid_idx(local);
  if (idx < 0){
    ret = 0;
  } else {
  
    spin_lock_irqsave(&local->baplock, flags);
    res = hfa384x_setup_bap(dev, BAP0, local->txfid[idx], 0);
    if (!res)
      res = hfa384x_to_bap(dev, BAP0, txdesc, sizeof(struct hfa384x_tx_frame));
    if (!res){
      if (skb->len > 0){
	if (skb_shinfo(skb)->frag_list){
	  
	  struct sk_buff *skb2;
	  int tlen = 0;
	  // the data exists on multiple skbs
	  res = hfa384x_to_bap(dev, BAP0, skb->data, skb_headlen(skb));
	  tlen += skb_headlen(skb);
	  skb2 = skb_shinfo(skb)->frag_list;
	  while (skb2 && (tlen < skb->len)){
	    res = hfa384x_to_bap(dev, BAP0, skb2->data, skb2->len);
	    tlen += skb2->len;
	    skb2 = skb2->next;
	  }

	} else {
	  
	  res = hfa384x_to_bap(dev, BAP0, skb->data, skb->len);
	}
      }
    }
    spin_unlock_irqrestore(&local->baplock, flags);
    
    if (!res)
      res = prism2_transmit(dev, idx);
    if (res) {
      printk("%s: prism2_tx_kernel - to BAP0 failed\n", dev->name);
      local->intransmitfid[idx] = PRISM2_TXFID_EMPTY;
      prism2_hw_reset(dev);
      ret = 0;
    }
  }

  if (local->monitor_type == PRISM2_MONITOR_ON){

    // need to set it back to monitor mode
    // the card seems to need this delay, otherwise it complains about 
    // a timeout on the above transmitted laser beam
    udelay(250);

    res = prism2_set_word(dev, HFA384X_RID_CNFPORTTYPE,
			  HFA384X_PORTTYPE_PSEUDO_IBSS);
    if (res) {
      printk(KERN_DEBUG "80211_tx : Port type setting for monitor mode "
	     "failed\n");
      //      return ret;
    }
    hfa384x_cmd(dev, HFA384X_CMDCODE_TEST |
		(HFA384X_TEST_MONITOR << 8),
		0, NULL, NULL);

  }

  stats->tx_packets++;
  stats->tx_bytes += skb->len;

 fail:
    dev_kfree_skb(skb);
  
  return 0;
  


}
#else

int prism2_tx_kernel(struct sk_buff *skb, struct net_device *dev,
		     struct hfa384x_tx_frame *txdesc){

  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  int res, idx = -1, ret = 1;
  //  struct net_device_stats *stats;
  unsigned long flags;

  if ((local->func->card_present && !local->func->card_present(local)) ||
      !local->hw_ready) {
    printk(KERN_DEBUG "%s: prism2_tx: hw not ready - skipping\n",
	   dev->name);
    ret = 0;
    goto fail;
  }
  

  if (local->monitor_type == PRISM2_MONITOR_ON){
    
    // need to get it out of monitor mode before attempting to transmit
    
    hfa384x_cmd(dev, HFA384X_CMDCODE_TEST |
		(HFA384X_TEST_STOP << 8),
		0, NULL, NULL);
    ret = prism2_set_word(dev, HFA384X_RID_CNFPORTTYPE,
			  HFA384X_PORTTYPE_HOSTAP);
  }
  
  idx = prism2_get_txfid_idx(local);
  if (idx < 0){
    ret = 0;
  } else {
  
    spin_lock_irqsave(&local->baplock, flags);
    res = hfa384x_setup_bap(dev, BAP0, local->txfid[idx], 0);
    if (!res)
      res = hfa384x_to_bap(dev, BAP0, txdesc, sizeof(struct hfa384x_tx_frame));
    if (!res){
      if (skb->len > 0){
	if (skb_shinfo(skb)->frag_list){
	  
	  struct sk_buff *skb2;
	  int tlen = 0;
	  // the data exists on multiple skbs
	  res = hfa384x_to_bap(dev, BAP0, skb->data, skb_headlen(skb));
	  tlen += skb_headlen(skb);
	  skb2 = skb_shinfo(skb)->frag_list;
	  while (skb2 && (tlen < skb->len)){
	    res = hfa384x_to_bap(dev, BAP0, skb2->data, skb2->len);
	    tlen += skb2->len;
	    skb2 = skb2->next;
	  }

	} else {
	  
	  res = hfa384x_to_bap(dev, BAP0, skb->data, skb->len);
	}
      }
    }
    spin_unlock_irqrestore(&local->baplock, flags);
    
    if (!res)
      res = prism2_transmit(dev, idx);
    if (res) {
      printk("%s: prism2_tx_kernel - to BAP0 failed\n", dev->name);
      local->intransmitfid[idx] = PRISM2_TXFID_EMPTY;
      prism2_hw_reset(dev);
      ret = 0;
    }
  }

  if (local->monitor_type == PRISM2_MONITOR_ON){

    // need to set it back to monitor mode
    // the card seems to need this delay, otherwise it complains about 
    // a timeout on the above transmitted laser beam
    udelay(250);

    res = prism2_set_word(dev, HFA384X_RID_CNFPORTTYPE,
			  HFA384X_PORTTYPE_PSEUDO_IBSS);
    if (res) {
      printk(KERN_DEBUG "80211_tx : Port type setting for monitor mode "
	     "failed\n");
      //      return ret;
    }
    hfa384x_cmd(dev, HFA384X_CMDCODE_TEST |
		(HFA384X_TEST_MONITOR << 8),
		0, NULL, NULL);

  }

 fail:
    dev_kfree_skb(skb);
  
  return 0;
  

}

#endif // ASAP_MOD

#ifndef ASAP_MOD
int process_am_frame(struct hfa384x_rx_frame *, char *);


int prism2_rx_kernel(struct net_device *dev, 
		     struct hfa384x_rx_frame *rxdesc,
		     int flag){

  // two skbs need to be created, 
  // one for the packet to the forwarding engine
  // and another for the header/frame to AM
  
#define GRE_HEADROOM 24

  struct sk_buff *skb;
  int hdrlen, len, flen;
  int res = 0;
  //  static __u8 abuf[2048];
  

  hdrlen = sizeof(*rxdesc);
  len = __le16_to_cpu(rxdesc->data_len);
  flen = sizeof(struct ieee80211_hdr) + len;
  if (len > PRISM2_DATA_MAXLEN) {
    printk(KERN_WARNING "%s: prism2_rx_80211: len(%d) > "
    	   "MAX(%d)\n", dev->name, len, PRISM2_DATA_MAXLEN);
    
    return 0;
  }


  
  // not in monitor mode
  if (len > 0){
    
    __u8 *pos;
    
    skb = dev_alloc_skb(GRE_HEADROOM + flen);
    if (!skb) {
      printk(KERN_WARNING "%s: prism2_rx_kernel: cannot allocate buffer\n", 
	     dev->name);
      return 0;
    }
    
    dev->last_rx = jiffies;
    
    skb_reserve(skb, GRE_HEADROOM);
    skb->nh.raw = skb->data;
    pos = skb_put(skb, flen);
    memcpy(pos, &rxdesc->frame_control, sizeof(struct ieee80211_hdr));
    pos += sizeof(struct ieee80211_hdr);
    res = hfa384x_from_bap(dev, BAP1, pos, len);
    if (res) {
      printk(KERN_WARNING "%s: prism2_rx_kernel: rx from_bap failed\n",
	     dev->name);
      dev_kfree_skb_irq(skb);
      return 0;
    }
    
    if (!(flag & 1)){
      process_am_frame(rxdesc, skb->data+sizeof(struct ieee80211_hdr));
    }
    
    skb->dev = dev;

    skb_queue_tail(&sap_readq, skb);
    tasklet_schedule(&sap_tasklet);
  }
  
  
  return 0;
}


#endif // ASAP_MOD
