
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wireless.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>

#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>


#include "prism2_config.h"
#include "prism2_wlan.h"

static int backend_mode = 0;
u8 mx_mac[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

struct sap_proc_data {
  char *rbuf;
  char *wbuf;
  int rlen;
  int wlen;
  int max_wlen;
  int flag;
  void (*when_closed)(struct inode *, struct file *);
};

#define NUM_MODES 3
struct _mode_d {
  int mode;
  char *descr;
} mode_d[NUM_MODES] = {
  {0, "Undefined"},
  {1, "GRE"},
  {2, "L2"}
};



static struct proc_dir_entry *sap_proc = NULL;

static ssize_t sap_proc_read(struct file *file,
			 char *buf,
			 size_t len,
			 loff_t *offset){

  struct sap_proc_data *pdata = (struct sap_proc_data *)file->private_data;
  int rlen;

  if (!pdata || !pdata->rbuf)
    return -EINVAL;

  rlen = (pdata->rlen > len) ? len : pdata->rlen;
  if (rlen > 0){
    if (copy_to_user(buf, pdata->rbuf, rlen))
      return -EFAULT;
  }

  pdata->rlen = 0;

  return rlen;
}

static ssize_t sap_proc_write(struct file *file,
			      const char *buf,
			      size_t len,
			      loff_t *offset){
  struct sap_proc_data *pdata = (struct sap_proc_data *)file->private_data;
  int wlen;

  if (!pdata || !pdata->wbuf)
    return -EINVAL;

  wlen = (pdata->max_wlen > len) ? len : pdata->wlen;

  if (copy_from_user(pdata->wbuf, buf, wlen))
    return -EFAULT;

  pdata->wlen = wlen;

  return wlen;

}

static int sap_proc_close(struct inode *inode, struct file *file){

  struct sap_proc_data *pdata = (struct sap_proc_data *)file->private_data;

  if (pdata->when_closed) pdata->when_closed(inode, file);
  if (pdata->rbuf) kfree(pdata->rbuf);
  if (pdata->wbuf) kfree(pdata->wbuf);
  kfree(pdata);
  return 0;
}

static void sap_mode_close(struct inode *inode, struct file *file){

  struct sap_proc_data *pdata = (struct sap_proc_data *)file->private_data;
  int new_mode;
  
  if (pdata->wlen == 0)
    return; // nothing was written

  sscanf(pdata->wbuf, "%d", &new_mode);
  if ((new_mode >= NUM_MODES) && (new_mode < 1)){
    printk("Undefined mode %d\n", new_mode);
  } else {
    printk("New mode to change to %d\n", new_mode);
    backend_mode = new_mode;
  }

  return;

}

static int sap_mode_open(struct inode *inode, struct file *file){

#define MODE_RLEN 256
#define MODE_WLEN 256

  struct sap_proc_data *pdata;
  char *p;
  int ii;

  file->private_data = kmalloc(sizeof(struct sap_proc_data), GFP_KERNEL);
  if (file->private_data == NULL)
    return -ENOMEM;

  pdata = (struct sap_proc_data *)file->private_data;
  memset(pdata, 0, sizeof(*pdata));
  
  pdata->rbuf = kmalloc(MODE_RLEN, GFP_KERNEL);
  if (!pdata->rbuf){
    kfree(pdata);
    return -ENOMEM;
  }

  pdata->wbuf = kmalloc(MODE_WLEN, GFP_KERNEL);
  if (!pdata->wbuf){
    kfree(pdata->rbuf);
    kfree(pdata);
    return -ENOMEM;
  }

  pdata->rlen = MODE_RLEN;
  memset(pdata->rbuf, 0, pdata->rlen);
  pdata->max_wlen = MODE_WLEN;
  memset(pdata->wbuf, 0, pdata->max_wlen);

  pdata->when_closed = sap_mode_close;

  p = pdata->rbuf;
  p += sprintf(p, "Current mode: %d (%s)\nPossible modes: \n", 
	  backend_mode, mode_d[backend_mode].descr);
  for(ii=0;ii<NUM_MODES;ii++){
    p += sprintf(p, "\t%d: %s\n", mode_d[ii].mode, mode_d[ii].descr);
  }
  pdata->rlen = strlen(pdata->rbuf);

  return 0;

}

static void sap_mx_mac_close(struct inode *inode, struct file *file){

  struct sap_proc_data *pdata = (struct sap_proc_data *)file->private_data;
  
  if (pdata->wlen == 0)
    return; // nothing was written
  
  sscanf(pdata->wbuf, "%02x%02x%02x%02x%02x%02x",
	 mx_mac, mx_mac+1, mx_mac+2,
	 mx_mac+3, mx_mac+4, mx_mac+5);


  printk("New Mswitch MAC address: %02x%02x%02x%02x%02x%02x\n",
	 mx_mac[0], mx_mac[1], mx_mac[2],
	 mx_mac[3], mx_mac[4], mx_mac[5]);

  return;
  

}

static int sap_mx_mac_open(struct inode *inode, struct file *file){

#define MX_MAC_RLEN 64
#define MX_MAC_WLEN 64

  struct sap_proc_data *pdata;
  char *p;

  file->private_data = kmalloc(sizeof(struct sap_proc_data), GFP_KERNEL);
  if (file->private_data == NULL)
    return -ENOMEM;

  pdata = (struct sap_proc_data *)file->private_data;
  memset(pdata, 0, sizeof(*pdata));
  
  pdata->rbuf = kmalloc(MX_MAC_RLEN, GFP_KERNEL);
  if (!pdata->rbuf){
    kfree(pdata);
    return -ENOMEM;
  }

  pdata->wbuf = kmalloc(MX_MAC_WLEN, GFP_KERNEL);
  if (!pdata->wbuf){
    kfree(pdata->rbuf);
    kfree(pdata);
    return -ENOMEM;
  }

  pdata->rlen = MX_MAC_RLEN;
  memset(pdata->rbuf, 0, pdata->rlen);
  pdata->max_wlen = MX_MAC_WLEN;
  memset(pdata->wbuf, 0, pdata->max_wlen);

  pdata->when_closed = sap_mx_mac_close;

  p = pdata->rbuf;
  p += sprintf(p, "%02x%02x%02x%02x%02x%02x\n",
	       mx_mac[0], mx_mac[1], mx_mac[2], 
	       mx_mac[3], mx_mac[4], mx_mac[5]);
  pdata->rlen = strlen(pdata->rbuf);

  return 0;
}


static int sap_channel_close(struct inode *inode, struct file *file){

  return 0;

}

static int sap_channel_open(struct inode *inode, struct file *file){

#define CH_RLEN 64
#define CH_WLEN 64

  struct sap_proc_data *pdata;
  char *p;

  file->private_data = kmalloc(sizeof(struct sap_proc_data), GFP_KERNEL);
  if (file->private_data == NULL)
    return -ENOMEM;

  pdata = (struct sap_proc_data *)file->private_data;
  memset(pdata, 0, sizeof(*pdata));
  
  pdata->rbuf = kmalloc(CH_RLEN, GFP_KERNEL);
  if (!pdata->rbuf){
    kfree(pdata);
    return -ENOMEM;
  }

  pdata->wbuf = kmalloc(CH_WLEN, GFP_KERNEL);
  if (!pdata->wbuf){
    kfree(pdata->rbuf);
    kfree(pdata);
    return -ENOMEM;
  }

  pdata->rlen = CH_RLEN;
  memset(pdata->rbuf, 0, pdata->rlen);
  pdata->max_wlen = CH_WLEN;
  memset(pdata->wbuf, 0, pdata->max_wlen);

  pdata->when_closed = sap_channel_close;

  p = pdata->rbuf;
  p += sprintf(p, "%02x%02x%02x%02x%02x%02x\n",
	       mx_mac[0], mx_mac[1], mx_mac[2], 
	       mx_mac[3], mx_mac[4], mx_mac[5]);
  pdata->rlen = strlen(pdata->rbuf);



  return 0;

}

static int sap_monitor_open(struct inode *inode, struct file *file){

  return 0;


}

static int sap_monitor_close(struct inode *inode, struct file *file){


  return 0;

}

static int sap_essid_open(struct inode *inode, struct file *file){


  return 0;

}

static int sap_essid_close(struct inode *inode, struct file *file){


  return 0;

}

static int sap_bssid_open(struct inode *inode, struct file *file){


  return 0;

}

static int sap_bssid_close(struct inode *inode, struct file *file){


  return 0;

}

static int sap_power_open(struct inode *inode, struct file *file){


  return 0;

}

static int sap_power_close(struct inode *inode, struct file *file){


  return 0;

}

static struct file_operations sap_mode_ops = {
  read: sap_proc_read,
  write: sap_proc_write,
  open: sap_mode_open,
  release: sap_proc_close
};

static struct file_operations sap_mx_mac_ops = {
  read: sap_proc_read,
  write: sap_proc_write,
  open: sap_mx_mac_open,
  release: sap_proc_close
};

static struct file_operations sap_channel_ops = {
  read: sap_proc_read,
  write: sap_proc_write,
  open: sap_channel_open,
  release: sap_proc_close
};

static struct file_operations sap_monitor_ops = {
  read: sap_proc_read,
  write: sap_proc_write,
  open: sap_monitor_open,
  release: sap_proc_close
};

static struct file_operations sap_essid_ops = {
  read: sap_proc_read,
  write: sap_proc_write,
  open: sap_essid_open,
  release: sap_proc_close
};

static struct file_operations sap_bssid_ops = {
  read: sap_proc_read,
  write: sap_proc_write,
  open: sap_bssid_open,
  release: sap_proc_close
};

static struct file_operations sap_power_ops = {
  read: sap_proc_read,
  write: sap_proc_write,
  open: sap_power_open,
  release: sap_proc_close
};



void softap_init_proc(struct net_device *dev){

  struct proc_dir_entry *entry;

  if (proc_net != NULL){
    if (sap_proc == NULL)
      sap_proc = proc_mkdir("sap", proc_net);
  }

  if (sap_proc == NULL){
    printk("sap_proc: failed to create /proc/net/sap\n");
    return;
  }

  entry = create_proc_entry("mode", S_IFREG | 0644, sap_proc);
  if (entry == NULL){
    printk("sap_proc: /proc/net/sap/mode failed\n");
    return;
  }
  entry->uid = 0;
  entry->gid = 0;
  entry->data = dev;
  entry->proc_fops = &sap_mode_ops;

  entry = create_proc_entry("mx_mac", S_IFREG | 0644, sap_proc);
  if (entry == NULL){
    printk("sap_proc: /proc/net/sap/mx_mac failed\n");
    return;
  }
  entry->uid = 0;
  entry->gid = 0;
  entry->data = dev;
  entry->proc_fops = &sap_mx_mac_ops;

  return;
}
  
void softap_takedown_proc(){

  remove_proc_entry("mx_mac", sap_proc);
  remove_proc_entry("mode", sap_proc);
  remove_proc_entry("sap", proc_net);
  return;

}
