/*
 * sibyte.c: Linux driver for the Broadcom SiByte network processor.
 * The Sibyte is a PCI device.  This driver downloads firmware to
 * the SiByte using the CFE pcihost interface which uses mailbox0 of
 * the SiByte.
 *
 * Copyright 2002 by John Taylor, <taylor@arubanetworks.com>.
 *
 * Thanks to Broadcom for providing hardware and documentation
 * enabling me to write this driver.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Additional credits:
 */

#undef DEBUG_FRAG

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/sockios.h>
#include <linux/pkt_sched.h>
#ifdef CONFIG_MSWITCH
#include <linux/mswitch.h>
#endif
#ifdef CONFIG_CABERNET
#include <linux/cabernet.h>
#endif
#ifdef CONFIG_SYRAH
#include <linux/syrah.h>
#endif
#ifdef CONFIG_MALBEC
#include <linux/malbec.h>
#endif
#include <linux/compile.h>
#ifdef DEBUG_FRAG	    
#include <linux/ip.h>
#endif

#ifdef SIOCETHTOOL
#include <linux/ethtool.h>
#endif

#include <net/sock.h>
#include <net/ip.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>


#undef INDEX_DEBUG
#define DEBUG  0
#define DEBUG1 0

#ifndef PCI_VENDOR_ID_SIBYTE
#define PCI_VENDOR_ID_SIBYTE		0x166d
#endif
#ifndef PCI_DEVICE_ID_SIBYTE
#define PCI_DEVICE_ID_SIBYTE            0x0001
#endif


#ifndef wmb
#define wmb()	mb()
#endif

#ifndef __exit
#define __exit
#endif

#ifndef __devinit
#define __devinit	__init
#endif

#ifndef min
#define min(a,b)	(((a)<(b))?(a):(b))
#endif

#ifndef SMP_CACHE_BYTES
#define SMP_CACHE_BYTES	L1_CACHE_BYTES
#endif

#ifndef SET_MODULE_OWNER
#define SET_MODULE_OWNER(dev)		{do{} while(0);}
#define SIBYTE_MOD_INC_USE_COUNT	MOD_INC_USE_COUNT
#define SIBYTE_MOD_DEC_USE_COUNT	MOD_DEC_USE_COUNT
#else
#define SIBYTE_MOD_INC_USE_COUNT	{do{} while(0);}
#define SIBYTE_MOD_DEC_USE_COUNT	{do{} while(0);}
#endif

#define SIBYTE_MBOX_SET_BIT  0x8

#define NEW_NETINIT
#define SIBYTE_PROBE_ARG			void

#define BOARD_IDX_STATIC	0
#define BOARD_IDX_OVERFLOW	-1

/* margot includes */
#include "../../../cmn/util/sibyte_types.h"
#include "../../../cmn/util/sibyte_msg.h"
#include "../../../cmn/util/sibyte_l2.h"

#include "sibyte.h"

/*
 * New interrupt handler strategy:
 *
 * An old interrupt handler worked using the traditional method of
 * replacing an skbuff with a new one when a packet arrives. However
 * the rx rings do not need to contain a static number of buffer
 * descriptors, thus it makes sense to move the memory allocation out
 * of the main interrupt handler and do it in a bottom half handler
 * and only allocate new buffers when the number of buffers in the
 * ring is below a certain threshold. In order to avoid starving the
 * NIC under heavy load it is however necessary to force allocation
 * when hitting a minimum threshold. The strategy for alloction is as
 * follows:
 *
 *     RX_LOW_BUF_THRES    - allocate buffers in the bottom half
 *     RX_PANIC_LOW_THRES  - we are very low on buffers, allocate
 *                           the buffers in the interrupt handler
 *     RX_RING_THRES       - maximum number of buffers in the rx ring
 *
 * One advantagous side effect of this allocation approach is that the
 * entire rx processing can be done without holding any spin lock
 * since the rx rings and registers are totally independant of the tx
 * ring and its registers.  This of course includes the kmalloc's of
 * new skb's. Thus start_xmit can run in parallel with rx processing
 * and the memory allocation on SMP systems.
 *
 * Note that running the skb reallocation in a bottom half opens up
 * another can of races which needs to be handled properly. In
 * particular it can happen that the interrupt handler tries to run
 * the reallocation while the bottom half is either running on another
 * CPU or was interrupted on the same CPU. To get around this the
 * driver uses bitops to prevent the reallocation routines from being
 * reentered.
 *
 * TX handling can also be done without holding any spin lock, wheee
 * this is fun! since tx_csm is only written to by the interrupt
 * handler.
 */

/*
 * Threshold values for RX buffer allocation - the low water marks for
 * when to start refilling the rings are set to 75% of the ring
 * sizes. It seems to make sense to refill the rings entirely from the
 * intrrupt handler once it gets below the panic threshold, that way
 * we don't risk that the refilling is moved to another CPU when the
 * one running the interrupt handler just got the slab code hot in its
 * cache.
 */
#define RX_RING_SIZE		SIBYTE_DESCR_ENTRIES
#define RX_PANIC_THRES	        (RX_RING_SIZE/4)
#define RX_LOW_THRES	        ((3*RX_RING_SIZE)/4)
#define RX_OFFSET               2

#define SIBYTE_BUFSIZE		(SIBYTE_MTU + ETH_HLEN + \
				 RX_OFFSET+sizeof(sibyte_msg_t))

#ifdef MODULE
MODULE_AUTHOR("John Taylor <taylor@arubanetworks.com>");
MODULE_DESCRIPTION("Broadcom SiByte Ethernet driver");
#endif

#if DEBUG
static char version[] __initdata = 
  "sibyte.c: v0.01 2002/08/12  taylor@arubanetworks.com\n";
#endif /* DEBUG */

static struct net_device *root_dev;
static int probed __initdata = 0;

static void sibyte_pci_dump(struct net_device *dev);

static inline u32 swap32(u32 x)
{
  return ((x >> 24) |
	  (x << 24) |
	  ((x >> 8) & 0x00ff00) |
	  ((x << 8) & 0xff0000));
}

static inline u16 swap16(u16 x)
{
  return (((x >> 8) & 0x00ff) |
	  ((x << 8) & 0xff00));
}

int __devinit sibyte_probe (SIBYTE_PROBE_ARG)
{
#ifdef NEW_NETINIT
    struct net_device *dev;
#endif

    struct sibyte_private *sp;
    struct pci_dev *pdev = NULL;
    int boards_found = 0;
    int version_disp;
    u16 tmp16;
    u32 tmp32;

    if (probed)
        return -ENODEV;
    probed++;

    if (!pci_present())		/* is PCI support present? */
        return -ENODEV;

    version_disp = 0;
	
    while ((pdev = pci_find_device(PCI_VENDOR_ID_SIBYTE,
				   PCI_DEVICE_ID_SIBYTE, pdev)) != NULL) {
      
	dev = init_etherdev(NULL, sizeof(struct sibyte_private));

	if (dev == NULL) {
	    printk(KERN_ERR "sibyte: Unable to allocate "
		   "net_device structure!\n");
	    break;
	}

	SET_MODULE_OWNER(dev);

	if (!dev->priv)
	    dev->priv = kmalloc(sizeof(*sp), GFP_KERNEL);
	if (!dev->priv) {
	    printk(KERN_ERR "sibyte: Unable to allocate memory\n");
	    return -ENOMEM;
	}

	sp = dev->priv;
	sp->pdev = pdev;

	dev->irq = pdev->irq;
	dev->open = &sibyte_open;
	dev->hard_start_xmit = &sibyte_start_xmit;
	dev->features |= NETIF_F_SG | NETIF_F_HIGHDMA;
	dev->stop = &sibyte_close;
	dev->get_stats = &sibyte_get_stats;
	dev->do_ioctl = &sibyte_ioctl;
	dev->vlan = 1;

	/* display version info if adapter is found */
	if (!version_disp) {
	    /* set display flag to TRUE so that */
	    /* we only display this string ONCE */
	    version_disp = 1;
#if DEBUG
	    printk(version);
#endif /* DEBUG */
	}

	/*
	 * Enable master mode before we start playing with the
	 * pci_command word since pci_set_master() will modify
	 * it.
	 */
	pci_set_master(pdev);

	pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &sp->pci_latency);
	if (sp->pci_latency <= 0x40) {
	    sp->pci_latency = 0x40;
	    pci_write_config_byte(pdev, PCI_LATENCY_TIMER, sp->pci_latency);
	}

	/* enable error reporting */
	pci_read_config_word(pdev, PCI_COMMAND, (u16 *)&tmp16);
	tmp16 |= PCI_COMMAND_PARITY|PCI_COMMAND_SERR;
	pci_write_config_word(pdev, PCI_COMMAND, tmp16);
	sp->pci_cmdstatus = tmp16; /* save of later */

	pci_read_config_dword(pdev, PCI_ADD_STATUS, (u32 *)&tmp32);
	tmp32 |= PCI_ADD_STATUS_TRDY|PCI_ADD_STATUS_RETRY;
	pci_write_config_dword(pdev, PCI_ADD_STATUS, tmp32);
	
	/* save base addresses for later */
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0, (u32 *)&dev->base_addr);
	dev->base_addr &= PCI_BASE_ADDRESS_MEM_MASK;
	sp->base_addr0 = dev->base_addr; /* save BAR0 */
	dev->base_addr = (u32) ioremap(dev->base_addr, 0x1000000);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_2, &sp->mb[0]);
	sp->mb[0] &= PCI_BASE_ADDRESS_MEM_MASK;
	sp->base_addr2 = sp->mb[0]; /* save BAR2 (mailbox 0) */
	sp->mb[0] = (u32) ioremap(sp->mb[0], 0x10);
#ifdef SIBYTE_MB1
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_3, &sp->mb[1]);
	sp->mb[1] &= PCI_BASE_ADDRESS_MEM_MASK;
	sp->base_addr3 = sp->mb[1]; /* save BAR3 (mailbox 1) */
	sp->mb[1] = (u32) ioremap(sp->mb[1], 0x10);
#endif /* SIBYTE_MB1 */

	/* announce our arrival */
	switch(pdev->vendor) {
	case PCI_VENDOR_ID_SIBYTE:
	    strncpy(sp->name, "Broadcom SiByte Ethernet",
		    sizeof (sp->name));
#if DEBUG
	    printk(KERN_INFO "%s: Broadcom ", dev->name);
#endif /* DEBUG */
	    break;
	default:
	    strncpy(sp->name, "Unknown SiByte Ethernet",
		    sizeof (sp->name));
	    printk(KERN_INFO "%s: Unknown Broadcom ", dev->name);
	    break;
	}
	sp->name [sizeof (sp->name) - 1] = '\0';
#if DEBUG
	printk("SiByte Ethernet, IRQ %i\n", dev->irq);
#endif /* DEBUG */

	dev->dev_addr[0] = 0x00; /* configured by fpapps */
	dev->dev_addr[1] = 0x0B;
	dev->dev_addr[2] = 0x86;
	dev->dev_addr[3] = 0x00;
	dev->dev_addr[4] = 0x00;
	dev->dev_addr[5] = 0x00;

#ifdef MODULE
	if (boards_found >= 2)
	    sp->board_idx = BOARD_IDX_OVERFLOW;
	else
	    sp->board_idx = boards_found;
#else
	sp->board_idx = BOARD_IDX_STATIC;
#endif

	if (sibyte_init(dev)) {
	    /*
	     * sibyte_init() calls sibyte_init_cleanup() on error.
	     */
	    kfree(dev);
	    continue;
	}

	/*
	 * Register the device
	 */
	sp->next = root_dev;
	root_dev = dev;

	boards_found++;
    }

    /*
     * If we're at this point we're going through sibyte_probe() for
     * the first time.  Return success (0) if we've initialized 1
     * or more boards. Otherwise, return failure (-ENODEV).
     */

    if (boards_found > 0) {
#if defined(CONFIG_MSWITCH)
        /* initialize mswitch control bus */
        mx_init();
#endif
#if defined(CONFIG_CABERNET)
        /* initialize mswitch control bus */
        cab_init();
#endif
#if defined(CONFIG_MALBEC)
        /* initialize mswitch control bus */
        mal_init();
#endif
#if defined(CONFIG_SYRAH)
        /* initialize mswitch control bus */
        syr_init();
#endif
        return 0;
    } else {
        return -ENODEV;
    }
}


static void sibyte_pci_dump(struct net_device *dev)
{
    struct sibyte_private *sp = dev->priv;
    u32 tmp;
    int i;

    printk("Sibyte PCI dump\n");
    for (i = 0; i < 0x9c; i += 4) {
	pci_read_config_dword(sp->pdev, i, (u32 *)&tmp);
	printk("%04x: %08x\n", i, tmp);
    }
}


static void __exit sibyte_module_cleanup(void)
{
    struct sibyte_private *sp;
    struct net_device *next;

    while (root_dev) {
        sp = root_dev->priv;
	next = sp->next;

	sibyte_init_cleanup(root_dev);
	unregister_netdev(root_dev);
	kfree(root_dev);
	root_dev = next;
    }
}


int __init sibyte_module_init(void)
{
    int status;

    root_dev = NULL;

#ifdef NEW_NETINIT
    status = sibyte_probe();
#else
    status = sibyte_probe(NULL);
#endif
    return status;
}


module_init(sibyte_module_init);
module_exit(sibyte_module_cleanup);


static void sibyte_free_descriptors(struct net_device *dev)
{
    struct sibyte_private *sp = dev->priv;
    int size;

    if (sp->rx_ring[0] != NULL) {
        size = (sizeof(sibyte_descr_t) * SIBYTE_DESCR_ENTRIES * SIBYTE_QUEUES);
	pci_free_consistent(sp->pdev, size,
			    (void *)sp->rx_ring[0],
			    sp->ring_base_dma);
	sp->rx_ring[0] = NULL;
	sp->rx_ring[1] = NULL;
	sp->tx_ring[0] = NULL;
	sp->tx_ring[1] = NULL;
    }
}


static int sibyte_allocate_descriptors(struct net_device *dev)
{
    struct sibyte_private *sp = dev->priv;
    int size;

    size = (sizeof(sibyte_descr_t) * (SIBYTE_DESCR_ENTRIES * SIBYTE_QUEUES));
    sp->rx_ring[0] = pci_alloc_consistent(sp->pdev, size, &sp->ring_base_dma);
    if (sp->rx_ring[0] == NULL)
        return 1;

    memset((void *)sp->rx_ring[0], 0, size);

    sp->rx_ring[1] = sp->rx_ring[0] + SIBYTE_DESCR_ENTRIES;
    sp->tx_ring[0] = sp->rx_ring[1] + SIBYTE_DESCR_ENTRIES;
    sp->tx_ring[1] = sp->tx_ring[0] + SIBYTE_DESCR_ENTRIES;

    return 0;
}


/*
 * Generic cleanup handling data allocated during init. Used when the
 * module is unloaded or if an error occurs during initialization
 */
static void sibyte_init_cleanup(struct net_device *dev)
{
    struct sibyte_private *sp;
    struct sk_buff *skb;
    int i, j;

    sp = dev->priv;

    sibyte_free_descriptors(dev);

    for (i = 0; i < SIBYTE_PRIOS; i++) {
        if (sp->rx_skb[i]) {
	    for (j = 0; j < SIBYTE_DESCR_ENTRIES; j++) {
	        skb = sp->rx_skb[i][j];
		if (skb) {
		    sp->rx_skb[i][j] = NULL;
		    dev_kfree_skb(skb);
		}
	    }
	    kfree(sp->rx_skb[i]);
	    sp->rx_skb[i] = NULL;
	}

        if (sp->tx_skb[i]) {
	    for (j = 0; j < SIBYTE_DESCR_ENTRIES; j++) {
		skb = sp->tx_skb[i][j];
		if (skb) {
		    sp->tx_skb[i][j] = NULL;
		    dev_kfree_skb(skb);
		}
	    }
	    kfree(sp->tx_skb[i]);
	    sp->tx_skb[i] = NULL;
	}
    }

    if (dev->irq) {
        free_irq(dev->irq, dev);
    }
}


static int __init sibyte_init(struct net_device *dev)
{
    struct sibyte_private *sp;
    short i;
    int ecode;

    sp = dev->priv;

    /*
     * Allocate descriptors
     */
    if (sibyte_allocate_descriptors(dev))
        goto init_error;

    /*
     * write base address of descriptors and num descrs to mb 1
     */
#ifdef SIBYTE_MB1
    writel(swap32(SIBYTE_DESCR_ENTRIES), sp->mb[1]|SIBYTE_MBOX_SET_BIT);
    wmb();
    writel(swap32(sp->ring_base_dma), (sp->mb[1]+4)|SIBYTE_MBOX_SET_BIT);
    wmb();
#if DEBUG
    printk("%s: mailbox %x %x %x\n", dev->name,
	   swap32(readl(sp->mb[1])),
	   swap32(readl(sp->mb[1]+4)),
	   sp->ring_base_dma);
#endif /* DEBUG */
#endif /* SIBYTE_MB1 */

    /*
     * Get the memory for the skb rings.
     */
    for (i = 0; i < SIBYTE_PRIOS; i++) {
        sp->rx_skb[i] = kmalloc(sizeof(struct sk_buff *) * SIBYTE_DESCR_ENTRIES,
				GFP_KERNEL);
	if (!(sp->rx_skb[i])) {
	    ecode = -EAGAIN;
	    goto init_error;
	}
	memset(sp->rx_skb[i], 0, sizeof(struct sk_buff *) * SIBYTE_DESCR_ENTRIES);

        sp->tx_skb[i] = kmalloc(sizeof(struct sk_buff *) * SIBYTE_DESCR_ENTRIES,
				GFP_KERNEL);
	if (!(sp->tx_skb[i])) {
	    ecode = -EAGAIN;
	    goto init_error;
	}
	memset(sp->tx_skb[i], 0, sizeof(struct sk_buff *) * SIBYTE_DESCR_ENTRIES);
    }

    /*
     * Get the IRQ
     */
    ecode = request_irq(dev->irq, sibyte_interrupt, SA_SHIRQ, dev->name, dev);
    if (ecode) {
        printk(KERN_WARNING "%s: Requested IRQ %d is busy\n",
	       dev->name, dev->irq);
	goto init_error;
    }

    /* 
     * Initialize firmware
     */
    sp->fw_running = 0;

    /*
     * Set tx_csm before we start receiving interrupts, otherwise
     * the interrupt handler might think it is supposed to process
     * tx ints before we are up and running, which may cause a null
     * pointer access in the int handler.
     */
    for (i = 0; i < SIBYTE_PRIOS; i++) {
        sp->rx_skbprd[i] = 0;
        sp->cur_rx[i] = 0;
	sp->tx_prd[i] = 0;
	sp->tx_csm[i] = 0;
    }

    wmb();

    /*
     * Zero the stats before starting the interface
     */
    memset(&sp->stats, 0, sizeof(sp->stats));

    /*
     * We load the ring here as there seem to be no way to tell the
     * firmware to wipe the ring without re-initializing it.
     */
    for (i = 0; i < SIBYTE_PRIOS; i++) {
        sibyte_load_rx_ring(sp, i, RX_RING_SIZE);
    }
    return 0;

 init_error:
    sibyte_init_cleanup(dev);
    return ecode;
}

/*
 * Load the rx ring.
 *
 * Loading rings is safe without holding the spin lock since this is
 * done only before the device is enabled, thus no interrupts are
 * generated and by the interrupt handler/tasklet handler.
 */
static void sibyte_load_rx_ring(struct sibyte_private *sp,
				u32 ring, int nr_bufs)
{
    short i, idx;

    idx = sp->rx_skbprd[ring];

    for (i = 0; i < nr_bufs; i++) {
        dma_addr_t mapping;
        struct sk_buff *skb;
	sibyte_descr_t *rd;

	if (sp->rx_skb[ring][idx]) {
#if DEBUG1
	    printk(KERN_INFO "sibyte rx refill full\n");
#endif /* DEBUG */
	    break;
	}

	skb = alloc_skb(SIBYTE_BUFSIZE, GFP_ATOMIC);
	if (!skb)
	    break;

	/*
	 * Make sure IP header starts on a fresh cache line.
	 */
	skb_reserve(skb, RX_OFFSET);
	sp->rx_skb[ring][idx] = skb;

	mapping = pci_map_page(sp->pdev, virt_to_page(skb->data),
			       ((unsigned long)skb->data & ~PAGE_MASK),
			       SIBYTE_BUFSIZE - RX_OFFSET,
			       PCI_DMA_FROMDEVICE);

	rd = (sibyte_descr_t *) &sp->rx_ring[ring][idx];
	rd->addr = mapping;
	rd->descr = SIBYTE_DESCR_OWNER | (SIBYTE_BUFSIZE - RX_OFFSET);
	idx = DSC_NEXT(idx);
    }

    if (!i) {
#if DEBUG1
        printk(KERN_INFO "Out of memory when allocating "
	       "standard receive buffers\n");
#endif /* DEBUG */
    } else {
	sp->rx_skbprd[ring] = idx;
    }

    return;
}

static void sibyte_rx_int(struct net_device *dev, u32 ring)
{
    struct sibyte_private *sp = dev->priv;
    u32 idx, start;

    idx = sp->cur_rx[ring];
    start = idx;
    do {
	struct sk_buff *skb;
	struct sk_buff *new_skb;
	sibyte_descr_t *rxdesc;
        dma_addr_t mapping;
	sibyte_msg_t *m;
	short length;
	u32 descr;
#ifdef DEBUG_FRAG	    
	struct iphdr *iph;
	int offset;
	int *iptr;
	int i;
#endif

	rxdesc = &sp->rx_ring[ring][idx];
	descr = rxdesc->descr;
	if (descr & SIBYTE_DESCR_OWNER) {
	    /* SiByte owns descriptor */
#if DEBUG
	  printk("siByte owns descriptor\n");
#endif /* DEBUG */
	    break;
	}

	skb = sp->rx_skb[ring][idx];
	if (!skb) {
#if DEBUG1
	    printk("%s: no skb %d %d\n", dev->name, ring, idx);
#endif /* DEBUG */
	    break;
	}

	length = descr & SIBYTE_DESCR_LENGTH;
	if ((length < sizeof(sibyte_msg_t)) ||
	    (length > (SIBYTE_BUFSIZE - RX_OFFSET))) {
	    printk("%s: rx len error %d\n", dev->name, length);
	    sp->stats.rx_dropped++;
            rxdesc->descr = SIBYTE_DESCR_OWNER | (SIBYTE_BUFSIZE - RX_OFFSET);
	    goto next;
	}

	new_skb = alloc_skb(SIBYTE_BUFSIZE, GFP_ATOMIC);
        if (!new_skb) {
	    sp->stats.rx_dropped++;
            rxdesc->descr = SIBYTE_DESCR_OWNER | (SIBYTE_BUFSIZE - RX_OFFSET);
            goto next;
        }
	sp->rx_skb[ring][idx] = new_skb;
	mapping = pci_map_page(sp->pdev, virt_to_page(new_skb->data),
			       ((unsigned long)new_skb->data & ~PAGE_MASK),
			       SIBYTE_BUFSIZE - RX_OFFSET,
			       PCI_DMA_FROMDEVICE);

	rxdesc->addr = mapping;
        rxdesc->descr = SIBYTE_DESCR_OWNER | (SIBYTE_BUFSIZE - RX_OFFSET);

	m = (sibyte_msg_t *)skb->data;
#if DEBUG
	printk("%s: rx int %d %d %d %x %x\n", dev->name, ring, start, idx,
	       rxdesc->descr, rxdesc->addr);
#endif /* DEBUG */
	skb->dev = dev;
#if defined(CONFIG_MSWITCH) || defined(CONFIG_CABERNET) || defined(CONFIG_SYRAH) || defined(CONFIG_MALBEC)
	skb->flags = m->flags;
	skb->vlan = m->vlan;
#endif
	if (m->opcode == SIBYTE_MSG_OPCODE_FORWARD) {
	    /* send up linux stack */
	    /* skip message header */
	    skb_reserve(skb, sizeof(sibyte_msg_t));
	    skb_put(skb, (length - sizeof(sibyte_msg_t)));
	    /* set protocol (802.1Q) */
	    skb->protocol = eth_type_trans(skb, dev);
#ifdef DEBUG_FRAG	    
	    if (skb->len >= (4 + sizeof(struct iphdr))) {
	      iph= (struct iphdr *) (skb->data+4);
	      offset = ntohs(iph->frag_off);
	      if ( (iph->version==4) && 
		   ((offset & IP_MF) || (offset & IP_OFFSET)) ) {
		printk("rx IP frag from: %x, id: %d offset: %d \n", 
		       iph->saddr, iph->id, (offset & IP_OFFSET) << 3);
	      }
	      else if ( iph->version==4 ) {
		printk("rx IP from: %x, id: %d offset: %d \n",
		       iph->saddr, iph->id, (offset & IP_OFFSET) << 3);
	      }
	      else if (skb->len >= ((4*4)+sizeof(struct iphdr))) {
		iptr= (int*)(skb->data);
		for (i =0; i < 4*4 ; i+=4) {
		  printk("%4i: %8x %8x %8x %8x\n", i, 
			 *(iptr+i), *(iptr+i+1), *(iptr+i+3),*(iptr+i+3));
		  ;
		}
	      }
	    }
#endif	    
	    netif_rx(skb);    /* send it up */
	} else {
	    /* include message header */
	    skb_put(skb, length);
	    /* set frame data */
	    skb->mac.raw = skb->data;
	    /* set protocol */
	    skb->protocol = SIBYTE_MSG_PROTO(m->opcode);
	    netif_rx(skb);          /* send it up */
	}

	dev->last_rx = jiffies;
	sp->stats.rx_packets++;
	sp->stats.rx_bytes += length;
    next:
	idx = DSC_NEXT(idx);
    } while (start != idx);

    sp->cur_rx[ring] = idx;
#if DEBUG
    printk("%s: rx done\n", dev->name);
#endif /* DEBUG */
}


static void sibyte_tx_int(struct net_device *dev, u32 ring)
{
    struct sibyte_private *sp = dev->priv;
    u32 count = 0;
    u32 idx;

    idx = sp->tx_csm[ring];
    while (idx != sp->tx_prd[ring]) {
        struct sk_buff *skb;
	sibyte_descr_t *txdesc;
	u32 descr;

	txdesc = &sp->tx_ring[ring][idx];
#if DEBUG
	printk("%s: tx int %d %d %d %d %x %x\n", dev->name, ring,
	       sp->tx_csm[ring], idx, sp->tx_prd[ring],
	       txdesc->descr, txdesc->addr);
#endif /* DEBUG */
	descr = txdesc->descr;
	if (descr & SIBYTE_DESCR_OWNER) {
	    /* sibyte still owns descr */
	    break;
	}

	count++;
	skb = sp->tx_skb[ring][idx];
	if (skb) {
	    sp->stats.tx_packets++;
	    sp->stats.tx_bytes += skb->len;
	    dev_kfree_skb_irq(skb);
	    sp->tx_skb[ring][idx] = NULL;
	} else {
	    sp->stats.tx_bytes += (descr & SIBYTE_DESCR_LENGTH);
	}
	idx = DSC_NEXT(idx);
    }

    if (count) {
	if (netif_queue_stopped(dev))
	    netif_wake_queue(dev);
	wmb();
	sp->tx_csm[ring] = idx;
    }
}


static void sibyte_interrupt(int irq, void *dev_id, struct pt_regs *ptregs)
{
    struct sibyte_private *sp;
    struct net_device *dev = (struct net_device *)dev_id;
    u32 ring;

    sp = dev->priv;

#if DEBUG
    printk("%s: interrupt\n", dev->name);
#endif /* DEBUG */

    /*
     * There is no conflict between transmit handling in
     * start_xmit and receive processing, thus there is no reason
     * to take a spin lock for RX handling. Wait until we start
     * working on the other stuff - hey we don't need a spin lock
     * anymore.
     */
    for (ring = 0; ring < SIBYTE_PRIOS; ring++) {
#if DEBUG
        printk("%s: %d %d %08x\n", dev->name, ring,
	       sp->cur_rx[ring],
	       sp->rx_ring[ring][sp->cur_rx[ring]].descr);
#endif /* DEBUG */
        if (!(sp->rx_ring[ring][sp->cur_rx[ring]].descr & SIBYTE_DESCR_OWNER)) {
	    if (!sp->fw_running) {
	        sp->fw_running = 1;
	    }
	    sibyte_rx_int(dev, ring); 
	}
    }

    /*
     * ACK intr after rx processing
     *
     * Tell SiByte to turn off interrupt by writing to mailbox 
     */
    if (sp->fw_running) {
        writel(swap32(0x1 << SIBYTE_PRIOS), (sp->mb[0]+4)|SIBYTE_MBOX_SET_BIT);
    	wmb();
    }


    for (ring = 0; ring < SIBYTE_PRIOS; ring++) {
        if (sp->tx_csm[ring] != sp->tx_prd[ring]) {
	    sibyte_tx_int(dev, ring);
	}
    }
}


static int sibyte_open(struct net_device *dev)
{
    struct sibyte_private *sp;

    sp = dev->priv;
    sp->promisc = 1;

    dev->mtu = SIBYTE_MTU;
    netif_start_queue(dev);

    SIBYTE_MOD_INC_USE_COUNT;

    return 0;
}


static int sibyte_close(struct net_device *dev)
{
    struct sibyte_private *sp;
    int i, j;

    /*
     * Without (or before) releasing irq and stopping hardware, this
     * is an absolute non-sense, by the way. It will be reset instantly
     * by the first irq.
     */
    netif_stop_queue(dev);

    sp = dev->priv;
    sp->promisc = 0;

    for (i = 0; i < SIBYTE_PRIOS; i++) {
        for (j = 0; j < SIBYTE_DESCR_ENTRIES; j++) {
	    struct sk_buff *skb;
	    sibyte_descr_t *txdesc;

	    txdesc = &sp->tx_ring[i][j];
	    txdesc->descr = 0;

	    skb = sp->tx_skb[i][j];
	    if (skb) {
	        dev_kfree_skb(skb);
		sp->tx_skb[i][j] = NULL;
	    }
	}
    }

    SIBYTE_MOD_DEC_USE_COUNT;
    return 0;
}

static int sibyte_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
    struct sibyte_private *sp = dev->priv;
    dma_addr_t mapping;
    sibyte_descr_t *desc;
    sibyte_msg_t *m;
    u32 ring, idx, flagsize;

    if (!sp->fw_running) {
	sp->stats.tx_dropped++;
	return 1;
    }

#if DEBUG
    printk(KERN_DEBUG "%s: tx pkt\n", dev->name);
#endif /* DEBUG */

    if (skb->protocol < SIBYTE_MSG_PROTO_BASE) {
        ring = (skb->priority >= TC_PRIO_CONTROL)? 1: 0;
    } else {
	m = (sibyte_msg_t *)skb->data;
	ring = (m->flags & SIBYTE_MSG_FLAG_HIGH_PRIO);
    }
    idx = sp->tx_prd[ring];

    if (tx_ring_full(sp->tx_csm[ring], idx)) {
        sp->stats.tx_dropped++;
	return 1;
    }

    if (skb->protocol < SIBYTE_MSG_PROTO_BASE) {
       /*
	* Prepend message header
	*/
        struct sk_buff *skm;

	skm = alloc_skb(sizeof(sibyte_msg_t), GFP_ATOMIC);
	if (!skm) {
#if DEBUG1
            printk(KERN_DEBUG "%s: tx alloc_skb failed\n", dev->name);
#endif /* DEBUG */
	    sp->stats.tx_dropped++;
	    return 1;
	}
	skb_put(skm, sizeof(sibyte_msg_t));
	m = (sibyte_msg_t *)skm->data;
	memset(m, 0, sizeof(*m));
	if (ring) {
	    m->flags |= SIBYTE_MSG_FLAG_HIGH_PRIO;
	}
#if defined(CONFIG_MSWITCH) || defined(CONFIG_CABERNET) || defined(CONFIG_SYRAH) || defined(CONFIG_MALBEC)
	m->flags |= skb->flags;
#endif
	m->opcode = SIBYTE_MSG_OPCODE_BRIDGE;
	if (skb->data[12] == 0x81 && skb->data[13] == 0x00) {
	    m->vlan =  ((skb->data[14] << 8 | skb->data[15]) & 
			0xfff);
#if defined(CONFIG_MSWITCH) || defined(CONFIG_CABERNET) || defined(CONFIG_SYRAH) || defined(CONFIG_MALBEC)
	} else {
	    m->vlan = skb->vlan & 0xfff;
#endif
	}
	sp->tx_skb[ring][idx] = skm;
	flagsize = skm->len | (SIBYTE_DESCR_START|SIBYTE_DESCR_OWNER);
	desc = &sp->tx_ring[ring][idx];
	mapping = pci_map_page(sp->pdev, virt_to_page(skm->data),
			       ((unsigned long)skm->data & ~PAGE_MASK),
			       skm->len,
			       PCI_DMA_TODEVICE);
	desc->addr = mapping;
	desc->descr = flagsize;
	idx = DSC_NEXT(idx);
	flagsize = 0;
    } else {
	flagsize = SIBYTE_DESCR_START;
    }

#if 0
    printk("TX vlan %d, opcode %x\n", m->vlan, m->opcode);
#endif

    /*
     *  Data
     */
#if MAX_SKB_FRAGS
    if (!skb_shinfo(skb)->nr_frags)
#endif
    {
	sp->tx_skb[ring][idx] = skb;
	flagsize |= skb->len | (SIBYTE_DESCR_STOP|SIBYTE_DESCR_OWNER);
	desc = &sp->tx_ring[ring][idx];
	mapping = pci_map_page(sp->pdev, virt_to_page(skb->data),
			       ((unsigned long)skb->data & ~PAGE_MASK),
			       skb->len,
			       PCI_DMA_TODEVICE);
	desc->addr = mapping;
	desc->descr = flagsize;
	idx = DSC_NEXT(idx);
    }
#if MAX_SKB_FRAGS
    else {
        int i;

#if DEBUG
	printk(KERN_DEBUG "%s: xmit frags\n", dev->name);
#endif /* DEBUG */
	flagsize |= (skb->len - skb->data_len) | SIBYTE_DESCR_OWNER;
	desc = &sp->tx_ring[ring][idx];
	desc->addr = (u32) skb->data;
	desc->descr = flagsize;
	idx = DSC_NEXT(idx);

	for (i = 0; i < skb_shinfo(skb)->nr_frags; i++) {
	    skb_frag_t *frag = &skb_shinfo(skb)->frags[i];

	    desc = &sp->tx_ring[ring][idx];
	    flagsize = frag->size | SIBYTE_DESCR_OWNER;
	    mapping = pci_map_page(sp->pdev, frag->page,
				   frag->page_offset, frag->size,
				   PCI_DMA_TODEVICE);
	    desc->addr = mapping;
	    if (i == skb_shinfo(skb)->nr_frags-1) {
		flagsize |= SIBYTE_DESCR_STOP;
		/*
		 * Only the last fragment frees
		 * the skb!
		 */
		sp->tx_skb[ring][idx] = skb;
	    }
	    desc->descr = flagsize;
	    idx = DSC_NEXT(idx);
	}
    }
#endif

    /* generate interrupt on SiByte by writing mailbox reg */
    writel(swap32(0x1 << ring), (sp->mb[0]+4)|SIBYTE_MBOX_SET_BIT);
    wmb();

    sp->tx_prd[ring] = idx;

    dev->trans_start = jiffies;
    return 0;
}

static int sibyte_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    struct sibyte_private *sp = dev->priv;
    u32 remap;
    int i;

    switch (cmd) {
    case SIOCDEVPRIVATE: {
        struct sibyte_cmd scmd;

        if (copy_from_user(&scmd, ifr->ifr_data, sizeof(scmd)))
	    return -EFAULT;

#if DEBUG
	printk("%s: ioctl devprivate c=%d a=%x l=%d m=%d d=%x,%x\n",
	       dev->name, scmd.cmd,
	       scmd.address, scmd.length,
	       scmd.mailbox, scmd.data[0], scmd.data[1]);
#endif /* DEBUG */

	switch (scmd.cmd) {
	case SIBYTE_DOWNLOAD:
	    sp->desc.length = scmd.length;
	    sp->desc.mapping = pci_alloc_consistent(sp->pdev,
						    scmd.length,
						    &sp->desc.address);
	    if (!sp->desc.mapping)
	        return -EFAULT;
	    if (copy_from_user(sp->desc.mapping, (char *)scmd.address,
			       scmd.length))
	        return -EFAULT;

	    remap = pci_map_page(sp->pdev,
				 virt_to_page(&sp->desc),
				 ((unsigned long)&sp->desc & ~PAGE_MASK),
				 sizeof(sp->desc),
				 PCI_DMA_TODEVICE);

	    writel(swap32(remap),
		   sp->mb[0]|SIBYTE_MBOX_SET_BIT);
            wmb();
	    writel(swap32(scmd.data[1]),
		   (sp->mb[0]+4)|SIBYTE_MBOX_SET_BIT);
            wmb();
	    break;

	case SIBYTE_FREE:
	    if (sp->desc.mapping) {
	        pci_free_consistent(sp->pdev,
				    sp->desc.length,
				    sp->desc.mapping,
				    sp->desc.address);
		sp->desc.mapping = NULL;
	    }
	    break;

	case SIBYTE_SYNC:
#ifndef SIBYTE_MB1
	    /* now tell sibyte our base address and number of descriptors */
	    writel(swap32(SIBYTE_DESCR_ENTRIES), sp->mb[0]|SIBYTE_MBOX_SET_BIT);
            wmb();
	    writel(swap32(sp->ring_base_dma), (sp->mb[0]+4)|SIBYTE_MBOX_SET_BIT);
            wmb();
#if DEBUG
	    printk("%s: mailbox %x %x %x\n", dev->name,
		   swap32(readl(sp->mb[0])),
		   swap32(readl(sp->mb[0]+4)),
		   sp->ring_base_dma);
#endif /* DEBUG */
#endif /* SIBYTE_MB1 */
	    break;

	case SIBYTE_MBGET:
	    if (scmd.mailbox > 1)
	        return -EPERM;
	    scmd.data[0] = swap32(readl(sp->mb[scmd.mailbox]));
	    scmd.data[1] = swap32(readl(sp->mb[scmd.mailbox]+4));
	    if(copy_to_user(ifr->ifr_data, &scmd, sizeof(scmd)))
	        return -EFAULT;
	    break;

	case SIBYTE_MBSET:
	    if (scmd.mailbox > 1)
	        return -EPERM;
	    writel(swap32(scmd.data[0]),
		   sp->mb[scmd.mailbox]|SIBYTE_MBOX_SET_BIT);
            wmb();
	    writel(swap32(scmd.data[1]),
		   (sp->mb[scmd.mailbox]+4)|SIBYTE_MBOX_SET_BIT);
            wmb();
	    break;

	case SIBYTE_PCI_RECONFIG:

	    pci_write_config_byte(sp->pdev, PCI_CACHE_LINE_SIZE,
				  SMP_CACHE_BYTES >> 2);

	    pci_write_config_byte(sp->pdev, PCI_LATENCY_TIMER, sp->pci_latency);
	    pci_write_config_dword(sp->pdev, PCI_BASE_ADDRESS_0, sp->base_addr0);
	    pci_write_config_dword(sp->pdev, PCI_BASE_ADDRESS_2, sp->base_addr2);
	    pci_write_config_dword(sp->pdev, PCI_BASE_ADDRESS_3, sp->base_addr3);
	    pci_write_config_word(sp->pdev, PCI_COMMAND, sp->pci_cmdstatus);

#if 0
	    PANICS
	    sibyte_init_cleanup(dev);
	    sibyte_init(dev);
#endif
	    break;

	case SIBYTE_PCI_DUMP:
	    sibyte_pci_dump(dev);
	    break;

	case SIBYTE_TEST_HANG:
	    while(1);
	    break;

	case SIBYTE_TEST_PANIC:
	    panic("sibyte test panic");
	    break;

	case SIBYTE_TEST_NULLPTR:
	    {
		char *nullPtr, c;
		nullPtr = NULL;
		c = *nullPtr;
		printk("NULLPTR ptr 0x%x val = 0x%x\n", (int) nullPtr, *nullPtr);
	    }
	    break;

	case SIBYTE_READ_DATA:
	    if(scmd.length==4){
		scmd.data[0] = *((u32*)scmd.address);
	    } else if(scmd.length==2) {
		scmd.data[0] = *((u16*)scmd.address);
	    } else if (scmd.length==1) {
		scmd.data[0] = *((u8*)scmd.address);
	    } else {
		return -EOPNOTSUPP;
	    }
	    if(copy_to_user(ifr->ifr_data, &scmd, sizeof(scmd)))
	        return -EFAULT;
	    break;

	case SIBYTE_WRITE_DATA:
	    if(scmd.length==4){
		*((u32*)scmd.address) = scmd.data[0];
	    } else if(scmd.length==2) {
		*((u16*)scmd.address) = scmd.data[0];
	    } else if (scmd.length==1) {
		*((u8*)scmd.address) = scmd.data[0];
	    } else {
		return -EOPNOTSUPP;
	    }
	    wmb();
	    break;

	// this is the version that the user sees
	case SIBYTE_GET_VERSION:
	    sprintf((char*) &scmd, "%s", ARUBA_VERSION);
	    if(copy_to_user(ifr->ifr_data, &scmd, sizeof(scmd)))
	        return -EFAULT;
	    break;

	// the base version is used for the configuration revision
	case SIBYTE_GET_BASE_VERSION:
	    sprintf((char*) &scmd, "%s", ARUBA_BASE_VERSION);
	    if(copy_to_user(ifr->ifr_data, &scmd, sizeof(scmd)))
	        return -EFAULT;
	    break;

        case SIBYTE_GET_STATE:
            for (i = 0; i < SIBYTE_PRIOS; i++) {
                scmd.u.dsc.cur_rx[i] = sp->cur_rx[i];
                scmd.u.dsc.tx_prd[i] = sp->tx_prd[i];
                scmd.u.dsc.tx_csm[i] = sp->tx_csm[i];
            }
	    if(copy_to_user(ifr->ifr_data, &scmd, sizeof(scmd)))
	        return -EFAULT;
            break;

        case SIBYTE_GET_STATS:
            memcpy(&scmd.u.stats, &sp->stats, sizeof(scmd.u.stats));
	    if(copy_to_user(ifr->ifr_data, &scmd, sizeof(scmd)))
	        return -EFAULT;
            break;

	default:
	    return -EOPNOTSUPP;
	}
	return 0;
    }
      
#ifdef SIOCETHTOOL
    case SIOCETHTOOL: { 
        struct ethtool_cmd ecmd;

        if (copy_from_user(&ecmd, ifr->ifr_data, sizeof(ecmd)))
	    return -EFAULT;

	if (ecmd.cmd == ETHTOOL_GSET) {
	    ecmd.supported = SUPPORTED_1000baseT_Half;
	    ecmd.port = PORT_TP;
	    ecmd.transceiver = XCVR_INTERNAL;
	    ecmd.phy_address = 0;
	    ecmd.speed = SPEED_1000;
	    ecmd.duplex = DUPLEX_HALF;
	    ecmd.autoneg = AUTONEG_DISABLE;

	    if (copy_to_user(ifr->ifr_data, &ecmd, sizeof(ecmd)))
	        return -EFAULT;
	    return 0;

	} else if (ecmd.cmd == ETHTOOL_SSET) {
	    return -EPERM;
	}
    }
#endif

    case SIOCSIFHWADDR:
        if (copy_from_user(dev->dev_addr, ifr->ifr_data, sizeof(dev->dev_addr)))
	    return -EFAULT;
	return 0;

    case SIOCGIFHWADDR:
        if (copy_to_user(ifr->ifr_data, dev->dev_addr, sizeof(dev->dev_addr)))
	    return -EFAULT;
	return 0;

    default:
        break;
    }
    return -EOPNOTSUPP;
}

static struct net_device_stats *sibyte_get_stats(struct net_device *dev)
{
    struct sibyte_private *sp = dev->priv;
    return &sp->stats;
}

/*
 * Local variables:
 * compile-command: "gcc -D__SMP__ -D__KERNEL__ -DMODULE -I../../include -Wall -Wstrict-prototypes -O2 -fomit-frame-pointer -pipe -fno-strength-reduce -DMODVERSIONS -include ../../include/linux/modversions.h   -c -o sibyte.o sibyte.c"
 * End:
 */
