/*
 * BRIEF MODULE DESCRIPTION
 *	MERLOT interrupt routines.
 *
 * Copyright 2002 MontaVista Software Inc.
 * Author: MontaVista Software, Inc.
 *		stevel@mvista.com or source@mvista.com
 *
 *  This program is free software; you can redistribute	 it and/or modify it
 *  under  the terms of	 the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED	  ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,	  INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY	  DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED	  TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN	 CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/module.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timex.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/delay.h>

#include <asm/bitops.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/aruba/merlot.h>

#include <asm/irq.h>

//#define DEBUG_IRQ
#ifdef DEBUG_IRQ
/* note: prints function name for you */
#define DPRINTK(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#ifdef CONFIG_REMOTE_DEBUG
extern void breakpoint(void);
#endif

extern asmlinkage void idtIRQ(void);
extern void set_debug_traps(void);
extern irq_cpustat_t irq_stat [NR_CPUS];
unsigned int local_bh_count[NR_CPUS];
unsigned int local_irq_count[NR_CPUS];

static unsigned int startup_irq(unsigned int irq);
static void end_irq(unsigned int irq_nr);
static void mask_and_ack_irq(unsigned int irq_nr);
static void merlot_enable_irq(unsigned int irq_nr);
static void merlot_disable_irq(unsigned int irq_nr);

extern unsigned int do_IRQ(int irq, struct pt_regs *regs);
extern void __init init_generic_irq(void);

#ifdef CONFIG_PM
extern void counter0_irq(int irq, void *dev_id, struct pt_regs *regs);
#endif

typedef struct {
  u32 mask;       /* mask of valid bits in pending/mask registers */
  volatile u32 *base_addr;
} intr_group_t;

#define IC_GROUP_OFFSET 0x0C

#define READ_PEND_MERLOT(base) (*((volatile unsigned long *)base))
#define READ_MASK_MERLOT(base) (*((volatile unsigned long *)base + 1))
#define WRITE_MASK_MERLOT(base, val) ((*((volatile unsigned long *)base + 1)) = (val))

static const intr_group_t intr_group_muscat[NUM_INTR_GROUPS] = {
  { 0x0000efff, (u32 *)KSEG1ADDR(0x18038000 + 0 * IC_GROUP_OFFSET) },
  { 0x00001fff, (u32 *)KSEG1ADDR(0x18038000 + 1 * IC_GROUP_OFFSET) },
  { 0x00000007, (u32 *)KSEG1ADDR(0x18038000 + 2 * IC_GROUP_OFFSET) },
  { 0x0003ffff, (u32 *)KSEG1ADDR(0x18038000 + 3 * IC_GROUP_OFFSET) },
  { 0xffffffff, (u32 *)KSEG1ADDR(0x18038000 + 4 * IC_GROUP_OFFSET) }
};

#define READ_PEND_MUSCAT(base) (*(base))
#define READ_MASK_MUSCAT(base) (*(base + 2))
#define WRITE_MASK_MUSCAT(base, val) (*(base + 2) = (val))

static inline int irq_to_group(unsigned int irq_nr)
{
  switch (mips_chip_vendor) {
  case MACH_VENDOR_ATHEROS:
    // always group 0
  default:
    return 0;
  case MACH_VENDOR_IDT:
    return ((irq_nr - GROUP0_IRQ_BASE) >> 5);
  }
}

static inline int group_to_ip(unsigned int group)
{
  switch (mips_chip_vendor) {
  case MACH_VENDOR_ATHEROS:
  default:
    return mips_local_ip;
  case MACH_VENDOR_IDT:
    return group+2;
  }
}

static inline void enable_local_irq(unsigned int ip)
{
  int ipnum = 0x100 << ip;
  clear_cp0_cause(ipnum);
  set_cp0_status(ipnum);
}

static inline void disable_local_irq(unsigned int ip)
{
  int ipnum = 0x100 << ip;
  clear_cp0_status(ipnum);
}

static inline void ack_local_irq(unsigned int ip)
{
  int ipnum = 0x100 << ip;
  clear_cp0_cause(ipnum);
}

static void merlot_enable_irq(unsigned int irq_nr)
{
  unsigned long flags;
  int           ip = irq_nr - GROUP0_IRQ_BASE;
  unsigned int  group, intr_bit;
  volatile unsigned int  *addr;

#ifdef DEBUG_IRQ
  printk("enable irq %d ip %d\n", irq_nr, ip);
#endif
  save_and_cli(flags);
  
  if (ip < 0) {
    enable_local_irq(irq_nr);
  } else {
    // calculate group

    switch (mips_chip_vendor) {
    case MACH_VENDOR_ATHEROS:
    default:
      group = 0;
      break;
    case MACH_VENDOR_IDT:
      group = ip >> 5;
      break;
    }
      
    
    // calc interrupt bit 
    ip -= (group << 5);
    intr_bit = 1 << ip;
    
    // first enable the IP mapped to this IRQ
    enable_local_irq(group_to_ip(group));
    
    switch (mips_chip_vendor) {
    case MACH_VENDOR_ATHEROS:
    default:
      addr = mips_intr_status;
      WRITE_MASK_MERLOT(addr, (READ_MASK_MERLOT(addr) | intr_bit));
    // unmask intr within group
#ifdef DEBUG_IRQ
     printk("done enable addr %x irq %d mask %x status %x intr_bit %x\n", 
		    addr, irq_nr, 
		    READ_MASK_MERLOT(addr), READ_PEND_MERLOT(addr), intr_bit);
#endif
      break;

    case MACH_VENDOR_IDT:
      addr = intr_group_muscat[group].base_addr;
      WRITE_MASK_MUSCAT(addr, READ_MASK_MUSCAT(addr) & ~intr_bit);
    // unmask intr within group
#ifdef DEBUG_IRQ
      printk("done enable addr %x irq %d mask %x status %x intr_bit %x\n", 
             addr, irq_nr, 
             READ_MASK_MUSCAT(addr), READ_PEND_MUSCAT(addr), intr_bit);
#endif
      break;
    }
  }

  restore_flags(flags);

}

static void merlot_disable_irq(unsigned int irq_nr)
{
  unsigned long flags;
  int           ip = irq_nr - GROUP0_IRQ_BASE;
  unsigned int  group, intr_bit, mask;
  volatile unsigned int  *addr;

#if 0
  printk("merlot_disable_irq %x\n", irq_nr);
#endif
  save_and_cli(flags);
  
  if (ip < 0) {
    disable_local_irq(irq_nr);
  } else {
    // calculate group

    switch (mips_chip_vendor) {
    case MACH_VENDOR_ATHEROS:
    default:
      group = 0;
      break;
    case MACH_VENDOR_IDT:
      group = ip >> 5;
      break;
     }

    // calc interrupt bit
    ip -= group << 5;
    intr_bit = 1 << ip;

    switch (mips_chip_vendor) {
    case MACH_VENDOR_ATHEROS:
    default:
      addr = mips_intr_status;
      // mask intr within group
      WRITE_MASK_MERLOT(addr, (READ_MASK_MERLOT(addr) & ~intr_bit));
    
      /*
       * if there are no more interrupts enabled in this
       * group, disable corresponding IP
       */
      if (READ_MASK_MERLOT(addr)) 
        disable_local_irq(group_to_ip(group));
      break;

    case MACH_VENDOR_IDT:
        addr = intr_group_muscat[group].base_addr;
        // mask intr within group
        mask = READ_MASK_MUSCAT(addr);
        mask |= intr_bit;
        WRITE_MASK_MUSCAT(addr, mask);
    
        /*
         * if there are no more interrupts enabled in this
         * group, disable corresponding IP
         */
        if (mask == intr_group_muscat[group].mask)
          disable_local_irq(group_to_ip(group));
        break;
    }
  }
  
  restore_flags(flags);
}

static unsigned int startup_irq(unsigned int irq_nr)
{
#ifdef DEBUG_IRQ
  printk("startup_irq %x\n", irq_nr);
#endif
  merlot_enable_irq(irq_nr);
  return 0; 
}

static void shutdown_irq(unsigned int irq_nr)
{
#ifdef DEBUG_IRQ
  printk("shutdown_irq %x\n", irq_nr);
#endif
  merlot_disable_irq(irq_nr);
  return;
}

static void mask_and_ack_irq(unsigned int irq_nr)
{
#if 0
  printk("mask_and_ack_irq %x\n", irq_nr);
#endif
  merlot_disable_irq(irq_nr);
  ack_local_irq(group_to_ip(irq_to_group(irq_nr)));
}

static void end_irq(unsigned int irq_nr)
{

  unsigned long flags;
  int ip = irq_nr - GROUP0_IRQ_BASE;
  unsigned int intr_bit, group;
  volatile unsigned int *addr;
#if 0
  printk("end_irq %x\n", irq_nr);
#endif
  save_and_cli(flags);
  if (!(irq_desc[irq_nr].status & (IRQ_DISABLED | IRQ_INPROGRESS))) {
    if (ip < 0) {
      enable_local_irq(irq_nr);
    } else {
      switch (mips_chip_vendor) {
      case MACH_VENDOR_ATHEROS:
      default:
        group = 0;

        // calc interrupt bit within group
        intr_bit = 1 << ip;

        // first enable the IP mapped to this IRQ
        enable_local_irq(group_to_ip(group));
  
        addr = mips_intr_status;
        // unmask intr within group
        WRITE_MASK_MERLOT(addr, (READ_MASK_MERLOT(addr) | intr_bit));
        break;

      case MACH_VENDOR_IDT:
        if (irq_nr == GROUP4_IRQ_BASE + 9)       gpio->gpioistat &= 0xfffffdff; /* AP70 Enet1 */
        else if (irq_nr == GROUP4_IRQ_BASE + 10) gpio->gpioistat &= 0xfffffbff; /* AP70 USB */
        else if (irq_nr == GROUP4_IRQ_BASE + 11) gpio->gpioistat &= 0xfffff7ff; /* AP70 PCI0 */
        else if (irq_nr == GROUP4_IRQ_BASE + 12) gpio->gpioistat &= 0xffffefff; /* AP70 PCI1 */
        else if (irq_nr == GROUP4_IRQ_BASE + 2) gpio->gpioistat &= 0xfffffffb; /* AP85 PCI0 */
        else if (irq_nr == GROUP4_IRQ_BASE + 3) gpio->gpioistat &= 0xfffffff7; /* AP85 PCI1 */

        group = ip >> 5;

        // calc interrupt bit within group
        ip -= (group << 5);
        intr_bit = 1 << ip;
        //  if (!intr_bit)
        //    return;

        // first enable the IP mapped to this IRQ
        enable_local_irq(group_to_ip(group));
  
        addr = intr_group_muscat[group].base_addr;
        // unmask intr within group
        WRITE_MASK_MUSCAT(addr, READ_MASK_MUSCAT(addr) & ~intr_bit);
        break;
      }
    }
  } 
  else {
#if 1
    printk("warning: end_irq %d did not enable (%x)\n", 
	   irq_nr, irq_desc[irq_nr].status);
#endif
  }
  restore_flags(flags);
}

static struct hw_interrupt_type merlot_irq_type = {
  "MERLOT",
  startup_irq,
  shutdown_irq,
  merlot_enable_irq,
  merlot_disable_irq,
  mask_and_ack_irq,
  end_irq,
  NULL
};


void __init init_IRQ(void)
{
  int i;
  unsigned long cp0_status;
  unsigned long *addr;

#ifdef DEBUG_IRQ
  printk("Initializing IRQ's: %d out of %d\n", MERLOT_NR_IRQS, NR_IRQS);  
#endif
  cp0_status = read_32bit_cp0_register(CP0_STATUS);
  memset(irq_desc, 0, sizeof(irq_desc));
  set_except_vector(0, idtIRQ);
  
  set_cp0_status(0xFF00|cp0_status);
  //  init_generic_irq();
  
  for (i = 0; i < MERLOT_NR_IRQS; i++) {
    irq_desc[i].status = IRQ_DISABLED;
    irq_desc[i].action = NULL;
    irq_desc[i].depth = 1;
    irq_desc[i].handler = &merlot_irq_type;
  }

  switch (mips_chip_vendor) {
  case MACH_VENDOR_ATHEROS:
  default:
    WRITE_MASK_MERLOT(mips_intr_status,0);
    switch (mips_chip) {
    case MACH_ATHEROS_AR5312:
    case MACH_ATHEROS_AR2313:
      /* enable UART (SEEMS UNEEDED) */
      WRITE_MASK_MERLOT(mips_intr_status,0x10);
      break;
    case MACH_ATHEROS_AR2315:
      /* enable watchdog interrupt */
      WRITE_MASK_MERLOT(mips_intr_status,0x80);
      break;
    }
    break;
  case MACH_VENDOR_IDT:
    break;
  }

#ifdef CONFIG_REMOTE_DEBUG
  /* If local serial I/O used for debug port, enter kgdb at once */
  puts("Waiting for kgdb to connect...");
  set_debug_traps();
  breakpoint(); 
#endif
}

/* Main Interrupt dispatcher */
asmlinkage void merlot_irqdispatch(unsigned long cp0_cause, struct pt_regs *regs)
{
  unsigned int ip, pend, group;
  volatile unsigned int *addr;
  unsigned int irq;

  switch (mips_chip_vendor) {
  case MACH_VENDOR_ATHEROS:
  default:
    if(cp0_cause == 0) {
	    return;
    }

    switch (mips_chip) {
    case MACH_ATHEROS_AR5312:
    case MACH_ATHEROS_AR2313:
      if (cp0_cause & (1<<(8+MERLOT_MISC_IRQ))) {
        // Misc Interrupt
        group = 0;
        addr = mips_intr_status;
        pend = READ_PEND_MERLOT(addr);
        pend &= READ_MASK_MERLOT(addr); // only unmasked interrupts
        /* handle one misc interrupt at a time */
        while(pend) {
          unsigned int intr_bit, irq_nr;
          intr_bit = pend ^ (pend-1);
          irq_nr = ((31 - merlot_clz(pend))+GROUP0_IRQ_BASE);
          do_IRQ(irq_nr, regs);
          pend &= ~intr_bit; 
        }
      } 

      if (cp0_cause & (1<<(8+MERLOT_WLAN_IRQ))) {
        do_IRQ(MERLOT_WLAN_IRQ, regs);
      }

      if (cp0_cause & (1<<(8+MERLOT_WLAN1_IRQ))) {
        if (mips_chip == MACH_ATHEROS_AR5312) {
          do_IRQ(MERLOT_WLAN1_IRQ, regs);
        }
      }

      if (cp0_cause & (1<<(8+MERLOT_ENET_IRQ))) {
        do_IRQ(MERLOT_ENET_IRQ, regs);
      }
      if (cp0_cause & (1<<(8+MERLOT_ENET_IRQ+1))) {
        do_IRQ(MERLOT_ENET_IRQ+1, regs);
      }
      break;

    case MACH_ATHEROS_AR2315:
      if (cp0_cause & (1<<(8+2))) {
        // Misc Interrupt
        group = 0;
        addr = mips_intr_status;
        pend = READ_PEND_MERLOT(addr);
        pend &= READ_MASK_MERLOT(addr); // only unmasked interrupts
        if (pend & 0x80) {
          char str[80];
          sprintf(str, "Watchdog (AHB %x) Interrupt!\n",
                  *((unsigned int *)0xb1000050));
          __die(str, regs, __FUNCTION__, __LINE__);
        }
        /* handle one misc interrupt at a time */
        while(pend) {
          unsigned int intr_bit, irq_nr;
          intr_bit = pend ^ (pend-1);
          irq_nr = ((31 - merlot_clz(pend))+GROUP0_IRQ_BASE);
          do_IRQ(irq_nr, regs);
          pend &= ~intr_bit; 
        }
      } 
      if (cp0_cause & (1<<(8+3))) {
        do_IRQ(3, regs);
      }
      if (cp0_cause & (1<<(8+4))) {
        do_IRQ(4, regs);
      }
      break;
    }
    break;

  case MACH_VENDOR_IDT:
    if ((ip = (cp0_cause & 0x7c00))) {
      group = 21 - merlot_clz(ip);
      addr = intr_group_muscat[group].base_addr;
      pend = READ_PEND_MUSCAT(addr);
      pend &= ~READ_MASK_MUSCAT(addr); // only unmasked interrupts
      //  if (!pend)
      //    return; // no interrupts pending in this group
      pend = 39 - merlot_clz(pend);
      irq = (group<<5)+pend;

      do_IRQ((group << 5) + pend, regs);
      return;
    } 
    irq = 23 - merlot_clz(ip);
    do_IRQ(23 - merlot_clz(ip), regs);
    break;
  }

  return;
}
