#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/notifier.h>
#include <linux/mtd/mtd.h>
#include <linux/interrupt.h>
#include <asm/bootinfo.h>
#include <sap/flash/ap60_flash_common.h>
#include <sap/flash/fl_cache.h>

extern void lock_ap60_70_flash(void);
extern void unlock_ap60_70_flash(void);

static void
apfc_compute_digest(__u8 *digest, __u8 *buf, int size)
{
    __u32 sum;
    __u32 *p;
    int i;
    int nwords = size / sizeof(__u32);

    p = (__u32 *) buf;
    sum = 0;
    for (i = 0; i < nwords; i++) {
        sum += *p++;
    }
    memset(digest, 0, APFC_DIGEST_SIZE);
    memcpy(digest, &sum, sizeof(sum));
}

volatile static int erase_done;

static void
erase_callback(struct erase_info *instr)
{
    erase_done = 1;
}

void
aruba_save_crash_reason(char *buf)
{
    static int entries = 0;
    struct mtd_info *mtd;
    loff_t cache_addr;
    __u32 cache_size;
    size_t retlen;
    int ret;
    apfc_hdr_t *hdr;
    static __u8 cache_buf[AP60_FLASH_SECTOR2_SIZE];
    __u8 odigest[APFC_DIGEST_SIZE];
    __u8 ndigest[APFC_DIGEST_SIZE];
    apfc_entry_hdr_t *entry;
    int msglen = strlen(buf);
    struct timeval tv;
    static struct erase_info erase;

    /*
     * XXX
     * As of now, we can't do this in interrupt context because the
     * flash device functions may sleep.
     */

    if (in_interrupt()) {
        return;
    }

    /*
     * Prevent repeated calls in case of additional panics/page faults.
     */

    if (++entries > 1) {
        return;
    }

    switch (mips_machtype) {
    case MACH_ARUBA_ALFRESCO:
    case MACH_ORTRONICS_WJD:
    case MACH_ORTRONICS_DUO:
        cache_addr = AP80_TOT_FLASH_SIZE - 3 * AP60_FLASH_SECTOR2_SIZE;
        cache_size = AP60_FLASH_SECTOR2_SIZE;
        break;
    case MACH_ARUBA_MERLOT:
    case MACH_ARUBA_MUSCAT:
    case MACH_ARUBA_MATARO:
        cache_addr = AP60_TOT_FLASH_SIZE - 3 * AP60_FLASH_SECTOR2_SIZE;
        cache_size = AP60_FLASH_SECTOR2_SIZE;
        break;
    default:
        /* other types not supported */
        return;
    }

    mtd = get_mtd_device(NULL, 0);
    if (!mtd) {
        printk("%s: unable to get mtd device\n", __FUNCTION__);
        return;
    }

    unlock_ap60_70_flash();
    ret = mtd->read(mtd, cache_addr, cache_size, &retlen, cache_buf);
    lock_ap60_70_flash();

    if (ret != 0) {
        printk("%s: mtd read failed (%d)\n", __FUNCTION__, ret);
        return;
    }
    
    hdr = (apfc_hdr_t *) cache_buf;

    entry = NULL;

    if (strncmp(hdr->magic, APFC_MAGIC_STRING, APFC_MAGIC_SIZE) == 0) {
        memcpy(odigest, hdr->digest, APFC_DIGEST_SIZE);
        memset(hdr->digest, 0, APFC_DIGEST_SIZE);
        apfc_compute_digest(ndigest, cache_buf, cache_size);
        if (memcmp(odigest, ndigest, APFC_DIGEST_SIZE) == 0) {
            /*
             * Remove existing reboot entry if present, and find end.
             */
            apfc_entry_hdr_t *eh = (apfc_entry_hdr_t *) (hdr + 1);
            apfc_entry_hdr_t *endp = (apfc_entry_hdr_t *) (cache_buf +
                    cache_size - sizeof(apfc_entry_hdr_t));
            while (eh <= endp) {
                if (eh->type[0] == '\0') {
                    entry = eh;
                    break;
                }
                if (strncmp(eh->type, "reboot", APFC_ENTRY_TYPE_SIZE) == 0) {
                    __u32 todelete = APFC_TOTAL_ENTRY_SIZE(eh);
                    apfc_entry_hdr_t *neh = APFC_NEXT_ENTRY(eh);
                    if (neh > endp) {
                        /* error */
                        break;
                    }
                    memmove(eh, neh, (cache_buf + cache_size) - ((__u8 *) neh));
                    memset(cache_buf + cache_size - todelete, 0, todelete);
                }
                else {
                    eh = APFC_NEXT_ENTRY(eh);
                }
            }
        }
        else {
            printk("%s: digest doesn't match", __FUNCTION__);
        }
    }

    if (!entry) {
        memset(cache_buf, 0, cache_size);
        memcpy(hdr->magic, APFC_MAGIC_STRING, APFC_MAGIC_SIZE);
        entry = (apfc_entry_hdr_t *) (hdr + 1);
    }

    memset(entry, 0, sizeof(apfc_entry_hdr_t));
    strncpy(entry->type, "reboot", APFC_ENTRY_TYPE_SIZE);
    entry->size = msglen;
    if (APFC_NEXT_ENTRY(entry) >
            ((apfc_entry_hdr_t *) (cache_buf + cache_size))) {
        printk("%s: No room to store reboot reason");
        return;
    }
    memcpy((entry + 1), buf, msglen);

    do_gettimeofday(&tv);
    hdr->last_write_time = tv.tv_sec;
    hdr->write_count++;

    apfc_compute_digest(ndigest, cache_buf, cache_size);
    memcpy(hdr->digest, ndigest, APFC_DIGEST_SIZE);

    unlock_ap60_70_flash();

    memset(&erase, 0, sizeof(struct erase_info));
    erase.addr = cache_addr;
    erase.len = cache_size;
    erase.mtd = mtd;
    erase.callback = erase_callback;
    erase_done = 0;
    ret = mtd->erase(mtd, &erase);
    if (ret != 0) {
        printk("%s: mtd erase failed (%d)\n", __FUNCTION__, ret);
        lock_ap60_70_flash();
        return;
    }
    while (erase_done == 0)
        ;

    ret = mtd->write(mtd, cache_addr, cache_size, &retlen, cache_buf);

    lock_ap60_70_flash();

    if (ret != 0) {
        printk("%s: mtd write failed (%d)\n", __FUNCTION__, ret);
    }
}

static int
aruba_notify_panic(struct notifier_block *self, unsigned long unused1,
			void *ptr)
{
    char *buf = ptr;
    char msgbuf[256];

    snprintf(msgbuf, sizeof(msgbuf), "Reboot caused by kernel panic: %s", buf);
    aruba_save_crash_reason(msgbuf);

    return 0;
}

static struct notifier_block aruba_panic_notifier = {
    .notifier_call = aruba_notify_panic,
};

void
aruba_panic_init(void)
{
    notifier_chain_register(&panic_notifier_list, &aruba_panic_notifier);
}
