/*****************************************************************
**  File : sar_linux_specific.c
**
**		Includes linux dependent functions for SOFTSAR. 
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
**  09/22/2001   sadik              Ported from sar_vxworks_specific.c
******************************************************************/

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/delay.h>  /* udelay */
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/tqueue.h>  
#include <asm/rc32300/rc32355.h>

/* Include files */
#include "sys_ss_def.h"
#include "sar_if.h"
#include "sardrv.h"
#include "sar_linux_specific.h"
#include "banyan.h"
#include "timer.h"
#include "sysinteg.h"
#include "dma.h"
#include "atm.h"

#define MAX_DMA_INTRS   8 /* DMA channels 0 through 8 are used by ATM */

#define SAR_COUNTER_TIMER_IRQ   GROUP0_IRQ_BASE + 1
#define SAR_DMA_IRQ   GROUP1_IRQ_BASE + 0
#define SAR_ATM_IRQ   GROUP2_IRQ_BASE + 0


UInt32  dma_interrupt_vectors[MAX_DMA_INTRS];

extern void sar_timer_isr(int irq, void *dev_id, struct pt_regs * regs);
void        sar_dma_isr(int irq, void *dev_id, struct pt_regs * regs);
extern void sar_atm_isr(int irq, void *dev_id, struct pt_regs * regs);
extern void sar_status_task(pATMIF_st );
extern pATM_VC_st  crunch_pvc_array[];
extern UInt32      atm_crunch_state;
void        init_dma_intr_vector_tbl(void);

static struct tq_struct sarStatusTask;

extern void *gATM;

/***************************************************************************************
Device IDs used for IRQ.
***************************************************************************************/
struct sar_device{
  unsigned char *name;
  unsigned int irq;
  void (*handler)(int, void *, struct pt_regs *);
};

static struct sar_device sar_counter_timer_dev =
  {
    "sar_counter_timer", SAR_COUNTER_TIMER_IRQ, sar_timer_isr
  }; 

static struct sar_device sar_dma_dev =
  {
    "sar_dma", SAR_DMA_IRQ, sar_dma_isr
  }; 

static struct sar_device sar_atm_dev =
  {
    "sar_atm", SAR_ATM_IRQ, sar_atm_isr
  }; 


/***************************************************************************************
Linux don't have bzero!
***************************************************************************************/
void bzero(volatile char* s,size_t count)
{
  volatile char *xs = (char *) s;

  while (count--)
    *xs++ = 0;

  return;
} /* bzero */  

/******************************************************************
** FUNCTION : set_dma_intr_vector
**
** DESCRIPTION:
**  Set the DMA interrupt vector table entry to the handler routine.
**
** NOTE:
**
******************************************************************/
void set_dma_intr_vector(void (*fptr)(), unsigned int index)
{
  dma_interrupt_vectors[index] = (unsigned int)fptr;

} /* set_dma_intr_vector */

/******************************************************************
** FUNCTION : init_dma_intr_vector_tbl
**
** DESCRIPTION:
**  Initialize the DMA interrupt vector table.
**
** NOTE:
**
******************************************************************/
void init_dma_intr_vector_tbl()
{
  int i;

  for (i = 0; i < MAX_DMA_INTRS; i++)
    dma_interrupt_vectors[i] = 0;

} /* init_dma_intr_vector_tbl */

/******************************************************************
** FUNCTION : sar_dma_isr
**
** DESCRIPTION:
** Only Chain interrupts need to be processed.
**
** NOTE:
**      The function assumes that each of the received PDU is held in a
**      single buffer. This is ensured by attaching MTU sized buffers to
**      each of the descriptors.
**
**   
******************************************************************/
static UInt32   gDMA_int_last = 0;
void sar_dma_isr(int irq, void *dev_id, struct pt_regs * regs)
{
  register pATMIF_st    pATM = (pATMIF_st )gATM;
  UInt32                status, dummy_read;
  volatile DMA_CHAN     *pdma_channel;
  pDMA_st               pchain;
  pATM_VC_st            pVC;
  unsigned int chan = (irq - GROUP1_IRQ_BASE);
  unsigned long key;

  save_and_cli(key);

  //  printk(__FUNCTION__" %d %d\n",irq,chan);
  /* Read the status register of the interrupting channel */
  pdma_channel = (DMA_CHAN *)((UInt32 )&dma.ch0 + chan * sizeof(DMA_CHAN));
  status = pdma_channel->dmas;
  pVC = pATM->pVC_array[chan];

  /* Process the interrupt */
  if (status & DMAxS_C)
    {   
      if (pVC->control & VPCI_ADMIN_UP)
        { 
          /* Chaining has taken effect - create a new descriptor list */

          pchain = get_2k_buf_desc_list(pATM, pVC, NUM_BDS_PER_LIST);

          if (pchain == (pDMA_st )NULL)
            { 
              pVC->state |= CHAIN_PENDING;
              crunch_pvc_array[chan - 1] = pVC;
              atm_crunch_state |= (1 << (chan - 1));
            }
          else
            { 
              dummy_read = pdma_channel->dmandptr;
              /* Attach the list to the DMA channel via NDPTR register */
              pdma_channel->dmandptr = VIRT_TO_PHYS((UInt32 )pchain);
            }
        }
    } /* if - CHAIN status */

  /* Clear the source of the interrupt */
  pdma_channel->dmas &= ~status;

  restore_flags(key);

} /* sar_dma_isr */

/******************************************************************
** FUNCTION : free_atmif_rtos_resources
**
** DESCRIPTION:
**	Free up all the resources allocated for the SOFTSAR.
**
** NOTE:
**
******************************************************************/
void free_atmif_rtos_resources(void *p)
{
  struct sar_device *dev;

  /* counter timer */
  dev = &sar_counter_timer_dev;
  free_irq(dev->irq, dev);

  /* dma */
  dev = &sar_dma_dev;
  free_irq(dev->irq, dev);

  /* atm */

  dev = &sar_atm_dev;
  free_irq(dev->irq, dev);

} /* free_atmif_rtos_resources */

/******************************************************************
** FUNCTION : rtos_dependent_init
**
** DESCRIPTION:
**	RTOS dependent initialization of SOFTSAR.
**
** NOTE:
**
******************************************************************/
void *rtos_dependent_init(pATMIF_st parm)
{
/*******************************************************************
  Initialise the sarStatusTask parameters.
*******************************************************************/
  sarStatusTask.routine =(void (*)(void *)) sar_status_task;
  sarStatusTask.data    = (pATMIF_st) parm;

  return(&sarStatusTask);
                                                                       

} /* rtos_dependent_init */

/******************************************************************
** FUNCTION : install_sar_int_hdlr()
**
** DESCRIPTION:
**	Install the interrupt handlers.
**
** NOTE:
**
******************************************************************/
int install_sar_int_hdlr(pATMIF_st pATM)
{
  struct sar_device *dev;

  dev = &sar_counter_timer_dev;
  if (request_irq(dev->irq, dev->handler, SA_SHIRQ | SA_INTERRUPT, dev->name, dev))
    {
      printk ("Unable to get IRQ %d for %s\n", dev->irq, dev->name);
      return EAGAIN;
    }

  dev = &sar_dma_dev;

 {
   int i;
   /* IRQs from 15 to 20 are used for VC cache. */
   for(i = GROUP1_IRQ_BASE + 1; i < GROUP1_IRQ_BASE + 7; i++)
     if (request_irq(i, &sar_dma_isr, SA_SHIRQ | SA_INTERRUPT, dev->name, dev))
       {
	 printk ("Unable to get IRQ %d for %s\n", dev->irq, dev->name);
	 return EAGAIN;
       }


   /* If DMA channel 7 is NOT used for FRAME Mode output, use it for VC cache */
   if (!(pATM->control & ATMIF_FRAME_MODE))
     if (request_irq(i++, &sar_dma_isr, SA_SHIRQ | SA_INTERRUPT, dev->name, dev))
       {
	 printk ("Unable to get IRQ %d for %s\n",dev->irq,dev->name);
	 return EAGAIN;
       }

   if ((pATM->control & ATMIF_TYPE_MASK) == ATM_UTOPIA2)
     { 
   /* If DMA channel 8 is NOT used for FRAME Mode output, use it for VC cache */
       if (!(pATM->control & ATMIF_FRAME_MODE))

	 if (request_irq(i, &sar_dma_isr, SA_SHIRQ | SA_INTERRUPT, dev->name, dev))
	   {
	     printk ("Unable to get IRQ %d for %s\n", dev->irq, dev->name);
	     return EAGAIN;
	   }
     }
   
 }
     
  dev = &sar_atm_dev;
 {
   int i;
   for(i = GROUP2_IRQ_BASE + 0; i < GROUP2_IRQ_BASE + 2; i++)
     if (request_irq(i, dev->handler, SA_SHIRQ | SA_INTERRUPT, dev->name, dev))
       {
	 printk ("Unable to get IRQ %d for %s\n", dev->irq, dev->name);
	 return EAGAIN;
       }
 }

#if 0
  if (request_irq(dev->irq, dev->handler, SA_INTERRUPT, dev->name, dev))
    {
      printk ("Unable to get IRQ %d for %s\n", dev->irq, dev->name);
      return EAGAIN;
    }
#endif
  return(1);

} /* install_sar_int_hdlr */

void post_status(void *p)
{
  queue_task(&sarStatusTask, &tq_immediate);
  mark_bh(IMMEDIATE_BH);
}

void taskDelay(unsigned long timeout)
{
    set_current_state(TASK_INTERRUPTIBLE);
    schedule_timeout(timeout);
} 

