/*******************************************************************************
 *
 * rng.c
 * 
 * rc32365 RNG Device Driver
 *
 * Author : calvin.young@idt.com
 * Date   : 06/30/2003
 * Update :
 *
 ******************************************************************************/
#include <asm/uaccess.h>
#include <asm/mipsregs.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/init.h>

#include <asm/system.h>															 
#include <asm/uaccess.h>
#include <asm/io.h>

#include "rng_reg.h"
#include <linux/idt_rng_interface.h>
#include <asm/dma.h>
#include <asm/rc32300/rc32365.h>

#define DMA_INTR_NUM	(GROUP1_IRQ_BASE + 8)

#define RNG_DEBUG 0
#ifdef RNG_DEBUG
#define dbg(lvl, format, arg...) \
	if (RNG_DEBUG > lvl) \
	printk(KERN_INFO "rng: " format, ## arg)
#else
#define dbg(lvl, format, arg...) do {} while (0)
#endif
#define err(format, arg...) \
	printk(KERN_ERR "rng: " format, ## arg)
#define info(format, arg...) \
	printk(KERN_INFO "rng: " format, ## arg)
#define warn(format, arg...) \
	printk(KERN_WARNING "rng: " format, ## arg)

static ssize_t rng_open(void *, void *);
static ssize_t rng_init(void);
static ssize_t rng_release(void *, void *) ;
static ssize_t rng_reset(void);
ssize_t rng_init_module(void);
void rng_cleanup_module(void);
static void rng_int_handler(int, void *, struct pt_regs *);
static int init_rng_dma_desc(void);
static void rng_dma_bh(void);
static int rng_go(void);


static DECLARE_TASKLET(rng_tasklet, (void *)rng_dma_bh, 0);

static idt_RNG_DEVICE rngDev = {
	name: "rc32365rng",
	open: rng_open,
	close: rng_release,
	reset: rng_reset,
	state: 0,
	rnFlag: 0,
	rnBuf: NULL,
	start: rng_go
};

static RNG_DEVICE *dev;

module_init(rng_init_module);
module_exit(rng_cleanup_module);

static inline void rng_halt_dma(void) 
{
	rc32365_halt_dma(dev->dma_regs); 
}

static inline void rng_start_dma(volatile DMAD_t id) 
{
	rc32365_start_dma(dev->dma_regs, virt_to_phys(id));
}

ssize_t __init rng_init_module( void )
{
	int rc;

	printk( "Random Number Generator present\n" );

	if(!(dev = (RNG_DEVICE *)kmalloc(sizeof(RNG_DEVICE), GFP_KERNEL))) {
		err("can't allocate RNG device memory\n"); 
		rc = -ENOMEM;
		goto rng_init_exit; 
	} 

	memset(dev, 0, sizeof(RNG_DEVICE));

	if (!(dev->rng_regs = ioremap_nocache(RNG_PhysicalAddress, sizeof(RNG_REG)))) {
		err("can't remap RNG registers\n");
		rc = -ENXIO;
		goto rng_init_exit; 
	}

	if(!(dev->dma_regs = ioremap_nocache(DMA0_PhysicalAddress + 8*DMA_CHAN_OFFSET,
				   sizeof(struct DMA_Chan_s)))) {
	   err("can't remap RNG dma registers\n");
	   rc = -ENXIO;
	   goto rng_init_exit; 
	}

	dev->dma_irq = DMA_INTR_NUM; 

	if (!(dev->dma_desc = (DMAD_t)kmalloc(RNG_DESC_RING_SIZE, GFP_KERNEL))) { 
		err("can't allocate RNG dma desc memory\n"); 
		rc = -ENOMEM;
		goto rng_init_exit; 
	}

	dma_cache_inv((unsigned long)(dev->dma_desc), RNG_DESC_RING_SIZE);
	memset((void *)dev->dma_desc, 0, RNG_DESC_RING_SIZE);
	dev->dma_desc = (DMAD_t )KSEG1ADDR(dev->dma_desc);
	
	/*
	 * allocate buffer for RNG input data
	 */
	if (!(dev->dma_buf = (u8*)kmalloc(RNG_DESC_BUF_SIZE, GFP_KERNEL))) {
		err("can't allocate RNG dma input buffers\n");
		rc = -ENOMEM; 	          
		goto rng_init_exit;
	}
		
	dma_cache_inv((unsigned long)(dev->dma_buf), RNG_DESC_BUF_SIZE);

	rngDev.rnBuf = dev->dma_buf; 
	register_rng_device(&rngDev);
	return (0);

rng_init_exit:
	return (1);
}

void rng_cleanup_module( void )
{
   return ;
}

static void start_dma(void)
{
	enable_irq(DMA_INTR_NUM);
	dev->rng_regs->rngc = 0x0;

	init_rng_dma_desc();
	writel(0, &dev->dma_regs->dmas);
	writel(0, &dev->dma_regs->dmasm);
	writel(0x08, &dev->dma_regs->dmac);
	writel(0, &dev->dma_regs->dmandptr);

	rng_reset();
	rng_start_dma(dev->dma_desc);
}

static ssize_t rng_open(void *irng, void *frng)
{

#ifdef  MODULE
	if(GET_USE_COUNT(THIS_MODULE));
#else
	if(0);
#endif
	else {
		if(rng_init()) {
			err("can't initialize RNG.");
			return(-1);
		}

		writel(0xFFFFFFFFU, &dev->dma_regs->dmas);
		writel(0xFFFFFFFFU, &dev->dma_regs->dmasm);
		if(request_irq(DMA_INTR_NUM, rng_int_handler, SA_INTERRUPT, "rng_dma_intr", 0)) {
			err("can't install RNG interrupt handler.\n");
			return (-1);
		}
		disable_irq(DMA_INTR_NUM);
	}
	start_dma();

#ifdef  MODULE
	MOD_INC_USE_COUNT;
#endif
	return (0);
}

static ssize_t rng_init(void)
{
	if(rng_reset())
		return(1);

	return (0);
}

static ssize_t rng_reset(void)
{

	dev->rng_regs->rngcfg = ((unsigned int)0x126254 << 11) | (unsigned int)RNG_CFG;
	dev->rng_regs->rngocfg = (unsigned int)RNG_OCFG;
	dev->rng_regs->rngseed0 = (unsigned int)RNG_SEED0;
	dev->rng_regs->rngseed1 = (unsigned int)RNG_SEED1;
	dev->rng_regs->rngc = 0x03;	
	dev->rng_regs->rngs = 0;
	dev->rng_regs->rngsm = 0xFFFFFFFFU;

	return (0);
}

static int rng_startup(void)
{
	rng_halt_dma();
	writel(0, &dev->rng_regs->rngc);
	rng_reset(); 
	writel(RNGC_en_m | RNGC_f_m, &dev->rng_regs->rngc);
	init_rng_dma_desc();
	writel(0, &dev->dma_regs->dmas);
	rng_start_dma(&dev->dma_desc[0]);
	writel(readl(&dev->dma_regs->dmasm) & 0x0,	
				&dev->dma_regs->dmasm); 
	return (0);
}

static int rng_go(void)
{	
	init_rng_dma_desc();
	writel(0, &dev->dma_regs->dmas);	
	rng_start_dma(&dev->dma_desc[0]);
	writel(readl(&dev->dma_regs->dmasm) & 0x0,	
			&dev->dma_regs->dmasm); 
    	return (0);
}

void rng_stop(void)
{
	rng_halt_dma();
	return;
}

void rng_restart(void)
{
	disable_irq(dev->dma_irq);
	writel(readl(&dev->dma_regs->dmasm) | (DMAS_f_m | DMAS_d_m),
			&dev->dma_regs->dmasm);

	rng_startup();
	enable_irq(dev->dma_irq); 
	return;
}

static void rng_int_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	volatile U32 dmas, rngs;	
	int i;

	if(!(dev->dma_regs->dmas & 0x01FU)) {
		err( "rng interrupt handler: Spurious interrupt.\n" );
		return;
	}

	dmas = dev->dma_regs->dmas;
	dev->dma_regs->dmas = ~dmas;
	rngs = dev->rng_regs->rngs;

	for (i=0; i<RNG_DESC_NUM; i++) {
		if (!(dev->dma_desc[i].control & (DMAD_d_m | DMAD_f_m)) ||
			(dev->dma_desc[i].control & DMAD_count_m)) {
			err("rng dma error. desc= %d, control= 0x%08x\n", i, dev->dma_desc[i].control);
			goto rng_int_exit;
		}
	}

	if (rngs & 0x078U) {
		err("error. rngs= 0x%08x\n", rngs);
		goto rng_int_exit;
	}

	rng_stop();
	tasklet_schedule(&rng_tasklet);
	return;

rng_int_exit:
	rng_restart();
	return;
}

static void rng_dma_bh(void)
{
	dma_cache_inv((unsigned long)dev->dma_buf, RNG_DESC_BUF_SIZE);
	rngDev.rnFlag = 1;
	return;
}

static int init_rng_dma_desc(void)
{
	int i;

	for (i=0; i< RNG_DESC_NUM; i++) {
		dev->dma_desc[i].control = 
			DMAD_iof_m | DMAD_iod_m | DMA_COUNT(RNG_DESC_BUF_SIZE);
		dev->dma_desc[i].devcs = 0;
		dev->dma_desc[i].ca =
			virt_to_phys(&dev->dma_buf[i * RNG_DESC_BUF_SIZE]);
		dev->dma_desc[i].link = virt_to_phys(&dev->dma_desc[i+1]);
	}
	dev->dma_desc[RNG_DESC_NUM-1].link = 0;
	return (0);
}

static ssize_t rng_release(void *irng, void *frng)
{
#ifdef  MODULE
	if(GET_USE_COUNT(THIS_MODULE)) {
		MOD_DEC_USE_COUNT;
#else
	if(0) {
#endif
	} else {
		if(rng_reset())
			return(1);
	}

	return(0);
}
