/*******************************************************************************
 *
 * se.c
 *
 * rc32365 security engine device driver
 *
 * Author : calvin.young@idt.com
 * Date   : 06/30/2003
 * Update :
 *
 ******************************************************************************/
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/init.h>
#include <asm/system.h>															 
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/if_arp.h>	
#include <linux/ip.h> 
#include <linux/config.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/tqueue.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <asm/delay.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/processor.h>
#include <asm/cache.h>
#include <asm/bootinfo.h>
#include <asm/pgtable.h>
#include <asm/segment.h>
#include <asm/dma.h>
#include  <asm/rc32300/rc32365.h>
#include "se_regs.h"
#include <linux/idt_se_interface.h>
#include "se.h"
#include "se_error.h"

/*
 * Prototypes
 */
static void se_interrupt(int, void *, struct pt_regs *);
static void se_bh(void);
static void se_cleanup(void); 
static int se_restart(int);
static inline void se_halt_out(void);
static inline void se_halt_in(void);
static inline void se_start_out(volatile DMAD_t);
static inline void se_start_in(volatile DMAD_t); 
static inline void se_chain_out(volatile DMAD_t);
static inline void se_chain_in(volatile DMAD_t);
static int se_startup(void);
static int init_se(SECURITY_ENGINE_UNIT_PARAM *);
static int shutdown_se(void);
static int se_out_process(CRYPTO_REQ *);

static SE_DEVICE *dev; 
static CONTEXT_CACHE contextCache[2] = {{INVALID_ID, {}}, {INVALID_ID, {}}};
static idt_CRYPTO_DEVICE cryptoDev = {
	name: "rc32365se",
	open: init_se,
	close: shutdown_se,
	process: se_out_process,
	state: 0
};
static DECLARE_TASKLET(se_tasklet, (void *)se_bh, 0);

static int se_read_proc(char *buf, char **start, off_t fpos,
			    int length, int *eof, void *data)
{
	int len = 0;

	len += sprintf(buf + len, "\n\tSecurity Engine Statistics\n\n");
	len += sprintf (buf + len,
			"total pkt count      = %10d, total cmd count      = %10d\n",
			dev->pkt_cnt, dev->cmd_cnt);

	len += sprintf (buf + len,
			"DMA full count       = %10d, context swap count   = %10d\n",
			dev->dma_full_cnt, dev->cntx_swap_cnt);
	if (fpos >= len) {
		*start = buf;
		*eof = 1;
		return 0;
	}
	*start = buf + fpos;
	if ((len -= fpos) > length)
		return length;
	*eof = 1;
	return len;

}

static void flush_context_cache(void)
{
	contextCache[0].contextId = INVALID_ID;
	contextCache[1].contextId = INVALID_ID;
}

static void flush_request(void)
{
	int i;
	CRYPTO_REQ *req = NULL;

	i = dev->s_head;
	for (; i<dev->s_tail;) {
		if (dev->s_ring[i].req && (dev->s_ring[i].req != req)) {
		    	req = dev->s_ring[i].req;
			req->result = SE_RESET_ERR; 
			idt_en_queue((QHEAD *)&dev->crypto_done_q,(QTYPE *)req);
		}
		i = ++i & SE_S_MASK;
	}

	if (req)
		tasklet_schedule(&se_tasklet);

}

/*
 * Security engine output processing.
 */
static int se_out_process(CRYPTO_REQ *req)	
{
	unsigned long flags;
	volatile DMAD_t od, od2;
	volatile struct DMAD_s lod, lod2;
	int next;
	u32 buf;
	int cmd = 0, intCmd = 0;
	int exec = 0;
	int io, count = 0;
#ifdef USE_IO_BUF
	int next_i_tail;
#endif

	spin_lock_irqsave(&dev->lock, flags);
	if (dev->o_count >= SE_O_MASK) {
		dbg(2, "output descriptor ring full\n");
		dev->o_full = 1;
		dev->dma_full_cnt++;
		crypto_stop_queue(&cryptoDev);	
		spin_unlock_irqrestore(&dev->lock, flags);
		return (SE_DESC_FULL); 
	}

	next = (dev->o_tail + 1) & SE_O_MASK;
	od = &dev->o_ring[dev->o_tail]; 
	if (!IS_DMA_USED(od->control)) {
		err("device owns descriptor.\n");
                dev->stats.o_errors++;
                se_restart(0);
                spin_unlock_irqrestore(&dev->lock, flags);
                return (SE_ERR);
	}
	io = (req->op & SE_SECURITY_CONTEXT_OP_ENCODE) ? 0 : 1;
	if (req->reqType == SE_COMMAND) {
		exec |= COMMAND;
		count++;
#ifdef SE_PROC_DEBUG
		dev->cmd_cnt++;	
#endif
	} else {
		/*
		 * if context id not the same as the last, context swap.
		 */
		exec |= DATA;
		if ((!io && (req->contextID != contextCache[0].contextId)) ||
		    (io && (req->contextID != contextCache[1].contextId))) {
			exec |= COMMAND;
			intCmd |= SE_CONTEXT_COMMAND_LOAD;
#ifdef SE_PROC_DEBUG
			dev->cntx_swap_cnt++;	
			dev->cmd_cnt++;	
#endif
			count++;

		}

		/*
		 * check if LOAD_PB needed.
		 */
		if ((req->packetContext->opFlag & PAD_BYTE) &&
			(req->packetContext->padByte != contextCache[io].pb)) {
				exec |= COMMAND;
				intCmd |= SE_CONTEXT_COMMAND_LOAD_PB;
				if (!(intCmd & SE_CONTEXT_COMMAND_LOAD)) {
					count++;
#ifdef SE_PROC_DEBUG
					dev->cmd_cnt++;	
#endif
				}

		}

		if ((dev->o_count + count) >= SE_O_MASK) {
			dbg(2, "output descriptor ring full\n");
			dev->o_full = 1;
			dev->dma_full_cnt++;
			crypto_stop_queue(&cryptoDev);
			spin_unlock_irqrestore(&dev->lock, flags);
			return (SE_DESC_FULL); 
		}
#ifdef SE_PROC_DEBUG
		dev->pkt_cnt++;
#endif
	}

	if (exec & COMMAND) {
		intCmd |= req->cmd;

		/*
		 * context command setup
		 */
		switch (intCmd) {
		case SE_CONTEXT_PB_COMMAND_LOAD:
			if (req->reqType == SE_COMMAND) {
				*((byte *)(req->commandContext->context) + 7) = req->commandContext->param.padByte;
			} else {
				*((byte *)(req->packetContext->externalSecurityRAM) + 7) = req->packetContext->padByte;
			}
		case SE_CONTEXT_COMMAND_LOAD:
			lod.devcs = (SEDEVCSCMD_cntxt_m & io);
			if (req->reqType == SE_COMMAND) {
				buf = (u32) req->commandContext->context;
			} else {
				buf = (u32) req->packetContext->externalSecurityRAM;
			}

			lod.ca = virt_to_phys((void *)buf);
			contextCache[io].pb = *((byte *)buf + 7);
			dma_cache_wback_inv((unsigned long)buf, sizeof(SECURITY_RAM_CONTEXT));

			lod.control = DMAD_iof_m | DMAD_iod_m | DMA_COUNT(sizeof(SECURITY_RAM_CONTEXT));
			contextCache[io].contextId = req->contextID;

			cmd = SEDEVCSCMD_cmd_ld_cntxt_v;
			break;
		case SE_CONTEXT_COMMAND_LOAD_PB:
			{
			byte pb;
			pb = (req->reqType == SE_COMMAND) ? req->commandContext->param.padByte :
							req->packetContext->padByte;
			lod.devcs = (SEDEVCSCMD_cntxt_m & io) |
				(SEDEVCSCMD_cmd_m & (SEDEVCSCMD_cmd_ld_pb_v << SEDEVCSCMD_cmd_b)) |
				(SEDEVCSCMD_imm_m & (pb << SEDEVCSCMD_imm_b));
			lod.ca = 0;    
			lod.control = DMAD_iof_m | DMAD_iod_m | DMA_COUNT(0);
			cmd = SEDEVCSCMD_cmd_ld_pb_v;
			contextCache[io].pb = pb;
			}
			break;
		default:
			dev->stats.o_errors++;
			err("invalid context command %d\n", intCmd);
			return (1);
			break;
		}
#ifndef DMA_CHAINING
		lod.link = (u32)NULL;
#endif
		dev->s_ring[dev->s_tail].req = req;
		dev->s_ring[dev->s_tail].type = COMMAND;
		dev->s_ring[dev->s_tail].cmd = cmd;
		dev->s_tail = (dev->s_tail + 1) & SE_S_MASK;
	}
       
	if (exec & DATA) {	/* SE DATA */
		count++;
		if (exec & COMMAND) {
			od2 = &dev->o_ring[next];
			next = ++next & SE_O_MASK;
		}
		buf = virt_to_phys(req->inBuf);  

		dma_cache_wback_inv((unsigned long)req->inBuf, req->inLen);

		lod2.devcs = (SEDEVCSOUT_cntxt_m & io) |
			    	((req->packetContext->opFlag & OVR_MODE) ? (SEDEVCSOUT_ovr_m |
					   ((req->packetContext->offset << SEDEVCSOUT_f_b) & SEDEVCSOUT_f_m) |
					   ((req->packetContext->length << SEDEVCSOUT_l_b) & SEDEVCSOUT_l_m)) : 0);
		lod2.devcs |= (req->packetContext->opFlag & DIS_MFP) ? (SEDEVCSOUT_dm_m & (0x1 << SEDEVCSOUT_dm_b)) : 0;
		lod2.devcs |= (SEDEVCSOUT_cntxt_m & io); 

		lod2.ca = buf;    
		lod2.control = (DMAD_devcmd_m & (SEDMAD_devcmd_first_last_data_v << DMAD_devcmd_b)) |
				(io? DMAD_iof_m | DMAD_iod_m | DMA_COUNT(req->inLen) :
				DMAD_iof_m | DMAD_iod_m | DMA_COUNT(req->packetContext->length));
		
		dev->stats.o_packets++;	
		dev->stats.o_bytes += req->packetContext->length;
#ifndef DMA_CHAINING
		lod2.link = (u32)NULL; 
#endif

		if (exec & COMMAND) { 
		    *od2 = lod2;
		    lod.link = virt_to_phys(&dev->o_ring[(next-1) & SE_O_MASK]);
		}
		else 
		    lod = lod2;
#ifdef USE_IO_BUF
		/*
		 * setup input descriptor buffer
		 */
		next_i_tail = (dev->i_tail + count - 1) & SE_I_MASK; 
		dev->i_ring[next_i_tail].ca = buf;
		dev->i_ring[next_i_tail].control &= ~DMAD_count_m; 
		dev->i_ring[next_i_tail].control |= DMA_COUNT(req->inLen);
		if(count > 1) { 
			dev->i_ring[dev->i_tail].link = virt_to_phys(&dev->i_ring[(dev->i_tail + 1) & SE_I_MASK]);
			dev->i_ring[(dev->i_tail + 1) & SE_I_MASK].link = (u32)NULL;
		} else {
			dev->i_ring[dev->i_tail].link = (u32) NULL;	
		}
#endif
		dev->s_ring[dev->s_tail].req = req;
		dev->s_ring[dev->s_tail].type = DATA;
		dev->s_ring[dev->s_tail].cmd = -1;	
		dev->s_ring[dev->s_tail].mode = io;
		dev->s_ring[dev->s_tail].buf = (u32 *)(req->inBuf);
#ifdef USE_IO_BUF
		dev->s_ring[dev->s_tail].bufLen = req->inLen;	
#endif
		dev->s_tail = (dev->s_tail + 1) & SE_S_MASK;
	}
#ifdef USE_IO_BUF
	se_start_in(&dev->i_ring[dev->i_tail]);	
#endif

#ifdef DMA_CHAINING
	if(dev->o_dma_regs->dmandptr == 0)
	{
		*od = lod;
		((struct DMAD_s *)od + count - 1)->control |= DMAD_cof_m | DMAD_cod_m;
		dev->o_dma_regs->dmandptr = virt_to_phys(&dev->o_ring[dev->o_chain]);
	}	 
#else	
		*od = lod;
		se_start_out(od);
#endif

	/*
	 * adjust indices, counter
	 */
#ifdef DMA_CHAINING
	dev->o_chain = dev->o_tail = next;
#else
	dev->o_tail = next;
#endif
#ifdef USE_IO_BUF
	dev->i_tail = (dev->i_tail + count) & SE_I_MASK; 
#endif
	dev->o_count += count;
	if ((dev->o_count) == SE_NUM_O - 1) {
		dev->o_full = 1;
		dev->dma_full_cnt++;
		crypto_stop_queue(&cryptoDev);	
		dbg(1, "output ring is full.\n");
	} 

	spin_unlock_irqrestore(&dev->lock, flags);
	return (SE_OK);
}

/*
 * SE buttom half
 */
static void se_bh(void)
{
	CRYPTO_REQ *req;

#ifndef BYPASS_REQ_QUEUEING
	idt_run_crypto_request();
#endif
	while(1) { 
		req = (CRYPTO_REQ *)idt_de_queue((QHEAD *)&dev->crypto_done_q);
		if(req == NULL) {
			return;
		}

		if (req->pCallback) 
			req->pCallback(
				req->contextID, req->cb,
				req->newbuf,	
				req->callHandle, req->result);
#ifndef PKT_API_BUF_INCLUDED
		se_free_crypto_request(req);
#endif
	}
}	

/*
 * SE error interrupt handler
 */
static void se_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 seSes;

	if (dev_id == NULL) {
		err("spurious interrupt. irq %d for unknown device.\n", irq);
		return;
	}

	seSes = readl(&dev->se_regs->ses);
	writel(0, &dev->se_regs->ses);
	if (IS_SESES_MIDS_ERR(seSes)) {
		err("malformed input data stream\n");
	} else if (IS_SESES_RME_ERR(seSes)) {
		err("reserved mode error\n");
	} else if (IS_SESES_DME_ERR(seSes)) {
		err("disabled mode error\n");
	} else if (IS_SESES_CDE_ERR(seSes)) {
		err("command data error\n");
	} else {
		err("spurious interrupt. No error detected.\n");
		return;
	}
	dev->stats.i_errors++;
	spin_lock(&dev->lock);
	se_restart(1); 
	spin_unlock(&dev->lock);
}

static int se_in_dma(void)
{
	volatile DMAD_t id = &dev->i_ring[dev->i_head];
	u32 devcs, reg;
	u32 cmd;
	SE_RET_CODE rc = SE_OK;
	CRYPTO_REQ *req;
	SE_DMAD_S *sd;
	u32 count = 0;
	u8 *data = NULL;
#ifdef USE_SK_BUF
	struct sk_buff *skb, *skb_new = NULL;
#endif
	while (IS_DMA_USED(id->control)) {
		devcs = id->devcs;
		sd = &dev->s_ring[dev->s_head];
		if (sd->type == COMMAND) {
			if (IS_SE_DATA(id->control)) {
				dbg(3, "command request, data response.\n");	
				goto se_in_dma_exit; 
			}

			cmd = (devcs & SEDEVCSCMD_cmd_m) >> SEDEVCSCMD_cmd_b;
			/*
			 * check errors 
			 */
			if (IS_CMD_CXT_ERR(devcs)) { 
				dev->stats.i_errors++;
			 	err("command 0x%x: Invalid context number 0x%x (ctrl= 0x%x : devcs= 0x%x)\n",
					cmd, CMD_CONTEXT(devcs),
					(unsigned int)id->control, (unsigned int)id->devcs);
				rc |= SE_INVALID_CONTEXT_ID; 
			}
			reg = readl(&dev->se_regs->ses);
			if (IS_CMD_PROC_ERR(devcs) && IS_SESES_CDE_ERR(reg)) {
				dev->stats.i_errors++;
				err("command 0x%x: Invalid context size tranfered\n", cmd);
				rc |= SE_INVALID_CONTEXT_SIZE; 
			}
			
			req = sd->req;

			if (sd->cmd != cmd) {
				err("command 0x%x not matched, out of sync.\n", cmd);
				rc |= SE_CMD_NOT_MATCHED; 
				goto se_in_dma_exit;
			}

			count = SE_I_DATA_BUF_SIZE - (u32)DMA_COUNT(id->control);
			if (((id->control & DMAD_devcmd_m) >> DMAD_devcmd_b) != SEDMAD_devcmd_cmd_v) {	
				err("input desc 0x%x not the command.\n", (id->control & DMAD_devcmd_m) >> DMAD_devcmd_b);
				goto se_in_dma_exit;
			}

			switch (cmd) {
			case SEDEVCSCMD_cmd_st_cntxt_v:
			case SEDEVCSCMD_cmd_rp_cntxt_v:
				if (count != sizeof(SECURITY_RAM_CONTEXT)) {
					rc |= SE_INVALID_RETURN_CONTEXT_DATA;
					goto se_in_dma_exit;
				}
				break;
			case SEDEVCSCMD_cmd_st_iv_v:
				if (count != 16) { 
					rc |= SE_INVALID_RETURN_IV_DATA; 
					goto se_in_dma_exit;	
				}
				break;
			}

			switch (cmd) {
			case SEDEVCSCMD_cmd_st_cntxt_v:
			case SEDEVCSCMD_cmd_rp_cntxt_v:
			case SEDEVCSCMD_cmd_st_iv_v:
#if defined(USE_SK_BUF)
				data = dev->i_skb[dev->s_head]->tail;
#elif defined(USE_IO_BUF)
				data = &dev->i_data_buf[dev->s_head * SE_I_DATA_BUF_SIZE];
#else
				data = &dev->i_data_buf[dev->s_head * SE_I_DATA_BUF_SIZE];
#endif
				dma_cache_inv((unsigned long) data, count);
				memcpy(req->outBuf, data, count); 
			}
			req->result = rc;
		} else {
			if (!IS_SE_DATA(id->control)) {
				dev->stats.i_errors++;
				dbg(3, "data request, command response id->control= 0x%08x\n",
				    id->control);	
				goto se_in_dma_exit;
			}

			/*
			 * check errors
			 */
			if (IS_ANY_ERROR(devcs)) {
				if (IS_DATA_CXT_ERR(devcs)) {
					rc |= SE_DATA_CXT_ERR;
					err("data context error\n");
				}
				if (IS_DATA_ERR(devcs)) {
					rc |= SE_DATA_ERR;
					err("malformed input data error\n");
				}
				if (IS_DATA_EME_ERR(devcs)) {
					rc |= SE_DATA_EME_ERR;
					err("encryption unit misconfiguration error\n");
				}
				if (IS_DATA_EBE_ERR(devcs)) {
					rc |= SE_DATA_EBE_ERR;
					err("encryption unit block error\n");
				}
				if (IS_DATA_HME_ERR(devcs)) {
					rc |= SE_DATA_HME_ERR;
					err("hash unit misconfiguration error\n");
				}
				if (IS_DATA_HTM_ERR(devcs)) {
					rc |= SE_DATA_HTM_ERR;
					err("hash mismatch error\n");
				}
				if (IS_DATA_PME_ERR(devcs)) {
					rc |= SE_DATA_PME_ERR;
					err("pad unit misconfiguration error\n");
				}
				if (IS_DATA_PE_ERR(devcs)) {
					rc |= SE_DATA_PE_ERR;
					err("pad error\n");
				}
				/*
				 * next header error should be ignored for decrypted inbound packet
				 */
				if (IS_DATA_NHE_ERR(devcs)) {
					if(sd->mode == 0) {
						rc |= SE_DATA_NHE_ERR;
						err("next header error\n");
					}
				}
				if (IS_DATA_AHM_ERR(devcs)) {
					rc |= SE_DATA_AHM_ERR;
					err("ah mismatch error\n");
				}
				dev->stats.i_errors++;
			}
#if defined(USE_IO_BUF) 
			count = sd->bufLen - (u32)DMA_COUNT(id->control);	
			if (((id->control & DMAD_devcmd_m) >> DMAD_devcmd_b) != SEDMAD_devcmd_first_last_data_v) {
				dev->stats.i_errors++;
				err("input data desc 0x%x not the first and last desc.\n", (id->control & DMAD_devcmd_m) >> DMAD_devcmd_b);
			}
			data = (u8 *)(sd->buf);
			dma_cache_inv((unsigned long) data, count);
			req = sd->req;
			req->result = rc;	

			if (count > 0) {
 				if (req->packetContext->opFlag & HASH_TRAILER)
					memcpy(req->packetContext->hashTrailer,
					        data + req->outLen - 12,
						req->packetContext->hashTrailerSize);

				idt_en_queue((QHEAD *)&dev->crypto_done_q,(QTYPE *)req);
			}
#elif defined(USE_SK_BUF)
			count = SE_I_DATA_BUF_SIZE - (u32)DMA_COUNT(id->control);
			skb = dev->i_skb[dev->s_head];
			if (((id->control & DMAD_devcmd_m) >> DMAD_devcmd_b) != SEDMAD_devcmd_first_last_data_v) {
				dev->stats.i_errors++;
				err("input data desc 0x%x not the first and last desc.\n", (id->control & DMAD_devcmd_m) >> DMAD_devcmd_b);
				goto se_in_dma_exit;
			}
			data = skb->tail;

			/*
			 * refill new sk buffer
			 */
			skb_new = dev_alloc_skb(SE_I_DATA_BUF_SIZE + SE_I_DATA_HDR_SIZE);
			if (!skb_new) {
				err("allocate sk buffer failed\n");
				goto se_in_dma_exit;
			}
			skb_reserve(skb_new, SE_I_DATA_HDR_SIZE);
			dev->i_skb[dev->s_head] = skb_new;

			req = sd->req;
			req->result = rc;	

			if (count > 0) {
				dma_cache_inv((unsigned long) data, count);
                                if (req->packetContext->opFlag & HASH_TRAILER) {
					memcpy(req->packetContext->hashTrailer,
						data + req->outLen,
						req->packetContext->hashTrailerSize);
					skb_put(skb, count-req->packetContext->hashTrailerSize);
				} else {
					skb_put(skb, count);
				}
				req->newbuf = (void *)skb;		

				idt_en_queue((QHEAD *)&dev->crypto_done_q,(QTYPE *)req);
			}
#else /* use memory copy */
			/*
			 * copy returned data to user buffer
			 */
			count = SE_I_DATA_BUF_SIZE - (u32)DMA_COUNT(id->control);
			if (((id->control & DMAD_devcmd_m) >> DMAD_devcmd_b) != SEDMAD_devcmd_first_last_data_v) {
				dev->stats.i_errors++;
				err("input data desc 0x%x not the first and last desc.\n", (id->control & DMAD_devcmd_m) >> DMAD_devcmd_b);
				goto se_in_dma_exit;
			}
			data = &dev->i_data_buf[dev->s_head * SE_I_DATA_BUF_SIZE];
			dma_cache_inv((unsigned long) data, count);
			req = sd->req;
			req->result = rc;	

			if (count > 0) {
				memcpy(req->outBuf,
					data,
					req->outLen);

				if (req->packetContext->opFlag & HASH_TRAILER)
					memcpy(req->packetContext->hashTrailer,
					        data + req->outLen,
						req->packetContext->hashTrailerSize);

				idt_en_queue((QHEAD *)&dev->crypto_done_q,(QTYPE *)req);
			} 
#endif
		}

		dev->stats.i_packets++;
		dev->stats.i_bytes += count;

#if defined(USE_SK_BUF)
		if (skb_new) {
		    	id->ca = virt_to_phys(skb_new->tail);
		} else {
			id->ca = virt_to_phys(dev->i_skb[dev->i_head]->tail);	
		}
#elif defined(USE_IO_BUF)
		id->ca = virt_to_phys(&dev->i_data_buf[dev->i_head * SE_I_DATA_BUF_SIZE]);
#else
		id->ca = virt_to_phys(&dev->i_data_buf[dev->i_head * SE_I_DATA_BUF_SIZE]); 
#endif
		id->control = DMAD_iod_m | DMA_COUNT(SE_I_DATA_BUF_SIZE);
		id->devcs = 0;
#ifdef DMA_CHAINING
		dev->i_tail = (dev->i_tail + 1) & SE_I_MASK;
		dev->i_chain_empty = 0;
#endif
		dev->i_head = (dev->i_head + 1) & SE_I_MASK;
		dev->s_head = (dev->s_head + 1) & SE_S_MASK;
		id = &dev->i_ring[dev->i_head];

		writel(0, &dev->i_dma_regs->dmas);
	} 
	return (0);

se_in_dma_exit:
	dbg(0, "se error= %x. se restated.\n", rc);
	dev->stats.i_errors++;
	se_restart(1);
	return (0);
}

/*
 * SE input DMA interrupt handler
 */
static void se_in_dma_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 dmas;
	
	if (!dev_id) {
		err("spurious interrupt. irq %d for unknown device.\n", irq);
		return;
	}

	spin_lock(&dev->lock);
	dmas = readl(&dev->i_dma_regs->dmas);
	writel(~dmas, &dev->i_dma_regs->dmas); 
	if (dmas & DMAS_f_m) {
		err("input dma finished? reset SE.\n");
		se_restart(1); 
	} 
	
#ifdef DMA_CHAINING
	if (dev->i_dma_regs->dmandptr == 0) {
		if(dev->i_chain_empty == 0)
		{
			dev->i_ring[(dev->i_tail-1) & SE_I_MASK].control |= DMAD_cof_m | DMAD_cod_m;
			dev->i_dma_regs->dmandptr  = virt_to_phys(&dev->i_ring[dev->i_chain]);
			dev->i_chain = dev->i_tail;
			dev->i_chain_empty = 1;
		}	 
	}
#endif
	if (dmas & (DMAS_d_m | DMAS_f_m)) { 
		writel(0, &dev->i_dma_regs->dmas);  
		se_in_dma();
	} else if (dmas & DMAS_e_m) {
		err("dma error\n");
		writel(0, &dev->i_dma_regs->dmas);  
		dev->stats.i_errors++;
	}
	spin_unlock(&dev->lock);
	tasklet_schedule(&se_tasklet);
}

static void se_out_dma(void)
{
	volatile DMAD_t od = &dev->o_ring[dev->o_head];
	volatile struct DMAD_s lod;

	while (dev->o_count && IS_DMA_USED(od->control)) {
		lod = *od;
		od->devcs = 0;
		od->ca = 0;
#ifndef DMA_CHAINING
		od->link = 0;
#endif
		
		/*
		 * start output processing if descriptor available
		 */
		if (dev->o_full) {
			dev->o_full = 0;
			crypto_wake_queue(&cryptoDev);
			tasklet_schedule(&se_tasklet);
		}
		
		dev->o_count--;
		dev->o_head = (dev->o_head + 1) & SE_O_MASK;
		od = &dev->o_ring[dev->o_head];
	}
}
/*
 * SE output DMA interrupt handler
 */
static void se_out_dma_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 dmas;

	if (!dev_id) {
		err("spurious interrupt irq %d for unknown device.\n", irq);
		return;
	}

	spin_lock(&dev->lock);

	dmas = readl(&dev->o_dma_regs->dmas);
	writel(0, &dev->o_dma_regs->dmas); 

	if (dmas & DMAS_e_m)
		err("dma error\n");
#ifdef DMA_CHAINING
	if(dev->o_dma_regs->dmandptr == 0)
	{
		if(dev->o_chain != dev->o_tail) 
		{
			dev->o_ring[(dev->o_tail-1) & SE_O_MASK].control |= DMAD_cof_m  | DMAD_cod_m; //chain-on-finish to terminate txNextIn chain
			dev->o_dma_regs->dmandptr = virt_to_phys(&dev->o_ring[dev->o_chain]);
			dev->o_chain = dev->o_tail;
		}	 
	}	 
#endif
	if (dmas & DMAS_f_m)
		se_out_dma();
	spin_unlock(&dev->lock);
}

/*
 * DMA operations
 */
static inline void se_halt_out(void)
{
	if (rc32365_halt_dma(dev->o_dma_regs))
		err(__FUNCTION__": timeout!\n");
}
static inline void se_halt_in(void)
{
	if (rc32365_halt_dma(dev->i_dma_regs))
		err(__FUNCTION__": timeout!\n");
}

static inline void se_start_out(volatile DMAD_t od)
{
	rc32365_start_dma(dev->o_dma_regs, virt_to_phys(od));
}
static inline void se_start_in(volatile DMAD_t id) 
{
	rc32365_start_dma(dev->i_dma_regs, virt_to_phys(id));
}

static inline void se_chain_out(volatile DMAD_t od)
{
	rc32365_chain_dma(dev->o_dma_regs, virt_to_phys(od));
}
static inline void se_chain_in(volatile DMAD_t id)
{
	rc32365_chain_dma(dev->i_dma_regs, virt_to_phys(id));
}

/*
 * Shutdown SE
 */
static int shutdown_se(void)
{
	u32 tmp;
	unregister_crypto_device(&cryptoDev);
    
	disable_irq(dev->o_irq); 
	disable_irq(dev->i_irq); 
	disable_irq(dev->irq);

	tmp = readl(&dev->o_dma_regs->dmasm);
	tmp = tmp | DMASM_f_m | DMASM_e_m;
	writel(tmp, &dev->o_dma_regs->dmasm);
     
	tmp = readl(&dev->i_dma_regs->dmasm);
	tmp = tmp | DMASM_d_m | DMASM_h_m | DMASM_e_m;
	writel(tmp, &dev->i_dma_regs->dmasm);

	free_irq(dev->i_irq, dev);
	free_irq(dev->o_irq, dev);
	free_irq(dev->irq, dev);
	return (0);
}

/*
 * Restart SE
 */
static int se_restart(int flush)
{
	crypto_stop_queue(&cryptoDev);
	disable_irq(dev->o_irq); 
	disable_irq(dev->i_irq);
	disable_irq(dev->irq); 
		
	writel(readl(&dev->o_dma_regs->dmasm) | DMAS_f_m,
	       &dev->o_dma_regs->dmasm);
     
	writel(readl(&dev->i_dma_regs->dmasm) | DMAS_d_m,
	       &dev->i_dma_regs->dmasm);

	if (flush)
		flush_request();
	flush_context_cache();
	se_startup();

	enable_irq(dev->irq);
	enable_irq(dev->i_irq); 
	enable_irq(dev->o_irq);
	crypto_wake_queue(&cryptoDev);
	return (0);
}

/*
 * Start up SE
 */
static int se_startup(void)
{
	int i;

	se_halt_out();
	se_halt_in();

	writel(0, &dev->se_regs->sec);
	writel(SESEC_en_m | SESEC_vc_m, &dev->se_regs->sec); 

	/*
	 * initialize descriptors.
	 */
	for (i = 0; i < SE_NUM_O; i++) {
		dev->o_ring[i].control = DMAD_f_m;
		dev->o_ring[i].devcs = 0;
		dev->o_ring[i].ca = 0;
#ifdef DMA_CHAINING
#else
		dev->o_ring[i].link = 0;
#endif
	}
#ifdef DMA_CHAINING
	dev->o_ring[SE_O_MASK].link = 0;
#endif
	dev->o_tail = dev->o_head = 0; 
	dev->s_tail = dev->s_head = 0;
	dev->o_count = 0;
	dev->o_full = 0;

#ifdef DMA_CHAINING
	dev->i_chain = 0;
	dev->i_chain_empty = 1;
#endif
	for (i=0; i< SE_NUM_I; i++) {
#ifdef USE_SK_BUF
		if (dev->i_skb[i]) 
			dev_kfree_skb_any(dev->i_skb[i]);

		dev->i_skb[i] = dev_alloc_skb(SE_I_DATA_BUF_SIZE + SE_I_DATA_HDR_SIZE);
		if (dev->i_skb[i] == NULL) {
			err("alloc sk buffers failed\n");
		}
		skb_reserve(dev->i_skb[i], SE_I_DATA_HDR_SIZE);
#endif
		dev->i_ring[i].control = 
			DMAD_iod_m | DMAD_cof_m | DMA_COUNT(SE_I_DATA_BUF_SIZE); 
		dev->i_ring[i].devcs = 0;

#ifdef USE_SK_BUF
		dev->i_ring[i].ca = virt_to_phys(dev->i_skb[i]->tail);
#else
		dev->i_ring[i].ca =
			virt_to_phys(&dev->i_data_buf[i * SE_I_DATA_BUF_SIZE]);
#endif
#ifdef USE_IO_BUF
		dev->i_ring[i].link = (U32)NULL; 
#else
		dev->i_ring[i].link = virt_to_phys(&dev->i_ring[i+1]);
#endif
	}
#ifdef USE_IO_BUF
	dev->i_ring[SE_NUM_I-1].link = (U32)NULL;
#else
	dev->i_ring[SE_NUM_I-1].link = virt_to_phys(&dev->i_ring[0]);
	#ifdef DMA_CHAINING
       		dev->i_ring[SE_NUM_I-1].control |= DMAD_cof_m | DMAD_cod_m; 
	#endif
#endif

	dev->i_head = dev->i_tail = 0;
	writel(0, &dev->i_dma_regs->dmas);

#ifndef USE_IO_BUF
	se_start_in(&dev->i_ring[0]);
#endif
	writel(readl(&dev->o_dma_regs->dmasm) & 0x0,	
				&dev->o_dma_regs->dmasm); 

	writel(readl(&dev->i_dma_regs->dmasm) & 0x0,	
				&dev->i_dma_regs->dmasm);

	writel(readl(&dev->se_regs->sesm) & 
		   ~(SESESM_mids_m | SESESM_rme_m | SESESM_dme_m | SESESM_cde_m),
		   		&dev->se_regs->sesm);
	return (0);
}

/*
 * SE subsequent initialization 
 */
static int init_se(SECURITY_ENGINE_UNIT_PARAM *pParam)
{
	if (request_irq(dev->i_irq, se_in_dma_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"se dma output", dev)) {
		err("unable to get se dma output irq %d\n",
		    dev->i_irq);
		return -EAGAIN;
	}
	if (request_irq(dev->o_irq, se_out_dma_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"se dma input", dev)) {
		err("unable to get se dma input irq %d\n",
		    dev->o_irq);
		free_irq(dev->i_irq, dev);
		return -EAGAIN;
	}
	if (request_irq(dev->irq, se_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"se", dev)) {
		err("unable to get se irq %d\n",
		    dev->irq);
		free_irq(dev->i_irq, dev);
		free_irq(dev->o_irq, dev);
		return -EAGAIN;
	}
	se_startup();
	return (0);
}

/*
 * SE initialization
 */
int
se_probe(void)
{
	int descSize, rc;

	request_region(SE_PhysicalAddress, sizeof(SE_REG), "rc32365se");

	if(!(dev = (SE_DEVICE *)kmalloc(sizeof(SE_DEVICE), GFP_KERNEL))) {
		err("can't allocate SE memory\n"); 
		rc = -ENOMEM;
		goto se_probe_exit; 
	} 
	memset(dev, 0, sizeof(SE_DEVICE));

	dev->i_irq = SE_DMA_I_IRQ;
	dev->o_irq = SE_DMA_O_IRQ;
	dev->irq = SE_IRQ;

	if (!(dev->se_regs = ioremap_nocache(SE_PhysicalAddress, sizeof(SE_REG)))) {
		err("can't remap SE registers\n");
		rc = -ENXIO;
		goto se_probe_exit; 
	}

	if(!(dev->i_dma_regs = ioremap_nocache(DMA0_PhysicalAddress + 7*DMA_CHAN_OFFSET,
 				 		sizeof(struct DMA_Chan_s)))) {
		err("can't remap SE input dma registers\n");
		rc = -ENXIO;
		goto se_probe_exit; 
	}
		
	if(!(dev->o_dma_regs = ioremap_nocache(DMA0_PhysicalAddress + 6*DMA_CHAN_OFFSET,
				 		sizeof(struct DMA_Chan_s)))) {
		err("can't remap SE output dma registers\n");
		rc = -ENXIO;
		goto se_probe_exit; 
	}

	/*
	 * allocate memory for descriptors
	 */
	descSize = 3 * SE_O_RING_SIZE;
	if (!(dev->o_ring = (DMAD_t)kmalloc(descSize, GFP_KERNEL))) { 
		err("can't allocate memory\n"); 
		rc = -ENOMEM;
		goto se_probe_exit; 
	}
	dma_cache_inv((unsigned long)(dev->o_ring), descSize);
	memset((void *)dev->o_ring, 0, descSize);
	dev->o_ring = (DMAD_t )KSEG1ADDR(dev->o_ring);
	dev->i_ring = &dev->o_ring[SE_NUM_O];
	dev->s_ring = (SE_DMAD_S *)&dev->i_ring[SE_NUM_I];

#ifndef USE_SK_BUF
	/*
	 * allocate buffer for input data stream.
	 */
	if (!(dev->i_data_buf = (u8*)kmalloc(SE_NUM_I * SE_I_DATA_BUF_SIZE,
				    GFP_KERNEL))) {
		err("can't allocate input dma input buffers\n");
		rc = -ENOMEM; 	          
		goto se_probe_exit;
	} 
	dma_cache_inv((unsigned long)(dev->i_data_buf),
			    	SE_NUM_I * SE_I_DATA_BUF_SIZE);
#endif

	spin_lock_init(&dev->lock);

	idt_queue_init(&dev->crypto_done_q);
	return (SE_OK);

se_probe_exit:
	se_cleanup();
	err("failed. returns %d\n", rc);

	return (SE_ERR);
};


/* 
 * Driver Initialization
 */
int __init se_init(void) 
{
	if(se_probe())
		goto se_init_exit;

	printk("Security Engine Present\n");

	if (register_crypto_device(&cryptoDev))
		goto se_init_exit;

#ifdef SE_PROC_DEBUG
	dev->ps = create_proc_read_entry ("driver/se", 0, NULL, 
					  se_read_proc, dev);
#endif
	return (0);

se_init_exit:
	err("security Engine initialization failed\n");
	return (1);
};

static void se_cleanup(void)
{
	if (dev->se_regs) {
		iounmap((void *)dev->se_regs);
	}
	if (dev->i_dma_regs) {
		iounmap((void *)dev->i_dma_regs);
	}
	if (dev->o_dma_regs) {
		iounmap((void *)dev->o_dma_regs);
	}
	if (dev->o_ring) {
		kfree((void *)KSEG0ADDR(dev->o_ring));
	}
#ifdef USE_SK_BUF
	if (dev->i_skb[i]) 
		dev_kfree_skb_any(dev->i_skb[i]);
#endif
	if (dev->i_data_buf) {
		kfree((const void *)dev->i_data_buf);
	}
#ifdef SE_PROC_DEBUG
	if (dev->ps)
		remove_proc_entry("driver/se", NULL); 
#endif
       	if (dev) {
		kfree((const void *)dev);
	}
	return;
}

module_init(se_init);
module_exit(se_cleanup);

