/*
   	i2c-rc32438.c
  
   	Author: Integrated Device Technology, Inc.
			hkou@idt.com
  
   	Linux i2c driver for IDT 79RC32438 on-board I2C.
   
   	This driver is modified from IDT:
  
  
   	Haofeng Kou:			Design modify and maintain the code. 
      						The author may be reached at hkou@idt.com or
						IDT-IPD, 2975 Stender Way, Santa Clara, CA 95054 


	Changelog:

	Rev 1.0			  	Haofeng Kou			3/6/03
						: This module contains the low level I2C driver for 79RC32438
						: Driver(i2c_driver) Adapter(i2c_adapter) and algorithm(i2c_algorithm) 
						: Support /dev interface
						: Enable /proc interface
*/


#include <linux/module.h>
#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/completion.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/i2c-dev.h>
#include <asm/io.h>

#include "i2c-rc32438.h"

MODULE_AUTHOR("hfkou");
MODULE_DESCRIPTION("I2C driver for IDT's 79RC32438 board");
MODULE_LICENSE("GPL");


struct  i2c_adapter* rc32438_adapter;
static struct i2c_rc32438_iface *rc32438_iface = NULL;
struct resource *rc32438_io_region;
unsigned int rc32438_master_done;
unsigned int rc32438_slave_done;
unsigned int rc32438_num_master_data_bytes_txd;
unsigned int rc32438_num_master_data_bytes_rxd;
unsigned int rc32438_num_master_done_ints;
unsigned int rc32438_num_master_lost_arb_ints;
unsigned int rc32438_num_master_err_ints;
unsigned int rc32438_num_acks;
unsigned int rc32438_num_naks;


static u32 i2c_rc32438_func(struct i2c_adapter * adapter)
{
	return I2C_FUNC_I2C;
}

void i2c_rc32438_master_isr (void)
{
  	unsigned int master_status;


  	// Read the Master Status Regs
  	master_status = i2c.i2cms;

  	if (master_status & I2CMS_ERR) 
	{
    		rc32438_num_master_err_ints++;
    		printk("\nI2C Master ERR Detected!\n");
  	}

  	while (!(master_status & I2CMS_D));

  	if (master_status & I2CMS_D)
    		rc32438_num_master_done_ints++;

  	if (master_status & I2CMS_LA) 
	{
    		rc32438_num_master_lost_arb_ints++;
    		printk ("\nI2C Master LA Detected!\n");
  	}

	/*printk("master.state: %x\n", master.state);*/
  	// Master is done with the current operation.
  	switch (rc32438_master.state) 
	{
    		case MASTER_IDLE:
    		// No need to do anything...
      			break;

		case MASTER_START:
    		// DONE sending START, begin sending address
      			i2c.i2cdo = rc32438_master.dest_addr[0];
      			i2c.i2cmcmd = I2CMCMD_CMD(WD);
      			rc32438_master.state = MASTER_ADDR;
      			break;

		case MASTER_ADDR:
    		// Count ACKs & NAKs - note, an ACK occurs when the ACK bit is cleared
    		// (Because SDA is driven low)
      			if (!(master_status & I2CMS_NA))
        			rc32438_num_acks++;
      			else
        			rc32438_num_naks++;
      			if (master_status & I2CMS_NA) 
			{
      				// No Slave ACknowledged the Address byte, so generate STOP if desired
        			if (rc32438_master.stop_when_done) 
				{
          				i2c.i2cmcmd = I2CMCMD_CMD(STOP);
          				rc32438_master.state = MASTER_STOP;
        			}
        			else 
				{
  					// No STOP desired, so go to IDLE and set global variable
          				rc32438_master.state = MASTER_IDLE;
          				rc32438_master_done = TRUE;
  					// Mask ALL Master Interrupts
          				i2c.i2cmsm = 0xFFFFFFFF;
        			}
        			break;
      			}
				
		      	if (rc32438_master.data_len == 0) 
			{
		  		// Data Length is zero, so skip Write / Read Stage
		        	if (rc32438_master.stop_when_done) 
				{
		          		i2c.i2cmcmd = I2CMCMD_CMD(STOP);
		          		rc32438_master.state = MASTER_STOP;
		        	}
		        	else 
				{
		    			// No STOP desired, so go to IDLE and set global variable
		          		rc32438_master.state = MASTER_IDLE;
		          		rc32438_master_done = TRUE;
		    			// Mask ALL Master Interrupts
		          		i2c.i2cmsm = 0xFFFFFFFF;
		        	}
		        	break;
		      	}

			if (rc32438_master.transfer_type == MASTER_WRITE) 
			{
		  		// DONE sending address, now send data
		        	rc32438_master.state = MASTER_WRITE_DATA;
		        	rc32438_num_master_data_bytes_txd++;
		        	i2c.i2cdo =  *rc32438_master.data_ptr++;
		        	i2c.i2cmcmd = I2CMCMD_CMD(WD);
		      	}
		      	else 
			{
		  		// DONE sending address, now read data
		        	rc32438_master.state = MASTER_READ_DATA;
		        	if (rc32438_num_master_data_bytes_rxd == (rc32438_master.data_len - 1))
		    			// Almost done reading data, now send RD (not RDACK!)
		          		i2c.i2cmcmd = I2CMCMD_CMD(RD);
		        	else
		    			// Read Another Data Byte (And ACK)
		          		i2c.i2cmcmd = I2CMCMD_CMD(RDACK);
		        	rc32438_num_master_data_bytes_rxd++;
		      	}
		      	break;
			
	    	case MASTER_WRITE_DATA:
    			// Count ACKs & NAKs - note, an ACK occurs when the ACK bit is cleared
    			// (Because SDA is driven low)
      			if (!(master_status & I2CMS_NA))
        			rc32438_num_acks++;
      			else
        			rc32438_num_naks++;
      			if (rc32438_num_master_data_bytes_txd >= rc32438_master.data_len) 
			{
  				// done sending data, now send STOP if desired.
        			if (rc32438_master.stop_when_done) 
				{
          				i2c.i2cmcmd = I2CMCMD_CMD(STOP);
          				rc32438_master.state = MASTER_STOP;
        			}
        			else 
				{
    					// No STOP desired, so go to IDLE and set global variable
          				rc32438_master.state = MASTER_IDLE;
          				rc32438_master_done = TRUE;
  					// Mask ALL Master Interrupts
          				i2c.i2cmsm = 0xFFFFFFFF;
        			}
      			}
      			else 
			{
  				// Send next data byte
        			i2c.i2cdo = *rc32438_master.data_ptr++;
        			i2c.i2cmcmd = I2CMCMD_CMD(WD);
        			rc32438_num_master_data_bytes_txd++;
      			}
			break;
			 
		case MASTER_READ_DATA:
    			// Write Incoming Read data to buffer
      			*rc32438_master.data_ptr = (unsigned char)i2c.i2cdi;
      			rc32438_master.data_ptr++;
    			// Count ACKs & NAKs - note, an ACK occurs when the ACK bit is cleared
    			// (Because SDA is driven low)
      			if (!(master_status & I2CMS_NA))
        			rc32438_num_acks++;
      			else
        			rc32438_num_naks++;
				
      			if (rc32438_num_master_data_bytes_rxd >= rc32438_master.data_len) 
			{
      				// done sending data, now send STOP if desired.
        			if (rc32438_master.stop_when_done) 
				{
          				i2c.i2cmcmd = I2CMCMD_CMD(STOP);
          				rc32438_master.state = MASTER_STOP;
        			}
        			else 
				{
	  				// No STOP desired, so go to IDLE and set global variable
	          			rc32438_master.state = MASTER_IDLE;
	          			rc32438_master_done = TRUE;
	  				// Mask ALL Master Interrupts
	          			i2c.i2cmsm = 0xFFFFFFFF;
	        		}
      			}				
      			else       // Almost done reading data, now send RD (not RDACK!)
			{
        			if (rc32438_num_master_data_bytes_rxd == (rc32438_master.data_len - 1)) 
				{
      					// Almost done reading data, now send RD (not RDACK!)
          				i2c.i2cmcmd = I2CMCMD_CMD(RD);
        			}
        			else 
				{
      					// Read Another Data Byte (And ACK)
          				i2c.i2cmcmd = I2CMCMD_CMD(RDACK);
        			}
			}
      			rc32438_num_master_data_bytes_rxd++;
      			break;
	  
    		case MASTER_STOP:
    			// Done with packet, set global variable, write NOP command, and go to idle
    			// Mask ALL Master Interrupts
      			i2c.i2cmsm = 0xFFFFFFFF;
      			rc32438_master.state = MASTER_IDLE;
      			rc32438_master_done = TRUE;
      			i2c.i2cmcmd = I2CMCMD_CMD(NOP);
      			break;
		
    		default:
      			printk ("\nErr in Default\n");
      			break;
  	}
}



/*
 * Generic i2c master transfer entrypoint
 */
static int i2c_rc32438_xfer(	struct i2c_adapter *adap,
		struct i2c_msg msgs[], 
		int num)
{
	struct i2c_msg *pmsg;
	int i, completed, rc = 0;
	unsigned char addr;

	completed = 0;
	for (i = 0; rc >= 0 && i < num; i ++) 
	{		
		pmsg = &msgs[i];
		if (pmsg->flags & I2C_M_TEN) 
		{
			printk(KERN_ERR "i2c-device(EPROM): 10 bits addr not supported !\n");
			rc = -EINVAL;
			break;
		}

		/*
		{
			printk("xfer: doing %s %d bytes to 0x%02x - %d of %d messages from %x. \n",
				     	pmsg->flags & I2C_M_RD ? "1 read" : "0 write",
		                     pmsg->len, pmsg->addr, i, num, pmsg->buf);
		}
		*/
		
		rc32438_master.stop_when_done = TRUE;
		rc32438_master.data_ptr = pmsg->buf;
		rc32438_num_master_data_bytes_txd = 0;
		rc32438_num_master_data_bytes_rxd = 0;
		rc32438_master.data_len = pmsg->len;
		rc32438_master_done = FALSE;

		if (pmsg->flags & I2C_M_RD)
			rc32438_master.transfer_type = MASTER_READ;
		else
			rc32438_master.transfer_type = MASTER_WRITE;

		rc32438_master.state = MASTER_IDLE;
		rc32438_master_done = FALSE;


                if (pmsg->addr != SLAVE_ADDR)
                {
                        //printk("Wrong Slave Address, use the default %x\n", SLAVE_ADDR);
                        pmsg->addr = SLAVE_ADDR;
                }
		addr = pmsg->addr << 1;
		if (rc32438_master.transfer_type == MASTER_WRITE)
			rc32438_master.dest_addr[0] = (unsigned char)addr;
		else
			rc32438_master.dest_addr[0] = (unsigned char)(addr |0x1);

		// Update Master State
		rc32438_master.state = MASTER_START;
		i2c.i2cmcmd = I2CMCMD_CMD(START);

	  	while (1) 
		{
	    		if (rc32438_master_done)
	      			break;
	    		if ((i2c.i2cms &~I2CMS_NA) != 0)
	      			i2c_rc32438_master_isr();
	  	}
		
	}

	return completed;
}



static struct i2c_algorithm rc32438_algorithm = 
{
	name:		"RC32438 I2C",
	id:			I2C_ALGO_SMBUS,
	smbus_xfer:	NULL,
	master_xfer:	i2c_rc32438_xfer,
	functionality:	i2c_rc32438_func,
};


int i2c_rc32438_create()
{
	unsigned int tsize, rc = 0;

	tsize = sizeof(struct i2c_rc32438_iface);
	rc32438_iface = (struct i2c_rc32438_iface *) kmalloc(tsize, GFP_KERNEL);
	if (rc32438_iface == NULL) 
	{
		printk(KERN_ERR "i2c-rc32438: can't allocate inteface !\n");
		return -ENOMEM;
	}
	memset((void*)rc32438_iface, 0, tsize);
	rc32438_iface->state = MASTER_IDLE;
	
	rc32438_iface->base = (unsigned long)I2C_BASE;	
	rc32438_io_region = request_region(rc32438_iface->base, I2C_REGION, "rc32438-i2c IO");

	rc32438_adapter = &rc32438_iface->adapter;
	
	sprintf(rc32438_adapter->name, "%s ", "RC32438-I2C");
	rc32438_adapter->id = I2C_ALGO_SMBUS;
	rc32438_adapter->algo = &rc32438_algorithm;
	rc32438_adapter->algo_data = NULL;
	rc32438_adapter->inc_use = NULL;
	rc32438_adapter->dec_use = NULL;
	rc32438_adapter->client_register = NULL;
	rc32438_adapter->client_unregister = NULL;
	rc32438_adapter->data = rc32438_iface;

	rc = i2c_add_adapter(rc32438_adapter);
	if (rc) 
	{
		printk("i2c-rc32438.c: Adapter %s registration failed\n", rc32438_adapter->name);
		kfree(rc32438_iface);
		rc32438_adapter->data = NULL;
	}
	
    	// Enable Master & Slave interfaces
  	i2c.i2cc = I2CC_MEN;

    	// Make sure to mask all unused bits.
  	i2c.i2ccp =  PRESCALER_VAL;
  	i2c.i2cmsm = 0xF;
  	i2c.i2cssm = 0x7F;

	return 0;
}

void i2c_rc32438_cleanup()
{
	if (rc32438_io_region != NULL)
		release_region(rc32438_iface->base, I2C_REGION);
	i2c_del_adapter(rc32438_adapter);
	kfree(rc32438_iface);		
}


int __init i2c_rc32438_init(void)
{
	int rc = -ENODEV;

	printk("i2c-rc32438.c: driver module\n");

	rc = i2c_rc32438_create();

	return rc;
}


void __exit i2c_rc32438_exit(void)
{
	i2c_rc32438_cleanup();
}



module_init(i2c_rc32438_init);
module_exit(i2c_rc32438_exit);


