/*
   	banyan.c
  
   	Author: Integrated Device Technology, Inc.
  			
   	Author: MontaVista Software, Inc.
           		stevel@mvista.com or source@mvista.com
  
   	A driver for the IDT RC32355 on-chip ethernet controller.
   
   	This driver is modified from IDT:
  

   	B Maruthanayagam: 	Ported from sonic.c 
   	cgg: 				Modified code. 
   	P Sadik: 				Applied filters.
   	Haofeng Kou:			Design modify and maintain the code. 
      						The author may be reached at hkou@idt.com or
						IDT-IPD, 2975 Stender Way, Santa Clara, CA 95054 


	Changelog:

	Rev 1.0			  	P Sadik				3/24/02
	
	Rev 1.1				Haofeng Kou			6/16/03
  						: Ported from the acacia.c(6/7/03 release)
  						: Add in code to handle the TX Underflow.



   ########################################################################
  
    This program is free software; you can redistribute  it and/or modify it
    under  the terms of  the GNU General  Public License as published by the
    Free Software Foundation;  either version 2 of the  License, or (at your
    option) any later version.
  
    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  
    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.
  
   ########################################################################
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/proc_fs.h>
#include <linux/in.h>
#include <linux/slab.h> 
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <asm/bootinfo.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/pgtable.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/dma.h>


//#define BANYAN_DEBUG 0xff //this must be before include "banyan.h"
#undef BANYAN_DEBUG

#include "banyan.h"

#undef BANYAN_PROC_DEBUG

/* use 0 for production, 1 for verification, >2 for debug */
#ifdef BANYAN_DEBUG
static int banyan_debug = BANYAN_DEBUG;
#endif

#define STATION_ADDRESS_HIGH(dev) (((dev)->dev_addr[0] << 8) | \
				   ((dev)->dev_addr[1]))
#define STATION_ADDRESS_LOW(dev)  (((dev)->dev_addr[2] << 24) | \
				   ((dev)->dev_addr[3] << 16) | \
				   ((dev)->dev_addr[4] << 8)  | \
				   ((dev)->dev_addr[5]))

#define MII_CLOCK 1250000 					/* no more than 2.5MHz */
#define NINTFC		2 						/* number of interface */
static char mac[18] = "08:00:06:05:40:01"; 

MODULE_PARM(mac, "c18");
MODULE_PARM_DESC(mac0, "MAC address for RC32355 ethernet");

static struct banyan_if_t 
{
	struct net_device *dev;
  	char* mac_str;
  	u32 iobase;
  	int rx_dma_irq;
  	int tx_dma_irq;
  	int rx_ovr_irq;
  	int tx_und_irq;		
} banyan_iflist[] = 
{
	{NULL, mac, RC32355_ETH_BASE, ETH_DMA_RX_IRQ, ETH_DMA_TX_IRQ, ETH_RX_OVR_IRQ, ETH_TX_UND_IRQ},
	{NULL, NULL, 0, 0}
};


/* Index to functions, as function prototypes. */
int banyan_init_module(void);
static void banyan_cleanup_module(void);
int banyan_probe(int port_num);

static int parse_mac_addr(struct net_device *dev, char* macstr)
{
  	int i, j;
  	unsigned char result, value;
	
  	for (i=0; i<6; i++) 
	{
    		result = 0;
    		if (i != 5 && *(macstr+2) != ':') 
		{
      			err(__FILE__ "invalid mac address format: %d %c\n", i, *(macstr+2));
      			return -EINVAL;
    		}
    
	    	for (j=0; j<2; j++) 
		{
	      		if (isxdigit(*macstr) && (value = isdigit(*macstr) ? *macstr-'0' : toupper(*macstr)-'A'+10) < 16) 
			{
				result = result*16 + value;
				macstr++;
	      		} 
			else 
			{
				err(__FILE__ "invalid mac address character: %c\n", *macstr);
				return -EINVAL;
	      		}
	    	}
    
    		macstr++;
    		dev->dev_addr[i] = result;
  	}
  
  	return 0;
}


static inline void banyan_halt_tx(struct net_device *dev)
{
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  	if (rc32355_halt_dma(lp->tx_dma_regs))
    		err(__FUNCTION__ ": timeout!\n");
}
static inline void banyan_halt_rx(struct net_device *dev)
{
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  	if (rc32355_halt_dma(lp->rx_dma_regs))
    		err(__FUNCTION__ ": timeout!\n");
}

static inline void banyan_start_tx(struct banyan_local *lp, volatile rc32355_dma_desc_t* td)
{
  	rc32355_start_dma(lp->tx_dma_regs, virt_to_phys(td));
}
static inline void banyan_start_rx(struct banyan_local *lp, volatile rc32355_dma_desc_t* rd)
{
  	rc32355_start_dma(lp->rx_dma_regs, virt_to_phys(rd));
}

static inline void banyan_chain_tx(struct banyan_local *lp, volatile rc32355_dma_desc_t* td)
{
  	rc32355_chain_dma(lp->tx_dma_regs, virt_to_phys(td));
}
static inline void banyan_chain_rx(struct banyan_local *lp, volatile rc32355_dma_desc_t* rd)
{
  	rc32355_chain_dma(lp->rx_dma_regs, virt_to_phys(rd));
}

#ifdef BANYAN_PROC_DEBUG
static int banyan_read_proc(char *buf, char **start, off_t fpos,
			   int length, int *eof, void *data)
{
  	struct net_device *dev = (struct net_device *)data;
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  	int len = 0;
  
  	/* print out header */
  	len += sprintf(buf + len, "\n\tRC32355 Ethernet Debug\n\n");
  
  	len += sprintf (buf + len,
		  "DMA halt count      = %10d, total pkt cnt = %10d\n",
		  lp->dma_halt_cnt, lp->halt_tx_count);
  	len += sprintf (buf + len,
		  "DMA run count       = %10d, total pkt cnt = %10d\n",
		  lp->dma_run_cnt, lp->run_tx_count);
  	len += sprintf (buf + len,
		  "DMA race count      = %10d, total pkt cnt = %10d\n",
		  lp->dma_race_cnt, lp->race_tx_count);
  	len += sprintf (buf + len,
		  "DMA collision count = %10d, total pkt cnt = %10d\n",
		  lp->dma_collide_cnt, lp->collide_tx_count);

  	if (fpos >= len) 
	{
    		*start = buf;
    		*eof = 1;
    		return 0;
  	}
  	*start = buf + fpos;
  	if ((len -= fpos) > length)
    		return length;
	
  	*eof = 1;
  	return len;
  
}
#endif


/*
 * Restart the BANYAN ethernet controller. Hold a spin lock
 * before calling.
 */
static int banyan_restart(struct net_device *dev)
{
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  
  	/*
   	* Disable interrupts
   	*/
  	disable_irq(lp->rx_irq);
  	disable_irq(lp->tx_irq);
  	disable_irq(lp->ovr_irq);
    	disable_irq(lp->und_irq);

  	/* Mask F bit in Tx DMA */
  	writel(readl(&lp->tx_dma_regs->dmasm) | DMAS_F, &lp->tx_dma_regs->dmasm);
    	/* Mask D bit in Rx DMA */
  	writel(readl(&lp->rx_dma_regs->dmasm) | DMAS_D, &lp->rx_dma_regs->dmasm);

	banyan_init(dev);
	banyan_multicast_list(dev);

  	enable_irq(lp->und_irq);
  	enable_irq(lp->ovr_irq);
  	enable_irq(lp->tx_irq);
  	enable_irq(lp->rx_irq);
  
  	return 0;
}

int banyan_init_module(void)
{
  	int i, retval=0;
  
  	for (i = 0; banyan_iflist[i].iobase; i++) 
	{ 
    		retval |= banyan_probe(i);
  	} 
  
  	return retval;
}

int banyan_probe(int port_num)
{
  	struct banyan_local *lp = NULL;
  	struct banyan_if_t *bif = NULL;
  	struct net_device *dev = NULL;
  	int i, retval;

	bif = &banyan_iflist[port_num];

	request_region(bif->iobase, 0x24C, "BANYAN");
	
  	/* Allocate a new 'dev' if needed */
  	dev = init_etherdev(0, sizeof(struct banyan_local));
  	bif->dev = dev;
  
	info("RC32355 ethernet found at 0x%08x\n", bif->iobase);
  
  	/* Fill in the 'dev' fields. */
  	dev->base_addr = bif->iobase;
	/* just use the rx dma irq */
  	dev->irq = bif->rx_dma_irq; 	
  
  	if ((retval = parse_mac_addr(dev, bif->mac_str))) 
	{
    		err(__FUNCTION__ ": MAC address parse failed\n");
    		retval = -EINVAL;
    		goto probe1_err_out;
  	}
  
  	info("HW Address ");
  	for (i = 0; i < 6; i++) 
	{
    		printk("%2.2x", dev->dev_addr[i]);
    		if (i<5)
      			printk(":");
  	}
  	printk("\n");
  
  	info("Rx IRQ %d, Tx IRQ %d\n", bif->rx_dma_irq, bif->tx_dma_irq);
  
  	/* Initialize the device structure. */
  	if (dev->priv == NULL) 
	{
    		lp = (struct banyan_local *)kmalloc(sizeof(*lp), GFP_KERNEL);
   		memset(lp, 0, sizeof(struct banyan_local));
  	} 
	else 
	{
    		lp = (struct banyan_local *)dev->priv;
  	}
  
  	dev->priv = lp;
  
  	lp->rx_irq = bif->rx_dma_irq;
  	lp->tx_irq = bif->tx_dma_irq;
    	lp->ovr_irq = bif->rx_ovr_irq;
  	lp->und_irq = bif->tx_und_irq;

  	lp->eth_regs = ioremap_nocache(bif->iobase, sizeof(*lp->eth_regs));
  
  	if (!lp->eth_regs) 
	{
    		err("Can't remap eth registers\n");
    		retval = -ENXIO;
    		goto probe1_err_out;
  	}
  
	lp->rx_dma_regs =
		ioremap_nocache(RC32355_DMA_BASE + 9*DMA_CHAN_OFFSET,
				sizeof(rc32355_dma_ch_t));
	if (!lp->rx_dma_regs) {
		err("Can't remap Rx DMA registers\n");
		retval = -ENXIO;
		goto probe1_err_out;
	}
		
	lp->tx_dma_regs =
		ioremap_nocache(RC32355_DMA_BASE + 10*DMA_CHAN_OFFSET,
				sizeof(rc32355_dma_ch_t));
	if (!lp->tx_dma_regs) {
		err("Can't remap Tx DMA registers\n");
		retval = -ENXIO;
		goto probe1_err_out;
	}
  
  	lp->td_ring = (rc32355_dma_desc_t*)kmalloc(TD_RING_SIZE + RD_RING_SIZE, GFP_KERNEL);
  	if (!lp->td_ring) 
	{
    		err("Can't allocate descriptors\n");
    		retval = -ENOMEM;
    		goto probe1_err_out;
  	}

  	dma_cache_inv((unsigned long)(lp->td_ring), TD_RING_SIZE + RD_RING_SIZE);
  
  	/* now convert TD_RING pointer to KSEG1 */
  	lp->td_ring = (rc32355_dma_desc_t *)KSEG1ADDR(lp->td_ring);
  	lp->rd_ring = &lp->td_ring[BANYAN_NUM_TDS];

  	spin_lock_init(&lp->lock);
  
  	dev->open = banyan_open;
  	dev->stop = banyan_close;
  	dev->hard_start_xmit = banyan_send_packet;
  	dev->get_stats	= banyan_get_stats;
  	dev->set_multicast_list = &banyan_multicast_list;
  	dev->tx_timeout = banyan_tx_timeout;
  	dev->watchdog_timeo = BANYAN_TX_TIMEOUT;
  
#ifdef BANYAN_PROC_DEBUG
  	lp->ps = create_proc_read_entry ("net/rc32355", 0, NULL,
				   banyan_read_proc, dev);
#endif
  
	/* Fill in the fields of the device structure with ethernet values. */
  	ether_setup(dev);
  	return 0;
  
 probe1_err_out:
  	banyan_cleanup_module();
  	err(__FUNCTION__ " failed.  Returns %d\n", retval);
  	return retval;
	
}


/*
 * Open/initialize the BANYAN controller.
 *
 * This routine should set everything up anew at each open, even
 *  registers that "should" only need to be set once at boot, so that
 *  there is non-reboot way to recover if something goes wrong.
 */
static int banyan_open(struct net_device *dev)
{
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  
  	MOD_INC_USE_COUNT;
	
  	/*
   	* Initialize
   	*/
  	if (banyan_init(dev)) 
  	{
    		err("Erroe: cannot open the Ethernet device\n");
    		return -EAGAIN;
  	}

	/*
	  * Install the interrupt handler that handles the dma Done and
	  * Finished Events.
	  */
    	if (request_irq(lp->rx_irq, &banyan_rx_dma_interrupt,
		  SA_SHIRQ | SA_INTERRUPT,
		  "banyan ethernet Rx", dev)) 
	{
    		err(__FUNCTION__ ": unable to get Rx DMA IRQ %d\n", lp->rx_irq);
    		MOD_DEC_USE_COUNT;
    		return -EAGAIN;
  	}
  	if (request_irq(lp->tx_irq, &banyan_tx_dma_interrupt,
		  SA_SHIRQ | SA_INTERRUPT,
		  "banyan ethernet Tx", dev)) 
	{
    		err(__FUNCTION__ ": unable to get Tx DMA IRQ %d\n", lp->tx_irq);
    		free_irq(lp->rx_irq, dev);
    		MOD_DEC_USE_COUNT;
    		return -EAGAIN;
  	}

  	/* Install handler for overrun error. */
  	if (request_irq(lp->ovr_irq, &banyan_ovr_interrupt,
		  SA_SHIRQ | SA_INTERRUPT,
		  "Ethernet Overflow", dev)) 
	{
    		err(__FUNCTION__ ": unable to get OVR IRQ %d\n",	lp->ovr_irq);
    		free_irq(lp->rx_irq, dev);
    		free_irq(lp->tx_irq, dev);
    		MOD_DEC_USE_COUNT;
    		return -EAGAIN;
  	}
	
  	/* Install handler for underflow error. */
	if (request_irq(lp->und_irq, &banyan_und_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"Ethernet Underflow", dev)) {
		err(__FUNCTION__ ": unable to get UND IRQ %d\n",
		    lp->und_irq);
		free_irq(lp->rx_irq, dev);
		free_irq(lp->tx_irq, dev);
		free_irq(lp->ovr_irq, dev);
		MOD_DEC_USE_COUNT;
		return -EAGAIN;
	}

  	return 0;
}



/*
 * Close the BANYAN device
 */
static int banyan_close(struct net_device *dev)
{
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  	u32 tmp;
  
	/*
	 * Disable interrupts
	 */
  	disable_irq(lp->rx_irq);
  	disable_irq(lp->tx_irq);
	disable_irq(lp->ovr_irq);
    	disable_irq(lp->und_irq);

  	tmp = readl(&lp->tx_dma_regs->dmasm);
  	tmp = tmp | DMAS_F | DMAS_E;
  	writel(tmp, &lp->tx_dma_regs->dmasm);

  	tmp = readl(&lp->rx_dma_regs->dmasm);
	tmp = tmp | DMAS_D | DMAS_H | DMAS_E;
	writel(tmp, &lp->rx_dma_regs->dmasm);
  
  	free_irq(lp->rx_irq, dev);
  	free_irq(lp->tx_irq, dev);
  	free_irq(lp->ovr_irq, dev);
    	free_irq(lp->und_irq, dev);	

  	MOD_DEC_USE_COUNT;
  	return 0;
}


/* transmit packet */
static int banyan_send_packet(struct sk_buff *skb, struct net_device *dev)
{
 	struct banyan_local	*lp = (struct banyan_local *)dev->priv;
  	unsigned long 	flags;
  	u32			length;
	volatile rc32355_dma_desc_t * td;

	spin_lock_irqsave(&lp->lock, flags);

	td = &lp->td_ring[lp->tx_chain_tail];

 	//stop queue when full, drop pkts if queue already full
  	if(lp->tx_count >= (BANYAN_NUM_TDS - 2))
  	{
		lp->tx_full = 1;

	  	if(lp->tx_count == (BANYAN_NUM_TDS - 2))
	  	{
			//this pkt is about to fill the queue
			//err("Tx Ring now full, queue stopped.\n");			
			netif_stop_queue(dev);
	  	}else{
			//this pkt cannot be added to the full queue
			//err("Tx ring full, packet dropped\n");
			lp->stats.tx_dropped++;
			dev_kfree_skb_any(skb);
			spin_unlock_irqrestore(&lp->lock, flags);
			return 1;
	  	}	   
  	}	   
  	lp->tx_count ++;
  
  	/* make sure payload gets written to memory */
  	dma_cache_wback_inv((unsigned long)skb->data, skb->len);
  
 	if (lp->tx_skb[lp->tx_chain_tail] != NULL)
  		dev_kfree_skb_any(lp->tx_skb[lp->tx_chain_tail]);

	lp->tx_skb[lp->tx_chain_tail] = skb;
  
  	length = skb->len; 

  	//Setup the transmit descriptor.
  	td->curr_addr = virt_to_phys(skb->data);

	//hkou: Using the NDPTR to handl the DMA Race Condition
   	if(lp->tx_dma_regs->dmandptr == 0) 
  	{
		if( lp->tx_chain_status == empty ) 
		{
			td->cmdstat = DMA_COUNT(length) |DMADESC_COF |DMADESC_IOF; 
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & BANYAN_TDS_MASK;
			lp->tx_dma_regs->dmandptr  = virt_to_phys(&lp->td_ring[lp->tx_chain_head]);
			lp->tx_chain_head = lp->tx_chain_tail; 
		}
		else
		{
			lp->td_ring[(lp->tx_chain_tail-1)& BANYAN_TDS_MASK].cmdstat &=  ~(DMADESC_COF); 	
			lp->td_ring[(lp->tx_chain_tail-1)& BANYAN_TDS_MASK].link =  virt_to_phys(td);
			td->cmdstat = DMA_COUNT(length) |DMADESC_COF|DMADESC_IOF;
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & BANYAN_TDS_MASK;
			lp->tx_chain_status = empty;
			lp->tx_dma_regs->dmandptr  = virt_to_phys(&lp->td_ring[lp->tx_chain_head]);
			lp->tx_chain_head = lp->tx_chain_tail; 
		}
   	}
	else
	{
		if( lp->tx_chain_status == empty ) 
		{
			td->cmdstat = DMA_COUNT(length) |DMADESC_COF |DMADESC_IOF;
			lp->td_ring[(lp->tx_chain_tail-1)& BANYAN_TDS_MASK].link =  virt_to_phys(td);	
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & BANYAN_TDS_MASK;
			lp->tx_chain_status = filled;		
		}
		else
		{
			lp->td_ring[(lp->tx_chain_tail-1)& BANYAN_TDS_MASK].cmdstat &=  ~(DMADESC_COF); 	
			lp->td_ring[(lp->tx_chain_tail-1)& BANYAN_TDS_MASK].link =  virt_to_phys(td);	
			td->cmdstat = DMA_COUNT(length) |DMADESC_COF |DMADESC_IOF; 
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & BANYAN_TDS_MASK;
		}
	}	   

	dev->trans_start = jiffies;
  
     	spin_unlock_irqrestore(&lp->lock, flags);

  
  	return 0;
}


/* Ethernet Rx Overflow interrupt */
static void banyan_ovr_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
  	struct net_device *dev = (struct net_device *)dev_id;
  	struct banyan_local *lp;
  	unsigned int i;
	
  	if (dev == NULL) 
	{
    		err(__FUNCTION__ ": irq %d for unknown device.\n", irq);
		return;
  	}
  
  	lp = (struct banyan_local *)dev->priv;
	netif_stop_queue(dev);

	//err("Rx overflow\n");
  
  	spin_lock(&lp->lock);

  	//clear OVR int (sticky bit)
  	i = readl(&lp->eth_regs->ethintfc);
  	i &= ~ETHERDMA_IN_OVR;
  	writel(i, &lp->eth_regs->ethintfc);

  	// Restart interface
  	banyan_restart(dev);	  
	
  	spin_unlock(&lp->lock);
  
}

/* Ethernet Tx Underflow interrupt */
static void banyan_und_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct banyan_local *lp;
	unsigned int i;
	
	if (dev == NULL) {
		err(__FUNCTION__ ": irq %d for unknown device.\n", irq);
		return;
	}
	
	netif_stop_queue(dev);	
	lp = (struct banyan_local *)dev->priv;

	//err("Tx underflow - i/f reset\n");
	spin_lock(&lp->lock);

	i = readl(&lp->eth_regs->ethintfc);
  	i &= ~ETHERDMA_OUT_UND;
	writel(i, &lp->eth_regs->ethintfc);

	/* Restart interface */
	banyan_restart(dev);     
	
	spin_unlock(&lp->lock);


}

/* Ethernet Rx DMA interrupt */
static void
banyan_rx_dma_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
  	struct net_device *dev = (struct net_device *)dev_id;
  	struct banyan_local *lp;
  	u32 dmas;
	
  	if (dev == NULL) 
	{
    		err(__FUNCTION__ ": irq %d for unknown device.\n", irq);
		return;
  	}
  
  	lp = (struct banyan_local *)dev->priv;

  	spin_lock(&lp->lock);
  
  	dmas = readl(&lp->rx_dma_regs->dmas);
  	writel(~dmas, &lp->rx_dma_regs->dmas);
  
  	if(dmas & (DMAS_D|DMAS_E))
  	{
	  	if(dmas & DMAS_E)
	  	{
	    		err(__FUNCTION__ ": DMAS_E\n");
	    		lp->stats.rx_errors++;
	  	}
	
	    	banyan_rx(dev);
  	}

  	spin_unlock(&lp->lock);
}

/* Ethernet Tx DMA interrupt */
static void
banyan_tx_dma_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
  	struct net_device	*dev = (struct net_device *)dev_id;
  	struct banyan_local	*lp;
  	u32			dmas;

    	if (dev == NULL) 
	{
    		err(__FUNCTION__ ": irq %d for unknown device.\n", irq);
		return;
  	}

   	lp = (struct banyan_local *)dev->priv;
  
  	spin_lock(&lp->lock);
 
  	dmas = readl(&lp->tx_dma_regs->dmas);
  	writel(~dmas, &lp->tx_dma_regs->dmas);

  	if (dmas & DMAS_E)
    		err(__FUNCTION__ ": DMA error\n");
  	if (dmas & DMAS_F)
    		banyan_tx(dev);

  	spin_unlock(&lp->lock);
}


static void
banyan_tx(struct net_device *dev)
{
  	struct banyan_local* lp = (struct banyan_local *)dev->priv;
  	volatile rc32355_dma_desc_t* td = &lp->td_ring[lp->tx_next_done];
    	u32			dmas, devcs;
  
  	dmas = readl(&lp->tx_dma_regs->dmas);

  	//process all desc that are done
  	while(IS_DMA_FINISHED(td->cmdstat)) 
  	{

  		if (	lp->tx_full == 1)
  		{
			netif_wake_queue(dev);
			lp->tx_full = 0;
		}

		devcs = lp->td_ring[lp->tx_next_done].devcs;    
		if ((devcs & (ETHERDMA_OUT_FD | ETHERDMA_OUT_LD)) != (ETHERDMA_OUT_FD | ETHERDMA_OUT_LD)) 
		{
	      		lp->stats.tx_errors++;
	      		lp->stats.tx_dropped++;				

	      		/* should never happen */
	      		dbg(1, __FUNCTION__ ": split tx ignored\n");
	    	} 
		else if (IS_TX_TOK(devcs)) 
		{
	      		/* transmit OK */
	      		lp->stats.tx_packets++;
	    	} 
		else 
		{

	      		lp->stats.tx_dropped++;				
			  
		      	/* underflow */
		  	if (IS_TX_UND_ERR(devcs)) 
				lp->stats.tx_fifo_errors++;

		      	/* oversized frame */
		      	if (IS_TX_OF_ERR(devcs))
				lp->stats.tx_aborted_errors++;

		      	/* excessive deferrals */
		      	if (IS_TX_ED_ERR(devcs))
				lp->stats.tx_carrier_errors++;

		      	/* collisions: medium busy */
		      	if (IS_TX_EC_ERR(devcs))
				lp->stats.collisions++;

		      	/* late collision */
		      	if (IS_TX_LC_ERR(devcs))
				lp->stats.tx_window_errors++;

    		}

	    	/* We must always free the original skb */
	    	if (lp->tx_skb[lp->tx_next_done] != NULL) 
		{
	      		dev_kfree_skb_irq(lp->tx_skb[lp->tx_next_done]);
	      		lp->tx_skb[lp->tx_next_done] = NULL;
	    	}

		lp->td_ring[lp->tx_next_done].cmdstat = 0;
	    	lp->tx_count --;

	    	/* go on to next transmission */
	    	lp->tx_next_done = (lp->tx_next_done + 1) & BANYAN_TDS_MASK;
	    	td = &lp->td_ring[lp->tx_next_done];
			
  	}
	
}


/*
 * We have a good packet(s), get it/them out of the buffers.
 *
 * 	 This driver works by creating (once) a circular list of receiver
 *	 DMA descriptors that will be used serially by the Cedar.
 *	 Because the descriptors are never unlinked from the list _they
 *	 are always live_.  We are counting on Linux (and the chosen number
 *	 of buffers) to keep ahead of the hardware otherwise the same
 *	 descriptor might be used for more than one reception.
 */
static void
banyan_rx(struct net_device *dev)
{
  	struct banyan_local* lp = (struct banyan_local *)dev->priv;
  	volatile rc32355_dma_desc_t*  rd = &lp->rd_ring[lp->rx_next_done];
  	struct sk_buff *skb, *skb_new;
	struct sk_buff 	 *skb_new_1, *skb_new_2;
  	u8* pkt_buf;
  	u32 devcs, count, pkt_len;
  
  	/* keep going while we have received into more descriptors */
  	while (IS_DMA_DONE(rd->cmdstat)) 
    	{
	  
      		/* init the var. used for the later operations within the while loop */
      		skb_new = NULL;
      		devcs = rd->devcs;
      		pkt_len = RCVPKT_LENGTH(devcs);
      		skb = lp->rx_skb[lp->rx_next_done];

      		count = BANYAN_RBSIZE - (u32)DMA_COUNT(rd->cmdstat);

	     	if( count != pkt_len) 
		{	
	          /*
	             * Due to a bug in banyan processor, the packet length
	             * given by devcs field and count field sometimes differ.
	             * If that is the case, report Error.
	           */				
	        	lp->stats.rx_errors++;
			lp->stats.rx_dropped++;
			  
	      	}
      		else if (count < 64) 
		{
	  		lp->stats.rx_errors++;
			lp->stats.rx_dropped++;
		} 
      		else if ((devcs & ( ETHERDMA_IN_LD)) !=	ETHERDMA_IN_LD)
		{
			  /* Check that this is a whole packet */
			  /* WARNING: DMA_FD bit incorrectly set in Cedar (errata ref #077) */
			lp->stats.rx_errors++;
			lp->stats.rx_dropped++;
		} 
      		else if (devcs & ETHERDMA_IN_ROK) 
		{	
		
			  /* must be the (first and) last descriptor then */
	  		pkt_buf = (u8*)lp->rx_skb[lp->rx_next_done]->data;
	  
			  /* invalidate the cache before copying the buffer */
	  		dma_cache_inv((unsigned long)pkt_buf, (pkt_len-4));
	  
	    		/* Malloc up new buffer. */					  
	  		skb_new = dev_alloc_skb(BANYAN_RBSIZE + 2);							

	  		if (skb_new != NULL)
	    		{

				skb_new->dev = dev;
				skb_reserve(skb_new, 2);	/* 16 bit align */
	      			//skb_new->ip_summed = CHECKSUM_HW;
				
		  		/* Make room */
		  		skb_put(skb_new, (pkt_len-4));		    
				
				eth_copy_and_sum(skb_new, skb->data, pkt_len-4, 0);
				
				skb_new->protocol = eth_type_trans(skb_new, dev);

				/* pass the packet to upper layers */
				netif_rx(skb_new);
				
				dev->last_rx = jiffies;
				lp->stats.rx_packets++;
	      			lp->stats.rx_bytes += (pkt_len-4);
								
				if (IS_RCV_MP(devcs))
					lp->stats.multicast++;
	    		}
		  	else
		    	{
		      		//err("no memory, dropping rx packet.\n");
			  	lp->stats.rx_errors++;				
				lp->stats.rx_dropped++;				
		    	}

		}				
      		else 
		{

			/* This should only happen if we enable accepting broken packets */
	  		lp->stats.rx_errors++;
			lp->stats.rx_dropped++;

	  		/* added statistics counters */
	  		if (IS_RCV_CRC_ERR(devcs)) 
	    		{			  			
	      			dbg(2, "RX CRC error\n");
	      			lp->stats.rx_crc_errors++;
	    		}
	  		else if (IS_RCV_LOR_ERR(devcs)) 
	    		{			
				dbg(2, "RX LOR error\n");
	      			lp->stats.rx_length_errors++;
	    		}
	  		else if (IS_RCV_LE_ERR(devcs)) 
	    		{			
	      			dbg(2, "RX LE error\n");
	      			lp->stats.rx_length_errors++;
	    		}
	  		else if (IS_RCV_OVR_ERR(devcs)) 
	    		{
			      /*
			       * The overflow errors are handled through
			       * an interrupt handler.
			       */

	      			lp->stats.rx_over_errors++;
	    		}
	  		else if (IS_RCV_CV_ERR(devcs)) 
		    	{			
		     		 /* code violation */
		      		dbg(2, "RX CV error\n");
		  		lp->stats.rx_errors++;					
		    	}
	  		else if (IS_RCV_CES_ERR(devcs)) 
	    		{
	      			dbg(2, "RX Preamble error\n");
		  		lp->stats.rx_errors++;					
	    		}
		}
      
		rd->cmdstat = DMA_COUNT(BANYAN_RBSIZE) |DMADESC_COD |DMADESC_IOD;

	      /*
	       * clear the bits that let us see whether this
	       * descriptor has been used or not & reset reception
	       * length.
	       */		 
      		rd->devcs = 0;
	  
      		/* restore descriptor's curr_addr */
      		if (skb_new == NULL)
      		{
			rd->curr_addr = virt_to_phys(skb->data);
      		}  
      
		lp->rx_next_done = (lp->rx_next_done + 1) & BANYAN_RDS_MASK;
      		rd = &lp->rd_ring[lp->rx_next_done];

	      /*
	       * we'll deal with all possible interrupts up to the last
	       * used descriptor - so cancel any interrupts that may have
	       * arrisen while we've been processing.
	       */
      		writel(0, &lp->rx_dma_regs->dmas);
		  
    }	

}

/*
 * Get the current statistics.
 * This may be called with the device open or closed.
 */
static struct net_device_stats *
banyan_get_stats(struct net_device *dev)
{
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  
	return &lp->stats;
}


/*
 * Set or clear the multicast filter for this adaptor.
 */
static void
banyan_multicast_list(struct net_device *dev)
{   	
	/* changed to listen to broadcasts always and to treat	*/
  	/*	   IFF bits independantly	*/
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  	unsigned long flags;
  	u32 recognise = ETHERARC_AB; 			/* always accept broadcasts */
  
  	if (dev->flags & IFF_PROMISC) 				/* set promiscuous mode */
    		recognise |= ETHERARC_PRO;
  
  	if ((dev->flags & IFF_ALLMULTI) || (dev->mc_count > 15))
    		recognise |= ETHERARC_AM;		  	/* all multicast & bcast */
  
  	else if (dev->mc_count > 0) 
	{
    		dbg(2, __FUNCTION__ ": mc_count %d\n", dev->mc_count);
    
    		recognise |= ETHERARC_AM;		  	/* for the time being */
  	}
  
  	spin_lock_irqsave(&lp->lock, flags);
	
  	writel(recognise, &lp->eth_regs->etharc);
	
  	spin_unlock_irqrestore(&lp->lock, flags);
	
}


static void
banyan_tx_timeout(struct net_device *dev)
{
  	struct banyan_local *lp = (struct banyan_local *)dev->priv;
  	unsigned long flags;
  
  	spin_lock_irqsave(&lp->lock, flags);
  	banyan_restart(dev);
  	spin_unlock_irqrestore(&lp->lock, flags);

}


/*
 * Initialize the BANYAN ethernet controller.
 */
static int banyan_init(struct net_device *dev)
{
	struct banyan_local *lp = (struct banyan_local *)dev->priv;
	int i, j;

	/* Disable DMA */       
	banyan_halt_tx(dev);
	banyan_halt_rx(dev); 
	
	/* reset ethernet logic */ 
	writel(0, &lp->eth_regs->ethintfc);
	while((readl(&lp->eth_regs->ethintfc) & ETHERINTFC_RIP));
	/* Enable Ethernet Interface */ 
	writel(ETHERINTFC_EN, &lp->eth_regs->ethintfc); 


	/* Initialize the transmit Descriptors */
	for (i = 0; i < BANYAN_NUM_TDS; i++) {
    		lp->td_ring[i].cmdstat = DMADESC_IOF;
    		lp->td_ring[i].devcs = ETHERDMA_OUT_FD | ETHERDMA_OUT_LD;
    		lp->td_ring[i].curr_addr = 0;
    		lp->td_ring[i].link = 0;
    		if (lp->tx_skb[i] != NULL) 
		{
      			/* free dangling skb */
      			dev_kfree_skb_any(lp->tx_skb[i]);
      			lp->tx_skb[i] = NULL;
    		}
	}
    	lp->tx_next_done = lp->tx_chain_head = lp->tx_chain_tail = 
    		lp->tx_count = lp->tx_full = 0;
	lp->	tx_chain_status = empty;
	
	/*
	 * Initialize the receive descriptors so that they
	 * become a circular linked list, ie. let the last
	 * descriptor point to the first again.
	 */
  	for (i=0; i<BANYAN_NUM_RDS; i++) 
  	{
		struct sk_buff *skb = lp->rx_skb[i];

		if (lp->rx_skb[i] == NULL) 
		{
			skb = dev_alloc_skb(BANYAN_RBSIZE + 2);
			if (skb == NULL)
			{
		 		//err("No memory in the system\n");
				for (j = 0; j < BANYAN_NUM_RDS; j ++)
					if (lp->rx_skb[j] != NULL) 
						dev_kfree_skb_any(lp->rx_skb[j]);
				
	              	return 1;
			}
			else
			{
				skb->dev = dev;
				skb_reserve(skb, 2);
				/*skb->ip_summed = CHECKSUM_HW;*/				
				lp->rx_skb[i] = skb;
		             	lp->rd_ring[i].curr_addr = virt_to_phys(skb->data); 
				
			}
		}
		lp->rd_ring[i].cmdstat =	DMADESC_IOD | DMA_COUNT(BANYAN_RBSIZE);
		lp->rd_ring[i].devcs = 0;
              lp->rd_ring[i].curr_addr = virt_to_phys(skb->data);
		lp->rd_ring[i].link = virt_to_phys(&lp->rd_ring[i+1]);

  	}
    	/* loop back */
  	lp->rd_ring[BANYAN_NUM_RDS-1].link = virt_to_phys(&lp->rd_ring[0]);
    	lp->rx_next_done   = 0;
  	lp->rd_ring[BANYAN_NUM_RDS-1].cmdstat |= DMADESC_COD;
  	lp->rx_chain_head = 0;
  	lp->rx_chain_tail = 0;
  	lp->rx_chain_status = empty;


	writel(0, &lp->rx_dma_regs->dmas);
	/* Start Rx DMA */
	banyan_start_rx(lp, &lp->rd_ring[0]);
	
  	/* Enable F bit in Tx DMA */
  	writel(readl(&lp->tx_dma_regs->dmasm) & ~(DMAS_F | DMAS_E), &lp->tx_dma_regs->dmasm); 
    	/* Enable D bit in Rx DMA */
  	writel(readl(&lp->rx_dma_regs->dmasm) & ~(DMAS_D | DMAS_E), &lp->rx_dma_regs->dmasm); 


	/* Accept only packets destined for this Ethernet device address */
	writel(ETHERARC_AB, &lp->eth_regs->etharc); 
 
	/* Set all Ether station address registers to their initial values */ 
	writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal0); 
	writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah0);
	
	writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal1); 
	writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah1);
	
	writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal2); 
	writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah2);
	
	writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal3); 
	writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah3); 
 
	/* Frame Length Checking, Pad Enable, CRC Enable, Full Duplex set */ 
	writel(ETHERMAC2_FLC | ETHERMAC2_PE | ETHERMAC2_CEN | ETHERMAC2_FD,
	       &lp->eth_regs->ethmac2);  

	/* Back to back inter-packet-gap */ 
	writel(0x15, &lp->eth_regs->ethipgt); 
	/* Non - Back to back inter-packet-gap */ 
	writel(0x12, &lp->eth_regs->ethipgr); 
     
	/* Management Clock Prescaler Divisor */
	/* changed from clock independent setting:
	   writel(40, &lp->eth_regs->ethmcp); */
	writel(((IDT_BUS_FREQ * 1000 * 1000)/MII_CLOCK+1) & ~1, &lp->eth_regs->ethmcp);

	/* don't transmit until fifo contains 48b */
	writel(48, &lp->eth_regs->ethfifott);

	writel(ETHERMAC1_RE, &lp->eth_regs->ethmac1);

	netif_start_queue(dev);

	/* Enable Ethernet Interface */ 
	writel(ETHERINTFC_EN, &lp->eth_regs->ethintfc); 


	return 0; 

}


static void banyan_cleanup_module(void)
{
  	int i;
  
  	for (i = 0; banyan_iflist[i].iobase; i++) 
	{
    		struct banyan_if_t * bif = &banyan_iflist[i];
    		if (bif->dev != NULL) 
		{
      			struct banyan_local *lp = (struct banyan_local *)bif->dev->priv;
      			if (lp != NULL) 
			{
				if (lp->eth_regs)
					iounmap((void*)lp->eth_regs);
				if (lp->rx_dma_regs)
					iounmap((void*)lp->rx_dma_regs);
				if (lp->tx_dma_regs)
					iounmap((void*)lp->tx_dma_regs);
				if (lp->td_ring)
	  				kfree((void*)KSEG0ADDR(lp->td_ring));

#ifdef BANYAN_PROC_DEBUG
				if (lp->ps)
	  				remove_proc_entry("net/rc32355", NULL);
#endif
				kfree(lp);
      			}

      			unregister_netdev(bif->dev);
     			kfree(bif->dev);
      			release_region(bif->iobase, 0x24C);
    		}
  	}
}


#ifndef MODULE

static int __init banyan_setup(char *options)
{
  	/* no options yet */
  	return 1;
}

static int __init banyan_setup_ethaddr(char *options)
{
  	memcpy(mac, options, 17);
  	mac[17]= '\0';
  	return 1;
}

__setup("rc3255eth=", banyan_setup);

__setup("ethaddr=", banyan_setup_ethaddr);

#endif /* !MODULE */

module_init(banyan_init_module);
module_exit(banyan_cleanup_module);

