/**************************************************************************
 *
 *  BRIEF MODULE DESCRIPTION
 *     Driver for the IDT KORINA on-chip ethernet controller.
 *
 *  Copyright 2004 IDT Inc. (rischelp@idt.com)
 *         
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **************************************************************************
 * May 2004 rkt, neb
 *
 * Based on the driver developed by B. Maruthanayakam, H. Kou and others.
 *
 * Aug 2004 Sadik
 *
 * Added NAPI
 *
 **************************************************************************
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/proc_fs.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <linux/mii.h>
#include <linux/ethtool.h>
#include <asm/bootinfo.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/pgtable.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/uaccess.h>
#include "korina.h"

#define DRIVER_VERSION "(mar2904)"

#define DRIVER_NAME "korina Ethernet driver. " DRIVER_VERSION


#define STATION_ADDRESS_HIGH(dev) (((dev)->dev_addr[0] << 8) | \
			           ((dev)->dev_addr[1]))
#define STATION_ADDRESS_LOW(dev)  (((dev)->dev_addr[2] << 24) | \
				   ((dev)->dev_addr[3] << 16) | \
				   ((dev)->dev_addr[4] << 8)  | \
				   ((dev)->dev_addr[5]))

#define MII_CLOCK 1250000 				/* no more than 2.5MHz */
static char mac[18] = "08:00:06:05:40:01"; 
char *ifname = "eth0";

MODULE_PARM(mac, "c18");
MODULE_PARM_DESC(mac, "MAC address for KORINA ethernet0");

static struct korina_if_t {
	char *name;
	struct net_device *dev;
	char* mac_str;
	int weight;
	u32 iobase;
	u32 rxdmabase;
	u32 txdmabase;
	int rx_dma_irq;
	int tx_dma_irq;
	int rx_ovr_irq;
	int tx_und_irq;			
} korina_iflist[] = 
{
	{
		"eth0",      NULL,       mac, 
		64,
		ETH0_PhysicalAddress,
		ETH0_RX_DMA_ADDR,
		ETH0_TX_DMA_ADDR,
		ETH0_DMA_RX_IRQ,
		ETH0_DMA_TX_IRQ,
		ETH0_RX_OVR_IRQ,
		ETH0_TX_UND_IRQ
	}
};

static int korina_mdio_read(struct net_device *dev, int phy_id, int reg);
static int korina_mdio_write(struct net_device *dev, int phy_id, int reg, int value);
static int korina_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);
static int korina_setup_timer(struct net_device *dev);
static void korina_check_link(struct net_device *dev);
static void korina_link_timer_fn(unsigned long data);

static int parse_mac_addr(struct net_device *dev, char* macstr)
{
	int i, j;
	unsigned char result, value;
	
	for (i=0; i<6; i++) {
		result = 0;
		if (i != 5 && *(macstr+2) != ':') {
			ERR("invalid mac address format: %d %c\n",
			    i, *(macstr+2));
			return -EINVAL;
		}				
		for (j=0; j<2; j++) {
			if (isxdigit(*macstr) && (value = isdigit(*macstr) ? *macstr-'0' : 
						  toupper(*macstr)-'A'+10) < 16) {
				result = result*16 + value;
				macstr++;
			} 
			else {
				ERR("invalid mac address "
				    "character: %c\n", *macstr);
				return -EINVAL;
			}
		}
		
		macstr++; 
		dev->dev_addr[i] = result;
	}
	
	return 0;
}



static inline void korina_abort_tx(struct net_device *dev)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
	korina_abort_dma(dev, lp->tx_dma_regs);
	
}

static inline void korina_abort_rx(struct net_device *dev)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
	korina_abort_dma(dev, lp->rx_dma_regs);
	
}

static inline void korina_start_tx(struct korina_local *lp,  volatile DMAD_t td)
{
	rc32434_start_dma(lp->tx_dma_regs, CPHYSADDR(td));
}

static inline void korina_start_rx(struct korina_local *lp, volatile DMAD_t rd)
{
	rc32434_start_dma(lp->rx_dma_regs, CPHYSADDR(rd));
}

static inline void korina_chain_tx(struct korina_local *lp, volatile DMAD_t td)
{
	rc32434_chain_dma(lp->tx_dma_regs, CPHYSADDR(td));
}

static inline void korina_chain_rx(struct korina_local *lp, volatile DMAD_t rd)
{
	rc32434_chain_dma(lp->rx_dma_regs, CPHYSADDR(rd));
}

#ifdef KORINA_PROC_DEBUG
static int rc32434_read_proc(char *buf, char **start, off_t fpos,
			     int length, int *eof, void *data)
{
	struct net_device *dev = (struct net_device *)data;
	struct korina_local *lp = (struct korina_local *)dev->priv;
	int len = 0;
	
	/* print out header */
	len += sprintf(buf + len, "\n\tKORINA Ethernet Debug\n\n");
	len += sprintf (buf + len,
			"DMA halt count      = %10d, DMA run count = %10d\n",
			lp->dma_halt_cnt, lp->dma_run_cnt);
	
	if (fpos >= len) {
		*start = buf;
		*eof = 1;
		return 0;
	}
	*start = buf + fpos;
	
	if ((len -= fpos) > length) 
		return length;	
	*eof = 1;
	
	return len;
	
}
#endif


/*
 * Restart the KORINA ethernet controller. 
 */
static int korina_restart(struct net_device *dev)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;

	/*
	 * Disable interrupts
	 */
	disable_irq(lp->rx_irq);
	disable_irq(lp->tx_irq);
#ifdef	KORINA_REVISION
	disable_irq(lp->ovr_irq);
#endif	
	disable_irq(lp->und_irq);
	
	/* Mask F E bit in Tx DMA */
	rc32434_writel(rc32434_readl(&lp->tx_dma_regs->dmasm) | DMASM_f_m | DMASM_e_m, &lp->tx_dma_regs->dmasm);
	/* Mask D H E bit in Rx DMA */
	rc32434_writel(rc32434_readl(&lp->rx_dma_regs->dmasm) | DMASM_d_m | DMASM_h_m | DMASM_e_m, &lp->rx_dma_regs->dmasm);
	
	korina_init(dev);
	korina_multicast_list(dev);
	
	enable_irq(lp->und_irq);
#ifdef	KORINA_REVISION
	enable_irq(lp->ovr_irq);
#endif
	enable_irq(lp->tx_irq);
	enable_irq(lp->rx_irq);
	
	return 0;
}

int korina_init_module(void)
{
        switch (mips_machtype) {
        case MACH_ARUBA_MUSCAT:
            ifname = "eth0";
            break;
        case MACH_ARUBA_DELFRESCO:
            ifname = "bond0";
            break;
        default:
            return 1;
        }

	return korina_probe(0);
}

static int korina_probe(int port_num)
{
	struct korina_if_t *bif = &korina_iflist[port_num];
	struct korina_local *lp = NULL;
	struct net_device *dev = NULL;
	int i, retval,err;

#if 0	
	dev = alloc_etherdev(sizeof(struct korina_local));
#else
        dev = (struct net_device *)init_netdev(NULL, sizeof(struct korina_local),
                                               ifname, ether_setup);
#endif
	if(!dev) {
		ERR("korina_eth: alloc_etherdev failed\n");
		return -1;
	}
	
	SET_MODULE_OWNER(dev);
	bif->dev = dev;

	if ((retval = parse_mac_addr(dev, bif->mac_str))) {
		ERR("MAC address parse failed\n");
		return -1;
	}
	
	/* Initialize the device structure. */
	if (dev->priv == NULL) {
		lp = (struct korina_local *)kmalloc(sizeof(*lp), GFP_KERNEL);
		memset(lp, 0, sizeof(struct korina_local));
	} 
	else {
		lp = (struct korina_local *)dev->priv;
	}
	
	lp->rx_irq = bif->rx_dma_irq;
	lp->tx_irq = bif->tx_dma_irq;
	lp->ovr_irq = bif->rx_ovr_irq;
	lp->und_irq = bif->tx_und_irq;
	
	lp->eth_regs = ioremap_nocache(bif->iobase, sizeof(*lp->eth_regs));

	if (!lp->eth_regs) {
		ERR("Can't remap eth registers\n");
		retval = -ENXIO;
		goto probe_err_out;
	}
	
	lp->rx_dma_regs = ioremap_nocache(bif->rxdmabase, sizeof(struct DMA_Chan_s));
	
	if (!lp->rx_dma_regs) {
		ERR("Can't remap Rx DMA registers\n");
		retval = -ENXIO;
		goto probe_err_out;
	}
	lp->tx_dma_regs = ioremap_nocache(bif->txdmabase,sizeof(struct DMA_Chan_s));
	
	if (!lp->tx_dma_regs) {
		ERR("Can't remap Tx DMA registers\n");
		retval = -ENXIO;
		goto probe_err_out;
	}
	
#ifdef KORINA_PROC_DEBUG
	lp->ps = create_proc_read_entry (bif->name, 0, proc_net,
					 rc32434_read_proc, dev);
#endif
	
	lp->td_ring =	(DMAD_t)kmalloc(TD_RING_SIZE + RD_RING_SIZE, GFP_KERNEL);
	if (!lp->td_ring) {
		ERR("Can't allocate descriptors\n");
		retval = -ENOMEM;
		goto probe_err_out;
	}
	
	dma_cache_inv((unsigned long)(lp->td_ring), TD_RING_SIZE + RD_RING_SIZE);
	
	/* now convert TD_RING pointer to KSEG1 */
	lp->td_ring = (DMAD_t )KSEG1ADDR(lp->td_ring);
	lp->rd_ring = &lp->td_ring[KORINA_NUM_TDS];
	
	
	spin_lock_init(&lp->lock);
	
	dev->base_addr = bif->iobase | 0xA0000000;
	/* just use the rx dma irq */
	dev->irq = bif->rx_dma_irq; 
	
	dev->priv = lp;
	
	dev->open = korina_open;
	dev->stop = korina_close;
	dev->hard_start_xmit = korina_send_packet;
	dev->get_stats	= korina_get_stats;
	dev->set_multicast_list = &korina_multicast_list;
	dev->do_ioctl = korina_ioctl;
	dev->tx_timeout = korina_tx_timeout;
	dev->watchdog_timeo = KORINA_TX_TIMEOUT;

#ifdef CONFIG_IDT_USE_NAPI
	dev->poll = korina_poll;
	dev->weight = bif->weight;
	printk("Using NAPI with weight %d\n",dev->weight);
#else
	lp->rx_tasklet = kmalloc(sizeof(struct tasklet_struct), GFP_KERNEL);
	tasklet_init(lp->rx_tasklet, korina_rx_tasklet, (unsigned long)dev);
#endif
	lp->tx_tasklet = kmalloc(sizeof(struct tasklet_struct), GFP_KERNEL);
	tasklet_init(lp->tx_tasklet, korina_tx_tasklet, (unsigned long)dev);

#if 0	
	if ((err = register_netdev(dev))) {
		printk(KERN_ERR "korina ethernet. Cannot register net device %d\n", err);
		retval = -EINVAL;
		goto probe_err_out;
	}
#endif
	
	INFO("RC32434 Ethernet at 0x%08x, ", bif->iobase);
	for (i = 0; i < 6; i++) {
		printk("%2.2x", dev->dev_addr[i]);
		if (i<5)
			printk(":");
	}
	printk(", Rx IRQ %d, Tx IRQ %d\n", bif->rx_dma_irq, bif->tx_dma_irq);
	
	return 0;
	
 probe_err_out:
	korina_cleanup_module();
	ERR(" failed.  Returns %d\n", retval);
	return retval;
	
}
/*  
 *  MII transceiver control section.  
 *  The 83815 series has an internal transceiver, 
 *  and we present the management registers as 
 *  if they were MII connected. 
 */
static int korina_mdio_read(struct net_device *dev, int phy_id, int reg)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
	unsigned long flags;
	int retval =  0xffff;

	save_flags(flags);
	cli();
	if (phy_id == 1 && reg < 32) { 

		// register address to read
		rc32434_writel((MIIMADDR_PHYADDR(phy_id) | MIIMADDR_REGADDR(reg)), &lp->eth_regs->miimaddr); 

		// command READ
		rc32434_writel(MIIMCMD_RD, &lp->eth_regs->miimcmd); 

		// wait for busy and NV to be turned off
		while(rc32434_readl(&lp->eth_regs->miimind) & (MIIMIND_NV|MIIMIND_BSY));

		// double clutch from IDTBOOT code */
		rc32434_writel(0, &lp->eth_regs->miimcmd); 

		// command READ
		rc32434_writel(MIIMCMD_RD, &lp->eth_regs->miimcmd); 

		// wait for busy and NV to be turned off
		while(rc32434_readl(&lp->eth_regs->miimind) & (MIIMIND_NV|MIIMIND_BSY));

		rc32434_writel(0, &lp->eth_regs->miimcmd); 

		retval = rc32434_readl(&lp->eth_regs->miimrdd)&0xffff;
	} 
	restore_flags(flags);
	return retval;
}

static int korina_mdio_write(struct net_device *dev, int phy_id, int reg, int value)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
	unsigned long flags;
	int retval = 0;

	save_flags(flags);
	cli();

	if (phy_id == 1 && reg < 32) { 

		// wait for busy and NV to be turned off
		while(rc32434_readl(&lp->eth_regs->miimind) & MIIMIND_BSY);

		// register address to write
		rc32434_writel((MIIMADDR_PHYADDR(phy_id) | MIIMADDR_REGADDR(reg)), &lp->eth_regs->miimaddr); 

		// command write
		rc32434_writel(value, &lp->eth_regs->miimwtd); 
                retval = 1;
	} 
	restore_flags(flags);
	return retval;
}

static int netdev_get_ecmd(struct net_device *dev, struct ethtool_cmd *ecmd)
{
	struct korina_local *np = (struct korina_local *)dev->priv;
	u32 tmp;
 	unsigned long flags;

	save_flags(flags);
	cli();

	ecmd->supported = 
		(SUPPORTED_10baseT_Half | SUPPORTED_10baseT_Full |
		SUPPORTED_100baseT_Half | SUPPORTED_100baseT_Full |
		SUPPORTED_Autoneg | SUPPORTED_TP | SUPPORTED_MII);
	
        ecmd->port = PORT_TP;
	/* only supports internal transceiver */
	ecmd->transceiver = XCVR_INTERNAL;
	/* not sure what this is for */
	ecmd->phy_address = 1;

	ecmd->advertising = ADVERTISED_MII;
	tmp = korina_mdio_read(dev, np->phy, MII_ADVERTISE);
	if (tmp & ADVERTISE_10HALF)
		ecmd->advertising |= ADVERTISED_10baseT_Half;
	if (tmp & ADVERTISE_10FULL)
		ecmd->advertising |= ADVERTISED_10baseT_Full;
	if (tmp & ADVERTISE_100HALF)
		ecmd->advertising |= ADVERTISED_100baseT_Half;
	if (tmp & ADVERTISE_100FULL)
		ecmd->advertising |= ADVERTISED_100baseT_Full;

	tmp = korina_mdio_read(dev, np->phy, MII_BMCR);
	if (tmp & BMCR_ANENABLE) {
		ecmd->advertising |= ADVERTISED_Autoneg;
		ecmd->autoneg = AUTONEG_ENABLE;
	} else {
		ecmd->autoneg = AUTONEG_DISABLE;
	}
        if (ecmd->autoneg == AUTONEG_ENABLE) {
            tmp = korina_mdio_read(dev, np->phy, MII_LPA);
            if (tmp & (LPA_100FULL|LPA_10FULL)) {
                ecmd->duplex = DUPLEX_FULL;
            } else {
                ecmd->duplex = DUPLEX_HALF;
            }
            if (tmp & (LPA_100FULL|LPA_100HALF)) {
		ecmd->speed = SPEED_100;
            } else {
		ecmd->speed = SPEED_10;
            }
        } else {
            if (tmp & BMCR_FULLDPLX) {
                ecmd->duplex = DUPLEX_FULL;
            } else {
                ecmd->duplex = DUPLEX_HALF;
            }
            if (tmp & BMCR_SPEED100) {
		ecmd->speed = SPEED_100;
            } else {
		ecmd->speed = SPEED_10;
            }
	}
	restore_flags(flags);
	/* ignore maxtxpkt, maxrxpkt for now */

	return 0;
}
static int netdev_set_ecmd(struct net_device *dev, struct ethtool_cmd *ecmd)
{
	struct korina_local *np = dev->priv;
	u32 tmp;
	int enable_auto = 0;
    unsigned long flags;

	if (ecmd->speed != SPEED_10 && ecmd->speed != SPEED_100)
		return -EINVAL;
	if (ecmd->duplex != DUPLEX_HALF && ecmd->duplex != DUPLEX_FULL)
		return -EINVAL;
	if (ecmd->port != PORT_TP)
		return -EINVAL;
	if (ecmd->transceiver != XCVR_INTERNAL)
		return -EINVAL;
	if (ecmd->autoneg != AUTONEG_DISABLE && ecmd->autoneg != AUTONEG_ENABLE)
		return -EINVAL;
	/* ignore phy_address, maxtxpkt, maxrxpkt for now */

    save_flags(flags);
    cli();
    
	/* WHEW! now lets bang some bits */
	
	tmp = korina_mdio_read(dev, np->phy, MII_BMCR);
	if (ecmd->autoneg == AUTONEG_ENABLE) {
		/* turn on autonegotiation */
	        if (tmp & BMCR_ANENABLE) {
                        enable_auto = 1;
                } else {
                        printk("%s: Enabling auto-neg\n", dev->name);
                        tmp |= BMCR_ANENABLE;
                }
	} else {
		/* turn off auto negotiation, set speed and duplexity */
		tmp &= ~(BMCR_ANENABLE | BMCR_SPEED100 | BMCR_FULLDPLX);
		if (ecmd->speed == SPEED_100)
			tmp |= BMCR_SPEED100;
		if (ecmd->duplex == DUPLEX_FULL)
			tmp |= BMCR_FULLDPLX;
                printk("%s: Hard coding %d/%s\n", dev->name, 
                       (ecmd->speed == SPEED_100)? 100:10,
                       (ecmd->duplex == DUPLEX_FULL)? "full":"half");
	}
    np->phyData = 0;
        
    if(!enable_auto)
      korina_mdio_write(dev, np->phy, MII_BMCR, tmp);

    restore_flags(flags);
    
	return 0;
}

static int netdev_ethtool_ioctl(struct net_device *dev, void *useraddr)
{
	struct korina_local *np = dev->priv;
	u32 cmd;
    unsigned long flags;
	
	if (get_user(cmd, (u32 *)useraddr))
		return -EFAULT;

        switch (cmd) {
	/* get settings */
	case ETHTOOL_GSET: {
		struct ethtool_cmd ecmd = { ETHTOOL_GSET };
		spin_lock_irqsave(&np->lock, flags);
		netdev_get_ecmd(dev, &ecmd);
		spin_unlock_irqrestore(&np->lock, flags);
		if (copy_to_user(useraddr, &ecmd, sizeof(ecmd)))
			return -EFAULT;
		return 0;
	}
	/* set settings */
	case ETHTOOL_SSET: {
		struct ethtool_cmd ecmd;
		int r;
		if (copy_from_user(&ecmd, useraddr, sizeof(ecmd)))
			return -EFAULT;
		spin_lock_irqsave(&np->lock, flags);
		r = netdev_set_ecmd(dev, &ecmd);
		spin_unlock_irqrestore(&np->lock, flags);
		return r;
	}
	/* restart autonegotiation */
	case ETHTOOL_NWAY_RST: {
		int tmp;
		int r = -EINVAL;
		/* if autoneg is off, it's an error */
                printk("%s: Restarting auto-negotiation\n", dev->name);
		tmp = korina_mdio_read(dev, np->phy, MII_BMCR);
		if (tmp & BMCR_ANENABLE) {
			tmp |= (BMCR_ANRESTART);
			korina_mdio_write(dev, np->phy, MII_BMCR, tmp);
			r = 0;
		}
		return r;
	}
	/* get link status */
	case ETHTOOL_GLINK: {
		struct ethtool_value edata = {ETHTOOL_GLINK};
		edata.data = (korina_mdio_read(dev, np->phy, MII_BMSR)&BMSR_LSTATUS) ? 1:0;
		if (copy_to_user(useraddr, &edata, sizeof(edata)))
			return -EFAULT;
		return 0;
	}
        }
	
	return -EOPNOTSUPP;
}

static int korina_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	struct mii_ioctl_data *data = (struct mii_ioctl_data *)&rq->ifr_data;

	switch(cmd) {
	case SIOCETHTOOL:
		return netdev_ethtool_ioctl(dev, (void *) rq->ifr_data);
	case SIOCGMIIPHY:		/* Get address of MII PHY in use. */
	case SIOCDEVPRIVATE:		/* for binary compat, remove in 2.5 */
		data->phy_id = 1;
		/* Fall Through */

	case SIOCGMIIREG:		/* Read MII PHY register. */
	case SIOCDEVPRIVATE+1:		/* for binary compat, remove in 2.5 */
		data->val_out = korina_mdio_read(dev, data->phy_id & 0x1f, 
			data->reg_num & 0x1f);
		return 0;
	case SIOCSMIIREG:		/* Write MII PHY register. */
	case SIOCDEVPRIVATE+2:		/* for binary compat, remove in 2.5 */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		korina_mdio_write(dev, data->phy_id & 0x1f, 
			data->reg_num & 0x1f, data->val_in);
		return 0;
	default:
		return -EOPNOTSUPP;
	}
}


static void korina_cleanup_module(void)
{
	int i;
	
	for (i = 0; korina_iflist[i].iobase; i++) {
		struct korina_if_t * bif = &korina_iflist[i];
		if (bif->dev != NULL) {
			struct korina_local *lp = (struct korina_local *)bif->dev->priv;
			if (lp != NULL) {
				if (lp->eth_regs)
					iounmap((void*)lp->eth_regs);
				if (lp->rx_dma_regs)
					iounmap((void*)lp->rx_dma_regs);
				if (lp->tx_dma_regs)
					iounmap((void*)lp->tx_dma_regs);
				if (lp->td_ring)
					kfree((void*)KSEG0ADDR(lp->td_ring));
				
#ifdef KORINA_PROC_DEBUG
				if (lp->ps) {
					remove_proc_entry(bif->name, proc_net);
				}
#endif
				kfree(lp);
			}
			
			unregister_netdev(bif->dev);
			kfree(bif->dev);
		}
	}
}



static int korina_open(struct net_device *dev)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
#if 0
	/* Initialize */
	if (korina_init(dev)) {
		ERR("Error: cannot open the Ethernet device\n");
		lp->open = 0;
		return -EAGAIN;
	}
#endif
	
	/* Install the interrupt handler that handles the Done Finished Ovr and Und Events */	
	if (request_irq(lp->rx_irq, &korina_rx_dma_interrupt,
		  SA_SHIRQ | SA_INTERRUPT,
			"korina ethernet Rx", dev)) {
		ERR(": unable to get Rx DMA IRQ %d\n",
		    lp->rx_irq);
		return -EAGAIN;
	}
	if (request_irq(lp->tx_irq, &korina_tx_dma_interrupt,
		  SA_SHIRQ | SA_INTERRUPT,
			"korina ethernet Tx", dev)) {
		ERR(": unable to get Tx DMA IRQ %d\n",
		    lp->tx_irq);
		free_irq(lp->rx_irq, dev);
		return -EAGAIN;
	}
	
#ifdef	KORINA_REVISION
	/* Install handler for overrun error. */
	if (request_irq(lp->ovr_irq, &korina_ovr_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"Ethernet Overflow", dev)) {
		ERR(": unable to get OVR IRQ %d\n",
		    lp->ovr_irq);
		free_irq(lp->rx_irq, dev);
		free_irq(lp->tx_irq, dev);
		return -EAGAIN;
	}
#endif
	
	/* Install handler for underflow error. */
	if (request_irq(lp->und_irq, &korina_und_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"Ethernet Underflow", dev)) {
		ERR(": unable to get UND IRQ %d\n",
		    lp->und_irq);
		free_irq(lp->rx_irq, dev);
		free_irq(lp->tx_irq, dev);
#ifdef	KORINA_REVISION		
		free_irq(lp->ovr_irq, dev);		
#endif
		return -EAGAIN;
	}
	
	lp->open = 1;
#if 1
	/*
	 * Initialize
	 */
	if (korina_init(dev)) {
      ERR("Error: cannot open the Ethernet device\n");
      lp->open = 0;
      return -EAGAIN;
	}
#endif
    
    /* start link poll timer */
    korina_setup_timer(dev);

	return 0;
}




static int korina_close(struct net_device *dev)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
	u32 tmp;

	lp->open = 0;
	
	/* Disable interrupts */
	disable_irq(lp->rx_irq);
	disable_irq(lp->tx_irq);
#ifdef	KORINA_REVISION
	disable_irq(lp->ovr_irq);
#endif
	disable_irq(lp->und_irq);
	
	tmp = rc32434_readl(&lp->tx_dma_regs->dmasm);
	tmp = tmp | DMASM_f_m | DMASM_e_m;
	rc32434_writel(tmp, &lp->tx_dma_regs->dmasm);
	
	tmp = rc32434_readl(&lp->rx_dma_regs->dmasm);
	tmp = tmp | DMASM_d_m | DMASM_h_m | DMASM_e_m;
	rc32434_writel(tmp, &lp->rx_dma_regs->dmasm);
	
	free_irq(lp->rx_irq, dev);
	free_irq(lp->tx_irq, dev);
#ifdef	KORINA_REVISION	
	free_irq(lp->ovr_irq, dev);
#endif
	free_irq(lp->und_irq, dev);
	return 0;
}


/* transmit packet */
static int korina_send_packet(struct sk_buff *skb, struct net_device *dev)
{
	struct korina_local		*lp = (struct korina_local *)dev->priv;
	unsigned long 			flags;
	u32					length;
	DMAD_t				td;
	
	
	spin_lock_irqsave(&lp->lock, flags);
	
	td = &lp->td_ring[lp->tx_chain_tail];
	
	/* stop queue when full, drop pkts if queue already full */
	if(lp->tx_count >= (KORINA_NUM_TDS - 2)) {
		lp->tx_full = 1;
		
		if(lp->tx_count == (KORINA_NUM_TDS - 2)) {
			netif_stop_queue(dev);
		}
		else {
			lp->stats.tx_dropped++;
			//dev_kfree_skb_any(skb);
			spin_unlock_irqrestore(&lp->lock, flags);
			return 1;
		}	   
	}	 
	
	lp->tx_count ++;
	
	lp->tx_skb[lp->tx_chain_tail] = skb;
	
	length = skb->len;
	lp->stats.tx_bytes += length;
	
	/* Setup the transmit descriptor. */
	td->ca = CPHYSADDR(skb->data);
	
	if(rc32434_readl(&(lp->tx_dma_regs->dmandptr)) == 0) {
		if( lp->tx_chain_status == empty ) {
			td->control = DMA_COUNT(length) |DMAD_cof_m |DMAD_iof_m;                                /*  Update tail      */
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & KORINA_TDS_MASK;                          /*   Move tail       */
			rc32434_writel(CPHYSADDR(&lp->td_ring[lp->tx_chain_head]), &(lp->tx_dma_regs->dmandptr)); /* Write to NDPTR    */
			lp->tx_chain_head = lp->tx_chain_tail;                                                  /* Move head to tail */
		}
		else {
			td->control = DMA_COUNT(length) |DMAD_cof_m|DMAD_iof_m;                                 /* Update tail */
			lp->td_ring[(lp->tx_chain_tail-1)& KORINA_TDS_MASK].control &=  ~(DMAD_cof_m);          /* Link to prev */
			lp->td_ring[(lp->tx_chain_tail-1)& KORINA_TDS_MASK].link =  CPHYSADDR(td);              /* Link to prev */
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & KORINA_TDS_MASK;                          /* Move tail */
			rc32434_writel(CPHYSADDR(&lp->td_ring[lp->tx_chain_head]), &(lp->tx_dma_regs->dmandptr)); /* Write to NDPTR */
			lp->tx_chain_head = lp->tx_chain_tail;                                                  /* Move head to tail */
			lp->tx_chain_status = empty;
		}
	}
	else {
		if( lp->tx_chain_status == empty ) {
			td->control = DMA_COUNT(length) |DMAD_cof_m |DMAD_iof_m;                                /* Update tail */
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & KORINA_TDS_MASK;                          /* Move tail */
			lp->tx_chain_status = filled;
		}
		else {
			td->control = DMA_COUNT(length) |DMAD_cof_m |DMAD_iof_m;                                /* Update tail */
			lp->td_ring[(lp->tx_chain_tail-1)& KORINA_TDS_MASK].control &=  ~(DMAD_cof_m);          /* Link to prev */
			lp->td_ring[(lp->tx_chain_tail-1)& KORINA_TDS_MASK].link =  CPHYSADDR(td);              /* Link to prev */
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & KORINA_TDS_MASK;                          /* Move tail */
		}
	}
	
	dev->trans_start = jiffies;				
	
	spin_unlock_irqrestore(&lp->lock, flags);
	
	return 0;
}


/* Ethernet MII-PHY Handler */
static void korina_mii_handler(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;		
	struct korina_local *lp = (struct korina_local *)dev->priv;
	unsigned long 	flags;
	unsigned long duplex_status;
	int port_addr = (lp->rx_irq == 0x2c? 1:0) << 8;
	
	spin_lock_irqsave(&lp->lock, flags);
	
	/* Two ports are using the same MII, the difference is the PHY address */
	rc32434_writel(0, &korina_eth0_regs->miimcfg);  
	rc32434_writel(0, &korina_eth0_regs->miimcmd);  
	rc32434_writel(port_addr |0x05, &korina_eth0_regs->miimaddr);  
	rc32434_writel(MIIMCMD_scn_m, &korina_eth0_regs->miimcmd);  
	while(rc32434_readl(&korina_eth0_regs->miimind) & MIIMIND_nv_m);
	
	ERR("irq:%x		port_addr:%x	RDD:%x\n", 
	    lp->rx_irq, port_addr, rc32434_readl(&korina_eth0_regs->miimrdd));
	duplex_status = (rc32434_readl(&korina_eth0_regs->miimrdd) & 0x140)? ETHMAC2_fd_m: 0;
	if(duplex_status != lp->duplex_mode) {
		ERR("The MII-PHY is Auto-negotiated to %s-Duplex mode for Eth-%x\n", duplex_status? "Full":"Half", lp->rx_irq == 0x2c? 1:0);		
		lp->duplex_mode = duplex_status;
		korina_restart(dev);		
	}
	
	lp->mii_phy_timer.expires = jiffies + 10 * HZ;	
	add_timer(&lp->mii_phy_timer);
	
	spin_unlock_irqrestore(&lp->lock, flags);
	
}

#ifdef	KORINA_REVISION	
/* Ethernet Rx Overflow interrupt */
static void
korina_ovr_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct korina_local *lp;
	unsigned int ovr;
	
	ASSERT(dev != NULL);
	
	lp = (struct korina_local *)dev->priv;
	spin_lock(&lp->lock);
        lp->stats.rx_errors++;
        lp->stats.rx_over_errors++;
	ovr = rc32434_readl(&lp->eth_regs->ethintfc);
	
	if(ovr & ETHINTFC_ovr_m) {
		netif_stop_queue(dev);
		
		/* clear OVR bit */
		rc32434_writel((ovr & ~ETHINTFC_ovr_m), &lp->eth_regs->ethintfc);
		
		/* Restart interface */
		korina_restart(dev);
	}
	spin_unlock(&lp->lock);
	
	return;
}

#endif


/* Ethernet Tx Underflow interrupt */
static void
korina_und_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct korina_local *lp;
	unsigned int und;
	
	ASSERT(dev != NULL);
	
	lp = (struct korina_local *)dev->priv;
	
	spin_lock(&lp->lock);
	lp->stats.tx_errors++;	
	und = rc32434_readl(&lp->eth_regs->ethintfc);
	
	if(und & ETHINTFC_und_m) {
		netif_stop_queue(dev);
		
		rc32434_writel((und & ~ETHINTFC_und_m), &lp->eth_regs->ethintfc);
		
		/* Restart interface */
		korina_restart(dev);
	}
	
	spin_unlock(&lp->lock);
	
	return;
}


/* Ethernet Rx DMA interrupt */
static void
korina_rx_dma_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct korina_local* lp;
	volatile u32 dmas,dmasm;
	
	ASSERT(dev != NULL);
	
	lp = (struct korina_local *)dev->priv;
	
	spin_lock(&lp->lock);
	dmas = rc32434_readl(&lp->rx_dma_regs->dmas);
	if(dmas & (DMAS_d_m|DMAS_h_m|DMAS_e_m)) {
		/* Mask D H E bit in Rx DMA */
		dmasm = rc32434_readl(&lp->rx_dma_regs->dmasm);
		rc32434_writel(dmasm | (DMASM_d_m | DMASM_h_m | DMASM_e_m), &lp->rx_dma_regs->dmasm);
#ifdef CONFIG_IDT_USE_NAPI
		if(netif_rx_schedule_prep(dev))
                        __netif_rx_schedule(dev);
#else
		tasklet_hi_schedule(lp->rx_tasklet);
#endif
		
		if (dmas & DMAS_e_m)
			ERR(": DMA error\n");
		
	}
	
	spin_unlock(&lp->lock);
	return;
}

#ifdef CONFIG_IDT_USE_NAPI
static int korina_poll(struct net_device *rx_data_dev, int *budget)
#else
static void korina_rx_tasklet(unsigned long rx_data_dev)
#endif
{
	struct net_device *dev = (struct net_device *)rx_data_dev;	
	struct korina_local* lp = dev->priv;
	volatile DMAD_t  rd = &lp->rd_ring[lp->rx_next_done];
	struct sk_buff *skb, *skb_new;
	u8* pkt_buf;
	u32 devcs, count, pkt_len, pktuncrc_len;
	volatile u32 dmas;
#ifdef CONFIG_IDT_USE_NAPI
	u32 received = 0;
	int rx_work_limit = min(*budget,dev->quota);
#else
	unsigned long 	flags;
	spin_lock_irqsave(&lp->lock, flags);
        if (!lp->open) {
                spin_unlock_irqrestore(&lp->lock, flags);
                return;
        }
#endif
	
	while ( (count = KORINA_RBSIZE - (u32)DMA_COUNT(rd->control)) != 0) {
#ifdef CONFIG_IDT_USE_NAPI
		if(--rx_work_limit <0)
                {
                        break;
                }
#endif
		/* init the var. used for the later operations within the while loop */
		skb_new = NULL;
		devcs = rd->devcs;
		pkt_len = RCVPKT_LENGTH(devcs);
		skb = lp->rx_skb[lp->rx_next_done];
      
		if (count < 64) {
			lp->stats.rx_errors++;
			lp->stats.rx_dropped++;			
		}
		else if ((devcs & ( ETHRX_ld_m)) !=	ETHRX_ld_m) {
			/* check that this is a whole packet */
			/* WARNING: DMA_FD bit incorrectly set in Korina (errata ref #077) */
			lp->stats.rx_errors++;
			lp->stats.rx_dropped++;
		}
		else if ( (devcs & ETHRX_rok_m)  ) {
			
			{
				/* must be the (first and) last descriptor then */
				pkt_buf = (u8*)lp->rx_skb[lp->rx_next_done]->data;
				
				pktuncrc_len = pkt_len - 4;
				/* invalidate the cache */
				dma_cache_inv((unsigned long)pkt_buf, pktuncrc_len);
				
				/* Malloc up new buffer. */					  
				skb_new = dev_alloc_skb(KORINA_RBSIZE + 2 + 128);					             	
				
				if (skb_new != NULL){
					/* Make room */
					skb_put(skb, pktuncrc_len);		    
#ifdef CONFIG_MERLOT
                                  if (((dev->am_pkt_handler == NULL) || 
                                       (dev->am_pkt_handler(skb, dev) == 0))){
#endif
                                    
	
					skb->protocol = eth_type_trans(skb, dev);
					
					/* pass the packet to upper layers */
#ifdef CONFIG_IDT_USE_NAPI
					netif_receive_skb(skb);
#else
#ifdef CONFIG_MERLOT
                                      if (dev->asap_netif_rx)
                                        dev->asap_netif_rx(skb);
                                      else
                                        netif_rx(skb);
#else
			              netif_rx(skb);
#endif
#endif
#ifdef CONFIG_MERLOT
                                  }
#endif
					
					dev->last_rx = jiffies;
					lp->stats.rx_packets++;
					lp->stats.rx_bytes += pktuncrc_len;
					
					if (IS_RCV_MP(devcs))
						lp->stats.multicast++;
					
					/* 16 bit align */						  
					skb_reserve(skb_new, 2 + 32);	
					
					skb_new->dev = dev;
					lp->rx_skb[lp->rx_next_done] = skb_new;
				}
				else {
					ERR("no memory, dropping rx packet.\n");
					lp->stats.rx_errors++;		
					lp->stats.rx_dropped++;					
				}
			}
			
		}			
		else {
			/* This should only happen if we enable accepting broken packets */
			lp->stats.rx_errors++;
			lp->stats.rx_dropped++;
			
			/* add statistics counters */
			if (IS_RCV_CRC_ERR(devcs)) {
				DBG(2, "RX CRC error\n");
				lp->stats.rx_crc_errors++;
			} 
			else if (IS_RCV_LOR_ERR(devcs)) {
				DBG(2, "RX LOR error\n");
				lp->stats.rx_length_errors++;
			}				
			else if (IS_RCV_LE_ERR(devcs)) {
				DBG(2, "RX LE error\n");
				lp->stats.rx_length_errors++;
			}
			else if (IS_RCV_OVR_ERR(devcs)) {
				lp->stats.rx_over_errors++;
			}
			else if (IS_RCV_CV_ERR(devcs)) {
				/* code violation */
				DBG(2, "RX CV error\n");
				lp->stats.rx_frame_errors++;
			}
			else if (IS_RCV_CES_ERR(devcs)) {
				DBG(2, "RX Preamble error\n");
			}
		}
		
		rd->devcs = 0;
		
		/* restore descriptor's curr_addr */
		if(skb_new)
			rd->ca = CPHYSADDR(skb_new->data); 
		else
			rd->ca = CPHYSADDR(skb->data);
		
		rd->control = DMA_COUNT(KORINA_RBSIZE) |DMAD_cod_m |DMAD_iod_m;
		lp->rd_ring[(lp->rx_next_done-1)& KORINA_RDS_MASK].control &=  ~(DMAD_cod_m); 	
		
		lp->rx_next_done = (lp->rx_next_done + 1) & KORINA_RDS_MASK;
		rd = &lp->rd_ring[lp->rx_next_done];
		rc32434_writel( ~DMAS_d_m, &lp->rx_dma_regs->dmas);
	}	
#ifdef CONFIG_IDT_USE_NAPI
        dev->quota -= received;
        *budget =- received;
        if(rx_work_limit < 0)
                goto not_done;
#endif
	
	dmas = rc32434_readl(&lp->rx_dma_regs->dmas);
	
	if(dmas & DMAS_h_m) {
		rc32434_writel( ~(DMAS_h_m | DMAS_e_m), &lp->rx_dma_regs->dmas);
#ifdef KORINA_PROC_DEBUG
		lp->dma_halt_cnt++;
#endif
		rd->devcs = 0;
		skb = lp->rx_skb[lp->rx_next_done];
		rd->ca = CPHYSADDR(skb->data);
		korina_chain_rx(lp,rd);
	}
	
#ifdef CONFIG_IDT_USE_NAPI
	netif_rx_complete(dev);
#endif
	/* Enable D H E bit in Rx DMA */
	rc32434_writel(rc32434_readl(&lp->rx_dma_regs->dmasm) & ~(DMASM_d_m | DMASM_h_m |DMASM_e_m), &lp->rx_dma_regs->dmasm); 
#ifdef CONFIG_IDT_USE_NAPI
	return 0;
 not_done:
	return 1;
#else
	spin_unlock_irqrestore(&lp->lock, flags);
	return;
#endif

	
}	



/* Ethernet Tx DMA interrupt */
static void
korina_tx_dma_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct korina_local *lp;
	volatile u32 dmas,dmasm;
	
	ASSERT(dev != NULL);
	
	lp = (struct korina_local *)dev->priv;
	
	spin_lock(&lp->lock);
	
	dmas = rc32434_readl(&lp->tx_dma_regs->dmas);
	
	if (dmas & (DMAS_f_m | DMAS_e_m)) {
		dmasm = rc32434_readl(&lp->tx_dma_regs->dmasm);
		/* Mask F E bit in Tx DMA */
		rc32434_writel(dmasm | (DMASM_f_m | DMASM_e_m), &lp->tx_dma_regs->dmasm);
		
		tasklet_hi_schedule(lp->tx_tasklet);
		
		if(lp->tx_chain_status == filled && (rc32434_readl(&(lp->tx_dma_regs->dmandptr)) == 0)) {
			rc32434_writel(CPHYSADDR(&lp->td_ring[lp->tx_chain_head]), &(lp->tx_dma_regs->dmandptr));			
			lp->tx_chain_status = empty;
			lp->tx_chain_head = lp->tx_chain_tail;
			dev->trans_start = jiffies;
		}
		
		if (dmas & DMAS_e_m)
			ERR(": DMA error\n");
		
	}
	spin_unlock(&lp->lock);
	
	return;
}


static void korina_tx_tasklet(unsigned long tx_data_dev)
{
	struct net_device *dev = (struct net_device *)tx_data_dev;	
	struct korina_local* lp = (struct korina_local *)dev->priv;
	volatile DMAD_t td = &lp->td_ring[lp->tx_next_done];
	u32 devcs;
	unsigned long 	flags;
	volatile u32 dmas;
	
	spin_lock_irqsave(&lp->lock, flags);
	
	/* process all desc that are done */
	while(IS_DMA_FINISHED(td->control)) {
		if(lp->tx_full == 1) {
			netif_wake_queue(dev);
			lp->tx_full = 0;
		}
		
		devcs = lp->td_ring[lp->tx_next_done].devcs;    
		if ((devcs & (ETHTX_fd_m | ETHTX_ld_m)) != (ETHTX_fd_m | ETHTX_ld_m)) {
			lp->stats.tx_errors++;
			lp->stats.tx_dropped++;				
			
			/* should never happen */
			DBG(1, __FUNCTION__ ": split tx ignored\n");
		}
		else if (IS_TX_TOK(devcs)) {
			lp->stats.tx_packets++;
		}
		else {
			lp->stats.tx_errors++;
			lp->stats.tx_dropped++;				
			
			/* underflow */
			if (IS_TX_UND_ERR(devcs)) 
				lp->stats.tx_fifo_errors++;
			
			/* oversized frame */
			if (IS_TX_OF_ERR(devcs))
				lp->stats.tx_aborted_errors++;
			
			/* excessive deferrals */
			if (IS_TX_ED_ERR(devcs))
				lp->stats.tx_carrier_errors++;
			
			/* collisions: medium busy */
			if (IS_TX_EC_ERR(devcs))
				lp->stats.collisions++;
			
			/* late collision */
			if (IS_TX_LC_ERR(devcs))
				lp->stats.tx_window_errors++;
			
		}
		
		/* We must always free the original skb */
		if (lp->tx_skb[lp->tx_next_done] != NULL) {
			dev_kfree_skb_any(lp->tx_skb[lp->tx_next_done]);
			lp->tx_skb[lp->tx_next_done] = NULL;
		}
		
		lp->td_ring[lp->tx_next_done].control = DMAD_iof_m;
		lp->td_ring[lp->tx_next_done].devcs = ETHTX_fd_m | ETHTX_ld_m;	
		lp->td_ring[lp->tx_next_done].link = 0;
		lp->td_ring[lp->tx_next_done].ca = 0;
		lp->tx_count --;
		
		/* go on to next transmission */
		lp->tx_next_done = (lp->tx_next_done + 1) & KORINA_TDS_MASK;
		td = &lp->td_ring[lp->tx_next_done];
		
	}
	
	dmas = rc32434_readl(&lp->tx_dma_regs->dmas);
	rc32434_writel( ~dmas, &lp->tx_dma_regs->dmas);
	
	/* Enable F E bit in Tx DMA */
	rc32434_writel(rc32434_readl(&lp->tx_dma_regs->dmasm) & ~(DMASM_f_m | DMASM_e_m), &lp->tx_dma_regs->dmasm); 
	spin_unlock_irqrestore(&lp->lock, flags);
	
}


static struct net_device_stats * korina_get_stats(struct net_device *dev)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
	return &lp->stats;
}


/*
 * Set or clear the multicast filter for this adaptor.
 */
static void korina_multicast_list(struct net_device *dev)
{   
	/* listen to broadcasts always and to treat 	*/
	/*       IFF bits independantly	*/
	struct korina_local *lp = (struct korina_local *)dev->priv;
	unsigned long flags;
	u32 recognise = ETHARC_ab_m; 		/* always accept broadcasts */
	
	if (dev->flags & IFF_PROMISC)         		/* set promiscuous mode */
		recognise |= ETHARC_pro_m;
	
	if ((dev->flags & IFF_ALLMULTI) || (dev->mc_count > 15))
		recognise |= ETHARC_am_m;    	  	/* all multicast & bcast */
	else if (dev->mc_count > 0) {
		DBG(2, __FUNCTION__ ": mc_count %d\n", dev->mc_count);
		recognise |= ETHARC_am_m;    	  	/* for the time being */
	}
	
	spin_lock_irqsave(&lp->lock, flags);
	rc32434_writel(recognise, &lp->eth_regs->etharc);
	spin_unlock_irqrestore(&lp->lock, flags);
}


static void korina_tx_timeout(struct net_device *dev)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
	unsigned long flags;
	
	spin_lock_irqsave(&lp->lock, flags);
	lp->stats.tx_errors++;
	korina_restart(dev);
	spin_unlock_irqrestore(&lp->lock, flags);
	
}


/*
 * Initialize the KORINA ethernet controller.
 */
static int korina_init(struct net_device *dev)
{
	struct korina_local *lp = (struct korina_local *)dev->priv;
	int i, j;
	
	/* Disable DMA */       
	korina_abort_tx(dev);
	korina_abort_rx(dev); 
	
	/* reset ethernet logic */ 
	rc32434_writel(0, &lp->eth_regs->ethintfc);
	while((rc32434_readl(&lp->eth_regs->ethintfc) & ETHINTFC_rip_m))
		dev->trans_start = jiffies;	
	
	/* Enable Ethernet Interface */ 
	//rc32434_writel(ETHINTFC_en_m, &lp->eth_regs->ethintfc); 
	
#ifndef CONFIG_IDT_USE_NAPI
	tasklet_disable(lp->rx_tasklet);
#endif
	tasklet_disable(lp->tx_tasklet);
	
	/* Initialize the transmit Descriptors */
	for (i = 0; i < KORINA_NUM_TDS; i++) {
		lp->td_ring[i].control = DMAD_iof_m;
		lp->td_ring[i].devcs = ETHTX_fd_m | ETHTX_ld_m;
		lp->td_ring[i].ca = 0;
		lp->td_ring[i].link = 0;
		if (lp->tx_skb[i] != NULL) {
			dev_kfree_skb_any(lp->tx_skb[i]);
			lp->tx_skb[i] = NULL;
		}
	}
	lp->tx_next_done = lp->tx_chain_head = lp->tx_chain_tail = 	lp->tx_full = lp->tx_count = 0;
	lp->	tx_chain_status = empty;
	
	rc32434_writel(0, &lp->tx_dma_regs->dmandptr);
		
	/*
	 * Initialize the receive descriptors so that they
	 * become a circular linked list, ie. let the last
	 * descriptor point to the first again.
	 */
	for (i=0; i<KORINA_NUM_RDS; i++) {
		struct sk_buff *skb = lp->rx_skb[i];
		
		if (lp->rx_skb[i] == NULL) {
			skb = dev_alloc_skb(KORINA_RBSIZE + 2 + 128);
			if (skb == NULL) {
				ERR("No memory in the system\n");
				for (j = 0; j < KORINA_NUM_RDS; j ++)
					if (lp->rx_skb[j] != NULL) 
						dev_kfree_skb_any(lp->rx_skb[j]);
				
				return 1;
			}
			else {
				skb->dev = dev;
				skb_reserve(skb, 2 + 32);
				lp->rx_skb[i] = skb;
				lp->rd_ring[i].ca = CPHYSADDR(skb->data); 
				
			}
		}
		lp->rd_ring[i].control =	DMAD_iod_m | DMA_COUNT(KORINA_RBSIZE);
		lp->rd_ring[i].devcs = 0;
		lp->rd_ring[i].ca = CPHYSADDR(skb->data);
		lp->rd_ring[i].link = CPHYSADDR(&lp->rd_ring[i+1]);
		
	}
	/* loop back */
	lp->rd_ring[KORINA_NUM_RDS-1].link = CPHYSADDR(&lp->rd_ring[0]);
	lp->rx_next_done   = 0;
	
	lp->rd_ring[KORINA_NUM_RDS-1].control |= DMAD_cod_m;
	lp->rx_chain_head = 0;
	lp->rx_chain_tail = 0;
	lp->rx_chain_status = empty;
	
	rc32434_writel(0, &lp->rx_dma_regs->dmas);
	/* Start Rx DMA */
	korina_start_rx(lp, &lp->rd_ring[0]);
	
	/* Enable F E bit in Tx DMA */
	rc32434_writel(rc32434_readl(&lp->tx_dma_regs->dmasm) & ~(DMASM_f_m | DMASM_e_m), &lp->tx_dma_regs->dmasm); 
	/* Enable D H E bit in Rx DMA */
	rc32434_writel(rc32434_readl(&lp->rx_dma_regs->dmasm) & ~(DMASM_d_m | DMASM_h_m | DMASM_e_m), &lp->rx_dma_regs->dmasm); 
	
	/* Accept only packets destined for this Ethernet device address */
	rc32434_writel(ETHARC_ab_m, &lp->eth_regs->etharc); 
	
	/* Set all Ether station address registers to their initial values */ 
	rc32434_writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal0); 
	rc32434_writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah0);
	
	rc32434_writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal1); 
	rc32434_writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah1);
	
	rc32434_writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal2); 
	rc32434_writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah2);
	
	rc32434_writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal3); 
	rc32434_writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah3); 
	
	 rc32434_writel(0, &lp->eth_regs->miimcfg);	
	/* Frame Length Checking, Pad Enable, CRC Enable, Full Duplex set */ 
	rc32434_writel(ETHMAC2_flc_m | ETHMAC2_pe_m | ETHMAC2_cen_m | ETHMAC2_fd_m | ETHMAC2_hfe_m, &lp->eth_regs->ethmac2);  
	//ETHMAC2_flc_m		ETHMAC2_fd_m	lp->duplex_mode

	/* set Eth MTU */
        rc32434_writel(1600, &lp->eth_regs->ethmaxf);
	
	/* Back to back inter-packet-gap */ 
	rc32434_writel(0x15, &lp->eth_regs->ethipgt); 
	/* Non - Back to back inter-packet-gap */ 
	rc32434_writel(0x12, &lp->eth_regs->ethipgr); 
	
	/* Management Clock Prescaler Divisor */
	/* Clock independent setting */
    	rc32434_writel(((100 * 1000 * 1000)/MII_CLOCK+1) & ~1,
	       &lp->eth_regs->ethmcp);
	
	/* don't transmit until fifo contains 0x40 */
	rc32434_writel(0x40, &lp->eth_regs->ethfifott);
	
	rc32434_writel(ETHMAC1_re_m, &lp->eth_regs->ethmac1);
	
#ifndef CONFIG_IDT_USE_NAPI
	tasklet_enable(lp->rx_tasklet);
#endif
	tasklet_enable(lp->tx_tasklet);
	
	netif_wake_queue(dev);
	
	/* Enable Ethernet Interface */ 
	rc32434_writel(ETHINTFC_en_m, &lp->eth_regs->ethintfc); 
	
	return 0; 
	
}
static int korina_setup_timer(struct net_device *dev)
{
    struct korina_local *sp = dev->priv; 

    init_timer(&sp->link_timer);

    sp->link_timer.function = korina_link_timer_fn;
    sp->link_timer.data = (int) dev;
    sp->link_timer.expires = jiffies + HZ;
    sp->phy = 1;

    add_timer(&sp->link_timer);
    return 0;

}

static void korina_link_timer_fn(unsigned long data)
{
    struct net_device *dev = (struct net_device *) data;
    struct korina_local *sp = dev->priv; 

    // see if the link status changed
    // This was needed to make sure we set the PHY to the
    // autonegotiated value of half or full duplex.
    korina_check_link(dev);
    
    // Loop faster when we don't have link. 
    // This was needed to speed up the AP bootstrap time.
    if(sp->link == 0) {
    	mod_timer(&sp->link_timer, jiffies + HZ/2);
    } else {
    	mod_timer(&sp->link_timer, jiffies + LINK_TIMER);
    }
}

static void korina_check_link(struct net_device *dev)
{
    struct korina_local *lp = dev->priv;
    u16 phyData;
    unsigned long flags;

    save_flags(flags);
    cli();

    phyData = korina_mdio_read(dev, lp->phy, MII_BMSR);
    if (lp->phyData != phyData) {
	if (phyData & BMSR_LSTATUS) {
            /* link is present */
            u16 reg;
            u32 mac2;
            int duplex = 0;

            lp->link = 1;
            /* Read PHY (Kendin KS87218) to determine duplexity for MAC */
            reg = korina_mdio_read(dev, lp->phy, MII_BMCR);
            if (reg & BMCR_ANENABLE) {
                /* auto neg enabled */
                reg = korina_mdio_read(dev, lp->phy, MII_LPA);
                duplex = (reg & (LPA_100FULL|LPA_10FULL))? 1:0;
            } else {
                /* no auto neg, just read duplex config */
                duplex = (reg & BMCR_FULLDPLX)? 1:0;
            }
	
            if (!duplex) {
                printk(KERN_INFO "%s: Configuring MAC for %s duplex\n", dev->name,
                       duplex? "full":"half");
            }

            mac2 = rc32434_readl(&lp->eth_regs->ethmac2);
            if (duplex) {
                /* full duplex */
                rc32434_writel(mac2 | ETHMAC2_fd_m, &lp->eth_regs->ethmac2);  
            } else {
                /* half duplex */
                rc32434_writel(mac2 & ~ETHMAC2_fd_m, &lp->eth_regs->ethmac2);  
            }
    	} else {
            /* no link */
            lp->link=0;
    	}
        lp->phyData = phyData;
    }
    restore_flags(flags);
}


#ifndef MODULE

static int __init korina_setup(char *options)
{
	/* no options yet */
	return 1;
}

static int __init korina_setup_ethaddr(char *options)
{
	memcpy(mac, options, 17);
	mac[17]= '\0';
	return 1;
}

__setup("rc32434eth=", korina_setup);
__setup("ethaddr=", korina_setup_ethaddr);


#endif /* MODULE */

module_init(korina_init_module);
module_exit(korina_cleanup_module);














