/**************************************************************************
 *
 *  BRIEF MODULE DESCRIPTION
 *     Definitions for IDT KORINA on-chip ethernet controller.
 *
 *  Copyright 2004 IDT Inc. (rischelp@idt.com)
 *         
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **************************************************************************
 * May 2004 rkt, neb
 *
 * Initial Release
 *
 * Aug 2004
 *
 * Added NAPI
 *
 **************************************************************************
 */


#include  <asm/rc32434/rc32434.h>
#include  <asm/rc32434/dma_v.h>
#include  <asm/rc32434/eth_v.h>

#define KORINA_REVISION
#define CPHYSADDR virt_to_phys
//MIIMCFG
#define MIIMCFG_RSV(i)          ((i&0x3)<<2)
#define MIIMCFG_R              (1<<15)

//MIIMCMD
#define MIIMCMD_RD             (1<<0)
#define MIIMCMD_SCN            (1<<1)

//MIIMADDR
#define MIIMADDR_REGADDR(i)    ((i&0x1f)<<0)
#define MIIMADDR_PHYADDR(i)    ((i&0x1f)<<8)

//MIIMIND
#define MIIMIND_BSY            (1<<0)
#define MIIMIND_SCN            (1<<1)
#define MIIMIND_NV             (1<<2)

#define KORINA_DEBUG	2
//#define KORINA_PROC_DEBUG
#undef	KORINA_DEBUG

#ifdef KORINA_DEBUG

/* use 0 for production, 1 for verification, >2 for debug */
static int korina_debug = KORINA_DEBUG;
#define ASSERT(expr) \
	if(!(expr)) {	\
		printk( "Assertion failed! %s,%s,%s,line=%d\n",	\
		#expr,__FILE__,__FUNCTION__,__LINE__);		}
#define DBG(lvl, format, arg...) if (korina_debug > lvl) printk(KERN_INFO "%s: " format, dev->name , ## arg)
#else
#define ASSERT(expr) do {} while (0)
#define DBG(lvl, format, arg...) do {} while (0)
#endif

#define INFO(format, arg...) printk(KERN_INFO "%s: " format, dev->name , ## arg)
#define ERR(format, arg...) printk(KERN_ERR "%s: " format, dev->name , ## arg)
#define WARN(format, arg...) printk(KERN_WARNING "%s: " format, dev->name , ## arg)		

#define ETH0_DMA_RX_IRQ   	GROUP1_IRQ_BASE + 0
#define ETH0_DMA_TX_IRQ   	GROUP1_IRQ_BASE + 1 
#define ETH0_RX_OVR_IRQ   	GROUP3_IRQ_BASE + 9
#define ETH0_TX_UND_IRQ   	GROUP3_IRQ_BASE + 10

#define ETH0_RX_DMA_ADDR  (DMA0_PhysicalAddress + 0*DMA_CHAN_OFFSET)
#define ETH0_TX_DMA_ADDR  (DMA0_PhysicalAddress + 1*DMA_CHAN_OFFSET)

/* the following must be powers of two */
#ifdef CONFIG_IDT_USE_NAPI
#define KORINA_NUM_RDS    64    		/* number of receive descriptors */
#define KORINA_NUM_TDS    64    		/* number of transmit descriptors */
#else
#define KORINA_NUM_RDS    128    		/* number of receive descriptors */
#define KORINA_NUM_TDS    128    		/* number of transmit descriptors */
#endif

#define KORINA_RBSIZE     1600  		/* size of one resource buffer = Ether MTU */
#define KORINA_RDS_MASK   (KORINA_NUM_RDS-1)
#define KORINA_TDS_MASK   (KORINA_NUM_TDS-1)
#define RD_RING_SIZE (KORINA_NUM_RDS * sizeof(struct DMAD_s))
#define TD_RING_SIZE (KORINA_NUM_TDS * sizeof(struct DMAD_s))

#define KORINA_TX_TIMEOUT HZ * 100

#define korina_eth0_regs ((ETH_t)(ETH0_VirtualAddress))
#define korina_eth1_regs ((ETH_t)(ETH1_VirtualAddress))

enum status	{ filled,	empty};
#define IS_DMA_FINISHED(X)   (((X) & (DMAD_f_m)) != 0)
#define IS_DMA_DONE(X)   (((X) & (DMAD_d_m)) != 0)

/*
 * probe link timer - 5 secs
 */
#define LINK_TIMER    (5*HZ) 

/* Information that need to be kept for each board. */
struct korina_local {
	ETH_t  eth_regs;
	DMA_Chan_t  rx_dma_regs;
	DMA_Chan_t  tx_dma_regs;
	volatile DMAD_t   td_ring;			/* transmit descriptor ring */ 
	volatile DMAD_t   rd_ring;			/* receive descriptor ring  */
	
	struct sk_buff* tx_skb[KORINA_NUM_TDS]; 	/* skbuffs for pkt to trans */
	struct sk_buff* rx_skb[KORINA_NUM_RDS]; 	/* skbuffs for pkt to trans */
	
#ifndef CONFIG_IDT_USE_NAPI
	struct tasklet_struct * rx_tasklet;
#endif
	struct tasklet_struct * tx_tasklet;
	
	int	rx_next_done;
	int	rx_chain_head;
	int	rx_chain_tail;
	enum status	rx_chain_status;
	
	int	tx_next_done;
	int	tx_chain_head;
	int	tx_chain_tail;
	enum status	tx_chain_status;
	int tx_count;			
	int	tx_full;
	
	struct timer_list    mii_phy_timer;
	unsigned long duplex_mode;
	
	int   	rx_irq;
	int    tx_irq;
	int    ovr_irq;
	int    und_irq;
	struct timer_list link_timer;
        u16 phyData;
        u16 phy;
        int link;
	
	struct net_device_stats stats;
	spinlock_t lock; 
	int open;
	
	/* debug /proc entry */
	struct proc_dir_entry *ps;
	int dma_halt_cnt;  int dma_run_cnt;
};

extern unsigned int idt_cpu_freq;

/* Index to functions, as function prototypes. */
static int korina_open(struct net_device *dev);
static int korina_send_packet(struct sk_buff *skb, struct net_device *dev);
static void korina_mii_handler(unsigned long data);
static void  korina_und_interrupt(int irq, void *dev_id, struct pt_regs * regs);
static void korina_rx_dma_interrupt(int irq, void *dev_id, struct pt_regs * regs);
static void korina_tx_dma_interrupt(int irq, void *dev_id, struct pt_regs * regs);
#ifdef	KORINA_REVISION	
static void korina_ovr_interrupt(int irq, void *dev_id, struct pt_regs * regs);
#endif
static int  korina_close(struct net_device *dev);
static struct net_device_stats *korina_get_stats(struct net_device *dev);
static void korina_multicast_list(struct net_device *dev);
static int  korina_init(struct net_device *dev);
static void korina_tx_timeout(struct net_device *dev);

static void korina_tx_tasklet(unsigned long tx_data_dev);
#ifdef CONFIG_IDT_USE_NAPI
static int korina_poll(struct net_device *rx_data_dev, int *budget);
#else
static void korina_rx_tasklet(unsigned long rx_data_dev);
#endif
static void korina_cleanup_module(void);
static int korina_probe(int port_num);
int korina_init_module(void);


static inline void korina_abort_dma(struct net_device *dev, DMA_Chan_t ch)
{
	if (rc32434_readl(&ch->dmac) & DMAC_run_m) {
		rc32434_writel(0x10, &ch->dmac); 
		
		while (!(rc32434_readl(&ch->dmas) & DMAS_h_m))
			dev->trans_start = jiffies;		
		
		rc32434_writel(0, &ch->dmas);  
	}
	
	rc32434_writel(0, &ch->dmadptr); 
	rc32434_writel(0, &ch->dmandptr); 
}
