/*****************************************************************************
 * 
 *  USB Slave Support Software 
 *  Linux Version
 *
 *  Developed by: 
 *  Jungo Software Technologies - Residential Gateway Software Division
 *  www.jungo.com/openrg
 *  info@jungo.com
 *  
 *  
 *  Jungo Software Technologies (Jungo) and Integarated Device Technology (IDT) 
 *  jointly own all rights in the USB Slave Support Software images and source 
 *  code. 
 *  You may not, without express written permission from either Jungo or IDT,
 *  a. Distribute, sell, reproduce or copy the source code in any way.
 *  b. Rent or lease the USB Slave Support Software or otherwise transfer or 
 *  assign the right to use the USB Slave Support Software.
 *
 *
 *  
 ****************************************************************************/


#include <linux/kernel.h> 
#include <linux/module.h> 
#include <asm/system.h> 
#include <linux/stddef.h> 
#include <linux/mm.h>
#include <linux/slab.h>

#include "usb_slave_common.h"
#include "usb_slave.h"

/* Fixed sizes */
#define USB_DEV_DESC_LEN 0x12
#define USB_SETUP_TOKEN_SIZE 8
#define USB_DEFAULT_CTRL_MAX_SIZE 8

/* Control request characteristic */
/* Type */
#define TYPE_STD	0x0
#define TYPE_CLASS	0x1
#define TYPE_VENDOR	0x2
#define TYPE_RSVD	0x3

/* Control standard device requests */ 
#define GET_STATUS	0x00
#define CLEAR_FEATURE	0x01
#define SET_FEATURE	0x03
#define SET_ADDRESS	0x05
#define GET_DESCRIPTOR  0x06
#define SET_DESCRIPTOR  0x07
#define GET_CONFIGURATION 0x08
#define SET_CONFIGURATION 0x09
#define GET_INTERFACE	  0x0A
#define SET_INTERFACE 	  0x0B
#define SYNCH_FRAME	  0x0C

/* Control Communication device class requests */
#define SEND_ENCAPSULATED_COMMAND	0x00
#define GET_ENCAPSULATED_RESPONSE	0x01

typedef enum {
    ATTACHED = 1,
    POWERED = 2,
    DEFAULT = 3,
    ADDRRESS = 4,
    CONFIGURED = 5,
    SUSPENDED = 6,
} usb_device_state_t;

typedef struct {
    __u8 bmRequestType;
    __u8 bRequest;
    __u16 wValue;
    __u16 wIndex;
    __u16 wLength;
} ctrl_setup_pkt_t;

typedef struct ctrl_req_info_t {
  ctrl_setup_pkt_t setup;
  __u8  data_buff[MAX_CTRL_DATA_PKT_LEN];
  __u8  class_data_buff[MAX_GET_ENCAPSULATED_RESPONSE_DATA_LEN];
  __u32 data_count;
  __u32 left_data_count;
  __u8  send_zero_len_pkt;
  control_phase_t req_phase;
  struct ctrl_req_info_t *next;
} ctrl_req_info_t;

typedef struct {
    endpoint_t control_ep;
    ctrl_req_info_t ctrl_req;
    __u8 dev_desc_p[USB_DEV_DESC_LEN];
    __u8 *cfg_desc_p;
    int cfg_desc_len;
    volatile int *connection_state;
    string_desc_t lang_desc;
    string_desc_t mfg_desc;
    string_desc_t prod_desc;
    __u8 curr_cfg_num;
    int num_of_cfgs;
    __u8 curr_alt_if_num;
    __u8 curr_frame_num;
    __u8 curr_if_status;
    usb_device_state_t current_state;
    endpoint_t *eps_pool;
    int eps_pool_size;
    usb_device_ops *ops;
} usb_device_t;

/* Local function decleration */
static void usb_slave_generic_interrupt_handler(__u32 status, int ep_addr);
static int handle_control_ep_setup_interrupt(__u32 status);
static int usb_slave_init(void);
#ifdef MODULE
static void usb_slave_on_device_class_closed(void);
static int usb_slave_soft_init(void);
static int usb_slave_soft_uninit(void);
#endif
static int usb_slave_init_control_ep(void);
static int usb_slave_setup_eps(int is_init);
static int usb_slave_get_ep_by_addr(int ep_addr, endpoint_t **ep_data);
static int usb_slave_ep_send(__u8 *buff, int len, endpoint_t *ep, 
    void *context1, void *context2);
static void usb_slave_on_send_data_pkt(void *context);
static int usb_slave_start_send_operation(endpoint_t *ep);
static void usb_slave_on_receive_data_pkt(void *context);
static int handle_get_descriptor_req(__u8 **data_b, __u8 **setup_b, 
    int *count, __u8 *initiate_write, __u8 *special_win_pkt);
static int handle_set_configuration_req(void);
static void usb_slave_handle_control_request(void *context);
void usb_slave_on_tx_buffer_ready(endpoint_t *ep);
static control_phase_t usb_slave_on_current_ctrl_pkt(void *context);
static void usb_slave_swap_setup_token(void);

/* Stubs functions */
static void usb_handle_send_encapsulated_stub(__u8 *req_block, __u32 req_size);
static void usb_handle_get_encapsulated_stub(__u8 *resp_buf, __u32 *resp_size);
#ifdef FIX_ZERO_LEN_PKT_BUG	 
static int usb_slave_get_data_msg_len_stub(__u8 *hdr, int hdr_size);
#endif
static void usb_device_uninitialized_stub(void);
static void  get_configuration_stub(__u8 cfg_num, __u32 if_num, 
    endpoint_t **cfg, int *eps_num);
static int init_descriptors_stub(__u8 *dev_desc_p, int dev_desc_s, 
	__u8 *cfg_desc_p, int cfg_desc_s, string_desc_t *prod_desc, 
	string_desc_t *mfg_desc, string_desc_t *lang_desc); 

/* Globals */
usb_device_t usb_device;
usb_device_ops usb_ops = {
#ifdef FIX_ZERO_LEN_PKT_BUG	 
    usb_slave_get_data_msg_len_stub, /* get_data_msg_len */
#endif    
    init_descriptors_stub, /* init_descriptors */
    get_configuration_stub, /* get_configuration */
    usb_handle_send_encapsulated_stub, /* handle_send_encapsulated_cmd */
    usb_handle_get_encapsulated_stub, /* handle_get_encapsulated_resp */
    usb_device_uninitialized_stub, /* device_class_uninitialized */
    usb_slave_ep_send, /* ep_send */
    usb_slave_terminate_current_transfers, /* terminate_current_transfers */
    usb_slave_on_receive_data_pkt,
    usb_slave_on_send_data_pkt,
#ifdef MODULE    
    usb_slave_on_device_class_closed, /* usb_slave_device_class_closed */
#endif    
};

int usb_slave_module_init(void)
{
    usb_device.ops = &usb_ops;
    return 0; 
}

#ifdef MODULE
EXPORT_SYMBOL(usb_slave_on_device_class_initialized);
int init_module(void)
{
    return usb_slave_module_init(); 
}

static int usb_slave_soft_uninit(void)
{  
    unsigned long flags;
    
    save_and_cli(flags);
    usb_slave_disable_ep(&usb_device.control_ep);

   /* TODO: free all allocated memory:
    * usb_device.control_ep->abs_info.
    * usb_device.cfg_descriptor.
    * etc.
    */
    restore_flags(flags);
    return 0;
}

void cleanup_module(void)
{
   usb_slave_soft_uninit();
   usb_slave_phys_uninit();
}

static void usb_slave_on_device_class_closed(void)
{
    unsigned long flags;
    
    save_and_cli(flags);
    
    /* TODO: unmark this after fix cleanup_module */
    /* MOD_DEC_USE_COUNT;*/
    usb_device.ops = &usb_ops;
    usb_device.connection_state = NULL;
    usb_device.curr_cfg_num = 1;
    usb_device.num_of_cfgs = 1;
    usb_device.curr_alt_if_num = 0;
    usb_device.curr_frame_num = 0;
    usb_device.curr_if_status = 0;
    usb_device.eps_pool = NULL;
    usb_device.eps_pool_size = 0;
    restore_flags(flags);
}
#endif

/* Called from usb_slave_handle_control_request to set up 
 * the various endpoints upon a SET_CONFIGURATION/SET_INTERFACE request from 
 * host.  
 * Get from configuration callback the endpoint_t array.
 * Call init_eps_phys_info to setup the physical endpoints information.
 */
static int usb_slave_setup_eps(int is_init)
{
    int i;
    
    /* initialize the pool */
    usb_device.eps_pool = NULL;
    usb_device.eps_pool_size = 0;
    
    /* Get the configuration from device class layer */
    usb_device.ops->get_configuration(usb_device.curr_cfg_num, 
	usb_device.curr_alt_if_num, &usb_device.eps_pool, 
	&usb_device.eps_pool_size);
    if (usb_device.eps_pool_size <= 0 || !usb_device.eps_pool)
    {
	printk(KERN_ERR __FUNCTION__": No configuration found for "
	    "configuration num %d and interface alternate setting num %d\n", 
	    usb_device.curr_cfg_num, usb_device.curr_alt_if_num);
	return -1;
    }
    /* Setup the physical endpoints configuration */
    if (usb_slave_init_eps_phys_info(is_init ? &usb_device.control_ep : NULL, 
	usb_device.eps_pool, usb_device.eps_pool_size) < 0)
    {
	printk(KERN_ERR __FUNCTION__": failed initializing physical endpoint\n"); 
	return -1;
    }
    if (!is_init)
    {
	/* Enable the endpoints in the hardware */
	for (i = 0; i < usb_device.eps_pool_size; i++)
	{
	    endpoint_t *ep = usb_device.eps_pool + i;
	    usb_slave_enable_ep(ep);
	    ep->abs_info->mode = USB_ENABLED;
            ep->abs_info->transfer_mode = USB_ACK;
	}
    }
    return 0;
}

/* Initialize the control endpoint abstract information */
static int usb_slave_init_control_ep(void)
{
    usb_device.control_ep.abs_info = (abstract_ep_t *)kmalloc(
	sizeof(abstract_ep_t), GFP_ATOMIC);
    if (!usb_device.control_ep.abs_info)
    {
	printk(KERN_ERR __FUNCTION__": failed allocating memory\n");
	return -1;
    }
    memset(usb_device.control_ep.abs_info, 0, sizeof(abstract_ep_t));
    usb_device.control_ep.abs_info->type = RG_USB_CONTROL;
    usb_device.control_ep.abs_info->direction = RG_USB_BIDIR;
    usb_device.control_ep.abs_info->addr = 0;
    usb_device.control_ep.abs_info->mode = USB_DISABLED;
    usb_device.control_ep.abs_info->max_pkt_size = USB_DEFAULT_CTRL_MAX_SIZE;
    usb_device.control_ep.abs_info->buff = NULL;
    usb_device.control_ep.abs_info->buff_pos = 0;
    usb_device.control_ep.abs_info->buff_size = 0;
    usb_device.control_ep.abs_info->is_busy = 0;
#ifdef CONFIG_USB_SIMPLE_DMA    
    usb_device.control_ep.abs_info->is_dma = 0;
#else    
    usb_device.control_ep.abs_info->is_dma = 1;
#endif    
    usb_device.control_ep.abs_info->ep_operation = 
	usb_slave_on_current_ctrl_pkt ;
    usb_device.control_ep.abs_info->end_func = 
	usb_slave_handle_control_request;
    usb_device.ctrl_req.req_phase = NOT_STARTED;
    usb_device.ctrl_req.send_zero_len_pkt = 0;
    usb_device.ctrl_req.data_count = 0;
    usb_device.ctrl_req.left_data_count = 0;
    memset(&usb_device.ctrl_req.setup, 0, sizeof(ctrl_setup_pkt_t));
    memset(&usb_device.ctrl_req.data_buff, 0, MAX_CTRL_DATA_PKT_LEN);
    memset(&usb_device.ctrl_req.class_data_buff, 0,
	MAX_GET_ENCAPSULATED_RESPONSE_DATA_LEN);
    return 0;
}

static int usb_slave_soft_init(void)
{
   if (usb_slave_init_control_ep() < 0)
   {
       printk(KERN_ERR __FUNCTION__": failed initializing control endpoint\n");
       return -1;
   }
   if (usb_slave_setup_eps(1) < 0)
       return -1;
   usb_device.cfg_desc_p  = (__u8 *)kmalloc(usb_device.cfg_desc_len,
       GFP_ATOMIC);
   if (!usb_device.cfg_desc_p)
   {
       printk(KERN_ERR __FUNCTION__":failed allocating memory\n");
       return -1;
   }
   if (usb_device.ops->init_descriptors(usb_device.dev_desc_p, 
       USB_DEV_DESC_LEN, usb_device.cfg_desc_p, usb_device.cfg_desc_len,  
       &usb_device.prod_desc, &usb_device.mfg_desc, &usb_device.lang_desc) < 0)
   {
       return -1;
   }
   return 0;
}

static int usb_slave_init(void)
{
   if ((usb_slave_soft_init() < 0) || 
       (usb_slave_phys_init(usb_slave_generic_interrupt_handler) < 0))
   {    
       goto Error;
   }
   return 0;
Error:
   if (usb_device.control_ep.abs_info)
   {
       kfree(usb_device.control_ep.abs_info);
       usb_device.control_ep.abs_info = NULL;
   }
   if (usb_device.cfg_desc_p)
   {
       kfree(usb_device.cfg_desc_p);
       usb_device.cfg_desc_p = NULL;
   }
   return -1;
}

int usb_slave_on_device_class_initialized(usb_device_ops *ops, int num_of_cfgs, 
     int cfg_desc_len, volatile int *is_device_connected)
{
    unsigned long flags;
    int rc = -1;
    
    save_flags(flags);
    cli();
    
    usb_device.connection_state = is_device_connected;
    if (!ops)
    {
	printk(KERN_ERR __FUNCTION__"upper layer api is empty\n");
	goto Exit;
    }
    usb_device.ops = ops;
    ops->terminate_current_transfers = usb_slave_terminate_current_transfers;
    ops->on_receive_usb_data_pkt = usb_slave_on_receive_data_pkt;
    ops->on_send_usb_data_pkt = usb_slave_on_send_data_pkt;
    ops->ep_send = usb_slave_ep_send;
#ifdef MODULE    
    ops->on_device_class_closed = usb_slave_on_device_class_closed;
#endif 
    usb_device.cfg_desc_len = cfg_desc_len;
    usb_device.num_of_cfgs = num_of_cfgs;
    usb_device.curr_cfg_num = 1;
    usb_device.eps_pool = NULL;
    if (usb_slave_init() < 0)
	goto Exit;
#ifdef MODULE    
    MOD_INC_USE_COUNT;
#endif    
    rc = 0;
Exit: 
    restore_flags(flags);    
    return rc;
}

static int handle_set_configuration_req(void)
{
    int is_allready_called = 0;
    
    /* Check validity */
    if(usb_device.ctrl_req.setup.wValue > usb_device.num_of_cfgs)
    {
	printk(KERN_ERR __FUNCTION__": bad configuration number\n"); 
	return -1;
    }
    if (!is_allready_called || 
	(usb_device.curr_cfg_num != usb_device.ctrl_req.setup.wValue))
    {
	 is_allready_called = 1;
         usb_device.curr_cfg_num = usb_device.ctrl_req.setup.wValue;
         usb_device.curr_alt_if_num = 0;
	 if (usb_slave_setup_eps(0) < 0)
	     return -1;
    }
    usb_device.ctrl_req.send_zero_len_pkt = 1;
    return 0;
}

static int handle_get_descriptor_req(__u8 **data_b, __u8 **setup_b, 
    int *count, __u8 *initiate_write, __u8 *special_win_pkt)
{
    switch ((usb_device.ctrl_req.setup.wValue & 0x0f00) >> 8)
    {
    case DEVICE_DESC:
        USB_DPRINTK(__FUNCTION__": got DEV_DESC request\n");
	*setup_b = &(usb_device.dev_desc_p[0]);
	*data_b = &(usb_device.ctrl_req.data_buff[0]);
	*count = usb_device.ctrl_req.setup.wLength != USB_DEV_DESC_LEN ?
	    USB_DEFAULT_CTRL_MAX_SIZE : USB_DEV_DESC_LEN;
	*initiate_write = 1;
	if ((*count) == USB_DEFAULT_CTRL_MAX_SIZE)
	    *special_win_pkt = 1;
	break;
    case CONFIGURATION_DESC:
        USB_DPRINTK(__FUNCTION__": got CONF_DESC request\n");
	if(usb_device.ctrl_req.setup.wValue & 0xf)
	{
	    usb_device.control_ep.abs_info->transfer_mode = USB_STALL;
	    break;
	}
	*setup_b = &(usb_device.cfg_desc_p[0]);
	*data_b = &(usb_device.ctrl_req.data_buff[0]);
	*count = (usb_device.ctrl_req.setup.wLength < 
	    usb_device.cfg_desc_len) ? 
	    usb_device.ctrl_req.setup.wLength : 
	    usb_device.cfg_desc_len;
	*initiate_write = 1;
	break;
    case STRING_DESC :
        USB_DPRINTK(__FUNCTION__": got STRING_DESC request\n");
	
	/* language descriptor */
	if (!(usb_device.ctrl_req.setup.wValue & 0xff))
	{
            USB_DPRINTK(__FUNCTION__": got language string request\n");
	    *setup_b = (__u8 *)&usb_device.lang_desc;
	    *data_b = &usb_device.ctrl_req.data_buff[0];
	    *count = (usb_device.ctrl_req.setup.wLength 
		< usb_device.lang_desc.length) ? 
		usb_device.ctrl_req.setup.wLength : 
		usb_device.lang_desc.length; 
	    *initiate_write = 1;
	    break;
	}

	/* manufacturer descriptor */
	if ((usb_device.ctrl_req.setup.wValue & 0xff) == 
	    usb_device.dev_desc_p[14])
	{
            USB_DPRINTK(__FUNCTION__": got manufacturer string request\n");
	    *setup_b = (__u8 *)(&usb_device.mfg_desc);
	    *data_b = &usb_device.ctrl_req.data_buff[0];
	    *count = (usb_device.ctrl_req.setup.wLength < 
		usb_device.mfg_desc.length) ?  
		usb_device.ctrl_req.setup.wLength : 
		usb_device.mfg_desc.length;
	    *initiate_write = 1;
	    break;
	}

	/* product descriptor */
	if ((usb_device.ctrl_req.setup.wValue & 0xff) == 
	    usb_device.dev_desc_p[15])
	{
            USB_DPRINTK(__FUNCTION__": got product string request\n");
	    *setup_b = (__u8 *)(&usb_device.prod_desc);
	    *data_b = &usb_device.ctrl_req.data_buff[0];
	    *count = (usb_device.ctrl_req.setup.wLength < 
		usb_device.prod_desc.length) ?
		usb_device.ctrl_req.setup.wLength : 
		usb_device.prod_desc.length;
	    *initiate_write = 1;
	    break;
	}
	break;
    default :
	printk(KERN_ERR __FUNCTION__": Can't get this type of descriptor "
	    "independently\n");
	return -1;	
    }
    return 0;
}

static int handle_ctrl_device_request(__u8 **data_b, __u8 **setup_b, 
    int *count, __u8 *initiate_write, __u8 *special_win_pkt)
{
    switch (usb_device.ctrl_req.setup.bRequest)
    {
    case SET_ADDRESS:
	if (usb_slave_handle_set_address_req(
	    usb_device.ctrl_req.setup.wValue) < 0)
	{
	    return -1;
	}
	break;
    case GET_DESCRIPTOR:
	USB_DPRINTK("Got GET_DESCRIPTOR message\n");
	if (handle_get_descriptor_req(data_b, setup_b, count, 
	    initiate_write, special_win_pkt) < 0)
	{
	    return -1;
	}
	break;
    case SET_CONFIGURATION:
	USB_DPRINTK(__FUNCTION__" Got SET_CONFIGURATION\n");
	*data_b = &usb_device.ctrl_req.data_buff[0];
	if (handle_set_configuration_req() < 0)
	    return -1;
	break;
    case CLEAR_FEATURE:
	/* wValue is not appropriate to recipient */
	if (usb_device.ctrl_req.setup.wValue != RG_DEVICE_REMOTE_WAKEUP || 
	    (usb_slave_handle_clear_feature_req(
	    usb_device.ctrl_req.setup.wValue, NULL) < 0))
	{
	    return -1;
	}		
	break;
    case SET_FEATURE:
	/* wValue is not appropriate to recipient */
	if (usb_device.ctrl_req.setup.wValue != RG_DEVICE_REMOTE_WAKEUP || 
	    (usb_slave_handle_set_feature_req(
	    usb_device.ctrl_req.setup.wValue, NULL) < 0))
	{
	    return -1;
	}		
	break;
    case GET_STATUS:
	USB_DPRINTK(__FUNCTION__" Got GET_STATUS for device "
	    "recipient\n");
	usb_slave_handle_get_status_req(RCPT_DEV, NULL, setup_b);
	if (usb_device.ctrl_req.setup.wValue ||
	    (usb_device.ctrl_req.setup.wLength != 2)) /* __u16 */
	{
	    return 0;
	}
	*data_b = &usb_device.ctrl_req.data_buff[0];
	*initiate_write = 1;
	*count = usb_device.ctrl_req.setup.wLength;
	break;
    case GET_CONFIGURATION :
	USB_DPRINTK(__FUNCTION__" Got GET_CONFIGURATION\n");
	*setup_b = &usb_device.curr_cfg_num;
	*data_b = &usb_device.ctrl_req.data_buff[0];
	*count = 1;
	*initiate_write = 1;
	break;
    case SET_DESCRIPTOR :
	USB_DPRINTK(__FUNCTION__" Got SET_DESCRIPTOR\n");
	*data_b = &usb_device.ctrl_req.data_buff[0];
	if ((usb_device.ctrl_req.setup.wValue  & 0xff00)
	    != 0x0100)
	{
	    return -1;
	}

	/* Setup phase of OUT packet */
	if (usb_device.ctrl_req.req_phase == SETUP_TOKEN)
	{
	    usb_device.ctrl_req.left_data_count = 
		usb_device.ctrl_req.setup.wLength;
	    usb_device.ctrl_req.data_count = 0;
	    break;
	}

	/* Data phase, after getting all data from host, so copy the
	 * data and break in order to return zero packet length to 
	 * host.
	 */
	memcpy(usb_device.dev_desc_p, usb_device.ctrl_req.data_buff, 
	    USB_DEV_DESC_LEN);
	usb_device.ctrl_req.send_zero_len_pkt = 1;
	break;
    default:
	printk(KERN_ERR __FUNCTION__": Unknown standard device request, or "
	    "request should be handled by hardware;\n bRequest = %d\n",
	    usb_device.ctrl_req.setup.bRequest);
	return -1;
    }
    return 0;
}

static int handle_ctrl_interface_request(__u8 ** data_b, __u8 **setup_b, 
    int *count, __u8 *initiate_write, __u8 *special_win_pkt)
{
    USB_DPRINTK(__FUNCTION__" Got RCP_IF\n");
    switch(usb_device.ctrl_req.setup.bRequest)
    {
    case SET_INTERFACE:
	USB_DPRINTK(__FUNCTION__" Got SET_INTERFACE\n");

	/* TODO Update configuration and interface descriptors 
	 * and what ever needed 
	 */

	usb_device.ctrl_req.send_zero_len_pkt = 1;
	break;
    case GET_INTERFACE:
	USB_DPRINTK(__FUNCTION__" Got GET_INTERFACE\n");
	*setup_b = &usb_device.curr_alt_if_num;
	*data_b = &usb_device.ctrl_req.data_buff[0];
	*count = 1;
	*initiate_write = 1;
	break;
    case CLEAR_FEATURE:
	USB_DPRINTK(__FUNCTION__" Got CLEAR_FEATURE\n");
        return -1;
    case SET_FEATURE:
	USB_DPRINTK(__FUNCTION__" Got SET_FEATURE\n");
	return -1;
    case GET_STATUS:
	USB_DPRINTK(__FUNCTION__" Got GET_STATUS for interface %d "
	    "recipient\n",usb_device.ctrl_req.setup.wIndex);
	usb_slave_handle_get_status_req(RCPT_IF, NULL, setup_b);
	if (usb_device.ctrl_req.setup.wValue ||
	    (usb_device.ctrl_req.setup.wLength != 2)) /* __u16 */
	{
	    return -1;
	}
	*data_b = &usb_device.ctrl_req.data_buff[0];
	*initiate_write = 1;
	*count = usb_device.ctrl_req.setup.wLength;
	break;
    default:
	printk(KERN_ERR __FUNCTION__": Unknown standard interface request, "
	    "bRequest = %d\n", usb_device.ctrl_req.setup.bRequest);
	return -1;
    }
    return 0;
}

static int handle_ctrl_endpoint_request(__u8 **data_b, __u8 **setup_b, 
    int *count, __u8 *initiate_write, __u8 *special_win_pkt)
{
    endpoint_t *rec_ep;

    switch(usb_device.ctrl_req.setup.bRequest)
    {
    case  SYNCH_FRAME:
	USB_DPRINTK(__FUNCTION__": Got SYNC_FRAME\n");
	*setup_b = &usb_device.curr_frame_num;
	*data_b = &usb_device.ctrl_req.data_buff[0];
	*count = 2;
	*initiate_write = 1;
	break;
    case CLEAR_FEATURE:
	usb_slave_get_ep_by_addr(usb_device.ctrl_req.setup.wIndex, 
	    &rec_ep);

	/* wValue is not appropriate to recipient */
	if (usb_device.ctrl_req.setup.wValue || !rec_ep || 
	    (usb_slave_handle_clear_feature_req(
	    usb_device.ctrl_req.setup.wValue, rec_ep) < 0))
	{
	    return -1;
	}		
	break;
    case SET_FEATURE:
	usb_slave_get_ep_by_addr(usb_device.ctrl_req.setup.wIndex, 
	    &rec_ep);

	/* wValue is not appropriate to recipient */
	if (usb_device.ctrl_req.setup.wValue || !rec_ep || 
	    (usb_slave_handle_set_feature_req(
	    usb_device.ctrl_req.setup.wValue, rec_ep) < 0))
	{
	    return -1;
	}		
	break;
    case GET_STATUS:
	USB_DPRINTK(__FUNCTION__" Got GET_STATUS for endpoint %d "
	    "recipient\n", usb_device.ctrl_req.setup.wIndex);
	usb_slave_get_ep_by_addr(usb_device.ctrl_req.setup.wIndex, 
	    &rec_ep);
	usb_slave_handle_get_status_req(RCPT_EP, rec_ep, setup_b);
	if (usb_device.ctrl_req.setup.wValue ||
	    ((usb_device.ctrl_req.setup.wLength != 2))) /* __u16 */
	{
	    return -1;
	}
	*data_b = &usb_device.ctrl_req.data_buff[0];
	*initiate_write = 1;
	*count = usb_device.ctrl_req.setup.wLength;
	break;
    default:
	printk(KERN_ERR __FUNCTION__": Unknown standard endpoint request, "
	    "bRequest = %d\n", usb_device.ctrl_req.setup.bRequest);
	return -1;
    }
    return 0;
}

static int handle_ctrl_standard_request(__u8 req_recipient, __u8 **data_b, 
    __u8 **setup_b, int *count, __u8 *initiate_write, __u8 *special_win_pkt)
{
    switch (req_recipient)
    {
    /* General request - recipient = device */
    case RCPT_DEV:
	if (handle_ctrl_device_request(data_b, setup_b, count, initiate_write,
	    special_win_pkt) < 0)
	{
	    return -1;
	}
	break;

    /* Request for a specific interface - recipient = interface */
    case RCPT_IF:
	if (handle_ctrl_interface_request(data_b, setup_b, count, 
	    initiate_write, special_win_pkt) < 0)
	{
	    return -1;
	}
	break;

    /* Request for a specific endpoint - recipient = endpoint */
    case RCPT_EP:
	if (handle_ctrl_endpoint_request(data_b, setup_b, count, 
	    initiate_write, special_win_pkt) < 0)
	{
	    return -1;
	}
	break;
    case RCPT_OTHER:
	printk(KERN_INFO __FUNCTION__": Nothing to do for recipient OTHER "
	    "request\n");
	break;
    default:
	printk(KERN_WARNING __FUNCTION__": Unknown standard request "
	    "recipient, recipient num  = %d\n", req_recipient);
	return -1;
    }
    return 0;
}

/* translate setup token fileds from Little-endian to cpu byte order.
 * (USB messages are in Little-endian order).
 */
static void usb_slave_swap_setup_token(void)
{
    usb_device.ctrl_req.setup.wValue = 
	__le16_to_cpu(usb_device.ctrl_req.setup.wValue);
    usb_device.ctrl_req.setup.wIndex = 
        __le16_to_cpu(usb_device.ctrl_req.setup.wIndex);
    usb_device.ctrl_req.setup.wLength = 
        __le16_to_cpu(usb_device.ctrl_req.setup.wLength);
}

/* Called when a setup or data packet arrive at the control endpoint to 
 * handle the request.
 * To use as control endpoint ep_operation call-back.
 * It will be called in two stages: 
 * 1. control SETUP token has arrived
 * 2. control data of data OUT phase has been "finished to arrive"
 */
static void usb_slave_handle_control_request(void *context)
{
    __u8 req_direction, req_type, req_recipient;
    __u8 *setup_b = NULL;
    __u8 *data_b = NULL;
    __u8 initiate_write = 0, special_win_pkt = 0;
    int count = 0;

    if (usb_device.ctrl_req.req_phase == NOT_STARTED)
	return;
    
#ifdef __BIG_ENDIAN
    /* Do the swap only once for each setup packet */
    if (usb_device.ctrl_req.req_phase == SETUP_TOKEN)
	usb_slave_swap_setup_token();
#endif    
    
    /* Parse setup token bmRequestType to get request type, direction and 
     * recipient
     */
    req_recipient = usb_device.ctrl_req.setup.bmRequestType & 0x1f; 
    req_type = (usb_device.ctrl_req.setup.bmRequestType >> 5) & 0x3; 
    req_direction = (usb_device.ctrl_req.setup.bmRequestType >> 7) & 0x1; 
    switch (req_type)
    {
    /* Standard request */
    case TYPE_STD:
	if (handle_ctrl_standard_request(req_recipient, &data_b, &setup_b, 
	    &count, &initiate_write, &special_win_pkt) < 0)
	{
	    usb_device.control_ep.abs_info->transfer_mode = USB_STALL;
	    goto Error;
	}
	break;
   
    /* Request of specific device class */
    case  TYPE_CLASS:
	if ((req_recipient == RCPT_IF) && (req_direction == RG_USB_OUT) && 
	    (usb_device.ctrl_req.setup.bRequest == SEND_ENCAPSULATED_COMMAND)) 
	{
	   USB_DPRINTK(__FUNCTION__": SEND_ENCAPSULATED_CMD\n");
	    
	    /* Setup phase */
	    if (usb_device.ctrl_req.req_phase == SETUP_TOKEN)
	    {
		/* bytes left to read */
		usb_device.ctrl_req.left_data_count = 
		    usb_device.ctrl_req.setup.wLength;
		
		/* bytes that was read */
		usb_device.ctrl_req.data_count = 0;
	        data_b = &usb_device.ctrl_req.data_buff[0];	
		break;
	    }
	    
	    /* Second call - data phase */
	    /* Put the data in the response queue */
	    usb_device.ops->handle_send_encapsulated_cmd(
		&usb_device.ctrl_req.data_buff[0], 
		usb_device.control_ep.abs_info->buff_pos);
	    
	    /* send ack to host */
	    usb_device.ctrl_req.send_zero_len_pkt = 1; 
	    break;
	}
	if ((req_recipient == RCPT_IF) && (req_direction == RG_USB_IN) && 
	    (usb_device.ctrl_req.setup.bRequest == GET_ENCAPSULATED_RESPONSE)) 
	{
	    /* XXX - Changed from USB_DPRINTK to printk in order to avoid
	     * kinpo-samsung weird problems. See B2958. */
	    USB_DPRINTK(__FUNCTION__": GET_ENCAPSULATED_RESP\n");
	    
	    /* No data is sent so it is called only once */
	    /* Take the first response from the queue and put in 
	     * usb_device.ctrl_req.class_data_buff)
	     */
	    data_b = &usb_device.ctrl_req.class_data_buff[0]; /* The response */
	    usb_device.ops->handle_get_encapsulated_resp(data_b, &count);
	    initiate_write = 1;
	    break;
	}
	break;

	/* Specific vendor Request */
    case TYPE_VENDOR:
	printk(KERN_WARNING __FUNCTION__": Not supporting VENDOR control "
	    "request\n");
	usb_device.ctrl_req.req_phase = NOT_STARTED;
	return;
	/* TODO:  add handle_vendor_request to slave api */
    case TYPE_RSVD:
	printk(KERN_WARNING __FUNCTION__": Not supporting RESERVED control "
	    "request\n");
	usb_device.ctrl_req.req_phase = NOT_STARTED;
	return;
    default:
	printk(KERN_WARNING __FUNCTION__": Unknown request type %d\n", 
	    req_type); 
	usb_device.ctrl_req.req_phase = NOT_STARTED;
	return;
    }

    /* We need to send some response to host */
    if (initiate_write)
    {
	if (data_b && setup_b)
	    memcpy(data_b, setup_b, count);

	/* At the end of transfer we need to send one more zero length to
	 * indicate end of transfer
	 */
	if (!(count % usb_device.control_ep.abs_info->max_pkt_size) && 
	    (count != usb_device.ctrl_req.setup.wLength) && !special_win_pkt)
	{
	    usb_device.ctrl_req.send_zero_len_pkt = 1;
	}
	usb_device.ctrl_req.left_data_count = count;
	usb_device.ctrl_req.data_count = 0;
    }
    /* Setup phase */
    if (usb_device.ctrl_req.req_phase == SETUP_TOKEN)
    {
	usb_device.control_ep.abs_info->buff = data_b;
	usb_device.control_ep.abs_info->buff_pos = 0;
	usb_device.control_ep.abs_info->buff_size = 
	    usb_device.ctrl_req.left_data_count;

	/* There is data to transmit/receive */
	if (usb_device.ctrl_req.left_data_count > 0)
	{
	    /* There is data to transmit */
	    if (req_direction == RG_USB_IN)
		usb_device.ctrl_req.req_phase = DATA_TRANSMIT;

	    /* Expect data from host */
	    else if (req_direction == RG_USB_OUT)
	    {
		usb_device.ctrl_req.req_phase = DATA_RECEIVE;
		return;
	    }
	}

	/* No data to transmit/recieve, see if zero lengh ack is needed */
	else
	{
	    /* Nothing to transmit */
	    if (!usb_device.ctrl_req.send_zero_len_pkt)
	    {
		usb_device.ctrl_req.req_phase = NOT_STARTED;
		return;
	    }

	    /* Only zero length packet to transmit */
	    usb_device.ctrl_req.req_phase = DATA_SEND_ZERO_LEN_PKT;
	    usb_device.ctrl_req.send_zero_len_pkt = 0;
	}

	/* On some architectures there is no interrupt when DMA
	 * Controller has sent zero length packet, so we are updating
	 * state before the actual transmit.
	 */
	if (usb_device.control_ep.abs_info->is_dma)
	    usb_device.ctrl_req.req_phase = NOT_STARTED;

	/* Start data transfer */
	usb_slave_start_send_operation(&usb_device.control_ep);
	return;
    }

    /* Finished receiving data OUT packet (data phase) */
    if (usb_device.ctrl_req.req_phase == DATA_RECEIVE)
    {
	/* Receive ended and no need to send ack */
	if (!usb_device.ctrl_req.send_zero_len_pkt)
	{
	    usb_device.ctrl_req.req_phase = NOT_STARTED;
	    return;
	}
	
	/* Need to send ack */
	usb_device.ctrl_req.req_phase = DATA_SEND_ZERO_LEN_PKT;
	usb_device.ctrl_req.send_zero_len_pkt = 0;
	
	/* On some architectures ther is no interrupt when DMA
	 * Controler has send zero length packet, so we are updating
	 * state before the actual transmit.
	 */
	if (usb_device.control_ep.abs_info->is_dma)
	    usb_device.ctrl_req.req_phase = NOT_STARTED;
	
	/* Start ack transfer */
	usb_slave_start_send_operation(&usb_device.control_ep);
    }
    return;
Error:    
    /* We need to stall control endpoint */
    if (usb_device.control_ep.abs_info->transfer_mode == USB_STALL)
    {
	printk(KERN_INFO __FUNCTION__"control endpoint stalled\n");
	usb_slave_stall_ep(&usb_device.control_ep);
	usb_device.control_ep.abs_info->transfer_mode = USB_ACK;
	usb_device.ctrl_req.req_phase = NOT_STARTED;
    }
}

static control_phase_t usb_slave_on_current_ctrl_pkt(void *context)
{
    endpoint_t *ep = (endpoint_t *)context;
    static int received_count = 0;
    int current_pkt_size = 0;
    
    if(!ep || !ep->abs_info || (ep->abs_info->type != RG_USB_CONTROL) ||
	(ep->abs_info->direction != RG_USB_BIDIR))
    {
	printk(KERN_ERR __FUNCTION__": Invalid endpoint\n");
	goto Exit;
    }

    if (usb_device.ctrl_req.req_phase == NOT_STARTED)
	goto Exit;
    if (usb_device.ctrl_req.req_phase == DATA_SEND_ZERO_LEN_PKT)
    {
	usb_device.ctrl_req.req_phase = NOT_STARTED;
	goto Exit;
    }	

    /* Data IN phase */
    if (usb_device.ctrl_req.req_phase == DATA_TRANSMIT)
    {
	/* Not end of transfer */
	if (ep->abs_info->buff_size)
	{
	    usb_slave_start_send_operation(&usb_device.control_ep);
	    goto Exit;
	}
	
	/* Finished data transmit */
        ep->abs_info->buff_pos = 0;

	/* No need for zero length packet */
	if (!usb_device.ctrl_req.send_zero_len_pkt)
	{
	    usb_device.ctrl_req.req_phase = NOT_STARTED;
	    goto Exit;
	}

	/* Need to send zero length packet */
	usb_device.ctrl_req.req_phase = DATA_SEND_ZERO_LEN_PKT;
	usb_device.ctrl_req.send_zero_len_pkt = 0;
        if (ep->abs_info->is_dma)
	    usb_device.ctrl_req.req_phase = NOT_STARTED;
	usb_slave_start_send_operation(&usb_device.control_ep);
	goto Exit;
    }

    /* Data OUT phase */
    if (ep->abs_info->buff_pos < received_count)
    {
	printk(KERN_ERR __FUNCTION__": Invalid possition in endpoint buffer\n");
	goto Exit;
    }
    current_pkt_size = ep->abs_info->buff_pos - received_count;
    received_count = ep->abs_info->buff_pos;
    
    /* End of transfer: a packet of size less then max_pkt_size or transfer
     * of current control transfer size.
     */
    if (!ep->abs_info->buff_size ||
	current_pkt_size < ep->abs_info->max_pkt_size)
    {
	received_count = 0;
	ep->abs_info->end_func((void *)ep);
	ep->abs_info->buff_pos = 0;
    }
Exit:
    return usb_device.ctrl_req.req_phase;
}

/* Called when USB bulk data packet arrived at a BULK-OUT endpoint.
 * To use as BULK-OUT endpoint ep_operation call-back.
 */
static void usb_slave_on_receive_data_pkt(void *context)
{
    endpoint_t *ep = (endpoint_t *)context;
    static int received_count = 0;
    int current_pkt_size = 0;
#ifdef FIX_ZERO_LEN_PKT_BUG	 
    static int device_msg_size = 0; 
#endif    
    
    *(usb_device.connection_state) = USB_CONNECTED;
    if(!ep || !ep->abs_info || (ep->abs_info->direction != RG_USB_OUT))
    {
	printk(KERN_ERR __FUNCTION__": Invalid endpoint\n");
        received_count = 0;
	return;
    }
    /* Data packet */
    if (ep->abs_info->buff_pos < received_count)
    {
	printk(KERN_ERR __FUNCTION__": Invalid possition in endpoint "
	    "buffer\n");
	received_count = 0;
#ifdef FIX_ZERO_LEN_PKT_BUG	 
	device_msg_size = 0;
#endif    
	ep->abs_info->buff_pos = 0;
	return;
    }
    current_pkt_size = ep->abs_info->buff_pos - received_count;
    
    /* XXX: device call-back for message size is a temporary patch to overcome
     * Conexant's UDC problem of non initiating interrupt on zero length 
     * packet.
     */
#ifdef FIX_ZERO_LEN_PKT_BUG	 
    if (!received_count && current_pkt_size)
    {
        device_msg_size = usb_device.ops->get_data_msg_len(ep->abs_info->buff, 
	    current_pkt_size);
	if (device_msg_size < 0)
	{
	    /* "Leftovers" from last device packet, no need to transfer to 
	     * upper layer.
	     */
	    received_count = 0;
	    device_msg_size = 0;
	    ep->abs_info->buff_pos = 0;
	    return;
	}
    }
#endif    
    received_count = ep->abs_info->buff_pos;
    
    /* End of transfer: a packet of size less then max_pkt_size or transfer
     * of size max_transfer_size.
     */
    if (!ep->abs_info->buff_size ||
	current_pkt_size < ep->abs_info->max_pkt_size  
#ifdef FIX_ZERO_LEN_PKT_BUG	 
	|| received_count >= device_msg_size
#endif	
	)
    {
	received_count = 0;
#ifdef FIX_ZERO_LEN_PKT_BUG	 
	device_msg_size = 0;
#endif	
	ep->abs_info->buff_size = MAX_ALLOWED_TRANSFER_SIZE;
	ep->abs_info->end_func((void *)ep);
	ep->abs_info->buff_pos = 0;
    }
}

void usb_slave_on_tx_buffer_ready(endpoint_t *ep)
{
    if (!ep || !ep->abs_info)
	return;
    
    /* Transfer is interrupt driven */
    if (ep->abs_info->is_interrupt & USB_INTERRUPT_ON_TX_START_PKT)
    {
	usb_slave_enable_transfer(ep);
	return;
    }
    if (!ep->abs_info->addr)
    {
	__u8 dir = 
	    ((usb_device.ctrl_req.setup.bmRequestType >> 7) & 0x1);
	usb_slave_handle_control_ep(&usb_device.control_ep, dir, 
	    usb_device.ctrl_req.req_phase);
	return;
    }
    usb_slave_handle_non_control_ep(ep, 
	ep->abs_info->buff + ep->abs_info->buff_pos, (ep->abs_info->buff_size 
	<= ep->abs_info->max_pkt_size) ? ep->abs_info->buff_size : 
	ep->abs_info->max_pkt_size);
}


/* Called when current USB BULK data packet on a BULK-IN endpoint had been
 * transfered to indicate that a new USB packet can be sent (only in non DMA 
 * state). To use as a BULK-IN endpoint ep_operation callback.
 * Only when end_func is called, a new upper layer packet can be sent, so no
 * problem of mutual exlution of transfer information of endpoint (buff, 
 * buff_pos, buff_size, is_busy etc.)
 */
static void usb_slave_on_send_data_pkt(void *context)
{
    endpoint_t *ep = (endpoint_t *)context;
    static int send_zero_len_pkt = 0;
    
    if (!ep || !ep->abs_info || ep->abs_info->direction != RG_USB_IN)
    {
	printk(KERN_ERR __FUNCTION__": Invalid endpoint\n");
	return;
    }
    /* We are not inside a transfer */
    if (!ep->abs_info->is_busy)
    {
	ep->abs_info->end_arg1 = USB_TRANSFER_ERROR;
        goto Exit;
    }
    
    /* Not end of transfer */
    if (ep->abs_info->buff_size)
    {
	/* Transfer size is devided into endpoint max_packet_size, so according
	 * to protocol we need to send another zero length packet
	 */
        if (ep->abs_info->buff_size == ep->abs_info->max_pkt_size)
	    send_zero_len_pkt = 1;
        usb_slave_on_tx_buffer_ready(ep);
	return;
    }
    /* End of transfer and we need to send zero length packet */
    if (send_zero_len_pkt)
    {
	usb_slave_handle_non_control_ep(ep, ep->abs_info->buff, 0);
	send_zero_len_pkt = 0;
    }
    ep->abs_info->end_arg1 = USB_TRANSFER_SUCCESS;
Exit:  
    /* End of transfer */
    if (ep->abs_info->is_interrupt & USB_INTERRUPT_ON_TX_START_PKT)
	usb_slave_disable_transfer(ep);
    ep->abs_info->end_func(ep);
}

static int usb_slave_start_send_operation(endpoint_t *ep)
{
    ep->abs_info->is_busy = 1;
    if (ep->abs_info->is_dma)
	return usb_slave_start_dma_transfer(ep);
    ep->abs_info->end_arg1 = USB_TRANSFER_ERROR;
    usb_slave_on_tx_buffer_ready(ep); 
    return 0;
}


/* Called from upper layer (Remote NDIS) to send data through USB 
 * or to send interrupt endpoint transfer
 */
static int usb_slave_ep_send(__u8 *buff, int len, endpoint_t *ep, 
    void *context1, void *context2)
{

    if (!ep || !ep->abs_info || (ep->abs_info->direction != RG_USB_IN))
    {
	printk(KERN_ERR __FUNCTION__": invalid endpoint\n");
	return -1;
    }
    if (ep->abs_info->is_busy)
    {
	printk(KERN_ERR __FUNCTION__": endpoint is busy\n");
	return -1;
    }
     
    /* Current buffer pointer to free at the end of transfer */
    ep->abs_info->end_context1 = context1; 
    
    /* Not in use at the moment */
    ep->abs_info->end_context2 = context2; 
    
    /* Transfer status */
    ep->abs_info->end_arg1 = USB_TRANSFER_SUCCESS;

   /* Current buffer length for statistics at the end */
    ep->abs_info->end_arg2 = len; 

    /* Transfer information */
    ep->abs_info->buff = buff;
    ep->abs_info->buff_size = len;
    ep->abs_info->buff_pos = 0;
    return (usb_slave_start_send_operation(ep));
}

static int handle_control_ep_setup_interrupt(__u32 status)
{
   if (usb_device.ctrl_req.req_phase == SETUP_TOKEN)
       return 0;
   if (usb_device.ctrl_req.req_phase > SETUP_TOKEN)
    {
	printk(KERN_WARNING __FUNCTION__": Host sent pre-mature control "
	    "request, req_phase num = %d\n", usb_device.ctrl_req.req_phase);
	usb_slave_stall_ep(&usb_device.control_ep);
	usb_device.control_ep.abs_info->mode = USB_ENABLED;
	usb_device.ctrl_req.req_phase = NOT_STARTED;
	return -1;
    }
    usb_device.ctrl_req.req_phase = SETUP_TOKEN;
    usb_device.ctrl_req.send_zero_len_pkt = 0;
    usb_device.ctrl_req.data_count = 0;
    usb_device.ctrl_req.left_data_count = 0;
    memset(&usb_device.ctrl_req.setup, 0, sizeof(ctrl_setup_pkt_t));
    memset(&usb_device.ctrl_req.data_buff, 0, MAX_CTRL_DATA_PKT_LEN);
    memset(&usb_device.ctrl_req.class_data_buff, 0,
	MAX_GET_ENCAPSULATED_RESPONSE_DATA_LEN);
    usb_device.control_ep.abs_info->buff = 
	(__u8 *)(&usb_device.ctrl_req.setup);
    usb_device.control_ep.abs_info->buff_pos = 0;
    usb_device.control_ep.abs_info->buff_size = USB_SETUP_TOKEN_SIZE;
    return 0;
}

static void usb_slave_generic_interrupt_handler(__u32 status, int ep_addr)
{
    endpoint_t *ep;
   
    if (status & INTERRUPT_ENDPOINT)
    {
	if (ep_addr < 0)
	{
	    printk(KERN_WARNING __FUNCTION__": Got endpoint interrupt but "
		"endpoint address is invalid\n");
	    return;
	}
	
	/* Control endpoint */
	if (!ep_addr)
	{
	    if (status & INTERRUPT_CTRL_EP_SETUP)
		handle_control_ep_setup_interrupt(status);
	    if (usb_device.control_ep.abs_info->is_interrupt)
	    {
		__u8 dir = 
		    ((usb_device.ctrl_req.setup.bmRequestType >> 7) & 0x1);
		usb_slave_handle_control_ep(&usb_device.control_ep, dir, 
		    usb_device.ctrl_req.req_phase);
	    }
	}
	else
	{
	   if (usb_slave_get_ep_by_addr(ep_addr, &ep) < 0)
	       return;
	   if (ep->abs_info->is_interrupt)
	   {
               usb_slave_handle_non_control_ep(ep, ep->abs_info->buff + 
		   ep->abs_info->buff_pos, (ep->abs_info->buff_size <= 
	           ep->abs_info->max_pkt_size) ? ep->abs_info->buff_size : 
	           ep->abs_info->max_pkt_size);
	   }
	}
    }
    if (status & INTERRUPT_RESET)
    {
	USB_DPRINTK(__FUNCTION__": Got RESET interrupt\n");
        usb_device.ctrl_req.req_phase = NOT_STARTED;
	*(usb_device.connection_state) = USB_CONNECTED;
	usb_device.ops->device_class_uninitialized();
	usb_slave_on_reset_cmd();
    }
    if (status & INTERRUPT_SUSPEND)
    {   
	USB_DPRINTK(__FUNCTION__": Got  SUSPEND interrupt\n");
        usb_device.ctrl_req.req_phase = NOT_STARTED;
	*(usb_device.connection_state) = USB_DISCONNECTED;
	
	/* It's not clear whether we're supposed to uninitialize the class
	 * driver in case we enter the Suspend state. For example, the RNDIS 
	 * spec. is not clear on whether the device should enter 
	 * RNDIS_UNINITIALIZED_STATE in this case. But because entering a
	 * Suspend state is the best indication we got for the USB bus 
	 * physical disconnection, we uninitialize the class driver. 
	 * This mainly helps the user-mode applications to detect and indicate 
	 * physical disconnection on the USB bus. */
	 usb_device.ops->device_class_uninitialized();
    }
    if (status & INTERRUPT_RESUME)
    {
	USB_DPRINTK(__FUNCTION__": Got RESUME interrupt\n");
        usb_device.ctrl_req.req_phase = NOT_STARTED;
	*(usb_device.connection_state) = USB_CONNECTED;
    }
    if (status & INTERRUPT_SOF)
	*(usb_device.connection_state) = USB_CONNECTED;
}


/* Insert to ep_data the data of the suitable endpoint from the pool, if
 * there is no endpoint with this address in the pool return -1.
 */
static int usb_slave_get_ep_by_addr(int ep_addr, endpoint_t **ep)
{
    int i;
    
    if (!usb_device.eps_pool || !usb_device.eps_pool_size)
    {
	printk(KERN_WARNING __FUNCTION__": Endpoints pool is empty\n");
	return -1;
    }
    for (i = 0; i < usb_device.eps_pool_size; i++)
    {
	if (usb_device.eps_pool[i].abs_info->addr == ep_addr)
	{
	    *ep = &usb_device.eps_pool[i];
	    return 0;
	}
    }
    printk(KERN_ERR __FUNCTION__": Couldn't find endpoint of address %d in "
	"the pool\n", ep_addr);
    return -1;
}

static void usb_handle_send_encapsulated_stub(__u8 *req_block, __u32 req_size)
{
    printk(KERN_WARNING __FUNCTION__":stub  should never be called\n");
}

static void usb_handle_get_encapsulated_stub(__u8 *resp_buff, __u32 *size_p)
{
    printk(KERN_WARNING __FUNCTION__":stub  should never be called\n");
}

#ifdef FIX_ZERO_LEN_PKT_BUG	 
static int usb_slave_get_data_msg_len_stub(__u8 *hdr, int hdr_size)
{
    printk(KERN_WARNING __FUNCTION__":stub  should never be called\n");
    return -1;
}
#endif

static void usb_device_uninitialized_stub(void)
{
    printk(KERN_WARNING __FUNCTION__":stub  should never be called\n");
}

static void  get_configuration_stub(__u8 cfg_num, __u32 if_num, 
    endpoint_t **cfg, int *eps_num)
{
    printk(KERN_WARNING __FUNCTION__":stub  should never be called\n");
}

static int init_descriptors_stub(__u8 *dev_desc_p, int dev_desc_s, 
	__u8 *cfg_desc_p, int cfg_desc_s, string_desc_t *prod_desc, 
	string_desc_t *mfg_desc, string_desc_t *lang_desc)
{
    printk(KERN_WARNING __FUNCTION__":stub  should never be called\n");
    return -1;
}

