/*****************************************************************************
 * 
 *  USB Slave Support Software 
 *  Linux Version
 *
 *  Developed by: 
 *  Jungo Software Technologies - Residential Gateway Software Division
 *  www.jungo.com/openrg
 *  info@jungo.com
 *  
 *  
 *  Jungo Software Technologies (Jungo) and Integarated Device Technology (IDT) 
 *  jointly own all rights in the USB Slave Support Software images and source 
 *  code. 
 *  You may not, without express written permission from either Jungo or IDT,
 *  a. Distribute, sell, reproduce or copy the source code in any way.
 *  b. Rent or lease the USB Slave Support Software or otherwise transfer or 
 *  assign the right to use the USB Slave Support Software.
 *
 *
 *  
 ****************************************************************************/


#ifndef USB_SLAVE_COMMON_H
#define USB_SLAVE_COMMON_H

#include "usb_device_class_defs.h"
/* This file: Contain all definitions that are common to slave layer and device
 * (class) layer.
 */
#define MAX_DEBUG 5000
#define MAX_ALLOWED_TRANSFER_SIZE 0x4000
#define CTRL_SETUP_PKT_LEN 8
#define MAX_CTRL_DATA_PKT_LEN 256
#define MAX_GET_ENCAPSULATED_RESPONSE_DATA_LEN 1024

#undef USB_SLAVE_DEBUG
 
//#ifndef USB_SLAVE_DEBUG	
//#define USB_SLAVE_DEBUG 
//#endif

#ifdef USB_SLAVE_DEBUG
#define USB_DPRINTK(args...) printk(KERN_DEBUG args)
#else
#define USB_DPRINTK(args...)
#endif

typedef enum {
    RG_USB_CONTROL = 0,
    RG_USB_ISO = 1,
    RG_USB_BULK = 2,
    RG_USB_INTERRUPT = 3,
} usb_ep_type_t;

typedef enum {
    RG_USB_OUT = 0,
    RG_USB_IN = 1,
    RG_USB_BIDIR = 2,
} usb_ep_dir_t;

typedef enum {
    USB_DISABLED = 0,
    USB_BUSY = 2,
    USB_ENABLED = 3,
} usb_ep_mode_t;

typedef enum {
    USB_NAK = 0,
    USB_STALL = 1, 
    USB_ACK = 2,
} usb_ep_transfer_mode_t;

/* endpoint transfer interrupt mode bits to be used at is_interrupt field in 
 * endpoint_t structure
 */
typedef enum {
    USB_NO_INTERRUPT = 0x00000000,
    USB_INTERRUPT_ON_RX_START_PKT = 0x00000001, 
    USB_INTERRUPT_ON_RX_END_PKT = 0x00000002, 
    USB_INTERRUPT_ON_TX_START_PKT = 0x00000004,
    USB_INTERRUPT_ON_TX_END_PKT = 0x00000008,
} usb_ep_interrupt_mode;

/* Connection states */
enum {
    USB_UNKNOWN =  0,
    USB_CONNECTED = 1,
    USB_DISCONNECTED = 2,
};

/* Control endpoint transfer states */
typedef enum {
    NOT_STARTED = 0,
    SETUP_TOKEN = 1,
    DATA_TRANSMIT = 2,
    DATA_RECEIVE = 3,
    DATA_SEND_ZERO_LEN_PKT = 4,
} control_phase_t;

/* All endpoint information that is needed to perform transfers through a
 * USB pipe 
 */
typedef struct { 
    usb_ep_type_t type; /* (CONTROL, BULK, ISO, INTERRUPT) */
    usb_ep_dir_t direction; /* (IN, OUT, BIDIR) */
    int addr; /* 0, .... */
    int interface_num;
    usb_ep_mode_t mode; /* (DISABLED, ENABLED, BUSY) */
    usb_ep_transfer_mode_t transfer_mode; /* (NAK, ACK, STALL) */
    int max_pkt_size; /* Limited by hardware and USB protocol */ 
    char *buff; /* point to current transfer buffer */
    int buff_size; /* Left to handle */
    int buff_pos; /* current position on buff */
    
    /* to use at the end of handling current USB packet */
    control_phase_t (*ep_operation)(void *context); 
    
    /* To use at the end of current USB transfer */
    void (*end_func)(void *ep);
    void *end_context1;
    void *end_context2;
    int end_arg1;
    int end_arg2;
    int is_busy;
    int is_dma;
    int is_interrupt;
} abstract_ep_t;

typedef struct endpoint_t { 
    void *phys_info;
    abstract_ep_t *abs_info;
} endpoint_t;

/* endpoint addres in the descriptor- first 4 bits: endpoint number 
 * last 4 bits endpoint direction: {0x0 - Out, 0x8 - In}
 */
#define GET_ENDPOINT_ADDR(ep)                                      \
       (ep.abs_info->direction == RG_USB_OUT) ? (__u8)ep.abs_info->addr : \
	    (0x80 + (__u8)ep.abs_info->addr)                             


/* API "handshake" between usb-slave layer and the upper usb-device (class)
 * layer
 */
typedef struct usb_device_ops
{
#ifdef FIX_ZERO_LEN_PKT_BUG
    /* Device-class layer, to return size of transfer */
    int (*get_data_msg_len)(__u8 *msg_hdr, int hdr_size);
#endif    
    
    /* Device-class layer, to fill USB descriptors */
    int (*init_descriptors)(__u8 *dev_desc_p, int dev_desc_s, 
	__u8 *cfg_desc_p, int cfg_desc_s, string_desc_t *prod_desc, 
	string_desc_t *mfg_desc, string_desc_t *lang_desc); 
   
    /* Device-class layer, to return the number of endpoints setup uppon
     * a SET_CONFIGURATION  request from the host 
     */
    void (*get_configuration)(__u8 cfg_num, __u32 if_num, 
        endpoint_t **cfg, int *eps_num);

    /* Device-class layer, to handle control class specific set command */
    void (*handle_send_encapsulated_cmd)(__u8 *req_block, __u32 req_size);
    
    /* Device-class layer, to handle control class specific get response */
    void (*handle_get_encapsulated_resp)(__u8 *resp_buff, __u32 *resp_size_p);
    
    /* Device-class layer to return itself to uninitialized state */
    void (*device_class_uninitialized)(void);

    /* Slave layer, to send device layer packet through a interrupt/bulk-in 
     * pipe
     */
    int (*ep_send)(__u8 *buff, int buff_len, endpoint_t *ep, void *context1,  
        void *context2); 
    
    /* Slave layer, to stop and cancle all current transfers (on all pipes) */ 
    void (*terminate_current_transfers)(void); 
    
    /* Slave layer, to be called when finished handling one USB pkt of OUT
     * transfer (device can use it as OUT endpoint ep_operation call-back).
     */ 
    void (*on_receive_usb_data_pkt)(void *context);
    
    /* Slave layer, to be called when finished handling one USB pkt of IN
     * transfer (device can use it as IN endpoint ep_operation call-back).
     */ 
    void (*on_send_usb_data_pkt)(void *context);
    
#ifdef CONFIG_RG_USB_SLAVE_MODULE    
    /* Slave layer, to return the slave to "annonimus" USB device state,
     * On closing of device class module
     */
    void (*on_device_class_closed)(void);
#endif    
} usb_device_ops;

/* This is the only function that is exported from usb_slave.c to perform the
 * initial handshake between the usb_slave layer and the upper usb-device
 * (class) layer.
 */
int usb_slave_on_device_class_initialized(struct usb_device_ops *usb_upper_ops, 
    int num_of_cfgs, int cfg_desc_s, volatile int *is_device_connected); 

void usb_slave_dump_buffer(__u8 *buff, int size);
#endif

