/*
 * (C) Copyright 2000
 * Rob Taylor, Flying Pig Systems. robt@flyingpig.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <mpc824x.h>
#include <watchdog.h>
#include <i2c.h>

struct manufact_hw_info_t {
  char magic[4];  /* 0x38323435 "8245" */
  char rev[4];
  char productName[60];
  char cardID[4];
  char cardRev[4];
  char flags[4]; /* what is present?  WLAN, ETH? */
  char leds[4];  /* assignments */
  char macAddress[6];
  char serialNumber[26];
};

void watchdog_reset(void);
void printManufacHwInfo (struct manufact_hw_info_t* mInfo);
void init_gpio(void);
void set_led(char reg);

/* flash stuff */
int board_pre_init(ulong bootflag)
{
    return 0;
}

int checkboard(void)
{
    ulong busfreq  = get_bus_freq(0);
    char  buf[16];

    printf("Airespace AP1200, Local Bus at %s MHz\n", strmhz(buf, busfreq));

    /* turn on POWER LED */
    init_gpio();

    return 0;
}

long int initdram(int board_type)
{
    volatile int		 i, cnt;
    volatile uchar * base= CFG_SDRAM_BASE;
    volatile ulong * addr;
    volatile ulong		 save[32];
    volatile ulong		 val, ret  = 0;

    for (i=0, cnt=(CFG_MAX_RAM_SIZE / sizeof(long)) >> 1; cnt > 0; cnt >>= 1) {
        addr = (volatile ulong *)base + cnt;
        save[i++] = *addr;
        *addr = ~cnt;
    }

    addr = (volatile ulong *)base;
    save[i] = *addr;
    *addr = 0;

    if (*addr != 0) {
        *addr = save[i];
        goto Done;
    }

    for (cnt = 1; cnt < CFG_MAX_RAM_SIZE / sizeof(long); cnt <<= 1) {
        addr = (volatile ulong *)base + cnt;
        val = *addr;
        *addr = save[--i];
        if (val != ~cnt) {
            ulong new_bank0_end = cnt * sizeof(long) - 1;
            ulong mear1  = mpc824x_mpc107_getreg(MEAR1);
            ulong emear1 = mpc824x_mpc107_getreg(EMEAR1);

            mear1 =  (mear1  & 0xFFFFFF00) |
                ((new_bank0_end & MICR_ADDR_MASK) >> MICR_ADDR_SHIFT);
            emear1 = (emear1 & 0xFFFFFF00) |
                ((new_bank0_end & MICR_ADDR_MASK) >> MICR_EADDR_SHIFT);
            mpc824x_mpc107_setreg(MEAR1,  mear1);
            mpc824x_mpc107_setreg(EMEAR1, emear1);
            
            ret = cnt * sizeof(long);

            goto Done;
        }
    }

    ret = CFG_MAX_RAM_SIZE;

Done:

    return ret;
}

static int dram_test(int test, ulong *base, ulong step, ulong length,
		     ulong pattern)
{
    ulong *ptr = base;
    ulong i;
    printf("mem %d ", test);
    for (i = 0; i < length; i += step, ptr += step) {
        *ptr = pattern ^ (ulong) ptr;
    }
    ptr = base;
    for (i = 0; i < length; i += step, ptr += step) {
        if (*ptr != (pattern ^ (ulong) ptr)) {
	    printf("Mismatch at %08lx: Expected %08lx got %08lx",
		   (ulong) ptr, pattern ^ (ulong) ptr, *ptr);
	    return 1;
	}
    }
    printf("\b\b\b\b\b\b");
    return 0;
}

#define GPIO_DEV  0x20

#define GPIO_ANT_2INT  0x0a
#define GPIO_ANT_2EXT  0x05
#define GPIO_ANT_5TOP  0x40
#define GPIO_ANT_5BOT  0x80

#define GPIO_LED_ALARM 0x02
#define GPIO_LED_2GHZ  0x04
#define GPIO_LED_POWER 0x08
#define GPIO_LED_5GHZ  0x10
#define GPIO_LED_MASK  0x1e

#define GPIO_REG_OUT0  2
#define GPIO_REG_OUT1  3
#define GPIO_REG_CONF0 6
#define GPIO_REG_CONF1 7

void set_led(char reg)
{
    i2c_send(GPIO_DEV, GPIO_REG_OUT1, 1, &reg);
}

void init_gpio(void)
{
    /* GPIO hangs off I2C bus and is supported by the
     * Philips PCA9555, I2C address 0x20
     */
    char reg;

    /* intialize I2C bus */
    i2c_init();

    /* set I/O to output */
    reg = 0; 
    udelay(1000);
    i2c_send(GPIO_DEV, GPIO_REG_CONF0, 1, &reg);
    udelay(1000);
    i2c_send(GPIO_DEV, GPIO_REG_CONF1, 1, &reg);

    /* select internal antenas */
    udelay(1000);
    reg = GPIO_ANT_2INT | GPIO_ANT_5TOP;
    i2c_send(GPIO_DEV, GPIO_REG_OUT0, 1, &reg);

    /* turn on POWER LED */
    udelay(1000);
    reg = GPIO_LED_2GHZ | GPIO_LED_5GHZ; /* leaves only POWER LED on */
    i2c_send(GPIO_DEV, GPIO_REG_OUT1, 1, &reg);

}

int testdram(ulong dram_size)
{
    int test;
    ulong *ptr;

    printf("POST:  ");

    ptr = (ulong *)CFG_SDRAM_BASE;
    test = 0;
    /* test entire first 128KB */
    if (dram_test(test++, ptr, 1, (128 * 1024) / sizeof(*ptr),
		  0x55555555)) {
        goto failure;
    }
    WATCHDOG_RESET();
    if (dram_test(test++, ptr, 1, (128 * 1024) / sizeof(*ptr),
		  0xaaaaaaaa)) {
        goto failure;
    }
    WATCHDOG_RESET();
    /* test each 128KB region */
    if (dram_test(test++, ptr, (128 * 1024) / sizeof(*ptr), dram_size/sizeof(*ptr),
		  0xff00ff00)) {
        goto failure;
    }
    WATCHDOG_RESET();
    if (dram_test(test++, ptr, (128 * 1024) / sizeof(*ptr), dram_size/sizeof(*ptr),
		  0x00ff00ff)) {
        goto failure;
    }

    printf("passed\n");

    return 0;

 failure:
    /* turn on alarm LED */
    set_led(GPIO_LED_ALARM | GPIO_LED_2GHZ | GPIO_LED_5GHZ);
    printf("FAILED!!!!!!!\n");
    return 1;
}
/*
 * Initialize PCI Devices, report devices found.
 */
void pci_init(bd_t *board)
{
    extern int pci_dev_init(int busNo);

    /* initialize PCI devices on bus 0, the only bus */
    pci_dev_init(0);
}

int last_stage_init(bd_t *bd)
{
    struct manufact_hw_info_t * mInfo;
    char ethaddr[18];

    /* setup environment variables */
    if (!getenv("ethaddr")) {
        mInfo = (struct manufact_hw_info_t *)OLD_MANUFAC_DATA_LOC;
        sprintf(ethaddr, "%02x:%02x:%02x:%02x:%02x:%02x",
                mInfo->macAddress[0],
                mInfo->macAddress[1],
                mInfo->macAddress[2],
                mInfo->macAddress[3],
                mInfo->macAddress[4],
                mInfo->macAddress[5]);
        setenv ("ethaddr", ethaddr);
        printf("Env:   saving\n");
        saveenv();
    }

    return 0;
}

void watchdog_reset(void)
{
    /* Watchdog is supported by the TI TPS3823.
     * Create a transition on ROM Chip Select 1 to
     * refresh watchdog.  It expires in about 1 sec.
     */
    (void) *(volatile char *) 0xff000000;
}

void printManufacHwInfo (struct manufact_hw_info_t* mInfo)
{
    int i=0;
    printf("Hardware Manufacturing Info as stored in NV: \n\n");
    printf("\tName      : ");
    for (i=0; i < 16; i++) printf("%c", mInfo->productName[i]);printf("\n");
    printf("\tEthaddr   : ");
    for(i=0; i< 6; i++) printf("%x ", mInfo->macAddress[i]);printf("\n");
}
