/*
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <watchdog.h>
#include <malloc.h>
#include <net.h>

#include "pci.h"

#undef DEBUG

#if (CONFIG_COMMANDS & CFG_CMD_NET) && defined(CONFIG_NET_MULTI) \
	&& defined(CONFIG_TULIP)

/* PCI Registers.
 */
#define PCI_CFDA_PSM		0x43

#define CFRV_RN		0x000000f0	/* Revision Number */
#define CBIO_MASK	0xffffff80

#define WAKEUP		0x00		/* Power Saving Wakeup */
#define SLEEP		0x80		/* Power Saving Sleep Mode */


#define AN983B_VID	0x1317		/* ADMtek Manufacturer */
#define AN983B_DID	0x0985		/* Unique Device ID # */

/* Ethernet chip registers.
	 */
#define AN983B_PAR	0x000		/* PCI Access Register */
#define AN983B_TDR	0x008		/* Transmit Poll Demand Reg */
#define AN983B_RDB	0x018		/* RX Ring Base Address Reg */
#define AN983B_TDB	0x020		/* TX Ring Base Address Reg */
#define AN983B_SR	0x028		/* Status Register */
#define AN983B_NAR	0x030		/* Operation Mode Register */
#define AN983B_SPR	0x048		/* Ethernet Address PROM */
#define AN983B_PAR0	0x0A4		/* Physical Address Register 0 */
#define AN983B_PAR1	0x0A8		/* Physical Address Register 1 */

/* Register bits.
	 */
#define PAR_SWR		0x00000001	/* Software Reset */
#define SR_TS		0x00700000	/* Transmit Process State */
#define SR_RS		0x000e0000	/* Receive Process State */
#define NAR_SR		0x00000002	/* Start/Stop Receive */
#define NAR_PR		0x00000040	/* Promiscuous mode */
#define NAR_PM		0x00000080	/* Pass All Multicast */
#define NAR_ST		0x00002000	/* Start/Stop Transmission Command */
#define NAR_PS		0x00040000	/* Port Select */
#define NAR_SQE		0x00080000	/* disable SQE function */
#define NAR_SDP		0x02000000	/* SD Polarity - MUST BE ASSERTED */

/* Descriptor bits.
	 */
#define R_OWN		0x80000000	/* Own Bit */
#define RD_RER		0x02000000	/* Receive End Of Ring */
#define RD_LS		0x00000100	/* Last Descriptor */
#define RD_ES		0x00008000	/* Error Summary */
#define TD_TER		0x02000000	/* Transmit End Of Ring */
#define T_OWN		0x80000000	/* Own Bit */
#define TD_LS		0x40000000	/* Last Segment */
#define TD_FS		0x20000000	/* First Segment */
#define TD_ES		0x00008000	/* Error Summary */
#define TD_SET		0x08000000	/* Setup Packet */


#define SROM_HWADD	    0x0008	/* Hardware Address offset in SROM */
#define SROM_RD		0x00004000	/* Read from Boot ROM */
#define SROM_SR		0x00000800	/* Select Serial ROM when set */

#define DT_IN		0x00000004	/* Serial Data In */
#define DT_CLK		0x00000002	/* Serial ROM Clock */
#define DT_CS		0x00000001	/* Serial ROM Chip Select */

#define POLL_DEMAND	1

#define RESET_AN983B(dev) {\
    int i;\
    i=INL(dev, AN983B_PAR);\
    udelay(1000);\
    OUTL(dev, i | PAR_SWR, AN983B_PAR);\
    udelay(1000);\
    OUTL(dev, i, AN983B_PAR);\
    udelay(1000);\
    for (i=0;i<5;i++) {INL(dev, AN983B_PAR); udelay(10000);}\
    udelay(1000);\
}

#define START_AN983B(dev) {\
    s32 nar; \
    nar = INL(dev, AN983B_NAR);\
    nar |= NAR_ST | NAR_SR;\
    OUTL(dev, nar, AN983B_NAR);		/* Enable the TX and/or RX */\
}

#define STOP_AN983B(dev) {\
    s32 nar; \
    nar = INL(dev, AN983B_NAR);\
    nar &= ~(NAR_ST|NAR_SR);\
    OUTL(dev, nar, AN983B_NAR);		/* Disable the TX and/or RX */ \
}

#define NUM_RX_DESC PKTBUFSRX           /* currently 4 */
#define NUM_TX_DESC 2			/* Number of TX descriptors   */
#define RX_BUFF_SZ  PKTSIZE_ALIGN

#define TOUT_LOOP   1000000

#define SETUP_FRAME_LEN 192
#define ETH_ALEN	6


struct an983b_desc {
	volatile s32 status;
	u32 des1;
	u32 buf;
	u32 next;
};

static struct an983b_desc rx_ring[NUM_RX_DESC]; /* RX descriptor ring         */
static struct an983b_desc tx_ring[NUM_TX_DESC]; /* TX descriptor ring         */
static int rx_new;                             /* RX descriptor ring pointer */
static int tx_new;                             /* TX descriptor ring pointer */

static char rxRingSize;
static char txRingSize;

static void  read_hw_addr(struct eth_device* dev, bd_t * bis);
#if 0
static short srom_rd(struct eth_device* dev, u_long address, u_char offset);
static void  srom_latch(struct eth_device* dev, u_int command, u_long address);
static void  srom_command(struct eth_device* dev, u_int command, u_long address);
static void  srom_address(struct eth_device* dev, u_int command, u_long address, u_char offset);
static short srom_data(struct eth_device* dev, u_int command, u_long address);
static void  sendto_srom(struct eth_device* dev, u_int command, u_long addr);
static int   getfrom_srom(struct eth_device* dev, u_long addr);
#endif

static int   an983b_init(struct eth_device* dev, bd_t* bis);
static int   an983b_send(struct eth_device* dev, volatile void *packet, int length);
static int   an983b_recv(struct eth_device* dev);
static void  an983b_halt(struct eth_device* dev);

#define pci_mem_to_phys(x) (x)
#define pci_ram_to_mem(x) (x)

static int INL(struct eth_device* dev, u_long addr)
{
	return le32_to_cpu(*(volatile u_long *)pci_mem_to_phys(addr + dev->iobase));
}

static void OUTL(struct eth_device* dev, int command, u_long addr)
{
	*(volatile u_long *)pci_mem_to_phys(addr + dev->iobase) = cpu_to_le32(command);
}

int an983b_initialize(bd_t *bis)
{
	int             	card_number, i, status = 0;
	short                   command;
	int                     bus, device, func;
	unsigned char   	timer;
	uint32			iobase;
	struct eth_device* 	dev;

	for(card_number=i=0; ; i++) {
		status = pci_dev_find(AN983B_VID, AN983B_DID, i,
				      &bus, &device, &func);
		if (status == ERROR) {
			break;
		}

		pci_config_inw(bus, device, func, PCI_CFG_COMMAND, &command);
		command |= PCI_CMD_MASTER_ENABLE | PCI_CMD_IO_ENABLE | PCI_CMD_MEM_ENABLE;
		pci_config_outw(bus, device, func, PCI_CFG_COMMAND, command);

		/* Check the latency timer for values >= 0x60. */
		pci_config_inb(bus, device, func, PCI_CFG_LATENCY_TIMER, &timer);
		if (timer < 0x60) {
		        pci_config_outb(bus, device, func, PCI_CFG_LATENCY_TIMER, 0x60);
		}

		/* read BAR for memory space access */
		pci_config_inl(bus, device, func, PCI_CFG_BASE_ADDRESS_1, &iobase);
		iobase &= CBIO_MASK;

#ifdef DEBUG
		printf("an983b: ADMtek AN983B PCI Device @0x%x\n", iobase);
#endif

		dev = (struct eth_device*) malloc(sizeof *dev);

		sprintf(dev->name, "an983b#%d", i);
		dev->iobase = iobase;
		dev->priv   = (void*) ((bus << 16) | (device << 11) | (func << 8));
		dev->init   = an983b_init;
		dev->halt   = an983b_halt;
		dev->send   = an983b_send;
		dev->recv   = an983b_recv;

		/* Ensure we're not sleeping. */
		pci_config_outb(bus, device, func, PCI_CFDA_PSM, WAKEUP);

		udelay(10 * 1000);

                read_hw_addr(dev, bis);

		eth_register(dev);

		card_number++;
	}

	return card_number;
}

#ifdef DEBUG
static void an983b_dump(struct eth_device* dev)
{
    int i;
    for (i = 0; i < 0xac; i+= 4) {
        printf("   %02x: %08x", i, INL(dev, i));
	if (i & 0x4) printf("\n");
    }
}
#endif /* DEBUG */

static int an983b_init(struct eth_device* dev, bd_t* bis)
{
	int		i;
	int		devbusfn = (int) dev->priv;

	/* Ensure we're not sleeping. */
	pci_config_outb(devbusfn>>16, devbusfn>>11&0x1f, devbusfn>>8&0x7,
			PCI_CFDA_PSM, WAKEUP);			

	RESET_AN983B(dev);

	if ((INL(dev, AN983B_SR) & (SR_TS | SR_RS)) != 0) {
		printf("Error: Cannot reset ethernet controller.\n");
		return 0;
	}

	OUTL(dev, NAR_SQE | NAR_SDP | NAR_PS, AN983B_NAR);

        /* set Eth Addr */
        OUTL(dev,
             (dev->enetaddr[0] |
              (dev->enetaddr[1] << 8) |
              (dev->enetaddr[2] << 16) |
              (dev->enetaddr[3] << 24)),
             AN983B_PAR0);
        OUTL(dev,
             (dev->enetaddr[4] |
              (dev->enetaddr[5] << 8)),
             AN983B_PAR1);

	for (i = 0; i < NUM_RX_DESC; i++) {
		rx_ring[i].status = cpu_to_le32(R_OWN);
		rx_ring[i].des1 = cpu_to_le32(RX_BUFF_SZ);
		rx_ring[i].buf = cpu_to_le32(pci_ram_to_mem((u32) NetRxPackets[i]));
		rx_ring[i].next = 0;
	}

	for (i=0; i < NUM_TX_DESC; i++) {
		tx_ring[i].status = 0;
		tx_ring[i].des1 = 0;
		tx_ring[i].buf = 0;
		tx_ring[i].next = 0;
	}

	rxRingSize = NUM_RX_DESC;
	txRingSize = NUM_TX_DESC;

	/* Write the end of list marker to the descriptor lists. */
	rx_ring[rxRingSize - 1].des1 |= cpu_to_le32(RD_RER);
	tx_ring[txRingSize - 1].des1 |= cpu_to_le32(TD_TER);

	/* Tell the adapter where the TX/RX rings are located. */
	OUTL(dev, pci_ram_to_mem((u32) &rx_ring), AN983B_RDB);
	OUTL(dev, pci_ram_to_mem((u32) &tx_ring), AN983B_TDB);

	START_AN983B(dev);

	tx_new = 0;
	rx_new = 0;

	return 1;
}

static int an983b_send(struct eth_device* dev, volatile void *packet, int length)
{
	int		status = -1;
	int		i;

	if (length <= 0) {
		printf("%s: bad packet size: %d\n", dev->name, length);
		goto Done;
	}

	for(i = 0; tx_ring[tx_new].status & cpu_to_le32(T_OWN); i++) {
		if (i >= TOUT_LOOP) {
			printf("%s: prev tx buffer not ready\n", dev->name);
			goto Done;
		}
	}

	tx_ring[tx_new].buf    = cpu_to_le32(pci_ram_to_mem((u32) packet));
	tx_ring[tx_new].des1   = cpu_to_le32(((tx_new == txRingSize - 1)?TD_TER:0) |
					     TD_LS | TD_FS | length);
	tx_ring[tx_new].status = cpu_to_le32(T_OWN);

	OUTL(dev, POLL_DEMAND, AN983B_TDR);

	for(i = 0; tx_ring[tx_new].status & cpu_to_le32(T_OWN); i++) {
		if (i >= TOUT_LOOP) {
			printf("%s: tx buffer not ready\n", dev->name);
#ifdef DEBUG
			an983b_dump(dev);
#endif /* DEBUG */
			goto Done;
		}
	}

#ifdef DEBUG
	/* test-only */
	if (le32_to_cpu(tx_ring[tx_new].status) & TD_ES) {
		printf("TX error status = 0x%08X\n",
		       le32_to_cpu(tx_ring[tx_new].status));
		goto Done;
	}
#endif /* DEBUG */

	tx_new = (tx_new + 1) % txRingSize;
	status = length;

 Done:
	return status;
}

static int g_unmanaged_rx_errors = 0;

static int an983b_recv(struct eth_device* dev)
{
	s32		status;
	int		length    = 0;

	for ( ; ; ) {
		WATCHDOG_RESET();

		status = (s32)le32_to_cpu(rx_ring[rx_new].status);

		if (status & R_OWN) {
			break;
		}

		if (status & RD_LS) {
			/* Valid frame status.
			 */
			if (status & RD_ES) {

				/* There was an error.
				 */
				if (g_unmanaged_rx_errors++ < 2)
				    printf("\nRX error status = 0x%08X/if errs cont, box will reset!\n", status);
				else {                 
				    printf("\nToo many RX errors: Resetting the box.\n");
				    udelay(5);
				    for (;;) ;
				}
			} else {
				/* A valid frame received.
				 */
				length = (le32_to_cpu(rx_ring[rx_new].status) >> 16);

				/* Pass the packet up to the protocol
				 * layers.
				 */
				NetReceive(NetRxPackets[rx_new], length - 4);
			}

			/* Change buffer ownership for this frame, back
			 * to the adapter.
			 */
			rx_ring[rx_new].status = cpu_to_le32(R_OWN);
		}

		/* Update entry information.
		 */
		rx_new = (rx_new + 1) % rxRingSize;
	}

	return length;
}

static void an983b_halt(struct eth_device* dev)
{
	int		devbusfn = (int) dev->priv;

	STOP_AN983B(dev);
	if (devbusfn > 0) {
	          pci_config_outb(devbusfn>>16, devbusfn>>11&0x1f, devbusfn>>8&0x7,
				  PCI_CFDA_PSM, SLEEP);
	}
}

static void read_hw_addr(struct eth_device *dev, bd_t *bis)
{
#if 0
	u_short tmp, *p = (short *)(&dev->enetaddr[0]);
	int i, j = 0;

	for (i = 0; i < (ETH_ALEN >> 1); i++) {
		tmp = srom_rd(dev, AN983B_SPR, (SROM_HWADD >> 1) + i);
		*p = le16_to_cpu(tmp);
		j += *p++;
	}

	if ((j == 0) || (j == 0x2fffd)) {
		printf("Warning: can't read HW address from SROM %x\n", j);
		goto Done;
	}

	for (i = 0; i < ETH_ALEN; i++) {
#ifdef DEBUG
		if (dev->enetaddr[i] != bis->bi_enetaddr[i]) {
			printf("Warning: HW addresses don't match:\n");
			printf("Address in SROM is         "
			       "%02X:%02X:%02X:%02X:%02X:%02X\n",
			       dev->enetaddr[0], dev->enetaddr[1],
			       dev->enetaddr[2], dev->enetaddr[3],
			       dev->enetaddr[4], dev->enetaddr[5]);
			printf("Address used by ppcboot is "
			       "%02X:%02X:%02X:%02X:%02X:%02X\n",
			       bis->bi_enetaddr[0], bis->bi_enetaddr[1],
			       bis->bi_enetaddr[2], bis->bi_enetaddr[3],
			       bis->bi_enetaddr[4], bis->bi_enetaddr[5]);
			goto Done;
		}
#endif
	        bis->bi_enetaddr[i] = dev->enetaddr[i];
	}

Done:
#else
        int i;
        char *c;
        int octet;

        c = getenv("ethaddr");
        if (!c) {
                /* read ethaddr from flash */
                last_stage_init(bis);
                c = getenv("ethaddr");
                if (!c) {
                        printf("Warning: No HW address found for ");
                        return;
                }
        }
	for (i = 0; i < ETH_ALEN; i++) {
                octet = simple_strtoul(c, &c, 16);
                c++; /* skip ':' */
                dev->enetaddr[i] = octet;
	        bis->bi_enetaddr[i] = dev->enetaddr[i];
        }
          
#endif
}

#if 0
/* SROM Read.
 */
static short
srom_rd(struct eth_device* dev, u_long addr, u_char offset)
{
	sendto_srom(dev, SROM_RD | SROM_SR, addr);

	srom_latch(dev, SROM_RD | SROM_SR | DT_CS, addr);
	srom_command(dev, SROM_RD | SROM_SR | DT_IN | DT_CS, addr);
	srom_address(dev, SROM_RD | SROM_SR | DT_CS, addr, offset);

	return srom_data(dev, SROM_RD | SROM_SR | DT_CS, addr);
}

static void
srom_latch(struct eth_device* dev, u_int command, u_long addr)
{
	sendto_srom(dev, command, addr);
	sendto_srom(dev, command | DT_CLK, addr);
	sendto_srom(dev, command, addr);
}

static void
srom_command(struct eth_device* dev, u_int command, u_long addr)
{
	srom_latch(dev, command, addr);
	srom_latch(dev, command, addr);
	srom_latch(dev, (command & 0x0000ff00) | DT_CS, addr);
}

static void
srom_address(struct eth_device *dev, u_int command, u_long addr, u_char offset)
{
	int i;
	signed char a;

	a = (char)(offset << 2);
	for (i=0; i<6; i++, a <<= 1) {
		srom_latch(dev, command | ((a < 0) ? DT_IN : 0), addr);
	}
	udelay(1);

	i = (getfrom_srom(dev, addr) >> 3) & 0x01;

	return;
}

static short
srom_data(struct eth_device *dev, u_int command, u_long addr)
{
	int i;
	short word = 0;
	s32 tmp;

	for (i=0; i<16; i++) {
		sendto_srom(dev, command  | DT_CLK, addr);
		tmp = getfrom_srom(dev, addr);
		sendto_srom(dev, command, addr);

		word = (word << 1) | ((tmp >> 3) & 0x01);
	}

	sendto_srom(dev, command & 0x0000ff00, addr);

	return word;
}

static void
sendto_srom(struct eth_device* dev, u_int command, u_long addr)
{
	OUTL(dev, command, addr);
	udelay(1);
}

static int
getfrom_srom(struct eth_device* dev, u_long addr)
{
	s32 tmp;

	tmp = INL(dev, addr);
	udelay(1);

	return tmp;
}
#endif
#endif
