#include <ppcboot.h>
#include <command.h>
#include <i2c.h>
#include "tps2383.h"

#include "../../../cmn/util/cab_cpld.h"
#include "../../../cmn/util/tps2383.h"

char Ctrl[TPS_PORTS] = {0};
short calib_factor = 1000;

#define PortMode(nTps, Port, Mode) \
    WriteTps(nTps, PORT_CONTROL + Port, Ctrl[Port] = (Ctrl[Port] & ~ENABLE_MODES_MASK) | Mode)
#define Null(nTps) \
    WriteTps(nTps, CONTROL, DISCOVERY_FAULT_DIS_MASK | BYPASS_CLASS_MASK | \
             BYPASS_SAMPLE_MASK | BYPASS_DISCOVERY_MASK | DISCONNECT_DIS_MASK | \
             BYPASS_RAMP_MASK | JOG_MODE_MASK)
#define Ramp(nTps) \
    WriteTps(nTps, CONTROL, DISCOVERY_FAULT_DIS_MASK | BYPASS_CLASS_MASK | \
             BYPASS_SAMPLE_MASK | BYPASS_DISCOVERY_MASK | \
             DISCONNECT_DIS_MASK | JOG_MODE_MASK)
#define JogTps(nTps) \
    WriteTps(nTps, CONTROL, DISCOVERY_FAULT_DIS_MASK | BYPASS_CLASS_MASK | \
             BYPASS_SAMPLE_MASK | BYPASS_DISCOVERY_MASK | \
             DISCONNECT_DIS_MASK | BYPASS_RAMP_MASK | JOG_MASK | \
             JOG_MODE_MASK)
#define Discovery(nTps) \
    WriteTps(nTps, CONTROL, DISCOVERY_FAULT_DIS_MASK | BYPASS_CLASS_MASK | \
             BYPASS_SAMPLE_MASK | DISCONNECT_DIS_MASK | BYPASS_RAMP_MASK | \
             JOG_MODE_MASK)
#define Current(nTps) \
    WriteTps(nTps, PORT_CONTROL + A2D_INPUT_SELECT, \
             Ctrl[A2D_INPUT_SELECT] &= ~PORT_CONTROL_X_MASK)
#define Voltage(nTps) \
    WriteTps(nTps, PORT_CONTROL + A2D_INPUT_SELECT, \
             Ctrl[A2D_INPUT_SELECT] |= PORT_CONTROL_X_MASK)
#define Sample(nTps) \
    WriteTps(nTps, CONTROL, DISCOVERY_FAULT_DIS_MASK | BYPASS_CLASS_MASK | \
             BYPASS_DISCOVERY_MASK | DISCONNECT_DIS_MASK | BYPASS_RAMP_MASK | \
             JOG_MODE_MASK)
#define PortModeOpen(Port, nTps, Mode) \
    WriteTps(nTps, PORT_CONTROL + Port, (Ctrl[Port] & ~ENABLE_MODES_MASK) | Mode)
#define PortModeClose(Port, nTps) \
    WriteTps(nTps, PORT_CONTROL + port, Ctrl[Port])
#define ClassLim(port, nTps, Limit1, Limit2) \
    WriteTps(nTps, PORT_CONTROL + CLASS_LIMIT1, \
             Ctrl[CLASS_LIMIT1] = ((Ctrl[CLASS_LIMIT1] & ~PORT_CONTROL_X_MASK) | \
	                           (Limit1 << PORT_CONTROL_X_POS))); \
    WriteTps(nTps, PORT_CONTROL + CLASS_LIMIT2, \
             Ctrl[CLASS_LIMIT2] = ((Ctrl[CLASS_LIMIT2] & ~PORT_CONTROL_X_MASK) | \
                                   (Limit2 << PORT_CONTROL_X_POS)))
#define Class(nTps) \
    WriteTps(nTps, CONTROL, DISCOVERY_FAULT_DIS_MASK | BYPASS_SAMPLE_MASK | \
             BYPASS_DISCOVERY_MASK | DISCONNECT_DIS_MASK | BYPASS_RAMP_MASK | \
             JOG_MODE_MASK)


#define HOLDPORT(port) \
    WriteTps(nTps, PORT_CONTROL + port, Ctrl[port] = (Ctrl[port] & ~NO_HOLD_INT) | HOLD_INT)
#define RELEASEPORT(port) \
    WriteTps(nTps, PORT_CONTROL + port, Ctrl[port] = (Ctrl[port] & ~HOLD_INT) | NO_HOLD_INT)

#define HOLDRAMP()    HOLDPORT(1)
#define RELEASERAMP() RELEASEPORT(1)

#define HOLDINT()     HOLDPORT(0)
#define RELEASEINT()  RELEASEPORT(0)

#define TPS_I2C_USDELAY 15


static char WriteTps(char addr, char reg, char data)
{
    char data_addr[8];
    data_addr[0] = data;

    //printf("i2c_send %x %x %x %x\n", addr, reg, 1, data);
    udelay(TPS_I2C_USDELAY);
    return i2c_send (addr, reg, 1, data_addr);
}

static int ReadTps(char addr, char reg, char *data)
{
    udelay(TPS_I2C_USDELAY);
    return i2c_receive (addr, reg, 1, data);
}

static int DelayMs(int ms)
{
    int us = ms*1000;
    udelay(us);
    return us;
}

void tps_init(int nTps)
{
    int port;

    i2c_init();

    Null(nTps);     // Reset common control register to null state
    for (port = 0; port < TPS_PORTS; port++) {
        WriteTps(nTps, PORT_CONTROL + port, 0);
    }
    RELEASEINT();
    RELEASERAMP();
}

int tps_jog(ulong port, int nTps)
{
    int i;
    int loops = 0;
    char data;

    /* stop scan */
    Null(nTps);
    while (1) {
        /* read port status */
        for (i = 0; i < TPS_PORTS; i++) {
	    if (ReadTps(nTps, PORT_STATUS + i, &data)) {
	        printf ("i2c_receive FAILED\n");
	        return 1;
	    }
	    if (data & PORT_SERVICE_MASK) {
	        if (port == i) {
		    return 0;
		}
	        break;
	    }
	}
	JogTps(nTps);
	if (loops++ > 16) {
	    break;
	}
    }
    return 1;
}

int tps_ramp_up(ulong port, int nTps, ulong ms)
{
    /* Hold ramp */
    HOLDRAMP();
    /* Disable port OL and UV/OV faults */
    WriteTps(nTps, PORT_CONTROL + port, Ctrl[port] | PORT_FAULT_DISABLE_MASK);
    /* Set control register back to null */
    Null(nTps);
    /* Set control register to ramp mode */
    Ramp(nTps);
    /* Set port control to ramp up */
    PortMode(nTps, port, DISC_CLASS_PWR);
    /* Release ramp */
    RELEASERAMP();
    DelayMs(10);
    while (ms--) {
        DelayMs(1);
	/* Set control register to ramp mode */
	Ramp(nTps);
	/* Set control register back to null */
	Null(nTps);
    } 
    PortMode(nTps, port, PORT_OFF);
    PortMode(nTps, port, DISC_CLASS_PWR);
    DelayMs(10);
    
    /* Enable port timed faults */
    WriteTps(nTps, PORT_CONTROL + port, Ctrl[port] & ~PORT_FAULT_DISABLE_MASK);

    return Ctrl[port];
}

int tps_ramp_down(ulong port, int nTps)
{
    /* Set control register to ramp mode */
    Ramp(nTps);
    /* Set the port to ramp down */
    PortMode(nTps, port, PWR_DOWN);
    /* Allow 90mS to ramp down */
    DelayMs(90);
    Null(nTps);
    PortMode(nTps, port, PORT_OFF);

    return Ctrl[port];
}

static int ZeroCrossUp(int nTps)
{
    /* Limit the delay waiting for CINT to rise thru zero cross (to the maximum interval timeout [of 65mSec]) */
    int i;
    char data;
    int us = 0;
    for (i = 0; i < 65; i++) {
        us += DelayMs(1);
	if (ReadTps(nTps, A2D_HIGH, &data)) {
            printf("i2c read failed\n");
	    return us;
	}
	if (data & A2D_DONE_MASK) {
	    break;
	}
    }
    return us;
}

static int ZeroCrossDown(int nTps)
{
    /* Limit the delay waiting for CINT to rise thru zero cross (to the maximum interval timeout [of 65mSec]) */
    int i;
    char data;
    int us = 0;
    for (i = 0; i < 65; i++) {
        us += DelayMs(1);
	if (ReadTps(nTps, A2D_HIGH, &data)) {
            printf("i2c read failed\n");
	    return us;
	}
	if (!(data & A2D_DONE_MASK)) {
	    return us;
	}
    }
    printf("timed out\n");
    return us;
}

static int A2d(int nTps, short *A2d1, short *A2d2)
{
  char data;
  int us;
  int a2d;
  
  /* 4.4V settling time (4.4V typ settling is 1.8mS).
   * Allow 5mS before releasing CINT hold switch. */
  us = DelayMs(5);
  /* Wait for rising edge */
  us += ZeroCrossUp(nTps);
  
  /* Wait until at least 2mS past zero cross to start checking for next event */
  us += DelayMs(2);
  /* Wait for falling edge */
  us += ZeroCrossUp(nTps);

  /* Read first value */
  ReadTps(nTps, A2D_HIGH, &data);
  a2d = (data & A2D_VALUE_MASK) << A2D_VALUE_POS;
  ReadTps(nTps, A2D_LOW, &data);
  a2d |= data;
  *A2d1 = (short) (a2d * calib_factor / 1000);

  /* Allow 7mS of line settling time to reach 8.8V */
  us += DelayMs(7);
  /* Wait until at least 2mS past zero cross to start checking for next event */
  us += DelayMs(2);
  us += ZeroCrossUp(nTps);
  /* Wait until at least 2mS past zero cross to start checking for next event */
  us += DelayMs(2);
  /* Wait for falling edge */
  us += ZeroCrossUp(nTps);

  // Read second value
  ReadTps(nTps, A2D_HIGH, &data);
  a2d = (data & A2D_VALUE_MASK) << A2D_VALUE_POS;
  ReadTps(nTps, A2D_LOW, &data);
  a2d |= data;
  *A2d2 = (short) (a2d * calib_factor / 1000);

  return us;
}

int tps_discover_resistor(ulong port, int nTps)
{
    long Value;
    short A2d1, A2d2;

    /* Set A/D input to current mode. Voltage mode would force off Discovery 2 level */
    Current(nTps);
    Null(nTps);
    /* Force off first */
    PortMode(nTps, port, PORT_OFF);
    /* Set common control to DISCOVER. DISCOVER MUST be before a
       mode change or the A/D will not trigger */
    Discovery(nTps);
    /* Mode change to start the A/D */
    PortMode(nTps, port, DISC_CLASS_PWR);
    /* 802.3af spec requires 60mS prior to discovery start for line to settle */
    DelayMs(65);
    /* Perform conversions. Results in Tps.A2d1 and Tps.A2d2 */
    A2d(nTps, &A2d1, &A2d2);
    printf("A2D %u %u\n", A2d1, A2d2);
    /* Return common control register back to inactive state  */
    Null(nTps);
    /* Return port mode control back to off state */
    PortMode(nTps, port, PORT_OFF);
    Value = (1321 - A2d2)/25 + 25;
    printf("Value %ld\n", Value);

    if ((Value >= 28) || (Value <= 19)) {
        return 0;
    } else {
        return 1;
    }
}

int tps_discover_capacitor(ulong port, int nTps)
{
    int present = 0;
#if 1
    int mV[2];

    mV[0] = tps_sample(port, nTps, 1);
    if (mV[0] < 3000) {
        ClassLim(port, nTps, 1, 1); /* 20 mA */
	WriteTps(nTps, PORT_CONTROL + port, 0);
	Class(nTps);
	WriteTps(nTps, PORT_CONTROL + port, 1);
	mV[0] = tps_sample(port, nTps, 1);
	DelayMs(1000);
	mV[1] = tps_sample(port, nTps, 1);
	printf("Voltages %d %d mV\n", mV[0], mV[1]);
	if ((mV[0] > mV[1]) && (mV[1] > 3000)) {
	    present = 1;
	}
    }
#else
    short A2d1, A2d2;
    short Volts1, Volts2;
    short DeltaVolts;
    short Cap;

    Voltage(nTps);
    /* Set common control to sample. Sample MUST be before a
     * mode change or the A/D will not trigger */
    Sample(nTps);
    /* Mode change to start the A/D */
    PortModeOpen(port, nTps, SAMPLE_PWR_CURRENT);
    A2d(nTps, &A2d1, &A2d2);
    /* Return common control register back to inactive state  */
    Null(nTps);
    /* Return port mode control back to state prior to this routine */
    PortModeClose(port, nTps);
    /* Not scaled by 9 to maintain higher resolution */
    Volts1 = A2d2;
    /* Calibration correction factor */
    Volts1 = (Volts1 * 1000) / 929;
    printf("Volts %d %d\n", A2d2, Volts1);
    if (Volts1 * 3 / 10 < (3 * 9)) {  
        /* Check to see that the port has discharged below 3 volts before pulsing */
        int Time;
	/* Set A/D input to voltage mode.  */
	Voltage(nTps);
	/* Set class current compliance to 20mA */
	ClassLim(port, nTps, 1, 1);
	/* Set common control to CLASS. CLASS MUST be before
	 * a mode change or the A/D will not trigger */
	Class(nTps);
	/* Mode change to start the A/D */
	PortMode(nTps, port, DISC_CLASS_PWR);
	/* Perform conversions */
	Time = A2d(nTps, &A2d1, &A2d2);
	/* Return common control register back to inactive state */
	Null(nTps);
	/* Return port mode control back to off state */
	PortMode(nTps, port, PORT_OFF);
	/* Not scaled down by 9, to maintain higher resolution */
	Volts2 = A2d2;
	Volts2 = (Volts2 * 1000);
	/* To get delta volts, multiply by 3/(10*9) */
	DeltaVolts = Volts2 - Volts1;
	printf("Volts %d %d %d, Time %d\n", Volts2, Volts1, DeltaVolts, Time);
	/* Continue if the cap final voltage is in the range os 1/9 to 17 *9*10/3 Volts */
	if ((DeltaVolts > 1) && (DeltaVolts < (17 * 30))) {
#define Current_20mA 25460
            Cap = (Time  * (Current_20mA)) / (118605 * DeltaVolts);
	    printf("Cap %d\n", Cap);
	    /* Legacy1 detect if Cap > 100uF and less than 765uF (3X value due to GUI) */
	    if ((Cap > 33) && (Cap < 255)) {
	        present = 1;
	    }
	}
    }
#endif
    return present;
}

int tps_sample(ulong port, int nTps, int voltage)
{
    short A2d1, A2d2;
    if (voltage) {
        Voltage(nTps);
    } else {
        Current(nTps);
    }
    /* Set common control to sample. Sample MUST be before a 
       mode change or the A/D will not trigger */
    Sample(nTps);
    /* Mode change to start the A/D */
    PortModeOpen(port, nTps, SAMPLE_PWR_CURRENT);
    Null(nTps);
    /* Return port mode control back to state prior to this routine */
    PortModeClose(port, nTps);
    DelayMs(1000);
    A2d(nTps, &A2d1, &A2d2);
    //printf("A2ds %u %d\n", A2d1, A2d2);
    if (!voltage) {
        /* current */
        int Current;
	Current = A2d1 * 100 / 472;
	//printf("Current %d mA\n", Current);
	return Current; /* mA */
    } else {
        /* Voltage */
        int Voltage;
	Voltage = A2d1 * 100000 / 3123 ;
	//printf("Voltage %d mV\n", Voltage);
	return Voltage; /* mV */
    }
}

int tps_calibrate(ulong port, int nTps)
{
#define SAMPLES 3
    int mV[SAMPLES];
    int mean_mV;
    int i;

    /* set calibration factor to default */
    calib_factor = 1000;

    /* ramp port up */
    tps_ramp_up(port, nTps, 16);
    DelayMs(1000);
    for (i = 0; i < SAMPLES; i++) {
	mV[i] = tps_sample(port, nTps, 1);
    }
    tps_ramp_down(port, nTps);

    /* compute mean of samples */
    mean_mV = (mV[0] < mV[1])? mV[1]: mV[0];
    mean_mV = (mean_mV < mV[2])? mean_mV : mV[2];

    calib_factor = 47900000/mean_mV;

    printf("Calibration factor %d\n", calib_factor);

    return calib_factor;
}

int tps_reg_dump(int nTps)
{
    ulong port;
    char data;

    for (port = 0; port < TPS_PORTS; port++) {
        ReadTps(nTps, PORT_STATUS + port, &data);
        printf("Port %d: Control %02x, Status %02x\n",
               port, Ctrl[port], data);
    }
    ReadTps(nTps, A2D_HIGH, &data);
    printf("A2D high: %02x\n", data);
    ReadTps(nTps, A2D_LOW, &data);
    printf("A2D low:  %02x\n", data);
    ReadTps(nTps, CHIP_ID_REV, &data);
    printf("Chip Id:  %02x\n", data);

    return 0;
}
