#include <ppcboot.h>
#include <command.h>
#include <net.h>
#include <watchdog.h>
#include "cmd_malbec.h"
#include "cmd_elf.h"
#include "cmd_boot.h"
#include "ns16550.h"
#include "../disk/part_dos.h"
#include "flash.h"

#include "../../../cmn/util/cab_cpld.h"
#include "../../../cmn/util/cab_fpga.h"
#include "../../../cmn/util/robo.h"
#include "../../../cmn/util/aruba_image.h"

#define TIMEOUT	       2		/* Seconds before failing */
#define PKTSIZEMIN     64
#define PHY_REGS       32
#define PHY_DEVS       64
#define PHY_PREAMBLE   32

#define CHAR(c)       (((c >= 0x20) && (c <= 0x7e))? (char)c: '.')

extern struct NS16550 *COM_PORTS[];
static u_long loopback_length;
static u_long loopback_count;
static u_long loopback_vlan;
static u_long loopback_incr;
static u_long loopback_received;
static int loopback_rcv;
void LoopbackRequest(void);
static unsigned long adjustSectors(unsigned long lba512);

static void
LoopbackTimeout (void)
{
    puts ("Timeout expired\n");
    NetState = NETLOOP_FAIL;
}

static void
LoopbackHandler (uchar * pkt, unsigned dest, unsigned src, unsigned len)
{
    int i;

    if (len != loopback_length) {
        printf("Error: length mismatch %d != %ld\n",
	       len, loopback_length);
	goto error;
    }
    for (i = 0; i < len; i++, pkt++) {
        if (*pkt != NetTxPacket[i]) {
	    printf("Error: byte %d mismatch: %02x != %02x\n",
		   i, *pkt, NetTxPacket[i]);
	    goto error;
	}
    }
    loopback_received++;
    
    if (loopback_incr) {
        if (loopback_length++ >= 1514) {
	    loopback_length = PKTSIZEMIN;
	}
    }

    if (serial_tstc() || 
	(loopback_count && (loopback_count-- == 1))) {
        NetState = NETLOOP_SUCCESS;
    } else {
        NetSendPacket(NetTxPacket, loopback_length);
	NetSetTimeout (TIMEOUT * CFG_HZ, LoopbackTimeout);
    }
    return;

 error:
    NetState = NETLOOP_FAIL;
}

void LoopbackRequest(void)
{
    /* called when a packet is received during ether test */
    int i;
    volatile uchar *pkt;

    pkt = NetTxPacket;

    NetSetEther(pkt, NetBcastAddr, PROT_TAG);

    pkt += ETHER_HDR_SIZE;
    *((ushort *)pkt) = loopback_vlan; /* VLAN ID */
    pkt += sizeof(ushort);

    for (i = 0; i < PKTSIZE - ETHER_HDR_SIZE - sizeof(ushort); i++) {
        pkt[i] = i & 0xff;
    }

    if (!loopback_rcv) {
        while (loopback_count--) {
	    NetSendPacket(NetTxPacket, loopback_length);
	}
        NetState = NETLOOP_SUCCESS;
    } else {

        NetSetHandler(LoopbackHandler);
	NetSetTimeout(TIMEOUT * CFG_HZ, LoopbackTimeout);

	NetSendPacket(NetTxPacket, loopback_length);
    }
}

int do_malbec_ether(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong length = PKTSIZEMIN, count = 0, incr = 0, vlan = 1;

    if ((argc < 2) || (argc > 6)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (eth_set(argv[1])) {
        printf("unknown device %s\n", argv[1]);
	return 1;
    }

    if (argc > 2) {
        if (argv[2][0] == 'i') {
  	    incr = 1;
	} else {
	    length = simple_strtoul(argv[2], NULL, 16);
	    if ((length < PKTSIZEMIN) || (length > 1514)) {
	        printf("length is invalid: %ld\n", length);
		return 1;
	    }
	}
    }

    if (argc > 3) {
        count = simple_strtoul(argv[3], NULL, 16);
	if (count == 0) {
	    printf("count is invalid: %ld\n", count);
	    return 1;
	}
    }

    if (argc > 4) {
        vlan = simple_strtoul(argv[4], NULL, 16);
	if (vlan == 0 || vlan > 4095) {
	    printf("vlan is invalid: %ld\n", vlan);
	    return 1;
	}
    }

    if (argc > 5) {
        /* do bother receiving */
        loopback_rcv = 0;
    } else {
        loopback_rcv = 1;
    }

    loopback_length = length;
    loopback_incr = incr;
    loopback_count = count;
    loopback_vlan = vlan;
    loopback_received = 0;

    NetLoop(bd, LOOPBACK);

    printf("%lx packets successfully received\n", loopback_received);

    return 0;
}

static void
uart_putc(int uart, const char c)
{
    if (c == '\n') {
        NS16550_putc(COM_PORTS[uart], '\r');
    }
    NS16550_putc(COM_PORTS[uart], c);
}

static void
uart_puts (int uart, const char *s)
{
    while (*s) {
        uart_putc (uart, *s++);
    }
}

static int
uart_getc(int uart)
{
    return NS16550_getc(COM_PORTS[uart]);
}

static int
uart_tstc(int uart)
{
    return NS16550_tstc(COM_PORTS[uart]);
}

static int
uart_space(int uart)
{
    return NS16550_space(COM_PORTS[uart]);
}

#define BUF_SIZE 2048
#define UART_NEXT(I) (((I) + 1) & (BUF_SIZE-1))
char uart_buffer[UARTS][BUF_SIZE];


int do_malbec_uartb(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    char c;
    int i;
    int prod[UARTS] = {0};
    int cons[UARTS] = {0};

    NS16550_init(1, CONFIG_BUS_CLK_FREQ / 16 / 115200);

    if (argc > 1) {
        for (i = 1; i < argc; i++) {
          uart_puts(1, argv[i]);
          uart_putc(1, ' ');
        }
	uart_putc(1, '\r');
        return 0;
    }

    printf("Connecting UART A to B...\nPress 'Q' to quit\n");
    while (1) {
        WATCHDOG_RESET();
	for (i = 0; i < UARTS; i++) {
	    if (uart_tstc(i) &&
		(UART_NEXT(prod[i]) != cons[i])) {
	        c = uart_getc(i);
		if ((c == 'Q') && (i == 0)) return 0;
		uart_buffer[i][prod[i]] = c;
		prod[i] = UART_NEXT(prod[i]);
	    }
	    if ((prod[i] != cons[i]) &&
		uart_space(i^1)) {
	        uart_putc(i^1, uart_buffer[i][cons[i]]);
		cons[i] = UART_NEXT(cons[i]);
	    }
	}
    }

    return 0;
}

#if (CONFIG_COMMANDS & CFG_CMD_IDE)
extern block_dev_desc_t * ide_get_dev(int dev);

static int le32_to_int(unsigned char *le32)
{
    return ((le32[3] << 24) +
	    (le32[2] << 16) +
	    (le32[1] << 8) +
	     le32[0]
	   );
}

static int upgrade_device;
static int upgrade_block;
static int upgrade_end;
static int upgrade_cksum;
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */

int ide_tftp_write(ulong offset, u8 *src, unsigned len)
{
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    block_dev_desc_t *dev;
    int i, *ptr;

    ptr = (int *) src;
    for (i = 0; i < len/sizeof(int); i++, ptr++) {
        upgrade_cksum += *ptr;
    }

    dev = ide_get_dev(upgrade_device);
    for (; len && (upgrade_block < upgrade_end); upgrade_block++) {
        if (dev->block_write(dev->dev, upgrade_block, 1, (ulong*)src) != 1) {
	    printf("%d write failed\n", upgrade_block);
	}
	if (len < dev->blksz) {
	    len = 0;
	} else {
	    len -= dev->blksz;
	}
	src += dev->blksz;
    }
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */
    return len? ERR_INVAL: ERR_OK;
}

#define DOS_PARTITIONS 4

int do_malbec_upgrade(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    block_dev_desc_t *dev;
    dos_partition_t *pt;
    aruba_image_t *ar;
    int block, count;
    unsigned device;
    ulong lba512;
    unsigned partition;
    char c;

    if (argc == 2) {
        if (argv[1][0] == 'i') {
            run_command("fpga p", bd, 0);
            run_command("cab i", bd, 0);
            run_command("robo i b", bd, 0);
            run_command("uartb eforw eth1 eth2", bd, 0);
            return 0;
        }
        return 1;
    }

    if (argc != 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][1] == ':') {
        device = simple_strtoul(argv[1], NULL, 16);
        partition = simple_strtoul(&argv[1][2], NULL, 16);
    } else {
        device = 0;
        partition = simple_strtoul(argv[1], NULL, 16);
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("%x: Device invalid\n", device);
	return 1;
    }
    if (partition >= DOS_PARTITIONS) { /* max 4 partitions */
        printf("%x: Partition invalid\n", partition);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("%x: Device invalid\n", device);
	return 1;
    }

    /* valid device? */
    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("%x: Device not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    /* valid partition? */
    if (dev->part_type != PART_TYPE_DOS) {
        printf("Partition type (if any) unrecognized\n");
	return 1;
    }

    /* Read partition information */
    if (dev->block_read(dev->dev, 0, 1, (ulong *)buffer) != 1) {
        printf("%d: Read failed\n", device);
	return 1;
    }

    /* The first two partitions can contain images */
    pt = (dos_partition_t *) (buffer + DOS_PART_TBL_OFFSET) + partition;
    if (pt->sys_ind == 0) {
        printf("%x: Partition not created\n", partition);
    }

    block = le32_to_int (pt->start4);
    count = block + le32_to_int (pt->size4);

    /* save state */
    upgrade_device = device;
    upgrade_block = block;
    upgrade_end = count;
    upgrade_cksum = 0;
    copy_filename (BootFile, argv[2], sizeof(BootFile));
    load_addr = CFG_ATA_BASE_ADDR;

    /* fire off TFTP */
    eth_set("i82559#0");
    if ((ar->size = NetLoop(bd, TFTP)) == 0) {
        return 1;
    }

    memcpy(buffer, (u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET, CFG_ENV_SECT_SIZE);
    if (buffer[0] != (device<<4|partition)) {
        printf("Try to boot this partition first ([y]/n)? ");
	if ((c = uart_getc(0)) != 'n') {
	    buffer[0] = device<<4|partition;
	    flash_write(buffer,
			CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET,
			CFG_ENV_SECT_SIZE);
	}
	printf("%c\n", c);
    }

#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */
    return (0);
}

#define FLASH_PARTS 4
#define FLASH_MARGIN 136
dos_partition_t part[FLASH_PARTS] = {
  { 0x00, 0x01, 0x01, 0x00, 0x83, 0x07, 0x60, 0x7f,
    { 0x20, 0x00, 0x00, 0x00 }, { 0xe0, 0x7f, 0x01, 0x00 } },
  { 0x00, 0x00, 0x41, 0x80, 0x83, 0x07, 0xe0, 0x00,
    { 0x00, 0x80, 0x01, 0x00 }, { 0x00, 0x81, 0x01, 0x00 } },
  { 0x00, 0x00, 0xc1, 0x01, 0x83, 0x07, 0xe0, 0xd1,
    { 0x00, 0x01, 0x03, 0x00 }, { 0x00, 0xd1, 0x00, 0x00 } },
  { 0x00 },
};

static unsigned long adjustSectors(unsigned long lba512)
{
    // printf("old lba512 = %d\n", (int) lba512);
    if((lba512 > 500000) && ( lba512 < 512000)) {
        //
        // 256 MB flash
        //
        // AVED comes in with 500400
        // WINTEC comes in with 500736
        //
        // Some web sites mention there are two types of 256 MB flashes
        // One that has 1954 blocks and one that has 1956 blocks.
        // To be safe and not trust the numbers passed in by the device 
        // we go with the smaller number of blocks i.e. 1954.
        //
        // 1954 did not work well with WINTEC flashes we try 1950
        // number of sectors = number of blocks(1950) * sectors(32) * heads(8)
        //                   = 499200
        //
        if(lba512 > 500224) {
                lba512 = 499200; // 1950 * 32 * 8;
        }
    } else {
        // 128 MB flash and 
        // all other sizes except 256MB 
        lba512 -= FLASH_MARGIN; /* SAMEER - 2/24/04 - remove sectors as margin */
    }
    // printf("new lba512 = %d\n", (int) lba512);
    return lba512;
}

int do_malbec_part(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    block_dev_desc_t *dev;
    ulong lba512;
    ulong size;
    unsigned device;
    int j;

    if (argc > 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        device = simple_strtoul(argv[1], NULL, 16);
    } else {
        device = 0;
    }
    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("Device %x not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    /* disable interrupts */
    disable_interrupts();
    for (j = 0; j < FLASH_PARTS; j++) {
        if (j < 2) {
            size = (part[j].size4[0] |
                    (part[j].size4[1] << 8) |
                    (part[j].size4[2] << 16) |
                    (part[j].size4[3] << 24));
            lba512 -= size;
        } else if (j == 2) {
            part[j].size4[0] = lba512 & 0xff;
            part[j].size4[1] = (lba512 >>  8) & 0xff;
            part[j].size4[2] = (lba512 >> 16) & 0xff;
            part[j].size4[3] = (lba512 >> 24) & 0xff;
        }

        set_dos_partition(dev, &part[j], j);
    }
    /* re-enable interrupts */
    enable_interrupts();

    print_part_dos(dev);
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */

    return (0);
}

#if (CONFIG_COMMANDS & CFG_CMD_IDE)
static int verify_sector(block_dev_desc_t *dev, int sector, char *wbuf)
{
    int i;
    char rbuf[512];

    if (dev->block_read(dev->dev, sector, 1, (ulong*)rbuf) != 1) {
        printf("%d read failed\n", sector);
        return 0;
    }
    for (i = 0; i < 512; i++) {
        if (wbuf[i] != rbuf[i]) {
            printf("%d read data mismatch\n", sector);
            return 0;
        }
    }
    return 1;
}

static int format_sector(block_dev_desc_t *dev, int sector, char *wbuf)
{
    if (dev->block_write(dev->dev, sector, 1, (ulong*)wbuf) != 1) {
        printf("%d write failed\n", sector);
        return 0;
    }
    return 1;
}

static void format_partition(block_dev_desc_t *dev, int start, int end,
                             char *wbuf, int verify)
{
#define PUTC_MASK 0x7ff
    int i;

    for(i = start; i < end; i++){
        format_sector(dev, i, wbuf);
        if (!(i & PUTC_MASK)) {
            serial_putc('.');
        }
    }
    if (verify) {
        for(i = start; i < end; i++){
            if (!verify_sector(dev, i, wbuf)) {
                return;
            }
            if (!(i & PUTC_MASK)) {
                serial_putc(',');
            }
        }
    }
}
#endif

int do_malbec_format(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    block_dev_desc_t *dev;
    ulong lba512;
    unsigned int device,partition;
    char c,src[512];
    char formatPart0=0;
    char formatPart1=0;
    char formatPart2=0;
    int verify = 0;
    int start, end;
    dos_partition_t *pt;

    if (argc > 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        if (argv[1][1] == ':') {
	    device = simple_strtoul(argv[1], NULL, 16);
	    partition = simple_strtoul(&argv[1][2], NULL, 16);

	    switch(partition){
		    case 0: formatPart0=1; break;
		    case 1: formatPart1=1; break;
		    case 2: formatPart2=1; break;
		    default:
			printf("Invalid device %s\n", argv[1]);
			return 1;
	    }
    	    printf("Format will erase everything on the %s flash's partition %d.\n", (device?"external":"internal"), partition);

	} else {
	    device = simple_strtoul(argv[1], NULL, 16);
	    formatPart0=1;
	    formatPart1=1;
	    formatPart2=1;
    	    printf("Format will erase the entire %s flash.\n", (device?"external":"internal"));
	} 

        if (argc > 2) {
            verify = 1;
        }

    } else {
	/* default format device 0 all three partitions. */
	formatPart0=1;
	formatPart1=1;
	formatPart2=1;
        device = 0;
    	printf("Format will erase the entire %s flash.\n", (device?"external":"internal"));
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("Device %x not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    printf("Are you sure you want to continue (y/n)?");
    c=getc();
    if(c!='y' && c!='Y') {
	    printf("\nFormat Cancelled\n");
	    return 1;
    }
    printf("%c\nFormatting", c);

    memset(src,0xFF,sizeof(src));

    if(formatPart0) {
        pt = &part[0];
        start = le32_to_int (pt->start4);
        end = start + le32_to_int (pt->size4);
        format_partition(dev, start, end, src, verify);
    } 
    if(formatPart1) {
        pt = &part[1];
        start = le32_to_int (pt->start4);
        end = start + le32_to_int (pt->size4);
        format_partition(dev, start, end, src, verify);
    } 
    if(formatPart2) {
        pt = &part[2];
        start = le32_to_int (pt->start4);
        end = lba512;
        format_partition(dev, start, end, src, verify);
    } 
    printf("\ndone\n");

    /* partition it */
    return (do_malbec_part(cmdtp, bd, flag, 1, argv));
#else /* (CONFIG_COMMANDS & CFG_CMD_IDE) */

    return (0);
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */
}

static u8 *load_partition(bd_t *bd, unsigned device, unsigned partition,
			  u32 *len)
{
    u8 *addr = NULL;
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    block_dev_desc_t *dev;
    dos_partition_t *pt;
    aruba_image_t *ar;
    ulong lba512;
    int block, count;
    int cksum, i, *ptr;

    dev = ide_get_dev(device);
    if (!dev) return NULL;

    /* valid device? */
    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) return NULL;
    lba512 = adjustSectors(lba512);

    /* valid partition? */
    if (dev->part_type != PART_TYPE_DOS) {
        printf("Device %d partition type (if any) unrecognized\n", device);
	return NULL;
    }

    /* Read partition information */
    if (dev->block_read(dev->dev, 0, 1, (ulong *)buffer) != 1) {
        printf("%d:%d:%d Read failed\n", device, partition, 0);
	return NULL;
    }

    /* valid partition? */
    pt = (dos_partition_t *) (buffer + DOS_PART_TBL_OFFSET) + partition;
    if (pt->sys_ind == 0) {
        return NULL;
    }

    block = le32_to_int (pt->start4);
    if (dev->block_read(dev->dev, block, 1, (ulong *)buffer) != 1) {
        printf("%d:%d:%d Read failed\n", device, partition, block);
	return NULL;
    }

    /* First block of partition contains aruba image header */
    ar = (aruba_image_t *) buffer;
    if (strcmp(ARUBA_IMAGE_MAGIC, ar->magic)) {
        printf("%d:%d Image not present\n", device, partition);
	return NULL;
    }
    if (!ar->valid) {
        printf("%d:%d Image not valid\n", device, partition);
	return NULL;
    }
    if (ar->machineType != ARUBA_IMAGE_MALBEC) {
        printf("%d:%d Image not for this platform\n", device, partition);
	return NULL;
    }

    /* remaining blocks contain actual image */
    addr = (u8 *) ((CFG_MEMTEST_END - ar->size) & 0xfff00000);
    count = (ar->size + dev->blksz - 1) / dev->blksz;
    /* read image into memory */
    printf("Loading image %d:%d", device, partition);
    if (dev->block_read(dev->dev, block + 1, count,
			(ulong *)addr) != count) {
        printf("%d:%d:%d read failed\n", device, partition, block+1);
	return NULL;
    }
    if (ctrlc())
        return NULL;

    if (ar->chksum) {
        /* verify checksum */
        printf("\nVerifying checksum... ");

	cksum = 0;
	ptr = (int *) buffer;
	for (i = 0; i < sizeof(buffer)/sizeof(int); i++, ptr++) {
	    cksum += *ptr;
	}

	ptr = (int *) addr;
	for (i = 0; i < ar->size/sizeof(int); i++, ptr++) {
	    cksum += *ptr;
	}

	if (cksum) {
	    /* invalid cksum */
	    printf("failed!\n");
	    return NULL;
	}

    }

    if (len) {
        *len = ar->size;
    }
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */

    return addr;
}

static void boot_partition(bd_t *bd, unsigned device, unsigned partition)
{
    u8 *addr;

    /* disable interrupts while accessing compact FLASH */
    disable_interrupts();
    addr = load_partition(bd, device, partition, NULL);
    enable_interrupts();
    if (addr) {
        printf("\nHalting active components...");
        CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET_ENABLE,
                           CAB_CPLD_RESET_ENABLE_VALUE);
        CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET,
                           CAB_CPLD_RESET_CFLASH);
        udelay (10 * 1000);
        CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET, 0);
        CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET_ENABLE, 0);
        printf("\nBooting image...\n");
	/* set load addr */
	load_addr = (ulong)addr;
	/* boot it (DOES NOT RETURN if successful) */
	do_bootelf(NULL, bd, 0, 0, NULL);
    }
}

int do_malbec_bootf(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    unsigned partition, part;
    unsigned device, dev;
    int search;

    if (argc > 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        search = 0;
        if (argv[1][1] == ':') {
	    device = simple_strtoul(argv[1], NULL, 16);
	    partition = simple_strtoul(&argv[1][2], NULL, 16);
	} else {
	    device = 0;
	    partition = simple_strtoul(argv[1], NULL, 16);
	}
    } else {
        device = *((u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET);
	partition = device & 0xf;
	device >>= 4;
	if ((device >= CFG_IDE_MAXDEVICE) ||
	    (partition >= DOS_PARTITIONS)) {
	    device = 0;
	    partition = 0;
	}
        search = 1;
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }
    if (partition >= DOS_PARTITIONS) { /* max 4 partitions */
        printf("Partition %x invalid\n", partition);
	return 1;
    }

    if (!search) {
        boot_partition(bd, device, partition);
	printf("Failed\n");
	return 1;
    }

    /* try to boot primary partition */
    boot_partition(bd, device, partition);

    /* failed, try to boot from any partition */
    for (dev = 0; dev < CFG_IDE_MAXDEVICE; dev++) {
        for (part = 0; part < DOS_PARTITIONS; part++) {
	    if ((part != partition) || (dev != device)) {
	        boot_partition(bd, dev, part);
	    }
	}
    }

    return 1;
}
