/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <mpc824x.h>

#define ROM_CS0_START	0xFF800000
#define ROM_CS1_START	0xFF000000

flash_info_t flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips    */

#if defined(CFG_ENV_IS_IN_FLASH)
# ifndef  CFG_ENV_ADDR
#  define CFG_ENV_ADDR  (CFG_FLASH_BASE + CFG_ENV_OFFSET)
# endif
# ifndef  CFG_ENV_SIZE
#  define CFG_ENV_SIZE  CFG_ENV_SECT_SIZE
# endif
# ifndef  CFG_ENV_SECT_SIZE
#  define CFG_ENV_SECT_SIZE  CFG_ENV_SIZE
# endif
#endif

/*-----------------------------------------------------------------------
 * Functions
 */
static int write_word (flash_info_t *info, ulong dest, ulong data);

/*flash command address offsets*/
#define ADDR0		0x5555
#define ADDR1		0x2AAA

#define FLASH_WORD_SIZE unsigned char

/*-----------------------------------------------------------------------
 */

static ulong flash_id(unsigned char mfct, unsigned char chip) __attribute__ ((const));

typedef struct
{
  FLASH_WORD_SIZE extval;
  ushort intval;
} map_entry;


static ulong
flash_id(unsigned char mfct, unsigned char chip)
{
    static const map_entry mfct_map[] = {
        {(FLASH_WORD_SIZE) AMD_MANUFACT,  (ushort) ((ulong) FLASH_MAN_AMD >> 16)},
	{(FLASH_WORD_SIZE) ATMEL_MANUFACT,(ushort) ((ulong) FLASH_MAN_ATMEL >> 16)},
	{(FLASH_WORD_SIZE) FUJ_MANUFACT,  (ushort) ((ulong) FLASH_MAN_FUJ >> 16)},
	{(FLASH_WORD_SIZE) STM_MANUFACT,  (ushort) ((ulong) FLASH_MAN_STM >> 16)},
	{(FLASH_WORD_SIZE) ST_MANUFACT,	  (ushort) ((ulong) FLASH_MAN_STM >> 16)},
	{(FLASH_WORD_SIZE) MT_MANUFACT,	  (ushort) ((ulong) FLASH_MAN_MT >> 16)},
	{(FLASH_WORD_SIZE) INTEL_MANUFACT,(ushort) ((ulong) FLASH_MAN_INTEL >> 16)},
	{(FLASH_WORD_SIZE) INTEL_ALT_MANU,(ushort) ((ulong) FLASH_MAN_INTEL >> 16)}
    };

    static const map_entry chip_map[] = {
        {ATMEL_ID_29LV020,	FLASH_AT29LV020},
	{ATMEL_ID_49xV002,	FLASH_AT49xV002},
	{ATMEL_ID_49xV040,	FLASH_AT49xV040},
    };

    const map_entry *p;
    ulong result = FLASH_UNKNOWN;

    /* find chip id */
    for (p = &chip_map[0]; p < &chip_map[sizeof chip_map / sizeof chip_map[0]]; p++) {
        if (p->extval == chip) {
	    result = FLASH_VENDMASK | p->intval;
	    break;
	}
    }

    /* find vendor id */
    for (p = &mfct_map[0]; p < &mfct_map[sizeof mfct_map / sizeof mfct_map[0]]; p++) {
        if (p->extval == mfct) {
	    result &= ~FLASH_VENDMASK;
	    result |= (ulong) p->intval << 16;
	    break;
	}
    }

    return result;
}



ulong
flash_init(void)
{
    ulong i;
    static const ulong flash_banks[] = CFG_FLASH_BANKS;
    ulong size = 0;

    /* Init: no FLASHes known */
    for (i = 0; i < CFG_MAX_FLASH_BANKS; i++) {
        flash_info_t * const pflinfo = &flash_info[i];
	pflinfo->flash_id = FLASH_UNKNOWN;
	pflinfo->size = 0;
	pflinfo->sector_count = 0;
    }

    for(i = 0; i < sizeof flash_banks / sizeof flash_banks[0]; i++) {
        flash_info_t * const pflinfo = &flash_info[i];
	const ulong base_address = flash_banks[i];
	volatile FLASH_WORD_SIZE * const flash = (FLASH_WORD_SIZE *) base_address;

	/* enter software product identification */
	flash[ADDR0] = (FLASH_WORD_SIZE)0x00AA00AA;
	flash[ADDR1] = (FLASH_WORD_SIZE)0x00550055;
	flash[ADDR0] = (FLASH_WORD_SIZE)0x00900090;
	__asm__ __volatile__("sync");
	udelay(20000);

	pflinfo->flash_id = flash_id(flash[0x0], flash[0x1]);

	/* exit software product identification */
	flash[ADDR0] = (FLASH_WORD_SIZE)0x00AA00AA;
	flash[ADDR1] = (FLASH_WORD_SIZE)0x00550055;
	flash[ADDR0] = (FLASH_WORD_SIZE)0x00F000F0;
	__asm__ __volatile__("sync");
      
	
	switch(pflinfo->flash_id & FLASH_TYPEMASK) {
	case FLASH_AT29LV020:
	    pflinfo->size = 256 << 10;
	    pflinfo->sector_count = 1; /* really 1024, but too many */
	    pflinfo->start[0] = base_address;
	    break;

	case FLASH_AT49xV002:
	    pflinfo->size = 256 << 10;
	    pflinfo->sector_count = 5;
	    pflinfo->start[0] = base_address;           /* main */
	    pflinfo->start[1] = base_address + 0x20000; /* main */
	    pflinfo->start[2] = base_address + 0x38000; /* param */
	    pflinfo->start[3] = base_address + 0x3a000; /* param */
	    pflinfo->start[4] = base_address + 0x3c000; /* boot */
	    break;

	case FLASH_AT49xV040:
	    pflinfo->size = 512 << 10;
	    pflinfo->sector_count = 1; /* ignoring boot block */
	    pflinfo->start[0] = base_address;
	    break;
	}

	size += pflinfo->size;
    }

    return size;
}

/*-----------------------------------------------------------------------
 */
void
flash_print_info(flash_info_t *info)
{
    static const char unk[] = "Unknown";
    const char *mfct = unk, *type = unk;
    unsigned int i;

    if (info->flash_id != FLASH_UNKNOWN) {
        switch(info->flash_id & FLASH_VENDMASK) {
	case FLASH_MAN_AMD:	mfct = "AMD";			break;
	case FLASH_MAN_FUJ:	mfct = "FUJITSU";		break;
	case FLASH_MAN_STM:	mfct = "STM";			break;
	case FLASH_MAN_SST:	mfct = "SST";			break;
	case FLASH_MAN_ST:	mfct = "ST";			break;
	case FLASH_MAN_BM:	mfct = "Bright Microelectonics";break;
	case FLASH_MAN_INTEL:	mfct = "Intel";			break;
	case FLASH_MAN_ATMEL:	mfct = "Atmel";			break;
	}

	switch(info->flash_id & FLASH_TYPEMASK) {
	case FLASH_AT29LV020:	type = "AT29LV020 (2 Mbit, uniform sector size)"; break;
	case FLASH_AT49xV002:	type = "AT49LV002 (2 Mbit, variable sector size)"; break;
	case FLASH_AT49xV040:	type = "AT49LV040 (4 Mbit, variable sector size)"; break;
	}
    }

    printf("\n  Brand: %s Type: %s\n"
	   "  Size: %lu KB in %d Sectors\n",
	   mfct,
	   type,
	   info->size >> 10,
	   info->sector_count
	   );

    printf ("  Sector Start Addresses:");

    for (i = 0; i < info->sector_count; i++) {
        ulong size;
	unsigned int erased;
	ulong * flash = (ulong *) info->start[i];

	/*
	 * Check if whole sector is erased
	 */
	size =
	  (i != (info->sector_count - 1)) ?
	  (info->start[i + 1] - info->start[i]) >> 2 :
	  (info->start[0] + info->size - info->start[i]) >> 2;

	for(
	    flash = (ulong *) info->start[i], erased = 1;
	    (flash != (ulong *) info->start[i] + size) && erased;
	    flash++
	    )
	  erased = *flash == ~0x0UL;

	printf(
	       "%s %08lX %s %s",
	       (i % 5) ? "" : "\n   ",
	       info->start[i],
	       erased ? "E" : " ",
	       info->protect[i] ? "RO" : "  "
	       );
    }

    puts("\n");
}

int
flash_erase(flash_info_t *info, int s_first, int s_last)
{
    volatile FLASH_WORD_SIZE * const flash = (volatile FLASH_WORD_SIZE *)info->start[0];

    switch(info->flash_id & FLASH_TYPEMASK) {
    case FLASH_AT29LV020:
        return (ERR_INVAL);
    default:
        /* erase entire chip */
        flash[ADDR0] = (FLASH_WORD_SIZE)0x00AA00AA;
        flash[ADDR1] = (FLASH_WORD_SIZE)0x00550055;
        flash[ADDR0] = (FLASH_WORD_SIZE)0x00800080;
        flash[ADDR0] = (FLASH_WORD_SIZE)0x00AA00AA;
        flash[ADDR1] = (FLASH_WORD_SIZE)0x00550055;
        flash[ADDR0] = (FLASH_WORD_SIZE)0x00100010;
	return (ERR_OK);
    }
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
    ulong data;
    int i, rc;


    switch(info->flash_id & FLASH_TYPEMASK) {
    case FLASH_AT29LV020:
        /* must write 256 B at a time */
        if (addr & 0xff) {
	    return (ERR_ALIGN);
	}
	cnt = (cnt + 0xff) & ~0xff;
	break;
    default:
        if (addr & 0x3) {
	    return (ERR_ALIGN);
	}
        cnt = (cnt + 0x3) & ~0x3;
        break;
    }

    while (cnt) {
        data = 0;
        for (i=0; i<4; ++i) {
            data = (data << 8) | *src++;
        }
        if ((rc = write_word(info, addr, data)) != ERR_OK) {
            return (rc);
        }
        addr += 4;
        cnt  -= 4;
    }

    return (ERR_OK);
}

/*-----------------------------------------------------------------------
 * Write a word to Flash
 */
static int write_word (flash_info_t *info, ulong dest, ulong data)
{
    volatile FLASH_WORD_SIZE *addr2 = (FLASH_WORD_SIZE *) info->start[0];
    volatile FLASH_WORD_SIZE *dest2 = (FLASH_WORD_SIZE *) dest;
    volatile FLASH_WORD_SIZE *data2 = (FLASH_WORD_SIZE *) &data;
    ulong start;
    int i;

    if (dest + 4 > info->start[0] + info->size) {
        return (ERR_INVAL);
    }

    switch(info->flash_id & FLASH_TYPEMASK) {
    case FLASH_AT29LV020:
        if (!(dest & 0xff)) {
	    udelay(20000);
            addr2[ADDR0] = (FLASH_WORD_SIZE)0x00AA00AA;
	    addr2[ADDR1] = (FLASH_WORD_SIZE)0x00550055;
	    addr2[ADDR0] = (FLASH_WORD_SIZE)0x00A000A0;
	}
	for (i=0; i<4/sizeof(FLASH_WORD_SIZE); i++) {
	    dest2[i] = data2[i];
	}
	break;

    default:
        /* Check if FLASH is (sufficiently) erased */
        if ((*dest2 & (FLASH_WORD_SIZE)data) != (FLASH_WORD_SIZE)data) {
            return (ERR_NOT_ERASED);
	}
	for (i=0; i<4/sizeof(FLASH_WORD_SIZE); i++) {
            addr2[ADDR0] = (FLASH_WORD_SIZE)0x00AA00AA;
	    addr2[ADDR1] = (FLASH_WORD_SIZE)0x00550055;
	    addr2[ADDR0] = (FLASH_WORD_SIZE)0x00A000A0;
	    dest2[i] = data2[i];

	    /* data polling for D7 */
	    start = get_timer (0);
	    while ((dest2[i] & (FLASH_WORD_SIZE)0x00800080) !=
		   (data2[i] & (FLASH_WORD_SIZE)0x00800080)) {
	        if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
                    return (ERR_TIMOUT);
		}
	    }
	}
	break;
    }
    return (ERR_OK);
}

/*-----------------------------------------------------------------------
 */
