/*
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <watchdog.h>
#include <malloc.h>
#include <net.h>

#include "pci.h"

#undef DEBUG

#if (CONFIG_COMMANDS & CFG_CMD_NET) && defined(CONFIG_NET_MULTI) \
	&& defined(CONFIG_NATSEMI)

/* PCI Registers.
 */
#define CBIO_MASK	0xffffff80

#define DP83816_VID	0x100B		/* National Manufacturer */
#define DP83816_DID	0x0020		/* Unique Device ID # */

/* Ethernet chip registers.
	 */
#define DP83816_CR	0x000		/* Command Register */
#define DP83816_CFG	0x004		/* Config Register */
#define DP83816_MEAR	0x008		/* EEPROM Access Register */
#define DP83816_TXDP	0x020		/* TX Descriptor Pointer Register */
#define DP83816_TXCFG	0x024		/* TX Config Register */
#define DP83816_RXDP	0x030		/* RX Descriptor Pointer Register */
#define DP83816_RXCFG	0x034		/* RX Config Register */
#define DP83816_RFCR	0x048		/* RX Config Register */

/* Register bits.
	 */
#define	CR_RST		0x00000100	/* software reset */
#define	CR_SWI		0x00000080	/* software interrupt */
#define	CR_RXR		0x00000020	/* receiver reset */
#define	CR_TXR		0x00000010	/* transmit reset */
#define	CR_RXD		0x00000008	/* receiver disable */
#define	CR_RXE		0x00000004	/* receiver enable */
#define	CR_TXD		0x00000002	/* transmit disable */
#define	CR_TXE		0x00000001	/* transmit enable */

#define	CFG_LNKSTS	0x80000000	/* link status (83815) */
#define	CFG_SPEED100	0x40000000	/* 100Mb/s (83815) */
#define	CFG_FDUP	0x20000000	/* full duplex (83815) */
#define	CFG_POL		0x10000000	/* 10Mb/s polarity (83815) */
#define	CFG_ANEG_DN	0x08000000	/* autonegotiation done (83815) */
#define	CFG_PHY_CFG	0x00fc0000	/* PHY configuration (83815) */
#define	CFG_PINT_ACEN	0x00020000	/* PHY interrupt auto clear (83815) */
#define	CFG_PAUSE_ADV	0x00010000	/* pause advertise (83815) */
#define	CFG_ANEG_SEL	0x0000e000	/* autonegotiation select (83815) */
#define	CFG_PHY_RST	0x00000400	/* PHY reset (83815) */
#define	CFG_PHY_DIS	0x00000200	/* PHY disable (83815) */
#define	CFG_EUPHCOMP	0x00000100	/* 83810 descriptor compat (83815) */
#define	CFG_REQALG	0x00000080	/* PCI bus request alg. */
#define	CFG_SB		0x00000040	/* single backoff */
#define	CFG_POW		0x00000020	/* program out of window timer */
#define	CFG_EXD		0x00000010	/* excessive defferal timer disable */
#define	CFG_PESEL	0x00000008	/* parity error detection action */
#define	CFG_BEM		0x00000001	/* big-endian mode */

#define	TXCFG_CSI	0x80000000	/* carrier sense ignore */
#define	TXCFG_HBI	0x40000000	/* heartbeat ignore */
#define	TXCFG_MLB	0x20000000	/* MAC loopback */
#define	TXCFG_ATP	0x10000000	/* automatic transmit padding */
#define	TXCFG_MXDMA	0x00700000	/* max DMA burst size */
#define	TXCFG_MXDMA_512	0x00000000	/*     512 bytes */
#define	TXCFG_MXDMA_4	0x00100000	/*       4 bytes */
#define	TXCFG_MXDMA_8	0x00200000	/*       8 bytes */
#define	TXCFG_MXDMA_16	0x00300000	/*      16 bytes */
#define	TXCFG_MXDMA_32	0x00400000	/*      32 bytes */
#define	TXCFG_MXDMA_64	0x00500000	/*      64 bytes */
#define	TXCFG_MXDMA_128	0x00600000	/*     128 bytes */
#define	TXCFG_MXDMA_256	0x00700000	/*     256 bytes */
#define TXCFG_SBOA      0x00040000      /* standard back-off algorithm */
#define	TXCFG_FLTH	0x00003f00	/* Tx fill threshold */
#define	TXCFG_FLTH_SHIFT 8
#define	TXCFG_DRTH	0x0000003f	/* Tx drain threshold */
#define	TXCFG_DRTH_SHIFT 0

#define	RXCFG_AEP	0x80000000	/* accept error packets */
#define	RXCFG_ARP	0x40000000	/* accept runt packets */
#define	RXCFG_ATX	0x10000000	/* accept transmit packets */
#define	RXCFG_ALP	0x08000000	/* accept long packets */
#define	RXCFG_MXDMA	0x00700000	/* max DMA burst size */
#define	RXCFG_MXDMA_512	0x00000000	/*     512 bytes */
#define	RXCFG_MXDMA_4	0x00100000	/*       4 bytes */
#define	RXCFG_MXDMA_8	0x00200000	/*       8 bytes */
#define	RXCFG_MXDMA_16	0x00300000	/*      16 bytes */
#define	RXCFG_MXDMA_32	0x00400000	/*      32 bytes */
#define	RXCFG_MXDMA_64	0x00500000	/*      64 bytes */
#define	RXCFG_MXDMA_128	0x00600000	/*     128 bytes */
#define	RXCFG_MXDMA_256	0x00700000	/*     256 bytes */
#define	RXCFG_DRTH	0x0000003e
#define	RXCFG_DRTH_SHIFT 1

#define	RFCR_RFEN	0x80000000	/* Rx filter enable */
#define	RFCR_AAB	0x40000000	/* accept all broadcast */
#define	RFCR_AAM	0x20000000	/* accept all multicast */
#define	RFCR_AAP	0x10000000	/* accept all physical */
#define	RFCR_APM	0x08000000	/* accept perfect match (83815) */
#define	RFCR_APAT	0x07800000	/* accept pattern match (83815) */
#define	RFCR_AARP	0x00400000	/* accept ARP (83815) */
#define	RFCR_MHEN	0x00200000	/* multicast hash enable (83815) */
#define	RFCR_UHEN	0x00100000	/* unicast hash enable (83815) */
#define	RFCR_ULM	0x00080000	/* U/L bit mask (83815) */
#define	RFCR_NS_RFADDR	0x000003ff	/* Rx filter ext reg address (83815) */
#define	RFCR_RFADDR	0x000f0000	/* Rx filter address */

/* Descriptor bits.
	 */
#define R_OWN		0x80000000	/* Own Bit */
#define RD_MORE		0x40000000	/* More descriptors */
#define RD_OK		0x08000000	/* packet OK */
#define RD_SIZE		0x000007ff	/* size mask */

#define T_OWN		0x80000000	/* Own Bit */
#define TD_MORE		0x40000000	/* More descriptors */
#define TD_OK		0x08000000	/* packet OK */
#define TD_SIZE		0x000007ff	/* size mask */

#define RST_TIMEOUT 10000

#define RESET_DP83816(dev) {\
    int j;\
    OUTL(dev, CR_RST, DP83816_CR);\
    for (j = 0; j < RST_TIMEOUT; j++) {\
       if (!(INL(dev, DP83816_CR) & CR_RST)) break;\
    }\
}

#define START_DP83816(dev) {\
    OUTL(dev, CR_RXE|CR_TXE, DP83816_CR);		/* Enable the TX and/or RX */\
}

#define STOP_DP83816(dev) {\
    OUTL(dev, CR_RXD|CR_TXD, DP83816_CR);		/* Disable the TX and/or RX */ \
}

#define NUM_RX_DESC PKTBUFSRX           /* currently 4 */
#define NUM_TX_DESC 2			/* Number of TX descriptors   */
#define RX_BUFF_SZ  PKTSIZE_ALIGN

#define TOUT_LOOP   1000000

#define SETUP_FRAME_LEN 192
#define ETH_ALEN	6


struct dp83816_desc {
	u32 next;
	volatile u32 cmdsts;
	u32 buf;
};

static struct dp83816_desc rx_ring[NUM_RX_DESC]; /* RX descriptor ring         */
static struct dp83816_desc tx_ring[NUM_TX_DESC]; /* TX descriptor ring         */
static u32 rx_new;                               /* RX descriptor ring pointer */
static u32 tx_new;                               /* TX descriptor ring pointer */

static void  read_hw_addr(struct eth_device* dev, bd_t * bis);
static int   dp83816_init(struct eth_device* dev, bd_t* bis);
static int   dp83816_send(struct eth_device* dev, volatile void *packet, int length);
static int   dp83816_recv(struct eth_device* dev);
static void  dp83816_halt(struct eth_device* dev);

#define pci_mem_to_phys(x) (x)
#define pci_ram_to_mem(x) (x)

static int INL(struct eth_device* dev, u_long addr)
{
	return le32_to_cpu(*(volatile u_long *)pci_mem_to_phys(addr + dev->iobase));
}

static void OUTL(struct eth_device* dev, int command, u_long addr)
{
	*(volatile u_long *)pci_mem_to_phys(addr + dev->iobase) = cpu_to_le32(command);
}

int dp83816_initialize(bd_t *bis)
{
	int             	card_number, i, status = 0;
	short                   command;
	int                     bus, device, func;
	unsigned char   	timer;
	uint32			iobase;
	struct eth_device* 	dev;

	for(card_number=i=0; ; i++) {
		status = pci_dev_find(DP83816_VID, DP83816_DID, i,
				      &bus, &device, &func);
		if (status == ERROR) {
			break;
		}

		pci_config_inw(bus, device, func, PCI_CFG_COMMAND, &command);
		command |= PCI_CMD_MASTER_ENABLE | PCI_CMD_IO_ENABLE | PCI_CMD_MEM_ENABLE;
		pci_config_outw(bus, device, func, PCI_CFG_COMMAND, command);

		/* Check the latency timer for values >= 0x60. */
		pci_config_inb(bus, device, func, PCI_CFG_LATENCY_TIMER, &timer);
		if (timer < 0x60) {
		        pci_config_outb(bus, device, func, PCI_CFG_LATENCY_TIMER, 0x60);
		}

		/* read BAR for memory space access */
		pci_config_inl(bus, device, func, PCI_CFG_BASE_ADDRESS_1, &iobase);
		iobase &= CBIO_MASK;

#ifdef DEBUG
		printf("dp83816: National Semiconductor DP83816 PCI Device @0x%x\n", iobase);
#endif

		dev = (struct eth_device*) malloc(sizeof *dev);

		sprintf(dev->name, "dp83816#%d", i);
		dev->iobase = iobase;
		dev->priv   = (void*) ((bus << 16) | (device << 11) | (func << 8));
		dev->init   = dp83816_init;
		dev->halt   = dp83816_halt;
		dev->send   = dp83816_send;
		dev->recv   = dp83816_recv;

		udelay(10 * 1000);

		if (0) {
		    /* NO EEPROM */
		    read_hw_addr(dev, bis);
		}

		eth_register(dev);

		card_number++;
	}

	return card_number;
}

#ifdef DEBUG
static void dp83816_dump(struct eth_device* dev)
{
    int i;
    for (i = 0; i < 0x88; i+= 4) {
        printf("   %02x: %08x", i, INL(dev, i));
	if (i & 0x4) printf("\n");
    }
}
#endif /* DEBUG */

static int dp83816_init(struct eth_device* dev, bd_t* bis)
{
	int		i;

	RESET_DP83816(dev);

	if (INL(dev, DP83816_CR) & CR_RST) {
		printf("Error: Cannot reset ethernet controller.\n");
		return 0;
	}

        /* global configuration */
        OUTL(dev, CFG_ANEG_SEL, DP83816_CFG);

	for (i = 0; i < NUM_RX_DESC; i++) {
                rx_ring[i].next = cpu_to_le32(pci_ram_to_mem((u32) &rx_ring[(i+1) % NUM_RX_DESC]));
		rx_ring[i].cmdsts = cpu_to_le32(RX_BUFF_SZ);
		rx_ring[i].buf = cpu_to_le32(pci_ram_to_mem((u32) NetRxPackets[i]));
	}

	for (i=0; i < NUM_TX_DESC; i++) {
		tx_ring[i].next = cpu_to_le32(0);
		tx_ring[i].cmdsts = cpu_to_le32(0);
		tx_ring[i].buf = cpu_to_le32(0);
	}

	/* Tell the adapter where the TX/RX rings are located. */
	OUTL(dev, pci_ram_to_mem((u32) &rx_ring), DP83816_RXDP);
	OUTL(dev, pci_ram_to_mem((u32) &tx_ring), DP83816_TXDP);

        /* configure TX/RX engines */
        OUTL(dev, 
             (TXCFG_CSI|TXCFG_HBI|TXCFG_ATP|TXCFG_MXDMA_64|TXCFG_SBOA|
              (5 << TXCFG_FLTH_SHIFT)|
              (2 << TXCFG_DRTH_SHIFT)), 
             DP83816_TXCFG);
        OUTL(dev,
             (RXCFG_ATX|RXCFG_MXDMA_64|
              (16 << RXCFG_DRTH_SHIFT)),
             DP83816_RXCFG);
        OUTL(dev, (RFCR_RFEN|RFCR_AAB|RFCR_AAM|RFCR_AAP), DP83816_RFCR);

	START_DP83816(dev);

	tx_new = 0;
	rx_new = 0;

	return 1;
}

static int dp83816_send(struct eth_device* dev, volatile void *packet, int length)
{
	int		status = -1;
	int		i;

	if (length <= 0) {
		printf("%s: bad packet size: %d\n", dev->name, length);
		goto Done;
	}

	for(i = 0; tx_ring[tx_new].cmdsts & le32_to_cpu(T_OWN); i++) {
		if (i >= TOUT_LOOP) {
			printf("%s: prev tx buffer not ready\n", dev->name);
			goto Done;
		}
	}

	tx_ring[tx_new].buf    = cpu_to_le32(pci_ram_to_mem((u32) packet));
	tx_ring[tx_new].cmdsts = cpu_to_le32(T_OWN | length);

        /* set TX descriptor pointer */
	OUTL(dev, pci_ram_to_mem((u32) &tx_ring[tx_new]), DP83816_TXDP);
        /* kick transmitter */
        OUTL(dev, CR_TXE, DP83816_CR);

	for(i = 0; tx_ring[tx_new].cmdsts & le32_to_cpu(T_OWN); i++) {
		if (i >= TOUT_LOOP) {
			printf("%s: tx buffer not ready\n", dev->name);
#ifdef DEBUG
			dp83816_dump(dev);
#endif /* DEBUG */
			goto Done;
		}
	}

	tx_new = (tx_new + 1) % NUM_TX_DESC;
	status = length;

 Done:
	return status;
}

static int dp83816_recv(struct eth_device* dev)
{
        u32		status;
	int		length    = 0;

	for ( ; ; ) {
		WATCHDOG_RESET();

		status = (u32)le32_to_cpu(rx_ring[rx_new].cmdsts);

		if (!(status & R_OWN)) {
			break;
		}

                if (status & RD_OK) {
                  
                        /* A valid frame received.
                         */
                        length = (le32_to_cpu(rx_ring[rx_new].cmdsts) & RD_SIZE);

                        /* Pass the packet up to the protocol
                         * layers.
                         */
                        NetReceive(NetRxPackets[rx_new], length - 4);
                } else {
                        printf("\nRX error status = 0x%08X\n", status);
                }

                /* Change buffer ownership for this frame, back
                 * to the adapter.
                 */
                rx_ring[rx_new].cmdsts = cpu_to_le32(RX_BUFF_SZ);

		/* Update entry information.
		 */
		rx_new = (rx_new + 1) % NUM_RX_DESC;
	}

	return length;
}

static void dp83816_halt(struct eth_device* dev)
{
	STOP_DP83816(dev);
}

static void read_hw_addr(struct eth_device *dev, bd_t *bis)
{
	int i;

	for (i = 0; i < ETH_ALEN; i++) {
            dev->enetaddr[i] = bis->bi_enetaddr[i];
	}
}

#endif
