/*
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <malloc.h>
#include <net.h>

#include "pci.h"

#undef DEBUG

#if (CONFIG_COMMANDS & CFG_CMD_NET) && defined(CONFIG_NET_MULTI) \
	&& defined(CONFIG_SIBYTE)

/* PCI Registers.
 */

#define SIBYTE_VID	0x166d		/* Broadcom Manufacturer */
#define SIBYTE_DID	0x0001		/* Unique Device ID # */

#define SIBYTE_MB0      0x01000000
#define SIBYTE_MB1      0x01001000
#define SIBYTE_MB_SET_BIT  0x8

/* Descriptor bits.
 */
#define D_OWNER		0x80000000	/* Own Bit */
#define D_FAIL		0x00800000	/* Error Summary */
#define D_LENGTH        0x0000ffff      /* lenght mask */

/* Mailbox bits.
 */
#define M_DOWNLOAD      0x00000003      /* download prod image */
#define M_TXPKT		0x00000010	/* TX pkt */
#define M_RESET		0x00000020	/* Reset Bit */

#define RX_BUFF_SZ  PKTSIZE_ALIGN

#define TOUT_LOOP   1000000

#define SETUP_FRAME_LEN 192
#define ETH_ALEN	6

static void  read_hw_addr(struct eth_device* dev, bd_t * bis);

static int   sibyte_init(struct eth_device* dev, bd_t* bis);
static int   sibyte_send(struct eth_device* dev, volatile void *packet, int length);
static int   sibyte_recv(struct eth_device* dev);
static void  sibyte_halt(struct eth_device* dev);
static int   rx_next = 0;

#define pci_mem_to_phys(x) (x)
#define pci_ram_to_mem(x) (x)

#define RX_DESCRS 4

#if (PKTBUFSRX != RX_DESCRS)
#error "descriptor count does not match packet count"
#endif

typedef struct {
    volatile uint32 addr;
    volatile uint32 status;
} sibyte_descr_t;

typedef struct {
    volatile sibyte_descr_t tx_descr;
    volatile sibyte_descr_t rx_descr[RX_DESCRS];
} sibyte_drv_t;

sibyte_drv_t sd;

static void OUTL(struct eth_device* dev, int command, u_long addr)
{
    *(volatile u_long *)pci_mem_to_phys(addr + dev->iobase) = command;
}

int sibyte_initialize(bd_t *bis)
{
    int            	card_number, i, status = 0;
    short               command;
    int                 bus, device, func;
    unsigned char   	timer;
    uint32		mbase;
    uint32		mbbase[2];
    struct eth_device* 	dev;

    for(card_number=i=0; ; i++) {
        status = pci_dev_find(SIBYTE_VID, SIBYTE_DID, i,
			      &bus, &device, &func);
	if (status == ERROR) {
	    break;
	}

	pci_config_inw(bus, device, func, PCI_CFG_COMMAND, &command);
	command |= PCI_CMD_MASTER_ENABLE | PCI_CMD_MEM_ENABLE;
	pci_config_outw(bus, device, func, PCI_CFG_COMMAND, command);

	/* Check the latency timer for values >= 0x60. */
	pci_config_inb(bus, device, func, PCI_CFG_LATENCY_TIMER, &timer);
	if (timer < 0x60) {
	    pci_config_outb(bus, device, func, PCI_CFG_LATENCY_TIMER, 0x60);
	}

	/* read BARs for memory space access */
	pci_config_inl(bus, device, func, PCI_CFG_BASE_ADDRESS_0, &mbase);
	mbase &= PCI_BASE_ADDRESS_MEM_MASK;
	pci_config_inl(bus, device, func, PCI_CFG_BASE_ADDRESS_2, &mbbase[0]);
	mbbase[0] &= PCI_BASE_ADDRESS_MEM_MASK;
	pci_config_inl(bus, device, func, PCI_CFG_BASE_ADDRESS_3, &mbbase[1]);
	mbbase[1] &= PCI_BASE_ADDRESS_MEM_MASK;

#ifdef DEBUG
	printf("sibyte: PCI Device @0x%x, 0x%x, 0x%x\n",
	       mbase, mbbase[0], mbbase[1]);
#endif

	if ((mbbase[0] != mbase + SIBYTE_MB0) ||
	    (mbbase[1] != mbase + SIBYTE_MB1)) {
	    printf("sibyte: MB offsets incorrect 0x%x 0x%x 0x%x\n",
		   mbase, mbbase[0], mbbase[1]);
	}
	
	dev = (struct eth_device*) malloc(sizeof *dev);

	sprintf(dev->name, "sibyte#%d", i);
	dev->iobase = mbase;
	dev->priv   = (void*) ((bus << 16) | (device << 11) | (func << 8));
	dev->init   = sibyte_init;
	dev->halt   = sibyte_halt;
	dev->send   = sibyte_send;
	dev->recv   = sibyte_recv;
	dev->try    = 0;

	read_hw_addr(dev, bis);

	eth_register(dev);

	card_number++;
    }

    return card_number;
}

static int sibyte_init(struct eth_device* dev, bd_t* bis)
{
    int i;

    /* set rx descr index on cp and sibyte to 0 */
    rx_next = 0;
    OUTL(dev, M_RESET, (SIBYTE_MB0+4)|SIBYTE_MB_SET_BIT);

    /* initialize descriptors */
    sd.tx_descr.addr = 0;
    sd.tx_descr.status = 0;
    for (i = 0; i < RX_DESCRS; i++) {
        sd.rx_descr[i].addr = (u32)NetRxPackets[i];
        sd.rx_descr[i].status = D_OWNER|RX_BUFF_SZ;
    }
    OUTL(dev, (u32)&sd, SIBYTE_MB0|SIBYTE_MB_SET_BIT);

    return 1;
}

static int sibyte_send(struct eth_device* dev, volatile void *packet, int length)
{
    int		status = -1;
    int		i;

    if (length <= 0) {
        printf("%s: bad packet size: %d\n", dev->name, length);
	goto Done;
    }

    for(i = 0;; i++) {
	if (!sd.tx_descr.status) break;
        if (i >= TOUT_LOOP) {
	    printf("%s: tx error, buffer not ready\n", dev->name);
	    goto Done;
	}
    }

    sd.tx_descr.addr = (u32) packet;
    sd.tx_descr.status = D_OWNER|length;
    OUTL(dev, M_TXPKT, (SIBYTE_MB0+4)|SIBYTE_MB_SET_BIT);

    for(i = 0;; i++) {
	if (!sd.tx_descr.status) break;
        if (i >= TOUT_LOOP) {
	    printf("%s: tx buffer not transmitted\n", dev->name);
	    goto Done;
	}
    }

    status = length;

 Done:
    return status;
}

static int sibyte_recv(struct eth_device* dev)
{
    u32	status;
    int	length = 0;

    while (1) {
        status = sd.rx_descr[rx_next].status;
	if (status & D_OWNER) {
	    break;
	}
	length = status & D_LENGTH;
	    
	/* Valid frame status.
	 */
	if (status & D_FAIL) {
	    /* There was an error.
	     */
	    printf("RX error status = 0x%08x\n", status);
	} else if (length) {
	    /* A valid frame received.
	     * Pass the packet up to the protocol
	     * layers.
	     */
	    NetReceive(NetRxPackets[rx_next], length);
	}

	/* Enqueue packet again
	 */
	sd.rx_descr[rx_next].status = D_OWNER|RX_BUFF_SZ;
	rx_next = (rx_next + 1) % PKTBUFSRX;
    }

    return length;
}

static void sibyte_halt(struct eth_device* dev)
{
}

#include "../../../cmn/util/nvram.h"
#include "../../../cmn/util/aruba_prom.h"

static void read_hw_addr(struct eth_device *dev, bd_t *bis)
{
    int i;
    aruba_prom_t *prom = (aruba_prom_t *) (NVRAM_BASE_ADDRESS +
					   NVRAM_MANUF_OFFSET);

    if ((prom->card.bp.mac[0] != 0x00) ||
	(prom->card.bp.mac[1] != 0x0b) ||
	(prom->card.bp.mac[2] != 0x86)) {
        printf("(Invalid MAC address, please program) ");
	return;
    }
				
    for (i = 0; i < ETH_ALEN; i++) {
        dev->enetaddr[i] = prom->card.bp.mac[i];
	bis->bi_enetaddr[i] = dev->enetaddr[i];
    }
}

#endif
