/***********************************************************************
*
* debug.c
*
* Implementation of user-space PPPoE redirector for Linux.
*
* Functions for printing debugging information
*
* Copyright (C) 2000 by Roaring Penguin Software Inc.
*
* This program may be distributed according to the terms of the GNU
* General Public License, version 2 or (at your option) any later version.
*
* LIC: GPL
*
***********************************************************************/

static char const RCSID[] =
"$Id: debug.c,v 1.5 2002/05/31 14:28:18 dfs Exp $";

#include "pppoe.h"
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>

/**********************************************************************
*%FUNCTION: dumpHex
*%ARGUMENTS:
* fp -- file to dump to
* buf -- buffer to dump
* len -- length of data
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Dumps buffer to fp in an easy-to-read format
***********************************************************************/
void
dumpHex(FILE *fp, unsigned char const *buf, int len)
{
    int i;
    int base;

    if (!fp) return;

    /* do NOT dump PAP packets */
    if (len >= 2 && buf[0] == 0xC0 && buf[1] == 0x23) {
	fprintf(fp, "(PAP Authentication Frame -- Contents not dumped)\n");
	return;
    }

    for (base=0; base<len; base += 16) {
	for (i=base; i<base+16; i++) {
	    if (i < len) {
		fprintf(fp, "%02x ", (unsigned) buf[i]);
	    } else {
		fprintf(fp, "   ");
	    }
	}
	fprintf(fp, "  ");
	for (i=base; i<base+16; i++) {
	    if (i < len) {
		if (isprint(buf[i])) {
		    fprintf(fp, "%c", buf[i]);
		} else {
		    fprintf(fp, ".");
		}
	    } else {
		break;
	    }
	}
	fprintf(fp, "\n");
    }
}

/**********************************************************************
*%FUNCTION: dumpPacket
*%ARGUMENTS:
* packet -- a PPPoE packet
* dir -- either SENT or RCVD
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Dumps the PPPoE packet to arubaLog in an easy-to-read format
***********************************************************************/
void
dumpPacket(PPPoEPacket *packet, char const *dir)
{
int len = ntohs(packet->length);
char buf[256], *ptr;

    ptr = buf;
    ptr += sprintf(ptr, "%s: ", dir);
    UINT16_t type = etherType(packet);
    if (type == Eth_PPPOE_Discovery) {
	ptr += sprintf(ptr, "Discovery (%x) ", (unsigned) type);
    } else if (type == Eth_PPPOE_Session) {
	ptr += sprintf(ptr, "Session (%x) ", (unsigned) type);
    } else {
	ptr += sprintf(ptr, "Unknown (%x) ", (unsigned) type);
    }

    switch(packet->code) {
    case CODE_PADI: ptr += sprintf(ptr, "PADI. "); break;
    case CODE_PADO: ptr += sprintf(ptr, "PADO. "); break;
    case CODE_PADR: ptr += sprintf(ptr, "PADR. "); break;
    case CODE_PADS: ptr += sprintf(ptr, "PADS. "); break;
    case CODE_PADT: ptr += sprintf(ptr, "PADT. "); break;
    case CODE_PADM: ptr += sprintf(ptr, "PADM. "); break;
    case CODE_PADN: ptr += sprintf(ptr, "PADN. "); break;
    case CODE_SESS: ptr += sprintf(ptr, "SESS. "); break;
    }

    ptr += sprintf(ptr, "sess-id %d length %d. ",
	    (int) ntohs(packet->session),
	    len);

    ptr += sprintf(ptr,
	    "SourceAddr %02x:%02x:%02x:%02x:%02x:%02x "
	    "DestAddr %02x:%02x:%02x:%02x:%02x:%02x\n",
	    (unsigned) packet->ethHdr.h_source[0],
	    (unsigned) packet->ethHdr.h_source[1],
	    (unsigned) packet->ethHdr.h_source[2],
	    (unsigned) packet->ethHdr.h_source[3],
	    (unsigned) packet->ethHdr.h_source[4],
	    (unsigned) packet->ethHdr.h_source[5],
	    (unsigned) packet->ethHdr.h_dest[0],
	    (unsigned) packet->ethHdr.h_dest[1],
	    (unsigned) packet->ethHdr.h_dest[2],
	    (unsigned) packet->ethHdr.h_dest[3],
	    (unsigned) packet->ethHdr.h_dest[4],
	    (unsigned) packet->ethHdr.h_dest[5]);
    arubaLog(LOG_INFO, buf);
}
